// EnergyPlus, Copyright (c) 1996-2019, The Board of Trustees of the University of Illinois,
// The Regents of the University of California, through Lawrence Berkeley National Laboratory
// (subject to receipt of any required approvals from the U.S. Dept. of Energy), Oak Ridge
// National Laboratory, managed by UT-Battelle, Alliance for Sustainable Energy, LLC, and other
// contributors. All rights reserved.
//
// NOTICE: This Software was developed under funding from the U.S. Department of Energy and the
// U.S. Government consequently retains certain rights. As such, the U.S. Government has been
// granted for itself and others acting on its behalf a paid-up, nonexclusive, irrevocable,
// worldwide license in the Software to reproduce, distribute copies to the public, prepare
// derivative works, and perform publicly and display publicly, and to permit others to do so.
//
// Redistribution and use in source and binary forms, with or without modification, are permitted
// provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright notice, this list of
//     conditions and the following disclaimer.
//
// (2) Redistributions in binary form must reproduce the above copyright notice, this list of
//     conditions and the following disclaimer in the documentation and/or other materials
//     provided with the distribution.
//
// (3) Neither the name of the University of California, Lawrence Berkeley National Laboratory,
//     the University of Illinois, U.S. Dept. of Energy nor the names of its contributors may be
//     used to endorse or promote products derived from this software without specific prior
//     written permission.
//
// (4) Use of EnergyPlus(TM) Name. If Licensee (i) distributes the software in stand-alone form
//     without changes from the version obtained under this License, or (ii) Licensee makes a
//     reference solely to the software portion of its product, Licensee must refer to the
//     software as "EnergyPlus version X" software, where "X" is the version number Licensee
//     obtained under this License and may not use a different name for the software. Except as
//     specifically required in this Section (4), Licensee shall not use in a company name, a
//     product name, in advertising, publicity, or other promotional activities any name, trade
//     name, trademark, logo, or other designation of "EnergyPlus", "E+", "e+" or confusingly
//     similar designation, without the U.S. Department of Energy's prior written consent.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

// Google test headers
#include <gtest/gtest.h>

// EnergyPlus Headers
#include "Fixtures/EnergyPlusFixture.hh"
#include <IntegratedHeatPump.hh>
#include <VariableSpeedCoils.hh>

namespace EnergyPlus {

namespace IntegratedHeatPump {

    TEST_F(EnergyPlusFixture, ASIHP_GetCurWorkMode)
    {

        // Don't get any inputs
        GetCoilsInputFlag = false;

        IntegratedHeatPumpData IHP;

        // don't size for this test
        IHP.IHPCoilsSized = true;
        IHP.CurMode = IHPOperationMode::DWHMode;

        IntegratedHeatPumps.push_back(IHP);

        EXPECT_ANY_THROW(GetCurWorkMode(0););

        EXPECT_EQ(IHPOperationMode::DWHMode, GetCurWorkMode(1));
    }

    TEST_F(EnergyPlusFixture, ASIHP_GetCoilIndexIHP)
    {

        // Don't get any inputs
        GetCoilsInputFlag = false;
        bool ErrorsFound = false;

        IntegratedHeatPumpData IHP;
        IHP.Name = "ASIHP";
        // don't size for this test
        IHP.IHPCoilsSized = true;

        IntegratedHeatPumps.push_back(IHP);

        EXPECT_EQ(0, GetCoilIndexIHP("", "", ErrorsFound));
        EXPECT_TRUE(ErrorsFound);

        ErrorsFound = false;
        EXPECT_EQ(1, GetCoilIndexIHP("", "ASIHP", ErrorsFound));
        EXPECT_FALSE(ErrorsFound);
    }

    TEST_F(EnergyPlusFixture, ASIHP_GetCoilInletNodeIHP)
    {

        // Don't get any inputs
        GetCoilsInputFlag = false;
        bool ErrorsFound = false;

        IntegratedHeatPumpData IHP;
        IHP.Name = "ASIHP";
        IHP.AirCoolInletNodeNum = 999;
        // don't size for this test
        IHP.IHPCoilsSized = true;

        IntegratedHeatPumps.push_back(IHP);

        EXPECT_EQ(0, GetCoilInletNodeIHP("", "", ErrorsFound));
        EXPECT_TRUE(ErrorsFound);

        ErrorsFound = false;
        EXPECT_EQ(999, GetCoilInletNodeIHP("", "ASIHP", ErrorsFound));
        EXPECT_FALSE(ErrorsFound);
    }

    TEST_F(EnergyPlusFixture, ASIHP_GetLowSpeedNumIHP)
    {

        // Don't get any inputs
        GetCoilsInputFlag = false;

        IntegratedHeatPumpData IHP;
        IHP.Name = "ASIHP";
        IHP.MinSpedSCWH = 9;
        IHP.MinSpedSCDWH = 99;
        IHP.MinSpedSHDWH = 999;
        // don't size for this test
        IHP.IHPCoilsSized = true;

        IntegratedHeatPumps.push_back(IHP);

        EXPECT_ANY_THROW(GetLowSpeedNumIHP(0););

        auto &instance(IntegratedHeatPumps(1));

        instance.CurMode = IHPOperationMode::IdleMode;
        EXPECT_EQ(1, GetLowSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::SCMode;
        EXPECT_EQ(1, GetLowSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::SHMode;
        EXPECT_EQ(1, GetLowSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::DWHMode;
        EXPECT_EQ(1, GetLowSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::SCWHMatchSCMode;
        EXPECT_EQ(IHP.MinSpedSCWH, GetLowSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::SCWHMatchWHMode;
        EXPECT_EQ(IHP.MinSpedSCWH, GetLowSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::SCDWHMode;
        EXPECT_EQ(IHP.MinSpedSCDWH, GetLowSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::SHDWHElecHeatOffMode;
        EXPECT_EQ(IHP.MinSpedSHDWH, GetLowSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::SHDWHElecHeatOnMode;
        EXPECT_EQ(IHP.MinSpedSHDWH, GetLowSpeedNumIHP(1));
    }

    TEST_F(EnergyPlusFixture, ASIHP_GetMaxSpeedNumIHP)
    {

        // Don't get any inputs
        GetCoilsInputFlag = false;

        VariableSpeedCoils::VariableSpeedCoilData vsCoil;
        vsCoil.NumOfSpeeds = 999;
        VariableSpeedCoils::VarSpeedCoil.push_back(vsCoil);

        IntegratedHeatPumpData IHP;
        IHP.Name = "ASIHP";
        IHP.MinSpedSCWH = 9;
        IHP.MinSpedSCDWH = 99;
        IHP.MinSpedSHDWH = 999;
        IHP.SCCoilIndex = 1;
        IHP.SHCoilIndex = 1;
        IHP.DWHCoilIndex = 1;
        IHP.SCWHCoilIndex = 1;
        IHP.SCDWHCoolCoilIndex = 1;
        IHP.SHDWHHeatCoilIndex = 1;
        // don't size for this test
        IHP.IHPCoilsSized = true;

        IntegratedHeatPumps.push_back(IHP);

        EXPECT_ANY_THROW(GetMaxSpeedNumIHP(0););

        auto &instance(IntegratedHeatPumps(1));

        instance.CurMode = IHPOperationMode::IdleMode;
        EXPECT_EQ(vsCoil.NumOfSpeeds, GetMaxSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::SCMode;
        EXPECT_EQ(vsCoil.NumOfSpeeds, GetMaxSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::SHMode;
        EXPECT_EQ(vsCoil.NumOfSpeeds, GetMaxSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::DWHMode;
        EXPECT_EQ(vsCoil.NumOfSpeeds, GetMaxSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::SCWHMatchSCMode;
        EXPECT_EQ(vsCoil.NumOfSpeeds, GetMaxSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::SCWHMatchWHMode;
        EXPECT_EQ(vsCoil.NumOfSpeeds, GetMaxSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::SCDWHMode;
        EXPECT_EQ(vsCoil.NumOfSpeeds, GetMaxSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::SHDWHElecHeatOffMode;
        EXPECT_EQ(vsCoil.NumOfSpeeds, GetMaxSpeedNumIHP(1));

        instance.CurMode = IHPOperationMode::SHDWHElecHeatOnMode;
        EXPECT_EQ(vsCoil.NumOfSpeeds, GetMaxSpeedNumIHP(1));
    }

    TEST_F(EnergyPlusFixture, ASIHP_GetIHPInput)
    {

        std::string const idf_objects = delimited_string({
            "  CoilSystem:IntegratedHeatPump:AirSource,",
            "	ASIHP1,     !- Name of an air-source integrated heat pump",
            "	HPOutdoorWaterOutletNode, !- Name of a water node to monitor the supply water flow amount",
            "	Heat Pump ACDXCoil 1,    !- Coil object Name of space cooling mode ",
            "	Heat Pump DX Heating Coil 1,    !- Coil object Name of space heating mode ",
            "	HPWHOutdoorDXCoilVS,   !- Coil object Name of dedicated water heating mode ",
            "	SCWHCoil1,  !- Coil object Name for combined space cooling and water heating mode w full condensing",
            "	SCDWHCoolCoil1,  !- Coil object Name for cooling part of combined space cooling and water heating mode w desuperheating",
            "	SCDWHWHCoil1,  !- Coil object Name for Water heating part of combined space cooling and water heating mode w desuperheating",
            "	SHDWHHeatCoil1,  !- Coil object Name for space heating part of combined space heating and water heating mode w desuperheating",
            "	SHDWHWHCoil1,  !- Coil object Name for WH part of combined space heating and water heating mode w desuperheating",
            "	23.0,       !-[C],indoor temperature above which indoor overcooling is allowed, ",
            "	28.0,       !-[C],ambient temperature above which indoor overcooling is allowed, ",
            "	20.0,       !-[C],indoor temperature above which water heating has the higher priority",
            "	16.0,       !-[C],ambient temperature above which water heating has the higher priority",
            "	0,          !-0: match space cooling load, 1: match water heating load, in SCWH mode",
            "	1,          !- minimum speed level for SCWH mode",
            "	3.0,        !- [m3], limit of heated water volume before switching from SCDWH to SCWH mode",
            "	1,          !- minimum speed level for SCDWH mode",
            "	600,        !- [s], time limit before turning on electric element in SHDWH mode",
            "	1;          !- minimum speed level for SHDWH mode",

            "  WaterHeater:HeatPump:PumpedCondenser,",
            "    OutdoorHeatPumpWaterHeater,  !- Name",
            "    PlantHPWHSch,            !- Availability Schedule Name",
            "    HPWHTempSch,             !- Compressor Setpoint Temperature Schedule Name",
            "    2.0,                     !- Dead Band Temperature Difference {deltaC}",
            "    HPOutdoorWaterInletNode, !- Condenser Water Inlet Node Name",
            "    HPOutdoorWaterOutletNode,!- Condenser Water Outlet Node Name",
            "    0.00016,                 !- Condenser Water Flow Rate {m3/s}",
            "    0.2685,                  !- Evaporator Air Flow Rate {m3/s}",
            "    OutdoorAirOnly,          !- Inlet Air Configuration",
            "    ,                        !- Air Inlet Node Name",
            "    ,                        !- Air Outlet Node Name",
            "    HPOutdoorAirInletNode,   !- Outdoor Air Node Name",
            "    HPOutdoorAirOutletNode,  !- Exhaust Air Node Name",
            "    ,                        !- Inlet Air Temperature Schedule Name",
            "    ,                        !- Inlet Air Humidity Schedule Name",
            "    ,                        !- Inlet Air Zone Name",
            "    WaterHeater:Mixed,       !- Tank Object Type",
            "    HPWHOutdoorTank,         !- Tank Name",
            "    ,                        !- Tank Use Side Inlet Node Name",
            "    ,                        !- Tank Use Side Outlet Node Name",
            "    CoilSystem:IntegratedHeatPump:AirSource,  !- DX Coil Object Type",
            "    ASIHP1,    				 !- DX Coil Name",
            "    -5.0,                   !- Minimum Inlet Air Temperature for Compressor Operation {C}",
            "    ,                        !- Maximum Inlet Air Temperature for Compressor Operation {C}",
            "    Outdoors,                !- Compressor Location",
            "    ,                        !- Compressor Ambient Temperature Schedule Name",
            "    Fan:OnOff,               !- Fan Object Type, Fan:OnOff",
            "    HPWHOutdoorFan,          !- Fan Name",
            "    BlowThrough,             !- Fan Placement, BlowThrough",
            "    ,                        !- On Cycle Parasitic Electric Load {W}",
            "    ,                        !- Off Cycle Parasitic Electric Load {W}",
            "    ;                        !- Parasitic Heat Rejection Location",

            "  WaterHeater:Mixed,",
            "    HPWHOutdoorTank,         !- Name",
            "    0.3785,                  !- Tank Volume {m3}",
            "    Hot Water Setpoint Temp Schedule,  !- Setpoint Temperature Schedule Name",
            "    2.0,                     !- Deadband Temperature Difference {deltaC}",
            "    82.2222,                 !- Maximum Temperature Limit {C}",
            "    CYCLE,                   !- Heater Control Type",
            "    5000,                    !- Heater Maximum Capacity {W}",
            "    0,                       !- Heater Minimum Capacity {W}",
            "    ,                        !- Heater Ignition Minimum Flow Rate {m3/s}",
            "    ,                        !- Heater Ignition Delay {s}",
            "    ELECTRICITY,             !- Heater Fuel Type",
            "    0.95,                    !- Heater Thermal Efficiency",
            "    ,                        !- Part Load Factor Curve Name",
            "    10,                      !- Off Cycle Parasitic Fuel Consumption Rate {W}",
            "    ELECTRICITY,             !- Off Cycle Parasitic Fuel Type",
            "    0,                       !- Off Cycle Parasitic Heat Fraction to Tank",
            "    30,                      !- On Cycle Parasitic Fuel Consumption Rate {W}",
            "    ELECTRICITY,             !- On Cycle Parasitic Fuel Type",
            "    0,                       !- On Cycle Parasitic Heat Fraction to Tank",
            "    Outdoors,                !- Ambient Temperature Indicator",
            "    ,                        !- Ambient Temperature Schedule Name",
            "    ,                        !- Ambient Temperature Zone Name",
            "    HPWHOutdoorTank OA Node, !- Ambient Temperature Outdoor Air Node Name",
            "    2.0,                     !- Off Cycle Loss Coefficient to Ambient Temperature {W/K}",
            "    1.0,                     !- Off Cycle Loss Fraction to Zone",
            "    2.0,                     !- On Cycle Loss Coefficient to Ambient Temperature {W/K}",
            "    1.0,                     !- On Cycle Loss Fraction to Zone",
            "    0.000379,                !- Peak Use Flow Rate {m3/s}",
            "    Hot Water Demand Schedule,  !- Use Flow Rate Fraction Schedule Name",
            "    ,                        !- Cold Water Supply Temperature Schedule Name",
            "    ,                        !- Use Side Inlet Node Name",
            "    ,                        !- Use Side Outlet Node Name",
            "    ,                        !- Use Side Effectiveness",
            "    HPOutdoorWaterOutletNode,!- Source Side Inlet Node Name",
            "    HPOutdoorWaterInletNode, !- Source Side Outlet Node Name",
            "    0.95;                    !- Source Side Effectiveness",

            "  OutdoorAir:Node,",
            "    HPWHOutdoorTank OA Node; !- Name",

            "  OutdoorAir:Node,",
            "    HPOUTDOORAIRINLETNODE; !- Name",

            "  OutdoorAir:Node,",
            "    HPOutdoorAirOutletNode; !- Name",

            "  Coil:WaterHeating:AirToWaterHeatPump:VariableSpeed,",
            "    HPWHOutdoorDXCoilVS,     !- Name",
            "	10,						 !- Number of Speeds",
            "	10,						 !- Nominal speed level",
            "    4000.0,                  !- Rated Heating Capacity {W}",
            "    29.44,                   !- Rated Evaporator Inlet Air Dry-Bulb Temperature {C}",
            "    22.22,                   !- Rated Evaporator Inlet Air Wet-Bulb Temperature {C}",
            "    55.72,                   !- Rated Condenser Inlet Water Temperature {C}",
            "    0.2685,                  !- Rated Evaporator Air Flow Rate {m3/s}",
            "    0.00016,                 !- Rated Condenser Water Flow Rate {m3/s}",
            "    No,                      !- Evaporator Fan Power Included in Rated COP",
            "    No,                      !- Condenser Pump Power Included in Rated COP",
            "    No,                      !- Condenser Pump Heat Included in Rated Heating Capacity and Rated COP",
            "    0.1,                     !- Fraction of Condenser Pump Heat to Water",
            "    HPOutdoorFanAirOutletNode,   !- Evaporator Air Inlet Node Name",
            "    HPOutdoorAirOutletNode,  	  !- Evaporator Air Outlet Node Name",
            "    HPOutdoorWaterInletNode, 	 !- Condenser Water Inlet Node Name",
            "    HPOutdoorWaterOutletNode,	 !- Condenser Water Outlet Node Name",
            "    100.0,                   !- Crankcase Heater Capacity {W}",
            "    5.0,                     !- Maximum Ambient Temperature for Crankcase Heater Operation {C}",
            "    WetBulbTemperature,      !- Evaporator Air Temperature Type for Curve Objects",
            "    HPWHPLFFPLR,             !- Part Load Fraction Correlation Curve Name",
            "	400.00,					 !- Speed 1 Water Heating capacity {W} of Reference Unit",
            "	5.0, 					 !- Speed 1 Water Heating COP {W/W} of Reference Unit",
            "	0.80, 					 !- Speed 1 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.020140,                !- Speed 1 Air Flow Rate of Reference Unit",
            "	0.000018,                !- Speed 1 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 1 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	800.00,					 !- Speed 2 Water Heating capacity {W} of Reference Unit",
            "	4.8, 					 !- Speed 2 Water Heating COP {W/W} of Reference Unit",
            "	0.79, 					 !- Speed 2 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.040280,                !- Speed 2 Air Flow Rate of Reference Unit",
            "	0.000036,                !- Speed 2 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 2 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	1200.00,				 !- Speed 3 Water Heating capacity {W} of Reference Unit",
            "	4.4, 					 !- Speed 3 Water Heating COP {W/W} of Reference Unit",
            "	0.78, 					 !- Speed 3 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.060420,                !- Speed 3 Air Flow Rate of Reference Unit",
            "	0.000054,                !- Speed 3 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 3 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	1600.00,				 !- Speed 4 Water Heating capacity {W} of Reference Unit",
            "	4.0, 					 !- Speed 4 Water Heating COP {W/W} of Reference Unit",
            "	0.77, 					 !- Speed 4 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.080560,                !- Speed 4 Air Flow Rate of Reference Unit",
            "	0.000072,                !- Speed 4 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 4 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	2000.00,				 !- Speed 5 Water Heating capacity {W} of Reference Unit",
            "	3.8, 					 !- Speed 5 Water Heating COP {W/W} of Reference Unit",
            "	0.76, 					 !- Speed 5 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.100700,                !- Speed 5 Air Flow Rate of Reference Unit",
            "	0.000090,                !- Speed 5 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 5 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	2400.00,				 !- Speed 6 Water Heating capacity {W} of Reference Unit",
            "	3.4, 					 !- Speed 6 Water Heating COP {W/W} of Reference Unit",
            "	0.75, 					 !- Speed 6 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.120840,                !- Speed 6 Air Flow Rate of Reference Unit",
            "	0.000108,                !- Speed 6 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 6 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	2800.00,				 !- Speed 7 Water Heating capacity {W} of Reference Unit",
            "	3.0, 					 !- Speed 7 Water Heating COP {W/W} of Reference Unit",
            "	0.74, 					 !- Speed 7 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.140980,                !- Speed 7 Air Flow Rate of Reference Unit",
            "	0.000126,                !- Speed 7 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 7 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	3200.00,				 !- Speed 8 Water Heating capacity {W} of Reference Unit",
            "	2.6, 					 !- Speed 8 Water Heating COP {W/W} of Reference Unit",
            "	0.73, 					 !- Speed 8 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.161120,                !- Speed 8 Air Flow Rate of Reference Unit",
            "	0.000144,                !- Speed 8 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 8 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	3600.00,				 !- Speed 9 Water Heating capacity {W} of Reference Unit",
            "	2.3, 					 !- Speed 9 Water Heating COP {W/W} of Reference Unit",
            "	0.72, 					 !- Speed 9 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.181260,                !- Speed 9 Air Flow Rate of Reference Unit",
            "	0.000162,                !- Speed 9 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 9 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	4000.00,				 !- Speed 10 Water Heating capacity {W} of Reference Unit",
            "	2.0, 					 !- Speed 10 Water Heating COP {W/W} of Reference Unit",
            "	0.70, 					 !- Speed 10 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.201400,                !- Speed 10 Air Flow Rate of Reference Unit",
            "	0.000179,                !- Speed 10 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 10 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic;           !- Heating COP Function of Water Flow Fraction Curve Name",

            " Coil:WaterHeating:AirToWaterHeatPump:VariableSpeed,",
            "    SCWHCoil1,               !- Name",
            "	10,						 !- Number of Speeds",
            "	10,						 !- Nominal speed level",
            "    4000.0,                  !- Rated Heating Capacity {W} at the nominal speed level",
            "    29.44,                   !- Rated Evaporator Inlet Air Dry-Bulb Temperature {C}",
            "    22.22,                   !- Rated Evaporator Inlet Air Wet-Bulb Temperature {C}",
            "    55.72,                   !- Rated Condenser Inlet Water Temperature {C}",
            "    autocalculate,           !- Rated Evaporator Air Flow Rate {m3/s}",
            "    autocalculate,           !- Rated Condenser Water Flow Rate {m3/s}",
            "    No,                      !- Evaporator Fan Power Included in Rated COP",
            "    No,                      !- Condenser Pump Power Included in Rated COP",
            "    No,                      !- Condenser Pump Heat Included in Rated Heating Capacity and Rated COP",
            "    0.1,                     !- Fraction of Condenser Pump Heat to Water",
            "    DX Cooling Coil Air Inlet Node,   	!- Evaporator Air Inlet Node Name",
            "    Heating Coil Air Inlet Node, 		!- Evaporator Air Outlet Node Name",
            "    HPOutdoorWaterInletNode,           !- Condenser Water Inlet Node Name",
            "    HPOutdoorWaterOutletNode,          !- Condenser Water Outlet Node Name",
            "    100.0,                   !- Crankcase Heater Capacity {W}",
            "    5.0,                     !- Maximum Ambient Temperature for Crankcase Heater Operation {C}",
            "    WetBulbTemperature,      !- Evaporator Air Temperature Type for Curve Objects",
            "	HPWHPLFFPLR,             !- Part Load Fraction Correlation Curve Name",
            "	400.00,					 !- Speed 1 Water Heating capacity {W} of Reference Unit",
            "	5.0, 					 !- Speed 1 Water Heating COP {W/W} of Reference Unit",
            "	0.80, 					 !- Speed 1 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.020140,                !- Speed 1 Air Flow Rate of Reference Unit",
            "	0.000018,                !- Speed 1 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 1 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	800.00,					 !- Speed 2 Water Heating capacity {W} of Reference Unit",
            "	4.8, 					 !- Speed 2 Water Heating COP {W/W} of Reference Unit",
            "	0.79, 					 !- Speed 2 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.040280,                !- Speed 2 Air Flow Rate of Reference Unit",
            "	0.000036,                !- Speed 2 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 2 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	1200.00,				 !- Speed 3 Water Heating capacity {W} of Reference Unit",
            "	4.4, 					 !- Speed 3 Water Heating COP {W/W} of Reference Unit",
            "	0.78, 					 !- Speed 3 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.060420,                !- Speed 3 Air Flow Rate of Reference Unit",
            "	0.000054,                !- Speed 3 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 3 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	1600.00,				 !- Speed 4 Water Heating capacity {W} of Reference Unit",
            "	4.0, 					 !- Speed 4 Water Heating COP {W/W} of Reference Unit",
            "	0.77, 					 !- Speed 4 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.080560,                !- Speed 4 Air Flow Rate of Reference Unit",
            "	0.000072,                !- Speed 4 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 4 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	2000.00,				 !- Speed 5 Water Heating capacity {W} of Reference Unit",
            "	3.8, 					 !- Speed 5 Water Heating COP {W/W} of Reference Unit",
            "	0.76, 					 !- Speed 5 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.100700,                !- Speed 5 Air Flow Rate of Reference Unit",
            "	0.000090,                !- Speed 5 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 5 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	2400.00,				 !- Speed 6 Water Heating capacity {W} of Reference Unit",
            "	3.4, 					 !- Speed 6 Water Heating COP {W/W} of Reference Unit",
            "	0.75, 					 !- Speed 6 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.120840,                !- Speed 6 Air Flow Rate of Reference Unit",
            "	0.000108,                !- Speed 6 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 6 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	2800.00,				 !- Speed 7 Water Heating capacity {W} of Reference Unit",
            "	3.0, 					 !- Speed 7 Water Heating COP {W/W} of Reference Unit",
            "	0.74, 					 !- Speed 7 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.140980,                !- Speed 7 Air Flow Rate of Reference Unit",
            "	0.000126,                !- Speed 7 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 7 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	3200.00,				 !- Speed 8 Water Heating capacity {W} of Reference Unit",
            "	2.6, 					 !- Speed 8 Water Heating COP {W/W} of Reference Unit",
            "	0.73, 					 !- Speed 8 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.161120,                !- Speed 8 Air Flow Rate of Reference Unit",
            "	0.000144,                !- Speed 8 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 8 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	3600.00,				 !- Speed 9 Water Heating capacity {W} of Reference Unit",
            "	2.3, 					 !- Speed 9 Water Heating COP {W/W} of Reference Unit",
            "	0.72, 					 !- Speed 9 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.181260,                !- Speed 9 Air Flow Rate of Reference Unit",
            "	0.000162,                !- Speed 9 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 9 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	4000.00,				 !- Speed 10 Water Heating capacity {W} of Reference Unit",
            "	2.0, 					 !- Speed 10 Water Heating COP {W/W} of Reference Unit",
            "	0.70, 					 !- Speed 10 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.201400,                !- Speed 10 Air Flow Rate of Reference Unit",
            "	0.000179,                !- Speed 10 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 10 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic;           !- Heating COP Function of Water Flow Fraction Curve Name",

            "  Coil:WaterHeating:AirToWaterHeatPump:VariableSpeed,",
            "    SCDWHWHCoil1,            !- Name",
            "	10,						 !- Number of Speeds",
            "	10,						 !- Nominal speed level",
            "    4000.0,                  !- Rated Heating Capacity {W} at the nominal speed level",
            "    29.44,                   !- Rated Evaporator Inlet Air Dry-Bulb Temperature {C}",
            "    22.22,                   !- Rated Evaporator Inlet Air Wet-Bulb Temperature {C}",
            "    55.72,                   !- Rated Condenser Inlet Water Temperature {C}",
            "    autocalculate,           !- Rated Evaporator Air Flow Rate {m3/s}",
            "    autocalculate,           !- Rated Condenser Water Flow Rate {m3/s}",
            "    No,                      !- Evaporator Fan Power Included in Rated COP",
            "    No,                      !- Condenser Pump Power Included in Rated COP",
            "    No,                      !- Condenser Pump Heat Included in Rated Heating Capacity and Rated COP",
            "    0.1,                     !- Fraction of Condenser Pump Heat to Water",
            "    HPOutdoorFanAirOutletNode,   !- Outdoor Air Inlet Node Name",
            "    HPOutdoorAirOutletNode,  	 !- Outdoor Air Outlet Node Name",
            "    HPOutdoorWaterInletNode,    !- Condenser Water Inlet Node Name",
            "    HPOutdoorWaterOutletNode,   !- Condenser Water Outlet Node Name",
            "    100.0,                   !- Crankcase Heater Capacity {W}",
            "    5.0,                     !- Maximum Ambient Temperature for Crankcase Heater Operation {C}",
            "    WetBulbTemperature,      !- Evaporator Air Temperature Type for Curve Objects",
            "	HPWHPLFFPLR,             !- Part Load Fraction Correlation Curve Name",
            "	400.00,					 !- Speed 1 Water Heating capacity {W} of Reference Unit",
            "	5.0, 					 !- Speed 1 Water Heating COP {W/W} of Reference Unit",
            "	0.80, 					 !- Speed 1 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.020140,                !- Speed 1 Air Flow Rate of Reference Unit",
            "	0.000018,                !- Speed 1 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 1 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	800.00,					 !- Speed 2 Water Heating capacity {W} of Reference Unit",
            "	4.8, 					 !- Speed 2 Water Heating COP {W/W} of Reference Unit",
            "	0.79, 					 !- Speed 2 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.040280,                !- Speed 2 Air Flow Rate of Reference Unit",
            "	0.000036,                !- Speed 2 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 2 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	1200.00,				 !- Speed 3 Water Heating capacity {W} of Reference Unit",
            "	4.4, 					 !- Speed 3 Water Heating COP {W/W} of Reference Unit",
            "	0.78, 					 !- Speed 3 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.060420,                !- Speed 3 Air Flow Rate of Reference Unit",
            "	0.000054,                !- Speed 3 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 3 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	1600.00,				 !- Speed 4 Water Heating capacity {W} of Reference Unit",
            "	4.0, 					 !- Speed 4 Water Heating COP {W/W} of Reference Unit",
            "	0.77, 					 !- Speed 4 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.080560,                !- Speed 4 Air Flow Rate of Reference Unit",
            "	0.000072,                !- Speed 4 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 4 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	2000.00,				 !- Speed 5 Water Heating capacity {W} of Reference Unit",
            "	3.8, 					 !- Speed 5 Water Heating COP {W/W} of Reference Unit",
            "	0.76, 					 !- Speed 5 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.100700,                !- Speed 5 Air Flow Rate of Reference Unit",
            "	0.000090,                !- Speed 5 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 5 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	2400.00,				 !- Speed 6 Water Heating capacity {W} of Reference Unit",
            "	3.4, 					 !- Speed 6 Water Heating COP {W/W} of Reference Unit",
            "	0.75, 					 !- Speed 6 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.120840,                !- Speed 6 Air Flow Rate of Reference Unit",
            "	0.000108,                !- Speed 6 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 6 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	2800.00,				 !- Speed 7 Water Heating capacity {W} of Reference Unit",
            "	3.0, 					 !- Speed 7 Water Heating COP {W/W} of Reference Unit",
            "	0.74, 					 !- Speed 7 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.140980,                !- Speed 7 Air Flow Rate of Reference Unit",
            "	0.000126,                !- Speed 7 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 7 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	3200.00,				 !- Speed 8 Water Heating capacity {W} of Reference Unit",
            "	2.6, 					 !- Speed 8 Water Heating COP {W/W} of Reference Unit",
            "	0.73, 					 !- Speed 8 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.161120,                !- Speed 8 Air Flow Rate of Reference Unit",
            "	0.000144,                !- Speed 8 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 8 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	3600.00,				 !- Speed 9 Water Heating capacity {W} of Reference Unit",
            "	2.3, 					 !- Speed 9 Water Heating COP {W/W} of Reference Unit",
            "	0.72, 					 !- Speed 9 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.181260,                !- Speed 9 Air Flow Rate of Reference Unit",
            "	0.000162,                !- Speed 9 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 9 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	4000.00,				 !- Speed 10 Water Heating capacity {W} of Reference Unit",
            "	2.0, 					 !- Speed 10 Water Heating COP {W/W} of Reference Unit",
            "	0.70, 					 !- Speed 10 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.201400,                !- Speed 10 Air Flow Rate of Reference Unit",
            "	0.000179,                !- Speed 10 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 10 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic;           !- Heating COP Function of Water Flow Fraction Curve Name",

            " Coil:WaterHeating:AirToWaterHeatPump:VariableSpeed,",
            "    SHDWHWHCoil1,     !- Name",
            "	10,						 !- Number of Speeds",
            "	10,						 !- Nominal speed level",
            "    4000.0,                  !- Rated Heating Capacity {W}",
            "    29.44,                   !- Rated Evaporator Inlet Air Dry-Bulb Temperature {C}",
            "    22.22,                   !- Rated Evaporator Inlet Air Wet-Bulb Temperature {C}",
            "    55.72,                   !- Rated Condenser Inlet Water Temperature {C}",
            "    0.2685,                  !- Rated Evaporator Air Flow Rate {m3/s}",
            "    0.00016,                 !- Rated Condenser Water Flow Rate {m3/s}",
            "    No,                      !- Evaporator Fan Power Included in Rated COP",
            "    No,                      !- Condenser Pump Power Included in Rated COP",
            "    No,                      !- Condenser Pump Heat Included in Rated Heating Capacity and Rated COP",
            "    0.1,                     !- Fraction of Condenser Pump Heat to Water",
            "    HPOutdoorFanAirOutletNode,  	!- Outdoor Air Inlet Node Name",
            "    HPOutdoorAirOutletNode,  		!- Outdoor Air Outlet Node Name",
            "    HPOutdoorWaterInletNode,		!- Condenser Water Inlet Node Name",
            "    HPOutdoorWaterOutletNode,		!- Condenser Water Outlet Node Name",
            "    100.0,                   !- Crankcase Heater Capacity {W}",
            "    5.0,                     !- Maximum Ambient Temperature for Crankcase Heater Operation {C}",
            "    WetBulbTemperature,      !- Evaporator Air Temperature Type for Curve Objects",
            "    HPWHPLFFPLR,             !- Part Load Fraction Correlation Curve Name",
            "	400.00,					 !- Speed 1 Water Heating capacity {W} of Reference Unit",
            "	5.0, 					 !- Speed 1 Water Heating COP {W/W} of Reference Unit",
            "	0.80, 					 !- Speed 1 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.020140,                !- Speed 1 Air Flow Rate of Reference Unit",
            "	0.000018,                !- Speed 1 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 1 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	800.00,					 !- Speed 2 Water Heating capacity {W} of Reference Unit",
            "	4.8, 					 !- Speed 2 Water Heating COP {W/W} of Reference Unit",
            "	0.79, 					 !- Speed 2 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.040280,                !- Speed 2 Air Flow Rate of Reference Unit",
            "	0.000036,                !- Speed 2 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 2 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	1200.00,				 !- Speed 3 Water Heating capacity {W} of Reference Unit",
            "	4.4, 					 !- Speed 3 Water Heating COP {W/W} of Reference Unit",
            "	0.78, 					 !- Speed 3 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.060420,                !- Speed 3 Air Flow Rate of Reference Unit",
            "	0.000054,                !- Speed 3 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 3 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	1600.00,				 !- Speed 4 Water Heating capacity {W} of Reference Unit",
            "	4.0, 					 !- Speed 4 Water Heating COP {W/W} of Reference Unit",
            "	0.77, 					 !- Speed 4 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.080560,                !- Speed 4 Air Flow Rate of Reference Unit",
            "	0.000072,                !- Speed 4 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 4 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	2000.00,				 !- Speed 5 Water Heating capacity {W} of Reference Unit",
            "	3.8, 					 !- Speed 5 Water Heating COP {W/W} of Reference Unit",
            "	0.76, 					 !- Speed 5 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.100700,                !- Speed 5 Air Flow Rate of Reference Unit",
            "	0.000090,                !- Speed 5 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 5 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	2400.00,				 !- Speed 6 Water Heating capacity {W} of Reference Unit",
            "	3.4, 					 !- Speed 6 Water Heating COP {W/W} of Reference Unit",
            "	0.75, 					 !- Speed 6 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.120840,                !- Speed 6 Air Flow Rate of Reference Unit",
            "	0.000108,                !- Speed 6 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 6 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	2800.00,				 !- Speed 7 Water Heating capacity {W} of Reference Unit",
            "	3.0, 					 !- Speed 7 Water Heating COP {W/W} of Reference Unit",
            "	0.74, 					 !- Speed 7 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.140980,                !- Speed 7 Air Flow Rate of Reference Unit",
            "	0.000126,                !- Speed 7 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 7 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	3200.00,				 !- Speed 8 Water Heating capacity {W} of Reference Unit",
            "	2.6, 					 !- Speed 8 Water Heating COP {W/W} of Reference Unit",
            "	0.73, 					 !- Speed 8 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.161120,                !- Speed 8 Air Flow Rate of Reference Unit",
            "	0.000144,                !- Speed 8 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 8 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	3600.00,				 !- Speed 9 Water Heating capacity {W} of Reference Unit",
            "	2.3, 					 !- Speed 9 Water Heating COP {W/W} of Reference Unit",
            "	0.72, 					 !- Speed 9 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.181260,                !- Speed 9 Air Flow Rate of Reference Unit",
            "	0.000162,                !- Speed 9 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 9 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Water Flow Fraction Curve Name",
            "	4000.00,				 !- Speed 10 Water Heating capacity {W} of Reference Unit",
            "	2.0, 					 !- Speed 10 Water Heating COP {W/W} of Reference Unit",
            "	0.70, 					 !- Speed 10 Sensible Heat Transfer Ratio of Reference Unit",
            "	0.201400,                !- Speed 10 Air Flow Rate of Reference Unit",
            "	0.000179,                !- Speed 10 water Air Flow Rate of Reference Unit",
            "	10.0,                    !- Speed 10 Water Pump Power of Reference Unit",
            "    HPWHHeatingCapFTemp,     !- Heating Capacity Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Air Flow Fraction Curve Name",
            "    ConstantCubic,           !- Heating Capacity Function of Water Flow Fraction Curve Name",
            "    HPWHHeatingCOPFTemp,     !- Heating COP Function of Temperature Curve Name",
            "    ConstantCubic,           !- Heating COP Function of Air Flow Fraction Curve Name",
            "    ConstantCubic;           !- Heating COP Function of Water Flow Fraction Curve Name",

            "  Fan:OnOff,",
            "    HPWHOutdoorFan,          !- Name",
            "    PlantHPWHSch,            !- Availability Schedule Name",
            "    0.7,                     !- Fan Total Efficiency",
            "    100.0,                   !- Pressure Rise {Pa}",
            "    2.2685,                  !- Maximum Flow Rate {m3/s}",
            "    0.9,                     !- Motor Efficiency",
            "    1.0,                     !- Motor In Airstream Fraction",
            "    HPOutdoorAirInletNode,   !- Air Inlet Node Name",
            "    HPOutdoorFanAirOutletNode;  !- Air Outlet Node Name",

            "  Schedule:Compact,",
            "    PlantHPWHSch,            !- Name",
            "    Any Number,              !- Schedule Type Limits Name",
            "    Through: 12/31,          !- Field 1",
            "    For: AllDays,            !- Field 2",
            "    Until: 24:00,1.0;        !- Field 3",

            "  Schedule:Compact,",
            "    HPWHTempSch,             !- Name",
            "    Any Number,              !- Schedule Type Limits Name",
            "    Through: 12/31,          !- Field 1",
            "    For: AllDays,            !- Field 2",
            "    Until: 24:00,60.0;       !- Field 3",

            "  Schedule:Compact,",
            "    Hot Water Demand Schedule,  !- Name",
            "    Fraction,                !- Schedule Type Limits Name",
            "    THROUGH: 12/31,          !- Field 1",
            "    FOR: AllDays,            !- Field 2",
            "    UNTIL: 5:00,0,           !- Field 3",
            "    UNTIL: 7:00,0.003,       !- Field 5",
            "    UNTIL: 13:00,0,          !- Field 7",
            "    UNTIL: 13:10,0.001,      !- Field 9",
            "    UNTIL: 18:00,0,          !- Field 11",
            "    UNTIL: 18:30,0.003,      !- Field 13",
            "    UNTIL: 24:00,0;          !- Field 15",

            "  Schedule:Compact,",
            "    Hot Water Setpoint Temp Schedule,  !- Name",
            "    Any Number,              !- Schedule Type Limits Name",
            "    Through: 12/31,          !- Field 1",
            "    For: AllDays,            !- Field 2",
            "    Until: 24:00,50.0;       !- Field 3",

            "  Curve:Biquadratic,",
            "    HPWHHeatingCapFTemp,     !- Name",
            "    0.369827,                !- Coefficient1 Constant",
            "    0.043341,                !- Coefficient2 x",
            "    -0.00023,                !- Coefficient3 x**2",
            "    0.000466,                !- Coefficient4 y",
            "    0.000026,                !- Coefficient5 y**2",
            "    -0.00027,                !- Coefficient6 x*y",
            "    0.0,                     !- Minimum Value of x",
            "    40.0,                    !- Maximum Value of x",
            "    20.0,                    !- Minimum Value of y",
            "    90.0,                    !- Maximum Value of y",
            "    ,                        !- Minimum Curve Output",
            "    ,                        !- Maximum Curve Output",
            "    Temperature,             !- Input Unit Type for X",
            "    Temperature,             !- Input Unit Type for Y",
            "    Dimensionless;           !- Output Unit Type",

            "  Curve:Biquadratic,",
            "    HPWHHeatingCOPFTemp,     !- Name",
            "    1.19713,                 !- Coefficient1 Constant",
            "    0.077849,                !- Coefficient2 x",
            "    -0.0000016,              !- Coefficient3 x**2",
            "    -0.02675,                !- Coefficient4 y",
            "    0.000296,                !- Coefficient5 y**2",
            "    -0.00112,                !- Coefficient6 x*y",
            "    0.0,                     !- Minimum Value of x",
            "    40.0,                    !- Maximum Value of x",
            "    20.0,                    !- Minimum Value of y",
            "    90.0,                    !- Maximum Value of y",
            "    ,                        !- Minimum Curve Output",
            "    ,                        !- Maximum Curve Output",
            "    Temperature,             !- Input Unit Type for X",
            "    Temperature,             !- Input Unit Type for Y",
            "    Dimensionless;           !- Output Unit Type",

            "  Curve:Quadratic,",
            "    HPWHPLFFPLR,             !- Name",
            "    0.75,                    !- Coefficient1 Constant",
            "    0.25,                    !- Coefficient2 x",
            "    0.0,                     !- Coefficient3 x**2",
            "    0.0,                     !- Minimum Value of x",
            "    1.0;                     !- Maximum Value of x",

            "  Curve:Cubic,",
            "   ConstantCubic,   !- Name",
            "   1,               !- Coefficient1 Constant",
            "   0,               !- Coefficient2 x",
            "   0,               !- Coefficient3 x",
            "   0,               !- Coefficient4 x",
            "   -100,            !- Minimum Value of x",
            "   100;             !- Maximum Value of x",

            "  AirLoopHVAC:UnitaryHeatPump:AirToAir,",
            "    DXAC Heat Pump 1,        !- Name",
            "    FanAndCoilAvailSched,    !- Availability Schedule Name",
            "    Mixed Air Node,          !- Air Inlet Node Name",
            "    Air Loop Outlet Node,    !- Air Outlet Node Name",
            "    1.7,                     !- Supply Air Flow Rate During Cooling Operation {m3/s}",
            "    1.7,                     !- Supply Air Flow Rate During Heating Operation {m3/s}",
            "    0.0,                     !- Supply Air Flow Rate When No Cooling or Heating is Needed {m3/s}",
            "    East Zone,               !- Controlling Zone or Thermostat Location",
            "    Fan:OnOff,               !- Supply Air Fan Object Type",
            "    Supply Fan 1,            !- Supply Air Fan Name",
            "    CoilSystem:IntegratedHeatPump:AirSource,  !- Heating Coil Object Type",
            "    ASIHP1,    				 !- Heating DX Coil Name",
            "    CoilSystem:IntegratedHeatPump:AirSource,  !- Coolinng Coil Object Type",
            "    ASIHP1,    				 !- Cooling DX Coil Name",
            "    Coil:Heating:Fuel,        !- Supplemental Heating Coil Object Type",
            "    Heat Pump DX Supp Heating Coil 1,  !- Supplemental Heating Coil Name",
            "    50,                      !- Maximum Supply Air Temperature from Supplemental Heater {C}",
            "    21,                      !- Maximum Outdoor Dry-Bulb Temperature for Supplemental Heater Operation {C}",
            "    BlowThrough,             !- Fan Placement",
            "    CyclingFanSchedule;      !- Supply Air Fan Operating Mode Schedule Name",

            "  Coil:Heating:DX:VariableSpeed,",
            "    Heat Pump DX Heating Coil 1,  !- Name",
            "    Heating Coil Air Inlet Node,  !- Indoor Air Inlet Node Name",
            "    SuppHeating Coil Air Inlet Node,  !- Indoor Air Outlet Node Name",
            "    10.0,                    !- Number of Speeds {dimensionless}",
            "    10.0,                    !- Nominal Speed Level {dimensionless}",
            "    35000,                   !- Rated Heating Capacity At Selected Nominal Speed Level {w}",
            "    1.7,                     !- Rated Air Flow Rate At Selected Nominal Speed Level {m3/s}",
            "    HPACCOOLPLFFPLR,         !- Energy Part Load Fraction Curve Name",
            "    ,                        !- Defrost Energy Input Ratio Function of Temperature Curve Name",
            "    -5.0,                    !- Minimum Outdoor Dry-Bulb Temperature for Compressor Operation {C}",
            "    ,                        !- Outdoor Dry-Bulb Temperature to Turn On Compressor {C}",
            "    5.0,                     !- Maximum Outdoor Dry-Bulb Temperature for Defrost Operation {C}",
            "    200.0,                   !- Crankcase Heater Capacity {W}",
            "    10.0,                    !- Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater Operation {C}",
            "    Resistive,               !- Defrost Strategy",
            "    TIMED,                   !- Defrost Control",
            "    0.166667,                !- Defrost Time Period Fraction",
            "    20000,                   !- Resistive Defrost Heater Capacity {W}",
            "    1838.7,                  !- Speed 1 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 1 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.1661088,               !- Speed 1 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 1 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 1 Total  Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 1 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 1 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    2295.5,                  !- Speed 2 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 2 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.179322,                !- Speed 2 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 2 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 2 Total  Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 2 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 2 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    2751.3,                  !- Speed 3 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 3 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.1925352,               !- Speed 3 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 3 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 3 Total  Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 3 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 3 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    3659.6,                  !- Speed 4 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 4 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.2189616,               !- Speed 4 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 4 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 4 Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 4 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 4 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    4563.7,                  !- Speed 5 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 5 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.245388,                !- Speed 5 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 5 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 5 Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 5 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 5 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    5463.3,                  !- Speed 6 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 6 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.2718144,               !- Speed 6 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 6 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 6 Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 6 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 6 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    6358.4,                  !- Speed 7 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 7 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.2982408,               !- Speed 7 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 7 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 7 Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 7 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 7 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    7248.5,                  !- Speed 8 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 8 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.3246672,               !- Speed 8 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 8 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 8 Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 8 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 8 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    8133.6,                  !- Speed 9 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 9 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.3510936,               !- Speed 9 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 9 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 9 Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 9 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 9 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    9013.2,                  !- Speed 10 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 10 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.37752,                 !- Speed 10 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 10 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 10 Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 10 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF;          !- Speed 10 Energy Input Ratio Function of Air Flow Fraction Curve Name",

            "  Coil:Heating:DX:VariableSpeed,",
            "    SHDWHHeatCoil1,  		!- Name",
            "    Heating Coil Air Inlet Node,  	!- Indoor Air Inlet Node Name",
            "    SuppHeating Coil Air Inlet Node,  	!- Indoor Air Outlet Node Name",
            "    10.0,                    !- Number of Speeds {dimensionless}",
            "    10.0,                    !- Nominal Speed Level {dimensionless}",
            "    35000,                   !- Rated Heating Capacity At Selected Nominal Speed Level {w}",
            "    1.7,                     !- Rated Air Flow Rate At Selected Nominal Speed Level {m3/s}",
            "    HPACCOOLPLFFPLR,         !- Energy Part Load Fraction Curve Name",
            "    ,                        !- Defrost Energy Input Ratio Function of Temperature Curve Name",
            "    -5.0,                    !- Minimum Outdoor Dry-Bulb Temperature for Compressor Operation {C}",
            "    ,                        !- Outdoor Dry-Bulb Temperature to Turn On Compressor {C}",
            "    5.0,                     !- Maximum Outdoor Dry-Bulb Temperature for Defrost Operation {C}",
            "    200.0,                   !- Crankcase Heater Capacity {W}",
            "    10.0,                    !- Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater Operation {C}",
            "    Resistive,               !- Defrost Strategy",
            "    TIMED,                   !- Defrost Control",
            "    0.166667,                !- Defrost Time Period Fraction",
            "    20000,                   !- Resistive Defrost Heater Capacity {W}",
            "    1838.7,                  !- Speed 1 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 1 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.1661088,               !- Speed 1 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 1 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 1 Total  Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 1 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 1 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    2295.5,                  !- Speed 2 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 2 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.179322,                !- Speed 2 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 2 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 2 Total  Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 2 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 2 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    2751.3,                  !- Speed 3 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 3 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.1925352,               !- Speed 3 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 3 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 3 Total  Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 3 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 3 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    3659.6,                  !- Speed 4 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 4 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.2189616,               !- Speed 4 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 4 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 4 Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 4 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 4 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    4563.7,                  !- Speed 5 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 5 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.245388,                !- Speed 5 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 5 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 5 Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 5 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 5 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    5463.3,                  !- Speed 6 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 6 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.2718144,               !- Speed 6 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 6 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 6 Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 6 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 6 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    6358.4,                  !- Speed 7 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 7 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.2982408,               !- Speed 7 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 7 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 7 Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 7 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 7 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    7248.5,                  !- Speed 8 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 8 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.3246672,               !- Speed 8 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 8 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 8 Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 8 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 8 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    8133.6,                  !- Speed 9 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 9 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.3510936,               !- Speed 9 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 9 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 9 Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 9 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF,          !- Speed 9 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    9013.2,                  !- Speed 10 Reference Unit Gross Rated Heating Capacity {w}",
            "    5.0,                     !- Speed 10 Reference Unit Gross Rated Heating COP {dimensionless}",
            "    0.37752,                 !- Speed 10 Reference Unit Rated Air Flow Rate {m3/s}",
            "    HPACHeatCapFT,           !- Speed 10 Heating Capacity Function of Temperature Curve Name",
            "    HPACHeatCapFFF,          !- Speed 10 Heating Capacity Function of Air Flow Fraction Curve Name",
            "    HPACHeatEIRFT,           !- Speed 10 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACHeatEIRFFF;          !- Speed 10 Energy Input Ratio Function of Air Flow Fraction Curve Name",

            "  Curve:Biquadratic,",
            "    HPACHeatCapFT,           !- Name",
            "    0.8529681407,            !- Coefficient1 Constant",
            "    -0.0004847169,           !- Coefficient2 x",
            "    -0.0000010693,           !- Coefficient3 x**2",
            "    0.0185542164,            !- Coefficient4 y",
            "    0.0000872425,            !- Coefficient5 y**2",
            "    -0.0000166868,           !- Coefficient6 x*y",
            "    17.78,                   !- Minimum Value of x",
            "    23.33,                   !- Maximum Value of x",
            "    -28.89,                  !- Minimum Value of y",
            "    17.22,                   !- Maximum Value of y",
            "    0.3799,                  !- Minimum Curve Output",
            "    1.1896,                  !- Maximum Curve Output",
            "    Temperature,             !- Input Unit Type for X",
            "    Temperature,             !- Input Unit Type for Y",
            "    Dimensionless;           !- Output Unit Type",

            "  Curve:Biquadratic,",
            "    HPACHeatEIRFT,           !- Name",
            "    0.7077081462,            !- Coefficient1 Constant",
            "    0.0148163478,            !- Coefficient2 x",
            "    0.0002622589,            !- Coefficient3 x**2",
            "    -0.0113239622,           !- Coefficient4 y",
            "    0.0002939277,            !- Coefficient5 y**2",
            "    -0.0003605284,           !- Coefficient6 x*y",
            "    17.78,                   !- Minimum Value of x",
            "    23.33,                   !- Maximum Value of x",
            "    -28.89,                  !- Minimum Value of y",
            "    17.22,                   !- Maximum Value of y",
            "    0.8266,                  !- Minimum Curve Output",
            "    2.0277,                  !- Maximum Curve Output",
            "    Temperature,             !- Input Unit Type for X",
            "    Temperature,             !- Input Unit Type for Y",
            "    Dimensionless;           !- Output Unit Type",

            "  Coil:Cooling:DX:VariableSpeed,",
            "    Heat Pump ACDXCoil 1,    !- Name",
            "    DX Cooling Coil Air Inlet Node,   	!- Indoor Air Inlet Node Name",
            "    Heating Coil Air Inlet Node, 		!- Indoor Air Outlet Node Name",
            "    10.0,                    !- Number of Speeds {dimensionless}",
            "    10.0,                    !- Nominal Speed Level {dimensionless}",
            "    32000,                   !- Gross Rated Total Cooling Capacity At Selected Nominal Speed Level {w}",
            "    1.7,                     !- Rated Air Flow Rate At Selected Nominal Speed Level {m3/s}",
            "    0.0,                     !- Nominal Time for Condensate to Begin Leaving the Coil {s}",
            "    0.0,                     !- Initial Moisture Evaporation Rate Divided by Steady-State AC Latent Capacity {dimensionless}",
            "    HPACCOOLPLFFPLR,         !- Energy Part Load Fraction Curve Name",
            "    ,                        !- Condenser Air Inlet Node Name",
            "    AirCooled,               !- Condenser Type",
            "    ,                        !- Evaporative Condenser Pump Rated Power Consumption {W}",
            "    200.0,                   !- Crankcase Heater Capacity {W}",
            "    10.0,                    !- Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater Operation {C}",
            "    ,                        !- Minimum Outdoor Dry-Bulb Temperature for Compressor Operation {C}",
            "    ,                        !- Supply Water Storage Tank Name",
            "    ,                        !- Condensate Collection Water Storage Tank Name",
            "    ,                        !- Basin Heater Capacity {W/K}",
            "    ,                        !- Basin Heater Setpoint Temperature {C}",
            "    ,                        !- Basin Heater Operating Schedule Name",
            "    1524.1,                  !- Speed 1 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 1 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 1 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.1359072,               !- Speed 1 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.26,                    !- Speed 1 Reference Unit Rated Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 1 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 1 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 1 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 1 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 1 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    1877.9,                  !- Speed 2 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 2 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 2 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.151008,                !- Speed 2 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.30,                    !- Speed 2 Reference Unit Rated Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 2 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 2 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 2 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 2 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 2 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    2226.6,                  !- Speed 3 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 3 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 3 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.1661088,               !- Speed 3 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.33,                    !- Speed 3 Reference Unit Rated Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 3 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 3 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 3 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 3 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 3 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    2911.3,                  !- Speed 4 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 4 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 4 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.1963104,               !- Speed 4 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.38,                    !- Speed 4 Reference Unit Rated Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 4 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 4 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 4 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 4 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 4 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    3581.7,                  !- Speed 5 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 5 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 5 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.226512,                !- Speed 5 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.44,                    !- Speed 5 Reference Unit Rated Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 5 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 5 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 5 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 5 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 5 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    4239.5,                  !- Speed 6 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 6 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 6 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.2567136,               !- Speed 6 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.50,                    !- Speed 6 Reference Unit Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 6 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 6 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 6 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 6 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 6 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    4885.7,                  !- Speed 7 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 7 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 7 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.2869152,               !- Speed 7 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.57,                    !- Speed 7 Reference Unit Condenser Flow Rate {m3/s}",
            "    ,                        !- Speed 7 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 7 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 7 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 7 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 7 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    5520.7,                  !- Speed 8 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 8 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 8 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.3171168,               !- Speed 8 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.63,                    !- Speed 8 Reference Unit Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 8 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 8 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 8 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 8 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 8 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    6144.8,                  !- Speed 9 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 9 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 9 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.3473184,               !- Speed 9 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.69,                    !- Speed 9 Reference Unit Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 9 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 9 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 9 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 9 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 9 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    6758.0,                  !- Speed 10 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 10 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 10 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.37752,                 !- Speed 10 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.74,                    !- Speed 10 Reference Unit Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 10 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 10 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 10 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 10 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF;          !- Speed 10 Energy Input Ratio Function of Air Flow Fraction Curve Name",

            "  Coil:Cooling:DX:VariableSpeed,",
            "    SCDWHCoolCoil1,    		!- Name",
            "    DX Cooling Coil Air Inlet Node,    !- Indoor Air Inlet Node Name",
            "    Heating Coil Air Inlet Node, 	    !- Indoor Air Outlet Node Name",
            "    10.0,                    !- Number of Speeds {dimensionless}",
            "    10.0,                    !- Nominal Speed Level {dimensionless}",
            "    32000,                   !- Gross Rated Total Cooling Capacity At Selected Nominal Speed Level {w}",
            "    1.7,                     !- Rated Air Flow Rate At Selected Nominal Speed Level {m3/s}",
            "    0.0,                     !- Nominal Time for Condensate to Begin Leaving the Coil {s}",
            "    0.0,                     !- Initial Moisture Evaporation Rate Divided by Steady-State AC Latent Capacity {dimensionless}",
            "    HPACCOOLPLFFPLR,         !- Energy Part Load Fraction Curve Name",
            "    ,                        !- Condenser Air Inlet Node Name",
            "    AirCooled,               !- Condenser Type",
            "    ,                        !- Evaporative Condenser Pump Rated Power Consumption {W}",
            "    200.0,                   !- Crankcase Heater Capacity {W}",
            "    10.0,                    !- Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater Operation {C}",
            "    ,                        !- Minimum Outdoor Dry-Bulb Temperature for Compressor Operation {C}",
            "    ,                        !- Supply Water Storage Tank Name",
            "    ,                        !- Condensate Collection Water Storage Tank Name",
            "    ,                        !- Basin Heater Capacity {W/K}",
            "    ,                        !- Basin Heater Setpoint Temperature {C}",
            "    ,                        !- Basin Heater Operating Schedule Name",
            "    1524.1,                  !- Speed 1 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 1 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 1 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.1359072,               !- Speed 1 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.26,                    !- Speed 1 Reference Unit Rated Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 1 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 1 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 1 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 1 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 1 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    1877.9,                  !- Speed 2 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 2 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 2 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.151008,                !- Speed 2 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.30,                    !- Speed 2 Reference Unit Rated Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 2 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 2 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 2 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 2 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 2 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    2226.6,                  !- Speed 3 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 3 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 3 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.1661088,               !- Speed 3 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.33,                    !- Speed 3 Reference Unit Rated Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 3 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 3 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 3 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 3 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 3 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    2911.3,                  !- Speed 4 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 4 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 4 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.1963104,               !- Speed 4 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.38,                    !- Speed 4 Reference Unit Rated Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 4 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 4 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 4 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 4 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 4 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    3581.7,                  !- Speed 5 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 5 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 5 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.226512,                !- Speed 5 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.44,                    !- Speed 5 Reference Unit Rated Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 5 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 5 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 5 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 5 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 5 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    4239.5,                  !- Speed 6 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 6 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 6 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.2567136,               !- Speed 6 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.50,                    !- Speed 6 Reference Unit Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 6 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 6 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 6 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 6 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 6 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    4885.7,                  !- Speed 7 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 7 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 7 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.2869152,               !- Speed 7 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.57,                    !- Speed 7 Reference Unit Condenser Flow Rate {m3/s}",
            "    ,                        !- Speed 7 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 7 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 7 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 7 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 7 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    5520.7,                  !- Speed 8 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 8 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 8 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.3171168,               !- Speed 8 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.63,                    !- Speed 8 Reference Unit Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 8 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 8 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 8 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 8 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 8 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    6144.8,                  !- Speed 9 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 9 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 9 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.3473184,               !- Speed 9 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.69,                    !- Speed 9 Reference Unit Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 9 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 9 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 9 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 9 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF,          !- Speed 9 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "    6758.0,                  !- Speed 10 Reference Unit Gross Rated Total Cooling Capacity {w}",
            "    0.75,                    !- Speed 10 Reference Unit Gross Rated Sensible Heat Ratio {dimensionless}",
            "    4.0,                     !- Speed 10 Reference Unit Gross Rated Cooling COP {dimensionless}",
            "    0.37752,                 !- Speed 10 Reference Unit Rated Air Flow Rate {m3/s}",
            "    0.74,                    !- Speed 10 Reference Unit Condenser Air Flow Rate {m3/s}",
            "    ,                        !- Speed 10 Reference Unit Rated Pad Effectiveness of Evap Precooling {dimensionless}",
            "    HPACCoolCapFT,           !- Speed 10 Total Cooling Capacity Function of Temperature Curve Name",
            "    HPACCoolCapFFF,          !- Speed 10 Total Cooling Capacity Function of Air Flow Fraction Curve Name",
            "    HPACCOOLEIRFT,           !- Speed 10 Energy Input Ratio Function of Temperature Curve Name",
            "    HPACCOOLEIRFFF;          !- Speed 10 Energy Input Ratio Function of Air Flow Fraction Curve Name",
            "  Curve:Biquadratic,",
            "    HPACCoolCapFT,           !- Name",
            "    0.766956,                !- Coefficient1 Constant",
            "    0.0107756,               !- Coefficient2 x",
            "    -0.0000414703,           !- Coefficient3 x**2",
            "    0.00134961,              !- Coefficient4 y",
            "    -0.000261144,            !- Coefficient5 y**2",
            "    0.000457488,             !- Coefficient6 x*y",
            "    12.77778,                !- Minimum Value of x",
            "    23.88889,                !- Maximum Value of x",
            "    18.0,                    !- Minimum Value of y",
            "    46.11111,                !- Maximum Value of y",
            "    ,                        !- Minimum Curve Output",
            "    ,                        !- Maximum Curve Output",
            "    Temperature,             !- Input Unit Type for X",
            "    Temperature,             !- Input Unit Type for Y",
            "    Dimensionless;           !- Output Unit Type",

            "  Curve:Biquadratic,",
            "    HPACCOOLEIRFT,           !- Name",
            "    0.297145,                !- Coefficient1 Constant",
            "    0.0430933,               !- Coefficient2 x",
            "    -0.000748766,            !- Coefficient3 x**2",
            "    0.00597727,              !- Coefficient4 y",
            "    0.000482112,             !- Coefficient5 y**2",
            "    -0.000956448,            !- Coefficient6 x*y",
            "    12.77778,                !- Minimum Value of x",
            "    23.88889,                !- Maximum Value of x",
            "    18.0,                    !- Minimum Value of y",
            "    46.11111,                !- Maximum Value of y",
            "    ,                        !- Minimum Curve Output",
            "    ,                        !- Maximum Curve Output",
            "    Temperature,             !- Input Unit Type for X",
            "    Temperature,             !- Input Unit Type for Y",
            "    Dimensionless;           !- Output Unit Type",

            "  Curve:Cubic,",
            "    HPACHeatCapFFF,          !- Name",
            "    0.84,                    !- Coefficient1 Constant",
            "    0.16,                    !- Coefficient2 x",
            "    0.0,                     !- Coefficient3 x**2",
            "    0.0,                     !- Coefficient4 x**3",
            "    0.5,                     !- Minimum Value of x",
            "    1.5;                     !- Maximum Value of x",

            "  Curve:Quadratic,",
            "    HPACCoolCapFFF,          !- Name",
            "    0.8,                     !- Coefficient1 Constant",
            "    0.2,                     !- Coefficient2 x",
            "    0.0,                     !- Coefficient3 x**2",
            "    0.5,                     !- Minimum Value of x",
            "    1.5;                     !- Maximum Value of x",

            "  Curve:Quadratic,",
            "    HPACCOOLEIRFFF,          !- Name",
            "    1.156,                   !- Coefficient1 Constant",
            "    -0.1816,                 !- Coefficient2 x",
            "    0.0256,                  !- Coefficient3 x**2",
            "    0.5,                     !- Minimum Value of x",
            "    1.5;                     !- Maximum Value of x",

            "  Curve:Quadratic,",
            "    HPACCOOLPLFFPLR,         !- Name",
            "    0.85,                    !- Coefficient1 Constant",
            "    0.15,                    !- Coefficient2 x",
            "    0.0,                     !- Coefficient3 x**2",
            "    0.0,                     !- Minimum Value of x",
            "    1.0;                     !- Maximum Value of x",

            "  Curve:Quadratic,",
            "    HPACHeatEIRFFF,          !- Name",
            "    1.3824,                  !- Coefficient1 Constant",
            "    -0.4336,                 !- Coefficient2 x",
            "    0.0512,                  !- Coefficient3 x**2",
            "    0.0,                     !- Minimum Value of x",
            "    1.0;                     !- Maximum Value of x",
        });

        ASSERT_TRUE(process_idf(idf_objects));

        EXPECT_NO_THROW(GetIHPInput());
        compare_err_stream("");
    }

} // namespace IntegratedHeatPump

} // namespace EnergyPlus
