PROGRAM epfilter

! PROGRAM INFORMATION:
!   AUTHOR         Jason Glazer of GARD Analytics, Inc.
!   DATE WRITTEN   January 2005
!   MODIFIED       na
!   RE-ENGINEERED  na
!
! PURPOSE OF THIS PROGRAM:
!
!   Program reads an EnergyPlus file and changes some of the objects into
!   other objects.  It does it in one pass. While reading the file it loads
!   arrays and comments out selected objects.  Then new objects are created
!   at the end of the file.  The IDD is read.  It does not change just
!   fields in objects instead it comments out an entire object and creates
!   new ones.
!
!   The specific processing this program was made for originally was to process
!   CompactHVAC objects into conventional objects in the building energy simulation
!   program EnergyPlus.
!
! NOTE TO FUTURE AUTHORS:
!   If you are using this to create a new filter the routines that you are most likely
!   going to need to modify are:
!     SetUpObjectsToProcess
!     CreateNewObjects
!   For a new object type:
!     1.  Add block of offset parameter definitions (use spreadsheet)
!     2.  Add call to AddObjToProcess (copy another one)
!     3.  Add variable declarations for compactXYZBase, numCompactXYZ, ptCompactXYZ (copy another one an rename)
!     4.  Add GetInputXYZ routine (copy another one and search/replace) - if this is a new zone type that is served
!         by a system, then add a new do loop in the system GetInput... routine.
!     5.  In the GetInputXYZ paste in lines to fill in defaults (use spreadsheet)
!     6.  Add call to GetInputXYZ (if needed)
!     7.  In subroutine MakeSpecialIntList, add appropriate new do loops for plenums, DOAS, etc.
!         If this is making new lists (because of a new system type, for example), then add new list name varaible
!         declarations (search for zlUnitAllRetPlen)
!     8.  Add CreateXYZ routine (copy another one and search/replace/edit as needed
!         For airloop systems, one routine creates the air loop and its terminal units, and in some cases, one
!         routine makes more than one type of airloop if they are similar to each other
!         a. Search and replace the string used in the compactXYZBase, numCompactXYZ variables
!            e.g. FanCoil --> Baseboard
!         b. Search and replace the offset parameter prefix, e.g. fcz --> bbz
!     9.  Add call to CreateXYZ (if needed)
!
!   Field values are stored in the fldVal array and everything is a string
!
!   Use CountOldObj('OBJ',objPt,objCount) to determine how many of those objects appear in the file
!
!   Use NextOldObj(objPt,firstFldVal,lastFldVal) this will get the next one and continue
!       to search from the beginning of the array.  Returns the indexes to the fldVal array.
!
!   Use CreateNewObj('OBJ',lastFlag) to make a new object
!   Use AddToObjStr(comment,charactervariable,lastFlag)
!   Use AddToObjNum(comment,realvariable,lastFlag)
!   Use AddToObjFld(comment,indexFldVal,suffix, lastFlag)
!   Use WriteComment() to write out a comment
!   Use StringToReal()  convert the character variable to real
!
! METHODOLOGY EMPLOYED:
!
! REFERENCES:
!   None.
!
! OTHER NOTES:
!
!
IMPLICIT NONE

! Array that holds the field values of the objects that are saved until end
! all values are stored as alpha
CHARACTER(len=1), PARAMETER :: tabchar=CHAR(9)    ! tab character

INTEGER, PARAMETER :: MaxAlphaLength = 100
CHARACTER(len=MaxAlphaLength), ALLOCATABLE, DIMENSION(:) :: FldVal !  = ''
CHARACTER(len=MaxAlphaLength), ALLOCATABLE, DIMENSION(:) :: FldValCopy ! = ''
INTEGER                                                  :: numFldVal = 0
INTEGER                                                  :: sizeFldVal = 1000

! Array that is used to read an entire object from input IDF file before
! it is written to output IDF file either with or without commenting it out
INTEGER, PARAMETER :: MaxLineLength = 500
CHARACTER(len=MaxLineLength), ALLOCATABLE, DIMENSION(:)  :: InBuffer ! = ''
CHARACTER(len=MaxLineLength), ALLOCATABLE, DIMENSION(:)  :: InBufferCopy  ! = ''
INTEGER                                                  :: numInBuffer = 0
INTEGER                                                  :: sizeInBuffer = 100

! Array that is used to hold error messages during execution so that they
! all get written at the end (not in the middle of any object)
INTEGER, PARAMETER :: MaxErrorLength = 500
TYPE ErrMsgType
  CHARACTER(len=MaxErrorLength)  :: msgText
  INTEGER                        :: msgKind
END TYPE
TYPE (ErrMsgType), ALLOCATABLE, DIMENSION(:)  :: ErrMsgs
TYPE (ErrMsgType), ALLOCATABLE, DIMENSION(:)  :: ErrMsgsCopy
INTEGER                                                   :: numErrMsgs = 0
INTEGER                                                   :: sizeErrMsgs = 100
INTEGER, PARAMETER :: msgError = 1
INTEGER, PARAMETER :: msgWarning = 2

! Array that holds a list of integers that can be very large - see
INTEGER, ALLOCATABLE, DIMENSION(:)  :: IntList
INTEGER, ALLOCATABLE, DIMENSION(:)  :: IntListCopy
INTEGER                             :: numIntList = 0
INTEGER                             :: sizeIntList = 100

! Array that holds a real value to be sorted and brings an associated pointer with it
TYPE NumSortListType
  REAL     :: numToSort
  INTEGER  :: sortedPtr
END TYPE
TYPE (NumSortListType), ALLOCATABLE, DIMENSION(:) :: NumSortList
TYPE (NumSortListType), ALLOCATABLE, DIMENSION(:) :: NumSortListCopy
INTEGER                                           :: numNumSortList  = 0
INTEGER                                           :: sizeNumSortList  = 100

! Array of strings to be used later
! this array is different than the others because "handle" is used to save ones from
! multiple lists
TYPE StrListType
  CHARACTER(len=MaxAlphaLength) :: StrToSave
  INTEGER                       :: Handle
END TYPE
TYPE (StrListType), ALLOCATABLE, DIMENSION(:) :: StrList
TYPE (StrListType), ALLOCATABLE, DIMENSION(:) :: StrListCopy
INTEGER                                       :: numStrList = 0
INTEGER                                       :: sizeStrList = 100
INTEGER, PARAMETER                            :: handleHeatingCoilBranch = 1
INTEGER, PARAMETER                            :: handleCoolingCoilBranch = 2
INTEGER, PARAMETER                            :: handleCondenserBranch   = 3
INTEGER, PARAMETER                            :: handleChillerChWOutletNode = 4
INTEGER, PARAMETER                            :: handleBoilerOutletNode = 5
INTEGER, PARAMETER                            :: handleTowerOutletNode = 6
INTEGER, PARAMETER                            :: handleHWSupplyBranch = 7
INTEGER, PARAMETER                            :: handleCondenserSupplyBranch = 8
INTEGER, PARAMETER                            :: handleMixedSupplyBranch = 9
INTEGER, PARAMETER                            :: handleMixedDemanBranch = 10
INTEGER, PARAMETER                            :: handleMixedTowerOutletNode = 11
INTEGER, PARAMETER                            :: handleMixedBoilerOutletNode = 12
!Use AddToStrList(ToList,SubListHandle)

!   The objToProcess array stores the last object held.
!   The heldObj array stores the field pointers to the fldVal array for each instance `
!   of an object.  It links to the previous instance of the object to enable quick searching

! Array that holds definition of selected objects found during the reading process
! the actual field values are in the fldVal array
TYPE HeldObjType
  INTEGER :: ObjToProcessIndex             = 0  ! Link to object to process index to identify the type of object
  INTEGER :: FirstFldVal                   = 0  ! Index to fldVal array - beginning of values
  INTEGER :: LastFldVal                    = 0  ! Index to fldVal array - end of values
END TYPE
TYPE (HeldObjType), ALLOCATABLE, DIMENSION(:)   :: HeldObj
TYPE (HeldObjType), ALLOCATABLE, DIMENSION(:)   :: HeldObjCopy
INTEGER                                         :: numHeldObj = 0
INTEGER                                         :: sizeHeldObj = 100
INTEGER                                         :: HeldObjSearchIndex = 0

! Array holds the names of the objects that are to be held until the end
! for further processing
TYPE ObjToProcessType
  CHARACTER(len=MaxAlphaLength) :: ucName = ''   !all uppercase name
  CHARACTER(len=MaxAlphaLength) :: ccName = ''   !camel case name
  LOGICAL                       :: commentOut = .FALSE.
  INTEGER                       :: lastObjHeld = 0
  INTEGER                       :: minFldCnt = 0
  INTEGER                       :: maxFldCnt = 0
  INTEGER                       :: expFldCnt = 0
  INTEGER                       :: iddObjIndex = 0
END TYPE
TYPE (ObjToProcessType),ALLOCATABLE, DIMENSION(:) :: objToProcess
TYPE (ObjToProcessType),ALLOCATABLE, DIMENSION(:) :: objToProcessCopy
INTEGER                                           :: numObjToProcess = 0
INTEGER                                           :: sizeObjToProcess = 100

CHARACTER(len=MaxAlphaLength), ALLOCATABLE, DIMENSION(:) :: alwaysSchedule !holds value of schedule
CHARACTER(len=MaxAlphaLength), ALLOCATABLE, DIMENSION(:) :: alwaysScheduleCopy
INTEGER                            :: numAlwaysSchedule = 0
INTEGER                            :: sizeAlwaysSchedule = 10

CHARACTER(200) :: InputFileName=' '
CHARACTER(200) :: OutputFileName=' '
CHARACTER(200) :: ErrorFileName=' '
CHARACTER(200) :: IniFileName=' '
CHARACTER(200) :: IddFileName=' '
CHARACTER(256) :: IOErrorMsg=' '
INTEGER        :: LenInputFileName=0

INTEGER,PARAMETER :: InFH   = 1 !input IDF file handle
INTEGER,PARAMETER :: OutFH  = 2 !output IDF file handle
INTEGER,PARAMETER :: ErrFH  = 3 !error file name
INTEGER,PARAMETER :: IddFH  = 4 !Energy+.IDD file handle
INTEGER,PARAMETER :: IniFH  = 5 !Energy+.INI file handle

LOGICAL :: isChillerWithWaterCooled = .FALSE.
LOGICAL :: errorCondition = .FALSE.
LOGICAL :: foundCompactObject = .FALSE.
LOGICAL :: iddCanBeRead =.FALSE.

REAL :: startTime=0.0
REAL :: endTime=0.0

REAL :: heatingSpds=0.0
REAL :: coolingSpds=0.0
REAL :: maxSpd=0.0

!Arrays related to reading the IDD file

TYPE iddObjType
  CHARACTER(len=MaxAlphaLength) :: name = ''
  LOGICAL                       :: memoPresent = .false.
  CHARACTER(len=1000)           :: memo = ''
  LOGICAL                       :: unique = .false.
  LOGICAL                       :: required = .false.
  LOGICAL                       :: minFieldsPresent = .false.
  INTEGER                       :: minFields
  LOGICAL                       :: obsolete = .false.
  LOGICAL                       :: extensible = .false.
  LOGICAL                       :: beginextensiblefound = .false.
  LOGICAL                       :: spFormatPresent = .false.
  INTEGER                       :: spFormat = 0
  INTEGER                       :: firstFld = 0
  INTEGER                       :: lastFld = 0
END TYPE
TYPE (iddObjType), ALLOCATABLE, DIMENSION(:) :: iddObj
TYPE (iddObjType), ALLOCATABLE, DIMENSION(:) :: iddObjCopy
INTEGER :: objCount = 0
INTEGER :: sizeIddObj = 0

INTEGER, PARAMETER ::  spFmtSingleLine = 1
INTEGER, PARAMETER ::  spFmtVertices = 2
INTEGER, PARAMETER ::  spFmtCompactSchedule = 3

TYPE iddFldType
  CHARACTER(len=MaxAlphaLength) :: name = ''
  LOGICAL                       :: notePresent = .false.
  CHARACTER(len=200)            :: note = ''
  LOGICAL                       :: required = .false.
  LOGICAL                       :: AorN = .false.  ! A=true, N=false
  LOGICAL                       :: unitsPresent = .false.
  CHARACTER(len=MaxAlphaLength) :: units = ''
  LOGICAL                       :: ipUnitsPresent = .false.
  CHARACTER(len=MaxAlphaLength) :: ipUnits = ''
  LOGICAL                       :: minimumPresent = .false.
  LOGICAL                       :: minOrEqual     = .false.
  REAL                          :: minimum = 0.
  LOGICAL                       :: maximumPresent = .false.
  LOGICAL                       :: maxOrEqual     = .false.
  REAL                          :: maximum = 0.
  LOGICAL                       :: defaultPresent = .false.
  REAL                          :: default = 0.
  LOGICAL                       :: autoSize = .false.
  LOGICAL                       :: autoCalc = .false.
  LOGICAL                       :: typeOfDataPresent = .false.
  INTEGER                       :: typeOfData = 0
  LOGICAL                       :: retainCase = .false.
  LOGICAL                       :: keyPresent = .false.
  LOGICAL                       :: deprecated = .false.
  INTEGER                       :: firstKey = 0
  INTEGER                       :: lastKey = 0
  LOGICAL                       :: objListPresent = .false.
  CHARACTER(len=MaxAlphaLength) :: objList = ''
  LOGICAL                       :: referencePresent = .false.
  CHARACTER(len=MaxAlphaLength) :: reference = ''
END TYPE
TYPE (iddFldType), ALLOCATABLE, DIMENSION(:) :: iddFld
TYPE (iddFldType), ALLOCATABLE, DIMENSION(:) :: iddFldCopy
INTEGER :: fldCount = 0
INTEGER :: sizeIddFld = 0
INTEGER, PARAMETER :: todInteger = 1
INTEGER, PARAMETER :: todReal = 2
INTEGER, PARAMETER :: todAlpha = 3
INTEGER, PARAMETER :: todChoice = 4
INTEGER, PARAMETER :: todObjList = 5

TYPE iddKeyType
  CHARACTER(len=MaxAlphaLength) :: name = ''
END TYPE
TYPE(iddKeyType),ALLOCATABLE, DIMENSION(:) :: iddKey
TYPE(iddKeyType),ALLOCATABLE, DIMENSION(:) :: iddKeyCopy
INTEGER :: keyCount = 0
INTEGER :: sizeiddKey = 0

! The following is a flag used by developers only to enable
! code that checks if the field names that are written out
! to the expanded.idf file use fields that exist in the IDD.
! It also enables the entire IDD being read instead of only
! the HVACTemplate objects. It will affect speed so it should
! be set to off for distribution to users.
LOGICAL :: checkOutputFieldNamesAgainstIDD = .false.
INTEGER :: checkObjectIndex = 0

INTEGER :: iFH !file handle for the input file
INTEGER :: oFH !file handle for the output file

!Constant offsets for accessing fldval array

INTEGER,PARAMETER :: tNameOff = 1
INTEGER,PARAMETER :: tHeatSchedNameOff = 2
INTEGER,PARAMETER :: tHeatSetPointOff = 3
INTEGER,PARAMETER :: tCoolSchedNameOff = 4
INTEGER,PARAMETER :: tCoolSetPointOff = 5

INTEGER,PARAMETER :: vzNameOff = 1
INTEGER,PARAMETER :: vzAirHandlerNameOff = 2
INTEGER,PARAMETER :: vzTStatNameOff = 3
INTEGER,PARAMETER :: vzSupplyMaxRateOff = 4
INTEGER,PARAMETER :: vzZoneHeatSizingFactorOff = 5
INTEGER,PARAMETER :: vzZoneCoolSizingFactorOff = 6
INTEGER,PARAMETER :: vzMinFlowMethodOff = 7
INTEGER,PARAMETER :: vzConstantMinFracOff = 8
INTEGER,PARAMETER :: vzFixedMinFlowOff = 9
INTEGER,PARAMETER :: vzMinFracSchedNameOff = 10
INTEGER,PARAMETER :: vzOutAirTypeOff = 11
INTEGER,PARAMETER :: vzOutAirFlowPersonOff = 12
INTEGER,PARAMETER :: vzOutAirFlowAreaOff = 13
INTEGER,PARAMETER :: vzOutAirFlowZoneOff = 14
INTEGER,PARAMETER :: vzReheatTypeOff = 15
INTEGER,PARAMETER :: vzReheatSchedNameOff = 16
INTEGER,PARAMETER :: vzDamperActionOff = 17
INTEGER,PARAMETER :: vzMaxReheatFlowPerAreaOff = 18
INTEGER,PARAMETER :: vzMaxReheatFlowFractionOff = 19
INTEGER,PARAMETER :: vzMaxReheatAirTempOff = 20
INTEGER,PARAMETER :: vzDesignSpecOANameCtrlOff = 21
INTEGER,PARAMETER :: vzZoneSupplyPlenumNameOff = 22
INTEGER,PARAMETER :: vzZoneReturnPlenumNameOff = 23
INTEGER,PARAMETER :: vzBaseboardTypeOff = 24
INTEGER,PARAMETER :: vzBaseboardAvailSchedNameOff = 25
INTEGER,PARAMETER :: vzBaseboardCapOff = 26
INTEGER,PARAMETER :: vzCoolDesignTempMethodOff = 27
INTEGER,PARAMETER :: vzCoolDesignTempOff = 28
INTEGER,PARAMETER :: vzCoolDesignTempDiffOff = 29
INTEGER,PARAMETER :: vzHeatDesignTempMethodOff = 30
INTEGER,PARAMETER :: vzHeatDesignTempOff = 31
INTEGER,PARAMETER :: vzHeatDesignTempDiffOff = 32
INTEGER,PARAMETER :: vzDesSpecOANameSizOff = 33
INTEGER,PARAMETER :: vzDesSpecOADistNameOff = 34

INTEGER,PARAMETER :: fpvzNameOff = 1
INTEGER,PARAMETER :: fpvzAirHandlerNameOff = 2
INTEGER,PARAMETER :: fpvzTStatNameOff = 3
INTEGER,PARAMETER :: fpvzPriSupplyMaxRateOff = 4
INTEGER,PARAMETER :: fpvzZoneHeatSizingFactorOff = 5
INTEGER,PARAMETER :: fpvzZoneCoolSizingFactorOff = 6
INTEGER,PARAMETER :: fpvzPriSupplyMinFracOff = 7
INTEGER,PARAMETER :: fpvzSecSupplyMaxRateOff = 8
INTEGER,PARAMETER :: fpvzFlowTypeOff = 9
INTEGER,PARAMETER :: fpvzFanOnFlowFracOff = 10
INTEGER,PARAMETER :: fpvzOutAirTypeOff = 11
INTEGER,PARAMETER :: fpvzOutAirFlowPersonOff = 12
INTEGER,PARAMETER :: fpvzOutAirFlowAreaOff = 13
INTEGER,PARAMETER :: fpvzOutAirFlowZoneOff = 14
INTEGER,PARAMETER :: fpvzReheatTypeOff = 15
INTEGER,PARAMETER :: fpvzReheatSchedNameOff = 16
INTEGER,PARAMETER :: fpvzFanTotEfficiencyOff = 17
INTEGER,PARAMETER :: fpvzFanPressureOff = 18
INTEGER,PARAMETER :: fpvzFanMotorEfficiencyOff = 19
INTEGER,PARAMETER :: fpvzZoneSupplyPlenumNameOff = 20
INTEGER,PARAMETER :: fpvzZoneReturnPlenumNameOff = 21
INTEGER,PARAMETER :: fpvzBaseboardTypeOff = 22
INTEGER,PARAMETER :: fpvzBaseboardAvailSchedNameOff = 23
INTEGER,PARAMETER :: fpvzBaseboardCapOff = 24
INTEGER,PARAMETER :: fpvzCoolDesignTempMethodOff = 25
INTEGER,PARAMETER :: fpvzCoolDesignTempOff = 26
INTEGER,PARAMETER :: fpvzCoolDesignTempDiffOff = 27
INTEGER,PARAMETER :: fpvzHeatDesignTempMethodOff = 28
INTEGER,PARAMETER :: fpvzHeatDesignTempOff = 29
INTEGER,PARAMETER :: fpvzHeatDesignTempDiffOff = 30
INTEGER,PARAMETER :: fpvzZoneFanSchedNameOff = 31
INTEGER,PARAMETER :: fpvzDesSpecOANameOff = 32
INTEGER,PARAMETER :: fpvzDesSpecOADistNameOff = 33

INTEGER,PARAMETER :: hcvzNameOff = 1
INTEGER,PARAMETER :: hcvzAirHandlerNameOff = 2
INTEGER,PARAMETER :: hcvzTStatNameOff = 3
INTEGER,PARAMETER :: hcvzSupplyMaxRateOff = 4
INTEGER,PARAMETER :: hcvzZoneHeatSizingFactorOff = 5
INTEGER,PARAMETER :: hcvzZoneCoolSizingFactorOff = 6
INTEGER,PARAMETER :: hcvzConstantMinFracOff = 7
INTEGER,PARAMETER :: hcvzOutAirTypeOff = 8
INTEGER,PARAMETER :: hcvzOutAirFlowPersonOff = 9
INTEGER,PARAMETER :: hcvzOutAirFlowAreaOff = 10
INTEGER,PARAMETER :: hcvzOutAirFlowZoneOff = 11
INTEGER,PARAMETER :: hcvzDesignSpecOANameOff = 12
INTEGER,PARAMETER :: hcvzDesSpecOADistNameOff = 13
INTEGER,PARAMETER :: hcvzReheatTypeOff = 14
INTEGER,PARAMETER :: hcvzReheatSchedNameOff = 15
INTEGER,PARAMETER :: hcvzMaxReheatAirTempOff = 16
INTEGER,PARAMETER :: hcvzZoneSupplyPlenumNameOff = 17
INTEGER,PARAMETER :: hcvzZoneReturnPlenumNameOff = 18
INTEGER,PARAMETER :: hcvzBaseboardTypeOff = 19
INTEGER,PARAMETER :: hcvzBaseboardAvailSchedNameOff = 20
INTEGER,PARAMETER :: hcvzBaseboardCapOff = 21
INTEGER,PARAMETER :: hcvzCoolDesignTempMethodOff = 22
INTEGER,PARAMETER :: hcvzCoolDesignTempOff = 23
INTEGER,PARAMETER :: hcvzCoolDesignTempDiffOff = 24
INTEGER,PARAMETER :: hcvzHeatDesignTempMethodOff = 25
INTEGER,PARAMETER :: hcvzHeatDesignTempOff = 26
INTEGER,PARAMETER :: hcvzHeatDesignTempDiffOff = 27

INTEGER,PARAMETER :: vsAirHandlerNameOff = 1
INTEGER,PARAMETER :: vsSysAvailSchedNameOff = 2
INTEGER,PARAMETER :: vsSupplyMaxRateOff = 3
INTEGER,PARAMETER :: vsSupplyMinRateOff = 4
INTEGER,PARAMETER :: vsSupplyEfficiencyOff = 5
INTEGER,PARAMETER :: vsSupplyPressureOff = 6
INTEGER,PARAMETER :: vsSupplyFanMotorEffOff = 7
INTEGER,PARAMETER :: vsSupplyFanFractionOff = 8
INTEGER,PARAMETER :: vsCoolCoilKindOff = 9
INTEGER,PARAMETER :: vsCoolAvailSchedNameOff = 10
INTEGER,PARAMETER :: vsCoolSetPtSchedNameOff = 11
INTEGER,PARAMETER :: vsCoolSetPtDesignOff = 12
INTEGER,PARAMETER :: vsHeatCoilKindOff = 13
INTEGER,PARAMETER :: vsHeatAvailSchedNameOff = 14
INTEGER,PARAMETER :: vsHeatSetPtSchedNameOff = 15
INTEGER,PARAMETER :: vsHeatSetPtDesignOff = 16
INTEGER,PARAMETER :: vsHeatEffiencyOff = 17
INTEGER,PARAMETER :: vsHeatParasiticOff = 18
INTEGER,PARAMETER :: vsPreheatCoilKindOff = 19
INTEGER,PARAMETER :: vsPreheatAvailSchedNameOff = 20
INTEGER,PARAMETER :: vsPreheatSetPtSchedNameOff = 21
INTEGER,PARAMETER :: vsPreheatSetPtConstantOff = 22
INTEGER,PARAMETER :: vsPreheatEffiencyOff = 23
INTEGER,PARAMETER :: vsPreheatParasiticOff = 24
INTEGER,PARAMETER :: vsMaxOutsideFlowOff = 25
INTEGER,PARAMETER :: vsMinOutsideFlowOff = 26
INTEGER,PARAMETER :: vsMinOutsideControlOff = 27
INTEGER,PARAMETER :: vsMinOutFlowSchedNameOff = 28
INTEGER,PARAMETER :: vsEconomizerKindOff = 29
INTEGER,PARAMETER :: vsEconoLockoutKindOff = 30
INTEGER,PARAMETER :: vsEconoUpLimitOff = 31
INTEGER,PARAMETER :: vsEconoLowLimitOff = 32
INTEGER,PARAMETER :: vsEconoEnthLimitOff = 33
INTEGER,PARAMETER :: vsEconoUpDewLimitOff = 34
INTEGER,PARAMETER :: vsSupplyPlenumNameOff = 35
INTEGER,PARAMETER :: vsReturnPlenumNameOff = 36
INTEGER,PARAMETER :: vsSupplyFanPlaceOff = 37
INTEGER,PARAMETER :: vsSupplyFanPLRCoeffOff = 38
INTEGER,PARAMETER :: vsNightCycleControlOff = 39
INTEGER,PARAMETER :: vsNightCycleCtrlZoneNameOff = 40
INTEGER,PARAMETER :: vsHeatRecTypeOff = 41
INTEGER,PARAMETER :: vsHeatRecSenEffOff = 42
INTEGER,PARAMETER :: vsHeatRecLatEffOff = 43
INTEGER,PARAMETER :: vsCoolSetPtResetTypeOff = 44
INTEGER,PARAMETER :: vsHeatSetPtResetTypeOff = 45
INTEGER,PARAMETER :: vsDehumCtrlTypeOff = 46
INTEGER,PARAMETER :: vsDehumCtrlZoneOff = 47
INTEGER,PARAMETER :: vsDehumSetPtOff = 48
INTEGER,PARAMETER :: vsHumidCtrlTypeOff = 49
INTEGER,PARAMETER :: vsHumidAvailSchedNameOff = 50
INTEGER,PARAMETER :: vsHumidCapOff = 51
INTEGER,PARAMETER :: vsHumidRatedElecOff = 52
INTEGER,PARAMETER :: vsHumidCtrlZoneOff = 53
INTEGER,PARAMETER :: vsHumidSetPtOff = 54
INTEGER,PARAMETER :: vsSizingOptionOff = 55
INTEGER,PARAMETER :: vsReturnFanPresentOff = 56
INTEGER,PARAMETER :: vsReturnEfficiencyOff = 57
INTEGER,PARAMETER :: vsReturnPressureOff = 58
INTEGER,PARAMETER :: vsReturnFanMotorEffOff = 59
INTEGER,PARAMETER :: vsReturnFanFractionOff = 60
INTEGER,PARAMETER :: vsReturnFanPLRCoeffOff = 61

INTEGER,PARAMETER :: uzNameOff = 1
INTEGER,PARAMETER :: uzAirHandlerNameOff = 2
INTEGER,PARAMETER :: uzTStatNameOff = 3
INTEGER,PARAMETER :: uzSupplyMaxRateOff = 4
INTEGER,PARAMETER :: uzZoneHeatSizingFactorOff = 5
INTEGER,PARAMETER :: uzZoneCoolSizingFactorOff = 6
INTEGER,PARAMETER :: uzOutAirTypeOff = 7
INTEGER,PARAMETER :: uzOutAirFlowPersonOff = 8
INTEGER,PARAMETER :: uzOutAirFlowAreaOff = 9
INTEGER,PARAMETER :: uzOutAirFlowZoneOff = 10
INTEGER,PARAMETER :: uzZoneSupplyPlenumNameOff = 11
INTEGER,PARAMETER :: uzZoneReturnPlenumNameOff = 12
INTEGER,PARAMETER :: uzBaseboardTypeOff = 13
INTEGER,PARAMETER :: uzBaseboardAvailSchedNameOff = 14
INTEGER,PARAMETER :: uzBaseboardCapOff = 15
INTEGER,PARAMETER :: uzCoolDesignTempMethodOff = 16
INTEGER,PARAMETER :: uzCoolDesignTempOff = 17
INTEGER,PARAMETER :: uzCoolDesignTempDiffOff = 18
INTEGER,PARAMETER :: uzHeatDesignTempMethodOff = 19
INTEGER,PARAMETER :: uzHeatDesignTempOff = 20
INTEGER,PARAMETER :: uzHeatDesignTempDiffOff = 21
INTEGER,PARAMETER :: uzDesSpecOANameOff = 22
INTEGER,PARAMETER :: uzDesSpecOADistNameOff = 23

INTEGER,PARAMETER :: usAirHandlerNameOff = 1
INTEGER,PARAMETER :: usSysAvailSchedNameOff = 2
INTEGER,PARAMETER :: usControlZoneOff = 3
INTEGER,PARAMETER :: usSupplyMaxRateOff = 4
INTEGER,PARAMETER :: usSupplyFanModeOff = 5
INTEGER,PARAMETER :: usSupplyEfficiencyOff = 6
INTEGER,PARAMETER :: usSupplyPressureOff = 7
INTEGER,PARAMETER :: usSupplyFanMotorEffOff = 8
INTEGER,PARAMETER :: usSupplyFanFractionOff = 9
INTEGER,PARAMETER :: usCoolCoilKindOff = 10
INTEGER,PARAMETER :: usCoolAvailSchedNameOff = 11
INTEGER,PARAMETER :: usCoolDesignSupplyTempOff = 12
INTEGER,PARAMETER :: usCoolCoilCapOff = 13
INTEGER,PARAMETER :: usCoolCoilSHROff = 14
INTEGER,PARAMETER :: usCoolCoilCOPOff = 15
INTEGER,PARAMETER :: usHeatCoilKindOff = 16
INTEGER,PARAMETER :: usHeatAvailSchedNameOff = 17
INTEGER,PARAMETER :: usHeatDesignSupplyTempOff = 18
INTEGER,PARAMETER :: usHeatCoilCapOff = 19
INTEGER,PARAMETER :: usHeatEffiencyOff = 20
INTEGER,PARAMETER :: usHeatParasiticOff = 21
INTEGER,PARAMETER :: usMaxOutsideFlowOff = 22
INTEGER,PARAMETER :: usMinOutsideFlowOff = 23
INTEGER,PARAMETER :: usMinOutFlowSchedNameOff = 24
INTEGER,PARAMETER :: usEconomizerKindOff = 25
INTEGER,PARAMETER :: usEconoLockoutKindOff = 26
INTEGER,PARAMETER :: usEconoUpLimitOff = 27
INTEGER,PARAMETER :: usEconoLowLimitOff = 28
INTEGER,PARAMETER :: usEconoEnthLimitOff = 29
INTEGER,PARAMETER :: usEconoUpDewLimitOff = 30
INTEGER,PARAMETER :: usSupplyPlenumNameOff = 31
INTEGER,PARAMETER :: usReturnPlenumNameOff = 32
INTEGER,PARAMETER :: usSupplyFanPlaceOff = 33
INTEGER,PARAMETER :: usNightCycleControlOff = 34
INTEGER,PARAMETER :: usNightCycleCtrlZoneNameOff = 35
INTEGER,PARAMETER :: usHeatRecTypeOff = 36
INTEGER,PARAMETER :: usHeatRecSenEffOff = 37
INTEGER,PARAMETER :: usHeatRecLatEffOff = 38
INTEGER,PARAMETER :: usDehumCtrlTypeOff = 39
INTEGER,PARAMETER :: usDehumSetPtOff = 40
INTEGER,PARAMETER :: usHumidCtrlTypeOff = 41
INTEGER,PARAMETER :: usHumidAvailSchedNameOff = 42
INTEGER,PARAMETER :: usHumidCapOff = 43
INTEGER,PARAMETER :: usHumidRatedElecOff = 44
INTEGER,PARAMETER :: usHumidCtrlZoneOff = 45
INTEGER,PARAMETER :: usHumidSetPtOff = 46
INTEGER,PARAMETER :: usReturnFanPresentOff = 47
INTEGER,PARAMETER :: usReturnEfficiencyOff = 48
INTEGER,PARAMETER :: usReturnPressureOff = 49
INTEGER,PARAMETER :: usReturnFanMotorEffOff = 50
INTEGER,PARAMETER :: usReturnFanFractionOff = 51

INTEGER,PARAMETER :: cvzNameOff = 1
INTEGER,PARAMETER :: cvzAirHandlerNameOff = 2
INTEGER,PARAMETER :: cvzTStatNameOff = 3
INTEGER,PARAMETER :: cvzSupplyMaxRateOff = 4
INTEGER,PARAMETER :: cvzZoneHeatSizingFactorOff = 5
INTEGER,PARAMETER :: cvzZoneCoolSizingFactorOff = 6
INTEGER,PARAMETER :: cvzOutAirTypeOff = 7
INTEGER,PARAMETER :: cvzOutAirFlowPersonOff = 8
INTEGER,PARAMETER :: cvzOutAirFlowAreaOff = 9
INTEGER,PARAMETER :: cvzOutAirFlowZoneOff = 10
INTEGER,PARAMETER :: cvzDesignSpecOANameOff = 11
INTEGER,PARAMETER :: cvzDesSpecOADistNameOff = 12
INTEGER,PARAMETER :: cvzReheatTypeOff = 13
INTEGER,PARAMETER :: cvzReheatSchedNameOff = 14
INTEGER,PARAMETER :: cvzMaxReheatAirTempOff = 15
INTEGER,PARAMETER :: cvzZoneSupplyPlenumNameOff = 16
INTEGER,PARAMETER :: cvzZoneReturnPlenumNameOff = 17
INTEGER,PARAMETER :: cvzBaseboardTypeOff = 18
INTEGER,PARAMETER :: cvzBaseboardAvailSchedNameOff = 19
INTEGER,PARAMETER :: cvzBaseboardCapOff = 20
INTEGER,PARAMETER :: cvzCoolDesignTempMethodOff = 21
INTEGER,PARAMETER :: cvzCoolDesignTempOff = 22
INTEGER,PARAMETER :: cvzCoolDesignTempDiffOff = 23
INTEGER,PARAMETER :: cvzHeatDesignTempMethodOff = 24
INTEGER,PARAMETER :: cvzHeatDesignTempOff = 25
INTEGER,PARAMETER :: cvzHeatDesignTempDiffOff = 26

INTEGER,PARAMETER :: cvsAirHandlerNameOff = 1
INTEGER,PARAMETER :: cvsSysAvailSchedNameOff = 2
INTEGER,PARAMETER :: cvsSupplyRateOff = 3
INTEGER,PARAMETER :: cvsSupplyEfficiencyOff = 4
INTEGER,PARAMETER :: cvsSupplyPressureOff = 5
INTEGER,PARAMETER :: cvsSupplyFanMotorEffOff = 6
INTEGER,PARAMETER :: cvsSupplyFanFractionOff = 7
INTEGER,PARAMETER :: cvsSupplyFanPlaceOff = 8
INTEGER,PARAMETER :: cvsCoolCoilKindOff = 9
INTEGER,PARAMETER :: cvsCoolAvailSchedNameOff = 10
INTEGER,PARAMETER :: cvsCoolSetPtCtrlTypeOff = 11
INTEGER,PARAMETER :: cvsCoolControlZoneOff = 12
INTEGER,PARAMETER :: cvsCoolSetPtDesignOff = 13
INTEGER,PARAMETER :: cvsCoolSetPtSchedNameOff = 14
INTEGER,PARAMETER :: cvsCoolSetPtAtODBLoOff = 15
INTEGER,PARAMETER :: cvsCoolResetODBLoOff = 16
INTEGER,PARAMETER :: cvsCoolSetPtAtODBHiOff = 17
INTEGER,PARAMETER :: cvsCoolResetODBHiOff = 18
INTEGER,PARAMETER :: cvsHeatCoilKindOff = 19
INTEGER,PARAMETER :: cvsHeatAvailSchedNameOff = 20
INTEGER,PARAMETER :: cvsHeatSetPtCtrlTypeOff = 21
INTEGER,PARAMETER :: cvsHeatControlZoneOff = 22
INTEGER,PARAMETER :: cvsHeatSetPtDesignOff = 23
INTEGER,PARAMETER :: cvsHeatSetPtSchedNameOff = 24
INTEGER,PARAMETER :: cvsHeatSetPtAtODBLoOff = 25
INTEGER,PARAMETER :: cvsHeatResetODBLoOff = 26
INTEGER,PARAMETER :: cvsHeatSetPtAtODBHiOff = 27
INTEGER,PARAMETER :: cvsHeatResetODBHiOff = 28
INTEGER,PARAMETER :: cvsHeatCoilCapOff = 29
INTEGER,PARAMETER :: cvsHeatEffiencyOff = 30
INTEGER,PARAMETER :: cvsHeatParasiticOff = 31
INTEGER,PARAMETER :: cvsPreheatCoilKindOff = 32
INTEGER,PARAMETER :: cvsPreheatAvailSchedNameOff = 33
INTEGER,PARAMETER :: cvsPreheatSetPtConstantOff = 34
INTEGER,PARAMETER :: cvsPreheatSetPtSchedNameOff = 35
INTEGER,PARAMETER :: cvsPreheatEffiencyOff = 36
INTEGER,PARAMETER :: cvsPreheatParasiticOff = 37
INTEGER,PARAMETER :: cvsMaxOutsideFlowOff = 38
INTEGER,PARAMETER :: cvsMinOutsideFlowOff = 39
INTEGER,PARAMETER :: cvsMinOutFlowSchedNameOff = 40
INTEGER,PARAMETER :: cvsEconomizerKindOff = 41
INTEGER,PARAMETER :: cvsEconoUpLimitOff = 42
INTEGER,PARAMETER :: cvsEconoLowLimitOff = 43
INTEGER,PARAMETER :: cvsEconoEnthLimitOff = 44
INTEGER,PARAMETER :: cvsEconoUpDewLimitOff = 45
INTEGER,PARAMETER :: cvsSupplyPlenumNameOff = 46
INTEGER,PARAMETER :: cvsReturnPlenumNameOff = 47
INTEGER,PARAMETER :: cvsNightCycleControlOff = 48
INTEGER,PARAMETER :: cvsNightCycleCtrlZoneNameOff = 49
INTEGER,PARAMETER :: cvsHeatRecTypeOff = 50
INTEGER,PARAMETER :: cvsHeatRecSenEffOff = 51
INTEGER,PARAMETER :: cvsHeatRecLatEffOff = 52
INTEGER,PARAMETER :: cvsHXTypeOff = 53
INTEGER,PARAMETER :: cvsFrostControlOff = 54
INTEGER,PARAMETER :: cvsDehumCtrlTypeOff = 55
INTEGER,PARAMETER :: cvsDehumCtrlZoneOff = 56
INTEGER,PARAMETER :: cvsDehumSetPtOff = 57
INTEGER,PARAMETER :: cvsDehumSetPtSchedNameOff = 58
INTEGER,PARAMETER :: cvsHumidCtrlTypeOff = 59
INTEGER,PARAMETER :: cvsHumidAvailSchedNameOff = 60
INTEGER,PARAMETER :: cvsHumidCapOff = 61
INTEGER,PARAMETER :: cvsHumidRatedElecOff = 62
INTEGER,PARAMETER :: cvsHumidCtrlZoneOff = 63
INTEGER,PARAMETER :: cvsHumidSetPtOff = 64
INTEGER,PARAMETER :: cvsHumidSetPtSchedNameOff = 65
INTEGER,PARAMETER :: cvsReturnFanPresentOff = 66
INTEGER,PARAMETER :: cvsReturnEfficiencyOff = 67
INTEGER,PARAMETER :: cvsReturnPressureOff = 68
INTEGER,PARAMETER :: cvsReturnFanMotorEffOff = 69
INTEGER,PARAMETER :: cvsReturnFanFractionOff = 70

INTEGER,PARAMETER :: fczNameOff = 1
INTEGER,PARAMETER :: fczTStatNameOff = 2
INTEGER,PARAMETER :: fczSupplyMaxRateOff = 3
INTEGER,PARAMETER :: fczZoneHeatSizingFactorOff = 4
INTEGER,PARAMETER :: fczZoneCoolSizingFactorOff = 5
INTEGER,PARAMETER :: fczOutAirTypeOff = 6
INTEGER,PARAMETER :: fczOutAirFlowPersonOff = 7
INTEGER,PARAMETER :: fczOutAirFlowAreaOff = 8
INTEGER,PARAMETER :: fczOutAirFlowZoneOff = 9
INTEGER,PARAMETER :: fczSysAvailSchedNameOff = 10
INTEGER,PARAMETER :: fczSupplyEfficiencyOff = 11
INTEGER,PARAMETER :: fczSupplyPressureOff = 12
INTEGER,PARAMETER :: fczSupplyFanMotorEffOff = 13
INTEGER,PARAMETER :: fczSupplyFanMotorFracOff = 14
INTEGER,PARAMETER :: fczCoolCoilKindOff = 15
INTEGER,PARAMETER :: fczCoolAvailSchedNameOff = 16
INTEGER,PARAMETER :: fczCoolSetPtDesignOff = 17
INTEGER,PARAMETER :: fczHeatCoilKindOff = 18
INTEGER,PARAMETER :: fczHeatAvailSchedNameOff = 19
INTEGER,PARAMETER :: fczHeatSetPtDesignOff = 20
INTEGER,PARAMETER :: fczDedOutAirSysNameOff = 21
INTEGER,PARAMETER :: fczCoolDesignTempMethodOff = 22
INTEGER,PARAMETER :: fczCoolDesignTempDiffOff = 23
INTEGER,PARAMETER :: fczHeatDesignTempMethodOff = 24
INTEGER,PARAMETER :: fczHeatDesignTempDiffOff = 25
INTEGER,PARAMETER :: fczDesSpecOANameOff = 26
INTEGER,PARAMETER :: fczDesSpecOADistNameOff = 27
INTEGER,PARAMETER :: fczCapCtrlTypeOff = 28
INTEGER,PARAMETER :: fczLowSpdSupplyRatioOff = 29
INTEGER,PARAMETER :: fczMedSpdSupplyRatioOff = 30
INTEGER,PARAMETER :: fczOutAirSchedNameOff = 31
INTEGER,PARAMETER :: fczBaseboardTypeOff = 32
INTEGER,PARAMETER :: fczBaseboardAvailSchedNameOff = 33
INTEGER,PARAMETER :: fczBaseboardCapOff = 34

INTEGER,PARAMETER :: pazNameOff = 1
INTEGER,PARAMETER :: pazTStatNameOff = 2
INTEGER,PARAMETER :: pazSysAvailSchedNameOff = 3
INTEGER,PARAMETER :: pazMaxHeatSupplyTempOff = 4
INTEGER,PARAMETER :: pazMinCoolSupplyTempOff = 5
INTEGER,PARAMETER :: pazMaxHeatSupplyHumRatOff = 6
INTEGER,PARAMETER :: pazMinCoolSupplyHumRatOff = 7
INTEGER,PARAMETER :: pazHeatLimitOff = 8
INTEGER,PARAMETER :: pazMaxHeatAirFlowRateOff = 9
INTEGER,PARAMETER :: pazMaxSensHeatCapOff = 10
INTEGER,PARAMETER :: pazCoolLimitOff = 11
INTEGER,PARAMETER :: pazMaxCoolAirFlowRateOff = 12
INTEGER,PARAMETER :: pazMaxTotalCoolCapOff = 13
INTEGER,PARAMETER :: pazHeatAvailSchedNameOff = 14
INTEGER,PARAMETER :: pazCoolAvailScheduleNameOff = 15
INTEGER,PARAMETER :: pazDehumCtrlTypeOff = 16
INTEGER,PARAMETER :: pazCoolSHROff = 17
INTEGER,PARAMETER :: pazDehumSetPtOff = 18
INTEGER,PARAMETER :: pazHumidCtrlTypeOff = 19
INTEGER,PARAMETER :: pazHumidSetPtOff = 20
INTEGER,PARAMETER :: pazOutAirTypeOff = 21
INTEGER,PARAMETER :: pazOutAirFlowPersonOff = 22
INTEGER,PARAMETER :: pazOutAirFlowAreaOff = 23
INTEGER,PARAMETER :: pazOutAirFlowZoneOff = 24
INTEGER,PARAMETER :: pazDesSpecOANameOff = 25
INTEGER,PARAMETER :: pazDCVTypeOff = 26
INTEGER,PARAMETER :: pazEconomizerTypeOff = 27
INTEGER,PARAMETER :: pazHeatRecTypeOff = 28
INTEGER,PARAMETER :: pazHeatRecSenEffOff = 29
INTEGER,PARAMETER :: pazHeatRecLatEffOff = 30

INTEGER,PARAMETER :: bbzNameOff = 1
INTEGER,PARAMETER :: bbzTStatNameOff = 2
INTEGER,PARAMETER :: bbzZoneHeatSizingFactorOff = 3
INTEGER,PARAMETER :: bbzBaseboardTypeOff = 4
INTEGER,PARAMETER :: bbzBaseboardAvailSchedNameOff = 5
INTEGER,PARAMETER :: bbzBaseboardCapOff = 6
INTEGER,PARAMETER :: bbzDedOutAirSysNameOff = 7
INTEGER,PARAMETER :: bbzOutAirTypeOff = 8
INTEGER,PARAMETER :: bbzOutAirFlowPersonOff = 9
INTEGER,PARAMETER :: bbzOutAirFlowAreaOff = 10
INTEGER,PARAMETER :: bbzOutAirFlowZoneOff = 11
INTEGER,PARAMETER :: bbzDesSpecOANameOff = 12
INTEGER,PARAMETER :: bbzDesSpecOADistNameOff = 13

INTEGER,PARAMETER :: ptaczNameOff = 1
INTEGER,PARAMETER :: ptaczTStatNameOff = 2
INTEGER,PARAMETER :: ptaczSupplyCoolFlowRateOff = 3
INTEGER,PARAMETER :: ptaczSupplyHeatFlowRateOff = 4
INTEGER,PARAMETER :: ptaczSupplyNoLoadFlowRateOff = 5
INTEGER,PARAMETER :: ptaczZoneHeatSizingFactorOff = 6
INTEGER,PARAMETER :: ptaczZoneCoolSizingFactorOff = 7
INTEGER,PARAMETER :: ptaczOutAirTypeOff = 8
INTEGER,PARAMETER :: ptaczOutAirFlowPersonOff = 9
INTEGER,PARAMETER :: ptaczOutAirFlowAreaOff = 10
INTEGER,PARAMETER :: ptaczOutAirFlowZoneOff = 11
INTEGER,PARAMETER :: ptaczSysAvailSchedNameOff = 12
INTEGER,PARAMETER :: ptaczFanModeOff = 13
INTEGER,PARAMETER :: ptaczFanPlacementOff = 14
INTEGER,PARAMETER :: ptaczFanTotEfficiencyOff = 15
INTEGER,PARAMETER :: ptaczFanPressureOff = 16
INTEGER,PARAMETER :: ptaczFanMotorEfficiencyOff = 17
INTEGER,PARAMETER :: ptaczCoolCoilKindOff = 18
INTEGER,PARAMETER :: ptaczCoolAvailSchedNameOff = 19
INTEGER,PARAMETER :: ptaczCoolCoilCapOff = 20
INTEGER,PARAMETER :: ptaczCoolCoilSHROff = 21
INTEGER,PARAMETER :: ptaczCoolCoilCOPOff = 22
INTEGER,PARAMETER :: ptaczHeatCoilKindOff = 23
INTEGER,PARAMETER :: ptaczHeatAvailSchedNameOff = 24
INTEGER,PARAMETER :: ptaczHeatCoilCapOff = 25
INTEGER,PARAMETER :: ptaczHeatEfficiencyOff = 26
INTEGER,PARAMETER :: ptaczHeatParasiticOff = 27
INTEGER,PARAMETER :: ptaczDedOutAirSysNameOff = 28
INTEGER,PARAMETER :: ptaczCoolDesignTempMethodOff = 29
INTEGER,PARAMETER :: ptaczCoolDesignTempOff = 30
INTEGER,PARAMETER :: ptaczCoolDesignTempDiffOff = 31
INTEGER,PARAMETER :: ptaczHeatDesignTempMethodOff = 32
INTEGER,PARAMETER :: ptaczHeatDesignTempOff = 33
INTEGER,PARAMETER :: ptaczHeatDesignTempDiffOff = 34
INTEGER,PARAMETER :: ptaczDesSpecOANameOff = 35
INTEGER,PARAMETER :: ptaczDesSpecOADistNameOff = 36
INTEGER,PARAMETER :: ptaczBaseboardTypeOff = 37
INTEGER,PARAMETER :: ptaczBaseboardAvailSchedNameOff = 38
INTEGER,PARAMETER :: ptaczBaseboardCapOff = 39
INTEGER,PARAMETER :: ptaczCapCtrlTypeOff = 40

INTEGER,PARAMETER :: pthpzNameOff = 1
INTEGER,PARAMETER :: pthpzTStatNameOff = 2
INTEGER,PARAMETER :: pthpzSupplyCoolFlowRateOff = 3
INTEGER,PARAMETER :: pthpzSupplyHeatFlowRateOff = 4
INTEGER,PARAMETER :: pthpzSupplyNoLoadFlowRateOff = 5
INTEGER,PARAMETER :: pthpzZoneHeatSizingFactorOff = 6
INTEGER,PARAMETER :: pthpzZoneCoolSizingFactorOff = 7
INTEGER,PARAMETER :: pthpzOutAirTypeOff = 8
INTEGER,PARAMETER :: pthpzOutAirFlowPersonOff = 9
INTEGER,PARAMETER :: pthpzOutAirFlowAreaOff = 10
INTEGER,PARAMETER :: pthpzOutAirFlowZoneOff = 11
INTEGER,PARAMETER :: pthpzSysAvailSchedNameOff = 12
INTEGER,PARAMETER :: pthpzFanModeOff = 13
INTEGER,PARAMETER :: pthpzFanPlacementOff = 14
INTEGER,PARAMETER :: pthpzFanTotEfficiencyOff = 15
INTEGER,PARAMETER :: pthpzFanPressureOff = 16
INTEGER,PARAMETER :: pthpzFanMotorEfficiencyOff = 17
INTEGER,PARAMETER :: pthpzCoolCoilKindOff = 18
INTEGER,PARAMETER :: pthpzCoolAvailSchedNameOff = 19
INTEGER,PARAMETER :: pthpzCoolCoilCapOff = 20
INTEGER,PARAMETER :: pthpzCoolCoilSHROff = 21
INTEGER,PARAMETER :: pthpzCoolCoilCOPOff = 22
INTEGER,PARAMETER :: pthpzHPHeatCoilKindOff = 23
INTEGER,PARAMETER :: pthpzHeatAvailSchedNameOff = 24
INTEGER,PARAMETER :: pthpzHPHeatCoilCapOff = 25
INTEGER,PARAMETER :: pthpzHPHeatCoilCOPOff = 26
INTEGER,PARAMETER :: pthpzHPHeatMinODBOff = 27
INTEGER,PARAMETER :: pthpzHPDefrMaxODBOff = 28
INTEGER,PARAMETER :: pthpzHPDefrTypeOff = 29
INTEGER,PARAMETER :: pthpzHPDefrCtrlTypeOff = 30
INTEGER,PARAMETER :: pthpzHPDefrTimeFracOff = 31
INTEGER,PARAMETER :: pthpzSuppHeatCoilKindOff = 32
INTEGER,PARAMETER :: pthpzSuppHeatAvailSchedNameOff = 33
INTEGER,PARAMETER :: pthpzSuppHeatCoilCapOff = 34
INTEGER,PARAMETER :: pthpzSuppHeatMaxODBOff = 35
INTEGER,PARAMETER :: pthpzSuppHeatEfficiencyOff = 36
INTEGER,PARAMETER :: pthpzSuppHeatParasiticOff = 37
INTEGER,PARAMETER :: pthpzDedOutAirSysNameOff = 38
INTEGER,PARAMETER :: pthpzCoolDesignTempMethodOff = 39
INTEGER,PARAMETER :: pthpzCoolDesignTempOff = 40
INTEGER,PARAMETER :: pthpzCoolDesignTempDiffOff = 41
INTEGER,PARAMETER :: pthpzHeatDesignTempMethodOff = 42
INTEGER,PARAMETER :: pthpzHeatDesignTempOff = 43
INTEGER,PARAMETER :: pthpzHeatDesignTempDiffOff = 44
INTEGER,PARAMETER :: pthpzDesSpecOANameOff = 45
INTEGER,PARAMETER :: pthpzDesSpecOADistNameOff = 46
INTEGER,PARAMETER :: pthpzBaseboardTypeOff = 47
INTEGER,PARAMETER :: pthpzBaseboardAvailSchedNameOff = 48
INTEGER,PARAMETER :: pthpzBaseboardCapOff = 49
INTEGER,PARAMETER :: pthpzCapCtrlTypeOff = 50

INTEGER,PARAMETER :: vrfzNameOff = 1
INTEGER,PARAMETER :: vrfzSystemNameOff = 2
INTEGER,PARAMETER :: vrfzTStatNameOff = 3
INTEGER,PARAMETER :: vrfzZoneHeatSizingFactorOff = 4
INTEGER,PARAMETER :: vrfzZoneCoolSizingFactorOff = 5
INTEGER,PARAMETER :: vrfzHeatCapSizRatOff = 6
INTEGER,PARAMETER :: vrfzSupplyCoolFlowRateOff = 7
INTEGER,PARAMETER :: vrfzSupplyNoCoolFlowRateOff = 8
INTEGER,PARAMETER :: vrfzSupplyHeatFlowRateOff = 9
INTEGER,PARAMETER :: vrfzSupplyNoHeatFlowRateOff = 10
INTEGER,PARAMETER :: vrfzOACoolFlowRateOff = 11
INTEGER,PARAMETER :: vrfzOAHeatFlowRateOff = 12
INTEGER,PARAMETER :: vrfzOANoLoadFlowRateOff = 13
INTEGER,PARAMETER :: vrfzOutAirTypeOff = 14
INTEGER,PARAMETER :: vrfzOutAirFlowPersonOff = 15
INTEGER,PARAMETER :: vrfzOutAirFlowAreaOff = 16
INTEGER,PARAMETER :: vrfzOutAirFlowZoneOff = 17
INTEGER,PARAMETER :: vrfzDesSpecOANameSizOff = 18
INTEGER,PARAMETER :: vrfzDesSpecOADistNameOff = 19
INTEGER,PARAMETER :: vrfzSysAvailSchedNameOff = 20
INTEGER,PARAMETER :: vrfzFanModeOff = 21
INTEGER,PARAMETER :: vrfzFanPlacementOff = 22
INTEGER,PARAMETER :: vrfzFanTotEfficiencyOff = 23
INTEGER,PARAMETER :: vrfzFanPressureOff = 24
INTEGER,PARAMETER :: vrfzFanMotorEfficiencyOff = 25
INTEGER,PARAMETER :: vrfzCoolCoilKindOff = 26
INTEGER,PARAMETER :: vrfzCoolAvailSchedNameOff = 27
INTEGER,PARAMETER :: vrfzCoolCoilCapOff = 28
INTEGER,PARAMETER :: vrfzCoolCoilSHROff = 29
INTEGER,PARAMETER :: vrfzHPHeatCoilKindOff = 30
INTEGER,PARAMETER :: vrfzHeatAvailSchedNameOff = 31
INTEGER,PARAMETER :: vrfzHeatCoilCapOff = 32
INTEGER,PARAMETER :: vrfzTermUnitOnParasiticElecOff = 33
INTEGER,PARAMETER :: vrfzTermUnitOffParasiticElecOff = 34
INTEGER,PARAMETER :: vrfzDedOutAirSysNameOff = 35
INTEGER,PARAMETER :: vrfzCoolDesignTempMethodOff = 36
INTEGER,PARAMETER :: vrfzCoolDesignTempOff = 37
INTEGER,PARAMETER :: vrfzCoolDesignTempDiffOff = 38
INTEGER,PARAMETER :: vrfzHeatDesignTempMethodOff = 39
INTEGER,PARAMETER :: vrfzHeatDesignTempOff = 40
INTEGER,PARAMETER :: vrfzHeatDesignTempDiffOff = 41
INTEGER,PARAMETER :: vrfzBaseboardTypeOff = 42
INTEGER,PARAMETER :: vrfzBaseboardAvailSchedNameOff = 43
INTEGER,PARAMETER :: vrfzBaseboardCapOff = 44

INTEGER,PARAMETER :: vrfsNameOff = 1
INTEGER,PARAMETER :: vrfsSysAvailSchedNameOff = 2
INTEGER,PARAMETER :: vrfsCoolCoilCapOff = 3
INTEGER,PARAMETER :: vrfsCoolCoilCOPOff = 4
INTEGER,PARAMETER :: vrfsMinTempCoolingOff = 5
INTEGER,PARAMETER :: vrfsMaxTempCoolingOff = 6
INTEGER,PARAMETER :: vrfsHPHeatCoilCapOff = 7
INTEGER,PARAMETER :: vrfsHeatCapSizRatOff = 8
INTEGER,PARAMETER :: vrfsHPHeatCoilCOPOff = 9
INTEGER,PARAMETER :: vrfsHPHeatMinODBOff = 10
INTEGER,PARAMETER :: vrfsHPHeatMaxODBOff = 11
INTEGER,PARAMETER :: vrfsHPMinPLROff = 12
INTEGER,PARAMETER :: vrfsTstatZoneNameOff = 13
INTEGER,PARAMETER :: vrfsTstatPriorityCtrlTypeOff = 14
INTEGER,PARAMETER :: vrfsTstatPrioritySchedNameOff = 15
INTEGER,PARAMETER :: vrfsHPWasteHeatRecoveryOff = 16
INTEGER,PARAMETER :: vrfsEquivPipLenCoolOff = 17
INTEGER,PARAMETER :: vrfsEquivPipHeightOff = 18
INTEGER,PARAMETER :: vrfsEquivPipLenHeatOff = 19
INTEGER,PARAMETER :: vrfsCrankHtPwrOff = 20
INTEGER,PARAMETER :: vrfsNumComprOff = 21
INTEGER,PARAMETER :: vrfsComprSizeRatioOff = 22
INTEGER,PARAMETER :: vrfsCrankHtMaxODBOff = 23
INTEGER,PARAMETER :: vrfsHPDefrTypeOff = 24
INTEGER,PARAMETER :: vrfsHPDefrCtrlTypeOff = 25
INTEGER,PARAMETER :: vrfsHPDefrTimeFracOff = 26
INTEGER,PARAMETER :: vrfsHPDefrHeatCapOff = 27
INTEGER,PARAMETER :: vrfsHPDefrMaxODBOff = 28
INTEGER,PARAMETER :: vrfsCondenserTypeOff = 29
INTEGER,PARAMETER :: vrfsWaterCondVolFlowRateOff = 30
INTEGER,PARAMETER :: vrfsEvapCondEffectOff = 31
INTEGER,PARAMETER :: vrfsEvapCondAirFlowRateOff = 32
INTEGER,PARAMETER :: vrfsEvapCondPumpPwrOff = 33
INTEGER,PARAMETER :: vrfsBasinHeatCapOff = 34
INTEGER,PARAMETER :: vrfsBasinHeatSetPtOff = 35
INTEGER,PARAMETER :: vrfsBasinHeatOperSchedNameOff = 36
INTEGER,PARAMETER :: vrfsFuelTypeOff = 37
INTEGER,PARAMETER :: vrfsHeatRecMinODBOff = 38
INTEGER,PARAMETER :: vrfsHeatRecMaxODBOff = 39

INTEGER,PARAMETER :: pvavsAirHandlerNameOff = 1
INTEGER,PARAMETER :: pvavsSysAvailSchedNameOff = 2
INTEGER,PARAMETER :: pvavsSupplyMaxRateOff = 3
INTEGER,PARAMETER :: pvavsSupplyMinRateOff = 4
INTEGER,PARAMETER :: pvavsSupplyFanPlaceOff = 5
INTEGER,PARAMETER :: pvavsSupplyEfficiencyOff = 6
INTEGER,PARAMETER :: pvavsSupplyPressureOff = 7
INTEGER,PARAMETER :: pvavsSupplyFanMotorEffOff = 8
INTEGER,PARAMETER :: pvavsSupplyFanFractionOff = 9
INTEGER,PARAMETER :: pvavsCoolCoilKindOff = 10
INTEGER,PARAMETER :: pvavsCoolAvailSchedNameOff = 11
INTEGER,PARAMETER :: pvavsCoolSetPtSchedNameOff = 12
INTEGER,PARAMETER :: pvavsCoolSetPtDesignOff = 13
INTEGER,PARAMETER :: pvavsCoolCoilCapOff = 14
INTEGER,PARAMETER :: pvavsCoolCoilSHROff = 15
INTEGER,PARAMETER :: pvavsCoolCoilCOPOff = 16
INTEGER,PARAMETER :: pvavsHeatCoilKindOff = 17
INTEGER,PARAMETER :: pvavsHeatAvailSchedNameOff = 18
INTEGER,PARAMETER :: pvavsHeatSetPtSchedNameOff = 19
INTEGER,PARAMETER :: pvavsHeatSetPtDesignOff = 20
INTEGER,PARAMETER :: pvavsHeatCoilCapOff = 21
INTEGER,PARAMETER :: pvavsHeatEffiencyOff = 22
INTEGER,PARAMETER :: pvavsHeatParasiticOff = 23
INTEGER,PARAMETER :: pvavsMaxOutsideFlowOff = 24
INTEGER,PARAMETER :: pvavsMinOutsideFlowOff = 25
INTEGER,PARAMETER :: pvavsMinOutsideControlOff = 26
INTEGER,PARAMETER :: pvavsMinOutFlowSchedNameOff = 27
INTEGER,PARAMETER :: pvavsEconomizerKindOff = 28
INTEGER,PARAMETER :: pvavsEconoLockoutKindOff = 29
INTEGER,PARAMETER :: pvavsEconoUpTempLimitOff = 30
INTEGER,PARAMETER :: pvavsEconoUpEnthLimitOff = 31
INTEGER,PARAMETER :: pvavsEconoUpDewLimitOff = 32
INTEGER,PARAMETER :: pvavsEconoLowTempLimitOff = 33
INTEGER,PARAMETER :: pvavsSupplyPlenumNameOff = 34
INTEGER,PARAMETER :: pvavsReturnPlenumNameOff = 35
INTEGER,PARAMETER :: pvavsSupplyFanPLRCoeffOff = 36
INTEGER,PARAMETER :: pvavsNightCycleControlOff = 37
INTEGER,PARAMETER :: pvavsNightCycleCtrlZoneNameOff = 38
INTEGER,PARAMETER :: pvavsHeatRecTypeOff = 39
INTEGER,PARAMETER :: pvavsHeatRecSenEffOff = 40
INTEGER,PARAMETER :: pvavsHeatRecLatEffOff = 41
INTEGER,PARAMETER :: pvavsCoolSetPtResetTypeOff = 42
INTEGER,PARAMETER :: pvavsHeatSetPtResetTypeOff = 43
INTEGER,PARAMETER :: pvavsDehumCtrlTypeOff = 44
INTEGER,PARAMETER :: pvavsDehumCtrlZoneOff = 45
INTEGER,PARAMETER :: pvavsDehumSetPtOff = 46
INTEGER,PARAMETER :: pvavsHumidCtrlTypeOff = 47
INTEGER,PARAMETER :: pvavsHumidAvailSchedNameOff = 48
INTEGER,PARAMETER :: pvavsHumidCapOff = 49
INTEGER,PARAMETER :: pvavsHumidRatedElecOff = 50
INTEGER,PARAMETER :: pvavsHumidCtrlZoneOff = 51
INTEGER,PARAMETER :: pvavsHumidSetPtOff = 52
INTEGER,PARAMETER :: pvavsSizingOptionOff = 53
INTEGER,PARAMETER :: pvavsReturnFanPresentOff = 54
INTEGER,PARAMETER :: pvavsReturnEfficiencyOff = 55
INTEGER,PARAMETER :: pvavsReturnPressureOff = 56
INTEGER,PARAMETER :: pvavsReturnFanMotorEffOff = 57
INTEGER,PARAMETER :: pvavsReturnFanFractionOff = 58
INTEGER,PARAMETER :: pvavsReturnFanPLRCoeffOff = 59

INTEGER,PARAMETER :: uhpsAirHandlerNameOff = 1
INTEGER,PARAMETER :: uhpsSysAvailSchedNameOff = 2
INTEGER,PARAMETER :: uhpsControlZoneOff = 3
INTEGER,PARAMETER :: uhpsSupplyCoolFlowRateOff = 4
INTEGER,PARAMETER :: uhpsSupplyHeatFlowRateOff = 5
INTEGER,PARAMETER :: uhpsSupplyNoLoadFlowRateOff = 6
INTEGER,PARAMETER :: uhpsSupplyFanOpModeSchedNameOff = 7
INTEGER,PARAMETER :: uhpsSupplyFanPlaceOff = 8
INTEGER,PARAMETER :: uhpsSupplyEfficiencyOff = 9
INTEGER,PARAMETER :: uhpsSupplyPressureOff = 10
INTEGER,PARAMETER :: uhpsSupplyFanMotorEffOff = 11
INTEGER,PARAMETER :: uhpsSupplyFanFractionOff = 12
INTEGER,PARAMETER :: uhpsCoolCoilKindOff = 13
INTEGER,PARAMETER :: uhpsCoolAvailSchedNameOff = 14
INTEGER,PARAMETER :: uhpsCoolDesignSupplyTempOff = 15
INTEGER,PARAMETER :: uhpsCoolCoilCapOff = 16
INTEGER,PARAMETER :: uhpsCoolCoilSHROff = 17
INTEGER,PARAMETER :: uhpsCoolCoilCOPOff = 18
INTEGER,PARAMETER :: uhpsHPHeatCoilKindOff = 19
INTEGER,PARAMETER :: uhpsHeatAvailSchedNameOff = 20
INTEGER,PARAMETER :: uhpsHeatDesignSupplyTempOff = 21
INTEGER,PARAMETER :: uhpsHPHeatCoilCapOff = 22
INTEGER,PARAMETER :: uhpsHPHeatCoilCOPOff = 23
INTEGER,PARAMETER :: uhpsHPHeatMinODBOff = 24
INTEGER,PARAMETER :: uhpsHPDefrMaxODBOff = 25
INTEGER,PARAMETER :: uhpsHPDefrStrategyOff = 26
INTEGER,PARAMETER :: uhpsHPDefrCtrlTypeOff = 27
INTEGER,PARAMETER :: uhpsHPDefrTimeFracOff = 28
INTEGER,PARAMETER :: uhpsSuppHeatCoilKindOff = 29
INTEGER,PARAMETER :: uhpsSuppHeatAvailSchedNameOff = 30
INTEGER,PARAMETER :: uhpsSuppHeatCoilCapOff = 31
INTEGER,PARAMETER :: uhpsSuppHeatMaxODBOff = 32
INTEGER,PARAMETER :: uhpsSuppHeatEfficiencyOff = 33
INTEGER,PARAMETER :: uhpsSuppHeatParasiticOff = 34
INTEGER,PARAMETER :: uhpsMaxOutsideFlowOff = 35
INTEGER,PARAMETER :: uhpsMinOutsideFlowOff = 36
INTEGER,PARAMETER :: uhpsMinOutFlowSchedNameOff = 37
INTEGER,PARAMETER :: uhpsEconomizerKindOff = 38
INTEGER,PARAMETER :: uhpsEconoLockoutKindOff = 39
INTEGER,PARAMETER :: uhpsEconoUpTempLimitOff = 40
INTEGER,PARAMETER :: uhpsEconoUpEnthLimitOff = 41
INTEGER,PARAMETER :: uhpsEconoUpDewLimitOff = 42
INTEGER,PARAMETER :: uhpsEconoLowTempLimitOff = 43
INTEGER,PARAMETER :: uhpsSupplyPlenumNameOff = 44
INTEGER,PARAMETER :: uhpsReturnPlenumNameOff = 45
INTEGER,PARAMETER :: uhpsNightCycleControlOff = 46
INTEGER,PARAMETER :: uhpsNightCycleCtrlZoneNameOff = 47
INTEGER,PARAMETER :: uhpsHeatRecTypeOff = 48
INTEGER,PARAMETER :: uhpsHeatRecSenEffOff = 49
INTEGER,PARAMETER :: uhpsHeatRecLatEffOff = 50
INTEGER,PARAMETER :: uhpsHumidCtrlTypeOff = 51
INTEGER,PARAMETER :: uhpsHumidAvailSchedNameOff = 52
INTEGER,PARAMETER :: uhpsHumidCapOff = 53
INTEGER,PARAMETER :: uhpsHumidRatedElecOff = 54
INTEGER,PARAMETER :: uhpsHumidCtrlZoneOff = 55
INTEGER,PARAMETER :: uhpsHumidSetPtOff = 56
INTEGER,PARAMETER :: uhpsReturnFanPresentOff = 57
INTEGER,PARAMETER :: uhpsReturnEfficiencyOff = 58
INTEGER,PARAMETER :: uhpsReturnPressureOff = 59
INTEGER,PARAMETER :: uhpsReturnFanMotorEffOff = 60
INTEGER,PARAMETER :: uhpsReturnFanFractionOff = 61

INTEGER,PARAMETER :: ussAirHandlerNameOff = 1
INTEGER,PARAMETER :: ussSysAvailSchedNameOff = 2
INTEGER,PARAMETER :: ussSystemControlTypeOff = 3
INTEGER,PARAMETER :: ussControlZoneOff = 4
INTEGER,PARAMETER :: ussSupplyCoolFlowRateOff = 5
INTEGER,PARAMETER :: ussSupplyHeatFlowRateOff = 6
INTEGER,PARAMETER :: ussSupplyNoLoadFlowRateOff = 7
INTEGER,PARAMETER :: ussFanModeSchedOff = 8
INTEGER,PARAMETER :: ussSupplyFanPlaceOff = 9
INTEGER,PARAMETER :: ussSupplyEfficiencyOff = 10
INTEGER,PARAMETER :: ussSupplyPressureOff = 11
INTEGER,PARAMETER :: ussSupplyFanMotorEffOff = 12
INTEGER,PARAMETER :: ussSupplyFanFractionOff = 13
INTEGER,PARAMETER :: ussCoolCoilKindOff = 14
INTEGER,PARAMETER :: ussCoolCoilNumSpeedOff = 15
INTEGER,PARAMETER :: ussCoolAvailSchedNameOff = 16
INTEGER,PARAMETER :: ussCoolDesignSupplyTempOff = 17
INTEGER,PARAMETER :: ussCoolCoilCapOff = 18
INTEGER,PARAMETER :: ussCoolCoilSHROff = 19
INTEGER,PARAMETER :: ussCoolCoilCOPOff = 20
INTEGER,PARAMETER :: ussHeatCoilKindOff = 21
INTEGER,PARAMETER :: ussHeatCoilNumSpeedOff = 22
INTEGER,PARAMETER :: ussHeatAvailSchedNameOff = 23
INTEGER,PARAMETER :: ussHeatDesignSupplyTempOff = 24
INTEGER,PARAMETER :: ussHeatCoilCapOff = 25
INTEGER,PARAMETER :: ussHeatEfficiencyOff = 26
INTEGER,PARAMETER :: ussHeatParasiticOff = 27
INTEGER,PARAMETER :: ussHPHeatCoilCOPOff = 28
INTEGER,PARAMETER :: ussHPHeatMinODBOff = 29
INTEGER,PARAMETER :: ussHPDefrMaxODBOff = 30
INTEGER,PARAMETER :: ussHPDefrStrategyOff = 31
INTEGER,PARAMETER :: ussHPDefrCtrlTypeOff = 32
INTEGER,PARAMETER :: ussHPDefrTimeFracOff = 33
INTEGER,PARAMETER :: ussSuppReHeatCoilKindOff = 34
INTEGER,PARAMETER :: ussSuppReHeatAvailSchedNameOff = 35
INTEGER,PARAMETER :: ussSuppReHeatCoilCapOff = 36
INTEGER,PARAMETER :: ussSuppReHeatMaxODBOff = 37
INTEGER,PARAMETER :: ussSuppReHeatEfficiencyOff = 38
INTEGER,PARAMETER :: ussSuppReHeatParasiticOff = 39
INTEGER,PARAMETER :: ussMaxOutsideFlowOff = 40
INTEGER,PARAMETER :: ussMinOutsideFlowOff = 41
INTEGER,PARAMETER :: ussMinOutFlowSchedNameOff = 42
INTEGER,PARAMETER :: ussEconomizerKindOff = 43
INTEGER,PARAMETER :: ussEconoLockoutKindOff = 44
INTEGER,PARAMETER :: ussEconoUpTempLimitOff = 45
INTEGER,PARAMETER :: ussEconoUpEnthLimitOff = 46
INTEGER,PARAMETER :: ussEconoUpDewLimitOff = 47
INTEGER,PARAMETER :: ussEconoLowTempLimitOff = 48
INTEGER,PARAMETER :: ussSupplyPlenumNameOff = 49
INTEGER,PARAMETER :: ussReturnPlenumNameOff = 50
INTEGER,PARAMETER :: ussHeatRecTypeOff = 51
INTEGER,PARAMETER :: ussHeatRecSenEffOff = 52
INTEGER,PARAMETER :: ussHeatRecLatEffOff = 53
INTEGER,PARAMETER :: ussHXTypeOff = 54
INTEGER,PARAMETER :: ussFrostControlOff = 55
INTEGER,PARAMETER :: ussDehumCtrlTypeOff = 56
INTEGER,PARAMETER :: ussDehumSetPtOff = 57
INTEGER,PARAMETER :: ussDehumSetPtSchedNameOff = 58
INTEGER,PARAMETER :: ussHumidCtrlTypeOff = 59
INTEGER,PARAMETER :: ussHumidAvailSchedNameOff = 60
INTEGER,PARAMETER :: ussHumidCapOff = 61
INTEGER,PARAMETER :: ussHumidRatedElecOff = 62
INTEGER,PARAMETER :: ussHumidCtrlZoneOff = 63
INTEGER,PARAMETER :: ussHumidSetPtOff = 64
INTEGER,PARAMETER :: ussHumidSetPtSchedNameOff = 65
INTEGER,PARAMETER :: ussSizingOptionOff = 66
INTEGER,PARAMETER :: ussReturnFanPresentOff = 67
INTEGER,PARAMETER :: ussReturnEfficiencyOff = 68
INTEGER,PARAMETER :: ussReturnPressureOff = 69
INTEGER,PARAMETER :: ussReturnFanMotorEffOff = 70
INTEGER,PARAMETER :: ussReturnFanFractionOff = 71

INTEGER,PARAMETER :: ddzNameOff = 1
INTEGER,PARAMETER :: ddzAirHandlerNameOff = 2
INTEGER,PARAMETER :: ddzTStatNameOff = 3
INTEGER,PARAMETER :: ddzSupplyMaxRateOff = 4
INTEGER,PARAMETER :: ddzZoneHeatSizingFactorOff = 5
INTEGER,PARAMETER :: ddzZoneCoolSizingFactorOff = 6
INTEGER,PARAMETER :: ddzConstantMinFracOff = 7
INTEGER,PARAMETER :: ddzOutAirTypeOff = 8
INTEGER,PARAMETER :: ddzOutAirFlowPersonOff = 9
INTEGER,PARAMETER :: ddzOutAirFlowAreaOff = 10
INTEGER,PARAMETER :: ddzOutAirFlowZoneOff = 11
INTEGER,PARAMETER :: ddzDesSpecOANameSizOff = 12
INTEGER,PARAMETER :: ddzDesSpecOADistNameOff = 13
INTEGER,PARAMETER :: ddzDesSpecOANameCtrlOff = 14
INTEGER,PARAMETER :: ddzZoneColdSupplyPlenumNameOff = 15
INTEGER,PARAMETER :: ddzZoneHotSupplyPlenumNameOff = 16
INTEGER,PARAMETER :: ddzZoneReturnPlenumNameOff = 17
INTEGER,PARAMETER :: ddzBaseboardTypeOff = 18
INTEGER,PARAMETER :: ddzBaseboardAvailSchedNameOff = 19
INTEGER,PARAMETER :: ddzBaseboardCapOff = 20
INTEGER,PARAMETER :: ddzCoolDesignTempMethodOff = 21
INTEGER,PARAMETER :: ddzCoolDesignTempOff = 22
INTEGER,PARAMETER :: ddzCoolDesignTempDiffOff = 23
INTEGER,PARAMETER :: ddzHeatDesignTempMethodOff = 24
INTEGER,PARAMETER :: ddzHeatDesignTempOff = 25
INTEGER,PARAMETER :: ddzHeatDesignTempDiffOff = 26

INTEGER,PARAMETER :: ddsAirHandlerNameOff = 1
INTEGER,PARAMETER :: ddsSysAvailSchedNameOff = 2
INTEGER,PARAMETER :: ddsSupplyFanConfigTypeOff = 3
INTEGER,PARAMETER :: ddsSupplyMaxRateOff = 4
INTEGER,PARAMETER :: ddsSupplyMinFracOff = 5
INTEGER,PARAMETER :: ddsSupplyEfficiencyOff = 6
INTEGER,PARAMETER :: ddsSupplyPressureOff = 7
INTEGER,PARAMETER :: ddsSupplyFanMotorEffOff = 8
INTEGER,PARAMETER :: ddsSupplyFanFractionOff = 9
INTEGER,PARAMETER :: ddsSupplyFanPLRCoeffOff = 10
INTEGER,PARAMETER :: ddsColdMaxRateOff = 11
INTEGER,PARAMETER :: ddsColdMinFracOff = 12
INTEGER,PARAMETER :: ddsColdEfficiencyOff = 13
INTEGER,PARAMETER :: ddsColdPressureOff = 14
INTEGER,PARAMETER :: ddsColdFanMotorEffOff = 15
INTEGER,PARAMETER :: ddsColdFanFractionOff = 16
INTEGER,PARAMETER :: ddsColdFanPLRCoeffOff = 17
INTEGER,PARAMETER :: ddsColdFanPlaceOff = 18
INTEGER,PARAMETER :: ddsHotMaxRateOff = 19
INTEGER,PARAMETER :: ddsHotMinFracOff = 20
INTEGER,PARAMETER :: ddsHotEfficiencyOff = 21
INTEGER,PARAMETER :: ddsHotPressureOff = 22
INTEGER,PARAMETER :: ddsHotFanMotorEffOff = 23
INTEGER,PARAMETER :: ddsHotFanFractionOff = 24
INTEGER,PARAMETER :: ddsHotFanPLRCoeffOff = 25
INTEGER,PARAMETER :: ddsHotFanPlaceOff = 26
INTEGER,PARAMETER :: ddsCoolCoilKindOff = 27
INTEGER,PARAMETER :: ddsCoolAvailSchedNameOff = 28
INTEGER,PARAMETER :: ddsCoolSetPtCtrlTypeOff = 29
INTEGER,PARAMETER :: ddsCoolSetPtDesignOff = 30
INTEGER,PARAMETER :: ddsCoolSetPtSchedNameOff = 31
INTEGER,PARAMETER :: ddsCoolSetPtAtODBLoOff = 32
INTEGER,PARAMETER :: ddsCoolResetODBLoOff = 33
INTEGER,PARAMETER :: ddsCoolSetPtAtODBHiOff = 34
INTEGER,PARAMETER :: ddsCoolResetODBHiOff = 35
INTEGER,PARAMETER :: ddsHeatCoilKindOff = 36
INTEGER,PARAMETER :: ddsHeatAvailSchedNameOff = 37
INTEGER,PARAMETER :: ddsHeatSetPtCtrlTypeOff = 38
INTEGER,PARAMETER :: ddsHeatSetPtDesignOff = 39
INTEGER,PARAMETER :: ddsHeatSetPtSchedNameOff = 40
INTEGER,PARAMETER :: ddsHeatSetPtAtODBLoOff = 41
INTEGER,PARAMETER :: ddsHeatResetODBLoOff = 42
INTEGER,PARAMETER :: ddsHeatSetPtAtODBHiOff = 43
INTEGER,PARAMETER :: ddsHeatResetODBHiOff = 44
INTEGER,PARAMETER :: ddsHeatCoilCapOff = 45
INTEGER,PARAMETER :: ddsHeatEffiencyOff = 46
INTEGER,PARAMETER :: ddsHeatParasiticOff = 47
INTEGER,PARAMETER :: ddsPreheatCoilKindOff = 48
INTEGER,PARAMETER :: ddsPreheatAvailSchedNameOff = 49
INTEGER,PARAMETER :: ddsPreheatSetPtConstantOff = 50
INTEGER,PARAMETER :: ddsPreheatSetPtSchedNameOff = 51
INTEGER,PARAMETER :: ddsPreheatEffiencyOff = 52
INTEGER,PARAMETER :: ddsPreheatParasiticOff = 53
INTEGER,PARAMETER :: ddsMaxOutsideFlowOff = 54
INTEGER,PARAMETER :: ddsMinOutsideFlowOff = 55
INTEGER,PARAMETER :: ddsMinOutsideControlOff = 56
INTEGER,PARAMETER :: ddsMinOutFlowSchedNameOff = 57
INTEGER,PARAMETER :: ddsEconomizerKindOff = 58
INTEGER,PARAMETER :: ddsEconoLockoutKindOff = 59
INTEGER,PARAMETER :: ddsEconoUpLimitOff = 60
INTEGER,PARAMETER :: ddsEconoLowLimitOff = 61
INTEGER,PARAMETER :: ddsEconoEnthLimitOff = 62
INTEGER,PARAMETER :: ddsEconoUpDewLimitOff = 63
INTEGER,PARAMETER :: ddsColdSupplyPlenumNameOff = 64
INTEGER,PARAMETER :: ddsHotSupplyPlenumNameOff = 65
INTEGER,PARAMETER :: ddsReturnPlenumNameOff = 66
INTEGER,PARAMETER :: ddsNightCycleControlOff = 67
INTEGER,PARAMETER :: ddsNightCycleCtrlZoneNameOff = 68
INTEGER,PARAMETER :: ddsHeatRecTypeOff = 69
INTEGER,PARAMETER :: ddsHeatRecSenEffOff = 70
INTEGER,PARAMETER :: ddsHeatRecLatEffOff = 71
INTEGER,PARAMETER :: ddsHXTypeOff = 72
INTEGER,PARAMETER :: ddsFrostControlOff = 73
INTEGER,PARAMETER :: ddsDehumCtrlTypeOff = 74
INTEGER,PARAMETER :: ddsDehumCtrlZoneOff = 75
INTEGER,PARAMETER :: ddsDehumSetPtOff = 76
INTEGER,PARAMETER :: ddsDehumSetPtSchedNameOff = 77
INTEGER,PARAMETER :: ddsHumidCtrlTypeOff = 78
INTEGER,PARAMETER :: ddsHumidAvailSchedNameOff = 79
INTEGER,PARAMETER :: ddsHumidCapOff = 80
INTEGER,PARAMETER :: ddsHumidRatedElecOff = 81
INTEGER,PARAMETER :: ddsHumidCtrlZoneOff = 82
INTEGER,PARAMETER :: ddsHumidSetPtOff = 83
INTEGER,PARAMETER :: ddsHumidSetPtSchedNameOff = 84
INTEGER,PARAMETER :: ddsSizingOptionOff = 85
INTEGER,PARAMETER :: ddsReturnFanPresentOff = 86
INTEGER,PARAMETER :: ddsReturnEfficiencyOff = 87
INTEGER,PARAMETER :: ddsReturnPressureOff = 88
INTEGER,PARAMETER :: ddsReturnFanMotorEffOff = 89
INTEGER,PARAMETER :: ddsReturnFanFractionOff = 90
INTEGER,PARAMETER :: ddsReturnFanPLRCoeffOff = 91

INTEGER,PARAMETER :: cwpNameOff = 1
INTEGER,PARAMETER :: cwpPumpSchedNameOff = 2
INTEGER,PARAMETER :: cwpPumpControlTypeOff = 3
INTEGER,PARAMETER :: cwpChWOperTypeOff = 4
INTEGER,PARAMETER :: cwpChWOperNameOff = 5
INTEGER,PARAMETER :: cwpChWSetPtSchedNameOff = 6
INTEGER,PARAMETER :: cwpChWSetPtDesignOff = 7
INTEGER,PARAMETER :: cwpChWPumpConfigOff = 8
INTEGER,PARAMETER :: cwpPriChWPumpHeadOff = 9
INTEGER,PARAMETER :: cwpSecChWPumpHeadOff = 10
INTEGER,PARAMETER :: cwpCndOperTypeOff = 11
INTEGER,PARAMETER :: cwpCndOperNameOff = 12
INTEGER,PARAMETER :: cwpCndTempControlTypeOff = 13
INTEGER,PARAMETER :: cwpCndSetPtSchedNameOff = 14
INTEGER,PARAMETER :: cwpCndSetPtDesignOff = 15
INTEGER,PARAMETER :: cwpCndPumpHeadOff = 16
INTEGER,PARAMETER :: cwpChWSetPtResetTypeOff = 17
INTEGER,PARAMETER :: cwpChWSetPtAtODBLoOff = 18
INTEGER,PARAMETER :: cwpChWODBLoOff = 19
INTEGER,PARAMETER :: cwpChWSetPtAtODBHiOff = 20
INTEGER,PARAMETER :: cwpChWODBHiOff = 21
INTEGER,PARAMETER :: cwpPriChWPumpTypeOff = 22
INTEGER,PARAMETER :: cwpSecChWPumpTypeOff = 23
INTEGER,PARAMETER :: cwpCndPumpTypeOff = 24
INTEGER,PARAMETER :: cwpChWSupplyBypassOff = 25
INTEGER,PARAMETER :: cwpChwDemandBypassOff = 26
INTEGER,PARAMETER :: cwpCndSupplyBypassOff = 27
INTEGER,PARAMETER :: cwpCndDemandBypassOff = 28
INTEGER,PARAMETER :: cwpFluidTypeOff = 29
INTEGER,PARAMETER :: cwpDesDeltaTOff = 30
INTEGER,PARAMETER :: cwpMinTOutOff = 31
INTEGER,PARAMETER :: cwpChwLdDistSchmOff = 32
INTEGER,PARAMETER :: cwpCndLdDistSchmOff = 33

INTEGER,PARAMETER :: chlNameOff = 1
INTEGER,PARAMETER :: chlTypeOff = 2
INTEGER,PARAMETER :: chlCapacityOff = 3
INTEGER,PARAMETER :: chlCOPOff = 4
INTEGER,PARAMETER :: chlCondTypeOff = 5
INTEGER,PARAMETER :: chlPriorityOff = 6
INTEGER,PARAMETER :: chlSizeFactorOff = 7
INTEGER,PARAMETER :: chlMinPLROff = 8
INTEGER,PARAMETER :: chlMaxPLROff = 9
INTEGER,PARAMETER :: chlOptPLROff = 10
INTEGER,PARAMETER :: chlMinUnloadRatOff = 11
INTEGER,PARAMETER :: chlLeavingLowLimOff = 12

INTEGER,PARAMETER :: chlorNameOff = 1
INTEGER,PARAMETER :: chlorObjTypeOff = 2
INTEGER,PARAMETER :: chlorObjNameOff = 3
INTEGER,PARAMETER :: chlorPriorityOff = 4

INTEGER,PARAMETER :: twrNameOff = 1
INTEGER,PARAMETER :: twrTypeOff = 2
INTEGER,PARAMETER :: twrHiSpdCapacityOff = 3
INTEGER,PARAMETER :: twrHiSpdFanPowerOff = 4
INTEGER,PARAMETER :: twrLoSpdCapacityOff = 5
INTEGER,PARAMETER :: twrLoSpdFanPowerOff = 6
INTEGER,PARAMETER :: twrFreeConvCapacityOff = 7
INTEGER,PARAMETER :: twrPriorityOff = 8
INTEGER,PARAMETER :: twrSizeFactorOff = 9
INTEGER,PARAMETER :: twrLoopTypeOff = 10

INTEGER,PARAMETER :: twrorNameOff = 1
INTEGER,PARAMETER :: twrorObjTypeOff = 2
INTEGER,PARAMETER :: twrorObjNameOff = 3
INTEGER,PARAMETER :: twrorPriorityOff = 4
INTEGER,PARAMETER :: twrorLoopTypeOff = 5

INTEGER,PARAMETER :: hwpNameOff = 1
INTEGER,PARAMETER :: hwpPumpSchedNameOff = 2
INTEGER,PARAMETER :: hwpPumpControlTypeOff = 3
INTEGER,PARAMETER :: hwpHWOperTypeOff = 4
INTEGER,PARAMETER :: hwpHWOperNameOff = 5
INTEGER,PARAMETER :: hwpHWSetPtSchedNameOff = 6
INTEGER,PARAMETER :: hwpHWSetPtDesignOff = 7
INTEGER,PARAMETER :: hwpHWPumpConfigOff = 8
INTEGER,PARAMETER :: hwpHWPumpHeadOff = 9
INTEGER,PARAMETER :: hwpHWSetPtResetTypeOff = 10
INTEGER,PARAMETER :: hwpHWSetPtAtODBLoOff = 11
INTEGER,PARAMETER :: hwpHWODBLoOff = 12
INTEGER,PARAMETER :: hwpHWSetPtAtODBHiOff = 13
INTEGER,PARAMETER :: hwpHWODBHiOff = 14
INTEGER,PARAMETER :: hwpHWPumpTypeOff = 15
INTEGER,PARAMETER :: hwpSupplyBypassOff = 16
INTEGER,PARAMETER :: hwpDemandBypassOff = 17
INTEGER,PARAMETER :: hwpFluidTypeOff = 18
INTEGER,PARAMETER :: hwpDesDeltaTOff = 19
INTEGER,PARAMETER :: hwpMaxTOutOff = 20
INTEGER,PARAMETER :: hwpLdDistSchmOff = 21

INTEGER,PARAMETER :: blrNameOff = 1
INTEGER,PARAMETER :: blrTypeOff = 2
INTEGER,PARAMETER :: blrCapacityOff = 3
INTEGER,PARAMETER :: blrEfficOff = 4
INTEGER,PARAMETER :: blrFuelTypeOff = 5
INTEGER,PARAMETER :: blrPriorityOff = 6
INTEGER,PARAMETER :: blrSizeFactorOff = 7
INTEGER,PARAMETER :: blrMinPLROff = 8
INTEGER,PARAMETER :: blrMaxPLROff = 9
INTEGER,PARAMETER :: blrOptPLROff = 10
INTEGER,PARAMETER :: blrOutletUpLimOff = 11
INTEGER,PARAMETER :: blrLoopTypeOff = 12

INTEGER,PARAMETER :: blrorNameOff = 1
INTEGER,PARAMETER :: blrorObjTypeOff = 2
INTEGER,PARAMETER :: blrorObjNameOff = 3
INTEGER,PARAMETER :: blrorPriorityOff = 4
INTEGER,PARAMETER :: blrorLoopTypeOff = 5

INTEGER,PARAMETER :: wahpNameOff = 1
INTEGER,PARAMETER :: wahpTStatNameOff = 2
INTEGER,PARAMETER :: wahpSupplyCoolFlowRateOff = 3
INTEGER,PARAMETER :: wahpSupplyHeatFlowRateOff = 4
INTEGER,PARAMETER :: wahpSupplyNoLoadFlowRateOff = 5
INTEGER,PARAMETER :: wahpZoneHeatSizingFactorOff = 6
INTEGER,PARAMETER :: wahpZoneCoolSizingFactorOff = 7
INTEGER,PARAMETER :: wahpOutAirTypeOff = 8
INTEGER,PARAMETER :: wahpOutAirFlowPersonOff = 9
INTEGER,PARAMETER :: wahpOutAirFlowAreaOff = 10
INTEGER,PARAMETER :: wahpOutAirFlowZoneOff = 11
INTEGER,PARAMETER :: wahpSysAvailSchedNameOff = 12
INTEGER,PARAMETER :: wahpFanModeOff = 13
INTEGER,PARAMETER :: wahpFanPlacementOff = 14
INTEGER,PARAMETER :: wahpFanTotEfficiencyOff = 15
INTEGER,PARAMETER :: wahpFanPressureOff = 16
INTEGER,PARAMETER :: wahpFanMotorEfficiencyOff = 17
INTEGER,PARAMETER :: wahpCoolCoilKindOff = 18
INTEGER,PARAMETER :: wahpCoolCoilCapOff = 19
INTEGER,PARAMETER :: wahpCoolCoilSHROff = 20
INTEGER,PARAMETER :: wahpCoolCoilCOPOff = 21
INTEGER,PARAMETER :: wahpHPHeatCoilKindOff = 22
INTEGER,PARAMETER :: wahpHPHeatCoilCapOff = 23
INTEGER,PARAMETER :: wahpHPHeatCoilCOPOff = 24
INTEGER,PARAMETER :: wahpSuppHeatAvailSchedNameOff = 25
INTEGER,PARAMETER :: wahpSuppHeatCoilCapOff = 26
INTEGER,PARAMETER :: wahpMaxCycRateOff = 27
INTEGER,PARAMETER :: wahpTimeConstOff = 28
INTEGER,PARAMETER :: wahpFractOnPowerOff = 29
INTEGER,PARAMETER :: wahpFanDelayOff = 30
INTEGER,PARAMETER :: wahpDedOutAirSysNameOff = 31
INTEGER,PARAMETER :: wahpSuppHeatCoilKindOff = 32
INTEGER,PARAMETER :: wahpCoolDesignTempMethodOff = 33
INTEGER,PARAMETER :: wahpCoolDesignTempOff = 34
INTEGER,PARAMETER :: wahpCoolDesignTempDiffOff = 35
INTEGER,PARAMETER :: wahpHeatDesignTempMethodOff = 36
INTEGER,PARAMETER :: wahpHeatDesignTempOff = 37
INTEGER,PARAMETER :: wahpHeatDesignTempDiffOff = 38
INTEGER,PARAMETER :: wahpHPCoilWaterFlowMode = 39
INTEGER,PARAMETER :: wahpDesSpecOANameOff = 40
INTEGER,PARAMETER :: wahpDesSpecOADistNameOff = 41
INTEGER,PARAMETER :: wahpBaseboardTypeOff = 42
INTEGER,PARAMETER :: wahpBaseboardAvailSchedNameOff = 43
INTEGER,PARAMETER :: wahpBaseboardCapOff = 44

INTEGER,PARAMETER :: mwpNameOff = 1
INTEGER,PARAMETER :: mwpPumpSchedNameOff = 2
INTEGER,PARAMETER :: mwpPumpControlTypeOff = 3
INTEGER,PARAMETER :: mwpOperTypeOff = 4
INTEGER,PARAMETER :: mwpOperNameOff = 5
INTEGER,PARAMETER :: mwpHiSetPtSchedNameOff = 6
INTEGER,PARAMETER :: mwpHiSetPtDesignOff = 7
INTEGER,PARAMETER :: mwpLoSetPtSchedNameOff = 8
INTEGER,PARAMETER :: mwpLoSetPtDesignOff = 9
INTEGER,PARAMETER :: mwpPumpConfigOff = 10
INTEGER,PARAMETER :: mwpPumpHeadOff = 11
INTEGER,PARAMETER :: mwpPumpTypeOff = 12
INTEGER,PARAMETER :: mwpSupplyBypassOff = 13
INTEGER,PARAMETER :: mwpDemandBypassOff = 14
INTEGER,PARAMETER :: mwpFluidTypeOff = 15
INTEGER,PARAMETER :: mwpDesDeltaTOff = 16
INTEGER,PARAMETER :: mwpLdDistSchmOff = 17

INTEGER,PARAMETER :: doasNameOff = 1
INTEGER,PARAMETER :: doasSysAvailSchedNameOff = 2
INTEGER,PARAMETER :: doasAirOutletTypeOff = 3
INTEGER,PARAMETER :: doasSupplyRateOff = 4
INTEGER,PARAMETER :: doasSupplyEfficiencyOff = 5
INTEGER,PARAMETER :: doasSupplyPressureOff = 6
INTEGER,PARAMETER :: doasSupplyFanMotorEffOff = 7
INTEGER,PARAMETER :: doasSupplyFanFractionOff = 8
INTEGER,PARAMETER :: doasSupplyFanPlaceOff = 9
INTEGER,PARAMETER :: doasCoolCoilKindOff = 10
INTEGER,PARAMETER :: doasCoolAvailSchedNameOff = 11
INTEGER,PARAMETER :: doasCoolSetPtCtrlTypeOff = 12
INTEGER,PARAMETER :: doasCoolSetPtDesignOff = 13
INTEGER,PARAMETER :: doasCoolSetPtSchedNameOff = 14
INTEGER,PARAMETER :: doasCoolSetPtAtODBLoOff = 15
INTEGER,PARAMETER :: doasCoolResetODBLoOff = 16
INTEGER,PARAMETER :: doasCoolSetPtAtODBHiOff = 17
INTEGER,PARAMETER :: doasCoolResetODBHiOff = 18
INTEGER,PARAMETER :: doasCoolCoilCapOff = 19
INTEGER,PARAMETER :: doasCoolCoilSHROff = 20
INTEGER,PARAMETER :: doasCoolCoilCOPOff = 21
INTEGER,PARAMETER :: doasHeatCoilKindOff = 22
INTEGER,PARAMETER :: doasHeatAvailSchedNameOff = 23
INTEGER,PARAMETER :: doasHeatSetPtCtrlTypeOff = 24
INTEGER,PARAMETER :: doasHeatSetPtDesignOff = 25
INTEGER,PARAMETER :: doasHeatSetPtSchedNameOff = 26
INTEGER,PARAMETER :: doasHeatSetPtAtODBLoOff = 27
INTEGER,PARAMETER :: doasHeatResetODBLoOff = 28
INTEGER,PARAMETER :: doasHeatSetPtAtODBHiOff = 29
INTEGER,PARAMETER :: doasHeatResetODBHiOff = 30
INTEGER,PARAMETER :: doasHeatEffiencyOff = 31
INTEGER,PARAMETER :: doasHeatParasiticOff = 32
INTEGER,PARAMETER :: doasHeatRecTypeOff = 33
INTEGER,PARAMETER :: doasHeatRecSenEffOff = 34
INTEGER,PARAMETER :: doasHeatRecLatEffOff = 35
INTEGER,PARAMETER :: doasHXTypeOff = 36
INTEGER,PARAMETER :: doasFrostControlOff = 37
INTEGER,PARAMETER :: doasDehumCtrlTypeOff = 38
INTEGER,PARAMETER :: doasDehumSetPtOff = 39
INTEGER,PARAMETER :: doasHumidCtrlTypeOff = 40
INTEGER,PARAMETER :: doasHumidAvailSchedNameOff = 41
INTEGER,PARAMETER :: doasHumidCapOff = 42
INTEGER,PARAMETER :: doasHumidRatedElecOff = 43
INTEGER,PARAMETER :: doasHumidSetPtOff = 44
INTEGER,PARAMETER :: doasDehumSetPtSchedNameOff = 45
INTEGER,PARAMETER :: doasHumidSetPtSchedNameOff = 46

INTEGER,PARAMETER :: ghtCtrlNameOff = 1
INTEGER,PARAMETER :: ghtCtrlBasementOff = 2
INTEGER,PARAMETER :: ghtCtrlSlabOff = 3

!only selected fields in the following objects are used
INTEGER,PARAMETER :: buildSurfNameOff = 1
INTEGER,PARAMETER :: buildSurfTypeOff = 2
INTEGER,PARAMETER :: buildSurfConstructionOff = 3
INTEGER,PARAMETER :: buildSurfZoneOff = 4
INTEGER,PARAMETER :: buildSurfOutBndCondOff = 5
INTEGER,PARAMETER :: buildSurfOutBndCondNameOff = 6
INTEGER,PARAMETER :: buildSurfSunExpOff = 7
INTEGER,PARAMETER :: buildSurfWindExpOff = 8 !last required field
INTEGER,PARAMETER :: buildSurfViewGroundOff = 9
INTEGER,PARAMETER :: buildSurfNumVertOff = 10
INTEGER,PARAMETER :: buildSurfX1Off = 11
INTEGER,PARAMETER :: buildSurfY1Off = 12
INTEGER,PARAMETER :: buildSurfZ1Off = 13
INTEGER,PARAMETER :: buildSurfLastZVertexOff = 370 !up to 120 vertices

INTEGER,PARAMETER :: wallDetlNameOff = 1
INTEGER,PARAMETER :: wallDetlConstructionOff = 2
INTEGER,PARAMETER :: wallDetlZoneOff = 3
INTEGER,PARAMETER :: wallDetlOutBndCondOff = 4
INTEGER,PARAMETER :: wallDetlOutBndCondNameOff = 5
INTEGER,PARAMETER :: wallDetlSunExpOff = 6
INTEGER,PARAMETER :: wallDetlWindExpOff = 7 !last required field
INTEGER,PARAMETER :: wallDetlViewGroundOff = 8
INTEGER,PARAMETER :: wallDetlNumVertOff = 9
INTEGER,PARAMETER :: wallDetlX1Off = 10
INTEGER,PARAMETER :: wallDetlY1Off = 11
INTEGER,PARAMETER :: wallDetlZ1Off = 12
INTEGER,PARAMETER :: wallDetlLastZVertexOff = 39 !up to 10 vertices

INTEGER,PARAMETER :: rfClDetlNameOff = 1
INTEGER,PARAMETER :: rfClDetlConstructionOff = 2
INTEGER,PARAMETER :: rfClDetlZoneOff = 3
INTEGER,PARAMETER :: rfClDetlOutBndCondOff = 4
INTEGER,PARAMETER :: rfClDetlOutBndCondNameOff = 5
INTEGER,PARAMETER :: rfClDetlSunExpOff = 6
INTEGER,PARAMETER :: rfClDetlWindExpOff = 7 !last required field
INTEGER,PARAMETER :: rfClDetlViewGroundOff = 8
INTEGER,PARAMETER :: rfClDetlNumVertOff = 9
INTEGER,PARAMETER :: rfClDetlX1Off = 10
INTEGER,PARAMETER :: rfClDetlY1Off = 11
INTEGER,PARAMETER :: rfClDetlZ1Off = 12
INTEGER,PARAMETER :: rfClDetlLastZVertexOff = 39 !up to 10 vertices

INTEGER,PARAMETER :: flrDetlNameOff = 1
INTEGER,PARAMETER :: flrDetlConstructionOff = 2
INTEGER,PARAMETER :: flrDetlZoneOff = 3
INTEGER,PARAMETER :: flrDetlOutBndCondOff = 4
INTEGER,PARAMETER :: flrDetlOutBndCondNameOff = 5
INTEGER,PARAMETER :: flrDetlSunExpOff = 6
INTEGER,PARAMETER :: flrDetlWindExpOff = 7 !last required field
INTEGER,PARAMETER :: flrDetlViewGroundOff = 8
INTEGER,PARAMETER :: flrDetlNumVertOff = 9
INTEGER,PARAMETER :: flrDetlX1Off = 10
INTEGER,PARAMETER :: flrDetlY1Off = 11
INTEGER,PARAMETER :: flrDetlZ1Off = 12
INTEGER,PARAMETER :: flrDetlLastZVertexOff = 39 !up to 10 vertices

INTEGER,PARAMETER :: gndTempSurfJanOff = 1
INTEGER,PARAMETER :: gndTempSurfFebOff = 2
INTEGER,PARAMETER :: gndTempSurfMarOff = 3
INTEGER,PARAMETER :: gndTempSurfAprOff = 4
INTEGER,PARAMETER :: gndTempSurfMayOff = 5
INTEGER,PARAMETER :: gndTempSurfJunOff = 6
INTEGER,PARAMETER :: gndTempSurfJulOff = 7
INTEGER,PARAMETER :: gndTempSurfAugOff = 8
INTEGER,PARAMETER :: gndTempSurfSepOff = 9
INTEGER,PARAMETER :: gndTempSurfOctOff = 10
INTEGER,PARAMETER :: gndTempSurfNovOff = 11
INTEGER,PARAMETER :: gndTempSurfDecOff = 12

  ! Boiler:HotWater (for use with HVACTemplate:Plant:Boiler:ObjectReference)
INTEGER,PARAMETER :: blrhwNameOff = 1
INTEGER,PARAMETER :: blrhwInletNodeOff = 11
INTEGER,PARAMETER :: blrhwOutletNodeOff = 12
INTEGER,PARAMETER :: blrhwLastFieldOff = 17

  ! Chiller:Electric:EIR (for use with HVACTemplate:Plant:Chiller:ObjectReference)
INTEGER,PARAMETER :: chleirNameOff = 1
INTEGER,PARAMETER :: chleirChWInletNodeOff = 15
INTEGER,PARAMETER :: chleirChWOutletNodeOff = 16
INTEGER,PARAMETER :: chleirCondInletNodeOff = 17
INTEGER,PARAMETER :: chleirCondOutletNodeOff = 18
INTEGER,PARAMETER :: chleirCondTypeOff = 19
INTEGER,PARAMETER :: chleirLastFieldOff = 34

  ! Chiller:Electric:ReformulatedEIR (for use with HVACTemplate:Plant:Chiller:ObjectReference)
INTEGER,PARAMETER :: chlreirNameOff = 1
INTEGER,PARAMETER :: chlreirChWInletNodeOff = 16
INTEGER,PARAMETER :: chlreirChWOutletNodeOff = 17
INTEGER,PARAMETER :: chlreirCondInletNodeOff = 18
INTEGER,PARAMETER :: chlreirCondOutletNodeOff = 19
INTEGER,PARAMETER :: chlreirLastFieldOff = 30

  ! CoolingTower:SingleSpeed (for use with HVACTemplate:Plant:Tower:ObjectReference)
INTEGER,PARAMETER :: twrssNameOff = 1
INTEGER,PARAMETER :: twrssInletNodeOff = 2
INTEGER,PARAMETER :: twrssOutletNodeOff = 3
INTEGER,PARAMETER :: twrssLastFieldOff = 38

  ! CoolingTower:TwoSpeed (for use with HVACTemplate:Plant:Tower:ObjectReference)
INTEGER,PARAMETER :: twrtsNameOff = 1
INTEGER,PARAMETER :: twrtsInletNodeOff = 2
INTEGER,PARAMETER :: twrtsOutletNodeOff = 3
INTEGER,PARAMETER :: twrtsLastFieldOff = 45

  ! CoolingTower:VariableSpeed (for use with HVACTemplate:Plant:Tower:ObjectReference)
INTEGER,PARAMETER :: twrvsNameOff = 1
INTEGER,PARAMETER :: twrvsInletNodeOff = 2
INTEGER,PARAMETER :: twrvsOutletNodeOff = 3
INTEGER,PARAMETER :: twrvsLastFieldOff = 31

!following objects are identified just by the first and last fields
INTEGER,PARAMETER :: ghtBsSimParamFirstOff = 1
INTEGER,PARAMETER :: ghtBsSimParamLastOff = 2

INTEGER,PARAMETER :: ghtBsMatlPropsFirstOff = 1
INTEGER,PARAMETER :: ghtBsMatlPropsLastOff = 19

INTEGER,PARAMETER :: ghtBsInsulFirstOff = 1
INTEGER,PARAMETER :: ghtBsInsulLastOff = 2

INTEGER,PARAMETER :: ghtBsSurfPropsFirstOff = 1
INTEGER,PARAMETER :: ghtBsSurfPropsLastOff = 7

INTEGER,PARAMETER :: ghtBsBldgDataFirstOff = 1
INTEGER,PARAMETER :: ghtBsBldgDataLastOff = 5

INTEGER,PARAMETER :: ghtBsInteriorFirstOff = 1
INTEGER,PARAMETER :: ghtBsInteriorLastOff = 7

INTEGER,PARAMETER :: ghtBsComBldgFirstOff = 1
INTEGER,PARAMETER :: ghtBsComBldgLastOff = 13

INTEGER,PARAMETER :: ghtBsEqvSlbFirstOff = 1
INTEGER,PARAMETER :: ghtBsEqvSlbLastOff = 2

INTEGER,PARAMETER :: ghtBsEqvAutoGridFirstOff = 1
INTEGER,PARAMETER :: ghtBsEqvAutoGridLastOff = 3

INTEGER,PARAMETER :: ghtBsAutoGridFirstOff = 1
INTEGER,PARAMETER :: ghtBsAutoGridLastOff = 6

INTEGER,PARAMETER :: ghtBsManGridFirstOff = 1
INTEGER,PARAMETER :: ghtBsManGridLastOff = 7

INTEGER,PARAMETER :: ghtBsXFaceFirstOff = 1
INTEGER,PARAMETER :: ghtBsXFaceLastOff = 44

INTEGER,PARAMETER :: ghtBsYFaceFirstOff = 1
INTEGER,PARAMETER :: ghtBsYFaceLastOff = 44

INTEGER,PARAMETER :: ghtBsZFaceFirstOff = 1
INTEGER,PARAMETER :: ghtBsZFaceLastOff = 40

INTEGER,PARAMETER :: ghtSlbMaterialFirstOff = 1
INTEGER,PARAMETER :: ghtSlbMaterialLastOff = 9

INTEGER,PARAMETER :: ghtSlbMatlPropsFirstOff = 1
INTEGER,PARAMETER :: ghtSlbMatlPropsLastOff = 6

INTEGER,PARAMETER :: ghtSlbBoundCondsFirstOff = 1
INTEGER,PARAMETER :: ghtSlbBoundCondsLastOff = 5

INTEGER,PARAMETER :: ghtSlbBldgPropsFirstOff = 1
INTEGER,PARAMETER :: ghtSlbBldgPropsLastOff = 17

INTEGER,PARAMETER :: ghtSlbInsulFirstOff = 1
INTEGER,PARAMETER :: ghtSlbInsulLastOff = 5

INTEGER,PARAMETER :: ghtSlbEqvSlbFirstOff = 1
INTEGER,PARAMETER :: ghtSlbEqvSlbLastOff = 4

INTEGER,PARAMETER :: ghtSlbAutoGridFirstOff = 1
INTEGER,PARAMETER :: ghtSlbAutoGridLastOff = 5

INTEGER,PARAMETER :: ghtSlbManGridFirstOff = 1
INTEGER,PARAMETER :: ghtSlbManGridLastOff = 5

INTEGER,PARAMETER :: ghtSlbXFaceFirstOff = 1
INTEGER,PARAMETER :: ghtSlbXFaceLastOff = 40

INTEGER,PARAMETER :: ghtSlbYFaceFirstOff = 1
INTEGER,PARAMETER :: ghtSlbYFaceLastOff = 40

INTEGER,PARAMETER :: ghtSlbZFaceFirstOff = 1
INTEGER,PARAMETER :: ghtSlbZFaceLastOff = 25

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactStatBase !holds fldVal index for start of object
INTEGER                            :: numCompactStat=0
INTEGER                            :: ptCompactStat=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactZoneVAVbase
INTEGER                            :: numCompactZoneVAV=0
INTEGER                            :: ptCompactZoneVAV=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactZoneFPVAVbase
INTEGER                            :: numCompactZoneFPVAV=0
INTEGER                            :: ptCompactZoneFPVAV=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactZoneHCVAVbase
INTEGER                            :: numCompactZoneHCVAV=0
INTEGER                            :: ptCompactZoneHCVAV=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactSysVAVbase
INTEGER                            :: numCompactSysVAV=0
INTEGER                            :: ptCompactSysVAV=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactSysPVAVbase
INTEGER                            :: numCompactSysPVAV=0
INTEGER                            :: ptCompactSysPVAV=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactZoneUnitBase
INTEGER                            :: numCompactZoneUnit=0
INTEGER                            :: ptCompactZoneUnit=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactSysUnitBase
INTEGER                            :: numCompactSysUnit=0
INTEGER                            :: ptCompactSysUnit=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactSysUnitHPBase
INTEGER                            :: numCompactSysUnitHP=0
INTEGER                            :: ptCompactSysUnitHP=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactSysUnitarySystemBase
INTEGER                            :: numCompactSysUnitarySystem=0
INTEGER                            :: ptCompactSysUnitarySystem=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactZoneConstVolBase
INTEGER                            :: numCompactZoneConstVol=0
INTEGER                            :: ptCompactZoneConstVol=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactSysConstVolBase
INTEGER                            :: numCompactSysConstVol=0
INTEGER                            :: ptCompactSysConstVol=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactZoneDualDuctBase
INTEGER                            :: numCompactZoneDualDuct=0
INTEGER                            :: ptCompactZoneDualDuct=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactSysDualDuctBase
INTEGER                            :: numCompactSysDualDuct=0
INTEGER                            :: ptCompactSysDualDuct=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactZoneVRFBase
INTEGER                            :: numCompactZoneVRF=0
INTEGER                            :: ptCompactZoneVRF=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactSysVRFBase
INTEGER                            :: numCompactSysVRF=0
INTEGER                            :: ptCompactSysVRF=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactFanCoilBase
INTEGER                            :: numCompactFanCoil=0
INTEGER                            :: ptCompactFanCoil=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactPurchAirBase
INTEGER                            :: numCompactPurchAir=0
INTEGER                            :: ptCompactPurchAir=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactBaseboardBase
INTEGER                            :: numCompactBaseboard=0
INTEGER                            :: ptCompactBaseboard=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactPTACBase
INTEGER                            :: numCompactPTAC=0
INTEGER                            :: ptCompactPTAC=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactPTHPBase
INTEGER                            :: numCompactPTHP=0
INTEGER                            :: ptCompactPTHP=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactBoilerBase
INTEGER                            :: numCompactBoiler=0
INTEGER                            :: ptCompactBoiler=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactBoilerORBase
INTEGER                            :: numCompactBoilerOR=0
INTEGER                            :: ptCompactBoilerOR=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: detBoilerHWBase
INTEGER                            :: numDetBoilerHW=0
INTEGER                            :: ptDetBoilerHW=0

INTEGER                            :: compactHotLoopBase = 0
INTEGER                            :: ptCompactHotLoop=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactChillerBase
INTEGER                            :: numCompactChiller=0
INTEGER                            :: ptCompactChiller=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactChillerORBase
INTEGER                            :: numCompactChillerOR=0
INTEGER                            :: ptCompactChillerOR=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: detChillerEIRBase
INTEGER                            :: numDetChillerEIR=0
INTEGER                            :: ptDetChillerEIR=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: detChillerREIRBase
INTEGER                            :: numDetChillerREIR=0
INTEGER                            :: ptDetChillerREIR=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactTowerBase
INTEGER                            :: numCompactTower=0
INTEGER                            :: ptCompactTower=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactTowerORBase
INTEGER                            :: numCompactTowerOR=0
INTEGER                            :: ptCompactTowerOR=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: detTowerSSBase
INTEGER                            :: numDetTowerSS=0
INTEGER                            :: ptDetTowerSS=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: detTowerTSBase
INTEGER                            :: numDetTowerTS=0
INTEGER                            :: ptDetTowerTS=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: detTowerVSBase
INTEGER                            :: numDetTowerVS=0
INTEGER                            :: ptDetTowerVS=0

INTEGER                            :: compactChldLoopBase = 0
INTEGER                            :: ptCompactChldLoop=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactWaterAirHPBase
INTEGER                            :: numCompactWaterAirHP=0
INTEGER                            :: ptCompactWaterAirHP=0

INTEGER                            :: compactMixedLoopBase = 0
INTEGER                            :: ptCompactMixedLoop=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: compactDedOutAirBase
INTEGER                            :: numCompactDedOutAir=0
INTEGER                            :: ptCompactDedOutAir=0

INTEGER                            :: ghtControlBase=0
INTEGER                            :: ptGhtControl=0

INTEGER                            :: gndTempBuildSurfBase=0
INTEGER                            :: ptGndTempBuildSurf=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: surfBldgDetlBase
INTEGER                            :: numSurfBldgDetl=0
INTEGER                            :: ptSurfBldgDetl=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: surfWallDetlBase
INTEGER                            :: numSurfWallDetl=0
INTEGER                            :: ptSurfWallDetl=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: surfRfClDetlBase
INTEGER                            :: numSurfRfClDetl=0
INTEGER                            :: ptSurfRfClDetl=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: surfFlrDetlBase
INTEGER                            :: numSurfFlrDetl=0
INTEGER                            :: ptSurfFlrDetl=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtGndTempSurfBase
INTEGER                            :: numGhtGndTempSurf=0
INTEGER                            :: ptGhtGndTempSurf=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtBsSimParamBase
INTEGER                            :: numGhtBsSimParam=0
INTEGER                            :: ptGhtBsSimParam=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtBaseBsMatlProps
INTEGER                            :: numGhtBsMatlProps=0
INTEGER                            :: ptGhtBsMatlProps=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtBsInsulBase
INTEGER                            :: numGhtBsInsul=0
INTEGER                            :: ptGhtBsInsul=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtBsSurfPropsBase
INTEGER                            :: numGhtBsSurfProps=0
INTEGER                            :: ptGhtBsSurfProps=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtBsBldgDataBase
INTEGER                            :: numGhtBsBldgData=0
INTEGER                            :: ptGhtBsBldgData=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtBsInteriorBase
INTEGER                            :: numGhtBsInterior=0
INTEGER                            :: ptGhtBsInterior=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtBsComBldgBase
INTEGER                            :: numGhtBsComBldg=0
INTEGER                            :: ptGhtBsComBldg=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtBsEqvSlbBase
INTEGER                            :: numGhtBsEqvSlb=0
INTEGER                            :: ptGhtBsEqvSlb=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtBsEqvAutoGridBase
INTEGER                            :: numGhtBsEqvAutoGrid=0
INTEGER                            :: ptGhtBsEqvAutoGrid=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtBsAutoGridBase
INTEGER                            :: numGhtBsAutoGrid=0
INTEGER                            :: ptGhtBsAutoGrid=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtBsManGridBase
INTEGER                            :: numGhtBsManGrid=0
INTEGER                            :: ptGhtBsManGrid=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtBsXFaceBase
INTEGER                            :: numGhtBsXFace=0
INTEGER                            :: ptGhtBsXFace=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtBsYFaceBase
INTEGER                            :: numGhtBsYFace=0
INTEGER                            :: ptGhtBsYFace=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtBsZFaceBase
INTEGER                            :: numGhtBsZFace=0
INTEGER                            :: ptGhtBsZFace=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtSlbMaterialBase
INTEGER                            :: numGhtSlbMaterial=0
INTEGER                            :: ptGhtSlbMaterial=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtSlbMatlPropsBase
INTEGER                            :: numGhtSlbMatlProps=0
INTEGER                            :: ptGhtSlbMatlProps=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtSlbBoundCondsBase
INTEGER                            :: numGhtSlbBoundConds=0
INTEGER                            :: ptGhtSlbBoundConds=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtSlbBldgPropsBase
INTEGER                            :: numGhtSlbBldgProps=0
INTEGER                            :: ptGhtSlbBldgProps=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtSlbInsulBase
INTEGER                            :: numGhtSlbInsul=0
INTEGER                            :: ptGhtSlbInsul=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtSlbEqvSlbBase
INTEGER                            :: numGhtSlbEqvSlb=0
INTEGER                            :: ptGhtSlbEqvSlb=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtSlbAutoGridBase
INTEGER                            :: numGhtSlbAutoGrid=0
INTEGER                            :: ptGhtSlbAutoGrid=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtSlbManGridBase
INTEGER                            :: numGhtSlbManGrid=0
INTEGER                            :: ptGhtSlbManGrid=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtSlbXFaceBase
INTEGER                            :: numGhtSlbXFace=0
INTEGER                            :: ptGhtSlbXFace=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtSlbYFaceBase
INTEGER                            :: numGhtSlbYFace=0
INTEGER                            :: ptGhtSlbYFace=0

INTEGER, ALLOCATABLE, DIMENSION(:) :: ghtSlbZFaceBase
INTEGER                            :: numGhtSlbZFace=0
INTEGER                            :: ptGhtSlbZFace=0

! types of integer zone name lists
INTEGER, PARAMETER :: zlUnitAll = 11
INTEGER, PARAMETER :: zlUnitAllSupPlen = 12
INTEGER, PARAMETER :: zlUnitAllRetPlen = 13
INTEGER, PARAMETER :: zlVavAll = 21
INTEGER, PARAMETER :: zlVavAllSupPlen = 22
INTEGER, PARAMETER :: zlVavAllRetPlen = 23
INTEGER, PARAMETER :: zlVavFPfromSysRetPlen = 24
INTEGER, PARAMETER :: zlDOAS = 31
INTEGER, PARAMETER :: zlConstVolAll = 41
INTEGER, PARAMETER :: zlConstVolAllSupPlen = 42
INTEGER, PARAMETER :: zlConstVolAllRetPlen = 43
INTEGER, PARAMETER :: zlDualDuctAll = 51
INTEGER, PARAMETER :: zlDualDuctColdSupPlen = 52
INTEGER, PARAMETER :: zlDualDuctHotSupPlen = 53
INTEGER, PARAMETER :: zlDualDuctAllRetPlen = 54
INTEGER, PARAMETER :: zlVRFAll = 61

LOGICAL :: infileExist=.false.
INTEGER :: ios=0
LOGICAL :: fileerrorsExist=.false.
LOGICAL :: outfileOpened=.false.
LOGICAL :: gatherSurfaces=.true.

!----------------------------------------------------------------------------------
! Main routine
!
! Quick Overview of Algorithm
! ---------------------------
! initialize
! open file
! loop through file reading a line at a time
! if in object
!    if yes, if object to process add items to current object
!    comment line if in object is flagged for comment
! if not in object and if an object is found
!    determine if it should be be processed
!      if yes, determine if it should be commented out
! at end append objects to file based on object saved during read
!
PRINT "(A)", "ExpandObjects Started."
CALL CPU_TIME(startTime)
IF (checkOutputFieldNamesAgainstIDD) PRINT "(A)", "******* checkOutputFieldNamesAgainstIDD Enabled ***********"
InputFileName = 'in.idf'
OutputFileName = 'expanded.idf'
ErrorFileName = 'expandedidf.err'
fileerrorsExist=.false.
INQUIRE(file=Trim(InputFileName),exist=InfileExist)
if (infileexist) then
  OPEN (unit=InFH, file=TRIM(InputFileName), action='READ',iostat=ios,iomsg=IOErrorMsg)
!  OPEN (unit=InFH, file=TRIM(InputFileName), action='READ',iostat=ios)
  if (ios /= 0) then
    call AddToErrMsg('error opening in.idf= runtime error#'//trim(ADJUSTL(IntToStr(ios)))//' '//trim(IOErrorMsg))
    fileerrorsExist=.true.
  endif
  OPEN (unit=OutFH, file=TRIM(OutputFileName), action='WRITE',iostat=ios,iomsg=IOErrorMsg)
!  OPEN (unit=OutFH, file=TRIM(OutputFileName), action='WRITE',iostat=ios)
  if (ios /= 0) then
!    IOErrorMsg=ADJUSTL(IntToStr(ios))
    call AddToErrMsg('error opening expanded.idf= runtime error#'//trim(ADJUSTL(IntToStr(ios)))//' '//trim(IOErrorMsg))
    fileerrorsExist=.true.
  endif
  OPEN (unit=ErrFH, file=TRIM(ErrorFileName), action='WRITE',iostat=ios,iomsg=IOErrorMsg)
!  OPEN (unit=ErrFH, file=TRIM(ErrorFileName), action='WRITE',iostat=ios)
  if (ios /= 0) then
!    IOErrorMsg=ADJUSTL(IntToStr(ios))
    call AddToErrMsg('error opening expandedidf.err= runtime error#'//trim(ADJUSTL(IntToStr(ios)))//' '//trim(IOErrorMsg))
    fileerrorsExist=.true.
  endif
  if (.not. fileerrorsExist) then
    !these are fixed since the number of objects and fields should be know based on IDD changes.
    !turn on the report writing in ReadIDDfile to see how much of each array is used.
    gatherSurfaces = DoesGroundHeatTransferExist()
    CLOSE(InFH) !close and reopen the file
    OPEN (unit=InFH, file=TRIM(InputFileName), action='READ',iostat=ios,iomsg=IOErrorMsg)
    CALL SetUpObjectsToProcess(gatherSurfaces)
    CALL ReadWriteFileAndProcess
    CALL GetInputAll
    IF (foundCompactObject) THEN
      IF (.NOT. checkOutputFieldNamesAgainstIDD) THEN
        ALLOCATE (iddObj(50))
        ALLOCATE (iddFld(1500))
        ALLOCATE (iddKey(1500))
      ELSE
        ALLOCATE (iddObj(1000))
        ALLOCATE (iddFld(20000))
        ALLOCATE (iddKey(20000))
      END IF
      CALL locateIDDfile
      IF (iddCanBeRead) THEN
        CALL readIDDfile
        CALL AssociateIDDwithObjToProc
        CALL CheckMaxMinKeys
      ELSE
        CALL WriteError('Since the Energy+.IDD file cannot be read no range or choice checking was performed.',msgWarning)
      END IF
      CALL CreateNewObjects
      IF (.NOT. checkOutputFieldNamesAgainstIDD) THEN !don't deallocate if checking agaist field names - this will happen at end of program
        DEALLOCATE (iddKey)
        DEALLOCATE (iddFld)
        DEALLOCATE (iddObj)
      END IF
    ENDIF
  endif
  CLOSE(InFH)
else
  call AddToErrMsg('error opening in.idf= no file found')
  fileerrorsExist=.true.
  INQUIRE(unit=OutFH, OPENED=outfileopened)
!  OPEN (unit=OutFH, file=TRIM(OutputFileName), action='WRITE',iostat=ios)
  OPEN (unit=OutFH, file=TRIM(OutputFileName), action='WRITE',iostat=ios,iomsg=IOErrorMsg)
  if (ios /= 0) then
!    IOErrorMsg=ADJUSTL(IntToStr(ios))
    call AddToErrMsg('error opening expanded.idf= runtime error#'//trim(ADJUSTL(IntToStr(ios)))//' '//trim(IOErrorMsg))
    PRINT "(A)",'error opening expanded.idf= runtime error#'//trim(IOErrorMsg)
  endif
  errorCondition=.true.
endif
!*The following lines allows seeing main array values when debugging (needs a reference to show them as local?)
!IF (ALLOCATED(FldVal)) THEN
!  FldVal(1) = FldVal(1)
!END IF
!*The following lines are for testing the word wrapping for error messages
!CALL WriteError('TEST ERROR MESSAGE REMOVE BEFORE RELEASE You must specify at least one COMPACT HVAC:SYSTEM:VAV or COMPACT HVAC:ZONE:FAN COIL if a COMPACT HVAC:PLANT:HOT WATER LOOP is defined.')
!CALL WriteError('TEST ERROR MESSAGE REMOVE BEFORE RELEASE Invalid choice in COMPACT HVAC:SYSTEM:VAV in the Economizer Type field Invalid choice in COMPACT HVAC:SYSTEM:VAV in the Economizer Type field')
!CALL WriteError('TEST ERROR MESSAGE REMOVE BEFORE RELEASE The quick brown fox jumped over the lazy dog The quick brown fox jumped over the lazy dog The quick brown fox jumped over the lazy dog.')
!CALL WriteError('TEST-ERROR-MESSAGE-REMOVE-BEFORE-RELEASE-The-quick-brown-fox-jumped-over-the-lazy-dog-The-quick-brown-fox jumped over the lazy dog The quick brown fox jumped over the lazy dog.')
!if no compact objects found then delete the output file and exit
IF (foundCompactObject .or. fileerrorsexist) THEN
  IF (errorCondition) THEN
    CALL writeErrorObjects
  END IF
  CLOSE(OutFH)
ELSE
  PRINT "(A)", "No expanded file generated."
  CLOSE(OutFH,STATUS='DELETE') !get rid of file if no changes are made
END IF
!if errors were found keep the file and report errors otherwise just delete the file
IF (errorCondition) THEN
  CLOSE(ErrFH)
  CALL CPU_TIME(endTime)
  PRINT "(A,F10.3)", "ExpandObjects Finished with Error(s). Time:", endTime - startTime
ELSE
  CLOSE(ErrFH,STATUS='DELETE')  !get rid of error file if no errors found
  CALL CPU_TIME(endTime)
  PRINT "(A,F10.3)", "ExpandObjects Finished. Time:", endTime - startTime
END IF

CONTAINS

!----------------------------------------------------------------------------------
LOGICAL FUNCTION DoesGroundHeatTransferExist()
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   September 2012
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Preprocesses the file and check if GroundHeatTransfer
          !    objects exist.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: read_stat
CHARACTER(len=MaxLineLength) :: LineIn
INTEGER :: semiPos  !position of the semicolon
INTEGER :: exclPos  !position of the exclamation mark
INTEGER :: commaPos !position of the comma
INTEGER :: tpos     ! position of tab character in input line
LOGICAL :: isGHTexist !true if ground heat transfer object found

CALL ClearBuffer
read_stat=0
isGHTexist = .FALSE.
DO WHILE (.not. is_iostat_end(read_stat)) !end of file
  READ(UNIT=InFH, FMT="(A)", IOSTAT=read_stat) LineIn
  IF (read_stat /= 0) LineIn = ''
  tpos=INDEX(LineIn,tabchar)
  DO WHILE (tpos > 0)
    LineIn(tpos:tpos) = ' '
    tpos=INDEX(LineIn,tabchar)
  END DO
  ! find the comment character, the object ending character and the field delimiter
  CALL FindDelimiters(LineIn,exclPos,semiPos,commaPos)
  CALL AddToBuffer(LineIn)  !add the single line to the buffer
  IF (semiPos .GT. 0) THEN
    ! if the buffer contains an object that should be held
    CALL testBufferForGroundHeatTrans(isGHTexist)
    CALL ClearBuffer
  END IF
  LineIn = ''
  IF (isGHTexist) EXIT
END DO
DoesGroundHeatTransferExist = isGHTexist
END FUNCTION DoesGroundHeatTransferExist

!----------------------------------------------------------------------------------
SUBROUTINE testBufferForGroundHeatTrans(doesGHT)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Check if the buffer includes an object that should be held and
          !    if the object should be commented out.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
LOGICAL, INTENT(OUT) :: doesGHT

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
INTEGER :: iBuffer
INTEGER :: iObjProc
INTEGER :: semiPos  !position of the semicolon
INTEGER :: exclPos  !position of the exclamation mark
INTEGER :: commaPos !position of the comma
CHARACTER(len=MaxAlphaLength) :: firstField
LOGICAL :: foundFirstField


! first assume that the buffer does not contain an object to hold
! until later
foundFirstField = .FALSE.
firstField = ''
doesGHT = .FALSE.
DO iBuffer = 1, numInBuffer
  ! find the comment character, the object ending character and the field delimiter
  CALL FindDelimiters(InBuffer(iBuffer),exclPos,semiPos,commaPos)
  ! if end of object or delimiter is on the line then found the first
  ! field of the object which is the object name
  ! exit the look because we are just looking for the first field
  IF (commaPos .GT. 1) THEN
    firstField = ADJUSTL(InBuffer(iBuffer)(1:commaPos - 1))
    foundFirstField = .TRUE.
    EXIT
  ELSE IF (semiPos .GT. 1) THEN
    firstField = ADJUSTL(InBuffer(iBuffer)(1:semiPos - 1))
    foundFirstField = .TRUE.
    EXIT
  END IF
END DO
! if a first field was found then look for it in the objects to process list
IF  (foundFirstField) THEN
  !                                12345678901234567890
  IF (SameString(firstField(1:18),'GroundHeatTransfer')) THEN
    doesGHT = .TRUE.
  END IF
END IF
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE SetUpObjectsToProcess(doGatherSurfaces)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !   Routine that sets all the objects that should be processed including
          !   objects that should be commented out.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          ! The min-fields designation for each of the objects is used in error messages only.
          ! HVACTemplate:Thermostat,
          !       \min-fields 5
          ! HVACTemplate:Zone:IdealLoadsAirSystem,
          !       \min-fields 2
          ! HVACTemplate:Zone:FanCoil,
          !       \min-fields 20
          ! HVACTemplate:Zone:PTAC,
          !       \min-fields 27
          ! HVACTemplate:Zone:PTHP,
          !       \min-fields 37
          ! HVACTemplate:Zone:Unitary,
          !       \min-fields 15
          ! HVACTemplate:Zone:VAV,
          !       \min-fields 26
          ! HVACTemplate:Zone:VAV:FanPowered,
          !       \min-fields 24
          ! HVACTemplate:Zone:WaterToAirHeatPump,
          !       \min-fields 39
          ! HVACTemplate:System:Unitary,
          !       \min-fields 51
          ! HVACTemplate:System:UnitarySystem,
          !       \min-fields 71
          ! HVACTemplate:System:UnitaryHeatPump:AirToAir,
          !       \min-fields 54
          ! HVACTemplate:System:VAV,
          !       \min-fields 53
          ! HVACTemplate:System:PackagedVAV,
          !       \min-fields 52
          ! HVACTemplate:System:DedicatedOutdoorAir,
          !       \min-fields 44
          ! HVACTemplate:Plant:ChilledWaterLoop,
          !       \min-fields 21
          ! HVACTemplate:Plant:Chiller,
          !       \min-fields 7
          ! HVACTemplate:Plant:Tower,
          !       \min-fields 9
          ! HVACTemplate:Plant:HotWaterLoop,
          !       \min-fields 14
          ! HVACTemplate:Plant:Boiler,
          !       \min-fields 7
          ! HVACTemplate:Plant:MixedWaterLoop,
          !       \min-fields 11
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
LOGICAL,INTENT(IN)                       :: doGatherSurfaces

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
! SUBROUTINE AddObjToProcess(objName,isCommentOut,countFieldMin,countFieldMax,numberExpected)
!   countFieldMin should be the last required field. An object with fewer fields will be an error
!   countFieldMax should be the last possible field. An object with more fields will be an error
!   numberExpected should be min-fields (I think, MJW)
CALL AddObjToProcess('HVACTemplate:Thermostat',.TRUE.,              tCoolSchedNameOff,           tCoolSetPointOff,           5)
CALL AddObjToProcess('HVACTemplate:Zone:VAV',.TRUE.,                vzConstantMinFracOff,        vzDesSpecOADistNameOff,    34)
CALL AddObjToProcess('HVACTemplate:Zone:VAV:FanPowered',.TRUE.,     fpvzAirHandlerNameOff,       fpvzDesSpecOADistNameOff,  33)
CALL AddObjToProcess('HVACTemplate:Zone:VAV:HeatAndCool',.TRUE.,    hcvzAirHandlerNameOff,       hcvzHeatDesignTempDiffOff, 27)
CALL AddObjToProcess('HVACTemplate:System:Unitary',.TRUE.,          usHeatCoilKindOff,           usReturnFanFractionOff,    51)
CALL AddObjToProcess('HVACTemplate:System:UnitaryHeatPump:AirToAir',.TRUE.,uhpsHPHeatCoilKindOff,uhpsReturnFanFractionOff,  61)
CALL AddObjToProcess('HVACTemplate:System:UnitarySystem',.TRUE.,    ussAirHandlerNameOff,        ussReturnFanFractionOff,   71)
CALL AddObjToProcess('HVACTemplate:Zone:Unitary',.TRUE.,            uzAirHandlerNameOff,         uzDesSpecOADistNameOff,    23)
CALL AddObjToProcess('HVACTemplate:System:VAV',.TRUE.,              vsAirHandlerNameOff,         vsReturnFanPLRCoeffOff,    61)
CALL AddObjToProcess('HVACTemplate:System:PackagedVAV',.TRUE.,      pvavsAirHandlerNameOff,      pvavsReturnFanPLRCoeffOff, 59)
CALL AddObjToProcess('HVACTemplate:Zone:ConstantVolume',.TRUE.,     cvzAirHandlerNameOff,        cvzHeatDesignTempDiffOff,  24)
CALL AddObjToProcess('HVACTemplate:System:ConstantVolume',.TRUE.,   cvsAirHandlerNameOff,        cvsReturnFanFractionOff,   70)
CALL AddObjToProcess('HVACTemplate:Zone:VRF',.TRUE.,                vrfzSystemNameOff,           vrfzBaseboardCapOff,       44)
CALL AddObjToProcess('HVACTemplate:System:VRF',.TRUE.,              vrfsNameOff,                 vrfsHeatRecMaxODBOff,      39)
CALL AddObjToProcess('HVACTemplate:Zone:DualDuct',.TRUE.,           ddzAirHandlerNameOff,        ddzHeatDesignTempDiffOff,  26)
CALL AddObjToProcess('HVACTemplate:System:DualDuct',.TRUE.,         ddsAirHandlerNameOff,        ddsReturnFanPLRCoeffOff,   91)
CALL AddObjToProcess('HVACTemplate:Zone:FanCoil',.TRUE.,            fczNameOff,                  fczBaseboardCapOff,        34)
CALL AddObjToProcess('HVACTemplate:Zone:IdealLoadsAirSystem',.TRUE.,pazNameOff,                  pazHeatRecLatEffOff,        2)
CALL AddObjToProcess('HVACTemplate:Zone:BaseboardHeat',.TRUE.,      bbzNameOff,                  bbzDesSpecOADistNameOff,   11)
CALL AddObjToProcess('HVACTemplate:Zone:PTAC',.TRUE.,               ptaczTStatNameOff,           ptaczCapCtrlTypeOff,       40)
CALL AddObjToProcess('HVACTemplate:Zone:PTHP',.TRUE.,               pthpzTStatNameOff,           pthpzCapCtrlTypeOff,       50)
CALL AddObjToProcess('HVACTemplate:Plant:ChilledWaterLoop',.TRUE.,  cwpNameOff,                  cwpCndLdDistSchmOff,       33)
CALL AddObjToProcess('HVACTemplate:Plant:Chiller',.TRUE.,           chlPriorityOff,              chlLeavingLowLimOff,       12)
CALL AddObjToProcess('HVACTemplate:Plant:Chiller:ObjectReference',.TRUE.,  chlorPriorityOff,     chlorPriorityOff,           4)
CALL AddObjToProcess('HVACTemplate:Plant:Tower',.TRUE.,             twrPriorityOff,              twrLoopTypeOff,             9)
CALL AddObjToProcess('HVACTemplate:Plant:Tower:ObjectReference',.TRUE.,  twrorPriorityOff,       twrorLoopTypeOff,           4)
CALL AddObjToProcess('HVACTemplate:Plant:HotWaterLoop',.TRUE.,      hwpNameOff,                  hwpLdDistSchmOff,          21)
CALL AddObjToProcess('HVACTemplate:Plant:Boiler',.TRUE.,            blrPriorityOff,              blrLoopTypeOff,            11)
CALL AddObjToProcess('HVACTemplate:Plant:Boiler:ObjectReference',.TRUE.,  blrorPriorityOff,      blrorLoopTypeOff,           4)
CALL AddObjToProcess('HVACTemplate:Zone:WaterToAirHeatPump',.TRUE., wahpTStatNameOff,            wahpBaseboardCapOff,       44)
CALL AddObjToProcess('HVACTemplate:Plant:MixedWaterLoop',.TRUE.,    mwpNameOff,                  mwpLdDistSchmOff,          17)
CALL AddObjToProcess('HVACTemplate:System:DedicatedOutdoorAir',.TRUE.,doasNameOff,              doasHumidSetPtSchedNameOff, 46)
! Detailed Plant Objects
CALL AddObjToProcess('Boiler:HotWater',.FALSE.,                     blrhwOutletNodeOff,          blrhwLastFieldOff,         17)
CALL AddObjToProcess('Chiller:Electric:EIR',.FALSE.,                chleirCondTypeOff,           chleirLastFieldOff,        30)
CALL AddObjToProcess('Chiller:Electric:ReformulatedEIR',.FALSE.,    chlreirCondOutletNodeOff,    chlreirLastFieldOff,       26)
CALL AddObjToProcess('CoolingTower:SingleSpeed',.FALSE.,            twrssOutletNodeOff,          twrssLastFieldOff,         29)
CALL AddObjToProcess('CoolingTower:TwoSpeed',.FALSE.,               twrtsOutletNodeOff,          twrtsLastFieldOff,         32)
CALL AddObjToProcess('CoolingTower:VariableSpeed',.FALSE.,          twrvsOutletNodeOff,          twrvsLastFieldOff,         30)
! Ground Heat Transfer
CALL AddObjToProcess('GroundHeatTransfer:Control',.TRUE.,           ghtCtrlNameOff,              ghtCtrlSlabOff,             3)
IF (doGatherSurfaces) THEN !only gather these surfaces if groundheattransfer has been found.
  CALL AddObjToProcess('BuildingSurface:Detailed',.TRUE.,             buildSurfWindExpOff,         buildSurfLastZVertexOff,  370)
  CALL AddObjToProcess('Wall:Detailed',.TRUE.,                        wallDetlWindExpOff,          wallDetlLastZVertexOff,    39)
  CALL AddObjToProcess('RoofCeiling:Detailed',.TRUE.,                 rfClDetlWindExpOff,          rfClDetlLastZVertexOff,    39)
  CALL AddObjToProcess('Floor:Detailed',.TRUE.,                       flrDetlWindExpOff,           flrDetlLastZVertexOff,     39)
END IF
CALL AddObjToProcess('Site:GroundTemperature:BuildingSurface',.TRUE.,gndTempSurfJanOff,          gndTempSurfDecOff,         12)
! Items to echo to basement or slab input files
CALL AddObjToProcess('GroundHeatTransfer:Basement:SimParameters',.TRUE.,ghtBsSimParamFirstOff,   ghtBsSimParamLastOff,       2)
CALL AddObjToProcess('GroundHeatTransfer:Basement:MatlProps',.TRUE.,ghtBsMatlPropsFirstOff,      ghtBsMatlPropsLastOff,     19)
CALL AddObjToProcess('GroundHeatTransfer:Basement:Insulation',.TRUE.,ghtBsInsulFirstOff,         ghtBsInsulLastOff,          2)
CALL AddObjToProcess('GroundHeatTransfer:Basement:SurfaceProps',.TRUE.,ghtBsSurfPropsFirstOff,   ghtBsSurfPropsLastOff,      7)
CALL AddObjToProcess('GroundHeatTransfer:Basement:BldgData',.TRUE., ghtBsBldgDataFirstOff,       ghtBsBldgDataLastOff,       5)
CALL AddObjToProcess('GroundHeatTransfer:Basement:Interior',.TRUE., ghtBsInteriorFirstOff,       ghtBsInteriorLastOff,       7)
CALL AddObjToProcess('GroundHeatTransfer:Basement:ComBldg',.TRUE.,  ghtBsComBldgFirstOff,        ghtBsComBldgLastOff,       13)
CALL AddObjToProcess('GroundHeatTransfer:Basement:EquivSlab',.TRUE.,ghtBsEqvSlbFirstOff,         ghtBsEqvSlbLastOff,         2)
CALL AddObjToProcess('GroundHeatTransfer:Basement:EquivAutoGrid',.TRUE.,ghtBsEqvAutoGridFirstOff,ghtBsEqvAutoGridLastOff,    3)
CALL AddObjToProcess('GroundHeatTransfer:Basement:AutoGrid',.TRUE.,ghtBsAutoGridFirstOff,        ghtBsAutoGridLastOff,       6)
CALL AddObjToProcess('GroundHeatTransfer:Basement:ManualGrid',.TRUE.,ghtBsManGridFirstOff,       ghtBsManGridLastOff,        7)
CALL AddObjToProcess('GroundHeatTransfer:Basement:XFACE',.TRUE.,    ghtBsXFaceFirstOff,          ghtBsXFaceLastOff,         44)
CALL AddObjToProcess('GroundHeatTransfer:Basement:YFACE',.TRUE.,    ghtBsYFaceFirstOff,          ghtBsYFaceLastOff,         44)
CALL AddObjToProcess('GroundHeatTransfer:Basement:ZFACE',.TRUE.,    ghtBsZFaceFirstOff,          ghtBsZFaceLastOff,         40)
CALL AddObjToProcess('GroundHeatTransfer:Slab:Materials',.TRUE.,    ghtSlbMaterialFirstOff,      ghtSlbMaterialLastOff,      9)
CALL AddObjToProcess('GroundHeatTransfer:Slab:MatlProps',.TRUE.,    ghtSlbMatlPropsFirstOff,     ghtSlbMatlPropsLastOff,     6)
CALL AddObjToProcess('GroundHeatTransfer:Slab:BoundConds',.TRUE.,   ghtSlbBoundCondsFirstOff,    ghtSlbBoundCondsLastOff,    5)
CALL AddObjToProcess('GroundHeatTransfer:Slab:BldgProps',.TRUE.,    ghtSlbBldgPropsFirstOff,     ghtSlbBldgPropsLastOff,    17)
CALL AddObjToProcess('GroundHeatTransfer:Slab:Insulation',.TRUE.,   ghtSlbInsulFirstOff,         ghtSlbInsulLastOff,         5)
CALL AddObjToProcess('GroundHeatTransfer:Slab:EquivalentSlab',.TRUE.,ghtSlbEqvSlbFirstOff,       ghtSlbEqvSlbLastOff,        4)
CALL AddObjToProcess('GroundHeatTransfer:Slab:AutoGrid',.TRUE.,     ghtSlbAutoGridFirstOff,      ghtSlbAutoGridLastOff,      5)
CALL AddObjToProcess('GroundHeatTransfer:Slab:ManualGrid',.TRUE.,   ghtSlbManGridFirstOff,       ghtSlbManGridLastOff,       5)
CALL AddObjToProcess('GroundHeatTransfer:Slab:XFACE',.TRUE.,        ghtSlbXFaceFirstOff,         ghtSlbXFaceLastOff,        40)
CALL AddObjToProcess('GroundHeatTransfer:Slab:YFACE',.TRUE.,        ghtSlbYFaceFirstOff,         ghtSlbYFaceLastOff,        40)
CALL AddObjToProcess('GroundHeatTransfer:Slab:ZFACE',.TRUE.,        ghtSlbZFaceFirstOff,         ghtSlbZFaceLastOff,        25)

END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE AddObjToProcess(objName,isCommentOut,countFieldMin,countFieldMax,numberExpected)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !   Adds the object and the comment flag to the ObjToProcess array that
          !   is saved for the creation of new objects

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*),INTENT(IN)              :: objName
LOGICAL,INTENT(IN)                       :: isCommentOut
INTEGER,INTENT(IN)                       :: countFieldMin
INTEGER,INTENT(IN)                       :: countFieldMax
INTEGER,INTENT(IN)                       :: numberExpected

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na

IF (.NOT. ALLOCATED(objToProcess)) THEN
  ALLOCATE(objToProcess(sizeObjToProcess))
  numObjToProcess = 1
ELSE
  numObjToProcess = numObjToProcess + 1
  ! if larger then current size then make a temporary array of the same
  ! type and put stuff into it while reallocating the main array
  IF (numObjToProcess .GT. sizeObjToProcess) THEN
    ALLOCATE(objToProcessCopy(sizeObjToProcess))
    objToProcessCopy = objToProcess
    DEALLOCATE(objToProcess)
    ! double the size of the array
    ALLOCATE(objToProcess(sizeObjToProcess * 2))
    objToProcess(1:sizeObjToProcess) = objToProcessCopy
    DEALLOCATE(objToProcessCopy)
    sizeObjToProcess = sizeObjToProcess * 2
  END IF
END IF
objToProcess(numObjToProcess)%ucName = MakeUPPERCase(objName)
objToProcess(numObjToProcess)%ccName = objName
objToProcess(numObjToProcess)%commentOut = isCommentOut
objToProcess(numObjToProcess)%minFldCnt = countFieldMin
objToProcess(numobjToProcess)%maxFldCnt = countFieldMax
objToProcess(numobjToProcess)%expFldCnt = numberExpected
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE ReadWriteFileAndProcess
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    One of the main working routines of the program. Reads the
          !    input IDF one object at a time and writes it out with or
          !    without comment symbols. If one of the objects to hold
          !    it stores them in the main arrays.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: read_stat
CHARACTER(len=MaxLineLength) :: LineIn
INTEGER :: semiPos  !position of the semicolon
INTEGER :: exclPos  !position of the exclamation mark
INTEGER :: commaPos !position of the comma
LOGICAL :: isCommentedObject !flag if buffer test indicates buffer should be commented
LOGICAL :: isHeldObject      !flag if buffer test indicates should be saved for later
INTEGER :: objProcIndex !returned index for the objProc array for found object
INTEGER :: tpos     ! position of tab character in input line

CALL ClearBuffer
read_stat=0
DO WHILE (.not. is_iostat_end(read_stat)) !end of file
  READ(UNIT=InFH, FMT="(A)", IOSTAT=read_stat) LineIn
  IF (read_stat /= 0) LineIn = ''
  tpos=INDEX(LineIn,tabchar)
  DO WHILE (tpos > 0)
    LineIn(tpos:tpos) = ' '
    tpos=INDEX(LineIn,tabchar)
  END DO
  ! find the comment character, the object ending character and the field delimiter
  CALL FindDelimiters(LineIn,exclPos,semiPos,commaPos)
  CALL AddToBuffer(LineIn)  !add the single line to the buffer
  IF (semiPos .GT. 0) THEN
    ! if the buffer contains an object that should be held
    CALL testObjectInBuffer(isCommentedObject,isHeldObject,objProcIndex)
    CALL WriteBuffer(isCommentedObject)
    IF (isHeldObject) THEN
      CALL holdObjectInBuffer(objProcIndex)
    END IF
    CALL ClearBuffer
  END IF
  LineIn = ''
END DO
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE AddToBuffer(ToBuffer)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Add a line to the input buffer array. Simple except
          !    that the buffer can be resized for a large object.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=MaxLineLength), INTENT(IN)  :: ToBuffer

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
IF (.NOT. ALLOCATED(InBuffer)) THEN
  ALLOCATE(InBuffer(sizeInBuffer))
  numInBuffer = 1
ELSE
  numInBuffer = numInBuffer + 1
  ! if larger then current size then make a temporary array of the same
  ! type and put stuff into it while reallocating the main array
  IF (numInBuffer .GT. sizeInBuffer) THEN
    ALLOCATE(InBufferCopy(sizeInBuffer))
    InBufferCopy = InBuffer
    DEALLOCATE(InBuffer)
    ! double the size of the array
    ALLOCATE(InBuffer(sizeInBuffer * 2))
    InBuffer(1:sizeInBuffer) = InBufferCopy
    DEALLOCATE(InBufferCopy)
    sizeInBuffer = sizeInBuffer * 2
  END IF
END IF
InBuffer(numInBuffer) = ToBuffer
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE AddToErrMsg(ToErrMsgs,KindOfMessage)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   October 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Add a line to the error buffer.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN)  :: ToErrMsgs
INTEGER,INTENT(IN),OPTIONAL   :: KindOfMessage

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
IF (.NOT. ALLOCATED(ErrMsgs)) THEN
  ALLOCATE(ErrMsgs(sizeErrMsgs))
  numErrMsgs = 1
ELSE
  numErrMsgs = numErrMsgs + 1
  ! if larger then current size then make a temporary array of the same
  ! type and put stuff into it while reallocating the main array
  IF (numErrMsgs .GT. sizeErrMsgs) THEN
    ALLOCATE(ErrMsgsCopy(sizeErrMsgs))
    ErrMsgsCopy = ErrMsgs
    DEALLOCATE(ErrMsgs)
    ! double the size of the array
    ALLOCATE(ErrMsgs(sizeErrMsgs * 2))
    ErrMsgs(1:sizeErrMsgs) = ErrMsgsCopy
    DEALLOCATE(ErrMsgsCopy)
    sizeErrMsgs = sizeErrMsgs * 2
  END IF
END IF
ErrMsgs(numErrMsgs)%msgText = ToErrMsgs
IF (PRESENT(KindOfMessage)) THEN
  ErrMsgs(numErrMsgs)%msgKind = KindOfMessage
ELSE
  ErrMsgs(numErrMsgs)%msgKind = msgError
END IF
END SUBROUTINE

!-------------- --------------------------------------------------------------------
SUBROUTINE ClearBuffer
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Reset the size of the buffer and clear the buffer

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
numInBuffer = 0
IF (ALLOCATED(InBuffer)) THEN
  InBuffer = ''
END IF
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE WriteBuffer(isCommentOut)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    To

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
LOGICAL, INTENT(IN) :: isCommentOut

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
INTEGER :: iBuffer

DO iBuffer = 1, numInBuffer
  IF (isCommentOut) THEN
    !if line starts with a ! don't bother with another comment
    IF (InBuffer(iBuffer)(1:1) .NE. '!') THEN
      WRITE(UNIT=OutFH, FMT="(A)") '! ' // TRIM(InBuffer(iBuffer))
    ELSE
      WRITE(UNIT=OutFH, FMT="(A)") TRIM(InBuffer(iBuffer))
    END IF
  ELSE
    WRITE(UNIT=OutFH, FMT="(A)") TRIM(InBuffer(iBuffer))
  END IF
END DO
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE testObjectInBuffer(doComment,doHold,foundObjToProcess)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Check if the buffer includes an object that should be held and
          !    if the object should be commented out.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
LOGICAL, INTENT(OUT) :: doComment
LOGICAL, INTENT(OUT) :: doHold
INTEGER, INTENT(OUT) :: foundObjToProcess

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
INTEGER :: iBuffer
INTEGER :: iObjProc
INTEGER :: semiPos  !position of the semicolon
INTEGER :: exclPos  !position of the exclamation mark
INTEGER :: commaPos !position of the comma
CHARACTER(len=MaxAlphaLength) :: firstField
LOGICAL :: foundFirstField


! first assume that the buffer does not contain an object to hold
! until later
doHold = .FALSE.
doComment = .FALSE.
foundFirstField = .FALSE.
foundObjToProcess = 0
firstField = ''
DO iBuffer = 1, numInBuffer
  ! find the comment character, the object ending character and the field delimiter
  CALL FindDelimiters(InBuffer(iBuffer),exclPos,semiPos,commaPos)
  ! if end of object or delimiter is on the line then found the first
  ! field of the object which is the object name
  ! exit the look because we are just looking for the first field
  IF (commaPos .GT. 1) THEN
    firstField = ADJUSTL(InBuffer(iBuffer)(1:commaPos - 1))
    foundFirstField = .TRUE.
    EXIT
  ELSE IF (semiPos .GT. 1) THEN
    firstField = ADJUSTL(InBuffer(iBuffer)(1:semiPos - 1))
    foundFirstField = .TRUE.
    EXIT
  END IF
END DO
! if a first field was found then look for it in the objects to process list
IF  (foundFirstField) THEN
  DO iObjProc = 1, numObjToProcess
    IF (SameString(firstField,objToProcess(iObjProc)%ucName)) THEN
      foundObjToProcess = iObjProc
      EXIT
    END IF
  END DO
  !found the object in the list of objects to process
  IF (foundObjToProcess .GT. 0) THEN
    doHold = .TRUE.
    doComment = objToProcess(foundObjToProcess)%commentOut
  END IF
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE holdObjectInBuffer(foundObjToProcess)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    To move the field values from the buffer into the fldval array.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
INTEGER, INTENT(IN) :: foundObjToProcess

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
CHARACTER(len=MaxLineLength)  :: curLine
CHARACTER(len=MaxLineLength)  :: curLineNoComment
CHARACTER(len=MaxLineLength*numInBuffer)  :: EntireBuffer
INTEGER :: EntireBufferLen
INTEGER :: iBuffer
INTEGER :: exclPos
INTEGER :: commaPos
INTEGER :: semiPos
INTEGER :: wordStart
INTEGER :: wordEnd
CHARACTER(len=MaxAlphaLength) :: fldString
INTEGER :: jChar
INTEGER :: tabPos
INTEGER :: numberOfFields
INTEGER :: iPadFldVal
LOGICAL :: inObject

CALL incrementHeldObj
! set the active held object to the one found
HeldObj(numHeldObj)%objToProcessIndex = foundObjToProcess
! set the current held object as the active object
objToProcess(foundObjToProcess)%lastObjHeld = numHeldObj
! set the beginning of the fields to one plus the current last field
HeldObj(numHeldObj)%FirstFldVal = numFldVal + 1
! now process the object currently in the buffer
! put the entire object into a single string without comments
EntireBuffer = ''
inObject = .FALSE.
DO iBuffer = 1, numInBuffer
  curLine = ADJUSTL(InBuffer(iBuffer))
  ! remove tabs from line
  DO
    tabPos = INDEX(curLine, CHAR(9))
    IF (tabPos .GT. 1) THEN
      curLine(tabPos:tabPos) = ' '
    ELSE
      EXIT
    END IF
  END DO
  ! Fine the exclamation point, semicolon and comma
  CALL FindDelimiters(curLine,exclPos,semiPos,commaPos)
  ! remove portion after the comment character
  IF (exclPos .GT. 0) THEN
    curLineNoComment = curLine(1:exclPos - 1)
  ELSE
    curLineNoComment = curLine
  END IF
  ! if the first line of the object is found then
  ! set flag that it is in object
  IF ((semiPos .GT. 0) .OR. (commaPos .GT. 0)) THEN
    inObject = .TRUE.
  END IF
  IF (inObject) THEN
    IF (LEN_TRIM(curLineNoComment) .GT. 0) THEN
      IF ((commaPos .EQ. 0) .AND. (semiPos .EQ. 0)) THEN
        CALL WriteError('WARNING: comma or semicolon missing: ' // TRIM(curLineNoComment),msgWarning)
        curLineNoComment = TRIM(curLineNoComment) // ','
      END IF
    END IF
  END IF
  EntireBuffer = TRIM(EntireBuffer) // TRIM(curLineNoComment)
END DO
!get rid of the trailing semicolon
EntireBufferLen = LEN_TRIM(EntireBuffer)
IF (EntireBuffer(EntireBufferLen:EntireBufferLen) .EQ. ';') THEN
  EntireBuffer(EntireBufferLen:EntireBufferLen) = ' '
ELSE
  PRINT "(A)","Semicolon not found at end of the line: " // TRIM(curLine)
END IF
!now separate by comma as the delimiter
wordStart = 1
DO
  commaPos = INDEX(EntireBuffer, ',')
  IF (commaPos .GT. 1) THEN
    wordEnd = commaPos - 1
    fldString = TRIM(EntireBuffer(wordStart:wordEnd))
    !the next word will start after the comma
    wordStart = commaPos + 1
    !get rid of comma
    EntireBuffer(commaPos:commaPos) = ' '
    HeldObj(numHeldObj)%LastFldVal = AddFldVal(fldString)
  ELSE
    !no more commas
    fldString = TRIM(EntireBuffer(wordStart:))
    HeldObj(numHeldObj)%LastFldVal = AddFldVal(fldString)
    EXIT
  END IF
END DO
! test of number of fields found is between the minimum and maximum
! number of fields and if less than minimum fill object
! Note: Can subtract FirstFldVal from LastFldVal without adding one because the
! name of the object takes up one position in the FldVal array and is counted as the
! first field.  The offsets all start at one.
numberOfFields = HeldObj(numHeldObj)%LastFldVal - HeldObj(numHeldObj)%FirstFldVal
!old error message: 'WARNING: A HVACTemplate:Zone:VAVFanPowered object was found with the wrong number of fields.'
IF (numberOfFields .LT. objToProcess(foundObjToProcess)%minFldCnt) THEN
  CALL WriteError('WARNING: Not enough fields were found for the object: ' // TRIM(objToProcess(foundObjToProcess)%ccName) //  &
    ', Number expected=['//trim(IntToStr(objToProcess(foundObjToProcess)%expFldCnt))//  &
    '], Number found=['//trim(IntToStr(numberOfFields))//']' ,msgWarning)
ELSE IF (numberOfFields .GT. objToProcess(foundObjToProcess)%maxFldCnt) THEN
  CALL WriteError('SEVERE: Too many fields were found for the object: ' // TRIM(objToProcess(foundObjToProcess)%ccName) //   &
    ', Number expected=['//trim(IntToStr(objToProcess(foundObjToProcess)%expFldCnt))//  &
    '], Number found=['//trim(IntToStr(numberOfFields))//']' ,msgError)
ELSE
  ! Add blank fields to match the count with the maximum number of fields
  DO iPadFldVal = numberOfFields + 1, objToProcess(foundObjToProcess)%maxFldCnt
    HeldObj(numHeldObj)%LastFldVal = AddFldVal('')
  END DO
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE incrementHeldObj
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !   Increments and manages the HeldObj array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na

IF (.NOT. ALLOCATED(HeldObj)) THEN
  ALLOCATE(HeldObj(sizeHeldObj))
  numHeldObj = 1
ELSE
  numHeldObj = numHeldObj + 1
  ! if larger then current size then make a temporary array of the same
  ! type and put stuff into it while reallocating the main array
  IF (numHeldObj .GT. sizeHeldObj) THEN
    ALLOCATE(HeldObjCopy(sizeHeldObj))
    HeldObjCopy = HeldObj
    DEALLOCATE(HeldObj)
    ! double the size of the array
    ALLOCATE(HeldObj(sizeHeldObj * 2))
    HeldObj(1:sizeHeldObj) = HeldObjCopy
    DEALLOCATE(HeldObjCopy)
    sizeHeldObj = sizeHeldObj * 2
  END IF
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
INTEGER FUNCTION AddFldVal(stringIn)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !   Increments and manages the HeldObj array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN) :: stringIn    ! Input String

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
CHARACTER(len=LEN(stringIn)) :: inShiftLeft

inShiftLeft = ADJUSTL(stringIn)
IF (.NOT. ALLOCATED(FldVal)) THEN
  ALLOCATE(FldVal(sizeFldVal))
  numFldVal = 1
ELSE
  numFldVal = numFldVal + 1
  ! if larger then current size then make a temporary array of the same
  ! type and put stuff into it while reallocating the main array
  IF (numFldVal .GT. sizeFldVal) THEN
    ALLOCATE(FldValCopy(sizeFldVal))
    FldValCopy = FldVal
    DEALLOCATE(FldVal)
    ! double the size of the array
    ALLOCATE(FldVal(sizeFldVal +10000))
    FldVal(1:sizeFldVal) = FldValCopy
    DEALLOCATE(FldValCopy)
    sizeFldVal = sizeFldVal +10000 !* 2
  END IF
END IF
FldVal(numFldVal) = inShiftLeft !finally put the string in the array
AddFldVal = numFldVal        !return the index of the fldstring array
END FUNCTION

!----------------------------------------------------------------------------------
SUBROUTINE AddToIntList(ToList)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Adds an item to a list of integers. The list can be any sized
          !    so it is allocatable. The list of integers can be used for any
          !    reason but is usually used when a group of one type of object
          !    is referenced by another object (zones by systems).

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
INTEGER, INTENT(IN)  :: ToList

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
IF (.NOT. ALLOCATED(IntList)) THEN
  ALLOCATE(IntList(sizeIntList))
  IntList = 0
  numIntList = 1
ELSE
  numIntList = numIntList + 1
  ! if larger then current size then make a temporary array of the same
  ! type and put stuff into it while reallocating the main array
  IF (numIntList .GT. sizeIntList) THEN
    ALLOCATE(IntListCopy(sizeIntList))
    IntListCopy = IntList
    DEALLOCATE(IntList)
    ! double the size of the array
    ALLOCATE(IntList(sizeIntList * 2))
    IntList = 0
    IntList(1:sizeIntList) = IntListCopy
    DEALLOCATE(IntListCopy)
    sizeIntList = sizeIntList * 2
  END IF
END IF
IntList(numIntList) = ToList
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE ClearIntList
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Clears list of integers. The list can be any sized
          !    so it is allocatable. The list of integers can be used for any
          !    reason but is usually used when a group of one type of object
          !    is referenced by another object (zones by systems).

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
IF (ALLOCATED(IntList)) THEN
  numIntList = 0
  IntList = 0
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
INTEGER FUNCTION repeatStringInIntList()
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2012
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Returns zero if all of the integers in the IntList are values of indexes to
          !    the FldVal array that contain strings that are not duplicated. Used primarily
          !    to make sure names of objects are not repeated.
          !    If a repeat does occur, returns index value of FldVal that has repeated string.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine

          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
INTEGER :: i
INTEGER :: j
LOGICAL :: isFldValIndex

repeatStringInIntList = 0 !assume all are unique unless a duplicate is found
! first check if all items in IntList are reasonable indexes of FldVal array
isFldValIndex = .TRUE.
DO i = 1,numIntList
  IF ((IntList(i) .LT. 1) .OR. (IntList(i) .GT. numFldVal)) THEN
    isFldValIndex = .FALSE.
    EXIT
  ENDIF
END DO
IF (.NOT. isFldValIndex) THEN
  ! this should not normally occur so the error message is aimed at developers.
  CALL WriteError('Invalid indices found when searching for repeated object names.')
ELSE
  DO i = 1,numIntList
    DO j = i + 1, numIntList
      IF (SameString(FldVal(IntList(i)),FldVal(IntList(j)))) THEN
        repeatStringInIntList = IntList(j)
        EXIT !stop comparing already found a repeat. Leave inner loop.
      END IF
    END DO
    !get out of outter loop
    IF (repeatStringInIntList .NE. 0) EXIT
  END DO
END IF
END FUNCTION

!----------------------------------------------------------------------------------
SUBROUTINE AddToNumSortList(NumToList,PtrToList)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Adds an item to a list of real numbers that will be later sorted.
          !    The list can be any sized so it is allocatable.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
REAL, INTENT(IN)     :: NumToList
INTEGER, INTENT(IN)  :: PtrToList

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
IF (.NOT. ALLOCATED(NumSortList)) THEN
  ALLOCATE(NumSortList(sizeNumSortList))
  NumSortList%numToSort = 0
  NumSortList%sortedPtr = 0
  numNumSortList = 1
ELSE
  numNumSortList = numNumSortList + 1
  ! if larger then current size then make a temporary array of the same
  ! type and put stuff into it while reallocating the main array
  IF (numNumSortList .GT. sizeNumSortList) THEN
    ALLOCATE(NumSortListCopy(sizeNumSortList))
    NumSortListCopy = NumSortList
    DEALLOCATE(NumSortList)
    ! double the size of the array
    ALLOCATE(NumSortList(sizeNumSortList * 2))
    NumSortList%numToSort = 0
    NumSortList%sortedPtr = 0
    NumSortList(1:sizeNumSortList) = NumSortListCopy
    DEALLOCATE(NumSortListCopy)
    sizeNumSortList = sizeNumSortList * 2
  END IF
END IF
NumSortList(numNumSortList)%numToSort = NumToList
NumSortList(numNumSortList)%sortedPtr = PtrToList
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE ClearNumSortList
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Clears list of numbers to sort.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
IF (ALLOCATED(NumSortList)) THEN
  numNumSortList = 0
  NumSortList%numToSort = 0.0
  NumSortList%sortedPtr = 0
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE SortIncreasingNumList
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Sort the number sort list in increasing order

          ! METHODOLOGY EMPLOYED:
          !    Simple bubble sort because don't expect the
          !    list to be very long.

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
LOGICAL :: exchanged
REAL    :: tempReal
INTEGER :: tempInteger
INTEGER :: i

IF (numNumSortList .GE. 2) THEN
  DO
    exchanged = .FALSE.
    DO i = 1, numNumSortList - 1
      IF (NumSortList(i)%numToSort .GT. NumSortList(i + 1)%numToSort) THEN
        !put higher value in temp variables
        tempReal = NumSortList(i)%numToSort
        tempInteger = NumSortList(i)%sortedPtr
        !move lower number into lower position
        NumSortList(i)%numToSort = NumSortList(i + 1)%numToSort
        NumSortList(i)%sortedPtr = NumSortList(i + 1)%sortedPtr
        !put higher number into higher position
        NumSortList(i + 1)%numToSort = tempReal
        NumSortList(i + 1)%sortedPtr = tempInteger
        exchanged = .TRUE.
      END IF
    END DO
    ! if no more exhanges then the list must be in order
    IF (.NOT. exchanged) EXIT
  END DO
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE AddToStrList(ToList,SubListHandle)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Adds an item to a list of strings. The list can be any sized
          !    so it is allocatable. Used to hold names of coils.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(*), INTENT(IN)  :: ToList
INTEGER,INTENT(IN)                             :: SubListHandle

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
IF (.NOT. ALLOCATED(StrList)) THEN
  ALLOCATE(StrList(sizeStrList))
  StrList%StrToSave = ''
  StrList%Handle = 0
  numStrList = 1
ELSE
  numStrList = numStrList + 1
  ! if larger then current size then make a temporary array of the same
  ! type and put stuff into it while reallocating the main array
  IF (numStrList .GT. sizeStrList) THEN
    ALLOCATE(StrListCopy(sizeStrList))
    StrListCopy = StrList
    DEALLOCATE(StrList)
    ! double the size of the array
    ALLOCATE(StrList(sizeStrList * 2))
    StrList%StrToSave = ''
    StrList%Handle = 0
    StrList(1:sizeStrList) = StrListCopy
    DEALLOCATE(StrListCopy)
    sizeStrList = sizeStrList * 2
  END IF
END IF
StrList(numStrList)%StrToSave = ToList // REPEAT(' ',MaxAlphaLength)
StrList(numStrList)%Handle = SubListHandle
END SUBROUTINE

!----------------------------------------------------------------------------------
REAL FUNCTION StringToReal(stringIn)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    To convert from a string to a real

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine

          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN) :: stringIn

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na

IF (LEN_TRIM(stringIn) .GE. 1) THEN
  IF (VERIFY(TRIM(stringIn),'-0123456789.E+') .EQ. 0) THEN
    READ(stringIn,*,ERR=2222) StringToReal
  ELSE
    StringToReal = 0
  END IF
ELSE
  StringToReal = 0
END IF
RETURN

2222 StringToReal = 0
END FUNCTION

!----------------------------------------------------------------------------------
FUNCTION IntToStr(intIn) RESULT (stringOut)
          ! SUBROUTINE INFORMATION:
          !       AUTHOR         Jason Glazer
          !       DATE WRITTEN   August 2003
          !       MODIFIED       na
          !       RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !   Abstract away the internal write concept

IMPLICIT NONE

INTEGER, INTENT(IN)    :: intIn
CHARACTER(LEN=12)      :: stringOut
WRITE(FMT=*, UNIT=stringOut) intIn
stringOut=Adjustl(stringOut)
END FUNCTION

!----------------------------------------------------------------------------------
FUNCTION RealToStr(realIn) RESULT (stringOut)
          ! SUBROUTINE INFORMATION:
          !       AUTHOR         Jason Glazer
          !       DATE WRITTEN   February 2012
          !       MODIFIED       na
          !       RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !   Abstract away the internal write concept

IMPLICIT NONE

REAL, INTENT(IN)      :: realIn
CHARACTER(LEN=20)      :: stringOut
WRITE(FMT=*, UNIT=stringOut) realIn
stringOut=Adjustl(stringOut)
END FUNCTION

!----------------------------------------------------------------------------------
FUNCTION StringToInt(valueString) RESULT (valueInt)
! Convert string containing a number into an
! integer.
!      Jason Glazer - July 2007
IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine
CHARACTER(len=*), INTENT(IN) :: valueString
INTEGER                      :: valueInt
IF (LEN_TRIM(valueString) .GE. 1) THEN
  IF (VERIFY(TRIM(valueString),'+-0123456789') .EQ. 0) THEN
    READ(valueString,*) valueInt
  ELSE
    valueInt = 0
  END IF
ELSE
  valueInt = 0
END IF
END FUNCTION

!----------------------------------------------------------------------------------
LOGICAL FUNCTION isNumeric(valueString)
! Check to see if the characters in the string were all numeric
IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine
CHARACTER(len=*), INTENT(IN) :: valueString
IF (LEN_TRIM(valueString) .GE. 1) THEN
  IF (VERIFY(TRIM(valueString),'+-0123456789E.') .EQ. 0) THEN
    isNumeric = .TRUE.
  ELSE
    isNumeric = .FALSE.
  END IF
ELSE
  isNumeric = .FALSE.
END IF
END FUNCTION

!----------------------------------------------------------------------------------
FUNCTION MakeUPPERCase(InputString) RESULT (ResultString)

          ! FUNCTION INFORMATION:
          !       AUTHOR         Linda K. Lawrie
          !       DATE WRITTEN   September 1997
          !       MODIFIED       na
          !       RE-ENGINEERED  July 2010

          ! PURPOSE OF THIS SUBROUTINE:
          ! This function returns the Upper Case representation of the InputString.

          ! METHODOLOGY EMPLOYED:
          ! Uses the Intrinsic SCAN function to scan the lowercase representation of
          ! characters (DataStringGlobals) for each character in the given string.

          ! REFERENCES:
          ! na

          ! USE STATEMENTS:
          ! na

  IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! FUNCTION ARGUMENT DEFINITIONS:
   CHARACTER(len=*), INTENT(IN) :: InputString    ! Input String
   CHARACTER(len=len(InputString)) ResultString ! Result String, string is limited to
                                                  ! MaxInputLineLength because of PowerStation Compiler
                                                  ! otherwise could say (CHARACTER(len=LEN(InputString))


          ! FUNCTION PARAMETER DEFINITIONS:
          ! na

          ! INTERFACE BLOCK SPECIFICATIONS
          ! na

          ! DERIVED TYPE DEFINITIONS
          ! na

          ! FUNCTION LOCAL VARIABLE DECLARATIONS:
  INTEGER i                  ! Loop Counter
  INTEGER Pos                ! Position in String representation
  INTEGER LengthInputString  ! Length (trimmed) of InputString
  INTEGER :: CurCharVal

    ResultString=InputString

    do i = 1, LEN_TRIM(InputString)
      curCharVal = ICHAR(InputString(i:i))
      SELECT CASE (curCharVal)
        CASE (97:122,224:255) !lowercase ASCII and accented characters
          ResultString(i:i) = CHAR(curCharVal-32)
        CASE DEFAULT
      END SELECT
    end do

  RETURN

END FUNCTION MakeUPPERCase

!----------------------------------------------------------------------------------
LOGICAL FUNCTION SameString(TestString1,TestString2)
          ! FUNCTION INFORMATION:
          !       AUTHOR         Linda K. Lawrie
          !       DATE WRITTEN   November 1997
          !       MODIFIED       na
          !       RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          ! This function returns true if the two strings are equal (case insensitively)

          ! METHODOLOGY EMPLOYED:
          ! Make both strings uppercase.  Do internal compare.

          ! REFERENCES:
          ! na

          ! USE STATEMENTS:
          ! na

  IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! FUNCTION ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN) :: TestString1  ! First String to Test
CHARACTER(len=*), INTENT(IN) :: TestString2  ! Second String to Test


          ! FUNCTION PARAMETER DEFINITIONS:
          ! na

          ! INTERFACE BLOCK SPECIFICATIONS
          ! na

          ! DERIVED TYPE DEFINITIONS
          ! na

          ! FUNCTION LOCAL VARIABLE DECLARATIONS:
          ! na

  IF (LEN_TRIM(TestString1) /= LEN_TRIM(TestString2)) THEN
    SameString=.false.
  ELSEIF (LEN(TestString1) <= MaxLineLength .and. LEN(TestString2) <= MaxLineLength) THEN
    ! This test (MaxInputLineLength) is necessary because of PowerStation Compiler
    SameString=MakeUPPERCase(TestString1) == MakeUPPERCase(TestString2)
  ELSE
!    CALL ShowFatalError('SameString aborting -- input strings too long')
    SameString=.false.
  ENDIF
END FUNCTION SameString

!----------------------------------------------------------------------------------
SUBROUTINE FindDelimiters(StringIn,posExcl,posSemi,posComma)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    To look for delimiters in a string of characters

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN) :: StringIn  ! string to search
INTEGER,INTENT(OUT)          :: posExcl   ! position of the exclamation point
INTEGER,INTENT(OUT)          :: posSemi   ! position of the semicolon
INTEGER,INTENT(OUT)          :: posComma  ! position of the comma


          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
posExcl = INDEX(StringIn, '!')  ! comment
posSemi = INDEX(StringIn,';')   ! object end
posComma = INDEX(StringIn, ',') ! field delimit
! if the delimiter is after the
IF (posExcl .GE. 1) THEN
  IF (posSemi .GT. posExcl) THEN
    posSemi = 0
  END IF
  IF (posComma .GT. posExcl) THEN
    posComma = 0
  END IF
END IF
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE AddAlwaysSchedule(alwaysValue)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    To

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN) :: alwaysValue

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
INTEGER :: found
INTEGER :: iSch
CHARACTER(5) :: valString

IF (.NOT. ALLOCATED(alwaysSchedule)) THEN
  ALLOCATE(alwaysSchedule(sizeAlwaysSchedule))
  numAlwaysSchedule = 0
  !Make generic schedule type
  CALL CreateNewObj('ScheduleTypeLimits')
  CALL AddToObjStr('Name','HVACTemplate Any Number',.TRUE.)
END IF
found = 0
DO iSch = 1, numAlwaysSchedule
  IF (alwaysSchedule(iSch) .EQ. alwaysValue) THEN
    found = iSch
    EXIT
  END IF
END DO
! if it was not found in the existing array then create a new
! schedule
IF (found .EQ. 0) THEN
  numAlwaysSchedule = numAlwaysSchedule + 1
  ! if larger then current size then make a temporary array of the same
  ! type and put stuff into it while reallocating the main array
  IF (numAlwaysSchedule .GT. sizeAlwaysSchedule) THEN
    ALLOCATE(alwaysScheduleCopy(sizeAlwaysSchedule))
    alwaysScheduleCopy = alwaysSchedule
    DEALLOCATE(alwaysSchedule)
    ! double the size of the array
    ALLOCATE(alwaysSchedule(sizeAlwaysSchedule * 2))
    alwaysSchedule(1:sizeAlwaysSchedule) = alwaysScheduleCopy
    DEALLOCATE(alwaysScheduleCopy)
    sizeAlwaysSchedule = sizeAlwaysSchedule * 2
  END IF
  alwaysSchedule(numAlwaysSchedule) = alwaysValue
  ! Create a schedule that always has the same value
  CALL CreateNewObj('Schedule:Compact')
  CALL AddToObjStr('Name','HVACTemplate-Always ' // alwaysValue)
  CALL AddToObjStr('Schedule Type Limits Name','HVACTemplate Any Number')
  CALL AddToObjStr('Field 1','Through: 12/31')
  CALL AddToObjStr('Field 2','For: AllDays')
  CALL AddToObjStr('Field 3','Until: 24:00')
  CALL AddToObjStr('Field 4',alwaysValue,.TRUE.)
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE CreateNewObj(nameOfObj, isLastItem)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    routine that prints to the IDF file the name of the object
          !    this is abstracted to a simple procedure so that the formatting
          !    can be easily changed

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN) :: nameOfObj
LOGICAL, OPTIONAL,INTENT(IN) :: isLastItem

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
LOGICAL :: useSemi
INTEGER :: iObj
INTEGER :: found

IF (PRESENT(isLastItem)) THEN
  useSemi = isLastItem
ELSE
  useSemi = .FALSE.
END IF
WRITE(UNIT=OutFH, FMT="(A)") ''
IF (useSemi) THEN
  WRITE(UNIT=OutFH, FMT="(A)") TRIM(nameOfObj) // ';'
ELSE
  WRITE(UNIT=OutFH, FMT="(A)") TRIM(nameOfObj) // ','
END IF
! if checking output is enabled (usually for developers only)
! then generate a warning if the name of the object is not
! a valid name in the IDD.
IF (checkOutputFieldNamesAgainstIDD) THEN
  found = 0
  DO iObj = 1,objCount
    IF (SameString(iddObj(iObj)%name,nameOfObj)) THEN
      found = iObj
      EXIT
    END IF
  END DO
  IF (found .EQ. 0) THEN !did not find the object name
    CALL WriteError('Object name not found: ' // TRIM(nameOfObj))
  END IF
  checkObjectIndex =  found !save the object index for when adding field names
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE AddToObjStr(fieldName,fieldValue,isLastItem)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    routine that prints to the IDF file the value of a field and the
          !    field name (as a commment) this is for string values
          !    this is abstracted to a simple procedure so that the formatting
          !    can be easily changed

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN) :: fieldName
CHARACTER(len=*), INTENT(IN) :: fieldValue
LOGICAL, OPTIONAL,INTENT(IN) :: isLastItem

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
LOGICAL :: useSemi

IF (PRESENT(isLastItem)) THEN
  useSemi = isLastItem
ELSE
  useSemi = .FALSE.
END IF
IF (len_trim(fieldValue) < 55) THEN
  IF (useSemi) THEN
    WRITE(UNIT=OutFH, FMT="(2X,A,T60,A)") TRIM(fieldValue) // ';','!- ' // TRIM(fieldName)
  ELSE
    WRITE(UNIT=OutFH, FMT="(2X,A,T60,A)") TRIM(fieldValue) // ',','!- ' // TRIM(fieldName)
  END IF
ELSE
  IF (useSemi) THEN
    WRITE(UNIT=OutFH, FMT="(2X,A,2X,A)") TRIM(fieldValue) // ';','!- ' // TRIM(fieldName)
  ELSE
    WRITE(UNIT=OutFH, FMT="(2X,A,2X,A)") TRIM(fieldValue) // ',','!- ' // TRIM(fieldName)
  END IF
ENDIF
CALL validateFieldNameInIDD(fieldName)
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE AddToObjFld(fieldName,indexFldVal,suffix, isLastItem)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    routine that prints to the IDF file the item out of the FldVal array
          !    plus a possible suffix and the field name (as a commment) this is for
          !    string values.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN) :: fieldName
INTEGER, INTENT(IN)          :: indexFldVal
CHARACTER(len=*), INTENT(IN) :: suffix
LOGICAL, OPTIONAL,INTENT(IN) :: isLastItem

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
LOGICAL :: useSemi

IF (PRESENT(isLastItem)) THEN
  useSemi = isLastItem
ELSE
  useSemi = .FALSE.
END IF
IF (len_trim(TRIM(FldVal(indexFldVal))//TRIM(suffix)) < 55) THEN
  IF (useSemi) THEN
    WRITE(UNIT=OutFH, FMT="(2X,A,T60,A)") TRIM(FldVal(indexFldVal)) // &
                 TRIM(suffix) // ';','!- ' // TRIM(fieldName)
  ELSE
    WRITE(UNIT=OutFH, FMT="(2X,A,T60,A)") TRIM(FldVal(indexFldVal)) // &
                 TRIM(suffix) // ',','!- ' // TRIM(fieldName)
  END IF
ELSE
  IF (useSemi) THEN
    WRITE(UNIT=OutFH, FMT="(2X,A,2X,A)") TRIM(FldVal(indexFldVal)) // &
                 TRIM(suffix) // ';','!- ' // TRIM(fieldName)
  ELSE
    WRITE(UNIT=OutFH, FMT="(2X,A,2X,A)") TRIM(FldVal(indexFldVal)) // &
                 TRIM(suffix) // ',','!- ' // TRIM(fieldName)
  END IF
END IF
CALL validateFieldNameInIDD(fieldName)
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE AddToObjNum(fieldName,fieldNum,isLastItem)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !   routine that prints to the IDF file the value of a field and the
          !   field name (as a commment) this is for numeric values
          !   this is abstracted to a simple procedure so that the formatting
          !   can be easily changed

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN) :: fieldName
REAL, INTENT(IN)             :: fieldNum
LOGICAL, OPTIONAL,INTENT(IN) :: isLastItem

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
LOGICAL :: useSemi

IF (PRESENT(isLastItem)) THEN
  useSemi = isLastItem
ELSE
  useSemi = .FALSE.
END IF
IF (useSemi) THEN
  WRITE(UNIT=OutFH, FMT="(2X,A,A,T60,A)") trim(rRoundSigDigits(fieldNum,6)),';','!- ' // TRIM(fieldName)
ELSE
  WRITE(UNIT=OutFH, FMT="(2X,A,A,T60,A)") trim(rRoundSigDigits(fieldNum,6)),',','!- ' // TRIM(fieldName)
END IF
CALL validateFieldNameInIDD(fieldName)
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE validateFieldNameInIDD(fieldName)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   March 2012
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !   Used for developers only, this routine flags when
          !   field names are not consistent with the IDD. It is
          !   only called when checkOutputFieldNamesAgainstIDD
          !   is set to TRUE.
          !   Uses checkObjectIndex which is set to the latest
          !   object index in CreateNewObj.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN) :: fieldName

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
INTEGER :: jFld
INTEGER :: found
CHARACTER(len=MaxAlphaLength) :: FldNoUnits
INTEGER :: findBracket

IF (checkOutputFieldNamesAgainstIDD) THEN
  IF (checkObjectIndex .NE. 0) THEN
    IF ((iddObj(checkObjectIndex)%firstFld .GE. 1) .AND. (iddObj(checkObjectIndex)%lastFld .GE. 1)) THEN
      findBracket = INDEX(fieldName,'{')
      IF (findBracket .GT.1) THEN
        FldNoUnits = TRIM(fieldName(1:(findBracket-1)))
      ELSE
        FldNoUnits = fieldName
      END IF
      found = 0
      DO jFld = iddObj(checkObjectIndex)%firstFld,iddObj(checkObjectIndex)%lastFld
        IF (SameString(iddFld(jFld)%name,FldNoUnits)) THEN
          found = jFld
          EXIT
        END IF
      END DO
      IF (found .EQ. 0) THEN
        CALL WriteError('For object: "' // TRIM(iddObj(checkObjectIndex)%name) // '" field name not found: "' // &
         TRIM(FldNoUnits) // '"')
      END IF
    END IF
  END IF
END IF
END SUBROUTINE


FUNCTION rRoundSigDigits(RealValue,SigDigits) RESULT(OutputString)

          ! FUNCTION INFORMATION:
          !       AUTHOR         Linda K. Lawrie
          !       DATE WRITTEN   March 2002
          !       MODIFIED       na
          !       RE-ENGINEERED  na

          ! PURPOSE OF THIS FUNCTION:
          ! This function accepts a number as parameter as well as the number of
          ! significant digits after the decimal point to report and returns a string
          ! that is appropriate.

          ! METHODOLOGY EMPLOYED:
          ! na

          ! REFERENCES:
          ! na

          ! USE STATEMENTS:
          ! na

  IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine

          ! FUNCTION ARGUMENT DEFINITIONS:
  REAL, INTENT(IN) :: RealValue
  INTEGER, INTENT(IN) :: SigDigits
  CHARACTER(len=30) OutputString

          ! FUNCTION PARAMETER DEFINITIONS:
  CHARACTER(len=11) :: DigitChar='01234567890'

          ! INTERFACE BLOCK SPECIFICATIONS
          ! na

          ! DERIVED TYPE DEFINITIONS
          ! na

          ! FUNCTION LOCAL VARIABLE DECLARATIONS:
  INTEGER DotPos               ! Position of decimal point in original string
  INTEGER EPos                 ! Position of E in original string format xxEyy
  INTEGER TPos                 ! Position of Testchar in Digit string
  INTEGER NPos                 ! Position of "next" char in Digit String
  INTEGER TPos1                ! Position of "next" char rounded in Digit string
  INTEGER SPos                 ! Actual string position being replaced
  INTEGER SLen                 ! Length of String (w/o E part)
  CHARACTER(len=30) String     ! Working string
  CHARACTER(len=10) EString    ! E string retained from original string
  CHARACTER(len=1) TestChar    ! Test character (digit) for rounding, if position in digit string > 5 (digit is 5 or greater)
                               ! then will round
  CHARACTER(len=1) Char2Rep    ! Character (digit) to be replaced
  LOGICAL IncludeDot           ! True when decimal point output

  IF (RealValue /= 0.0) THEN
    WRITE(String,*) RealValue
  ELSE
    String='0.000000000000000000000000000'
  ENDIF
  EPos=INDEX(String,'E')
  IF (EPos > 0) THEN
    EString=String(EPos:)
    String(Epos:)=' '
  ELSE
    EString=' '
  ENDIF

  DotPos=INDEX(String,'.')
  TestChar=String(DotPos+SigDigits+1:DotPos+SigDigits+1)
  TPos=INDEX(DigitChar,TestChar)

  IF (SigDigits == 0) THEN
    SPos=DotPos-1
  ELSE
    SPos=DotPos+SigDigits
  ENDIF

  IF (TPos > 5) THEN  ! Must round to next Digit
    Char2Rep=String(SPos:SPos)
    NPos=INDEX(DigitChar,Char2Rep)
    String(SPos:SPos)=DigitChar(NPos+1:NPos+1)
    DO WHILE (NPos == 10)
        ! Must change other char too
      IF (SigDigits == 1) THEN
        TestChar=String(SPos-2:SPos-2)
        IF (TestChar == '.') THEN
          TestChar=String(SPos-3:SPos-3)
          SPos=SPos-2
        ENDIF
        IF (TestChar == ' ') TestChar='0'  ! all 999s
        TPos1=INDEX(DigitChar,TestChar)
        String(SPos-2:SPos-2)=DigitChar(TPos1+1:TPos1+1)
      ELSE
        TestChar=String(SPos-1:SPos-1)
        IF (TestChar == '.') THEN
          TestChar=String(SPos-2:SPos-2)
          SPos=SPos-1
        ENDIF
        IF (TestChar == ' ') TestChar='0'  ! all 999s
        TPos1=INDEX(DigitChar,TestChar)
        String(SPos-1:SPos-1)=DigitChar(TPos1+1:TPos1+1)
      ENDIF
      SPos=SPos-1
      NPos=TPos1
    ENDDO
  ENDIF

  SLen=LEN_TRIM(String)
  IF (SigDigits > 0 .or. EString /= ' ') THEN
    IncludeDot=.true.
  ELSE
    IncludeDot=.false.
  ENDIF
  IF (IncludeDot) THEN
    String=String(1:MIN(DotPos+SigDigits,SLen))//EString
  ELSE
    String=String(1:DotPos-1)
  ENDIF
  IF (IsNAN(RealValue)) THEN
    String='NAN'
  ENDIF
  OutputString=ADJUSTL(String)

  RETURN

END FUNCTION rRoundSigDigits


!----------------------------------------------------------------------------------
SUBROUTINE WriteComment(commentString)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Write a comment to the IDF output file.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN) :: commentString

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na

WRITE(UNIT=OutFH, FMT="(A)") '! ' // TRIM(commentString)
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE WriteError(errorString,warningOrError)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Write a comment to the IDF output file.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN) :: errorString
INTEGER, INTENT(IN), OPTIONAL :: warningOrError

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na

WRITE(UNIT=OutFH, FMT="(A)") '! '
WRITE(UNIT=OutFH, FMT="(A)") '! ' // TRIM(ErrorString)
WRITE(UNIT=ErrFH, FMT="(A)") 'ExpandObjects: ' // TRIM(ErrorString)
IF (PRESENT (warningOrError)) THEN
  CALL AddToErrMsg(ErrorString,warningOrError)
ELSE
  CALL AddToErrMsg(ErrorString)
END IF
errorCondition = .TRUE.
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE validChoice(fldValIndx,choice1,choice2,objName,fldName)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Write a comment to the IDF output file.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
INTEGER, INTENT(IN) ::          fldValIndx
CHARACTER(len=*), INTENT(IN) :: choice1
CHARACTER(len=*), INTENT(IN) :: choice2
CHARACTER(len=*), INTENT(IN) :: objName
CHARACTER(len=*), INTENT(IN) :: fldName

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
LOGICAL :: isValid

isValid = .FALSE.
IF (SameString(FldVal(fldValIndx),choice1)) isValid = .TRUE.
IF (SameString(FldVal(fldValIndx),choice2)) isValid = .TRUE.
IF (.NOT. isValid) THEN
  CALL WriteError('Invalid string in HVACTemplate:' // TRIM(objName) // ' for field ' // TRIM(fldName) // &
       '.  The value entered is: '// TRIM(FldVal(fldValIndx)) // ' and the choices are:' // TRIM(choice1) // &
       ' and ' // TRIM(choice2))
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE CountOldObj(objName, objPointer,objCount)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    To count the number of instances of an object with a specific
          !    name that occured in the original file.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine

          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN) :: objName
INTEGER, INTENT(OUT)         :: objPointer
INTEGER, INTENT(OUT)         :: objCount

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
INTEGER :: foundObjToProc
INTEGER :: countOfHeld
INTEGER :: iObjToProc
INTEGER :: iHeldObj

foundObjToProc = 0
DO iObjToProc = 1, numObjToProcess
  IF (SameString(objName,objToProcess(iObjToProc)%ucName)) THEN
    foundObjToProc = iObjToProc
    EXIT !only need to find it then no longer need to loop
  END IF
END DO
! if it is found then search through the held object array for
! instances of that index.
IF (foundObjToProc .GT. 0) THEN
  countOfHeld = 0
  DO iHeldObj = 1, numHeldObj
    IF (HeldObj(iHeldObj)%ObjToProcessIndex .EQ. foundObjToProc) THEN
      countOfHeld = countOfHeld + 1 !if one is found increment counter
    END IF
  END DO
  objPointer = foundObjToProc ! return the objToProcess index to use to search
  objCount = countOfHeld
ELSE
  objPointer = 0
  objCount = 0
END IF
HeldObjSearchIndex = 0 !used as a pointer for the array
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE NextOldObj(objPointer,startFldVal,endFldVal)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    To retrieve the next instance of an object in the held obj array
          !    and return the first and last fields for that instance.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
INTEGER, INTENT(IN)  :: objPointer
INTEGER, INTENT(OUT) :: startFldVal
INTEGER, INTENT(OUT) :: endFldVal

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
INTEGER  :: countThroughLoop

!initialize the return values to zero to indicate not found
startFldVal = 0
endFldVal = 0
! loop through the array only once
DO countThroughLoop = 1, numHeldObj
  !increment the held object array and circle around to the beginning
  !of the array if reach the end.
  HeldObjSearchIndex = HeldObjSearchIndex + 1
  IF (HeldObjSearchIndex .GT. numHeldObj) THEN
    HeldObjSearchIndex = 1
  END IF
  IF (objPointer .EQ. HeldObj(HeldObjSearchIndex)%ObjToProcessIndex) THEN
    startFldVal = HeldObj(HeldObjSearchIndex)%FirstFldVal
    endFldVal = HeldObj(HeldObjSearchIndex)%LastFldVal
    EXIT
  END IF
END DO
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE SetIfBlank(fldValIndex,valueIfBlank)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   October 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    If a field is blank set the default value.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
INTEGER, INTENT(IN)  :: fldValIndex
CHARACTER(len=*), INTENT(IN) :: valueIfBlank

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
IF (TRIM(FldVal(fldValIndex)) .EQ. '') THEN
  FldVal(fldValIndex) = valueIfBlank
END IF
END SUBROUTINE

SUBROUTINE WriteOrigObj(nameOfObj, fileHandle, startIndex, endIndex)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2010
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Echo all the fields of an object to an output file
          !    with no field names.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine

          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN) :: nameOfObj
INTEGER, INTENT(IN)  :: fileHandle
INTEGER, INTENT(IN)  :: startIndex
INTEGER, INTENT(IN)  :: endIndex

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
INTEGER :: i

WRITE(UNIT=fileHandle, FMT="(A)") ''
WRITE(UNIT=fileHandle, FMT="(A)") TRIM(nameOfObj) // ','
IF ((endIndex - 1) .GE. startIndex) THEN
  DO i = startIndex,endIndex - 1
    WRITE(UNIT=fileHandle, FMT="(2X,A)") TRIM(FldVal(i)) // ','
  END DO
END IF
WRITE(UNIT=fileHandle, FMT="(2X,A)") TRIM(FldVal(endIndex)) // ';'
END SUBROUTINE

SUBROUTINE WriteGroupOfOrigObj(nameOfObj, fileHandle, baseArray, numObjects, lastFieldCount)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2010
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    For a given kind of object, echo all the fields of an object to an output file
          !    with no field names.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine

          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN) :: nameOfObj
INTEGER, INTENT(IN)  :: fileHandle
INTEGER, DIMENSION(:), INTENT(IN)  :: baseArray
INTEGER, INTENT(IN)  :: numObjects
INTEGER, INTENT(IN)  :: lastFieldCount

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
INTEGER :: base
INTEGER :: i

DO i = 1, numObjects
  base = baseArray(i)
  CALL WriteOrigObj(nameOfObj, fileHandle, base + 1, base + lastFieldCount)
END DO
END SUBROUTINE


!----------------------------------------------------------------------------------
!----------------------------------------------------------------------------------
!
!   READ IDD FILE AND CONFIRM RANGE ROUTINES
!
!----------------------------------------------------------------------------------
!----------------------------------------------------------------------------------

!----------------------------------------------------------------------------------
SUBROUTINE locateIDDfile
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2012
          !    MODIFIED       na
          !    RE-ENGINEERED

          ! PURPOSE OF THIS SUBROUTINE:
          !   Locate the Energy+.IDD file. Look for the Energy+.ini first and use
          !   the directory indicated by [program]. If this fails assume the
          !   Energy+.idd file is located in the current directory.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine

          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
INTEGER :: status = 0
LOGICAL :: iniExists = .FALSE.
LOGICAL :: iddExists = .FALSE.
INTEGER :: readStatus
CHARACTER(len=200) :: iddPath=' '
LOGICAL :: iddPathFound = .FALSE.
INTEGER :: iddPathLength = 0
CHARACTER(len=200) :: inputLine

!assume the idd cannot be read unless it is successfully found
iddCanBeRead = .FALSE.
!the default location of the Energy+.idd would be same directory as ExpandObjects
!this might get changed if ini indicates a different directory
IddFileName = 'Energy+.idd'
!the Energy+.ini, if it exists, should be in the same directory as ExpandObjects
IniFileName = 'Energy+.ini'
!see if the ini file exists
INQUIRE(FILE=IniFileName,EXIST=iniExists)
IF (iniExists) THEN
!  OPEN (unit=IniFH, file=TRIM(IniFileName), action='READ',iostat=status)
  OPEN (unit=IniFH, file=TRIM(IniFileName), action='READ',iostat=status, iomsg=IOErrorMsg)
  IF (status .NE. 0) THEN
!    IOErrorMsg=ADJUSTL(IntToStr(status))
    CALL WriteError('Cannot open Energy+.ini. Runtime error #'//trim(ADJUSTL(IntToStr(status)))//' '//trim(IOErrorMsg),msgWarning)
    iniExists = .FALSE.
  ENDIF
END IF
!if the ini file exists and can be opened without raising an error
readstatus=0
IF (iniExists) THEN
  DO WHILE (.not. is_iostat_end(readStatus))
    READ(UNIT=IniFH, FMT="(A)", IOSTAT=readStatus) inputLine
    IF (SameString(inputLine,'[program]')) THEN
      IF (readStatus .NE. -1) THEN
        DO WHILE (readStatus /= -1)
          READ(UNIT=IniFH, FMT="(A)", IOSTAT=readStatus) inputLine
          IF (SameString(inputLine(1:4),'dir=')) THEN
            iddPath = TRIM(inputLine(5:))
            IF (LEN_TRIM(iddPath) .GE. 1) THEN
              iddPathFound = .TRUE.
              EXIT
            END IF
          END IF
        ENDDO
      END IF
    END IF
  END DO
  CLOSE (unit=IniFH)
END IF
!if path is found to ini file, merge it with file name
IF (iddPathFound) THEN
  iddPathLength = LEN_TRIM(iddPath)
  !make sure the path ends with a trailing slash
  IF ( iddPathLength .GT. 0 ) THEN
    IF ( INDEX(iddPath,'\') > 0 ) THEN
      IF (iddPath(iddPathLength:iddPathLength) .NE. '\') iddPath(iddPathLength+1:iddPathLength+1) = '\'
    ELSE IF ( INDEX(iddPath,'/') > 0 ) THEN
      IF (iddPath(iddPathLength:iddPathLength) .NE. '/') iddPath(iddPathLength+1:iddPathLength+1) = '/'
    ELSE ! Try to detect platform
#ifdef _WIN32
      IF (iddPath(iddPathLength:iddPathLength) .NE. '\') iddPath(iddPathLength+1:iddPathLength+1) = '\'
#else
      IF (iddPath(iddPathLength:iddPathLength) .NE. '/') iddPath(iddPathLength+1:iddPathLength+1) = '/'
#endif
    END IF
    IddFileName = TRIM(iddPath) // TRIM(IddFileName)
  END IF
END IF
!see if the idd file exists
INQUIRE(FILE=IddFileName,EXIST=iddExists)
!if it does not exist
IF (.NOT. iddExists) THEN
  CALL WriteError('Cannot find Energy+.idd as specified in Energy+.ini.',msgWarning)
  !try using just the file name looking in the current directory.
  INQUIRE(FILE='Energy+.idd',EXIST=iddExists)
  IF (iddExists) THEN
    IddFileName = 'Energy+.idd'
  END IF
END IF
IF (iddExists) THEN
  OPEN (unit=IddFH, file=TRIM(IddFileName), action='READ',iostat=ios,iomsg=IOErrorMsg)
!  OPEN (unit=IddFH, file=TRIM(IddFileName), action='READ',iostat=ios)
  IF (status .EQ. 0) THEN
    iddCanBeRead = .TRUE.
  ELSE
!    IOErrorMsg=ADJUSTL(IntToStr(status))
    CALL WriteError('Error opening Energy+.idd. Runtime error #'//trim(ADJUSTL(IntToStr(status)))//  &
       ' '//trim(IOErrorMsg),msgWarning)
    iddCanBeRead = .FALSE.
  ENDIF
ELSE
  iddCanBeRead = .FALSE.
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE readIDDfile
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2012
          !    MODIFIED       na
          !    RE-ENGINEERED  Originally from idd2f90 by Jason Glazer (c)2007

          ! PURPOSE OF THIS SUBROUTINE:
          !   Read the HVACTemplate objects from the IDD to provide valid ranges for numeric values.
          !   Read object by object. An object starts with a line with a comma that is not in a comment.
          !   An object ends after the last line with a slashcode after the semicolon that is not in a comment.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine

          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
INTEGER :: readStatus
CHARACTER(len=1000) :: inputLine
CHARACTER(len=1000) :: ainputLine
CHARACTER(len=MaxAlphaLength) :: curName
CHARACTER(len=MaxAlphaLength) :: slashCodeString
CHARACTER(len=MaxAlphaLength) :: lcSlashCodeString
CHARACTER(len=MaxAlphaLength) :: curSlashValue
CHARACTER(len=MaxAlphaLength) :: lcCurSlashValue
INTEGER :: pSemicolon
INTEGER :: pComma
INTEGER :: pSlash
INTEGER :: pExclaim
INTEGER :: pSpaceAfterSlash
INTEGER :: readObjectLoc
INTEGER,PARAMETER :: rolInObj = 1
INTEGER,PARAMETER :: rolOutObj = 2
INTEGER,PARAMETER :: rolNearEnd = 3
INTEGER :: iObj
INTEGER :: jFld
LOGICAL :: keepObject = .FALSE.
INTEGER :: foundNumber
IF (.NOT. foundCompactObject) RETURN
WRITE(*,*) 'Begin reading Energy+.idd file.'
readObjectLoc = rolOutObj
objCount = 0
fldCount = 0
keyCount = 0
readStatus=0
!skip to \Group HVAC Templates before starting to read
IF (.NOT. checkOutputFieldNamesAgainstIDD) THEN !read entire IDD if checking output field names
  DO WHILE (.not. is_iostat_end(readStatus))
    READ(UNIT=IddFH, FMT="(A)", IOSTAT=readStatus) inputLine
    IF (inputLine(1:1) .EQ. '\') THEN
      IF (SameString(inputLine(1:21),'\Group HVAC Templates')) EXIT
    END IF
  END DO
END IF
!now that in group start parsing
DO WHILE (.not. is_iostat_end(readStatus))
  READ(UNIT=IddFH, FMT="(A)", IOSTAT=readStatus) inputLine
  pSemicolon = INDEX(inputLine,';')
  pComma = INDEX(inputLine,',')
  pSlash = INDEX(inputLine,'\')
  pExclaim = INDEX(inputLine,'!')
  !if puntuation after comment then ignore it
  IF (pExclaim .GE. 1) THEN
    IF (pSlash .GT. pExclaim) pSlash = 0
    IF (pSemicolon .GT. pExclaim) pSemicolon = 0
    IF (pComma .GT. pExclaim) pComma = 0
  END IF
  !if puncuation after slash (in memo) then ignore it
  IF (pSlash .GE. 1) THEN
    IF (pSemicolon .GT. pSlash) pSemicolon = 0
    IF (pComma .GT. pSlash) pComma = 0
  END IF
  !if slash code exists on the line get the slashcode and
  !the argument
  IF (keepObject) THEN
    IF (objCount .GE. 1) THEN !must have an object to start assigning slashcodes
      IF (pSlash .GE. 1) THEN
        pSpaceAfterSlash = INDEX(inputLine(pSlash:),' ') + pSlash !find the first space after the slash
        IF (pSpaceAfterSlash .GE. 1) THEN
          slashCodeString = TRIM(inputLine(pSlash+1:pSpaceAfterSlash-1))
          curSlashValue = TRIM(ADJUSTL(inputLine(pSpaceAfterSlash:)))
        ELSE
          slashCodeString = TRIM(inputLine(pSlash+1:))
          curSlashValue = ''
        END IF
        !assign slash code
        lcSlashCodeString = makeLowerCase(slashCodeString)
        lcCurSlashValue = makeLowerCase(curSlashValue)
        !WRITE (*,*) '    [' // TRIM(lcSlashCodeString) // '][' // TRIM(lcCurSlashValue) // ']'
        !field level slashcodes
        IF (lcSlashCodeString .EQ. 'field') THEN
          fldCount = fldCount + 1
          iddFld(fldCount)%name = curSlashValue
          IF (iddObj(objCount)%firstFld .EQ. 0) THEN !the object has no fields yet
            iddObj(objCount)%firstFld = fldCount
          END IF
          IF ((pComma > 0 .or. pSemicolon > 0) .and. iddObj(objCount)%firstFld > 0) THEN
            ainputLine=ADJUSTL(inputLine)
            IF (ainputLine(1:1) == 'A') THEN
              iddFld(fldCount)%AorN = .true.
            ELSE
              iddFld(fldCount)%AorN = .false.
            ENDIF
          ENDIF
          iddObj(objCount)%lastFld = fldCount
        ELSEIF (lcSlashCodeString .EQ. 'note') THEN
          iddFld(fldCount)%note = TRIM(iddFld(fldCount)%note) // TRIM(lcCurSlashValue) // ' '
          iddFld(fldCount)%notePresent = .true.
        ELSEIF (lcSlashCodeString .EQ. 'required-field') THEN
          iddFld(fldCount)%required = .true.
        ELSEIF (lcSlashCodeString .EQ. 'units') THEN
          iddFld(fldCount)%units = lcCurSlashValue
          iddFld(fldCount)%unitsPresent = .true.
        ELSEIF (lcSlashCodeString .EQ. 'ip-units') THEN
          iddFld(fldCount)%ipUnits = lcCurSlashValue
          iddFld(fldCount)%ipUnitsPresent = .true.
        ELSEIF (lcSlashCodeString .EQ. 'minimum') THEN
          iddFld(fldCount)%minimum = StringToReal(lcCurSlashValue)
          iddFld(fldCount)%minimumPresent = .true.
          iddFld(fldCount)%minOrEqual = .true.
        ELSEIF (lcSlashCodeString .EQ. 'minimum>') THEN
          iddFld(fldCount)%minimum = StringToReal(lcCurSlashValue)
          iddFld(fldCount)%minimumPresent = .true.
          iddFld(fldCount)%minOrEqual = .false.
        ELSEIF (lcSlashCodeString .EQ. 'maximum') THEN
          iddFld(fldCount)%maximum = StringToReal(lcCurSlashValue)
          iddFld(fldCount)%maximumPresent = .true.
          iddFld(fldCount)%maxOrEqual = .true.
        ELSEIF (lcSlashCodeString .EQ. 'maximum<') THEN
          iddFld(fldCount)%maximum = StringToReal(lcCurSlashValue)
          iddFld(fldCount)%maximumPresent = .true.
          iddFld(fldCount)%maxOrEqual = .false.
        ELSEIF (lcSlashCodeString .EQ. 'default') THEN
          iddFld(fldCount)%default = StringToReal(lcCurSlashValue)
          iddFld(fldCount)%defaultPresent = .true.
        ELSEIF (lcSlashCodeString .EQ. 'autosizable') THEN
          iddFld(fldCount)%autoSize = .true.
        ELSEIF (lcSlashCodeString .EQ. 'autocalculatable') THEN
          iddFld(fldCount)%autoCalc = .true.
        ELSEIF (lcSlashCodeString .EQ. 'type') THEN
          IF (lcCurSlashValue .EQ. 'integer') THEN
            iddFld(fldCount)%typeOfData = todInteger
          ELSEIF (lcCurSlashValue .EQ.  'real') THEN
            iddFld(fldCount)%typeOfData = todReal
          ELSEIF (lcCurSlashValue .EQ.  'alpha') THEN
            iddFld(fldCount)%typeOfData = todAlpha
          ELSEIF (lcCurSlashValue .EQ.  'choice') THEN
            iddFld(fldCount)%typeOfData = todChoice
          ELSEIF (lcCurSlashValue .EQ.  'object-list') THEN
            iddFld(fldCount)%typeOfData = todObjList
          ELSEIF (lcCurSlashValue .EQ.  'external-list') THEN
            iddFld(fldCount)%typeOfData = todObjList
          ELSEIF (lcCurSlashValue .EQ.  'node') THEN
            iddFld(fldCount)%typeOfData = todAlpha
          ELSE
            WRITE (*,*) 'WARNING:  Do not recognize the type of data: ' // trim(lcCurSlashValue)
          END IF
          iddFld(fldCount)%typeOfDataPresent = .true.
        ELSEIF (lcSlashCodeString .EQ. 'retaincase') THEN
          iddFld(fldCount)%retainCase = .true.
        ELSEIF (lcSlashCodeString .EQ. 'key') THEN
          keyCount = keyCount + 1
          iddKey(keyCount)%name = curSlashValue
          IF (iddFld(fldCount)%firstKey .EQ. 0) THEN !the object has no fields yet
            iddFld(fldCount)%firstKey = keyCount
          END IF
          iddFld(fldCount)%lastKey = keyCount
        ELSEIF (lcSlashCodeString .EQ. 'object-list') THEN
          iddFld(fldCount)%objList = lcCurSlashValue
          iddFld(fldCount)%objListPresent = .true.
        ELSEIF (lcSlashCodeString .EQ. 'external-list') THEN
          iddFld(fldCount)%objList = lcCurSlashValue
          iddFld(fldCount)%objListPresent = .true.
        ELSEIF (lcSlashCodeString .EQ. 'reference') THEN
          iddFld(fldCount)%reference = lcCurSlashValue
          iddFld(fldCount)%referencePresent = .true.
        ELSEIF (lcSlashCodeString .eq. 'deprecated') then
          iddFld(objCount)%deprecated = .true.
        !---------------
        !object level slashcode
        ELSEIF (lcSlashCodeString .EQ. 'memo') THEN
          iddObj(objCount)%memo = TRIM(iddObj(objCount)%memo) // TRIM(lcCurSlashValue) // ' '
          iddObj(objCount)%memoPresent = .true.
        ELSEIF (lcSlashCodeString .EQ. 'unique-object') THEN
          iddObj(objCount)%unique = .true.
        ELSEIF (lcSlashCodeString .EQ. 'required-object') THEN
          iddObj(objCount)%required = .true.
        ELSEIF (lcSlashCodeString .EQ. 'min-fields') THEN
          iddObj(objCount)%minFields = StringToInt(lcCurSlashValue)
          iddObj(objCount)%minFieldsPresent = .true.
        ELSEIF (lcSlashCodeString .EQ. 'obsolete') THEN
          iddObj(objCount)%obsolete = .true.
        ELSEIF (lcSlashCodeString(1:10) .EQ. 'extensible') THEN
          iddObj(objCount)%extensible = .true.
        ELSEIF (lcSlashCodeString .EQ. 'begin-extensible') THEN
          iddObj(objCount)%beginextensiblefound = .true.
        ELSEIF (lcSlashCodeString .EQ. 'format') THEN
          IF (lcCurSlashValue .EQ. 'singleline') THEN
            iddObj(objCount)%spFormat = spFmtSingleLine
          ELSEIF (lcCurSlashValue .EQ.  'vertices') THEN
            iddObj(objCount)%spFormat = spFmtVertices
          ELSEIF (lcCurSlashValue .EQ.  'compactschedule') THEN
            iddObj(objCount)%spFormat = spFmtCompactSchedule
          ELSEIF (lcCurSlashValue .EQ.  'fluidproperty') THEN
          ELSEIF (lcCurSlashValue .EQ.  'spectral') THEN
          ELSEIF (lcCurSlashValue .EQ.  'viewfactor') THEN
          ELSE
            WRITE (*,*) 'WARNING:  Do not recognize the kind of special format: ' // lcCurSlashValue
          END IF
          iddObj(objCount)%spFormatPresent = .true.
        ELSEIF (lcSlashCodeString .EQ. 'group') THEN
          !group level slashcode - don't do anything.
        ELSEIF (lcSlashCodeString .EQ. 'begin-extensible') THEN
          !begin-extensible slashcode - don't do anything.
        ELSEIF (lcSlashCodeString .EQ. 'unitsbasedonfield') THEN
          !unitsBasedOnField slashcode - don't do anything.
        ELSE
          WRITE (*,*) 'WARNING:  Do not recognize the slashcode: ' // lcSlashCodeString
        END IF
      END IF
    END IF
  END IF
  !depending on what parsing location when reading an object
  !(in, out, near end) does different actions.
  SELECT CASE (readObjectLoc)
    CASE (rolInObj) !in object
      IF (pSemicolon .GE. 1) THEN
        readObjectLoc = rolNearEnd
        IF (pSlash .EQ. 0) THEN
          readObjectLoc = rolOutObj
          keepObject = .FALSE.
        END IF
      END IF
    CASE (rolOutObj) !out of object
      IF (pComma .GE. 1) THEN
        readObjectLoc = rolInObj
        curName = adjustl(inputLine(1:pComma-1))
        IF (.NOT. checkOutputFieldNamesAgainstIDD) THEN
          IF (SameString(curName(1:13), 'HVACTemplate:')) THEN !only keep objects that start with HVACTemplate:
            keepObject = .TRUE.
            objCount = objCount + 1
            iddObj(objCount)%name = curName
          END IF
        ELSE !if checking output filed names against the IDD store them all
            keepObject = .TRUE.
            objCount = objCount + 1
            iddObj(objCount)%name = curName
        END IF
      END IF
      IF (pSlash .EQ. 1) THEN    !don't read the rest of file if not in an object and \group is found
        IF (.NOT. checkOutputFieldNamesAgainstIDD) THEN !only still the rest of file if not checking output field names
          IF (SameString(inputLine(1:6),'\group')) EXIT
        END IF
      END IF
    CASE (rolNearEnd) !near the end of an object (after semicolon but before last slashcode)
      IF (pSlash .EQ. 0) THEN
        readObjectLoc = rolOutObj
        keepObject = .FALSE.
      END IF
  END SELECT
END DO
! for fields that are the same except a number, remove the number "2" so that
! when scanning for valid fields without numbers that will work.
DO jFld = 1, fldCount
  foundNumber = INDEX(iddFld(jFld)%name,'2 ')
  IF (foundNumber .GT. 1) THEN
    iddFld(jFld)%name = iddFld(jFld)%name(:foundNumber-1) // iddFld(jFld)%name((foundNumber+2):)
  END IF
END DO
IF (.false.) THEN  !toggle on and off putting echo of IDD into expand.idf
  ! write out what has been read as confirmation
  WRITE(OutFH,'(A)') '! ******************** Echo of summary from reading Energy+.idd file ********************'
  WRITE(OutFH,'(A,I6)') '! Count of objects: ',objCount
  WRITE(OutFH,'(A,I6)') '! Count of fields:  ',fldCount
  WRITE(OutFH,'(A,I6)') '! Count of keys:    ',keyCount
  WRITE(OutFH,'(A)') '! '
  WRITE(OutFH,'(A,T86,A,8X,A,8X,A)') '!','Min','Def','Max'
  DO iObj = 1,objCount
    WRITE(OutFH,'(A,A)') '!',TRIM(iddObj(iObj)%name)
    IF ((iddObj(iObj)%firstFld .GE. 1) .AND. (iddObj(iObj)%lastFld .GE. 1)) THEN
      DO jFld = iddObj(iObj)%firstFld,iddObj(iObj)%lastFld
        WRITE(OutFH,'(A,2X,A,T80,F9.2,2X,F9.2,2X,F9.2)') '!',TRIM(iddFld(jFld)%name),iddFld(jFld)%minimum, &
         iddFld(jFld)%default,iddFld(jFld)%maximum
      END DO
    END IF
    WRITE(OutFH,'(A)') '!'
  END DO
  WRITE(OutFH,'(A)') '! ******************** End of summary from reading Energy+.idd file ********************'
  WRITE(OutFH,'(A)') '!'
END IF
WRITE(*,*) 'Done reading Energy+.idd file.'
END SUBROUTINE

!----------------------------------------------------------------------------------
FUNCTION makeLowerCase(phraseIn) RESULT (phraseOut)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2012
          !    MODIFIED       na
          !    RE-ENGINEERED  Originally from idd2f90 by Jason Glazer (c)2007

          ! PURPOSE OF THIS SUBROUTINE:
          !   Convert a string into all lower case

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine

          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
CHARACTER(len=*), INTENT(IN) :: phraseIn
CHARACTER(len=LEN(phraseIn)) :: phraseOut
INTEGER :: curCharVal
INTEGER :: convCharVal
INTEGER :: i
phraseOut = ''
DO i=1,LEN(phraseIn)
  curCharVal = ICHAR(phraseIn(i:i))
  SELECT CASE (curCharVal)
    CASE (65:90) !uppercase ASCII
      convCharVal = curCharVal + 32
    CASE DEFAULT
      convCharVal = curCharVal
  END SELECT
  phraseOut(i:i) = CHAR(convCharVal)
END DO
END FUNCTION

!----------------------------------------------------------------------------------
SUBROUTINE AssociateIDDwithObjToProc
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2012
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Add to objToProcess array an index to the iddObj array.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
INTEGER :: iObjToProc
INTEGER :: jIddObj

IF (.NOT. foundCompactObject) RETURN
! first check all objects that are HVACTemplates to see if any names are repeated
DO iObjToProc = 1, numObjToProcess
  !only test objects that are HVACTemplate objects, skip the rest
  IF (SameString(objToProcess(iObjToProc)%ucName(1:12),'HVACTEMPLATE')) THEN
    DO jIddObj = 1,objCount
      IF (SameString(objToProcess(iObjToProc)%ucName,iddObj(jIddObj)%name)) THEN
        objToProcess(iObjToProc)%iddObjIndex = jIddObj
        EXIT
      END IF
    END DO
    ! show error if no associated values
    IF (objToProcess(iObjToProc)%iddObjIndex .EQ. 0) THEN
      CALL WriteError('No IDD object found for: ' //TRIM(objToProcess(iObjToProc)%ccName))
    END IF
  END IF
END DO
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE CheckMaxMinKeys
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2012
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Using the values for min and max from the IDD file, make sure
          !    all minimum and maximum values are valid. Also check if the choice
          !    key used in the objects are one of the options shown in the IDD
          !    file.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
INTEGER :: iHeldObj
INTEGER :: curFirstFldVal
INTEGER :: curLastFldVal
INTEGER :: numFldVals
INTEGER :: curObjToProc
INTEGER :: curIddObj
INTEGER :: curIddFirstField
INTEGER :: curIddLastField
INTEGER :: numIddFields
CHARACTER(len=MaxAlphaLength) :: curFldValString
REAL :: curFldVal
INTEGER :: jFld
INTEGER :: actIddField
LOGICAL :: found
INTEGER :: kKey

DO iHeldObj = 1,numHeldObj
  curObjToProc = HeldObj(iHeldObj)%ObjToProcessIndex
  curFirstFldVal = HeldObj(iHeldObj)%firstFldVal + 1 !since the first value is actually the name of the held object
  curLastFldVal = HeldObj(iHeldObj)%lastFldVal
  numFldVals = curLastFldVal - curFirstFldVal + 1
  IF ((curObjToProc .GE. 1) .AND. (curObjToProc .LE. numObjToProcess)) THEN
    curIddObj = objToProcess(curObjToProc)%iddObjIndex
    IF ((curIddObj .GE. 1) .AND. (curIddObj .LE. objCount)) THEN
      curIddFirstField = iddObj(curIddObj)%firstFld
      curIddLastField = iddObj(curIddObj)%lastFld
      numIddFields= 1 + curIddLastField - curIddFirstField
      IF (numFldVals .EQ. numIddFields) THEN
        DO jFld = 0,numIddFields - 1 !base this on zero because using it as an offset from first field
          curFldValString = FldVal(curFirstFldVal + jFld)
          actIddField = curIddFirstField + jFld
          IF (.NOT. iddFld(actIddField)%AorN) THEN !it is an N (numeric) field
            IF (isNumeric(curFldValString)) THEN
              curFldVal = StringToReal(curFldValString)
              !test for minimum
              IF (iddFld(actIddField)%minimumPresent) THEN
                IF (iddFld(actIddField)%minOrEqual) THEN
                  IF (curFldVal .LT. iddFld(actIddField)%minimum) THEN
                    CALL WriteError('The value ' // TRIM(FldVal(curFirstFldVal + jFld)) // ' is less than ' // &
                         'the minimum value expected of '// TRIM(RealToStr(iddFld(actIddField)%minimum)) // ' for object: ' // &
                         TRIM(iddObj(curIddObj)%name) // ' with the name: ' // TRIM(FldVal(curFirstFldVal)) // ' and field: ' // &
                         TRIM(iddFld(actIddField)%name),msgWarning)
                  END IF
                ELSE
                  IF (curFldVal .LE. iddFld(actIddField)%minimum) THEN
                    CALL WriteError('The value ' // TRIM(FldVal(curFirstFldVal + jFld)) // ' is less than or equal to ' // &
                         'the minimum value expected of '// TRIM(RealToStr(iddFld(actIddField)%minimum)) // ' for object: ' // &
                         TRIM(iddObj(curIddObj)%name) // ' with the name: ' // TRIM(FldVal(curFirstFldVal)) // ' and field: ' // &
                        TRIM(iddFld(actIddField)%name),msgWarning)
                  END IF
                END IF
              END IF
              !test for maximum
              IF (iddFld(actIddField)%maximumPresent) THEN
                IF (iddFld(actIddField)%maxOrEqual) THEN
                  IF (curFldVal .GT. iddFld(actIddField)%maximum) THEN
                    CALL WriteError('The value ' // TRIM(FldVal(curFirstFldVal + jFld)) // ' is greater than ' // &
                         'the maximum value expected of '// TRIM(RealToStr(iddFld(actIddField)%maximum)) // ' for object: ' // &
                         TRIM(iddObj(curIddObj)%name) // ' with the name: ' // TRIM(FldVal(curFirstFldVal)) // ' and field: ' // &
                         TRIM(iddFld(actIddField)%name),msgWarning)
                  END IF
                ELSE
                  IF (curFldVal .GE. iddFld(actIddField)%maximum) THEN
                    CALL WriteError('The value ' // TRIM(FldVal(curFirstFldVal + jFld)) // ' is greater than or equal to ' // &
                         'the maximum value expected of '// TRIM(RealToStr(iddFld(actIddField)%maximum)) // ' for object: ' // &
                         TRIM(iddObj(curIddObj)%name) // ' with the name: ' // TRIM(FldVal(curFirstFldVal)) // ' and field: ' // &
                         TRIM(iddFld(actIddField)%name),msgWarning)
                  END IF
                END IF
              END IF
            END IF
          ELSE !test for valid keys
            IF (iddFld(actIddField)%typeOfDataPresent) THEN
              IF (iddFld(actIddField)%typeOfData .EQ. todChoice) THEN
                IF (LEN_TRIM(curFldValString) .GE. 1) THEN
                  found = .FALSE.
                  DO kKey = iddFld(actIddField)%firstKey,iddFld(actIddField)%lastKey
                    IF (SameString(curFldValString,iddKey(kKey)%name)) THEN
                      found = .TRUE.
                      EXIT
                    END IF
                  END DO
                  IF (.NOT. found) THEN
                    CALL WriteError('The value ' // TRIM(curFldValString) // ' is not a valid choice for object: ' // &
                           TRIM(iddObj(curIddObj)%name) // ' with the name: ' // TRIM(FldVal(curFirstFldVal)) // ' and field: ' // &
                           TRIM(iddFld(actIddField)%name),msgWarning)
                  END IF
                END IF
              END IF
            ENDIF
          END  IF
        END DO
      ELSE
        CALL WriteError('The number of fields in the IDD do not match the number of fields assumed. ' // &
         TRIM(objToProcess(curObjToProc)%ccName))
      END IF
    END IF
  END IF
END DO
END SUBROUTINE

!----------------------------------------------------------------------------------
!----------------------------------------------------------------------------------
!
!   CREATE NEW OBJECTS
!
!----------------------------------------------------------------------------------
!----------------------------------------------------------------------------------

!----------------------------------------------------------------------------------
SUBROUTINE GetInputAll
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Move the held objects into arrays by object.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
!set up the base pointer arrays
CALL GetInputThermostat
CALL GetInputVAV
CALL GetInputUnitary
CALL GetInputConstVol
CALL GetInputDualDuct
CALL GetInputVRF
CALL GetInputFanCoil
CALL GetInputPurchasedAir
CALL GetInputBaseboard
CALL GetInputPTAC
CALL GetInputPTHP
CALL GetInputWaterAirHP
CALL GetInputDOAS
CALL GetInputMixedLoop
CALL GetInputHotLoop
CALL GetInputBoiler
CALL GetInputChiller
CALL GetInputTower
CALL GetInputChldLoop
CALL GetGroundHeatTransferControl
CALL GetInputBuildingSurfaces
CALL GetAllGroundHeatTransfer
CALL GetGroundTempSurf
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE CreateNewObjects
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    To create objects based on the objects that are held. Usually this
          !    routine just calls other routines that do the work.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
!create expanded objects if some compact inputs were found
IF (foundCompactObject) THEN
  CALL WriteComment('')
  CALL WriteComment('-------------------------------------------------------------')
  CALL WriteComment('New objects created from ExpandObjects')
  CALL WriteComment('-------------------------------------------------------------')
  CALL WriteComment('')
  CALL CheckIfNamesAreUnique
  CALL CreateNewThermostat
  CALL CreateNewVAV
  CALL CreateNewUnitary
  CALL CreateNewConstVol
  CALL CreateNewDualDuct
  CALL CreateNewVRF
  CALL CreateFanCoil
  CALL CreatePurchasedAir
  CALL CreateBaseboard
  CALL CreatePTAC
  CALL CreatePTHP
  CALL CreateWaterAirHP
  CALL CreateDOAS
  ! Plant supply components must be called before loops
  CALL CreateBoiler
  CALL CreateChiller
  CALL CreateTower
  CALL CreateHotLoop
  CALL CreateChldLoop
  CALL CreateMixedLoop
  CALL WriteBasement
  CALL WriteSlab
  CALL CreateDetailedBuildingSurface
  CALL CreateGroundTemperatures
END IF
END SUBROUTINE


!----------------------------------------------------------------------------------
!----------------------------------------------------------------------------------
!
!   GET INPUT ROUTINES
!
!----------------------------------------------------------------------------------
!----------------------------------------------------------------------------------

!----------------------------------------------------------------------------------
SUBROUTINE GetInputVAV
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the VAV objects into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iZone
INTEGER    :: iSys

! ZONE
CALL CountOldObj('HVACTemplate:Zone:VAV',ptCompactZoneVAV,prelimCount)
ALLOCATE(compactZoneVAVbase(prelimCount))
actCount = 0
DO iZone= 1, prelimCount
  CALL NextOldObj(ptCompactZoneVAV,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. vzDesSpecOADistNameOff) THEN
    actCount = actCount + 1
    compactZoneVAVbase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + vzMinFlowMethodOff,       'Constant')
    CALL SetIfBlank(fldValStart + vzConstantMinFracOff,     '0.2')
    CALL SetIfBlank(fldValStart + vzSupplyMaxRateOff,       'autosize')
    CALL SetIfBlank(fldValStart + vzOutAirTypeOff,          'Flow/Person')
    CALL SetIfBlank(fldValStart + vzOutAirFlowPersonOff,    '0.00944')
    CALL SetIfBlank(fldValStart + vzReheatTypeOff,          'None')
    CALL SetIfBlank(fldValStart + vzDamperActionOff,        'Reverse')
    CALL SetIfBlank(fldValStart + vzBaseboardTypeOff,       'None')
    CALL SetIfBlank(fldValStart + vzBaseboardCapOff,        'autosize')
    CALL SetIfBlank(fldValStart + vzCoolDesignTempMethodOff, 'SystemSupplyAirTemperature')
    CALL SetIfBlank(fldValStart + vzCoolDesignTempOff,       '12.8')
    CALL SetIfBlank(fldValStart + vzCoolDesignTempDiffOff,   '11.11')
    CALL SetIfBlank(fldValStart + vzHeatDesignTempMethodOff, 'SupplyAirTemperature')
    CALL SetIfBlank(fldValStart + vzHeatDesignTempOff,       '50.0')
    CALL SetIfBlank(fldValStart + vzCoolDesignTempDiffOff,   '30.0')
  END IF
END DO
numCompactZoneVAV = actCount
! FAN POWERD ZONE
CALL CountOldObj('HVACTemplate:Zone:VAV:FanPowered',ptCompactZoneFPVAV,prelimCount)
ALLOCATE(compactZoneFPVAVbase(prelimCount))
actCount = 0
DO iZone= 1, prelimCount
  CALL NextOldObj(ptCompactZoneFPVAV,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. fpvzDesSpecOADistNameOff) THEN
    actCount = actCount + 1
    compactZoneFPVAVbase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + fpvzPriSupplyMaxRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + fpvzPriSupplyMinFracOff, 'autosize')
    CALL SetIfBlank(fldValStart + fpvzSecSupplyMaxRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + fpvzFlowTypeOff,         'parallel')
    CALL SetIfBlank(fldValStart + fpvzFanOnFlowFracOff,    'autosize')
    CALL SetIfBlank(fldValStart + fpvzOutAirTypeOff,       'Flow/Person')
    CALL SetIfBlank(fldValStart + fpvzOutAirFlowPersonOff, '0.00944')
    CALL SetIfBlank(fldValStart + fpvzOutAirFlowAreaOff,   '0.0')
    CALL SetIfBlank(fldValStart + fpvzOutAirFlowZoneOff,   '0.0')
    CALL SetIfBlank(fldValStart + fpvzReheatTypeOff,       'Electric')
    CALL SetIfBlank(fldValStart + fpvzFanTotEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + fpvzFanPressureOff,      '1000')
    CALL SetIfBlank(fldValStart + fpvzFanMotorEfficiencyOff,'0.9')
    CALL SetIfBlank(fldValStart + fpvzBaseboardTypeOff,    'None')
    CALL SetIfBlank(fldValStart + fpvzBaseboardCapOff,     'autosize')
    CALL SetIfBlank(fldValStart + fpvzCoolDesignTempMethodOff, 'SystemSupplyAirTemperature')
    CALL SetIfBlank(fldValStart + fpvzCoolDesignTempOff,       '12.8')
    CALL SetIfBlank(fldValStart + fpvzCoolDesignTempDiffOff,   '11.11')
    CALL SetIfBlank(fldValStart + fpvzHeatDesignTempMethodOff, 'SupplyAirTemperature')
    CALL SetIfBlank(fldValStart + fpvzHeatDesignTempOff,       '50.0')
    CALL SetIfBlank(fldValStart + fpvzCoolDesignTempDiffOff,   '30.0')
  END IF
END DO
numCompactZoneFPVAV = actCount
! FAN POWERD ZONE
CALL CountOldObj('HVACTemplate:Zone:VAV:HeatAndCool',ptCompactZoneHCVAV,prelimCount)
ALLOCATE(compactZoneHCVAVbase(prelimCount))
actCount = 0
DO iZone= 1, prelimCount
  CALL NextOldObj(ptCompactZoneHCVAV,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. hcvzHeatDesignTempDiffOff) THEN
    actCount = actCount + 1
    compactZoneHCVAVbase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + hcvzSupplyMaxRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + hcvzConstantMinFracOff, '0.2')
    CALL SetIfBlank(fldValStart + hcvzOutAirTypeOff, 'flow/person')
    CALL SetIfBlank(fldValStart + hcvzOutAirFlowPersonOff, '0.00944')
    CALL SetIfBlank(fldValStart + hcvzOutAirFlowAreaOff, '0')
    CALL SetIfBlank(fldValStart + hcvzOutAirFlowZoneOff, '0')
    CALL SetIfBlank(fldValStart + hcvzReheatTypeOff, 'None')
    CALL SetIfBlank(fldValStart + hcvzBaseboardTypeOff, 'None')
    CALL SetIfBlank(fldValStart + hcvzBaseboardCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + hcvzCoolDesignTempMethodOff, 'SystemSupplyAirTemperature')
    CALL SetIfBlank(fldValStart + hcvzCoolDesignTempOff, '12.8')
    CALL SetIfBlank(fldValStart + hcvzCoolDesignTempDiffOff, '11.11')
    CALL SetIfBlank(fldValStart + hcvzHeatDesignTempMethodOff, 'SupplyAirTemperature')
    CALL SetIfBlank(fldValStart + hcvzHeatDesignTempOff, '50')
    CALL SetIfBlank(fldValStart + hcvzHeatDesignTempDiffOff, '30')
  END IF
END DO
numCompactZoneHCVAV = actCount
! SYSTEM
CALL CountOldObj('HVACTemplate:System:VAV',ptCompactSysVAV,prelimCount)
ALLOCATE(compactSysVAVbase(prelimCount))
actCount = 0
DO iSys = 1, prelimCount
  CALL NextOldObj(ptCompactSysVAV,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. vsReturnFanPLRCoeffOff) THEN
    actCount = actCount + 1
    compactSysVAVbase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + vsSupplyMaxRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + vsSupplyMinRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + vsSupplyEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + vsSupplyPressureOff, '1000')
    CALL SetIfBlank(fldValStart + vsSupplyFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + vsSupplyFanFractionOff, '1.0')
    CALL SetIfBlank(fldValStart + vsCoolCoilKindOff, 'ChilledWater')
    CALL SetIfBlank(fldValStart + vsCoolSetPtDesignOff, '12.8')
    CALL SetIfBlank(fldValStart + vsHeatCoilKindOff, 'None')
    CALL SetIfBlank(fldValStart + vsHeatSetPtDesignOff, '10.0')
    CALL SetIfBlank(fldValStart + vsHeatEffiencyOff, '0.8')
    CALL SetIfBlank(fldValStart + vsHeatParasiticOff, '0.0')
    CALL SetIfBlank(fldValStart + vsPreheatCoilKindOff, 'None')
    CALL SetIfBlank(fldValStart + vsPreheatSetPtConstantOff, '7.2')
    CALL SetIfBlank(fldValStart + vsPreheatEffiencyOff, '0.8')
    CALL SetIfBlank(fldValStart + vsPreheatParasiticOff, '0.0')
    CALL SetIfBlank(fldValStart + vsMaxOutsideFlowOff, 'autosize')
    CALL SetIfBlank(fldValStart + vsMinOutsideFlowOff, 'autosize')
    CALL SetIfBlank(fldValStart + vsMinOutsideControlOff, 'ProportionalMinimum')
    CALL SetIfBlank(fldValStart + vsEconomizerKindOff, 'None')
    CALL SetIfBlank(fldValStart + vsEconoLockoutKindOff, 'NoLockout')
    CALL SetIfBlank(fldValStart + vsSupplyFanPlaceOff, 'DrawThrough')
    CALL SetIfBlank(fldValStart + vsSupplyFanPLRCoeffOff, 'InletVaneDampers')
    CALL SetIfBlank(fldValStart + vsNightCycleControlOff, 'StayOff')
    CALL SetIfBlank(fldValStart + vsHeatRecTypeOff, 'None')
    CALL SetIfBlank(fldValStart + vsHeatRecSenEffOff, '0.70')
    CALL SetIfBlank(fldValStart + vsHeatRecLatEffOff, '0.65')
    CALL SetIfBlank(fldValStart + vsCoolSetPtResetTypeOff, 'None')
    CALL SetIfBlank(fldValStart + vsHeatSetPtResetTypeOff, 'None')
    CALL SetIfBlank(fldValStart + vsDehumCtrlTypeOff,'None')
    CALL SetIfBlank(fldValStart + vsDehumSetPtOff,'60')
    CALL SetIfBlank(fldValStart + vsHumidCtrlTypeOff,'None')
    CALL SetIfBlank(fldValStart + vsHumidCapOff,'0.000001')
    CALL SetIfBlank(fldValStart + vsHumidRatedElecOff,'autosize')
    CALL SetIfBlank(fldValStart + vsHumidSetPtOff,'30.0')
    CALL SetIfBlank(fldValStart + vsSizingOptionOff,'NonCoincident')
    CALL SetIfBlank(fldValStart + vsReturnFanPresentOff, 'No')
    CALL SetIfBlank(fldValStart + vsReturnEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + vsReturnPressureOff, '500')
    CALL SetIfBlank(fldValStart + vsReturnFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + vsReturnFanFractionOff, '1')
    CALL SetIfBlank(fldValStart + vsReturnFanPLRCoeffOff, 'InletVaneDampers')
  END IF
END DO
numCompactSysVAV = actCount
! Packaged VAV SYSTEM
CALL CountOldObj('HVACTemplate:System:PackagedVAV',ptCompactSysPVAV,prelimCount)
ALLOCATE(compactSysPVAVbase(prelimCount))
actCount = 0
DO iSys = 1, prelimCount
  CALL NextOldObj(ptCompactSysPVAV,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. pvavsReturnFanPLRCoeffOff) THEN
    actCount = actCount + 1
    compactSysPVAVbase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    ! No defaults for zone names, schedule names,econo limits,
    CALL SetIfBlank(fldValStart + pvavsSupplyMaxRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + pvavsSupplyMinRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + pvavsSupplyFanPlaceOff, 'DrawThrough')
    CALL SetIfBlank(fldValStart + pvavsSupplyEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + pvavsSupplyPressureOff, '1000')
    CALL SetIfBlank(fldValStart + pvavsSupplyFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + pvavsSupplyFanFractionOff, '1.0')
    CALL SetIfBlank(fldValStart + pvavsCoolCoilKindOff, 'TwoSpeedDX')
    CALL SetIfBlank(fldValStart + pvavsCoolSetPtDesignOff, '12.8')
    CALL SetIfBlank(fldValStart + pvavsCoolCoilCapOff, 'Autosize')
    CALL SetIfBlank(fldValStart + pvavsCoolCoilSHROff, 'Autosize')
    CALL SetIfBlank(fldValStart + pvavsCoolCoilCOPOff, '3.0')
    CALL SetIfBlank(fldValStart + pvavsHeatCoilKindOff, 'None')
    CALL SetIfBlank(fldValStart + pvavsHeatSetPtDesignOff, '10.0')
    CALL SetIfBlank(fldValStart + pvavsHeatCoilCapOff, 'Autosize')
    CALL SetIfBlank(fldValStart + pvavsHeatEffiencyOff, '0.8')
    CALL SetIfBlank(fldValStart + pvavsHeatParasiticOff, '0.0')
    CALL SetIfBlank(fldValStart + pvavsMaxOutsideFlowOff, 'Autosize')
    CALL SetIfBlank(fldValStart + pvavsMinOutsideFlowOff, 'Autosize')
    CALL SetIfBlank(fldValStart + pvavsMinOutsideControlOff, 'ProportionalMinimum')
    CALL SetIfBlank(fldValStart + pvavsEconomizerKindOff, 'NoEconomizer')
    CALL SetIfBlank(fldValStart + pvavsEconoLockoutKindOff, 'NoLockout')
    CALL SetIfBlank(fldValStart + pvavsSupplyFanPLRCoeffOff, 'InletVaneDampers')
    CALL SetIfBlank(fldValStart + pvavsNightCycleControlOff, 'StayOff')
    CALL SetIfBlank(fldValStart + pvavsHeatRecTypeOff, 'None')
    CALL SetIfBlank(fldValStart + pvavsHeatRecSenEffOff, '0.70')
    CALL SetIfBlank(fldValStart + pvavsHeatRecLatEffOff, '0.65')
    CALL SetIfBlank(fldValStart + pvavsCoolSetPtResetTypeOff, 'None')
    CALL SetIfBlank(fldValStart + pvavsHeatSetPtResetTypeOff, 'None')
    CALL SetIfBlank(fldValStart + pvavsDehumCtrlTypeOff, 'None')
    CALL SetIfBlank(fldValStart + pvavsDehumSetPtOff, '60.0')
    CALL SetIfBlank(fldValStart + pvavsHumidCtrlTypeOff, ' None')
    CALL SetIfBlank(fldValStart + pvavsHumidCapOff, '0.000001')
    CALL SetIfBlank(fldValStart + pvavsHumidRatedElecOff, 'autosize')
    CALL SetIfBlank(fldValStart + pvavsHumidSetPtOff, '30.0')
    CALL SetIfBlank(fldValStart + pvavsSizingOptionOff, 'NonCoincident')
    CALL SetIfBlank(fldValStart + pvavsReturnFanPresentOff, 'No')
    CALL SetIfBlank(fldValStart + pvavsReturnEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + pvavsReturnPressureOff, '500')
    CALL SetIfBlank(fldValStart + pvavsReturnFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + pvavsReturnFanFractionOff, '1')
    CALL SetIfBlank(fldValStart + pvavsReturnFanPLRCoeffOff, 'InletVaneDampers')
  END IF
END DO
numCompactSysPVAV = actCount
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputThermostat
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the compact hvac thermostat object into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iObj

CALL CountOldObj('HVACTemplate:Thermostat',ptCompactStat,prelimCount)
ALLOCATE(compactStatBase(prelimCount))
actCount = 0
DO iObj= 1, prelimCount
  CALL NextOldObj(ptCompactStat,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. tCoolSetPointOff) THEN
    actCount = actCount + 1
    compactStatBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numCompactStat = actCount
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE GetInputUnitary
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the unitary objects into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iZone
INTEGER    :: iSys

! ZONE
CALL CountOldObj('HVACTemplate:Zone:Unitary',ptCompactZoneUnit,prelimCount)
ALLOCATE(compactZoneUnitBase(prelimCount))
actCount = 0
DO iZone= 1, prelimCount
  CALL NextOldObj(ptCompactZoneUnit,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. uzDesSpecOADistNameOff) THEN
    actCount = actCount + 1
    compactZoneUnitBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + uzSupplyMaxRateOff,   'autosize')
    CALL SetIfBlank(fldValStart + uzOutAirTypeOff,      'Flow/Person')
    CALL SetIfBlank(fldValStart + uzOutAirFlowPersonOff,'0.00944')
    CALL SetIfBlank(fldValStart + uzOutAirFlowAreaOff,  '0.0')
    CALL SetIfBlank(fldValStart + uzOutAirFlowZoneOff,  '0.0')
    CALL SetIfBlank(fldValStart + uzBaseboardTypeOff,   'None')
    CALL SetIfBlank(fldValStart + uzBaseboardCapOff,    'autosize')
    CALL SetIfBlank(fldValStart + uzCoolDesignTempMethodOff, 'SystemSupplyAirTemperature')
    CALL SetIfBlank(fldValStart + uzCoolDesignTempOff,       '12.8')
    CALL SetIfBlank(fldValStart + uzCoolDesignTempDiffOff,   '11.11')
    CALL SetIfBlank(fldValStart + uzHeatDesignTempMethodOff, 'SystemSupplyAirTemperature')
    CALL SetIfBlank(fldValStart + uzHeatDesignTempOff,       '50.0')
    CALL SetIfBlank(fldValStart + uzCoolDesignTempDiffOff,   '30.0')
  END IF
END DO
numCompactZoneUnit = actCount
! Unitary SYSTEM
CALL CountOldObj('HVACTemplate:System:Unitary',ptCompactSysUnit,prelimCount)
ALLOCATE(compactSysUnitBase(prelimCount))
actCount = 0
DO iSys = 1, prelimCount
  CALL NextOldObj(ptCompactSysUnit,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. usReturnFanFractionOff) THEN
    actCount = actCount + 1
    compactSysUnitBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + usSupplyMaxRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + usSupplyEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + usSupplyPressureOff, '600')
    CALL SetIfBlank(fldValStart + usSupplyFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + usSupplyFanFractionOff, '1.0')
    CALL SetIfBlank(fldValStart + usCoolCoilKindOff, 'SingleSpeedDX')
    CALL SetIfBlank(fldValStart + usCoolDesignSupplyTempOff, '12.8')
    CALL SetIfBlank(fldValStart + usCoolCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + usCoolCoilSHROff, 'autosize')
    CALL SetIfBlank(fldValStart + usCoolCoilCOPOff, '3.0')
    CALL SetIfBlank(fldValStart + usHeatCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + usHeatDesignSupplyTempOff, '50.0')
    CALL SetIfBlank(fldValStart + usHeatEffiencyOff, '0.8')
    CALL SetIfBlank(fldValStart + usHeatParasiticOff, '0.0')
    CALL SetIfBlank(fldValStart + usMaxOutsideFlowOff, 'autosize')
    CALL SetIfBlank(fldValStart + usMinOutsideFlowOff, 'autosize')
    CALL SetIfBlank(fldValStart + usEconomizerKindOff, 'None')
    CALL SetIfBlank(fldValStart + usEconoLockoutKindOff, 'NoLockout')
    CALL SetIfBlank(fldValStart + usSupplyFanPlaceOff, 'BlowThrough')
    CALL SetIfBlank(fldValStart + usNightCycleControlOff, 'StayOff')
    CALL SetIfBlank(fldValStart + usHeatRecTypeOff, 'None')
    CALL SetIfBlank(fldValStart + usHeatRecSenEffOff, '0.70')
    CALL SetIfBlank(fldValStart + usHeatRecLatEffOff, '0.65')
    CALL SetIfBlank(fldValStart + usDehumCtrlTypeOff, 'None')
    CALL SetIfBlank(fldValStart + usDehumSetPtOff, '60.0')
    CALL SetIfBlank(fldValStart + usHumidCtrlTypeOff, 'None')
    CALL SetIfBlank(fldValStart + usHumidCapOff, '0.000001')
    CALL SetIfBlank(fldValStart + usHumidRatedElecOff, 'autosize')
    CALL SetIfBlank(fldValStart + usHumidSetPtOff, '30.0')
    CALL SetIfBlank(fldValStart + usHumidSetPtOff, '30')
    CALL SetIfBlank(fldValStart + usReturnFanPresentOff, 'No')
    CALL SetIfBlank(fldValStart + usReturnEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + usReturnPressureOff, '300')
    CALL SetIfBlank(fldValStart + usReturnFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + usReturnFanFractionOff, '1')
  END IF
END DO
numCompactSysUnit = actCount
! Unitary Heat Pump SYSTEM
CALL CountOldObj('HVACTemplate:System:UnitaryHeatPump:AirToAir',ptCompactSysUnitHP,prelimCount)
ALLOCATE(compactSysUnitHPBase(prelimCount))
actCount = 0
DO iSys = 1, prelimCount
  CALL NextOldObj(ptCompactSysUnitHP,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. uhpsReturnFanFractionOff) THEN
    actCount = actCount + 1
    compactSysUnitHPBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + uhpsSupplyCoolFlowRateOff, 'Autosize')
    CALL SetIfBlank(fldValStart + uhpsSupplyHeatFlowRateOff, 'Autosize')
    CALL SetIfBlank(fldValStart + uhpsSupplyNoLoadFlowRateOff, 'Autosize')
    CALL SetIfBlank(fldValStart + uhpsSupplyFanPlaceOff, 'BlowThrough')
    CALL SetIfBlank(fldValStart + uhpsSupplyEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + uhpsSupplyPressureOff, '600')
    CALL SetIfBlank(fldValStart + uhpsSupplyFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + uhpsSupplyFanFractionOff, '1.0')
    CALL SetIfBlank(fldValStart + uhpsCoolCoilKindOff, 'SingleSpeedDX')
    CALL SetIfBlank(fldValStart + uhpsCoolDesignSupplyTempOff, '12.8')
    CALL SetIfBlank(fldValStart + uhpsCoolCoilCapOff, 'Autosize')
    CALL SetIfBlank(fldValStart + uhpsCoolCoilSHROff, 'Autosize')
    CALL SetIfBlank(fldValStart + uhpsCoolCoilCOPOff, '3.0')
    CALL SetIfBlank(fldValStart + uhpsHPHeatCoilKindOff, 'SingleSpeedDXHeatPump')
    CALL SetIfBlank(fldValStart + uhpsHeatDesignSupplyTempOff, '50.0')
    CALL SetIfBlank(fldValStart + uhpsHPHeatCoilCapOff, 'Autosize')
    CALL SetIfBlank(fldValStart + uhpsHPHeatCoilCOPOff, '2.75')
    CALL SetIfBlank(fldValStart + uhpsHPHeatMinODBOff, '-8.0')
    CALL SetIfBlank(fldValStart + uhpsHPDefrMaxODBOff, '5.0')
    CALL SetIfBlank(fldValStart + uhpsHPDefrStrategyOff, 'ReverseCycle')
    CALL SetIfBlank(fldValStart + uhpsHPDefrCtrlTypeOff, 'Timed')
    CALL SetIfBlank(fldValStart + uhpsHPDefrTimeFracOff, '0.058333')
    CALL SetIfBlank(fldValStart + uhpsSuppHeatCoilKindOff, 'Electric')
    CALL SetIfBlank(fldValStart + uhpsSuppHeatCoilCapOff, 'Autosize')
    CALL SetIfBlank(fldValStart + uhpsSuppHeatMaxODBOff, '21.0')
    CALL SetIfBlank(fldValStart + uhpsSuppHeatEfficiencyOff, '0.8')
    CALL SetIfBlank(fldValStart + uhpsSuppHeatParasiticOff, '0.0')
    CALL SetIfBlank(fldValStart + uhpsMaxOutsideFlowOff, 'Autosize')
    CALL SetIfBlank(fldValStart + uhpsMinOutsideFlowOff, 'Autosize')
    CALL SetIfBlank(fldValStart + uhpsEconomizerKindOff, 'NoEconomizer')
    CALL SetIfBlank(fldValStart + uhpsEconoLockoutKindOff, 'NoLockout')
    CALL SetIfBlank(fldValStart + uhpsNightCycleControlOff, 'StayOff')
    CALL SetIfBlank(fldValStart + uhpsHeatRecTypeOff, 'None')
    CALL SetIfBlank(fldValStart + uhpsHeatRecSenEffOff, '0.70')
    CALL SetIfBlank(fldValStart + uhpsHeatRecLatEffOff, '0.65')
    CALL SetIfBlank(fldValStart + uhpsHumidCtrlTypeOff, 'None')
    CALL SetIfBlank(fldValStart + uhpsHumidCapOff, '0.000001')
    CALL SetIfBlank(fldValStart + uhpsHumidRatedElecOff, 'autosize')
    CALL SetIfBlank(fldValStart + uhpsHumidSetPtOff, '30.0')
    CALL SetIfBlank(fldValStart + uhpsHumidSetPtOff, '30')
    CALL SetIfBlank(fldValStart + uhpsReturnFanPresentOff, 'No')
    CALL SetIfBlank(fldValStart + uhpsReturnEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + uhpsReturnPressureOff, '300')
    CALL SetIfBlank(fldValStart + uhpsReturnFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + uhpsReturnFanFractionOff, '1')
  END IF
END DO
numCompactSysUnitHP = actCount

! Unitary System SYSTEM
CALL CountOldObj('HVACTemplate:System:UnitarySystem',ptCompactSysUnitarySystem,prelimCount)
ALLOCATE(compactSysUnitarySystemBase(prelimCount))
actCount = 0
DO iSys = 1, prelimCount
  CALL NextOldObj(ptCompactSysUnitarySystem,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. ussReturnFanFractionOff) THEN
    actCount = actCount + 1
    compactSysUnitarySystemBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + ussSystemControlTypeOff, 'Load')
    CALL SetIfBlank(fldValStart + ussSupplyCoolFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + ussSupplyHeatFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + ussSupplyNoLoadFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + ussSupplyFanPlaceOff, 'BlowThrough')
    CALL SetIfBlank(fldValStart + ussSupplyEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + ussSupplyPressureOff, '600')
    CALL SetIfBlank(fldValStart + ussSupplyFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + ussSupplyFanFractionOff, '1')
    CALL SetIfBlank(fldValStart + ussCoolCoilKindOff, 'SingleSpeedDX')
    CALL SetIfBlank(fldValStart + ussCoolCoilNumSpeedOff, '1')
    CALL SetIfBlank(fldValStart + ussCoolDesignSupplyTempOff, '12.8')
    CALL SetIfBlank(fldValStart + ussCoolCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + ussCoolCoilSHROff, 'autosize')
    CALL SetIfBlank(fldValStart + ussCoolCoilCOPOff, '3')
    CALL SetIfBlank(fldValStart + ussHeatCoilKindOff, 'Gas')
    CALL SetIfBlank(fldValStart + ussHeatCoilNumSpeedOff, '1')
    CALL SetIfBlank(fldValStart + ussHeatDesignSupplyTempOff, '12.2')
    CALL SetIfBlank(fldValStart + ussHeatCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + ussHeatEfficiencyOff, '0.8')
    CALL SetIfBlank(fldValStart + ussHeatParasiticOff, '0')
    CALL SetIfBlank(fldValStart + ussHPHeatCoilCOPOff, '2.75')
    CALL SetIfBlank(fldValStart + ussHPHeatMinODBOff, '-8')
    CALL SetIfBlank(fldValStart + ussHPDefrMaxODBOff, '5')
    CALL SetIfBlank(fldValStart + ussHPDefrStrategyOff, 'ReverseCycle')
    CALL SetIfBlank(fldValStart + ussHPDefrCtrlTypeOff, 'Timed')
    CALL SetIfBlank(fldValStart + ussHPDefrTimeFracOff, '0.058333')
    CALL SetIfBlank(fldValStart + ussSuppReHeatCoilKindOff, 'Electric')
    CALL SetIfBlank(fldValStart + ussSuppReHeatCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + ussSuppReHeatMaxODBOff, '21')
    CALL SetIfBlank(fldValStart + ussSuppReHeatEfficiencyOff, '0.8')
    CALL SetIfBlank(fldValStart + ussSuppReHeatParasiticOff, '0')
    CALL SetIfBlank(fldValStart + ussMaxOutsideFlowOff, 'autosize')
    CALL SetIfBlank(fldValStart + ussMinOutsideFlowOff, 'autosize')
    CALL SetIfBlank(fldValStart + ussEconomizerKindOff, 'NoEconomizer')
    CALL SetIfBlank(fldValStart + ussEconoLockoutKindOff, 'NoLockout')
    CALL SetIfBlank(fldValStart + ussHeatRecTypeOff, 'None')
    CALL SetIfBlank(fldValStart + ussHeatRecSenEffOff, '0.7')
    CALL SetIfBlank(fldValStart + ussHeatRecLatEffOff, '0.65')
    CALL SetIfBlank(fldValStart + ussHXTypeOff, 'Plate')
    CALL SetIfBlank(fldValStart + ussFrostControlOff, 'None')
    CALL SetIfBlank(fldValStart + ussDehumCtrlTypeOff, 'None')
    CALL SetIfBlank(fldValStart + ussDehumSetPtOff, '60')
    CALL SetIfBlank(fldValStart + ussHumidCtrlTypeOff, 'None')
    CALL SetIfBlank(fldValStart + ussHumidCapOff, '0.000001')
    CALL SetIfBlank(fldValStart + ussHumidRatedElecOff, 'autosize')
    CALL SetIfBlank(fldValStart + ussHumidSetPtOff, '30')
    CALL SetIfBlank(fldValStart + ussSizingOptionOff, 'NonCoincident')
    CALL SetIfBlank(fldValStart + ussReturnFanPresentOff, 'No')
    CALL SetIfBlank(fldValStart + ussReturnEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + ussReturnPressureOff, '300')
    CALL SetIfBlank(fldValStart + ussReturnFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + ussReturnFanFractionOff, '1')
  END IF
END DO
numCompactSysUnitarySystem = actCount
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputConstVol
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Michael Witte of GARD Analytics, Inc.
          !    DATE WRITTEN   September 2013
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the constant volume objects into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! cvsE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iZone
INTEGER    :: iSys

! ZONE
CALL CountOldObj('HVACTemplate:Zone:ConstantVolume',ptCompactZoneConstVol,prelimCount)
ALLOCATE(compactZoneConstVolBase(prelimCount))
actCount = 0
DO iZone= 1, prelimCount
  CALL NextOldObj(ptCompactZoneConstVol,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. cvzHeatDesignTempDiffOff) THEN
    actCount = actCount + 1
    compactZoneConstVolBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + cvzSupplyMaxRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + cvzOutAirTypeOff, 'flow/person')
    CALL SetIfBlank(fldValStart + cvzOutAirFlowPersonOff, '0.00944')
    CALL SetIfBlank(fldValStart + cvzOutAirFlowAreaOff, '0')
    CALL SetIfBlank(fldValStart + cvzOutAirFlowZoneOff, '0')
    CALL SetIfBlank(fldValStart + cvzReheatTypeOff, 'None')
    CALL SetIfBlank(fldValStart + cvzBaseboardTypeOff, 'None')
    CALL SetIfBlank(fldValStart + cvzBaseboardCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + cvzCoolDesignTempMethodOff, 'SystemSupplyAirTemperature')
    CALL SetIfBlank(fldValStart + cvzCoolDesignTempOff, '12.8')
    CALL SetIfBlank(fldValStart + cvzCoolDesignTempDiffOff, '11.11')
    CALL SetIfBlank(fldValStart + cvzHeatDesignTempMethodOff, 'SupplyAirTemperature')
    CALL SetIfBlank(fldValStart + cvzHeatDesignTempOff, '50')
    CALL SetIfBlank(fldValStart + cvzHeatDesignTempDiffOff, '30')
  END IF
END DO
numCompactZoneConstVol = actCount
! ConstVol SYSTEM
CALL CountOldObj('HVACTemplate:System:ConstantVolume',ptCompactSysConstVol,prelimCount)
ALLOCATE(compactSysConstVolBase(prelimCount))
actCount = 0
DO iSys = 1, prelimCount
  CALL NextOldObj(ptCompactSysConstVol,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. cvsReturnFanFractionOff) THEN
    actCount = actCount + 1
    compactSysConstVolBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + cvsSupplyRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + cvsSupplyEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + cvsSupplyPressureOff, '600')
    CALL SetIfBlank(fldValStart + cvsSupplyFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + cvsSupplyFanFractionOff, '1')
    CALL SetIfBlank(fldValStart + cvsSupplyFanPlaceOff, 'DrawThrough')
    CALL SetIfBlank(fldValStart + cvsCoolCoilKindOff, 'ChilledWater')
    CALL SetIfBlank(fldValStart + cvsCoolSetPtCtrlTypeOff, 'FixedSetpoint')
    CALL SetIfBlank(fldValStart + cvsCoolSetPtDesignOff, '12.8')
    CALL SetIfBlank(fldValStart + cvsCoolSetPtAtODBLoOff, '15.6')
    CALL SetIfBlank(fldValStart + cvsCoolResetODBLoOff, '15.6')
    CALL SetIfBlank(fldValStart + cvsCoolSetPtAtODBHiOff, '12.8')
    CALL SetIfBlank(fldValStart + cvsCoolResetODBHiOff, '23.3')
    CALL SetIfBlank(fldValStart + cvsHeatCoilKindOff, 'None')
    CALL SetIfBlank(fldValStart + cvsHeatSetPtCtrlTypeOff, 'FixedSetpoint')
    CALL SetIfBlank(fldValStart + cvsHeatSetPtDesignOff, '10')
    CALL SetIfBlank(fldValStart + cvsHeatSetPtAtODBLoOff, '15')
    CALL SetIfBlank(fldValStart + cvsHeatResetODBLoOff, '7.8')
    CALL SetIfBlank(fldValStart + cvsHeatSetPtAtODBHiOff, '12.2')
    CALL SetIfBlank(fldValStart + cvsHeatResetODBHiOff, '12.2')
    CALL SetIfBlank(fldValStart + cvsHeatCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + cvsHeatEffiencyOff, '0.8')
    CALL SetIfBlank(fldValStart + cvsHeatParasiticOff, '0')
    CALL SetIfBlank(fldValStart + cvsPreheatCoilKindOff, 'None')
    CALL SetIfBlank(fldValStart + cvsPreheatSetPtConstantOff, '7.2')
    CALL SetIfBlank(fldValStart + cvsPreheatEffiencyOff, '0.8')
    CALL SetIfBlank(fldValStart + cvsPreheatParasiticOff, '0')
    CALL SetIfBlank(fldValStart + cvsMaxOutsideFlowOff, 'autosize')
    CALL SetIfBlank(fldValStart + cvsMinOutsideFlowOff, 'autosize')
    CALL SetIfBlank(fldValStart + cvsEconomizerKindOff, 'None')
    CALL SetIfBlank(fldValStart + cvsNightCycleControlOff, 'StayOff')
    CALL SetIfBlank(fldValStart + cvsHeatRecTypeOff, 'None')
    CALL SetIfBlank(fldValStart + cvsHeatRecSenEffOff, '0.7')
    CALL SetIfBlank(fldValStart + cvsHeatRecLatEffOff, '0.65')
    CALL SetIfBlank(fldValStart + cvsHXTypeOff, 'Plate')
    CALL SetIfBlank(fldValStart + cvsFrostControlOff, 'None')
    CALL SetIfBlank(fldValStart + cvsDehumCtrlTypeOff, 'None')
    CALL SetIfBlank(fldValStart + cvsDehumSetPtOff, '60')
    CALL SetIfBlank(fldValStart + cvsHumidCtrlTypeOff, 'None')
    CALL SetIfBlank(fldValStart + cvsHumidCapOff, '0.000001')
    CALL SetIfBlank(fldValStart + cvsHumidRatedElecOff, 'autosize')
    CALL SetIfBlank(fldValStart + cvsHumidSetPtOff, '30')
    CALL SetIfBlank(fldValStart + cvsReturnFanPresentOff, 'No')
    CALL SetIfBlank(fldValStart + cvsReturnEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + cvsReturnPressureOff, '300')
    CALL SetIfBlank(fldValStart + cvsReturnFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + cvsReturnFanFractionOff, '1')
  END IF
END DO
numCompactSysConstVol = actCount
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputDualDuct
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Michael Witte of GARD Analytics, Inc.
          !    DATE WRITTEN   September 2013
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the constant volume objects into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! ddsE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iZone
INTEGER    :: iSys

! ZONE
CALL CountOldObj('HVACTemplate:Zone:DualDuct',ptCompactZoneDualDuct,prelimCount)
ALLOCATE(compactZoneDualDuctBase(prelimCount))
actCount = 0
DO iZone= 1, prelimCount
  CALL NextOldObj(ptCompactZoneDualDuct,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. ddzHeatDesignTempDiffOff) THEN
    actCount = actCount + 1
    compactZoneDualDuctBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + ddzSupplyMaxRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + ddzConstantMinFracOff, '0.2')
    CALL SetIfBlank(fldValStart + ddzOutAirTypeOff, 'flow/person')
    CALL SetIfBlank(fldValStart + ddzOutAirFlowPersonOff, '0.00944')
    CALL SetIfBlank(fldValStart + ddzOutAirFlowAreaOff, '0')
    CALL SetIfBlank(fldValStart + ddzOutAirFlowZoneOff, '0')
    CALL SetIfBlank(fldValStart + ddzBaseboardTypeOff, 'None')
    CALL SetIfBlank(fldValStart + ddzBaseboardCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + ddzCoolDesignTempMethodOff, 'SystemSupplyAirTemperature')
    CALL SetIfBlank(fldValStart + ddzCoolDesignTempOff, '12.8')
    CALL SetIfBlank(fldValStart + ddzCoolDesignTempDiffOff, '11.11')
    CALL SetIfBlank(fldValStart + ddzHeatDesignTempMethodOff, 'SystemSupplyAirTemperature')
    CALL SetIfBlank(fldValStart + ddzHeatDesignTempOff, '50')
    CALL SetIfBlank(fldValStart + ddzHeatDesignTempDiffOff, '30')
  END IF
END DO
numCompactZoneDualDuct = actCount
! DualDuct SYSTEM
CALL CountOldObj('HVACTemplate:System:DualDuct',ptCompactSysDualDuct,prelimCount)
ALLOCATE(compactSysDualDuctBase(prelimCount))
actCount = 0
DO iSys = 1, prelimCount
  CALL NextOldObj(ptCompactSysDualDuct,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. ddsReturnFanPLRCoeffOff) THEN
    actCount = actCount + 1
    compactSysDualDuctBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + ddsSupplyFanConfigTypeOff, 'SingleConstantVolume')
    CALL SetIfBlank(fldValStart + ddsSupplyMaxRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + ddsSupplyMinFracOff, '0.2')
    CALL SetIfBlank(fldValStart + ddsSupplyEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + ddsSupplyPressureOff, '1000')
    CALL SetIfBlank(fldValStart + ddsSupplyFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + ddsSupplyFanFractionOff, '1')
    CALL SetIfBlank(fldValStart + ddsSupplyFanPLRCoeffOff, 'InletVaneDampers')
    CALL SetIfBlank(fldValStart + ddsColdMaxRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + ddsColdMinFracOff, '0.2')
    CALL SetIfBlank(fldValStart + ddsColdEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + ddsColdPressureOff, '1000')
    CALL SetIfBlank(fldValStart + ddsColdFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + ddsColdFanFractionOff, '1')
    CALL SetIfBlank(fldValStart + ddsColdFanPLRCoeffOff, 'InletVaneDampers')
    CALL SetIfBlank(fldValStart + ddsColdFanPlaceOff, 'DrawThrough')
    CALL SetIfBlank(fldValStart + ddsHotMaxRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + ddsHotMinFracOff, '0.2')
    CALL SetIfBlank(fldValStart + ddsHotEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + ddsHotPressureOff, '1000')
    CALL SetIfBlank(fldValStart + ddsHotFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + ddsHotFanFractionOff, '1')
    CALL SetIfBlank(fldValStart + ddsHotFanPLRCoeffOff, 'InletVaneDampers')
    CALL SetIfBlank(fldValStart + ddsHotFanPlaceOff, 'DrawThrough')
    CALL SetIfBlank(fldValStart + ddsCoolCoilKindOff, 'ChilledWater')
    CALL SetIfBlank(fldValStart + ddsCoolSetPtCtrlTypeOff, 'FixedSetpoint')
    CALL SetIfBlank(fldValStart + ddsCoolSetPtDesignOff, '12.8')
    CALL SetIfBlank(fldValStart + ddsCoolSetPtAtODBLoOff, '15.6')
    CALL SetIfBlank(fldValStart + ddsCoolResetODBLoOff, '15.6')
    CALL SetIfBlank(fldValStart + ddsCoolSetPtAtODBHiOff, '12.8')
    CALL SetIfBlank(fldValStart + ddsCoolResetODBHiOff, '23.3')
    CALL SetIfBlank(fldValStart + ddsHeatCoilKindOff, 'HotWater')
    CALL SetIfBlank(fldValStart + ddsHeatSetPtCtrlTypeOff, 'FixedSetpoint')
    CALL SetIfBlank(fldValStart + ddsHeatSetPtDesignOff, '50')
    CALL SetIfBlank(fldValStart + ddsHeatSetPtAtODBLoOff, '50')
    CALL SetIfBlank(fldValStart + ddsHeatResetODBLoOff, '7.8')
    CALL SetIfBlank(fldValStart + ddsHeatSetPtAtODBHiOff, '12.2')
    CALL SetIfBlank(fldValStart + ddsHeatResetODBHiOff, '20.0')
    CALL SetIfBlank(fldValStart + ddsHeatCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + ddsHeatEffiencyOff, '0.8')
    CALL SetIfBlank(fldValStart + ddsHeatParasiticOff, '0')
    CALL SetIfBlank(fldValStart + ddsPreheatCoilKindOff, 'None')
    CALL SetIfBlank(fldValStart + ddsPreheatSetPtConstantOff, '7.2')
    CALL SetIfBlank(fldValStart + ddsPreheatEffiencyOff, '0.8')
    CALL SetIfBlank(fldValStart + ddsPreheatParasiticOff, '0')
    CALL SetIfBlank(fldValStart + ddsMaxOutsideFlowOff, 'autosize')
    CALL SetIfBlank(fldValStart + ddsMinOutsideFlowOff, 'autosize')
    CALL SetIfBlank(fldValStart + ddsMinOutsideControlOff, 'ProportionalMinimum')
    CALL SetIfBlank(fldValStart + ddsEconomizerKindOff, 'None')
    CALL SetIfBlank(fldValStart + ddsEconoLockoutKindOff, 'NoLockout')
    CALL SetIfBlank(fldValStart + ddsNightCycleControlOff, 'StayOff')
    CALL SetIfBlank(fldValStart + ddsHeatRecTypeOff, 'None')
    CALL SetIfBlank(fldValStart + ddsHeatRecSenEffOff, '0.7')
    CALL SetIfBlank(fldValStart + ddsHeatRecLatEffOff, '0.65')
    CALL SetIfBlank(fldValStart + ddsHXTypeOff, 'Plate')
    CALL SetIfBlank(fldValStart + ddsFrostControlOff, 'None')
    CALL SetIfBlank(fldValStart + ddsDehumCtrlTypeOff, 'None')
    CALL SetIfBlank(fldValStart + ddsDehumSetPtOff, '60')
    CALL SetIfBlank(fldValStart + ddsHumidCtrlTypeOff, 'None')
    CALL SetIfBlank(fldValStart + ddsHumidCapOff, '0.000001')
    CALL SetIfBlank(fldValStart + ddsHumidRatedElecOff, 'autosize')
    CALL SetIfBlank(fldValStart + ddsHumidSetPtOff, '30')
    CALL SetIfBlank(fldValStart + ddsSizingOptionOff, 'NonCoincident')
    CALL SetIfBlank(fldValStart + ddsReturnFanPresentOff, 'No')
    CALL SetIfBlank(fldValStart + ddsReturnEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + ddsReturnPressureOff, '300')
    CALL SetIfBlank(fldValStart + ddsReturnFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + ddsReturnFanFractionOff, '1')
    CALL SetIfBlank(fldValStart + ddsReturnFanPLRCoeffOff, 'InletVaneDampers')
  END IF
END DO
numCompactSysDualDuct = actCount
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputVRF
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Michael Witte of GARD Analytics, Inc.
          !    DATE WRITTEN   September 2013
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the constant volume objects into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! vrfsE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iZone
INTEGER    :: iSys

! ZONE
CALL CountOldObj('HVACTemplate:Zone:VRF',ptCompactZoneVRF,prelimCount)
ALLOCATE(compactZoneVRFBase(prelimCount))
actCount = 0
DO iZone= 1, prelimCount
  CALL NextOldObj(ptCompactZoneVRF,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. vrfzBaseboardCapOff) THEN
    actCount = actCount + 1
    compactZoneVRFBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + vrfzHeatCapSizRatOff, '1')
    CALL SetIfBlank(fldValStart + vrfzSupplyCoolFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + vrfzSupplyNoCoolFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + vrfzSupplyHeatFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + vrfzSupplyNoHeatFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + vrfzOACoolFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + vrfzOAHeatFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + vrfzOANoLoadFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + vrfzOutAirTypeOff, 'flow/person')
    CALL SetIfBlank(fldValStart + vrfzOutAirFlowPersonOff, '0.00944')
    CALL SetIfBlank(fldValStart + vrfzOutAirFlowAreaOff, '0')
    CALL SetIfBlank(fldValStart + vrfzOutAirFlowZoneOff, '0')
    CALL SetIfBlank(fldValStart + vrfzFanPlacementOff, 'BlowThrough')
    CALL SetIfBlank(fldValStart + vrfzFanTotEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + vrfzFanPressureOff, '75')
    CALL SetIfBlank(fldValStart + vrfzFanMotorEfficiencyOff, '0.9')
    CALL SetIfBlank(fldValStart + vrfzCoolCoilKindOff, 'VariableRefrigerantFlowDX')
    CALL SetIfBlank(fldValStart + vrfzHPHeatCoilKindOff, 'VariableRefrigerantFlowDX')
    CALL SetIfBlank(fldValStart + vrfzTermUnitOnParasiticElecOff, '0')
    CALL SetIfBlank(fldValStart + vrfzTermUnitOffParasiticElecOff, '0')
    CALL SetIfBlank(fldValStart + vrfzCoolDesignTempMethodOff, 'SupplyAirTemperature')
    CALL SetIfBlank(fldValStart + vrfzCoolDesignTempOff, '14')
    CALL SetIfBlank(fldValStart + vrfzCoolDesignTempDiffOff, '11.11')
    CALL SetIfBlank(fldValStart + vrfzHeatDesignTempMethodOff, 'SupplyAirTemperature')
    CALL SetIfBlank(fldValStart + vrfzHeatDesignTempOff, '50')
    CALL SetIfBlank(fldValStart + vrfzHeatDesignTempDiffOff, '30')
    CALL SetIfBlank(fldValStart + vrfzBaseboardTypeOff, 'None')
    CALL SetIfBlank(fldValStart + vrfzBaseboardCapOff, 'autosize')
  END IF
END DO
numCompactZoneVRF = actCount
! VRF SYSTEM
CALL CountOldObj('HVACTemplate:System:VRF',ptCompactSysVRF,prelimCount)
ALLOCATE(compactSysVRFBase(prelimCount))
actCount = 0
DO iSys = 1, prelimCount
  CALL NextOldObj(ptCompactSysVRF,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. vrfsHeatRecMaxODBOff) THEN
    actCount = actCount + 1
    compactSysVRFBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + vrfsCoolCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + vrfsCoolCoilCOPOff, '3.3')
    CALL SetIfBlank(fldValStart + vrfsMinTempCoolingOff, '-6')
    CALL SetIfBlank(fldValStart + vrfsMaxTempCoolingOff, '43')
    CALL SetIfBlank(fldValStart + vrfsHPHeatCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + vrfsHeatCapSizRatOff, '1')
    CALL SetIfBlank(fldValStart + vrfsHPHeatCoilCOPOff, '2.75')
    CALL SetIfBlank(fldValStart + vrfsHPHeatMinODBOff, '-20')
    CALL SetIfBlank(fldValStart + vrfsHPHeatMaxODBOff, '16')
    CALL SetIfBlank(fldValStart + vrfsHPMinPLROff, '0.15')
    CALL SetIfBlank(fldValStart + vrfsTstatPriorityCtrlTypeOff, 'MasterThermostatPriority')
    CALL SetIfBlank(fldValStart + vrfsHPWasteHeatRecoveryOff, 'No')
    CALL SetIfBlank(fldValStart + vrfsEquivPipLenCoolOff, '30')
    CALL SetIfBlank(fldValStart + vrfsEquivPipHeightOff, '10')
    CALL SetIfBlank(fldValStart + vrfsEquivPipLenHeatOff, '30')
    CALL SetIfBlank(fldValStart + vrfsCrankHtPwrOff, '33')
    CALL SetIfBlank(fldValStart + vrfsNumComprOff, '2')
    CALL SetIfBlank(fldValStart + vrfsComprSizeRatioOff, '0.5')
    CALL SetIfBlank(fldValStart + vrfsCrankHtMaxODBOff, '5')
    CALL SetIfBlank(fldValStart + vrfsHPDefrTypeOff, 'Resistive')
    CALL SetIfBlank(fldValStart + vrfsHPDefrCtrlTypeOff, 'Timed')
    CALL SetIfBlank(fldValStart + vrfsHPDefrTimeFracOff, '0.058333')
    CALL SetIfBlank(fldValStart + vrfsHPDefrHeatCapOff, '0')
    CALL SetIfBlank(fldValStart + vrfsHPDefrMaxODBOff, '5')
    CALL SetIfBlank(fldValStart + vrfsCondenserTypeOff, 'AirCooled')
    CALL SetIfBlank(fldValStart + vrfsWaterCondVolFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + vrfsEvapCondEffectOff, '0.9')
    CALL SetIfBlank(fldValStart + vrfsEvapCondAirFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + vrfsEvapCondPumpPwrOff, '0')
    CALL SetIfBlank(fldValStart + vrfsBasinHeatCapOff, '0')
    CALL SetIfBlank(fldValStart + vrfsBasinHeatSetPtOff, '2')
    CALL SetIfBlank(fldValStart + vrfsFuelTypeOff, 'Electricity')
    CALL SetIfBlank(fldValStart + vrfsHeatRecMinODBOff, '-15')
    CALL SetIfBlank(fldValStart + vrfsHeatRecMaxODBOff, '45')
  END IF
END DO
numCompactSysVRF = actCount
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputFanCoil
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   March 2006
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the fan coil objects into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iZone

! ZONE
CALL CountOldObj('HVACTemplate:Zone:FanCoil',ptCompactFanCoil,prelimCount)
ALLOCATE(compactFanCoilBase(prelimCount))
actCount = 0
DO iZone= 1, prelimCount
  CALL NextOldObj(ptCompactFanCoil,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. fczBaseboardCapOff) THEN
    actCount = actCount + 1
    compactFanCoilBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + fczSupplyMaxRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + fczOutAirTypeOff, 'Flow/Person')
    CALL SetIfBlank(fldValStart + fczOutAirFlowPersonOff, '0.00944')
    CALL SetIfBlank(fldValStart + fczOutAirFlowAreaOff, '0.0')
    CALL SetIfBlank(fldValStart + fczOutAirFlowZoneOff, '0.0')
    CALL SetIfBlank(fldValStart + fczSupplyEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + fczSupplyPressureOff, '75')
    CALL SetIfBlank(fldValStart + fczSupplyFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + fczSupplyFanMotorFracOff, '1.0')
    CALL SetIfBlank(fldValStart + fczCoolCoilKindOff, 'ChilledWater')
    CALL SetIfBlank(fldValStart + fczCoolSetPtDesignOff, '14')
    CALL SetIfBlank(fldValStart + fczHeatCoilKindOff, 'HotWater')
    CALL SetIfBlank(fldValStart + fczHeatSetPtDesignOff, '50')
    CALL SetIfBlank(fldValStart + fczCoolDesignTempMethodOff, 'SupplyAirTemperature')
    CALL SetIfBlank(fldValStart + fczCoolDesignTempDiffOff,   '11.11')
    CALL SetIfBlank(fldValStart + fczHeatDesignTempMethodOff, 'SupplyAirTemperature')
    CALL SetIfBlank(fldValStart + fczCoolDesignTempDiffOff,   '30.0')
    CALL SetIfBlank(fldValStart + fczLowSpdSupplyRatioOff, '0.33')
    CALL SetIfBlank(fldValStart + fczMedSpdSupplyRatioOff, '0.66')
    CALL SetIfBlank(fldValStart + fczBaseboardTypeOff, 'None')
    CALL SetIfBlank(fldValStart + fczBaseboardCapOff, 'autosize')
  END IF
END DO
numCompactFanCoil = actCount
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputPurchasedAir
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the compact hvac purchased air object into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iObj

CALL CountOldObj('HVACTemplate:Zone:IdealLoadsAirSystem',ptCompactPurchAir,prelimCount)

ALLOCATE(compactPurchAirBase(prelimCount))
actCount = 0
DO iObj= 1, prelimCount
  CALL NextOldObj(ptCompactPurchAir,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. pazHeatRecLatEffOff) THEN
    actCount = actCount + 1
    compactPurchAirBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + pazMaxHeatSupplyTempOff, '50')
    CALL SetIfBlank(fldValStart + pazMinCoolSupplyTempOff, '13')
    CALL SetIfBlank(fldValStart + pazMaxHeatSupplyHumRatOff, '0.0156')
    CALL SetIfBlank(fldValStart + pazMinCoolSupplyHumRatOff, '0.0077')
    CALL SetIfBlank(fldValStart + pazHeatLimitOff, 'NoLimit')
    CALL SetIfBlank(fldValStart + pazCoolLimitOff, 'NoLimit')
    CALL SetIfBlank(fldValStart + pazDehumCtrlTypeOff, 'ConstantSensibleHeatRatio ')
    CALL SetIfBlank(fldValStart + pazCoolSHROff, '0.7')
    CALL SetIfBlank(fldValStart + pazDehumSetPtOff, '60')
    CALL SetIfBlank(fldValStart + pazHumidCtrlTypeOff, 'None')
    CALL SetIfBlank(fldValStart + pazHumidSetPtOff, '30')
    CALL SetIfBlank(fldValStart + pazOutAirTypeOff, 'None')
    CALL SetIfBlank(fldValStart + pazOutAirFlowPersonOff, '0.00944')
    CALL SetIfBlank(fldValStart + pazOutAirFlowAreaOff, '0')
    CALL SetIfBlank(fldValStart + pazOutAirFlowZoneOff, '0')
    CALL SetIfBlank(fldValStart + pazDCVTypeOff, 'None')
    CALL SetIfBlank(fldValStart + pazEconomizerTypeOff, 'NoEconomizer')
    CALL SetIfBlank(fldValStart + pazHeatRecTypeOff, 'None')
    CALL SetIfBlank(fldValStart + pazHeatRecSenEffOff, '0.7')
    CALL SetIfBlank(fldValStart + pazHeatRecLatEffOff, '0.65')
  END IF
END DO
numCompactPurchAir = actCount
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputBaseboard
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Michael Witte of GARD Analytics, Inc.
          !    DATE WRITTEN   September 2013
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the fan coil objects into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iZone

! ZONE
CALL CountOldObj('HVACTemplate:Zone:BaseboardHeat',ptCompactBaseboard,prelimCount)
ALLOCATE(compactBaseboardBase(prelimCount))
actCount = 0
DO iZone= 1, prelimCount
  CALL NextOldObj(ptCompactBaseboard,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. bbzDesSpecOADistNameOff) THEN
    actCount = actCount + 1
    compactBaseboardBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + bbzBaseboardTypeOff, 'HotWater')
    CALL SetIfBlank(fldValStart + bbzBaseboardCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + bbzOutAirTypeOff, 'flow/person')
    CALL SetIfBlank(fldValStart + bbzOutAirFlowPersonOff, '0.00944')
    CALL SetIfBlank(fldValStart + bbzOutAirFlowAreaOff, '0')
    CALL SetIfBlank(fldValStart + bbzOutAirFlowZoneOff, '0')
  END IF
END DO
numCompactBaseboard = actCount
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputPTAC
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   September 2007
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the packaged terminal air conditioner objects into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iZone

CALL CountOldObj('HVACTemplate:Zone:PTAC',ptCompactPTAC,prelimCount)
ALLOCATE(compactPTACBase(prelimCount))
actCount = 0
DO iZone= 1, prelimCount
  CALL NextOldObj(ptCompactPTAC,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. ptaczCapCtrlTypeOff) THEN
    actCount = actCount + 1
    compactPTACBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + ptaczSupplyCoolFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + ptaczSupplyHeatFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + ptaczOutAirTypeOff, 'Flow/Person')
    CALL SetIfBlank(fldValStart + ptaczOutAirFlowPersonOff, '0.00944')
    CALL SetIfBlank(fldValStart + ptaczOutAirFlowAreaOff, '0.0')
    CALL SetIfBlank(fldValStart + ptaczOutAirFlowZoneOff, '0.0')
    CALL SetIfBlank(fldValStart + ptaczFanPlacementOff, 'DrawThrough')
    CALL SetIfBlank(fldValStart + ptaczFanTotEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + ptaczFanPressureOff, '75')
    CALL SetIfBlank(fldValStart + ptaczFanMotorEfficiencyOff, '0.9')
    CALL SetIfBlank(fldValStart + ptaczCoolCoilKindOff, 'SingleSpeedDX')
    CALL SetIfBlank(fldValStart + ptaczCoolCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + ptaczCoolCoilSHROff, 'autosize')
    CALL SetIfBlank(fldValStart + ptaczCoolCoilCOPOff, '3')
    CALL SetIfBlank(fldValStart + ptaczHeatCoilKindOff, 'Electric')
    CALL SetIfBlank(fldValStart + ptaczHeatCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + ptaczHeatEfficiencyOff, '0.8')
    CALL SetIfBlank(fldValStart + ptaczHeatParasiticOff, '0')
    CALL SetIfBlank(fldValStart + ptaczCoolDesignTempMethodOff, 'SupplyAirTemperature')
    CALL SetIfBlank(fldValStart + ptaczCoolDesignTempOff, '14.0')
    CALL SetIfBlank(fldValStart + ptaczCoolDesignTempDiffOff,   '11.11')
    CALL SetIfBlank(fldValStart + ptaczHeatDesignTempMethodOff, 'SupplyAirTemperature')
    CALL SetIfBlank(fldValStart + ptaczHeatDesignTempOff, '50.0')
    CALL SetIfBlank(fldValStart + ptaczCoolDesignTempDiffOff,   '30.0')
    CALL SetIfBlank(fldValStart + ptaczBaseboardTypeOff, 'None')
    CALL SetIfBlank(fldValStart + ptaczBaseboardCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + ptaczCapCtrlTypeOff, 'None')
  END IF
END DO
numCompactPTAC = actCount
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputPTHP
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   September 2007
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the packaged terminal heat pumps objects into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iZone

CALL CountOldObj('HVACTemplate:Zone:PTHP',ptCompactPTHP,prelimCount)
ALLOCATE(compactPTHPBase(prelimCount))
actCount = 0
DO iZone= 1, prelimCount
  CALL NextOldObj(ptCompactPTHP,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. pthpzCapCtrlTypeOff) THEN
    actCount = actCount + 1
    compactPTHPBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + pthpzSupplyCoolFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + pthpzSupplyHeatFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + pthpzOutAirTypeOff, 'Flow/Person')
    CALL SetIfBlank(fldValStart + pthpzOutAirFlowPersonOff, '0.00944')
    CALL SetIfBlank(fldValStart + pthpzOutAirFlowAreaOff, '0.0')
    CALL SetIfBlank(fldValStart + pthpzOutAirFlowZoneOff, '0.0')
    CALL SetIfBlank(fldValStart + pthpzFanPlacementOff, 'DrawThrough')
    CALL SetIfBlank(fldValStart + pthpzFanTotEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + pthpzFanPressureOff, '75')
    CALL SetIfBlank(fldValStart + pthpzFanMotorEfficiencyOff, '0.9')
    CALL SetIfBlank(fldValStart + pthpzCoolCoilKindOff, 'SingleSpeedDX')
    CALL SetIfBlank(fldValStart + pthpzCoolCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + pthpzCoolCoilSHROff, 'autosize')
    CALL SetIfBlank(fldValStart + pthpzCoolCoilCOPOff, '3')
    CALL SetIfBlank(fldValStart + pthpzHPHeatCoilKindOff, 'SingleSpeedDXHeatPump')
    CALL SetIfBlank(fldValStart + pthpzHPHeatCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + pthpzHPHeatCoilCOPOff, '2.75')
    CALL SetIfBlank(fldValStart + pthpzHPHeatMinODBOff, '-8')
    CALL SetIfBlank(fldValStart + pthpzHPDefrMaxODBOff, '5')
    CALL SetIfBlank(fldValStart + pthpzHPDefrTypeOff, 'ReverseCycle')
    CALL SetIfBlank(fldValStart + pthpzHPDefrCtrlTypeOff, 'Timed')
    CALL SetIfBlank(fldValStart + pthpzHPDefrTimeFracOff, '0.058333')
    CALL SetIfBlank(fldValStart + pthpzSuppHeatCoilKindOff, 'Electric')
    CALL SetIfBlank(fldValStart + pthpzSuppHeatCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + pthpzSuppHeatMaxODBOff, '21')
    CALL SetIfBlank(fldValStart + pthpzSuppHeatEfficiencyOff, '0.8')
    CALL SetIfBlank(fldValStart + pthpzSuppHeatParasiticOff, '0')
    CALL SetIfBlank(fldValStart + pthpzCoolDesignTempMethodOff, 'SupplyAirTemperature')
    CALL SetIfBlank(fldValStart + pthpzCoolDesignTempOff, '14.0')
    CALL SetIfBlank(fldValStart + pthpzCoolDesignTempDiffOff,   '11.11')
    CALL SetIfBlank(fldValStart + pthpzHeatDesignTempMethodOff, 'SupplyAirTemperature')
    CALL SetIfBlank(fldValStart + pthpzHeatDesignTempOff, '50.0')
    CALL SetIfBlank(fldValStart + pthpzCoolDesignTempDiffOff,   '30.0')
    CALL SetIfBlank(fldValStart + pthpzBaseboardTypeOff, 'None')
    CALL SetIfBlank(fldValStart + pthpzBaseboardCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + pthpzCapCtrlTypeOff, 'None')
  END IF
END DO
numCompactPTHP = actCount
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE GetInputBoiler
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the COMPACT HVAC:PLANT:BOILER object into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iObj

CALL CountOldObj('HVACTemplate:Plant:Boiler',ptCompactBoiler,prelimCount)
ALLOCATE(compactBoilerBase(prelimCount))
actCount = 0
DO iObj= 1, prelimCount
  CALL NextOldObj(ptCompactBoiler,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. blrLoopTypeOff) THEN
    actCount = actCount + 1
    compactBoilerBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + blrTypeOff, 'HotWaterBoiler')
    CALL SetIfBlank(fldValStart + blrCapacityOff, 'autosize')
    CALL SetIfBlank(fldValStart + blrEfficOff, '0.8')
    CALL SetIfBlank(fldValStart + blrSizeFactorOff, '1.0')
    CALL SetIfBlank(fldValStart + blrMinPLROff, '0.0')
    CALL SetIfBlank(fldValStart + blrMaxPLROff, '1.1')
    CALL SetIfBlank(fldValStart + blrOptPLROff, '1.0')
    CALL SetIfBlank(fldValStart + blrOutletUpLimOff, '100.0')
  END IF
END DO
numCompactBoiler = actCount

CALL CountOldObj('HVACTemplate:Plant:Boiler:ObjectReference',ptCompactBoilerOR,prelimCount)
ALLOCATE(compactBoilerORBase(prelimCount))
actCount = 0
DO iObj= 1, prelimCount
  CALL NextOldObj(ptCompactBoilerOR,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. blrorLoopTypeOff) THEN
    actCount = actCount + 1
    compactBoilerORBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + blrorObjTypeOff, 'Boiler:HotWater')
  END IF
END DO
numCompactBoilerOR = actCount

! If there are HVACTemplate:Plant:Boiler:ObjectReference present, then get detailed boiler objects
IF (numCompactBoilerOR > 0) THEN
  CALL CountOldObj('Boiler:HotWater',ptDetBoilerHW,prelimCount)
  ALLOCATE(detBoilerHWBase(prelimCount))
  actCount = 0
  DO iObj= 1, prelimCount
    CALL NextOldObj(ptDetBoilerHW,fldValStart,fldValEnd)
    ! test to make sure object is valid
    IF (fldValEnd - fldValStart .EQ. blrhwLastFieldOff) THEN
      actCount = actCount + 1
      detBoilerHWBase(actCount) = fldValStart
      !fill in defaults
    END IF
  END DO
  numDetBoilerHW = actCount
ENDIF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputHotLoop
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the compact hot water loop object into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iObj

CALL CountOldObj('HVACTemplate:Plant:HotWaterLoop',ptCompactHotLoop,prelimCount)
IF (prelimCount .GT. 1) THEN
  CALL WriteError('WARNING: Only one HVACTemplate:Plant:HotWaterLoop allowed per file.')
ELSEIF (prelimCount .EQ. 1) THEN
  CALL NextOldObj(ptCompactHotLoop,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. hwpLdDistSchmOff) THEN
    compactHotLoopBase = fldValStart
    foundCompactObject = .TRUE.
    !check if loop exist but system doesn't
    IF ((numCompactSysVAV .EQ. 0) .AND. (numCompactFanCoil .EQ. 0) .AND.  &
        (numCompactZoneUnit .EQ. 0) .AND. (numCompactPTAC .EQ. 0) .AND. &
         (numCompactPTHP .EQ. 0) .AND. (numCompactWaterAirHP .EQ. 0) .AND. (numCompactSysVRF .EQ. 0) .AND.&
         (numCompactDedOutAir .EQ. 0) .AND. (numCompactSysPVAV .EQ. 0) .AND. &
         (numCompactSysUnitHP .EQ. 0) .AND. (numCompactSysConstVol .EQ. 0) .AND. &
         (numCompactSysDualDuct .EQ. 0) .AND. ((numCompactSysUnitarySystem .EQ. 0) .AND. numCompactBaseBoard .EQ. 0)) THEN
      CALL WriteError('You must specify at least one HVACTemplate:System:VAV or HVACTemplate:Zone:FanCoil ' // &
                      'or HVACTemplate:Zone:Unitary or HVACTemplate:Zone:PTAC or HVACTemplate:Zone:PTHP ' // &
                      'or HVACTemplate:Zone:WaterToAirHeatPump or HVACTemplate:System:UnitaryHeatPump:AirToAir ' // &
                      'or HVACTemplate:System:PackagedVAV or HVACTemplate:System:DedicatedOutdoorAir' // &
                      'or HVACTemplate:System:ConstantVolume or HVACTemplate:System:DualDuct' // &
                      'or HVACTemplate:Zone:Baseboard or HVACTemplate:System:UnitarySystem' // &
                      'or HVACTemplate:System:VRF if a HVACTemplate:Plant:HotWaterLoop is defined.')
    END IF
    !fill in defaults
    CALL SetIfBlank(fldValStart + hwpPumpControlTypeOff, 'Intermittent')
    CALL SetIfBlank(fldValStart + hwpHWOperTypeOff, 'Default')
    CALL SetIfBlank(fldValStart + hwpHWSetPtDesignOff, '82.0')
    CALL SetIfBlank(fldValStart + hwpHWPumpConfigOff, 'ConstantFlow')
    CALL SetIfBlank(fldValStart + hwpHWPumpHeadOff, '179352')
    CALL SetIfBlank(fldValStart + hwpHWSetPtResetTypeOff, 'None')
    CALL SetIfBlank(fldValStart + hwpHWSetPtAtODBLoOff, '82.2')
    CALL SetIfBlank(fldValStart + hwpHWODBLoOff, '-6.7')
    CALL SetIfBlank(fldValStart + hwpHWSetPtAtODBHiOff, '65.6')
    CALL SetIfBlank(fldValStart + hwpHWODBHiOff, '10.0')
    CALL SetIfBlank(fldValStart + hwpHWPumpTypeOff, 'SinglePump')
    CALL SetIfBlank(fldValStart + hwpSupplyBypassOff, 'Yes')
    CALL SetIfBlank(fldValStart + hwpDemandBypassOff, 'Yes')
    CALL SetIfBlank(fldValStart + hwpFluidTypeOff, 'Water')
    CALL SetIfBlank(fldValStart + hwpDesDeltaTOff, '11.0')
    CALL SetIfBlank(fldValStart + hwpLdDistSchmOff, 'SequentialLoad')
  END IF
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputChldLoop
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the compact chilled water loop object into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iObj

CALL CountOldObj('HVACTemplate:Plant:ChilledWaterLoop',ptCompactChldLoop,prelimCount)
IF (prelimCount .GT. 1) THEN
  CALL WriteError('WARNING: Only one HVACTemplate:Plant:ChilledWaterLoop allowed per file.')
ELSEIF (prelimCount .EQ. 1) THEN
  CALL NextOldObj(ptCompactChldLoop,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. cwpCndLdDistSchmOff) THEN
    compactChldLoopBase = fldValStart
    foundCompactObject = .TRUE.
    !check if loop exist but system doesn't
    IF ((numCompactSysVAV .EQ. 0) .AND. (numCompactFanCoil .EQ. 0) .AND. &
        (numCompactDedOutAir .EQ. 0) .AND. (numCompactSysConstVol .EQ. 0) .AND. &
        (numCompactSysDualDuct .EQ. 0) .AND. (numCompactSysUnitarySystem .EQ. 0)) THEN
      CALL WriteError('You must specify at least one HVACTemplate:System:VAV or '//  &
         'HVACTemplate:Zone:FanCoil or HVACTemplate:System:DedicatedOutdoorAir or ' // &
         'HVACTemplate:System:ConstantVolume or HVACTemplate:System:DualDuct ' // &
         'or HVACTemplate:System:UnitarySystem if a HVACTemplate:Plant:ChilledWaterLoop is defined.')
    END IF
    !fill in defaults
    CALL SetIfBlank(fldValStart + cwpPumpControlTypeOff, 'Intermittent')
    CALL SetIfBlank(fldValStart + cwpChWOperTypeOff, 'Default')
    CALL SetIfBlank(fldValStart + cwpChWSetPtDesignOff, '7.22')
    CALL SetIfBlank(fldValStart + cwpChWPumpConfigOff, 'ConstantFlow')
    CALL SetIfBlank(fldValStart + cwpPriChWPumpHeadOff, '179352')
    CALL SetIfBlank(fldValStart + cwpSecChWPumpHeadOff, '179352')
    CALL SetIfBlank(fldValStart + cwpCndOperTypeOff, 'Default')
    CALL SetIfBlank(fldValStart + cwpCndSetPtDesignOff, '29.4')
    CALL SetIfBlank(fldValStart + cwpCndPumpHeadOff, '179352')
    CALL SetIfBlank(fldValStart + cwpChWSetPtResetTypeOff, 'None')
    CALL SetIfBlank(fldValStart + cwpChWSetPtAtODBLoOff, '12.2')
    CALL SetIfBlank(fldValStart + cwpChWODBLoOff, '15.6')
    CALL SetIfBlank(fldValStart + cwpChWSetPtAtODBHiOff, '6.7')
    CALL SetIfBlank(fldValStart + cwpChWODBHiOff, '26.7')
    CALL SetIfBlank(fldValStart + cwpPriChWPumpTypeOff, 'SinglePump')
    CALL SetIfBlank(fldValStart + cwpSecChWPumpTypeOff, 'SinglePump')
    CALL SetIfBlank(fldValStart + cwpCndPumpTypeOff, 'SinglePump')
    CALL SetIfBlank(fldValStart + cwpChWSupplyBypassOff, 'Yes')
    CALL SetIfBlank(fldValStart + cwpChwDemandBypassOff, 'Yes')
    CALL SetIfBlank(fldValStart + cwpCndSupplyBypassOff, 'Yes')
    CALL SetIfBlank(fldValStart + cwpCndDemandBypassOff, 'Yes')
    CALL SetIfBlank(fldValStart + cwpFluidTypeOff, 'Water')
    CALL SetIfBlank(fldValStart + cwpDesDeltaTOff, '6.67')
    CALL SetIfBlank(fldValStart + cwpChwLdDistSchmOff, 'SequentialLoad')
    CALL SetIfBlank(fldValStart + cwpCndLdDistSchmOff, 'SequentialLoad')
  END IF
END IF
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE GetInputChiller
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the COMPACT COMPACT HVAC:PLANT:CHILLER object into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iObj

CALL CountOldObj('HVACTemplate:Plant:Chiller',ptCompactChiller,prelimCount)
ALLOCATE(compactChillerBase(prelimCount))
actCount = 0
DO iObj= 1, prelimCount
  CALL NextOldObj(ptCompactChiller,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. chlLeavingLowLimOff) THEN
    actCount = actCount + 1
    compactChillerBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + chlTypeOff, 'ElectricReciprocatingChiller')
    CALL SetIfBlank(fldValStart + chlCapacityOff, 'autosize')
    CALL SetIfBlank(fldValStart + chlCondTypeOff, 'WaterCooled')
    CALL SetIfBlank(fldValStart + chlSizeFactorOff, '1.0')
    CALL SetIfBlank(fldValStart + chlMinPLROff, '0.0')
    CALL SetIfBlank(fldValStart + chlMaxPLROff, '1.0')
    CALL SetIfBlank(fldValStart + chlOptPLROff, '1.0')
    CALL SetIfBlank(fldValStart + chlMinUnloadRatOff, '0.25')
    CALL SetIfBlank(fldValStart + chlLeavingLowLimOff, '5.0')
  END IF
END DO
numCompactChiller = actCount

CALL CountOldObj('HVACTemplate:Plant:Chiller:ObjectReference',ptCompactChillerOR,prelimCount)
ALLOCATE(compactChillerORBase(prelimCount))
actCount = 0
DO iObj= 1, prelimCount
  CALL NextOldObj(ptCompactChillerOR,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. chlorPriorityOff) THEN
    actCount = actCount + 1
    compactChillerORBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + chlorObjTypeOff, 'Chiller:Electric:EIR')
  END IF
END DO
numCompactChillerOR = actCount

! If there are HVACTemplate:Plant:Chiller:ObjectReference present, then get detailed chiller objects
IF (numCompactChillerOR > 0) THEN
  CALL CountOldObj('Chiller:Electric:EIR',ptDetChillerEIR,prelimCount)
  ALLOCATE(detChillerEIRBase(prelimCount))
  actCount = 0
  DO iObj= 1, prelimCount
    CALL NextOldObj(ptDetChillerEIR,fldValStart,fldValEnd)
    ! test to make sure object is valid
    IF (fldValEnd - fldValStart .EQ. chleirLastFieldOff) THEN
      actCount = actCount + 1
      detChillerEIRBase(actCount) = fldValStart
      !fill in defaults
    END IF
  END DO
  numDetChillerEIR = actCount

  CALL CountOldObj('Chiller:Electric:ReformulatedEIR',ptDetChillerREIR,prelimCount)
  ALLOCATE(detChillerREIRBase(prelimCount))
  actCount = 0
  DO iObj= 1, prelimCount
    CALL NextOldObj(ptDetChillerREIR,fldValStart,fldValEnd)
    ! test to make sure object is valid
    IF (fldValEnd - fldValStart .EQ. chlreirLastFieldOff) THEN
      actCount = actCount + 1
      detChillerREIRBase(actCount) = fldValStart
      !fill in defaults
    END IF
  END DO
  numDetChillerREIR = actCount
ENDIF


END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputTower
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the COMPACT HVAC:PLANT:TOWER object into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iObj

CALL CountOldObj('HVACTemplate:Plant:Tower',ptCompactTower,prelimCount)
ALLOCATE(compactTowerBase(prelimCount))
actCount = 0
DO iObj= 1, prelimCount
  CALL NextOldObj(ptCompactTower,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. twrLoopTypeOff) THEN
    actCount = actCount + 1
    compactTowerBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + twrTypeOff, 'SingleSpeed')
    CALL SetIfBlank(fldValStart + twrHiSpdCapacityOff, 'autosize')
    CALL SetIfBlank(fldValStart + twrHiSpdFanPowerOff, 'autosize')
    CALL SetIfBlank(fldValStart + twrLoSpdCapacityOff, 'autosize')
    CALL SetIfBlank(fldValStart + twrLoSpdFanPowerOff, 'autosize')
    CALL SetIfBlank(fldValStart + twrFreeConvCapacityOff, 'autosize')
    CALL SetIfBlank(fldValStart + twrSizeFactorOff, '1.0')
  END IF
END DO
numCompactTower = actCount

CALL CountOldObj('HVACTemplate:Plant:Tower:ObjectReference',ptCompactTowerOR,prelimCount)
ALLOCATE(compactTowerORBase(prelimCount))
actCount = 0
DO iObj= 1, prelimCount
  CALL NextOldObj(ptCompactTowerOR,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. twrorLoopTypeOff) THEN
    actCount = actCount + 1
    compactTowerORBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + chlorObjTypeOff, 'CoolingTower:SingleSpeed')
  END IF
END DO
numCompactTowerOR = actCount

! If there are HVACTemplate:Plant:Tower:ObjectReference present, then get detailed tower objects
IF (numCompactTowerOR > 0) THEN
  CALL CountOldObj('CoolingTower:SingleSpeed',ptDetTowerSS,prelimCount)
  ALLOCATE(detTowerSSBase(prelimCount))
  actCount = 0
  DO iObj= 1, prelimCount
    CALL NextOldObj(ptDetTowerSS,fldValStart,fldValEnd)
    ! test to make sure object is valid
    IF (fldValEnd - fldValStart .EQ. twrssLastFieldOff) THEN
      actCount = actCount + 1
      detTowerSSBase(actCount) = fldValStart
      !fill in defaults
    END IF
  END DO
  numDetTowerSS = actCount

  CALL CountOldObj('CoolingTower:TwoSpeed',ptDetTowerTS,prelimCount)
  ALLOCATE(detTowerTSBase(prelimCount))
  actCount = 0
  DO iObj= 1, prelimCount
    CALL NextOldObj(ptDetTowerTS,fldValStart,fldValEnd)
    ! test to make sure object is valid
    IF (fldValEnd - fldValStart .EQ. twrtsLastFieldOff) THEN
      actCount = actCount + 1
      detTowerTSBase(actCount) = fldValStart
      !fill in defaults
    END IF
  END DO
  numDetTowerTS = actCount

  CALL CountOldObj('CoolingTower:VariableSpeed',ptDetTowerVS,prelimCount)
  ALLOCATE(detTowerVSBase(prelimCount))
  actCount = 0
  DO iObj= 1, prelimCount
    CALL NextOldObj(ptDetTowerVS,fldValStart,fldValEnd)
    ! test to make sure object is valid
    IF (fldValEnd - fldValStart .EQ. twrvsLastFieldOff) THEN
      actCount = actCount + 1
      detTowerVSBase(actCount) = fldValStart
      !fill in defaults
    END IF
  END DO
  numDetTowerVS = actCount

ENDIF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputWaterAirHP
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2010
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the water to air heat pumps objects into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iZone

CALL CountOldObj('HVACTemplate:Zone:WaterToAirHeatPump',ptCompactWaterAirHP,prelimCount)
ALLOCATE(compactWaterAirHPBase(prelimCount))
actCount = 0
DO iZone= 1, prelimCount
  CALL NextOldObj(ptCompactWaterAirHP,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. wahpBaseboardCapOff) THEN
    actCount = actCount + 1
    compactWaterAirHPBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + wahpSupplyCoolFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + wahpSupplyHeatFlowRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + wahpOutAirTypeOff, 'Flow/Person')
    CALL SetIfBlank(fldValStart + wahpOutAirFlowPersonOff, '0.00944')
    CALL SetIfBlank(fldValStart + wahpOutAirFlowAreaOff, '0.0')
    CALL SetIfBlank(fldValStart + wahpOutAirFlowZoneOff, '0.0')
    CALL SetIfBlank(fldValStart + wahpFanPlacementOff, 'DrawThrough')
    CALL SetIfBlank(fldValStart + wahpFanTotEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + wahpFanPressureOff, '75')
    CALL SetIfBlank(fldValStart + wahpFanMotorEfficiencyOff, '0.9')
    CALL SetIfBlank(fldValStart + wahpCoolCoilKindOff, 'Coil:Cooling:WaterToAirHeatPump:EquationFit')
    CALL SetIfBlank(fldValStart + wahpCoolCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + wahpCoolCoilSHROff, 'autosize')
    CALL SetIfBlank(fldValStart + wahpCoolCoilCOPOff, '3.5')
    CALL SetIfBlank(fldValStart + wahpHPHeatCoilKindOff, 'Coil:Heating:WaterToAirHeatPump:EquationFit')
    CALL SetIfBlank(fldValStart + wahpHPHeatCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + wahpHPHeatCoilCOPOff, '4.2')
    CALL SetIfBlank(fldValStart + wahpSuppHeatCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + wahpMaxCycRateOff, '2.5')
    CALL SetIfBlank(fldValStart + wahpTimeConstOff, '60')
    CALL SetIfBlank(fldValStart + wahpFractOnPowerOff, '0.01')
    CALL SetIfBlank(fldValStart + wahpFanDelayOff, '60')
    CALL SetIfBlank(fldValStart + wahpSuppHeatCoilKindOff, 'Electric')
    CALL SetIfBlank(fldValStart + wahpCoolDesignTempMethodOff, 'SupplyAirTemperature')
    CALL SetIfBlank(fldValStart + wahpCoolDesignTempOff, '14.0')
    CALL SetIfBlank(fldValStart + wahpCoolDesignTempDiffOff,   '11.11')
    CALL SetIfBlank(fldValStart + wahpHeatDesignTempMethodOff, 'SupplyAirTemperature')
    CALL SetIfBlank(fldValStart + wahpHeatDesignTempOff, '50.0')
    CALL SetIfBlank(fldValStart + wahpCoolDesignTempDiffOff,   '30.0')
    CALL SetIfBlank(fldValStart + wahpHPCoilWaterFlowMode, 'Cycling')
    CALL SetIfBlank(fldValStart + wahpBaseboardTypeOff, 'None')
    CALL SetIfBlank(fldValStart + wahpBaseboardCapOff, 'autosize')
  END IF
END DO
numCompactWaterAirHP = actCount
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputMixedLoop
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2010
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the compact mixed water loop object into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iObj

CALL CountOldObj('HVACTemplate:Plant:MixedWaterLoop',ptCompactMixedLoop,prelimCount)
IF (prelimCount .GT. 1) THEN
  CALL WriteError('WARNING: Only one HVACTemplate:Plant:MixedWaterLoop allowed per file.')
ELSEIF (prelimCount .EQ. 1) THEN
  CALL NextOldObj(ptCompactMixedLoop,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. mwpLdDistSchmOff) THEN
    compactMixedLoopBase = fldValStart
    foundCompactObject = .TRUE.
    !check if loop exist but system doesn't
    IF ((numCompactWaterAirHP .EQ. 0) .AND. (numCompactSysVRF .EQ. 0) .AND. (numCompactSysUnitarySystem .EQ. 0)) THEN
      CALL WriteError('You must specify at least one HVACTemplate:Zone:WaterToAirHeatPump or '//  &
         'HVACTemplate:System:VRF or HVACTemplate:System:UnitarySystem if a HVACTemplate:Plant:MixedWaterLoop is defined.')
    END IF
    !fill in defaults
    CALL SetIfBlank(fldValStart + mwpPumpControlTypeOff, 'Intermittent')
    CALL SetIfBlank(fldValStart + mwpOperTypeOff, 'Default')
    CALL SetIfBlank(fldValStart + mwpHiSetPtDesignOff, '33')
    CALL SetIfBlank(fldValStart + mwpLoSetPtDesignOff, '20')
    CALL SetIfBlank(fldValStart + mwpPumpConfigOff, 'ConstantFlow')
    CALL SetIfBlank(fldValStart + mwpPumpHeadOff, '179352')
    CALL SetIfBlank(fldValStart + mwpPumpTypeOff, 'SinglePump')
    CALL SetIfBlank(fldValStart + mwpSupplyBypassOff, 'Yes')
    CALL SetIfBlank(fldValStart + mwpDemandBypassOff, 'Yes')
    CALL SetIfBlank(fldValStart + mwpFluidTypeOff, 'Water')
    CALL SetIfBlank(fldValStart + mwpDesDeltaTOff, '5.6')
    CALL SetIfBlank(fldValStart + mwpLdDistSchmOff, 'SequentialLoad')
  END IF
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputDOAS
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2010
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the dedicated outside air objects into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iSys

! SYSTEM
CALL CountOldObj('HVACTemplate:System:DedicatedOutdoorAir',ptCompactDedOutAir,prelimCount)
ALLOCATE(compactDedOutAirBase(prelimCount))
actCount = 0
DO iSys = 1, prelimCount
  CALL NextOldObj(ptCompactDedOutAir,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. doasHumidSetPtSchedNameOff) THEN
    actCount = actCount + 1
    compactDedOutAirBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
    !fill in defaults
    CALL SetIfBlank(fldValStart + doasAirOutletTypeOff, 'DirectIntoZone')
    CALL SetIfBlank(fldValStart + doasSupplyRateOff, 'autosize')
    CALL SetIfBlank(fldValStart + doasSupplyEfficiencyOff, '0.7')
    CALL SetIfBlank(fldValStart + doasSupplyPressureOff, '1000')
    CALL SetIfBlank(fldValStart + doasSupplyFanMotorEffOff, '0.9')
    CALL SetIfBlank(fldValStart + doasSupplyFanFractionOff, '1.0')
    CALL SetIfBlank(fldValStart + doasSupplyFanPlaceOff, 'DrawThrough')
    CALL SetIfBlank(fldValStart + doasCoolCoilKindOff, 'ChilledWater')
    CALL SetIfBlank(fldValStart + doasCoolSetPtCtrlTypeOff, 'FixedSetpoint')
    CALL SetIfBlank(fldValStart + doasCoolSetPtDesignOff, '12.8')
    ! OutdoorAirTempearatureReset defaults are 15.6C (60F) at 15.6C (60F) to 12.8C (55F) at 23.3C (74F)
    ! based on Carrier 62D 100% Outdoor Air Unit, cooling on at 60F, full on at 74F
    CALL SetIfBlank(fldValStart + doasCoolSetPtAtODBLoOff, '15.6')
    CALL SetIfBlank(fldValStart + doasCoolResetODBLoOff, '15.6')
    CALL SetIfBlank(fldValStart + doasCoolSetPtAtODBHiOff, '12.8')
    CALL SetIfBlank(fldValStart + doasCoolResetODBHiOff, '23.3')
    CALL SetIfBlank(fldValStart + doasCoolCoilCapOff, 'autosize')
    CALL SetIfBlank(fldValStart + doasCoolCoilSHROff, 'autosize')
    CALL SetIfBlank(fldValStart + doasCoolCoilCOPOff, '3.0')
    CALL SetIfBlank(fldValStart + doasHeatCoilKindOff, 'HotWater')
    CALL SetIfBlank(fldValStart + doasHeatSetPtCtrlTypeOff, 'FixedSetpoint')
    ! OutdoorAirTempearatureReset defaults are 15.0C (60F) at 7.8C (46F) to 12.2C (54F) at 12.2C (54F)
    ! based on Carrier 62D 100% Outdoor Air Unit, heating on at 55F, full on at 46F
    ! For EnergyPlus to avoid coil battling with defaults, use 54F and set default design setpoint to 12.2C (54F)
    CALL SetIfBlank(fldValStart + doasHeatSetPtDesignOff, '12.2')
    CALL SetIfBlank(fldValStart + doasHeatSetPtAtODBLoOff, '15.0')
    CALL SetIfBlank(fldValStart + doasHeatResetODBLoOff, '7.8')
    CALL SetIfBlank(fldValStart + doasHeatSetPtAtODBHiOff, '12.2')
    CALL SetIfBlank(fldValStart + doasHeatResetODBHiOff, '12.2')
    CALL SetIfBlank(fldValStart + doasHeatEffiencyOff, '0.8')
    CALL SetIfBlank(fldValStart + doasHeatParasiticOff, '0.0')
    CALL SetIfBlank(fldValStart + doasHeatRecTypeOff, 'None')
    CALL SetIfBlank(fldValStart + doasHeatRecSenEffOff, '0.70')
    CALL SetIfBlank(fldValStart + doasHeatRecLatEffOff, '0.65')
    CALL SetIfBlank(fldValStart + doasHXTypeOff, 'Plate')
    CALL SetIfBlank(fldValStart + doasFrostControlOff, 'None')
    CALL SetIfBlank(fldValStart + doasDehumCtrlTypeOff,'None')
    ! Default is equivalent to dewpoint of 12.8C (55F)
    CALL SetIfBlank(fldValStart + doasDehumSetPtOff,'0.00924')
    CALL SetIfBlank(fldValStart + doasHumidCtrlTypeOff,'None')
    CALL SetIfBlank(fldValStart + doasHumidCapOff,'0.000001')
    CALL SetIfBlank(fldValStart + doasHumidRatedElecOff,'autosize')
    CALL SetIfBlank(fldValStart + doasHumidSetPtOff,'0.003')
  END IF
END DO
numCompactDedOutAir = actCount
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetGroundHeatTransferControl
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2010
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read the compact chilled water loop object into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iObj

CALL CountOldObj('GroundHeatTransfer:Control',ptGhtControl,prelimCount)
IF (prelimCount .GT. 1) THEN
  CALL WriteError('WARNING: Only one GroundHeatTransfer:Control allowed per file.')
ELSEIF (prelimCount .EQ. 1) THEN
  CALL NextOldObj(ptGhtControl,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. ghtCtrlSlabOff) THEN
    ghtControlBase = fldValStart
    !fill in defaults
    CALL SetIfBlank(fldValStart + ghtCtrlBasementOff, 'No')
    CALL SetIfBlank(fldValStart + ghtCtrlSlabOff, 'No')
  END IF
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetInputBuildingSurfaces
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2010
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read all of the building surface objects into the arrays

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iSurf

!don't worry about defaults since only reading one field and
!EnergyPlus will enforce defaults with rest of pasted through fields
!Also, no need to set the flag for foundCompactObject = .TRUE. since
!the presence of these objects alone should not trigger expansion.

!BuildingSurface:Detailed
CALL CountOldObj('BuildingSurface:Detailed',ptSurfBldgDetl,prelimCount)
ALLOCATE(surfBldgDetlBase(prelimCount))
actCount = 0
DO iSurf = 1, prelimCount
  CALL NextOldObj(ptSurfBldgDetl,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. buildSurfLastZVertexOff) THEN
    actCount = actCount + 1
    surfBldgDetlBase(actCount) = fldValStart
  END IF
END DO
numSurfBldgDetl = actCount
!Wall:Detailed
CALL CountOldObj('Wall:Detailed',ptSurfWallDetl,prelimCount)
ALLOCATE(surfWallDetlBase(prelimCount))
actCount = 0
DO iSurf = 1, prelimCount
  CALL NextOldObj(ptSurfWallDetl,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. wallDetlLastZVertexOff) THEN
    actCount = actCount + 1
    surfWallDetlBase(actCount) = fldValStart
  END IF
END DO
numSurfWallDetl = actCount
!RoofCeiling:Detailed
CALL CountOldObj('RoofCeiling:Detailed',ptSurfRfClDetl,prelimCount)
ALLOCATE(surfRfClDetlBase(prelimCount))
actCount = 0
DO iSurf = 1, prelimCount
  CALL NextOldObj(ptSurfRfClDetl,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. rfClDetlLastZVertexOff) THEN
    actCount = actCount + 1
    surfRfClDetlBase(actCount) = fldValStart
  END IF
END DO
numSurfRfClDetl = actCount
!Floor:Detailed
CALL CountOldObj('Floor:Detailed',ptSurfFlrDetl,prelimCount)
ALLOCATE(surfFlrDetlBase(prelimCount))
actCount = 0
DO iSurf = 1, prelimCount
  CALL NextOldObj(ptSurfFlrDetl,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. flrDetlLastZVertexOff) THEN
    actCount = actCount + 1
    surfFlrDetlBase(actCount) = fldValStart
  END IF
END DO
numSurfFlrDetl = actCount
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetAllGroundHeatTransfer
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2010
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read all of the GroundHeatTransfer:* objects into the arrays

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: i

!GroundHeatTransfer:Slab:Materials
CALL CountOldObj('GroundHeatTransfer:Slab:Materials',ptGhtSlbMaterial,prelimCount)
ALLOCATE(ghtSlbMaterialBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtSlbMaterial,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtSlbMaterialLastOff) THEN
    actCount = actCount + 1
    ghtSlbMaterialBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtSlbMaterial = actCount
!GroundHeatTransfer:Slab:MatlProps
CALL CountOldObj('GroundHeatTransfer:Slab:MatlProps',ptGhtSlbMatlProps,prelimCount)
ALLOCATE(ghtSlbMatlPropsBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtSlbMatlProps,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtSlbMatlPropsLastOff) THEN
    actCount = actCount + 1
    ghtSlbMatlPropsBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtSlbMatlProps = actCount
!GroundHeatTransfer:Slab:BoundConds
CALL CountOldObj('GroundHeatTransfer:Slab:BoundConds',ptGhtSlbBoundConds,prelimCount)
ALLOCATE(ghtSlbBoundCondsBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtSlbBoundConds,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtSlbBoundCondsLastOff) THEN
    actCount = actCount + 1
    ghtSlbBoundCondsBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtSlbBoundConds = actCount
!GroundHeatTransfer:Slab:BldgProps
CALL CountOldObj('GroundHeatTransfer:Slab:BldgProps',ptGhtSlbBldgProps,prelimCount)
ALLOCATE(ghtSlbBldgPropsBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtSlbBldgProps,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtSlbBldgPropsLastOff) THEN
    actCount = actCount + 1
    ghtSlbBldgPropsBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtSlbBldgProps = actCount
!GroundHeatTransfer:Slab:Insulation
CALL CountOldObj('GroundHeatTransfer:Slab:Insulation',ptGhtSlbInsul,prelimCount)
ALLOCATE(ghtSlbInsulBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtSlbInsul,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtSlbInsulLastOff) THEN
    actCount = actCount + 1
    ghtSlbInsulBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtSlbInsul = actCount
!GroundHeatTransfer:Slab:EquivalentSlab
CALL CountOldObj('GroundHeatTransfer:Slab:EquivalentSlab',ptGhtSlbEqvSlb,prelimCount)
ALLOCATE(ghtSlbEqvSlbBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtSlbEqvSlb,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtSlbEqvSlbLastOff) THEN
    actCount = actCount + 1
    ghtSlbEqvSlbBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtSlbEqvSlb = actCount
!GroundHeatTransfer:Slab:AutoGrid
CALL CountOldObj('GroundHeatTransfer:Slab:AutoGrid',ptGhtSlbAutoGrid,prelimCount)
ALLOCATE(ghtSlbAutoGridBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtSlbAutoGrid,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtSlbAutoGridLastOff) THEN
    actCount = actCount + 1
    ghtSlbAutoGridBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtSlbAutoGrid = actCount
!GroundHeatTransfer:Slab:ManualGrid
CALL CountOldObj('GroundHeatTransfer:Slab:ManualGrid',ptGhtSlbManGrid,prelimCount)
ALLOCATE(ghtSlbManGridBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtSlbManGrid,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtSlbManGridLastOff) THEN
    actCount = actCount + 1
    ghtSlbManGridBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtSlbManGrid = actCount
!GroundHeatTransfer:Slab:XFACE
CALL CountOldObj('GroundHeatTransfer:Slab:XFACE',ptGhtSlbXFace,prelimCount)
ALLOCATE(ghtSlbXFaceBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtSlbXFace,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtSlbXFaceLastOff) THEN
    actCount = actCount + 1
    ghtSlbXFaceBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtSlbXFace = actCount
!GroundHeatTransfer:Slab:YFACE
CALL CountOldObj('GroundHeatTransfer:Slab:YFACE',ptGhtSlbYFace,prelimCount)
ALLOCATE(ghtSlbYFaceBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtSlbYFace,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtSlbYFaceLastOff) THEN
    actCount = actCount + 1
    ghtSlbYFaceBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtSlbYFace = actCount
!GroundHeatTransfer:Slab:ZFACE
CALL CountOldObj('GroundHeatTransfer:Slab:ZFACE',ptGhtSlbZFace,prelimCount)
ALLOCATE(ghtSlbZFaceBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtSlbZFace,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtSlbZFaceLastOff) THEN
    actCount = actCount + 1
    ghtSlbZFaceBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtSlbZFace = actCount
!GroundHeatTransfer:Basement:SimParameters
CALL CountOldObj('GroundHeatTransfer:Basement:SimParameters',ptGhtBsSimParam,prelimCount)
ALLOCATE(ghtBsSimParamBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtBsSimParam,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtBsSimParamLastOff) THEN
    actCount = actCount + 1
    ghtBsSimParamBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtBsSimParam = actCount
!GroundHeatTransfer:Basement:MatlProps
CALL CountOldObj('GroundHeatTransfer:Basement:MatlProps',ptGhtBsMatlProps,prelimCount)
ALLOCATE(ghtBaseBsMatlProps(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtBsMatlProps,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtBsMatlPropsLastOff) THEN
    actCount = actCount + 1
    ghtBaseBsMatlProps(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtBsMatlProps = actCount
!GroundHeatTransfer:Basement:Insulation
CALL CountOldObj('GroundHeatTransfer:Basement:Insulation',ptGhtBsInsul,prelimCount)
ALLOCATE(ghtBsInsulBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtBsInsul,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtBsInsulLastOff) THEN
    actCount = actCount + 1
    ghtBsInsulBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtBsInsul = actCount
!GroundHeatTransfer:Basement:SurfaceProps
CALL CountOldObj('GroundHeatTransfer:Basement:SurfaceProps',ptGhtBsSurfProps,prelimCount)
ALLOCATE(ghtBsSurfPropsBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtBsSurfProps,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtBsSurfPropsLastOff) THEN
    actCount = actCount + 1
    ghtBsSurfPropsBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtBsSurfProps = actCount
!GroundHeatTransfer:Basement:BldgData
CALL CountOldObj('GroundHeatTransfer:Basement:BldgData',ptGhtBsBldgData,prelimCount)
ALLOCATE(ghtBsBldgDataBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtBsBldgData,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtBsBldgDataLastOff) THEN
    actCount = actCount + 1
    ghtBsBldgDataBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtBsBldgData = actCount
!GroundHeatTransfer:Basement:Interior
CALL CountOldObj('GroundHeatTransfer:Basement:Interior',ptGhtBsInterior,prelimCount)
ALLOCATE(ghtBsInteriorBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtBsInterior,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtBsInteriorLastOff) THEN
    actCount = actCount + 1
    ghtBsInteriorBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtBsInterior = actCount
!GroundHeatTransfer:Basement:ComBldg
CALL CountOldObj('GroundHeatTransfer:Basement:ComBldg',ptGhtBsComBldg,prelimCount)
ALLOCATE(ghtBsComBldgBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtBsComBldg,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtBsComBldgLastOff) THEN
    actCount = actCount + 1
    ghtBsComBldgBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtBsComBldg = actCount
!GroundHeatTransfer:Basement:EquivSlab
CALL CountOldObj('GroundHeatTransfer:Basement:EquivSlab',ptGhtBsEqvSlb,prelimCount)
ALLOCATE(ghtBsEqvSlbBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtBsEqvSlb,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtBsEqvSlbLastOff) THEN
    actCount = actCount + 1
    ghtBsEqvSlbBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtBsEqvSlb = actCount
!GroundHeatTransfer:Basement:EquivAutoGrid
CALL CountOldObj('GroundHeatTransfer:Basement:EquivAutoGrid',ptGhtBsEqvAutoGrid,prelimCount)
ALLOCATE(ghtBsEqvAutoGridBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtBsEqvAutoGrid,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtBsEqvAutoGridLastOff) THEN
    actCount = actCount + 1
    ghtBsEqvAutoGridBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtBsEqvAutoGrid = actCount
!GroundHeatTransfer:Basement:AutoGrid
CALL CountOldObj('GroundHeatTransfer:Basement:AutoGrid',ptGhtBsAutoGrid,prelimCount)
ALLOCATE(ghtBsAutoGridBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtBsAutoGrid,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtBsAutoGridLastOff) THEN
    actCount = actCount + 1
    ghtBsAutoGridBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtBsAutoGrid = actCount
!GroundHeatTransfer:Basement:ManualGrid
CALL CountOldObj('GroundHeatTransfer:Basement:ManualGrid',ptGhtBsManGrid,prelimCount)
ALLOCATE(ghtBsManGridBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtBsManGrid,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtBsManGridLastOff) THEN
    actCount = actCount + 1
    ghtBsManGridBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtBsManGrid = actCount
!GroundHeatTransfer:Basement:XFACE
CALL CountOldObj('GroundHeatTransfer:Basement:XFACE',ptGhtBsXFace,prelimCount)
ALLOCATE(ghtBsXFaceBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtBsXFace,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtBsXFaceLastOff) THEN
    actCount = actCount + 1
    ghtBsXFaceBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtBsXFace = actCount
!GroundHeatTransfer:Basement:YFACE
CALL CountOldObj('GroundHeatTransfer:Basement:YFACE',ptGhtBsYFace,prelimCount)
ALLOCATE(ghtBsYFaceBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtBsYFace,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtBsYFaceLastOff) THEN
    actCount = actCount + 1
    ghtBsYFaceBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtBsYFace = actCount
!GroundHeatTransfer:Basement:ZFACE
CALL CountOldObj('GroundHeatTransfer:Basement:ZFACE',ptGhtBsZFace,prelimCount)
ALLOCATE(ghtBsZFaceBase(prelimCount))
actCount = 0
DO i = 1, prelimCount
  CALL NextOldObj(ptGhtBsZFace,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (FldValEnd - FldValStart .EQ. ghtBsZFaceLastOff) THEN
    actCount = actCount + 1
    ghtBsZFaceBase(actCount) = fldValStart
    foundCompactObject = .TRUE.
  END IF
END DO
numGhtBsZFace = actCount
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE GetGroundTempSurf
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2010
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Read Site:GroundTemperature:BuildingSurface object into the array

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER    :: prelimCount !based on number objects found but not verified
INTEGER    :: actCount
INTEGER    :: fldValStart
INTEGER    :: fldValEnd
INTEGER    :: iObj

CALL CountOldObj('Site:GroundTemperature:BuildingSurface',ptGndTempBuildSurf,prelimCount)
IF (prelimCount .GT. 1) THEN
  CALL WriteError('WARNING: Only one Site:GroundTemperature:BuildingSurface allowed per file.')
ELSEIF (prelimCount .EQ. 1) THEN
  CALL NextOldObj(ptGndTempBuildSurf,fldValStart,fldValEnd)
  ! test to make sure object is valid
  IF (fldValEnd - fldValStart .EQ. gndTempSurfDecOff) THEN
    gndTempBuildSurfBase = fldValStart
  END IF
END IF
END SUBROUTINE



!----------------------------------------------------------------------------------
SUBROUTINE CheckIfNamesAreUnique
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2012
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Check all objects that have names to see if the names are unique.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS
INTEGER :: iObjToProc
INTEGER :: jHeldObj
INTEGER :: repeatFound

! first check all objects that are HVACTemplates to see if any names are repeated
DO iObjToProc = 1, numObjToProcess
  !only test objects that are HVACTemplate objects, skip the rest
  IF (SameString(objToProcess(iObjToProc)%ucName(1:12),'HVACTEMPLATE')) THEN
    !use the IntList to hold pointers to the fields containing the names.
    CALL ClearIntList
    !find all the held objects that match the current object to process
    DO jHeldObj = 1,numHeldObj
      IF (HeldObj(jHeldObj)%ObjToProcessIndex .EQ. iObjToProc) THEN
        !store the first field value plus one for each matching held object since
        !the first field is always the name, this will work.
        CALL AddToIntList(HeldObj(jHeldObj)%FirstFldVal + 1)
        !just to test write an error message
        !CALL WriteError('TEST names: '//TRIM(FldVal(HeldObj(jHeldObj)%FirstFldVal + 1)))
      END IF
    END DO
    repeatFound = RepeatStringInIntList()
    IF (repeatFound .NE. 0) THEN
      CALL WriteError('Repeated name found: '//TRIM(FldVal(repeatFound))//' found as the name of more than one '&
             //TRIM(objToProcess(iObjToProc)%ccName)//' object.' ,msgWarning)
    ENDIF
  END IF
END DO
! now check all the zone objects to see if any names are repeated
CALL ClearIntList
DO iObjToProc = 1, numObjToProcess
  !use the IntList to hold pointers to the fields containing the names.
  !only test objects that are HVACTemplate objects, skip the rest
  IF (SameString(objToProcess(iObjToProc)%ucName(1:17),'HVACTEMPLATE:ZONE')) THEN
    !find all the held objects that match the current object to process
    DO jHeldObj = 1,numHeldObj
      IF (HeldObj(jHeldObj)%ObjToProcessIndex .EQ. iObjToProc) THEN
        !store the first field value plus one for each matching held object since
        !the first field is always the name, this will work.
        CALL AddToIntList(HeldObj(jHeldObj)%FirstFldVal + 1)
      END IF
    END DO
  END IF
END DO
repeatFound = RepeatStringInIntList()
IF (repeatFound .NE. 0) THEN
  CALL WriteError('The same name is used for multiple HVACTemplate:Zone:* objects: '//TRIM(FldVal(repeatFound)) &
  // '. Names must be unique across all HVACTemplate:Zone:* objects.' ,msgWarning)
END IF
! now check all the system objects to see if any names are repeated
CALL ClearIntList
DO iObjToProc = 1, numObjToProcess
  !use the IntList to hold pointers to the fields containing the names.
  !only test objects that are HVACTemplate objects, skip the rest
  IF (SameString(objToProcess(iObjToProc)%ucName(1:19),'HVACTEMPLATE:SYSTEM')) THEN
    !find all the held objects that match the current object to process
    DO jHeldObj = 1,numHeldObj
      IF (HeldObj(jHeldObj)%ObjToProcessIndex .EQ. iObjToProc) THEN
        !store the first field value plus one for each matching held object since
        !the first field is always the name, this will work.
        CALL AddToIntList(HeldObj(jHeldObj)%FirstFldVal + 1)
      END IF
    END DO
  END IF
END DO
repeatFound = RepeatStringInIntList()
IF (repeatFound .NE. 0) THEN
  CALL WriteError('Repeated HVACTemplate:System:* name found: '//TRIM(FldVal(repeatFound)),msgWarning)
END IF
! now check all the loop objects to see if any names are repeated
CALL ClearIntList
DO iObjToProc = 1, numObjToProcess
  !use the IntList to hold pointers to the fields containing the names.
  !only test objects that are HVACTemplate objects, skip the rest
  IF ( SameString(objToProcess(iObjToProc)%ucName,'HVACTemplate:Plant:HotWaterLoop') .OR. &
       SameString(objToProcess(iObjToProc)%ucName,'HVACTemplate:Plant:ChilledWaterLoop') .OR. &
       SameString(objToProcess(iObjToProc)%ucName,'HVACTemplate:Plant:MixedWaterLoop') ) THEN
    !find all the held objects that match the current object to process
    DO jHeldObj = 1,numHeldObj
      IF (HeldObj(jHeldObj)%ObjToProcessIndex .EQ. iObjToProc) THEN
        !store the first field value plus one for each matching held object since
        !the first field is always the name, this will work.
        CALL AddToIntList(HeldObj(jHeldObj)%FirstFldVal + 1)
      END IF
    END DO
  END IF
END DO
repeatFound = RepeatStringInIntList()
IF (repeatFound .NE. 0) THEN
  CALL WriteError('Repeated HVACTemplate:Plant:*WaterLoop name found: '//TRIM(FldVal(repeatFound)),msgWarning)
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
!----------------------------------------------------------------------------------
!
!   CREATE SPECIFIC OBJECTS ROUTINES
!
!----------------------------------------------------------------------------------
!----------------------------------------------------------------------------------

!----------------------------------------------------------------------------------
SUBROUTINE CreateNewThermostat
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a compact thermostat

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for
INTEGER :: iStat
LOGICAL :: isHeatSchedName
LOGICAL :: isCoolSchedName

DO iStat = 1, numCompactStat
  base =  compactStatBase(iStat)
  ! first check if the schedule names are used or if values should be used.
  IF (FldVal(base + tHeatSchedNameOff) .EQ. '') THEN
    isHeatSchedName = .FALSE.
    CALL AddAlwaysSchedule(FldVal(base + tHeatSetPointOff))
  ELSE
    isHeatSchedName = .TRUE.
  END IF
  IF (FldVal(base + tCoolSchedNameOff) .EQ. '') THEN
    isCoolSchedName = .FALSE.
    CALL AddAlwaysSchedule(FldVal(base + tCoolSetPointOff))
  ELSE
    isCoolSchedName = .TRUE.
  END IF
  ! write object
  CALL CreateNewObj('ThermostatSetpoint:DualSetpoint')
  CALL AddToObjFld('Name', base + tNameOff,' Dual SP Control')
  !If  vzHeatSchedNameOff is blank, then create schedule of vzHeatSetPointOff
  IF (isHeatSchedName) THEN
    CALL AddToObjFld('Heating Setpoint Temperature Schedule Name', base +  tHeatSchedNameOff,' ')
  ELSE
    CALL AddToObjStr('Heating Setpoint Temperature Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + tHeatSetPointOff)))
  END IF
  IF (isCoolSchedName) THEN
    CALL AddToObjFld('Cooling Setpoint Temperature Schedule Name', base +  tCoolSchedNameOff,' ',.TRUE.)
  ELSE
    CALL AddToObjStr('Cooling Setpoint Temperature Schedule Name', 'HVACTemplate-Always ' //   &
       TRIM(FldVal(base + tCoolSetPointOff)),.TRUE.)
  END IF
END DO
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE CreateNewVAV
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2005
          !    MODIFIED       December 2008 to add Packaged VAV
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a VAV system.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for coil type (heating)
INTEGER,PARAMETER :: ctHotWater = 1
INTEGER,PARAMETER :: ctElectric = 2
INTEGER,PARAMETER :: ctGas      = 3
INTEGER,PARAMETER :: ctNone     = 4
          ! parameters for coil type (cooling)
INTEGER, PARAMETER :: ccChWater = 1
INTEGER, PARAMETER :: ccChWaterDetailed = 2
INTEGER, PARAMETER :: ccTwoSpeedDX = 3
INTEGER, PARAMETER :: ccTwoSpeedHumidControlDX = 4
!INTEGER, PARAMETER :: ccThreeSpeedDX = 5
!INTEGER, PARAMETER :: ccFourSpeedDX = 6
          ! parameters for economizer type
INTEGER,PARAMETER :: etFixedDryBulb                   = 1
INTEGER,PARAMETER :: etFixedEnthalpy                  = 2
INTEGER,PARAMETER :: etDifferentialDryBulb            = 3
INTEGER,PARAMETER :: etDifferentialEnthalpy           = 4
INTEGER,PARAMETER :: etFixedDewPointAndDryBulb        = 5
INTEGER,PARAMETER :: etElectronicEnthalpy             = 6
INTEGER,PARAMETER :: etDifferentialDryBulbAndEnthalpy = 7
INTEGER,PARAMETER :: etNoEconomizer                   = 8
          ! parameters for lockout type
INTEGER, PARAMETER :: loNo       = 1
INTEGER, PARAMETER :: loHeat     = 2
INTEGER, PARAMETER :: loComp     = 3
          ! parameters for cool setpoint reset type
INTEGER,PARAMETER :: csprNone    = 1
INTEGER,PARAMETER :: csprWarmest = 2
INTEGER,PARAMETER :: csprOAreset = 3
INTEGER,PARAMETER :: csprWarmestTempFirst = 4
          ! parameters for heat setpoint reset type
INTEGER,PARAMETER :: hsprNone    = 1
!INTEGER,PARAMETER :: hsprColdest = 2
INTEGER,PARAMETER :: hsprOAreset = 3
!INTEGER,PARAMETER :: hsprColdestTempFirst = 4
          ! parameters for supply fan coef type
INTEGER,PARAMETER :: sfcInletVane = 1
INTEGER,PARAMETER :: sfcDischgDamp = 2
INTEGER,PARAMETER :: sfcVSD = 3
INTEGER,PARAMETER :: sfcAppG = 4
INTEGER,PARAMETER :: sfcVSDPresReset = 5
          ! parameters for heat recovery type
INTEGER,PARAMETER :: htrecNone = 1
INTEGER,PARAMETER :: htrecSens = 2
INTEGER,PARAMETER :: htrecEnth = 3
          ! parameters for supply fan type
INTEGER,PARAMETER :: sfpBlowThru = 1
INTEGER,PARAMETER :: sfpDrawThru = 2
          ! parameters for dehumidification type
INTEGER, PARAMETER :: dehumidNone    = 1
INTEGER, PARAMETER :: dehumidCoolRht = 2
          ! parameters for humidification type
INTEGER, PARAMETER :: humidifyNone      = 1
INTEGER, PARAMETER :: humidifyElecSteam = 2
          ! parameters for baseboard type
INTEGER, PARAMETER :: baseboardNone           = 1
INTEGER, PARAMETER :: baseboardHotWater       = 2
INTEGER, PARAMETER :: baseboardElectric       = 3

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for
INTEGER :: iZone
INTEGER :: iSys
INTEGER :: i
LOGICAL :: isTStatNameBlank
LOGICAL :: isSupplyMaxRateAutosize
LOGICAL :: isDamperActionNormal
REAL    :: heatDesAirFlowRate
INTEGER :: reheatCoilType=0
INTEGER :: heatCoilType=0
INTEGER :: preHeatCoilType=0
INTEGER :: coolCoilKind=0
LOGICAL :: isSupplyMinRateAutosize
LOGICAL :: isSupplyPlenumBlank
LOGICAL :: isReturnPlenumBlank
REAL    :: minAirRatio
LOGICAL :: isZoneFanSchedBlank
INTEGER :: znBase
INTEGER :: sysBase
LOGICAL :: isSysAvailSchedBlank
LOGICAL :: isCoolStPtSchedBlank
LOGICAL :: isHeatStPtSchedBlank
LOGICAL :: isPreheatStPtSchedBlank
INTEGER :: econoKind=0
INTEGER :: lockoutKind=0
LOGICAL :: isZoneRetPlenBlank
LOGICAL :: isZoneSupPlenBlank
INTEGER :: coolSetPtReset=0
INTEGER :: heatSetPtReset=0
INTEGER :: supFanCoef=0
INTEGER :: retFanCoef=0
INTEGER :: heatRecovery=0
LOGICAL :: isEconoLowLimitBlank
LOGICAL :: isPriSupplyAuto
LOGICAL :: isSecSupplyAuto
REAL    :: heatDesAirRate
REAL    :: totalFlow
LOGICAL :: isFlowTypeSeries = .FALSE.
LOGICAL :: isFlowTypeFromZonePlenum = .FALSE.
LOGICAL :: isFlowTypeFromZone = .FALSE.
INTEGER :: iInt
INTEGER :: supFanPlacement=0
LOGICAL :: isMinOARateAutosize
REAL    :: htRecSens75
REAL    :: htRecLat75
LOGICAL :: isDehumidifyNone
INTEGER :: dehumidCtrlKind = 0
LOGICAL :: isHumidifierNone
INTEGER :: humidifierKind = 0
INTEGER :: baseboardKind = 0
LOGICAL :: isPackaged = .FALSE.
LOGICAL :: isCoolCapAutosize = .FALSE.
LOGICAL :: isHeatCoilCapAutosize = .FALSE.
LOGICAL :: isSHRAutosize = .FALSE.
LOGICAL :: doSuppressMsgzlVavAll = .FALSE.
LOGICAL :: usePreheatTempforHeatDesignSuppT = .FALSE.
LOGICAL :: isOATypeDetailed = .FALSE.
LOGICAL :: isReturnFanYes = .FALSE.

! Node Names
CHARACTER(len=MaxAlphaLength) :: returnInletToOAMIxer=''

!======================================================================================================
! create Zones
!======================================================================================================
DO iZone = 1, numCompactZoneVAV
  base =  compactZoneVAVbase(iZone)
  ! find base for system this zone belongs to
  sysBase = -1
  DO iSys = 1, numCompactSysVAV
    IF (SameString(FldVal(compactSysVAVbase(iSys) + vsAirHandlerNameOff), FldVal(base +  vzAirHandlerNameOff))) THEN
      sysBase = compactSysVAVBase(iSys)
      EXIT
    END IF
  END DO
  IF (sysBase .NE. -1) THEN
    isPackaged = .FALSE.
  ELSE
    DO iSys = 1, numCompactSysPVAV
      IF (SameString(FldVal(compactSysPVAVbase(iSys) + pvavsAirHandlerNameOff), FldVal(base +  vzAirHandlerNameOff))) THEN
        sysBase = compactSysPVAVBase(iSys)
        EXIT
      END IF
    END DO
    IF (sysBase .NE. -1) isPackaged = .TRUE.
  END IF
  IF (sysBase .EQ. -1) THEN
    CALL WriteError('Could not find air handler name referenced in HVACTemplate:Zone:VAV object: '// TRIM(FldVal(base + vzNameOff)))
    CYCLE
  END IF
  ! set flags
  isTStatNameBlank = (FldVal(base + vzTStatNameOff) .EQ. '')
  isSupplyMaxRateAutosize = SameString(FldVal(base + vzSupplyMaxRateOff),'autosize')
  isDamperActionNormal = SameString(FldVal(base + vzDamperActionOff),'Normal')
  isOATypeDetailed = SameString(FldVal(base + vzOutAirTypeOff),'DetailedSpecification')
  CALL validChoice(base + vzDamperActionOff,'Normal','Reverse','ZONE:VAV','Heating Damper Action')
  !set reheat coil indicator
  IF (SameString(FldVal(base + vzReheatTypeOff),'HotWater')) THEN
    reheatCoilType = ctHotWater
  ELSEIF (SameString(FldVal(base + vzReheatTypeOff),'Electric')) THEN
    reheatCoilType = ctElectric
  ELSEIF (SameString(FldVal(base + vzReheatTypeOff),'Gas')) THEN
    reheatCoilType = ctGas
  ELSEIF (SameString(FldVal(base + vzReheatTypeOff),'None')) THEN
    reheatCoilType = ctNone
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VAV in the Reheat Coil Type field')
  END IF
  isZoneRetPlenBlank = (FldVal(base + vzZoneReturnPlenumNameOff) .EQ. '')
  isZoneSupPlenBlank = (FldVal(base + vzZoneSupplyPlenumNameOff) .EQ. '')
  !set the baseboard indicator
  IF (SameString(FldVal(base +  vzBaseboardTypeOff),'None')) THEN
    baseboardKind = baseboardNone
  ELSEIF (SameString(FldVal(base +  vzBaseboardTypeOff),'HotWater')) THEN
    baseboardKind = baseboardHotWater
  ELSEIF (SameString(FldVal(base +  vzBaseboardTypeOff),'Electric')) THEN
    baseboardKind = baseboardElectric
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VAV "'//TRIM(FldVal(base + vzNameOff))//'"'// &
                    ' in the Baseboard Heating Type field: '//TRIM(FldVal(base + vzBaseboardTypeOff)))
  END IF
  !validate cooling design method
  IF (SameString(FldVal(base +  vzCoolDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  vzCoolDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  vzCoolDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VAV "'//TRIM(FldVal(base + vzNameOff))//'"'// &
                    ' in the Zone Cooling Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + vzCoolDesignTempMethodOff)))
  END IF
  !validate heating design method
  IF (SameString(FldVal(base +  vzHeatDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  vzHeatDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VAV "'//TRIM(FldVal(base + vzNameOff))//'"'// &
                    ' in the Zone Heating Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + vzCoolDesignTempMethodOff)))
  END IF
  !ZONE CONTROL:THERMOSTATIC ~ line 5
  !repeat per zone - only if vzTStatNameOff is not blank
  IF (.NOT. isTStatNameBlank) THEN
    CALL AddAlwaysSchedule('4')
    CALL CreateNewObj('ZoneControl:Thermostat')
    CALL AddToObjFld('Name', base +  vzNameOff,' Thermostat')
    CALL AddToObjFld('Zone or ZoneList Name', base +  vzNameOff,' ')
    !this schedule was previously created
    CALL AddToObjStr('Control Type Schedule Name','HVACTemplate-Always 4')
    CALL AddToObjStr('Control Object Type','ThermostatSetpoint:DualSetpoint')
    CALL AddToObjFld('Control Name', base + vzTStatNameOff,' Dual SP Control',.TRUE.)
  END IF
  !ZONE SIZING   ~ line 12
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:OutdoorAir')
    CALL AddToObjStr('Name', 'SZ DSOA '//FldVal(base+vzNameOff))
    CALL AddToObjFld('Outdoor Air Method', base + vzOutAirTypeOff,' ')
    CALL AddToObjFld('Outdoor Air Flow per Person {m3/s}', base + vzOutAirFlowPersonOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone Floor Area {m3/s-m2}', base + vzOutAirFlowAreaOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone {m3/s}', base + vzOutAirFlowZoneOff,'',.true.)
  END IF

  CALL CreateNewObj('Sizing:Zone')
  CALL AddToObjFld('Zone or ZoneList Name', base +  vzNameOff,'')
  IF (SameString(FldVal(base +  vzCoolDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
    CALL AddToObjStr('Zone Cooling Design Supply Air Temperature Input Method','SupplyAirTemperature')
    IF (isPackaged) THEN
      CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', sysBase + pvavsCoolSetPtDesignOff,'')
    ELSE
      CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', sysBase + vsCoolSetPtDesignOff,'')
    ENDIF
  ELSE
    CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Input Method',base + vzCoolDesignTempMethodOff,'')
    CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', base + vzCoolDesignTempOff,'')
  ENDIF
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Difference {delta C}',base + vzCoolDesignTempDiffOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Input Method',base + vzHeatDesignTempMethodOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', base + vzHeatDesignTempOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Difference {delta C}',base + vzHeatDesignTempDiffOff,'')
  CALL AddToObjStr('Zone Cooling Design Supply Air Humidity Ratio {kg-H20/kg-air}','0.008')
  CALL AddToObjStr('Zone Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + vzDesSpecOANameSizOff,'')
  ELSE
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','SZ DSOA '//FldVal(base+vzNameOff))
  END IF
  CALL AddToObjFld('Zone Heating Sizing Factor', base + vzZoneHeatSizingFactorOff,'')
  CALL AddToObjFld('Zone Cooling Sizing Factor', base + vzZoneCoolSizingFactorOff,'')
  IF (isSupplyMaxRateAutosize) THEN
    CALL AddToObjStr('Cooling Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Cooling Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  ELSE
    CALL AddToObjStr('Cooling Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Cooling Design Air Flow Rate {m3/s}', base + vzSupplyMaxRateOff,' ')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  END IF
  IF (isSupplyMaxRateAutosize) THEN
    CALL AddToObjStr('Heating Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Heating Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
!    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+vzNameOff),.true.)
  ELSE
    IF (isDamperActionNormal) THEN
      IF (SameString(FldVal(base + vzMinFlowMethodOff),'Constant')) THEN
        heatDesAirFlowRate = StringToReal(FldVal(base + vzSupplyMaxRateOff)) * &
                             StringToReal(FldVal(base + vzConstantMinFracOff))
        CALL AddToObjStr('Heating Design Air Flow Method','Flow/Zone')
        CALL AddToObjNum('Heating Design Air Flow Rate {m3/s}',heatDesAirFlowRate)
      ELSE IF (SameString(FldVal(base + vzMinFlowMethodOff),'FixedFlowRate')) THEN
        heatDesAirFlowRate = StringToReal(FldVal(base + vzFixedMinFlowOff))
        CALL AddToObjStr('Heating Design Air Flow Method','Flow/Zone')
        CALL AddToObjNum('Heating Design Air Flow Rate {m3/s}',heatDesAirFlowRate)
      ELSE ! Min flow is 'Scheduled'))
        ! If min flow is scheduled, can't know what the min flow frac will be, so autosize heating flow
        CALL AddToObjStr('Heating Design Air Flow Method','DesignDay')
        CALL AddToObjStr('Heating Design Air Flow Rate {m3/s}','0')
      END IF
      CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
      CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
      CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
 !     CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+vzNameOff),.true.)
    ELSE
      CALL AddToObjStr('Heating Design Air Flow Method','Flow/Zone')
      CALL AddToObjFld('Heating Design Air Flow Rate {m3/s}', base + vzSupplyMaxRateOff,'')
      CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
      CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
      CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
    END IF
  END IF
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Zone Air Distribution Object Name', base + vzDesSpecOADistNameOff,'',.true.)
  ELSE
    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+vzNameOff),.true.)
  END IF
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:ZoneAirDistribution')
    CALL AddToObjStr('Name','SZ DSZAD '//FldVal(base+vzNameOff))
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Cooling Mode {}','1')
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Heating Mode {}','1',.true.)
  END IF
  !CONTROLLED ZONE EQUIP CONFIGURATION  ~ line 41
  CALL CreateNewObj('ZoneHVAC:EquipmentConnections')
  CALL AddToObjFld('Zone Name', base +  vzNameOff,'')
  CALL AddToObjFld('Zone Conditioning Equipment List Name', base +  vzNameOff,' Equipment')
  CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base +  vzNameOff,' Supply Inlet')
  CALL AddToObjStr('Zone Air Exhaust Node or NodeList Name','')
  CALL AddToObjFld('Zone Air Node Name', base +  vzNameOff,' Zone Air Node')
  IF (isZoneRetPlenBlank) THEN
    CALL AddToObjFld('Zone Return Air Node Name', base +  vzNameOff,' Return Outlet',.TRUE.)
  ELSE
    CALL AddToObjFld('Zone Return Air Node Name', base + vzNameOff,' Zone Return Plenum Inlet',.TRUE.)
  END IF
  !ZONE SUPPLY PLENUM ~line 50
  IF (.NOT. isZoneSupPlenBlank) THEN
    CALL CreateNewObj('AirLoopHVAC:SupplyPlenum')
    CALL AddToObjFld('Name', base + vzNameOff,' Supply Plenum')
    CALL AddToObjFld('Zone Name', base + vzZoneSupplyPlenumNameOff,'')
    CALL AddToObjFld('Zone Node Name', base + vzZoneSupplyPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Inlet Node Name', base + vzNameOff,' Zone Equip Inlet')
    CALL AddToObjFld('Outlet Node Name', base + vzNameOff,' Terminal Unit Inlet',.TRUE.)
  END IF
  !ZONE RETURN PLENUM ~ line 57
  IF (.NOT. isZoneRetPlenBlank) THEN
    CALL CreateNewObj('AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Name', base + vzNameOff,' Return Plenum')
    CALL AddToObjFld('Zone Name', base + vzZoneReturnPlenumNameOff,'')
    CALL AddToObjFld('Zone Node Name', base + vzZoneReturnPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Outlet Node Name', base + vzNameOff,' Return Outlet')
    CALL AddToObjStr('Induced Air Outlet Node or NodeList Name','')
    CALL AddToObjFld('Inlet Node Name', base + vzNameOff,' Zone Return Plenum Inlet',.TRUE.)
  END IF
  !ZONE EQUIPMENT LIST ~ line 64
  CALL CreateNewObj('ZoneHVAC:EquipmentList')
  CALL AddToObjFld('Name', base +  vzNameOff,' Equipment')
  CALL AddToObjStr('Load Distribution Scheme','SequentialLoad')
  CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:AirDistributionUnit')
  CALL AddToObjFld('Zone Equipment Name', base +  vzNameOff,' ATU')
  CALL AddToObjStr('Zone Equipment Cooling Sequence','1')
  SELECT CASE (baseboardKind)
    CASE (baseboardNone)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1',.FALSE.)
    CASE (baseboardHotWater)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1')
      CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
      CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.FALSE.)
      CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Zone Equipment Name', base + vzNameOff,' Baseboard Heat')
      CALL AddToObjStr('Zone Equipment Cooling Sequence','2')
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','2',.FALSE.)
    CASE (baseboardElectric)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1')
      CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Zone Equipment Name', base + vzNameOff,' Baseboard Heat')
      CALL AddToObjStr('Zone Equipment Cooling Sequence','2')
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','2',.FALSE.)
  END SELECT
  CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.TRUE.)
!AIR DISTRIBUTION UNIT ~ line 72
  IF (reheatCoilType .EQ. ctNone) THEN
    CALL CreateNewObj('ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Name', base +  vzNameOff,' ATU')
    CALL AddToObjFld('Air Distribution Unit Outlet Node Name', base +  vzNameOff,' Supply Inlet')
    CALL AddToObjStr('Air Terminal Object Type','AirTerminal:SingleDuct:VAV:NoReheat')
    CALL AddToObjFld('Air Terminal Name', base +  vzNameOff,' VAV',.TRUE.)
  ELSE
    CALL CreateNewObj('ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Name', base +  vzNameOff,' ATU')
    CALL AddToObjFld('Air Distribution Unit Outlet Node Name', base +  vzNameOff,' Supply Inlet')
    CALL AddToObjStr('Air Terminal Object Type','AirTerminal:SingleDuct:VAV:Reheat')
    CALL AddToObjFld('Air Terminal Name', base +  vzNameOff,' VAV Reheat',.TRUE.)
  END IF
  !This object if vzReheatTypeOff=Hot Water
  IF (reheatCoilType .EQ. ctHotWater) THEN
    !SINGLE DUCT:VAV:REHEAT   ~ line 85
    CALL CreateNewObj('AirTerminal:SingleDuct:VAV:Reheat')
    CALL AddToObjFld('Name', base +  vzNameOff,' VAV Reheat')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Damper Air Outlet Node Name', base + vzNameOff,' Damper Outlet')
    IF (isZoneSupPlenBlank) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + vzNameOff,' Zone Equip Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + vzNameOff,' Terminal Unit Inlet')
    END IF
    !CR8001
    !CALL AddToObjFld('Maximum air flow rate {m3/s} ', base + vzSupplyMaxRateOff,' ')
    CALL AddToObjStr('Maximum air flow rate {m3/s} ', 'autosize')
    CALL AddToObjFld('Zone Minimum Air Flow Input Method', base + vzMinFlowMethodOff,' ')
    CALL AddToObjFld('Constant Minimum Air Flow Fraction', base + vzConstantMinFracOff,' ')
    CALL AddToObjFld('Fixed Minimum Air Flow Rate', base + vzFixedMinFlowOff,' ')
    CALL AddToObjFld('Minimum Air Flow Fraction Schedule Name', base + vzMinFracSchedNameOff,' ')
!    CALL AddToObjFld('Sensor Node Name', base + vzNameOff,' Reheat Coil HW Inlet')
    CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Reheat Coil Name', base + vzNameOff,' Reheat Coil')
    CALL AddToObjStr('Maximum Hot Water or Steam Flow Rate {m3/s}  ','autosize')
    CALL AddToObjStr('Minimum Hot Water or Steam Flow Rate {m3/s}  ','0')
    CALL AddToObjFld('Air Outlet Node Name', base + vzNameOff,' Supply Inlet')
    CALL AddToObjStr('Convergence Tolerance','0.001')
    CALL AddToObjFld('Damper Heating Action', base + vzDamperActionOff,' ')
    CALL AddToObjFld('Maximum Flow per Zone Floor Area During Reheat', base + vzMaxReheatFlowPerAreaOff,' ')
    CALL AddToObjFld('Maximum Flow Fraction During Reheat', base + vzMaxReheatFlowFractionOff,' ')
    CALL AddToObjFld('Maximum Reheat Air Temperature', base + vzMaxReheatAirTempOff,' ')
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + vzDesignSpecOANameCtrlOff,' ',.TRUE.)
    !COIL:Water:SimpleHeating  ~ line 104
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjFld('Name', base + vzNameOff,' Reheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + vzReheatSchedNameOff,' ')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name', base + vzNameOff,' Reheat Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + vzNameOff,' Reheat Coil HW Outlet')
    CALL AddToObjFld('Air Inlet Node Name', base + vzNameOff,' Damper Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + vzNameOff,' Supply Inlet')
    CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
    CALL AddToObjStr('Rated Capacity','autosize')
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
    !BRANCH ~ line 113
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + vzNameOff,' Reheat Coil HW Branch')
    !save this branch name for later use in plant
    CALL AddToStrList(TRIM(FldVal(base + vzNameOff)) // ' Reheat Coil HW Branch',handleHeatingCoilBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + vzNameOff,' Reheat Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + vzNameOff,' Reheat Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + vzNameOff,' Reheat Coil HW Outlet',.TRUE.)
  ELSEIF (reheatCoilType .EQ. ctElectric) THEN
    !SINGLE DUCT:VAV:REHEAT ~ line 122
    CALL CreateNewObj('AirTerminal:SingleDuct:VAV:Reheat')
    CALL AddToObjFld('Name', base +  vzNameOff,' VAV Reheat')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Damper Air Outlet Node Name', base + vzNameOff,' Damper Outlet')
    IF (isZoneSupPlenBlank) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + vzNameOff,' Zone Equip Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + vzNameOff,' Terminal Unit Inlet')
    END IF
    !CR8001
    !CALL AddToObjFld('Maximum air flow rate {m3/s}', base + vzSupplyMaxRateOff,' ')
    CALL AddToObjStr('Maximum air flow rate {m3/s}','autosize')
    CALL AddToObjFld('Zone Minimum Air Flow Input Method', base + vzMinFlowMethodOff,' ')
    CALL AddToObjFld('Constant Minimum Air Flow Fraction', base + vzConstantMinFracOff,' ')
    CALL AddToObjFld('Fixed Minimum Air Flow Rate', base + vzFixedMinFlowOff,' ')
    CALL AddToObjFld('Minimum Air Flow Fraction Schedule Name', base + vzMinFracSchedNameOff,' ')
!    CALL AddToObjStr('Sensor Node Name','')
    CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Electric')
    CALL AddToObjFld('Reheat Coil Name', base + vzNameOff,' Reheat Coil')
    CALL AddToObjStr('Maximum Hot Water or Steam Flow Rate {m3/s}','')
    CALL AddToObjStr('Minimum Hot Water or Steam Flow Rate {m3/s}','')
    CALL AddToObjFld('Air Outlet Node Name', base + vzNameOff,' Supply Inlet')
    CALL AddToObjStr('Convergence Tolerance','0.001')
    CALL AddToObjFld('Damper Heating Action', base + vzDamperActionOff,' ')
    CALL AddToObjFld('Maximum Flow per Zone Floor Area During Reheat', base + vzMaxReheatFlowPerAreaOff,' ')
    CALL AddToObjFld('Maximum Flow Fraction During Reheat', base + vzMaxReheatFlowFractionOff,' ')
    CALL AddToObjFld('Maximum Reheat Air Temperature', base + vzMaxReheatAirTempOff,' ')
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + vzDesignSpecOANameCtrlOff,' ',.TRUE.)
    !COIL:Electric:Heating ~ line 139
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjFld('Name', base + vzNameOff,' Reheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + vzReheatSchedNameOff,' ')
    CALL AddToObjStr('Efficiency','1')
    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    CALL AddToObjFld('Air Inlet Node Name', base + vzNameOff,' Damper Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + vzNameOff,' Supply Inlet',.TRUE.)
  ELSEIF (reheatCoilType .EQ. ctGas) THEN
    !SINGLE DUCT:VAV:REHEAT ~ line 122
    CALL CreateNewObj('AirTerminal:SingleDuct:VAV:Reheat')
    CALL AddToObjFld('Name', base +  vzNameOff,' VAV Reheat')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Damper Air Outlet Node Name', base + vzNameOff,' Damper Outlet')
    IF (isZoneSupPlenBlank) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + vzNameOff,' Zone Equip Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + vzNameOff,' Terminal Unit Inlet')
    END IF
    !CR8001
    CALL AddToObjStr('Maximum air flow rate {m3/s}','autosize')
    CALL AddToObjFld('Zone Minimum Air Flow Input Method', base + vzMinFlowMethodOff,' ')
    CALL AddToObjFld('Constant Minimum Air Flow Fraction', base + vzConstantMinFracOff,' ')
    CALL AddToObjFld('Fixed Minimum Air Flow Rate', base + vzFixedMinFlowOff,' ')
    CALL AddToObjFld('Minimum Air Flow Fraction Schedule Name', base + vzMinFracSchedNameOff,' ')
    CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Fuel')
    CALL AddToObjFld('Reheat Coil Name', base + vzNameOff,' Reheat Coil')
    CALL AddToObjStr('Maximum Hot Water or Steam Flow Rate {m3/s}','')
    CALL AddToObjStr('Minimum Hot Water or Steam Flow Rate {m3/s}','')
    CALL AddToObjFld('Air Outlet Node Name', base + vzNameOff,' Supply Inlet')
    CALL AddToObjStr('Convergence Tolerance','0.001')
    CALL AddToObjFld('Damper Heating Action', base + vzDamperActionOff,' ')
    CALL AddToObjFld('Maximum Flow per Zone Floor Area During Reheat', base + vzMaxReheatFlowPerAreaOff,' ')
    CALL AddToObjFld('Maximum Flow Fraction During Reheat', base + vzMaxReheatFlowFractionOff,' ')
    CALL AddToObjFld('Maximum Reheat Air Temperature', base + vzMaxReheatAirTempOff,' ')
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + vzDesignSpecOANameCtrlOff,' ',.TRUE.)
    !Coil:Heating:Fuel
    CALL CreateNewObj('Coil:Heating:Fuel')
    CALL AddToObjFld('Name', base + vzNameOff,' Reheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + vzReheatSchedNameOff,' ')
    CALL AddToObjStr('Fuel Type', 'Gas')
      CALL AddToObjStr('Gas Burner Efficiency','0.8')
      CALL AddToObjStr('Nominal Capacity {W}', 'autosize')
    CALL AddToObjFld('Air Inlet Node Name', base + vzNameOff,' Damper Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + vzNameOff,' Supply Inlet')
    CALL AddToObjStr('Temperature Setpoint Node Name','')
    CALL AddToObjStr('Parasitic Electric Load {W}', '0')
    CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
        base + vzNameOff,' Reheat Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
    !CURVE:CUBIC
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + vzNameOff,' Reheat Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  ELSE
    !SINGLE DUCT:VAV:NOREHEAT ~ line 148
    CALL CreateNewObj('AirTerminal:SingleDuct:VAV:NoReheat')
    CALL AddToObjFld('Name', base +  vzNameOff,' VAV')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Air Outlet Node Name', base + vzNameOff,' Supply Inlet')
    IF (isZoneSupPlenBlank) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + vzNameOff,' Zone Equip Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + vzNameOff,' Terminal Unit Inlet')
    END IF
    !CR8001
    !CALL AddToObjFld('Maximum air flow rate {m3/s}', base + vzSupplyMaxRateOff,' ')
    CALL AddToObjStr('Maximum air flow rate {m3/s}', 'autosize ')
    CALL AddToObjFld('Zone Minimum Air Flow Input Method', base + vzMinFlowMethodOff,'')
    CALL AddToObjFld('Constant Minimum Air Flow Fraction', base + vzConstantMinFracOff,'')
    CALL AddToObjFld('Fixed Minimum Air Flow Rate', base + vzFixedMinFlowOff,'')
    CALL AddToObjFld('Minimum Air Flow Fraction Schedule Name', base + vzMinFracSchedNameOff,'')
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + vzDesignSpecOANameCtrlOff,'',.TRUE.)
  END IF
  !Baseboards
  SELECT CASE (baseboardKind)
    CASE (baseboardHotWater)
      !ZoneHVAC:Baseboard:RadiantConvective:Water
      CALL CreateNewObj('ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Name', base + vzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + vzBaseboardAvailSchedNameOff,'')
      CALL AddToObjFld('Inlet Node Name', base + vzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Outlet Node Name', base + vzNameOff,' Baseboard HW Outlet')
      CALL AddToObjStr('Rated Average Water Temperature {C}','87.78')
      CALL AddToObjStr('Rated Water Mass Flow Rate {m3/s}','0.063')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + vzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Maximum Water Flow Rate','autosize')
      CALL AddToObjStr('Convergence Tolerance','0.001')
      CALL AddToObjStr('Fraction Radiant','0.0')
      CALL AddToObjStr('Fraction of Radiant Energy Incident on People','0.0',.TRUE.)
      !BRANCH (for baseboard)
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + vzNameOff,' Baseboard Heat HW Branch')
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Component Name', base + vzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Component Inlet Node Name', base + vzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + vzNameOff,' Baseboard HW Outlet',.TRUE.)
      CALL AddToStrList(TRIM(FldVal(base + vzNameOff)) // ' Baseboard Heat HW Branch',handleHeatingCoilBranch)
      IF ((compactHotLoopBase .EQ. 0) .OR. (numCompactBoiler .EQ. 0)) THEN
        CALL WriteError('Both a HVACTemplate:Plant:HotWaterLoop and a HVACTemplate:Plant:Boiler are needed ' // &
             'when using hot water baseboards.')
      END IF
    CASE (baseboardElectric)
      !BASEBOARD HEATER:Electric:Convective
      CALL CreateNewObj('ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Name', base + vzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + vzBaseboardAvailSchedNameOff,'')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + vzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Efficiency','1',.TRUE.)
  END SELECT
END DO
!======================================================================================================
! create Fan Powered Zones
!======================================================================================================
DO iZone = 1, numCompactZoneFPVAV
  base =  compactZoneFPVAVbase(iZone)
  ! find base for system this zone belongs to
  sysBase = -1
  DO iSys = 1, numCompactSysVAV
    IF (SameString(FldVal(compactSysVAVbase(iSys) + vsAirHandlerNameOff), FldVal(base +  fpvzAirHandlerNameOff))) THEN
      sysBase = compactSysVAVBase(iSys)
      EXIT
    END IF
  END DO
  IF (sysBase .NE. -1) THEN
    isPackaged = .FALSE.
  ELSE
    DO iSys = 1, numCompactSysPVAV
      IF (SameString(FldVal(compactSysPVAVbase(iSys) + pvavsAirHandlerNameOff), FldVal(base +  fpvzAirHandlerNameOff))) THEN
        sysBase = compactSysPVAVBase(iSys)
        EXIT
      END IF
    END DO
    IF (sysBase .NE. -1) isPackaged = .TRUE.
  END IF
  IF (sysBase .EQ. -1) THEN
    CALL WriteError('Could not find air handler name referenced in HVACTemplate:Zone:VAV:FanPowered object: '// &
     TRIM(FldVal(base + fpvzNameOff)))
    CYCLE
  END IF
  ! set flags
  isTStatNameBlank = (FldVal(base + fpvzTStatNameOff) .EQ. '')
  isPriSupplyAuto = SameString(FldVal(base + fpvzPriSupplyMaxRateOff),'autosize')
  isSecSupplyAuto = SameString(FldVal(base + fpvzSecSupplyMaxRateOff),'autosize')
  isZoneRetPlenBlank = (FldVal(base + fpvzZoneReturnPlenumNameOff) .EQ. '')
  isZoneSupPlenBlank = (FldVal(base + fpvzZoneSupplyPlenumNameOff) .EQ. '')
  isZoneFanSchedBlank = (FldVal(base + fpvzZoneFanSchedNameOff) .EQ. '')
  isOATypeDetailed = SameString(FldVal(base + fpvzOutAirTypeOff),'DetailedSpecification')
  IF (SameString(FldVal(base + fpvzReheatTypeOff),'HotWater')) THEN
    reheatCoilType = ctHotWater
  ELSEIF (SameString(FldVal(base + fpvzReheatTypeOff),'Electric')) THEN
    reheatCoilType = ctElectric
  ELSEIF (SameString(FldVal(base + fpvzReheatTypeOff),'Gas')) THEN
    reheatCoilType = ctGas
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VAV:FanPowered in the Reheat Coil Type field')
  END IF
  !set the baseboard indicator
  IF (SameString(FldVal(base +  fpvzBaseboardTypeOff),'None')) THEN
    baseboardKind = baseboardNone
  ELSEIF (SameString(FldVal(base +  fpvzBaseboardTypeOff),'HotWater')) THEN
    baseboardKind = baseboardHotWater
  ELSEIF (SameString(FldVal(base +  fpvzBaseboardTypeOff),'Electric')) THEN
    baseboardKind = baseboardElectric
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VAV:FanPowered "'//TRIM(FldVal(base + fpvzNameOff))//'"'// &
                    ' in the Baseboard Heating Type field: '//TRIM(FldVal(base + fpvzBaseboardTypeOff)))
  END IF
  !validate flow type and set flags
  isFlowTypeSeries = .FALSE.
  isFlowTypeFromZonePlenum = .FALSE.
  isFlowTypeFromZone = .FALSE.
  IF (SameString(FldVal(base +  fpvzFlowTypeOff),'Series')) THEN
    isFlowTypeSeries = .TRUE.
    isFlowTypeFromZone = .TRUE.
  ELSEIF (SameString(FldVal(base +  fpvzFlowTypeOff),'SeriesFromPlenum')) THEN
    isFlowTypeSeries = .TRUE.
    IF ( .NOT. isZoneRetPlenBlank) isFlowTypeFromZonePlenum = .TRUE.
  ELSEIF (SameString(FldVal(base +  fpvzFlowTypeOff),'Parallel')) THEN
    isFlowTypeFromZone = .TRUE.
  ELSEIF (SameString(FldVal(base +  fpvzFlowTypeOff),'ParallelFromPlenum')) THEN
    IF ( .NOT. isZoneRetPlenBlank) isFlowTypeFromZonePlenum = .TRUE.
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VAV:FanPowered "'//TRIM(FldVal(base + fpvzNameOff))//'"'// &
                    ' in the Flow Type field: '// &
                    TRIM(FldVal(base + fpvzFlowTypeOff)))
  END IF
  !validate cooling design method
  IF (SameString(FldVal(base +  fpvzCoolDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  fpvzCoolDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  fpvzCoolDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VAV:FanPowered "'//TRIM(FldVal(base + fpvzNameOff))//'"'// &
                    ' in the Zone Cooling Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + fpvzCoolDesignTempMethodOff)))
  END IF
  !validate heating design method
  IF (SameString(FldVal(base +  fpvzHeatDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  fpvzHeatDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VAV:FanPowered "'//TRIM(FldVal(base + fpvzNameOff))//'"'// &
                    ' in the Zone Heating Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + fpvzHeatDesignTempMethodOff)))
  END IF
  IF (.NOT. isTStatNameBlank) THEN
    !ZONE CONTROL:THERMOSTATIC ~line 4 of FPVAV tab
    CALL AddAlwaysSchedule('4')
    CALL CreateNewObj('ZoneControl:Thermostat')
    CALL AddToObjFld('Name', base + fpvzNameOff,' Thermostat')
    CALL AddToObjFld('Zone or ZoneList Name', base + fpvzNameOff,'')
    CALL AddToObjStr('Control Type Schedule Name','HVACTemplate-Always 4')
    CALL AddToObjStr('Control Object Type','ThermostatSetpoint:DualSetpoint')
    CALL AddToObjFld('Control Name', base + fpvzTStatNameOff,' Dual SP Control',.TRUE.)
  END IF
  !ZONE SIZING ~ line 11 of FPVAV tab
  ! OADCV changes
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:OutdoorAir')
    CALL AddToObjStr('Name', 'SZ DSOA '//FldVal(base+fpvzNameOff))
    CALL AddToObjFld('Outdoor Air Method', base + fpvzOutAirTypeOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Person {m3/s}', base + fpvzOutAirFlowPersonOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone Floor Area {m3/s-m2}', base + fpvzOutAirFlowAreaOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone {m3/s}', base + fpvzOutAirFlowZoneOff,'',.true.)
  END IF

  CALL CreateNewObj('Sizing:Zone')
  CALL AddToObjFld('Zone or ZoneList Name', base + fpvzNameOff,'')
  IF (SameString(FldVal(base +  fpvzCoolDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
    CALL AddToObjStr('Zone Cooling Design Supply Air Temperature Input Method','SupplyAirTemperature')
    IF (isPackaged) THEN
      CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', sysBase + pvavsCoolSetPtDesignOff,'')
    ELSE
      CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', sysBase + vsCoolSetPtDesignOff,'')
    ENDIF
  ELSE
    CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Input Method',base + fpvzCoolDesignTempMethodOff,'')
    CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', base + fpvzCoolDesignTempOff,'')
  ENDIF
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Difference {delta C}',base + fpvzCoolDesignTempDiffOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Input Method',base + fpvzHeatDesignTempMethodOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', base + fpvzHeatDesignTempOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Difference {delta C}',base + fpvzHeatDesignTempDiffOff,'')
  CALL AddToObjStr('Zone Cooling Design Supply Air Humidity Ratio {kg-H20/kg-air}','0.008')
  CALL AddToObjStr('Zone Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + fpvzDesSpecOANameOff,'')
  ELSE
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','SZ DSOA '//FldVal(base+fpvzNameOff))
  END IF
  CALL AddToObjFld('Zone Heating Sizing Factor', base + fpvzZoneHeatSizingFactorOff,'')
  CALL AddToObjFld('Zone Cooling Sizing Factor', base + fpvzZoneCoolSizingFactorOff,'')
  !CR8001 - separate IF for isPriSuplyAuto and isSecSupplyAuto
  IF (isPriSupplyAuto) THEN
    CALL AddToObjStr('Cooling Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Cooling Design Air Flow Rate {m3/s}','0')
  ELSE
    CALL AddToObjStr('Cooling Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Cooling Design Air Flow Rate {m3/s}', base + fpvzPriSupplyMaxRateOff,'')
  END IF
  CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
  CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}',' ')
  CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  IF (isSecSupplyAuto) THEN
    CALL AddToObjStr('Heating Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Heating Design Air Flow Rate {m3/s}','0')
  ELSE
    CALL AddToObjStr('Heating Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Heating Design Air Flow Rate {m3/s}', base + fpvzSecSupplyMaxRateOff,'')
    !heatDesAirRate = StringToReal(FldVal(base + fpvzPriSupplyMaxRateOff)) * &
    !                 StringToReal(FldVal(base + fpvzPriSupplyMinFracOff))
    !CALL AddToObjNum('heating design air flow rate {m3/s}', heatDesAirRate)
  END IF
  CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
  CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
  CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Zone Air Distribution Object Name', base + fpvzDesSpecOADistNameOff,'',.true.)
  ELSE
    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+fpvzNameOff),.true.)
  END IF

  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:ZoneAirDistribution')
    CALL AddToObjStr('Name','SZ DSZAD '//FldVal(base + fpvzNameOff))
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Cooling Mode {}','1')
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Heating Mode {}','1',.true.)
  END IF

  !CONTROLLED ZONE EQUIP CONFIGURATION ~ line 38 of FPVAV tab
  CALL CreateNewObj('ZoneHVAC:EquipmentConnections')
  CALL AddToObjFld('Zone Name', base + fpvzNameOff,'             ')
  CALL AddToObjFld('Zone Conditioning Equipment List Name', base + fpvzNameOff,' Equipment')
  CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base + fpvzNameOff,' Supply Inlet')
  IF (isFlowTypeFromZone) THEN
    CALL AddToObjFld('Zone Air Exhaust Node or NodeList Name', base + fpvzNameOff,' PIU Secondary Inlet')
  ELSE
    CALL AddToObjStr('Zone Air Exhaust Node or NodeList Name','')
  END IF
  CALL AddToObjFld('Zone Air Node Name', base + fpvzNameOff,' Zone Air Node')
  IF (isZoneRetPlenBlank) THEN
    CALL AddToObjFld('Zone Return Air Node Name', base + fpvzNameOff,' Return Outlet',.TRUE.)
  ELSE
    CALL AddToObjFld('Zone Return Air Node Name', base + fpvzNameOff,' Zone Return Plenum Inlet',.TRUE.)
  END IF
  !ZONE SUPPLY PLENUM ~ line 47 of FPVAV tab
  IF (.NOT. isZoneSupPlenBlank) THEN
    CALL CreateNewObj('AirLoopHVAC:SupplyPlenum')
    CALL AddToObjFld('Name', base + fpvzNameOff,' Supply Plenum')
    CALL AddToObjFld('Zone Name', base + fpvzZoneSupplyPlenumNameOff,'')
    CALL AddToObjFld('Zone Node Name', base + fpvzZoneSupplyPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Inlet Node Name', base + fpvzNameOff,' Zone Equip Inlet')
    CALL AddToObjFld('Outlet Node Name', base + fpvzNameOff,' Terminal Unit Inlet',.TRUE.)
  END IF
  !ZONE RETURN PLENUM ~ line 54 of FPVAV tab
  IF (.NOT.isZoneRetPlenBlank) THEN
    CALL CreateNewObj('AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Name', base + fpvzNameOff,' Return Plenum')
    CALL AddToObjFld('Zone Name', base + fpvzZoneReturnPlenumNameOff,'')
    CALL AddToObjFld('Zone Node Name', base + fpvzZoneReturnPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Outlet Node Name', base + fpvzNameOff,' Return Outlet')
    IF (isFlowTypeFromZonePlenum) THEN
      CALL AddToObjFld('Induced Air Outlet Node or NodeList Name', base + fpvzNameOff,' PIU Secondary Inlet')
    ELSE
      CALL AddToObjStr('Induced Air Outlet Node or NodeList Name','')
    END IF
    CALL AddToObjFld('Inlet Node Name', base + fpvzNameOff,' Zone Return Plenum Inlet',.TRUE.)
  END IF
  !ZONE EQUIPMENT LIST ~ line 61 of FPVAV tab
  CALL CreateNewObj('ZoneHVAC:EquipmentList')
  CALL AddToObjFld('Name', base + fpvzNameOff,' Equipment')
  CALL AddToObjStr('Load Distribution Scheme','SequentialLoad')
  CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:AirDistributionUnit')
  CALL AddToObjFld('Zone Equipment Name', base + fpvzNameOff,' ATU')
  CALL AddToObjStr('Zone Equipment Cooling Sequence','1')
  SELECT CASE (baseboardKind)
    CASE (baseboardNone)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1',.FALSE.)
    CASE (baseboardHotWater)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1')
      CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
      CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.FALSE.)
      CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Zone Equipment Name', base + fpvzNameOff,' Baseboard Heat')
      CALL AddToObjStr('Zone Equipment Cooling Sequence','2')
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','2',.FALSE.)
    CASE (baseboardElectric)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1')
      CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Zone Equipment Name', base + fpvzNameOff,' Baseboard Heat')
      CALL AddToObjStr('Zone Equipment Cooling Sequence','2')
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','2',.FALSE.)
  END SELECT
  CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.TRUE.)
  !AIR DISTRIBUTION UNIT ~ line 69 of FPVAV tab
  CALL CreateNewObj('ZoneHVAC:AirDistributionUnit')
  CALL AddToObjFld('Name', base + fpvzNameOff,' ATU')
  CALL AddToObjFld('Air Distribution Unit Outlet Node Name', base + fpvzNameOff,' Supply Inlet')
  IF (isFlowTypeSeries) THEN
    CALL AddToObjStr('Air Terminal Object Type','AirTerminal:SingleDuct:SeriesPIU:Reheat')
    CALL AddToObjFld('Air Terminal Name', base + fpvzNameOff,' VAV Series PIU Reheat',.TRUE.)
  ELSE
    CALL AddToObjStr('Air Terminal Object Type','AirTerminal:SingleDuct:ParallelPIU:Reheat')
    CALL AddToObjFld('Air Terminal Name', base + fpvzNameOff,' VAV Parallel PIU Reheat',.TRUE.)
  END IF
  IF (isFlowTypeSeries) THEN
    !SINGLE DUCT:SERIES PIU:REHEAT ~ line 78 of FPVAV tab
    CALL CreateNewObj('AirTerminal:SingleDuct:SeriesPIU:Reheat')
    CALL AddToObjFld('Name', base + fpvzNameOff,' VAV Series PIU Reheat')
    CALL AddToObjStr('Availability Schedule Name',' ')
    !CR8001
    !IF (isPriSupplyAuto .OR. isSecSupplyAuto) THEN
    CALL AddToObjStr('Maximum total air volume flow rate {m3/s}','autosize')
    !ELSE
    !  totalFlow = StringToReal(FldVal(base + fpvzPriSupplyMaxRateOff)) + &
    !              StringToReal(FldVal(base + fpvzSecSupplyMaxRateOff))
    !  CALL AddToObjNum('Maximum total air volume flow rate {m3/s}', totalFlow)
    !END IF
    !CR8001
    !CALL AddToObjFld('Maximum primary air volume flow rate {m3/s}', base + fpvzPriSupplyMaxRateOff,'')
    CALL AddToObjStr('Maximum primary air volume flow rate {m3/s}', 'autosize')
    CALL AddToObjFld('Minimum Primary Air Flow Fraction', base + fpvzPriSupplyMinFracOff,'')
    IF (isZoneSupPlenBlank) THEN
      CALL AddToObjFld('Supply Air Inlet Node Name', base + fpvzNameOff,' Zone Equip Inlet')
    ELSE
      CALL AddToObjFld('Supply Air Inlet Node Name', base + fpvzNameOff,' Terminal Unit Inlet')
    END IF
    CALL AddToObjFld('Secondary Air Inlet Node Name', base + fpvzNameOff,' PIU Secondary Inlet')
    CALL AddToObjFld('Outlet Node Name', base + fpvzNameOff,' Supply Inlet')
    CALL AddToObjFld('Reheat Coil Air Inlet Node Name', base + fpvzNameOff,' PIU Fan Outlet')
    CALL AddToObjFld('Zone Mixer Name', base + fpvzNameOff,' PIU Mixer')
    CALL AddToObjFld('Fan Name', base + fpvzNameOff,' PIU Fan')
    IF (reheatCoilType .EQ. ctHotWater) THEN
      CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Water')
    ELSEIF (reheatCoilType .EQ. ctElectric) THEN
      CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Electric')
    ELSEIF (reheatCoilType .EQ. ctGas) THEN
      CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Fuel')
    END IF
    CALL AddToObjFld('Reheat Coil Name', base + fpvzNameOff,' Reheat Coil')
    IF (reheatCoilType .EQ. ctHotWater) THEN
      CALL AddToObjStr('Max Reheat Water Flow {m3/s}','autosize')
      CALL AddToObjStr('Min Reheat Water Flow {m3/s}','0')
    ELSE
      CALL AddToObjStr('Max Reheat Water Flow {m3/s}  ','')
      CALL AddToObjStr('Min Reheat Water Flow {m3/s}  ','')
    END IF
    CALL AddToObjStr('Convergence Tolerance','0.001',.TRUE.)
  ELSE
    !SINGLE DUCT:PARALLEL PIU:REHEAT ~ line 104 of FPVAV tab
    CALL CreateNewObj('AirTerminal:SingleDuct:ParallelPIU:Reheat')
    CALL AddToObjFld('Name', base + fpvzNameOff,' VAV Parallel PIU Reheat')
    CALL AddToObjStr('Availability Schedule Name',' ')
    !CR8001
    !CALL AddToObjFld('Maximum primary air volume flow rate {m3/s}     ', base + fpvzPriSupplyMaxRateOff,'')
    !CALL AddToObjFld('Maximum secondary air volume flow rate {m3/s}     ', base + fpvzSecSupplyMaxRateOff,'')
    CALL AddToObjStr('Maximum primary air volume flow rate {m3/s}', 'autosize')
    CALL AddToObjStr('Maximum secondary air volume flow rate {m3/s}', 'autosize')
    CALL AddToObjFld('Minimum Primary Air Flow Fraction', base + fpvzPriSupplyMinFracOff,'')
    CALL AddToObjFld('Fan On Flow Fraction', base + fpvzFanOnFlowFracOff,'')
    IF (isZoneSupPlenBlank) THEN
      CALL AddToObjFld('Supply Air Inlet Node Name', base + fpvzNameOff,' Zone Equip Inlet')
    ELSE
      CALL AddToObjFld('Supply Air Inlet Node Name', base + fpvzNameOff,' Terminal Unit Inlet')
    END IF
    CALL AddToObjFld('Secondary Air Inlet Node Name', base + fpvzNameOff,' PIU Secondary Inlet')
    CALL AddToObjFld('Outlet Node Name', base + fpvzNameOff,' Supply Inlet')
    CALL AddToObjFld('Reheat Coil Air Inlet Node Name', base + fpvzNameOff,' PIU Mixer Outlet')
    CALL AddToObjFld('Zone Mixer Name', base + fpvzNameOff,' PIU Mixer')
    CALL AddToObjFld('Fan Name', base + fpvzNameOff,' PIU Fan')
    IF (reheatCoilType .EQ. ctHotWater) THEN
      CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Water')
    ELSEIF (reheatCoilType .EQ. ctElectric) THEN
      CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Electric')
    ELSEIF (reheatCoilType .EQ. ctGas) THEN
      CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Fuel')
    END IF
    CALL AddToObjFld('Reheat Coil Name', base + fpvzNameOff,' Reheat Coil')
    IF (reheatCoilType .EQ. ctHotWater) THEN
      CALL AddToObjStr('Max Reheat Water Flow {m3/s}  ','autosize')
      CALL AddToObjStr('Min Reheat Water Flow {m3/s}  ','0')
    ELSE
      CALL AddToObjStr('Max Reheat Water Flow {m3/s}  ','')
      CALL AddToObjStr('Min Reheat Water Flow {m3/s}  ','')
    END IF
    CALL AddToObjStr('Convergence Tolerance         ','0.001',.TRUE.)
  END IF
  !ZONE MIXER ~ line 129 of FPVAV tab
  CALL CreateNewObj('AirLoopHVAC:ZoneMixer')
  CALL AddToObjFld('Name', base + fpvzNameOff,' PIU Mixer')
  CALL AddToObjFld('Outlet Node Name', base + fpvzNameOff,' PIU Mixer Outlet')
  IF (isZoneSupPlenBlank) THEN
    CALL AddToObjFld('Inlet Node 1', base + fpvzNameOff,' Zone Equip Inlet')
  ELSE
    CALL AddToObjFld('Inlet Node 1', base + fpvzNameOff,' Terminal Unit Inlet')
  END IF
  IF (isFlowTypeSeries) THEN
    CALL AddToObjFld('Inlet Node 2', base + fpvzNameOff,' PIU Secondary Inlet',.TRUE.)
  ELSE
    CALL AddToObjFld('Inlet Node 2', base + fpvzNameOff,' PIU Fan Outlet',.TRUE.)
  END IF
  !FAN:SIMPLE:CONSTVOLUME ~ line 137 of FPVAV tab
  CALL CreateNewObj('Fan:ConstantVolume')
  CALL AddToObjFld('Name', base + fpvzNameOff,' PIU Fan')
  ! If zone fan schedule name is not specified then use the system availability schedule
  IF (isZoneFanSchedBlank) THEN
    IF (isPackaged) THEN
      CALL AddToObjFld('Availability Schedule Name', sysBase + pvavsSysAvailSchedNameOff,'')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', sysBase + vsSysAvailSchedNameOff,'')
    ENDIF
  ELSE
    CALL AddToObjFld('Availability Schedule Name', base + fpvzZoneFanSchedNameOff,'')
  END IF
  CALL AddToObjFld('Fan Efficiency', base + fpvzFanTotEfficiencyOff,'')
  CALL AddToObjFld('Pressure Rise {Pa}', base + fpvzFanPressureOff,'')
  CALL AddToObjStr('Maximum Flow Rate {m3/s}','autosize')
  CALL AddToObjFld('Motor Efficiency', base + fpvzFanMotorEfficiencyOff,'')
  CALL AddToObjStr('Motor in Airstream Fraction','1')
  IF (isFlowTypeSeries) THEN
    CALL AddToObjFld('Air Inlet Node Name', base + fpvzNameOff,' PIU Mixer Outlet')
  ELSE
    CALL AddToObjFld('Air Inlet Node Name', base + fpvzNameOff,' PIU Secondary Inlet')
  END IF
  CALL AddToObjFld('Air Outlet Node Name', base + fpvzNameOff,' PIU Fan Outlet',.TRUE.)
  IF (reheatCoilType .EQ. ctHotWater) THEN
    !FAN:SIMPLE:CONSTVOLUME ~ line 151 of FPVAV tab
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjFld('Name', base + fpvzNameOff,' Reheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + fpvzReheatSchedNameOff,'')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name', base + fpvzNameOff,' Reheat Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + fpvzNameOff,' Reheat Coil HW Outlet')
    IF (isFlowTypeSeries) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + fpvzNameOff,' PIU Fan Outlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + fpvzNameOff,' PIU Mixer Outlet')
    END IF
    CALL AddToObjFld('Air Outlet Node Name', base + fpvzNameOff,' Supply Inlet')
    CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
    CALL AddToObjStr('Rated Capacity','autosize')
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
   !BRANCH ~ line 163 of FPVAV tab
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + fpvzNameOff,' Reheat Coil HW Branch')
    !save this branch name for later use in plant
    CALL AddToStrList(TRIM(FldVal(base + fpvzNameOff)) // ' Reheat Coil HW Branch',handleHeatingCoilBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + fpvzNameOff,' Reheat Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + fpvzNameOff,' Reheat Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + fpvzNameOff,' Reheat Coil HW Outlet',.TRUE.)
  ELSEIF (reheatCoilType .EQ. ctElectric) THEN
    !COIL:Electric:Heating ~ line 163 of FPVAV tab
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjFld('Name', base + fpvzNameOff,' Reheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + fpvzReheatSchedNameOff,'')
    CALL AddToObjStr('Efficiency','1')
    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    IF (isFlowTypeSeries) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + fpvzNameOff,' PIU Fan Outlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + fpvzNameOff,' PIU Mixer Outlet')
    END IF
    CALL AddToObjFld('Air Outlet Node Name', base + fpvzNameOff,' Supply Inlet',.TRUE.)
  ELSEIF (reheatCoilType .EQ. ctGas) THEN
    !Coil:Heating:Fuel
    CALL CreateNewObj('Coil:Heating:Fuel')
    CALL AddToObjFld('Name', base + fpvzNameOff,' Reheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + fpvzReheatSchedNameOff,' ')
    CALL AddToObjStr('Fuel Type', 'Gas')
      CALL AddToObjStr('Gas Burner Efficiency','0.8')
      CALL AddToObjStr('Nominal Capacity {W}', 'autosize')
    IF (isFlowTypeSeries) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + fpvzNameOff,' PIU Fan Outlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + fpvzNameOff,' PIU Mixer Outlet')
    END IF
    CALL AddToObjFld('Air Outlet Node Name', base + fpvzNameOff,' Supply Inlet')
    CALL AddToObjStr('Temperature Setpoint Node Name','')
    CALL AddToObjStr('Parasitic Electric Load {W}', '0')
    CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
        base + fpvzNameOff,' Reheat Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
    !CURVE:CUBIC
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + fpvzNameOff,' Reheat Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  END IF
  !Baseboards
  SELECT CASE (baseboardKind)
    CASE (baseboardHotWater)
      !ZoneHVAC:Baseboard:RadiantConvective:Water
      CALL CreateNewObj('ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Name', base + fpvzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + fpvzBaseboardAvailSchedNameOff,'')
      CALL AddToObjFld('Inlet Node Name', base + fpvzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Outlet Node Name', base + fpvzNameOff,' Baseboard HW Outlet')
      CALL AddToObjStr('Rated Average Water Temperature {C}','87.78')
      CALL AddToObjStr('Rated Water Mass Flow Rate {m3/s}','0.063')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + fpvzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Maximum Water Flow Rate','autosize')
      CALL AddToObjStr('Convergence Tolerance','0.001')
      CALL AddToObjStr('Fraction Radiant','0.0')
      CALL AddToObjStr('Fraction of Radiant Energy Incident on People','0.0',.TRUE.)
      !BRANCH (for baseboard)
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + fpvzNameOff,' Baseboard Heat HW Branch')
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Component Name', base + fpvzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Component Inlet Node Name', base + fpvzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + fpvzNameOff,' Baseboard HW Outlet',.TRUE.)
      CALL AddToStrList(TRIM(FldVal(base + fpvzNameOff)) // ' Baseboard Heat HW Branch',handleHeatingCoilBranch)
      IF ((compactHotLoopBase .EQ. 0) .OR. (numCompactBoiler .EQ. 0)) THEN
        CALL WriteError('Both a HVACTemplate:Plant:HotWaterLoop and a HVACTemplate:Plant:Boiler are needed ' // &
             'when using hot water baseboards.')
      END IF
    CASE (baseboardElectric)
      !BASEBOARD HEATER:Electric:Convective
      CALL CreateNewObj('ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Name', base + fpvzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + fpvzBaseboardAvailSchedNameOff,'')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + fpvzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Efficiency','1',.TRUE.)
  END SELECT
END DO
!======================================================================================================
! create Heat and Cool VAV Zones
!======================================================================================================
DO iZone = 1, numCompactZoneHCVAV
  base =  compactZoneHCVAVbase(iZone)
  ! find base for system this zone belongs to
  sysBase = -1
  DO iSys = 1, numCompactSysVAV
    IF (SameString(FldVal(compactSysVAVbase(iSys) + vsAirHandlerNameOff), FldVal(base +  hcvzAirHandlerNameOff))) THEN
      sysBase = compactSysVAVBase(iSys)
      EXIT
    END IF
  END DO
  IF (sysBase .NE. -1) THEN
    isPackaged = .FALSE.
  ELSE
    DO iSys = 1, numCompactSysPVAV
      IF (SameString(FldVal(compactSysPVAVbase(iSys) + pvavsAirHandlerNameOff), FldVal(base +  hcvzAirHandlerNameOff))) THEN
        sysBase = compactSysPVAVBase(iSys)
        EXIT
      END IF
    END DO
    IF (sysBase .NE. -1) isPackaged = .TRUE.
  END IF
  IF (sysBase .EQ. -1) THEN
    CALL WriteError('Could not find air handler name referenced in HVACTemplate:Zone:VAV:HeatAndCool object: '// &
     TRIM(FldVal(base + hcvzNameOff)))
    CYCLE
  END IF
  ! set flags
  isTStatNameBlank = (FldVal(base + hcvzTStatNameOff) .EQ. '')
  isSupplyMaxRateAutosize = SameString(FldVal(base + hcvzSupplyMaxRateOff),'autosize')
  isOATypeDetailed = SameString(FldVal(base + hcvzOutAirTypeOff),'DetailedSpecification')
  !set reheat coil indicator
  IF (SameString(FldVal(base + hcvzReheatTypeOff),'HotWater')) THEN
    reheatCoilType = ctHotWater
  ELSEIF (SameString(FldVal(base + hcvzReheatTypeOff),'Electric')) THEN
    reheatCoilType = ctElectric
  ELSEIF (SameString(FldVal(base + hcvzReheatTypeOff),'Gas')) THEN
    reheatCoilType = ctGas
  ELSEIF (SameString(FldVal(base + hcvzReheatTypeOff),'None')) THEN
    reheatCoilType = ctNone
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VAV:HeatAndCool in the Reheat Coil Type field')
  END IF
  isZoneRetPlenBlank = (FldVal(base + hcvzZoneReturnPlenumNameOff) .EQ. '')
  isZoneSupPlenBlank = (FldVal(base + hcvzZoneSupplyPlenumNameOff) .EQ. '')
  !set the baseboard indicator
  IF (SameString(FldVal(base +  hcvzBaseboardTypeOff),'None')) THEN
    baseboardKind = baseboardNone
  ELSEIF (SameString(FldVal(base +  hcvzBaseboardTypeOff),'HotWater')) THEN
    baseboardKind = baseboardHotWater
  ELSEIF (SameString(FldVal(base +  hcvzBaseboardTypeOff),'Electric')) THEN
    baseboardKind = baseboardElectric
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VAV:HeatAndCool "'//TRIM(FldVal(base + hcvzNameOff))//'"'// &
                    ' in the Baseboard Heating Type field: '//TRIM(FldVal(base + hcvzBaseboardTypeOff)))
  END IF
  !validate cooling design method
  IF (SameString(FldVal(base +  hcvzCoolDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  hcvzCoolDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  hcvzCoolDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VAV:HeatAndCool "'//TRIM(FldVal(base + hcvzNameOff))//'"'// &
                    ' in the Zone Cooling Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + hcvzCoolDesignTempMethodOff)))
  END IF
  !validate heating design method
  IF (SameString(FldVal(base +  hcvzHeatDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  hcvzHeatDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  hcvzHeatDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VAV:HeatAndCool "'//TRIM(FldVal(base + hcvzNameOff))//'"'// &
                    ' in the Zone Heating Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + hcvzCoolDesignTempMethodOff)))
  END IF
  !ZONE CONTROL:THERMOSTATIC ~ line 5
  !repeat per zone - only if hcvzTStatNameOff is not blank
  IF (.NOT. isTStatNameBlank) THEN
    CALL AddAlwaysSchedule('4')
    CALL CreateNewObj('ZoneControl:Thermostat')
    CALL AddToObjFld('Name', base +  hcvzNameOff,' Thermostat')
    CALL AddToObjFld('Zone or ZoneList Name', base +  hcvzNameOff,' ')
    !this schedule was previously created
    CALL AddToObjStr('Control Type Schedule Name','HVACTemplate-Always 4')
    CALL AddToObjStr('Control Object Type','ThermostatSetpoint:DualSetpoint')
    CALL AddToObjFld('Control Name', base + hcvzTStatNameOff,' Dual SP Control',.TRUE.)
  END IF
  !ZONE SIZING   ~ line 12
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:OutdoorAir')
    CALL AddToObjStr('Name', 'SZ DSOA '//FldVal(base+hcvzNameOff))
    CALL AddToObjFld('Outdoor Air Method', base + hcvzOutAirTypeOff,' ')
    CALL AddToObjFld('Outdoor Air Flow per Person {m3/s}', base + hcvzOutAirFlowPersonOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone Floor Area {m3/s-m2}', base + hcvzOutAirFlowAreaOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone {m3/s}', base + hcvzOutAirFlowZoneOff,'',.true.)
  END IF

  CALL CreateNewObj('Sizing:Zone')
  CALL AddToObjFld('Zone or ZoneList Name', base +  hcvzNameOff,'')
  IF (SameString(FldVal(base +  hcvzCoolDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
    CALL AddToObjStr('Zone Cooling Design Supply Air Temperature Input Method','SupplyAirTemperature')
    IF (isPackaged) THEN
      CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', sysBase + pvavsCoolSetPtDesignOff,'')
    ELSE
      CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', sysBase + vsCoolSetPtDesignOff,'')
    ENDIF
  ELSE
    CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Input Method',base + hcvzCoolDesignTempMethodOff,'')
    CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', base + hcvzCoolDesignTempOff,'')
  ENDIF
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Difference {delta C}',base + hcvzCoolDesignTempDiffOff,'')
  IF (SameString(FldVal(base +  hcvzHeatDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
    CALL AddToObjStr('Zone Heating Design Supply Air Temperature Input Method','SupplyAirTemperature')
    IF (isPackaged) THEN
      CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', sysBase + pvavsHeatSetPtDesignOff,'')
    ELSE
      CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', sysBase + vsHeatSetPtDesignOff,'')
    ENDIF
  ELSE
    CALL AddToObjFld('Zone Heating Design Supply Air Temperature Input Method',base + hcvzHeatDesignTempMethodOff,'')
    CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', base + hcvzHeatDesignTempOff,'')
  END IF
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Difference {delta C}',base + hcvzHeatDesignTempDiffOff,'')
  CALL AddToObjStr('Zone Cooling Design Supply Air Humidity Ratio {kg-H20/kg-air}','0.008')
  CALL AddToObjStr('Zone Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + hcvzDesignSpecOANameOff,'')
  ELSE
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','SZ DSOA '//FldVal(base+hcvzNameOff))
  END IF
  CALL AddToObjFld('Zone Heating Sizing Factor', base + hcvzZoneHeatSizingFactorOff,'')
  CALL AddToObjFld('Zone Cooling Sizing Factor', base + hcvzZoneCoolSizingFactorOff,'')
  IF (isSupplyMaxRateAutosize) THEN
    CALL AddToObjStr('Cooling Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Cooling Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  ELSE
    CALL AddToObjStr('Cooling Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Cooling Design Air Flow Rate {m3/s}', base + hcvzSupplyMaxRateOff,' ')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  END IF
  IF (isSupplyMaxRateAutosize) THEN
    CALL AddToObjStr('Heating Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Heating Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
  ELSE
    CALL AddToObjStr('Heating Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Heating Design Air Flow Rate {m3/s}', base + hcvzSupplyMaxRateOff,'')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
  END IF
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Zone Air Distribution Object Name', base + hcvzDesSpecOADistNameOff,'',.true.)
  ELSE
    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+hcvzNameOff),.true.)
  END IF
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:ZoneAirDistribution')
    CALL AddToObjStr('Name','SZ DSZAD '//FldVal(base+hcvzNameOff))
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Cooling Mode {}','1')
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Heating Mode {}','1',.true.)
  END IF
  !CONTROLLED ZONE EQUIP CONFIGURATION  ~ line 41
  CALL CreateNewObj('ZoneHVAC:EquipmentConnections')
  CALL AddToObjFld('Zone Name', base +  hcvzNameOff,'')
  CALL AddToObjFld('Zone Conditioning Equipment List Name', base +  hcvzNameOff,' Equipment')
  CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base +  hcvzNameOff,' Supply Inlet')
  CALL AddToObjStr('Zone Air Exhaust Node or NodeList Name','')
  CALL AddToObjFld('Zone Air Node Name', base +  hcvzNameOff,' Zone Air Node')
  IF (isZoneRetPlenBlank) THEN
    CALL AddToObjFld('Zone Return Air Node Name', base +  hcvzNameOff,' Return Outlet',.TRUE.)
  ELSE
    CALL AddToObjFld('Zone Return Air Node Name', base + hcvzNameOff,' Zone Return Plenum Inlet',.TRUE.)
  END IF
  !ZONE SUPPLY PLENUM ~line 50
  IF (.NOT. isZoneSupPlenBlank) THEN
    CALL CreateNewObj('AirLoopHVAC:SupplyPlenum')
    CALL AddToObjFld('Name', base + hcvzNameOff,' Supply Plenum')
    CALL AddToObjFld('Zone Name', base + hcvzZoneSupplyPlenumNameOff,'')
    CALL AddToObjFld('Zone Node Name', base + hcvzZoneSupplyPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Inlet Node Name', base + hcvzNameOff,' Zone Equip Inlet')
    CALL AddToObjFld('Outlet Node Name', base + hcvzNameOff,' Terminal Unit Inlet',.TRUE.)
  END IF
  !ZONE RETURN PLENUM ~ line 57
  IF (.NOT. isZoneRetPlenBlank) THEN
    CALL CreateNewObj('AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Name', base + hcvzNameOff,' Return Plenum')
    CALL AddToObjFld('Zone Name', base + hcvzZoneReturnPlenumNameOff,'')
    CALL AddToObjFld('Zone Node Name', base + hcvzZoneReturnPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Outlet Node Name', base + hcvzNameOff,' Return Outlet')
    CALL AddToObjStr('Induced Air Outlet Node or NodeList Name','')
    CALL AddToObjFld('Inlet Node Name', base + hcvzNameOff,' Zone Return Plenum Inlet',.TRUE.)
  END IF
  !ZONE EQUIPMENT LIST ~ line 64
  CALL CreateNewObj('ZoneHVAC:EquipmentList')
  CALL AddToObjFld('Name', base +  hcvzNameOff,' Equipment')
  CALL AddToObjStr('Load Distribution Scheme','SequentialLoad')
  CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:AirDistributionUnit')
  CALL AddToObjFld('Zone Equipment Name', base +  hcvzNameOff,' ATU')
  CALL AddToObjStr('Zone Equipment Cooling Sequence','1')
  SELECT CASE (baseboardKind)
    CASE (baseboardNone)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1',.FALSE.)
    CASE (baseboardHotWater)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1')
      CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
      CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.FALSE.)
      CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Zone Equipment Name', base + hcvzNameOff,' Baseboard Heat')
      CALL AddToObjStr('Zone Equipment Cooling Sequence','2')
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','2',.FALSE.)
    CASE (baseboardElectric)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1')
      CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Zone Equipment Name', base + hcvzNameOff,' Baseboard Heat')
      CALL AddToObjStr('Zone Equipment Cooling Sequence','2')
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','2',.FALSE.)
  END SELECT
  CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.TRUE.)
  !AIR DISTRIBUTION UNIT ~ line 72
  IF (reheatCoilType .EQ. ctNone) THEN
    CALL CreateNewObj('ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Name', base +  hcvzNameOff,' ATU')
    CALL AddToObjFld('Air Distribution Unit Outlet Node Name', base +  hcvzNameOff,' Supply Inlet')
    CALL AddToObjStr('Air Terminal Object Type','AirTerminal:SingleDuct:VAV:HeatAndCool:NoReheat')
    CALL AddToObjFld('Air Terminal Name', base +  hcvzNameOff,' VAV Heat Cool',.TRUE.)
  ELSE
    CALL CreateNewObj('ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Name', base +  hcvzNameOff,' ATU')
    CALL AddToObjFld('Air Distribution Unit Outlet Node Name', base +  hcvzNameOff,' Supply Inlet')
    CALL AddToObjStr('Air Terminal Object Type','AirTerminal:SingleDuct:VAV:HeatAndCool:Reheat')
    CALL AddToObjFld('Air Terminal Name', base +  hcvzNameOff,' VAV Heat Cool Reheat',.TRUE.)
  END IF
  !This object if hcvzReheatTypeOff=Hot Water
  IF (reheatCoilType .EQ. ctHotWater) THEN
    !SINGLE DUCT:VAV:REHEAT   ~ line 85
    CALL CreateNewObj('AirTerminal:SingleDuct:VAV:HeatAndCool:Reheat')
    CALL AddToObjFld('Name', base +  hcvzNameOff,' VAV Heat Cool Reheat')
    !this schedule was previously created
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Damper Air Outlet Node Name', base + hcvzNameOff,' Damper Outlet')
    IF (isZoneSupPlenBlank) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + hcvzNameOff,' Zone Equip Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + hcvzNameOff,' Terminal Unit Inlet')
    END IF
    !CR8001
    !CALL AddToObjFld('Maximum air flow rate {m3/s} ', base + hcvzSupplyMaxRateOff,' ')
    CALL AddToObjStr('Maximum Air Flow Rate {m3/s} ', 'autosize')
    CALL AddToObjFld('Zone Minimum Air Flow Fraction', base + hcvzConstantMinFracOff,' ')
    CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Reheat Coil Name', base + hcvzNameOff,' Reheat Coil')
    CALL AddToObjStr('Maximum Hot Water or Steam Flow Rate {m3/s}  ','autosize')
    CALL AddToObjStr('Minimum Hot Water or Steam Flow Rate {m3/s}  ','0')
    CALL AddToObjFld('Air Outlet Node Name', base + hcvzNameOff,' Supply Inlet')
    CALL AddToObjStr('Convergence Tolerance','0.001')
    CALL AddToObjFld('Maximum Reheat Air Temperature', base + hcvzMaxReheatAirTempOff,' ',.TRUE.)
    !COIL:Water:SimpleHeating  ~ line 104
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjFld('Name', base + hcvzNameOff,' Reheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + hcvzReheatSchedNameOff,' ')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name', base + hcvzNameOff,' Reheat Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + hcvzNameOff,' Reheat Coil HW Outlet')
    CALL AddToObjFld('Air Inlet Node Name', base + hcvzNameOff,' Damper Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + hcvzNameOff,' Supply Inlet')
    CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
    CALL AddToObjStr('Rated Capacity','autosize')
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
    !BRANCH ~ line 113
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + hcvzNameOff,' Reheat Coil HW Branch')
    !save this branch name for later use in plant
    CALL AddToStrList(TRIM(FldVal(base + hcvzNameOff)) // ' Reheat Coil HW Branch',handleHeatingCoilBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + hcvzNameOff,' Reheat Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + hcvzNameOff,' Reheat Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + hcvzNameOff,' Reheat Coil HW Outlet',.TRUE.)
  ELSEIF (reheatCoilType .EQ. ctElectric) THEN
    !SINGLE DUCT:VAV:REHEAT ~ line 122
    CALL CreateNewObj('AirTerminal:SingleDuct:VAV:HeatAndCool:Reheat')
    CALL AddToObjFld('Name', base +  hcvzNameOff,' VAV Heat Cool Reheat')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Damper Air Outlet Node Name', base + hcvzNameOff,' Damper Outlet')
    IF (isZoneSupPlenBlank) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + hcvzNameOff,' Zone Equip Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + hcvzNameOff,' Terminal Unit Inlet')
    END IF
    !CR8001
    !CALL AddToObjFld('Maximum air flow rate {m3/s}', base + hcvzSupplyMaxRateOff,' ')
    CALL AddToObjStr('Maximum Air Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Zone Minimum Air Flow Fraction', base + hcvzConstantMinFracOff,' ')
    CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Electric')
    CALL AddToObjFld('Reheat Coil Name', base + hcvzNameOff,' Reheat Coil')
    CALL AddToObjStr('Maximum Hot Water or Steam Flow Rate {m3/s}','')
    CALL AddToObjStr('Minimum Hot Water or Steam Flow Rate {m3/s}','')
    CALL AddToObjFld('Air Outlet Node Name', base + hcvzNameOff,' Supply Inlet')
    CALL AddToObjStr('Convergence Tolerance','0.001')
    CALL AddToObjFld('Maximum Reheat Air Temperature', base + hcvzMaxReheatAirTempOff,' ',.TRUE.)
    !COIL:Electric:Heating ~ line 139
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjFld('Name', base + hcvzNameOff,' Reheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + hcvzReheatSchedNameOff,' ')
    CALL AddToObjStr('Efficiency','1')
    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    CALL AddToObjFld('Air Inlet Node Name', base + hcvzNameOff,' Damper Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + hcvzNameOff,' Supply Inlet',.TRUE.)
  ELSEIF (reheatCoilType .EQ. ctGas) THEN
    !SINGLE DUCT:VAV:REHEAT ~ line 122
    CALL CreateNewObj('AirTerminal:SingleDuct:VAV:HeatAndCool:Reheat')
    CALL AddToObjFld('Name', base +  hcvzNameOff,' VAV Heat Cool Reheat')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Damper Air Outlet Node Name', base + hcvzNameOff,' Damper Outlet')
    IF (isZoneSupPlenBlank) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + hcvzNameOff,' Zone Equip Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + hcvzNameOff,' Terminal Unit Inlet')
    END IF
    !CR8001
    CALL AddToObjStr('Maximum Air Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Zone Minimum Air Flow Fraction', base + hcvzConstantMinFracOff,' ')
    CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Fuel')
    CALL AddToObjFld('Reheat Coil Name', base + hcvzNameOff,' Reheat Coil')
    CALL AddToObjStr('Maximum Hot Water or Steam Flow Rate {m3/s}','')
    CALL AddToObjStr('Minimum Hot Water or Steam Flow Rate {m3/s}','')
    CALL AddToObjFld('Air Outlet Node Name', base + hcvzNameOff,' Supply Inlet')
    CALL AddToObjStr('Convergence Tolerance','0.001')
    CALL AddToObjFld('Maximum Reheat Air Temperature', base + hcvzMaxReheatAirTempOff,' ',.TRUE.)
    !Coil:Heating:Fuel
    CALL CreateNewObj('Coil:Heating:Fuel')
    CALL AddToObjFld('Name', base + hcvzNameOff,' Reheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + hcvzReheatSchedNameOff,' ')
      CALL AddToObjStr('Fuel Type', 'Gas')
      CALL AddToObjStr('Gas Burner Efficiency','0.8')
      CALL AddToObjStr('Nominal Capacity {W}', 'autosize')
    CALL AddToObjFld('Air Inlet Node Name', base + hcvzNameOff,' Damper Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + hcvzNameOff,' Supply Inlet')
    CALL AddToObjStr('Temperature Setpoint Node Name','')
    CALL AddToObjStr('Parasitic Electric Load {W}', '0')
    CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
        base + hcvzNameOff,' Reheat Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
    !CURVE:CUBIC
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + hcvzNameOff,' Reheat Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  ELSE
    !SINGLE DUCT:VAV:NOREHEAT ~ line 148
    CALL CreateNewObj('AirTerminal:SingleDuct:VAV:HeatAndCool:NoReheat')
    CALL AddToObjFld('Name', base +  hcvzNameOff,' VAV Heat Cool')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Air Outlet Node Name', base + hcvzNameOff,' Supply Inlet')
    IF (isZoneSupPlenBlank) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + hcvzNameOff,' Zone Equip Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + hcvzNameOff,' Terminal Unit Inlet')
    END IF
    !CR8001
    !CALL AddToObjFld('Maximum air flow rate {m3/s}', base + hcvzSupplyMaxRateOff,' ')
    CALL AddToObjStr('Maximum Air Flow Rate {m3/s}', 'autosize ')
    CALL AddToObjFld('Zone Minimum Air Flow Fraction', base + hcvzConstantMinFracOff,' ',.TRUE.)
  END IF
  !Baseboards
  SELECT CASE (baseboardKind)
    CASE (baseboardHotWater)
      !ZoneHVAC:Baseboard:RadiantConvective:Water
      CALL CreateNewObj('ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Name', base + hcvzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + hcvzBaseboardAvailSchedNameOff,'')
      CALL AddToObjFld('Inlet Node Name', base + hcvzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Outlet Node Name', base + hcvzNameOff,' Baseboard HW Outlet')
      CALL AddToObjStr('Rated Average Water Temperature {C}','87.78')
      CALL AddToObjStr('Rated Water Mass Flow Rate {m3/s}','0.063')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + hcvzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Maximum Water Flow Rate','autosize')
      CALL AddToObjStr('Convergence Tolerance','0.001')
      CALL AddToObjStr('Fraction Radiant','0.0')
      CALL AddToObjStr('Fraction of Radiant Energy Incident on People','0.0',.TRUE.)
      !BRANCH (for baseboard)
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + hcvzNameOff,' Baseboard Heat HW Branch')
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Component Name', base + hcvzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Component Inlet Node Name', base + hcvzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + hcvzNameOff,' Baseboard HW Outlet',.TRUE.)
      CALL AddToStrList(TRIM(FldVal(base + hcvzNameOff)) // ' Baseboard Heat HW Branch',handleHeatingCoilBranch)
      IF ((compactHotLoopBase .EQ. 0) .OR. (numCompactBoiler .EQ. 0)) THEN
        CALL WriteError('Both a HVACTemplate:Plant:HotWaterLoop and a HVACTemplate:Plant:Boiler are needed ' // &
             'when using hot water baseboards.')
      END IF
    CASE (baseboardElectric)
      !BASEBOARD HEATER:Electric:Convective
      CALL CreateNewObj('ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Name', base + hcvzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + hcvzBaseboardAvailSchedNameOff,'')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + hcvzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Efficiency','1',.TRUE.)
  END SELECT
END DO
!======================================================================================================
! create Systems (not packaged)
!======================================================================================================
DO iSys = 1, numCompactSysVAV
  base =  compactSysVAVbase(iSys)
  ! set flags
  isSupplyMaxRateAutosize = SameString(FldVal(base + vsSupplyMaxRateOff),'autosize')
  isSupplyMinRateAutosize = SameString(FldVal(base + vsSupplyMinRateOff),'autosize')
  isMinOARateAutosize = SameString(FldVal(base + vsMinOutsideFlowOff),'autosize')
  isSupplyPlenumBlank = (FldVal(base + vsSupplyPlenumNameOff) .EQ. '')
  isReturnPlenumBlank = (FldVal(base + vsReturnPlenumNameOff) .EQ. '')
  isSysAvailSchedBlank = (FldVal(base + vsSysAvailSchedNameOff) .EQ. '')
  isCoolStPtSchedBlank = (FldVal(base + vsCoolSetPtSchedNameOff) .EQ. '')
  isHeatStPtSchedBlank = (FldVal(base + vsHeatSetPtSchedNameOff) .EQ. '')
  isPreheatStPtSchedBlank = (FldVal(base + vsPreheatSetPtSchedNameOff) .EQ. '')
  isEconoLowLimitBlank = (FldVal(base + vsEconoLowLimitOff) .EQ. '')
  isReturnFanYes = (FldVal(base + vsReturnFanPresentOff) .EQ. 'Yes')

  doSuppressMsgzlVavAll = .FALSE.
  !check cooling coil kind
  IF (SameString(FldVal(base +  vsCoolCoilKindOff),'ChilledWater')) THEN
    coolCoilKind = ccChWater
  ELSEIF (SameString(FldVal(base + vsCoolCoilKindOff),'ChilledWaterDetailedFlatModel')) THEN
    coolCoilKind = ccChWaterDetailed
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:VAV "'//TRIM(FldVal(base + vsAirHandlerNameOff))//'"'// &
                    ' in the Cooling Coil Type field: '//TRIM(FldVal(base + vsCoolCoilKindOff)))
  END IF
  !set heating coil indicator
  IF (SameString(FldVal(base + vsHeatCoilKindOff),'HotWater')) THEN
    heatCoilType = ctHotWater
  ELSEIF (SameString(FldVal(base + vsHeatCoilKindOff),'Electric')) THEN
    heatCoilType = ctElectric
  ELSEIF (SameString(FldVal(base + vsHeatCoilKindOff),'Gas')) THEN
    heatCoilType = ctGas
  ELSEIF (SameString(FldVal(base + vsHeatCoilKindOff),'None')) THEN
    heatCoilType = ctNone
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:VAV in the Heating Coil Type field')
  END IF
  !set preheating coil indicator
  IF (SameString(FldVal(base +  vsPreheatCoilKindOff),'HotWater')) THEN
    preHeatCoilType = ctHotWater
  ELSEIF (SameString(FldVal(base +  vsPreheatCoilKindOff),'Electric')) THEN
    preHeatCoilType = ctElectric
  ELSEIF (SameString(FldVal(base +  vsPreheatCoilKindOff),'Gas')) THEN
    preHeatCoilType = ctGas
  ELSEIF (SameString(FldVal(base +  vsPreheatCoilKindOff),'None')) THEN
    preHeatCoilType = ctNone
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:VAV in the Preheat Coil Type field')
  END IF
  !validate the Minimum Outside Air Control Type
  CALL validChoice(base + vsMinOutsideControlOff,'FixedMinimum','ProportionalMinimum',  &
     'SYSTEM:VAV','Minimum Outdoor Air Control Type')
  !set the economizer indicator
  IF (SameString(FldVal(base +  vsEconomizerKindOff),'FixedDryBulb')) THEN
    econoKind = etFixedDryBulb
  ELSEIF (SameString(FldVal(base +  vsEconomizerKindOff),'FixedEnthalpy')) THEN
    econoKind = etFixedEnthalpy
  ELSEIF (SameString(FldVal(base +  vsEconomizerKindOff),'DifferentialDryBulb')) THEN
    econoKind = etDifferentialDryBulb
  ELSEIF (SameString(FldVal(base +  vsEconomizerKindOff),'DifferentialEnthalpy')) THEN
    econoKind = etDifferentialEnthalpy
  ELSEIF (SameString(FldVal(base +  vsEconomizerKindOff),'FixedDewPointAndDryBulb')) THEN
    econoKind = etFixedDewPointAndDryBulb
  ELSEIF (SameString(FldVal(base +  vsEconomizerKindOff),'ElectronicEnthalpy')) THEN
    econoKind = etElectronicEnthalpy
  ELSEIF (SameString(FldVal(base +  vsEconomizerKindOff),'DifferentialDryBulbAndEnthalpy')) THEN
    econoKind = etDifferentialDryBulbAndEnthalpy
  ELSEIF (SameString(FldVal(base +  vsEconomizerKindOff),'NoEconomizer')) THEN
    econoKind = etNoEconomizer
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:VAV in the Economizer Type field')
  END IF
  !set the economizer lockout indicator
  IF (SameString(FldVal(base +  vsEconoLockoutKindOff),'NoLockout')) THEN
    lockoutKind = loNo
  ELSEIF (SameString(FldVal(base +  vsEconoLockoutKindOff),'')) THEN
    lockoutKind = loNo
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:VAV in the Economizer Lockout field')
  END IF
  !set the CoolSetPtResetType indicator
  IF (SameString(FldVal(base + vsCoolSetPtResetTypeOff),'None')) THEN
    coolSetPtReset = csprNone
  ELSEIF (SameString(FldVal(base + vsCoolSetPtResetTypeOff),'Warmest')) THEN
    coolSetPtReset = csprWarmest
  ELSEIF (SameString(FldVal(base + vsCoolSetPtResetTypeOff),'OutdoorAirTemperatureReset')) THEN
    coolSetPtReset = csprOAreset
  ELSEIF (SameString(FldVal(base + vsCoolSetPtResetTypeOff),'WarmestTemperatureFirst')) THEN
    coolSetPtReset = csprWarmestTempFirst
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:VAV in the Cooling Coil Setpoint Reset Type field')
    coolSetPtReset = csprNone
  END IF
  !set the HeatSetPtResetType indicator
  IF (SameString(FldVal(base + vsHeatSetPtResetTypeOff),'None')) THEN
    heatSetPtReset = hsprNone
  ELSEIF (SameString(FldVal(base + vsHeatSetPtResetTypeOff),'OutdoorAirTemperatureReset')) THEN
    heatSetPtReset = hsprOAreset
!  This setpoint manager type not yet available in EnergyPlus, leave code here for future use
!  ELSEIF (SameString(FldVal(base + vsHeatSetPtResetTypeOff),'ColdestTemperatureFirst')) THEN
!    coolSetPtReset = hsprColdestTempFirst
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:VAV in the Heating Coil Setpoint Reset Type field')
    heatSetPtReset = hsprNone
  END IF
  !set the Supply Fan Part-Load Power Coefficients indicator
  IF (SameString(FldVal(base + vsSupplyFanPLRCoeffOff),'InletVaneDampers')) THEN
    supFanCoef = sfcInletVane
  ELSEIF (SameString(FldVal(base + vsSupplyFanPLRCoeffOff),'OutletDampers')) THEN
    supFanCoef = sfcDischgDamp
  ELSEIF (SameString(FldVal(base + vsSupplyFanPLRCoeffOff),'VariableSpeedMotor')) THEN
    supFanCoef = sfcVSD
  ELSEIF (SameString(FldVal(base + vsSupplyFanPLRCoeffOff),'ASHRAE90.1-2004AppendixG')) THEN
    supFanCoef = sfcAppG
  ELSEIF (SameString(FldVal(base + vsSupplyFanPLRCoeffOff),'VariableSpeedMotorPressureReset')) THEN
    supFanCoef = sfcVSDPresReset
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:VAV in the Supply Fan Part-Load Power Coefficients field')
    supFanCoef = sfcInletVane
  END IF
  !set the Return Fan Part-Load Power Coefficients indicator
  IF (SameString(FldVal(base + vsReturnFanPLRCoeffOff),'InletVaneDampers')) THEN
    retFanCoef = sfcInletVane
  ELSEIF (SameString(FldVal(base + vsReturnFanPLRCoeffOff),'OutletDampers')) THEN
    retFanCoef = sfcDischgDamp
  ELSEIF (SameString(FldVal(base + vsReturnFanPLRCoeffOff),'VariableSpeedMotor')) THEN
    retFanCoef = sfcVSD
  ELSEIF (SameString(FldVal(base + vsReturnFanPLRCoeffOff),'ASHRAE90.1-2004AppendixG')) THEN
    retFanCoef = sfcAppG
  ELSEIF (SameString(FldVal(base + vsReturnFanPLRCoeffOff),'VariableSpeedMotorPressureReset')) THEN
    retFanCoef = sfcVSDPresReset
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:VAV in the Return Fan Part-Load Power Coefficients field')
    retFanCoef = sfcInletVane
  END IF
  !set the Heat Recovery Type indicator
  IF (SameString(FldVal(base + vsHeatRecTypeOff),'Sensible')) THEN
    heatRecovery = htrecSens
  ELSEIF (SameString(FldVal(base + vsHeatRecTypeOff),'Enthalpy')) THEN
    heatRecovery = htrecEnth
  ELSEIF (SameString(FldVal(base + vsHeatRecTypeOff),'None')) THEN
    heatRecovery = htrecNone
  ELSE
    heatRecovery = htrecNone
    CALL WriteError('Invalid choice in HVACTemplate:System:VAV in the Heating Coil Setpoint Reset Type field')
  ENDIF
  IF (SameString(FldVal(base + vsSupplyFanPlaceOff), 'BlowThrough')) THEN
    supFanPlacement = sfpBlowThru
  ELSEIF (SameString(FldVal(base + vsSupplyFanPlaceOff), 'DrawThrough')) THEN
    supFanPlacement = sfpDrawThru
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:VAV "'//TRIM(FldVal(base + vsAirHandlerNameOff))//'"'// &
                    ' in the Supply Fan Placement field: '//TRIM(FldVal(base + vsSupplyFanPlaceOff)))
  END IF
  !set the dehumidification indicators
  isDehumidifyNone = SameString(FldVal(base + vsDehumCtrlTypeOff),'None')
  IF (SameString(FldVal(base +  vsDehumCtrlTypeOff),'None')) THEN
    dehumidCtrlKind = dehumidNone
  ELSEIF (SameString(FldVal(base +  vsDehumCtrlTypeOff),'CoolReheat')) THEN
    dehumidCtrlKind = dehumidCoolRht
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:VAV "'//TRIM(FldVal(base + vsAirHandlerNameOff))//'"'// &
                    ' in the Dehumidification Control Type field: '//TRIM(FldVal(base + vsDehumCtrlTypeOff)))
  END IF
  !set the humidifier indicators
  isHumidifierNone = SameString(FldVal(base + vsHumidCtrlTypeOff),'None')
  IF (SameString(FldVal(base +  vsHumidCtrlTypeOff),'None')) THEN
    humidifierKind = humidifyNone
  ELSEIF (SameString(FldVal(base +  vsHumidCtrlTypeOff),'ElectricSteam')) THEN
    humidifierKind = humidifyElecSteam
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:VAV "'//TRIM(FldVal(base + vsAirHandlerNameOff))//'"'// &
                    ' in the Humidifier Type field: '//TRIM(FldVal(base + vsHumidCtrlTypeOff)))
  END IF

  !set node name for return inlet to OA mixer
  IF (isReturnFanYes) THEN
    returnInletToOAMIxer = TRIM(FldVal(base + vsAirHandlerNameOff))//' Return Fan Outlet'
  ELSE
    returnInletToOAMIxer = TRIM(FldVal(base + vsAirHandlerNameOff))//' Air Loop Inlet'
  END IF
  !Fix for CR7356
  ! Warn if fixed heating or preheat setpoint is greater than fixed cooling setpoint, and no reset option is active
  IF (isCoolStPtSchedBlank .AND. (coolSetPtReset == csprNone)) THEN
    IF ((heatCoilType .NE. ctNone) .AND. isHeatStPtSchedBlank .AND. (heatSetPtReset == hsprNone)) THEN
      IF (StringToReal(FldVal(base + vsHeatSetPtDesignOff)) .GT. StringToReal(FldVal(base + vsCoolSetPtDesignOff))) THEN
       CALL WriteError('Warning:  In HVACTemplate:System:VAV "'//TRIM(FldVal(base + vsAirHandlerNameOff))//'"'// &
                      ' the Heating Coil Design Setpoint is greater than the Cooling Coil Design Setpoint.'// &
                      ' This may cause the heating coil and cooling coil to operate simultaneously.'// &
                      ' Check results carefully and adjust controls if needed.',msgWarning)
      ENDIF
    ENDIF
    IF ((preHeatCoilType .NE. ctNone) .AND. isPreheatStPtSchedBlank) THEN
      IF (StringToReal(FldVal(base + vsPreheatSetPtConstantOff)) .GT. StringToReal(FldVal(base + vsCoolSetPtDesignOff))) THEN
       CALL WriteError('Warning:  In HVACTemplate:System:VAV "'//TRIM(FldVal(base + vsAirHandlerNameOff))//'"'// &
                      ' the Preheat Coil Design Setpoint is greater than the Cooling Coil Design Setpoint.'// &
                      ' This may cause the preheat coil and cooling coil to operate simultaneously.'// &
                      ' Check results carefully and adjust controls if needed.',msgWarning)
      ENDIF
    ENDIF
  ENDIF
  ! check if there is no heating coil and heating design setpoint is > preheat design setpoint
  IF (heatCoilType == ctNone) THEN
    IF(StringToReal(FldVal(base + vsHeatSetPtDesignOff)) .GT. StringToReal(FldVal(base + vsPreheatSetPtConstantOff))) THEN
      usePreheatTempforHeatDesignSuppT = .TRUE.
      CALL WriteError('Warning:  In HVACTemplate:System:VAV "'//TRIM(FldVal(base + vsAirHandlerNameOff))//'"'// &
                    ' the Heating Coil Design Setpoint is greater than the Preheat Coil Design Setpoint,'// &
                    ' but Heating Coil Type=None. Using Preheat Coil Design Setpoint for the Sizing:System'// &
                    ' Central Heating Design Supply Air Temperature.',msgWarning)
    ENDIF
    IF (preHeatCoilType == ctNone) THEN
      usePreheatTempforHeatDesignSuppT = .TRUE.
      CALL WriteError('Warning:  In HVACTemplate:System:VAV "'//TRIM(FldVal(base + vsAirHandlerNameOff))//'"'// &
                    ' there is no Heating Coil and no Preheat Coil.'// &
                    ' The Preheat Coil Design Setpoint will be used for the Sizing:System'// &
                    ' Central Heating Design Supply Air Temperature. This will be the inlet air temperature for sizing'// &
                    ' reheat coils.',msgWarning)
    ENDIF
  ENDIF


  ! validate sizing option input
  IF (SameString(FldVal(base +  vsSizingOptionOff),'Coincident')) THEN
  ELSEIF (SameString(FldVal(base +  vsSizingOptionOff),'NonCoincident')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:VAV "'//TRIM(FldVal(base + vsAirHandlerNameOff))//'"'// &
                    ' in the Sizing Option field: '//TRIM(FldVal(base + vsSizingOptionOff)))
  END IF
  !SYSTEM SIZING  ~ line 158
  CALL CreateNewObj('Sizing:System')
  CALL AddToObjFld('AirLoop Name', base + vsAirHandlerNameOff,' ')
  CALL AddToObjStr('Type of Load to Size On','Sensible')
  CALL AddToObjFld('Design Outdoor Air Flow Rate {m3/s}', base + vsMinOutsideFlowOff,' ')
  !This line if vsSupplyMaxRateOff=autosize or vsSupplyMinRateOff=autosize
  IF (isSupplyMaxRateAutosize .OR. isSupplyMinRateAutosize) THEN
    CALL AddToObjStr('Central Heating Maximum System Air Flow Ratio','0.3')
  ELSE
    minAirRatio = StringToReal(FldVal(base + vsSupplyMinRateOff)) / &
                  StringToReal(FldVal(base + vsSupplyMaxRateOff))
    CALL AddToObjNum('Central Heating Maximum System Air Flow Ratio', minAirRatio)
  END IF
  CALL AddToObjFld('Preheat Design Temperature {C}', base + vsPreheatSetPtConstantOff,'')
  CALL AddToObjStr('Preheat Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjStr('Precool Design Temperature {C}','11')
  CALL AddToObjStr('Precool Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjFld('Central Cooling Design Supply Air Temperature {C}', base + vsCoolSetPtDesignOff,'')
  IF (usePreheatTempforHeatDesignSuppT) THEN
    CALL AddToObjFld('Central Heating Design Supply Air Temperature {C}', base + vsPreheatSetPtConstantOff,'')
  ELSE
    CALL AddToObjFld('Central Heating Design Supply Air Temperature {C}', base + vsHeatSetPtDesignOff,'')
  ENDIF
  CALL AddToObjFld('Type of Zone Sum to Use', base + vsSizingOptionOff,'')
  CALL AddToObjStr('100% Outdoor Air in Cooling','No')
  CALL AddToObjStr('100% Outdoor Air in Heating','No')
  CALL AddToObjStr('Central Cooling Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjStr('Central Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isSupplyMaxRateAutosize) THEN
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','DesignDay')
    CALL AddToObjStr('Cooling Supply Air Flow Rate {m3/s}','0')
  ELSE
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','Flow/System')
    CALL AddToObjFld('Cooling Supply Air Flow Rate {m3/s}', base + vsSupplyMaxRateOff,' ')
  END IF
  ! New fields for v8.2 next three rows
  CALL AddToObjStr('Cooling Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
  CALL AddToObjStr('Cooling Fraction of Autosized Cooling Supply Air Flow Rate','')
  CALL AddToObjStr('Cooling Supply Air Flow Rate Per Unit Cooling Capacity {m3/s-W}','')
  CALL AddToObjStr('Heating Supply Air Flow Rate Method','DesignDay')
  CALL AddToObjStr('Heating Supply Air Flow Rate {m3/s}','0')
  ! New fields for v8.2 next four rows
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
  CALL AddToObjStr('Heating Fraction of Autosized Heating Supply Air Flow Rate','')
  CALL AddToObjStr('Heating Fraction of Autosized Cooling Supply Air Flow Rate','')
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Unit Heating Capacity {m3/s-W}','')
  CALL AddToObjStr('System Outdoor Air Method','ZoneSum')
  CALL AddToObjStr('Zone Maximum Outdoor Air Fraction','1.0')
  CALL AddToObjStr('Cooling Design Capacity Method','CoolingDesignCapacity')
  CALL AddToObjStr('Cooling Design Capacity {W}','autosize')
  CALL AddToObjStr('Cooling Design Capacity Per Floor Area {W/m2}','')
  CALL AddToObjStr('Fraction of Autosized Cooling Design Capacity','')
  CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
  CALL AddToObjStr('Heating Design Capacity {W}','autosize')
  CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
  CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
  CALL AddToObjStr('Central Cooling Capacity Control Method','OnOff',.true.)
  !AIR PRIMARY LOOP ~ line 184
  CALL CreateNewObj('AirLoopHVAC')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' ')
  CALL AddToObjFld('Controller List Name', base + vsAirHandlerNameOff,' Controllers')
  CALL AddToObjFld('Availability Manager List Name', base + vsAirHandlerNameOff,' Availability Managers')
  CALL AddToObjStr('Design Supply Air Flow Rate {m3/s}', 'autosize')
  CALL AddToObjFld('Branch List Name', base + vsAirHandlerNameOff,' Branches')
  CALL AddToObjStr('Connector List Name','')
  CALL AddToObjFld('Supply Side Inlet Node Name', base + vsAirHandlerNameOff,' Air Loop Inlet')
  CALL AddToObjFld('Demand Side Outlet Node Name', base + vsAirHandlerNameOff,' Return Air Outlet')
  CALL AddToObjFld('Demand Side Inlet Node Names', base + vsAirHandlerNameOff,' Supply Path Inlet')
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    IF (heatCoilType .EQ. ctNone) THEN
      CALL AddToObjFld('Supply Side Outlet Node Names', base + vsAirHandlerNameOff,' Cooling Coil Outlet',.TRUE.)
    ELSE
      CALL AddToObjFld('Supply Side Outlet Node Names', base + vsAirHandlerNameOff,' Heating Coil Outlet',.TRUE.)
    END IF
  ELSE
    CALL AddToObjFld('Supply Side Outlet Node Names', base + vsAirHandlerNameOff,' Supply Fan Outlet',.TRUE.)
  END IF
  !CONTROLLER LIST ~ line 196       Updated section on Sept 28, 2006 to rearrange coil order for CR7094
  CALL CreateNewObj('AirLoopHVAC:ControllerList')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Controllers')
  !if vsHeatCoilKindOff=Hot Water
  IF (heatCoilType .EQ. ctHotWater) THEN
    CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
    CALL AddToObjFld('Controller Name', base + vsAirHandlerNameOff,' Cooling Coil Controller')
    CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
    CALL AddToObjFld('Controller Name', base + vsAirHandlerNameOff,' Heating Coil Controller',.TRUE.)
  ELSE
    CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
    CALL AddToObjFld('Controller Name', base + vsAirHandlerNameOff,' Cooling Coil Controller',.TRUE.)
  END IF
  !BRANCH LIST ~ line 220
  CALL CreateNewObj('BranchList')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Branches')
  CALL AddToObjFld('Branch Name', base + vsAirHandlerNameOff,' Main Branch',.TRUE.)
  !BRANCH ~ line 224
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Main Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  IF (isReturnFanYes) THEN
    CALL AddToObjStr('Component Object Type','Fan:VariableVolume')
    CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Return Fan')
    CALL AddToObjFld('Component Inlet Node Name', base + vsAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + vsAirHandlerNameOff,' Return Fan Outlet')
  END IF
  CALL AddToObjStr('Component Object Type','AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' OA System')
  CALL AddToObjStr('Component Inlet Node Name', TRIM(returnInletToOAMIxer))
  CALL AddToObjFld('Component Outlet Node Name', base + vsAirHandlerNameOff,' Mixed Air Outlet')
  IF (humidifierKind .EQ. humidifyElecSteam) THEN
    CALL AddToObjStr('Component Object Type','Humidifier:Steam:Electric')
    CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Humidifier')
    CALL AddToObjFld('Component Inlet Node Name', base + vsAirHandlerNameOff,' Mixed Air Outlet')
    CALL AddToObjFld('Component Outlet Node Name', base + vsAirHandlerNameOff,' Humidifier Outlet')
  END IF
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    CALL AddToObjStr('Component Object Type','Fan:VariableVolume')
    CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Supply Fan')
    IF (isHumidifierNone) THEN
      CALL AddToObjFld('Component Inlet Node Name', base + vsAirHandlerNameOff,' Mixed Air Outlet')
    ELSE
      CALL AddToObjFld('Component Inlet Node Name', base + vsAirHandlerNameOff,' Humidifier Outlet')
    END IF
    CALL AddToObjFld('Component Outlet Node Name', base + vsAirHandlerNameOff,' Supply Fan Outlet')
  END IF
! MJW 18 Aug 2005 - Switch to Detailed cooling coil model to avoid sizing problem
! MJW 27 Jun 2007 - Cooling coil type is a choice now, revert to COIL:WATER:COOLING as the default, sizing is better than it was
  IF (coolCoilKind .EQ. ccChWater) THEN
    CALL AddToObjStr('Component Object Type','Coil:Cooling:Water')
  ELSEIF (coolCoilKind .EQ. ccChWaterDetailed) THEN
    CALL AddToObjStr('Component Object Type','Coil:Cooling:Water:DetailedGeometry')
  END IF
  CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Cooling Coil')
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    CALL AddToObjFld('Component Inlet Node Name', base + vsAirHandlerNameOff,' Supply Fan Outlet')
  ELSE
    IF (isHumidifierNone) THEN
      CALL AddToObjFld('Component Inlet Node Name', base + vsAirHandlerNameOff,' Mixed Air Outlet')
    ELSE
      CALL AddToObjFld('Component Inlet Node Name', base + vsAirHandlerNameOff,' Humidifier Outlet')
    END IF
  END IF
  IF ((heatCoilType .EQ. ctNone) .AND. (supFanPlacement .NE. sfpDrawThru)) THEN
    CALL AddToObjFld('Component Outlet Node Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet',.TRUE.)
  ELSE
    CALL AddToObjFld('Component Outlet Node Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet')
  END IF
  IF (heatCoilType .NE. ctNone) THEN
    IF (heatCoilType .EQ. ctHotWater) THEN
      CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    ELSEIF (heatCoilType .EQ. ctElectric) THEN
      CALL AddToObjStr('Component Object Type','Coil:Heating:Electric')
    ELSEIF (heatCoilType .EQ. ctGas) THEN
      CALL AddToObjStr('Component Object Type','Coil:Heating:Fuel')
    END IF
    CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet')
    IF (supFanPlacement .EQ. sfpDrawThru) THEN
      CALL AddToObjFld('Component Outlet Node Name', base + vsAirHandlerNameOff,' Heating Coil Outlet')
    ELSE
      CALL AddToObjFld('Component Outlet Node Name', base + vsAirHandlerNameOff,' Heating Coil Outlet',.TRUE.)
    END IF
  END IF
  IF (supFanPlacement .EQ. sfpDrawThru) THEN
    CALL AddToObjStr('Component Object Type','Fan:VariableVolume')
    CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Supply Fan')
    IF (heatCoilType .NE. ctNone) THEN
      CALL AddToObjFld('Component Inlet Node Name', base + vsAirHandlerNameOff,' Heating Coil Outlet')
    ELSE
      CALL AddToObjFld('Component Inlet Node Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet')
    END IF
    CALL AddToObjFld('Component Outlet Node Name', base + vsAirHandlerNameOff,' Supply Fan Outlet',.TRUE.)
  END IF
  IF (isSupplyPlenumBlank) THEN
    !ZONE SUPPLY AIR PATH ~ line 236
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + vsAirHandlerNameOff,' Supply Path Inlet')
    CALL MakeSpecialIntList(zlVAVAllSupPlen, base + vsAirHandlerNameOff)
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneSplitter')
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Zone Splitter',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Zone Splitter')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Supply Plenum', .TRUE.)
      END IF
    END IF
    !ZONE SPLITTER ~ line 244
    CALL CreateNewObj('AirLoopHVAC:ZoneSplitter')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Zone Splitter')
    CALL AddToObjFld('Inlet Node Name', base + vsAirHandlerNameOff,' Supply Path Inlet')
    CALL MakeSpecialIntList(zlVavAll,base + vsAirHandlerNameOff)
    ! add logic so error message is not repeated.
    IF (numIntList .EQ. 0) THEN
      doSuppressMsgzlVavAll = .TRUE.
    END IF
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(i),' Zone Equip Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Inlet',.TRUE.)
    END IF
  ELSE
    !ZONE SUPPLY AIR PATH ~ line 250
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + vsAirHandlerNameOff,' Supply Path Inlet')
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
    CALL MakeSpecialIntList(zlVAVAllSupPlen, base + vsAirHandlerNameOff)
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Supply Plenum',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Supply Plenum')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Supply Plenum', .TRUE.)
      END IF
    END IF
    !ZONE SUPPLY PLENUM ~ line 258
    CALL CreateNewObj('AirLoopHVAC:SupplyPlenum')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Supply Plenum')
    CALL AddToObjFld('Zone Name', base + vsSupplyPlenumNameOff,' ')
    CALL AddToObjFld('Zone Node Name', base + vsSupplyPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Inlet Node Name', base + vsAirHandlerNameOff,' Supply Path Inlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlVavAll,base + vsAirHandlerNameOff)
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(i),' Zone Equip Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Inlet',.TRUE.)
    END IF
  END IF
  IF (isReturnPlenumBlank) THEN
    ! Check if any fan powered zones are requesting induced air from system plenum
    CALL MakeSpecialIntList(zlVavFPfromSysRetPlen, base + vsAirHandlerNameOff)
      !the IntList holds the list of pointers to the zone names for induced air nodes
    IF (numIntList .NE. 0) THEN
      CALL WriteError('In HVACTemplate:System:VAV "'//TRIM(FldVal(base + vsAirHandlerNameOff))//'"'// &
                      ' there are fan powered zones requesting secondary air from the system return plenum'// &
                      ' but the Return Plenum Name is blank.')
    END IF

    !ZONE RETURN AIR PATH ~ line 266
    CALL CreateNewObj('AirLoopHVAC:ReturnPath')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Return Path')
    CALL AddToObjFld('Return Air Path Outlet Node Name', base + vsAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlVavAllRetPlen, base + vsAirHandlerNameOff)
    IF (numIntList .NE. 0) THEN
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Return Plenum')
      END DO
    END IF
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneMixer')
    CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Zone Mixer',.TRUE.)
    !ZONE MIXER ~ line 274
    CALL CreateNewObj('AirLoopHVAC:ZoneMixer')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Zone Mixer')
    CALL AddToObjFld('Outlet Node Name', base + vsAirHandlerNameOff,' Return Air Outlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlVavAll,base + vsAirHandlerNameOff,doSuppressMsgzlVavAll)
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Inlet Node Name', IntList(i),' Return Outlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Inlet Node Name', IntList(numIntList),' Return Outlet',.TRUE.)
    END IF
  ELSE
    !ZONE RETURN AIR PATH ~ line 280
    CALL CreateNewObj('AirLoopHVAC:ReturnPath')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Return Path')
    CALL AddToObjFld('Return Air Path Outlet Node Name', base + vsAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlVavAllRetPlen, base + vsAirHandlerNameOff)
    IF (numIntList .NE. 0) THEN
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Return Plenum')
      END DO
    END IF
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Return Plenum',.TRUE.)
    !ZONE RETURN PLENUM ~ line 288
    CALL CreateNewObj('AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Return Plenum')
    CALL AddToObjFld('Zone Name', base + vsReturnPlenumNameOff,' ')
    CALL AddToObjFld('Zone Node Name', base + vsReturnPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Outlet Node Name', base + vsAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlVavFPfromSysRetPlen, base + vsAirHandlerNameOff)
      !the IntList holds the list of pointers to the zone names for induced air nodes
    IF (numIntList .NE. 0) THEN
      CALL AddToObjFld('Induced Air Outlet Node or NodeList Name', base + vsAirHandlerNameOff,' Induced Air Nodes')
    ELSE
      CALL AddToObjStr('Induced Air Outlet Node or NodeList Name','')
    END IF
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlVavAll,base + vsAirHandlerNameOff)
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Inlet Node Name', IntList(i),' Return Outlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Inlet Node Name', IntList(numIntList),' Return Outlet',.TRUE.)
    END IF
  END IF

    !Object ==> NodeList for induced air nodes
  CALL MakeSpecialIntList(zlVavFPfromSysRetPlen, base + vsAirHandlerNameOff)
  !the IntList holds the list of pointers to the zone names for induced air nodes
  IF (numIntList .NE. 0) THEN
    CALL CreateNewObj('NodeList')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Induced Air Nodes')
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Node Name', IntList(i),' PIU Secondary Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Node Name', IntList(numIntList),' PIU Secondary Inlet',.TRUE.)
    END IF
  END IF

  !SYSTEM AVAILABILITY MANAGER LIST ~ line 296
  CALL CreateNewObj('AvailabilityManagerAssignmentList')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Availability Managers')
  CALL AddToObjStr('Availability Manager Object Type','AvailabilityManager:NightCycle')
  CALL AddToObjFld('Availability Manager Name', base + vsAirHandlerNameOff,' Availability',.TRUE.)
  !SYSTEM AVAILABILITY MANAGER:NIGHT CYCLE line ~307
  CALL CreateNewObj('AvailabilityManager:NightCycle')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Availability')
  CALL AddToObjStr('Applicability Schedule Name','HVACTemplate-Always 1')
  IF (isSysAvailSchedBlank) THEN
    CALL AddToObjStr('Fan Schedule Name','HVACTemplate-Always 1')
  ELSE
    CALL AddToObjFld('Fan Schedule Name', base + vsSysAvailSchedNameOff,'')
  END IF
  CALL AddToObjFld('Control Type', base + vsNightCycleControlOff,'')
  CALL AddToObjStr('Thermostat Tolerance {deltaC}','0.2')
  CALL AddToObjStr('Cycling Run Time Control Type', 'FixedRunTime')
  CALL AddToObjStr('Cycling run time {s}','3600')
  CALL AddToObjFld('Control Zone Name', base + vsNightCycleCtrlZoneNameOff,'',.TRUE.)
  CALL AddAlwaysSchedule('1')
  SELECT CASE (coolSetPtReset)
    CASE (csprNone)
      !SET POINT MANAGER:SCHEDULED  ~ line 317
      IF (isCoolStPtSchedBlank) THEN
        CALL AddAlwaysSchedule(FldVal(base + vsCoolSetPtDesignOff))
      END IF
      CALL CreateNewObj('SetpointManager:Scheduled')
      CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      IF (isCoolStPtSchedBlank) THEN
        CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + vsCoolSetPtDesignOff)))
      ELSE
        CALL AddToObjFld('Schedule Name', base + vsCoolSetPtSchedNameOff,' ')
      END IF
      IF (supFanPlacement .EQ. sfpBlowThru) THEN
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet',.TRUE.)
      ELSE
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Supply Fan Outlet',.TRUE.)
      END IF
    CASE (csprWarmest)
      !SET POINT MANAGER:WARMEST ~ line 322
      CALL CreateNewObj('SetpointManager:Warmest')
      CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('HVAC Air Loop Name', base + vsAirHandlerNameOff,'')
      CALL AddToObjFld('Minimum Setpoint Temperature {C}', base + vsCoolSetPtDesignOff,'')
      IF ((StringToReal(FldVal(base + vsHeatSetPtDesignOff)) + 1.0) .GT. 18.0) THEN
        CALL AddToObjNum('Maximum Setpoint Temperature {C}',StringToReal(FldVal(base + vsHeatSetPtDesignOff)) + 1.0)
      ELSE
        CALL AddToObjStr('Maximum Setpoint Temperature {C}','18')
      END IF
      CALL AddToObjStr('Strategy','MaximumTemperature')
      IF (supFanPlacement .EQ. sfpBlowThru) THEN
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet',.TRUE.)
      ELSE
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Supply Fan Outlet',.TRUE.)
      END IF
    CASE (csprWarmestTempFirst)
      !SET POINT MANAGER:WarmestTemperatureFlow
      CALL CreateNewObj('SetpointManager:WarmestTemperatureFlow')
      CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('HVAC Air Loop Name', base + vsAirHandlerNameOff,'')
      CALL AddToObjFld('Minimum Setpoint Temperature {C}', base + vsCoolSetPtDesignOff,'')
      IF ((StringToReal(FldVal(base + vsHeatSetPtDesignOff)) + 1.0) .GT. 18.0) THEN
        CALL AddToObjNum('Maximum Setpoint Temperature {C}',StringToReal(FldVal(base + vsHeatSetPtDesignOff)) + 1.0)
      ELSE
        CALL AddToObjStr('Maximum Setpoint Temperature {C}','18')
      END IF
      CALL AddToObjStr('Strategy','TemperatureFirst')
      IF (supFanPlacement .EQ. sfpBlowThru) THEN
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet')
      ELSE
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Supply Fan Outlet')
      END IF
      IF (isSupplyMaxRateAutosize .OR. isSupplyMinRateAutosize) THEN
        CALL AddToObjStr('Minimum Turndown Ratio','0.2',.TRUE.)
      ELSE
        minAirRatio = StringToReal(FldVal(base + vsSupplyMinRateOff)) / &
                      StringToReal(FldVal(base + vsSupplyMaxRateOff))
        CALL AddToObjNum('Minimum Turndown Ratio', minAirRatio,.TRUE.)
      END IF
    CASE (csprOAreset)
      !SET POINT MANAGER:OUTSIDE AIR ~ line 330
      CALL CreateNewObj('SetpointManager:OutdoorAirReset')
      CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      !prior to fix of CR7971
      !CALL AddToObjStr('Set point at outside low {C}','18')
      CALL AddToObjNum('Set point at outside low {C}',StringToReal(FldVal(base + vsCoolSetPtDesignOff)) + 5.2)
      CALL AddToObjStr('Outside low {C}','15.6')
      CALL AddToObjFld('Set point at outside high {C}', base + vsCoolSetPtDesignOff,'')
      CALL AddToObjStr('Outside high {C}','26.7')
      IF (supFanPlacement .EQ. sfpBlowThru) THEN
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet',.TRUE.)
      ELSE
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Supply Fan Outlet',.TRUE.)
      END IF
  END SELECT
  IF (coolCoilKind .EQ. ccChWaterDetailed) THEN
    ! MJW 18 Aug 2005 - Switch to Detailed cooling coil model to avoid sizing problem
    ! MJW 27 Jun 2007 - Cooling coil type is a choice now, revert to COIL:WATER:COOLING as the default, sizing is better than it was
    !COIL:Water:DetailedFlatCooling ~ line 372
    CALL CreateNewObj('Coil:Cooling:Water:DetailedGeometry')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Cooling Coil')
    CALL AddToObjFld('Availability Schedule Name', base + vsCoolAvailSchedNameOff,' ')
    CALL AddToObjStr('Max Water Flow Rate of Coil {m3/s}','autosize')
    CALL AddToObjStr('Tube Outside Surf Area {m2}','autosize')
    CALL AddToObjStr('Total Tube Inside Area {m2}','autosize')
    CALL AddToObjStr('Fin Surface Area {m2}','autosize')
    CALL AddToObjStr('Minimum Air Flow Area {m2}','autosize')
    CALL AddToObjStr('Coil Depth {m}','autosize')
    CALL AddToObjStr('Fin Diameter {m}','autosize')
    CALL AddToObjStr('Fin Thickness {m}','0.0015')
    CALL AddToObjStr('Tube Inside Diameter {m}','0.01445')
    CALL AddToObjStr('Tube Outside Diameter {m}','0.0159')
    CALL AddToObjStr('Tube Thermal Conductivity {W/m-K}','386.0')
    CALL AddToObjStr('Fin Thermal Conductivity {W/m-K}','204.0')
    CALL AddToObjStr('Fin Spacing {m}','0.0018')
    CALL AddToObjStr('Tube Depth Spacing {m}','0.026')
    CALL AddToObjStr('Number of Tube Rows','4')
    CALL AddToObjStr('Number of Tubes per Row','autosize')
    CALL AddToObjFld('Coil Water Inlet Node', base + vsAirHandlerNameOff,' Cooling Coil ChW Inlet')
    CALL AddToObjFld('Coil Water Outlet Node', base + vsAirHandlerNameOff,' Cooling Coil ChW Outlet')
    IF (supFanPlacement .EQ. sfpBlowThru) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Supply Fan Outlet')
    ELSE
      IF (isHumidifierNone) THEN
        CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Mixed Air Outlet')
      ELSE
        CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Humidifier Outlet')
      END IF
    END IF
    CALL AddToObjFld('Air Outlet Node Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet',.TRUE.)
  ELSEIF (coolCoilKind .EQ. ccChWater) THEN
    !COIL:Water:Cooling ~ line 399
    CALL CreateNewObj('Coil:Cooling:Water')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Cooling Coil')
    CALL AddToObjFld('Availability Schedule Name', base + vsCoolAvailSchedNameOff,' ')
    CALL AddToObjStr('Design Water Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Design Air Flow Rate {m3/s}', 'autosize')
    CALL AddToObjStr('Design Inlet Water Temperature {C}','autosize')
    CALL AddToObjStr('Design Inlet Air Temperature {C}','autosize')
    CALL AddToObjStr('Design Outlet Air Temperature {C}','autosize')
    CALL AddToObjStr('Design Inlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
    CALL AddToObjStr('Design Outlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
    CALL AddToObjFld('Water Inlet Node Name', base + vsAirHandlerNameOff,' Cooling Coil ChW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + vsAirHandlerNameOff,' Cooling Coil ChW Outlet')
    IF (supFanPlacement .EQ. sfpBlowThru) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Supply Fan Outlet')
    ELSE
      IF (isHumidifierNone) THEN
        CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Mixed Air Outlet')
      ELSE
        CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Humidifier Outlet')
      END IF
    END IF
    CALL AddToObjFld('Air Outlet Node Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet')
    CALL AddToObjStr('Type of Analysis','DetailedAnalysis')
    CALL AddToObjStr('Heat Exchanger Configuration','CrossFlow',.TRUE.)
  END IF
  !CONTROLLER:SIMPLE ~ line 419
  CALL CreateNewObj('Controller:WaterCoil')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Cooling Coil Controller')
  IF (isDehumidifyNone) THEN
    CALL AddToObjStr('Control Variable','Temperature')
  ELSE
    CALL AddToObjStr('Control Variable','TemperatureAndHumidityRatio')
  END IF
  CALL AddToObjStr('Action','Reverse')
  CALL AddToObjStr('Actuator Variable','Flow')
  CALL AddToObjFld('Sensor Node Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet')
  CALL AddToObjFld('Actuator Node Name', base + vsAirHandlerNameOff,' Cooling Coil ChW Inlet')
  CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
  CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
  CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)

  ! Humdistat(s) if needed
  !    Single humidistat if humidification and dehumidification are both active and control zones are the same
  IF ((.NOT. isHumidifierNone) .AND. (.NOT. isDehumidifyNone) &
       .AND. SameString(FldVal(base +  vsDehumCtrlZoneOff),FldVal(base +  vsHumidCtrlZoneOff))) THEN
    CALL CreateNewObj('ZoneControl:Humidistat')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Humidistat')
    CALL AddToObjFld('Zone Name', base + vsDehumCtrlZoneOff,'')
    CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name', &
                     'HVACTemplate-Always ' // TRIM(FldVal(base + vsHumidSetPtOff)))
    CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name', &
                     'HVACTemplate-Always ' // TRIM(FldVal(base + vsDehumSetPtOff)),.TRUE.)

    CALL AddAlwaysSchedule(FldVal(base + vsHumidSetPtOff))
    CALL AddAlwaysSchedule(FldVal(base + vsDehumSetPtOff))

  ELSE
    IF (.NOT. isDehumidifyNone) THEN
  !   Dehumidification humidistat
      CALL CreateNewObj('ZoneControl:Humidistat')
      CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Dehumidification Humidistat')
      CALL AddToObjFld('Zone Name', base + vsDehumCtrlZoneOff,'')
      CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name','HVACTemplate-Always 1')
      CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name', &
                       'HVACTemplate-Always ' // TRIM(FldVal(base + vsDehumSetPtOff)),.TRUE.)

      CALL AddAlwaysSchedule('1')
      CALL AddAlwaysSchedule(FldVal(base + vsDehumSetPtOff))
    ENDIF
    IF (.NOT. isHumidifierNone) THEN
  !    Humidification humidistat
      CALL CreateNewObj('ZoneControl:Humidistat')
      CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Humidification Humidistat')
      CALL AddToObjFld('Zone Name', base + vsHumidCtrlZoneOff,'')
      CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name', &
                       'HVACTemplate-Always ' // TRIM(FldVal(base + vsHumidSetPtOff)))
      CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name','HVACTemplate-Always 100',.TRUE.)

      CALL AddAlwaysSchedule(FldVal(base + vsHumidSetPtOff))
      CALL AddAlwaysSchedule('100')
    ENDIF
  ENDIF

  !Set point manager for dehumidification
  IF (.NOT. isDehumidifyNone) THEN
    CALL CreateNewObj('SetpointManager:SingleZone:Humidity:Maximum')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Dehumidification Setpoint Manager')
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet')
    CALL AddToObjFld('Control Zone Air Node Name', base + vsDehumCtrlZoneOff,' Zone Air Node',.TRUE.)
  END IF

  !SET POINT MANAGER:MIXED AIR  ~ Line 443
  CALL CreateNewObj('SetpointManager:MixedAir')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Cooling Coil Air Temp Manager')
  CALL AddToObjStr('Control Variable','Temperature')
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    CALL AddToObjFld('Reference Setpoint Node Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet')
    IF (isHumidifierNone) THEN
      CALL AddToObjFld('Fan Inlet Node Name', base + vsAirHandlerNameOff,' Mixed Air Outlet')
    ELSE
      CALL AddToObjFld('Fan Inlet Node Name', base + vsAirHandlerNameOff,' Humidifier Outlet')
    END IF
  ELSE
    CALL AddToObjFld('Reference Setpoint Node Name', base + vsAirHandlerNameOff,' Supply Fan Outlet')
    IF (heatCoilType .NE. ctNone) THEN
      CALL AddToObjFld('Fan Inlet Node Name', base + vsAirHandlerNameOff,' Heating Coil Outlet')
    ELSE
      CALL AddToObjFld('Fan Inlet Node Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet')
    END IF
  END IF
  CALL AddToObjFld('Fan Outlet Node Name', base + vsAirHandlerNameOff,' Supply Fan Outlet')
  CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Mixed Air Nodes',.TRUE.)
  !NODE LIST ~ line 391
  CALL CreateNewObj('NodeList')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Mixed Air Nodes')
  IF (supFanPlacement .EQ. sfpDrawThru) THEN
    CALL AddToObjFld('Node Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet')
  END IF
  CALL AddToObjFld('Node Name', base + vsAirHandlerNameOff,' Mixed Air Outlet',.TRUE.)
  !BRANCH ~ line 396
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Cooling Coil ChW Branch')
  !store for later in the chilled water loop
  CALL AddToStrList(TRIM(FldVal(base + vsAirHandlerNameOff)) // ' Cooling Coil ChW Branch', handleCoolingCoilBranch)
  CALL AddToObjStr('Pressure Drop Curve Name','')
! MJW 18 Aug 2005 - Switch to Detailed cooling coil model to avoid sizing problem
! MJW 27 Jun 2007 - Cooling coil type is a choice now, revert to COIL:WATER:COOLING as the default, sizing is better than it was
  IF (coolCoilKind .EQ. ccChWater) THEN
    CALL AddToObjStr('Component Object Type','Coil:Cooling:Water')
  ELSEIF (coolCoilKind .EQ. ccChWaterDetailed) THEN
    CALL AddToObjStr('Component Object Type','Coil:Cooling:Water:DetailedGeometry')
  END IF
  CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Cooling Coil')
  CALL AddToObjFld('Component Inlet Node Name', base + vsAirHandlerNameOff,' Cooling Coil ChW Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + vsAirHandlerNameOff,' Cooling Coil ChW Outlet',.TRUE.)
  IF (heatCoilType .EQ. ctHotWater) THEN
    !COIL:Water:SimpleHeating ~ line 406
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + vsHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name ', base + vsAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + vsAirHandlerNameOff,' Heating Coil HW Outlet')
    CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + vsAirHandlerNameOff,' Heating Coil Outlet')
    CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
    CALL AddToObjStr('Rated Capacity','autosize')
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
    !BRANCH ~ line 416
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Heating Coil HW Branch')
    !save this branch name for later use in plan
    CALL AddToStrList(TRIM(FldVal(base + vsAirHandlerNameOff)) // ' Heating Coil HW Branch',handleHeatingCoilBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + vsAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + vsAirHandlerNameOff,' Heating Coil HW Outlet',.TRUE.)
    !CONTROLLER:SIMPLE ~line 425
    CALL CreateNewObj('Controller:WaterCoil')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Heating Coil Controller')
    CALL AddToObjStr('Control Variable','Temperature')
    CALL AddToObjStr('Action','Normal')
    CALL AddToObjStr('Actuator Variable','Flow')
    CALL AddToObjFld('Sensor Node Name', base + vsAirHandlerNameOff,' Heating Coil Outlet')
    CALL AddToObjFld('Actuator Node Name', base + vsAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
    CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
    CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)
  END IF
  !COIL:Electric:Heating ~ line 437
  IF (heatCoilType .EQ. ctElectric) THEN
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + vsHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Efficiency','1')
    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + vsAirHandlerNameOff,' Heating Coil Outlet')
    CALL AddToObjFld('Coil Temp Setpoint Node', base + vsAirHandlerNameOff,' Heating Coil Outlet',.TRUE.)
  END IF
  IF (heatCoilType .EQ. ctGas) THEN
    !COIL:Gas:Heating ~ line 448
    CALL CreateNewObj('Coil:Heating:Fuel')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + vsHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Fuel Type', 'Gas')
    CALL AddToObjFld('Gas Burner Efficiency', base + vsHeatEffiencyOff,' ')
    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + vsAirHandlerNameOff,' Heating Coil Outlet')
    CALL AddToObjFld('Coil Temp Setpoint Node', base + vsAirHandlerNameOff,' Heating Coil Outlet')
    CALL AddToObjFld('Parasitic Electric Load {W}', base + vsHeatParasiticOff,' ')
    CALL AddToObjFld('Part load fraction correlation (function of part load ratio)',   &
       base + vsAirHandlerNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
    !CURVE:CUBIC ~ line 461
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  END IF
  IF (heatCoilType .NE. ctNone) THEN
    !If  vsHeatSetPtSchedNameOff is blank, then create schedule of vsHeatSetPtDesignOff
    IF (isHeatStPtSchedBlank) THEN
      CALL AddAlwaysSchedule(FldVal(base + vsHeatSetPtDesignOff))
    END IF
    IF (supFanPlacement .EQ. sfpDrawThru) THEN
      CALL WriteComment('NOTE:  Heating supply air setpoint is placed on Supply Path ' // &
                      'Inlet node to allow separate heating and cooling setpoints.')
      CALL WriteComment('SetpointManager:MixedAir <system name> Heating Coil Air ' // &
                      'Temp Manager adjusts this setpoint for fan heat and places it on the Heating Coil Outlet')
    END IF
    SELECT CASE (heatSetPtReset)
      CASE (hsprNone)
        !SET POINT MANAGER:SCHEDULED ~ line 472
        IF (isHeatStPtSchedBlank) THEN
          CALL AddAlwaysSchedule(FldVal(base + vsHeatSetPtDesignOff))
        END IF
        CALL CreateNewObj('SetpointManager:Scheduled')
        CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Heating Supply Air Temp Manager')
        CALL AddToObjStr('Control Variable','Temperature')
        IF (isHeatStPtSchedBlank) THEN
          CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + vsHeatSetPtDesignOff)))
        ELSE
          CALL AddToObjFld('Schedule Name', base + vsHeatSetPtSchedNameOff,'')
        END IF
        IF (supFanPlacement .EQ. sfpBlowThru) THEN
          CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Heating Coil Outlet',.TRUE.)
        ELSE
          CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Supply Path Inlet',.TRUE.)
        END IF
!    CASE (hsprColdestTempFirst)
!  This setpoint manager type not yet available in EnergyPlus, leave code here for future use
      !SET POINT MANAGER:ColdestTemperatureFlow
!      CALL CreateNewObj('SetpointManager:ColdestTemperatureFlow')
!      CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Heating Supply Air Temp Manager')
!      CALL AddToObjStr('Control Variable','Temperature')
!      CALL AddToObjFld('HVAC Air Loop Name', base + vsAirHandlerNameOff,'')
!      IF ((StringToReal(FldVal(base + vsCoolSetPtDesignOff)) - 1.0) .LT. (StringToReal(FldVal(base + vsHeatSetPtDesignOff)) - 5.2)) THEN
!        CALL AddToObjNum('Minimum Setpoint Temperature {C}',StringToReal(FldVal(base + vsCoolSetPtDesignOff)) - 1.0)
!      ELSE
!        CALL AddToObjNum('Minimum Setpoint Temperature {C}',StringToReal(FldVal(base + vsHeatSetPtDesignOff)) - 5.2)
!      END IF
!      CALL AddToObjFld('Maximum Setpoint Temperature {C}', base + vsHeatSetPtDesignOff,'')
!      CALL AddToObjStr('Strategy','TemperatureFirst')
!      IF (supFanPlacement .EQ. sfpBlowThru) THEN
!        CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Heating Coil Outlet',.TRUE.)
!      ELSE
!        CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Supply Path Inlet',.TRUE.)
!      END IF
    CASE (hsprOAreset)
      !SET POINT MANAGER:OUTSIDE AIR ~ line 485
      CALL CreateNewObj('SetpointManager:OutdoorAirReset')
      CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Heating Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('Set point at outside low {C}', base + vsHeatSetPtDesignOff,'')
      CALL AddToObjStr('Outside low {C}','-6.7')
      !prior to fix of CR7971
      !CALL AddToObjStr('Set point at outside high {C}','20')
      CALL AddToObjNum('Set point at outside high {C}',StringToReal(FldVal(base + vsHeatSetPtDesignOff)) - 5.2)
      CALL AddToObjStr('Outside high {C}','10')
      IF (supFanPlacement .EQ. sfpBlowThru) THEN
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Heating Coil Outlet',.TRUE.)
      ELSE
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Supply Path Inlet',.TRUE.)
      END IF
    END SELECT
    !SET POINT MANAGER:MIXED AIR ~ line 494
    ! Only needed if system is draw through and there is a heating coil
    IF (supFanPlacement .EQ. sfpDrawThru) THEN
      CALL CreateNewObj('SetpointManager:MixedAir')
      CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Heating Coil Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('Reference Setpoint Node Name', base + vsAirHandlerNameOff,' Supply Path Inlet')
      CALL AddToObjFld('Fan Inlet Node Name', base + vsAirHandlerNameOff,' Heating Coil Outlet')
      CALL AddToObjFld('Fan Outlet Node Name', base + vsAirHandlerNameOff,' Supply Fan Outlet')
      CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Heating Coil Outlet',.TRUE.)
    END IF
  END IF
  ! Return fan
  IF (isReturnFanYes) THEN
    CALL CreateNewObj('Fan:VariableVolume')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Return Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + vsSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + vsReturnEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + vsReturnPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
    CALL AddToObjStr('Fan Power Minimum Flow Rate Input Method','Fraction')
    CALL AddToObjStr('Fan Power Minimum Flow Fraction','0.0')
    CALL AddToObjStr('Fan Power Minimum Air Flow Rate {m3/s}', '')
    CALL AddToObjFld('Motor Efficiency', base + vsReturnFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + vsReturnFanFractionOff,' ')
    SELECT CASE (retFanCoef)
      CASE (sfcInletVane)
        CALL AddToObjStr('Fan Power Coefficient 1','0.35071223')
        CALL AddToObjStr('Fan Power Coefficient 2','0.30850535')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.54137364')
        CALL AddToObjStr('Fan Power Coefficient 4','0.87198823')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcDischgDamp)
        CALL AddToObjStr('Fan Power Coefficient 1','0.37073425')
        CALL AddToObjStr('Fan Power Coefficient 2','0.97250253')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.34240761')
        CALL AddToObjStr('Fan Power Coefficient 4','0')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcVSD)
        CALL AddToObjStr('Fan Power Coefficient 1','0.0015302446')
        CALL AddToObjStr('Fan Power Coefficient 2','0.0052080574')
        CALL AddToObjStr('Fan Power Coefficient 3','1.1086242')
        CALL AddToObjStr('Fan Power Coefficient 4','-0.11635563')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcAppG)
        CALL AddToObjStr('Fan Power Coefficient 1','0.0013')
        CALL AddToObjStr('Fan Power Coefficient 2','0.147')
        CALL AddToObjStr('Fan Power Coefficient 3','0.9506')
        CALL AddToObjStr('Fan Power Coefficient 4','-0.0998')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcVSDPresReset) !based on Good Static Pressure Reset VSD Fan from EDR_DesignGuidelines_VAV.pdf
        CALL AddToObjStr('Fan Power Coefficient 1', '0.040759894')
        CALL AddToObjStr('Fan Power Coefficient 2', '0.08804497')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.07292612')
        CALL AddToObjStr('Fan Power Coefficient 4', '0.943739823')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
    END SELECT
    CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Air Outlet Node Name', base + vsAirHandlerNameOff,' Return Fan Outlet',.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF
  !FAN:SIMPLE:VariableVolume ~ line 574
  CALL CreateNewObj('Fan:VariableVolume')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Supply Fan')
  IF (isSysAvailSchedBlank) THEN
    CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
  ELSE
    CALL AddToObjFld('Availability Schedule Name', base + vsSysAvailSchedNameOff,' ')
  END IF
  CALL AddToObjFld('Fan Efficiency', base + vsSupplyEfficiencyOff,' ')
  CALL AddToObjFld('Pressure Rise {Pa}', base + vsSupplyPressureOff,' ')
  CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
  IF (SameString(FldVal(base + vsSupplyMinRateOff),'autosize')) THEN
    CALL AddToObjStr('Fan Power Minimum Flow Rate Input Method','Fraction')
    CALL AddToObjStr('Fan Power Minimum Flow Fraction','0.25')
    CALL AddToObjStr('Fan Power Minimum Air Flow Rate {m3/s}', '')
  ELSE
    CALL AddToObjStr('Fan Power Minimum Flow Rate Input Method','FixedFlowRate')
    CALL AddToObjStr('Fan Power Minimum Flow Fraction','')
    CALL AddToObjFld('Fan Power Minimum Air Flow Rate {m3/s}', base + vsSupplyMinRateOff,' ')
  ENDIF
  CALL AddToObjFld('Motor Efficiency', base + vsSupplyFanMotorEffOff,' ')
  CALL AddToObjFld('Motor in Airstream Fraction', base + vsSupplyFanFractionOff,' ')
  SELECT CASE (supFanCoef)
    CASE (sfcInletVane)
      CALL AddToObjStr('Fan Power Coefficient 1','0.35071223')
      CALL AddToObjStr('Fan Power Coefficient 2','0.30850535')
      CALL AddToObjStr('Fan Power Coefficient 3','-0.54137364')
      CALL AddToObjStr('Fan Power Coefficient 4','0.87198823')
      CALL AddToObjStr('Fan Power Coefficient 5','0')
    CASE (sfcDischgDamp)
      CALL AddToObjStr('Fan Power Coefficient 1','0.37073425')
      CALL AddToObjStr('Fan Power Coefficient 2','0.97250253')
      CALL AddToObjStr('Fan Power Coefficient 3','-0.34240761')
      CALL AddToObjStr('Fan Power Coefficient 4','0')
      CALL AddToObjStr('Fan Power Coefficient 5','0')
    CASE (sfcVSD)
      CALL AddToObjStr('Fan Power Coefficient 1','0.0015302446')
      CALL AddToObjStr('Fan Power Coefficient 2','0.0052080574')
      CALL AddToObjStr('Fan Power Coefficient 3','1.1086242')
      CALL AddToObjStr('Fan Power Coefficient 4','-0.11635563')
      CALL AddToObjStr('Fan Power Coefficient 5','0')
    CASE (sfcAppG)
      CALL AddToObjStr('Fan Power Coefficient 1','0.0013')
      CALL AddToObjStr('Fan Power Coefficient 2','0.147')
      CALL AddToObjStr('Fan Power Coefficient 3','0.9506')
      CALL AddToObjStr('Fan Power Coefficient 4','-0.0998')
      CALL AddToObjStr('Fan Power Coefficient 5','0')
    CASE (sfcVSDPresReset) !based on Good Static Pressure Reset VSD Fan from EDR_DesignGuidelines_VAV.pdf
      CALL AddToObjStr('Fan Power Coefficient 1', '0.040759894')
      CALL AddToObjStr('Fan Power Coefficient 2', '0.08804497')
      CALL AddToObjStr('Fan Power Coefficient 3','-0.07292612')
      CALL AddToObjStr('Fan Power Coefficient 4', '0.943739823')
      CALL AddToObjStr('Fan Power Coefficient 5','0')
  END SELECT
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    IF (isHumidifierNone) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Mixed Air Outlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Humidifier Outlet')
    END IF
  ELSE
    IF (heatCoilType .NE. ctNone) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Heating Coil Outlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Cooling Coil Outlet')
    END IF
  END IF
  CALL AddToObjFld('Air Outlet Node Name', base + vsAirHandlerNameOff,' Supply Fan Outlet',.TRUE.)
  IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')

  ! Humidifier set point manager and humidifier ~ line 610
  IF (.NOT. isHumidifierNone) THEN
    CALL CreateNewObj('SetpointManager:SingleZone:Humidity:Minimum')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Humidification Setpoint Manager')
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Humidifier Outlet')
    CALL AddToObjFld('Control Zone Air Node Name', base + vsHumidCtrlZoneOff,' Zone Air Node',.TRUE.)
    IF (humidifierKind .EQ. humidifyElecSteam) THEN
      CALL CreateNewObj('Humidifier:Steam:Electric')
      CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Humidifier')
      CALL AddToObjFld('Availability Schedule Name', base + vsHumidAvailSchedNameOff,' ')
      CALL AddToObjFld('Rated Capacity {m3/s}', base + vsHumidCapOff,'')
      CALL AddToObjFld('Rated Power {W}', base + vsHumidRatedElecOff,'')
      CALL AddToObjStr('rated fan power {W}','0')
      CALL AddToObjStr('Standby power {W}','0')
      CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Mixed Air Outlet')
      CALL AddToObjFld('Air Outlet Node Name', base + vsAirHandlerNameOff,' Humidifier Outlet',.TRUE.)
    END IF
  END IF
  !OUTSIDE AIR INLET NODE LIST ~ line 633
  CALL CreateNewObj('OutdoorAir:NodeList')
  CALL AddToObjFld('Node or NodeList Name 1', base + vsAirHandlerNameOff,' Outside Air Inlet',.TRUE.)
  !OUTSIDE AIR SYSTEM ~line 539
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' OA System')
  CALL AddToObjFld('Controller List Name', base + vsAirHandlerNameOff,' OA System Controllers')
  CALL AddToObjFld('Outdoor Air Equipment List Name', base + vsAirHandlerNameOff,' OA System Equipment')
  CALL AddToObjFld('Availability Manager List Name', base + vsAirHandlerNameOff,' Availability Managers',.TRUE.)
  !CONTROLLER LIST ~ line 545
  CALL CreateNewObj('AirLoopHVAC:ControllerList')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' OA System Controllers')
  IF (preHeatCoilType .EQ. ctHotWater) THEN
    CALL AddToObjStr('Controller Object Type','Controller:OutdoorAir')
    CALL AddToObjFld('Controller Name', base + vsAirHandlerNameOff,' OA Controller')
    CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
    CALL AddToObjFld('Controller Name', base + vsAirHandlerNameOff,' Preheat Coil Controller',.TRUE.)
  ELSE
    CALL AddToObjStr('Controller Object Type','Controller:OutdoorAir')
    CALL AddToObjFld('Controller Name', base + vsAirHandlerNameOff,' OA Controller',.TRUE.)
  END IF
  !AIR LOOP EQUIPMENT LIST ~ line 552
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem:EquipmentList')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' OA System Equipment')
  IF (preHeatCoilType .EQ. ctHotWater) THEN
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Preheat Coil')
  ELSEIF (preHeatCoilType .EQ. ctElectric) THEN
    CALL AddToObjStr('Component Object Type','Coil:Heating:Electric')
    CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Preheat Coil')
  ELSEIF (preHeatCoilType .EQ. ctGas) THEN
    CALL AddToObjStr('Component Object Type','Coil:Heating:Fuel')
    CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Preheat Coil')
  END IF
  IF (heatRecovery .EQ. htrecSens) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Heat Recovery')
  ELSEIF (heatRecovery .EQ. htrecEnth) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Heat Recovery')
  END IF
  CALL AddToObjStr('Component Object Type','OutdoorAir:Mixer')
  CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' OA Mixing Box',.TRUE.)
  !OUTSIDE AIR MIXER ~ line 567
  CALL CreateNewObj('OutdoorAir:Mixer')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' OA Mixing Box')
  CALL AddToObjFld('Mixed Air Node Name', base + vsAirHandlerNameOff,' Mixed Air Outlet')
  IF ((heatRecovery .EQ. htrecNone) .AND. (preHeatCoilType .EQ. ctNone)) THEN
    CALL AddToObjFld('Outside Air Stream Node Name', base + vsAirHandlerNameOff,' Outside Air Inlet')
  END IF
  IF (preHeatCoilType .NE. ctNone) THEN
    CALL AddToObjFld('Outside Air Stream Node Name', base + vsAirHandlerNameOff,' Preheat Coil Outlet')
  END IF
  IF ((heatRecovery .NE. htrecNone) .AND. (preHeatCoilType .EQ. ctNone)) THEN
    CALL AddToObjFld('Outside Air Stream Node Name', base + vsAirHandlerNameOff,' Heat Recovery Supply Outlet')
  ENDIF
  CALL AddToObjFld('Relief Air Stream Node Name', base + vsAirHandlerNameOff,' Relief Air Outlet')
  CALL AddToObjStr('Return Air Stream Node Name', TRIM(returnInletToOAMIxer),.TRUE.)
  !CONTROLLER:OUTSIDE AIR ~ line 576
  CALL CreateNewObj('Controller:OutdoorAir')
  CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' OA Controller')
  CALL AddToObjFld('Relief Air Outlet Node Name', base + vsAirHandlerNameOff,' Relief Air Outlet')
  CALL AddToObjStr('Return Air Node Name',TRIM(returnInletToOAMIxer))
  CALL AddToObjFld('Mixed Air Node Name', base + vsAirHandlerNameOff,' Mixed Air Outlet')
  CALL AddToObjFld('Actuator Node Name', base + vsAirHandlerNameOff,' Outside Air Inlet')
  CALL AddToObjFld('Minimum Outdoor Air Flow Rate {m3/s}', base + vsMinOutsideFlowOff,' ')
  CALL AddToObjFld('Maximum Outdoor Air Flow Rate {m3/s}', base + vsMaxOutsideFlowOff,' ')
  ! vsEconomizerKindOff is now a straight pass-thru to EconomizerChoice in
  ! the OA controller
  CALL AddToObjFld('Economizer Control Type', base + vsEconomizerKindOff,' ')
  ! If heat recovery is selected, then set economizer control to bypass
  ! otherwise, set to modulate OA flow
  IF (heatRecovery .EQ. htrecNone) THEN
    CALL AddtoObjStr('Economizer Control Action Type','ModulateFlow')
  ELSE
    CALL AddtoObjStr('Economizer Control Action Type','MinimumFlowWithBypass')
  ENDIF
  CALL AddToObjFld('Economizer Maximum Limit Dry-Bulb Temperature {C}', base + vsEconoUpLimitOff,' ')
  CALL AddToObjFld('Economizer Maximum Limit Enthalpy {J/kg}', base + vsEconoEnthLimitOff,' ')
  IF (econoKind .EQ. etFixedDewPointAndDryBulb) THEN
    CALL AddToObjFld('Economizer Maximum Limit Dewpoint Temperature (C)',base + vsEconoUpDewLimitOff,' ')
  ELSE
    CALL AddToObjStr('Economizer Maximum Limit Dewpoint Temperature (C)',' ')
  ENDIF
  IF (econoKind .EQ. etElectronicEnthalpy) THEN
    CALL AddToObjFld('Electronic Enthalpy Limit Curve Name', base + vsAirHandlerNameOff,' ElectrEnthCurve')
  ELSE
    CALL AddToObjStr('Electronic Enthalpy Limit Curve Name',' ')
  ENDIF
  CALL AddToObjFld('Economizer Minimum Limit Dry-Bulb Temperature {C}', base + vsEconoLowLimitOff,' ')
  CALL AddToObjFld('Lockout Type', base + vsEconoLockoutKindOff,' ')
  CALL AddToObjFld('Minimum Limit Type', base + vsMinOutsideControlOff,' ')
  CALL AddToObjFld('Minimum Outdoor Air Schedule Name', base + vsMinOutFlowSchedNameOff,' ')
  CALL AddToObjStr('Minimum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Maximum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Mechanical Ventilation Controller Name','',.TRUE.)
  ! If economizer type is ElectronicEnthalpy then create curve object
  IF (econoKind .EQ. etElectronicEnthalpy) THEN
    !CURVE:CUBIC
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' ElectrEnthCurve')
    CALL AddToObjStr('Coefficient1 Constant','0.01342704')
    CALL AddToObjStr('Coefficient2 x','-0.00047892')
    CALL AddToObjStr('Coefficient3 x**2','0.000053352')
    CALL AddToObjStr('Coefficient4 x**3','-0.0000018103')
    CALL AddToObjStr('Minimum Value of x','16.6')
    CALL AddToObjStr('Maximum Value of x','29.13',.TRUE.)
  ENDIF
  IF (preHeatCoilType .EQ. ctHotWater) THEN
    !COIL:Water:SimpleHeating  ~ line 614
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Preheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + vsPreheatAvailSchedNameOff,' ')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name ', base + vsAirHandlerNameOff,' Preheat Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + vsAirHandlerNameOff,' Preheat Coil HW Outlet')
    IF (heatRecovery .EQ. htrecNone) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Outside Air Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    END IF
    CALL AddToObjFld('Air Outlet Node Name', base + vsAirHandlerNameOff,' Preheat Coil Outlet')
    CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
    CALL AddToObjStr('Rated Capacity','autosize')
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
    !BRANCH ~ line 625
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Preheat Coil HW Branch')
    !save this branch name for later use in plan
    CALL AddToStrList(TRIM(FldVal(base + vsAirHandlerNameOff)) // ' Preheat Coil HW Branch',handleHeatingCoilBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + vsAirHandlerNameOff,' Preheat Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + vsAirHandlerNameOff,' Preheat Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + vsAirHandlerNameOff,' Preheat Coil HW Outlet',.TRUE.)
    !CONTROLLER:SIMPLE ~ line 634
    CALL CreateNewObj('Controller:WaterCoil')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Preheat Coil Controller')
    CALL AddToObjStr('Control Variable','Temperature')
    CALL AddToObjStr('Action','Normal')
    CALL AddToObjStr('Actuator Variable','Flow')
    CALL AddToObjFld('Sensor Node Name', base + vsAirHandlerNameOff,' Preheat Coil Outlet')
    CALL AddToObjFld('Actuator Node Name', base + vsAirHandlerNameOff,' Preheat Coil HW Inlet')
    CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
    CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
    CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)
  ELSEIF (preHeatCoilType .EQ. ctElectric) THEN
    !COIL:Electric:Heating  ~ line 646
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Preheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + vsPreheatAvailSchedNameOff,' ')
    CALL AddToObjStr('Efficiency','1')
    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    IF (heatRecovery .EQ. htrecNone) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Outside Air Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    END IF
    CALL AddToObjFld('Air Outlet Node Name', base + vsAirHandlerNameOff,' Preheat Coil Outlet')
    CALL AddToObjFld('Coil Temp Setpoint Node', base + vsAirHandlerNameOff,' Preheat Coil Outlet',.TRUE.)
  ELSEIF (preHeatCoilType .EQ. ctGas) THEN
    !COIL:Gas:Heating ~ line 657
    CALL CreateNewObj('Coil:Heating:Fuel')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Preheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + vsPreheatAvailSchedNameOff,' ')
    CALL AddToObjStr('Fuel Type', 'Gas')
    CALL AddToObjFld('Gas Burner Efficiency', base + vsPreheatEffiencyOff,' ')
    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    IF (heatRecovery .EQ. htrecNone) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Outside Air Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + vsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    END IF
    CALL AddToObjFld('Air Outlet Node Name', base + vsAirHandlerNameOff,' Preheat Coil Outlet')
    CALL AddToObjFld('Coil Temp Setpoint Node', base + vsAirHandlerNameOff,' Preheat Coil Outlet')
    CALL AddToObjFld('Parasitic Electric Load {W}', base + vsPreheatParasiticOff,' ')
    CALL AddToObjFld('Part load fraction correlation (function of part load ratio)',   &
       base + vsAirHandlerNameOff,' Preheat Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
    !CURVE:CUBIC ~ line 670
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Preheat Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  END IF
  IF (preHeatCoilType .NE. ctNone) THEN
    IF (isPreheatStPtSchedBlank) THEN
      CALL AddAlwaysSchedule(FldVal(base + vsPreheatSetPtConstantOff))
    END IF
    !SET POINT MANAGER:SCHEDULED ~ line 679
    IF (isPreheatStPtSchedBlank) THEN
      CALL AddAlwaysSchedule(FldVal(base + vsPreheatSetPtConstantOff))
    END IF
    CALL CreateNewObj('SetpointManager:Scheduled')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Preheat Coil Air Temp Manager')
    CALL AddToObjStr('Control Variable','Temperature')
    IF (isPreheatStPtSchedBlank) THEN
      CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + vsPreheatSetPtConstantOff)))
    ELSE
      CALL AddToObjFld('Schedule Name', base + vsPreheatSetPtSchedNameOff,'')
    END IF
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Preheat Coil Outlet',.TRUE.)
  END IF
  IF (heatRecovery .EQ. htrecSens) THEN
    !HEAT EXCHANGER:AIR TO AIR:GENERIC ~ line 685
    htRecSens75 = StringToReal(FldVal(base + vsHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + vsAirHandlerNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Nominal supply air flow rate {m3/s}', base + vsMinOutsideFlowOff,'')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + vsHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Heating Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Heating Air Flow','0')
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + vsHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Cooling Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Cooling Air Flow','0')
    CALL AddToObjFld('Supply Air Inlet Node Name', base + vsAirHandlerNameOff,' Outside Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + vsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + vsAirHandlerNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + vsAirHandlerNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','0')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjStr('Heat exchanger type','Plate')
    CALL AddToObjStr('Frost Control Type','ExhaustOnly')
    CALL AddToObjStr('Threshold Temperature','-1.1')
    CALL AddToObjStr('Initial Defrost Time Fraction','0.167')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','0.024',.TRUE.)
  ELSEIF (heatRecovery .EQ. htrecEnth) THEN
    !HEAT EXCHANGER:AIR TO AIR:GENERIC ~ line 708
    htRecSens75 = StringToReal(FldVal(base + vsHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    htRecLat75 = StringToReal(FldVal(base + vsHeatRecLatEffOff)) + 0.05
    IF (htRecLat75 .GT. 1.) htRecLat75 = 1.
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + vsAirHandlerNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Nominal supply air flow rate {m3/s}', base + vsMinOutsideFlowOff,'')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + vsHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Heating Air Flow', base + vsHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow', htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Heating Air Flow', htRecLat75)
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + vsHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Cooling Air Flow', base + vsHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow', htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Cooling Air Flow', htRecLat75)
    CALL AddToObjFld('Supply Air Inlet Node Name', base + vsAirHandlerNameOff,' Outside Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + vsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + vsAirHandlerNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + vsAirHandlerNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','0')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjStr('Heat exchanger type','Rotary')
    CALL AddToObjStr('Frost Control Type','MinimumExhaustTemperature')
    CALL AddToObjStr('Threshold Temperature','1.7')
    CALL AddToObjStr('Initial Defrost Time Fraction','')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','',.TRUE.)
  END IF
  !SET POINT MANAGER:SCHEDULED ~ line 734
  IF (heatRecovery .NE. htrecNone) THEN
    IF (isEconoLowLimitBlank) THEN
      CALL AddAlwaysSchedule('5')
    ELSE
      CALL AddAlwaysSchedule(FldVal(base + vsEconoLowLimitOff))
    END IF
    CALL CreateNewObj('SetpointManager:Scheduled')
    CALL AddToObjFld('Name', base + vsAirHandlerNameOff,' Heat Recovery Air Temp Manager')
    CALL AddToObjStr('Control Variable','Temperature')
    IF (isEconoLowLimitBlank) THEN
      CALL AddToObjStr('Schedule Name','HVACTemplate-Always 5')
    ELSE
      CALL AddToObjStr('Schedule Name','HVACTemplate-Always ' // FldVal(base + vsEconoLowLimitOff))
    END IF
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + vsAirHandlerNameOff,' Heat Recovery Supply Outlet',.TRUE.)
  END IF
END DO
!======================================================================================================
! create Packaged (DX) VAV Systems
!   not developed using the spreadsheet so no spreadsheet reference lines are included
!======================================================================================================
DO iSys = 1, numCompactSysPVAV
  base =  compactSysPVAVbase(iSys)
  ! set flags
  isSupplyMaxRateAutosize = SameString(FldVal(base + pvavsSupplyMaxRateOff),'autosize')
  isSupplyMinRateAutosize = SameString(FldVal(base + pvavsSupplyMinRateOff),'autosize')
  isMinOARateAutosize = SameString(FldVal(base + pvavsMinOutsideFlowOff),'autosize')
  isCoolCapAutosize = SameString(FldVal(base + pvavsCoolCoilCapOff),'autosize')
  isSHRAutosize = SameString(FldVal(base + pvavsCoolCoilSHROff),'autosize')
  isHeatCoilCapAutosize = SameString(FldVal(base + pvavsHeatCoilCapOff),'autosize')
  isSupplyPlenumBlank = (FldVal(base + pvavsSupplyPlenumNameOff) .EQ. '')
  isReturnPlenumBlank = (FldVal(base + pvavsReturnPlenumNameOff) .EQ. '')
  isSysAvailSchedBlank = (FldVal(base + pvavsSysAvailSchedNameOff) .EQ. '')
  isCoolStPtSchedBlank = (FldVal(base + pvavsCoolSetPtSchedNameOff) .EQ. '')
  isHeatStPtSchedBlank = (FldVal(base + pvavsHeatSetPtSchedNameOff) .EQ. '')
  isEconoLowLimitBlank = (FldVal(base + pvavsEconoUpTempLimitOff) .EQ. '')
  isReturnFanYes = (FldVal(base + pvavsReturnFanPresentOff) .EQ. 'Yes')
  !check cooling coil kind
  IF (SameString(FldVal(base +  pvavsCoolCoilKindOff),'TwoSpeedDX')) THEN
    coolCoilKind = ccTwoSpeedDX
  ELSEIF (SameString(FldVal(base + pvavsCoolCoilKindOff),'TwoSpeedHumidControlDX')) THEN
    coolCoilKind = ccTwoSpeedHumidControlDX
  !ELSEIF (SameString(FldVal(base + pvavsCoolCoilKindOff),'ThreeSpeedDX')) THEN
  !  coolCoilKind = ccThreeSpeedDX
  !ELSEIF (SameString(FldVal(base + pvavsCoolCoilKindOff),'FourSpeedDX')) THEN
  !  coolCoilKind = ccFourSpeedDX
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:PackagedVAV "'//TRIM(FldVal(base + pvavsAirHandlerNameOff))//'"'// &
                    ' in the Cooling Coil Type field: '//TRIM(FldVal(base + pvavsCoolCoilKindOff)))
  END IF
  !set heating coil indicator
  IF (SameString(FldVal(base + pvavsHeatCoilKindOff),'HotWater')) THEN
    heatCoilType = ctHotWater
  ELSEIF (SameString(FldVal(base + pvavsHeatCoilKindOff),'Electric')) THEN
    heatCoilType = ctElectric
  ELSEIF (SameString(FldVal(base + pvavsHeatCoilKindOff),'Gas')) THEN
    heatCoilType = ctGas
  ELSEIF (SameString(FldVal(base + pvavsHeatCoilKindOff),'None')) THEN
    heatCoilType = ctNone
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:PackagedVAV in the Heating Coil Type field')
  END IF
  !validate the Minimum Outside Air Control Type
  CALL validChoice(base + pvavsMinOutsideControlOff,'FixedMinimum','ProportionalMinimum',  &
     'System:PackagedVAV','Minimum Outdoor Air Control Type')
  !set the economizer indicator
  IF (SameString(FldVal(base +  pvavsEconomizerKindOff),'FixedDryBulb')) THEN
    econoKind = etFixedDryBulb
  ELSEIF (SameString(FldVal(base +  pvavsEconomizerKindOff),'FixedEnthalpy')) THEN
    econoKind = etFixedEnthalpy
  ELSEIF (SameString(FldVal(base +  pvavsEconomizerKindOff),'DifferentialDryBulb')) THEN
    econoKind = etDifferentialDryBulb
  ELSEIF (SameString(FldVal(base +  pvavsEconomizerKindOff),'DifferentialEnthalpy')) THEN
    econoKind = etDifferentialEnthalpy
  ELSEIF (SameString(FldVal(base +  pvavsEconomizerKindOff),'FixedDewPointAndDryBulb')) THEN
    econoKind = etFixedDewPointAndDryBulb
  ELSEIF (SameString(FldVal(base +  pvavsEconomizerKindOff),'ElectronicEnthalpy')) THEN
    econoKind = etElectronicEnthalpy
  ELSEIF (SameString(FldVal(base +  pvavsEconomizerKindOff),'DifferentialDryBulbAndEnthalpy')) THEN
    econoKind = etDifferentialDryBulbAndEnthalpy
  ELSEIF (SameString(FldVal(base +  pvavsEconomizerKindOff),'NoEconomizer')) THEN
    econoKind = etNoEconomizer
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:PackagedVAV in the Economizer Type field')
  END IF
  !set the economizer lockout indicator
  IF (SameString(FldVal(base +  pvavsEconoLockoutKindOff),'NoLockout')) THEN
    lockoutKind = loNo
  ELSEIF (SameString(FldVal(base +  pvavsEconoLockoutKindOff),'LockoutWithHeating')) THEN
    lockoutKind = loHeat
  ELSEIF (SameString(FldVal(base +  pvavsEconoLockoutKindOff),'LockoutWithCompressor')) THEN
    lockoutKind = loComp
  ELSEIF (SameString(FldVal(base +  pvavsEconoLockoutKindOff),'')) THEN
    lockoutKind = loNo
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:VAV in the Economizer Lockout field')
  END IF
  !set the CoolSetPtResetType indicator
  IF (SameString(FldVal(base + pvavsCoolSetPtResetTypeOff),'None')) THEN
    coolSetPtReset = csprNone
  ELSEIF (SameString(FldVal(base + pvavsCoolSetPtResetTypeOff),'Warmest')) THEN
    coolSetPtReset = csprWarmest
  ELSEIF (SameString(FldVal(base + pvavsCoolSetPtResetTypeOff),'OutdoorAirTemperatureReset')) THEN
    coolSetPtReset = csprOAreset
  ELSEIF (SameString(FldVal(base + pvavsCoolSetPtResetTypeOff),'WarmestTemperatureFirst')) THEN
    coolSetPtReset = csprWarmestTempFirst
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:PackagedVAV in the Cooling Coil Setpoint Reset Type field')
    coolSetPtReset = csprNone
  END IF
  !set the HeatSetPtResetType indicator
  IF (SameString(FldVal(base + pvavsHeatSetPtResetTypeOff),'None')) THEN
    heatSetPtReset = hsprNone
  ELSEIF (SameString(FldVal(base + pvavsHeatSetPtResetTypeOff),'OutdoorAirTemperatureReset')) THEN
    heatSetPtReset = hsprOAreset
!  This setpoint manager type not yet available in EnergyPlus, leave code here for future use
!  ELSEIF (SameString(FldVal(base + pvavsHeatSetPtResetTypeOff),'ColdestTemperatureFirst')) THEN
!    coolSetPtReset = hsprColdestTempFirst
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:PackagedVAV in the Heating Coil Setpoint Reset Type field')
    heatSetPtReset = hsprNone
  END IF
  !set the Supply Fan Part-Load Power Coefficients indicator
  IF (SameString(FldVal(base + pvavsSupplyFanPLRCoeffOff),'InletVaneDampers')) THEN
    supFanCoef = sfcInletVane
  ELSEIF (SameString(FldVal(base + pvavsSupplyFanPLRCoeffOff),'OutletDampers')) THEN
    supFanCoef = sfcDischgDamp
  ELSEIF (SameString(FldVal(base + pvavsSupplyFanPLRCoeffOff),'VariableSpeedMotor')) THEN
    supFanCoef = sfcVSD
  ELSEIF (SameString(FldVal(base + pvavsSupplyFanPLRCoeffOff),'ASHRAE90.1-2004AppendixG')) THEN
    supFanCoef = sfcAppG
  ELSEIF (SameString(FldVal(base + pvavsSupplyFanPLRCoeffOff),'VariableSpeedMotorPressureReset')) THEN
    supFanCoef = sfcVSDPresReset
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:PackagedVAV in the Supply Fan Part-Load Power Coefficients field')
    supFanCoef = sfcInletVane
  END IF
  !set the Return Fan Part-Load Power Coefficients indicator
  IF (SameString(FldVal(base + pvavsReturnFanPLRCoeffOff),'InletVaneDampers')) THEN
    retFanCoef = sfcInletVane
  ELSEIF (SameString(FldVal(base + pvavsReturnFanPLRCoeffOff),'OutletDampers')) THEN
    retFanCoef = sfcDischgDamp
  ELSEIF (SameString(FldVal(base + pvavsReturnFanPLRCoeffOff),'VariableSpeedMotor')) THEN
    retFanCoef = sfcVsD
  ELSEIF (SameString(FldVal(base + pvavsReturnFanPLRCoeffOff),'ASHRAE90.1-2004AppendixG')) THEN
    retFanCoef = sfcAppG
  ELSEIF (SameString(FldVal(base + pvavsReturnFanPLRCoeffOff),'VariableSpeedMotorPressureReset')) THEN
    retFanCoef = sfcVSDPresReset
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:PackagedVAV in the Return Fan Part-Load Power Coefficients field')
    retFanCoef = sfcInletVane
  END IF
  !set the Heat Recovery Type indicator
  IF (SameString(FldVal(base + pvavsHeatRecTypeOff),'Sensible')) THEN
    heatRecovery = htrecSens
  ELSEIF (SameString(FldVal(base + pvavsHeatRecTypeOff),'Enthalpy')) THEN
    heatRecovery = htrecEnth
  ELSEIF (SameString(FldVal(base + pvavsHeatRecTypeOff),'None')) THEN
    heatRecovery = htrecNone
  ELSE
    heatRecovery = htrecNone
    CALL WriteError('Invalid choice in HVACTemplate:System:PackagedVAV in the Heating Coil Setpoint Reset Type field')
  ENDIF
  IF (SameString(FldVal(base + pvavsSupplyFanPlaceOff), 'BlowThrough')) THEN
    supFanPlacement = sfpBlowThru
  ELSEIF (SameString(FldVal(base + pvavsSupplyFanPlaceOff), 'DrawThrough')) THEN
    supFanPlacement = sfpDrawThru
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:PackagedVAV "'//TRIM(FldVal(base + pvavsAirHandlerNameOff))//'"'// &
                    ' in the Supply Fan Placement field: '//TRIM(FldVal(base + pvavsSupplyFanPlaceOff)))
  END IF
  !set the dehumidification indicators
  isDehumidifyNone = SameString(FldVal(base + pvavsDehumCtrlTypeOff),'None')
  IF (SameString(FldVal(base +  pvavsDehumCtrlTypeOff),'None')) THEN
    dehumidCtrlKind = dehumidNone
  ELSEIF (SameString(FldVal(base +  pvavsDehumCtrlTypeOff),'CoolReheat')) THEN
    dehumidCtrlKind = dehumidCoolRht
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:PackagedVAV "'//TRIM(FldVal(base + pvavsAirHandlerNameOff))//'"'// &
                    ' in the Dehumidification Control Type field: '//TRIM(FldVal(base + pvavsDehumCtrlTypeOff)))
  END IF
  !set the humidifier indicators
  isHumidifierNone = SameString(FldVal(base + pvavsHumidCtrlTypeOff),'None')
  IF (SameString(FldVal(base +  pvavsHumidCtrlTypeOff),'None')) THEN
    humidifierKind = humidifyNone
  ELSEIF (SameString(FldVal(base +  pvavsHumidCtrlTypeOff),'ElectricSteam')) THEN
    humidifierKind = humidifyElecSteam
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:PackagedVAV "'//TRIM(FldVal(base + pvavsAirHandlerNameOff))//'"'// &
                    ' in the Humidifier Type field: '//TRIM(FldVal(base + pvavsHumidCtrlTypeOff)))
  END IF
  !set node name for return inlet to OA mixer
  IF (isReturnFanYes) THEN
    returnInletToOAMIxer = TRIM(FldVal(base + pvavsAirHandlerNameOff))//' Return Fan Outlet'
  ELSE
    returnInletToOAMIxer = TRIM(FldVal(base + pvavsAirHandlerNameOff))//' Air Loop Inlet'
  END IF
  !Fix for CR7356
  ! Warn if fixed heating or preheat setpoint is greater than fixed cooling setpoint, and no reset option is active
  IF (isCoolStPtSchedBlank .AND. (coolSetPtReset == csprNone)) THEN
    IF ((heatCoilType .NE. ctNone) .AND. isHeatStPtSchedBlank .AND. (heatSetPtReset == hsprNone)) THEN
      IF (FldVal(base + pvavsHeatSetPtDesignOff) .GT. FldVal(base + pvavsCoolSetPtDesignOff)) THEN
       CALL WriteError('Warning:  In HVACTemplate:System:PackagedVAV "'//TRIM(FldVal(base + pvavsAirHandlerNameOff))//'"'// &
                      ' the Heating Coil Design Setpoint is greater than the Cooling Coil Design Setpoint.'// &
                      ' This may cause the heating coil and cooling coil to operate simultaneously.'// &
                      ' Check results carefully and adjust controls if needed.',msgWarning)
      ENDIF
    ENDIF
  ENDIF
  ! validate sizing option input
  IF (SameString(FldVal(base +  pvavsSizingOptionOff),'Coincident')) THEN
  ELSEIF (SameString(FldVal(base +  pvavsSizingOptionOff),'NonCoincident')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:PackagedVAV "'//TRIM(FldVal(base + pvavsAirHandlerNameOff))//'"'// &
                    ' in the Sizing Option field: '//TRIM(FldVal(base + pvavsSizingOptionOff)))
  END IF
  !Object ==> Sizing:System
  CALL CreateNewObj('Sizing:System')
  CALL AddToObjFld('AirLoop Name', base + pvavsAirHandlerNameOff,' ')
  CALL AddToObjStr('Type of Load to Size On','Sensible')
  CALL AddToObjFld('Design Outdoor Air Flow Rate {m3/s}', base + pvavsMinOutsideFlowOff,' ')
  !This line if vsSupplyMaxRateOff=autosize or vsSupplyMinRateOff=autosize
  IF (isSupplyMaxRateAutosize .OR. isSupplyMinRateAutosize) THEN
    CALL AddToObjStr('Central Heating Maximum System Air Flow Ratio','0.3')
  ELSE
    minAirRatio = StringToReal(FldVal(base + pvavsSupplyMinRateOff)) / &
                  StringToReal(FldVal(base + pvavsSupplyMaxRateOff))
    CALL AddToObjNum('Central Heating Maximum System Air Flow Ratio', minAirRatio)
  END IF
  CALL AddToObjStr('Preheat Design Temperature {C}', '2')
  CALL AddToObjStr('Preheat Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjStr('Precool Design Temperature {C}','11')
  CALL AddToObjStr('Precool Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjFld('Central Cooling Design Supply Air Temperature {C}', base + pvavsCoolSetPtDesignOff,'')
  CALL AddToObjFld('Central Heating Design Supply Air Temperature {C}', base + pvavsHeatSetPtDesignOff,'')
  CALL AddToObjFld('Type of Zone Sum to Use', base + pvavsSizingOptionOff,'')
  CALL AddToObjStr('100% Outdoor Air in Cooling','No')
  CALL AddToObjStr('100% Outdoor Air in Heating','No')
  CALL AddToObjStr('Central Cooling Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjStr('Central Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isSupplyMaxRateAutosize) THEN
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','DesignDay')
    CALL AddToObjStr('Cooling Supply Air Flow Rate {m3/s}','0')
  ELSE
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','Flow/System')
    CALL AddToObjFld('Cooling Supply Air Flow Rate {m3/s}', base + pvavsSupplyMaxRateOff,' ')
  END IF
  ! New fields for v8.2 next three rows
  CALL AddToObjStr('Cooling Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
  CALL AddToObjStr('Cooling Fraction of Autosized Cooling Supply Air Flow Rate','')
  CALL AddToObjStr('Cooling Supply Air Flow Rate Per Unit Cooling Capacity {m3/s-W}','')
  CALL AddToObjStr('Heating Supply Air Flow Rate Method','DesignDay')
  CALL AddToObjStr('Heating Supply Air Flow Rate {m3/s}','0')
  ! New fields for v8.2 next four rows
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
  CALL AddToObjStr('Heating Fraction of Autosized Heating Supply Air Flow Rate','')
  CALL AddToObjStr('Heating Fraction of Autosized Cooling Supply Air Flow Rate','')
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Unit Heating Capacity {m3/s-W}','')
  CALL AddToObjStr('System Outdoor Air Method','ZoneSum')
  CALL AddToObjStr('Zone Maximum Outdoor Air Fraction','1.0')
  CALL AddToObjStr('Cooling Design Capacity Method','CoolingDesignCapacity')
  CALL AddToObjStr('Cooling Design Capacity {W}','autosize')
  CALL AddToObjStr('Cooling Design Capacity Per Floor Area {W/m2}','')
  CALL AddToObjStr('Fraction of Autosized Cooling Design Capacity','')
  CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
  CALL AddToObjStr('Heating Design Capacity {W}','autosize')
  CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
  CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
  CALL AddToObjStr('Central Cooling Capacity Control Method','OnOff',.true.)
  !Object ==> AirLoopHVAC
  CALL CreateNewObj('AirLoopHVAC')
  CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' ')
  IF (heatCoilType .EQ. ctHotWater) THEN
    CALL AddToObjFld('Controller List Name', base + pvavsAirHandlerNameOff,' Controllers')
  ELSE
    CALL AddToObjStr('Controller List Name',' ')
  END IF
  CALL AddToObjFld('Availability Manager List Name', base + pvavsAirHandlerNameOff,' Availability Managers')
  CALL AddToObjStr('Design Supply Air Flow Rate {m3/s}', 'autosize')
  CALL AddToObjFld('Branch List Name', base + pvavsAirHandlerNameOff,' Branches')
  CALL AddToObjStr('Connector List Name','')
  CALL AddToObjFld('Supply Side Inlet Node Name', base + pvavsAirHandlerNameOff,' Air Loop Inlet')
  CALL AddToObjFld('Demand Side Outlet Node Name', base + pvavsAirHandlerNameOff,' Return Air Outlet')
  CALL AddToObjFld('Demand Side Inlet Node Names', base + pvavsAirHandlerNameOff,' Supply Path Inlet')
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    IF (heatCoilType .EQ. ctNone) THEN
      CALL AddToObjFld('Supply Side Outlet Node Names', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet',.TRUE.)
    ELSE
      CALL AddToObjFld('Supply Side Outlet Node Names', base + pvavsAirHandlerNameOff,' Heating Coil Outlet',.TRUE.)
    END IF
  ELSE
    CALL AddToObjFld('Supply Side Outlet Node Names', base + pvavsAirHandlerNameOff,' Supply Fan Outlet',.TRUE.)
  END IF
  !Object ==> BranchList
  CALL CreateNewObj('BranchList')
  CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Branches')
  CALL AddToObjFld('Branch Name', base + pvavsAirHandlerNameOff,' Main Branch',.TRUE.)
  !Object ==> Branch
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Main Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  IF (isReturnFanYes) THEN
    CALL AddToObjStr('Component Object Type','Fan:VariableVolume')
    CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' Return Fan')
    CALL AddToObjFld('Component Inlet Node Name', base + pvavsAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + pvavsAirHandlerNameOff,' Return Fan Outlet')
  END IF
  CALL AddToObjStr('Component Object Type','AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' OA System')
  CALL AddToObjStr('Component Inlet Node Name', TRIM(returnInletToOAMIxer))
  CALL AddToObjFld('Component Outlet Node Name', base + pvavsAirHandlerNameOff,' Mixed Air Outlet')
  IF (humidifierKind .EQ. humidifyElecSteam) THEN
    CALL AddToObjStr('Component Object Type','Humidifier:Steam:Electric')
    CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' Humidifier')
    CALL AddToObjFld('Component Inlet Node Name', base + pvavsAirHandlerNameOff,' Mixed Air Outlet')
    CALL AddToObjFld('Component Outlet Node Name', base + pvavsAirHandlerNameOff,' Humidifier Outlet')
  END IF
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    CALL AddToObjStr('Component Object Type','Fan:VariableVolume')
    CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' Supply Fan')
    IF (isHumidifierNone) THEN
      CALL AddToObjFld('Component Inlet Node Name', base + pvavsAirHandlerNameOff,' Mixed Air Outlet')
    ELSE
      CALL AddToObjFld('Component Inlet Node Name', base + pvavsAirHandlerNameOff,' Humidifier Outlet')
    END IF
    CALL AddToObjFld('Component Outlet Node Name', base + pvavsAirHandlerNameOff,' Supply Fan Outlet')
  END IF
  CALL AddToObjStr('Component Object Type','CoilSystem:Cooling:DX')
  CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' System')
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    CALL AddToObjFld('Component Inlet Node Name', base + pvavsAirHandlerNameOff,' Supply Fan Outlet')
  ELSE
    IF (isHumidifierNone) THEN
      CALL AddToObjFld('Component Inlet Node Name', base + pvavsAirHandlerNameOff,' Mixed Air Outlet')
    ELSE
      CALL AddToObjFld('Component Inlet Node Name', base + pvavsAirHandlerNameOff,' Humidifier Outlet')
    END IF
  END IF
  IF ((heatCoilType .EQ. ctNone) .AND. (supFanPlacement .NE. sfpDrawThru)) THEN
    CALL AddToObjFld('Component Outlet Node Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet',.TRUE.)
  ELSE
    CALL AddToObjFld('Component Outlet Node Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet')
  END IF
  IF (heatCoilType .NE. ctNone) THEN
    IF (heatCoilType .EQ. ctHotWater) THEN
      CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    ELSEIF (heatCoilType .EQ. ctElectric) THEN
      CALL AddToObjStr('Component Object Type','Coil:Heating:Electric')
    ELSEIF (heatCoilType .EQ. ctGas) THEN
      CALL AddToObjStr('Component Object Type','Coil:Heating:Fuel')
    END IF
    CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet')
    IF (supFanPlacement .EQ. sfpDrawThru) THEN
     CALL AddToObjFld('Component Outlet Node Name', base + pvavsAirHandlerNameOff,' Heating Coil Outlet')
    ELSE
      CALL AddToObjFld('Component Outlet Node Name', base + pvavsAirHandlerNameOff,' Heating Coil Outlet',.TRUE.)
    END IF
  END IF
  IF (supFanPlacement .EQ. sfpDrawThru) THEN
    CALL AddToObjStr('Component Object Type','Fan:VariableVolume')
    CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' Supply Fan')
    IF (heatCoilType .NE. ctNone) THEN
      CALL AddToObjFld('Component Inlet Node Name', base + pvavsAirHandlerNameOff,' Heating Coil Outlet')
    ELSE
      CALL AddToObjFld('Component Inlet Node Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet')
    END IF
    CALL AddToObjFld('Component Outlet Node Name', base + pvavsAirHandlerNameOff,' Supply Fan Outlet',.TRUE.)
  END IF
  IF (isSupplyPlenumBlank) THEN
    !Object ==> AirLoopHVAC:SupplyPath
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + pvavsAirHandlerNameOff,' Supply Path Inlet')
    CALL MakeSpecialIntList(zlVAVAllSupPlen, base + pvavsAirHandlerNameOff)
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneSplitter')
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' Zone Splitter',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' Zone Splitter')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Supply Plenum', .TRUE.)
      END IF
    END IF
    !Object ==> AirLoopHVAC:ZoneSplitter
    CALL CreateNewObj('AirLoopHVAC:ZoneSplitter')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Zone Splitter')
    CALL AddToObjFld('Inlet Node Name', base + pvavsAirHandlerNameOff,' Supply Path Inlet')
    CALL MakeSpecialIntList(zlVavAll,base + pvavsAirHandlerNameOff)
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(i),' Zone Equip Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Inlet',.TRUE.)
    END IF
  ELSE
    !Object ==> AirLoopHVAC:SupplyPath
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + pvavsAirHandlerNameOff,' Supply Path Inlet')
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
    CALL MakeSpecialIntList(zlVAVAllSupPlen, base + pvavsAirHandlerNameOff)
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' Supply Plenum',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' Supply Plenum')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Supply Plenum', .TRUE.)
      END IF
    END IF
    !Object ==> AirLoopHVAC:SupplyPlenum
    CALL CreateNewObj('AirLoopHVAC:SupplyPlenum')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Supply Plenum')
    CALL AddToObjFld('Zone Name', base + pvavsSupplyPlenumNameOff,' ')
    CALL AddToObjFld('Zone Node Name', base + pvavsSupplyPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Inlet Node Name', base + pvavsAirHandlerNameOff,' Supply Path Inlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlVavAll,base + pvavsAirHandlerNameOff)
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(i),' Zone Equip Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Inlet',.TRUE.)
    END IF
  END IF
  IF (isReturnPlenumBlank) THEN
    ! Check if any fan powered zones are requesting induced air from system plenum
    CALL MakeSpecialIntList(zlVavFPfromSysRetPlen, base + pvavsAirHandlerNameOff)
      !the IntList holds the list of pointers to the zone names for induced air nodes
    IF (numIntList .NE. 0) THEN
      CALL WriteError('In HVACTemplate:System:VAV "'//TRIM(FldVal(base + pvavsAirHandlerNameOff))//'"'// &
                      ' there are fan powered zones requesting secondary air from the system return plenum'// &
                      ' but the Return Plenum Name is blank.')
    END IF
    !Object ==> AirLoopHVAC:ReturnPath
    CALL CreateNewObj('AirLoopHVAC:ReturnPath')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Return Path')
    CALL AddToObjFld('Return Air Path Outlet Node Name', base + pvavsAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlVavAllRetPlen, base + pvavsAirHandlerNameOff)
    IF (numIntList .NE. 0) THEN
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Return Plenum')
      END DO
    END IF
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneMixer')
    CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' Zone Mixer',.TRUE.)
    !Object ==> AirLoopHVAC:ZoneMixer
    CALL CreateNewObj('AirLoopHVAC:ZoneMixer')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Zone Mixer')
    CALL AddToObjFld('Outlet Node Name', base + pvavsAirHandlerNameOff,' Return Air Outlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlVavAll,base + pvavsAirHandlerNameOff)
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Inlet Node Name', IntList(i),' Return Outlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Inlet Node Name', IntList(numIntList),' Return Outlet',.TRUE.)
    END IF
  ELSE
    !Object ==> AirLoopHVAC:ReturnPath
    CALL CreateNewObj('AirLoopHVAC:ReturnPath')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Return Path')
    CALL AddToObjFld('Return Air Path Outlet Node Name', base + pvavsAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlVavAllRetPlen, base + pvavsAirHandlerNameOff)
    IF (numIntList .NE. 0) THEN
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Return Plenum')
      END DO
    END IF
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' Return Plenum',.TRUE.)
    !Object ==> AirLoopHVAC:ReturnPlenum
    CALL CreateNewObj('AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Return Plenum')
    CALL AddToObjFld('Zone Name', base + pvavsReturnPlenumNameOff,' ')
    CALL AddToObjFld('Zone Node Name', base + pvavsReturnPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Outlet Node Name', base + pvavsAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlVavFPfromSysRetPlen, base + pvavsAirHandlerNameOff)
      !the IntList holds the list of pointers to the zone names for induced air nodes
    IF (numIntList .NE. 0) THEN
      CALL AddToObjFld('Induced Air Outlet Node or NodeList Name', base + pvavsAirHandlerNameOff,' Induced Air Nodes')
    ELSE
      CALL AddToObjStr('Induced Air Outlet Node or NodeList Name','')
    END IF
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlVavAll,base + pvavsAirHandlerNameOff)
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Inlet Node Name', IntList(i),' Return Outlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Inlet Node Name', IntList(numIntList),' Return Outlet',.TRUE.)
    END IF
  END IF

    !Object ==> NodeList for induced air nodes
  CALL MakeSpecialIntList(zlVavFPfromSysRetPlen, base + pvavsAirHandlerNameOff)
  !the IntList holds the list of pointers to the zone names for induced air nodes
  IF (numIntList .NE. 0) THEN
    CALL CreateNewObj('NodeList')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Induced Air Nodes')
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Node Name', IntList(i),' PIU Secondary Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Node Name', IntList(numIntList),' PIU Secondary Inlet',.TRUE.)
    END IF
  END IF

  !Object ==> AvailabilityManagerAssignmentList
  CALL CreateNewObj('AvailabilityManagerAssignmentList')
  CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Availability Managers')
  CALL AddToObjStr('Availability Manager Object Type','AvailabilityManager:NightCycle')
  CALL AddToObjFld('Availability Manager Name', base + pvavsAirHandlerNameOff,' Availability',.TRUE.)
  !Object ==> AvailabilityManager:NightCycle
  CALL CreateNewObj('AvailabilityManager:NightCycle')
  CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Availability')
  CALL AddToObjStr('Applicability Schedule Name','HVACTemplate-Always 1')
  IF (isSysAvailSchedBlank) THEN
    CALL AddToObjStr('Fan Schedule Name','HVACTemplate-Always 1')
  ELSE
    CALL AddToObjFld('Fan Schedule Name', base + pvavsSysAvailSchedNameOff,'')
  END IF
  CALL AddToObjFld('Control Type', base + pvavsNightCycleControlOff,'')
  CALL AddToObjStr('Thermostat Tolerance {deltaC}','0.2')
  CALL AddToObjStr('Cycling Run Time Control Type', 'FixedRunTime')
  CALL AddToObjStr('Cycling run time {s}','3600')
  CALL AddToObjFld('Control Zone Name', base + pvavsNightCycleCtrlZoneNameOff,'',.TRUE.)
  CALL AddAlwaysSchedule('1')
  SELECT CASE (coolSetPtReset)
    CASE (csprNone)
      IF (isCoolStPtSchedBlank) THEN
        CALL AddAlwaysSchedule(FldVal(base + pvavsCoolSetPtDesignOff))
      END IF
      !Object ==> SetpointManager:Scheduled
      CALL CreateNewObj('SetpointManager:Scheduled')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      IF (isCoolStPtSchedBlank) THEN
        CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + pvavsCoolSetPtDesignOff)))
      ELSE
        CALL AddToObjFld('Schedule Name', base + pvavsCoolSetPtSchedNameOff,' ')
      END IF
      IF (supFanPlacement .EQ. sfpBlowThru) THEN
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet',.TRUE.)
      ELSE
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Supply Fan Outlet',.TRUE.)
      END IF
    CASE (csprWarmest)
      !Object ==> SetpointManager:Warmest
      CALL CreateNewObj('SetpointManager:Warmest')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('HVAC Air Loop Name', base + pvavsAirHandlerNameOff,'')
      CALL AddToObjFld('Minimum Setpoint Temperature {C}', base + pvavsCoolSetPtDesignOff,'')
      IF ((StringToReal(FldVal(base + pvavsHeatSetPtDesignOff)) + 1.0) .GT. 18.0) THEN
        CALL AddToObjNum('Maximum Setpoint Temperature {C}',StringToReal(FldVal(base + pvavsHeatSetPtDesignOff)) + 1.0)
      ELSE
        CALL AddToObjStr('Maximum Setpoint Temperature {C}','18')
      END IF
      CALL AddToObjStr('Strategy','MaximumTemperature')
      IF (supFanPlacement .EQ. sfpBlowThru) THEN
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet',.TRUE.)
      ELSE
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Supply Fan Outlet',.TRUE.)
      END IF
    CASE (csprWarmestTempFirst)
      !Object ==> SetpointManager:WarmestTemperatureFlow
      CALL CreateNewObj('SetpointManager:WarmestTemperatureFlow')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('HVAC Air Loop Name', base + pvavsAirHandlerNameOff,'')
      CALL AddToObjFld('Minimum Setpoint Temperature {C}', base + pvavsCoolSetPtDesignOff,'')
      IF ((StringToReal(FldVal(base + pvavsHeatSetPtDesignOff)) + 1.0) .GT. 18.0) THEN
        CALL AddToObjNum('Maximum Setpoint Temperature {C}',StringToReal(FldVal(base + pvavsHeatSetPtDesignOff)) + 1.0)
      ELSE
        CALL AddToObjStr('Maximum Setpoint Temperature {C}','18')
      END IF
      CALL AddToObjStr('Strategy','TemperatureFirst')
      IF (supFanPlacement .EQ. sfpBlowThru) THEN
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet')
      ELSE
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Supply Fan Outlet')
      END IF
      IF (isSupplyMaxRateAutosize .OR. isSupplyMinRateAutosize) THEN
        CALL AddToObjStr('Minimum Turndown Ratio','0.2',.TRUE.)
      ELSE
        minAirRatio = StringToReal(FldVal(base + pvavsSupplyMinRateOff)) / &
                      StringToReal(FldVal(base + pvavsSupplyMaxRateOff))
        CALL AddToObjNum('Minimum Turndown Ratio', minAirRatio,.TRUE.)
      END IF
    CASE (csprOAreset)
      !Object ==> SetpointManager:OutdoorAirReset
      CALL CreateNewObj('SetpointManager:OutdoorAirReset')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjNum('Set point at outside low {C}',StringToReal(FldVal(base + pvavsCoolSetPtDesignOff)) + 5.2)
      CALL AddToObjStr('Outside low {C}','15.6')
      CALL AddToObjFld('Set point at outside high {C}', base + pvavsCoolSetPtDesignOff,'')
      CALL AddToObjStr('Outside high {C}','26.7')
      IF (supFanPlacement .EQ. sfpBlowThru) THEN
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet',.TRUE.)
      ELSE
        CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Supply Fan Outlet',.TRUE.)
      END IF
  END SELECT
  !Object ==> CoilSystem:Cooling:DX
  CALL CreateNewObj('CoilSystem:Cooling:DX')
  CALL AddToObjFld('Name',base + pvavsAirHandlerNameOff,' System')
  CALL AddToObjFld('Availability Schedule Name', base + pvavsCoolAvailSchedNameOff,' ')
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    CALL AddToObjFld('DX Cooling Coil System Inlet Node Name', base + pvavsAirHandlerNameOff,' Supply Fan Outlet')
  ELSE
    IF (isHumidifierNone) THEN
      CALL AddToObjFld('DX Cooling Coil System Inlet Node Name', base + pvavsAirHandlerNameOff,' Mixed Air Outlet')
    ELSE
      CALL AddToObjFld('DX Cooling Coil System Inlet Node Name', base + pvavsAirHandlerNameOff,' Humidifier Outlet')
    END IF
  END IF
  CALL AddToObjFld('DX Cooling Coil System Outlet Node Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet')
  CALL AddToObjFld('DX Cooling Coil System Sensor Node Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet')
  SELECT CASE (coolCoilKind)
    CASE(ccTwoSpeedDX)
      CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:DX:TwoSpeed')
    CASE(ccTwoSpeedHumidControlDX)
      CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:DX:TwoStageWithHumidityControlMode')
!    CASE(ccThreeSpeedDX)
!      CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:DX:MultiSpeed')
!    CASE(ccFourSpeedDX)
!      CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:DX:MultiSpeed')
  END SELECT
  CALL AddToObjFld('Cooling Coil Name',base + pvavsAirHandlerNameOff,' Cooling Coil')
  IF (dehumidCtrlKind .EQ. dehumidCoolRht) THEN
    CALL AddToObjStr('Dehumidification Control Type','CoolReheat')
  ELSE
    CALL AddToObjStr('Dehumidification Control Type','None')
  END IF
  CALL AddToObjStr('Run on Sensible Load','Yes')
  ! CR8813 - MJW - Delete isHumidifierNone IF (.NOT. isDehumidifyNone .OR. .NOT. isHumidifierNone) THEN
  IF (.NOT. isDehumidifyNone) THEN
    CALL AddToObjStr('Run on Latent Load','Yes',.TRUE.)
  ELSE
    CALL AddToObjStr('Run on Latent Load','No',.TRUE.)
  END IF
  SELECT CASE (coolCoilKind)
    CASE(ccTwoSpeedDX)
      !Object ==> Coil:Cooling:DX:TwoSpeed
      CALL CreateNewObj('Coil:Cooling:DX:TwoSpeed')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + pvavsCoolAvailSchedNameOff,' ')
      CALL AddToObjFld('Rated High Speed Total Cooling Capacity', base + pvavsCoolCoilCapOff, '')
      CALL AddToObjFld('Rated High Speed Sensible Heat Ratio', base + pvavsCoolCoilSHROff, '')
      CALL AddToObjFld('Rated High Speed COP', base + pvavsCoolCoilCOPOff, '')
      CALL AddToObjStr('Rated High Speed Air Flow Rate', 'autosize')
      ! DX Ratings - Jul 2012
      CALL AddToObjStr('Unit Internal Static Air Pressure','')
      IF (supFanPlacement .EQ. sfpBlowThru) THEN
        CALL AddToObjFld('Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Supply Fan Outlet')
      ELSE
        IF (isHumidifierNone) THEN
          CALL AddToObjFld('Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Mixed Air Outlet')
        ELSE
          CALL AddToObjFld('Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Humidifier Outlet')
        END IF
      END IF
      CALL AddToObjFld('Air Outlet Node Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet')
      CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Cap-FT')
      CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Cap-FF')
      CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil EIR-FT')
      CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil EIR-FF')
      CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil PLF')
      IF (isCoolCapAutosize) THEN
        CALL AddToObjStr('Rated Low Speed Total Cooling Capacity','Autosize')
      ELSE
        ! If high speed capacity is specified, assume that low speed is 1/3 of high speed capacity, same assumption that autosizing uses
        CALL AddToObjNum('Rated Low Speed Total Cooling Capacity',StringToReal(FldVal(base + pvavsCoolCoilCapOff)) * 0.33)
      END IF
      ! assume low speed sensible heat ratio same as high speed
      CALL AddToObjFld('Rated Low Speed Sensible Heat Ratio', base + pvavsCoolCoilSHROff, '')
      CALL AddToObjNum('Rated Low Speed COP',StringToReal(FldVal(base + pvavsCoolCoilCOPOff)) * 1.5)
      ! assume low speed air flow rate is automatically sized
      CALL AddToObjStr('Rated Low Speed Air Flow Rate','Autosize')
      ! assume the same curves as full speed
      CALL AddToObjFld('Low Speed Total Cooling Capacity Function of Temperature Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Low Cap-FT')
      CALL AddToObjFld('Low Speed Energy Input Ratio Function of Temperature Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Low EIR-FT',.TRUE.)
      !Object ==> Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('DOE-2.1E, SDL-C4')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' DX Coil Cap-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.476428')
      CALL AddToObjStr('Coefficient2 x','0.401147E-01')
      CALL AddToObjStr('Coefficient3 x**2','0.226411E-03')
      CALL AddToObjStr('Coefficient4 y','-0.827136E-03')
      CALL AddToObjStr('Coefficient5 y**2','-0.732240E-05')
      CALL AddToObjStr('Coefficient6 x*y','-0.446278E-03')
      CALL AddToObjStr('Minimum Value of x','12.77778')
      CALL AddToObjStr('Maximum Value of x','23.88889')
      CALL AddToObjStr('Minimum Value of y','23.88889')
      CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
      !Object ==> Curve:Cubic
      CALL CreateNewObj('Curve:Cubic')
      CALL WriteComment('DOE-2.1E, SDL-C78')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' DX Coil Cap-FF')
      CALL AddToObjStr('Coefficient1 Constant','.47278589')
      CALL AddToObjStr('Coefficient2 x','1.2433415')
      CALL AddToObjStr('Coefficient3 x**2','-1.0387055')
      CALL AddToObjStr('Coefficient4 x**3','.32257813')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !Object ==> Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('DOE-2.1E, SDL-C82')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' DX Coil EIR-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.632475')
      CALL AddToObjStr('Coefficient2 x',' -0.121321E-01')
      CALL AddToObjStr('Coefficient3 x**2','0.507773E-03')
      CALL AddToObjStr('Coefficient4 y','0.155377E-01')
      CALL AddToObjStr('Coefficient5 y**2','0.272840E-03')
      CALL AddToObjStr('Coefficient6 x*y','-0.679201E-03')
      CALL AddToObjStr('Minimum Value of x','12.77778')
      CALL AddToObjStr('Maximum Value of x','23.88889')
      CALL AddToObjStr('Minimum Value of y','23.88889')
      CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
      !Object ==> Curve:Cubic
      CALL CreateNewObj('Curve:Cubic')
      CALL WriteComment('DOE-2.1E, SDL-C82')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' DX Coil EIR-FF')
      CALL AddToObjStr('Coefficient1 Constant','1.0079484')
      CALL AddToObjStr('Coefficient2 x','.34544129')
      CALL AddToObjStr('Coefficient3 x**2','-.6922891')
      CALL AddToObjStr('Coefficient4 x**3','.33889943')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !Object ==> Curve:Quadratic
      CALL CreateNewObj('Curve:Quadratic')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' DX Coil PLF')
      CALL AddToObjStr('Coefficient1 Constant','0.85')
      CALL AddToObjStr('Coefficient2 x','0.15')
      CALL AddToObjStr('Coefficient3 x**2','0.0')
      CALL AddToObjStr('Minimum Value of x','0.0')
      CALL AddToObjStr('Maximum Value of x','1.0',.TRUE.)
      !Object ==> Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('DOE-2.1E, SDL-C4')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' DX Coil Low Cap-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.476428E+00')
      CALL AddToObjStr('Coefficient2 x','0.401147E-01')
      CALL AddToObjStr('Coefficient3 x**2','0.226411E-03')
      CALL AddToObjStr('Coefficient4 y','-0.827136E-03')
      CALL AddToObjStr('Coefficient5 y**2','-0.732240E-05')
      CALL AddToObjStr('Coefficient6 x*y','-0.446278E-03')
      CALL AddToObjStr('Minimum Value of x','12.77778')
      CALL AddToObjStr('Maximum Value of x','23.88889')
      CALL AddToObjStr('Minimum Value of y','23.88889')
      CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
      !Object ==> Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' DX Coil Low EIR-FT')
      !fixed below from CR7661
      CALL AddToObjStr('Coefficient1 Constant','0.774645E+00')
      CALL AddToObjStr('Coefficient2 x','-0.343731E-01')
      CALL AddToObjStr('Coefficient3 x**2','0.783173E-03')
      CALL AddToObjStr('Coefficient4 y','0.146596E-01')
      CALL AddToObjStr('Coefficient5 y**2','0.488851E-03')
      CALL AddToObjStr('Coefficient6 x*y','-0.752036E-03')
      CALL AddToObjStr('Minimum Value of x','12.77778')
      CALL AddToObjStr('Maximum Value of x','23.88889')
      CALL AddToObjStr('Minimum Value of y','23.88889')
      CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
    CASE(ccTwoSpeedHumidControlDX)
      !Object ==> Coil:Cooling:DX:TwoStageWithHumidityControlMode
      CALL CreateNewObj('Coil:Cooling:DX:TwoStageWithHumidityControlMode')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + pvavsCoolAvailSchedNameOff,' ')
      IF (supFanPlacement .EQ. sfpBlowThru) THEN
        CALL AddToObjFld('Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Supply Fan Outlet')
      ELSE
        IF (isHumidifierNone) THEN
          CALL AddToObjFld('Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Mixed Air Outlet')
        ELSE
          CALL AddToObjFld('Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Humidifier Outlet')
        END IF
      END IF
      CALL AddToObjFld('Air Outlet Node Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet')
      CALL AddToObjStr('Crankcase Heater Capacity','')
      CALL AddToObjStr('Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater','')
      CALL AddToObjStr('Number of Capacity Stages','2')
      CALL AddToObjStr('Number of Enhanced Dehumidification Modes','1')
      CALL AddToObjStr('Normal Mode Stage 1 Coil Performance Object Type','CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Normal Mode Stage 1 Coil Performance Name',base + pvavsAirHandlerNameOff,' Standard Perf 1')
      CALL AddToObjStr('Normal Mode Stage 1+2 Coil Performance Object Type','CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Normal Mode Stage 1+2 Coil Performance Name',base + pvavsAirHandlerNameOff,' Standard Perf 1+2')
      CALL AddToObjStr('Dehumidification Mode 1 Stage 1 Coil Performance Object Type','CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Dehumidification Mode 1 Stage 1 Coil Performance Name',base + pvavsAirHandlerNameOff,' Dehumid Perf 1')
      CALL AddToObjStr('Dehumidification Mode 1 Stage 1+2 Coil Performance Object Type','CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Dehumidification Mode 1 Stage 1+2 Coil Performance Name',base + pvavsAirHandlerNameOff,' Dehumid Perf 1+2')
      CALL AddToObjStr('Supply Water Storage Tank Name','')
      CALL AddToObjStr('Condensate Collection Water Storage Tank Name','',.TRUE.)
      !Object ==>    CoilPerformance:DX:Cooling
      ! Standard - Stage 1
      CALL CreateNewObj('CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Name',base + pvavsAirHandlerNameOff,' Standard Perf 1')
      IF (isCoolCapAutosize) THEN
        CALL AddToObjStr('Rated Total Cooling Capacity', 'autosize')
      ELSE
        ! half the capacity if provided
        CALL AddToObjNum('Rated Total Cooling Capacity', 0.5 * StringToReal(FldVal(base + pvavsCoolCoilCapOff)))
      END IF
      CALL AddToObjFld('Gross Rated Sensible Heat Ratio', base + pvavsCoolCoilSHROff, '')
      CALL AddToObjFld('Rated COP', base + pvavsCoolCoilCOPOff, '')
      CALL AddToObjStr('Rated Air Flow Rate', 'autosize')
      CALL AddToObjStr('Fraction of Air Flow Bypassed Around Coil', '0.5')
      CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Std Cap-FT')
      CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Cap-FF')
      CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Std EIR-FT')
      CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil EIR-FF')
      CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil PLF')
      CALL AddToObjStr('Nominal Time for Condensate Removal to Begin', '')
      CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity', '')
      CALL AddToObjStr('Maximum Cycling Rate', '')
      CALL AddToObjStr('Latent Capacity Time Constant', '',.TRUE.)
      !Object ==>    CoilPerformance:DX:Cooling
      ! Standard - Stage 1 + 2
      CALL CreateNewObj('CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Name',base + pvavsAirHandlerNameOff,' Standard Perf 1+2')
      CALL AddToObjFld('Rated Total Cooling Capacity', base + pvavsCoolCoilCapOff,'')
      CALL AddToObjFld('Gross Rated Sensible Heat Ratio', base + pvavsCoolCoilSHROff, '')
      CALL AddToObjFld('Rated COP', base + pvavsCoolCoilCOPOff, '')
      CALL AddToObjStr('Rated Air Flow Rate', 'autosize')
      CALL AddToObjStr('Fraction of Air Flow Bypassed Around Coil', '0.0')
      CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Std Cap-FT')
      CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Cap-FF')
      CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Std EIR-FT')
      CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil EIR-FF')
      CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil PLF')
      CALL AddToObjStr('Nominal Time for Condensate Removal to Begin', '')
      CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity', '')
      CALL AddToObjStr('Maximum Cycling Rate', '')
      CALL AddToObjStr('Latent Capacity Time Constant', '',.TRUE.)
      !Object ==>    CoilPerformance:DX:Cooling
      ! Dehumidification (subcool) - Stage 1
      CALL CreateNewObj('CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Name',base + pvavsAirHandlerNameOff,' Dehumid Perf 1')
      IF (isCoolCapAutosize) THEN
        CALL AddToObjStr('Rated Total Cooling Capacity', 'autosize')
      ELSE
        ! half the capacity if provided
        CALL AddToObjNum('Rated Total Cooling Capacity', 0.5 * StringToReal(FldVal(base + pvavsCoolCoilCapOff)))
      END IF
      IF (isSHRAutosize) THEN
        CALL AddToObjStr('Gross Rated Sensible Heat Ratio','autosize')
      ELSE
        CALL AddToObjNum('Gross Rated Sensible Heat Ratio', 0.9 * StringToReal(FldVal(base + pvavsCoolCoilSHROff)))
      ENDIF
      CALL AddToObjNum('Rated COP', 0.9 * StringToReal(FldVal(base + pvavsCoolCoilCOPOff)))
      CALL AddToObjStr('Rated Air Flow Rate', 'autosize')
      CALL AddToObjStr('Fraction of Air Flow Bypassed Around Coil', '0.5')
      CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Dehum Cap-FT')
      CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Cap-FF')
      CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Dehum EIR-FT')
      CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil EIR-FF')
      CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil PLF')
      CALL AddToObjStr('Nominal Time for Condensate Removal to Begin', '')
      CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity', '')
      CALL AddToObjStr('Maximum Cycling Rate', '')
      CALL AddToObjStr('Latent Capacity Time Constant', '',.TRUE.)
      !Object ==>    CoilPerformance:DX:Cooling
      ! Dehumidification (subcool) - Stage 1 + 2
      CALL CreateNewObj('CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Name',base + pvavsAirHandlerNameOff,' Dehumid Perf 1+2')
      CALL AddToObjFld('Rated Total Cooling Capacity', base + pvavsCoolCoilCapOff,'')
      IF (isSHRAutosize) THEN
        CALL AddToObjStr('Gross Rated Sensible Heat Ratio','autosize')
      ELSE
        CALL AddToObjNum('Gross Rated Sensible Heat Ratio', 0.9 * StringToReal(FldVal(base + pvavsCoolCoilSHROff)))
      ENDIF
      CALL AddToObjNum('Rated COP', 0.9 * StringToReal(FldVal(base + pvavsCoolCoilCOPOff)))
      CALL AddToObjStr('Rated Air Flow Rate', 'autosize')
      CALL AddToObjStr('Fraction of Air Flow Bypassed Around Coil', '0.0')
      CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Dehum Cap-FT')
      CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Cap-FF')
      CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil Dehum EIR-FT')
      CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil EIR-FF')
      CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
         base + pvavsAirHandlerNameOff,' DX Coil PLF')
      CALL AddToObjStr('Nominal Time for Condensate Removal to Begin', '')
      CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity', '')
      CALL AddToObjStr('Maximum Cycling Rate', '')
      CALL AddToObjStr('Latent Capacity Time Constant', '',.TRUE.)
      !Object ==> Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' DX Coil Std Cap-FT')
      CALL AddToObjStr('Coefficient1 Constant','1.067939449')
      CALL AddToObjStr('Coefficient2 x','-0.031261829')
      CALL AddToObjStr('Coefficient3 x**2','0.001974308')
      CALL AddToObjStr('Coefficient4 y','-0.002726426')
      CALL AddToObjStr('Coefficient5 y**2','-5.52654E-05')
      CALL AddToObjStr('Coefficient6 x*y','-6.31169E-05')
      CALL AddToObjStr('Minimum Value of x','-100')
      CALL AddToObjStr('Maximum Value of x','100')
      CALL AddToObjStr('Minimum Value of y','-100')
      CALL AddToObjStr('Maximum Value of y','100',.TRUE.)
      !Object ==> Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' DX Coil Dehum Cap-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.596779741')
      CALL AddToObjStr('Coefficient2 x','0.034216637')
      CALL AddToObjStr('Coefficient3 x**2','0.000113924')
      CALL AddToObjStr('Coefficient4 y','-0.00375859')
      CALL AddToObjStr('Coefficient5 y**2','-9.17495E-05')
      CALL AddToObjStr('Coefficient6 x*y','-8.98373E-05')
      CALL AddToObjStr('Minimum Value of x','-100')
      CALL AddToObjStr('Maximum Value of x','100')
      CALL AddToObjStr('Minimum Value of y','-100')
      CALL AddToObjStr('Maximum Value of y','100',.TRUE.)
      !Object ==> Curve:Quadratic
      CALL CreateNewObj('Curve:Quadratic')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' DX Coil Cap-FF')
      CALL AddToObjStr('Coefficient1 Constant','0.8')
      CALL AddToObjStr('Coefficient2 x','0.2')
      CALL AddToObjStr('Coefficient3 x**2','0.0')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !Object ==> Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' DX Coil Std EIR-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.174059889')
      CALL AddToObjStr('Coefficient2 x','0.022281508')
      CALL AddToObjStr('Coefficient3 x**2','-0.000134077')
      CALL AddToObjStr('Coefficient4 y','0.028298025')
      CALL AddToObjStr('Coefficient5 y**2','0.00048510')
      CALL AddToObjStr('Coefficient6 x*y','-0.001677095')
      CALL AddToObjStr('Minimum Value of x','-100')
      CALL AddToObjStr('Maximum Value of x','100')
      CALL AddToObjStr('Minimum Value of y','-100')
      CALL AddToObjStr('Maximum Value of y','100',.TRUE.)
      !Object ==> Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' DX Coil Dehum EIR-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.435347586')
      CALL AddToObjStr('Coefficient2 x','0.004015641')
      CALL AddToObjStr('Coefficient3 x**2','0.000604235')
      CALL AddToObjStr('Coefficient4 y','0.015824043')
      CALL AddToObjStr('Coefficient5 y**2','0.000747287')
      CALL AddToObjStr('Coefficient6 x*y','-0.001779745')
      CALL AddToObjStr('Minimum Value of x','-100')
      CALL AddToObjStr('Maximum Value of x','100')
      CALL AddToObjStr('Minimum Value of y','-100')
      CALL AddToObjStr('Maximum Value of y','100',.TRUE.)
      !Object ==> Curve:Quadratic
      CALL CreateNewObj('Curve:Quadratic')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' DX Coil EIR-FF')
      CALL AddToObjStr('Coefficient1 Constant','1.1552')
      CALL AddToObjStr('Coefficient2 x','-0.1808')
      CALL AddToObjStr('Coefficient3 x**2','0.0256')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !Object ==> Curve:Quadratic
      CALL CreateNewObj('Curve:Quadratic')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' DX Coil PLF')
      CALL AddToObjStr('Coefficient1 Constant','0.85')
      CALL AddToObjStr('Coefficient2 x','0.15')
      CALL AddToObjStr('Coefficient3 x**2','0.0')
      CALL AddToObjStr('Minimum Value of x','0.0')
      CALL AddToObjStr('Maximum Value of x','1.0',.TRUE.)
  END SELECT
  ! Humdistat(s) if needed
  !    Single humidistat if humidification and dehumidification are both active and control zones are the same
  IF ((.NOT. isHumidifierNone) .AND. (.NOT. isDehumidifyNone) &
       .AND. SameString(FldVal(base +  pvavsDehumCtrlZoneOff),FldVal(base +  pvavsHumidCtrlZoneOff))) THEN
    !Object ==> ZoneControl:Humidistat
    CALL CreateNewObj('ZoneControl:Humidistat')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Humidistat')
    CALL AddToObjFld('Zone Name', base + pvavsDehumCtrlZoneOff,'')
    CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name', &
                     'HVACTemplate-Always ' // TRIM(FldVal(base + pvavsHumidSetPtOff)))
    CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name', &
                     'HVACTemplate-Always ' // TRIM(FldVal(base + pvavsDehumSetPtOff)),.TRUE.)
    CALL AddAlwaysSchedule(FldVal(base + pvavsHumidSetPtOff))
    CALL AddAlwaysSchedule(FldVal(base + pvavsDehumSetPtOff))
  ELSE
    IF (.NOT. isDehumidifyNone) THEN
      !   Dehumidification humidistat
      !Object ==> ZoneControl:Humidistat
      CALL CreateNewObj('ZoneControl:Humidistat')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Dehumidification Humidistat')
      CALL AddToObjFld('Zone Name', base + pvavsDehumCtrlZoneOff,'')
      CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name','HVACTemplate-Always 1')
      CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name', &
                       'HVACTemplate-Always ' // TRIM(FldVal(base + pvavsDehumSetPtOff)),.TRUE.)

      CALL AddAlwaysSchedule('1')
      CALL AddAlwaysSchedule(FldVal(base + pvavsDehumSetPtOff))
    ENDIF
    IF (.NOT. isHumidifierNone) THEN
      !    Humidification humidistat
      !Object ==> ZoneControl:Humidistat
      CALL CreateNewObj('ZoneControl:Humidistat')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Humidification Humidistat')
      CALL AddToObjFld('Zone Name', base + pvavsHumidCtrlZoneOff,'')
      CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name', &
                       'HVACTemplate-Always ' // TRIM(FldVal(base + pvavsHumidSetPtOff)))
      CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name','HVACTemplate-Always 100',.TRUE.)

      CALL AddAlwaysSchedule(FldVal(base + pvavsHumidSetPtOff))
      CALL AddAlwaysSchedule('100')
    ENDIF
  ENDIF
  !Set point manager for dehumidification
  IF (.NOT. isDehumidifyNone) THEN
    !Object ==> SetpointManager:SingleZone:Humidity:Maximum
    CALL CreateNewObj('SetpointManager:SingleZone:Humidity:Maximum')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Dehumidification Setpoint Manager')
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet')
    CALL AddToObjFld('Control Zone Air Node Name', base + pvavsDehumCtrlZoneOff,' Zone Air Node',.TRUE.)
  END IF
  !Object ==> SetpointManager:MixedAir
  CALL CreateNewObj('SetpointManager:MixedAir')
  CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Cooling Coil Air Temp Manager')
  CALL AddToObjStr('Control Variable','Temperature')
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    CALL AddToObjFld('Reference Setpoint Node Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet')
    IF (isHumidifierNone) THEN
      CALL AddToObjFld('Fan Inlet Node Name', base + pvavsAirHandlerNameOff,' Mixed Air Outlet')
    ELSE
      CALL AddToObjFld('Fan Inlet Node Name', base + pvavsAirHandlerNameOff,' Humidifier Outlet')
    END IF
  ELSE
    CALL AddToObjFld('Reference Setpoint Node Name', base + pvavsAirHandlerNameOff,' Supply Fan Outlet')
    IF (heatCoilType .NE. ctNone) THEN
      CALL AddToObjFld('Fan Inlet Node Name', base + pvavsAirHandlerNameOff,' Heating Coil Outlet')
    ELSE
      CALL AddToObjFld('Fan Inlet Node Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet')
    END IF
  END IF
  CALL AddToObjFld('Fan Outlet Node Name', base + pvavsAirHandlerNameOff,' Supply Fan Outlet')
  CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Mixed Air Nodes',.TRUE.)
  !Object ==> NodeList
  CALL CreateNewObj('NodeList')
  CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Mixed Air Nodes')
  IF (supFanPlacement .EQ. sfpDrawThru) THEN
    CALL AddToObjFld('Node Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet')
  END IF
  CALL AddToObjFld('Node Name', base + pvavsAirHandlerNameOff,' Mixed Air Outlet',.TRUE.)
  IF (heatCoilType .EQ. ctHotWater) THEN
    !Object ==> Coil:Heating:Water
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + pvavsHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name ', base + pvavsAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + pvavsAirHandlerNameOff,' Heating Coil HW Outlet')
    CALL AddToObjFld('Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + pvavsAirHandlerNameOff,' Heating Coil Outlet')
    IF (isHeatCoilCapAutosize) THEN
      CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
    ELSE
      CALL AddToObjStr('Performance Input Method','NominalCapacity')
    END IF
    CALL AddToObjFld('Rated Capacity',base + pvavsHeatCoilCapOff, ' ')
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
    !Object ==> Branch
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Heating Coil HW Branch')
    !save this branch name for later use in plan
    CALL AddToStrList(TRIM(FldVal(base + pvavsAirHandlerNameOff)) // ' Heating Coil HW Branch',handleHeatingCoilBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + pvavsAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + pvavsAirHandlerNameOff,' Heating Coil HW Outlet',.TRUE.)
    !Object ==> Controller:WaterCoil
    CALL CreateNewObj('Controller:WaterCoil')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Heating Coil Controller')
    CALL AddToObjStr('Control Variable','Temperature')
    CALL AddToObjStr('Action','Normal')
    CALL AddToObjStr('Actuator Variable','Flow')
    CALL AddToObjFld('Sensor Node Name', base + pvavsAirHandlerNameOff,' Heating Coil Outlet')
    CALL AddToObjFld('Actuator Node Name', base + pvavsAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
    CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
    CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)
    !CONTROLLER LIST - This system can only have a hot water coil controller
    CALL CreateNewObj('AirLoopHVAC:ControllerList')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Controllers')
    CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
    CALL AddToObjFld('Controller Name', base + pvavsAirHandlerNameOff,' Heating Coil Controller',.TRUE.)
  END IF
  !Object ==> Coil:Heating:Electric
  IF (heatCoilType .EQ. ctElectric) THEN
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + pvavsHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Efficiency','1')
!    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    CALL AddToObjFld('Nominal Capacity of the Coil {W}',base + pvavsHeatCoilCapOff, ' ')
    CALL AddToObjFld('Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + pvavsAirHandlerNameOff,' Heating Coil Outlet')
    CALL AddToObjFld('Coil Temp Setpoint Node', base + pvavsAirHandlerNameOff,' Heating Coil Outlet',.TRUE.)
  END IF
  IF (heatCoilType .EQ. ctGas) THEN
    !Object ==> Coil:Heating:Fuel
    CALL CreateNewObj('Coil:Heating:Fuel')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + pvavsHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Fuel Type', 'Gas')
    CALL AddToObjFld('Gas Burner Efficiency', base + pvavsHeatEffiencyOff,' ')
!    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    CALL AddToObjFld('Nominal Capacity of the Coil {W}',base + pvavsHeatCoilCapOff, ' ')
    CALL AddToObjFld('Air Inlet Node Name', base +pvavsAirHandlerNameOff,' Cooling Coil Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + pvavsAirHandlerNameOff,' Heating Coil Outlet')
    CALL AddToObjFld('Coil Temp Setpoint Node', base + pvavsAirHandlerNameOff,' Heating Coil Outlet')
    CALL AddToObjFld('Parasitic Electric Load {W}', base + pvavsHeatParasiticOff,' ')
    CALL AddToObjFld('Part load fraction correlation (function of part load ratio)',   &
       base + pvavsAirHandlerNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
    !Object ==> Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  END IF
  IF (heatCoilType .NE. ctNone) THEN
    !If  vsHeatSetPtSchedNameOff is blank, then create schedule of vsHeatSetPtDesignOff
    IF (isHeatStPtSchedBlank) THEN
      CALL AddAlwaysSchedule(FldVal(base + pvavsHeatSetPtDesignOff))
    END IF
    IF (supFanPlacement .EQ. sfpDrawThru) THEN
      CALL WriteComment('NOTE:  Heating supply air setpoint is placed on Supply Path ' // &
                      'Inlet node to allow separate heating and cooling setpoints.')
      CALL WriteComment('SetpointManager:MixedAir <system name> Heating Coil Air ' // &
                      'Temp Manager adjusts this setpoint for fan heat and places it on the Heating Coil Outlet')
    END IF
    SELECT CASE (heatSetPtReset)
      CASE (hsprNone)
        IF (isHeatStPtSchedBlank) THEN
          CALL AddAlwaysSchedule(FldVal(base + pvavsHeatSetPtDesignOff))
        END IF
        !Object ==> SetpointManager:Scheduled
        CALL CreateNewObj('SetpointManager:Scheduled')
        CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Heating Supply Air Temp Manager')
        CALL AddToObjStr('Control Variable','Temperature')
        IF (isHeatStPtSchedBlank) THEN
          CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + pvavsHeatSetPtDesignOff)))
        ELSE
          CALL AddToObjFld('Schedule Name', base + pvavsHeatSetPtSchedNameOff,'')
        END IF
        IF (supFanPlacement .EQ. sfpBlowThru) THEN
          CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Heating Coil Outlet',.TRUE.)
        ELSE
          CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Supply Path Inlet',.TRUE.)
        END IF
 !     CASE (hsprColdestTempFirst)
 !  This setpoint manager type not yet available in EnergyPlus, leave code here for future use
       !SET POINT MANAGER:ColdestTemperatureFlow
 !       CALL CreateNewObj('SetpointManager:ColdestTemperatureFlow')
 !       CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Heating Supply Air Temp Manager')
 !       CALL AddToObjStr('Control Variable','Temperature')
 !       CALL AddToObjFld('HVAC Air Loop Name', base + pvavsAirHandlerNameOff,'')
 !       IF ((StringToReal(FldVal(base + pvavsCoolSetPtDesignOff)) - 1.0) .LT. (StringToReal(FldVal(base + pvavsHeatSetPtDesignOff)) - 5.2)) THEN
 !         CALL AddToObjNum('Minimum Setpoint Temperature {C}',StringToReal(FldVal(base + pvavsCoolSetPtDesignOff)) - 1.0)
 !       ELSE
 !         CALL AddToObjNum('Minimum Setpoint Temperature {C}',StringToReal(FldVal(base + pvavsHeatSetPtDesignOff)) - 5.2)
 !       END IF
 !       CALL AddToObjFld('Maximum Setpoint Temperature {C}', base + pvavsHeatSetPtDesignOff,'')
 !       CALL AddToObjStr('Strategy','TemperatureFirst')
 !       IF (supFanPlacement .EQ. sfpBlowThru) THEN
 !         CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Heating Coil Outlet',.TRUE.)
 !       ELSE
 !         CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Supply Path Inlet',.TRUE.)
 !       END IF
      CASE (hsprOAreset)
        !Object ==> SetpointManager:OutdoorAirReset
        CALL CreateNewObj('SetpointManager:OutdoorAirReset')
        CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Heating Supply Air Temp Manager')
        CALL AddToObjStr('Control Variable','Temperature')
        CALL AddToObjFld('Set point at outside low {C}', base + pvavsHeatSetPtDesignOff,'')
        CALL AddToObjStr('Outside low {C}','-6.7')
        CALL AddToObjNum('Set point at outside high {C}',StringToReal(FldVal(base + pvavsHeatSetPtDesignOff)) - 5.2)
        CALL AddToObjStr('Outside high {C}','10')
        IF (supFanPlacement .EQ. sfpBlowThru) THEN
          CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Heating Coil Outlet',.TRUE.)
        ELSE
          CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Supply Path Inlet',.TRUE.)
        END IF
    END SELECT
    !Object ==> SetpointManager:MixedAir
    ! Only needed if system is draw through and there is a heating coil
    IF (supFanPlacement .EQ. sfpDrawThru) THEN
      CALL CreateNewObj('SetpointManager:MixedAir')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Heating Coil Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('Reference Setpoint Node Name', base + pvavsAirHandlerNameOff,' Supply Path Inlet')
      CALL AddToObjFld('Fan Inlet Node Name', base + pvavsAirHandlerNameOff,' Heating Coil Outlet')
      CALL AddToObjFld('Fan Outlet Node Name', base + pvavsAirHandlerNameOff,' Supply Fan Outlet')
      CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Heating Coil Outlet',.TRUE.)
    END IF
  END IF
  ! Return fan
  IF (isReturnFanYes) THEN
    CALL CreateNewObj('Fan:VariableVolume')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Return Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + pvavsSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + pvavsReturnEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + pvavsReturnPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
    CALL AddToObjStr('Fan Power Minimum Flow Rate Input Method','Fraction')
    CALL AddToObjStr('Fan Power Minimum Flow Fraction','0.0')
    CALL AddToObjStr('Fan Power Minimum Air Flow Rate {m3/s}', '')
    CALL AddToObjFld('Motor Efficiency', base + pvavsReturnFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + pvavsReturnFanFractionOff,' ')
    SELECT CASE (retFanCoef)
      CASE (sfcInletVane)
        CALL AddToObjStr('Fan Power Coefficient 1','0.35071223')
        CALL AddToObjStr('Fan Power Coefficient 2','0.30850535')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.54137364')
        CALL AddToObjStr('Fan Power Coefficient 4','0.87198823')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcDischgDamp)
        CALL AddToObjStr('Fan Power Coefficient 1','0.37073425')
        CALL AddToObjStr('Fan Power Coefficient 2','0.97250253')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.34240761')
        CALL AddToObjStr('Fan Power Coefficient 4','0')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcVSD)
        CALL AddToObjStr('Fan Power Coefficient 1','0.0015302446')
        CALL AddToObjStr('Fan Power Coefficient 2','0.0052080574')
        CALL AddToObjStr('Fan Power Coefficient 3','1.1086242')
        CALL AddToObjStr('Fan Power Coefficient 4','-0.11635563')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcAppG)
        CALL AddToObjStr('Fan Power Coefficient 1','0.0013')
        CALL AddToObjStr('Fan Power Coefficient 2','0.147')
        CALL AddToObjStr('Fan Power Coefficient 3','0.9506')
        CALL AddToObjStr('Fan Power Coefficient 4','-0.0998')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcVSDPresReset) !based on Good Static Pressure Reset pvavsD Fan from EDR_DesignGuidelines_VAV.pdf
        CALL AddToObjStr('Fan Power Coefficient 1', '0.040759894')
        CALL AddToObjStr('Fan Power Coefficient 2', '0.08804497')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.07292612')
        CALL AddToObjStr('Fan Power Coefficient 4', '0.943739823')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
    END SELECT
    CALL AddToObjFld('Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Air Outlet Node Name', base + pvavsAirHandlerNameOff,' Return Fan Outlet',.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF
  !Object ==> Fan:VariableVolume
  CALL CreateNewObj('Fan:VariableVolume')
  CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Supply Fan')
  IF (isSysAvailSchedBlank) THEN
    CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
  ELSE
    CALL AddToObjFld('Availability Schedule Name', base + pvavsSysAvailSchedNameOff,' ')
  END IF
  CALL AddToObjFld('Fan Efficiency', base + pvavsSupplyEfficiencyOff,' ')
  CALL AddToObjFld('Pressure Rise {Pa}', base + pvavsSupplyPressureOff,' ')
  CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
  IF (SameString(FldVal(base + pvavsSupplyMinRateOff),'autosize')) THEN
    CALL AddToObjStr('Fan Power Minimum Flow Rate Input Method','Fraction')
    CALL AddToObjStr('Fan Power Minimum Flow Fraction','0.25')
    CALL AddToObjStr('Fan Power Minimum Air Flow Rate {m3/s}', '')
  ELSE
    CALL AddToObjStr('Fan Power Minimum Flow Rate Input Method','FixedFlowRate')
    CALL AddToObjStr('Fan Power Minimum Flow Fraction','')
    CALL AddToObjFld('Fan Power Minimum Air Flow Rate {m3/s}', base + pvavsSupplyMinRateOff,' ')
  ENDIF
  CALL AddToObjFld('Motor Efficiency', base + pvavsSupplyFanMotorEffOff,' ')
  CALL AddToObjFld('Motor in Airstream Fraction', base + pvavsSupplyFanFractionOff,' ')
  SELECT CASE (supFanCoef)
    CASE (sfcInletVane)
      CALL AddToObjStr('Fan Power Coefficient 1','0.35071223')
      CALL AddToObjStr('Fan Power Coefficient 2','0.30850535')
      CALL AddToObjStr('Fan Power Coefficient 3','-0.54137364')
      CALL AddToObjStr('Fan Power Coefficient 4','0.87198823')
      CALL AddToObjStr('Fan Power Coefficient 5','0')
    CASE (sfcDischgDamp)
      CALL AddToObjStr('Fan Power Coefficient 1','0.37073425')
      CALL AddToObjStr('Fan Power Coefficient 2','0.97250253')
      CALL AddToObjStr('Fan Power Coefficient 3','-0.34240761')
      CALL AddToObjStr('Fan Power Coefficient 4','0')
      CALL AddToObjStr('Fan Power Coefficient 5','0')
    CASE (sfcVSD)
      CALL AddToObjStr('Fan Power Coefficient 1','0.0015302446')
      CALL AddToObjStr('Fan Power Coefficient 2','0.0052080574')
      CALL AddToObjStr('Fan Power Coefficient 3','1.1086242')
      CALL AddToObjStr('Fan Power Coefficient 4','-0.11635563')
      CALL AddToObjStr('Fan Power Coefficient 5','0')
    CASE (sfcAppG)
      CALL AddToObjStr('Fan Power Coefficient 1','0.0013')
      CALL AddToObjStr('Fan Power Coefficient 2','0.147')
      CALL AddToObjStr('Fan Power Coefficient 3','0.9506')
      CALL AddToObjStr('Fan Power Coefficient 4','-0.0998')
      CALL AddToObjStr('Fan Power Coefficient 5','0')
    CASE (sfcVSDPresReset) !based on Good Static Pressure Reset VSD Fan from EDR_DesignGuidelines_VAV.pdf
      CALL AddToObjStr('Fan Power Coefficient 1', '0.040759894')
      CALL AddToObjStr('Fan Power Coefficient 2', '0.08804497')
      CALL AddToObjStr('Fan Power Coefficient 3','-0.07292612')
      CALL AddToObjStr('Fan Power Coefficient 4', '0.943739823')
      CALL AddToObjStr('Fan Power Coefficient 5','0')
  END SELECT
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    IF (isHumidifierNone) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Mixed Air Outlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Humidifier Outlet')
    END IF
  ELSE
    IF (heatCoilType .NE. ctNone) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Heating Coil Outlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Cooling Coil Outlet')
    END IF
  END IF
  CALL AddToObjFld('Air Outlet Node Name', base + pvavsAirHandlerNameOff,' Supply Fan Outlet',.TRUE.)
  IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  !Object ==>  SetpointManager:SingleZone:Humidity:Minimum
  IF (.NOT. isHumidifierNone) THEN
    CALL CreateNewObj('SetpointManager:SingleZone:Humidity:Minimum')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Humidification Setpoint Manager')
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Humidifier Outlet')
    CALL AddToObjFld('Control Zone Air Node Name', base + pvavsHumidCtrlZoneOff,' Zone Air Node',.TRUE.)
    IF (humidifierKind .EQ. humidifyElecSteam) THEN
      !Object ==> Humidifier:Steam:Electric
      CALL CreateNewObj('Humidifier:Steam:Electric')
      CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Humidifier')
      CALL AddToObjFld('Availability Schedule Name', base + pvavsHumidAvailSchedNameOff,' ')
      CALL AddToObjFld('Rated Capacity {m3/s}', base + pvavsHumidCapOff,'')
      CALL AddToObjFld('Rated Power {W}', base + pvavsHumidRatedElecOff,'')
      CALL AddToObjStr('rated fan power {W}','0')
      CALL AddToObjStr('Standby power {W}','0')
      CALL AddToObjFld('Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Mixed Air Outlet')
      CALL AddToObjFld('Air Outlet Node Name', base + pvavsAirHandlerNameOff,' Humidifier Outlet',.TRUE.)
    END IF
  END IF
  !Object ==> OutdoorAir:NodeList
  CALL CreateNewObj('OutdoorAir:NodeList')
  CALL AddToObjFld('Node or NodeList Name 1', base + pvavsAirHandlerNameOff,' Outside Air Inlet',.TRUE.)
  !Object ==> AirLoopHVAC:OutdoorAirSystem
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' OA System')
  CALL AddToObjFld('Controller List Name', base + pvavsAirHandlerNameOff,' OA System Controllers')
  CALL AddToObjFld('Outdoor Air Equipment List Name', base + pvavsAirHandlerNameOff,' OA System Equipment')
  CALL AddToObjFld('Availability Manager List Name', base + pvavsAirHandlerNameOff,' Availability Managers',.TRUE.)
  !Object ==> AirLoopHVAC:ControllerList
  CALL CreateNewObj('AirLoopHVAC:ControllerList')
  CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' OA System Controllers')
  CALL AddToObjStr('Controller Object Type','Controller:OutdoorAir')
  CALL AddToObjFld('Controller Name', base + pvavsAirHandlerNameOff,' OA Controller',.TRUE.)
  !Object ==> AirLoopHVAC:OutdoorAirSystem:EquipmentList
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem:EquipmentList')
  CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' OA System Equipment')
  IF (heatRecovery .EQ. htrecSens) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' Heat Recovery')
  ELSEIF (heatRecovery .EQ. htrecEnth) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' Heat Recovery')
  END IF
  CALL AddToObjStr('Component Object Type','OutdoorAir:Mixer')
  CALL AddToObjFld('Component Name', base + pvavsAirHandlerNameOff,' OA Mixing Box',.TRUE.)
  !Object ==> OutdoorAir:Mixer
  CALL CreateNewObj('OutdoorAir:Mixer')
  CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' OA Mixing Box')
  CALL AddToObjFld('Mixed Air Node Name', base + pvavsAirHandlerNameOff,' Mixed Air Outlet')
  IF (heatRecovery .EQ. htrecNone) THEN
    CALL AddToObjFld('Outside Air Stream Node Name', base + pvavsAirHandlerNameOff,' Outside Air Inlet')
  ELSE
    CALL AddToObjFld('Outside Air Stream Node Name', base + pvavsAirHandlerNameOff,' Heat Recovery Supply Outlet')
  ENDIF
  CALL AddToObjFld('Relief Air Stream Node Name', base + pvavsAirHandlerNameOff,' Relief Air Outlet')
  CALL AddToObjStr('Return Air Stream Node Name', TRIM(returnInletToOAMIxer),.TRUE.)
  !Object ==> Controller:OutdoorAir
  CALL CreateNewObj('Controller:OutdoorAir')
  CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' OA Controller')
  CALL AddToObjFld('Relief Air Outlet Node Name', base + pvavsAirHandlerNameOff,' Relief Air Outlet')
  CALL AddToObjStr('Return Air Node Name',TRIM(returnInletToOAMIxer))
  CALL AddToObjFld('Mixed Air Node Name', base + pvavsAirHandlerNameOff,' Mixed Air Outlet')
  CALL AddToObjFld('Actuator Node Name', base + pvavsAirHandlerNameOff,' Outside Air Inlet')
  CALL AddToObjFld('Minimum Outdoor Air Flow Rate {m3/s}', base + pvavsMinOutsideFlowOff,' ')
  CALL AddToObjFld('Maximum Outdoor Air Flow Rate {m3/s}', base + pvavsMaxOutsideFlowOff,' ')
  ! vsEconomizerKindOff is now a straight pass-thru to EconomizerChoice in
  ! the OA controller
  CALL AddToObjFld('Economizer Control Type', base + pvavsEconomizerKindOff,' ')
  ! If heat recovery is selected, then set economizer control to bypass
  ! otherwise, set to modulate OA flow
  IF (heatRecovery .EQ. htrecNone) THEN
    CALL AddtoObjStr('Economizer Control Action Type','ModulateFlow')
  ELSE
    CALL AddtoObjStr('Economizer Control Action Type','MinimumFlowWithBypass')
  ENDIF
  CALL AddToObjFld('Economizer Maximum Limit Dry-Bulb Temperature {C}', base + pvavsEconoUpTempLimitOff,' ')
  CALL AddToObjFld('Economizer Maximum Limit Enthalpy {J/kg}', base + pvavsEconoUpEnthLimitOff,' ')
  IF (econoKind .EQ. etFixedDewPointAndDryBulb) THEN
    CALL AddToObjFld('Economizer Maximum Limit Dewpoint Temperature (C)',base + pvavsEconoUpDewLimitOff,' ')
  ELSE
    CALL AddToObjStr('Economizer Maximum Limit Dewpoint Temperature (C)',' ')
  ENDIF
  IF (econoKind .EQ. etElectronicEnthalpy) THEN
    CALL AddToObjFld('Electronic Enthalpy Limit Curve Name', base + pvavsAirHandlerNameOff,' ElectrEnthCurve')
  ELSE
    CALL AddToObjStr('Electronic Enthalpy Limit Curve Name',' ')
  ENDIF
  CALL AddToObjFld('Economizer Minimum Limit Dry-Bulb Temperature {C}', base + pvavsEconoLowTempLimitOff,' ')
  CALL AddToObjFld('Lockout Type', base + pvavsEconoLockoutKindOff,' ')
  CALL AddToObjFld('Minimum Limit Type', base + pvavsMinOutsideControlOff,' ')
  CALL AddToObjFld('Minimum Outdoor Air Schedule Name', base + pvavsMinOutFlowSchedNameOff,' ')
  CALL AddToObjStr('Minimum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Maximum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Mechanical Ventilation Controller Name','',.TRUE.)
  ! If economizer type is ElectronicEnthalpy then create curve object
  IF (econoKind .EQ. etElectronicEnthalpy) THEN
    !Object ==> Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' ElectrEnthCurve')
    CALL AddToObjStr('Coefficient1 Constant','0.01342704')
    CALL AddToObjStr('Coefficient2 x','-0.00047892')
    CALL AddToObjStr('Coefficient3 x**2','0.000053352')
    CALL AddToObjStr('Coefficient4 x**3','-0.0000018103')
    CALL AddToObjStr('Minimum Value of x','16.6')
    CALL AddToObjStr('Maximum Value of x','29.13',.TRUE.)
  ENDIF
  IF (heatRecovery .EQ. htrecSens) THEN
    htRecSens75 = StringToReal(FldVal(base + pvavsHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    !Object ==> HeatExchanger:AirToAir:SensibleAndLatent
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + pvavsAirHandlerNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Nominal supply air flow rate {m3/s}', base + pvavsMinOutsideFlowOff,'')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + pvavsHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Heating Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Heating Air Flow','0')
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + pvavsHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Cooling Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Cooling Air Flow','0')
    CALL AddToObjFld('Supply Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Outside Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + pvavsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + pvavsAirHandlerNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','0')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjStr('Heat exchanger type','Plate')
    CALL AddToObjStr('Frost Control Type','ExhaustOnly')
    CALL AddToObjStr('Threshold Temperature','-1.1')
    CALL AddToObjStr('Initial Defrost Time Fraction','0.167')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','0.024',.TRUE.)
  ELSEIF (heatRecovery .EQ. htrecEnth) THEN
    htRecSens75 = StringToReal(FldVal(base + pvavsHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    htRecLat75 = StringToReal(FldVal(base + pvavsHeatRecLatEffOff)) + 0.05
    IF (htRecLat75 .GT. 1.) htRecLat75 = 1.
    !Object ==> HeatExchanger:AirToAir:SensibleAndLatent
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + pvavsAirHandlerNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Nominal supply air flow rate {m3/s}', base + pvavsMinOutsideFlowOff,'')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + pvavsHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Heating Air Flow', base + pvavsHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow', htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Heating Air Flow', htRecLat75)
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + pvavsHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Cooling Air Flow', base + pvavsHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow', htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Cooling Air Flow', htRecLat75)
    CALL AddToObjFld('Supply Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Outside Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + pvavsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + pvavsAirHandlerNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + pvavsAirHandlerNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','0')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjStr('Heat exchanger type','Rotary')
    CALL AddToObjStr('Frost Control Type','MinimumExhaustTemperature')
    CALL AddToObjStr('Threshold Temperature','1.7')
    CALL AddToObjStr('Initial Defrost Time Fraction','')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','',.TRUE.)
  END IF
  IF (heatRecovery .NE. htrecNone) THEN
    IF (isEconoLowLimitBlank) THEN
      CALL AddAlwaysSchedule('5')
    ELSE
      CALL AddAlwaysSchedule(FldVal(base + pvavsEconoLowTempLimitOff))
    END IF
    !Object ==> SetpointManager:Scheduled
    CALL CreateNewObj('SetpointManager:Scheduled')
    CALL AddToObjFld('Name', base + pvavsAirHandlerNameOff,' Heat Recovery Air Temp Manager')
    CALL AddToObjStr('Control Variable','Temperature')
    IF (isEconoLowLimitBlank) THEN
      CALL AddToObjStr('Schedule Name','HVACTemplate-Always 5')
    ELSE
      CALL AddToObjStr('Schedule Name','HVACTemplate-Always ' // FldVal(base + pvavsEconoLowTempLimitOff))
    END IF
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + pvavsAirHandlerNameOff,' Heat Recovery Supply Outlet',.TRUE.)
  END IF
END DO
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE CreateNewUnitary
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       January 2009 to add Unitary Heat Pump
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a unitary (DX) system.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for cooling coil type
INTEGER, PARAMETER :: ccNone = 0
INTEGER, PARAMETER :: ccSingleSpeedDX = 1
INTEGER, PARAMETER :: ccTwoSpeedDX = 2
INTEGER, PARAMETER :: ccMultiSpeedDX = 3
INTEGER, PARAMETER :: ccTwoStageDX = 4
INTEGER, PARAMETER :: ccTwoStageHumidControlDX = 5
INTEGER, PARAMETER :: ccDXcoilSystemHX = 6
INTEGER, PARAMETER :: ccSingleSpeedDXWater = 7
INTEGER, PARAMETER :: ccChWater = 8
INTEGER, PARAMETER :: ccChWaterDetailed = 9
INTEGER, PARAMETER :: ccWaterCoilSystemHX = 10
          ! parameters for heating coil type
INTEGER, PARAMETER :: hcNone = 0
INTEGER, PARAMETER :: hcElectric = 1
INTEGER, PARAMETER :: hcGas = 2
INTEGER, PARAMETER :: hcWater = 3
INTEGER, PARAMETER :: hcSingleSpeedDXHPAir = 4
INTEGER, PARAMETER :: hcMultiSpeedDXHPAir = 5
INTEGER, PARAMETER :: hcSingleSpeedDXHPWater = 6
INTEGER, PARAMETER :: hcMultiStageElectric = 7
INTEGER, PARAMETER :: hcMultiStageGas = 8
INTEGER, PARAMETER :: hcDeSuperHeat = 9
          ! parameters for economizer type
INTEGER,PARAMETER :: etFixedDryBulb                   = 1
INTEGER,PARAMETER :: etFixedEnthalpy                  = 2
INTEGER,PARAMETER :: etDifferentialDryBulb            = 3
INTEGER,PARAMETER :: etDifferentialEnthalpy           = 4
INTEGER,PARAMETER :: etFixedDewPointAndDryBulb        = 5
INTEGER,PARAMETER :: etElectronicEnthalpy             = 6
INTEGER,PARAMETER :: etDifferentialDryBulbAndEnthalpy = 7
INTEGER,PARAMETER :: etNoEconomizer                   = 8
          ! parameters for lockout type
INTEGER, PARAMETER :: loNo       = 1
INTEGER, PARAMETER :: loHeat     = 2
INTEGER, PARAMETER :: loComp     = 3
          ! parameters for supply fan placement
INTEGER,PARAMETER :: sfpBlowThru = 1
INTEGER,PARAMETER :: sfpDrawThru = 2
          ! parameters for supply fan kind
INTEGER,PARAMETER :: sfkVAV = 1
INTEGER,PARAMETER :: sfkCAV = 2
          ! parameters for dehumidification type
INTEGER, PARAMETER :: dehumidNone           = 1
INTEGER, PARAMETER :: dehumidCoolRhtHtgCoil = 2
INTEGER, PARAMETER :: dehumidCoolRhtDesuper = 3
INTEGER, PARAMETER :: dehumidCoolRht        = 4
INTEGER, PARAMETER :: dehumidMultimode      = 5
          ! parameters for humidification type
INTEGER, PARAMETER :: humidifyNone      = 1
INTEGER, PARAMETER :: humidifyElecSteam = 2
          ! parameters for baseboard type
INTEGER, PARAMETER :: baseboardNone           = 1
INTEGER, PARAMETER :: baseboardHotWater       = 2
INTEGER, PARAMETER :: baseboardElectric       = 3

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for
INTEGER :: sysBase
INTEGER :: iZone
INTEGER :: iSys
INTEGER :: iInt
INTEGER :: znBase
LOGICAL :: isTStatNameBlank
LOGICAL :: isSupplyMaxRateAutosize
LOGICAL :: isSupplyCoolRateAutosize
LOGICAL :: isSupplyHeatRateAutosize
INTEGER :: coolCoilKind
LOGICAL :: isSupplyPlenumBlank
LOGICAL :: isReturnPlenumBlank
LOGICAL :: isSysAvailSchedBlank
LOGICAL :: isSupplyFanCycling
INTEGER :: heatCoilKind=0
INTEGER :: supHeatCoilKind=0
LOGICAL :: isHeatCoilCapAutosize
INTEGER :: econoKind=0
INTEGER :: lockoutKind=0
LOGICAL :: isZnSupPlenumBlank
LOGICAL :: isZnRetPlenumBlank
LOGICAL :: isNightCycleCtrlZoneNameBlank
LOGICAL :: isHeatRecSensible
LOGICAL :: isHeatRecEnthalpy
LOGICAL :: isHeatRecNone
INTEGER :: controlZoneBase
LOGICAL :: isControlZonePlenumBlank
LOGICAL :: isEconoLowLimitBlank
LOGICAL :: isMinOARateAutosize
REAL    :: htRecSens75
REAL    :: htRecLat75
CHARACTER(len=1) :: SchType
INTEGER :: supFanPlacement=0
INTEGER :: supFanKind=0
LOGICAL :: isDehumidifyNone
INTEGER :: dehumidCtrlKind = 0
LOGICAL :: isSuppHeatCoilCapAutosize
LOGICAL :: isHumidifierNone
LOGICAL :: isMultiSpeed
INTEGER :: humidifierKind = 0
INTEGER :: fanMode=0
CHARACTER(len=30) :: furnaceInlet=''
CHARACTER(len=30) :: furnaceOutlet=''
CHARACTER(len=30) :: coolCoilInlet=''
CHARACTER(len=30) :: coolCoilOutlet=''
CHARACTER(len=30) :: coolCoilUnitInlet=''
CHARACTER(len=30) :: coolCoilUnitOutlet=''
CHARACTER(len=30) :: heatCoilInlet=''
CHARACTER(len=30) :: heatCoilOutlet=''
CHARACTER(len=30) :: supFanInlet=''
CHARACTER(len=30) :: supFanOutlet=''
CHARACTER(len=30) :: reheatCoilInlet=''
CHARACTER(len=30) :: reheatCoilOutlet=''
CHARACTER(len=30) :: unitInlet=''
CHARACTER(len=30) :: unitOutlet=''
CHARACTER(len=30) :: supheatCoilInlet=''
CHARACTER(len=30) :: supheatCoilOutlet=''
CHARACTER(len=30) :: humidifierInlet=''
CHARACTER(len=30) :: humidifierOutlet=''
CHARACTER(len=30) :: unitarySystemInlet=''
CHARACTER(len=30) :: unitarySystemOutlet=''
CHARACTER(len=30) :: airLoopOutlet=''
CHARACTER(len=MaxAlphaLength) :: supFanObjectType=''
CHARACTER(len=MaxAlphaLength) :: coolCoilObjectType=''
CHARACTER(len=MaxAlphaLength) :: heatCoilObjectType=''
CHARACTER(len=MaxAlphaLength) :: supheatCoilObjectType=''
CHARACTER(len=MaxAlphaLength) :: coolCoilObjectName=''
CHARACTER(len=MaxAlphaLength) :: heatCoilObjectName=''
CHARACTER(len=MaxAlphaLength) :: supheatCoilObjectName=''
CHARACTER(len=MaxAlphaLength) :: lastOutlet=''
INTEGER :: baseboardKind = 0
LOGICAL :: isHeatPump = .FALSE.
LOGICAL :: isUnitarySys = .FALSE.
LOGICAL :: doSuppressMsgZlUnitAll = .FALSE.
LOGICAL :: isOATypeDetailed = .FALSE.
LOGICAL :: isReturnFanYes = .FALSE.
LOGICAL :: isCoolCapAutosize = .FALSE.
LOGICAL :: isHeatCapAutosize = .FALSE.
LOGICAL :: isSHRAutosize = .FALSE.
REAL    :: CapPerSpeed = 0.0
INTEGER :: numSpeeds
INTEGER :: iSpd
LOGICAL :: isNumSpeedOne = .FALSE.

! More Names
CHARACTER(len=MaxAlphaLength) :: returnInletToOAMIxer=''
CHARACTER(len=MaxAlphaLength) :: dehumscheduleName=''
CHARACTER(len=MaxAlphaLength) :: humscheduleName=''
CHARACTER(len=2) :: iSpdChar=''

!======================================================================================================
! create Zones
!======================================================================================================
coolCoilKind=0
DO iZone = 1, numCompactZoneUnit
  base =  compactZoneUnitBase(iZone)
 ! find base for system this zone belongs to
  sysBase = -1
  isHeatPump = .FALSE.
  isUnitarySys = .FALSE.
  DO iSys = 1, numCompactSysUnit
    IF (SameString(FldVal(compactSysUnitBase(iSys) + usAirHandlerNameOff), FldVal(base +  uzAirHandlerNameOff))) THEN
      sysBase = compactSysUnitBase(iSys)
      EXIT
    END IF
  END DO
  IF (sysBase .NE. -1) THEN
    isHeatPump = .FALSE.
    isUnitarySys = .FALSE.
  ELSE
    DO iSys = 1, numCompactSysUnitHP
      IF (SameString(FldVal(compactSysUnitHPBase(iSys) + uhpsAirHandlerNameOff), FldVal(base +  uzAirHandlerNameOff))) THEN
        sysBase = compactSysUnitHPBase(iSys)
        EXIT
      END IF
    END DO
    IF (sysBase .NE. -1) THEN
      isHeatPump = .TRUE.
      isUnitarySys = .FALSE.
    END IF
  END IF
  IF (sysBase .NE. -1) THEN
    isUnitarySys = .FALSE.
  ELSE
    DO iSys = 1, numCompactSysUnitarySystem
      IF (SameString(FldVal(compactSysUnitarySystemBase(iSys) + ussAirHandlerNameOff), FldVal(base +  uzAirHandlerNameOff))) THEN
        sysBase = compactSysUnitarySystemBase(iSys)
        EXIT
      END IF
    END DO
    IF (sysBase .NE. -1) THEN
      isHeatPump = .FALSE.
      isUnitarySys = .TRUE.
    END IF
  END IF
  IF (sysBase .EQ. -1) THEN
    CALL WriteError('Could not find air handler name referenced in HVACTemplate:Zone:Unitary object: '// &
     TRIM(FldVal(base + uzNameOff)))
    CYCLE
  END IF
  ! set flags
  isTStatNameBlank = (FldVal(base + uzTStatNameOff) .EQ. '')
  isZnSupPlenumBlank = (FldVal(base + uzZoneSupplyPlenumNameOff) .EQ. '')
  isZnRetPlenumBlank = (FldVal(base + uzZoneReturnPlenumNameOff) .EQ. '')
  isSupplyMaxRateAutosize = SameString(FldVal(base + uzSupplyMaxRateOff),'autosize')
  isOATypeDetailed = SameString(FldVal(base + uzOutAirTypeOff),'DetailedSpecification')
  !set the baseboard indicator
  IF (SameString(FldVal(base +  uzBaseboardTypeOff),'None')) THEN
    baseboardKind = baseboardNone
  ELSEIF (SameString(FldVal(base +  uzBaseboardTypeOff),'HotWater')) THEN
    baseboardKind = baseboardHotWater
  ELSEIF (SameString(FldVal(base +  uzBaseboardTypeOff),'Electric')) THEN
    baseboardKind = baseboardElectric
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:Unitary "'//TRIM(FldVal(base + uzNameOff))//'"'// &
                    ' in the Baseboard Heating Type field: '//TRIM(FldVal(base + uzBaseboardTypeOff)))
  END IF
    !validate cooling design method
  IF (SameString(FldVal(base +  uzCoolDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  uzCoolDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  uzCoolDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:Unitary "'//TRIM(FldVal(base + uzNameOff))//'"'// &
                    ' in the Zone Cooling Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + uzCoolDesignTempMethodOff)))
  END IF
  !validate heating design method
  IF (SameString(FldVal(base +  uzHeatDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  uzHeatDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  uzHeatDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:Unitary "'//TRIM(FldVal(base + uzNameOff))//'"'// &
                    ' in the Zone Heating Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + uzCoolDesignTempMethodOff)))
  END IF
!repeat per zone - only if uzTStatNameOff is not blank
  IF (.NOT. isTStatNameBlank) THEN
    !ZONE CONTROL:THERMOSTATIC
    CALL AddAlwaysSchedule('4')
    CALL CreateNewObj('ZoneControl:Thermostat')
    CALL AddToObjFld('Name', base +  uzNameOff,' Thermostat')
    CALL AddToObjFld('Zone or ZoneList Name', base +  uzNameOff,' ')
    !this schedule was previously created
    CALL AddToObjStr('Control Type Schedule Name','HVACTemplate-Always 4')
    CALL AddToObjStr('Control Object Type','ThermostatSetpoint:DualSetpoint')
    CALL AddToObjFld('Control Name', base + uzTStatNameOff,' Dual SP Control',.TRUE.)
  END IF
  !ZONE SIZING ~line 9
  ! OADCV changes
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:OutdoorAir')
    CALL AddToObjStr('Name', 'SZ DSOA '//FldVal(base+uzNameOff))
    CALL AddToObjFld('Outdoor Air Method', base + uzOutAirTypeOff,' ')
    CALL AddToObjFld('Outdoor Air Flow per Person {m3/s}', base + uzOutAirFlowPersonOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone Floor Area {m3/s-m2}', base + uzOutAirFlowAreaOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone {m3/s}', base + uzOutAirFlowZoneOff,'',.true.)
  END IF

  CALL CreateNewObj('Sizing:Zone')
  CALL AddToObjFld('Zone or ZoneList Name', base +  uzNameOff,'')
  IF (SameString(FldVal(base +  uzCoolDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
    CALL AddToObjStr('Zone Cooling Design Supply Air Temperature Input Method','SupplyAirTemperature')
    IF (isHeatPump) THEN
      CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', sysBase + uhpsCoolDesignSupplyTempOff,'')
    ELSEIF (isUnitarySys) THEN
      CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', sysBase + ussCoolDesignSupplyTempOff,'')
    ELSE
      CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', sysBase + usCoolDesignSupplyTempOff,'')
    ENDIF
  ELSE
    CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Input Method',base + uzCoolDesignTempMethodOff,'')
    CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', base + uzCoolDesignTempOff,'')
  ENDIF
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Difference {delta C}',base + uzCoolDesignTempDiffOff,'')
  IF (SameString(FldVal(base +  uzHeatDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
    CALL AddToObjStr('Zone Heating Design Supply Air Temperature Input Method','SupplyAirTemperature')
    IF (isHeatPump) THEN
      CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', sysBase + uhpsHeatDesignSupplyTempOff,'')
    ELSEIF (isUnitarySys) THEN
      CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', sysBase + ussHeatDesignSupplyTempOff,'')
    ELSE
      CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', sysBase + usHeatDesignSupplyTempOff,'')
    ENDIF
  ELSE
    CALL AddToObjFld('Zone Heating Design Supply Air Temperature Input Method',base + uzHeatDesignTempMethodOff,'')
    CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', base + uzHeatDesignTempOff,'')
  ENDIF
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Difference {delta C}',base + uzHeatDesignTempDiffOff,'')
  CALL AddToObjStr('Zone Cooling Design Supply Air Humidity Ratio {kg-H20/kg-air}','0.008')
  CALL AddToObjStr('Zone Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + uzDesSpecOANameOff,'')
  ELSE
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','SZ DSOA '//FldVal(base+uzNameOff))
  END IF
  CALL AddToObjFld('Zone Heating Sizing Factor', base + uzZoneHeatSizingFactorOff,'')
  CALL AddToObjFld('Zone Cooling Sizing Factor', base + uzZoneCoolSizingFactorOff,'')
  IF (isSupplyMaxRateAutosize) THEN
    CALL AddToObjStr('Cooling Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Cooling Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
    CALL AddToObjStr('Heating Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Heating Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
  ELSE
    CALL AddToObjStr('Cooling Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Cooling Design Air Flow Rate {m3/s}', base + uzSupplyMaxRateOff,' ')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
    CALL AddToObjStr('Heating Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Heating Design Air Flow Rate {m3/s}', base + uzSupplyMaxRateOff,'')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
  END IF
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Zone Air Distribution Object Name', base + uzDesSpecOADistNameOff,'',.true.)
  ELSE
    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+uzNameOff),.true.)
  END IF
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:ZoneAirDistribution')
    CALL AddToObjStr('Name','SZ DSZAD '//FldVal(base+uzNameOff))
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Cooling Mode {}','1')
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Heating Mode {}','1',.true.)
  END IF
  !CONTROLLED ZONE EQUIP CONFIGURATION
  CALL CreateNewObj('ZoneHVAC:EquipmentConnections')
  CALL AddToObjFld('Zone Name', base +  uzNameOff,'')
  CALL AddToObjFld('Zone Conditioning Equipment List Name', base +  uzNameOff,' Equipment')
  IF (isZnSupPlenumBlank) THEN
    CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base + uzNameOff,' Zone Equip Inlet')
  ELSE
    CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base +  uzNameOff,' Supply Inlet')
  END IF
  CALL AddToObjStr('Zone Air Exhaust Node or NodeList Name','')
  CALL AddToObjFld('Zone Air Node Name', base +  uzNameOff,' Zone Air Node')
  IF (isZnRetPlenumBlank) THEN
    CALL AddToObjFld('Zone Return Air Node Name', base + uzNameOff,' Return Outlet', .TRUE.)
  ELSE
    CALL AddToObjFld('Zone Return Air Node Name', base + uzNameOff,' Zone Return Plenum Inlet', .TRUE.)
  END IF
  !ZONE SUPPLY PLENUM
  IF (.NOT. isZnSupPlenumBlank) THEN
    CALL CreateNewObj('AirLoopHVAC:SupplyPlenum')
    CALL AddToObjFld('Name', base + uzNameOff,' Supply Plenum')
    CALL AddToObjFld('Zone Name', base + uzZoneSupplyPlenumNameOff,'')
    CALL AddToObjFld('Zone Node Name', base + uzZoneSupplyPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Inlet Node Name', base + uzNameOff,' Zone Equip Inlet')
    CALL AddToObjFld('Outlet Node Name', base + uzNameOff,' Supply Inlet',.TRUE.)
  END IF
  IF (.NOT. isZnRetPlenumBlank) THEN
    CALL CreateNewObj('AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Name', base + uzNameOff,' Return Plenum')
    CALL AddToObjFld('Zone Name', base + uzZoneReturnPlenumNameOff,'')
    CALL AddToObjFld('Zone Node Name', base + uzZoneReturnPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Outlet Node Name', base + uzNameOff,' Return Outlet')
    CALL AddToObjStr('Induced Air Outlet Node or NodeList Name','')
    CALL AddToObjFld('Inlet Node Name', base + uzNameOff,' Zone Return Plenum Inlet',.TRUE.)
  END IF
  !ZONE EQUIPMENT LIST
  CALL CreateNewObj('ZoneHVAC:EquipmentList')
  CALL AddToObjFld('Name', base +  uzNameOff,' Equipment')
  CALL AddToObjStr('Load Distribution Scheme','SequentialLoad')
  CALL AddToObjStr('Zone Equipment Object Type','AirTerminal:SingleDuct:Uncontrolled')
  CALL AddToObjFld('Zone Equipment Name', base +  uzNameOff,' Air Terminal')
  CALL AddToObjStr('Zone Equipment Cooling Sequence','1')
  SELECT CASE (baseboardKind)
    CASE (baseboardNone)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1',.FALSE.)
    CASE (baseboardHotWater)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1')
      CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
      CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.FALSE.)
      CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Zone Equipment Name', base + uzNameOff,' Baseboard Heat')
      CALL AddToObjStr('Zone Equipment Cooling Sequence','2')
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','2',.FALSE.)
    CASE (baseboardElectric)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1')
      CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Zone Equipment Name', base + uzNameOff,' Baseboard Heat')
      CALL AddToObjStr('Zone Equipment Cooling Sequence','2')
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','2',.FALSE.)
  END SELECT
  CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.TRUE.)
  !DIRECT AIR
  CALL CreateNewObj('AirTerminal:SingleDuct:Uncontrolled')
  CALL AddToObjFld('Name', base +  uzNameOff,' Air Terminal')
  CALL AddToObjStr('Availability Schedule Name',' ')
  IF (isZnSupPlenumBlank) THEN
    CALL AddToObjFld('Zone Supply Air Node Name', base + uzNameOff,' Zone Equip Inlet')
  ELSE
    CALL AddToObjFld('Zone Supply Air Node Name', base + uzNameOff,' Supply Inlet')
  END IF
  !CR8001
  !CALL AddToObjFld('Maximum air flow rate {m3/s}', base + uzSupplyMaxRateOff,' ',.TRUE.)
  CALL AddToObjStr('Maximum air flow rate {m3/s}', 'autosize',.TRUE.)
  !Baseboards
  SELECT CASE (baseboardKind)
    CASE (baseboardHotWater)
      !ZoneHVAC:Baseboard:RadiantConvective:Water
      CALL CreateNewObj('ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Name', base + uzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + uzBaseboardAvailSchedNameOff,'')
      CALL AddToObjFld('Inlet Node Name', base + uzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Outlet Node Name', base + uzNameOff,' Baseboard HW Outlet')
      CALL AddToObjStr('Rated Average Water Temperature {C}','87.78')
      CALL AddToObjStr('Rated Water Mass Flow Rate {m3/s}','0.063')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + uzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Maximum Water Flow Rate','autosize')
      CALL AddToObjStr('Convergence Tolerance','0.001')
      CALL AddToObjStr('Fraction Radiant','0.0')
      CALL AddToObjStr('Fraction of Radiant Energy Incident on People','0.0',.TRUE.)
      !BRANCH (for baseboard)
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + uzNameOff,' Baseboard Heat HW Branch')
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Component Name', base + uzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Component Inlet Node Name', base + uzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + uzNameOff,' Baseboard HW Outlet',.TRUE.)
      CALL AddToStrList(TRIM(FldVal(base + uzNameOff)) // ' Baseboard Heat HW Branch',handleHeatingCoilBranch)
      IF ((compactHotLoopBase .EQ. 0) .OR. (numCompactBoiler .EQ. 0)) THEN
        CALL WriteError('Both a HVACTemplate:Plant:HotWaterLoop and a HVACTemplate:Plant:Boiler are needed ' // &
             'when using hot water baseboards.')
      END IF
    CASE (baseboardElectric)
      !BASEBOARD HEATER:Electric:Convective
      CALL CreateNewObj('ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Name', base + uzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + uzBaseboardAvailSchedNameOff,'')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + uzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Efficiency','1',.TRUE.)
  END SELECT
END DO
!======================================================================================================
! create Systems for HVACTemplate:System:Unitary
!======================================================================================================
DO iSys = 1, numCompactSysUnit
  base =  compactSysUnitBase(iSys)
  ! find base for controlled zone used by this system
  controlZoneBase = -1
  DO iZone = 1, numCompactZoneUnit
    IF (SameString(FldVal(compactZoneUnitBase(iZone) + uzNameOff), FldVal(base +  usControlZoneOff))) THEN
      controlZoneBase = compactZoneUnitBase(iZone)
      EXIT
    END IF
  END DO
  IF (controlZoneBase .EQ. -1) THEN
    CALL WriteError('In HVACTemplate:System:Unitary "'//TRIM(FldVal(base + usAirHandlerNameOff))// &
     '" for the field Control Zone Name could not find a matching HVACTemplate:Zone:Unitary named: '// &
     TRIM(FldVal(base +  usControlZoneOff)))
  END IF
  ! set flags
  isSupplyMaxRateAutosize = SameString(FldVal(base + usSupplyMaxRateOff),'autosize')
  isMinOARateAutosize = SameString(FldVal(base + usMinOutsideFlowOff),'autosize')
  isSupplyPlenumBlank = (FldVal(base + usSupplyPlenumNameOff) .EQ. '')
  isReturnPlenumBlank = (FldVal(base + usReturnPlenumNameOff) .EQ. '')
  isSysAvailSchedBlank = (FldVal(base + usSysAvailSchedNameOff) .EQ. '')
  ! Supply fan mode may be blank (default to Cycling) or a schedule name
  isSupplyFanCycling = SameString(FldVal(base + usSupplyFanModeOff),'')
  !set the heating coil indicator
  IF (SameString(FldVal(base +  usHeatCoilKindOff),'Electric')) THEN
    heatCoilKind = hcElectric
  ELSEIF (SameString(FldVal(base +  usHeatCoilKindOff),'Gas')) THEN
    heatCoilKind = hcGas
  ELSEIF (SameString(FldVal(base +  usHeatCoilKindOff),'HotWater')) THEN
    heatCoilKind = hcWater
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:Unitary "'//TRIM(FldVal(base + usAirHandlerNameOff))//'"'// &
                    ' in the Heating Coil Type field: '//TRIM(FldVal(base + usHeatCoilKindOff)))
  END IF
  isHeatCoilCapAutosize = SameString(FldVal(base + usHeatCoilCapOff),'autosize')
  isNightCycleCtrlZoneNameBlank = (FldVal(base + usNightCycleCtrlZoneNameOff) .EQ. '')
  isHeatRecSensible = SameString(FldVal(base + usHeatRecTypeOff),'Sensible')
  isHeatRecEnthalpy = SameString(FldVal(base + usHeatRecTypeOff),'Enthalpy')
  isHeatRecNone = SameString(FldVal(base + usHeatRecTypeOff),'None')
  isControlZonePlenumBlank = (FldVal(controlZoneBase + uzZoneSupplyPlenumNameOff) .EQ. '') !uses control zone base
  isEconoLowLimitBlank = (FldVal(base + usEconoLowLimitOff) .EQ. '')
  isReturnFanYes = (FldVal(base + usReturnFanPresentOff) .EQ. 'Yes')
  IF (SameString(FldVal(base + usCoolCoilKindOff),'SingleSpeedDX')) THEN
    coolCoilKind = ccSingleSpeedDX
  ELSEIF (SameString(FldVal(base + usCoolCoilKindOff),'None')) THEN
    coolCoilKind = ccNone
  ELSE
!    CALL WriteError('Invalid choice in HVACTemplate:System:Unitary in the Cooling Coil Type field')
    CALL WriteError('Invalid choice in HVACTemplate:System:Unitary "'//TRIM(FldVal(base + usAirHandlerNameOff))//'"'// &
                   'in the Cooling Coil Type field: '//TRIM(FldVal(base + usCoolCoilKindOff)))
  END IF
  !set the economizer indicator
  IF (SameString(FldVal(base +  usEconomizerKindOff),'FixedDryBulb')) THEN
    econoKind = etFixedDryBulb
  ELSEIF (SameString(FldVal(base +  usEconomizerKindOff),'FixedEnthalpy')) THEN
    econoKind = etFixedEnthalpy
  ELSEIF (SameString(FldVal(base +  usEconomizerKindOff),'DifferentialDryBulb')) THEN
    econoKind = etDifferentialDryBulb
  ELSEIF (SameString(FldVal(base +  usEconomizerKindOff),'DifferentialEnthalpy')) THEN
    econoKind = etDifferentialEnthalpy
  ELSEIF (SameString(FldVal(base +  usEconomizerKindOff),'FixedDewPointAndDryBulb')) THEN
    econoKind = etFixedDewPointAndDryBulb
  ELSEIF (SameString(FldVal(base +  usEconomizerKindOff),'ElectronicEnthalpy')) THEN
    econoKind = etElectronicEnthalpy
  ELSEIF (SameString(FldVal(base +  usEconomizerKindOff),'DifferentialDryBulbAndEnthalpy')) THEN
    econoKind = etDifferentialDryBulbAndEnthalpy
  ELSEIF (SameString(FldVal(base +  usEconomizerKindOff),'NoEconomizer')) THEN
    econoKind = etNoEconomizer
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:Unitary in the Economizer Type field')
    econoKind = 0
  END IF
  !set the economizer lockout indicator
  IF (SameString(FldVal(base +  usEconoLockoutKindOff),'NoLockout')) THEN
    lockoutKind = loNo
  ELSEIF (SameString(FldVal(base +  usEconoLockoutKindOff),'LockoutWithHeating')) THEN
    lockoutKind = loHeat
  ELSEIF (SameString(FldVal(base +  usEconoLockoutKindOff),'LockoutWithCompressor')) THEN
    lockoutKind = loComp
  ELSEIF (SameString(FldVal(base +  usEconoLockoutKindOff),'')) THEN
    lockoutKind = loNo
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:Unitary in the Economizer Lockout field')
    lockoutKind = 0
  END IF
  !set the supply fan placement indicator
  IF (SameString(FldVal(base + usSupplyFanPlaceOff), 'BlowThrough')) THEN
    supFanPlacement = sfpBlowThru
  ELSEIF (SameString(FldVal(base + usSupplyFanPlaceOff), 'DrawThrough')) THEN
    supFanPlacement = sfpDrawThru
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:Unitary "'//TRIM(FldVal(base + usAirHandlerNameOff))//'"'// &
                    ' in the Supply Fan Placement field: '//TRIM(FldVal(base + usSupplyFanPlaceOff)))
    supFanPlacement = 0
  END IF
  !set the dehumidification indicators
  isDehumidifyNone = SameString(FldVal(base + usDehumCtrlTypeOff),'None')
  IF (SameString(FldVal(base +  usDehumCtrlTypeOff),'None')) THEN
    dehumidCtrlKind = dehumidNone
  ELSEIF (SameString(FldVal(base +  usDehumCtrlTypeOff),'CoolReheatHeatingCoil')) THEN
    dehumidCtrlKind = dehumidCoolRhtHtgCoil
  ELSEIF (SameString(FldVal(base +  usDehumCtrlTypeOff),'CoolReheatDesuperheater')) THEN
    dehumidCtrlKind = dehumidCoolRhtDesuper
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:Unitary "'//TRIM(FldVal(base + usAirHandlerNameOff))//'"'// &
                    ' in the Dehumidification Control Type field: '//TRIM(FldVal(base + usDehumCtrlTypeOff)))
    dehumidCtrlKind = 0
  END IF

  !set the humidifier indicators
  isHumidifierNone = SameString(FldVal(base + usHumidCtrlTypeOff),'None')
  IF (SameString(FldVal(base +  usHumidCtrlTypeOff),'None')) THEN
    humidifierKind = humidifyNone
  ELSEIF (SameString(FldVal(base +  usHumidCtrlTypeOff),'ElectricSteam')) THEN
    humidifierKind = humidifyElecSteam
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:Unitary "'//TRIM(FldVal(base + usAirHandlerNameOff))//'"'// &
                    ' in the Humidifier Type field: '//TRIM(FldVal(base + usHumidCtrlTypeOff)))
  END IF

  !set node name for return inlet to OA mixer
  IF (isReturnFanYes) THEN
    returnInletToOAMIxer = TRIM(FldVal(base + usAirHandlerNameOff))//' Return Fan Outlet'
  ELSE
    returnInletToOAMIxer = TRIM(FldVal(base + usAirHandlerNameOff))//' Air Loop Inlet'
  END IF

  !SYSTEM SIZING
  CALL CreateNewObj('Sizing:System')
  CALL AddToObjFld('AirLoop Name', base + usAirHandlerNameOff,' ')
  CALL AddToObjStr('Type of Load to Size On','Sensible')
  CALL AddToObjFld('Design Outdoor Air Flow Rate {m3/s}', base + usMinOutsideFlowOff,' ')
  CALL AddToObjStr('Central Heating Maximum System Air Flow Ratio','1')
  CALL AddToObjStr('Preheat Design Set Temperature {C}','7')
  CALL AddToObjStr('Preheat Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjStr('Precool Design Temperature {C}','11')
  CALL AddToObjStr('Precool Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjFld('Central Cooling Design Supply Air Temperature {C}', base + usCoolDesignSupplyTempOff,'')
  CALL AddToObjFld('Central Heating Design Supply Air Temperature {C}', base + usHeatDesignSupplyTempOff,'')
  CALL AddToObjStr('Type of Zone Sum to Use','NonCoincident')
  CALL AddToObjStr('100% Outdoor Air in Cooling','No')
  CALL AddToObjStr('100% Outdoor Air in Heating','No')
  CALL AddToObjStr('Central Cooling Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjStr('Central Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isSupplyMaxRateAutosize) THEN
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','DesignDay')
    CALL AddToObjStr('Cooling Supply Air Flow Rate {m3/s}','0')
    ! New fields for v8.2 next three rows
    CALL AddToObjStr('Cooling Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
    CALL AddToObjStr('Cooling Fraction of Autosized Cooling Supply Air Flow Rate','')
    CALL AddToObjStr('Cooling Supply Air Flow Rate Per Unit Cooling Capacity {m3/s-W}','')
    CALL AddToObjStr('Heating Supply Air Flow Rate Method','DesignDay')
    CALL AddToObjStr('Heating Supply Air Flow Rate {m3/s}','0')
    ! New fields for v8.2 next four rows
    CALL AddToObjStr('Heating Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
    CALL AddToObjStr('Heating Fraction of Autosized Heating Supply Air Flow Rate','')
    CALL AddToObjStr('Heating Fraction of Autosized Cooling Supply Air Flow Rate','')
    CALL AddToObjStr('Heating Supply Air Flow Rate Per Unit Heating Capacity {m3/s-W}','')
    CALL AddToObjStr('System Outdoor Air Method','ZoneSum')
    CALL AddToObjStr('Zone Maximum Outdoor Air Fraction','1.0')
    CALL AddToObjStr('Cooling Design Capacity Method','CoolingDesignCapacity')
    CALL AddToObjStr('Cooling Design Capacity {W}','autosize')
    CALL AddToObjStr('Cooling Design Capacity Per Floor Area {W/m2}','')
    CALL AddToObjStr('Fraction of Autosized Cooling Design Capacity','')
    CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
    CALL AddToObjStr('Heating Design Capacity {W}','autosize')
    CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
    CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
    CALL AddToObjStr('Central Cooling Capacity Control Method','OnOff',.true.)
  ELSE
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','Flow/System')
    CALL AddToObjFld('Cooling Supply Air Flow Rate {m3/s}', base + usSupplyMaxRateOff,' ')
    ! New fields for v8.2 next three rows
    CALL AddToObjStr('Cooling Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
    CALL AddToObjStr('Cooling Fraction of Autosized Cooling Supply Air Flow Rate','')
    CALL AddToObjStr('Cooling Supply Air Flow Rate Per Unit Cooling Capacity {m3/s-W}','')
    CALL AddToObjStr('Heating Supply Air Flow Rate Method','Flow/System')
    CALL AddToObjFld('Heating Supply Air Flow Rate {m3/s}', base + usSupplyMaxRateOff,' ')
    ! New fields for v8.2 next four rows
    CALL AddToObjStr('Heating Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
    CALL AddToObjStr('Heating Fraction of Autosized Heating Supply Air Flow Rate','')
    CALL AddToObjStr('Heating Fraction of Autosized Cooling Supply Air Flow Rate','')
    CALL AddToObjStr('Heating Supply Air Flow Rate Per Unit Heating Capacity {m3/s-W}','')
    CALL AddToObjStr('System Outdoor Air Method','ZoneSum')
    CALL AddToObjStr('Zone Maximum Outdoor Air Fraction','1.0')
    CALL AddToObjStr('Cooling Design Capacity Method','CoolingDesignCapacity')
    CALL AddToObjStr('Cooling Design Capacity {W}','autosize')
    CALL AddToObjStr('Cooling Design Capacity Per Floor Area {W/m2}','')
    CALL AddToObjStr('Fraction of Autosized Cooling Design Capacity','')
    CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
    CALL AddToObjStr('Heating Design Capacity {W}','autosize')
    CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
    CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
    CALL AddToObjStr('Central Cooling Capacity Control Method','OnOff',.true.)
  END IF
  !AIR PRIMARY LOOP
  CALL CreateNewObj('AirLoopHVAC')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' ')
  CALL AddToObjStr('Controller List Name','')
  CALL AddToObjFld('Availability Manager List Name', base + usAirHandlerNameOff,' Availability Managers')
  CALL AddToObjStr('Design Supply Air Flow Rate {m3/s}', 'autosize')
  CALL AddToObjFld('Branch List Name', base + usAirHandlerNameOff,' Branches')
  CALL AddToObjStr('Connector List Name','')
  CALL AddToObjFld('Supply Side Inlet Node Name', base + usAirHandlerNameOff,' Air Loop Inlet')
  CALL AddToObjFld('Demand Side Outlet Node Name', base + usAirHandlerNameOff,' Return Air Outlet')
  CALL AddToObjFld('Demand Side Inlet Node Names', base + usAirHandlerNameOff,' Supply Path Inlet')
  CALL AddToObjFld('Supply Side Outlet Node Names', base + usAirHandlerNameOff,' Air Loop Outlet',.TRUE.)
  !BRANCH LIST
  CALL CreateNewObj('BranchList')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Branches')
  CALL AddToObjFld('Branch Name', base + usAirHandlerNameOff,' Main Branch',.TRUE.)
  !BRANCH ~ line 97
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Main Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  IF (isReturnFanYes) THEN
    CALL AddToObjStr('Component Object Type','Fan:ConstantVolume')
    CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' Return Fan')
    CALL AddToObjFld('Component Inlet Node Name', base + usAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + usAirHandlerNameOff,' Return Fan Outlet')
  END IF
  CALL AddToObjStr('Component Object Type','AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' OA System')
  CALL AddToObjStr('Component Inlet Node Name', TRIM(returnInletToOAMIxer))
  CALL AddToObjFld('Component Outlet Node Name', base + usAirHandlerNameOff,' Mixed Air Outlet')
  IF (humidifierKind .EQ. humidifyElecSteam) THEN
    CALL AddToObjStr('Component Object Type','Humidifier:Steam:Electric')
    CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' Humidifier')
    CALL AddToObjFld('Component Inlet Node Name', base + usAirHandlerNameOff,' Mixed Air Outlet')
    CALL AddToObjFld('Component Outlet Node Name', base + usAirHandlerNameOff,' Humidifier Outlet')
  END IF
  IF (coolCoilKind .EQ. ccNone) THEN
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:Unitary:Furnace:HeatOnly')
    CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' Furnace')
  ELSE
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:Unitary:Furnace:HeatCool')
    CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' Furnace with DX Cooling')
  END IF
  IF (isHumidifierNone) THEN
    CALL AddToObjFld('Component Inlet Node Name', base + usAirHandlerNameOff,' Mixed Air Outlet')
  ELSE
    CALL AddToObjFld('Component Inlet Node Name', base + usAirHandlerNameOff,' Humidifier Outlet')
  END IF
  CALL AddToObjFld('Component Outlet Node Name', base + usAirHandlerNameOff,' Air Loop Outlet',.TRUE.)
  IF (isSupplyPlenumBlank) THEN
    !ZONE SUPPLY AIR PATH
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + usAirHandlerNameOff,' Supply Path Inlet')
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneSplitter')
    CALL MakeSpecialIntList(zlUnitAllSupPlen, base + usAirHandlerNameOff)
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' Zone Splitter',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' Zone Splitter')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Supply Plenum', .TRUE.)
      END IF
    END IF
    !ZONE SPLITTER
    CALL CreateNewObj('AirLoopHVAC:ZoneSplitter')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Zone Splitter')
    CALL AddToObjFld('Inlet Node Name', base + usAirHandlerNameOff,' Supply Path Inlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlUnitAll,base + usAirHandlerNameOff)
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(iInt),' Zone Equip Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Inlet',.TRUE.)
    END IF
  ELSE
    !ZONE SUPPLY AIR PATH
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + usAirHandlerNameOff,' Supply Path Inlet')
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
    CALL MakeSpecialIntList(zlUnitAllSupPlen, base + usAirHandlerNameOff)
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' Supply Plenum',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' Supply Plenum')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Supply Plenum', .TRUE.)
      END IF
    END IF
    !ZONE SUPPLY PLENUM
    CALL CreateNewObj('AirLoopHVAC:SupplyPlenum')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Supply Plenum')
    CALL AddToObjFld('Zone Name', base + usSupplyPlenumNameOff,' ')
    CALL AddToObjFld('Zone Node Name', base + usSupplyPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Inlet Node Name', base + usAirHandlerNameOff,' Supply Path Inlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlUnitAll,base + usAirHandlerNameOff)
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(iInt),' Zone Equip Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Inlet',.TRUE.)
    END IF
  END IF
  IF (isReturnPlenumBlank) THEN
    !ZONE RETURN AIR PATH
    CALL CreateNewObj('AirLoopHVAC:ReturnPath')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Return Path')
    CALL AddToObjFld('Return Air Path Outlet Node Name', base + usAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlUnitAllRetPlen, base + usAirHandlerNameOff)
    IF (numIntList .NE. 0) THEN
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Return Plenum')
      END DO
    END IF
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneMixer')
    CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' Zone Mixer',.TRUE.)
    !ZONE MIXER
    CALL CreateNewObj('AirLoopHVAC:ZoneMixer')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Zone Mixer')
    CALL AddToObjFld('Outlet Node Name', base + usAirHandlerNameOff,' Return Air Outlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlUnitAll,base + usAirHandlerNameOff)
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Inlet Node Name', IntList(iInt),' Return Outlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Inlet Node Name', IntList(numIntList),' Return Outlet',.TRUE.)
    END IF
  ELSE
    !ZONE RETURN AIR PATH
    CALL CreateNewObj('AirLoopHVAC:ReturnPath')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Return Path')
    CALL AddToObjFld('Return Air Path Outlet Node Name', base + usAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlUnitAllRetPlen, base + usAirHandlerNameOff)
    IF (numIntList .NE. 0) THEN
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Return Plenum')
      END DO
    END IF
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' Return Plenum',.TRUE.)
    !ZONE RETURN PLENUM
    CALL CreateNewObj('AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Return Plenum')
    CALL AddToObjFld('Zone Name', base + usReturnPlenumNameOff,' ')
    CALL AddToObjFld('Zone Node Name', base + usReturnPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Outlet Node Name', base + usAirHandlerNameOff,' Return Air Outlet')
    CALL AddToObjStr('Induced Air Outlet Node or NodeList Name','')
    CALL MakeSpecialIntList(zlUnitAll,base + usAirHandlerNameOff)
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Inlet Node Name', IntList(iInt),' Return Outlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Inlet Node Name', IntList(numIntList),' Return Outlet',.TRUE.)
    END IF
  END IF
  !SYSTEM AVAILABILITY MANAGER LIST ~ line 165
  CALL CreateNewObj('AvailabilityManagerAssignmentList')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Availability Managers')
  CALL AddToObjStr('Availability Manager Object Type','AvailabilityManager:NightCycle')
  CALL AddToObjFld('Availability Manager Name', base + usAirHandlerNameOff,' Availability',.TRUE.)
  !SYSTEM AVAILABILITY MANAGER:NIGHT CYCLE
  CALL CreateNewObj('AvailabilityManager:NightCycle')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Availability')
  CALL AddToObjStr('Applicability Schedule Name','HVACTemplate-Always 1')
  IF (.NOT. isSysAvailSchedBlank) THEN
    CALL AddToObjFld('Fan Schedule Name', base + usSysAvailSchedNameOff,'')
  ELSE
    CALL AddToObjStr('Fan Schedule Name','HVACTemplate-Always 1')
  END IF
  CALL AddToObjFld('Control Type', base + usNightCycleControlOff,'')
  CALL AddToObjStr('Thermostat Tolerance {deltaC}','0.2')
  CALL AddToObjStr('Cycling Run Time Control Type', 'FixedRunTime')
  CALL AddToObjStr('Cycling run time {s}','3600')
  CALL AddToObjFld('Control Zone Name', base + usNightCycleCtrlZoneNameOff,'',.TRUE.)
  CALL AddAlwaysSchedule('1')
  !check if zone is specified if it should be.
  IF (isNightCycleCtrlZoneNameBlank) THEN
    IF (SameString(FldVal(base + usNightCycleControlOff),'CycleOnControlZone')) THEN
     CALL WriteError('Warning: A zone name must be specified when Night Cycle Control is set to Cycle on Control Zone')
    END IF
  END IF

  !Set up node name suffixes for various combinations of fan placement, dehumidication and humidification controls
  ! Component orders are (note - these will change for v2.1):
  !    Blow Through heat-only                  = fan, heating coil
  !    Blow Through w/o  dehumidification ctrl = fan, heating coil, dx coil
  !    Blow Through with dehumidification ctrl = fan, dx coil, heating coil, reheat coil
  !    Draw Through heat-only                  = heating coil, fan
  !    Draw Through w/o  dehumidification ctrl = heating coil, dx coil, fan
  !    Draw Through with dehumidification ctrl = dx coil, heating coil, fan, reheat coil
  IF (isHumidifierNone) THEN
    furnaceInlet = ' Mixed Air Outlet'
  ELSE
    furnaceInlet = ' Humidifier Outlet'
  END IF
  furnaceOutlet  = ' Air Loop Outlet'
  IF (coolCoilKind .EQ. ccNone) THEN
    IF (supFanPlacement .EQ. sfpBlowThru) THEN
      supFanInlet    = TRIM(furnaceInlet)
      supFanOutlet   = ' Supply Fan Outlet'
      heatCoilInlet  = ' Supply Fan Outlet'
      heatCoilOutlet = TRIM(furnaceOutlet)
    ELSE IF (supFanPlacement .EQ. sfpDrawThru) THEN
      heatCoilInlet  = TRIM(furnaceInlet)
      heatCoilOutlet = ' Heating Coil Outlet'
      supFanInlet    = ' Heating Coil Outlet'
      supFanOutlet   = TRIM(furnaceOutlet)
    END IF
  ELSE
    IF (supFanPlacement .EQ. sfpBlowThru) THEN
      IF (isDehumidifyNone) THEN
        supFanInlet      = TRIM(furnaceInlet)
        supFanOutlet     = ' Supply Fan Outlet'
        heatCoilInlet    = ' Supply Fan Outlet'
        heatCoilOutlet   = ' Heating Coil Outlet'
        coolCoilInlet    = ' Heating Coil Outlet'
        coolCoilOutlet   = TRIM(furnaceOutlet)
      ELSE
        supFanInlet      = TRIM(furnaceInlet)
        supFanOutlet     = ' Supply Fan Outlet'
        coolCoilInlet    = ' Supply Fan Outlet'
        coolCoilOutlet   = ' Cooling Coil Outlet'
        heatCoilInlet    = ' Cooling Coil Outlet'
        heatCoilOutlet   = ' Heating Coil Outlet'
        reheatCoilInlet  = ' Heating Coil Outlet'
        reheatCoilOutlet = TRIM(furnaceOutlet)
      END IF
    ELSE IF (supFanPlacement .EQ. sfpDrawThru) THEN
      IF (isDehumidifyNone) THEN
        heatCoilInlet    = TRIM(furnaceInlet)
        heatCoilOutlet   = ' Heating Coil Outlet'
        coolCoilInlet    = ' Heating Coil Outlet'
        coolCoilOutlet   = ' Cooling Coil Outlet'
        supFanInlet      = ' Cooling Coil Outlet'
        supFanOutlet     = TRIM(furnaceOutlet)
      ELSE
        coolCoilInlet    = TRIM(furnaceInlet)
        coolCoilOutlet   = ' Cooling Coil Outlet'
        heatCoilInlet    = ' Cooling Coil Outlet'
        heatCoilOutlet   = ' Heating Coil Outlet'
        supFanInlet      = ' Heating Coil Outlet'
        supFanOutlet     = ' Supply Fan Outlet'
        reheatCoilInlet  = ' Supply Fan Outlet'
        reheatCoilOutlet = TRIM(furnaceOutlet)
      END IF
    END IF
  END IF

  ! Humdistat(s) if needed
  !    Single humidistat if humidification and dehumidification are both active and control zones are the same
  IF ((.NOT. isHumidifierNone) .AND. (.NOT. isDehumidifyNone) &
       .AND. SameString(FldVal(base +  usControlZoneOff),FldVal(base +  usHumidCtrlZoneOff))) THEN
    CALL CreateNewObj('ZoneControl:Humidistat')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Humidistat')
    CALL AddToObjFld('Zone Name', base + usControlZoneOff,'')
    CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name', &
                     'HVACTemplate-Always ' // TRIM(FldVal(base + usHumidSetPtOff)))
    CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name', &
                     'HVACTemplate-Always ' // TRIM(FldVal(base + usDehumSetPtOff)),.TRUE.)

    CALL AddAlwaysSchedule(FldVal(base + usHumidSetPtOff))
    CALL AddAlwaysSchedule(FldVal(base + usDehumSetPtOff))

  ELSE
    IF ((coolCoilKind .NE. ccNone) .AND. (.NOT. isDehumidifyNone)) THEN
  !   Dehumidification humidistat
      CALL CreateNewObj('ZoneControl:Humidistat')
      CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Dehumidification Humidistat')
      CALL AddToObjFld('Zone Name', base + usControlZoneOff,'')
      CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name','HVACTemplate-Always 1')
      CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name', &
                       'HVACTemplate-Always ' // TRIM(FldVal(base + usDehumSetPtOff)),.TRUE.)

      CALL AddAlwaysSchedule('1')
      CALL AddAlwaysSchedule(FldVal(base + usDehumSetPtOff))
    ENDIF
    IF (.NOT. isHumidifierNone) THEN
  !    Humidification humidistat
      CALL CreateNewObj('ZoneControl:Humidistat')
      CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Humidification Humidistat')
      CALL AddToObjFld('Zone Name', base + usHumidCtrlZoneOff,'')
      CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name', &
                       'HVACTemplate-Always ' // TRIM(FldVal(base + usHumidSetPtOff)))
      CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name','HVACTemplate-Always 100',.TRUE.)

      CALL AddAlwaysSchedule(FldVal(base + usHumidSetPtOff))
      CALL AddAlwaysSchedule('100')
    ENDIF
  ENDIF

  ! Humidifier if humidification active ~ line 233
  IF (.NOT. isHumidifierNone) THEN
    CALL CreateNewObj('SetpointManager:SingleZone:Humidity:Minimum')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Humidification Setpoint Manager')
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + usAirHandlerNameOff,' Humidifier Outlet')
    CALL AddToObjFld('Control Zone Air Node Name', base + usHumidCtrlZoneOff,' Zone Air Node',.TRUE.)

    IF (humidifierKind .EQ. humidifyElecSteam) THEN
      CALL CreateNewObj('Humidifier:Steam:Electric')
      CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Humidifier')
      CALL AddToObjFld('Availability Schedule Name', base + usHumidAvailSchedNameOff,' ')
      CALL AddToObjFld('Rated Capacity {m3/s}', base + usHumidCapOff,'')
      CALL AddToObjFld('Rated Power {W}', base + usHumidRatedElecOff,'')
      CALL AddToObjStr('rated fan power {W}','0')
      CALL AddToObjStr('Standby power {W}','0')
      CALL AddToObjFld('Air Inlet Node Name', base + usAirHandlerNameOff,' Mixed Air Outlet')
      CALL AddToObjFld('Air Outlet Node Name', base + usAirHandlerNameOff,' Humidifier Outlet',.TRUE.)
    END IF
  END IF
  !FURNACE:HEATCOOL or FURNACE:HEATONLY ~ line 256
  SchType=' '
  IF (coolCoilKind .EQ. ccNone) THEN
    CALL CreateNewObj('AirLoopHVAC:Unitary:Furnace:HeatOnly')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Furnace')
  ELSE
    CALL CreateNewObj('AirLoopHVAC:Unitary:Furnace:HeatCool')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Furnace with DX Cooling')
  END IF
  CALL AddToObjStr('Availability Schedule Name',' ')
  CALL AddToObjFld('Furnace Air Inlet Node Name', base + usAirHandlerNameOff, TRIM(furnaceInlet))
  CALL AddToObjFld('Furnace Air Outlet Node Name', base + usAirHandlerNameOff, TRIM(furnaceOutlet))
  IF (isSupplyFanCycling) THEN
    CALL AddToObjStr('Supply Air Fan Operating Mode Schedule Name','HVACTemplate-Always 0')
    SchType='0'
  ELSE
    CALL AddToObjFld('Supply Air Fan Operating Mode Schedule Name', base + usSupplyFanModeOff,'')
  END IF
  CALL AddToObjStr('Maximum Supply Air Temperature {C}','80')
  IF (coolCoilKind .EQ. ccNone) THEN
    CALL AddToObjStr('Heating Supply Air Flow Rate {m3/s}', 'autosize')
  ELSE
    CALL AddToObjStr('Cooling Supply Air Flow Rate {m3/s}', 'autosize')
    CALL AddToObjStr('Heating Supply Air Flow Rate {m3/s}', 'autosize')
    CALL AddToObjStr('No Load Supply Air Flow Rate {m3/s}', 'autosize')
  ENDIF
  CALL AddToObjFld('Controlling Zone or Thermostat Location', base + usControlZoneOff,' ')
  !CALL AddToObjStr('Fraction of Supply Air Flow That Goes Through the Controlling Zone','autosize')
  CALL AddToObjStr('Supply Fan Object Type','Fan:OnOff')
  CALL AddToObjFld('Supply Fan Name', base + usAirHandlerNameOff,' Supply Fan')
  CALL AddToObjFld('Fan Placement', base + usSupplyFanPlaceOff,'')
  IF (heatCoilKind .EQ. hcElectric) THEN
    CALL AddToObjStr('Heating Coil Object Type','Coil:Heating:Electric')
  ELSE IF (heatCoilKind .EQ. hcGas) THEN
    CALL AddToObjStr('Heating Coil Object Type','Coil:Heating:Fuel')
  ELSE IF (heatCoilKind .EQ. hcWater) THEN
    CALL AddToObjStr('Heating Coil Object Type','Coil:Heating:Water')
  END IF
  IF (coolCoilKind .EQ. ccNone) THEN
    CALL AddToObjFld('Heating Coil Name', base + usAirHandlerNameOff,' Heating Coil',.TRUE.)
  ELSE
    CALL AddToObjFld('Heating Coil Name', base + usAirHandlerNameOff,' Heating Coil')
    CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:DX:SingleSpeed')
    IF (dehumidCtrlKind .EQ. dehumidNone) THEN
      CALL AddToObjFld('Cooling Coil Name', base + usAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjStr('Dehumidification Control Type','None')
      CALL AddToObjStr('Reheat Coil Object Type','')
      CALL AddToObjStr('Reheat Coil Name','',.TRUE.)
    ELSE IF (dehumidCtrlKind .EQ. dehumidCoolRhtDesuper) THEN
      CALL AddToObjFld('Cooling Coil Name', base + usAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjStr('Dehumidification Control Type','CoolReheat')
      CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Desuperheater')
      CALL AddToObjFld('Reheat Coil Name', base + usAirHandlerNameOff,' Reheat Coil',.TRUE.)
    ELSE IF (dehumidCtrlKind .EQ. dehumidCoolRhtHtgCoil) THEN
      CALL AddToObjFld('Cooling Coil Name', base + usAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjStr('Dehumidification Control Type','CoolReheat')
      IF (heatCoilKind .EQ. hcElectric) THEN
        CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Electric')
      ELSE IF (heatCoilKind .EQ. hcGas) THEN
        CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Fuel')
      ELSE IF (heatCoilKind .EQ. hcWater) THEN
        CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Water')
      END IF
      CALL AddToObjFld('Reheat Coil Name', base + usAirHandlerNameOff,' Reheat Coil',.TRUE.)
    ELSE
      CALL AddToObjFld('Cooling Coil Name', base + usAirHandlerNameOff,' Cooling Coil',.TRUE.)
    END IF
  END IF
  ! add new schedule if needed (from Supply Fan Operating Mode Schedule, above)
  IF (SchType /= ' ') THEN
    CALL AddAlwaysSchedule(SchType)
  ENDIF
  ! Main Heating Coil - line 271
  IF (heatCoilKind .EQ. hcElectric) THEN
    !COIL:Electric:Heating ~ line 271
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + usHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Efficiency','1')
    CALL AddToObjFld('Nominal Capacity of the Coil {W}', base + usHeatCoilCapOff,'')
    CALL AddToObjFld('Air Inlet Node Name', base + usAirHandlerNameOff, TRIM(heatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + usAirHandlerNameOff, TRIM(heatCoilOutlet))
    CALL AddToObjStr('Coil Temp Setpoint Node','',.TRUE.)
  ELSE IF (heatCoilKind .EQ. hcGas) THEN
    !COIL:Gas:Heating
    CALL CreateNewObj('Coil:Heating:Fuel')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + usHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Fuel Type', 'Gas')
    CALL AddToObjFld('Gas Burner Efficiency', base + usHeatEffiencyOff,' ')
    CALL AddToObjFld('Nominal Capacity of the Coil {W}', base + usHeatCoilCapOff,'')
    CALL AddToObjFld('Air Inlet Node Name', base + usAirHandlerNameOff, TRIM(heatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + usAirHandlerNameOff, TRIM(heatCoilOutlet))
    CALL AddToObjStr('Coil Temp Setpoint Node','')
    CALL AddToObjFld('Parasitic Electric Load {W}', base + usHeatParasiticOff,' ')
    CALL AddToObjFld('Part load fraction correlation (function of part load ratio)',   &
       base + usAirHandlerNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
    !CURVE:CUBIC
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  ELSE IF (heatCoilKind .EQ. hcWater) THEN
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + usHeatAvailSchedNameOff,'')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name', base + usAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + usAirHandlerNameOff,' Heating Coil HW Outlet')
    CALL AddToObjFld('Air Inlet Node Name', base + usAirHandlerNameOff, TRIM(heatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + usAirHandlerNameOff, TRIM(heatCoilOutlet))
    IF (isHeatCoilCapAutosize) THEN
      CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
    ELSE
      CALL AddToObjStr('Performance Input Method','NominalCapacity')
    END IF
    CALL AddToObjFld('Rated Capacity', base + usHeatCoilCapOff,'')
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
    !CURVE:CUBIC ~ line 254
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Heating Coil HW Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + usAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + usAirHandlerNameOff,' Heating Coil HW Outlet',.TRUE.)
    CALL AddToStrList(TRIM(FldVal(base + usAirHandlerNameOff)) // ' Heating Coil HW Branch',handleHeatingCoilBranch)
  END IF

  ! Reheat coil - if there is a cooling coil and if dehumidification is active
  IF ((coolCoilKind .NE. ccNone) .AND. (.NOT. isDehumidifyNone)) THEN
    IF (dehumidCtrlKind .EQ. dehumidCoolRhtDesuper) THEN
      CALL CreateNewObj('Coil:Heating:Desuperheater')
      CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Reheat Coil')
      CALL AddToObjStr('Availability Schedule Name',' ')
      CALL AddToObjStr('Heat Reclaim Recovery Efficiency','0.3')
      CALL AddToObjFld('Air Inlet Node Name', base + usAirHandlerNameOff,TRIM(reheatCoilInlet))
      CALL AddToObjFld('Air Outlet Node Name', base + usAirHandlerNameOff,TRIM(reheatCoilOutlet))
      CALL AddToObjStr('Heating Source Object Type','Coil:Cooling:DX:SingleSpeed')
      CALL AddToObjFld('Heating Source Name', base + usAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjStr('Temperature Setpoint Node Name','')
      CALL AddToObjStr('Parasitic Electric Load {W}','',.TRUE.)
    ELSE IF (dehumidCtrlKind .EQ. dehumidCoolRhtHtgCoil) THEN
      IF (heatCoilKind .EQ. hcElectric) THEN
        CALL CreateNewObj('Coil:Heating:Electric')
        CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Reheat Coil')
        CALL AddToObjFld('Availability Schedule Name', base + usHeatAvailSchedNameOff,' ')
        CALL AddToObjStr('Efficiency','1')
        CALL AddToObjFld('Nominal Capacity of the Coil {W}', base + usHeatCoilCapOff,'')
        CALL AddToObjFld('Air Inlet Node Name', base + usAirHandlerNameOff,TRIM(reheatCoilInlet))
        CALL AddToObjFld('Air Outlet Node Name', base + usAirHandlerNameOff,TRIM(reheatCoilOutlet))
        CALL AddToObjStr('Coil Temp Setpoint Node','',.TRUE.)
      ELSE IF (heatCoilKind .EQ. hcGas) THEN
        CALL CreateNewObj('Coil:Heating:Fuel')
        CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Reheat Coil')
        CALL AddToObjFld('Availability Schedule Name', base + usHeatAvailSchedNameOff,' ')
        CALL AddToObjStr('Fuel Type', 'Gas')
        CALL AddToObjFld('Gas Burner Efficiency', base + usHeatEffiencyOff,'')
        CALL AddToObjFld('Nominal Capacity of the Coil {W}', base + usHeatCoilCapOff,'')
        CALL AddToObjFld('Air Inlet Node Name', base + usAirHandlerNameOff,TRIM(reheatCoilInlet))
        CALL AddToObjFld('Air Outlet Node Name', base + usAirHandlerNameOff,TRIM(reheatCoilOutlet))
        CALL AddToObjStr('Coil Temp Setpoint Node','')
        CALL AddToObjFld('Parasitic Electric Load {W}', base + usHeatParasiticOff,'')
        CALL AddToObjFld('Part load fraction correlation (function of part load ratio)',   &
           base + usAirHandlerNameOff,' Reheat Coil PLF-FPLR')
        CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)

        CALL CreateNewObj('Curve:Cubic')
        CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Reheat Coil PLF-FPLR')
        CALL AddToObjStr('Coefficient1 Constant','0.8')
        CALL AddToObjStr('Coefficient2 x','0.2')
        CALL AddToObjStr('Coefficient3 x**2','0')
        CALL AddToObjStr('Coefficient4 x**3','0')
        CALL AddToObjStr('Minimum Value of x','0')
        CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
      ELSE IF (heatCoilKind .EQ. hcWater) THEN
        CALL CreateNewObj('Coil:Heating:Water')
        CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Reheat Coil')
        CALL AddToObjFld('Availability Schedule Name', base + usHeatAvailSchedNameOff,'')
        CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
        CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
        CALL AddToObjFld('Water Inlet Node Name', base + usAirHandlerNameOff,' Reheat Coil HW Inlet')
        CALL AddToObjFld('Water Outlet Node Name', base + usAirHandlerNameOff,' Reheat Coil HW Outlet')
        CALL AddToObjFld('Air Inlet Node Name', base + usAirHandlerNameOff,TRIM(reheatCoilInlet))
        CALL AddToObjFld('Air Outlet Node Name', base + usAirHandlerNameOff,TRIM(reheatCoilOutlet))
        IF (isHeatCoilCapAutosize) THEN
          CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
        ELSE
          CALL AddToObjStr('Performance Input Method','NominalCapacity')
        END IF
        CALL AddToObjFld('Rated Capacity', base + usHeatCoilCapOff,'')
        CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
        CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
        CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
        CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
        CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
        !CURVE:CUBIC ~ line 254
        CALL CreateNewObj('Branch')
        CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Reheat Coil HW Branch')
        CALL AddToObjStr('Pressure Drop Curve Name','')
        CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
        CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' Reheat Coil')
        CALL AddToObjFld('Component Inlet Node Name', base + usAirHandlerNameOff,' Reheat Coil HW Inlet')
        CALL AddToObjFld('Component Outlet Node Name', base + usAirHandlerNameOff,' Reheat Coil HW Outlet',.TRUE.)
        CALL AddToStrList(TRIM(FldVal(base + usAirHandlerNameOff)) // ' Reheat Coil HW Branch',handleHeatingCoilBranch)
      END IF
    END IF
  END IF
  ! Cooling coil
  IF (coolCoilKind .NE. ccNone) THEN
    !COIL:DX:CoolingBypassFactorEmpirical ~ line 366
    CALL CreateNewObj('Coil:Cooling:DX:SingleSpeed')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Cooling Coil')
    CALL AddToObjFld('Availability Schedule Name', base + usCoolAvailSchedNameOff,' ')
    CALL AddToObjFld('Rated Total Cooling Capacity {W}', base + usCoolCoilCapOff,' ')
    CALL AddToObjFld('Gross Rated Sensible Heat Ratio', base + usCoolCoilSHROff,' ')
    CALL AddToObjFld('Rated COP', base + usCoolCoilCOPOff,' ')
    CALL AddToObjStr('Rated Air Flow Rate {m3/s}', 'autosize')
    CALL AddToObjStr('Rated Evaporator Fan Power per Volume Flow Rate {W/(m3/s)}','')
    CALL AddToObjFld('Air Inlet Node Name', base + usAirHandlerNameOff, TRIM(coolCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + usAirHandlerNameOff, TRIM(coolCoilOutlet))
    CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
       base + usAirHandlerNameOff,' DX Coil Cap-FT')
    CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
       base + usAirHandlerNameOff,' DX Coil Cap-FF')
    CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
       base + usAirHandlerNameOff,' DX Coil EIR-FT')
    CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
       base + usAirHandlerNameOff,' DX Coil EIR-FF')
    CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
       base + usAirHandlerNameOff,' DX Coil PLF')
    CALL AddToObjStr('Minimum Outdoor Dry-Bulb Temperature for Compressor Operation {C}','')
    CALL AddToObjStr('Nominal Time for Condensate Removal to Begin','0')
    CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity','0')
    CALL AddToObjStr('Maximum Cycling Rate','0')
    CALL AddToObjStr('Latent Capacity Time Constant','0')
    CALL AddToObjFld('Condenser Air Inlet Node Name',base + usAirHandlerNameOff,' Cooling Coil Condenser Inlet')
    CALL AddToObjStr('Condenser Type','AirCooled')
    CALL AddToObjStr('Evaporative Condenser Effectiveness','0')
    CALL AddToObjStr('Evaporative Condenser Air Flow Rate','')
    CALL AddToObjStr('Evaporative Condenser Pump Rated Power Consumption','0')
    CALL AddToObjStr('Crankcase Heater Capacity','0')
    CALL AddToObjStr('Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater Operation','10',.TRUE.)
    !CURVE:BIQUADRATIC
    CALL CreateNewObj('Curve:Biquadratic')
    CALL WriteComment('DOE-2.1E, COOL-CAP-FT for PTAC w/ SI temps')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' DX Coil Cap-FT')
    CALL AddToObjStr('Coefficient1 Constant','0.942587793')
    CALL AddToObjStr('Coefficient2 x','0.009543347')
    CALL AddToObjStr('Coefficient3 x**2','0.00068377')
    CALL AddToObjStr('Coefficient4 y','-0.011042676')
    CALL AddToObjStr('Coefficient5 y**2','0.000005249')
    CALL AddToObjStr('Coefficient6 x*y','-0.00000972')
    CALL AddToObjStr('Minimum Value of x','12.77778')
    CALL AddToObjStr('Maximum Value of x','23.88889')
    CALL AddToObjStr('Minimum Value of y','18.0')
    CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
    !CURVE:QUADRATIC
    CALL CreateNewObj('Curve:Quadratic')
    CALL WriteComment('DOE-2.1E, RATED-CCAP-FFLOW for PTAC')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' DX Coil Cap-FF')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Minimum Value of x','0.5')
    CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
    !CURVE:BIQUADRATIC
    CALL CreateNewObj('Curve:Biquadratic')
    CALL WriteComment('DOE-2.1E, COOL-EIR-FT for PTAC w/ SI temps')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' DX Coil EIR-FT')
    CALL AddToObjStr('Coefficient1 Constant','0.342414409')
    CALL AddToObjStr('Coefficient2 x','0.034885008')
    CALL AddToObjStr('Coefficient3 x**2','-0.0006237')
    CALL AddToObjStr('Coefficient4 y','0.004977216')
    CALL AddToObjStr('Coefficient5 y**2','0.000437951')
    CALL AddToObjStr('Coefficient6 x*y','-0.000728028')
    CALL AddToObjStr('Minimum Value of x','12.77778')
    CALL AddToObjStr('Maximum Value of x','23.88889')
    CALL AddToObjStr('Minimum Value of y','18.0')
    CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
    !CURVE:QUADRATIC
    CALL CreateNewObj('Curve:Quadratic')
    CALL WriteComment('DOE-2.1E, RATED-CEIR-FFLOW for PTAC')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' DX Coil EIR-FF')
    CALL AddToObjStr('Coefficient1 Constant','1.1552')
    CALL AddToObjStr('Coefficient2 x','-0.1808')
    CALL AddToObjStr('Coefficient3 x**2','0.0256')
    CALL AddToObjStr('Minimum Value of x','0.5')
    CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
    !CURVE:QUADRATIC
    CALL CreateNewObj('Curve:Quadratic')
    CALL WriteComment('PLF = l.- Cd(1.-PLR) where Cd = 0.15')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' DX Coil PLF')
    CALL AddToObjStr('Coefficient1 Constant','0.85')
    CALL AddToObjStr('Coefficient2 x','0.15')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
    !OutAir Node
    !LKL- 26Jan2007
    CALL CreateNewObj('OutdoorAir:Node')
    CALL AddToObjFld('Name',base + usAirHandlerNameOff,' Cooling Coil Condenser Inlet')
    CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
  END IF
  ! Return fan
  IF (isReturnFanYes) THEN
    CALL CreateNewObj('Fan:ConstantVolume')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Return Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + usSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + usReturnEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + usReturnPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
    CALL AddToObjFld('Motor Efficiency', base + usReturnFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + usReturnFanFractionOff,' ')
    CALL AddToObjFld('Air Inlet Node Name', base + usAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Air Outlet Node Name', base + usAirHandlerNameOff,' Return Fan Outlet',.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF
  !always use FAN:SIMPLE:ONOFF
  CALL CreateNewObj('Fan:OnOff')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Supply Fan')
  IF (isSysAvailSchedBlank) THEN
    CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
  ELSE
    CALL AddToObjFld('Availability Schedule Name', base + usSysAvailSchedNameOff,' ')
  END IF
  CALL AddToObjFld('Fan Efficiency', base + usSupplyEfficiencyOff,' ')
  CALL AddToObjFld('Pressure Rise {Pa}', base + usSupplyPressureOff,' ')
  CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
  CALL AddToObjFld('Motor Efficiency', base + usSupplyFanMotorEffOff,' ')
  CALL AddToObjFld('Motor in Airstream Fraction', base + usSupplyFanFractionOff,' ')
  CALL AddToObjFld('Air Inlet Node Name', base + usAirHandlerNameOff, TRIM(supFanInlet))
  CALL AddToObjFld('Air Outlet Node Name', base + usAirHandlerNameOff, TRIM(supFanOutlet),.TRUE.)
  IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  !OUTSIDE AIR INLET NODE LIST
  CALL CreateNewObj('OutdoorAir:NodeList')
  CALL AddToObjFld('Node or NodeList Name 1', base + usAirHandlerNameOff,' Outside Air Inlet',.TRUE.)
  !OUTSIDE AIR SYSTEM
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' OA System')
  CALL AddToObjFld('Controller List Name', base + usAirHandlerNameOff,' OA System Controllers')
  CALL AddToObjFld('Outdoor Air Equipment List Name', base + usAirHandlerNameOff,' OA System Equipment')
  CALL AddToObjFld('Availability Manager List Name', base + usAirHandlerNameOff,' Availability Managers',.TRUE.)
  !CONTROLLER LIST
  CALL CreateNewObj('AirLoopHVAC:ControllerList')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' OA System Controllers')
  CALL AddToObjStr('Controller Object Type','Controller:OutdoorAir')
  CALL AddToObjFld('Controller Name', base + usAirHandlerNameOff,' OA Controller',.TRUE.)
  !AIR LOOP EQUIPMENT LIST
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem:EquipmentList')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' OA System Equipment')
  IF (isHeatRecSensible) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' Heat Recovery')
  END IF
  IF (isHeatRecEnthalpy) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' Heat Recovery')
  END IF
  CALL AddToObjStr('Component Object Type','OutdoorAir:Mixer')
  CALL AddToObjFld('Component Name', base + usAirHandlerNameOff,' OA Mixing Box',.TRUE.)
  !OUTSIDE AIR MIXER
  CALL CreateNewObj('OutdoorAir:Mixer')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' OA Mixing Box')
  CALL AddToObjFld('Mixed Air Node Name', base + usAirHandlerNameOff,' Mixed Air Outlet')
  IF (isHeatRecNone) THEN
    CALL AddToObjFld('Outside Air Stream Node Name', base + usAirHandlerNameOff,' Outside Air Inlet')
  ELSEIF (isHeatRecSensible) THEN
    CALL AddToObjFld('Outside Air Stream Node Name', base + usAirHandlerNameOff,' Heat Recovery Supply Outlet')
  ELSEIF (isHeatRecEnthalpy) THEN
    CALL AddToObjFld('Outside Air Stream Node Name', base + usAirHandlerNameOff,' Heat Recovery Supply Outlet')
  ELSE
    CALL AddToObjFld('Outside Air Stream Node Name', base + usAirHandlerNameOff,' Outside Air Inlet')
    CALL WriteError('Warning: Heat recovery option invalid. Assuming none.')
    isHeatRecNone = .TRUE.
  END IF
  CALL AddToObjFld('Relief Air Stream Node Name', base + usAirHandlerNameOff,' Relief Air Outlet')
  CALL AddToObjStr('Return Air Stream Node Name', TRIM(returnInletToOAMIxer),.TRUE.)
  !SetpointManager:SingleZone:Cooling
  CALL CreateNewObj('SetpointManager:SingleZone:Cooling')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Economizer Supply Air Temp Manager')
  CALL AddToObjStr('Control Variable','Temperature')
  CALL AddToObjStr('minimum supply air temperature {C}','13')
  CALL AddToObjStr('maximum supply air temperature {C}','45')
  CALL AddToObjFld('Control Zone Name', base + usControlZoneOff,'')
  CALL AddToObjFld('Zone Node Name', base + usControlZoneOff,' Zone Air Node')
  IF (isControlZonePlenumBlank) THEN
    CALL AddToObjFld('Zone Inlet Node Name', base + usControlZoneOff,' Zone Equip Inlet')
  ELSE
    CALL AddToObjFld('Zone Inlet Node Name', base + usControlZoneOff,' Supply Inlet')
  END IF
  CALL AddToObjFld('Setpoint Node or NodeList Name', base + usAirHandlerNameOff,' Air Loop Outlet',.TRUE.)
  !SET POINT MANAGER:MIXED AIR
  CALL CreateNewObj('SetpointManager:MixedAir')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Cooling Coil Air Temp Manager')
  CALL AddToObjStr('Control Variable','Temperature')
  CALL AddToObjFld('Reference Setpoint Node Name', base + usAirHandlerNameOff,' Air Loop Outlet')
  CALL AddToObjFld('Fan Inlet Node Name', base + usAirHandlerNameOff, TRIM(supFanInlet))
  CALL AddToObjFld('Fan Outlet Node Name', base + usAirHandlerNameOff, TRIM(supFanOutlet))
  CALL AddToObjFld('Setpoint Node or NodeList Name', base + usAirHandlerNameOff,' Mixed Air Outlet',.TRUE.)
  !CONTROLLER:OUTSIDE AIR ~ line 387
  CALL CreateNewObj('Controller:OutdoorAir')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' OA Controller')
  CALL AddToObjFld('Relief Air Outlet Node Name', base + usAirHandlerNameOff,' Relief Air Outlet')
  CALL AddToObjStr('Return Air Node Name',TRIM(returnInletToOAMIxer))
  CALL AddToObjFld('Mixed Air Node Name', base + usAirHandlerNameOff,' Mixed Air Outlet')
  CALL AddToObjFld('Actuator Node Name', base + usAirHandlerNameOff,' Outside Air Inlet')
  CALL AddToObjFld('Minimum Outdoor Air Flow Rate {m3/s}', base + usMinOutsideFlowOff,' ')
  CALL AddToObjFld('Maximum Outdoor Air Flow Rate {m3/s}', base + usMaxOutsideFlowOff,' ')
  ! usEconomizerKindOff is now a straight pass-thru to EconomizerChoice in
  ! the OA controller
  CALL AddToObjFld('Economizer Control Type', base + usEconomizerKindOff,' ')
  ! If heat recovery is selected, then set economizer control to bypass
  ! otherwise, set to modulate OA flow
  IF (isHeatRecNone) THEN
    CALL AddtoObjStr('Economizer Control Action Type','ModulateFlow')
  ELSE
    CALL AddtoObjStr('Economizer Control Action Type','MinimumFlowWithBypass')
  ENDIF
  CALL AddToObjFld('Economizer Maximum Limit Dry-Bulb Temperature {C}', base + usEconoUpLimitOff,' ')
  CALL AddToObjFld('Economizer Maximum Limit Enthalpy {J/kg}', base + usEconoEnthLimitOff,' ')
  IF (econoKind .EQ. etFixedDewPointAndDryBulb) THEN
    CALL AddToObjFld('Economizer Maximum Limit Dewpoint Temperature (C)',base + usEconoUpDewLimitOff, ' ')
  ELSE
    CALL AddToObjStr('Economizer Maximum Limit Dewpoint Temperature (C)',' ')
  ENDIF
  IF (econoKind .EQ. etElectronicEnthalpy) THEN
    CALL AddToObjFld('Electronic Enthalpy Limit Curve Name',base + usAirHandlerNameOff,' ElectrEnthCurve')
  ELSE
    CALL AddToObjStr('Electronic Enthalpy Limit Curve Name',' ')
  ENDIF
  CALL AddToObjFld('Economizer Minimum Limit Dry-Bulb Temperature {C}', base + usEconoLowLimitOff,' ')
  CALL AddToObjFld('Lockout Type', base + usEconoLockoutKindOff,' ')
  CALL AddToObjStr('Minimum Limit Type', 'FixedMinimum')
  CALL AddToObjFld('Minimum Outdoor Air Schedule Name', base + usMinOutFlowSchedNameOff,' ')
  CALL AddToObjStr('Minimum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Maximum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Mechanical Ventilation Controller Name','',.TRUE.)
  IF (econoKind .EQ. etElectronicEnthalpy) THEN
    !CURVE:CUBIC
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' ElectrEnthCurve')
    CALL AddToObjStr('Coefficient1 Constant','0.01342704')
    CALL AddToObjStr('Coefficient2 x','-0.00047892')
    CALL AddToObjStr('Coefficient3 x**2','0.000053352')
    CALL AddToObjStr('Coefficient4 x**3','-0.0000018103')
    CALL AddToObjStr('Minimum Value of x','16.6')
    CALL AddToObjStr('Maximum Value of x','29.13',.TRUE.)
  ENDIF
  !HEAT EXCHANGER:AIR TO AIR:GENERIC
  IF (isHeatRecSensible) THEN
    htRecSens75 = StringToReal(FldVal(base + usHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + usAirHandlerNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Nominal supply air flow rate {m3/s}', base + usMinOutsideFlowOff,'')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + usHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Heating Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Heating Air Flow','0')
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + usHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Cooling Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Cooling Air Flow','0')
    CALL AddToObjFld('Supply Air Inlet Node Name', base + usAirHandlerNameOff,' Outside Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + usAirHandlerNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + usAirHandlerNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + usAirHandlerNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','0')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjStr('Heat exchanger type','Plate')
    CALL AddToObjStr('Frost Control Type','ExhaustOnly')
    CALL AddToObjStr('Threshold Temperature','-1.1')
    CALL AddToObjStr('Initial Defrost Time Fraction','0.167')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','0.024',.TRUE.)
  END IF
  IF (isHeatRecEnthalpy) THEN
    htRecSens75 = StringToReal(FldVal(base + usHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    htRecLat75 = StringToReal(FldVal(base + usHeatRecLatEffOff)) + 0.05
    IF (htRecLat75 .GT. 1.) htRecLat75 = 1.
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + usAirHandlerNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Nominal supply air flow rate {m3/s}', base + usMinOutsideFlowOff,'')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + usHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Heating Air Flow', base + usHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow',htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Heating Air Flow', htRecLat75)
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + usHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Cooling Air Flow', base + usHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow',htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Cooling Air Flow', htRecLat75)
    CALL AddToObjFld('Supply Air Inlet Node Name', base + usAirHandlerNameOff,' Outside Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + usAirHandlerNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + usAirHandlerNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + usAirHandlerNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','0')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjStr('Heat exchanger type','Rotary')
    CALL AddToObjStr('Frost Control Type','MinimumExhaustTemperature')
    CALL AddToObjStr('Threshold Temperature','1.7')
    CALL AddToObjStr('Initial Defrost Time Fraction','')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','',.TRUE.)
  END IF
  IF (.NOT. isHeatRecNone) THEN
    IF (isEconoLowLimitBlank) THEN
      CALL AddAlwaysSchedule('5')
    ELSE
      CALL AddAlwaysSchedule(FldVal(base + usEconoLowLimitOff))
    END IF
    CALL CreateNewObj('SetpointManager:Scheduled')
    CALL AddToObjFld('Name', base + usAirHandlerNameOff,' Heat Recovery Air Temp Manager')
    CALL AddToObjStr('Control Variable','Temperature')
    IF (isEconoLowLimitBlank) THEN
      CALL AddToObjStr('Schedule Name','HVACTemplate-Always 5')
    ELSE
      CALL AddToObjStr('Schedule Name','HVACTemplate-Always ' // FldVal(base + usEconoLowLimitOff))
    END IF
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + usAirHandlerNameOff,' Heat Recovery Supply Outlet',.TRUE.)
  END IF
END DO
!======================================================================================================
! create Systems for HVACTemplate:System:UnitaryHeatPump:AirToAir
!   not developed using the spreadsheet so no spreadsheet reference lines are included
!======================================================================================================
DO iSys = 1, numCompactSysUnitHP
  base =  compactSysUnitHPBase(iSys)
  ! find base for controlled zone used by this system
  controlZoneBase = -1
  DO iZone = 1, numCompactZoneUnit
    IF (SameString(FldVal(compactZoneUnitBase(iZone) + uzNameOff), FldVal(base +  uhpsControlZoneOff))) THEN
      controlZoneBase = compactZoneUnitBase(iZone)
      EXIT
    END IF
  END DO
  IF (controlZoneBase .EQ. -1) THEN
    CALL WriteError('In HVACTemplate:System:UnitaryHeatPump:AirToAir "'//TRIM(FldVal(base + uhpsAirHandlerNameOff))// &
     '" for the field Control Zone Name could not find a matching HVACTemplate:Zone:Unitary named: '// &
     TRIM(FldVal(base +  uhpsControlZoneOff)))
  END IF
  ! set flags
  isSupplyCoolRateAutosize = SameString(FldVal(base + uhpsSupplyCoolFlowRateOff),'autosize')
  isSupplyHeatRateAutosize = SameString(FldVal(base + uhpsSupplyHeatFlowRateOff),'autosize')
  isMinOARateAutosize = SameString(FldVal(base + uhpsMinOutsideFlowOff),'autosize')
  isSupplyPlenumBlank = (FldVal(base + uhpsSupplyPlenumNameOff) .EQ. '')
  isReturnPlenumBlank = (FldVal(base + uhpsReturnPlenumNameOff) .EQ. '')
  isSysAvailSchedBlank = (FldVal(base + uhpsSysAvailSchedNameOff) .EQ. '')
  ! Supply fan mode may be blank (default to Cycling) or a schedule name
  isSupplyFanCycling = SameString(FldVal(base + uhpsSupplyFanOpModeSchedNameOff),'')
  doSuppressMsgZlUnitAll = .FALSE.
  !set the supplemental heating indicator
  IF (SameString(FldVal(base +  uhpsSuppHeatCoilKindOff),'Electric')) THEN
    supHeatCoilKind = hcElectric
  ELSEIF (SameString(FldVal(base +  uhpsSuppHeatCoilKindOff),'Gas')) THEN
    supHeatCoilKind = hcGas
  ELSEIF (SameString(FldVal(base +  uhpsSuppHeatCoilKindOff),'HotWater')) THEN
    supHeatCoilKind = hcWater
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:UnitaryHeatPump:AirToAir "'// &
     TRIM(FldVal(base + uhpsAirHandlerNameOff))//'" in the Supplemental Heating Coil Type: '// &
     TRIM(FldVal(base + uhpsSuppHeatCoilKindOff)))
  END IF
  isSuppHeatCoilCapAutosize = SameString(FldVal(base + uhpsSuppHeatCoilCapOff),'autosize')
  isNightCycleCtrlZoneNameBlank = (FldVal(base + uhpsNightCycleCtrlZoneNameOff) .EQ. '')
  isHeatRecSensible = SameString(FldVal(base + uhpsHeatRecTypeOff),'Sensible')
  isHeatRecEnthalpy = SameString(FldVal(base + uhpsHeatRecTypeOff),'Enthalpy')
  isHeatRecNone = SameString(FldVal(base + uhpsHeatRecTypeOff),'None')
  isControlZonePlenumBlank = (FldVal(controlZoneBase + uzZoneSupplyPlenumNameOff) .EQ. '') !uses control zone base
  isEconoLowLimitBlank = (FldVal(base + uhpsEconoLowTempLimitOff) .EQ. '')
  isReturnFanYes = (FldVal(base + uhpsReturnFanPresentOff) .EQ. 'Yes')
  IF (SameString(FldVal(base + uhpsCoolCoilKindOff),'SingleSpeedDX')) THEN
    coolCoilKind = ccSingleSpeedDX
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:UnitaryHeatPump:AirToAir in the Cooling Coil Type field')
  END IF
  !set the economizer indicator
  IF (SameString(FldVal(base +  uhpsEconomizerKindOff),'FixedDryBulb')) THEN
    econoKind = etFixedDryBulb
  ELSEIF (SameString(FldVal(base +  uhpsEconomizerKindOff),'FixedEnthalpy')) THEN
    econoKind = etFixedEnthalpy
  ELSEIF (SameString(FldVal(base +  uhpsEconomizerKindOff),'DifferentialDryBulb')) THEN
    econoKind = etDifferentialDryBulb
  ELSEIF (SameString(FldVal(base +  uhpsEconomizerKindOff),'DifferentialEnthalpy')) THEN
    econoKind = etDifferentialEnthalpy
  ELSEIF (SameString(FldVal(base +  uhpsEconomizerKindOff),'FixedDewPointAndDryBulb')) THEN
    econoKind = etFixedDewPointAndDryBulb
  ELSEIF (SameString(FldVal(base +  uhpsEconomizerKindOff),'ElectronicEnthalpy')) THEN
    econoKind = etElectronicEnthalpy
  ELSEIF (SameString(FldVal(base +  uhpsEconomizerKindOff),'DifferentialDryBulbAndEnthalpy')) THEN
    econoKind = etDifferentialDryBulbAndEnthalpy
  ELSEIF (SameString(FldVal(base +  uhpsEconomizerKindOff),'NoEconomizer')) THEN
    econoKind = etNoEconomizer
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:UnitaryHeatPump:AirToAir in the Economizer Type field')
  END IF
  !set the economizer lockout indicator
  IF (SameString(FldVal(base +  uhpsEconoLockoutKindOff),'NoLockout')) THEN
    lockoutKind = loNo
  ELSEIF (SameString(FldVal(base +  uhpsEconoLockoutKindOff),'LockoutWithHeating')) THEN
    lockoutKind = loHeat
  ELSEIF (SameString(FldVal(base +  uhpsEconoLockoutKindOff),'LockoutWithCompressor')) THEN
    lockoutKind = loComp
  ELSEIF (SameString(FldVal(base +  uhpsEconoLockoutKindOff),'')) THEN
    lockoutKind = loNo
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:UnitaryHeatPump:AirToAir in the Economizer Lockout field')
  END IF
  !set the supply fan placement indicator
  IF (SameString(FldVal(base + uhpsSupplyFanPlaceOff), 'BlowThrough')) THEN
    supFanPlacement = sfpBlowThru
  ELSEIF (SameString(FldVal(base + uhpsSupplyFanPlaceOff), 'DrawThrough')) THEN
    supFanPlacement = sfpDrawThru
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:UnitaryHeatPump:AirToAir "'// &
     TRIM(FldVal(base + uhpsAirHandlerNameOff))//'" in the Supply Fan Placement field: '// &
     TRIM(FldVal(base + uhpsSupplyFanPlaceOff)))
  END IF
  !set the humidifier indicators
  isHumidifierNone = SameString(FldVal(base + uhpsHumidCtrlTypeOff),'None')
  IF (SameString(FldVal(base +  uhpsHumidCtrlTypeOff),'None')) THEN
    humidifierKind = humidifyNone
  ELSEIF (SameString(FldVal(base +  uhpsHumidCtrlTypeOff),'ElectricSteam')) THEN
    humidifierKind = humidifyElecSteam
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:UnitaryHeatPump:AirToAir "'// &
     TRIM(FldVal(base + uhpsAirHandlerNameOff))//'" in the Humidifier Type field: '// &
     TRIM(FldVal(base + uhpsHumidCtrlTypeOff)))
  END IF

    !set node name for return inlet to OA mixer
  IF (isReturnFanYes) THEN
    returnInletToOAMIxer = TRIM(FldVal(base + uhpsAirHandlerNameOff))//' Return Fan Outlet'
  ELSE
    returnInletToOAMIxer = TRIM(FldVal(base + uhpsAirHandlerNameOff))//' Air Loop Inlet'
  END IF

!Object ==> Sizing:System
  CALL CreateNewObj('Sizing:System')
  CALL AddToObjFld('AirLoop Name', base + uhpsAirHandlerNameOff,' ')
  CALL AddToObjStr('Type of Load to Size On','Sensible')
  CALL AddToObjFld('Design Outdoor Air Flow Rate {m3/s}', base + uhpsMinOutsideFlowOff,' ')
  CALL AddToObjStr('Central Heating Maximum System Air Flow Ratio','1')
  CALL AddToObjStr('Preheat Design Set Temperature {C}','7')
  CALL AddToObjStr('Preheat Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjStr('Precool Design Temperature {C}','11')
  CALL AddToObjStr('Precool Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjFld('Central Cooling Design Supply Air Temperature {C}', base + uhpsCoolDesignSupplyTempOff,'')
  CALL AddToObjFld('Central Heating Design Supply Air Temperature {C}', base + uhpsHeatDesignSupplyTempOff,'')
  CALL AddToObjStr('Sizing Option','NonCoincident')
  CALL AddToObjStr('100% Outdoor Air in Cooling','No')
  CALL AddToObjStr('100% Outdoor Air in Heating','No')
  CALL AddToObjStr('Central Cooling Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjStr('Central Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isSupplyCoolRateAutosize) THEN
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','DesignDay')
    CALL AddToObjStr('Cooling Supply Air Flow Rate {m3/s}','0')
  ELSE
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','Flow/System')
    CALL AddToObjFld('Cooling Supply Air Flow Rate {m3/s}', base + uhpsSupplyCoolFlowRateOff,' ')
  END IF
  ! New fields for v8.2 next three rows
  CALL AddToObjStr('Cooling Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
  CALL AddToObjStr('Cooling Fraction of Autosized Cooling Supply Air Flow Rate','')
  CALL AddToObjStr('Cooling Supply Air Flow Rate Per Unit Cooling Capacity {m3/s-W}','')
  IF (isSupplyHeatRateAutosize) THEN
    CALL AddToObjStr('Heating Supply Air Flow Rate Method','DesignDay')
    CALL AddToObjStr('Heating Supply Air Flow Rate {m3/s}','0')
  ELSE
    CALL AddToObjStr('Heating Supply Air Flow Rate Method','Flow/System')
    CALL AddToObjFld('Heating Supply Air Flow Rate {m3/s}', base + uhpsSupplyHeatFlowRateOff,' ')
  END IF
  ! New fields for v8.2 next four rows
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
  CALL AddToObjStr('Heating Fraction of Autosized Heating Supply Air Flow Rate','')
  CALL AddToObjStr('Heating Fraction of Autosized Cooling Supply Air Flow Rate','')
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Unit Heating Capacity {m3/s-W}','')
  CALL AddToObjStr('System Outdoor Air Method','ZoneSum')
  CALL AddToObjStr('Zone Maximum Outdoor Air Fraction','1.0')
  CALL AddToObjStr('Cooling Design Capacity Method','CoolingDesignCapacity')
  CALL AddToObjStr('Cooling Design Capacity {W}','autosize')
  CALL AddToObjStr('Cooling Design Capacity Per Floor Area {W/m2}','')
  CALL AddToObjStr('Fraction of Autosized Cooling Design Capacity','')
  CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
  CALL AddToObjStr('Heating Design Capacity {W}','autosize')
  CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
  CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
  CALL AddToObjStr('Central Cooling Capacity Control Method','OnOff',.true.)
  !Object ==> AirLoopHVAC
  CALL CreateNewObj('AirLoopHVAC')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' ')
  CALL AddToObjStr('Controller List Name','')
  CALL AddToObjFld('Availability Manager List Name', base + uhpsAirHandlerNameOff,' Availability Managers')
  CALL AddToObjStr('Design Supply Air Flow Rate {m3/s}', 'autosize')
  CALL AddToObjFld('Branch List Name', base + uhpsAirHandlerNameOff,' Branches')
  CALL AddToObjStr('Connector List Name','')
  CALL AddToObjFld('Supply Side Inlet Node Name', base + uhpsAirHandlerNameOff,' Air Loop Inlet')
  CALL AddToObjFld('Demand Side Outlet Node Name', base + uhpsAirHandlerNameOff,' Return Air Outlet')
  CALL AddToObjFld('Demand Side Inlet Node Names', base + uhpsAirHandlerNameOff,' Supply Path Inlet')
  CALL AddToObjFld('Supply Side Outlet Node Names', base + uhpsAirHandlerNameOff,' Air Loop Outlet',.TRUE.)
  !Object ==> BranchList
  CALL CreateNewObj('BranchList')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Branches')
  CALL AddToObjFld('Branch Name', base + uhpsAirHandlerNameOff,' Main Branch',.TRUE.)
  !Object ==> Branch
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Main Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  IF (isReturnFanYes) THEN
    CALL AddToObjStr('Component Object Type','Fan:ConstantVolume')
    CALL AddToObjFld('Component Name', base + uhpsAirHandlerNameOff,' Return Fan')
    CALL AddToObjFld('Component Inlet Node Name', base + uhpsAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + uhpsAirHandlerNameOff,' Return Fan Outlet')
  END IF
  CALL AddToObjStr('Component Object Type','AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Component Name', base + uhpsAirHandlerNameOff,' OA System')
  CALL AddToObjStr('Component Inlet Node Name', TRIM(returnInletToOAMIxer))
  CALL AddToObjFld('Component Outlet Node Name', base + uhpsAirHandlerNameOff,' Mixed Air Outlet')
  IF (humidifierKind .EQ. humidifyElecSteam) THEN
    CALL AddToObjStr('Component Object Type','Humidifier:Steam:Electric')
    CALL AddToObjFld('Component Name', base + uhpsAirHandlerNameOff,' Humidifier')
    CALL AddToObjFld('Component Inlet Node Name', base + uhpsAirHandlerNameOff,' Mixed Air Outlet')
    CALL AddToObjFld('Component Outlet Node Name', base + uhpsAirHandlerNameOff,' Humidifier Outlet')
  END IF
  CALL AddToObjStr('Component Object Type','AirLoopHVAC:UnitaryHeatPump:AirToAir')
  CALL AddToObjFld('Component Name', base + uhpsAirHandlerNameOff,' Heat Pump')
  IF (isHumidifierNone) THEN
    CALL AddToObjFld('Component Inlet Node Name', base + uhpsAirHandlerNameOff,' Mixed Air Outlet')
  ELSE
    CALL AddToObjFld('Component Inlet Node Name', base + uhpsAirHandlerNameOff,' Humidifier Outlet')
  END IF
  CALL AddToObjFld('Component Outlet Node Name', base + uhpsAirHandlerNameOff,' Air Loop Outlet',.TRUE.)
  IF (isSupplyPlenumBlank) THEN
    !Object ==> AirLoopHVAC:SupplyPath
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + uhpsAirHandlerNameOff,' Supply Path Inlet')
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneSplitter')
    CALL MakeSpecialIntList(zlUnitAllSupPlen, base + uhpsAirHandlerNameOff)
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + uhpsAirHandlerNameOff,' Zone Splitter',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + uhpsAirHandlerNameOff,' Zone Splitter')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Supply Plenum', .TRUE.)
      END IF
    END IF
    !Object ==> AirLoopHVAC:ZoneSplitter
    CALL CreateNewObj('AirLoopHVAC:ZoneSplitter')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Zone Splitter')
    CALL AddToObjFld('Inlet Node Name', base + uhpsAirHandlerNameOff,' Supply Path Inlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlUnitAll,base + uhpsAirHandlerNameOff)
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(iInt),' Zone Equip Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Inlet',.TRUE.)
    END IF
  ELSE
    !Object ==> AirLoopHVAC:SupplyPath
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + uhpsAirHandlerNameOff,' Supply Path Inlet')
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
    CALL MakeSpecialIntList(zlUnitAllSupPlen, base + uhpsAirHandlerNameOff)
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + uhpsAirHandlerNameOff,' Supply Plenum',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + uhpsAirHandlerNameOff,' Supply Plenum')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Supply Plenum', .TRUE.)
      END IF
    END IF
    !Object ==> AirLoopHVAC:SupplyPlenum
    CALL CreateNewObj('AirLoopHVAC:SupplyPlenum')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Supply Plenum')
    CALL AddToObjFld('Zone Name', base + uhpsSupplyPlenumNameOff,' ')
    CALL AddToObjFld('Zone Node Name', base + uhpsSupplyPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Inlet Node Name', base + uhpsAirHandlerNameOff,' Supply Path Inlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlUnitAll,base + uhpsAirHandlerNameOff)
    ! since the previous line is called again for the return plenum blank case, logic was added
    ! to not allow the same error message to be displayed more than once.
    IF (numIntList .EQ. 0) THEN
      doSuppressMsgZlUnitAll = .TRUE.
    END IF
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(iInt),' Zone Equip Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Inlet',.TRUE.)
    END IF
  END IF
  IF (isReturnPlenumBlank) THEN
    !Object ==> AirLoopHVAC:ReturnPath
    CALL CreateNewObj('AirLoopHVAC:ReturnPath')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Return Path')
    CALL AddToObjFld('Return Air Path Outlet Node Name', base + uhpsAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlUnitAllRetPlen, base + uhpsAirHandlerNameOff)
    IF (numIntList .NE. 0) THEN
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Return Plenum')
      END DO
    END IF
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneMixer')
    CALL AddToObjFld('Component Name', base + uhpsAirHandlerNameOff,' Zone Mixer',.TRUE.)
    !Object ==> AirLoopHVAC:ZoneMixer
    CALL CreateNewObj('AirLoopHVAC:ZoneMixer')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Zone Mixer')
    CALL AddToObjFld('Outlet Node Name', base + uhpsAirHandlerNameOff,' Return Air Outlet')
    !the IntList holds the list of pointers to the zone names
    ! logic added to avoid duplicate error from being shown since also in supply plenum blank case
    CALL MakeSpecialIntList(zlUnitAll,base + uhpsAirHandlerNameOff,doSuppressMsgZlUnitAll)
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Inlet Node Name', IntList(iInt),' Return Outlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Inlet Node Name', IntList(numIntList),' Return Outlet',.TRUE.)
    END IF
  ELSE
    !Object ==> AirLoopHVAC:ReturnPath
    CALL CreateNewObj('AirLoopHVAC:ReturnPath')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Return Path')
    CALL AddToObjFld('Return Air Path Outlet Node Name', base + uhpsAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlUnitAllRetPlen, base + uhpsAirHandlerNameOff)
    IF (numIntList .NE. 0) THEN
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Return Plenum')
      END DO
    END IF
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Component Name', base + uhpsAirHandlerNameOff,' Return Plenum',.TRUE.)
    !Object ==> AirLoopHVAC:ReturnPlenum
    CALL CreateNewObj('AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Return Plenum')
    CALL AddToObjFld('Zone Name', base + uhpsReturnPlenumNameOff,' ')
    CALL AddToObjFld('Zone Node Name', base + uhpsReturnPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Outlet Node Name', base + uhpsAirHandlerNameOff,' Return Air Outlet')
    CALL AddToObjStr('Induced Air Outlet Node or NodeList Name','')
    CALL MakeSpecialIntList(zlUnitAll,base + uhpsAirHandlerNameOff)
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Inlet Node Name', IntList(iInt),' Return Outlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Inlet Node Name', IntList(numIntList),' Return Outlet',.TRUE.)
    END IF
  END IF
  !Object ==> AvailabilityManagerAssignmentList
  CALL CreateNewObj('AvailabilityManagerAssignmentList')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Availability Managers')
  CALL AddToObjStr('Availability Manager Object Type','AvailabilityManager:NightCycle')
  CALL AddToObjFld('Availability Manager Name', base + uhpsAirHandlerNameOff,' Availability',.TRUE.)
  !Object ==> AvailabilityManager:NightCycle
  CALL CreateNewObj('AvailabilityManager:NightCycle')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Availability')
  CALL AddToObjStr('Applicability Schedule Name','HVACTemplate-Always 1')
  IF (.NOT. isSysAvailSchedBlank) THEN
    CALL AddToObjFld('Fan Schedule Name', base + uhpsSysAvailSchedNameOff,'')
  ELSE
    CALL AddToObjStr('Fan Schedule Name','HVACTemplate-Always 1')
  END IF
  CALL AddToObjFld('Control Type', base + uhpsNightCycleControlOff,'')
  CALL AddToObjStr('Thermostat Tolerance {deltaC}','0.2')
  CALL AddToObjStr('Cycling Run Time Control Type', 'FixedRunTime')
  CALL AddToObjStr('Cycling run time {s}','3600')
  CALL AddToObjFld('Control Zone Name', base + uhpsNightCycleCtrlZoneNameOff,'',.TRUE.)
  CALL AddAlwaysSchedule('1')
  !check if zone is specified if it should be.
  IF (isNightCycleCtrlZoneNameBlank) THEN
    IF (SameString(FldVal(base + uhpsNightCycleControlOff),'CycleOnControlZone')) THEN
     CALL WriteError('Warning: A zone name must be specified when Night Cycle Control is set to Cycle on Control Zone')
    END IF
  END IF
  !Set up node name suffixes for various combinations of fan placement
  !    Blow Through  = fan, dx coil, heating coil, sup heat coil
  !    Draw Through  = dx coil, heating coil, fan, sup heat coil
  IF (isHumidifierNone) THEN
    unitInlet = ' Mixed Air Outlet'
  ELSE
    unitInlet = ' Humidifier Outlet'
  END IF
  unitOutlet  = ' Air Loop Outlet'
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    supFanInlet       = TRIM(unitInlet)
    supFanOutlet      = ' Supply Fan Outlet'
    coolCoilInlet     = ' Supply Fan Outlet'
    coolCoilOutlet    = ' Cooling Coil Outlet'
    heatCoilInlet     = ' Cooling Coil Outlet'
    heatCoilOutlet    = ' Heating Coil Outlet'
    supheatCoilInlet  = ' Heating Coil Outlet'
    supheatCoilOutlet = TRIM(unitOutlet)
  ELSE IF (supFanPlacement .EQ. sfpDrawThru) THEN
    coolCoilInlet     = TRIM(unitInlet)
    coolCoilOutlet    = ' Cooling Coil Outlet'
    heatCoilInlet     = ' Cooling Coil Outlet'
    heatCoilOutlet    = ' Heating Coil Outlet'
    supFanInlet       = ' Heating Coil Outlet'
    supFanOutlet      = ' Supply Fan Outlet'
    supheatCoilInlet  = ' Supply Fan Outlet'
    supheatCoilOutlet = TRIM(unitOutlet)
  END IF
  ! Humdistat if needed
  !    Single humidistat if humidification and dehumidification are both active and control zones are the same
  IF (.NOT. isHumidifierNone) THEN
    !Object ==> ZoneControl:Humidistat
    CALL CreateNewObj('ZoneControl:Humidistat')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Humidistat')
    CALL AddToObjFld('Zone Name', base + uhpsHumidCtrlZoneOff,'')
    CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name', &
                     'HVACTemplate-Always ' // TRIM(FldVal(base + uhpsHumidSetPtOff)))
    CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name', ' ',.TRUE.)
    CALL AddAlwaysSchedule(FldVal(base + uhpsHumidSetPtOff))
    !Object ==> SetpointManager:SingleZone:Humidity:Minimum
    CALL CreateNewObj('SetpointManager:SingleZone:Humidity:Minimum')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Humidification Setpoint Manager')
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + uhpsAirHandlerNameOff,' Humidifier Outlet')
    CALL AddToObjFld('Control Zone Air Node Name', base + uhpsHumidCtrlZoneOff,' Zone Air Node',.TRUE.)
    IF (humidifierKind .EQ. humidifyElecSteam) THEN
      !Object ==> Humidifier:Steam:Electric
      CALL CreateNewObj('Humidifier:Steam:Electric')
      CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Humidifier')
      CALL AddToObjFld('Availability Schedule Name', base + uhpsHumidAvailSchedNameOff,' ')
      CALL AddToObjFld('Rated Capacity {m3/s}', base + uhpsHumidCapOff,'')
      CALL AddToObjFld('Rated Power {W}', base + uhpsHumidRatedElecOff,'')
      CALL AddToObjStr('rated fan power {W}','0')
      CALL AddToObjStr('Standby power {W}','0')
      CALL AddToObjFld('Air Inlet Node Name', base + uhpsAirHandlerNameOff,' Mixed Air Outlet')
      CALL AddToObjFld('Air Outlet Node Name', base + uhpsAirHandlerNameOff,' Humidifier Outlet',.TRUE.)
    END IF
  END IF
  SchType=' '
  !Object ==> AirLoopHVAC:UnitaryHeatPump:AirToAir
  CALL CreateNewObj('AirLoopHVAC:UnitaryHeatPump:AirToAir')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Heat Pump')
  CALL AddToObjStr('Availability Schedule Name', ' ')
  CALL AddToObjFld('Air Inlet Node Name', base + uhpsAirHandlerNameOff, TRIM(unitInlet))
  CALL AddToObjFld('Air Outlet Node Name', base + uhpsAirHandlerNameOff, TRIM(unitOutlet))
  CALL AddToObjFld('Cooling Supply Air Flow Rate', base + uhpsSupplyCoolFlowRateOff,'')
  CALL AddToObjFld('Heating Supply Air Flow Rate', base + uhpsSupplyHeatFlowRateOff,'')
  CALL AddToObjFld('No Load Supply Air Flow Rate', base + uhpsSupplyNoLoadFlowRateOff,'')
  CALL AddToObjFld('Controlling Zone or Thermostat Location', base + uhpsControlZoneOff,'')
  !CALL AddToObjStr('Fraction of Supply Air Flow That Goes Through the Controlling Zone','autosize')
  CALL AddToObjStr('Supply Air Fan Object Type','Fan:OnOff')
  CALL AddToObjFld('Supply Air Fan Name', base + uhpsAirHandlerNameOff,' Supply Fan')
  CALL AddToObjStr('Heating Coil Object Type', 'Coil:Heating:DX:SingleSpeed')
  CALL AddToObjFld('Heating Coil Name', base + uhpsAirHandlerNameOff,' HP Heating Coil')
  CALL AddToObjStr('Cooling Coil Object Type', 'Coil:Cooling:DX:SingleSpeed')
  CALL AddToObjFld('Cooling Coil Name', base + uhpsAirHandlerNameOff,' Cooling Coil')
  IF (supHeatCoilKind .EQ. hcGas) THEN
    CALL AddToObjStr('Supplemental Heating Coil Object Type','Coil:Heating:Fuel')
    CALL AddToObjFld('Supplemental Heating Coil Name', base + uhpsAirHandlerNameOff,' Sup Heat Coil')
  ELSEIF (supHeatCoilKind .EQ. hcWater) THEN
    CALL AddToObjStr('Supplemental Heating Coil Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Supplemental Heating Coil Name', base + uhpsAirHandlerNameOff,' Sup Heat Coil')
  ELSE
    CALL AddToObjStr('Supplemental Heating Coil Object Type','Coil:Heating:Electric')
    CALL AddToObjFld('Supplemental Heating Coil Name', base + uhpsAirHandlerNameOff,' Sup Heat Coil')
  END IF
  CALL AddToObjStr('Maximum Supply Air Temperature from Supplemental Heater', 'Autosize')
  CALL AddToObjFld('Maximum Outdoor Dry-Bulb Temperature for Supplemental Heater Operation', base + uhpsSuppHeatMaxODBOff,'')
  CALL AddToObjFld('Fan Placement', base + uhpsSupplyFanPlaceOff,'')
  IF (isSupplyFanCycling) THEN
    CALL AddToObjStr('Supply Air Fan Operating Mode Schedule Name','HVACTemplate-Always 0',.TRUE.)
    SchType='0'
  ELSE
    CALL AddToObjFld('Supply Air Fan Operating Mode Schedule Name', base + uhpsSupplyFanOpModeSchedNameOff,'',.TRUE.)
  END IF
  ! add new schedule if needed (from Supply Fan Operating Mode Schedule, above)
  IF (SchType /= ' ') THEN
    CALL AddAlwaysSchedule(SchType)
  ENDIF
  !Object ==> Coil:Heating:DX:SingleSpeed
  CALL CreateNewObj('Coil:Heating:DX:SingleSpeed')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' HP Heating Coil')
  CALL AddToObjFld('Availability Schedule Name', base + uhpsHeatAvailSchedNameOff,'')
  CALL AddToObjFld('Rated Total Heating Capacity {W}', base + uhpsHPHeatCoilCapOff,'')
  CALL AddToObjFld('Rated COP', base + uhpsHPHeatCoilCOPOff,'')
  CALL AddToObjFld('Rated Air Flow Rate {m3/s}', base + uhpsSupplyHeatFlowRateOff,'')
  CALL AddToObjStr('Rated Evaporator Fan Power Per Volume Flow Rate','')
  CALL AddToObjFld('Air Inlet Node Name', base + uhpsAirHandlerNameOff,' Cooling Coil Outlet')
  CALL AddToObjFld('Air Outlet Node Name', base + uhpsAirHandlerNameOff,' Heating Coil Outlet')
  CALL AddToObjFld('Total Heating Capacity Function of Temperature Curve Name',   &
     base + uhpsAirHandlerNameOff,' HP Heating Coil Cap-FT')
  CALL AddToObjFld('Total Heating Capacity Function of Flow Fraction Curve Name',   &
     base + uhpsAirHandlerNameOff,' HP Heating Coil Cap-FF')
  CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
     base + uhpsAirHandlerNameOff,' HP Heating Coil EIR-FT')
  CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
     base + uhpsAirHandlerNameOff,' HP Heating Coil EIR-FF')
  CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
     base + uhpsAirHandlerNameOff,' HP Heating Coil PLF')
  CALL AddToObjFld('Defrost Energy Input Ratio Function of Temperature Curve Name',   &
     base + uhpsAirHandlerNameOff,' HP Heating Coil DefrEIR-FT')
  CALL AddToObjFld('Minimum Outdoor Dry-Bulb Temperature for Compressor Operation {C}', base + uhpsHPHeatMinODBOff,'')
  CALL AddToObjStr('Outdoor Dry-Bulb Temperature to Turn On Compressor','')
  CALL AddToObjFld('Maximum Outdoor Dry-Bulb Temperature for Defrost Operation {C}', base + uhpsHPDefrMaxODBOff,'')
  CALL AddToObjStr('Crankcase Heater Capacity {W}','0')
  CALL AddToObjStr('Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater Operation {C}','0')
  CALL AddToObjFld('Defrost Strategy', base + uhpsHPDefrStrategyOff,'')
  CALL AddToObjFld('Defrost Control', base + uhpsHPDefrCtrlTypeOff,'')
  CALL AddToObjFld('Defrost Time Period Fraction', base + uhpsHPDefrTimeFracOff,'')
  CALL AddToObjStr('Resistive Defrost Heater Capacity {W}','autosize',.TRUE.)
  !Object ==> Curve:Cubic
  CALL CreateNewObj('Curve:Cubic')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' HP Heating Coil Cap-FT')
  CALL AddToObjStr('Coefficient1 Constant','0.758746')
  CALL AddToObjStr('Coefficient2 x','0.027626')
  CALL AddToObjStr('Coefficient3 x**2','0.000148716')
  CALL AddToObjStr('Coefficient4 x**3','0.0000034992')
  CALL AddToObjStr('Minimum Value of x','-20.0')
  CALL AddToObjStr('Maximum Value of x','20.0',.TRUE.)
  !Object ==> Curve:Cubic
  CALL CreateNewObj('Curve:Cubic')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' HP Heating Coil Cap-FF')
  CALL AddToObjStr('Coefficient1 Constant','0.84')
  CALL AddToObjStr('Coefficient2 x','0.16')
  CALL AddToObjStr('Coefficient3 x**2','0.0')
  CALL AddToObjStr('Coefficient4 x**3','0.0')
  CALL AddToObjStr('Minimum Value of x','0.5')
  CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
  !Object ==> Curve:Cubic
  CALL CreateNewObj('Curve:Cubic')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' HP Heating Coil EIR-FT')
  CALL AddToObjStr('Coefficient1 Constant','1.19248')
  CALL AddToObjStr('Coefficient2 x','-0.0300438')
  CALL AddToObjStr('Coefficient3 x**2','0.00103745')
  CALL AddToObjStr('Coefficient4 x**3','-0.000023328')
  CALL AddToObjStr('Minimum Value of x','-20.0')
  CALL AddToObjStr('Maximum Value of x','20.0',.TRUE.)
  !Object ==> Curve:Quadratic
  CALL CreateNewObj('Curve:Quadratic')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' HP Heating Coil EIR-FF')
  CALL AddToObjStr('Coefficient1 Constant','1.3824')
  CALL AddToObjStr('Coefficient2 x','-0.4336')
  CALL AddToObjStr('Coefficient3 x**2','0.0512')
  CALL AddToObjStr('Minimum Value of x','0.0')
  CALL AddToObjStr('Maximum Value of x','1.0',.TRUE.)
  !Object ==> Curve:Quadratic
  CALL CreateNewObj('Curve:Quadratic')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' HP Heating Coil PLF')
  CALL AddToObjStr('Coefficient1 Constant','0.75')
  CALL AddToObjStr('Coefficient2 x','0.25')
  CALL AddToObjStr('Coefficient3 x**2','0.0')
  CALL AddToObjStr('Minimum Value of x','0.0')
  CALL AddToObjStr('Maximum Value of x','1.0',.TRUE.)
  !Object ==> Curve:Biquadratic
  CALL CreateNewObj('Curve:Biquadratic')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' HP Heating Coil DefrEIR-FT')
  CALL AddToObjStr('Coefficient1 Constant','1')
  CALL AddToObjStr('Coefficient2 x','0')
  CALL AddToObjStr('Coefficient3 x**2','0')
  CALL AddToObjStr('Coefficient4 y','0')
  CALL AddToObjStr('Coefficient5 y**2','0')
  CALL AddToObjStr('Coefficient6 x*y','0')
  CALL AddToObjStr('Minimum Value of x','0')
  CALL AddToObjStr('Maximum Value of x','50')
  CALL AddToObjStr('Minimum Value of y','0')
  CALL AddToObjStr('Maximum Value of y','50',.TRUE.)
  ! Supplemental Heating Coil
  IF (supHeatCoilKind .EQ. hcElectric) THEN
    !Object ==> Coil:Heating:Electric
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Sup Heat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + uhpsSuppHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Efficiency','1')
    CALL AddToObjFld('Nominal Capacity of the Coil {W}', base + uhpsSuppHeatCoilCapOff,'')
    CALL AddToObjFld('Air Inlet Node Name', base + uhpsAirHandlerNameOff, TRIM(supheatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + uhpsAirHandlerNameOff, TRIM(supheatCoilOutlet))
    CALL AddToObjStr('Coil Temp Setpoint Node','',.TRUE.)
  ELSE IF (supHeatCoilKind .EQ. hcGas) THEN
    !Object ==> Coil:Heating:Fuel
    CALL CreateNewObj('Coil:Heating:Fuel')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Sup Heat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + uhpsSuppHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Fuel Type', 'Gas')
    CALL AddToObjFld('Gas Burner Efficiency', base + uhpsSuppHeatEfficiencyOff,' ')
    CALL AddToObjFld('Nominal Capacity of the Coil {W}', base + uhpsSuppHeatCoilCapOff,'')
    CALL AddToObjFld('Air Inlet Node Name', base + uhpsAirHandlerNameOff, TRIM(supheatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + uhpsAirHandlerNameOff, TRIM(supheatCoilOutlet))
    CALL AddToObjStr('Coil Temp Setpoint Node','')
    CALL AddToObjFld('Parasitic Electric Load {W}', base + uhpsSuppHeatParasiticOff,' ')
    CALL AddToObjFld('Part load fraction correlation (function of part load ratio)',   &
       base + uhpsAirHandlerNameOff,' Sup Heating Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
    !Object ==> Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Sup Heating Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  ELSE IF (supHeatCoilKind .EQ. hcWater) THEN
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Sup Heat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + uhpsHeatAvailSchedNameOff,'')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name', base + uhpsAirHandlerNameOff,' Sup Heat Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + uhpsAirHandlerNameOff,' Sup Heat Coil HW Outlet')
    CALL AddToObjFld('Air Inlet Node Name', base + uhpsAirHandlerNameOff, TRIM(supheatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + uhpsAirHandlerNameOff, TRIM(supheatCoilOutlet))
    IF (isSuppHeatCoilCapAutosize) THEN
      CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
    ELSE
      CALL AddToObjStr('Performance Input Method','NominalCapacity')
    END IF
    CALL AddToObjFld('Rated Capacity', base + uhpsSuppHeatCoilCapOff,'')
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Sup Heat Coil HW Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + uhpsAirHandlerNameOff,' Sup Heat Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + uhpsAirHandlerNameOff,' Sup Heat Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + uhpsAirHandlerNameOff,' Sup Heat Coil HW Outlet',.TRUE.)
    CALL AddToStrList(TRIM(FldVal(base + uhpsAirHandlerNameOff)) // ' Sup Heat Coil HW Branch',handleHeatingCoilBranch)
  END IF
  ! Cooling coil
  !Object ==> Coil:Cooling:DX:SingleSpeed
  CALL CreateNewObj('Coil:Cooling:DX:SingleSpeed')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Cooling Coil')
  CALL AddToObjFld('Availability Schedule Name', base + uhpsCoolAvailSchedNameOff,' ')
  CALL AddToObjFld('Gross Rated Total Cooling Capacity {W}', base + uhpsCoolCoilCapOff,' ')
  CALL AddToObjFld('Gross Rated Sensible Heat Ratio', base + uhpsCoolCoilSHROff,' ')
  CALL AddToObjFld('Rated COP', base + uhpsCoolCoilCOPOff,' ')
  CALL AddToObjFld('Rated Air Flow Rate {m3/s}', base + uhpsSupplyCoolFlowRateOff,' ')
  CALL AddToObjStr('Rated Evaporator Fan Power per Volume Flow Rate {W/(m3/s)}','')
  CALL AddToObjFld('Air Inlet Node Name', base + uhpsAirHandlerNameOff, TRIM(coolCoilInlet))
  CALL AddToObjFld('Air Outlet Node Name', base + uhpsAirHandlerNameOff, TRIM(coolCoilOutlet))
  CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
     base + uhpsAirHandlerNameOff,' Cool Coil Cap-FT')
  CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
     base + uhpsAirHandlerNameOff,' Cool Coil Cap-FF')
  CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
     base + uhpsAirHandlerNameOff,' Cool Coil EIR-FT')
  CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
     base + uhpsAirHandlerNameOff,' Cool Coil EIR-FF')
  CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
     base + uhpsAirHandlerNameOff,' Cool Coil PLF')
  CALL AddToObjStr('Minimum Outdoor Dry-Bulb Temperature for Compressor Operation {C}','')
  CALL AddToObjStr('Nominal Time for Condensate Removal to Begin','0')
  CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity','0')
  CALL AddToObjStr('Maximum Cycling Rate','0')
  CALL AddToObjStr('Latent Capacity Time Constant','0')
  CALL AddToObjFld('Condenser Air Inlet Node Name',base + uhpsAirHandlerNameOff,' Cooling Coil Condenser Inlet')
  CALL AddToObjStr('Condenser Type','AirCooled')
  CALL AddToObjStr('Evaporative Condenser Effectiveness','0')
  CALL AddToObjStr('Evaporative Condenser Air Flow Rate','')
  CALL AddToObjStr('Evaporative Condenser Pump Rated Power Consumption','0')
  CALL AddToObjStr('Crankcase Heater Capacity','0')
  CALL AddToObjStr('Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater Operation','10',.TRUE.)
  !Object ==> Curve:Biquadratic
  CALL CreateNewObj('Curve:Biquadratic')
  CALL WriteComment('DOE-2.1E, COOL-CAP-FT for PTAC w/ SI temps')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Cool Coil Cap-FT')
  CALL AddToObjStr('Coefficient1 Constant','0.942587793')
  CALL AddToObjStr('Coefficient2 x','0.009543347')
  CALL AddToObjStr('Coefficient3 x**2','0.00068377')
  CALL AddToObjStr('Coefficient4 y','-0.011042676')
  CALL AddToObjStr('Coefficient5 y**2','0.000005249')
  CALL AddToObjStr('Coefficient6 x*y','-0.00000972')
  CALL AddToObjStr('Minimum Value of x','12.77778')
  CALL AddToObjStr('Maximum Value of x','23.88889')
  CALL AddToObjStr('Minimum Value of y','18.0')
  CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
  !Object ==> Curve:Quadratic
  CALL CreateNewObj('Curve:Quadratic')
  CALL WriteComment('DOE-2.1E, RATED-CCAP-FFLOW for PTAC')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Cool Coil Cap-FF')
  CALL AddToObjStr('Coefficient1 Constant','0.8')
  CALL AddToObjStr('Coefficient2 x','0.2')
  CALL AddToObjStr('Coefficient3 x**2','0')
  CALL AddToObjStr('Minimum Value of x','0.5')
  CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
  !Object ==> Curve:Biquadratic
  CALL CreateNewObj('Curve:Biquadratic')
  CALL WriteComment('DOE-2.1E, COOL-EIR-FT for PTAC w/ SI temps')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Cool Coil EIR-FT')
  CALL AddToObjStr('Coefficient1 Constant','0.342414409')
  CALL AddToObjStr('Coefficient2 x','0.034885008')
  CALL AddToObjStr('Coefficient3 x**2','-0.0006237')
  CALL AddToObjStr('Coefficient4 y','0.004977216')
  CALL AddToObjStr('Coefficient5 y**2','0.000437951')
  CALL AddToObjStr('Coefficient6 x*y','-0.000728028')
  CALL AddToObjStr('Minimum Value of x','12.77778')
  CALL AddToObjStr('Maximum Value of x','23.88889')
  CALL AddToObjStr('Minimum Value of y','18.0')
  CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
  !Object ==> Curve:Quadratic
  CALL CreateNewObj('Curve:Quadratic')
  CALL WriteComment('DOE-2.1E, RATED-CEIR-FFLOW for PTAC')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Cool Coil EIR-FF')
  CALL AddToObjStr('Coefficient1 Constant','1.1552')
  CALL AddToObjStr('Coefficient2 x','-0.1808')
  CALL AddToObjStr('Coefficient3 x**2','0.0256')
  CALL AddToObjStr('Minimum Value of x','0.5')
  CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
  !Object ==> Curve:Quadratic
  CALL CreateNewObj('Curve:Quadratic')
  CALL WriteComment('PLF = l.- Cd(1.-PLR) where Cd = 0.15')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Cool Coil PLF')
  CALL AddToObjStr('Coefficient1 Constant','0.85')
  CALL AddToObjStr('Coefficient2 x','0.15')
  CALL AddToObjStr('Coefficient3 x**2','0')
  CALL AddToObjStr('Minimum Value of x','0')
  CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  !Object ==> OutdoorAir:Node
  CALL CreateNewObj('OutdoorAir:Node')
  CALL AddToObjFld('Name',base + uhpsAirHandlerNameOff,' Cooling Coil Condenser Inlet')
  CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
  ! Return fan
  IF (isReturnFanYes) THEN
    CALL CreateNewObj('Fan:ConstantVolume')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Return Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + uhpsSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + uhpsReturnEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + uhpsReturnPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Motor Efficiency', base + uhpsReturnFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + uhpsReturnFanFractionOff,' ')
    CALL AddToObjFld('Air Inlet Node Name', base + uhpsAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Air Outlet Node Name', base + uhpsAirHandlerNameOff,' Return Fan Outlet',.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF
  !Object ==> Fan:OnOff
  CALL CreateNewObj('Fan:OnOff')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Supply Fan')
  IF (isSysAvailSchedBlank) THEN
    CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
  ELSE
    CALL AddToObjFld('Availability Schedule Name', base + uhpsSysAvailSchedNameOff,' ')
  END IF
  CALL AddToObjFld('Fan Efficiency', base + uhpsSupplyEfficiencyOff,' ')
  CALL AddToObjFld('Pressure Rise {Pa}', base + uhpsSupplyPressureOff,' ')
  CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
  CALL AddToObjFld('Motor Efficiency', base + uhpsSupplyFanMotorEffOff,' ')
  CALL AddToObjFld('Motor in Airstream Fraction', base + uhpsSupplyFanFractionOff,' ')
  CALL AddToObjFld('Air Inlet Node Name', base + uhpsAirHandlerNameOff, TRIM(supFanInlet))
  CALL AddToObjFld('Air Outlet Node Name', base + uhpsAirHandlerNameOff, TRIM(supFanOutlet),.TRUE.)
  IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  !Object ==> OutdoorAir:NodeList
  CALL CreateNewObj('OutdoorAir:NodeList')
  CALL AddToObjFld('Node or NodeList Name 1', base + uhpsAirHandlerNameOff,' Outside Air Inlet',.TRUE.)
  !OObject ==> AirLoopHVAC:OutdoorAirSystem
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' OA System')
  CALL AddToObjFld('Controller List Name', base + uhpsAirHandlerNameOff,' OA System Controllers')
  CALL AddToObjFld('Outdoor Air Equipment List Name', base + uhpsAirHandlerNameOff,' OA System Equipment')
  CALL AddToObjFld('Availability Manager List Name', base + uhpsAirHandlerNameOff,' Availability Managers',.TRUE.)
  !Object ==> AirLoopHVAC:ControllerList
  CALL CreateNewObj('AirLoopHVAC:ControllerList')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' OA System Controllers')
  CALL AddToObjStr('Controller Object Type','Controller:OutdoorAir')
  CALL AddToObjFld('Controller Name', base + uhpsAirHandlerNameOff,' OA Controller',.TRUE.)
  !Object ==> AirLoopHVAC:OutdoorAirSystem:EquipmentList
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem:EquipmentList')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' OA System Equipment')
  IF (isHeatRecSensible) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + uhpsAirHandlerNameOff,' Heat Recovery')
  END IF
  IF (isHeatRecEnthalpy) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + uhpsAirHandlerNameOff,' Heat Recovery')
  END IF
  CALL AddToObjStr('Component Object Type','OutdoorAir:Mixer')
  CALL AddToObjFld('Component Name', base + uhpsAirHandlerNameOff,' OA Mixing Box',.TRUE.)
  !Object ==> OutdoorAir:Mixer
  CALL CreateNewObj('OutdoorAir:Mixer')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' OA Mixing Box')
  CALL AddToObjFld('Mixed Air Node Name', base + uhpsAirHandlerNameOff,' Mixed Air Outlet')
  IF (isHeatRecNone) THEN
    CALL AddToObjFld('Outside Air Stream Node Name', base + uhpsAirHandlerNameOff,' Outside Air Inlet')
  ELSEIF (isHeatRecSensible) THEN
    CALL AddToObjFld('Outside Air Stream Node Name', base + uhpsAirHandlerNameOff,' Heat Recovery Supply Outlet')
  ELSEIF (isHeatRecEnthalpy) THEN
    CALL AddToObjFld('Outside Air Stream Node Name', base + uhpsAirHandlerNameOff,' Heat Recovery Supply Outlet')
  ELSE
    CALL AddToObjFld('Outside Air Stream Node Name', base + uhpsAirHandlerNameOff,' Outside Air Inlet')
    CALL WriteError('Warning: Heat recovery option invalid. Assuming none.')
    isHeatRecNone = .TRUE.
  END IF
  CALL AddToObjFld('Relief Air Stream Node Name', base + uhpsAirHandlerNameOff,' Relief Air Outlet')
  CALL AddToObjStr('Return Air Stream Node Name', TRIM(returnInletToOAMIxer),.TRUE.)
  !Object ==> SetpointManager:SingleZone:Cooling
  CALL CreateNewObj('SetpointManager:SingleZone:Cooling')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Economizer Supply Air Temp Manager')
  CALL AddToObjStr('Control Variable','Temperature')
  CALL AddToObjStr('minimum supply air temperature {C}','13')
  CALL AddToObjStr('maximum supply air temperature {C}','45')
  CALL AddToObjFld('Control Zone Name', base + uhpsControlZoneOff,'')
  CALL AddToObjFld('Zone Node Name', base + uhpsControlZoneOff,' Zone Air Node')
  IF (isControlZonePlenumBlank) THEN
    CALL AddToObjFld('Zone Inlet Node Name', base + uhpsControlZoneOff,' Zone Equip Inlet')
  ELSE
    CALL AddToObjFld('Zone Inlet Node Name', base + uhpsControlZoneOff,' Supply Inlet')
  END IF
  CALL AddToObjFld('Setpoint Node or NodeList Name', base + uhpsAirHandlerNameOff,' Air Loop Outlet',.TRUE.)
  !Object ==> SetpointManager:MixedAir
  CALL CreateNewObj('SetpointManager:MixedAir')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Cooling Coil Air Temp Manager')
  CALL AddToObjStr('Control Variable','Temperature')
  CALL AddToObjFld('Reference Setpoint Node Name', base + uhpsAirHandlerNameOff,' Air Loop Outlet')
  CALL AddToObjFld('Fan Inlet Node Name', base + uhpsAirHandlerNameOff, TRIM(supFanInlet))
  CALL AddToObjFld('Fan Outlet Node Name', base + uhpsAirHandlerNameOff, TRIM(supFanOutlet))
  CALL AddToObjFld('Setpoint Node or NodeList Name', base + uhpsAirHandlerNameOff,' Mixed Air Outlet',.TRUE.)
  !Object ==> Controller:OutdoorAir
  CALL CreateNewObj('Controller:OutdoorAir')
  CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' OA Controller')
  CALL AddToObjFld('Relief Air Outlet Node Name', base + uhpsAirHandlerNameOff,' Relief Air Outlet')
  CALL AddToObjStr('Return Air Node Name',TRIM(returnInletToOAMIxer))
  CALL AddToObjFld('Mixed Air Node Name', base + uhpsAirHandlerNameOff,' Mixed Air Outlet')
  CALL AddToObjFld('Actuator Node Name', base + uhpsAirHandlerNameOff,' Outside Air Inlet')
  CALL AddToObjFld('Minimum Outdoor Air Flow Rate {m3/s}', base + uhpsMinOutsideFlowOff,' ')
  CALL AddToObjFld('Maximum Outdoor Air Flow Rate {m3/s}', base + uhpsMaxOutsideFlowOff,' ')
  ! usEconomizerKindOff is now a straight pass-thru to EconomizerChoice in
  ! the OA controller
  CALL AddToObjFld('Economizer Control Type', base + uhpsEconomizerKindOff,' ')
  ! If heat recovery is selected, then set economizer control to bypass
  ! otherwise, set to modulate OA flow
  IF (isHeatRecNone) THEN
    CALL AddtoObjStr('Economizer Control Action Type','ModulateFlow')
  ELSE
    CALL AddtoObjStr('Economizer Control Action Type','MinimumFlowWithBypass')
  ENDIF
  CALL AddToObjFld('Economizer Maximum Limit Dry-Bulb Temperature {C}', base + uhpsEconoUpTempLimitOff,' ')
  CALL AddToObjFld('Economizer Maximum Limit Enthalpy {J/kg}', base + uhpsEconoUpEnthLimitOff,' ')
  IF (econoKind .EQ. etFixedDewPointAndDryBulb) THEN
    CALL AddToObjFld('Economizer Maximum Limit Dewpoint Temperature (C)',base + uhpsEconoUpDewLimitOff, ' ')
  ELSE
    CALL AddToObjStr('Economizer Maximum Limit Dewpoint Temperature (C)',' ')
  ENDIF
  IF (econoKind .EQ. etElectronicEnthalpy) THEN
    CALL AddToObjFld('Electronic Enthalpy Limit Curve Name',base + uhpsAirHandlerNameOff,' ElectrEnthCurve')
  ELSE
    CALL AddToObjStr('Electronic Enthalpy Limit Curve Name',' ')
  ENDIF
  CALL AddToObjFld('Economizer Minimum Limit Dry-Bulb Temperature {C}', base + uhpsEconoLowTempLimitOff,' ')
  CALL AddToObjFld('Lockout Type', base + uhpsEconoLockoutKindOff,' ')
  CALL AddToObjStr('Minimum Limit Type', 'FixedMinimum')
  CALL AddToObjFld('Minimum Outdoor Air Schedule Name', base + uhpsMinOutFlowSchedNameOff,' ')
  CALL AddToObjStr('Minimum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Maximum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Mechanical Ventilation Controller Name','',.TRUE.)
  IF (econoKind .EQ. etElectronicEnthalpy) THEN
    !Object ==> Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' ElectrEnthCurve')
    CALL AddToObjStr('Coefficient1 Constant','0.01342704')
    CALL AddToObjStr('Coefficient2 x','-0.00047892')
    CALL AddToObjStr('Coefficient3 x**2','0.000053352')
    CALL AddToObjStr('Coefficient4 x**3','-0.0000018103')
    CALL AddToObjStr('Minimum Value of x','16.6')
    CALL AddToObjStr('Maximum Value of x','29.13',.TRUE.)
  ENDIF
  IF (isHeatRecSensible) THEN
    htRecSens75 = StringToReal(FldVal(base + uhpsHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    !Object ==> HeatExchanger:AirToAir:SensibleAndLatent
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + uhpsAirHandlerNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Nominal supply air flow rate {m3/s}', base + uhpsMinOutsideFlowOff,'')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + uhpsHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Heating Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Heating Air Flow','0')
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + uhpsHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Cooling Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Cooling Air Flow','0')
    CALL AddToObjFld('Supply Air Inlet Node Name', base + uhpsAirHandlerNameOff,' Outside Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + uhpsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + uhpsAirHandlerNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + uhpsAirHandlerNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','0')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjStr('Heat exchanger type','Plate')
    CALL AddToObjStr('Frost Control Type','ExhaustOnly')
    CALL AddToObjStr('Threshold Temperature','-1.1')
    CALL AddToObjStr('Initial Defrost Time Fraction','0.167')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','0.024',.TRUE.)
  END IF
  IF (isHeatRecEnthalpy) THEN
    htRecSens75 = StringToReal(FldVal(base + uhpsHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    htRecLat75 = StringToReal(FldVal(base + uhpsHeatRecLatEffOff)) + 0.05
    IF (htRecLat75 .GT. 1.) htRecLat75 = 1.
    !Object ==> HeatExchanger:AirToAir:SensibleAndLatent
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + uhpsAirHandlerNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Nominal supply air flow rate {m3/s}', base + uhpsMinOutsideFlowOff,'')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + uhpsHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Heating Air Flow', base + uhpsHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow',htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Heating Air Flow', htRecLat75)
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + uhpsHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Cooling Air Flow', base + uhpsHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow',htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Cooling Air Flow', htRecLat75)
    CALL AddToObjFld('Supply Air Inlet Node Name', base + uhpsAirHandlerNameOff,' Outside Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + uhpsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + uhpsAirHandlerNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + uhpsAirHandlerNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','0')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjStr('Heat exchanger type','Rotary')
    CALL AddToObjStr('Frost Control Type','MinimumExhaustTemperature')
    CALL AddToObjStr('Threshold Temperature','1.7')
    CALL AddToObjStr('Initial Defrost Time Fraction','')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','',.TRUE.)
  END IF
  IF (.NOT. isHeatRecNone) THEN
    IF (isEconoLowLimitBlank) THEN
      CALL AddAlwaysSchedule('5')
    ELSE
      CALL AddAlwaysSchedule(FldVal(base + uhpsEconoLowTempLimitOff))
    END IF
    !Object ==> SetpointManager:Scheduled
    CALL CreateNewObj('SetpointManager:Scheduled')
    CALL AddToObjFld('Name', base + uhpsAirHandlerNameOff,' Heat Recovery Air Temp Manager')
    CALL AddToObjStr('Control Variable','Temperature')
    IF (isEconoLowLimitBlank) THEN
      CALL AddToObjStr('Schedule Name','HVACTemplate-Always 5')
    ELSE
      CALL AddToObjStr('Schedule Name','HVACTemplate-Always ' // FldVal(base + uhpsEconoLowTempLimitOff))
    END IF
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + uhpsAirHandlerNameOff,' Heat Recovery Supply Outlet',.TRUE.)
  END IF
END DO
!======================================================================================================
! create Systems for HVACTemplate:System:UnitarySystem
!   not developed using the spreadsheet so no spreadsheet reference lines are included
!======================================================================================================
DO iSys = 1, numCompactSysUnitarySystem
  base =  compactSysUnitarySystemBase(iSys)
  ! find base for controlled zone used by this system
  controlZoneBase = -1
  DO iZone = 1, numCompactZoneUnit
    IF (SameString(FldVal(compactZoneUnitBase(iZone) + uzNameOff), FldVal(base +  ussControlZoneOff))) THEN
      controlZoneBase = compactZoneUnitBase(iZone)
      EXIT
    END IF
  END DO
  IF (controlZoneBase .EQ. -1) THEN
    CALL WriteError('In HVACTemplate:System:UnitarySystem "'//TRIM(FldVal(base + ussAirHandlerNameOff))// &
     '" for the field Control Zone Name could not find a matching HVACTemplate:Zone:Unitary named: '// &
     TRIM(FldVal(base +  ussControlZoneOff)))
  END IF
  ! set flags
  isSupplyCoolRateAutosize = SameString(FldVal(base + ussSupplyCoolFlowRateOff),'autosize')
  isSupplyHeatRateAutosize = SameString(FldVal(base + ussSupplyHeatFlowRateOff),'autosize')
  isMinOARateAutosize = SameString(FldVal(base + ussMinOutsideFlowOff),'autosize')
  isSupplyPlenumBlank = (FldVal(base + ussSupplyPlenumNameOff) .EQ. '')
  isReturnPlenumBlank = (FldVal(base + ussReturnPlenumNameOff) .EQ. '')
  isSysAvailSchedBlank = (FldVal(base + ussSysAvailSchedNameOff) .EQ. '')
  ! Supply fan mode may be blank (default to Cycling) or a schedule name
  isSupplyFanCycling = SameString(FldVal(base + ussFanModeSchedOff),'')
  doSuppressMsgZlUnitAll = .FALSE.
  ! set fan to constant volume on/off initially, reset to variable volume if a multispeed or multi-stage coil is selected
  supFanKind = sfkCAV
  supFanObjectType = 'Fan:OnOff'
  ! set multi-speed flag initially to false, reset to true if a multispeed or multi-stage coil is selected
  isMultiSpeed = .FALSE.
  coolCoilObjectName = TRIM(FldVal(base + ussAirHandlerNameOff))//' Cooling Coil'
  ! set cooling coil type
 IF (SameString(FldVal(base + ussCoolCoilKindOff),'None')) THEN
    coolCoilKind = ccNone
    coolCoilObjectType = ''
    coolCoilObjectName = ''
  ELSEIF (SameString(FldVal(base + ussCoolCoilKindOff),'SingleSpeedDX')) THEN
    coolCoilKind = ccSingleSpeedDX
    coolCoilObjectType = 'Coil:Cooling:DX:SingleSpeed'
  ELSEIF (SameString(FldVal(base + ussCoolCoilKindOff),'TwoSpeedDX')) THEN
    coolCoilKind = ccTwoSpeedDX
    coolCoilObjectType = 'Coil:Cooling:DX:TwoSpeed'
    supFanKind = sfkVAV
    supFanObjectType = 'Fan:VariableVolume'
    isMultiSpeed = .FALSE. ! because the twospeed coil isn't really multi-speed
  ELSEIF (SameString(FldVal(base + ussCoolCoilKindOff),'MultiSpeedDX')) THEN
    coolCoilKind = ccMultiSpeedDX
    coolCoilObjectType = 'Coil:Cooling:DX:MultiSpeed'
    supFanKind = sfkVAV
    supFanObjectType = 'Fan:VariableVolume'
    isMultiSpeed = .TRUE.
  ELSEIF (SameString(FldVal(base + ussCoolCoilKindOff),'TwoStageDX')) THEN
    coolCoilKind = ccTwoStageDX
    coolCoilObjectType = 'Coil:Cooling:DX:TwoStageWithHumidityControlMode'
  ELSEIF (SameString(FldVal(base + ussCoolCoilKindOff),'TwoStageHumidityControlDX')) THEN
    coolCoilKind = ccTwoStageHumidControlDX
    coolCoilObjectType = 'Coil:Cooling:DX:TwoStageWithHumidityControlMode'
  ELSEIF (SameString(FldVal(base + ussCoolCoilKindOff),'HeatExchangerAssistedDX')) THEN
    coolCoilKind = ccDXcoilSystemHX
    coolCoilObjectType = 'CoilSystem:Cooling:DX:HeatExchangerAssisted'
    coolCoilObjectName = TRIM(FldVal(base + ussAirHandlerNameOff))//'Heat Exchanger Assisted Cooling Coil'
  ELSEIF (SameString(FldVal(base + ussCoolCoilKindOff),'SingleSpeedDXWaterCooled')) THEN
    coolCoilKind = ccSingleSpeedDXWater
    coolCoilObjectType = 'Coil:Cooling:WaterToAirHeatPump:EquationFit'

  ELSEIF (SameString(FldVal(base + ussCoolCoilKindOff),'ChilledWater')) THEN
    coolCoilKind = ccChWater
    coolCoilObjectType = 'Coil:Cooling:Water'
  ELSEIF (SameString(FldVal(base + ussCoolCoilKindOff),'ChilledWaterDetailedFlatModel')) THEN
    coolCoilKind = ccChWaterDetailed
    coolCoilObjectType = 'Coil:Cooling:Water:DetailedGeometry'
  ELSEIF (SameString(FldVal(base + ussCoolCoilKindOff),'HeatExchangerAssistedChilledWater')) THEN
    coolCoilKind = ccWaterCoilSystemHX
    coolCoilObjectType = 'CoilSystem:Cooling:Water:HeatExchangerAssisted'
    coolCoilObjectName = TRIM(FldVal(base + ussAirHandlerNameOff))//'Heat Exchanger Assisted Cooling Coil'
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:UnitarySystem "'//TRIM(FldVal(base + ussAirHandlerNameOff))//'"'// &
                    ' in the Cooling Coil Type field: '//TRIM(FldVal(base + ussCoolCoilKindOff)))
  END IF
  ! set heating coil type
  heatCoilObjectName = TRIM(FldVal(base + ussAirHandlerNameOff))//' Heating Coil'
  IF (SameString(FldVal(base + ussHeatCoilKindOff),'None')) THEN
    heatCoilKind = hcNone
    heatCoilObjectType = ''
    heatCoilObjectName = ''
  ELSEIF (SameString(FldVal(base + ussHeatCoilKindOff),'Electric')) THEN
    heatCoilKind = hcElectric
    heatCoilObjectType = 'Coil:Heating:Electric'
  ELSEIF (SameString(FldVal(base + ussHeatCoilKindOff),'Gas')) THEN
    heatCoilKind = hcGas
    heatCoilObjectType = 'Coil:Heating:Fuel'
  ELSEIF (SameString(FldVal(base + ussHeatCoilKindOff),'HotWater')) THEN
    heatCoilKind = hcWater
    heatCoilObjectType = 'Coil:Heating:Water'
  ELSEIF (SameString(FldVal(base + ussHeatCoilKindOff),'SingleSpeedDXHeatPumpAirSource')) THEN
    heatCoilKind = hcSingleSpeedDXHPAir
    heatCoilObjectType = 'Coil:Heating:DX:SingleSpeed'
  ELSEIF (SameString(FldVal(base + ussHeatCoilKindOff),'MultiSpeedDXHeatPumpAirSource')) THEN
    heatCoilKind = hcMultiSpeedDXHPAir
    heatCoilObjectType = 'Coil:Heating:DX:MultiSpeed'
    supFanKind = sfkVAV
    supFanObjectType = 'Fan:VariableVolume'
    isMultiSpeed = .TRUE.
  ELSEIF (SameString(FldVal(base + ussHeatCoilKindOff),'SingleSpeedDXHeatPumpWaterSource')) THEN
    heatCoilKind = hcSingleSpeedDXHPWater
    heatCoilObjectType = 'Coil:Heating:WaterToAirHeatPump:EquationFit'
  ELSEIF (SameString(FldVal(base + ussHeatCoilKindOff),'MultiStageElectric')) THEN
    heatCoilKind = hcMultiStageElectric
    heatCoilObjectType = 'Coil:Heating:Electric:MultiStage'
    supFanKind = sfkVAV
    supFanObjectType = 'Fan:VariableVolume'
    isMultiSpeed = .TRUE.
  ELSEIF (SameString(FldVal(base + ussHeatCoilKindOff),'MultiStageGas')) THEN
    heatCoilKind = hcMultiStageGas
    heatCoilObjectType = 'Coil:Heating:Gas:MultiStage'
    supFanKind = sfkVAV
    supFanObjectType = 'Fan:VariableVolume'
    isMultiSpeed = .TRUE.
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:UnitarySystem "'//TRIM(FldVal(base + ussAirHandlerNameOff))//'"'// &
                    ' in the Heating Coil Type field: '//TRIM(FldVal(base + ussHeatCoilKindOff)))
  END IF
  !set the supplemental heat or reheat indicator
  supHeatCoilObjectName = TRIM(FldVal(base + ussAirHandlerNameOff))//' Supp Heat or Reheat Coil'
  IF (SameString(FldVal(base + ussSuppReHeatCoilKindOff),'None')) THEN
    supHeatCoilKind = hcNone
    supheatCoilObjectType = ''
    supHeatCoilObjectName = ''
  ELSEIF (SameString(FldVal(base + ussSuppReHeatCoilKindOff),'Electric')) THEN
    supHeatCoilKind = hcElectric
    supheatCoilObjectType = 'Coil:Heating:Electric'
  ELSEIF (SameString(FldVal(base + ussSuppReHeatCoilKindOff),'Gas')) THEN
    supHeatCoilKind = hcGas
    supheatCoilObjectType = 'Coil:Heating:Fuel'
  ELSEIF (SameString(FldVal(base + ussSuppReHeatCoilKindOff),'HotWater')) THEN
    supHeatCoilKind = hcWater
    supheatCoilObjectType = 'Coil:Heating:Water'
  ELSEIF (SameString(FldVal(base + ussSuppReHeatCoilKindOff),'DesuperHeater')) THEN
    supHeatCoilKind = hcDeSuperHeat
    supheatCoilObjectType = 'Coil:Heating:Desuperheater'
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:UnitarySystem "'//TRIM(FldVal(base + ussAirHandlerNameOff))//'"'// &
                    ' in the Supplemental Heating or Reheat Coil Type: '//TRIM(FldVal(base + ussSuppReHeatCoilKindOff)))
  END IF
  isSuppHeatCoilCapAutosize = SameString(FldVal(base + ussSuppReHeatCoilCapOff),'autosize')
  isHeatCapAutosize = SameString(FldVal(base + ussHeatCoilCapOff),'autosize')
  isCoolCapAutosize = SameString(FldVal(base + ussCoolCoilCapOff),'autosize')
  isSHRAutosize = SameString(FldVal(base + ussCoolCoilSHROff),'autosize')
  isHeatRecSensible = SameString(FldVal(base + ussHeatRecTypeOff),'Sensible')
  isHeatRecEnthalpy = SameString(FldVal(base + ussHeatRecTypeOff),'Enthalpy')
  isHeatRecNone = SameString(FldVal(base + ussHeatRecTypeOff),'None')
  isControlZonePlenumBlank = (FldVal(controlZoneBase + uzZoneSupplyPlenumNameOff) .EQ. '') !uses control zone base
  isEconoLowLimitBlank = (FldVal(base + ussEconoLowTempLimitOff) .EQ. '')
  isReturnFanYes = (FldVal(base + ussReturnFanPresentOff) .EQ. 'Yes')
  !set the economizer indicator
  IF (SameString(FldVal(base +  ussEconomizerKindOff),'FixedDryBulb')) THEN
    econoKind = etFixedDryBulb
  ELSEIF (SameString(FldVal(base +  ussEconomizerKindOff),'FixedEnthalpy')) THEN
    econoKind = etFixedEnthalpy
  ELSEIF (SameString(FldVal(base +  ussEconomizerKindOff),'DifferentialDryBulb')) THEN
    econoKind = etDifferentialDryBulb
  ELSEIF (SameString(FldVal(base +  ussEconomizerKindOff),'DifferentialEnthalpy')) THEN
    econoKind = etDifferentialEnthalpy
  ELSEIF (SameString(FldVal(base +  ussEconomizerKindOff),'FixedDewPointAndDryBulb')) THEN
    econoKind = etFixedDewPointAndDryBulb
  ELSEIF (SameString(FldVal(base +  ussEconomizerKindOff),'ElectronicEnthalpy')) THEN
    econoKind = etElectronicEnthalpy
  ELSEIF (SameString(FldVal(base +  ussEconomizerKindOff),'DifferentialDryBulbAndEnthalpy')) THEN
    econoKind = etDifferentialDryBulbAndEnthalpy
  ELSEIF (SameString(FldVal(base +  ussEconomizerKindOff),'NoEconomizer')) THEN
    econoKind = etNoEconomizer
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:UnitarySystem in the Economizer Type field')
  END IF
  !set the economizer lockout indicator
  IF (SameString(FldVal(base +  ussEconoLockoutKindOff),'NoLockout')) THEN
    lockoutKind = loNo
  ELSEIF (SameString(FldVal(base +  ussEconoLockoutKindOff),'LockoutWithHeating')) THEN
    lockoutKind = loHeat
  ELSEIF (SameString(FldVal(base +  ussEconoLockoutKindOff),'LockoutWithCompressor')) THEN
    lockoutKind = loComp
  ELSEIF (SameString(FldVal(base +  ussEconoLockoutKindOff),'')) THEN
    lockoutKind = loNo
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:UnitarySystem in the Economizer Lockout field')
  END IF
  !set the supply fan placement indicator
  IF (SameString(FldVal(base + ussSupplyFanPlaceOff), 'BlowThrough')) THEN
    supFanPlacement = sfpBlowThru
  ELSEIF (SameString(FldVal(base + ussSupplyFanPlaceOff), 'DrawThrough')) THEN
    supFanPlacement = sfpDrawThru
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:UnitarySystem "'//TRIM(FldVal(base + ussAirHandlerNameOff))//'"'// &
                    ' in the Supply Fan Placement field: '//TRIM(FldVal(base + ussSupplyFanPlaceOff)))
  END IF
  !set the dehumidification indicators
  isDehumidifyNone = SameString(FldVal(base + ussDehumCtrlTypeOff),'None')
  IF (SameString(FldVal(base +  ussDehumCtrlTypeOff),'None')) THEN
    dehumidCtrlKind = dehumidNone
  ELSEIF (SameString(FldVal(base +  ussDehumCtrlTypeOff),'CoolReheat')) THEN
    dehumidCtrlKind = dehumidCoolRht
  ELSEIF (SameString(FldVal(base +  ussDehumCtrlTypeOff),'Multimode')) THEN
    dehumidCtrlKind = dehumidMultimode
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:UnitarySystem "'//TRIM(FldVal(base + ussAirHandlerNameOff))//'"'// &
                    ' in the Dehumidification Control Type field: '//TRIM(FldVal(base + ussDehumCtrlTypeOff)))
    dehumidCtrlKind = 0
  END IF
  !set the humidifier indicators
  isHumidifierNone = SameString(FldVal(base + ussHumidCtrlTypeOff),'None')
  IF (SameString(FldVal(base +  ussHumidCtrlTypeOff),'None')) THEN
    humidifierKind = humidifyNone
  ELSEIF (SameString(FldVal(base +  ussHumidCtrlTypeOff),'ElectricSteam')) THEN
    humidifierKind = humidifyElecSteam
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:UnitarySystem "'//TRIM(FldVal(base + ussAirHandlerNameOff))//'"'// &
                    ' in the Humidifier Type field: '//TRIM(FldVal(base + ussHumidCtrlTypeOff)))
  END IF

    !set node name for return inlet to OA mixer
  IF (isReturnFanYes) THEN
    returnInletToOAMIxer = TRIM(FldVal(base + ussAirHandlerNameOff))//' Return Fan Outlet'
  ELSE
    returnInletToOAMIxer = TRIM(FldVal(base + ussAirHandlerNameOff))//' Air Loop Inlet'
  END IF

  !set component inlet and outlet node names, including system name from cvsAirHandlerNameOff as prefix
  ! First component will always be AirLoopHVAC:OutdoorAirSystem with outlet node name <ussAirHandlerNameOff> Mixed Air Outlet
  lastOutlet = ' Mixed Air Outlet'

  !set node name for return inlet to OA mixer
  IF (isReturnFanYes) THEN
    returnInletToOAMIxer = ' Return Fan Outlet'
  ELSE
    returnInletToOAMIxer = ' Air Loop Inlet'
  END IF

  unitInlet         = TRIM(lastOutlet)
  ! blow thru fan
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    supFanInlet         = TRIM(lastOutlet)
    supFanOutlet        = ' Supply Fan Outlet'
    lastOutlet       = TRIM(supFanOutlet)
  END IF

  ! cooling coil
  IF (coolCoilKind .NE. ccNone) THEN
    coolCoilUnitInlet = TRIM(lastOutlet)
    IF ((coolCoilKind .EQ. ccWaterCoilSystemHX) .OR. (coolCoilKind .EQ. ccDXcoilSystemHX)) THEN
      ! cooling coil with heat exchanger assist
      coolCoilUnitOutlet = ' Cooling Coil HX Unit Outlet'
    ELSE
      ! simple cooling coil (without heat exchanger assist)
      coolCoilUnitOutlet = ' Cooling Coil Outlet'
    END IF
    lastOutlet       = TRIM(coolCoilUnitOutlet)
  END IF

  ! heating coil
  IF (heatCoilKind .NE. hcNone) THEN
    heatCoilInlet    = TRIM(lastOutlet)
    heatCoilOutlet   = ' Heating Coil Outlet'
    lastOutlet       = TRIM(heatCoilOutlet)
  END IF

  ! draw thru fan
  IF (supFanPlacement .EQ. sfpDrawThru) THEN
    supFanInlet         = TRIM(lastOutlet)
    supFanOutlet        = ' Supply Fan Outlet'
    lastOutlet       = TRIM(supFanOutlet)
  END IF

  ! supplemental heating or reheat coil
  IF (supHeatCoilKind .NE. hcNone) THEN
    supheatCoilInlet = TRIM(lastOutlet)
    supheatCoilOutlet = ' Supp Heat or Reheat Coil Outlet'
    lastOutlet       = TRIM(supheatCoilOutlet)
  END IF

  unitOutlet         = TRIM(lastOutlet)

  ! humidifier
  IF (.NOT. isHumidifierNone) THEN
    humidifierInlet  = TRIM(lastOutlet)
    humidifierOutlet = ' Humidifier Outlet'
    lastOutlet       = TRIM(humidifierOutlet)
  END IF

  ! set AirLoopHVAC Supply Side Outlet Node Name
  airLoopOutlet      = TRIM(lastOutlet)

!Object ==> Sizing:System
  CALL CreateNewObj('Sizing:System')
  CALL AddToObjFld('AirLoop Name', base + ussAirHandlerNameOff,' ')
  CALL AddToObjStr('Type of Load to Size On','Sensible')
  CALL AddToObjFld('Design Outdoor Air Flow Rate {m3/s}', base + ussMinOutsideFlowOff,' ')
  CALL AddToObjStr('Central Heating Maximum System Air Flow Ratio','1')
  CALL AddToObjStr('Preheat Design Set Temperature {C}','7')
  CALL AddToObjStr('Preheat Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjStr('Precool Design Temperature {C}','11')
  CALL AddToObjStr('Precool Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjFld('Central Cooling Design Supply Air Temperature {C}', base + ussCoolDesignSupplyTempOff,'')
  CALL AddToObjFld('Central Heating Design Supply Air Temperature {C}', base + ussHeatDesignSupplyTempOff,'')
  CALL AddToObjFld('Sizing Option', base + ussSizingOptionOff,'')
  CALL AddToObjStr('100% Outdoor Air in Cooling','No')
  CALL AddToObjStr('100% Outdoor Air in Heating','No')
  CALL AddToObjStr('Central Cooling Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjStr('Central Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isSupplyCoolRateAutosize) THEN
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','DesignDay')
    CALL AddToObjStr('Cooling Supply Air Flow Rate {m3/s}','0')
  ELSE
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','Flow/System')
    CALL AddToObjFld('Cooling Supply Air Flow Rate {m3/s}', base + ussSupplyCoolFlowRateOff,' ')
  END IF
  ! New fields for v8.2 next three rows
  CALL AddToObjStr('Cooling Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
  CALL AddToObjStr('Cooling Fraction of Autosized Cooling Supply Air Flow Rate','')
  CALL AddToObjStr('Cooling Supply Air Flow Rate Per Unit Cooling Capacity {m3/s-W}','')
  IF (isSupplyHeatRateAutosize) THEN
    CALL AddToObjStr('Heating Supply Air Flow Rate Method','DesignDay')
    CALL AddToObjStr('Heating Supply Air Flow Rate {m3/s}','0')
  ELSE
    CALL AddToObjStr('Heating Supply Air Flow Rate Method','Flow/System')
    CALL AddToObjFld('Heating Supply Air Flow Rate {m3/s}', base + ussSupplyHeatFlowRateOff,' ')
  END IF
  ! New fields for v8.2 next four rows
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
  CALL AddToObjStr('Heating Fraction of Autosized Heating Supply Air Flow Rate','')
  CALL AddToObjStr('Heating Fraction of Autosized Cooling Supply Air Flow Rate','')
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Unit Heating Capacity {m3/s-W}','')
  CALL AddToObjStr('System Outdoor Air Method','ZoneSum')
  CALL AddToObjStr('Zone Maximum Outdoor Air Fraction','1.0')
  CALL AddToObjStr('Cooling Design Capacity Method','CoolingDesignCapacity')
  CALL AddToObjStr('Cooling Design Capacity {W}','autosize')
  CALL AddToObjStr('Cooling Design Capacity Per Floor Area {W/m2}','')
  CALL AddToObjStr('Fraction of Autosized Cooling Design Capacity','')
  CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
  CALL AddToObjStr('Heating Design Capacity {W}','autosize')
  CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
  CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
  CALL AddToObjStr('Central Cooling Capacity Control Method','OnOff',.true.)
  !Object ==> AirLoopHVAC
  CALL CreateNewObj('AirLoopHVAC')
  CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' ')
  CALL AddToObjStr('Controller List Name','')
  CALL AddToObjFld('Availability Manager List Name', base + ussAirHandlerNameOff,' Availability Managers')
  CALL AddToObjStr('Design Supply Air Flow Rate {m3/s}', 'autosize')
  CALL AddToObjFld('Branch List Name', base + ussAirHandlerNameOff,' Branches')
  CALL AddToObjStr('Connector List Name','')
  CALL AddToObjFld('Supply Side Inlet Node Name', base + ussAirHandlerNameOff,' Air Loop Inlet')
  CALL AddToObjFld('Demand Side Outlet Node Name', base + ussAirHandlerNameOff,' Return Air Outlet')
  CALL AddToObjFld('Demand Side Inlet Node Names', base + ussAirHandlerNameOff,' Supply Path Inlet')
  CALL AddToObjFld('Supply Side Outlet Node Names', base + ussAirHandlerNameOff,TRIM(airLoopOutlet),.TRUE.)
  !Object ==> BranchList
  CALL CreateNewObj('BranchList')
  CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Branches')
  CALL AddToObjFld('Branch Name', base + ussAirHandlerNameOff,' Main Branch',.TRUE.)
  !Object ==> Branch
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Main Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  IF (isReturnFanYes) THEN
    IF (supFanKind .EQ. sfkVAV) THEN
      CALL AddToObjStr('Component Object Type','Fan:VariableVolume')
    ELSE
      CALL AddToObjStr('Component Object Type','Fan:ConstantVolume')
    END IF
    CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' Return Fan')
    CALL AddToObjFld('Component Inlet Node Name', base + ussAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + ussAirHandlerNameOff,' Return Fan Outlet')
  END IF
  CALL AddToObjStr('Component Object Type','AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' OA System')
  CALL AddToObjFld('Component Inlet Node Name', base + ussAirHandlerNameOff, TRIM(returnInletToOAMIxer))
  CALL AddToObjFld('Component Outlet Node Name', base + ussAirHandlerNameOff,' Mixed Air Outlet')
  CALL AddToObjStr('Component Object Type','AirLoopHVAC:UnitarySystem')
  CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' Unitary System')
  CALL AddToObjFld('Component Inlet Node Name', base + ussAirHandlerNameOff, TRIM(unitInlet))
  CALL AddToObjFld('Component Outlet Node Name', base + ussAirHandlerNameOff, TRIM(unitOutlet),isHumidifierNone)
  IF (humidifierKind .EQ. humidifyElecSteam) THEN
    CALL AddToObjStr('Component Object Type','Humidifier:Steam:Electric')
    CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' Humidifier')
    CALL AddToObjFld('Component Inlet Node Name', base + ussAirHandlerNameOff,TRIM(humidifierInlet))
    CALL AddToObjFld('Component Outlet Node Name', base + ussAirHandlerNameOff,TRIM(humidifierOutlet),.TRUE.)
  END IF
  IF (isSupplyPlenumBlank) THEN
    !Object ==> AirLoopHVAC:SupplyPath
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + ussAirHandlerNameOff,' Supply Path Inlet')
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneSplitter')
    CALL MakeSpecialIntList(zlUnitAllSupPlen, base + ussAirHandlerNameOff)
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' Zone Splitter',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' Zone Splitter')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Supply Plenum', .TRUE.)
      END IF
    END IF
    !Object ==> AirLoopHVAC:ZoneSplitter
    CALL CreateNewObj('AirLoopHVAC:ZoneSplitter')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Zone Splitter')
    CALL AddToObjFld('Inlet Node Name', base + ussAirHandlerNameOff,' Supply Path Inlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlUnitAll,base + ussAirHandlerNameOff)
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(iInt),' Zone Equip Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Inlet',.TRUE.)
    END IF
  ELSE
    !Object ==> AirLoopHVAC:SupplyPath
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + ussAirHandlerNameOff,' Supply Path Inlet')
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
    CALL MakeSpecialIntList(zlUnitAllSupPlen, base + ussAirHandlerNameOff)
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' Supply Plenum',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' Supply Plenum')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Supply Plenum', .TRUE.)
      END IF
    END IF
    !Object ==> AirLoopHVAC:SupplyPlenum
    CALL CreateNewObj('AirLoopHVAC:SupplyPlenum')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Supply Plenum')
    CALL AddToObjFld('Zone Name', base + ussSupplyPlenumNameOff,' ')
    CALL AddToObjFld('Zone Node Name', base + ussSupplyPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Inlet Node Name', base + ussAirHandlerNameOff,' Supply Path Inlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlUnitAll,base + ussAirHandlerNameOff)
    ! since the previous line is called again for the return plenum blank case, logic was added
    ! to not allow the same error message to be displayed more than once.
    IF (numIntList .EQ. 0) THEN
      doSuppressMsgZlUnitAll = .TRUE.
    END IF
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(iInt),' Zone Equip Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Inlet',.TRUE.)
    END IF
  END IF
  IF (isReturnPlenumBlank) THEN
    !Object ==> AirLoopHVAC:ReturnPath
    CALL CreateNewObj('AirLoopHVAC:ReturnPath')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Return Path')
    CALL AddToObjFld('Return Air Path Outlet Node Name', base + ussAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlUnitAllRetPlen, base + ussAirHandlerNameOff)
    IF (numIntList .NE. 0) THEN
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Return Plenum')
      END DO
    END IF
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneMixer')
    CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' Zone Mixer',.TRUE.)
    !Object ==> AirLoopHVAC:ZoneMixer
    CALL CreateNewObj('AirLoopHVAC:ZoneMixer')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Zone Mixer')
    CALL AddToObjFld('Outlet Node Name', base + ussAirHandlerNameOff,' Return Air Outlet')
    !the IntList holds the list of pointers to the zone names
    ! logic added to avoid duplicate error from being shown since also in supply plenum blank case
    CALL MakeSpecialIntList(zlUnitAll,base + ussAirHandlerNameOff,doSuppressMsgZlUnitAll)
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Inlet Node Name', IntList(iInt),' Return Outlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Inlet Node Name', IntList(numIntList),' Return Outlet',.TRUE.)
    END IF
  ELSE
    !Object ==> AirLoopHVAC:ReturnPath
    CALL CreateNewObj('AirLoopHVAC:ReturnPath')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Return Path')
    CALL AddToObjFld('Return Air Path Outlet Node Name', base + ussAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlUnitAllRetPlen, base + ussAirHandlerNameOff)
    IF (numIntList .NE. 0) THEN
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Return Plenum')
      END DO
    END IF
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' Return Plenum',.TRUE.)
    !Object ==> AirLoopHVAC:ReturnPlenum
    CALL CreateNewObj('AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Return Plenum')
    CALL AddToObjFld('Zone Name', base + ussReturnPlenumNameOff,' ')
    CALL AddToObjFld('Zone Node Name', base + ussReturnPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Outlet Node Name', base + ussAirHandlerNameOff,' Return Air Outlet')
    CALL AddToObjStr('Induced Air Outlet Node or NodeList Name','')
    CALL MakeSpecialIntList(zlUnitAll,base + ussAirHandlerNameOff)
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Inlet Node Name', IntList(iInt),' Return Outlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Inlet Node Name', IntList(numIntList),' Return Outlet',.TRUE.)
    END IF
  END IF
  !Object ==> AvailabilityManagerAssignmentList
  CALL CreateNewObj('AvailabilityManagerAssignmentList')
  CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Availability Managers')
  CALL AddToObjStr('Availability Manager Object Type','AvailabilityManager:Scheduled')
  CALL AddToObjFld('Availability Manager Name', base + ussAirHandlerNameOff,' Availability',.TRUE.)
  !Object ==> AvailabilityManager
  CALL CreateNewObj('AvailabilityManager:Scheduled')
  CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Availability')
  IF (.NOT. isSysAvailSchedBlank) THEN
    CALL AddToObjFld('Schedule Name', base + ussSysAvailSchedNameOff,'',.TRUE.)
  ELSE
    CALL AddToObjStr('Schedule Name','HVACTemplate-Always 1',.TRUE.)
    CALL AddAlwaysSchedule('1')
  END IF
  ! Humdistat(s) if needed
  ! first check if the schedule names are used or if values should be used.
  IF (.NOT. isDehumidifyNone) THEN
    IF (FldVal(base + ussDehumSetPtSchedNameOff) .EQ. '') THEN
      dehumscheduleName = 'HVACTemplate-Always ' // TRIM(FldVal(base + ussDehumSetPtOff))
      CALL AddAlwaysSchedule(FldVal(base + ussDehumSetPtOff))
    ELSE
      dehumscheduleName = TRIM(FldVal(base + ussDehumSetPtSchedNameOff))
    ENDIF
  END IF
  IF (.NOT. isHumidifierNone) THEN
    ! first check if the schedule name is used or if value should be used.
    IF (FldVal(base + ussHumidSetPtSchedNameOff) .EQ. '') THEN
      humscheduleName = 'HVACTemplate-Always ' // TRIM(FldVal(base + ussHumidSetPtOff))
      CALL AddAlwaysSchedule(FldVal(base + ussHumidSetPtOff))
    ELSE
      humscheduleName = TRIM(FldVal(base + ussHumidSetPtSchedNameOff))
    ENDIF
  END IF

  !    Single humidistat if humidification and dehumidification are both active and control zones are the same
  IF ((.NOT. isHumidifierNone) .AND. (.NOT. isDehumidifyNone) &
       .AND. SameString(FldVal(base +  ussControlZoneOff),FldVal(base +  ussHumidCtrlZoneOff))) THEN
    CALL CreateNewObj('ZoneControl:Humidistat')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Humidistat')
    CALL AddToObjFld('Zone Name', base + ussControlZoneOff,'')
    CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name', TRIM(humscheduleName))
    CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name', TRIM(dehumscheduleName),.TRUE.)
  ELSE
    IF (.NOT. isDehumidifyNone) THEN
  !   Dehumidification humidistat
      CALL CreateNewObj('ZoneControl:Humidistat')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Dehumidification Humidistat')
      CALL AddToObjFld('Zone Name', base + ussControlZoneOff,'')
      CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name','HVACTemplate-Always 1')
      CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name', TRIM(dehumscheduleName),.TRUE.)

      CALL AddAlwaysSchedule('1')
    ENDIF
    IF (.NOT. isHumidifierNone) THEN
  !    Humidification humidistat
      CALL CreateNewObj('ZoneControl:Humidistat')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Humidification Humidistat')
      CALL AddToObjFld('Zone Name', base + ussHumidCtrlZoneOff,'')
      CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name', TRIM(humscheduleName))
      CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name','HVACTemplate-Always 100',.TRUE.)

      CALL AddAlwaysSchedule('100')
    ENDIF
  ENDIF

  SchType=' '
  !Object ==> AirLoopHVAC:UnitarySystem
  CALL CreateNewObj('AirLoopHVAC:UnitarySystem')
  CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Unitary System')
  CALL AddToObjFld('Control Type', base + ussSystemControlTypeOff,'')
  CALL AddToObjFld('Controlling Zone or Thermostat Location', base + ussControlZoneOff,'')
  CALL AddToObjFld('Dehumidification Control Type', base + ussDehumCtrlTypeOff,'')
  CALL AddToObjStr('Availability Schedule Name', ' ')
  CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(unitInlet))
  CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(unitOutlet))
  CALL AddToObjStr('Supply Fan Object Type',TRIM(supFanObjectType))
  CALL AddToObjFld('Supply Fan Name', base + ussAirHandlerNameOff,' Supply Fan')
  CALL AddToObjFld('Fan Placement', base + ussSupplyFanPlaceOff,'')
  IF (isSupplyFanCycling) THEN
    CALL AddToObjStr('Supply Air Fan Operating Mode Schedule Name','HVACTemplate-Always 0')
    SchType='0'
  ELSE
    CALL AddToObjFld('Supply Air Fan Operating Mode Schedule Name', base + ussFanModeSchedOff,'')
  END IF
  CALL AddToObjStr('Heating Coil Object Type',TRIM(heatCoilObjectType))
  CALL AddToObjStr('Heating Coil Name',  TRIM(heatCoilObjectName))
  CALL AddToObjStr('DX Heating Coil Sizing Ratio','1.0')
  CALL AddToObjStr('Cooling Coil Object Type',TRIM(coolCoilObjectType))
  CALL AddToObjStr('Cooling Coil Name', TRIM(coolCoilObjectName))
  CALL AddToObjStr('Use DOAS DX Cooling Coil','')
  CALL AddToObjStr('DOAS DX Cooling Coil Leaving Minimum Air Temperature {C}','')
  CALL AddToObjStr('Latent Load Control','')
  CALL AddToObjStr('Supplemental Heating Coil Object Type',TRIM(supheatCoilObjectType))
  CALL AddToObjStr('Supplemental Heating Coil Name',TRIM(supheatCoilObjectName))
  CALL AddToObjStr('Cooling Supply Air Flow Rate Method','SupplyAirFlowRate')
  CALL AddToObjFld('Cooling Supply Air Flow Rate {m3/s}', base + ussSupplyCoolFlowRateOff,'')
  CALL AddToObjStr('Cooling Supply Air Flow Rate Per Floor Area {m3/s-m2','')
  CALL AddToObjStr('Cooling Fraction of Autosized Cooling Supply Air Flow Rate','')
  CALL AddToObjStr('Cooling Supply Air Flow Rate Per Unit of Capacity {m3/s-W','')
  CALL AddToObjStr('Heating Supply Air Flow Rate Method','SupplyAirFlowRate')
  CALL AddToObjFld('Heating Supply Air Flow Rate {m3/s}', base + ussSupplyHeatFlowRateOff,'')
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Floor Area {m3/s-m2','')
  CALL AddToObjStr('Heating Fraction of Autosized Heating Supply Air Flow Rate','')
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Unit of Capacity {m3/s-W','')
  CALL AddToObjStr('No Load Supply Air Flow Rate Method','SupplyAirFlowRate')
  CALL AddToObjFld('No Load Supply Air Flow Rate {m3/s}', base + ussSupplyNoLoadFlowRateOff,'')
  CALL AddToObjStr('No Load Supply Air Flow Rate Per Floor Area {m3/s-m2','')
  CALL AddToObjStr('No Load Fraction of Autosized Cooling Supply Air Flow Rate','')
  CALL AddToObjStr('No Load Fraction of Autosized Heating Supply Air Flow Rate','')
  CALL AddToObjStr('No Load Supply Air Flow Rate Per Unit of Capacity During Cooling Operation {m3/s-W','')
  CALL AddToObjStr('No Load Supply Air Flow Rate Per Unit of Capacity During Heating Operation {m3/s-W','')
  CALL AddToObjStr('Maximum Supply Air Temperature {C}', 'Autosize')
  CALL AddToObjFld('Maximum Outdoor Dry-Bulb Temperature for Supplemental Heater Operation {C}', base + ussSuppReHeatMaxODBOff,'')
  CALL AddToObjStr('Outdoor Dry-Bulb Temperature Sensor Node Name', '')
  CALL AddToObjStr('Maximum Cycling Rate', '')
  CALL AddToObjStr('Heat Pump Time Constant', '')
  CALL AddToObjStr('Fraction of On-Cycle Power Use', '')
  CALL AddToObjStr('Heat Pump Fan Delay Time', '')
  CALL AddToObjStr('Ancilliary On-Cycle Electric Power', '')
  CALL AddToObjStr('Ancilliary Off-Cycle Electric Power', '')
  CALL AddToObjStr('Design Heat Recovery Water Flow Rate', '')
  CALL AddToObjStr('Maximum Temperature for Heat Recovery', '')
  CALL AddToObjStr('Heat Recovery Water Inlet Node Name', '')
  CALL AddToObjStr('Heat Recovery Water Outlet Node Name', '')
  IF (.NOT. isMultiSpeed) THEN
    CALL AddToObjStr('Design Specification Multispeed Object Type', '')
    CALL AddToObjStr('Design Specification Multispeed Object Name', '',.TRUE.)
  ELSE
    CALL AddToObjStr('Design Specification Multispeed Object Type', 'UnitarySystemPerformance:Multispeed')
    CALL AddToObjFld('Design Specification Multispeed Object Name', base + ussAirHandlerNameOff, &
                     ' Unitary System MultiSpeed Performance',.TRUE.)

    !Object ==> UnitarySystemPerformance:Multispeed
    CALL CreateNewObj('UnitarySystemPerformance:Multispeed')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Unitary System MultiSpeed Performance')
    CALL AddToObjFld('Number of Speeds for Heating', base + ussHeatCoilNumSpeedOff,'')
    CALL AddToObjFld('Number of Speeds for Cooling', base + ussCoolCoilNumSpeedOff,'')
    CALL AddToObjStr('Single Mode Operation', 'No')
    heatingSpds = StringToReal(FldVal(base + ussHeatCoilNumSpeedOff))
    coolingSpds = StringToReal(FldVal(base + ussCoolCoilNumSpeedOff))
    maxSpd = max(heatingSpds, coolingSpds)
    CALL AddToObjStr('No Load Supply Air Flow Rate Ratio', RealToStr(1.0/maxSpd))
    CALL AddToObjStr('Heating Speed 1 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Cooling Speed 1 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Heating Speed 2 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Cooling Speed 2 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Heating Speed 3 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Cooling Speed 3 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Heating Speed 4 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Cooling Speed 4 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Heating Speed 5 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Cooling Speed 5 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Heating Speed 6 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Cooling Speed 6 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Heating Speed 7 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Cooling Speed 7 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Heating Speed 8 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Cooling Speed 8 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Heating Speed 9 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Cooling Speed 9 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Heating Speed 10 Supply Air Flow Ratio', 'autosize')
    CALL AddToObjStr('Cooling Speed 10 Supply Air Flow Ratio', 'autosize',.TRUE.)
  END IF

  ! add new schedule if needed (from Supply Fan Operating Mode Schedule, above)
  IF (SchType /= ' ') THEN
    CALL AddAlwaysSchedule(SchType)
  ENDIF
  ! Cooling Coil
  SELECT CASE (coolCoilKind)
    CASE (ccChWaterDetailed)
      !***Coil:Cooling:Water:DetailedGeometry
      CALL CreateNewObj('Coil:Cooling:Water:DetailedGeometry')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + ussCoolAvailSchedNameOff,' ')
      CALL AddToObjStr('Max Water Flow Rate of Coil {m3/s}','autosize')
      CALL AddToObjStr('Tube Outside Surf Area {m2}','autosize')
      CALL AddToObjStr('Total Tube Inside Area {m2}','autosize')
      CALL AddToObjStr('Fin Surface Area {m2}','autosize')
      CALL AddToObjStr('Minimum Air Flow Area {m2}','autosize')
      CALL AddToObjStr('Coil Depth {m}','autosize')
      CALL AddToObjStr('Fin Diameter {m}','autosize')
      CALL AddToObjStr('Fin Thickness {m}','0.0015')
      CALL AddToObjStr('Tube Inside Diameter {m}','0.01445')
      CALL AddToObjStr('Tube Outside Diameter {m}','0.0159')
      CALL AddToObjStr('Tube Thermal Conductivity {W/m-K}','386.0')
      CALL AddToObjStr('Fin Thermal Conductivity {W/m-K}','204.0')
      CALL AddToObjStr('Fin Spacing {m}','0.0018')
      CALL AddToObjStr('Tube Depth Spacing {m}','0.026')
      CALL AddToObjStr('Number of Tube Rows','4')
      CALL AddToObjStr('Number of Tubes per Row','autosize')
      CALL AddToObjFld('Coil Water Inlet Node', base + ussAirHandlerNameOff,' Cooling Coil ChW Inlet')
      CALL AddToObjFld('Coil Water Outlet Node', base + ussAirHandlerNameOff,' Cooling Coil ChW Outlet')
      CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(coolCoilUnitInlet))
      CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(coolCoilUnitOutlet),.TRUE.)
    CASE (ccChWater)
      !***Coil:Cooling:Water
      CALL CreateNewObj('Coil:Cooling:Water')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + ussCoolAvailSchedNameOff,' ')
      CALL AddToObjStr('Design Water Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Design Air Flow Rate {m3/s}', 'autosize')
      CALL AddToObjStr('Design Inlet Water Temperature {C}','autosize')
      CALL AddToObjStr('Design Inlet Air Temperature {C}','autosize')
      CALL AddToObjStr('Design Outlet Air Temperature {C}','autosize')
      CALL AddToObjStr('Design Inlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
      CALL AddToObjStr('Design Outlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
      CALL AddToObjFld('Water Inlet Node Name', base + ussAirHandlerNameOff,' Cooling Coil ChW Inlet')
      CALL AddToObjFld('Water Outlet Node Name', base + ussAirHandlerNameOff,' Cooling Coil ChW Outlet')
      CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(coolCoilUnitInlet))
      CALL AddToObjFld('Air Outlet Node Name',base + ussAirHandlerNameOff, TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('Type of Analysis','DetailedAnalysis')
      CALL AddToObjStr('Heat Exchanger Configuration','CrossFlow',.TRUE.)
    CASE (ccSingleSpeedDX)
      ! Cooling coil
      !Object ==> Coil:Cooling:DX:SingleSpeed
      CALL CreateNewObj('Coil:Cooling:DX:SingleSpeed')
        CALL AddToObjStr('Name', TRIM(coolCoilObjectName))
      CALL AddToObjFld('Availability Schedule Name', base + ussCoolAvailSchedNameOff,' ')
      CALL AddToObjFld('Gross Rated Total Cooling Capacity {W}', base + ussCoolCoilCapOff,' ')
      CALL AddToObjFld('Gross Rated Sensible Heat Ratio', base + ussCoolCoilSHROff,' ')
      CALL AddToObjFld('Rated COP', base + ussCoolCoilCOPOff,' ')
      CALL AddToObjStr('Rated Air Flow Rate {m3/s}', 'autosize')
      CALL AddToObjStr('Rated Evaporator Fan Power per Volume Flow Rate {W/(m3/s)}','')
      CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(coolCoilUnitInlet))
      CALL AddToObjFld('Air Outlet Node Name',base + ussAirHandlerNameOff, TRIM(coolCoilUnitOutlet))
      CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
         base + ussAirHandlerNameOff,' Cool Coil Cap-FT')
      CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
         base + ussAirHandlerNameOff,' Cool Coil Cap-FF')
      CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
         base + ussAirHandlerNameOff,' Cool Coil EIR-FT')
      CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
         base + ussAirHandlerNameOff,' Cool Coil EIR-FF')
      CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
         base + ussAirHandlerNameOff,' Cool Coil PLF')
      CALL AddToObjStr('Minimum Outdoor Dry-Bulb Temperature for Compressor Operation {C}','')
      CALL AddToObjStr('Nominal Time for Condensate Removal to Begin','0')
      CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity','0')
      CALL AddToObjStr('Maximum Cycling Rate','0')
      CALL AddToObjStr('Latent Capacity Time Constant','0')
      CALL AddToObjFld('Condenser Air Inlet Node Name',base + ussAirHandlerNameOff,' Cooling Coil Condenser Inlet')
      CALL AddToObjStr('Condenser Type','AirCooled')
      CALL AddToObjStr('Evaporative Condenser Effectiveness','0')
      CALL AddToObjStr('Evaporative Condenser Air Flow Rate','')
      CALL AddToObjStr('Evaporative Condenser Pump Rated Power Consumption','0')
      CALL AddToObjStr('Crankcase Heater Capacity','0')
      CALL AddToObjStr('Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater Operation','10',.TRUE.)
      !Object ==> Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('DOE-2.1E, COOL-CAP-FT for PTAC w/ SI temps')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cool Coil Cap-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.942587793')
      CALL AddToObjStr('Coefficient2 x','0.009543347')
      CALL AddToObjStr('Coefficient3 x**2','0.00068377')
      CALL AddToObjStr('Coefficient4 y','-0.011042676')
      CALL AddToObjStr('Coefficient5 y**2','0.000005249')
      CALL AddToObjStr('Coefficient6 x*y','-0.00000972')
      CALL AddToObjStr('Minimum Value of x','12.77778')
      CALL AddToObjStr('Maximum Value of x','23.88889')
      CALL AddToObjStr('Minimum Value of y','18.0')
      CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
      !Object ==> Curve:Quadratic
      CALL CreateNewObj('Curve:Quadratic')
      CALL WriteComment('DOE-2.1E, RATED-CCAP-FFLOW for PTAC')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cool Coil Cap-FF')
      CALL AddToObjStr('Coefficient1 Constant','0.8')
      CALL AddToObjStr('Coefficient2 x','0.2')
      CALL AddToObjStr('Coefficient3 x**2','0')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !Object ==> Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('DOE-2.1E, COOL-EIR-FT for PTAC w/ SI temps')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cool Coil EIR-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.342414409')
      CALL AddToObjStr('Coefficient2 x','0.034885008')
      CALL AddToObjStr('Coefficient3 x**2','-0.0006237')
      CALL AddToObjStr('Coefficient4 y','0.004977216')
      CALL AddToObjStr('Coefficient5 y**2','0.000437951')
      CALL AddToObjStr('Coefficient6 x*y','-0.000728028')
      CALL AddToObjStr('Minimum Value of x','12.77778')
      CALL AddToObjStr('Maximum Value of x','23.88889')
      CALL AddToObjStr('Minimum Value of y','18.0')
      CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
      !Object ==> Curve:Quadratic
      CALL CreateNewObj('Curve:Quadratic')
      CALL WriteComment('DOE-2.1E, RATED-CEIR-FFLOW for PTAC')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cool Coil EIR-FF')
      CALL AddToObjStr('Coefficient1 Constant','1.1552')
      CALL AddToObjStr('Coefficient2 x','-0.1808')
      CALL AddToObjStr('Coefficient3 x**2','0.0256')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !Object ==> Curve:Quadratic
      CALL CreateNewObj('Curve:Quadratic')
      CALL WriteComment('PLF = l.- Cd(1.-PLR) where Cd = 0.15')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cool Coil PLF')
      CALL AddToObjStr('Coefficient1 Constant','0.85')
      CALL AddToObjStr('Coefficient2 x','0.15')
      CALL AddToObjStr('Coefficient3 x**2','0')
      CALL AddToObjStr('Minimum Value of x','0')
      CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
      !Object ==> OutdoorAir:Node
      CALL CreateNewObj('OutdoorAir:Node')
      CALL AddToObjFld('Name',base + ussAirHandlerNameOff,' Cooling Coil Condenser Inlet')
      CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
    CASE (ccTwoSpeedDX)
      !***Coil:Cooling:DX:TwoSpeed
      CALL CreateNewObj('Coil:Cooling:DX:TwoSpeed')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + ussCoolAvailSchedNameOff,' ')
      CALL AddToObjFld('Rated High Speed Total Cooling Capacity', base + ussCoolCoilCapOff, '')
      CALL AddToObjFld('Rated High Speed Sensible Heat Ratio', base + ussCoolCoilSHROff, '')
      CALL AddToObjFld('Rated High Speed COP', base + ussCoolCoilCOPOff, '')
      CALL AddToObjStr('Rated High Speed Air Flow Rate', 'autosize')
      ! DX Ratings - Jul 2012
      CALL AddToObjStr('Unit Internal Static Air Pressure','')
      CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(coolCoilUnitInlet))
      CALL AddToObjFld('Air Outlet Node Name',base + ussAirHandlerNameOff, TRIM(coolCoilUnitOutlet))
      CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil Cap-FT')
      CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil Cap-FF')
      CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil EIR-FT')
      CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil EIR-FF')
      CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil PLF')
      IF (isCoolCapAutosize) THEN
        CALL AddToObjStr('Rated Low Speed Total Cooling Capacity','Autosize')
      ELSE
        ! If high speed capacity is specified, assume that low speed is 1/3 of high speed capacity, same assumption that autosizing uses
        CALL AddToObjNum('Rated Low Speed Total Cooling Capacity',StringToReal(FldVal(base + ussCoolCoilCapOff)) * 0.33)
      END IF
      ! assume low speed sensible heat ratio same as high speed
      CALL AddToObjFld('Rated Low Speed Sensible Heat Ratio', base + ussCoolCoilSHROff, '')
      CALL AddToObjNum('Rated Low Speed COP',StringToReal(FldVal(base + ussCoolCoilCOPOff)) * 1.5)
      ! assume low speed air flow rate is automatically sized
      CALL AddToObjStr('Rated Low Speed Air Flow Rate','Autosize')
      ! assume the same curves as full speed
      CALL AddToObjFld('Low Speed Total Cooling Capacity Function of Temperature Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil Low Cap-FT')
      CALL AddToObjFld('Low Speed Energy Input Ratio Function of Temperature Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil Low EIR-FT',.TRUE.)
      !***Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('DOE-2.1E, SDL-C4')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil Cap-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.476428')
      CALL AddToObjStr('Coefficient2 x','0.401147E-01')
      CALL AddToObjStr('Coefficient3 x**2','0.226411E-03')
      CALL AddToObjStr('Coefficient4 y','-0.827136E-03')
      CALL AddToObjStr('Coefficient5 y**2','-0.732240E-05')
      CALL AddToObjStr('Coefficient6 x*y','-0.446278E-03')
      CALL AddToObjStr('Minimum Value of x','0.0')
      CALL AddToObjStr('Maximum Value of x','50.0')
      CALL AddToObjStr('Minimum Value of y','0.0')
      CALL AddToObjStr('Maximum Value of y','50.0')
      CALL AddToObjStr('Minimum Curve Output','0.0')
      CALL AddToObjStr('Maximum Curve Output','5.0')
      CALL AddToObjStr('Input Unit Type for X','Temperature')
      CALL AddToObjStr('Input Unit Type for Y','Temperature')
      CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
      !***Curve:Cubic
      CALL CreateNewObj('Curve:Cubic')
      CALL WriteComment('DOE-2.1E, SDL-C78')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil Cap-FF')
      CALL AddToObjStr('Coefficient1 Constant','.47278589')
      CALL AddToObjStr('Coefficient2 x','1.2433415')
      CALL AddToObjStr('Coefficient3 x**2','-1.0387055')
      CALL AddToObjStr('Coefficient4 x**3','.32257813')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !***Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('DOE-2.1E, SDL-C82')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil EIR-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.632475')
      CALL AddToObjStr('Coefficient2 x',' -0.121321E-01')
      CALL AddToObjStr('Coefficient3 x**2','0.507773E-03')
      CALL AddToObjStr('Coefficient4 y','0.155377E-01')
      CALL AddToObjStr('Coefficient5 y**2','0.272840E-03')
      CALL AddToObjStr('Coefficient6 x*y','-0.679201E-03')
      CALL AddToObjStr('Minimum Value of x','0.0')
      CALL AddToObjStr('Maximum Value of x','50.0')
      CALL AddToObjStr('Minimum Value of y','0.0')
      CALL AddToObjStr('Maximum Value of y','50.0')
      CALL AddToObjStr('Minimum Curve Output','0.0')
      CALL AddToObjStr('Maximum Curve Output','5.0')
      CALL AddToObjStr('Input Unit Type for X','Temperature')
      CALL AddToObjStr('Input Unit Type for Y','Temperature')
      CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
      !***Curve:Cubic
      CALL CreateNewObj('Curve:Cubic')
      CALL WriteComment('DOE-2.1E, SDL-C82')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil EIR-FF')
      CALL AddToObjStr('Coefficient1 Constant','1.0079484')
      CALL AddToObjStr('Coefficient2 x','.34544129')
      CALL AddToObjStr('Coefficient3 x**2','-.6922891')
      CALL AddToObjStr('Coefficient4 x**3','.33889943')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !***Curve:Quadratic
      CALL CreateNewObj('Curve:Quadratic')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil PLF')
      CALL AddToObjStr('Coefficient1 Constant','0.85')
      CALL AddToObjStr('Coefficient2 x','0.15')
      CALL AddToObjStr('Coefficient3 x**2','0.0')
      CALL AddToObjStr('Minimum Value of x','0.0')
      CALL AddToObjStr('Maximum Value of x','1.0',.TRUE.)
      !***Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('DOE-2.1E, SDL-C4')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil Low Cap-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.476428E+00')
      CALL AddToObjStr('Coefficient2 x','0.401147E-01')
      CALL AddToObjStr('Coefficient3 x**2','0.226411E-03')
      CALL AddToObjStr('Coefficient4 y','-0.827136E-03')
      CALL AddToObjStr('Coefficient5 y**2','-0.732240E-05')
      CALL AddToObjStr('Coefficient6 x*y','-0.446278E-03')
      CALL AddToObjStr('Minimum Value of x','0.0')
      CALL AddToObjStr('Maximum Value of x','50.0')
      CALL AddToObjStr('Minimum Value of y','0.0')
      CALL AddToObjStr('Maximum Value of y','50.0')
      CALL AddToObjStr('Minimum Curve Output','0.0')
      CALL AddToObjStr('Maximum Curve Output','5.0')
      CALL AddToObjStr('Input Unit Type for X','Temperature')
      CALL AddToObjStr('Input Unit Type for Y','Temperature')
      CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
      !***Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil Low EIR-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.774645E+00')
      CALL AddToObjStr('Coefficient2 x','-0.343731E-01')
      CALL AddToObjStr('Coefficient3 x**2','0.783173E-03')
      CALL AddToObjStr('Coefficient4 y','0.146596E-01')
      CALL AddToObjStr('Coefficient5 y**2','0.488851E-03')
      CALL AddToObjStr('Coefficient6 x*y','-0.752036E-03')
      CALL AddToObjStr('Minimum Value of x','0.0')
      CALL AddToObjStr('Maximum Value of x','50.0')
      CALL AddToObjStr('Minimum Value of y','0.0')
      CALL AddToObjStr('Maximum Value of y','50.0')
      CALL AddToObjStr('Minimum Curve Output','0.0')
      CALL AddToObjStr('Maximum Curve Output','5.0')
      CALL AddToObjStr('Input Unit Type for X','Temperature')
      CALL AddToObjStr('Input Unit Type for Y','Temperature')
      CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
    CASE (ccTwoStageDX, ccTwoStageHumidControlDX)
      !***Coil:Cooling:DX:TwoStageWithHumidityControlMode
      CALL CreateNewObj('Coil:Cooling:DX:TwoStageWithHumidityControlMode')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + ussCoolAvailSchedNameOff,' ')
      CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(coolCoilUnitInlet))
      CALL AddToObjFld('Air Outlet Node Name',base + ussAirHandlerNameOff, TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('Crankcase Heater Capacity','')
      CALL AddToObjStr('Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater','')
      CALL AddToObjStr('Number of Capacity Stages','2')
      IF (coolCoilKind .EQ. ccTwoStageHumidControlDX) THEN
        CALL AddToObjStr('Number of Enhanced Dehumidification Modes','1')
      ELSE
        CALL AddToObjStr('Number of Enhanced Dehumidification Modes','0')
      END IF
      CALL AddToObjStr('Normal Mode Stage 1 Coil Performance Object Type','CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Normal Mode Stage 1 Coil Performance Name',base + ussAirHandlerNameOff,' Standard Perf 1')
      CALL AddToObjStr('Normal Mode Stage 1+2 Coil Performance Object Type','CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Normal Mode Stage 1+2 Coil Performance Name',base + ussAirHandlerNameOff,' Standard Perf 1+2')
      IF (coolCoilKind .EQ. ccTwoStageHumidControlDX) THEN
        CALL AddToObjStr('Dehumidification Mode 1 Stage 1 Coil Performance Object Type','CoilPerformance:DX:Cooling')
        CALL AddToObjFld('Dehumidification Mode 1 Stage 1 Coil Performance Name',base + ussAirHandlerNameOff,' Dehumid Perf 1')
        CALL AddToObjStr('Dehumidification Mode 1 Stage 1+2 Coil Performance Object Type','CoilPerformance:DX:Cooling')
        CALL AddToObjFld('Dehumidification Mode 1 Stage 1+2 Coil Performance Name',base + ussAirHandlerNameOff,' Dehumid Perf 1+2')
      ELSE
        CALL AddToObjStr('Dehumidification Mode 1 Stage 1 Coil Performance Object Type',' ')
        CALL AddToObjStr('Dehumidification Mode 1 Stage 1 Coil Performance Name',' ')
        CALL AddToObjStr('Dehumidification Mode 1 Stage 1+2 Coil Performance Object Type',' ')
        CALL AddToObjStr('Dehumidification Mode 1 Stage 1+2 Coil Performance Name',' ')
      END IF
      CALL AddToObjStr('Supply Water Storage Tank Name','')
      CALL AddToObjStr('Condensate Collection Water Storage Tank Name','',.TRUE.)
      !***CoilPerformance:DX:Cooling
      ! Standard - Stage 1
      CALL CreateNewObj('CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Name',base + ussAirHandlerNameOff,' Standard Perf 1')
      IF (isCoolCapAutosize) THEN
        CALL AddToObjStr('Rated Total Cooling Capacity', 'autosize')
      ELSE
        ! half the capacity if provided
        CALL AddToObjNum('Rated Total Cooling Capacity', 0.5 * StringToReal(FldVal(base + ussCoolCoilCapOff)))
      END IF
      CALL AddToObjFld('Rated Sensible Heat Ratio', base + ussCoolCoilSHROff, '')
      CALL AddToObjFld('Rated COP', base + ussCoolCoilCOPOff, '')
      CALL AddToObjStr('Rated Air Flow Rate', 'autosize')
      CALL AddToObjStr('Fraction of Air Flow Bypassed Around Coil', '0.5')
      CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil Std Cap-FT')
      CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil Cap-FF')
      CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil Std EIR-FT')
      CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil EIR-FF')
      CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil PLF')
      CALL AddToObjStr('Nominal Time for Condensate Removal to Begin', '')
      CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity', '')
      CALL AddToObjStr('Maximum Cycling Rate', '')
      CALL AddToObjStr('Latent Capacity Time Constant', '',.TRUE.)
      !***CoilPerformance:DX:Cooling
      ! Standard - Stage 1 + 2
      CALL CreateNewObj('CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Name',base + ussAirHandlerNameOff,' Standard Perf 1+2')
      CALL AddToObjFld('Rated Total Cooling Capacity', base + ussCoolCoilCapOff,'')
      CALL AddToObjFld('Gross Rated Sensible Heat Ratio', base + ussCoolCoilSHROff, '')
      CALL AddToObjFld('Rated COP', base + ussCoolCoilCOPOff, '')
      CALL AddToObjStr('Rated Air Flow Rate', 'autosize')
      CALL AddToObjStr('Fraction of Air Flow Bypassed Around Coil', '0.0')
      CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil Std Cap-FT')
      CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil Cap-FF')
      CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil Std EIR-FT')
      CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil EIR-FF')
      CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil PLF')
      CALL AddToObjStr('Nominal Time for Condensate Removal to Begin', '')
      CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity', '')
      CALL AddToObjStr('Maximum Cycling Rate', '')
      CALL AddToObjStr('Latent Capacity Time Constant', '',.TRUE.)

      IF (coolCoilKind .EQ. ccTwoStageHumidControlDX) THEN
        !***CoilPerformance:DX:Cooling
        ! Dehumidification (subcool) - Stage 1
        CALL CreateNewObj('CoilPerformance:DX:Cooling')
        CALL AddToObjFld('Name',base + ussAirHandlerNameOff,' Dehumid Perf 1')
        IF (isCoolCapAutosize) THEN
          CALL AddToObjStr('Rated Total Cooling Capacity', 'autosize')
        ELSE
          ! half the capacity if provided
          CALL AddToObjNum('Rated Total Cooling Capacity', 0.5 * StringToReal(FldVal(base + ussCoolCoilCapOff)))
        END IF
        IF (isSHRAutosize) THEN
          CALL AddToObjStr('Gross Rated Sensible Heat Ratio','autosize')
        ELSE
          CALL AddToObjNum('Gross Rated Sensible Heat Ratio', 0.9 * StringToReal(FldVal(base + ussCoolCoilSHROff)))
        ENDIF
        CALL AddToObjNum('Rated COP', 0.9 * StringToReal(FldVal(base + ussCoolCoilCOPOff)))
        CALL AddToObjStr('Rated Air Flow Rate', 'autosize')
        CALL AddToObjStr('Fraction of Air Flow Bypassed Around Coil', '0.5')
        CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
           base + ussAirHandlerNameOff,' DX Coil Dehum Cap-FT')
        CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
           base + ussAirHandlerNameOff,' DX Coil Cap-FF')
        CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
           base + ussAirHandlerNameOff,' DX Coil Dehum EIR-FT')
        CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
           base + ussAirHandlerNameOff,' DX Coil EIR-FF')
        CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
           base + ussAirHandlerNameOff,' DX Coil PLF')
        CALL AddToObjStr('Nominal Time for Condensate Removal to Begin', '')
        CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity', '')
        CALL AddToObjStr('Maximum Cycling Rate', '')
        CALL AddToObjStr('Latent Capacity Time Constant', '',.TRUE.)
        !***CoilPerformance:DX:Cooling
        ! Dehumidification (subcool) - Stage 1 + 2
        CALL CreateNewObj('CoilPerformance:DX:Cooling')
        CALL AddToObjFld('Name',base + ussAirHandlerNameOff,' Dehumid Perf 1+2')
        CALL AddToObjFld('Rated Total Cooling Capacity', base + ussCoolCoilCapOff,'')
        IF (isSHRAutosize) THEN
          CALL AddToObjStr('Gross Rated Sensible Heat Ratio','autosize')
        ELSE
          CALL AddToObjNum('Gross Rated Sensible Heat Ratio', 0.9 * StringToReal(FldVal(base + ussCoolCoilSHROff)))
        ENDIF
        CALL AddToObjNum('Rated COP', 0.9 * StringToReal(FldVal(base + ussCoolCoilCOPOff)))
        CALL AddToObjStr('Rated Air Flow Rate', 'autosize')
        CALL AddToObjStr('Fraction of Air Flow Bypassed Around Coil', '0.0')
        CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
           base + ussAirHandlerNameOff,' DX Coil Dehum Cap-FT')
        CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
           base + ussAirHandlerNameOff,' DX Coil Cap-FF')
        CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
           base + ussAirHandlerNameOff,' DX Coil Dehum EIR-FT')
        CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
           base + ussAirHandlerNameOff,' DX Coil EIR-FF')
        CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
           base + ussAirHandlerNameOff,' DX Coil PLF')
        CALL AddToObjStr('Nominal Time for Condensate Removal to Begin', '')
        CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity', '')
        CALL AddToObjStr('Maximum Cycling Rate', '')
        CALL AddToObjStr('Latent Capacity Time Constant', '',.TRUE.)
      ENDIF
      !***Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil Std Cap-FT')
      CALL AddToObjStr('Coefficient1 Constant','1.067939449')
      CALL AddToObjStr('Coefficient2 x','-0.031261829')
      CALL AddToObjStr('Coefficient3 x**2','0.001974308')
      CALL AddToObjStr('Coefficient4 y','-0.002726426')
      CALL AddToObjStr('Coefficient5 y**2','-5.52654E-05')
      CALL AddToObjStr('Coefficient6 x*y','-6.31169E-05')
      CALL AddToObjStr('Minimum Value of x','-100')
      CALL AddToObjStr('Maximum Value of x','100')
      CALL AddToObjStr('Minimum Value of y','-100')
      CALL AddToObjStr('Maximum Value of y','100')
      CALL AddToObjStr('Minimum Curve Output','0.0')
      CALL AddToObjStr('Maximum Curve Output','5.0')
      CALL AddToObjStr('Input Unit Type for X','Temperature')
      CALL AddToObjStr('Input Unit Type for Y','Temperature')
      CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
      IF (coolCoilKind .EQ. ccTwoStageHumidControlDX) THEN
        !***Curve:Biquadratic
        CALL CreateNewObj('Curve:Biquadratic')
        CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil Dehum Cap-FT')
        CALL AddToObjStr('Coefficient1 Constant','0.596779741')
        CALL AddToObjStr('Coefficient2 x','0.034216637')
        CALL AddToObjStr('Coefficient3 x**2','0.000113924')
        CALL AddToObjStr('Coefficient4 y','-0.00375859')
        CALL AddToObjStr('Coefficient5 y**2','-9.17495E-05')
        CALL AddToObjStr('Coefficient6 x*y','-8.98373E-05')
        CALL AddToObjStr('Minimum Value of x','-100')
        CALL AddToObjStr('Maximum Value of x','100')
        CALL AddToObjStr('Minimum Value of y','-100')
        CALL AddToObjStr('Maximum Value of y','100')
        CALL AddToObjStr('Minimum Curve Output','0.0')
        CALL AddToObjStr('Maximum Curve Output','5.0')
        CALL AddToObjStr('Input Unit Type for X','Temperature')
        CALL AddToObjStr('Input Unit Type for Y','Temperature')
        CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
      ENDIF
      !***Curve:Quadratic
      CALL CreateNewObj('Curve:Quadratic')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil Cap-FF')
      CALL AddToObjStr('Coefficient1 Constant','0.8')
      CALL AddToObjStr('Coefficient2 x','0.2')
      CALL AddToObjStr('Coefficient3 x**2','0.0')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !***Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil Std EIR-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.174059889')
      CALL AddToObjStr('Coefficient2 x','0.022281508')
      CALL AddToObjStr('Coefficient3 x**2','-0.000134077')
      CALL AddToObjStr('Coefficient4 y','0.028298025')
      CALL AddToObjStr('Coefficient5 y**2','0.00048510')
      CALL AddToObjStr('Coefficient6 x*y','-0.001677095')
      CALL AddToObjStr('Minimum Value of x','-100')
      CALL AddToObjStr('Maximum Value of x','100')
      CALL AddToObjStr('Minimum Value of y','-100')
      CALL AddToObjStr('Maximum Value of y','100')
      CALL AddToObjStr('Minimum Curve Output','0.0')
      CALL AddToObjStr('Maximum Curve Output','5.0')
      CALL AddToObjStr('Input Unit Type for X','Temperature')
      CALL AddToObjStr('Input Unit Type for Y','Temperature')
      CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
      IF (coolCoilKind .EQ. ccTwoStageHumidControlDX) THEN
        !***Curve:Biquadratic
        CALL CreateNewObj('Curve:Biquadratic')
        CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil Dehum EIR-FT')
        CALL AddToObjStr('Coefficient1 Constant','0.435347586')
        CALL AddToObjStr('Coefficient2 x','0.004015641')
        CALL AddToObjStr('Coefficient3 x**2','0.000604235')
        CALL AddToObjStr('Coefficient4 y','0.015824043')
        CALL AddToObjStr('Coefficient5 y**2','0.000747287')
        CALL AddToObjStr('Coefficient6 x*y','-0.001779745')
        CALL AddToObjStr('Minimum Value of x','-100')
        CALL AddToObjStr('Maximum Value of x','100')
        CALL AddToObjStr('Minimum Value of y','-100')
        CALL AddToObjStr('Maximum Value of y','100')
        CALL AddToObjStr('Minimum Curve Output','0.0')
        CALL AddToObjStr('Maximum Curve Output','5.0')
        CALL AddToObjStr('Input Unit Type for X','Temperature')
        CALL AddToObjStr('Input Unit Type for Y','Temperature')
        CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
      ENDIF
      !***Curve:Quadratic
      CALL CreateNewObj('Curve:Quadratic')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil EIR-FF')
      CALL AddToObjStr('Coefficient1 Constant','1.1552')
      CALL AddToObjStr('Coefficient2 x','-0.1808')
      CALL AddToObjStr('Coefficient3 x**2','0.0256')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !***Curve:Quadratic
      CALL CreateNewObj('Curve:Quadratic')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil PLF')
      CALL AddToObjStr('Coefficient1 Constant','0.85')
      CALL AddToObjStr('Coefficient2 x','0.15')
      CALL AddToObjStr('Coefficient3 x**2','0.0')
      CALL AddToObjStr('Minimum Value of x','0.0')
      CALL AddToObjStr('Maximum Value of x','1.0',.TRUE.)
    CASE (ccWaterCoilSystemHX)
      !***CoilSystem:Cooling:Water:HeatExchangerAssisted
      CALL CreateNewObj('CoilSystem:Cooling:Water:HeatExchangerAssisted')
      CALL AddToObjStr('Name', TRIM(coolCoilObjectName))
      CALL AddToObjStr('Heat Exchanger Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
      CALL AddToObjFld('Heat Exchanger Name', base + ussAirHandlerNameOff,' Cooling Coil Heat Exchanger')
      CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:Water')
      CALL AddToObjFld('Cooling Coil Name', base + ussAirHandlerNameOff, ' Cooling Coil',.TRUE.)
      !***HeatExchanger:AirToAir:SensibleAndLatent
      CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cooling Coil Heat Exchanger')
      CALL AddToObjStr('Availability Schedule Name',' ')
      CALL AddToObjStr('Nominal Supply Air Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Sensible Effectiveness at 100% Heating Air Flow {dimensionless}','0.7')
      CALL AddToObjStr('Latent Effectiveness at 100% Heating Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Sensible Effectiveness at 75% Heating Air Flow {dimensionless}','0.75')
      CALL AddToObjStr('Latent Effectiveness at 75% Heating Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Sensible Effectiveness at 100% Cooling Air Flow {dimensionless}','0.7')
      CALL AddToObjStr('Latent Effectiveness at 100% Cooling Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Sensible Effectiveness at 75% Cooling Air Flow {dimensionless}','0.75')
      CALL AddToObjStr('Latent Effectiveness at 75% Cooling Air Flow {dimensionless}','0.0')
      CALL AddToObjFld('Supply Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(coolCoilUnitInlet))
      CALL AddToObjFld('Supply Air Outlet Node Name', base + ussAirHandlerNameOff,' HX Supply Outlet Node')
      CALL AddToObjFld('Exhaust Air Inlet Node Name', base + ussAirHandlerNameOff,' Cooling Coil Outlet Node')
      CALL AddToObjFld('Exhaust Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('Nominal Electric Power {W}','0')
      CALL AddToObjStr('Supply Air Outlet Temperature Control','No')
      CALL AddToObjStr('Heat Exchanger Type','Plate')
      CALL AddToObjStr('Frost Control Type','None')
      CALL AddToObjStr('Threshold Temperature {C}','')
      CALL AddToObjStr('Initial Defrost Time Fraction {dimensionless}','')
      CALL AddToObjStr('Rate of Defrost Time Fraction Increase {1/K}','')
      ! MJW ???? - Not sure
      CALL AddToObjStr('Economizer Lockout','No',.TRUE.)
      !***Coil:Cooling:Water
      CALL CreateNewObj('Coil:Cooling:Water')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + ussCoolAvailSchedNameOff,' ')
      CALL AddToObjStr('Design Water Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Design Air Flow Rate {m3/s}', 'autosize')
      CALL AddToObjStr('Design Inlet Water Temperature {C}','autosize')
      CALL AddToObjStr('Design Inlet Air Temperature {C}','autosize')
      CALL AddToObjStr('Design Outlet Air Temperature {C}','autosize')
      CALL AddToObjStr('Design Inlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
      CALL AddToObjStr('Design Outlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
      CALL AddToObjFld('Water Inlet Node Name', base + ussAirHandlerNameOff,' Cooling Coil ChW Inlet')
      CALL AddToObjFld('Water Outlet Node Name', base + ussAirHandlerNameOff,' Cooling Coil ChW Outlet')
      CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff,' HX Supply Outlet Node')
      CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff,' Cooling Coil Outlet Node')
      CALL AddToObjStr('Type of Analysis','DetailedAnalysis')
      CALL AddToObjStr('Heat Exchanger Configuration','CrossFlow',.TRUE.)
    CASE (ccDXcoilSystemHX)
      !***CoilSystem:Cooling:DX:HeatExchangerAssisted
      CALL CreateNewObj('CoilSystem:Cooling:DX:HeatExchangerAssisted')
      CALL AddToObjStr('Name', TRIM(coolCoilObjectName))
      CALL AddToObjStr('Heat Exchanger Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
      CALL AddToObjFld('Heat Exchanger Name', base + ussAirHandlerNameOff,' Cooling Coil Heat Exchanger')
      CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:DX:SingleSpeed')
      CALL AddToObjFld('Cooling Coil Name', base + ussAirHandlerNameOff,' Cooling Coil',.TRUE.)
      !***HeatExchanger:AirToAir:SensibleAndLatent
      CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cooling Coil Heat Exchanger')
      CALL AddToObjStr('Availability Schedule Name',' ')
      CALL AddToObjStr('Nominal Supply Air Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Sensible Effectiveness at 100% Heating Air Flow {dimensionless}','0.7')
      CALL AddToObjStr('Latent Effectiveness at 100% Heating Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Sensible Effectiveness at 75% Heating Air Flow {dimensionless}','0.75')
      CALL AddToObjStr('Latent Effectiveness at 75% Heating Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Sensible Effectiveness at 100% Cooling Air Flow {dimensionless}','0.7')
      CALL AddToObjStr('Latent Effectiveness at 100% Cooling Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Sensible Effectiveness at 75% Cooling Air Flow {dimensionless}','0.75')
      CALL AddToObjStr('Latent Effectiveness at 75% Cooling Air Flow {dimensionless}','0.0')
      CALL AddToObjFld('Supply Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(coolCoilUnitInlet))
      CALL AddToObjFld('Supply Air Outlet Node Name', base + ussAirHandlerNameOff,' HX Supply Outlet Node')
      CALL AddToObjFld('Exhaust Air Inlet Node Name', base + ussAirHandlerNameOff,' Cooling Coil Outlet Node')
      CALL AddToObjFld('Exhaust Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('Nominal Electric Power {W}','0')
      CALL AddToObjStr('Supply Air Outlet Temperature Control','No')
      CALL AddToObjStr('Heat Exchanger Type','Plate')
      CALL AddToObjStr('Frost Control Type','None')
      CALL AddToObjStr('Threshold Temperature {C}','')
      CALL AddToObjStr('Initial Defrost Time Fraction {dimensionless}','')
      CALL AddToObjStr('Rate of Defrost Time Fraction Increase {1/K}','')
      ! MJW ???? - Not sure
      CALL AddToObjStr('Economizer Lockout','No',.TRUE.)
      !***Coil:Cooling:DX:SingleSpeed
      CALL CreateNewObj('Coil:Cooling:DX:SingleSpeed')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + ussCoolAvailSchedNameOff,' ')
      CALL AddToObjFld('Rated Total Cooling Capacity', base + ussCoolCoilCapOff, '')
      CALL AddToObjFld('Gross Rated Sensible Heat Ratio', base + ussCoolCoilSHROff, '')
      CALL AddToObjFld('Rated COP', base + ussCoolCoilCOPOff, '')
      CALL AddToObjStr('Rated Air Flow Rate', 'autosize')
      CALL AddToObjStr('Rated Evaporator Fan Power per Volume Flow Rate {W/(m3/s)}','')
      CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff,' HX Supply Outlet Node')
      CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff,' Cooling Coil Outlet Node')
      CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil Cap-FT')
      CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil Cap-FF')
      CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil EIR-FT')
      CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil EIR-FF')
      CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
         base + ussAirHandlerNameOff,' DX Coil PLF')
      CALL AddToObjStr('Minimum Outdoor Dry-Bulb Temperature for Compressor Operation {C}','')
      CALL AddToObjStr('Nominal Time for Condensate Removal to Begin','0')
      CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity','0')
      CALL AddToObjStr('Maximum Cycling Rate','0')
      CALL AddToObjStr('Latent Capacity Time Constant','0')
      CALL AddToObjFld('Condenser Air Inlet Node Name',base + ussAirHandlerNameOff,' Cooling Coil Condenser Inlet')
      CALL AddToObjStr('Condenser Type','AirCooled')
      CALL AddToObjStr('Evaporative Condenser Effectiveness','0')
      CALL AddToObjStr('Evaporative Condenser Air Flow Rate','')
      CALL AddToObjStr('Evaporative Condenser Pump Rated Power Consumption','0')
      CALL AddToObjStr('Crankcase Heater Capacity','0')
      CALL AddToObjStr('Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater Operation','10',.TRUE.)
      !***Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('DOE-2.1E, COOL-CAP-FT for PTAC w/ SI temps')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil Cap-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.942587793')
      CALL AddToObjStr('Coefficient2 x','0.009543347')
      CALL AddToObjStr('Coefficient3 x**2','0.00068377')
      CALL AddToObjStr('Coefficient4 y','-0.011042676')
      CALL AddToObjStr('Coefficient5 y**2','0.000005249')
      CALL AddToObjStr('Coefficient6 x*y','-0.00000972')
      CALL AddToObjStr('Minimum Value of x','0.0')
      CALL AddToObjStr('Maximum Value of x','50.0')
      CALL AddToObjStr('Minimum Value of y','0.0')
      CALL AddToObjStr('Maximum Value of y','50.0')
      CALL AddToObjStr('Minimum Curve Output','0.0')
      CALL AddToObjStr('Maximum Curve Output','5.0')
      CALL AddToObjStr('Input Unit Type for X','Temperature')
      CALL AddToObjStr('Input Unit Type for Y','Temperature')
      CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
      !CURVE:QUADRATIC
      CALL CreateNewObj('Curve:Quadratic')
      CALL WriteComment('DOE-2.1E, RATED-CCAP-FFLOW for PTAC')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil Cap-FF')
      CALL AddToObjStr('Coefficient1 Constant','0.8')
      CALL AddToObjStr('Coefficient2 x','0.2')
      CALL AddToObjStr('Coefficient3 x**2','0')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !CURVE:BIQUADRATIC
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('DOE-2.1E, COOL-EIR-FT for PTAC w/ SI temps')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil EIR-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.342414409')
      CALL AddToObjStr('Coefficient2 x','0.034885008')
      CALL AddToObjStr('Coefficient3 x**2','-0.0006237')
      CALL AddToObjStr('Coefficient4 y','0.004977216')
      CALL AddToObjStr('Coefficient5 y**2','0.000437951')
      CALL AddToObjStr('Coefficient6 x*y','-0.000728028')
      CALL AddToObjStr('Minimum Value of x','0.0')
      CALL AddToObjStr('Maximum Value of x','50.0')
      CALL AddToObjStr('Minimum Value of y','0.0')
      CALL AddToObjStr('Maximum Value of y','50.0')
      CALL AddToObjStr('Minimum Curve Output','0.0')
      CALL AddToObjStr('Maximum Curve Output','5.0')
      CALL AddToObjStr('Input Unit Type for X','Temperature')
      CALL AddToObjStr('Input Unit Type for Y','Temperature')
      CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
      !CURVE:QUADRATIC
      CALL CreateNewObj('Curve:Quadratic')
      CALL WriteComment('DOE-2.1E, RATED-CEIR-FFLOW for PTAC')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil EIR-FF')
      CALL AddToObjStr('Coefficient1 Constant','1.1552')
      CALL AddToObjStr('Coefficient2 x','-0.1808')
      CALL AddToObjStr('Coefficient3 x**2','0.0256')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !CURVE:QUADRATIC
      CALL CreateNewObj('Curve:Quadratic')
      CALL WriteComment('PLF = l.- Cd(1.-PLR) where Cd = 0.15')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' DX Coil PLF')
      CALL AddToObjStr('Coefficient1 Constant','0.85')
      CALL AddToObjStr('Coefficient2 x','0.15')
      CALL AddToObjStr('Coefficient3 x**2','0')
      CALL AddToObjStr('Minimum Value of x','0')
      CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
      !OutAir Node
      CALL CreateNewObj('OutdoorAir:Node')
      CALL AddToObjFld('Name',base + ussAirHandlerNameOff,' Cooling Coil Condenser Inlet')
      CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
    CASE (ccMultiSpeedDX)
      ! Cooling coil
      IF (.NOT. isCoolCapAutosize) THEN
        CapPerSpeed = StringToReal(FldVal(base + ussCoolCoilCapOff))/StringToReal(FldVal(base + ussCoolCoilNumSpeedOff))
      END IF
      numSpeeds = StringToInt(FldVal(base + ussCoolCoilNumSpeedOff))
      isNumSpeedOne = (numSpeeds == 1)
      !Object ==> Coil:Cooling:DX:MultiSpeed
      CALL CreateNewObj('Coil:Cooling:DX:MultiSpeed')
      CALL AddToObjStr('Name', TRIM(coolCoilObjectName))
      CALL AddToObjFld('Availability Schedule Name', base + ussCoolAvailSchedNameOff,' ')
      CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(coolCoilUnitInlet))
      CALL AddToObjFld('Air Outlet Node Name',base + ussAirHandlerNameOff, TRIM(coolCoilUnitOutlet))
      CALL AddToObjFld('Condenser Air Inlet Node Name',base + ussAirHandlerNameOff,' Cooling Coil Condenser Inlet')
      CALL AddToObjStr('Condenser Type','AirCooled')
      CALL AddToObjStr('Minimum Outdoor Dry-Bulb Temperature for Compressor Operation {C}','')
      CALL AddToObjStr('Supply Water Storage Tank Name','')
      CALL AddToObjStr('Condensate Collection Water Storage Tank Name','')
      CALL AddToObjStr('Apply Part Load Fraction to Speeds Greater than 1','No')
      CALL AddToObjStr('Apply Latent Degradation to Speeds Greater than 1','No')
      CALL AddToObjStr('Crankcase Heater Capacity {W}','0.0')
      CALL AddToObjStr('Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater Operation {C}','10.0')
      CALL AddToObjStr('Basin Heater Capacity {W/K}','')
      CALL AddToObjStr('Basin Heater Setpoint Temperature {C}','')
      CALL AddToObjStr('Basin Heater Operating Schedule Name','')
      CALL AddToObjStr('Fuel Type','Electricity')
      CALL AddToObjFld('Number of Speeds', base + ussCoolCoilNumSpeedOff,'')
      IF (isCoolCapAutosize) THEN
        CALL AddToObjStr('Speed 1 Gross Rated Total Cooling Capacity {W}','autosize')
      ELSE
        CALL AddToObjNum('Speed 1 Gross Rated Total Cooling Capacity {W}',CapPerSpeed)
      END IF
      CALL AddToObjFld('Speed 1 Gross Rated Sensible Heat Ratio', base + ussCoolCoilSHROff,' ')
      CALL AddToObjFld('Speed 1 Gross Rated Cooling COP {W/W}', base + ussCoolCoilCOPOff,' ')
      CALL AddToObjStr('Speed 1 Rated Air Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Speed 1 Rated Evaporator Fan Power Per Volume Flow Rate {W/(m3/s)}','')
      CALL AddToObjFld('Speed 1 Total Cooling Capacity Function of Temperature Curve Name',   &
         base + ussAirHandlerNameOff,' Cool Coil Cap-FT')
      CALL AddToObjFld('Speed 1 Total Cooling Capacity Function of Flow Fraction Curve Name',   &
         base + ussAirHandlerNameOff,' Cool Coil Cap-FF')
      CALL AddToObjFld('Speed 1 Energy Input Ratio Function of Temperature Curve Name',   &
         base + ussAirHandlerNameOff,' Cool Coil EIR-FT')
      CALL AddToObjFld('Speed 1 Energy Input Ratio Function of Flow Fraction Curve Name',   &
         base + ussAirHandlerNameOff,' Cool Coil EIR-FF')
      CALL AddToObjFld('Speed 1 Part Load Fraction Correlation Curve Name',   &
         base + ussAirHandlerNameOff,' Cool Coil PLF')
      CALL AddToObjStr('Speed 1 Nominal Time for Condensate Removal to Begin','0')
      CALL AddToObjStr('Speed 1 Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity','0')
      CALL AddToObjStr('Speed 1 Maximum Cycling Rate','0')
      CALL AddToObjStr('Speed 1 Latent Capacity Time Constant','0')
      CALL AddToObjStr('Speed 1 Rated Waste Heat Fraction of Power Input {dimensionless}','0.2')
      CALL AddToObjFld('Speed 1 Waste Heat Function of Temperature Curve Name',base + ussAirHandlerNameOff,' Cool Coil WH-FT')
      CALL AddToObjStr('Speed 1 Evaporative Condenser Effectiveness {dimensionless}','')
      CALL AddToObjStr('Speed 1 Evaporative Condenser Air Flow Rate {m3/s}','')
      CALL AddToObjStr('Speed 1 Rated Evaporative Condenser Pump Power Consumption {W}','',isNumSpeedOne)
      IF (.NOT. isNumSpeedOne) THEN
        DO iSpd = 2, numSpeeds
          iSpdChar = IntToStr(iSpd)
          IF (isCoolCapAutosize) THEN
            CALL AddToObjStr('Speed '//TRIM(iSpdChar)//' Gross Rated Total Cooling Capacity {W}','autosize')
          ELSE
            CALL AddToObjNum('Speed '//TRIM(iSpdChar)//' Gross Rated Total Cooling Capacity {W}',CapPerSpeed*iSpd)
          END IF
          CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Gross Rated Sensible Heat Ratio', base + ussCoolCoilSHROff,' ')
          CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Gross Rated Cooling COP {W/W}', base + ussCoolCoilCOPOff,' ')
          CALL AddToObjStr('Speed '//TRIM(iSpdChar)//' Rated Air Flow Rate {m3/s}','autosize')
          CALL AddToObjStr('Speed '//TRIM(iSpdChar)//' Rated Evaporator Fan Power Per Volume Flow Rate {W/(m3/s)}','')
          CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Total Cooling Capacity Function of Temperature Curve Name',   &
             base + ussAirHandlerNameOff,' Cool Coil Cap-FT')
          CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Total Cooling Capacity Function of Flow Fraction Curve Name',   &
             base + ussAirHandlerNameOff,' Cool Coil Cap-FF')
          CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Energy Input Ratio Function of Temperature Curve Name',   &
             base + ussAirHandlerNameOff,' Cool Coil EIR-FT')
          CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Energy Input Ratio Function of Flow Fraction Curve Name',   &
             base + ussAirHandlerNameOff,' Cool Coil EIR-FF')
          CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Part Load Fraction Correlation Curve Name',   &
             base + ussAirHandlerNameOff,' Cool Coil PLF')
          CALL AddToObjStr('Speed '//TRIM(iSpdChar)//' Nominal Time for Condensate Removal to Begin','0')
          CALL AddToObjStr('Speed '//TRIM(iSpdChar)// &
           ' Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity','0')
          CALL AddToObjStr('Speed '//TRIM(iSpdChar)//' Maximum Cycling Rate','0')
          CALL AddToObjStr('Speed '//TRIM(iSpdChar)//' Latent Capacity Time Constant','0',isNumSpeedOne)
          CALL AddToObjStr('Speed '//TRIM(iSpdChar)//' Rated Waste Heat Fraction of Power Input {dimensionless}','0.2')
          CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Waste Heat Function of Temperature Curve Name',base + ussAirHandlerNameOff, &
           ' Cool Coil WH-FT')
          CALL AddToObjStr('Speed '//TRIM(iSpdChar)//' Evaporative Condenser Effectiveness {dimensionless}','')
          CALL AddToObjStr('Speed '//TRIM(iSpdChar)//' Evaporative Condenser Air Flow Rate {m3/s}','')
          CALL AddToObjStr('Speed '//TRIM(iSpdChar)//' Rated Evaporative Condenser Pump Power Consumption {W}','', &
           (iSpd == numSpeeds))
        END DO
      END IF

      ! Performance curves, same for every speed
      CALL WriteComment('Curves from example file MultiSpeedHeatPump.idf, Sep 2013, same curves for all speeds.')
      !Object ==> Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cool Coil Cap-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.476428E+00')
      CALL AddToObjStr('Coefficient2 x','0.401147E-01')
      CALL AddToObjStr('Coefficient3 x**2','0.226411E-03')
      CALL AddToObjStr('Coefficient4 y','-0.827136E-03')
      CALL AddToObjStr('Coefficient5 y**2','-0.732240E-05')
      CALL AddToObjStr('Coefficient6 x*y','-0.446278E-03')
      CALL AddToObjStr('Minimum Value of x','0.0')
      CALL AddToObjStr('Maximum Value of x','50.0')
      CALL AddToObjStr('Minimum Value of y','0.0')
      CALL AddToObjStr('Maximum Value of y','50.0')
      CALL AddToObjStr('Minimum Curve Output','0.0')
      CALL AddToObjStr('Maximum Curve Output','5.0')
      CALL AddToObjStr('Input Unit Type for X','Temperature')
      CALL AddToObjStr('Input Unit Type for Y','Temperature')
      CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
      !Object ==> Curve:Cubic
      CALL CreateNewObj('Curve:Cubic')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cool Coil Cap-FF')
      CALL AddToObjStr('Coefficient1 Constant','.47278589')
      CALL AddToObjStr('Coefficient2 x','1.2433415')
      CALL AddToObjStr('Coefficient3 x**2','-1.0387055')
      CALL AddToObjStr('Coefficient4 x**3','.32257813')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !Object ==> Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cool Coil EIR-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.632475E+00')
      CALL AddToObjStr('Coefficient2 x','-0.121321E-01')
      CALL AddToObjStr('Coefficient3 x**2','0.507773E-03')
      CALL AddToObjStr('Coefficient4 y','0.155377E-01')
      CALL AddToObjStr('Coefficient5 y**2','0.272840E-03')
      CALL AddToObjStr('Coefficient6 x*y','-0.679201E-03')
      CALL AddToObjStr('Minimum Value of x','0.0')
      CALL AddToObjStr('Maximum Value of x','50.0')
      CALL AddToObjStr('Minimum Value of y','0.0')
      CALL AddToObjStr('Maximum Value of y','50.0')
      CALL AddToObjStr('Minimum Curve Output','0.0')
      CALL AddToObjStr('Maximum Curve Output','5.0')
      CALL AddToObjStr('Input Unit Type for X','Temperature')
      CALL AddToObjStr('Input Unit Type for Y','Temperature')
      CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
      !Object ==> Curve:Quadratic
      CALL CreateNewObj('Curve:Cubic')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cool Coil EIR-FF')
      CALL AddToObjStr('Coefficient1 Constant','.47278589')
      CALL AddToObjStr('Coefficient2 x','1.2433415')
      CALL AddToObjStr('Coefficient3 x**2','-1.0387055')
      CALL AddToObjStr('Coefficient4 x**3','.32257813')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
       !Object ==> Curve:Quadratic
      CALL WriteComment('PLF = l.- Cd(1.-PLR) where Cd = 0.15')
      CALL CreateNewObj('Curve:Quadratic')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cool Coil PLF')
      CALL AddToObjStr('Coefficient1 Constant','0.85')
      CALL AddToObjStr('Coefficient2 x','0.15')
      CALL AddToObjStr('Coefficient3 x**2','0')
      CALL AddToObjStr('Minimum Value of x','0')
      CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
      !waste heat curve
      CALL CreateNewObj('Curve:Biquadratic')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cool Coil WH-FT')
      CALL AddToObjStr('Coefficient1 Constant','1.0')
      CALL AddToObjStr('Coefficient2 x','0.0')
      CALL AddToObjStr('Coefficient3 x**2','0.0')
      CALL AddToObjStr('Coefficient4 y','0.0')
      CALL AddToObjStr('Coefficient5 y**2','0.0')
      CALL AddToObjStr('Coefficient6 x*y','0.0')
      CALL AddToObjStr('Minimum Value of x','0')
      CALL AddToObjStr('Maximum Value of x','50')
      CALL AddToObjStr('Minimum Value of y','0')
      CALL AddToObjStr('Maximum Value of y','50')
      CALL AddToObjStr('Minimum Curve Output','')
      CALL AddToObjStr('Maximum Curve Output','')
      CALL AddToObjStr('Input Unit Type for X','Temperature')
      CALL AddToObjStr('Input Unit Type for Y','Temperature')
      CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
      !Object ==> OutdoorAir:Node
      CALL CreateNewObj('OutdoorAir:Node')
      CALL AddToObjFld('Name',base + ussAirHandlerNameOff,' Cooling Coil Condenser Inlet')
      CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
    CASE (ccSingleSpeedDXWater)
       ! Warn if cooling capacity and SHR inputs are not consistent
      IF (isCoolCapAutosize .AND. (.NOT. isSHRAutosize)) THEN
        CALL WriteError('Warning:  In HVACTemplate:System:UnitarySystem "'//TRIM(FldVal(base + ussAirHandlerNameOff))//'"'// &
         ' the Cooling Coil Rated Capacity is autosized, so the Cooling Coil Gross Rated Sensible Heat Ratio '// &
         ' will also be autosized. The specified value for Cooling Coil Gross Rated Sensible Heat Ratio'// &
         ' will be ignored.  Autosize both or specify values for both.',msgWarning)
      ENDIF
      IF ((.NOT.isCoolCapAutosize) .AND. isSHRAutosize) THEN
        CALL WriteError('Warning:  In HVACTemplate:System:UnitarySystem "'//TRIM(FldVal(base + ussAirHandlerNameOff))//'"'// &
         ' the Cooling Coil Rated Capacity will not be used when the Cooling Coil Gross Rated Sensible Heat Ratio'// &
         ' is autosized. Autosize both or specify values for both.',msgWarning)
      ENDIF
      CALL WriteComment('Performance coefficients assume rated Capacities and COP are at:')
      CALL WriteComment('26.7C (80F) entering air dry-bulb temperature, 19.4C (67F) entering air wet-bulb temperature')
      CALL WriteComment('and 29.4C (85F) entering water temperature')
      CALL CreateNewObj('Coil:Cooling:WaterToAirHeatPump:EquationFit')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjFld('Water Inlet Node Name', base + ussAirHandlerNameOff,' WAHP Cooling Water Inlet Node')
      CALL AddToObjFld('Water Outlet Node Name', base + ussAirHandlerNameOff,' WAHP Cooling Water Outlet Node')
      CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(coolCoilUnitInlet))
      CALL AddToObjFld('Air Outlet Node Name',base + ussAirHandlerNameOff, TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('Rated Air Flow Rate {m3/s}','Autosize')
      CALL AddToObjStr('Rated Water Flow Rate {m3/s}','Autosize')
      CALL AddToObjFld('Rated Total Cooling Capacity {W}',base + ussCoolCoilCapOff,'')
      IF (isCoolCapAutosize .OR. isSHRAutosize) THEN
        CALL AddToObjStr('Rated Sensible Cooling Capacity {W}','Autosize')
      ELSE
        CALL AddToObjNum('Rated Sensible Cooling Capacity {W}',StringToReal(FldVal(base + ussCoolCoilCapOff)) * &
         StringToReal(FldVal(base + ussCoolCoilSHROff)))
      ENDIF
      CALL AddToObjFld('Rated Cooling Coefficient of Performance',base + ussCoolCoilCOPOff,'')
      CALL AddToObjStr('Total Cooling Capacity Coefficient 1','-9.149069561')
      CALL AddToObjStr('Total Cooling Capacity Coefficient 2','10.87814026')
      CALL AddToObjStr('Total Cooling Capacity Coefficient 3','-1.718780157')
      CALL AddToObjStr('Total Cooling Capacity Coefficient 4','0.746414818')
      CALL AddToObjStr('Total Cooling Capacity Coefficient 5','0.0')
      CALL AddToObjStr('Sensible Cooling Capacity Coefficient 1','-5.462690012')
      CALL AddToObjStr('Sensible Cooling Capacity Coefficient 2','17.95968138')
      CALL AddToObjStr('Sensible Cooling Capacity Coefficient 3','-11.87818402')
      CALL AddToObjStr('Sensible Cooling Capacity Coefficient 4','-0.980163419')
      CALL AddToObjStr('Sensible Cooling Capacity Coefficient 5','0.767285761')
      CALL AddToObjStr('Sensible Cooling Capacity Coefficient 6','0.0')
      CALL AddToObjStr('Cooling Power Consumption Coefficient 1','-3.205409884')
      CALL AddToObjStr('Cooling Power Consumption Coefficient 2','-0.976409399')
      CALL AddToObjStr('Cooling Power Consumption Coefficient 3','3.97892546')
      CALL AddToObjStr('Cooling Power Consumption Coefficient 4','0.938181818')
      CALL AddToObjStr('Cooling Power Consumption Coefficient 5','0.0')
      CALL AddToObjStr('Nominal Time for Condensate Removal to Begin {s}','0')
      CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity {dimensionless}','0',.TRUE.)
      !***Branch
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cooling Condenser Branch')
      CALL AddToStrList(TRIM(FldVal(base + ussAirHandlerNameOff)) // ' Cooling Condenser Branch',handleMixedDemanBranch)
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','Coil:Cooling:WaterToAirHeatPump:EquationFit')
      CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjFld('Component Inlet Node Name', base + ussAirHandlerNameOff,' WAHP Cooling Water Inlet Node')
      CALL AddToObjFld('Component Outlet Node Name', base + ussAirHandlerNameOff,' WAHP Cooling Water Outlet Node',.TRUE.)
    CASE (ccNone)
      ! nothing to do if no cooling coil specified
  END SELECT

  !***Chilled Water Branch
  IF ((coolCoilKind .EQ. ccChWater) .OR. (coolCoilKind .EQ. ccChWaterDetailed) .OR. (coolCoilKind .EQ. ccWaterCoilSystemHX)) THEN
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cooling Coil ChW Branch')
    !store for later in the chilled water loop
    CALL AddToStrList(TRIM(FldVal(base + ussAirHandlerNameOff)) // ' Cooling Coil ChW Branch', handleCoolingCoilBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF ((coolCoilKind .EQ. ccChWater) .OR. (coolCoilKind .EQ. ccWaterCoilSystemHX)) THEN
      CALL AddToObjStr('Component Object Type','Coil:Cooling:Water')
    ELSEIF (coolCoilKind .EQ. ccChWaterDetailed) THEN
      CALL AddToObjStr('Component Object Type','Coil:Cooling:Water:DetailedGeometry')
    END IF
    CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' Cooling Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + ussAirHandlerNameOff,' Cooling Coil ChW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + ussAirHandlerNameOff,' Cooling Coil ChW Outlet',.TRUE.)
  ENDIF


  ! Heating Coil
  IF (heatcoilKind .EQ. hcElectric) THEN
    !Object ==> Coil:Heating:Electric
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjStr('Name', TRIM(heatCoilObjectName))
    CALL AddToObjFld('Availability Schedule Name', base + ussHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Efficiency','1')
    CALL AddToObjFld('Nominal Capacity of the Coil {W}', base + ussHeatCoilCapOff,'')
    CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(heatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(heatCoilOutlet))
    CALL AddToObjStr('Coil Temp Setpoint Node','',.TRUE.)
  ELSE IF (heatcoilKind .EQ. hcGas) THEN
    !Object ==> Coil:Heating:Fuel
    CALL CreateNewObj('Coil:Heating:Fuel')
    CALL AddToObjStr('Name', TRIM(heatCoilObjectName))
    CALL AddToObjFld('Availability Schedule Name', base + ussHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Fuel Type', 'Gas')
    CALL AddToObjFld('Gas Burner Efficiency', base + ussHeatEfficiencyOff,' ')
    CALL AddToObjFld('Nominal Capacity of the Coil {W}', base + ussHeatCoilCapOff,'')
    CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(heatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(heatCoilOutlet))
    CALL AddToObjStr('Coil Temp Setpoint Node','')
    CALL AddToObjFld('Parasitic Electric Load {W}', base + ussHeatParasiticOff,' ')
    CALL AddToObjFld('Part load fraction correlation (function of part load ratio)',   &
       base + ussAirHandlerNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
    !Object ==> Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  ELSE IF (heatcoilKind .EQ. hcMultiStageElectric) THEN
    IF (.NOT. isHeatCapAutosize) THEN
      CapPerSpeed = StringToReal(FldVal(base + ussHeatCoilCapOff))/StringToReal(FldVal(base + ussHeatCoilNumSpeedOff))
    END IF
    numSpeeds = StringToInt(FldVal(base + ussHeatCoilNumSpeedOff))
    isNumSpeedOne = (numSpeeds == 1)
    !Object ==> Coil:Heating:Electric:Multistage
    CALL CreateNewObj('Coil:Heating:Electric:Multistage')
    CALL AddToObjStr('Name', TRIM(heatCoilObjectName))
    CALL AddToObjFld('Availability Schedule Name', base + ussHeatAvailSchedNameOff,' ')
    CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(heatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(heatCoilOutlet))
    CALL AddToObjStr('Temperature Setpoint Node Name','')
    CALL AddToObjFld('Number of Stages', base + ussHeatCoilNumSpeedOff,'')
    CALL AddToObjFld('Stage 1 Efficiency {W/W}', base + ussHeatEfficiencyOff,' ')
    IF (isHeatCapAutosize) THEN
      CALL AddToObjStr('Stage 1 Nominal Capacity {W}','autosize',isNumSpeedOne)
    ELSE
      CALL AddToObjNum('Stage 1 Nominal Capacity {W}',CapPerSpeed,isNumSpeedOne)
    END IF
    IF (.NOT. isNumSpeedOne) THEN
      DO iSpd = 2, numSpeeds
        iSpdChar = IntToStr(iSpd)
        CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Efficiency {W/W}', base + ussHeatEfficiencyOff,' ')
        IF (isHeatCapAutosize) THEN
          CALL AddToObjStr('Stage '//TRIM(iSpdChar)//' Nominal Capacity {W}','autosize',(iSpd == numSpeeds))
        ELSE
          CALL AddToObjNum('Stage '//TRIM(iSpdChar)//' Nominal Capacity {W}',CapPerSpeed*iSpd,(iSpd == numSpeeds))
        END IF
      END DO
    END IF
  ELSE IF (heatcoilKind .EQ. hcMultiStageGas) THEN
    IF (.NOT. isHeatCapAutosize) THEN
      CapPerSpeed = StringToReal(FldVal(base + ussHeatCoilCapOff))/StringToReal(FldVal(base + ussHeatCoilNumSpeedOff))
    END IF
    numSpeeds = StringToInt(FldVal(base + ussHeatCoilNumSpeedOff))
    isNumSpeedOne = (numSpeeds == 1)
    !Object ==> Coil:Heating:Gas:Multistage
    CALL CreateNewObj('Coil:Heating:Gas:Multistage')
    CALL AddToObjStr('Name', TRIM(heatCoilObjectName))
    CALL AddToObjFld('Availability Schedule Name', base + ussHeatAvailSchedNameOff,' ')
    CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(heatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(heatCoilOutlet))
    CALL AddToObjStr('Temperature Setpoint Node Name','')
    CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
       base + ussAirHandlerNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0')
    CALL AddToObjFld('Number of Stages', base + ussHeatCoilNumSpeedOff,'')
    CALL AddToObjFld('Stage 1 Gas Burner Efficiency {W/W}', base + ussHeatEfficiencyOff,' ')
    IF (isHeatCapAutosize) THEN
      CALL AddToObjStr('Stage 1 Nominal Capacity {W}','autosize')
    ELSE
      CALL AddToObjNum('Stage 1 Nominal Capacity {W}',CapPerSpeed)
    END IF
    CALL AddToObjFld('Stage 1 Parasitic Electric Load {W}', base + ussHeatParasiticOff,' ',isNumSpeedOne)
    IF (.NOT. isNumSpeedOne) THEN
      DO iSpd = 2, numSpeeds
        iSpdChar = IntToStr(iSpd)
        CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Gas Burner Efficiency {W/W}', base + ussHeatEfficiencyOff,' ')
        IF (isHeatCapAutosize) THEN
          CALL AddToObjStr('Stage '//TRIM(iSpdChar)//' Nominal Capacity {W}','autosize')
        ELSE
          CALL AddToObjNum('Stage '//TRIM(iSpdChar)//' Nominal Capacity {W}',CapPerSpeed*iSpd)
        END IF
        CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Parasitic Electric Load {W}', base + ussHeatParasiticOff,' ', &
         (iSpd == numSpeeds))
      END DO
    END IF
    !Object ==> Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  ELSE IF (heatcoilKind .EQ. hcWater) THEN
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjStr('Name', TRIM(heatCoilObjectName))
    CALL AddToObjFld('Availability Schedule Name', base + ussHeatAvailSchedNameOff,'')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name', base + ussAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + ussAirHandlerNameOff,' Heating Coil HW Outlet')
    CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(heatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(heatCoilOutlet))
    IF (isHeatCapAutosize) THEN
      CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
    ELSE
      CALL AddToObjStr('Performance Input Method','NominalCapacity')
    END IF
    CALL AddToObjFld('Rated Capacity', base + ussHeatCoilCapOff,'')
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Heating Coil HW Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + ussAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + ussAirHandlerNameOff,' Heating Coil HW Outlet',.TRUE.)
    CALL AddToStrList(TRIM(FldVal(base + ussAirHandlerNameOff)) // ' Heating Coil HW Branch',handleHeatingCoilBranch)
  ELSE IF (heatcoilKind .EQ. hcSingleSpeedDXHPAir) THEN
    !Object ==> Coil:Heating:DX:SingleSpeed
    CALL CreateNewObj('Coil:Heating:DX:SingleSpeed')
    CALL AddToObjStr('Name', TRIM(heatCoilObjectName))
    CALL AddToObjFld('Availability Schedule Name', base + ussHeatAvailSchedNameOff,'')
    CALL AddToObjFld('Rated Total Heating Capacity {W}', base + ussHeatCoilCapOff,'')
    CALL AddToObjFld('Rated COP', base + ussHPHeatCoilCOPOff,'')
    CALL AddToObjStr('Rated Air Flow Rate {m3/s}', 'autosize')
    CALL AddToObjStr('Rated Evaporator Fan Power Per Volume Flow Rate','')
    CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(heatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(heatCoilOutlet))
    CALL AddToObjFld('Total Heating Capacity Function of Temperature Curve Name',   &
       base + ussAirHandlerNameOff,' HP Heating Coil Cap-FT')
    CALL AddToObjFld('Total Heating Capacity Function of Flow Fraction Curve Name',   &
       base + ussAirHandlerNameOff,' HP Heating Coil Cap-FF')
    CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
       base + ussAirHandlerNameOff,' HP Heating Coil EIR-FT')
    CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
       base + ussAirHandlerNameOff,' HP Heating Coil EIR-FF')
    CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
       base + ussAirHandlerNameOff,' HP Heating Coil PLF')
    CALL AddToObjFld('Defrost Energy Input Ratio Function of Temperature Curve Name',   &
       base + ussAirHandlerNameOff,' HP Heating Coil DefrEIR-FT')
    CALL AddToObjFld('Minimum Outdoor Dry-Bulb Temperature for Compressor Operation {C}', base + ussHPHeatMinODBOff,'')
    CALL AddToObjStr('Outdoor Dry-Bulb Temperature to Turn On Compressor','')
    CALL AddToObjFld('Maximum Outdoor Dry-Bulb Temperature for Defrost Operation {C}', base + ussHPDefrMaxODBOff,'')
    CALL AddToObjStr('Crankcase Heater Capacity {W}','0')
    CALL AddToObjStr('Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater Operation {C}','0')
    CALL AddToObjFld('Defrost Strategy', base + ussHPDefrStrategyOff,'')
    CALL AddToObjFld('Defrost Control', base + ussHPDefrCtrlTypeOff,'')
    CALL AddToObjFld('Defrost Time Period Fraction', base + ussHPDefrTimeFracOff,'')
    CALL AddToObjStr('Resistive Defrost Heater Capacity {W}','autosize',.TRUE.)
    !Object ==> Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' HP Heating Coil Cap-FT')
    CALL AddToObjStr('Coefficient1 Constant','0.758746')
    CALL AddToObjStr('Coefficient2 x','0.027626')
    CALL AddToObjStr('Coefficient3 x**2','0.000148716')
    CALL AddToObjStr('Coefficient4 x**3','0.0000034992')
    CALL AddToObjStr('Minimum Value of x','-20.0')
    CALL AddToObjStr('Maximum Value of x','20.0',.TRUE.)
    !Object ==> Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' HP Heating Coil Cap-FF')
    CALL AddToObjStr('Coefficient1 Constant','0.84')
    CALL AddToObjStr('Coefficient2 x','0.16')
    CALL AddToObjStr('Coefficient3 x**2','0.0')
    CALL AddToObjStr('Coefficient4 x**3','0.0')
    CALL AddToObjStr('Minimum Value of x','0.5')
    CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
    !Object ==> Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' HP Heating Coil EIR-FT')
    CALL AddToObjStr('Coefficient1 Constant','1.19248')
    CALL AddToObjStr('Coefficient2 x','-0.0300438')
    CALL AddToObjStr('Coefficient3 x**2','0.00103745')
    CALL AddToObjStr('Coefficient4 x**3','-0.000023328')
    CALL AddToObjStr('Minimum Value of x','-20.0')
    CALL AddToObjStr('Maximum Value of x','20.0',.TRUE.)
    !Object ==> Curve:Quadratic
    CALL CreateNewObj('Curve:Quadratic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' HP Heating Coil EIR-FF')
    CALL AddToObjStr('Coefficient1 Constant','1.3824')
    CALL AddToObjStr('Coefficient2 x','-0.4336')
    CALL AddToObjStr('Coefficient3 x**2','0.0512')
    CALL AddToObjStr('Minimum Value of x','0.0')
    CALL AddToObjStr('Maximum Value of x','1.0',.TRUE.)
    !Object ==> Curve:Quadratic
    CALL CreateNewObj('Curve:Quadratic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' HP Heating Coil PLF')
    CALL AddToObjStr('Coefficient1 Constant','0.75')
    CALL AddToObjStr('Coefficient2 x','0.25')
    CALL AddToObjStr('Coefficient3 x**2','0.0')
    CALL AddToObjStr('Minimum Value of x','0.0')
    CALL AddToObjStr('Maximum Value of x','1.0',.TRUE.)
    !Object ==> Curve:Biquadratic
    CALL CreateNewObj('Curve:Biquadratic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' HP Heating Coil DefrEIR-FT')
    CALL AddToObjStr('Coefficient1 Constant','1')
    CALL AddToObjStr('Coefficient2 x','0')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 y','0')
    CALL AddToObjStr('Coefficient5 y**2','0')
    CALL AddToObjStr('Coefficient6 x*y','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','50')
    CALL AddToObjStr('Minimum Value of y','0')
    CALL AddToObjStr('Maximum Value of y','50',.TRUE.)
  ELSE IF (heatcoilKind .EQ. hcMultiSpeedDXHPAir) THEN
    ! Heating coil
    IF (.NOT. isHeatCapAutosize) THEN
      CapPerSpeed = StringToReal(FldVal(base + ussHeatCoilCapOff))/StringToReal(FldVal(base + ussHeatCoilNumSpeedOff))
    END IF
    numSpeeds = StringToInt(FldVal(base + ussHeatCoilNumSpeedOff))
    isNumSpeedOne = (numSpeeds == 1)
    !Object ==> Coil:Heating:DX:MultiSpeed
    CALL CreateNewObj('Coil:Heating:DX:MultiSpeed')
    CALL AddToObjStr('Name', TRIM(heatCoilObjectName))
    CALL AddToObjFld('Availability Schedule Name', base + ussHeatAvailSchedNameOff,' ')
    CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(heatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name',base + ussAirHandlerNameOff, TRIM(heatCoilOutlet))
    CALL AddToObjFld('Minimum Outdoor Dry-Bulb Temperature for Compressor Operation {C}', base + ussHPHeatMinODBOff,'')
    CALL AddToObjStr('Outdoor Dry-Bulb Temperature to Turn On Compressor','')
    CALL AddToObjStr('Crankcase Heater Capacity {W}','0.0')
    CALL AddToObjStr('Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater Operation {C}','10.0')
    CALL AddToObjFld('Defrost Energy Input Ratio Function of Temperature Curve Name',base + ussAirHandlerNameOff, &
                     ' Heat Coil Defr-FT')
    CALL AddToObjFld('Maximum Outdoor Dry-Bulb Temperature for Defrost Operation {C}', base + ussHPDefrMaxODBOff,'')
    CALL AddToObjFld('Defrost Strategy', base + ussHPDefrStrategyOff,'')
    CALL AddToObjFld('Defrost Control', base + ussHPDefrCtrlTypeOff,'')
    CALL AddToObjFld('Defrost Time Period Fraction', base + ussHPDefrTimeFracOff,'')
    CALL AddToObjStr('Resistive Defrost Heater Capacity {W}','autosize')
    CALL AddToObjStr('Apply Part Load Fraction to Speeds Greater than 1','No')
    CALL AddToObjStr('Fuel Type','Electricity')
    CALL AddToObjStr('Region number for Calculating HSPF','')
    CALL AddToObjFld('Number of Speeds', base + ussHeatCoilNumSpeedOff,'')
    IF (isHeatCapAutosize) THEN
      CALL AddToObjStr('Speed 1 Gross Rated Heating Capacity {W}','autosize')
    ELSE
      CALL AddToObjNum('Speed 1 Gross Rated Heating Capacity {W}',CapPerSpeed)
    END IF
    CALL AddToObjFld('Speed 1 Gross Rated Heating COP {W/W}', base + ussHPHeatCoilCOPOff,' ')
    CALL AddToObjStr('Speed 1 Rated Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Speed 1 Rated Supply Air Fan Power Per Volume Flow Rate {W/(m3/s)}','')
    CALL AddToObjFld('Speed 1 Heating Capacity Function of Temperature Curve Name',   &
        base + ussAirHandlerNameOff,' Heat Coil Cap-FT')
    CALL AddToObjFld('Speed 1 Heating Capacity Function of Flow Fraction Curve Name',   &
        base + ussAirHandlerNameOff,' Heat Coil Cap-FF')
    CALL AddToObjFld('Speed 1 Energy Input Ratio Function of Temperature Curve Name',   &
        base + ussAirHandlerNameOff,' Heat Coil EIR-FT')
    CALL AddToObjFld('Speed 1 Energy Input Ratio Function of Flow Fraction Curve Name',   &
        base + ussAirHandlerNameOff,' Heat Coil EIR-FF')
    CALL AddToObjFld('Speed 1 Part Load Fraction Correlation Curve Name',   &
        base + ussAirHandlerNameOff,' Heat Coil PLF')
    CALL AddToObjStr('Speed 1 Rated Waste Heat Fraction of Power Input {dimensionless}','0.2')
    CALL AddToObjFld('Speed 1 Waste Heat Function of Temperature Curve Name',base + ussAirHandlerNameOff,' Heat Coil WH-FT', &
     isNumSpeedOne)
    IF (.NOT. isNumSpeedOne) THEN
      DO iSpd = 2, numSpeeds
        iSpdChar = IntToStr(iSpd)
        IF (isHeatCapAutosize) THEN
          CALL AddToObjStr('Speed '//TRIM(iSpdChar)//' Gross Rated Heating Capacity {W}','autosize')
        ELSE
          CALL AddToObjNum('Speed '//TRIM(iSpdChar)//' Gross Rated Heating Capacity {W}',CapPerSpeed*iSpd)
        END IF
        CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Gross Rated Heating COP {W/W}', base + ussHPHeatCoilCOPOff,' ')
        CALL AddToObjStr('Speed '//TRIM(iSpdChar)//' Rated Air Flow Rate {m3/s}','autosize')
        CALL AddToObjStr('Speed '//TRIM(iSpdChar)//' Rated Supply Air Fan Power Per Volume Flow Rate {W/(m3/s)}','')
        CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Heating Capacity Function of Temperature Curve Name',   &
            base + ussAirHandlerNameOff,' Heat Coil Cap-FT')
        CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Heating Capacity Function of Flow Fraction Curve Name',   &
            base + ussAirHandlerNameOff,' Heat Coil Cap-FF')
        CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Energy Input Ratio Function of Temperature Curve Name',   &
            base + ussAirHandlerNameOff,' Heat Coil EIR-FT')
        CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Energy Input Ratio Function of Flow Fraction Curve Name',   &
            base + ussAirHandlerNameOff,' Heat Coil EIR-FF')
        CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Part Load Fraction Correlation Curve Name',   &
            base + ussAirHandlerNameOff,' Heat Coil PLF')
        CALL AddToObjStr('Speed '//TRIM(iSpdChar)//' Rated Waste Heat Fraction of Power Input {dimensionless}','0.2')
        CALL AddToObjFld('Speed '//TRIM(iSpdChar)//' Waste Heat Function of Temperature Curve Name',base + ussAirHandlerNameOff, &
                          ' Heat Coil WH-FT',(iSpd == numSpeeds))
      END DO
    END IF

    ! Performance curves, same for every speed
    CALL WriteComment('Curves from example file MultiSpeedHeatPump.idf, Sep 2013, same curves for all speeds.')
    !Object ==> Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Heat Coil Cap-FT')
    CALL AddToObjStr('Coefficient1 Constant','0.758746')
    CALL AddToObjStr('Coefficient2 x','0.027626')
    CALL AddToObjStr('Coefficient3 x**2','0.000148716')
    CALL AddToObjStr('Coefficient4 x**3','0.0000034992')
    CALL AddToObjStr('Minimum Value of x','-20.0')
    CALL AddToObjStr('Maximum Value of x','20.0')
    CALL AddToObjStr('Minimum Curve Output','')
    CALL AddToObjStr('Maximum Curve Output','')
    CALL AddToObjStr('Input Unit Type for X','Temperature')
    CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
    !Object ==> Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Heat Coil Cap-FF')
    CALL AddToObjStr('Coefficient1 Constant','0.84')
    CALL AddToObjStr('Coefficient2 x','0.16')
    CALL AddToObjStr('Coefficient3 x**2','0.0')
    CALL AddToObjStr('Coefficient4 x**3','0.0')
    CALL AddToObjStr('Minimum Value of x','0.5')
    CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
    !Object ==> Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Heat Coil EIR-FT')
    CALL AddToObjStr('Coefficient1 Constant','1.19248')
    CALL AddToObjStr('Coefficient2 x','-0.0300438')
    CALL AddToObjStr('Coefficient3 x**2','0.00103745')
    CALL AddToObjStr('Coefficient4 x**3','-0.000023328')
    CALL AddToObjStr('Minimum Value of x','-20.0')
    CALL AddToObjStr('Maximum Value of x','20.0')
    CALL AddToObjStr('Minimum Curve Output','')
    CALL AddToObjStr('Maximum Curve Output','')
    CALL AddToObjStr('Input Unit Type for X','Temperature')
    CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
    !Object ==> Curve:Quadratic
    CALL CreateNewObj('Curve:Quadratic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Heat Coil EIR-FF')
    CALL AddToObjStr('Coefficient1 Constant','1.3824')
    CALL AddToObjStr('Coefficient2 x','-0.4336')
    CALL AddToObjStr('Coefficient3 x**2','0.0512')
    CALL AddToObjStr('Minimum Value of x','0.0')
    CALL AddToObjStr('Maximum Value of x','1.0',.TRUE.)
      !Object ==> Curve:Quadratic
    CALL WriteComment('PLF = l.- Cd(1.-PLR) where Cd = 0.15')
    CALL CreateNewObj('Curve:Quadratic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Heat Coil PLF')
    CALL AddToObjStr('Coefficient1 Constant','0.85')
    CALL AddToObjStr('Coefficient2 x','0.15')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
    ! defrost curve
    CALL CreateNewObj('Curve:Biquadratic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Heat Coil Defr-FT')
    CALL AddToObjStr('Coefficient1 Constant','1.0')
    CALL AddToObjStr('Coefficient2 x','0.0')
    CALL AddToObjStr('Coefficient3 x**2','0.0')
    CALL AddToObjStr('Coefficient4 y','0.0')
    CALL AddToObjStr('Coefficient5 y**2','0.0')
    CALL AddToObjStr('Coefficient6 x*y','0.0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','50')
    CALL AddToObjStr('Minimum Value of y','0')
    CALL AddToObjStr('Maximum Value of y','50')
    CALL AddToObjStr('Minimum Curve Output','')
    CALL AddToObjStr('Maximum Curve Output','')
    CALL AddToObjStr('Input Unit Type for X','Temperature')
    CALL AddToObjStr('Input Unit Type for Y','Temperature')
    CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
    !waste heat curve
    CALL CreateNewObj('Curve:Biquadratic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Heat Coil WH-FT')
    CALL AddToObjStr('Coefficient1 Constant','1.0')
    CALL AddToObjStr('Coefficient2 x','0.0')
    CALL AddToObjStr('Coefficient3 x**2','0.0')
    CALL AddToObjStr('Coefficient4 y','0.0')
    CALL AddToObjStr('Coefficient5 y**2','0.0')
    CALL AddToObjStr('Coefficient6 x*y','0.0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','50')
    CALL AddToObjStr('Minimum Value of y','0')
    CALL AddToObjStr('Maximum Value of y','50')
    CALL AddToObjStr('Minimum Curve Output','')
    CALL AddToObjStr('Maximum Curve Output','')
    CALL AddToObjStr('Input Unit Type for X','Temperature')
    CALL AddToObjStr('Input Unit Type for Y','Temperature')
    CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
    !Object ==> OutdoorAir:Node
    CALL CreateNewObj('OutdoorAir:Node')
    CALL AddToObjFld('Name',base + ussAirHandlerNameOff,' Heating Coil Condenser Inlet')
    CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
  ELSE IF (heatcoilKind .EQ. hcSingleSpeedDXHPWater) THEN
    !***Coil:Heating:WaterToAirHeatPump:EquationFit
    CALL WriteComment('Performance coefficients assume rated Capacity and COP are at:')
    CALL WriteComment('21.11C (70F) entering air dry-bulb temperature and 21.11C (70F) entering water temperature')
    CALL CreateNewObj('Coil:Heating:WaterToAirHeatPump:EquationFit')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Water Inlet Node Name', base + ussAirHandlerNameOff,' WAHP Heating Water Inlet Node')
    CALL AddToObjFld('Water Outlet Node Name', base + ussAirHandlerNameOff,' WAHP Heating Water Outlet Node')
    CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(heatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(heatCoilOutlet))
    CALL AddToObjStr('Rated Air Flow Rate {m3/s}','Autosize')
    CALL AddToObjStr('Rated Water Flow Rate {m3/s}','Autosize')
    CALL AddToObjFld('Rated Heating Capacity {W}',base + ussHeatCoilCapOff,'')
    CALL AddToObjFld('Rated Heating Coefficient of Performance',base + ussHPHeatCoilCOPOff,'')
    CALL AddToObjStr('Heating Capacity Coefficient 1','-1.361311959')
    CALL AddToObjStr('Heating Capacity Coefficient 2','-2.471798046')
    CALL AddToObjStr('Heating Capacity Coefficient 3','4.173164514')
    CALL AddToObjStr('Heating Capacity Coefficient 4','0.640757401')
    CALL AddToObjStr('Heating Capacity Coefficient 5','0.0')
    CALL AddToObjStr('Heating Power Consumption Coefficient 1','-2.176941116')
    CALL AddToObjStr('Heating Power Consumption Coefficient 2','0.832114286')
    CALL AddToObjStr('Heating Power Consumption Coefficient 3','1.570743399')
    CALL AddToObjStr('Heating Power Consumption Coefficient 4','0.690793651')
    CALL AddToObjStr('Heating Power Consumption Coefficient 5','0.0',.TRUE.)
    !***Branch
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Heating Condenser Branch')
    CALL AddToStrList(TRIM(FldVal(base + ussAirHandlerNameOff)) // ' Heating Condenser Branch',handleMixedDemanBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:WaterToAirHeatPump:EquationFit')
    CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + ussAirHandlerNameOff,' WAHP Heating Water Inlet Node')
    CALL AddToObjFld('Component Outlet Node Name', base + ussAirHandlerNameOff,' WAHP Heating Water Outlet Node',.TRUE.)
  END IF
  ! Supplemental Heating Coil
  IF (supHeatCoilKind .EQ. hcElectric) THEN
    !Object ==> Coil:Heating:Electric
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjStr('Name', TRIM(supHeatCoilObjectName))
    CALL AddToObjFld('Availability Schedule Name', base + ussSuppReHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Efficiency','1')
    CALL AddToObjFld('Nominal Capacity of the Coil {W}', base + ussSuppReHeatCoilCapOff,'')
    CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(supheatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(supheatCoilOutlet))
    CALL AddToObjStr('Coil Temp Setpoint Node','',.TRUE.)
  ELSE IF (supHeatCoilKind .EQ. hcGas) THEN
    !Object ==> Coil:Heating:Fuel
    CALL CreateNewObj('Coil:Heating:Fuel')
    CALL AddToObjStr('Name', TRIM(supHeatCoilObjectName))
    CALL AddToObjFld('Availability Schedule Name', base + ussSuppReHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Fuel Type', 'Gas')
    CALL AddToObjFld('Gas Burner Efficiency', base + ussSuppReHeatEfficiencyOff,' ')
    CALL AddToObjFld('Nominal Capacity of the Coil {W}', base + ussSuppReHeatCoilCapOff,'')
    CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(supheatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(supheatCoilOutlet))
    CALL AddToObjStr('Coil Temp Setpoint Node','')
    CALL AddToObjFld('Parasitic Electric Load {W}', base + ussSuppReHeatParasiticOff,' ')
    CALL AddToObjFld('Part load fraction correlation (function of part load ratio)',   &
       base + ussAirHandlerNameOff,' Sup Heating Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
    !Object ==> Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Sup Heating Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  ELSE IF (supHeatCoilKind .EQ. hcWater) THEN
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjStr('Name', TRIM(supHeatCoilObjectName))
    CALL AddToObjFld('Availability Schedule Name', base + ussHeatAvailSchedNameOff,'')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name', base + ussAirHandlerNameOff,' Sup Heat Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + ussAirHandlerNameOff,' Sup Heat Coil HW Outlet')
    CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(supheatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(supheatCoilOutlet))
    IF (isSuppHeatCoilCapAutosize) THEN
      CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
    ELSE
      CALL AddToObjStr('Performance Input Method','NominalCapacity')
    END IF
    CALL AddToObjFld('Rated Capacity', base + ussSuppReHeatCoilCapOff,'')
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Sup Heat Coil HW Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjStr('Component Name', TRIM(supHeatCoilObjectName))
    CALL AddToObjFld('Component Inlet Node Name', base + ussAirHandlerNameOff,' Sup Heat Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + ussAirHandlerNameOff,' Sup Heat Coil HW Outlet',.TRUE.)
    CALL AddToStrList(TRIM(FldVal(base + ussAirHandlerNameOff)) // ' Sup Heat Coil HW Branch',handleHeatingCoilBranch)
  ELSE IF (supHeatCoilKind .EQ. hcDeSuperHeat) THEN
    CALL CreateNewObj('Coil:Heating:Desuperheater')
    CALL AddToObjStr('Name', TRIM(supHeatCoilObjectName))
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjStr('Heat Reclaim Recovery Efficiency','0.3')
    CALL AddToObjFld('Air Inlet Node Name', base + usAirHandlerNameOff,TRIM(supheatCoilInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + usAirHandlerNameOff,TRIM(supheatCoilOutlet))
    CALL AddToObjStr('Heating Source Object Type',TRIM(coolCoilObjectType))
    CALL AddToObjStr('Heating Source Name', TRIM(coolCoilObjectName))
    CALL AddToObjStr('Temperature Setpoint Node Name','')
    CALL AddToObjStr('Parasitic Electric Load {W}','',.TRUE.)
  END IF
  ! Return fan
  IF ((isReturnFanYes) .AND. (supFanKind .EQ. sfkVAV)) THEN
    CALL WriteComment('VAV Fan coefficients based on variable-speed drive.')
    CALL CreateNewObj('Fan:VariableVolume')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Return Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + ussSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + ussReturnEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + ussReturnPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Fan Power Minimum Flow Rate Input Method','Fraction')
    CALL AddToObjStr('Fan Power Minimum Flow Fraction', '0.0')
    CALL AddToObjStr('Fan Power Minimum Air Flow Rate {m3/s}', '')
    CALL AddToObjFld('Motor Efficiency', base + ussReturnFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + ussReturnFanFractionOff,' ')
    CALL WriteComment('VAV Fan coefficients based on variable-speed drive.')
    CALL AddToObjStr('Fan Power Coefficient 1','0.0015302446')
    CALL AddToObjStr('Fan Power Coefficient 2','0.0052080574')
    CALL AddToObjStr('Fan Power Coefficient 3','1.1086242')
    CALL AddToObjStr('Fan Power Coefficient 4','-0.11635563')
    CALL AddToObjStr('Fan Power Coefficient 5','0')
    CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff,' Return Fan Outlet',.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF
  IF ((isReturnFanYes) .AND. (supFanKind .EQ. sfkCAV)) THEN
    CALL CreateNewObj('Fan:ConstantVolume')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Return Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + ussSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + ussReturnEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + ussReturnPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Motor Efficiency', base + ussReturnFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + ussReturnFanFractionOff,' ')
    CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff,' Return Fan Outlet',.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF
  IF (supFanKind .EQ. sfkVAV) THEN
    CALL WriteComment('VAV Fan coefficients based on variable-speed drive.')
    CALL CreateNewObj('Fan:VariableVolume')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Supply Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + ussSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + ussSupplyEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + ussSupplyPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
    CALL AddToObjStr('Fan Power Minimum Flow Rate Input Method','Fraction')
    CALL AddToObjStr('Fan Power Minimum Flow Fraction', '0.0')
    CALL AddToObjStr('Fan Power Minimum Air Flow Rate {m3/s}', '')
    CALL AddToObjFld('Motor Efficiency', base + ussSupplyFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + ussSupplyFanFractionOff,' ')
    CALL AddToObjStr('Fan Power Coefficient 1','0.0015302446')
    CALL AddToObjStr('Fan Power Coefficient 2','0.0052080574')
    CALL AddToObjStr('Fan Power Coefficient 3','1.1086242')
    CALL AddToObjStr('Fan Power Coefficient 4','-0.11635563')
    CALL AddToObjStr('Fan Power Coefficient 5','0')
    CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(supFanInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(supFanOutlet),.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF
  IF (supFanKind .EQ. sfkCAV) THEN
    CALL CreateNewObj('Fan:OnOff')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Supply Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + ussSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + ussSupplyEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + ussSupplyPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
    CALL AddToObjFld('Motor Efficiency', base + ussSupplyFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + ussSupplyFanFractionOff,' ')
    CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff, TRIM(supFanInlet))
    CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff, TRIM(supFanOutlet),.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF

  IF (.NOT. isHumidifierNone) THEN
    !Object ==> SetpointManager:SingleZone:Humidity:Minimum
    CALL CreateNewObj('SetpointManager:SingleZone:Humidity:Minimum')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Humidification Setpoint Manager')
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + ussAirHandlerNameOff,TRIM(humidifierOutlet))
    CALL AddToObjFld('Control Zone Air Node Name', base + ussHumidCtrlZoneOff,' Zone Air Node',.TRUE.)
    IF (humidifierKind .EQ. humidifyElecSteam) THEN
      !Object ==> Humidifier:Steam:Electric
      CALL CreateNewObj('Humidifier:Steam:Electric')
      CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Humidifier')
      CALL AddToObjFld('Availability Schedule Name', base + ussHumidAvailSchedNameOff,' ')
      CALL AddToObjFld('Rated Capacity {m3/s}', base + ussHumidCapOff,'')
      CALL AddToObjFld('Rated Power {W}', base + ussHumidRatedElecOff,'')
      CALL AddToObjStr('rated fan power {W}','0')
      CALL AddToObjStr('Standby power {W}','0')
      CALL AddToObjFld('Air Inlet Node Name', base + ussAirHandlerNameOff,TRIM(humidifierInlet))
      CALL AddToObjFld('Air Outlet Node Name', base + ussAirHandlerNameOff,TRIM(humidifierOutlet),.TRUE.)
    END IF
  END IF

  !Object ==> OutdoorAir:NodeList
  CALL CreateNewObj('OutdoorAir:NodeList')
  CALL AddToObjFld('Node or NodeList Name 1', base + ussAirHandlerNameOff,' Outdoor Air Inlet',.TRUE.)
  !OObject ==> AirLoopHVAC:OutdoorAirSystem
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' OA System')
  CALL AddToObjFld('Controller List Name', base + ussAirHandlerNameOff,' OA System Controllers')
  CALL AddToObjFld('Outdoor Air Equipment List Name', base + ussAirHandlerNameOff,' OA System Equipment')
  CALL AddToObjFld('Availability Manager List Name', base + ussAirHandlerNameOff,' Availability Managers',.TRUE.)
  !Object ==> AirLoopHVAC:ControllerList
  CALL CreateNewObj('AirLoopHVAC:ControllerList')
  CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' OA System Controllers')
  CALL AddToObjStr('Controller Object Type','Controller:OutdoorAir')
  CALL AddToObjFld('Controller Name', base + ussAirHandlerNameOff,' OA Controller',.TRUE.)
  !Object ==> AirLoopHVAC:OutdoorAirSystem:EquipmentList
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem:EquipmentList')
  CALL AddToObjFld('Name', base + usAirHandlerNameOff,' OA System Equipment')
  IF (isHeatRecSensible) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' Heat Recovery')
  END IF
  IF (isHeatRecEnthalpy) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' Heat Recovery')
  END IF
  CALL AddToObjStr('Component Object Type','OutdoorAir:Mixer')
  CALL AddToObjFld('Component Name', base + ussAirHandlerNameOff,' OA Mixing Box',.TRUE.)
  !Object ==> OutdoorAir:Mixer
  CALL CreateNewObj('OutdoorAir:Mixer')
  CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' OA Mixing Box')
  CALL AddToObjFld('Mixed Air Node Name', base + ussAirHandlerNameOff,' Mixed Air Outlet')
  IF (isHeatRecNone) THEN
    CALL AddToObjFld('Outdoor Air Stream Node Name', base + ussAirHandlerNameOff,' Outdoor Air Inlet')
  ELSEIF (isHeatRecSensible) THEN
    CALL AddToObjFld('Outdoor Air Stream Node Name', base + ussAirHandlerNameOff,' Heat Recovery Supply Outlet')
  ELSEIF (isHeatRecEnthalpy) THEN
    CALL AddToObjFld('Outdoor Air Stream Node Name', base + ussAirHandlerNameOff,' Heat Recovery Supply Outlet')
  ELSE
    CALL AddToObjFld('Outdoor Air Stream Node Name', base + ussAirHandlerNameOff,' Outdoor Air Inlet')
    CALL WriteError('Warning: Heat recovery option invalid. Assuming none.')
    isHeatRecNone = .TRUE.
  END IF
  CALL AddToObjFld('Relief Air Stream Node Name', base + ussAirHandlerNameOff,' Relief Air Outlet')
  CALL AddToObjFld('Return Air Stream Node Name', base + ussAirHandlerNameOff, TRIM(returnInletToOAMIxer),.TRUE.)
  !Object ==> SetpointManager:SingleZone:Cooling
  CALL CreateNewObj('SetpointManager:SingleZone:Cooling')
  CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Economizer Supply Air Temp Manager')
  CALL AddToObjStr('Control Variable','Temperature')
  CALL AddToObjStr('minimum supply air temperature {C}','13')
  CALL AddToObjStr('maximum supply air temperature {C}','45')
  CALL AddToObjFld('Control Zone Name', base + ussControlZoneOff,'')
  CALL AddToObjFld('Zone Node Name', base + ussControlZoneOff,' Zone Air Node')
  IF (isControlZonePlenumBlank) THEN
    CALL AddToObjFld('Zone Inlet Node Name', base + ussControlZoneOff,' Zone Equip Inlet')
  ELSE
    CALL AddToObjFld('Zone Inlet Node Name', base + ussControlZoneOff,' Supply Inlet')
  END IF
  CALL AddToObjFld('Setpoint Node or NodeList Name', base + ussAirHandlerNameOff,TRIM(unitOutlet),.TRUE.)
  !Object ==> SetpointManager:MixedAir
  CALL CreateNewObj('SetpointManager:MixedAir')
  CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Cooling Coil Air Temp Manager')
  CALL AddToObjStr('Control Variable','Temperature')
  CALL AddToObjFld('Reference Setpoint Node Name', base + ussAirHandlerNameOff,TRIM(unitOutlet))
  CALL AddToObjFld('Fan Inlet Node Name', base + ussAirHandlerNameOff, TRIM(supFanInlet))
  CALL AddToObjFld('Fan Outlet Node Name', base + ussAirHandlerNameOff, TRIM(supFanOutlet))
  CALL AddToObjFld('Setpoint Node or NodeList Name', base + ussAirHandlerNameOff,' Mixed Air Outlet',.TRUE.)
  !Object ==> Controller:OutdoorAir
  CALL CreateNewObj('Controller:OutdoorAir')
  CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' OA Controller')
  CALL AddToObjFld('Relief Air Outlet Node Name', base + ussAirHandlerNameOff,' Relief Air Outlet')
  CALL AddToObjFld('Return Air Node Name', base + ussAirHandlerNameOff, TRIM(returnInletToOAMIxer))
  CALL AddToObjFld('Mixed Air Node Name', base + ussAirHandlerNameOff,' Mixed Air Outlet')
  CALL AddToObjFld('Actuator Node Name', base + ussAirHandlerNameOff,' Outdoor Air Inlet')
  CALL AddToObjFld('Minimum Outdoor Air Flow Rate {m3/s}', base + ussMinOutsideFlowOff,' ')
  CALL AddToObjFld('Maximum Outdoor Air Flow Rate {m3/s}', base + ussMaxOutsideFlowOff,' ')
  ! usEconomizerKindOff is now a straight pass-thru to EconomizerChoice in
  ! the OA controller
  CALL AddToObjFld('Economizer Control Type', base + ussEconomizerKindOff,' ')
  ! If heat recovery is selected, then set economizer control to bypass
  ! otherwise, set to modulate OA flow
  IF (isHeatRecNone) THEN
    CALL AddtoObjStr('Economizer Control Action Type','ModulateFlow')
  ELSE
    CALL AddtoObjStr('Economizer Control Action Type','MinimumFlowWithBypass')
  ENDIF
  CALL AddToObjFld('Economizer Maximum Limit Dry-Bulb Temperature {C}', base + ussEconoUpTempLimitOff,' ')
  CALL AddToObjFld('Economizer Maximum Limit Enthalpy {J/kg}', base + ussEconoUpEnthLimitOff,' ')
  IF (econoKind .EQ. etFixedDewPointAndDryBulb) THEN
    CALL AddToObjFld('Economizer Maximum Limit Dewpoint Temperature (C)',base + ussEconoUpDewLimitOff, ' ')
  ELSE
    CALL AddToObjStr('Economizer Maximum Limit Dewpoint Temperature (C)',' ')
  ENDIF
  IF (econoKind .EQ. etElectronicEnthalpy) THEN
    CALL AddToObjFld('Electronic Enthalpy Limit Curve Name',base + ussAirHandlerNameOff,' ElectrEnthCurve')
  ELSE
    CALL AddToObjStr('Electronic Enthalpy Limit Curve Name',' ')
  ENDIF
  CALL AddToObjFld('Economizer Minimum Limit Dry-Bulb Temperature {C}', base + ussEconoLowTempLimitOff,' ')
  CALL AddToObjFld('Lockout Type', base + ussEconoLockoutKindOff,' ')
  CALL AddToObjStr('Minimum Limit Type', 'FixedMinimum')
  CALL AddToObjFld('Minimum Outdoor Air Schedule Name', base + ussMinOutFlowSchedNameOff,' ')
  CALL AddToObjStr('Minimum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Maximum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Mechanical Ventilation Controller Name','',.TRUE.)
  IF (econoKind .EQ. etElectronicEnthalpy) THEN
    !Object ==> Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' ElectrEnthCurve')
    CALL AddToObjStr('Coefficient1 Constant','0.01342704')
    CALL AddToObjStr('Coefficient2 x','-0.00047892')
    CALL AddToObjStr('Coefficient3 x**2','0.000053352')
    CALL AddToObjStr('Coefficient4 x**3','-0.0000018103')
    CALL AddToObjStr('Minimum Value of x','16.6')
    CALL AddToObjStr('Maximum Value of x','29.13',.TRUE.)
  ENDIF
  IF (isHeatRecSensible) THEN
    htRecSens75 = StringToReal(FldVal(base + ussHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    !Object ==> HeatExchanger:AirToAir:SensibleAndLatent
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + ussAirHandlerNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Nominal supply air flow rate {m3/s}', base + ussMinOutsideFlowOff,'')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + ussHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Heating Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Heating Air Flow','0')
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + ussHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Cooling Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Cooling Air Flow','0')
    CALL AddToObjFld('Supply Air Inlet Node Name', base + ussAirHandlerNameOff,' Outdoor Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + ussAirHandlerNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + ussAirHandlerNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + ussAirHandlerNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','0')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjStr('Heat exchanger type','Plate')
    CALL AddToObjStr('Frost Control Type','ExhaustOnly')
    CALL AddToObjStr('Threshold Temperature','-1.1')
    CALL AddToObjStr('Initial Defrost Time Fraction','0.167')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','0.024',.TRUE.)
  END IF
  IF (isHeatRecEnthalpy) THEN
    htRecSens75 = StringToReal(FldVal(base + ussHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    htRecLat75 = StringToReal(FldVal(base + ussHeatRecLatEffOff)) + 0.05
    IF (htRecLat75 .GT. 1.) htRecLat75 = 1.
    !Object ==> HeatExchanger:AirToAir:SensibleAndLatent
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + ussAirHandlerNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Nominal supply air flow rate {m3/s}', base + ussMinOutsideFlowOff,'')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + ussHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Heating Air Flow', base + ussHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow',htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Heating Air Flow', htRecLat75)
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + ussHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Cooling Air Flow', base + ussHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow',htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Cooling Air Flow', htRecLat75)
    CALL AddToObjFld('Supply Air Inlet Node Name', base + ussAirHandlerNameOff,' Outdoor Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + ussAirHandlerNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + ussAirHandlerNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + ussAirHandlerNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','0')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjStr('Heat exchanger type','Rotary')
    CALL AddToObjStr('Frost Control Type','MinimumExhaustTemperature')
    CALL AddToObjStr('Threshold Temperature','1.7')
    CALL AddToObjStr('Initial Defrost Time Fraction','')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','',.TRUE.)
  END IF
  IF (.NOT. isHeatRecNone) THEN
    IF (isEconoLowLimitBlank) THEN
      CALL AddAlwaysSchedule('5')
    ELSE
      CALL AddAlwaysSchedule(FldVal(base + ussEconoLowTempLimitOff))
    END IF
    !Object ==> SetpointManager:Scheduled
    CALL CreateNewObj('SetpointManager:Scheduled')
    CALL AddToObjFld('Name', base + ussAirHandlerNameOff,' Heat Recovery Air Temp Manager')
    CALL AddToObjStr('Control Variable','Temperature')
    IF (isEconoLowLimitBlank) THEN
      CALL AddToObjStr('Schedule Name','HVACTemplate-Always 5')
    ELSE
      CALL AddToObjStr('Schedule Name','HVACTemplate-Always ' // FldVal(base + ussEconoLowTempLimitOff))
    END IF
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + ussAirHandlerNameOff,' Heat Recovery Supply Outlet',.TRUE.)
  END IF
END DO
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE MakeSpecialIntList(typeOfZoneList,fldPointMatch,suppressMessages)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   March 2006
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Several lists of integers need to constructed over and over again
          !    This routine allows those special lists to be assembled in
          !    one central place in the code.  The lists are usually gathering
          !    pointers to the FldVal array that contain names of zones.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
INTEGER,INTENT(IN) :: typeOfZoneList
INTEGER,INTENT(IN) :: fldPointMatch
LOGICAL,INTENT(IN),OPTIONAL   :: suppressMessages


          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: altBase
INTEGER :: i
LOGICAL :: showMessages

showMessages = .TRUE.
IF (PRESENT(suppressMessages)) THEN
  IF (suppressMessages) THEN
    showMessages = .FALSE.
  END IF
END IF

altbase=0
CALL ClearIntList
SELECT CASE (typeOfZoneList)
  CASE (zlUnitAll) ! all unitary zones
    DO i = 1, numCompactZoneUnit
      altBase = compactZoneUnitBase(i)
      IF (SameString(FldVal(altBase + uzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        CALL AddToIntList(altBase + uzNameOff)
      END IF
    END DO
    IF (numIntList .EQ. 0) THEN
      IF (showMessages) THEN
        CALL WriteError('Did not find any HVACTemplate:Zone:Unitary objects connected to HVACTemplate:System:Unitary='//  &
         '"'//trim(FldVal(fldPointMatch))//'". There must be at least one HVACTemplate:Zone:Unitary object which specifies '// &
         'this system as the Template Unitary System Name.')
      END IF
    END IF
  CASE (zlUnitAllSupPlen)  !all unitary zones with supply plenum
    DO i = 1, numCompactZoneUnit
      altBase = compactZoneUnitBase(i)
      IF (SameString(FldVal(altBase + uzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF (FldVal(altBase + uzZoneSupplyPlenumNameOff) .NE. '') THEN
          CALL AddToIntList(altBase + uzNameOff)
        END IF
      END IF
    END DO
  CASE (zlUnitAllRetPlen) !all unitary zones with return plenum
    DO i = 1, numCompactZoneUnit
      altBase = compactZoneUnitBase(i)
      IF (SameString(FldVal(altBase + uzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF (FldVal(altBase + uzZoneReturnPlenumNameOff) .NE. '') THEN
          CALL AddToIntList(altBase + uzNameOff)
        END IF
      END IF
    END DO
  CASE (zlConstVolAll) ! all constant volume zones
    DO i = 1, numCompactZoneConstVol
      altBase = compactZoneConstVolBase(i)
      IF (SameString(FldVal(altBase + cvzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        CALL AddToIntList(altBase + cvzNameOff)
      END IF
    END DO
    IF (numIntList .EQ. 0) THEN
      IF (showMessages) THEN
        CALL WriteError( &
         'Did not find any HVACTemplate:Zone:ConstantVolume objects connected to HVACTemplate:System:ConstantVolume='// &
         '"'//trim(FldVal(fldPointMatch))// &
         '". There must be at least one HVACTemplate:Zone:ConstantVolume object which specifies '// &
         'this system as the Template Constant Volume System Name.')
      END IF
    END IF
  CASE (zlConstVolAllSupPlen)  !all constant volume zones with supply plenum
    DO i = 1, numCompactZoneConstVol
      altBase = compactZoneConstVolBase(i)
      IF (SameString(FldVal(altBase + cvzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF (FldVal(altBase + cvzZoneSupplyPlenumNameOff) .NE. '') THEN
          CALL AddToIntList(altBase + cvzNameOff)
        END IF
      END IF
    END DO
  CASE (zlConstVolAllRetPlen) !all constant volume zones with return plenum
    DO i = 1, numCompactZoneConstVol
      altBase = compactZoneConstVolBase(i)
      IF (SameString(FldVal(altBase + cvzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF (FldVal(altBase + cvzZoneReturnPlenumNameOff) .NE. '') THEN
          CALL AddToIntList(altBase + cvzNameOff)
        END IF
      END IF
    END DO
  CASE (zlDualDuctAll) ! all dual duct zones
    DO i = 1, numCompactZoneDualDuct
      altBase = compactZoneDualDuctBase(i)
      IF (SameString(FldVal(altBase + ddzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        CALL AddToIntList(altBase + ddzNameOff)
      END IF
    END DO
    IF (numIntList .EQ. 0) THEN
      IF (showMessages) THEN
        CALL WriteError('Did not find any HVACTemplate:Zone:DualDuct objects connected to HVACTemplate:System:DualDuct='//  &
         '"'//trim(FldVal(fldPointMatch))//'". There must be at least one HVACTemplate:Zone:DualDuct object which specifies '// &
         'this system as the Template Dual Duct System Name.')
      END IF
    END IF
  CASE (zlDualDuctColdSupPlen)  !all dual duct zones with cold supply plenum
    DO i = 1, numCompactZoneDualDuct
      altBase = compactZoneDualDuctBase(i)
      IF (SameString(FldVal(altBase + ddzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF (FldVal(altBase + ddzZoneColdSupplyPlenumNameOff) .NE. '') THEN
          CALL AddToIntList(altBase + ddzNameOff)
        END IF
      END IF
    END DO
  CASE (zlDualDuctHotSupPlen)  !all dual duct zones with hot supply plenum
    DO i = 1, numCompactZoneDualDuct
      altBase = compactZoneDualDuctBase(i)
      IF (SameString(FldVal(altBase + ddzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF (FldVal(altBase + ddzZoneHotSupplyPlenumNameOff) .NE. '') THEN
          CALL AddToIntList(altBase + ddzNameOff)
        END IF
      END IF
    END DO
  CASE (zlDualDuctAllRetPlen) !all dual duct zones with return plenum
    DO i = 1, numCompactZoneDualDuct
      altBase = compactZoneDualDuctBase(i)
      IF (SameString(FldVal(altBase + ddzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF (FldVal(altBase + ddzZoneReturnPlenumNameOff) .NE. '') THEN
          CALL AddToIntList(altBase + ddzNameOff)
        END IF
      END IF
    END DO
  CASE (zlVRFAll) ! all VRF zones
    DO i = 1, numCompactZoneVRF
      altBase = compactZoneVRFBase(i)
      IF (SameString(FldVal(altBase + vrfzSystemNameOff),FldVal(fldPointMatch))) THEN
        CALL AddToIntList(altBase + vrfzNameOff)
      END IF
    END DO
    IF (numIntList .EQ. 0) THEN
      IF (showMessages) THEN
        CALL WriteError('Did not find any HVACTemplate:Zone:VRF objects connected to HVACTemplate:System:VRF='//  &
         '"'//trim(FldVal(fldPointMatch))//'". There must be at least one HVACTemplate:Zone:VRF object which specifies '// &
         'this system as the VRF System Name.')
      END IF
    END IF
  CASE (zlVavAll) !all zones for VAV system
    ! Make a combined list of all zone objects associated with this system.
    ! since three different vav and the unitary types of zones can be used need
    ! to scan all of them.
    !loop through the normal VAV zones to find the ones that are associated with this system
    DO i = 1, numCompactZoneVAV
      altBase = compactZoneVAVbase(i)
      IF (SameString(FldVal(altBase + vzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        CALL AddToIntList(altBase + vzNameOff)
      END IF
    END DO
    !loop through the fan power VAV zones to find the ones that are associated with this system
    DO i = 1, numCompactZoneFPVAV
      altBase = compactZoneFPVAVbase(i)
      IF (SameString(FldVal(altBase + fpvzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        CALL AddToIntList(altBase + fpvzNameOff)
      END IF
    END DO
    !loop through the heat and cool VAV zones to find the ones that are associated with this system
    DO i = 1, numCompactZoneHCVAV
      altBase = compactZoneHCVAVbase(i)
      IF (SameString(FldVal(altBase + hcvzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        CALL AddToIntList(altBase + hcvzNameOff)
      END IF
    END DO
    !loop through the unitary zones to find the ones that are associated with this system
    DO i = 1, numCompactZoneUnit
      altBase = compactZoneUnitBase(i)
      IF (SameString(FldVal(altBase + uzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        CALL AddToIntList(altBase + uzNameOff)
      END IF
    END DO
    IF (showMessages) THEN
      IF (numIntList .EQ. 0) THEN
        CALL WriteError('You must specify at least one HVACTemplate:Zone:VAV object for each defined HVACTemplate:System:VAV.'// &
         ' Did not find of match for HVACTemplate:System:VAV='//trim(FldVal(fldPointMatch)))
      END IF
    END IF
  CASE (zlVavAllSupPlen) !all zones for VAV system with supply plenum
    ! Make a combined list of all zone objects associated with this system.
    ! since two different vav and the unitary types of zones can be used need
    ! to scan all of them.
    !loop through the normal VAV zones to find the ones that are associated with this system
    DO i = 1, numCompactZoneVAV
      altBase = compactZoneVAVbase(i)
      IF (SameString(FldVal(altBase + vzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF (FldVal(altBase + vzZoneSupplyPlenumNameOff) .NE. '') THEN
          CALL AddToIntList(altBase + vzNameOff)
        END IF
      END IF
    END DO
    !loop through the fan power VAV zones to find the ones that are associated with this system
    DO i = 1, numCompactZoneFPVAV
      altBase = compactZoneFPVAVbase(i)
      IF (SameString(FldVal(altBase + fpvzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF (FldVal(altBase + fpvzZoneSupplyPlenumNameOff) .NE. '') THEN
          CALL AddToIntList(altBase + fpvzNameOff)
        END IF
      END IF
    END DO
     !loop through the heat and cool VAV zones to find the ones that are associated with this system
    DO i = 1, numCompactZoneHCVAV
      altBase = compactZoneHCVAVbase(i)
      IF (SameString(FldVal(altBase + hcvzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF (FldVal(altBase + hcvzZoneSupplyPlenumNameOff) .NE. '') THEN
          CALL AddToIntList(altBase + hcvzNameOff)
        END IF
      END IF
    END DO
    !loop through the unitary zones to find the ones that are associated with this system
    DO i = 1, numCompactZoneUnit
      altBase = compactZoneUnitBase(i)
      IF (SameString(FldVal(altBase + uzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF (FldVal(altBase + uzZoneSupplyPlenumNameOff) .NE. '') THEN
          CALL AddToIntList(altBase + uzNameOff)
        END IF
      END IF
    END DO
  CASE (zlVavAllRetPlen) !all zones for VAV system with return plenum
    ! Make a combined list of all zone objects associated with this system.
    ! since two different vav and the unitary types of zones can be used need
    ! to scan all of them.
    !loop through the normal VAV zones to find the ones that are associated with this system
    DO i = 1, numCompactZoneVAV
      altBase = compactZoneVAVbase(i)
      IF (SameString(FldVal(altBase + vzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF (FldVal(altBase + vzZoneReturnPlenumNameOff) .NE. '') THEN
          CALL AddToIntList(altBase + vzNameOff)
        END IF
      END IF
    END DO
    !loop through the fan power VAV zones to find the ones that are associated with this system
    DO i = 1, numCompactZoneFPVAV
      altBase = compactZoneFPVAVbase(i)
      IF (SameString(FldVal(altBase + fpvzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF (FldVal(altBase + fpvzZoneReturnPlenumNameOff) .NE. '') THEN
          CALL AddToIntList(altBase + fpvzNameOff)
        END IF
      END IF
    END DO
    !loop through the heat and cool VAV zones to find the ones that are associated with this system
    DO i = 1, numCompactZoneHCVAV
      altBase = compactZoneHCVAVbase(i)
      IF (SameString(FldVal(altBase + hcvzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF (FldVal(altBase + hcvzZoneReturnPlenumNameOff) .NE. '') THEN
          CALL AddToIntList(altBase + hcvzNameOff)
        END IF
      END IF
    END DO
    !loop through the unitary zones to find the ones that are associated with this system
    DO i = 1, numCompactZoneUnit
      altBase = compactZoneUnitBase(i)
      IF (SameString(FldVal(altBase + uzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF (FldVal(altBase + uzZoneReturnPlenumNameOff) .NE. '') THEN
          CALL AddToIntList(altBase + uzNameOff)
        END IF
      END IF
    END DO
  CASE (zlVavFPfromSysRetPlen) !all fan powered zones for VAV system with secondary flow from system return plenum
    ! Make a list of all fan powered vav zone objects associated with this system where the flow type is from plenum
    ! and the zone return plenum name is blank
    DO i = 1, numCompactZoneFPVAV
      altBase = compactZoneFPVAVbase(i)
      IF (SameString(FldVal(altBase + fpvzAirHandlerNameOff),FldVal(fldPointMatch))) THEN
        IF ((FldVal(altBase + fpvzZoneReturnPlenumNameOff) == '') .AND. &
            ((FldVal(altBase + fpvzFlowTypeOff) == 'SeriesFromPlenum') .OR. &
             (FldVal(altBase + fpvzFlowTypeOff) == 'ParallelFromPlenum')) )THEN
          CALL AddToIntList(altBase + fpvzNameOff)
        END IF
      END IF
    END DO
  CASE (zlDOAS)
    DO i = 1, numCompactFanCoil
      altBase = compactFanCoilBase(i)
      IF (SameString(FldVal(altBase + fczDedOutAirSysNameOff),FldVal(fldPointMatch))) THEN
        CALL AddToIntList(altBase + fczNameOff)
      END IF
    END DO
    DO i = 1, numCompactBaseboard
      altBase = compactBaseboardBase(i)
      IF (SameString(FldVal(altBase + bbzDedOutAirSysNameOff),FldVal(fldPointMatch))) THEN
        CALL AddToIntList(altBase + bbzNameOff)
      END IF
    END DO
    DO i = 1, numCompactPTAC
      altBase = compactPTACBase(i)
      IF (SameString(FldVal(altBase + ptaczDedOutAirSysNameOff),FldVal(fldPointMatch))) THEN
        CALL AddToIntList(altBase + ptaczNameOff)
      END IF
    END DO
    DO i = 1, numCompactPTHP
      altBase = compactPTHPBase(i)
      IF (SameString(FldVal(altBase + pthpzDedOutAirSysNameOff),FldVal(fldPointMatch))) THEN
        CALL AddToIntList(altBase + pthpzNameOff)
      END IF
    END DO
    DO i = 1, numCompactWaterAirHP
      altBase = compactWaterAirHPBase(i)
      IF (SameString(FldVal(altBase + wahpDedOutAirSysNameOff),FldVal(fldPointMatch))) THEN
        CALL AddToIntList(altBase + wahpNameOff)
      END IF
    END DO
    DO i = 1, numCompactZoneVRF
      altBase = compactZoneVRFBase(i)
      IF (SameString(FldVal(altBase + vrfzDedOutAirSysNameOff),FldVal(fldPointMatch))) THEN
        CALL AddToIntList(altBase + vrfzNameOff)
      END IF
    END DO
    IF (showMessages) THEN
      IF (numIntList .EQ. 0) THEN
        CALL WriteError('You must specify at least one HVACTemplate:Zone:FanCoil/PTAC/PTHP/VRF/WaterToAirHeatPump served by a '// &
         'HVACTemplate:System:DedicatedOutdoorAir. Did not find of match for '//trim(FldVal(fldPointMatch)))
      END IF
    END IF
END SELECT
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE CreateNewConstVol
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Michael Witte of GARD Analytics, Inc.
          !    DATE WRITTEN   October 2013
          !    MODIFIED
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a constant volume air handling system

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for coil type (heating)
INTEGER,PARAMETER :: ctHotWater = 1
INTEGER,PARAMETER :: ctElectric = 2
INTEGER,PARAMETER :: ctGas      = 3
INTEGER,PARAMETER :: ctNone     = 4
          ! parameters for baseboard type
INTEGER, PARAMETER :: baseboardNone           = 1
INTEGER, PARAMETER :: baseboardHotWater       = 2
INTEGER, PARAMETER :: baseboardElectric       = 3
          ! parameters for coil type (cooling)
INTEGER, PARAMETER :: ccChWater = 1
INTEGER, PARAMETER :: ccChWaterDetailed = 2
INTEGER, PARAMETER :: ccWaterCoilSystemHX = 6
INTEGER, PARAMETER :: ccNone = 8
          ! parameters for cool setpoint control type
INTEGER,PARAMETER :: cspFixed    = 1
INTEGER,PARAMETER :: cspScheduled = 2
INTEGER,PARAMETER :: cspOAreset = 3
INTEGER,PARAMETER :: cspWarmest = 4
INTEGER,PARAMETER :: cspControlZone = 5
          ! parameters for heat setpoint control type
INTEGER,PARAMETER :: hspFixed    = 1
INTEGER,PARAMETER :: hspScheduled = 2
INTEGER,PARAMETER :: hspOAreset = 3
INTEGER,PARAMETER :: hspControlZone = 5
          ! parameters for economizer type
INTEGER,PARAMETER :: etFixedDryBulb                   = 1
INTEGER,PARAMETER :: etFixedEnthalpy                  = 2
INTEGER,PARAMETER :: etDifferentialDryBulb            = 3
INTEGER,PARAMETER :: etDifferentialEnthalpy           = 4
INTEGER,PARAMETER :: etFixedDewPointAndDryBulb        = 5
INTEGER,PARAMETER :: etElectronicEnthalpy             = 6
INTEGER,PARAMETER :: etDifferentialDryBulbAndEnthalpy = 7
INTEGER,PARAMETER :: etNoEconomizer                   = 8
          ! parameters for heat recovery type
INTEGER,PARAMETER :: htrecNone = 1
INTEGER,PARAMETER :: htrecSens = 2
INTEGER,PARAMETER :: htrecEnth = 3
          ! parameters for supply fan type
INTEGER,PARAMETER :: sfpBlowThru = 1
INTEGER,PARAMETER :: sfpDrawThru = 2
          ! parameters for dehumidification type
INTEGER, PARAMETER :: dehumidNone           = 1
INTEGER, PARAMETER :: dehumidCoolRht        = 2
          ! parameters for humidification type
INTEGER, PARAMETER :: humidifyNone      = 1
INTEGER, PARAMETER :: humidifyElecSteam = 2

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for
INTEGER :: sysBase
INTEGER :: iZone
INTEGER :: iSys
INTEGER :: i
LOGICAL :: isTStatNameBlank
LOGICAL :: isSupplyMaxRateAutosize
LOGICAL :: isOATypeDetailed = .FALSE.
LOGICAL :: isZoneRetPlenBlank
LOGICAL :: isZoneSupPlenBlank
LOGICAL :: isSupplyPlenumBlank
LOGICAL :: isReturnPlenumBlank
INTEGER :: reheatCoilType=0
INTEGER :: baseboardKind = 0

LOGICAL :: isSupplyFlowRateAutosize
INTEGER :: heatCoilType=0
INTEGER :: preHeatCoilType=0
INTEGER :: coolCoilKind=0
! REAL    :: minAirRatio
LOGICAL :: isSysAvailSchedBlank
LOGICAL :: isCoolStPtSchedBlank
LOGICAL :: isHeatStPtSchedBlank
LOGICAL :: isPreheatStPtSchedBlank
INTEGER :: coolSetPtCtrlKind=0
INTEGER :: heatSetPtCtrlKind=0
! INTEGER :: supFanCoef=0
INTEGER :: heatRecovery=0
INTEGER :: iInt
INTEGER :: supFanPlacement=0
REAL    :: htRecSens75
REAL    :: htRecLat75
LOGICAL :: isDehumidifyNone
INTEGER :: dehumidCtrlKind = 0
LOGICAL :: isHumidifierNone
INTEGER :: humidifierKind = 0
LOGICAL :: isCoolCoilChW
LOGICAL :: isHeatCoilCapAutosize
LOGICAL :: usePreheatTempforHeatDesignSuppT = .FALSE.
LOGICAL :: isReturnFanYes = .FALSE.
INTEGER :: coolCtrlZoneBase = 0
INTEGER :: heatCtrlZoneBase = 0
LOGICAL :: isEconoLowLimitBlank
INTEGER :: econoKind=0

! Node Names
CHARACTER(len=MaxAlphaLength) :: lastOutlet=''
CHARACTER(len=MaxAlphaLength) :: fanInlet=''
CHARACTER(len=MaxAlphaLength) :: fanOutlet=''
CHARACTER(len=MaxAlphaLength) :: coolCoilUnitInlet=''
CHARACTER(len=MaxAlphaLength) :: coolCoilUnitOutlet=''
CHARACTER(len=MaxAlphaLength) :: heatCoilInlet=''
CHARACTER(len=MaxAlphaLength) :: heatCoilOutlet=''
CHARACTER(len=MaxAlphaLength) :: reheatCoilInlet=''
CHARACTER(len=MaxAlphaLength) :: reheatCoilOutlet=''
CHARACTER(len=MaxAlphaLength) :: humidifierInlet=''
CHARACTER(len=MaxAlphaLength) :: humidifierOutlet=''
CHARACTER(len=MaxAlphaLength) :: airloopOutlet=''
CHARACTER(len=MaxAlphaLength) :: dxCoilObjectType='None'
CHARACTER(len=MaxAlphaLength) :: scheduleName=''
CHARACTER(len=MaxAlphaLength) :: returnInletToOAMIxer=''
CHARACTER(len=MaxAlphaLength) :: dehumscheduleName=''
CHARACTER(len=MaxAlphaLength) :: humscheduleName=''

!======================================================================================================
! create Zones
!======================================================================================================
DO iZone = 1, numCompactZoneConstVol
  base =  compactZoneConstVolbase(iZone)
  ! find base for system this zone belongs to
  sysBase = -1
  DO iSys = 1, numCompactSysConstVol
    IF (SameString(FldVal(compactSysConstVolbase(iSys) + cvsAirHandlerNameOff), FldVal(base +  cvzAirHandlerNameOff))) THEN
      sysBase = compactSysConstVolBase(iSys)
      EXIT
    END IF
  END DO
  IF (sysBase .EQ. -1) THEN
    CALL WriteError('Could not find air handler name referenced in HVACTemplate:Zone:ConstantVolume object: '// &
                    TRIM(FldVal(base + cvzNameOff)))
    CYCLE
  END IF
  ! set flags
  isTStatNameBlank = (FldVal(base + cvzTStatNameOff) .EQ. '')
  isSupplyMaxRateAutosize = SameString(FldVal(base + cvzSupplyMaxRateOff),'autosize')
  isOATypeDetailed = SameString(FldVal(base + cvzOutAirTypeOff),'DetailedSpecification')
  !set reheat coil indicator
  IF (SameString(FldVal(base + cvzReheatTypeOff),'HotWater')) THEN
    reheatCoilType = ctHotWater
  ELSEIF (SameString(FldVal(base + cvzReheatTypeOff),'Electric')) THEN
    reheatCoilType = ctElectric
  ELSEIF (SameString(FldVal(base + cvzReheatTypeOff),'Gas')) THEN
    reheatCoilType = ctGas
  ELSEIF (SameString(FldVal(base + cvzReheatTypeOff),'None')) THEN
    reheatCoilType = ctNone
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:ConstantVolume in the Reheat Coil Type field')
  END IF
  isZoneRetPlenBlank = (FldVal(base + cvzZoneReturnPlenumNameOff) .EQ. '')
  isZoneSupPlenBlank = (FldVal(base + cvzZoneSupplyPlenumNameOff) .EQ. '')
  !set the baseboard indicator
  IF (SameString(FldVal(base +  cvzBaseboardTypeOff),'None')) THEN
    baseboardKind = baseboardNone
  ELSEIF (SameString(FldVal(base +  cvzBaseboardTypeOff),'HotWater')) THEN
    baseboardKind = baseboardHotWater
  ELSEIF (SameString(FldVal(base +  cvzBaseboardTypeOff),'Electric')) THEN
    baseboardKind = baseboardElectric
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:ConstantVolume "'//TRIM(FldVal(base + cvzNameOff))//'"'// &
                    ' in the Baseboard Heating Type field: '//TRIM(FldVal(base + cvzBaseboardTypeOff)))
  END IF
  !validate cooling design method
  IF (SameString(FldVal(base +  cvzCoolDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  cvzCoolDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  cvzCoolDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:ConstantVolume "'//TRIM(FldVal(base + cvzNameOff))//'"'// &
                    ' in the Zone Cooling Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + cvzCoolDesignTempMethodOff)))
  END IF
  !validate heating design method
  IF (SameString(FldVal(base +  cvzHeatDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  cvzHeatDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:ConstantVolume "'//TRIM(FldVal(base + cvzNameOff))//'"'// &
                    ' in the Zone Heating Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + cvzCoolDesignTempMethodOff)))
  END IF
  !ZONE CONTROL:THERMOSTATIC ~ line 5
  !repeat per zone - only if cvzTStatNameOff is not blank
  IF (.NOT. isTStatNameBlank) THEN
    CALL AddAlwaysSchedule('4')
    CALL CreateNewObj('ZoneControl:Thermostat')
    CALL AddToObjFld('Name', base +  cvzNameOff,' Thermostat')
    CALL AddToObjFld('Zone or ZoneList Name', base +  cvzNameOff,' ')
    !this schedule was previously created
    CALL AddToObjStr('Control Type Schedule Name','HVACTemplate-Always 4')
    CALL AddToObjStr('Control Object Type','ThermostatSetpoint:DualSetpoint')
    CALL AddToObjFld('Control Name', base + cvzTStatNameOff,' Dual SP Control',.TRUE.)
  END IF
  !ZONE SIZING   ~ line 12
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:OutdoorAir')
    CALL AddToObjStr('Name', 'SZ DSOA '//FldVal(base+cvzNameOff))
    CALL AddToObjFld('Outdoor Air Method', base + cvzOutAirTypeOff,' ')
    CALL AddToObjFld('Outdoor Air Flow per Person {m3/s}', base + cvzOutAirFlowPersonOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone Floor Area {m3/s-m2}', base + cvzOutAirFlowAreaOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone {m3/s}', base + cvzOutAirFlowZoneOff,'',.true.)
  END IF

  CALL CreateNewObj('Sizing:Zone')
  CALL AddToObjFld('Zone or ZoneList Name', base +  cvzNameOff,'')
  IF (SameString(FldVal(base +  cvzCoolDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
    CALL AddToObjStr('Zone Cooling Design Supply Air Temperature Input Method','SupplyAirTemperature')
    CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', sysBase + cvsCoolSetPtDesignOff,'')
  ELSE
    CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Input Method',base + cvzCoolDesignTempMethodOff,'')
    CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', base + cvzCoolDesignTempOff,'')
  ENDIF
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Difference {delta C}',base + cvzCoolDesignTempDiffOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Input Method',base + cvzHeatDesignTempMethodOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', base + cvzHeatDesignTempOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Difference {delta C}',base + cvzHeatDesignTempDiffOff,'')
  CALL AddToObjStr('Zone Cooling Design Supply Air Humidity Ratio {kg-H20/kg-air}','0.008')
  CALL AddToObjStr('Zone Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + cvzDesignSpecOANameOff,'')
  ELSE
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','SZ DSOA '//FldVal(base+cvzNameOff))
  END IF
  CALL AddToObjFld('Zone Heating Sizing Factor', base + cvzZoneHeatSizingFactorOff,'')
  CALL AddToObjFld('Zone Cooling Sizing Factor', base + cvzZoneCoolSizingFactorOff,'')
  IF (isSupplyMaxRateAutosize) THEN
    CALL AddToObjStr('Cooling Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Cooling Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  ELSE
    CALL AddToObjStr('Cooling Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Cooling Design Air Flow Rate {m3/s}', base + cvzSupplyMaxRateOff,' ')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  END IF
  IF (isSupplyMaxRateAutosize) THEN
    CALL AddToObjStr('Heating Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Heating Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
  ELSE
    CALL AddToObjStr('Heating Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Heating Design Air Flow Rate {m3/s}', base + cvzSupplyMaxRateOff,'')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
  END IF
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Zone Air Distribution Object Name', base + cvzDesSpecOADistNameOff,'',.true.)
  ELSE
    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+cvzNameOff),.true.)
  END IF
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:ZoneAirDistribution')
    CALL AddToObjStr('Name','SZ DSZAD '//FldVal(base+cvzNameOff))
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Cooling Mode {}','1')
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Heating Mode {}','1',.true.)
  END IF

  !CONTROLLED ZONE EQUIP CONFIGURATION  ~ line 41
  CALL CreateNewObj('ZoneHVAC:EquipmentConnections')
  CALL AddToObjFld('Zone Name', base +  cvzNameOff,'')
  CALL AddToObjFld('Zone Conditioning Equipment List Name', base +  cvzNameOff,' Equipment')
  IF ((reheatCoilType .EQ. ctNone) .AND. (isZoneSupPlenBlank)) THEN
    CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base +  cvzNameOff,' Zone Equip Inlet')
  ELSE
    CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base +  cvzNameOff,' Supply Inlet')
  END IF
  CALL AddToObjStr('Zone Air Exhaust Node or NodeList Name','')
  CALL AddToObjFld('Zone Air Node Name', base +  cvzNameOff,' Zone Air Node')
  IF (isZoneRetPlenBlank) THEN
    CALL AddToObjFld('Zone Return Air Node Name', base +  cvzNameOff,' Return Outlet',.TRUE.)
  ELSE
    CALL AddToObjFld('Zone Return Air Node Name', base + cvzNameOff,' Zone Return Plenum Inlet',.TRUE.)
  END IF
  !ZONE SUPPLY PLENUM ~line 50
  IF (.NOT. isZoneSupPlenBlank) THEN
    CALL CreateNewObj('AirLoopHVAC:SupplyPlenum')
    CALL AddToObjFld('Name', base + cvzNameOff,' Supply Plenum')
    CALL AddToObjFld('Zone Name', base + cvzZoneSupplyPlenumNameOff,'')
    CALL AddToObjFld('Zone Node Name', base + cvzZoneSupplyPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Inlet Node Name', base + cvzNameOff,' Zone Equip Inlet')
    IF (reheatCoilType .EQ. ctNone) THEN
      CALL AddToObjFld('Outlet Node Name', base + uzNameOff,' Supply Inlet',.TRUE.)
    ELSE
      CALL AddToObjFld('Outlet Node Name', base + cvzNameOff,' Terminal Unit Inlet',.TRUE.)
    END IF
  END IF
  !ZONE RETURN PLENUM ~ line 57
  IF (.NOT. isZoneRetPlenBlank) THEN
    CALL CreateNewObj('AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Name', base + cvzNameOff,' Return Plenum')
    CALL AddToObjFld('Zone Name', base + cvzZoneReturnPlenumNameOff,'')
    CALL AddToObjFld('Zone Node Name', base + cvzZoneReturnPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Outlet Node Name', base + cvzNameOff,' Return Outlet')
    CALL AddToObjStr('Induced Air Outlet Node or NodeList Name','')
    CALL AddToObjFld('Inlet Node Name', base + cvzNameOff,' Zone Return Plenum Inlet',.TRUE.)
  END IF
  !ZONE EQUIPMENT LIST ~ line 64
  CALL CreateNewObj('ZoneHVAC:EquipmentList')
  CALL AddToObjFld('Name', base +  cvzNameOff,' Equipment')
  CALL AddToObjStr('Load Distribution Scheme','SequentialLoad')
  IF (reheatCoilType .EQ. ctNone) THEN
    CALL AddToObjStr('Zone Equipment Object Type','AirTerminal:SingleDuct:Uncontrolled')
    CALL AddToObjFld('Zone Equipment Name', base +  cvzNameOff,' Air Terminal')
  ELSE
    CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Zone Equipment Name', base +  cvzNameOff,' ATU')
  END IF
  CALL AddToObjStr('Zone Equipment Cooling Sequence','1')
  SELECT CASE (baseboardKind)
    CASE (baseboardNone)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1',.FALSE.)
    CASE (baseboardHotWater)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1')
      CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
      CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.FALSE.)
      CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Zone Equipment Name', base + cvzNameOff,' Baseboard Heat')
      CALL AddToObjStr('Zone Equipment Cooling Sequence','2')
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','2',.FALSE.)
    CASE (baseboardElectric)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1')
      CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Zone Equipment Name', base + cvzNameOff,' Baseboard Heat')
      CALL AddToObjStr('Zone Equipment Cooling Sequence','2')
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','2',.FALSE.)
  END SELECT
  CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.TRUE.)
  !AIR DISTRIBUTION UNIT ~ line 72
  IF (reheatCoilType .EQ. ctNone) THEN
    CALL CreateNewObj('AirTerminal:SingleDuct:Uncontrolled')
    CALL AddToObjFld('Name', base +  cvzNameOff,' Air Terminal')
    CALL AddToObjStr('Availability Schedule Name','')
    IF (isZoneSupPlenBlank) THEN
      CALL AddToObjFld('Zone Supply Air Node Name', base + cvzNameOff,' Zone Equip Inlet')
    ELSE
      CALL AddToObjFld('Zone Supply Air Node Name', base + cvzNameOff,' Supply Inlet')
    END IF
    CALL AddToObjStr('Maximum air flow rate {m3/s}', 'autosize',.TRUE.)
  ELSE
    CALL CreateNewObj('ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Name', base +  cvzNameOff,' ATU')
    CALL AddToObjFld('Air Distribution Unit Outlet Node Name', base +  cvzNameOff,' Supply Inlet')
    CALL AddToObjStr('Air Terminal Object Type','AirTerminal:SingleDuct:ConstantVolume:Reheat')
    CALL AddToObjFld('Air Terminal Name', base +  cvzNameOff,' CV Reheat',.TRUE.)
  END IF
  !This object if cvzReheatTypeOff=Hot Water
  IF (.NOT. reheatCoilType .EQ. ctNone) THEN
    !AirTerminal:SingleDuct:ConstantVolume:Reheat
    CALL CreateNewObj('AirTerminal:SingleDuct:ConstantVolume:Reheat')
    CALL AddToObjFld('Name', base +  cvzNameOff,' CV Reheat')
    !this schedule was previously created
    CALL AddToObjStr('Availability Schedule Name','')
    CALL AddToObjFld('Air Outlet Node Name', base + cvzNameOff,' Supply Inlet')
    IF (isZoneSupPlenBlank) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + cvzNameOff,' Zone Equip Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + cvzNameOff,' Terminal Unit Inlet')
    END IF
    CALL AddToObjStr('Maximum air flow rate {m3/s} ', 'autosize')
    !CR8001
    IF (reheatCoilType .EQ. ctHotWater) THEN
      CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Water')
      CALL AddToObjFld('Reheat Coil Name', base + cvzNameOff,' Reheat Coil')
      CALL AddToObjStr('Maximum Hot Water or Steam Flow Rate {m3/s}  ','autosize')
      CALL AddToObjStr('Minimum Hot Water or Steam Flow Rate {m3/s}  ','0')
      CALL AddToObjStr('Convergence Tolerance','0.001')
      CALL AddToObjFld('Maximum Reheat Air Temperature', base + cvzMaxReheatAirTempOff,' ',.TRUE.)
      !COIL:Water:SimpleHeating  ~ line 104
      CALL CreateNewObj('Coil:Heating:Water')
      CALL AddToObjFld('Name', base + cvzNameOff,' Reheat Coil')
      CALL AddToObjFld('Availability Schedule Name', base + cvzReheatSchedNameOff,' ')
      CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
      CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Water Inlet Node Name', base + cvzNameOff,' Reheat Coil HW Inlet')
      CALL AddToObjFld('Water Outlet Node Name', base + cvzNameOff,' Reheat Coil HW Outlet')
      IF (isZoneSupPlenBlank) THEN
        CALL AddToObjFld('Air Inlet Node Name', base + cvzNameOff,' Zone Equip Inlet')
      ELSE
        CALL AddToObjFld('Air Inlet Node Name', base + cvzNameOff,' Terminal Unit Inlet')
      END IF
      CALL AddToObjFld('Air Outlet Node Name', base + cvzNameOff,' Supply Inlet')
      CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
      CALL AddToObjStr('Rated Capacity','autosize')
      CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
      CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
      CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
      CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
      CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
      !BRANCH ~ line 113
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + cvzNameOff,' Reheat Coil HW Branch')
      !save this branch name for later use in plant
      CALL AddToStrList(TRIM(FldVal(base + cvzNameOff)) // ' Reheat Coil HW Branch',handleHeatingCoilBranch)
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
      CALL AddToObjFld('Component Name', base + cvzNameOff,' Reheat Coil')
      CALL AddToObjFld('Component Inlet Node Name', base + cvzNameOff,' Reheat Coil HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + cvzNameOff,' Reheat Coil HW Outlet',.TRUE.)
    ELSEIF (reheatCoilType .EQ. ctElectric) THEN
      CALL AddToObjStr('Hot Water or Steam Inlet Node Name', '')
      CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Electric')
      CALL AddToObjFld('Reheat Coil Name', base + cvzNameOff,' Reheat Coil')
      CALL AddToObjStr('Maximum Hot Water or Steam Flow Rate {m3/s}','')
      CALL AddToObjStr('Minimum Hot Water or Steam Flow Rate {m3/s}','')
      CALL AddToObjStr('Convergence Tolerance','0.001')
      CALL AddToObjFld('Maximum Reheat Air Temperature', base + cvzMaxReheatAirTempOff,' ',.TRUE.)
      !COIL:Electric:Heating ~ line 139
      CALL CreateNewObj('Coil:Heating:Electric')
      CALL AddToObjFld('Name', base + cvzNameOff,' Reheat Coil')
      CALL AddToObjFld('Availability Schedule Name', base + cvzReheatSchedNameOff,' ')
      CALL AddToObjStr('Efficiency','1')
      CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
      IF (isZoneSupPlenBlank) THEN
        CALL AddToObjFld('Air Inlet Node Name', base + cvzNameOff,' Zone Equip Inlet')
      ELSE
        CALL AddToObjFld('Air Inlet Node Name', base + cvzNameOff,' Terminal Unit Inlet')
      END IF
      CALL AddToObjFld('Air Outlet Node Name', base + cvzNameOff,' Supply Inlet')
      CALL AddToObjStr('Temperature Setpoint Node Name','',.TRUE.)
    ELSEIF (reheatCoilType .EQ. ctGas) THEN
      CALL AddToObjStr('Hot Water or Steam Inlet Node Name', '')
      CALL AddToObjStr('Reheat Coil Object Type','Coil:Heating:Fuel')
      CALL AddToObjFld('Reheat Coil Name', base + cvzNameOff,' Reheat Coil')
      CALL AddToObjStr('Maximum Hot Water or Steam Flow Rate {m3/s}','')
      CALL AddToObjStr('Minimum Hot Water or Steam Flow Rate {m3/s}','')
      CALL AddToObjStr('Convergence Tolerance','0.001')
      CALL AddToObjFld('Maximum Reheat Air Temperature', base + cvzMaxReheatAirTempOff,' ',.TRUE.)
      !Coil:Heating:Fuel
      CALL CreateNewObj('Coil:Heating:Fuel')
      CALL AddToObjFld('Name', base + cvzNameOff,' Reheat Coil')
      CALL AddToObjFld('Availability Schedule Name', base + cvzReheatSchedNameOff,' ')
      CALL AddToObjStr('Fuel Type', 'Gas')
      CALL AddToObjStr('Gas Burner Efficiency','0.8')
      CALL AddToObjStr('Nominal Capacity {W}', 'autosize')
      IF (isZoneSupPlenBlank) THEN
        CALL AddToObjFld('Air Inlet Node Name', base + cvzNameOff,' Zone Equip Inlet')
      ELSE
        CALL AddToObjFld('Air Inlet Node Name', base + cvzNameOff,' Terminal Unit Inlet')
      END IF
      CALL AddToObjFld('Air Outlet Node Name', base + cvzNameOff,' Supply Inlet')
      CALL AddToObjStr('Temperature Setpoint Node Name','')
      CALL AddToObjStr('Parasitic Electric Load {W}', '0')
      CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
         base + cvzNameOff,' Reheat Coil PLF-FPLR')
      CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
      !CURVE:CUBIC
      CALL CreateNewObj('Curve:Cubic')
      CALL AddToObjFld('Name', base + cvzNameOff,' Reheat Coil PLF-FPLR')
      CALL AddToObjStr('Coefficient1 Constant','0.8')
      CALL AddToObjStr('Coefficient2 x','0.2')
      CALL AddToObjStr('Coefficient3 x**2','0')
      CALL AddToObjStr('Coefficient4 x**3','0')
      CALL AddToObjStr('Minimum Value of x','0')
      CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
    END IF
  END IF
  !Baseboards
  SELECT CASE (baseboardKind)
    CASE (baseboardHotWater)
      !ZoneHVAC:Baseboard:RadiantConvective:Water
      CALL CreateNewObj('ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Name', base + cvzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + cvzBaseboardAvailSchedNameOff,'')
      CALL AddToObjFld('Inlet Node Name', base + cvzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Outlet Node Name', base + cvzNameOff,' Baseboard HW Outlet')
      CALL AddToObjStr('Rated Average Water Temperature {C}','87.78')
      CALL AddToObjStr('Rated Water Mass Flow Rate {m3/s}','0.063')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + cvzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Maximum Water Flow Rate','autosize')
      CALL AddToObjStr('Convergence Tolerance','0.001')
      CALL AddToObjStr('Fraction Radiant','0.0')
      CALL AddToObjStr('Fraction of Radiant Energy Incident on People','0.0',.TRUE.)
      !BRANCH (for baseboard)
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + cvzNameOff,' Baseboard Heat HW Branch')
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Component Name', base + cvzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Component Inlet Node Name', base + cvzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + cvzNameOff,' Baseboard HW Outlet',.TRUE.)
      CALL AddToStrList(TRIM(FldVal(base + cvzNameOff)) // ' Baseboard Heat HW Branch',handleHeatingCoilBranch)
      IF ((compactHotLoopBase .EQ. 0) .OR. (numCompactBoiler .EQ. 0)) THEN
        CALL WriteError('Both a HVACTemplate:Plant:HotWaterLoop and a HVACTemplate:Plant:Boiler are needed ' // &
             'when using hot water baseboards.')
      END IF
    CASE (baseboardElectric)
      !BASEBOARD HEATER:Electric:Convective
      CALL CreateNewObj('ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Name', base + cvzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + cvzBaseboardAvailSchedNameOff,'')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + cvzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Efficiency','1',.TRUE.)
  END SELECT
END DO

!======================================================================================================
! create Systems
!======================================================================================================
DO iSys = 1, numCompactSysConstVol
  base =  compactSysConstVolBase(iSys)
  ! set flags
  isSupplyFlowRateAutosize = SameString(FldVal(base + cvsSupplyRateOff),'autosize')
  isSysAvailSchedBlank = (FldVal(base + cvsSysAvailSchedNameOff) .EQ. '')
  isSupplyPlenumBlank = (FldVal(base + cvsSupplyPlenumNameOff) .EQ. '')
  isReturnPlenumBlank = (FldVal(base + cvsReturnPlenumNameOff) .EQ. '')
  isCoolStPtSchedBlank = (FldVal(base + cvsCoolSetPtSchedNameOff) .EQ. '')
  isHeatStPtSchedBlank = (FldVal(base + cvsHeatSetPtSchedNameOff) .EQ. '')
  isPreheatStPtSchedBlank = (FldVal(base + cvsPreheatSetPtSchedNameOff) .EQ. '')
  isEconoLowLimitBlank = (FldVal(base + cvsEconoLowLimitOff) .EQ. '')
  isHeatCoilCapAutosize = SameString(FldVal(base + cvsHeatCoilCapOff),'autosize')
  isReturnFanYes = (FldVal(base + cvsReturnFanPresentOff) .EQ. 'Yes')
  !check cooling coil kind
  IF (SameString(FldVal(base +  cvsCoolCoilKindOff),'ChilledWater')) THEN
    coolCoilKind  = ccChWater
    isCoolCoilChW = .TRUE.
  ELSEIF (SameString(FldVal(base + cvsCoolCoilKindOff),'ChilledWaterDetailedFlatModel')) THEN
    coolCoilKind = ccChWaterDetailed
    isCoolCoilChW = .TRUE.
  ELSEIF (SameString(FldVal(base + cvsCoolCoilKindOff),'HeatExchangerAssistedChilledWater')) THEN
    coolCoilKind = ccWaterCoilSystemHX
    isCoolCoilChW = .TRUE.
  ELSEIF (SameString(FldVal(base + cvsCoolCoilKindOff),'None')) THEN
    coolCoilKind = ccNone
    isCoolCoilChW = .FALSE.
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:ConstantVolume  "'//TRIM(FldVal(base + cvsAirHandlerNameOff))//'"'// &
                    ' in the Cooling Coil Type field: '//TRIM(FldVal(base + cvsCoolCoilKindOff)))
  END IF
  !check cooling coil setpoint control type
  IF (SameString(FldVal(base +  cvsCoolSetPtCtrlTypeOff),'FixedSetpoint')) THEN
    coolSetPtCtrlKind = cspFixed
  ELSEIF (SameString(FldVal(base + cvsCoolSetPtCtrlTypeOff),'Scheduled')) THEN
    coolSetPtCtrlKind = cspScheduled
  ELSEIF (SameString(FldVal(base + cvsCoolSetPtCtrlTypeOff),'OutdoorAirTemperatureReset')) THEN
    coolSetPtCtrlKind = cspOAreset
  ELSEIF (SameString(FldVal(base + cvsCoolSetPtCtrlTypeOff),'Warmest')) THEN
    coolSetPtCtrlKind = cspWarmest
  ELSEIF (SameString(FldVal(base + cvsCoolSetPtCtrlTypeOff),'ControlZone')) THEN
    coolSetPtCtrlKind = cspControlZone
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:ConstantVolume  "'//TRIM(FldVal(base + cvsAirHandlerNameOff))//'"'// &
                    ' in the Cooling Coil Setpoint Control Type field: '//TRIM(FldVal(base + cvsCoolSetPtCtrlTypeOff)))
  END IF
  !set heating coil indicator
  IF (SameString(FldVal(base + cvsHeatCoilKindOff),'HotWater')) THEN
    heatCoilType = ctHotWater
  ELSEIF (SameString(FldVal(base + cvsHeatCoilKindOff),'Electric')) THEN
    heatCoilType = ctElectric
  ELSEIF (SameString(FldVal(base + cvsHeatCoilKindOff),'Gas')) THEN
    heatCoilType = ctGas
  ELSEIF (SameString(FldVal(base + cvsHeatCoilKindOff),'None')) THEN
    heatCoilType = ctNone
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:ConstantVolume  "'//TRIM(FldVal(base + cvsAirHandlerNameOff))//'"'// &
                    ' in the Heating Coil Type field: '//TRIM(FldVal(base + cvsHeatCoilKindOff)))
  END IF
  !check heating coil setpoint control type
  IF (SameString(FldVal(base +  cvsHeatSetPtCtrlTypeOff),'FixedSetpoint')) THEN
    heatSetPtCtrlKind = hspFixed
  ELSEIF (SameString(FldVal(base + cvsHeatSetPtCtrlTypeOff),'Scheduled')) THEN
    heatSetPtCtrlKind = hspScheduled
  ELSEIF (SameString(FldVal(base + cvsHeatSetPtCtrlTypeOff),'OutdoorAirTemperatureReset')) THEN
    heatSetPtCtrlKind = hspOAreset
  ELSEIF (SameString(FldVal(base + cvsHeatSetPtCtrlTypeOff),'ControlZone')) THEN
    heatSetPtCtrlKind = hspControlZone
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:ConstantVolume  "'//TRIM(FldVal(base + cvsAirHandlerNameOff))//'"'// &
                    ' in the Heating Coil Setpoint Control Type field: '//TRIM(FldVal(base + cvsHeatSetPtCtrlTypeOff)))
  END IF
  !set preheating coil indicator
  IF (SameString(FldVal(base +  cvsPreheatCoilKindOff),'HotWater')) THEN
    preHeatCoilType = ctHotWater
  ELSEIF (SameString(FldVal(base +  cvsPreheatCoilKindOff),'Electric')) THEN
    preHeatCoilType = ctElectric
  ELSEIF (SameString(FldVal(base +  cvsPreheatCoilKindOff),'Gas')) THEN
    preHeatCoilType = ctGas
  ELSEIF (SameString(FldVal(base +  cvsPreheatCoilKindOff),'None')) THEN
    preHeatCoilType = ctNone
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:ConstantVolume  "'//TRIM(FldVal(base + cvsAirHandlerNameOff))//'"'// &
                    ' in the Preheat Coil Type field: '//TRIM(FldVal(base + cvsPreheatCoilKindOff)))
  END IF
  !set the economizer indicator
  IF (SameString(FldVal(base +  cvsEconomizerKindOff),'FixedDryBulb')) THEN
    econoKind = etFixedDryBulb
  ELSEIF (SameString(FldVal(base +  cvsEconomizerKindOff),'FixedEnthalpy')) THEN
    econoKind = etFixedEnthalpy
  ELSEIF (SameString(FldVal(base +  cvsEconomizerKindOff),'DifferentialDryBulb')) THEN
    econoKind = etDifferentialDryBulb
  ELSEIF (SameString(FldVal(base +  cvsEconomizerKindOff),'DifferentialEnthalpy')) THEN
    econoKind = etDifferentialEnthalpy
  ELSEIF (SameString(FldVal(base +  cvsEconomizerKindOff),'FixedDewPointAndDryBulb')) THEN
    econoKind = etFixedDewPointAndDryBulb
  ELSEIF (SameString(FldVal(base +  cvsEconomizerKindOff),'ElectronicEnthalpy')) THEN
    econoKind = etElectronicEnthalpy
  ELSEIF (SameString(FldVal(base +  cvsEconomizerKindOff),'DifferentialDryBulbAndEnthalpy')) THEN
    econoKind = etDifferentialDryBulbAndEnthalpy
  ELSEIF (SameString(FldVal(base +  cvsEconomizerKindOff),'NoEconomizer')) THEN
    econoKind = etNoEconomizer
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:ConstantVolume in the Economizer Type field')
  END IF
  !set the Heat Recovery Type indicator
  IF (SameString(FldVal(base + cvsHeatRecTypeOff),'Sensible')) THEN
    heatRecovery = htrecSens
  ELSEIF (SameString(FldVal(base + cvsHeatRecTypeOff),'Enthalpy')) THEN
    heatRecovery = htrecEnth
  ELSEIF (SameString(FldVal(base + cvsHeatRecTypeOff),'None')) THEN
    heatRecovery = htrecNone
  ELSE
    heatRecovery = htrecNone
    CALL WriteError('Invalid choice in HVACTemplate:System:ConstantVolume in the Heating Coil Setpoint Reset Type field')
  ENDIF
  IF (SameString(FldVal(base + cvsSupplyFanPlaceOff), 'BlowThrough')) THEN
    supFanPlacement = sfpBlowThru
  ELSEIF (SameString(FldVal(base + cvsSupplyFanPlaceOff), 'DrawThrough')) THEN
    supFanPlacement = sfpDrawThru
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:ConstantVolume "'//TRIM(FldVal(base + cvsAirHandlerNameOff))//'"'// &
                    ' in the Supply Fan Placement field: '//TRIM(FldVal(base + cvsSupplyFanPlaceOff)))
  END IF
  !set the dehumidification indicators
  isDehumidifyNone = SameString(FldVal(base + cvsDehumCtrlTypeOff),'None')
  IF (SameString(FldVal(base +  cvsDehumCtrlTypeOff),'None')) THEN
    dehumidCtrlKind = dehumidNone
  ELSEIF (SameString(FldVal(base +  cvsDehumCtrlTypeOff),'CoolReheat')) THEN
    dehumidCtrlKind = dehumidCoolRht
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:ConstantVolume "'//TRIM(FldVal(base + cvsAirHandlerNameOff))//'"'// &
                    ' in the Dehumidification Control Type field: '//TRIM(FldVal(base + cvsDehumCtrlTypeOff)))
  END IF

  !set the humidifier indicators
  isHumidifierNone = SameString(FldVal(base + cvsHumidCtrlTypeOff),'None')
  IF (SameString(FldVal(base +  cvsHumidCtrlTypeOff),'None')) THEN
    humidifierKind = humidifyNone
  ELSEIF (SameString(FldVal(base +  cvsHumidCtrlTypeOff),'ElectricSteam')) THEN
    humidifierKind = humidifyElecSteam
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:ConstantVolume "'//TRIM(FldVal(base + cvsAirHandlerNameOff))//'"'// &
                    ' in the Humidifier Type field: '//TRIM(FldVal(base + cvsHumidCtrlTypeOff)))
  END IF

  !set component inlet and outlet node names, including system name from cvsAirHandlerNameOff as prefix
  ! First component will always be AirLoopHVAC:OutdoorAirSystem with outlet node name <cvsAirHandlerNameOff> Mixed Air Outlet
  lastOutlet = TRIM(FldVal(base + cvsAirHandlerNameOff))//' Mixed Air Outlet'

  !set node name for return inlet to OA mixer
  IF (isReturnFanYes) THEN
    returnInletToOAMIxer = TRIM(FldVal(base + cvsAirHandlerNameOff))//' Return Fan Outlet'
  ELSE
    returnInletToOAMIxer = TRIM(FldVal(base + cvsAirHandlerNameOff))//' Air Loop Inlet'
  END IF

  ! humidifier
  IF (.NOT. isHumidifierNone) THEN
    humidifierInlet  = TRIM(lastOutlet)
    humidifierOutlet = TRIM(FldVal(base + cvsAirHandlerNameOff))//' Humidifier Outlet'
    lastOutlet       = TRIM(humidifierOutlet)
  END IF

  ! blow thru fan
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    fanInlet         = TRIM(lastOutlet)
    fanOutlet        = TRIM(FldVal(base + cvsAirHandlerNameOff))//' Supply Fan Outlet'
    lastOutlet       = TRIM(fanOutlet)
  END IF

  ! cooling coil
  IF (coolCoilKind .NE. ccNone) THEN
    coolCoilUnitInlet = TRIM(lastOutlet)
    IF (coolCoilKind .EQ. ccWaterCoilSystemHX) THEN
      ! cooling coil with heat exchanger assist
      coolCoilUnitOutlet = TRIM(FldVal(base + cvsAirHandlerNameOff))//' Cooling Coil HX Unit Outlet'
    ELSE
      ! simple cooling coil (without heat exchanger assist)
      coolCoilUnitOutlet = TRIM(FldVal(base + cvsAirHandlerNameOff))//' Cooling Coil Outlet'
    END IF
    lastOutlet       = TRIM(coolCoilUnitOutlet)
  END IF

  ! heating coil
  IF (heatCoilType .NE. ctNone) THEN
    heatCoilInlet    = TRIM(lastOutlet)
    heatCoilOutlet   = TRIM(FldVal(base + cvsAirHandlerNameOff))//' Heating Coil Outlet'
    lastOutlet       = TRIM(heatCoilOutlet)
  END IF

  ! draw thru fan
  IF (supFanPlacement .EQ. sfpDrawThru) THEN
    fanInlet         = TRIM(lastOutlet)
    fanOutlet        = TRIM(FldVal(base + cvsAirHandlerNameOff))//' Supply Fan Outlet'
    lastOutlet       = TRIM(fanOutlet)
  END IF

  ! set AirLoopHVAC Supply Side Outlet Node Name
  airloopOutlet      = TRIM(lastOutlet)

  !Fix for CR7356
  ! Warn if fixed heating or preheat setpoint is greater than fixed cooling setpoint, and no reset option is active
  IF (isCoolStPtSchedBlank .AND. (coolSetPtCtrlKind == cspFixed)) THEN
    IF ((heatCoilType .NE. ctNone) .AND. isHeatStPtSchedBlank .AND. (heatSetPtCtrlKind == hspFixed)) THEN
      IF (StringToReal(FldVal(base + cvsHeatSetPtDesignOff)) .GT. StringToReal(FldVal(base + cvsCoolSetPtDesignOff))) THEN
       CALL WriteError('Warning:  In HVACTemplate:System:ConstantVolume "'//TRIM(FldVal(base + cvsAirHandlerNameOff))//'"'// &
                      ' the Heating Coil Design Setpoint is greater than the Cooling Coil Design Setpoint.'// &
                      ' This may cause the heating coil and cooling coil to operate simultaneously.'// &
                      ' Check results carefully and adjust controls if needed.',msgWarning)
      ENDIF
    ENDIF
    IF ((preHeatCoilType .NE. ctNone) .AND. isPreheatStPtSchedBlank) THEN
      IF (StringToReal(FldVal(base + cvsPreheatSetPtConstantOff)) .GT. StringToReal(FldVal(base + cvsCoolSetPtDesignOff))) THEN
       CALL WriteError('Warning:  In HVACTemplate:System:ConstantVolume "'//TRIM(FldVal(base + cvsAirHandlerNameOff))//'"'// &
                      ' the Preheat Coil Design Setpoint is greater than the Cooling Coil Design Setpoint.'// &
                      ' This may cause the preheat coil and cooling coil to operate simultaneously.'// &
                      ' Check results carefully and adjust controls if needed.',msgWarning)
      ENDIF
    ENDIF
  ENDIF
  ! check if there is no heating coil and heating design setpoint is > preheat design setpoint
  IF (heatCoilType == ctNone) THEN
    IF(StringToReal(FldVal(base + cvsHeatSetPtDesignOff)) .GT. StringToReal(FldVal(base + cvsPreheatSetPtConstantOff))) THEN
      usePreheatTempforHeatDesignSuppT = .TRUE.
      CALL WriteError('Warning:  In HVACTemplate:System:ConstantVolume "'//TRIM(FldVal(base + cvsAirHandlerNameOff))//'"'// &
                    ' the Heating Coil Design Setpoint is greater than the Preheat Coil Design Setpoint,'// &
                    ' but Heating Coil Type=None. Using Preheat Coil Design Setpoint for the Sizing:System'// &
                    ' Central Heating Design Supply Air Temperature.',msgWarning)
    ENDIF
    IF (preHeatCoilType == ctNone) THEN
      usePreheatTempforHeatDesignSuppT = .TRUE.
      CALL WriteError('Warning:  In HVACTemplate:System:ConstantVolume "'//TRIM(FldVal(base + cvsAirHandlerNameOff))//'"'// &
                    ' there is no Heating Coil and no Preheat Coil.'// &
                    ' The Preheat Coil Design Setpoint will be used for the Sizing:System'// &
                    ' Central Heating Design Supply Air Temperature. This will be the inlet air temperature for sizing'// &
                    ' reheat coils.',msgWarning)
    ENDIF
  ENDIF


  !SYSTEM SIZING  ~ line 158
  CALL CreateNewObj('Sizing:System')
  CALL AddToObjFld('AirLoop Name', base + cvsAirHandlerNameOff,' ')
  CALL AddToObjStr('Type of Load to Size On','Sensible')
  CALL AddToObjFld('Design Outdoor Air Flow Rate {m3/s}', base + cvsMinOutsideFlowOff,' ')
  CALL AddToObjStr('Central Heating Maximum System Air Flow Ratio','1.0')
  CALL AddToObjFld('Preheat Design Temperature {C}', base + cvsPreheatSetPtConstantOff,'')
  CALL AddToObjStr('Preheat Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjStr('Precool Design Temperature {C}','11')
  CALL AddToObjStr('Precool Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjFld('Central Cooling Design Supply Air Temperature {C}', base + cvsCoolSetPtDesignOff,'')
  IF (usePreheatTempforHeatDesignSuppT) THEN
    CALL AddToObjFld('Central Heating Design Supply Air Temperature {C}', base + cvsPreheatSetPtConstantOff,'')
  ELSE
    CALL AddToObjFld('Central Heating Design Supply Air Temperature {C}', base + cvsHeatSetPtDesignOff,'')
  ENDIF
  CALL AddToObjStr('Sizing Option', 'NonCoincident')
  CALL AddToObjStr('100% Outdoor Air in Cooling','No')
  CALL AddToObjStr('100% Outdoor Air in Heating','No')
  CALL AddToObjStr('Central Cooling Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjStr('Central Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isSupplyMaxRateAutosize) THEN
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','DesignDay')
    CALL AddToObjStr('Cooling Supply Air Flow Rate {m3/s}','0')
  ELSE
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','Flow/System')
    CALL AddToObjFld('Cooling Supply Air Flow Rate {m3/s}', base + cvsSupplyRateOff,' ')
  END IF
  ! New fields for v8.2 next three rows
  CALL AddToObjStr('Cooling Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
  CALL AddToObjStr('Cooling Fraction of Autosized Cooling Supply Air Flow Rate','')
  CALL AddToObjStr('Cooling Supply Air Flow Rate Per Unit Cooling Capacity {m3/s-W}','')
  CALL AddToObjStr('Heating Supply Air Flow Rate Method','DesignDay')
  CALL AddToObjStr('Heating Supply Air Flow Rate {m3/s}','0')
  ! New fields for v8.2 next four rows
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
  CALL AddToObjStr('Heating Fraction of Autosized Heating Supply Air Flow Rate','')
  CALL AddToObjStr('Heating Fraction of Autosized Cooling Supply Air Flow Rate','')
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Unit Heating Capacity {m3/s-W}','')
  CALL AddToObjStr('System Outdoor Air Method','ZoneSum')
  CALL AddToObjStr('Zone Maximum Outdoor Air Fraction','1.0')
  CALL AddToObjStr('Cooling Design Capacity Method','CoolingDesignCapacity')
  CALL AddToObjStr('Cooling Design Capacity {W}','autosize')
  CALL AddToObjStr('Cooling Design Capacity Per Floor Area {W/m2}','')
  CALL AddToObjStr('Fraction of Autosized Cooling Design Capacity','')
  CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
  CALL AddToObjStr('Heating Design Capacity {W}','autosize')
  CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
  CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
  CALL AddToObjStr('Central Cooling Capacity Control Method','OnOff',.true.)

  !***AirLoopHVAC
  CALL CreateNewObj('AirLoopHVAC')
  CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' ')
  IF ((coolCoilKind .EQ. ccChWater) .OR. (coolCoilKind .EQ. ccChWaterDetailed) .OR. (coolCoilKind .EQ. ccWaterCoilSystemHX) &
       .OR. (heatCoilType .EQ. ctHotWater)) THEN
    CALL AddToObjFld('Controller List Name', base + cvsAirHandlerNameOff,' Controllers')
  ELSE
    CALL AddToObjStr('Controller List Name','')
  ENDIF
  CALL AddToObjFld('Availability Manager List Name', base + cvsAirHandlerNameOff,' Availability Managers')
  CALL AddToObjStr('Design Supply Air Flow Rate {m3/s}','autosize')
  CALL AddToObjFld('Branch List Name', base + cvsAirHandlerNameOff,' Branches')
  CALL AddToObjStr('Connector List Name','')
  CALL AddToObjFld('Supply Side Inlet Node Name', base + cvsAirHandlerNameOff,' Air Loop Inlet')
  CALL AddToObjFld('Demand Side Outlet Node Name', base + cvsAirHandlerNameOff,' Return Air Outlet')
  CALL AddToObjFld('Demand Side Inlet Node Names', base + cvsAirHandlerNameOff,' Supply Path Inlet')
  CALL AddToObjStr('Supply Side Outlet Node Names', TRIM(airloopOutlet),.TRUE.)
  !if there is a chilled water coil
  IF ((coolCoilKind .EQ. ccChWater) .OR. (coolCoilKind .EQ. ccChWaterDetailed) .OR. (coolCoilKind .EQ. ccWaterCoilSystemHX)) THEN
  !if cvsHeatCoilKindOff=Hot Water
    IF (heatCoilType .EQ. ctHotWater) THEN
      !***AirLoopHVAC:ControllerList
      CALL CreateNewObj('AirLoopHVAC:ControllerList')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Controllers')
      CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
      CALL AddToObjFld('Controller Name', base + cvsAirHandlerNameOff,' Cooling Coil Controller')
      CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
      CALL AddToObjFld('Controller Name', base + cvsAirHandlerNameOff,' Heating Coil Controller',.TRUE.)
    ELSE
      !***AirLoopHVAC:ControllerList
      CALL CreateNewObj('AirLoopHVAC:ControllerList')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Controllers')
      CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
      CALL AddToObjFld('Controller Name', base + cvsAirHandlerNameOff,' Cooling Coil Controller',.TRUE.)
    END IF
  ELSE
    IF (heatCoilType .EQ. ctHotWater) THEN
      !***AirLoopHVAC:ControllerList
      CALL CreateNewObj('AirLoopHVAC:ControllerList')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Controllers')
      CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
      CALL AddToObjFld('Controller Name', base + cvsAirHandlerNameOff,' Heating Coil Controller',.TRUE.)
    END IF
  END IF
  !***BranchList
  CALL CreateNewObj('BranchList')
  CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Branches')
  CALL AddToObjFld('Branch Name', base + cvsAirHandlerNameOff,' Main Branch',.TRUE.)
  !***Branch
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Main Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  IF (isReturnFanYes) THEN
    CALL AddToObjStr('Component Object Type','Fan:ConstantVolume')
    CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Return Fan')
    CALL AddToObjFld('Component Inlet Node Name', base + cvsAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + cvsAirHandlerNameOff,' Return Fan Outlet')
  END IF
  CALL AddToObjStr('Component Object Type','AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' OA System')
  CALL AddToObjStr('Component Inlet Node Name', TRIM(returnInletToOAMIxer))
  CALL AddToObjFld('Component Outlet Node Name', base + cvsAirHandlerNameOff,' Mixed Air Outlet')
  IF (humidifierKind .EQ. humidifyElecSteam) THEN
    CALL AddToObjStr('Component Object Type','Humidifier:Steam:Electric')
    CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Humidifier')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(humidifierInlet))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(humidifierOutlet))
  END IF
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    CALL AddToObjStr('Component Object Type','Fan:ConstantVolume')
    CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Supply Fan')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(fanInlet))
    IF ((coolCoilKind .NE. ccNone) .OR. (heatCoilType .NE. ctNone)) THEN
      CALL AddToObjStr('Component Outlet Node Name', TRIM(fanOutlet))
    ELSE
      CALL AddToObjStr('Component Outlet Node Name', TRIM(fanOutlet),.TRUE.)
    END IF
  END IF
  IF (coolCoilKind .NE. ccNone) THEN
    IF (coolCoilKind .EQ. ccChWater) THEN
      CALL AddToObjStr('Component Object Type','Coil:Cooling:Water')
      CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Cooling Coil')
    ELSEIF (coolCoilKind .EQ. ccChWaterDetailed) THEN
      CALL AddToObjStr('Component Object Type','Coil:Cooling:Water:DetailedGeometry')
      CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Cooling Coil')
    ELSEIF (coolCoilKind .EQ. ccWaterCoilSystemHX) THEN
      CALL AddToObjStr('Component Object Type','CoilSystem:Cooling:Water:HeatExchangerAssisted')
      CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Heat Exchanger Assisted Cooling Coil')
    END IF
    CALL AddToObjStr('Component Inlet Node Name', TRIM(coolCoilUnitInlet))
    IF ((heatCoilType .EQ. ctNone) .AND. (supFanPlacement .NE. sfpDrawThru)) THEN
      CALL AddToObjStr('Component Outlet Node Name', TRIM(coolCoilUnitOutlet),.TRUE.)
    ELSE
      CALL AddToObjStr('Component Outlet Node Name', TRIM(coolCoilUnitOutlet))
    END IF
  END IF
  IF (heatCoilType .NE. ctNone) THEN
    IF (heatCoilType .EQ. ctHotWater) THEN
      CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    ELSEIF (heatCoilType .EQ. ctElectric) THEN
      CALL AddToObjStr('Component Object Type','Coil:Heating:Electric')
    ELSEIF (heatCoilType .EQ. ctGas) THEN
      CALL AddToObjStr('Component Object Type','Coil:Heating:Fuel')
    END IF
    CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(heatCoilInlet))
    IF (supFanPlacement .EQ. sfpDrawThru) THEN
     CALL AddToObjStr('Component Outlet Node Name', TRIM(heatCoilOutlet))
    ELSE
      CALL AddToObjStr('Component Outlet Node Name', TRIM(heatCoilOutlet),.TRUE.)
    END IF
  END IF
  IF (supFanPlacement .EQ. sfpDrawThru) THEN
    CALL AddToObjStr('Component Object Type','Fan:ConstantVolume')
    CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Supply Fan')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(fanInlet))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(fanOutlet),.TRUE.)
  END IF
  IF (isSupplyPlenumBlank) THEN
    !ZONE SUPPLY AIR PATH
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + cvsAirHandlerNameOff,' Supply Path Inlet')
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneSplitter')
    CALL MakeSpecialIntList(zlConstVolAllSupPlen, base + cvsAirHandlerNameOff)
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Zone Splitter',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Zone Splitter')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Supply Plenum', .TRUE.)
      END IF
    END IF
    !ZONE SPLITTER
    CALL CreateNewObj('AirLoopHVAC:ZoneSplitter')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Zone Splitter')
    CALL AddToObjFld('Inlet Node Name', base + cvsAirHandlerNameOff,' Supply Path Inlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlConstVolAll,base + cvsAirHandlerNameOff)
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(iInt),' Zone Equip Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Inlet',.TRUE.)
    END IF
  ELSE
    !ZONE SUPPLY AIR PATH
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + cvsAirHandlerNameOff,' Supply Path Inlet')
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
    CALL MakeSpecialIntList(zlConstVolAllSupPlen, base + cvsAirHandlerNameOff)
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Supply Plenum',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Supply Plenum')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Supply Plenum', .TRUE.)
      END IF
    END IF
    !ZONE SUPPLY PLENUM
    CALL CreateNewObj('AirLoopHVAC:SupplyPlenum')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Supply Plenum')
    CALL AddToObjFld('Zone Name', base + cvsSupplyPlenumNameOff,' ')
    CALL AddToObjFld('Zone Node Name', base + cvsSupplyPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Inlet Node Name', base + cvsAirHandlerNameOff,' Supply Path Inlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlConstVolAll,base + cvsAirHandlerNameOff)
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(iInt),' Zone Equip Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Inlet',.TRUE.)
    END IF
  END IF
  IF (isReturnPlenumBlank) THEN
    !ZONE RETURN AIR PATH
    CALL CreateNewObj('AirLoopHVAC:ReturnPath')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Return Path')
    CALL AddToObjFld('Return Air Path Outlet Node Name', base + cvsAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlConstVolAllRetPlen, base + cvsAirHandlerNameOff)
    IF (numIntList .NE. 0) THEN
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Return Plenum')
      END DO
    END IF
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneMixer')
    CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Zone Mixer',.TRUE.)
    !ZONE MIXER
    CALL CreateNewObj('AirLoopHVAC:ZoneMixer')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Zone Mixer')
    CALL AddToObjFld('Outlet Node Name', base + cvsAirHandlerNameOff,' Return Air Outlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlConstVolAll,base + cvsAirHandlerNameOff)
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Inlet Node Name', IntList(iInt),' Return Outlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Inlet Node Name', IntList(numIntList),' Return Outlet',.TRUE.)
    END IF
  ELSE
    !ZONE RETURN AIR PATH
    CALL CreateNewObj('AirLoopHVAC:ReturnPath')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Return Path')
    CALL AddToObjFld('Return Air Path Outlet Node Name', base + cvsAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlConstVolAllRetPlen, base + cvsAirHandlerNameOff)
    IF (numIntList .NE. 0) THEN
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Return Plenum')
      END DO
    END IF
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Return Plenum',.TRUE.)
    !ZONE RETURN PLENUM
    CALL CreateNewObj('AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Return Plenum')
    CALL AddToObjFld('Zone Name', base + cvsReturnPlenumNameOff,' ')
    CALL AddToObjFld('Zone Node Name', base + cvsReturnPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Outlet Node Name', base + cvsAirHandlerNameOff,' Return Air Outlet')
    CALL AddToObjStr('Induced Air Outlet Node or NodeList Name','')
    CALL MakeSpecialIntList(zlConstVolAll,base + cvsAirHandlerNameOff)
    DO iInt = 1,numIntList - 1
      CALL AddToObjFld('Inlet Node Name', IntList(iInt),' Return Outlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Inlet Node Name', IntList(numIntList),' Return Outlet',.TRUE.)
    END IF
  END IF
  !SYSTEM AVAILABILITY MANAGER LIST ~ line 296
  CALL CreateNewObj('AvailabilityManagerAssignmentList')
  CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Availability Managers')
  CALL AddToObjStr('Availability Manager Object Type','AvailabilityManager:NightCycle')
  CALL AddToObjFld('Availability Manager Name', base + cvsAirHandlerNameOff,' Availability',.TRUE.)
  !SYSTEM AVAILABILITY MANAGER:NIGHT CYCLE line ~307
  CALL CreateNewObj('AvailabilityManager:NightCycle')
  CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Availability')
  CALL AddToObjStr('Applicability Schedule Name','HVACTemplate-Always 1')
  IF (isSysAvailSchedBlank) THEN
    CALL AddToObjStr('Fan Schedule Name','HVACTemplate-Always 1')
  ELSE
    CALL AddToObjFld('Fan Schedule Name', base + cvsSysAvailSchedNameOff,'')
  END IF
  CALL AddToObjFld('Control Type', base + cvsNightCycleControlOff,'')
  CALL AddToObjStr('Thermostat Tolerance {deltaC}','0.2')
  CALL AddToObjStr('Cycling Run Time Control Type', 'FixedRunTime')
  CALL AddToObjStr('Cycling run time {s}','3600')
  CALL AddToObjFld('Control Zone Name', base + cvsNightCycleCtrlZoneNameOff,'',.TRUE.)
  CALL AddAlwaysSchedule('1')

  ! Cooling coil setpoint manager
  IF ((coolCoilKind .NE. ccNone) .OR. (econoKind .NE. etNoEconomizer)) THEN
    IF ((coolSetPtCtrlKind .EQ. cspFixed) .OR. (coolSetPtCtrlKind .EQ. cspScheduled)) THEN
      IF ((coolSetPtCtrlKind .EQ. cspFixed) .OR. (isCoolStPtSchedBlank)) THEN
        CALL AddAlwaysSchedule(FldVal(base + cvsCoolSetPtDesignOff))
      END IF
      !***SetpointManager:Scheduled
      CALL CreateNewObj('SetpointManager:Scheduled')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      IF ((coolSetPtCtrlKind .EQ. cspFixed) .OR. (isCoolStPtSchedBlank)) THEN
        CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + cvsCoolSetPtDesignOff)))
      ELSE
        CALL AddToObjFld('Schedule Name', base + cvsCoolSetPtSchedNameOff,' ')
      END IF
    ELSE IF (coolSetPtCtrlKind .EQ. cspOAreset) THEN
      !***SetpointManager:OutdoorAirReset
      CALL CreateNewObj('SetpointManager:OutdoorAirReset')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('Setpoint at Outdoor Low Temperature {C}', base + cvsCoolSetPtAtODBLoOff,'')
      CALL AddToObjFld('Outdoor Low Temperature {C}', base + cvsCoolResetODBLoOff,'')
      CALL AddToObjFld('Setpoint at Outdoor High Temperature {C}', base + cvsCoolSetPtAtODBHiOff,'')
      CALL AddToObjFld('Outdoor High Temperature {C}', base + cvsCoolResetODBHiOff,'')
    ELSE IF (coolSetPtCtrlKind .EQ. cspWarmest) THEN
      !SET POINT MANAGER:WARMEST ~ line 322
      CALL CreateNewObj('SetpointManager:Warmest')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('HVAC Air Loop Name', base + cvsAirHandlerNameOff,'')
      CALL AddToObjFld('Minimum Setpoint Temperature {C}', base + cvsCoolSetPtDesignOff,'')
      IF ((StringToReal(FldVal(base + cvsHeatSetPtDesignOff)) + 1.0) .GT. 18.0) THEN
        CALL AddToObjNum('Maximum Setpoint Temperature {C}',StringToReal(FldVal(base + cvsHeatSetPtDesignOff)) + 1.0)
      ELSE
        CALL AddToObjStr('Maximum Setpoint Temperature {C}','18')
      END IF
        CALL AddToObjStr('Strategy','MaximumTemperature')
    ELSE IF (coolSetPtCtrlKind .EQ. cspControlZone) THEN
      coolCtrlZoneBase = 0
      DO iZone = 1, numCompactZoneConstVol
        IF (SameString(FldVal(compactZoneConstVolBase(iZone) + cvzNameOff), FldVal(base +  cvsCoolControlZoneOff))) THEN
          coolCtrlZoneBase = compactZoneConstVolBase(iZone)
          EXIT
        END IF
      END DO
      IF (coolCtrlZoneBase .EQ. 0) THEN
        CALL WriteError('Could not find Cooling Coil Control Zone name referenced in HVACTemplate:System:ConstantVolume object: ' &
         //TRIM(FldVal(base + cvsAirHandlerNameOff)))
      END IF
      !***SetpointManager:SingleZone:Cooling
      CALL CreateNewObj('SetpointManager:SingleZone:Cooling')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('Minimum Supply Air Temperature {C}', base + cvsCoolSetPtDesignOff,'')
      IF ((StringToReal(FldVal(base + cvsHeatSetPtDesignOff)) + 1.0) .GT. 18.0) THEN
        CALL AddToObjNum('Maximum Supply Air Temperature {C}',StringToReal(FldVal(base + cvsHeatSetPtDesignOff)) + 1.0)
      ELSE
        CALL AddToObjStr('Maximum Supply Air Temperature {C}','18')
      END IF
      CALL AddToObjFld('Control Zone Name', base + cvsCoolControlZoneOff,'')
      CALL AddToObjFld('Zone Node Name', base + cvsCoolControlZoneOff,' Zone Air Node')
      IF ((FldVal(coolCtrlZoneBase + cvzZoneSupplyPlenumNameOff) .EQ. '') .AND. &
          (FldVal(coolCtrlZoneBase + cvzReheatTypeOff) .EQ. 'None')) THEN
        CALL AddToObjFld('Zone Inlet Node Name', base + cvsCoolControlZoneOff,' Zone Equip Inlet')
      ELSE
        CALL AddToObjFld('Zone Inlet Node Name', base + cvsCoolControlZoneOff,' Supply Inlet')
      END IF
    END IF
    IF (supFanPlacement .EQ. sfpDrawThru) THEN
      CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(airloopOutlet),.TRUE.)
    ELSE
      CALL AddToObjFld('Setpoint Node or NodeList Name', base + cvsAirHandlerNameOff,' Cooling Setpoint Nodes',.TRUE.)
      CALL CreateNewObj('NodeList')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Cooling Setpoint Nodes')
      CALL AddToObjStr('Node Name 1', TRIM(airloopOutlet))
      CALL AddToObjStr('Node Name 2', TRIM(coolCoilUnitOutlet),.TRUE.)
    END IF

    IF ((coolCoilKind .NE. ccNone) .AND. (supFanPlacement .EQ. sfpDrawThru)) THEN
      !***SetpointManager:MixedAir
      CALL CreateNewObj('SetpointManager:MixedAir')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Cooling Coil Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjStr('Reference Setpoint Node Name', TRIM(airloopOutlet))
      CALL AddToObjStr('Fan Inlet Node Name', TRIM(fanInlet))
      CALL AddToObjStr('Fan Outlet Node Name', TRIM(fanOutlet))
      CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(coolCoilUnitOutlet),.TRUE.)
    END IF
    IF (econoKind .NE. etNoEconomizer) THEN
      !***SetpointManager:MixedAir
      CALL CreateNewObj('SetpointManager:MixedAir')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Economizer Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjStr('Reference Setpoint Node Name', TRIM(airloopOutlet))
      CALL AddToObjStr('Fan Inlet Node Name', TRIM(fanInlet))
      CALL AddToObjStr('Fan Outlet Node Name', TRIM(fanOutlet))
      CALL AddToObjFld('Setpoint Node or NodeList Name', base + cvsAirHandlerNameOff,' Mixed Air Outlet',.TRUE.)
    END IF
  END IF

  SELECT CASE (coolCoilKind)
    CASE (ccChWaterDetailed)
      !***Coil:Cooling:Water:DetailedGeometry
      CALL CreateNewObj('Coil:Cooling:Water:DetailedGeometry')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + cvsCoolAvailSchedNameOff,' ')
      CALL AddToObjStr('Max Water Flow Rate of Coil {m3/s}','autosize')
      CALL AddToObjStr('Tube Outside Surf Area {m2}','autosize')
      CALL AddToObjStr('Total Tube Inside Area {m2}','autosize')
      CALL AddToObjStr('Fin Surface Area {m2}','autosize')
      CALL AddToObjStr('Minimum Air Flow Area {m2}','autosize')
      CALL AddToObjStr('Coil Depth {m}','autosize')
      CALL AddToObjStr('Fin Diameter {m}','autosize')
      CALL AddToObjStr('Fin Thickness {m}','0.0015')
      CALL AddToObjStr('Tube Inside Diameter {m}','0.01445')
      CALL AddToObjStr('Tube Outside Diameter {m}','0.0159')
      CALL AddToObjStr('Tube Thermal Conductivity {W/m-K}','386.0')
      CALL AddToObjStr('Fin Thermal Conductivity {W/m-K}','204.0')
      CALL AddToObjStr('Fin Spacing {m}','0.0018')
      CALL AddToObjStr('Tube Depth Spacing {m}','0.026')
      CALL AddToObjStr('Number of Tube Rows','4')
      CALL AddToObjStr('Number of Tubes per Row','autosize')
      CALL AddToObjFld('Coil Water Inlet Node', base + cvsAirHandlerNameOff,' Cooling Coil ChW Inlet')
      CALL AddToObjFld('Coil Water Outlet Node', base + cvsAirHandlerNameOff,' Cooling Coil ChW Outlet')
      CALL AddToObjStr('Air Inlet Node Name', TRIM(coolCoilUnitInlet))
      CALL AddToObjStr('Air Outlet Node Name',TRIM(coolCoilUnitOutlet),.TRUE.)
      !***Controller:WaterCoil
      CALL CreateNewObj('Controller:WaterCoil')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Cooling Coil Controller')
      IF (isDehumidifyNone) THEN
        CALL AddToObjStr('Control Variable','Temperature')
      ELSE
        CALL AddToObjStr('Control Variable','TemperatureAndHumidityRatio')
      END IF
      CALL AddToObjStr('Action','Reverse')
      CALL AddToObjStr('Actuator Variable','Flow')
      CALL AddToObjStr('Sensor Node Name', TRIM(coolCoilUnitOutlet))
      CALL AddToObjFld('Actuator Node Name', base + cvsAirHandlerNameOff,' Cooling Coil ChW Inlet')
      CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
      CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
      CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)
    CASE (ccChWater)
      !***Coil:Cooling:Water
      CALL CreateNewObj('Coil:Cooling:Water')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + cvsCoolAvailSchedNameOff,' ')
      CALL AddToObjStr('Design Water Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Design Air Flow Rate {m3/s}', 'autosize')
      CALL AddToObjStr('Design Inlet Water Temperature {C}','autosize')
      CALL AddToObjStr('Design Inlet Air Temperature {C}','autosize')
      CALL AddToObjStr('Design Outlet Air Temperature {C}','autosize')
      CALL AddToObjStr('Design Inlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
      CALL AddToObjStr('Design Outlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
      CALL AddToObjFld('Water Inlet Node Name', base + cvsAirHandlerNameOff,' Cooling Coil ChW Inlet')
      CALL AddToObjFld('Water Outlet Node Name', base + cvsAirHandlerNameOff,' Cooling Coil ChW Outlet')
      CALL AddToObjStr('Air Inlet Node Name', TRIM(coolCoilUnitInlet))
      CALL AddToObjStr('Air Outlet Node Name',TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('Type of Analysis','DetailedAnalysis')
      CALL AddToObjStr('Heat Exchanger Configuration','CrossFlow',.TRUE.)
      !***Controller:WaterCoil
      CALL CreateNewObj('Controller:WaterCoil')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Cooling Coil Controller')
      IF (isDehumidifyNone) THEN
        CALL AddToObjStr('Control Variable','Temperature')
      ELSE
        CALL AddToObjStr('Control Variable','TemperatureAndHumidityRatio')
      END IF
      CALL AddToObjStr('Action','Reverse')
      CALL AddToObjStr('Actuator Variable','Flow')
      CALL AddToObjStr('Sensor Node Name', TRIM(coolCoilUnitOutlet))
      CALL AddToObjFld('Actuator Node Name', base + cvsAirHandlerNameOff,' Cooling Coil ChW Inlet')
      CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
      CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
      CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)
    CASE (ccWaterCoilSystemHX)
      !***CoilSystem:Cooling:Water:HeatExchangerAssisted
      CALL CreateNewObj('CoilSystem:Cooling:Water:HeatExchangerAssisted')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Heat Exchanger Assisted Cooling Coil')
      CALL AddToObjStr('Heat Exchanger Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
      CALL AddToObjFld('Heat Exchanger Name', base + cvsAirHandlerNameOff,' Cooling Coil Heat Exchanger')
      CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:Water')
      CALL AddToObjFld('Cooling Coil Name', base + cvsAirHandlerNameOff, ' Cooling Coil',.TRUE.)
      !***HeatExchanger:AirToAir:SensibleAndLatent
      CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Cooling Coil Heat Exchanger')
      CALL AddToObjStr('Availability Schedule Name','')
      CALL AddToObjStr('Nominal Supply Air Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Sensible Effectiveness at 100% Heating Air Flow {dimensionless}','0.7')
      CALL AddToObjStr('Latent Effectiveness at 100% Heating Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Sensible Effectiveness at 75% Heating Air Flow {dimensionless}','0.75')
      CALL AddToObjStr('Latent Effectiveness at 75% Heating Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Sensible Effectiveness at 100% Cooling Air Flow {dimensionless}','0.7')
      CALL AddToObjStr('Latent Effectiveness at 100% Cooling Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Sensible Effectiveness at 75% Cooling Air Flow {dimensionless}','0.75')
      CALL AddToObjStr('Latent Effectiveness at 75% Cooling Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Supply Air Inlet Node Name', TRIM(coolCoilUnitInlet))
      CALL AddToObjFld('Supply Air Outlet Node Name', base + cvsAirHandlerNameOff,' HX Supply Outlet Node')
      CALL AddToObjFld('Exhaust Air Inlet Node Name', base + cvsAirHandlerNameOff,' Cooling Coil Outlet Node')
      CALL AddToObjStr('Exhaust Air Outlet Node Name', TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('Nominal Electric Power {W}','0')
      CALL AddToObjStr('Supply Air Outlet Temperature Control','No')
      CALL AddToObjStr('Heat Exchanger Type','Plate')
      CALL AddToObjStr('Frost Control Type','None')
      CALL AddToObjStr('Threshold Temperature {C}','')
      CALL AddToObjStr('Initial Defrost Time Fraction {dimensionless}','')
      CALL AddToObjStr('Rate of Defrost Time Fraction Increase {1/K}','')
      ! MJW ???? - Not sure
      CALL AddToObjStr('Economizer Lockout','No',.TRUE.)
      !***Coil:Cooling:Water
      CALL CreateNewObj('Coil:Cooling:Water')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + cvsCoolAvailSchedNameOff,' ')
      CALL AddToObjStr('Design Water Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Design Air Flow Rate {m3/s}', 'autosize')
      CALL AddToObjStr('Design Inlet Water Temperature {C}','autosize')
      CALL AddToObjStr('Design Inlet Air Temperature {C}','autosize')
      CALL AddToObjStr('Design Outlet Air Temperature {C}','autosize')
      CALL AddToObjStr('Design Inlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
      CALL AddToObjStr('Design Outlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
      CALL AddToObjFld('Water Inlet Node Name', base + cvsAirHandlerNameOff,' Cooling Coil ChW Inlet')
      CALL AddToObjFld('Water Outlet Node Name', base + cvsAirHandlerNameOff,' Cooling Coil ChW Outlet')
      CALL AddToObjFld('Air Inlet Node Name', base + cvsAirHandlerNameOff,' HX Supply Outlet Node')
      CALL AddToObjFld('Air Outlet Node Name', base + cvsAirHandlerNameOff,' Cooling Coil Outlet Node')
      CALL AddToObjStr('Type of Analysis','DetailedAnalysis')
      CALL AddToObjStr('Heat Exchanger Configuration','CrossFlow',.TRUE.)
      !***Controller:WaterCoil
      CALL CreateNewObj('Controller:WaterCoil')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Cooling Coil Controller')
      IF (isDehumidifyNone) THEN
        CALL AddToObjStr('Control Variable','Temperature')
      ELSE
        CALL AddToObjStr('Control Variable','TemperatureAndHumidityRatio')
      END IF
      CALL AddToObjStr('Action','Reverse')
      CALL AddToObjStr('Actuator Variable','Flow')
      CALL AddToObjStr('Sensor Node Name', TRIM(coolCoilUnitOutlet))
      CALL AddToObjFld('Actuator Node Name', base + cvsAirHandlerNameOff,' Cooling Coil ChW Inlet')
      CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
      CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
      CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)
    CASE (ccNone)
      ! nothing to do if no cooling coil specified
  END SELECT

  !***Branch
  IF ((coolCoilKind .EQ. ccChWater) .OR. (coolCoilKind .EQ. ccChWaterDetailed) .OR. (coolCoilKind .EQ. ccWaterCoilSystemHX)) THEN
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Cooling Coil ChW Branch')
    !store for later in the chilled water loop
    CALL AddToStrList(TRIM(FldVal(base + cvsAirHandlerNameOff)) // ' Cooling Coil ChW Branch', handleCoolingCoilBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF ((coolCoilKind .EQ. ccChWater) .OR. (coolCoilKind .EQ. ccWaterCoilSystemHX)) THEN
      CALL AddToObjStr('Component Object Type','Coil:Cooling:Water')
    ELSEIF (coolCoilKind .EQ. ccChWaterDetailed) THEN
      CALL AddToObjStr('Component Object Type','Coil:Cooling:Water:DetailedGeometry')
    END IF
    CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Cooling Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + cvsAirHandlerNameOff,' Cooling Coil ChW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + cvsAirHandlerNameOff,' Cooling Coil ChW Outlet',.TRUE.)
  ENDIF

  IF (heatCoilType .EQ. ctHotWater) THEN
    !***Coil:Heating:Water
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + cvsHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name ', base + cvsAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + cvsAirHandlerNameOff,' Heating Coil HW Outlet')
    CALL AddToObjStr('Air Inlet Node Name',  TRIM(heatCoilInlet))
    CALL AddToObjStr('Air Outlet Node Name',  TRIM(heatCoilOutlet))
    IF (isHeatCoilCapAutosize) THEN
      CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
      CALL AddToObjStr('Rated Capacity','')
    ELSE
      CALL AddToObjStr('Performance Input Method','NominalCapacity')
      CALL AddToObjFld('Rated Capacity',base + cvsHeatCoilCapOff,'')
    END IF
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
    !***Branch
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Heating Coil HW Branch')
    !save this branch name for later use in plan
    CALL AddToStrList(TRIM(FldVal(base + cvsAirHandlerNameOff)) // ' Heating Coil HW Branch',handleHeatingCoilBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + cvsAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + cvsAirHandlerNameOff,' Heating Coil HW Outlet',.TRUE.)
    !***Controller:WaterCoil
    CALL CreateNewObj('Controller:WaterCoil')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Heating Coil Controller')
    CALL AddToObjStr('Control Variable','Temperature')
    CALL AddToObjStr('Action','Normal')
    CALL AddToObjStr('Actuator Variable','Flow')
    CALL AddToObjStr('Sensor Node Name', TRIM(heatCoilOutlet))
    CALL AddToObjFld('Actuator Node Name', base + cvsAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
    CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
    CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)
  END IF
  IF (heatCoilType .EQ. ctElectric) THEN
    !***Coil:Heating:Electric
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + cvsHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Efficiency','1')
    CALL AddToObjFld('Nominal Capacity {W}',base + cvsHeatCoilCapOff,'')
    CALL AddToObjStr('Air Inlet Node Name',  TRIM(heatCoilInlet))
    CALL AddToObjStr('Air Outlet Node Name',  TRIM(heatCoilOutlet))
    CALL AddToObjStr('Temperature Setpoint Node Name', TRIM(heatCoilOutlet),.TRUE.)
  END IF
  IF (heatCoilType .EQ. ctGas) THEN
    !***Coil:Heating:Fuel
    CALL CreateNewObj('Coil:Heating:Fuel')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + cvsHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Fuel Type', 'Gas')
    CALL AddToObjFld('Gas Burner Efficiency', base + cvsHeatEffiencyOff,' ')
    CALL AddToObjFld('Nominal Capacity {W}',base + cvsHeatCoilCapOff,'')
    CALL AddToObjStr('Air Inlet Node Name',  TRIM(heatCoilInlet))
    CALL AddToObjStr('Air Outlet Node Name',  TRIM(heatCoilOutlet))
    CALL AddToObjStr('Temperature Setpoint Node Name', TRIM(heatCoilOutlet))
    CALL AddToObjFld('Parasitic Electric Load {W}', base + cvsHeatParasiticOff,' ')
    CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
       base + cvsAirHandlerNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
    !***Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  END IF

  ! Heating coil setpoint manager
  IF (heatCoilType .NE. ctNone) THEN
    IF ((heatSetPtCtrlKind .EQ. hspFixed) .OR. (heatSetPtCtrlKind .EQ. hspScheduled)) THEN
      IF ((heatSetPtCtrlKind .EQ. hspFixed) .OR. (isheatStPtSchedBlank)) THEN
        CALL AddAlwaysSchedule(FldVal(base + cvsHeatSetPtDesignOff))
      END IF
      !***SetpointManager:Scheduled
      IF (supFanPlacement .EQ. sfpDrawThru) THEN
        CALL WriteComment('NOTE:  Heating supply air setpoint is placed on Supply Path ' // &
                        'Inlet node to allow separate heating and cooling setpoints.')
        CALL WriteComment('SetpointManager:MixedAir <system name> Heating Coil Air ' // &
                        'Temp Manager adjusts this setpoint for fan heat and places it on the Heating Coil Outlet')
      END IF
      CALL CreateNewObj('SetpointManager:Scheduled')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Heating Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      IF ((heatSetPtCtrlKind .EQ. hspFixed) .OR. (isheatStPtSchedBlank)) THEN
        CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + cvsHeatSetPtDesignOff)))
      ELSE
        CALL AddToObjFld('Schedule Name', base + cvsHeatSetPtSchedNameOff,'')
      END IF
    ELSE IF (heatSetPtCtrlKind .EQ. hspOAreset) THEN
      !***SetpointManager:OutdoorAirReset
      IF (supFanPlacement .EQ. sfpDrawThru) THEN
        CALL WriteComment('NOTE:  Heating supply air setpoint is placed on Supply Path ' // &
                        'Inlet node to allow separate heating and cooling setpoints.')
        CALL WriteComment('SetpointManager:MixedAir <system name> Heating Coil Air ' // &
                        'Temp Manager adjusts this setpoint for fan heat and places it on the Heating Coil Outlet')
      END IF
      CALL CreateNewObj('SetpointManager:OutdoorAirReset')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Heating Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('Setpoint at Outdoor Low Temperature {C}', base + cvsHeatSetPtAtODBLoOff,'')
      CALL AddToObjFld('Outdoor Low Temperature {C}', base + cvsHeatResetODBLoOff,'')
      CALL AddToObjFld('Setpoint at Outdoor High Temperature {C}', base + cvsHeatSetPtAtODBHiOff,'')
      CALL AddToObjFld('Outdoor High Temperature {C}', base + cvsHeatResetODBHiOff,'')
    ELSE IF (heatSetPtCtrlKind .EQ. hspControlZone) THEN
      heatCtrlZoneBase = 0
      DO iZone = 1, numCompactZoneConstVol
        IF (SameString(FldVal(compactZoneConstVolBase(iZone) + cvzNameOff), FldVal(base +  cvsheatControlZoneOff))) THEN
          heatCtrlZoneBase = compactZoneConstVolBase(iZone)
          EXIT
        END IF
      END DO
      IF (heatCtrlZoneBase .EQ. 0) THEN
        CALL WriteError('Could not find Heating Coil Control Zone name referenced in HVACTemplate:System:ConstantVolume object: ' &
         //TRIM(FldVal(base + cvsAirHandlerNameOff)))
      END IF
      !***SetpointManager:SingleZone:Heating
      CALL CreateNewObj('SetpointManager:SingleZone:Heating')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Heating Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      IF ((StringToReal(FldVal(base + cvsCoolSetPtDesignOff)) - 1.0) .LT. &
       (StringToReal(FldVal(base + cvsHeatSetPtDesignOff)) - 5.2)) THEN
        CALL AddToObjNum('Minimum Supply Air Temperature {C}',StringToReal(FldVal(base + cvsCoolSetPtDesignOff)) - 1.0)
      ELSE
        CALL AddToObjNum('Minimum Supply Air Temperature {C}',StringToReal(FldVal(base + cvsHeatSetPtDesignOff)) - 5.2)
      END IF
      CALL AddToObjFld('Maximum Supply Air Temperature {C}', base + cvsHeatSetPtDesignOff,'')
      CALL AddToObjFld('Control Zone Name', base + cvsheatControlZoneOff,'')
      CALL AddToObjFld('Zone Node Name', base + cvsheatControlZoneOff,' Zone Air Node')
      IF ((FldVal(heatCtrlZoneBase + cvzZoneSupplyPlenumNameOff) .EQ. '') .AND. &
          (FldVal(heatCtrlZoneBase + cvzReheatTypeOff) .EQ. 'None')) THEN
        CALL AddToObjFld('Zone Inlet Node Name', base + cvsheatControlZoneOff,' Zone Equip Inlet')
      ELSE
        CALL AddToObjFld('Zone Inlet Node Name', base + cvsheatControlZoneOff,' Supply Inlet')
      END IF
    END IF
    IF (supFanPlacement .EQ. sfpDrawThru) THEN
      CALL AddToObjFld('Setpoint Node or NodeList Name', base + cvsAirHandlerNameOff,' Supply Path Inlet',.TRUE.)
    ELSE
      CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(heatCoilOutlet),.TRUE.)
    END IF
    ! Only needed if system is draw through and there is a heating coil
    IF (supFanPlacement .EQ. sfpDrawThru) THEN
      !***SetpointManager:MixedAir
      CALL CreateNewObj('SetpointManager:MixedAir')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Heating Coil Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('Reference Setpoint Node Name', base + cvsAirHandlerNameOff,' Supply Path Inlet')
      CALL AddToObjStr('Fan Inlet Node Name', TRIM(fanInlet))
      CALL AddToObjStr('Fan Outlet Node Name', TRIM(fanOutlet))
      CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(heatCoilOutlet),.TRUE.)
    END IF
  END IF
  ! Return fan
  IF (isReturnFanYes) THEN
    CALL CreateNewObj('Fan:ConstantVolume')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Return Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + cvsSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + cvsReturnEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + cvsReturnPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
    CALL AddToObjFld('Motor Efficiency', base + cvsReturnFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + cvsReturnFanFractionOff,' ')
    CALL AddToObjFld('Air Inlet Node Name', base + cvsAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Air Outlet Node Name', base + cvsAirHandlerNameOff,' Return Fan Outlet',.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF
  !***Fan:ConstantVolume
  CALL CreateNewObj('Fan:ConstantVolume')
  CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Supply Fan')
  IF (isSysAvailSchedBlank) THEN
    CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
  ELSE
    CALL AddToObjFld('Availability Schedule Name', base + cvsSysAvailSchedNameOff,' ')
  END IF
  CALL AddToObjFld('Fan Efficiency', base + cvsSupplyEfficiencyOff,' ')
  CALL AddToObjFld('Pressure Rise {Pa}', base + cvsSupplyPressureOff,' ')
  CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
  CALL AddToObjFld('Motor Efficiency', base + cvsSupplyFanMotorEffOff,' ')
  CALL AddToObjFld('Motor in Airstream Fraction', base + cvsSupplyFanFractionOff,' ')
  CALL AddToObjStr('Air Inlet Node Name', TRIM(fanInlet))
  CALL AddToObjStr('Air Outlet Node Name', TRIM(fanOutlet),.TRUE.)
  IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')

  ! Humdistat(s) if needed
  ! first check if the schedule names are used or if values should be used.
  IF (.NOT. isDehumidifyNone) THEN
    IF (FldVal(base + cvsDehumSetPtSchedNameOff) .EQ. '') THEN
      dehumscheduleName = 'HVACTemplate-Always ' // TRIM(FldVal(base + cvsDehumSetPtOff))
      CALL AddAlwaysSchedule(FldVal(base + cvsDehumSetPtOff))
    ELSE
      dehumscheduleName = TRIM(FldVal(base + cvsDehumSetPtSchedNameOff))
    ENDIF
  END IF
  IF (.NOT. isHumidifierNone) THEN
    ! first check if the schedule name is used or if value should be used.
    IF (FldVal(base + cvsHumidSetPtSchedNameOff) .EQ. '') THEN
      humscheduleName = 'HVACTemplate-Always ' // TRIM(FldVal(base + cvsHumidSetPtOff))
      CALL AddAlwaysSchedule(FldVal(base + cvsHumidSetPtOff))
    ELSE
      humscheduleName = TRIM(FldVal(base + cvsHumidSetPtSchedNameOff))
    ENDIF
  END IF

  !    Single humidistat if humidification and dehumidification are both active and control zones are the same
  IF ((.NOT. isHumidifierNone) .AND. (.NOT. isDehumidifyNone) &
       .AND. SameString(FldVal(base +  cvsDehumCtrlZoneOff),FldVal(base +  cvsHumidCtrlZoneOff))) THEN
    CALL CreateNewObj('ZoneControl:Humidistat')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Humidistat')
    CALL AddToObjFld('Zone Name', base + cvsDehumCtrlZoneOff,'')
    CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name', TRIM(humscheduleName))
    CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name', TRIM(dehumscheduleName),.TRUE.)
  ELSE
    IF (.NOT. isDehumidifyNone) THEN
  !   Dehumidification humidistat
      CALL CreateNewObj('ZoneControl:Humidistat')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Dehumidification Humidistat')
      CALL AddToObjFld('Zone Name', base + cvsDehumCtrlZoneOff,'')
      CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name','HVACTemplate-Always 1')
      CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name', TRIM(dehumscheduleName),.TRUE.)

      CALL AddAlwaysSchedule('1')
    ENDIF
    IF (.NOT. isHumidifierNone) THEN
  !    Humidification humidistat
      CALL CreateNewObj('ZoneControl:Humidistat')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Humidification Humidistat')
      CALL AddToObjFld('Zone Name', base + cvsHumidCtrlZoneOff,'')
      CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name', TRIM(humscheduleName))
      CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name','HVACTemplate-Always 100',.TRUE.)

      CALL AddAlwaysSchedule('100')
    ENDIF
  ENDIF

  !Set point manager for dehumidification
  IF (.NOT. isDehumidifyNone) THEN
    CALL CreateNewObj('SetpointManager:SingleZone:Humidity:Maximum')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Dehumidification Setpoint Manager')
    CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(coolCoilUnitOutlet))
    CALL AddToObjFld('Control Zone Air Node Name', base + cvsDehumCtrlZoneOff,' Zone Air Node',.TRUE.)
  END IF

  ! Humidifier set point manager and humidifier
  IF (.NOT. isHumidifierNone) THEN
    CALL CreateNewObj('SetpointManager:SingleZone:Humidity:Minimum')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Humidification Setpoint Manager')
    CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(humidifierOutlet))
    CALL AddToObjFld('Control Zone Air Node Name', base + cvsHumidCtrlZoneOff,' Zone Air Node',.TRUE.)
    IF (humidifierKind .EQ. humidifyElecSteam) THEN
      !***Humidifier:Steam:Electric
      CALL CreateNewObj('Humidifier:Steam:Electric')
      CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Humidifier')
      CALL AddToObjFld('Availability Schedule Name', base + cvsHumidAvailSchedNameOff,' ')
      CALL AddToObjFld('Rated Capacity {m3/s}', base + cvsHumidCapOff,'')
      CALL AddToObjFld('Rated Power {W}', base + cvsHumidRatedElecOff,'')
      CALL AddToObjStr('rated fan power {W}','0')
      CALL AddToObjStr('Standby power {W}','0')
      CALL AddToObjStr('Air Inlet Node Name', TRIM(humidifierInlet))
      CALL AddToObjStr('Air Outlet Node Name', TRIM(humidifierOutlet),.TRUE.)
    END IF
  END IF
  !***OutdoorAir:NodeList
  CALL CreateNewObj('OutdoorAir:NodeList')
  CALL AddToObjFld('Node or NodeList Name 1', base + cvsAirHandlerNameOff,' Outdoor Air Inlet',.TRUE.)
  !***AirLoopHVAC:OutdoorAirSystem
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' OA System')
  CALL AddToObjFld('Controller List Name', base + cvsAirHandlerNameOff,' OA System Controllers')
  CALL AddToObjFld('Outdoor Air Equipment List Name', base + cvsAirHandlerNameOff,' OA System Equipment')
  CALL AddToObjFld('Availability Manager List Name', base + cvsAirHandlerNameOff,' Availability Managers',.TRUE.)
  !***AirLoopHVAC:ControllerList
  CALL CreateNewObj('AirLoopHVAC:ControllerList')
  CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' OA System Controllers')
  IF (preHeatCoilType .EQ. ctHotWater) THEN
    CALL AddToObjStr('Controller Object Type','Controller:OutdoorAir')
    CALL AddToObjFld('Controller Name', base + cvsAirHandlerNameOff,' OA Controller')
    CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
    CALL AddToObjFld('Controller Name', base + cvsAirHandlerNameOff,' Preheat Coil Controller',.TRUE.)
  ELSE
    CALL AddToObjStr('Controller Object Type','Controller:OutdoorAir')
    CALL AddToObjFld('Controller Name', base + cvsAirHandlerNameOff,' OA Controller',.TRUE.)
  END IF
  !***AirLoopHVAC:OutdoorAirSystem:EquipmentList
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem:EquipmentList')
  CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' OA System Equipment')
  IF (heatRecovery .EQ. htrecSens) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Heat Recovery')
  ELSEIF (heatRecovery .EQ. htrecEnth) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Heat Recovery')
  END IF
  IF (preHeatCoilType .EQ. ctHotWater) THEN
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Preheat Coil')
  ELSEIF (preHeatCoilType .EQ. ctElectric) THEN
    CALL AddToObjStr('Component Object Type','Coil:Heating:Electric')
    CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Preheat Coil')
  ELSEIF (preHeatCoilType .EQ. ctGas) THEN
    CALL AddToObjStr('Component Object Type','Coil:Heating:Fuel')
    CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Preheat Coil')
  END IF
  CALL AddToObjStr('Component Object Type','OutdoorAir:Mixer')
  CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' OA Mixing Box',.TRUE.)
  !***OutdoorAir:Mixer
  CALL CreateNewObj('OutdoorAir:Mixer')
  CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' OA Mixing Box')
  CALL AddToObjFld('Mixed Air Node Name', base + cvsAirHandlerNameOff,' Mixed Air Outlet')
  IF ((heatRecovery .EQ. htrecNone) .AND. (preHeatCoilType .EQ. ctNone)) THEN
    CALL AddToObjFld('Outdoor Air Stream Node Name', base + cvsAirHandlerNameOff,' Outdoor Air Inlet')
  END IF
  IF (preHeatCoilType .NE. ctNone) THEN
    CALL AddToObjFld('Outdoor Air Stream Node Name', base + cvsAirHandlerNameOff,' Preheat Coil Outlet')
  END IF
  IF ((heatRecovery .NE. htrecNone) .AND. (preHeatCoilType .EQ. ctNone)) THEN
    CALL AddToObjFld('Outdoor Air Stream Node Name', base + cvsAirHandlerNameOff,' Heat Recovery Supply Outlet')
  ENDIF
  CALL AddToObjFld('Relief Air Stream Node Name', base + cvsAirHandlerNameOff,' Relief Air Outlet')
  CALL AddToObjStr('Return Air Stream Node Name', TRIM(returnInletToOAMIxer),.TRUE.)
  !***Controller:OutdoorAir
  CALL CreateNewObj('Controller:OutdoorAir')
  CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' OA Controller')
  CALL AddToObjFld('Relief Air Outlet Node Name', base + cvsAirHandlerNameOff,' Relief Air Outlet')
  CALL AddToObjStr('Return Air Node Name',TRIM(returnInletToOAMIxer))
  CALL AddToObjFld('Mixed Air Node Name', base + cvsAirHandlerNameOff,' Mixed Air Outlet')
  CALL AddToObjFld('Actuator Node Name', base + cvsAirHandlerNameOff,' Outdoor Air Inlet')
  ! the minimum and maximum are the same volume of air
  CALL AddToObjFld('Minimum Outdoor Air Flow Rate {m3/s}', base + cvsMinOutsideFlowOff,' ')
  CALL AddToObjFld('Maximum Outdoor Air Flow Rate {m3/s}', base + cvsMaxOutsideFlowOff,' ')
  CALL AddToObjFld('Economizer Control Type', base + cvsEconomizerKindOff,' ')
  ! If heat recovery is selected, then set economizer control to bypass
  ! otherwise, set to modulate OA flow
  IF (heatRecovery .EQ. htrecNone) THEN
    CALL AddtoObjStr('Economizer Control Action Type','ModulateFlow')
  ELSE
    CALL AddtoObjStr('Economizer Control Action Type','MinimumFlowWithBypass')
  ENDIF
  CALL AddToObjFld('Economizer Maximum Limit Dry-Bulb Temperature {C}', base + cvsEconoUpLimitOff,' ')
  CALL AddToObjFld('Economizer Maximum Limit Enthalpy {J/kg}', base + cvsEconoEnthLimitOff,' ')
  IF (econoKind .EQ. etFixedDewPointAndDryBulb) THEN
    CALL AddToObjFld('Economizer Maximum Limit Dewpoint Temperature (C)',base + cvsEconoUpDewLimitOff,' ')
  ELSE
    CALL AddToObjStr('Economizer Maximum Limit Dewpoint Temperature (C)',' ')
  ENDIF
  IF (econoKind .EQ. etElectronicEnthalpy) THEN
    CALL AddToObjFld('Electronic Enthalpy Limit Curve Name', base + cvsAirHandlerNameOff,' ElectrEnthCurve')
  ELSE
    CALL AddToObjStr('Electronic Enthalpy Limit Curve Name',' ')
  ENDIF
  CALL AddToObjFld('Economizer Minimum Limit Dry-Bulb Temperature {C}', base + cvsEconoLowLimitOff,' ')
  CALL AddToObjStr('Lockout Type', 'NoLockout')
  CALL AddToObjStr('Minimum Limit Type','ProportionalMinimum')
  CALL AddToObjFld('Minimum Outdoor Air Schedule Name', base + cvsMinOutFlowSchedNameOff,' ')
  CALL AddToObjStr('Minimum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Maximum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Mechanical Ventilation Controller Name','',.TRUE.)
  ! If economizer type is ElectronicEnthalpy then create curve object
  IF (econoKind .EQ. etElectronicEnthalpy) THEN
    !CURVE:CUBIC
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' ElectrEnthCurve')
    CALL AddToObjStr('Coefficient1 Constant','0.01342704')
    CALL AddToObjStr('Coefficient2 x','-0.00047892')
    CALL AddToObjStr('Coefficient3 x**2','0.000053352')
    CALL AddToObjStr('Coefficient4 x**3','-0.0000018103')
    CALL AddToObjStr('Minimum Value of x','16.6')
    CALL AddToObjStr('Maximum Value of x','29.13',.TRUE.)
  ENDIF
  IF (preHeatCoilType .EQ. ctHotWater) THEN
    !COIL:Water:SimpleHeating  ~ line 614
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Preheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + cvsPreheatAvailSchedNameOff,' ')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name ', base + cvsAirHandlerNameOff,' Preheat Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + cvsAirHandlerNameOff,' Preheat Coil HW Outlet')
    IF (heatRecovery .EQ. htrecNone) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + cvsAirHandlerNameOff,' Outdoor Air Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + cvsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    END IF
    CALL AddToObjFld('Air Outlet Node Name', base + cvsAirHandlerNameOff,' Preheat Coil Outlet')
    CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
    CALL AddToObjStr('Rated Capacity','autosize')
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
    !BRANCH ~ line 625
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Preheat Coil HW Branch')
    !save this branch name for later use in plan
    CALL AddToStrList(TRIM(FldVal(base + cvsAirHandlerNameOff)) // ' Preheat Coil HW Branch',handleHeatingCoilBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + cvsAirHandlerNameOff,' Preheat Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + cvsAirHandlerNameOff,' Preheat Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + cvsAirHandlerNameOff,' Preheat Coil HW Outlet',.TRUE.)
    !CONTROLLER:SIMPLE ~ line 634
    CALL CreateNewObj('Controller:WaterCoil')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Preheat Coil Controller')
    CALL AddToObjStr('Control Variable','Temperature')
    CALL AddToObjStr('Action','Normal')
    CALL AddToObjStr('Actuator Variable','Flow')
    CALL AddToObjFld('Sensor Node Name', base + cvsAirHandlerNameOff,' Preheat Coil Outlet')
    CALL AddToObjFld('Actuator Node Name', base + cvsAirHandlerNameOff,' Preheat Coil HW Inlet')
    CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
    CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
    CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)
  ELSEIF (preHeatCoilType .EQ. ctElectric) THEN
    !COIL:Electric:Heating  ~ line 646
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Preheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + cvsPreheatAvailSchedNameOff,' ')
    CALL AddToObjStr('Efficiency','1')
    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    IF (heatRecovery .EQ. htrecNone) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + cvsAirHandlerNameOff,' Outdoor Air Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + cvsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    END IF
    CALL AddToObjFld('Air Outlet Node Name', base + cvsAirHandlerNameOff,' Preheat Coil Outlet')
    CALL AddToObjFld('Coil Temp Setpoint Node', base + cvsAirHandlerNameOff,' Preheat Coil Outlet',.TRUE.)
  ELSEIF (preHeatCoilType .EQ. ctGas) THEN
    !COIL:Gas:Heating ~ line 657
    CALL CreateNewObj('Coil:Heating:Fuel')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Preheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + cvsPreheatAvailSchedNameOff,' ')
    CALL AddToObjStr('Fuel Type', 'Gas')
    CALL AddToObjFld('Gas Burner Efficiency', base + cvsPreheatEffiencyOff,' ')
    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    IF (heatRecovery .EQ. htrecNone) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + cvsAirHandlerNameOff,' Outdoor Air Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + cvsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    END IF
    CALL AddToObjFld('Air Outlet Node Name', base + cvsAirHandlerNameOff,' Preheat Coil Outlet')
    CALL AddToObjFld('Coil Temp Setpoint Node', base + cvsAirHandlerNameOff,' Preheat Coil Outlet')
    CALL AddToObjFld('Parasitic Electric Load {W}', base + cvsPreheatParasiticOff,' ')
    CALL AddToObjFld('Part load fraction correlation (function of part load ratio)',   &
       base + cvsAirHandlerNameOff,' Preheat Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
    !CURVE:CUBIC ~ line 670
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Preheat Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  END IF
  IF (preHeatCoilType .NE. ctNone) THEN
    IF (isPreheatStPtSchedBlank) THEN
      CALL AddAlwaysSchedule(FldVal(base + cvsPreheatSetPtConstantOff))
    END IF
    !SET POINT MANAGER:SCHEDULED ~ line 679
    IF (isPreheatStPtSchedBlank) THEN
      CALL AddAlwaysSchedule(FldVal(base + cvsPreheatSetPtConstantOff))
    END IF
    CALL CreateNewObj('SetpointManager:Scheduled')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Preheat Coil Air Temp Manager')
    CALL AddToObjStr('Control Variable','Temperature')
    IF (isPreheatStPtSchedBlank) THEN
      CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + cvsPreheatSetPtConstantOff)))
    ELSE
      CALL AddToObjFld('Schedule Name', base + cvsPreheatSetPtSchedNameOff,'')
    END IF
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + cvsAirHandlerNameOff,' Preheat Coil Outlet',.TRUE.)
  END IF
  IF (heatRecovery .EQ. htrecSens) THEN
    htRecSens75 = StringToReal(FldVal(base + cvsHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    !***HeatExchanger:AirToAir:SensibleAndLatent
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + cvsAirHandlerNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name','')
    CALL AddToObjStr('Nominal supply air flow rate {m3/s}', 'autosize')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + cvsHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Heating Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Heating Air Flow','0')
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + cvsHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Cooling Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Cooling Air Flow','0')
    CALL AddToObjFld('Supply Air Inlet Node Name', base + cvsAirHandlerNameOff,' Outdoor Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + cvsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + cvsAirHandlerNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + cvsAirHandlerNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','250')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjFld('Heat Exchanger Type', base + cvsHXTypeOff,'')
    CALL AddToObjFld('Frost Control Type', base + cvsFrostControlOff,'')
    CALL AddToObjStr('Threshold Temperature','1.7')
    CALL AddToObjStr('Initial Defrost Time Fraction','0.083')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','0.012')
    CALL AddToObjStr('Economizer Lockout','Yes',.TRUE.)
  ELSEIF (heatRecovery .EQ. htrecEnth) THEN
    htRecSens75 = StringToReal(FldVal(base + cvsHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    htRecLat75 = StringToReal(FldVal(base + cvsHeatRecLatEffOff)) + 0.05
    IF (htRecLat75 .GT. 1.) htRecLat75 = 1.
    !***HeatExchanger:AirToAir:SensibleAndLatent
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + cvsAirHandlerNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name','')
    CALL AddToObjFld('Nominal supply air flow rate {m3/s}', base + cvsMinOutsideFlowOff,'')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + cvsHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Heating Air Flow', base + cvsHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow', htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Heating Air Flow', htRecLat75)
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + cvsHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Cooling Air Flow', base + cvsHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow', htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Cooling Air Flow', htRecLat75)
    CALL AddToObjFld('Supply Air Inlet Node Name', base + cvsAirHandlerNameOff,' Outdoor Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + cvsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + cvsAirHandlerNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + cvsAirHandlerNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','0')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjFld('Heat Exchanger Type', base + cvsHXTypeOff,'')
    CALL AddToObjFld('Frost Control Type', base + cvsFrostControlOff,'')
    CALL AddToObjStr('Threshold Temperature','1.7')
    CALL AddToObjStr('Initial Defrost Time Fraction','0.083')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','0.012')
    CALL AddToObjStr('Economizer Lockout','Yes',.TRUE.)
  END IF
  IF (heatRecovery .NE. htrecNone) THEN
    IF (isEconoLowLimitBlank) THEN
      CALL AddAlwaysSchedule('5')
    ELSE
      CALL AddAlwaysSchedule(FldVal(base + cvsEconoLowLimitOff))
    END IF
    CALL CreateNewObj('SetpointManager:Scheduled')
    CALL AddToObjFld('Name', base + cvsAirHandlerNameOff,' Heat Recovery Air Temp Manager')
    CALL AddToObjStr('Control Variable','Temperature')
    IF (isEconoLowLimitBlank) THEN
      CALL AddToObjStr('Schedule Name','HVACTemplate-Always 5')
    ELSE
      CALL AddToObjStr('Schedule Name','HVACTemplate-Always ' // FldVal(base + cvsEconoLowLimitOff))
    END IF
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + cvsAirHandlerNameOff,' Heat Recovery Supply Outlet',.TRUE.)
  END IF
END DO
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE CreateNewDualDuct
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Michael Witte of GARD Analytics, Inc.
          !    DATE WRITTEN   September 2013
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a Dual Duct system.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for coil type (heating)
INTEGER,PARAMETER :: ctHotWater = 1
INTEGER,PARAMETER :: ctElectric = 2
INTEGER,PARAMETER :: ctGas      = 3
INTEGER,PARAMETER :: ctNone     = 4
          ! parameters for coil type (cooling)
INTEGER, PARAMETER :: ccChWater = 1
INTEGER, PARAMETER :: ccChWaterDetailed = 2
INTEGER, PARAMETER :: ccNone = 3
          ! parameters for economizer type
INTEGER,PARAMETER :: etFixedDryBulb                   = 1
INTEGER,PARAMETER :: etFixedEnthalpy                  = 2
INTEGER,PARAMETER :: etDifferentialDryBulb            = 3
INTEGER,PARAMETER :: etDifferentialEnthalpy           = 4
INTEGER,PARAMETER :: etFixedDewPointAndDryBulb        = 5
INTEGER,PARAMETER :: etElectronicEnthalpy             = 6
INTEGER,PARAMETER :: etDifferentialDryBulbAndEnthalpy = 7
INTEGER,PARAMETER :: etNoEconomizer                   = 8
          ! parameters for lockout type
INTEGER, PARAMETER :: loNo       = 1
INTEGER, PARAMETER :: loHeat     = 2
INTEGER, PARAMETER :: loComp     = 3
          ! parameters for cool setpoint control type
INTEGER,PARAMETER :: cspFixed    = 1
INTEGER,PARAMETER :: cspScheduled = 2
INTEGER,PARAMETER :: cspOAreset = 3
INTEGER,PARAMETER :: cspWarmest = 4
         ! parameters for heat setpoint control type
INTEGER,PARAMETER :: hspFixed    = 1
INTEGER,PARAMETER :: hspScheduled = 2
INTEGER,PARAMETER :: hspOAreset = 3
INTEGER,PARAMETER :: hspColdest = 4
          ! parameters for supply fan coef type
INTEGER,PARAMETER :: sfcInletVane = 1
INTEGER,PARAMETER :: sfcDischgDamp = 2
INTEGER,PARAMETER :: sfcVSD = 3
INTEGER,PARAMETER :: sfcAppG = 4
INTEGER,PARAMETER :: sfcVSDPresReset = 5
          ! parameters for heat recovery type
INTEGER,PARAMETER :: htrecNone = 1
INTEGER,PARAMETER :: htrecSens = 2
INTEGER,PARAMETER :: htrecEnth = 3
          ! parameters for supply fan placement
INTEGER,PARAMETER :: sfpBlowThru = 1
INTEGER,PARAMETER :: sfpDrawThru = 2
INTEGER,PARAMETER :: sfpNone = 3
          ! parameters for supply fan kind
INTEGER,PARAMETER :: sfkVAV = 1
INTEGER,PARAMETER :: sfkCAV = 2
INTEGER,PARAMETER :: sfkNone = 3
         ! parameters for dehumidification type
INTEGER, PARAMETER :: dehumidNone    = 1
INTEGER, PARAMETER :: dehumidCoolRht = 2
          ! parameters for humidification type
INTEGER, PARAMETER :: humidifyNone      = 1
INTEGER, PARAMETER :: humidifyElecSteam = 2
          ! parameters for baseboard type
INTEGER, PARAMETER :: baseboardNone           = 1
INTEGER, PARAMETER :: baseboardHotWater       = 2
INTEGER, PARAMETER :: baseboardElectric       = 3

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for
INTEGER :: iZone
INTEGER :: iSys
INTEGER :: i
LOGICAL :: isTStatNameBlank
LOGICAL :: isSupplyMaxRateAutosize
LOGICAL :: isHotSupplyMaxRateAutosize
LOGICAL :: isColdSupplyMaxRateAutosize
REAL    :: heatDesAirFlowRate
INTEGER :: heatCoilType=0
INTEGER :: preHeatCoilType=0
INTEGER :: coolCoilType=0
LOGICAL :: isColdSupplyPlenumBlank
LOGICAL :: isHotSupplyPlenumBlank
LOGICAL :: isReturnPlenumBlank
REAL    :: minAirRatio
INTEGER :: znBase
INTEGER :: sysBase
LOGICAL :: isSysAvailSchedBlank
LOGICAL :: isCoolStPtSchedBlank
LOGICAL :: isHeatStPtSchedBlank
LOGICAL :: isPreheatStPtSchedBlank
INTEGER :: econoKind=0
INTEGER :: lockoutKind=0
LOGICAL :: isZoneRetPlenBlank
LOGICAL :: isZoneColdSupPlenBlank
LOGICAL :: isZoneHotSupPlenBlank
INTEGER :: coolSetPtCtrlKind=0
INTEGER :: heatSetPtCtrlKind=0
INTEGER :: supFanCoef=0
INTEGER :: coldSupFanCoef=0
INTEGER :: hotSupFanCoef=0
INTEGER :: retFanCoef=0
INTEGER :: heatRecovery=0
LOGICAL :: isEconoLowLimitBlank
REAL    :: heatDesAirRate
REAL    :: totalFlow
INTEGER :: iInt
INTEGER :: coldSupFanPlacement=0
INTEGER :: hotSupFanPlacement=0
INTEGER :: supFanKind=0
INTEGER :: coldFanKind=0
INTEGER :: hotFanKind=0
LOGICAL :: isMinOARateAutosize
REAL    :: htRecSens75
REAL    :: htRecLat75
LOGICAL :: isDehumidifyNone
INTEGER :: dehumidCtrlKind = 0
LOGICAL :: isHumidifierNone
INTEGER :: humidifierKind = 0
INTEGER :: baseboardKind = 0
LOGICAL :: isHeatCoilCapAutosize = .FALSE.
LOGICAL :: doSuppressMsgzlDualDuctAll = .FALSE.
LOGICAL :: usePreheatTempforHeatDesignSuppT = .FALSE.
LOGICAL :: isOATypeDetailed = .FALSE.
LOGICAL :: isReturnFanYes = .FALSE.
LOGICAL :: isSystemVAV = .FALSE.

! Node Names
CHARACTER(len=MaxAlphaLength) :: returnInletToOAMIxer=''
CHARACTER(len=MaxAlphaLength) :: lastOutlet=''
CHARACTER(len=MaxAlphaLength) :: supfanInlet=''
CHARACTER(len=MaxAlphaLength) :: supfanOutlet=''
CHARACTER(len=MaxAlphaLength) :: coldfanInlet=''
CHARACTER(len=MaxAlphaLength) :: coldfanOutlet=''
CHARACTER(len=MaxAlphaLength) :: hotfanInlet=''
CHARACTER(len=MaxAlphaLength) :: hotfanOutlet=''
CHARACTER(len=MaxAlphaLength) :: coolCoilInlet=''
CHARACTER(len=MaxAlphaLength) :: coolCoilOutlet=''
CHARACTER(len=MaxAlphaLength) :: heatCoilInlet=''
CHARACTER(len=MaxAlphaLength) :: heatCoilOutlet=''
CHARACTER(len=MaxAlphaLength) :: humidifierInlet=''
CHARACTER(len=MaxAlphaLength) :: humidifierOutlet=''
CHARACTER(len=MaxAlphaLength) :: mainbranchOutlet=''
CHARACTER(len=MaxAlphaLength) :: coldbranchInlet=''
CHARACTER(len=MaxAlphaLength) :: hotbranchInlet=''
CHARACTER(len=MaxAlphaLength) :: coldbranchOutlet=''
CHARACTER(len=MaxAlphaLength) :: hotbranchOutlet=''
CHARACTER(len=MaxAlphaLength) :: dehumscheduleName=''
CHARACTER(len=MaxAlphaLength) :: humscheduleName=''

!======================================================================================================
! create Zones
!======================================================================================================
DO iZone = 1, numCompactZoneDualDuct
  base =  compactZoneDualDuctbase(iZone)
  ! find base for system this zone belongs to
  sysBase = -1
  DO iSys = 1, numCompactSysDualDuct
    IF (SameString(FldVal(compactSysDualDuctbase(iSys) + ddsAirHandlerNameOff), FldVal(base +  ddzAirHandlerNameOff))) THEN
      sysBase = compactSysDualDuctBase(iSys)
      EXIT
    END IF
  END DO
  IF (sysBase .EQ. -1) THEN
    CALL WriteError('Could not find air handler name referenced in HVACTemplate:Zone:DualDuct object: '// &
     TRIM(FldVal(base + ddzNameOff)))
    CYCLE
  END IF
  ! set flags
  isTStatNameBlank = (FldVal(base + ddzTStatNameOff) .EQ. '')
  isSupplyMaxRateAutosize = SameString(FldVal(base + ddzSupplyMaxRateOff),'autosize')
  isOATypeDetailed = SameString(FldVal(base + ddzOutAirTypeOff),'DetailedSpecification')
  isZoneRetPlenBlank = (FldVal(base + ddzZoneReturnPlenumNameOff) .EQ. '')
  isZoneColdSupPlenBlank = (FldVal(base + ddzZoneColdSupplyPlenumNameOff) .EQ. '')
  isZoneHotSupPlenBlank = (FldVal(base + ddzZoneHotSupplyPlenumNameOff) .EQ. '')
  isSystemVAV = (SameString(FldVal(sysBase + ddsSupplyFanConfigTypeOff),'SingleFanVariableVolume') .OR. &
                 SameString(FldVal(sysBase + ddsSupplyFanConfigTypeOff),'DualFanVariableVolume'))
  !set the baseboard indicator
  IF (SameString(FldVal(base +  ddzBaseboardTypeOff),'None')) THEN
    baseboardKind = baseboardNone
  ELSEIF (SameString(FldVal(base +  ddzBaseboardTypeOff),'HotWater')) THEN
    baseboardKind = baseboardHotWater
  ELSEIF (SameString(FldVal(base +  ddzBaseboardTypeOff),'Electric')) THEN
    baseboardKind = baseboardElectric
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:DualDuct "'//TRIM(FldVal(base + ddzNameOff))//'"'// &
                    ' in the Baseboard Heating Type field: '//TRIM(FldVal(base + ddzBaseboardTypeOff)))
  END IF
  !validate cooling design method
  IF (SameString(FldVal(base +  ddzCoolDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  ddzCoolDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  ddzCoolDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:DualDuct "'//TRIM(FldVal(base + ddzNameOff))//'"'// &
                    ' in the Zone Cooling Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + ddzCoolDesignTempMethodOff)))
  END IF
  !validate heating design method
  IF (SameString(FldVal(base +  ddzHeatDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  ddzHeatDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  ddzHeatDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:DualDuct "'//TRIM(FldVal(base + ddzNameOff))//'"'// &
                    ' in the Zone Heating Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + ddzCoolDesignTempMethodOff)))
  END IF
  !ZONE CONTROL:THERMOSTATIC ~ line 5
  !repeat per zone - only if ddzTStatNameOff is not blank
  IF (.NOT. isTStatNameBlank) THEN
    CALL AddAlwaysSchedule('4')
    CALL CreateNewObj('ZoneControl:Thermostat')
    CALL AddToObjFld('Name', base +  ddzNameOff,' Thermostat')
    CALL AddToObjFld('Zone or ZoneList Name', base +  ddzNameOff,' ')
    !this schedule was previously created
    CALL AddToObjStr('Control Type Schedule Name','HVACTemplate-Always 4')
    CALL AddToObjStr('Control Object Type','ThermostatSetpoint:DualSetpoint')
    CALL AddToObjFld('Control Name', base + ddzTStatNameOff,' Dual SP Control',.TRUE.)
  END IF
  !ZONE SIZING   ~ line 12
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:OutdoorAir')
    CALL AddToObjStr('Name', 'SZ DSOA '//FldVal(base+ddzNameOff))
    CALL AddToObjFld('Outdoor Air Method', base + ddzOutAirTypeOff,' ')
    CALL AddToObjFld('Outdoor Air Flow per Person {m3/s}', base + ddzOutAirFlowPersonOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone Floor Area {m3/s-m2}', base + ddzOutAirFlowAreaOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone {m3/s}', base + ddzOutAirFlowZoneOff,'',.true.)
  END IF

  CALL CreateNewObj('Sizing:Zone')
  CALL AddToObjFld('Zone or ZoneList Name', base +  ddzNameOff,'')
  IF (SameString(FldVal(base +  ddzCoolDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
    CALL AddToObjStr('Zone Cooling Design Supply Air Temperature Input Method','SupplyAirTemperature')
    CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', sysBase + ddsCoolSetPtDesignOff,'')
  ELSE
    CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Input Method',base + ddzCoolDesignTempMethodOff,'')
    CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', base + ddzCoolDesignTempOff,'')
  ENDIF
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Difference {delta C}',base + ddzCoolDesignTempDiffOff,'')
  IF (SameString(FldVal(base +  ddzHeatDesignTempMethodOff),'SystemSupplyAirTemperature')) THEN
    CALL AddToObjStr('Zone Heating Design Supply Air Temperature Input Method','SupplyAirTemperature')
    CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', sysBase + ddsHeatSetPtDesignOff,'')
  ELSE
    CALL AddToObjFld('Zone Heating Design Supply Air Temperature Input Method',base + ddzHeatDesignTempMethodOff,'')
    CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', base + ddzHeatDesignTempOff,'')
  ENDIF
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Difference {delta C}',base + ddzHeatDesignTempDiffOff,'')
  CALL AddToObjStr('Zone Cooling Design Supply Air Humidity Ratio {kg-H20/kg-air}','0.008')
  CALL AddToObjStr('Zone Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + ddzDesSpecOANameSizOff,'')
  ELSE
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','SZ DSOA '//FldVal(base+ddzNameOff))
  END IF
  CALL AddToObjFld('Zone Heating Sizing Factor', base + ddzZoneHeatSizingFactorOff,'')
  CALL AddToObjFld('Zone Cooling Sizing Factor', base + ddzZoneCoolSizingFactorOff,'')
  IF (isSupplyMaxRateAutosize) THEN
    CALL AddToObjStr('Cooling Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Cooling Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  ELSE
    CALL AddToObjStr('Cooling Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Cooling Design Air Flow Rate {m3/s}', base + ddzSupplyMaxRateOff,' ')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  END IF
  IF (isSupplyMaxRateAutosize) THEN
    CALL AddToObjStr('Heating Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Heating Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
  ELSE
    CALL AddToObjStr('Heating Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Heating Design Air Flow Rate {m3/s}', base + ddzSupplyMaxRateOff,'')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
  END IF
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Zone Air Distribution Object Name', base + ddzDesSpecOADistNameOff,'',.true.)
  ELSE
    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+ddzNameOff),.true.)
  END IF
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:ZoneAirDistribution')
    CALL AddToObjStr('Name','SZ DSZAD '//FldVal(base+ddzNameOff))
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Cooling Mode {}','1')
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Heating Mode {}','1',.true.)
  END IF
  !CONTROLLED ZONE EQUIP CONFIGURATION  ~ line 41
  CALL CreateNewObj('ZoneHVAC:EquipmentConnections')
  CALL AddToObjFld('Zone Name', base +  ddzNameOff,'')
  CALL AddToObjFld('Zone Conditioning Equipment List Name', base +  ddzNameOff,' Equipment')
  CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base +  ddzNameOff,' Supply Inlet')
  CALL AddToObjStr('Zone Air Exhaust Node or NodeList Name','')
  CALL AddToObjFld('Zone Air Node Name', base +  ddzNameOff,' Zone Air Node')
  IF (isZoneRetPlenBlank) THEN
    CALL AddToObjFld('Zone Return Air Node Name', base +  ddzNameOff,' Return Outlet',.TRUE.)
  ELSE
    CALL AddToObjFld('Zone Return Air Node Name', base + ddzNameOff,' Zone Return Plenum Inlet',.TRUE.)
  END IF
  !ZONE SUPPLY PLENUM ~line 50
  IF (.NOT. isZoneHotSupPlenBlank) THEN
    CALL CreateNewObj('AirLoopHVAC:SupplyPlenum')
    CALL AddToObjFld('Name', base + ddzNameOff,' Hot Supply Plenum')
    CALL AddToObjFld('Zone Name', base + ddzZoneHotSupplyPlenumNameOff,'')
    CALL AddToObjFld('Zone Node Name', base + ddzZoneHotSupplyPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Inlet Node Name', base + ddzNameOff,' Zone Equip Hot Inlet')
    CALL AddToObjFld('Outlet Node Name', base + ddzNameOff,' Terminal Unit Hot Inlet',.TRUE.)
  END IF
  IF (.NOT. isZoneColdSupPlenBlank) THEN
    CALL CreateNewObj('AirLoopHVAC:SupplyPlenum')
    CALL AddToObjFld('Name', base + ddzNameOff,' Cold Supply Plenum')
    CALL AddToObjFld('Zone Name', base + ddzZoneColdSupplyPlenumNameOff,'')
    CALL AddToObjFld('Zone Node Name', base + ddzZoneColdSupplyPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Inlet Node Name', base + ddzNameOff,' Zone Equip Cold Inlet')
    CALL AddToObjFld('Outlet Node Name', base + ddzNameOff,' Terminal Unit Cold Inlet',.TRUE.)
  END IF
  !ZONE RETURN PLENUM ~ line 57
  IF (.NOT. isZoneRetPlenBlank) THEN
    CALL CreateNewObj('AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Name', base + ddzNameOff,' Return Plenum')
    CALL AddToObjFld('Zone Name', base + ddzZoneReturnPlenumNameOff,'')
    CALL AddToObjFld('Zone Node Name', base + ddzZoneReturnPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Outlet Node Name', base + ddzNameOff,' Return Outlet')
    CALL AddToObjStr('Induced Air Outlet Node or NodeList Name','')
    CALL AddToObjFld('Inlet Node Name', base + ddzNameOff,' Zone Return Plenum Inlet',.TRUE.)
  END IF
  !ZONE EQUIPMENT LIST ~ line 64
  CALL CreateNewObj('ZoneHVAC:EquipmentList')
  CALL AddToObjFld('Name', base +  ddzNameOff,' Equipment')
  CALL AddToObjStr('Load Distribution Scheme','SequentialLoad')
  CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:AirDistributionUnit')
  CALL AddToObjFld('Zone Equipment Name', base +  ddzNameOff,' ATU')
  CALL AddToObjStr('Zone Equipment Cooling Sequence','1')
  SELECT CASE (baseboardKind)
    CASE (baseboardNone)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1',.FALSE.)
    CASE (baseboardHotWater)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1')
      CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
      CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.FALSE.)
      CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Zone Equipment Name', base + ddzNameOff,' Baseboard Heat')
      CALL AddToObjStr('Zone Equipment Cooling Sequence','2')
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','2',.FALSE.)
    CASE (baseboardElectric)
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1')
      CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Zone Equipment Name', base + ddzNameOff,' Baseboard Heat')
      CALL AddToObjStr('Zone Equipment Cooling Sequence','2')
      CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','2',.FALSE.)
  END SELECT
  CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.TRUE.)
  !AIR DISTRIBUTION UNIT
  IF (isSystemVAV) THEN
    CALL CreateNewObj('ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Name', base +  ddzNameOff,' ATU')
    CALL AddToObjFld('Air Distribution Unit Outlet Node Name', base +  ddzNameOff,' Supply Inlet')
    CALL AddToObjStr('Air Terminal Object Type','AirTerminal:DualDuct:VAV')
    CALL AddToObjFld('Air Terminal Name', base +  ddzNameOff,' Dual Duct',.TRUE.)
    ! DualDuct:VariableVolume
    CALL CreateNewObj('AirTerminal:DualDuct:VAV')
    CALL AddToObjFld('Name', base +  ddzNameOff,' Dual Duct')
    CALL AddToObjStr('Availability Schedule Name','')
    CALL AddToObjFld('Air Outlet Node Name', base + ddzNameOff,' Supply Inlet')
    IF (isZoneHotSupPlenBlank) THEN
      CALL AddToObjFld('Hot Air Inlet Node Name', base + ddzNameOff,' Zone Equip Hot Inlet')
    ELSE
      CALL AddToObjFld('Hot Air Inlet Node Name', base + ddzNameOff,' Terminal Unit Hot Inlet')
    END IF
    IF (isZoneColdSupPlenBlank) THEN
      CALL AddToObjFld('Cold Air Inlet Node Name', base + ddzNameOff,' Zone Equip Cold Inlet')
    ELSE
      CALL AddToObjFld('Cold Air Inlet Node Name', base + ddzNameOff,' Terminal Unit Cold Inlet')
    END IF
    CALL AddToObjStr('Maximum Damper Air Flow Rate {m3/s}', 'autosize')
    CALL AddToObjFld('Zone Minimum Air Flow Fraction', base + ddzConstantMinFracOff,' ')
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + ddzDesSpecOANameCtrlOff,'',.TRUE.)
  ELSE
    CALL CreateNewObj('ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Name', base +  ddzNameOff,' ATU')
    CALL AddToObjFld('Air Distribution Unit Outlet Node Name', base +  ddzNameOff,' Supply Inlet')
    CALL AddToObjStr('Air Terminal Object Type','AirTerminal:DualDuct:ConstantVolume')
    CALL AddToObjFld('Air Terminal Name', base +  ddzNameOff,' Dual Duct',.TRUE.)
    ! DualDuct:VariableVolume
    CALL CreateNewObj('AirTerminal:DualDuct:ConstantVolume')
    CALL AddToObjFld('Name', base +  ddzNameOff,' Dual Duct')
    CALL AddToObjStr('Availability Schedule Name','')
    CALL AddToObjFld('Air Outlet Node Name', base + ddzNameOff,' Supply Inlet')
    IF (isZoneHotSupPlenBlank) THEN
      CALL AddToObjFld('Hot Air Inlet Node Name', base + ddzNameOff,' Zone Equip Hot Inlet')
    ELSE
      CALL AddToObjFld('Hot Air Inlet Node Name', base + ddzNameOff,' Terminal Unit Hot Inlet')
    END IF
    IF (isZoneColdSupPlenBlank) THEN
      CALL AddToObjFld('Cold Air Inlet Node Name', base + ddzNameOff,' Zone Equip Cold Inlet')
    ELSE
      CALL AddToObjFld('Cold Air Inlet Node Name', base + ddzNameOff,' Terminal Unit Cold Inlet')
    END IF
    CALL AddToObjStr('Maximum Air Flow Rate {m3/s}', 'autosize',.TRUE.)
  END IF
  !Baseboards
  SELECT CASE (baseboardKind)
    CASE (baseboardHotWater)
      !ZoneHVAC:Baseboard:RadiantConvective:Water
      CALL CreateNewObj('ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Name', base + ddzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + ddzBaseboardAvailSchedNameOff,'')
      CALL AddToObjFld('Inlet Node Name', base + ddzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Outlet Node Name', base + ddzNameOff,' Baseboard HW Outlet')
      CALL AddToObjStr('Rated Average Water Temperature {C}','87.78')
      CALL AddToObjStr('Rated Water Mass Flow Rate {m3/s}','0.063')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + ddzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Maximum Water Flow Rate','autosize')
      CALL AddToObjStr('Convergence Tolerance','0.001')
      CALL AddToObjStr('Fraction Radiant','0.0')
      CALL AddToObjStr('Fraction of Radiant Energy Incident on People','0.0',.TRUE.)
      !BRANCH (for baseboard)
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + ddzNameOff,' Baseboard Heat HW Branch')
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Component Name', base + ddzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Component Inlet Node Name', base + ddzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + ddzNameOff,' Baseboard HW Outlet',.TRUE.)
      CALL AddToStrList(TRIM(FldVal(base + ddzNameOff)) // ' Baseboard Heat HW Branch',handleHeatingCoilBranch)
      IF ((compactHotLoopBase .EQ. 0) .OR. (numCompactBoiler .EQ. 0)) THEN
        CALL WriteError('Both a HVACTemplate:Plant:HotWaterLoop and a HVACTemplate:Plant:Boiler are needed ' // &
             'when using hot water baseboards.')
      END IF
    CASE (baseboardElectric)
      !BASEBOARD HEATER:Electric:Convective
      CALL CreateNewObj('ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Name', base + ddzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + ddzBaseboardAvailSchedNameOff,'')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + ddzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Efficiency','1',.TRUE.)
  END SELECT
END DO
!======================================================================================================
! create Systems (dualduct can be VAV or ConstantVolume)
!======================================================================================================
DO iSys = 1, numCompactSysDualDuct
  base =  compactSysDualDuctbase(iSys)
  ! set flags
  isSupplyMaxRateAutosize = SameString(FldVal(base + ddsSupplyMaxRateOff),'autosize')
  isHotSupplyMaxRateAutosize = SameString(FldVal(base + ddsHotMaxRateOff),'autosize')
  isColdSupplyMaxRateAutosize = SameString(FldVal(base + ddsColdMaxRateOff),'autosize')
  isHeatCoilCapAutosize = SameString(FldVal(base + ddsHeatCoilCapOff),'autosize')
  isMinOARateAutosize = SameString(FldVal(base + ddsMinOutsideFlowOff),'autosize')
  isColdSupplyPlenumBlank = (FldVal(base + ddsColdSupplyPlenumNameOff) .EQ. '')
  isHotSupplyPlenumBlank = (FldVal(base + ddsHotSupplyPlenumNameOff) .EQ. '')
  isReturnPlenumBlank = (FldVal(base + ddsReturnPlenumNameOff) .EQ. '')
  isSysAvailSchedBlank = (FldVal(base + ddsSysAvailSchedNameOff) .EQ. '')
  isCoolStPtSchedBlank = (FldVal(base + ddsCoolSetPtSchedNameOff) .EQ. '')
  isHeatStPtSchedBlank = (FldVal(base + ddsHeatSetPtSchedNameOff) .EQ. '')
  isPreheatStPtSchedBlank = (FldVal(base + ddsPreheatSetPtSchedNameOff) .EQ. '')
  isEconoLowLimitBlank = (FldVal(base + ddsEconoLowLimitOff) .EQ. '')
  isReturnFanYes = (FldVal(base + ddsReturnFanPresentOff) .EQ. 'Yes')

  doSuppressMsgzlDualDuctAll = .FALSE.
  !check cooling coil kind
  IF (SameString(FldVal(base +  ddsCoolCoilKindOff),'ChilledWater')) THEN
    coolCoilType = ccChWater
  ELSEIF (SameString(FldVal(base + ddsCoolCoilKindOff),'ChilledWaterDetailedFlatModel')) THEN
    coolCoilType = ccChWaterDetailed
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct "'//TRIM(FldVal(base + ddsAirHandlerNameOff))//'"'// &
                    ' in the Cooling Coil Type field: '//TRIM(FldVal(base + ddsCoolCoilKindOff)))
  END IF
  !set heating coil indicator
  IF (SameString(FldVal(base + ddsHeatCoilKindOff),'HotWater')) THEN
    heatCoilType = ctHotWater
  ELSEIF (SameString(FldVal(base + ddsHeatCoilKindOff),'Electric')) THEN
    heatCoilType = ctElectric
  ELSEIF (SameString(FldVal(base + ddsHeatCoilKindOff),'Gas')) THEN
    heatCoilType = ctGas
  ELSEIF (SameString(FldVal(base + ddsHeatCoilKindOff),'None')) THEN
    heatCoilType = ctNone
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct in the Heating Coil Type field')
  END IF
  !set preheating coil indicator
  IF (SameString(FldVal(base +  ddsPreheatCoilKindOff),'HotWater')) THEN
    preHeatCoilType = ctHotWater
  ELSEIF (SameString(FldVal(base +  ddsPreheatCoilKindOff),'Electric')) THEN
    preHeatCoilType = ctElectric
  ELSEIF (SameString(FldVal(base +  ddsPreheatCoilKindOff),'Gas')) THEN
    preHeatCoilType = ctGas
  ELSEIF (SameString(FldVal(base +  ddsPreheatCoilKindOff),'None')) THEN
    preHeatCoilType = ctNone
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct in the Preheat Coil Type field')
  END IF
  !validate the Minimum Outside Air Control Type
  CALL validChoice(base + ddsMinOutsideControlOff,'FixedMinimum','ProportionalMinimum',  &
     'SYSTEM:DualDuct','Minimum Outdoor Air Control Type')
  !set the economizer indicator
  IF (SameString(FldVal(base +  ddsEconomizerKindOff),'FixedDryBulb')) THEN
    econoKind = etFixedDryBulb
  ELSEIF (SameString(FldVal(base +  ddsEconomizerKindOff),'FixedEnthalpy')) THEN
    econoKind = etFixedEnthalpy
  ELSEIF (SameString(FldVal(base +  ddsEconomizerKindOff),'DifferentialDryBulb')) THEN
    econoKind = etDifferentialDryBulb
  ELSEIF (SameString(FldVal(base +  ddsEconomizerKindOff),'DifferentialEnthalpy')) THEN
    econoKind = etDifferentialEnthalpy
  ELSEIF (SameString(FldVal(base +  ddsEconomizerKindOff),'FixedDewPointAndDryBulb')) THEN
    econoKind = etFixedDewPointAndDryBulb
  ELSEIF (SameString(FldVal(base +  ddsEconomizerKindOff),'ElectronicEnthalpy')) THEN
    econoKind = etElectronicEnthalpy
  ELSEIF (SameString(FldVal(base +  ddsEconomizerKindOff),'DifferentialDryBulbAndEnthalpy')) THEN
    econoKind = etDifferentialDryBulbAndEnthalpy
  ELSEIF (SameString(FldVal(base +  ddsEconomizerKindOff),'NoEconomizer')) THEN
    econoKind = etNoEconomizer
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct in the Economizer Type field')
  END IF
  !set the economizer lockout indicator
  IF (SameString(FldVal(base +  ddsEconoLockoutKindOff),'NoLockout')) THEN
    lockoutKind = loNo
  ELSEIF (SameString(FldVal(base +  ddsEconoLockoutKindOff),'')) THEN
    lockoutKind = loNo
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct in the Economizer Lockout field')
  END IF
  !check cooling coil setpoint control type
  IF (SameString(FldVal(base +  ddsCoolSetPtCtrlTypeOff),'FixedSetpoint')) THEN
    coolSetPtCtrlKind = cspFixed
  ELSEIF (SameString(FldVal(base + ddsCoolSetPtCtrlTypeOff),'Scheduled')) THEN
    coolSetPtCtrlKind = cspScheduled
  ELSEIF (SameString(FldVal(base + ddsCoolSetPtCtrlTypeOff),'OutdoorAirTemperatureReset')) THEN
    coolSetPtCtrlKind = cspOAreset
  ELSEIF (SameString(FldVal(base + ddsCoolSetPtCtrlTypeOff),'Warmest')) THEN
    coolSetPtCtrlKind = cspWarmest
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct  "'//TRIM(FldVal(base + ddsAirHandlerNameOff))//'"'// &
                    ' in the Cooling Coil Setpoint Control Type field: '//TRIM(FldVal(base + ddsCoolSetPtCtrlTypeOff)))
  END IF
  !check heating coil setpoint control type
  IF (SameString(FldVal(base +  ddsHeatSetPtCtrlTypeOff),'FixedSetpoint')) THEN
    heatSetPtCtrlKind = hspFixed
  ELSEIF (SameString(FldVal(base + ddsHeatSetPtCtrlTypeOff),'Scheduled')) THEN
    heatSetPtCtrlKind = hspScheduled
  ELSEIF (SameString(FldVal(base + ddsHeatSetPtCtrlTypeOff),'OutdoorAirTemperatureReset')) THEN
    heatSetPtCtrlKind = hspOAreset
  ELSEIF (SameString(FldVal(base + ddsHeatSetPtCtrlTypeOff),'Coldest')) THEN
    heatSetPtCtrlKind = hspColdest
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct  "'//TRIM(FldVal(base + ddsAirHandlerNameOff))//'"'// &
                    ' in the Heating Coil Setpoint Control Type field: '//TRIM(FldVal(base + ddsHeatSetPtCtrlTypeOff)))
  END IF
  !set the Supply Fan Part-Load Power Coefficients indicator
  IF (SameString(FldVal(base + ddsSupplyFanPLRCoeffOff),'InletVaneDampers')) THEN
    supFanCoef = sfcInletVane
  ELSEIF (SameString(FldVal(base + ddsSupplyFanPLRCoeffOff),'OutletDampers')) THEN
    supFanCoef = sfcDischgDamp
  ELSEIF (SameString(FldVal(base + ddsSupplyFanPLRCoeffOff),'VariableSpeedMotor')) THEN
    supFanCoef = sfcVSD
  ELSEIF (SameString(FldVal(base + ddsSupplyFanPLRCoeffOff),'ASHRAE90.1-2004AppendixG')) THEN
    supFanCoef = sfcAppG
  ELSEIF (SameString(FldVal(base + ddsSupplyFanPLRCoeffOff),'VariableSpeedMotorPressureReset')) THEN
    supFanCoef = sfcVSDPresReset
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct  "'//TRIM(FldVal(base + ddsAirHandlerNameOff))//'"'// &
                    ' in the Supply Fan Part-Load Power Coefficients field: '//TRIM(FldVal(base + ddsSupplyFanPLRCoeffOff)))
    supFanCoef = sfcInletVane
  END IF
  !set the Cold Supply Fan Part-Load Power Coefficients indicator
  IF (SameString(FldVal(base + ddsColdFanPLRCoeffOff),'InletVaneDampers')) THEN
    coldSupFanCoef = sfcInletVane
  ELSEIF (SameString(FldVal(base + ddsColdFanPLRCoeffOff),'OutletDampers')) THEN
    coldSupFanCoef = sfcDischgDamp
  ELSEIF (SameString(FldVal(base + ddsColdFanPLRCoeffOff),'VariableSpeedMotor')) THEN
    coldSupFanCoef = sfcVSD
  ELSEIF (SameString(FldVal(base + ddsColdFanPLRCoeffOff),'ASHRAE90.1-2004AppendixG')) THEN
    coldSupFanCoef = sfcAppG
  ELSEIF (SameString(FldVal(base + ddsColdFanPLRCoeffOff),'VariableSpeedMotorPressureReset')) THEN
    coldSupFanCoef = sfcVSDPresReset
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct  "'//TRIM(FldVal(base + ddsAirHandlerNameOff))//'"'// &
                    ' in the Cold Supply Fan Part-Load Power Coefficients field: '//TRIM(FldVal(base + ddsColdFanPLRCoeffOff)))
    coldSupFanCoef = sfcInletVane
  END IF
  !set the Hot Supply Fan Part-Load Power Coefficients indicator
  IF (SameString(FldVal(base + ddsHotFanPLRCoeffOff),'InletVaneDampers')) THEN
    hotSupFanCoef = sfcInletVane
  ELSEIF (SameString(FldVal(base + ddsHotFanPLRCoeffOff),'OutletDampers')) THEN
    hotSupFanCoef = sfcDischgDamp
  ELSEIF (SameString(FldVal(base + ddsHotFanPLRCoeffOff),'VariableSpeedMotor')) THEN
    hotSupFanCoef = sfcVSD
  ELSEIF (SameString(FldVal(base + ddsHotFanPLRCoeffOff),'ASHRAE90.1-2004AppendixG')) THEN
    hotSupFanCoef = sfcAppG
  ELSEIF (SameString(FldVal(base + ddsHotFanPLRCoeffOff),'VariableSpeedMotorPressureReset')) THEN
    hotSupFanCoef = sfcVSDPresReset
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct  "'//TRIM(FldVal(base + ddsAirHandlerNameOff))//'"'// &
                    ' in the Hot Supply Fan Part-Load Power Coefficients field: '//TRIM(FldVal(base + ddsHotFanPLRCoeffOff)))
    hotSupFanCoef = sfcInletVane
  END IF
  !set the Return Fan Part-Load Power Coefficients indicator
  IF (SameString(FldVal(base + ddsReturnFanPLRCoeffOff),'InletVaneDampers')) THEN
    retFanCoef = sfcInletVane
  ELSEIF (SameString(FldVal(base + ddsReturnFanPLRCoeffOff),'OutletDampers')) THEN
    retFanCoef = sfcDischgDamp
  ELSEIF (SameString(FldVal(base + ddsReturnFanPLRCoeffOff),'VariableSpeedMotor')) THEN
    retFanCoef = sfcVSD
  ELSEIF (SameString(FldVal(base + ddsReturnFanPLRCoeffOff),'ASHRAE90.1-2004AppendixG')) THEN
    retFanCoef = sfcAppG
  ELSEIF (SameString(FldVal(base + ddsReturnFanPLRCoeffOff),'VariableSpeedMotorPressureReset')) THEN
    retFanCoef = sfcVSDPresReset
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct in the Return Fan Part-Load Power Coefficients field')
    retFanCoef = sfcInletVane
  END IF
  !set the Heat Recovery Type indicator
  IF (SameString(FldVal(base + ddsHeatRecTypeOff),'Sensible')) THEN
    heatRecovery = htrecSens
  ELSEIF (SameString(FldVal(base + ddsHeatRecTypeOff),'Enthalpy')) THEN
    heatRecovery = htrecEnth
  ELSEIF (SameString(FldVal(base + ddsHeatRecTypeOff),'None')) THEN
    heatRecovery = htrecNone
  ELSE
    heatRecovery = htrecNone
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct in the Heating Coil Setpoint Reset Type field')
  ENDIF
  ! set the fan placement indicators
  IF (SameString(FldVal(base + ddsColdFanPlaceOff), 'BlowThrough')) THEN
    coldSupFanPlacement = sfpBlowThru
  ELSEIF (SameString(FldVal(base + ddsColdFanPlaceOff), 'DrawThrough')) THEN
    coldSupFanPlacement = sfpDrawThru
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct "'//TRIM(FldVal(base + ddsAirHandlerNameOff))//'"'// &
                    ' in the Cold Supply Fan Placement field: '//TRIM(FldVal(base + ddsColdFanPlaceOff)))
  END IF
  IF (SameString(FldVal(base + ddsHotFanPlaceOff), 'BlowThrough')) THEN
    hotSupFanPlacement = sfpBlowThru
  ELSEIF (SameString(FldVal(base + ddsHotFanPlaceOff), 'DrawThrough')) THEN
    hotSupFanPlacement = sfpDrawThru
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct "'//TRIM(FldVal(base + ddsAirHandlerNameOff))//'"'// &
                    ' in the Hot Supply Fan Placement field: '//TRIM(FldVal(base + ddsHotFanPlaceOff)))
  END IF
  ! set the fan configuration indicators
  IF (SameString(FldVal(base + ddsSupplyFanConfigTypeOff), 'SingleFanConstantVolume')) THEN
    isSystemVAV = .FALSE.
    supFanKind=sfkCAV
    coldFanKind=sfkNone
    hotFanKind=sfkNone
    coldSupFanPlacement = sfpNone
    hotSupFanPlacement  = sfpNone
  ELSEIF (SameString(FldVal(base + ddsSupplyFanConfigTypeOff), 'DualFanConstantVolume')) THEN
    isSystemVAV = .FALSE.
    supFanKind=sfkNone
    coldFanKind=sfkCAV
    hotFanKind=sfkCAV
  ELSEIF (SameString(FldVal(base + ddsSupplyFanConfigTypeOff), 'SingleFanVariableVolume')) THEN
    isSystemVAV = .TRUE.
    supFanKind=sfkVAV
    coldFanKind=sfkNone
    hotFanKind=sfkNone
    coldSupFanPlacement = sfpNone
    hotSupFanPlacement  = sfpNone
  ELSEIF (SameString(FldVal(base + ddsSupplyFanConfigTypeOff), 'DualFanVariableVolume')) THEN
    isSystemVAV = .TRUE.
    supFanKind=sfkNone
    coldFanKind=sfkVAV
    hotFanKind=sfkVAV
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct "'//TRIM(FldVal(base + ddsAirHandlerNameOff))//'"'// &
                    ' in the System Configuration Type field: '//TRIM(FldVal(base + ddsSupplyFanConfigTypeOff)))
  END IF
  !set the dehumidification indicators
  isDehumidifyNone = SameString(FldVal(base + ddsDehumCtrlTypeOff),'None')
  IF (SameString(FldVal(base +  ddsDehumCtrlTypeOff),'None')) THEN
    dehumidCtrlKind = dehumidNone
  ELSEIF (SameString(FldVal(base +  ddsDehumCtrlTypeOff),'CoolReheat')) THEN
    dehumidCtrlKind = dehumidCoolRht
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct "'//TRIM(FldVal(base + ddsAirHandlerNameOff))//'"'// &
                    ' in the Dehumidification Control Type field: '//TRIM(FldVal(base + ddsDehumCtrlTypeOff)))
  END IF
  !set the humidifier indicators
  isHumidifierNone = SameString(FldVal(base + ddsHumidCtrlTypeOff),'None')
  IF (SameString(FldVal(base +  ddsHumidCtrlTypeOff),'None')) THEN
    humidifierKind = humidifyNone
  ELSEIF (SameString(FldVal(base +  ddsHumidCtrlTypeOff),'ElectricSteam')) THEN
    humidifierKind = humidifyElecSteam
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct "'//TRIM(FldVal(base + ddsAirHandlerNameOff))//'"'// &
                    ' in the Humidifier Type field: '//TRIM(FldVal(base + ddsHumidCtrlTypeOff)))
  END IF

  !set node name for return inlet to OA mixer
  IF (isReturnFanYes) THEN
    returnInletToOAMIxer = TRIM(FldVal(base + ddsAirHandlerNameOff))//' Return Fan Outlet'
  ELSE
    returnInletToOAMIxer = TRIM(FldVal(base + ddsAirHandlerNameOff))//' Air Loop Inlet'
  END IF
  ! check if there is no heating coil and heating design setpoint is > preheat design setpoint
  IF (heatCoilType == ctNone) THEN
    IF(StringToReal(FldVal(base + ddsHeatSetPtDesignOff)) .GT. StringToReal(FldVal(base + ddsPreheatSetPtConstantOff))) THEN
      usePreheatTempforHeatDesignSuppT = .TRUE.
      CALL WriteError('Warning:  In HVACTemplate:System:DualDuct "'//TRIM(FldVal(base + ddsAirHandlerNameOff))//'"'// &
                    ' the Heating Coil Design Setpoint is greater than the Preheat Coil Design Setpoint,'// &
                    ' but Heating Coil Type=None. Using Preheat Coil Design Setpoint for the Sizing:System'// &
                    ' Central Heating Design Supply Air Temperature.',msgWarning)
    ENDIF
    IF (preHeatCoilType == ctNone) THEN
      usePreheatTempforHeatDesignSuppT = .TRUE.
      CALL WriteError('Warning:  In HVACTemplate:System:DualDuct "'//TRIM(FldVal(base + ddsAirHandlerNameOff))//'"'// &
                    ' there is no Heating Coil and no Preheat Coil.'// &
                    ' The Preheat Coil Design Setpoint will be used for the Sizing:System'// &
                    ' Central Heating Design Supply Air Temperature. This will be the inlet air temperature for sizing'// &
                    ' reheat coils.',msgWarning)
    ENDIF
  ENDIF


  ! validate sizing option input
  IF (SameString(FldVal(base +  ddsSizingOptionOff),'Coincident')) THEN
  ELSEIF (SameString(FldVal(base +  ddsSizingOptionOff),'NonCoincident')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DualDuct "'//TRIM(FldVal(base + ddsAirHandlerNameOff))//'"'// &
                    ' in the Sizing Option field: '//TRIM(FldVal(base + ddsSizingOptionOff)))
  END IF

  !set component inlet and outlet node names for main, hot and cold ducts, including system name from ddsAirHandlerNameOff as prefix

  ! For main branch first component will always be AirLoopHVAC:OutdoorAirSystem with outlet node name <ddsAirHandlerNameOff> Mixed Air Outlet
  lastOutlet           = TRIM(FldVal(base + ddsAirHandlerNameOff))//' Mixed Air Outlet'

  ! main supply fan
  IF (supFanKind .NE. sfkNone) THEN
    supfanInlet        = TRIM(lastOutlet)
    supfanOutlet       = TRIM(FldVal(base + ddsAirHandlerNameOff))//' Main Supply Fan Outlet'
    lastOutlet         = TRIM(supfanOutlet)
  END IF
  mainbranchOutlet     = TRIM(lastOutlet)

  ! For cold branch inlet node will always be <ddsAirHandlerNameOff> Cold Duct Inlet
  lastOutlet           = TRIM(FldVal(base + ddsAirHandlerNameOff))//' Cold Duct Inlet'
  coldbranchInlet      = TRIM(lastOutlet)
  ! blow thru cold supply fan
  IF (coldSupFanPlacement .EQ. sfpBlowThru) THEN
    coldfanInlet       = TRIM(lastOutlet)
    coldfanOutlet      = TRIM(FldVal(base + ddsAirHandlerNameOff))//' Cold Supply Fan Outlet'
    lastOutlet         = TRIM(coldfanOutlet)
  END IF
  ! cooling coil
  IF (coolCoilType .NE. ccNone) THEN
    coolCoilInlet      = TRIM(lastOutlet)
    coolCoilOutlet     = TRIM(FldVal(base + ddsAirHandlerNameOff))//' Cooling Coil Outlet'
    lastOutlet         = TRIM(coolCoilOutlet)
  END IF
  ! draw thru cold supply fan
  IF (coldSupFanPlacement .EQ. sfpDrawThru) THEN
    coldfanInlet       = TRIM(lastOutlet)
    coldfanOutlet      = TRIM(FldVal(base + ddsAirHandlerNameOff))//' Cold Supply Fan Outlet'
    lastOutlet         = TRIM(coldfanOutlet)
  END IF
  ! set AirLoopHVAC Supply Side Cold Outlet Node Name
  coldbranchOutlet     = TRIM(lastOutlet)

  ! For hot branch inlet node will always be <ddsAirHandlerNameOff> Hot Duct Inlet
  lastOutlet           = TRIM(FldVal(base + ddsAirHandlerNameOff))//' Hot Duct Inlet'
  hotbranchInlet       = TRIM(lastOutlet)
  ! blow thru hot supply fan
  IF (hotSupFanPlacement .EQ. sfpBlowThru) THEN
    hotfanInlet        = TRIM(lastOutlet)
    hotfanOutlet       = TRIM(FldVal(base + ddsAirHandlerNameOff))//' Hot Supply Fan Outlet'
    lastOutlet         = TRIM(hotfanOutlet)
  END IF

  IF (heatCoilType .NE. ctNone) THEN
    heatCoilInlet      = TRIM(lastOutlet)
    heatCoilOutlet     = TRIM(FldVal(base + ddsAirHandlerNameOff))//' Heating Coil Outlet'
    lastOutlet         = TRIM(heatCoilOutlet)
  END IF
  ! draw thru hot supply fan
  IF (hotSupFanPlacement .EQ. sfpDrawThru) THEN
    hotfanInlet           = TRIM(lastOutlet)
    hotfanOutlet          = TRIM(FldVal(base + ddsAirHandlerNameOff))//' Hot Supply Fan Outlet'
    lastOutlet         = TRIM(hotfanOutlet)
  END IF
  ! humidifier
  IF (.NOT. isHumidifierNone) THEN
    humidifierInlet    = TRIM(lastOutlet)
    humidifierOutlet   = TRIM(FldVal(base + ddsAirHandlerNameOff))//' Humidifier Outlet'
    lastOutlet         = TRIM(humidifierOutlet)
  END IF
  ! set AirLoopHVAC Supply Side Hot Outlet Node Name
  hotbranchOutlet      = TRIM(lastOutlet)

  !SYSTEM SIZING  ~ line 158
  CALL CreateNewObj('Sizing:System')
  CALL AddToObjFld('AirLoop Name', base + ddsAirHandlerNameOff,' ')
  CALL AddToObjStr('Type of Load to Size On','Sensible')
  CALL AddToObjFld('Design Outdoor Air Flow Rate {m3/s}', base + ddsMinOutsideFlowOff,' ')
  !This line if ddsSupplyMaxRateOff=autosize or ddsSupplyMinRateOff=autosize
  IF (coldFanKind .EQ. sfkVAV) THEN
    CALL AddToObjFld('Central Heating Maximum System Air Flow Ratio',base + ddsColdMinFracOff,'')
  ELSEIF (supFanKind .EQ. sfkVAV) THEN
    CALL AddToObjFld('Central Heating Maximum System Air Flow Ratio',base + ddsSupplyMinFracOff,'')
  ELSE
    CALL AddToObjStr('Central Heating Maximum System Air Flow Ratio', '1.0')
  END IF
  CALL AddToObjFld('Preheat Design Temperature {C}', base + ddsPreheatSetPtConstantOff,'')
  CALL AddToObjStr('Preheat Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjStr('Precool Design Temperature {C}','11')
  CALL AddToObjStr('Precool Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjFld('Central Cooling Design Supply Air Temperature {C}', base + ddsCoolSetPtDesignOff,'')
  IF (usePreheatTempforHeatDesignSuppT) THEN
    CALL AddToObjFld('Central Heating Design Supply Air Temperature {C}', base + ddsPreheatSetPtConstantOff,'')
  ELSE
    CALL AddToObjFld('Central Heating Design Supply Air Temperature {C}', base + ddsHeatSetPtDesignOff,'')
  ENDIF
  CALL AddToObjFld('Sizing Option', base + ddsSizingOptionOff,'')
  CALL AddToObjStr('100% Outdoor Air in Cooling','No')
  CALL AddToObjStr('100% Outdoor Air in Heating','No')
  CALL AddToObjStr('Central Cooling Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjStr('Central Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isSupplyMaxRateAutosize) THEN
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','DesignDay')
    CALL AddToObjStr('Cooling Supply Air Flow Rate {m3/s}','0')
  ELSE
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','Flow/System')
    CALL AddToObjFld('Cooling Supply Air Flow Rate {m3/s}', base + ddsSupplyMaxRateOff,' ')
  END IF
  ! New fields for v8.2 next three rows
  CALL AddToObjStr('Cooling Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
  CALL AddToObjStr('Cooling Fraction of Autosized Cooling Supply Air Flow Rate','')
  CALL AddToObjStr('Cooling Supply Air Flow Rate Per Unit Cooling Capacity {m3/s-W}','')
  CALL AddToObjStr('Heating Supply Air Flow Rate Method','DesignDay')
  CALL AddToObjStr('Heating Supply Air Flow Rate {m3/s}','0')
  ! New fields for v8.2 next four rows
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
  CALL AddToObjStr('Heating Fraction of Autosized Heating Supply Air Flow Rate','')
  CALL AddToObjStr('Heating Fraction of Autosized Cooling Supply Air Flow Rate','')
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Unit Heating Capacity {m3/s-W}','')
  CALL AddToObjStr('System Outdoor Air Method','ZoneSum')
  CALL AddToObjStr('Zone Maximum Outdoor Air Fraction','1')
  CALL AddToObjStr('Cooling Design Capacity Method','CoolingDesignCapacity')
  CALL AddToObjStr('Cooling Design Capacity {W}','autosize')
  CALL AddToObjStr('Cooling Design Capacity Per Floor Area {W/m2}','')
  CALL AddToObjStr('Fraction of Autosized Cooling Design Capacity','')
  CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
  CALL AddToObjStr('Heating Design Capacity {W}','autosize')
  CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
  CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
  CALL AddToObjStr('Central Cooling Capacity Control Method','OnOff',.true.)
  !AIR PRIMARY LOOP ~ line 184
  CALL CreateNewObj('AirLoopHVAC')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' ')
  CALL AddToObjFld('Controller List Name', base + ddsAirHandlerNameOff,' Controllers')
  CALL AddToObjFld('Availability Manager List Name', base + ddsAirHandlerNameOff,' Availability Managers')
  CALL AddToObjStr('Design Supply Air Flow Rate {m3/s}','autosize')
  CALL AddToObjFld('Branch List Name', base + ddsAirHandlerNameOff,' Branches')
  CALL AddToObjFld('Connector List Name', base + ddsAirHandlerNameOff,' Connectors')
  CALL AddToObjFld('Supply Side Inlet Node Name', base + ddsAirHandlerNameOff,' Air Loop Inlet')
  CALL AddToObjFld('Demand Side Outlet Node Name', base + ddsAirHandlerNameOff,' Return Air Outlet')
  CALL AddToObjFld('Demand Side Inlet Node Names', base + ddsAirHandlerNameOff,' Supply Path Inlets')
  CALL AddToObjFld('Supply Side Outlet Node Names', base + ddsAirHandlerNameOff,' Supply Side Outlets',.TRUE.)

  ! Node lists for dual duct loop connections
  CALL CreateNewObj('NodeList')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Supply Path Inlets')
  CALL AddToObjFld('Node Name 1', base + ddsAirHandlerNameOff,' Hot Supply Path Inlet')
  CALL AddToObjFld('Node Name 2', base + ddsAirHandlerNameOff,' Cold Supply Path Inlet',.TRUE.)

  CALL CreateNewObj('NodeList')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Supply Side Outlets')
  CALL AddToObjStr('Node Name 1', TRIM(hotbranchOutlet))
  CALL AddToObjStr('Node Name 2', TRIM(coldbranchOutlet),.TRUE.)

  !CONTROLLER LIST ~ line 196       Updated section on Sept 28, 2006 to rearrange coil order for CR7094
  CALL CreateNewObj('AirLoopHVAC:ControllerList')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Controllers')
  !if ddsHeatCoilKindOff=Hot Water
  IF (heatCoilType .EQ. ctHotWater) THEN
    CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
    CALL AddToObjFld('Controller Name', base + ddsAirHandlerNameOff,' Heating Coil Controller',(coolCoilType .EQ. ctNone))
  ENDIF
  IF (coolCoilType .NE. ctNone) THEN
    CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
    CALL AddToObjFld('Controller Name', base + ddsAirHandlerNameOff,' Cooling Coil Controller',.TRUE.)
  END IF
  !BRANCH LIST
  CALL CreateNewObj('BranchList')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Branches')
  CALL AddToObjFld('Branch 1 Name', base + ddsAirHandlerNameOff,' Main Branch')
  CALL AddToObjFld('Branch 2 Name', base + ddsAirHandlerNameOff,' Hot Branch')
  CALL AddToObjFld('Branch 3 Name', base + ddsAirHandlerNameOff,' Cold Branch',.TRUE.)

  !CONNECTOR LIST
  CALL CreateNewObj('ConnectorList')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Connectors')
  CALL AddToObjStr('Connector 1 Object Type','Connector:Splitter')
  CALL AddToObjFld('Connector 1 Name', base + ddsAirHandlerNameOff,' Duct Splitter',.TRUE.)

  !SPLITTER
  CALL CreateNewObj('Connector:Splitter')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Duct Splitter')
  CALL AddToObjFld('Inlet Branch Name', base + ddsAirHandlerNameOff,' Main Branch')
  CALL AddToObjFld('Outlet Branch 1 Name', base + ddsAirHandlerNameOff,' Hot Branch')
  CALL AddToObjFld('Outlet Branch 2 Name', base + ddsAirHandlerNameOff,' Cold Branch',.TRUE.)

    !MAIN BRANCH
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Main Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  IF (isReturnFanYes) THEN
    IF (isSystemVAV) THEN
      CALL AddToObjStr('Component Object Type','Fan:VariableVolume')
    ELSE
      CALL AddToObjStr('Component Object Type','Fan:ConstantVolume')
    END IF
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Return Fan')
    CALL AddToObjFld('Component Inlet Node Name', base + ddsAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + ddsAirHandlerNameOff,' Return Fan Outlet')
  END IF
  CALL AddToObjStr('Component Object Type','AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' OA System')
  CALL AddToObjStr('Component Inlet Node Name', TRIM(returnInletToOAMIxer))
  CALL AddToObjFld('Component Outlet Node Name', base + ddsAirHandlerNameOff,' Mixed Air Outlet',(supFanKind .EQ. sfkNone))
  IF (supFanKind .NE. sfkNone) THEN
    IF (supFanKind .EQ. sfkVAV) THEN
      CALL AddToObjStr('Component Object Type','Fan:VariableVolume')
    ELSE
      CALL AddToObjStr('Component Object Type','Fan:ConstantVolume')
    END IF
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Main Supply Fan')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(supfanInlet))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(supfanOutlet),.TRUE.)
  END IF

  !HOT BRANCH
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Hot Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  IF (hotSupFanPlacement .EQ. sfpBlowThru) THEN
    IF (hotFanKind .EQ. sfkVAV) THEN
      CALL AddToObjStr('Component Object Type','Fan:VariableVolume')
    ELSE
      CALL AddToObjStr('Component Object Type','Fan:ConstantVolume')
    END IF
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Hot Supply Fan')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(hotfanInlet))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(hotfanOutlet),(isHumidifierNone .AND. heatCoilType == ctNone))
  END IF
  IF (heatCoilType .NE. ctNone) THEN
    IF (heatCoilType .EQ. ctHotWater) THEN
      CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    ELSEIF (heatCoilType .EQ. ctElectric) THEN
      CALL AddToObjStr('Component Object Type','Coil:Heating:Electric')
    ELSEIF (heatCoilType .EQ. ctGas) THEN
      CALL AddToObjStr('Component Object Type','Coil:Heating:Fuel')
    END IF
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(heatCoilInlet))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(heatCoilOutlet), &
                                             (isHumidifierNone .AND. (hotSupFanPlacement .NE. sfpDrawThru)))
  END IF
  IF (humidifierKind .EQ. humidifyElecSteam) THEN
    CALL AddToObjStr('Component Object Type','Humidifier:Steam:Electric')
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Humidifier')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(humidifierInlet))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(humidifierOutlet), &
                                            (hotSupFanPlacement .NE. sfpDrawThru))
  END IF
  IF (hotSupFanPlacement .EQ. sfpDrawThru) THEN
    IF (hotFanKind .EQ. sfkVAV) THEN
      CALL AddToObjStr('Component Object Type','Fan:VariableVolume')
    ELSE
      CALL AddToObjStr('Component Object Type','Fan:ConstantVolume')
    END IF
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Hot Supply Fan')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(hotfanInlet))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(hotfanOutlet),.TRUE.)
  END IF

  !COLD BRANCH
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cold Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  IF (coldSupFanPlacement .EQ. sfpBlowThru) THEN
    IF (coldFanKind .EQ. sfkVAV) THEN
      CALL AddToObjStr('Component Object Type','Fan:VariableVolume')
    ELSE
      CALL AddToObjStr('Component Object Type','Fan:ConstantVolume')
    END IF
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Cold Supply Fan')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(coldfanInlet))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(coldfanOutlet),(coolCoilType == ctNone))
  END IF
  IF (coolCoilType .NE. ctNone) THEN
    IF (coolCoilType .EQ. ccChWater) THEN
      CALL AddToObjStr('Component Object Type','Coil:Cooling:Water')
    ELSEIF (coolCoilType .EQ. ccChWaterDetailed) THEN
      CALL AddToObjStr('Component Object Type','Coil:Cooling:Water:DetailedGeometry')
    END IF
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Cooling Coil')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(coolCoilInlet))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(coolCoilOutlet),(coldSupFanPlacement .NE. sfpDrawThru))
  END IF
  IF (coldSupFanPlacement .EQ. sfpDrawThru) THEN
    IF (coldFanKind .EQ. sfkVAV) THEN
      CALL AddToObjStr('Component Object Type','Fan:VariableVolume')
    ELSE
      CALL AddToObjStr('Component Object Type','Fan:ConstantVolume')
    END IF
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Cold Supply Fan')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(coldfanInlet))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(coldfanOutlet),.TRUE.)
  END IF

    !HOT SUPPLY AIR PATH
  IF (isHotSupplyPlenumBlank) THEN
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Hot Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + ddsAirHandlerNameOff,' Hot Supply Path Inlet')
    CALL MakeSpecialIntList(zlDualDuctHotSupPlen, base + ddsAirHandlerNameOff)
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneSplitter')
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Hot Zone Splitter',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Hot Zone Splitter')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Hot Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Hot Supply Plenum', .TRUE.)
      END IF
    END IF
    !ZONE SPLITTER ~ line 244
    CALL CreateNewObj('AirLoopHVAC:ZoneSplitter')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Hot Zone Splitter')
    CALL AddToObjFld('Inlet Node Name', base + ddsAirHandlerNameOff,' Hot Supply Path Inlet')
    CALL MakeSpecialIntList(zlDualDuctAll,base + ddsAirHandlerNameOff)
    ! add logic so error message is not repeated.
    IF (numIntList .EQ. 0) THEN
      doSuppressMsgzlDualDuctAll = .TRUE.
    END IF
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(i),' Zone Equip Hot Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Hot Inlet',.TRUE.)
    END IF
  ELSE
    !ZONE SUPPLY AIR PATH ~ line 250
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Hot Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + ddsAirHandlerNameOff,' Hot Supply Path Inlet')
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
    CALL MakeSpecialIntList(zlDualDuctHotSupPlen, base + ddsAirHandlerNameOff)
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Hot Supply Plenum',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Hot Supply Plenum')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Hot Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Hot Supply Plenum', .TRUE.)
      END IF
    END IF
    !ZONE SUPPLY PLENUM ~ line 258
    CALL CreateNewObj('AirLoopHVAC:SupplyPlenum')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Hot Supply Plenum')
    CALL AddToObjFld('Zone Name', base + ddsHotSupplyPlenumNameOff,' ')
    CALL AddToObjFld('Zone Node Name', base + ddsHotSupplyPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Inlet Node Name', base + ddsAirHandlerNameOff,' Hot Supply Path Inlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlDualDuctAll,base + ddsAirHandlerNameOff)
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(i),' Zone Equip Hot Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Hot Inlet',.TRUE.)
    END IF
  END IF

    !COLD SUPPLY AIR PATH
  IF (isColdSupplyPlenumBlank) THEN
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cold Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + ddsAirHandlerNameOff,' Cold Supply Path Inlet')
    CALL MakeSpecialIntList(zlDualDuctColdSupPlen, base + ddsAirHandlerNameOff)
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneSplitter')
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Cold Zone Splitter',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Cold Zone Splitter')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Cold Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Cold Supply Plenum', .TRUE.)
      END IF
    END IF
    !ZONE SPLITTER ~ line 244
    CALL CreateNewObj('AirLoopHVAC:ZoneSplitter')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cold Zone Splitter')
    CALL AddToObjFld('Inlet Node Name', base + ddsAirHandlerNameOff,' Cold Supply Path Inlet')
    CALL MakeSpecialIntList(zlDualDuctAll,base + ddsAirHandlerNameOff)
    ! add logic so error message is not repeated.
    IF (numIntList .EQ. 0) THEN
      doSuppressMsgzlDualDuctAll = .TRUE.
    END IF
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(i),' Zone Equip Cold Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Cold Inlet',.TRUE.)
    END IF
  ELSE
    !ZONE SUPPLY AIR PATH ~ line 250
    CALL CreateNewObj('AirLoopHVAC:SupplyPath')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cold Supply Path')
    CALL AddToObjFld('Supply Air Path Inlet Node Name', base + ddsAirHandlerNameOff,' Cold Supply Path Inlet')
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
    CALL MakeSpecialIntList(zlDualDuctColdSupPlen, base + ddsAirHandlerNameOff)
    IF (numIntList .EQ. 0) THEN
      CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Cold Supply Plenum',.TRUE.)
    ELSE
      CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Cold Supply Plenum')
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList - 1
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Cold Supply Plenum')
      END DO
      IF (numIntList .GT. 0) THEN
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:SupplyPlenum')
        CALL AddToObjFld('Component Name', IntList(numIntList),' Cold Supply Plenum', .TRUE.)
      END IF
    END IF
    !ZONE SUPPLY PLENUM ~ line 258
    CALL CreateNewObj('AirLoopHVAC:SupplyPlenum')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cold Supply Plenum')
    CALL AddToObjFld('Zone Name', base + ddsColdSupplyPlenumNameOff,' ')
    CALL AddToObjFld('Zone Node Name', base + ddsColdSupplyPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Inlet Node Name', base + ddsAirHandlerNameOff,' Cold Supply Path Inlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlDualDuctAll,base + ddsAirHandlerNameOff)
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Outlet Node Name', IntList(i),' Zone Equip Cold Inlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Outlet Node Name', IntList(numIntList),' Zone Equip Cold Inlet',.TRUE.)
    END IF
  END IF

   ! RETURN PATH
  IF (isReturnPlenumBlank) THEN
   !ZONE RETURN AIR PATH ~ line 266
    CALL CreateNewObj('AirLoopHVAC:ReturnPath')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Return Path')
    CALL AddToObjFld('Return Air Path Outlet Node Name', base + ddsAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlDualDuctAllRetPlen, base + ddsAirHandlerNameOff)
    IF (numIntList .NE. 0) THEN
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Return Plenum')
      END DO
    END IF
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneMixer')
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Zone Mixer',.TRUE.)
    !ZONE MIXER ~ line 274
    CALL CreateNewObj('AirLoopHVAC:ZoneMixer')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Zone Mixer')
    CALL AddToObjFld('Outlet Node Name', base + ddsAirHandlerNameOff,' Return Air Outlet')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlDualDuctAll,base + ddsAirHandlerNameOff,doSuppressMsgzlDualDuctAll)
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Inlet Node Name', IntList(i),' Return Outlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Inlet Node Name', IntList(numIntList),' Return Outlet',.TRUE.)
    END IF
  ELSE
    !ZONE RETURN AIR PATH ~ line 280
    CALL CreateNewObj('AirLoopHVAC:ReturnPath')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Return Path')
    CALL AddToObjFld('Return Air Path Outlet Node Name', base + ddsAirHandlerNameOff,' Return Air Outlet')
    CALL MakeSpecialIntList(zlDualDuctAllRetPlen, base + ddsAirHandlerNameOff)
    IF (numIntList .NE. 0) THEN
      !the IntList holds the list of pointers to the zone names
      DO iInt = 1,numIntList
        CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
        CALL AddToObjFld('Component Name', IntList(iInt),' Return Plenum')
      END DO
    END IF
    CALL AddToObjStr('Component Object Type','AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Return Plenum',.TRUE.)
    !ZONE RETURN PLENUM ~ line 288
    CALL CreateNewObj('AirLoopHVAC:ReturnPlenum')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Return Plenum')
    CALL AddToObjFld('Zone Name', base + ddsReturnPlenumNameOff,' ')
    CALL AddToObjFld('Zone Node Name', base + ddsReturnPlenumNameOff,' Zone Air Node')
    CALL AddToObjFld('Outlet Node Name', base + ddsAirHandlerNameOff,' Return Air Outlet')
    CALL AddToObjStr('Induced Air Outlet Node or NodeList Name','')
    !the IntList holds the list of pointers to the zone names
    CALL MakeSpecialIntList(zlDualDuctAll,base + ddsAirHandlerNameOff)
    DO i = 1,numIntList - 1
      CALL AddToObjFld('Inlet Node Name', IntList(i),' Return Outlet')
    END DO
    IF (numIntList .GT. 0) THEN
      CALL AddToObjFld('Inlet Node Name', IntList(numIntList),' Return Outlet',.TRUE.)
    END IF
  END IF

  !SYSTEM AVAILABILITY MANAGER LIST
  CALL CreateNewObj('AvailabilityManagerAssignmentList')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Availability Managers')
  CALL AddToObjStr('Availability Manager Object Type','AvailabilityManager:NightCycle')
  CALL AddToObjFld('Availability Manager Name', base + ddsAirHandlerNameOff,' Availability',.TRUE.)
  !SYSTEM AVAILABILITY MANAGER:NIGHT CYCLE
  CALL CreateNewObj('AvailabilityManager:NightCycle')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Availability')
  CALL AddToObjStr('Applicability Schedule Name','HVACTemplate-Always 1')
  IF (isSysAvailSchedBlank) THEN
    CALL AddToObjStr('Fan Schedule Name','HVACTemplate-Always 1')
  ELSE
    CALL AddToObjFld('Fan Schedule Name', base + ddsSysAvailSchedNameOff,'')
  END IF
  CALL AddToObjFld('Control Type', base + ddsNightCycleControlOff,'')
  CALL AddToObjStr('Thermostat Tolerance {deltaC}','0.2')
  CALL AddToObjStr('Cycling Run Time Control Type', 'FixedRunTime')
  CALL AddToObjStr('Cycling run time {s}','3600')
  CALL AddToObjFld('Control Zone Name', base + ddsNightCycleCtrlZoneNameOff,'',.TRUE.)
  CALL AddAlwaysSchedule('1')

  ! Cooling coil setpoint manager
  IF (coolCoilType .NE. ccNone) THEN
    CALL CreateNewObj('NodeList')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cooling Setpoint Nodes')
    CALL AddToObjFld('Node Name 1', base + ddsAirHandlerNameOff,' Mixed Air Outlet')
    CALL AddToObjStr('Node Name 2', TRIM(coolCoilOutlet),.TRUE.)
    IF ((coolSetPtCtrlKind .EQ. cspFixed) .OR. (coolSetPtCtrlKind .EQ. cspScheduled)) THEN
      IF ((coolSetPtCtrlKind .EQ. cspFixed) .OR. (isCoolStPtSchedBlank)) THEN
        CALL AddAlwaysSchedule(FldVal(base + ddsCoolSetPtDesignOff))
      END IF
      !***SetpointManager:Scheduled
      CALL CreateNewObj('SetpointManager:Scheduled')
      CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      IF ((coolSetPtCtrlKind .EQ. cspFixed) .OR. (isCoolStPtSchedBlank)) THEN
        CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + ddsCoolSetPtDesignOff)))
      ELSE
        CALL AddToObjFld('Schedule Name', base + ddsCoolSetPtSchedNameOff,' ')
      END IF
    ELSE IF (coolSetPtCtrlKind .EQ. cspOAreset) THEN
      !***SetpointManager:OutdoorAirReset
      CALL CreateNewObj('SetpointManager:OutdoorAirReset')
      CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('Setpoint at Outdoor Low Temperature {C}', base + ddsCoolSetPtAtODBLoOff,'')
      CALL AddToObjFld('Outdoor Low Temperature {C}', base + ddsCoolResetODBLoOff,'')
      CALL AddToObjFld('Setpoint at Outdoor High Temperature {C}', base + ddsCoolSetPtAtODBHiOff,'')
      CALL AddToObjFld('Outdoor High Temperature {C}', base + ddsCoolResetODBHiOff,'')
    ELSE IF (coolSetPtCtrlKind .EQ. cspWarmest) THEN
      !SET POINT MANAGER:WARMEST
      CALL CreateNewObj('SetpointManager:Warmest')
      CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('HVAC Air Loop Name', base + ddsAirHandlerNameOff,'')
      CALL AddToObjFld('Minimum Setpoint Temperature {C}', base + ddsCoolSetPtDesignOff,'')
      IF (18.0 .GT. (StringToReal(FldVal(base + ddsCoolSetPtDesignOff)) + 5.2)) THEN
        CALL AddToObjStr('Maximum Setpoint Temperature {C}','18')
      ELSE
        CALL AddToObjNum('Maximum Supply Air Temperature {C}',StringToReal(FldVal(base + ddsCoolSetPtDesignOff)) + 5.2)
      END IF
      CALL AddToObjStr('Strategy','MaximumTemperature')
    END IF
    IF (coldSupFanPlacement .EQ. sfpDrawThru) THEN
      CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(coldbranchOutlet),.TRUE.)
    ELSE
      CALL AddToObjFld('Setpoint Node or NodeList Name', base + ddsAirHandlerNameOff,' Cooling Setpoint Nodes',.TRUE.)
    END IF
    IF (coldSupFanPlacement .EQ. sfpDrawThru) THEN
      !***SetpointManager:MixedAir
      CALL CreateNewObj('SetpointManager:MixedAir')
      CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cooling Coil Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjStr('Reference Setpoint Node Name', TRIM(coldbranchOutlet))
      CALL AddToObjStr('Fan Inlet Node Name', TRIM(coldfanInlet))
      CALL AddToObjStr('Fan Outlet Node Name', TRIM(coldfanOutlet))
      CALL AddToObjFld('Setpoint Node or NodeList Name', base + ddsAirHandlerNameOff,' Cooling Setpoint Nodes',.TRUE.)
    END IF
  END IF

  IF (coolCoilType .EQ. ccChWaterDetailed) THEN
    !COIL:Water:DetailedFlatCooling
    CALL CreateNewObj('Coil:Cooling:Water:DetailedGeometry')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cooling Coil')
    CALL AddToObjFld('Availability Schedule Name', base + ddsCoolAvailSchedNameOff,' ')
    CALL AddToObjStr('Max Water Flow Rate of Coil {m3/s}','autosize')
    CALL AddToObjStr('Tube Outside Surf Area {m2}','autosize')
    CALL AddToObjStr('Total Tube Inside Area {m2}','autosize')
    CALL AddToObjStr('Fin Surface Area {m2}','autosize')
    CALL AddToObjStr('Minimum Air Flow Area {m2}','autosize')
    CALL AddToObjStr('Coil Depth {m}','autosize')
    CALL AddToObjStr('Fin Diameter {m}','autosize')
    CALL AddToObjStr('Fin Thickness {m}','0.0015')
    CALL AddToObjStr('Tube Inside Diameter {m}','0.01445')
    CALL AddToObjStr('Tube Outside Diameter {m}','0.0159')
    CALL AddToObjStr('Tube Thermal Conductivity {W/m-K}','386.0')
    CALL AddToObjStr('Fin Thermal Conductivity {W/m-K}','204.0')
    CALL AddToObjStr('Fin Spacing {m}','0.0018')
    CALL AddToObjStr('Tube Depth Spacing {m}','0.026')
    CALL AddToObjStr('Number of Tube Rows','4')
    CALL AddToObjStr('Number of Tubes per Row','autosize')
    CALL AddToObjFld('Coil Water Inlet Node', base + ddsAirHandlerNameOff,' Cooling Coil ChW Inlet')
    CALL AddToObjFld('Coil Water Outlet Node', base + ddsAirHandlerNameOff,' Cooling Coil ChW Outlet')
    CALL AddToObjStr('Air Inlet Node Name', TRIM(coolCoilInlet))
    CALL AddToObjStr('Air Outlet Node Name',TRIM(coolCoilOutlet),.TRUE.)
  ELSEIF (coolCoilType .EQ. ccChWater) THEN
    !COIL:Water:Cooling ~ line 399
    CALL CreateNewObj('Coil:Cooling:Water')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cooling Coil')
    CALL AddToObjFld('Availability Schedule Name', base + ddsCoolAvailSchedNameOff,' ')
    CALL AddToObjStr('Design Water Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Design Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Design Inlet Water Temperature {C}','autosize')
    CALL AddToObjStr('Design Inlet Air Temperature {C}','autosize')
    CALL AddToObjStr('Design Outlet Air Temperature {C}','autosize')
    CALL AddToObjStr('Design Inlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
    CALL AddToObjStr('Design Outlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
    CALL AddToObjFld('Water Inlet Node Name', base + ddsAirHandlerNameOff,' Cooling Coil ChW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + ddsAirHandlerNameOff,' Cooling Coil ChW Outlet')
    CALL AddToObjStr('Air Inlet Node Name', TRIM(coolCoilInlet))
    CALL AddToObjStr('Air Outlet Node Name',TRIM(coolCoilOutlet))
    CALL AddToObjStr('Type of Analysis','DetailedAnalysis')
    CALL AddToObjStr('Heat Exchanger Configuration','CrossFlow',.TRUE.)
  END IF
  !CONTROLLER:SIMPLE ~ line 419
  CALL CreateNewObj('Controller:WaterCoil')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cooling Coil Controller')
  IF (isDehumidifyNone) THEN
    CALL AddToObjStr('Control Variable','Temperature')
  ELSE
    CALL AddToObjStr('Control Variable','TemperatureAndHumidityRatio')
  END IF
  CALL AddToObjStr('Action','Reverse')
  CALL AddToObjStr('Actuator Variable','Flow')
  CALL AddToObjStr('Sensor Node Name', TRIM(coolCoilOutlet))
  CALL AddToObjFld('Actuator Node Name', base + ddsAirHandlerNameOff,' Cooling Coil ChW Inlet')
  CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
  CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
  CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)

  ! Humdistat(s) if needed
  ! first check if the schedule names are used or if values should be used.
  IF (.NOT. isDehumidifyNone) THEN
    IF (FldVal(base + ddsDehumSetPtSchedNameOff) .EQ. '') THEN
      dehumscheduleName = 'HVACTemplate-Always ' // TRIM(FldVal(base + ddsDehumSetPtOff))
      CALL AddAlwaysSchedule(FldVal(base + ddsDehumSetPtOff))
    ELSE
      dehumscheduleName = TRIM(FldVal(base + ddsDehumSetPtSchedNameOff))
    ENDIF
  END IF
  IF (.NOT. isHumidifierNone) THEN
    ! first check if the schedule name is used or if value should be used.
    IF (FldVal(base + ddsHumidSetPtSchedNameOff) .EQ. '') THEN
      humscheduleName = 'HVACTemplate-Always ' // TRIM(FldVal(base + ddsHumidSetPtOff))
      CALL AddAlwaysSchedule(FldVal(base + ddsHumidSetPtOff))
    ELSE
      humscheduleName = TRIM(FldVal(base + ddsHumidSetPtSchedNameOff))
    ENDIF
  END IF

  !    Single humidistat if humidification and dehumidification are both active and control zones are the same
  IF ((.NOT. isHumidifierNone) .AND. (.NOT. isDehumidifyNone) &
       .AND. SameString(FldVal(base +  ddsDehumCtrlZoneOff),FldVal(base +  ddsHumidCtrlZoneOff))) THEN
    CALL CreateNewObj('ZoneControl:Humidistat')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Humidistat')
    CALL AddToObjFld('Zone Name', base + ddsDehumCtrlZoneOff,'')
    CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name', TRIM(humscheduleName))
    CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name', TRIM(dehumscheduleName),.TRUE.)
  ELSE
    IF (.NOT. isDehumidifyNone) THEN
  !   Dehumidification humidistat
      CALL CreateNewObj('ZoneControl:Humidistat')
      CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Dehumidification Humidistat')
      CALL AddToObjFld('Zone Name', base + ddsDehumCtrlZoneOff,'')
      CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name','HVACTemplate-Always 1')
      CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name', TRIM(dehumscheduleName),.TRUE.)

      CALL AddAlwaysSchedule('1')
    ENDIF
    IF (.NOT. isHumidifierNone) THEN
  !    Humidification humidistat
      CALL CreateNewObj('ZoneControl:Humidistat')
      CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Humidification Humidistat')
      CALL AddToObjFld('Zone Name', base + ddsHumidCtrlZoneOff,'')
      CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name', TRIM(humscheduleName))
      CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name','HVACTemplate-Always 100',.TRUE.)

      CALL AddAlwaysSchedule('100')
    ENDIF
  ENDIF

  !Set point manager for dehumidification
  IF (.NOT. isDehumidifyNone) THEN
    CALL CreateNewObj('SetpointManager:SingleZone:Humidity:Maximum')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Dehumidification Setpoint Manager')
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + ddsAirHandlerNameOff,' Cooling Coil Outlet')
    CALL AddToObjFld('Control Zone Air Node Name', base + ddsDehumCtrlZoneOff,' Zone Air Node',.TRUE.)
  END IF

  !BRANCH ~ line 396
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cooling Coil ChW Branch')
  !store for later in the chilled water loop
  CALL AddToStrList(TRIM(FldVal(base + ddsAirHandlerNameOff)) // ' Cooling Coil ChW Branch', handleCoolingCoilBranch)
  CALL AddToObjStr('Pressure Drop Curve Name','')
  IF (coolCoilType .EQ. ccChWater) THEN
    CALL AddToObjStr('Component Object Type','Coil:Cooling:Water')
  ELSEIF (coolCoilType .EQ. ccChWaterDetailed) THEN
    CALL AddToObjStr('Component Object Type','Coil:Cooling:Water:DetailedGeometry')
  END IF
  CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Cooling Coil')
  CALL AddToObjFld('Component Inlet Node Name', base + ddsAirHandlerNameOff,' Cooling Coil ChW Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + ddsAirHandlerNameOff,' Cooling Coil ChW Outlet',.TRUE.)
  IF (heatCoilType .EQ. ctHotWater) THEN
    !COIL:Water:SimpleHeating ~ line 406
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + ddsHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name ', base + ddsAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + ddsAirHandlerNameOff,' Heating Coil HW Outlet')
    CALL AddToObjStr('Air Inlet Node Name',  TRIM(heatCoilInlet))
    CALL AddToObjStr('Air Outlet Node Name',  TRIM(heatCoilOutlet))
    IF (isHeatCoilCapAutosize) THEN
      CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
      CALL AddToObjStr('Rated Capacity','')
    ELSE
      CALL AddToObjStr('Performance Input Method','NominalCapacity')
      CALL AddToObjFld('Rated Capacity',base + ddsHeatCoilCapOff,'')
    END IF
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
    !BRANCH ~ line 416
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Heating Coil HW Branch')
    !save this branch name for later use in plan
    CALL AddToStrList(TRIM(FldVal(base + ddsAirHandlerNameOff)) // ' Heating Coil HW Branch',handleHeatingCoilBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + ddsAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + ddsAirHandlerNameOff,' Heating Coil HW Outlet',.TRUE.)
    !CONTROLLER:SIMPLE ~line 425
    CALL CreateNewObj('Controller:WaterCoil')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Heating Coil Controller')
    CALL AddToObjStr('Control Variable','Temperature')
    CALL AddToObjStr('Action','Normal')
    CALL AddToObjStr('Actuator Variable','Flow')
    CALL AddToObjStr('Sensor Node Name', TRIM(heatCoilOutlet))
    CALL AddToObjFld('Actuator Node Name', base + ddsAirHandlerNameOff,' Heating Coil HW Inlet')
    CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
    CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
    CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)
  END IF
  !COIL:Electric:Heating ~ line 437
  IF (heatCoilType .EQ. ctElectric) THEN
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + ddsHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Efficiency','1')
    CALL AddToObjFld('Nominal Capacity {W}',base + ddsHeatCoilCapOff,'')
    CALL AddToObjStr('Air Inlet Node Name',  TRIM(heatCoilInlet))
    CALL AddToObjStr('Air Outlet Node Name',  TRIM(heatCoilOutlet))
    CALL AddToObjStr('Temperature Setpoint Node Name', TRIM(heatCoilOutlet),.TRUE.)
  END IF
  IF (heatCoilType .EQ. ctGas) THEN
    !COIL:Gas:Heating ~ line 448
    CALL CreateNewObj('Coil:Heating:Fuel')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + ddsHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Fuel Type', 'Gas')
    CALL AddToObjFld('Gas Burner Efficiency', base + ddsHeatEffiencyOff,' ')
    CALL AddToObjFld('Nominal Capacity {W}',base + ddsHeatCoilCapOff,'')
    CALL AddToObjStr('Air Inlet Node Name',  TRIM(heatCoilInlet))
    CALL AddToObjStr('Air Outlet Node Name',  TRIM(heatCoilOutlet))
    CALL AddToObjStr('Temperature Setpoint Node Name', TRIM(heatCoilOutlet))
    CALL AddToObjFld('Parasitic Electric Load {W}', base + ddsHeatParasiticOff,' ')
    CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
       base + ddsAirHandlerNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
    !CURVE:CUBIC ~ line 461
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  END IF
  ! Heating coil setpoint manager
  IF (heatCoilType .NE. ctNone) THEN
    IF ((heatSetPtCtrlKind .EQ. hspFixed) .OR. (heatSetPtCtrlKind .EQ. hspScheduled)) THEN
      IF ((heatSetPtCtrlKind .EQ. hspFixed) .OR. (isheatStPtSchedBlank)) THEN
        CALL AddAlwaysSchedule(FldVal(base + ddsHeatSetPtDesignOff))
      END IF
      !***SetpointManager:Scheduled
      CALL CreateNewObj('SetpointManager:Scheduled')
      CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Heating Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      IF ((heatSetPtCtrlKind .EQ. hspFixed) .OR. (isheatStPtSchedBlank)) THEN
        CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + ddsHeatSetPtDesignOff)))
      ELSE
        CALL AddToObjFld('Schedule Name', base + ddsHeatSetPtSchedNameOff,'')
      END IF
    ELSE IF (heatSetPtCtrlKind .EQ. hspOAreset) THEN
      !***SetpointManager:OutdoorAirReset
      CALL CreateNewObj('SetpointManager:OutdoorAirReset')
      CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Heating Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('Setpoint at Outdoor Low Temperature {C}', base + ddsHeatSetPtAtODBLoOff,'')
      CALL AddToObjFld('Outdoor Low Temperature {C}', base + ddsHeatResetODBLoOff,'')
      CALL AddToObjFld('Setpoint at Outdoor High Temperature {C}', base + ddsHeatSetPtAtODBHiOff,'')
      CALL AddToObjFld('Outdoor High Temperature {C}', base + ddsHeatResetODBHiOff,'')
    ELSE IF (heatSetPtCtrlKind .EQ. hspColdest) THEN
      !***SetpointManager:Coldest
      CALL CreateNewObj('SetpointManager:Coldest')
      CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Heating Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('HVAC Air Loop Name', base + ddsAirHandlerNameOff,'')
      IF (20.0 .LT. (StringToReal(FldVal(base + ddsHeatSetPtDesignOff)) - 5.2)) THEN
        CALL AddToObjStr('Minimum Setpoint Temperature {C}','20.0')
      ELSE
        CALL AddToObjNum('Minimum Setpoint Temperature {C}',StringToReal(FldVal(base + ddsHeatSetPtDesignOff)) - 5.2)
      END IF
      CALL AddToObjFld('Maximum Setpoint Temperature {C}', base + ddsHeatSetPtDesignOff,'')
      CALL AddToObjStr('Strategy','MinimumTemperature')
    END IF
    IF (hotsupFanPlacement .EQ. sfpDrawThru) THEN
      CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(hotbranchOutlet),.TRUE.)
    ELSE
      CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(heatCoilOutlet),.TRUE.)
    END IF
    ! Only needed if system is draw through and there is a heating coil
    IF (hotsupFanPlacement .EQ. sfpDrawThru) THEN
      !***SetpointManager:MixedAir
      CALL CreateNewObj('SetpointManager:MixedAir')
      CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Heating Coil Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjStr('Reference Setpoint Node Name', TRIM(hotbranchOutlet))
      CALL AddToObjStr('Fan Inlet Node Name', TRIM(hotfanInlet))
      CALL AddToObjStr('Fan Outlet Node Name', TRIM(hotfanOutlet))
      CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(heatCoilOutlet),.TRUE.)
    END IF
  END IF

  ! Return fan
  IF (isReturnFanYes .AND. isSystemVAV) THEN
    CALL CreateNewObj('Fan:VariableVolume')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Return Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + ddsSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + ddsReturnEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + ddsReturnPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
    CALL AddToObjStr('Fan Power Minimum Flow Rate Input Method','Fraction')
    CALL AddToObjStr('Fan Power Minimum Flow Fraction','0.0')
    CALL AddToObjStr('Fan Power Minimum Air Flow Rate {m3/s}', '')
    CALL AddToObjFld('Motor Efficiency', base + ddsReturnFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + ddsReturnFanFractionOff,' ')
    SELECT CASE (retFanCoef)
      CASE (sfcInletVane)
        CALL AddToObjStr('Fan Power Coefficient 1','0.35071223')
        CALL AddToObjStr('Fan Power Coefficient 2','0.30850535')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.54137364')
        CALL AddToObjStr('Fan Power Coefficient 4','0.87198823')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcDischgDamp)
        CALL AddToObjStr('Fan Power Coefficient 1','0.37073425')
        CALL AddToObjStr('Fan Power Coefficient 2','0.97250253')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.34240761')
        CALL AddToObjStr('Fan Power Coefficient 4','0')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcVSD)
        CALL AddToObjStr('Fan Power Coefficient 1','0.0015302446')
        CALL AddToObjStr('Fan Power Coefficient 2','0.0052080574')
        CALL AddToObjStr('Fan Power Coefficient 3','1.1086242')
        CALL AddToObjStr('Fan Power Coefficient 4','-0.11635563')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcAppG)
        CALL AddToObjStr('Fan Power Coefficient 1','0.0013')
        CALL AddToObjStr('Fan Power Coefficient 2','0.147')
        CALL AddToObjStr('Fan Power Coefficient 3','0.9506')
        CALL AddToObjStr('Fan Power Coefficient 4','-0.0998')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcVSDPresReset) !based on Good Static Pressure Reset ddsD Fan from EDR_DesignGuidelines_VAV.pdf
        CALL AddToObjStr('Fan Power Coefficient 1', '0.040759894')
        CALL AddToObjStr('Fan Power Coefficient 2', '0.08804497')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.07292612')
        CALL AddToObjStr('Fan Power Coefficient 4', '0.943739823')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
    END SELECT
    CALL AddToObjFld('Air Inlet Node Name', base + ddsAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Air Outlet Node Name', base + ddsAirHandlerNameOff,' Return Fan Outlet',.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF
  IF (isReturnFanYes .AND. .NOT. isSystemVAV) THEN
    CALL CreateNewObj('Fan:ConstantVolume')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Return Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + ddsSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + ddsReturnEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + ddsReturnPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
    CALL AddToObjFld('Motor Efficiency', base + ddsReturnFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + ddsReturnFanFractionOff,' ')
    CALL AddToObjFld('Air Inlet Node Name', base + ddsAirHandlerNameOff,' Air Loop Inlet')
    CALL AddToObjFld('Air Outlet Node Name', base + ddsAirHandlerNameOff,' Return Fan Outlet',.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF

  !Main Supply Fan
  IF (supFanKind .EQ. sfkVAV) THEN
    CALL CreateNewObj('Fan:VariableVolume')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Main Supply Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + ddsSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + ddsSupplyEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + ddsSupplyPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
    CALL AddToObjStr('Fan Power Minimum Flow Rate Input Method','Fraction')
    CALL AddToObjFld('Fan Power Minimum Flow Fraction', base + ddsSupplyMinFracOff,'')
    CALL AddToObjStr('Fan Power Minimum Air Flow Rate {m3/s}', '')
    CALL AddToObjFld('Motor Efficiency', base + ddsSupplyFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + ddsSupplyFanFractionOff,' ')
    SELECT CASE (supFanCoef)
      CASE (sfcInletVane)
        CALL AddToObjStr('Fan Power Coefficient 1','0.35071223')
        CALL AddToObjStr('Fan Power Coefficient 2','0.30850535')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.54137364')
        CALL AddToObjStr('Fan Power Coefficient 4','0.87198823')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcDischgDamp)
        CALL AddToObjStr('Fan Power Coefficient 1','0.37073425')
        CALL AddToObjStr('Fan Power Coefficient 2','0.97250253')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.34240761')
        CALL AddToObjStr('Fan Power Coefficient 4','0')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcVSD)
        CALL AddToObjStr('Fan Power Coefficient 1','0.0015302446')
        CALL AddToObjStr('Fan Power Coefficient 2','0.0052080574')
        CALL AddToObjStr('Fan Power Coefficient 3','1.1086242')
        CALL AddToObjStr('Fan Power Coefficient 4','-0.11635563')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcAppG)
        CALL AddToObjStr('Fan Power Coefficient 1','0.0013')
        CALL AddToObjStr('Fan Power Coefficient 2','0.147')
        CALL AddToObjStr('Fan Power Coefficient 3','0.9506')
        CALL AddToObjStr('Fan Power Coefficient 4','-0.0998')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcVSDPresReset) !based on Good Static Pressure Reset ddsD Fan from EDR_DesignGuidelines_VAV.pdf
        CALL AddToObjStr('Fan Power Coefficient 1', '0.040759894')
        CALL AddToObjStr('Fan Power Coefficient 2', '0.08804497')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.07292612')
        CALL AddToObjStr('Fan Power Coefficient 4', '0.943739823')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
    END SELECT
    CALL AddToObjStr('Air Inlet Node Name', TRIM(supfanInlet))
    CALL AddToObjStr('Air Outlet Node Name', TRIM(supfanOutlet),.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF
  IF (supFanKind .EQ. sfkCAV) THEN
    CALL CreateNewObj('Fan:ConstantVolume')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Main Supply Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + ddsSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + ddsSupplyEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + ddsSupplyPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
    CALL AddToObjFld('Motor Efficiency', base + ddsSupplyFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + ddsSupplyFanFractionOff,' ')
    CALL AddToObjStr('Air Inlet Node Name', TRIM(supfanInlet))
    CALL AddToObjStr('Air Outlet Node Name', TRIM(supfanOutlet),.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF

  !Hot Supply Fan
  IF (hotFanKind .EQ. sfkVAV) THEN
    CALL CreateNewObj('Fan:VariableVolume')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Hot Supply Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + ddsSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + ddsHotEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + ddsHotPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
    CALL AddToObjStr('Fan Power Minimum Flow Rate Input Method','Fraction')
    CALL AddToObjFld('Fan Power Minimum Flow Fraction', base + ddsHotMinFracOff,'')
    CALL AddToObjStr('Fan Power Minimum Air Flow Rate {m3/s}', '')
    CALL AddToObjFld('Motor Efficiency', base + ddsHotFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + ddsHotFanFractionOff,' ')
    SELECT CASE (hotsupFanCoef)
      CASE (sfcInletVane)
        CALL AddToObjStr('Fan Power Coefficient 1','0.35071223')
        CALL AddToObjStr('Fan Power Coefficient 2','0.30850535')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.54137364')
        CALL AddToObjStr('Fan Power Coefficient 4','0.87198823')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcDischgDamp)
        CALL AddToObjStr('Fan Power Coefficient 1','0.37073425')
        CALL AddToObjStr('Fan Power Coefficient 2','0.97250253')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.34240761')
        CALL AddToObjStr('Fan Power Coefficient 4','0')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcVSD)
        CALL AddToObjStr('Fan Power Coefficient 1','0.0015302446')
        CALL AddToObjStr('Fan Power Coefficient 2','0.0052080574')
        CALL AddToObjStr('Fan Power Coefficient 3','1.1086242')
        CALL AddToObjStr('Fan Power Coefficient 4','-0.11635563')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcAppG)
        CALL AddToObjStr('Fan Power Coefficient 1','0.0013')
        CALL AddToObjStr('Fan Power Coefficient 2','0.147')
        CALL AddToObjStr('Fan Power Coefficient 3','0.9506')
        CALL AddToObjStr('Fan Power Coefficient 4','-0.0998')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcVSDPresReset) !based on Good Static Pressure Reset ddsD Fan from EDR_DesignGuidelines_VAV.pdf
        CALL AddToObjStr('Fan Power Coefficient 1', '0.040759894')
        CALL AddToObjStr('Fan Power Coefficient 2', '0.08804497')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.07292612')
        CALL AddToObjStr('Fan Power Coefficient 4', '0.943739823')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
    END SELECT
    CALL AddToObjStr('Air Inlet Node Name', TRIM(hotfanInlet))
    CALL AddToObjStr('Air Outlet Node Name', TRIM(hotfanOutlet),.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF
  IF (hotFanKind .EQ. sfkCAV) THEN
    CALL CreateNewObj('Fan:ConstantVolume')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Hot Supply Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + ddsSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + ddsHotEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + ddsHotPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
    CALL AddToObjFld('Motor Efficiency', base + ddsHotFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + ddsHotFanFractionOff,' ')
    CALL AddToObjStr('Air Inlet Node Name', TRIM(hotfanInlet))
    CALL AddToObjStr('Air Outlet Node Name', TRIM(hotfanOutlet),.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF

  !Cold Supply Fan
  IF (coldFanKind .EQ. sfkVAV) THEN
    CALL CreateNewObj('Fan:VariableVolume')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cold Supply Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + ddsSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + ddsColdEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + ddsColdPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
    CALL AddToObjStr('Fan Power Minimum Flow Rate Input Method','Fraction')
    CALL AddToObjFld('Fan Power Minimum Flow Fraction', base + ddsColdMinFracOff,'')
    CALL AddToObjStr('Fan Power Minimum Air Flow Rate {m3/s}', '')
    CALL AddToObjFld('Motor Efficiency', base + ddsColdFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + ddsColdFanFractionOff,' ')
    SELECT CASE (coldsupFanCoef)
      CASE (sfcInletVane)
        CALL AddToObjStr('Fan Power Coefficient 1','0.35071223')
        CALL AddToObjStr('Fan Power Coefficient 2','0.30850535')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.54137364')
        CALL AddToObjStr('Fan Power Coefficient 4','0.87198823')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcDischgDamp)
        CALL AddToObjStr('Fan Power Coefficient 1','0.37073425')
        CALL AddToObjStr('Fan Power Coefficient 2','0.97250253')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.34240761')
        CALL AddToObjStr('Fan Power Coefficient 4','0')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcVSD)
        CALL AddToObjStr('Fan Power Coefficient 1','0.0015302446')
        CALL AddToObjStr('Fan Power Coefficient 2','0.0052080574')
        CALL AddToObjStr('Fan Power Coefficient 3','1.1086242')
        CALL AddToObjStr('Fan Power Coefficient 4','-0.11635563')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcAppG)
        CALL AddToObjStr('Fan Power Coefficient 1','0.0013')
        CALL AddToObjStr('Fan Power Coefficient 2','0.147')
        CALL AddToObjStr('Fan Power Coefficient 3','0.9506')
        CALL AddToObjStr('Fan Power Coefficient 4','-0.0998')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
      CASE (sfcVSDPresReset) !based on Good Static Pressure Reset ddsD Fan from EDR_DesignGuidelines_VAV.pdf
        CALL AddToObjStr('Fan Power Coefficient 1', '0.040759894')
        CALL AddToObjStr('Fan Power Coefficient 2', '0.08804497')
        CALL AddToObjStr('Fan Power Coefficient 3','-0.07292612')
        CALL AddToObjStr('Fan Power Coefficient 4', '0.943739823')
        CALL AddToObjStr('Fan Power Coefficient 5','0')
    END SELECT
    CALL AddToObjStr('Air Inlet Node Name', TRIM(coldfanInlet))
    CALL AddToObjStr('Air Outlet Node Name', TRIM(coldfanOutlet),.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF
  IF (coldFanKind .EQ. sfkCAV) THEN
    CALL CreateNewObj('Fan:ConstantVolume')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Cold Supply Fan')
    IF (isSysAvailSchedBlank) THEN
      CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
    ELSE
      CALL AddToObjFld('Availability Schedule Name', base + ddsSysAvailSchedNameOff,' ')
    END IF
    CALL AddToObjFld('Fan Efficiency', base + ddsColdEfficiencyOff,' ')
    CALL AddToObjFld('Pressure Rise {Pa}', base + ddsColdPressureOff,' ')
    CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
    CALL AddToObjFld('Motor Efficiency', base + ddsColdFanMotorEffOff,' ')
    CALL AddToObjFld('Motor in Airstream Fraction', base + ddsColdFanFractionOff,' ')
    CALL AddToObjStr('Air Inlet Node Name', TRIM(coldfanInlet))
    CALL AddToObjStr('Air Outlet Node Name', TRIM(coldfanOutlet),.TRUE.)
    IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  END IF

  ! Humidifier set point manager and humidifier ~ line 610
  IF (.NOT. isHumidifierNone) THEN
    CALL CreateNewObj('SetpointManager:SingleZone:Humidity:Minimum')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Humidification Setpoint Manager')
    CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(humidifierOutlet))
    CALL AddToObjFld('Control Zone Air Node Name', base + ddsHumidCtrlZoneOff,' Zone Air Node',.TRUE.)
    IF (humidifierKind .EQ. humidifyElecSteam) THEN
      CALL CreateNewObj('Humidifier:Steam:Electric')
      CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Humidifier')
      CALL AddToObjFld('Availability Schedule Name', base + ddsHumidAvailSchedNameOff,' ')
      CALL AddToObjFld('Rated Capacity {m3/s}', base + ddsHumidCapOff,'')
      CALL AddToObjFld('Rated Power {W}', base + ddsHumidRatedElecOff,'')
      CALL AddToObjStr('rated fan power {W}','0')
      CALL AddToObjStr('Standby power {W}','0')
      CALL AddToObjStr('Air Inlet Node Name', TRIM(humidifierInlet))
      CALL AddToObjStr('Air Outlet Node Name', TRIM(humidifierOutlet),.TRUE.)
    END IF
  END IF
  !OUTSIDE AIR INLET NODE LIST ~ line 633
  CALL CreateNewObj('OutdoorAir:NodeList')
  CALL AddToObjFld('Node or NodeList Name 1', base + ddsAirHandlerNameOff,' Outside Air Inlet',.TRUE.)
  !OUTSIDE AIR SYSTEM ~line 539
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' OA System')
  CALL AddToObjFld('Controller List Name', base + ddsAirHandlerNameOff,' OA System Controllers')
  CALL AddToObjFld('Outdoor Air Equipment List Name', base + ddsAirHandlerNameOff,' OA System Equipment')
  CALL AddToObjFld('Availability Manager List Name', base + ddsAirHandlerNameOff,' Availability Managers',.TRUE.)
  !CONTROLLER LIST ~ line 545
  CALL CreateNewObj('AirLoopHVAC:ControllerList')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' OA System Controllers')
  IF (preHeatCoilType .EQ. ctHotWater) THEN
    CALL AddToObjStr('Controller Object Type','Controller:OutdoorAir')
    CALL AddToObjFld('Controller Name', base + ddsAirHandlerNameOff,' OA Controller')
    CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
    CALL AddToObjFld('Controller Name', base + ddsAirHandlerNameOff,' Preheat Coil Controller',.TRUE.)
  ELSE
    CALL AddToObjStr('Controller Object Type','Controller:OutdoorAir')
    CALL AddToObjFld('Controller Name', base + ddsAirHandlerNameOff,' OA Controller',.TRUE.)
  END IF
  !AIR LOOP EQUIPMENT LIST ~ line 552
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem:EquipmentList')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' OA System Equipment')
  IF (preHeatCoilType .EQ. ctHotWater) THEN
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Preheat Coil')
  ELSEIF (preHeatCoilType .EQ. ctElectric) THEN
    CALL AddToObjStr('Component Object Type','Coil:Heating:Electric')
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Preheat Coil')
  ELSEIF (preHeatCoilType .EQ. ctGas) THEN
    CALL AddToObjStr('Component Object Type','Coil:Heating:Fuel')
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Preheat Coil')
  END IF
  IF (heatRecovery .EQ. htrecSens) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Heat Recovery')
  ELSEIF (heatRecovery .EQ. htrecEnth) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Heat Recovery')
  END IF
  CALL AddToObjStr('Component Object Type','OutdoorAir:Mixer')
  CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' OA Mixing Box',.TRUE.)
  !OUTSIDE AIR MIXER ~ line 567
  CALL CreateNewObj('OutdoorAir:Mixer')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' OA Mixing Box')
  CALL AddToObjFld('Mixed Air Node Name', base + ddsAirHandlerNameOff,' Mixed Air Outlet')
  IF ((heatRecovery .EQ. htrecNone) .AND. (preHeatCoilType .EQ. ctNone)) THEN
    CALL AddToObjFld('Outside Air Stream Node Name', base + ddsAirHandlerNameOff,' Outside Air Inlet')
  END IF
  IF (preHeatCoilType .NE. ctNone) THEN
    CALL AddToObjFld('Outside Air Stream Node Name', base + ddsAirHandlerNameOff,' Preheat Coil Outlet')
  END IF
  IF ((heatRecovery .NE. htrecNone) .AND. (preHeatCoilType .EQ. ctNone)) THEN
    CALL AddToObjFld('Outside Air Stream Node Name', base + ddsAirHandlerNameOff,' Heat Recovery Supply Outlet')
  ENDIF
  CALL AddToObjFld('Relief Air Stream Node Name', base + ddsAirHandlerNameOff,' Relief Air Outlet')
  CALL AddToObjStr('Return Air Stream Node Name', TRIM(returnInletToOAMIxer),.TRUE.)
  !CONTROLLER:OUTSIDE AIR ~ line 576
  CALL CreateNewObj('Controller:OutdoorAir')
  CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' OA Controller')
  CALL AddToObjFld('Relief Air Outlet Node Name', base + ddsAirHandlerNameOff,' Relief Air Outlet')
  CALL AddToObjStr('Return Air Node Name',TRIM(returnInletToOAMIxer))
  CALL AddToObjFld('Mixed Air Node Name', base + ddsAirHandlerNameOff,' Mixed Air Outlet')
  CALL AddToObjFld('Actuator Node Name', base + ddsAirHandlerNameOff,' Outside Air Inlet')
  CALL AddToObjFld('Minimum Outdoor Air Flow Rate {m3/s}', base + ddsMinOutsideFlowOff,' ')
  CALL AddToObjFld('Maximum Outdoor Air Flow Rate {m3/s}', base + ddsMaxOutsideFlowOff,' ')
  ! ddsEconomizerKindOff is now a straight pass-thru to EconomizerChoice in
  ! the OA controller
  CALL AddToObjFld('Economizer Control Type', base + ddsEconomizerKindOff,' ')
  ! If heat recovery is selected, then set economizer control to bypass
  ! otherwise, set to modulate OA flow
  IF (heatRecovery .EQ. htrecNone) THEN
    CALL AddtoObjStr('Economizer Control Action Type','ModulateFlow')
  ELSE
    CALL AddtoObjStr('Economizer Control Action Type','MinimumFlowWithBypass')
  ENDIF
  CALL AddToObjFld('Economizer Maximum Limit Dry-Bulb Temperature {C}', base + ddsEconoUpLimitOff,' ')
  CALL AddToObjFld('Economizer Maximum Limit Enthalpy {J/kg}', base + ddsEconoEnthLimitOff,' ')
  IF (econoKind .EQ. etFixedDewPointAndDryBulb) THEN
    CALL AddToObjFld('Economizer Maximum Limit Dewpoint Temperature (C)',base + ddsEconoUpDewLimitOff,' ')
  ELSE
    CALL AddToObjStr('Economizer Maximum Limit Dewpoint Temperature (C)',' ')
  ENDIF
  IF (econoKind .EQ. etElectronicEnthalpy) THEN
    CALL AddToObjFld('Electronic Enthalpy Limit Curve Name', base + ddsAirHandlerNameOff,' ElectrEnthCurve')
  ELSE
    CALL AddToObjStr('Electronic Enthalpy Limit Curve Name',' ')
  ENDIF
  CALL AddToObjFld('Economizer Minimum Limit Dry-Bulb Temperature {C}', base + ddsEconoLowLimitOff,' ')
  CALL AddToObjFld('Lockout Type', base + ddsEconoLockoutKindOff,' ')
  CALL AddToObjFld('Minimum Limit Type', base + ddsMinOutsideControlOff,' ')
  CALL AddToObjFld('Minimum Outdoor Air Schedule Name', base + ddsMinOutFlowSchedNameOff,' ')
  CALL AddToObjStr('Minimum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Maximum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Mechanical Ventilation Controller Name','',.TRUE.)
  ! If economizer type is ElectronicEnthalpy then create curve object
  IF (econoKind .EQ. etElectronicEnthalpy) THEN
    !CURVE:CUBIC
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' ElectrEnthCurve')
    CALL AddToObjStr('Coefficient1 Constant','0.01342704')
    CALL AddToObjStr('Coefficient2 x','-0.00047892')
    CALL AddToObjStr('Coefficient3 x**2','0.000053352')
    CALL AddToObjStr('Coefficient4 x**3','-0.0000018103')
    CALL AddToObjStr('Minimum Value of x','16.6')
    CALL AddToObjStr('Maximum Value of x','29.13',.TRUE.)
  ENDIF
  IF (preHeatCoilType .EQ. ctHotWater) THEN
    !COIL:Water:SimpleHeating  ~ line 614
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Preheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + ddsPreheatAvailSchedNameOff,' ')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name ', base + ddsAirHandlerNameOff,' Preheat Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + ddsAirHandlerNameOff,' Preheat Coil HW Outlet')
    IF (heatRecovery .EQ. htrecNone) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + ddsAirHandlerNameOff,' Outside Air Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + ddsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    END IF
    CALL AddToObjFld('Air Outlet Node Name', base + ddsAirHandlerNameOff,' Preheat Coil Outlet')
    CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
    CALL AddToObjStr('Rated Capacity','autosize')
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
    !BRANCH ~ line 625
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Preheat Coil HW Branch')
    !save this branch name for later use in plan
    CALL AddToStrList(TRIM(FldVal(base + ddsAirHandlerNameOff)) // ' Preheat Coil HW Branch',handleHeatingCoilBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + ddsAirHandlerNameOff,' Preheat Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + ddsAirHandlerNameOff,' Preheat Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + ddsAirHandlerNameOff,' Preheat Coil HW Outlet',.TRUE.)
    !CONTROLLER:SIMPLE ~ line 634
    CALL CreateNewObj('Controller:WaterCoil')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Preheat Coil Controller')
    CALL AddToObjStr('Control Variable','Temperature')
    CALL AddToObjStr('Action','Normal')
    CALL AddToObjStr('Actuator Variable','Flow')
    CALL AddToObjFld('Sensor Node Name', base + ddsAirHandlerNameOff,' Preheat Coil Outlet')
    CALL AddToObjFld('Actuator Node Name', base + ddsAirHandlerNameOff,' Preheat Coil HW Inlet')
    CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
    CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
    CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)
  ELSEIF (preHeatCoilType .EQ. ctElectric) THEN
    !COIL:Electric:Heating  ~ line 646
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Preheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + ddsPreheatAvailSchedNameOff,' ')
    CALL AddToObjStr('Efficiency','1')
    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    IF (heatRecovery .EQ. htrecNone) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + ddsAirHandlerNameOff,' Outside Air Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + ddsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    END IF
    CALL AddToObjFld('Air Outlet Node Name', base + ddsAirHandlerNameOff,' Preheat Coil Outlet')
    CALL AddToObjFld('Coil Temp Setpoint Node', base + ddsAirHandlerNameOff,' Preheat Coil Outlet',.TRUE.)
  ELSEIF (preHeatCoilType .EQ. ctGas) THEN
    !COIL:Gas:Heating ~ line 657
    CALL CreateNewObj('Coil:Heating:Fuel')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Preheat Coil')
    CALL AddToObjFld('Availability Schedule Name', base + ddsPreheatAvailSchedNameOff,' ')
    CALL AddToObjStr('Fuel Type', 'Gas')
    CALL AddToObjFld('Gas Burner Efficiency', base + ddsPreheatEffiencyOff,' ')
    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    IF (heatRecovery .EQ. htrecNone) THEN
      CALL AddToObjFld('Air Inlet Node Name', base + ddsAirHandlerNameOff,' Outside Air Inlet')
    ELSE
      CALL AddToObjFld('Air Inlet Node Name', base + ddsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    END IF
    CALL AddToObjFld('Air Outlet Node Name', base + ddsAirHandlerNameOff,' Preheat Coil Outlet')
    CALL AddToObjFld('Coil Temp Setpoint Node', base + ddsAirHandlerNameOff,' Preheat Coil Outlet')
    CALL AddToObjFld('Parasitic Electric Load {W}', base + ddsPreheatParasiticOff,' ')
    CALL AddToObjFld('Part load fraction correlation (function of part load ratio)',   &
       base + ddsAirHandlerNameOff,' Preheat Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
    !CURVE:CUBIC ~ line 670
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Preheat Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  END IF
  IF (preHeatCoilType .NE. ctNone) THEN
    IF (isPreheatStPtSchedBlank) THEN
      CALL AddAlwaysSchedule(FldVal(base + ddsPreheatSetPtConstantOff))
    END IF
    !SET POINT MANAGER:SCHEDULED ~ line 679
    IF (isPreheatStPtSchedBlank) THEN
      CALL AddAlwaysSchedule(FldVal(base + ddsPreheatSetPtConstantOff))
    END IF
    CALL CreateNewObj('SetpointManager:Scheduled')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Preheat Coil Air Temp Manager')
    CALL AddToObjStr('Control Variable','Temperature')
    IF (isPreheatStPtSchedBlank) THEN
      CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + ddsPreheatSetPtConstantOff)))
    ELSE
      CALL AddToObjFld('Schedule Name', base + ddsPreheatSetPtSchedNameOff,'')
    END IF
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + ddsAirHandlerNameOff,' Preheat Coil Outlet',.TRUE.)
  END IF
  IF (heatRecovery .EQ. htrecSens) THEN
    !HEAT EXCHANGER:AIR TO AIR:GENERIC ~ line 685
    htRecSens75 = StringToReal(FldVal(base + ddsHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + ddsAirHandlerNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Nominal supply air flow rate {m3/s}', base + ddsMinOutsideFlowOff,'')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + ddsHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Heating Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Heating Air Flow','0')
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + ddsHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Cooling Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Cooling Air Flow','0')
    CALL AddToObjFld('Supply Air Inlet Node Name', base + ddsAirHandlerNameOff,' Outside Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + ddsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + ddsAirHandlerNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + ddsAirHandlerNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','0')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjStr('Heat exchanger type','Plate')
    CALL AddToObjStr('Frost Control Type','ExhaustOnly')
    CALL AddToObjStr('Threshold Temperature','-1.1')
    CALL AddToObjStr('Initial Defrost Time Fraction','0.167')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','0.024',.TRUE.)
  ELSEIF (heatRecovery .EQ. htrecEnth) THEN
    !HEAT EXCHANGER:AIR TO AIR:GENERIC ~ line 708
    htRecSens75 = StringToReal(FldVal(base + ddsHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    htRecLat75 = StringToReal(FldVal(base + ddsHeatRecLatEffOff)) + 0.05
    IF (htRecLat75 .GT. 1.) htRecLat75 = 1.
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + ddsAirHandlerNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Nominal supply air flow rate {m3/s}', base + ddsMinOutsideFlowOff,'')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + ddsHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Heating Air Flow', base + ddsHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow', htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Heating Air Flow', htRecLat75)
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + ddsHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Cooling Air Flow', base + ddsHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow', htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Cooling Air Flow', htRecLat75)
    CALL AddToObjFld('Supply Air Inlet Node Name', base + ddsAirHandlerNameOff,' Outside Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + ddsAirHandlerNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + ddsAirHandlerNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + ddsAirHandlerNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','0')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjStr('Heat exchanger type','Rotary')
    CALL AddToObjStr('Frost Control Type','MinimumExhaustTemperature')
    CALL AddToObjStr('Threshold Temperature','1.7')
    CALL AddToObjStr('Initial Defrost Time Fraction','')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','',.TRUE.)
  END IF
  !SET POINT MANAGER:SCHEDULED ~ line 734
  IF (heatRecovery .NE. htrecNone) THEN
    IF (isEconoLowLimitBlank) THEN
      CALL AddAlwaysSchedule('5')
    ELSE
      CALL AddAlwaysSchedule(FldVal(base + ddsEconoLowLimitOff))
    END IF
    CALL CreateNewObj('SetpointManager:Scheduled')
    CALL AddToObjFld('Name', base + ddsAirHandlerNameOff,' Heat Recovery Air Temp Manager')
    CALL AddToObjStr('Control Variable','Temperature')
    IF (isEconoLowLimitBlank) THEN
      CALL AddToObjStr('Schedule Name','HVACTemplate-Always 5')
    ELSE
      CALL AddToObjStr('Schedule Name','HVACTemplate-Always ' // FldVal(base + ddsEconoLowLimitOff))
    END IF
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + ddsAirHandlerNameOff,' Heat Recovery Supply Outlet',.TRUE.)
  END IF
END DO
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE CreateNewVRF
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Michael Witte of GARD Analytics, Inc.
          !    DATE WRITTEN   September 2013
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a VRF heat pump (terminal units and condenser system)

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for fan mode type
INTEGER, PARAMETER :: fmCycling = 1
INTEGER, PARAMETER :: fmNeither = 3
          ! parameters for baseboard type
INTEGER, PARAMETER :: baseboardNone           = 1
INTEGER, PARAMETER :: baseboardHotWater       = 2
INTEGER, PARAMETER :: baseboardElectric       = 3

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for
INTEGER :: iZone
INTEGER :: iSys
INTEGER :: iInt
LOGICAL :: isTStatNameBlank
LOGICAL :: isSupplyCoolRateAutosize
LOGICAL :: isSupplyHeatRateAutosize
LOGICAL :: isSysAvailSchedBlank
INTEGER :: fanMode=0
LOGICAL :: isDrawThru
LOGICAL :: isDedOutAirNameBlank
CHARACTER(len=1) :: SchType
LOGICAL :: isOATypeDetailed = .FALSE.
LOGICAL :: isBaseboardNone = .FALSE.
LOGICAL :: isTstatCtrlTypeScheduled = .FALSE.
LOGICAL :: isCoolCoilNone = .FALSE.
LOGICAL :: isHeatCoilNone = .FALSE.
LOGICAL :: isCondWaterCooled = .FALSE.
INTEGER :: baseboardKind = 0
CHARACTER(len=1) :: NextSequenceNumber = '1'
INTEGER :: controlZoneBase
CHARACTER(len=MaxAlphaLength) :: lastOutlet=''
CHARACTER(len=MaxAlphaLength) :: fanInlet=''
CHARACTER(len=MaxAlphaLength) :: fanOutlet=''
CHARACTER(len=MaxAlphaLength) :: coolCoilUnitInlet=''
CHARACTER(len=MaxAlphaLength) :: coolCoilUnitOutlet=''
CHARACTER(len=MaxAlphaLength) :: heatCoilInlet=''
CHARACTER(len=MaxAlphaLength) :: heatCoilOutlet=''

!======================================================================================================
! create Zones
!======================================================================================================
DO iZone = 1, numCompactZoneVRF
  base =  compactZoneVRFBase(iZone)
  ! set flags
  isTStatNameBlank = (FldVal(base + vrfzTStatNameOff) .EQ. '')
  isSupplyCoolRateAutosize = SameString(FldVal(base + vrfzSupplyCoolFlowRateOff),'autosize')
  isSupplyHeatRateAutosize = SameString(FldVal(base + vrfzSupplyHeatFlowRateOff),'autosize')
  isSysAvailSchedBlank = (FldVal(base + vrfzSysAvailSchedNameOff) .EQ. '')
  isOATypeDetailed = SameString(FldVal(base + vrfzOutAirTypeOff),'DetailedSpecification')
  isBaseboardNone = SameString(FldVal(base + vrfzBaseboardTypeOff),'None')
  ! validate cooling coil type
  IF (SameString(FldVal(base + vrfzCoolCoilKindOff),'None')) THEN
    isCoolCoilNone = .TRUE.
  ELSEIF (SameString(FldVal(base + vrfzCoolCoilKindOff),'VariableRefrigerantFlowDX')) THEN
    isCoolCoilNone = .FALSE.
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VRF "'//TRIM(FldVal(base + vrfzNameOff))//'"'// &
                    ' in the Cooling Coil Type field: '//TRIM(FldVal(base + vrfzCoolCoilKindOff)))
  END IF
  ! validate heating coil type
  IF (SameString(FldVal(base + vrfzHPHeatCoilKindOff),'None')) THEN
    isHeatCoilNone = .TRUE.
  ELSEIF (SameString(FldVal(base + vrfzHPHeatCoilKindOff),'VariableRefrigerantFlowDX')) THEN
    isHeatCoilNone = .FALSE.
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VRF "'//TRIM(FldVal(base + vrfzNameOff))//'"'// &
                    ' in the Heating Coil Type field: '//TRIM(FldVal(base + vrfzHPHeatCoilKindOff)))
  END IF
  ! Supply fan mode may be blank (default to Cycling) or a schedule name
  IF (SameString(FldVal(base + vrfzFanModeOff),'')) THEN
    fanMode = fmCycling
  ELSE
    fanMode = fmNeither
  END IF
  isDrawThru = SameString(FldVal(base + vrfzFanPlacementOff),'DrawThrough')
  isDedOutAirNameBlank = (FldVal(base + vrfzDedOutAirSysNameOff) .EQ. '')
  !validate cooling design method
  IF (SameString(FldVal(base +  vrfzCoolDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  vrfzCoolDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VRF "'//TRIM(FldVal(base + vrfzNameOff))//'"'// &
                    ' in the Zone Cooling Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + vrfzCoolDesignTempMethodOff)))
  END IF
  !validate heating design method
  IF (SameString(FldVal(base +  vrfzHeatDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  vrfzHeatDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VRF "'//TRIM(FldVal(base + vrfzNameOff))//'"'// &
                    ' in the Zone Heating Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + vrfzCoolDesignTempMethodOff)))
  END IF
  !set the baseboard indicator
  IF (SameString(FldVal(base +  vrfzBaseboardTypeOff),'None')) THEN
    baseboardKind = baseboardNone
  ELSEIF (SameString(FldVal(base +  vrfzBaseboardTypeOff),'HotWater')) THEN
    baseboardKind = baseboardHotWater
  ELSEIF (SameString(FldVal(base +  vrfzBaseboardTypeOff),'Electric')) THEN
    baseboardKind = baseboardElectric
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:VRF "'//TRIM(FldVal(base + vrfzNameOff))//'"'// &
                    ' in the Baseboard Heating Type field: '//TRIM(FldVal(base + vrfzBaseboardTypeOff)))
  END IF

  !repeat per zone - only if vrfzTStatNameOff is not blank
  IF (.NOT. isTStatNameBlank) THEN
    !ZONE CONTROL:THERMOSTATIC ~ line 5
    CALL AddAlwaysSchedule('4')
    CALL CreateNewObj('ZoneControl:Thermostat')
    CALL AddToObjFld('Name', base + vrfzNameOff,' Thermostat')
    CALL AddToObjFld('Zone or ZoneList Name', base + vrfzNameOff,'')
    CALL AddToObjStr('Control Type Schedule Name','HVACTemplate-Always 4')
    CALL AddToObjStr('Control Object Type','ThermostatSetpoint:DualSetpoint')
    CALL AddToObjFld('Control Name', base + vrfzTStatNameOff,' Dual SP Control',.TRUE.)
  END IF
  !ZONE SIZING ~ line 12
! OADCV Changes
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:OutdoorAir')
    CALL AddToObjStr('Name', 'SZ DSOA '//FldVal(base+vrfzNameOff))
    CALL AddToObjFld('Outdoor Air Method', base + vrfzOutAirTypeOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Person {m3/s}', base + vrfzOutAirFlowPersonOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone Floor Area {m3/s-m2}', base + vrfzOutAirFlowAreaOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone {m3/s}', base + vrfzOutAirFlowZoneOff,'',.true.)
  END IF

  CALL CreateNewObj('Sizing:Zone')
  CALL AddToObjFld('Zone or ZoneList Name', base + vrfzNameOff,'')
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Input Method',base + vrfzCoolDesignTempMethodOff,'')
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', base + vrfzCoolDesignTempOff,'')
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Difference {delta C}',base + vrfzCoolDesignTempDiffOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Input Method',base + vrfzHeatDesignTempMethodOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', base + vrfzHeatDesignTempOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Difference {delta C}',base + vrfzHeatDesignTempDiffOff,'')
  CALL AddToObjStr('Zone Cooling Design Supply Air Humidity Ratio {kg-H20/kg-air}','0.008')
  CALL AddToObjStr('Zone Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + vrfzDesSpecOANameSizOff,'')
  ELSE
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','SZ DSOA '//FldVal(base+vrfzNameOff))
  END IF
  CALL AddToObjFld('Zone Heating Sizing Factor', base + vrfzZoneHeatSizingFactorOff,'')
  CALL AddToObjFld('Zone Cooling Sizing Factor', base + vrfzZoneCoolSizingFactorOff,'')
  IF (isSupplyCoolRateAutosize) THEN
    CALL AddToObjStr('Cooling Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Cooling Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}','')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  ELSE
    CALL AddToObjStr('Cooling Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Cooling Design Air Flow Rate {m3/s}', base + vrfzSupplyCoolFlowRateOff,'')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}','')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  END IF
  IF (isSupplyHeatRateAutosize) THEN
    CALL AddToObjStr('Heating Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Heating Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}','')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
!    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+vrfzNameOff),.true.)
  ELSE
    CALL AddToObjStr('Heating Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Heating Design Air Flow Rate {m3/s}', base + vrfzSupplyHeatFlowRateOff,'')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}','')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
!    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+vrfzNameOff),.true.)
  END IF
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Zone Air Distribution Object Name', base + vrfzDesSpecOADistNameOff,'',.true.)
  ELSE
    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+vrfzNameOff),.true.)
  END IF

  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:ZoneAirDistribution')
    CALL AddToObjStr('Name','SZ DSZAD '//FldVal(base+vrfzNameOff))
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Cooling Mode {}','1')
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Heating Mode {}','1',.true.)
  END IF
  !CONTROLLED ZONE EQUIP CONFIGURATION ~ line 46
  CALL CreateNewObj('ZoneHVAC:EquipmentConnections')
  CALL AddToObjFld('Zone Name', base + vrfzNameOff,' ')
  CALL AddToObjFld('Zone Conditioning Equipment List Name', base + vrfzNameOff,' Equipment')
  IF (isDedOutAirNameBlank) THEN
    CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base + vrfzNameOff,' VRF Supply Inlet')
  ELSE
    CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base + vrfzNameOff,' Inlets')
  END IF
  CALL AddToObjFld('Zone Air Exhaust Node or NodeList Name', base + vrfzNameOff,' VRF Return')
  CALL AddToObjFld('Zone Air Node Name', base + vrfzNameOff,' Zone Air Node')
  CALL AddToObjFld('Zone Return Air Node Name', base + vrfzNameOff,' Return Outlet',.TRUE.)
  !NODELIST (not in spreadsheet added with DOAS)
  IF (.NOT. isDedOutAirNameBlank) THEN
    CALL CreateNewObj('NodeList')
    CALL AddToObjFld('Name', base + vrfzNameOff,' Inlets')
    CALL AddToObjFld('Node Name', base + vrfzNameOff,' VRF Supply Inlet')
    CALL AddToObjFld('Node Name', base + vrfzNameOff,' DOAS Supply Inlet',.TRUE.)
  ENDIF
  !ZONE EQUIPMENT LIST ~ line 54
  CALL CreateNewObj('ZoneHVAC:EquipmentList')
  CALL AddToObjFld('Name', base + vrfzNameOff,' Equipment')
  CALL AddToObjStr('Load Distribution Scheme','SequentialLoad')

  NextSequenceNumber = '1'
  IF (.NOT. isDedOutAirNameBlank) THEN
    CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Zone Equipment Name', base + vrfzNameOff,' DOAS ATU')
    CALL AddToObjStr('Zone Equipment Cooling Sequence',trim(NextSequenceNumber))
    CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence',trim(NextSequenceNumber))
    CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
    CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.FALSE.)
  NextSequenceNumber = '2'
  END IF
  CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:TerminalUnit:VariableRefrigerantFlow')
  CALL AddToObjFld('Zone Equipment Name', base + vrfzNameOff,' VRF Terminal Unit')
  CALL AddToObjStr('Zone Equipment Cooling Sequence',trim(NextSequenceNumber))
  CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence',trim(NextSequenceNumber),.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',isBaseboardNone)
  IF (.NOT. isBaseboardNone) THEN
    IF (isDedOutAirNameBlank) THEN
      NextSequenceNumber = '2'
    ELSE
      NextSequenceNumber = '3'
    END IF
    SELECT CASE (baseboardKind)
      CASE (baseboardHotWater)
        CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CASE (baseboardElectric)
        CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:Convective:Electric')
    END SELECT
        CALL AddToObjFld('Zone Equipment Name', base + vrfzNameOff,' Baseboard Heat')
        CALL AddToObjStr('Zone Equipment Cooling Sequence',trim(NextSequenceNumber))
        CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence',trim(NextSequenceNumber),.FALSE.)
        CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
        CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.TRUE.)
    END IF

  IF (.NOT. isDedOutAirNameBlank) THEN
    !***ZoneHVAC:AirDistributionUnit
    CALL CreateNewObj('ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Name', base + vrfzNameOff,' DOAS ATU')
    CALL AddToObjFld('Air Distribution Unit Outlet Node Name', base + vrfzNameOff,' DOAS Supply Inlet')
    CALL AddToObjStr('Air Terminal Object Type','AirTerminal:SingleDuct:VAV:NoReheat')
    CALL AddToObjFld('Air Terminal Name', base + vrfzNameOff,' DOAS Air Terminal',.TRUE.)
    !***AirTerminal:SingleDuct:VAV:NoReheat
    CALL CreateNewObj('AirTerminal:SingleDuct:VAV:NoReheat')
    CALL AddToObjFld('Name', base + vrfzNameOff,' DOAS Air Terminal')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Air Outlet Node Name', base + vrfzNameOff,' DOAS Supply Inlet')
    CALL AddToObjFld('Air Inlet Node Name', base + vrfzNameOff,' Zone Equip Inlet')
    CALL AddToObjStr('Maximum Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Zone Minimum Air Flow Input Method','Constant')
    CALL AddToObjStr('Constant Minimum Air Flow Fraction','1.0')
    CALL AddToObjStr('Fixed Minimum Air Flow Rate {m3/s}','')
    CALL AddToObjStr('Minimum Air Flow Fraction Schedule Name','')
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','',.TRUE.)
  ENDIF

  ! Set component node names
  lastOutlet = TRIM(FldVal(base + vrfzNameOff))//' VRF Mixed Air Outlet'

  ! blow thru fan
  IF (.NOT. isDrawThru) THEN
    fanInlet         = TRIM(lastOutlet)
    fanOutlet        = TRIM(FldVal(base + vrfzNameOff))//' VRF Supply Fan Outlet'
    lastOutlet       = TRIM(fanOutlet)
  END IF

  ! cooling coil
  IF (.NOT. isCoolCoilNone) THEN
    coolCoilUnitInlet = TRIM(lastOutlet)
    IF ((.NOT. isHeatCoilNone) .OR. (isDrawThru)) THEN
      coolCoilUnitOutlet = TRIM(FldVal(base + vrfzNameOff))//' VRF Cooling Coil Outlet'
    ELSE
      coolCoilUnitOutlet = TRIM(FldVal(base + vrfzNameOff))//' VRF Supply Inlet'
    END IF
    lastOutlet       = TRIM(coolCoilUnitOutlet)
  END IF

  ! heating coil
  IF (.NOT. isHeatCoilNone) THEN
    heatCoilInlet    = TRIM(lastOutlet)
    IF (isDrawThru) THEN
      heatCoilOutlet   = TRIM(FldVal(base + vrfzNameOff))//' VRF Heating Coil Outlet'
    ELSE
      heatCoilOutlet   = TRIM(FldVal(base + vrfzNameOff))//' VRF Supply Inlet'
    END IF
    lastOutlet       = TRIM(heatCoilOutlet)
  END IF

  ! draw thru fan
  IF (isDrawThru) THEN
    fanInlet         = TRIM(lastOutlet)
    fanOutlet        = TRIM(FldVal(base + vrfzNameOff))//' VRF Supply Inlet'
    lastOutlet       = TRIM(fanOutlet)
  END IF


  !ZoneHVAC:PackagedTerminalHeatPump ~ line 61
  IF (isDedOutAirNameBlank) THEN
    CALL WriteComment('Set VRF outdoor air to zero, because zone is served by dedicated outdoor air system (DOAS)')
  END IF
  CALL CreateNewObj('ZoneHVAC:TerminalUnit:VariableRefrigerantFlow')
  CALL AddToObjFld('Zone Terminal Unit Name', base + vrfzNameOff,' VRF Terminal Unit')
  CALL AddToObjFld('Terminal Unit Availability Schedule', base + vrfzSysAvailSchedNameOff,'')
  CALL AddToObjFld('Terminal Unit Air Inlet Node Name', base + vrfzNameOff,' VRF Return')
  CALL AddToObjFld('Terminal Unit Air Outlet Node Name', base + vrfzNameOff,' VRF Supply Inlet')
  !CR8001
  !CALL AddToObjFld('Supply air volumetric flow rate during cooling operation {m3/s}',   &
  !   base + vrfzSupplyCoolFlowRateOff,'')
  CALL AddToObjStr('Cooling Supply Air Flow Rate {m3/s}','autosize')
  CALL AddToObjFld('No Cooling Supply Air Flow Rate {m3/s}',   &
     base + vrfzSupplyNoCoolFlowRateOff,'')
  CALL AddToObjStr('Heating Supply Air Flow Rate {m3/s}','autosize')
  CALL AddToObjFld('No Heating Supply Air Flow Rate {m3/s}',   &
     base + vrfzSupplyNoHeatFlowRateOff,'')
  IF (isDedOutAirNameBlank) THEN
    CALL AddToObjStr('Cooling Outdoor Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Heating Outdoor Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('No Load Outdoor Air Flow Rate {m3/s}','autosize')
  ELSE
    CALL AddToObjStr('Cooling Outdoor Air Flow Rate {m3/s}','0.0')
    CALL AddToObjStr('Heating Outdoor Air Flow Rate {m3/s}','0.0')
    CALL AddToObjStr('No Load Outdoor Air Flow Rate {m3/s}','0.0')
  END IF
  SchType=' '
  SELECT CASE (fanMode)
    CASE (fmCycling)
      CALL AddToObjStr('Supply Air Fan Operating Mode Schedule Name','HVACTemplate-Always 0')
      SchType='0'
    CASE (fmNeither)
      CALL AddToObjFld('Supply Air Fan Operating Mode Schedule Name', base + vrfzFanModeOff,'')
  END SELECT
  CALL AddToObjFld('Supply Air Fan Placement', base + vrfzFanPlacementOff,'')
  CALL AddToObjStr('Supply Air Fan Object Type','Fan:OnOff')
  CALL AddToObjFld('Supply Air Fan Object Name', base + vrfzNameOff,' VRF Supply Fan')
  CALL AddToObjStr('Outdoor Air Mixer Object Type','OutdoorAir:Mixer')
  CALL AddToObjFld('Outdoor Air Mixer Object Name', base + vrfzNameOff,' VRF OA Mixing Box')
  IF (isCoolCoilNone) THEN
    CALL AddToObjStr('Cooling Coil Object Type','')
    CALL AddToObjStr('Cooling Coil Object Name','')
  ELSE
    CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:DX:VariableRefrigerantFlow')
    CALL AddToObjFld('Cooling Coil Object Name', base + vrfzNameOff,' VRF Cooling Coil')
  END IF
  IF (isHeatCoilNone) THEN
    CALL AddToObjStr('Heating Coil Object Type','')
    CALL AddToObjStr('Heating Coil Object Name','')
  ELSE
    CALL AddToObjStr('Heating Coil Object Type','Coil:Heating:DX:VariableRefrigerantFlow')
    CALL AddToObjFld('Heating Coil Object Name', base + vrfzNameOff,' VRF Heating Coil')
  END IF
  CALL AddToObjFld('Zone Terminal Unit On Parasitic Electric Energy Use {W}',   &
     base + vrfzTermUnitOnParasiticElecOff,'')
  CALL AddToObjFld('Zone Terminal Unit Off Parasitic Electric Energy Use {W}',   &
     base + vrfzTermUnitOffParasiticElecOff,'',.TRUE.)
! add new schedule if needed, from Supply Air Fan Operating Mode Schedule Name above
  IF (SchType /= ' ') CALL AddAlwaysSchedule(SchType)
!end LKL changes
  !FAN:SIMPLE:ONOFF ~ line 95
  CALL CreateNewObj('Fan:OnOff')
  CALL AddToObjFld('Name', base + vrfzNameOff,' VRF Supply Fan')
  IF (.NOT. isSysAvailSchedBlank) THEN
    CALL AddToObjFld('Availability Schedule Name', base + vrfzSysAvailSchedNameOff,'')
  ELSE
    CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
  END IF
  CALL AddToObjFld('Fan Efficiency', base + vrfzFanTotEfficiencyOff,'')
  CALL AddToObjFld('Pressure Rise {Pa}', base + vrfzFanPressureOff,'')
  CALL AddToObjStr('Maximum Flow Rate {m3/s}','autosize')
  CALL AddToObjFld('Motor Efficiency', base + vrfzFanMotorEfficiencyOff,'')
  CALL AddToObjStr('Motor in Airstream Fraction','1')
  CALL AddToObjStr('Air Inlet Node Name', TRIM(fanInlet))
  CALL AddToObjStr('Air Outlet Node Name', TRIM(fanOutlet),.TRUE.)
  IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  !OUTSIDE AIR MIXER ~ line 108
  CALL CreateNewObj('OutdoorAir:Mixer')
  CALL AddToObjFld('Name', base + vrfzNameOff,' VRF OA Mixing Box')
  CALL AddToObjFld('Mixed Air Node Name', base + vrfzNameOff,' VRF Mixed Air Outlet')
  CALL AddToObjFld('Outdoor Air Stream Node Name', base + vrfzNameOff,' VRF Outdoor Air Inlet')
  CALL AddToObjFld('Relief Air Stream Node Name', base + vrfzNameOff,' VRF Relief Air Outlet')
  CALL AddToObjFld('Return Air Stream Node Name', base + vrfzNameOff,' VRF Return',.TRUE.)
  !OUTSIDE AIR NODE ~ line 115
  CALL CreateNewObj('OutdoorAir:Node')
  CALL AddToObjFld('Name', base + vrfzNameOff,' VRF Outdoor Air Inlet')
  CALL AddToObjStr('Height Above Ground','-1',.TRUE.)

  ! Cooling coil
  IF (.NOT. isCoolCoilNone) THEN
    CALL CreateNewObj('Coil:Cooling:DX:VariableRefrigerantFlow')
    CALL AddToObjFld('Name', base + vrfzNameOff,' VRF Cooling Coil')
    CALL AddToObjFld('Availability Schedule Name', base + vrfzCoolAvailSchedNameOff,'')
    CALL AddToObjFld('Gross Rated Total Cooling Capacity {W}', base + vrfzCoolCoilCapOff,'')
    CALL AddToObjFld('Gross Rated Sensible Heat Ratio', base + vrfzCoolCoilSHROff,'')
    CALL AddToObjStr('Rated Air Flow Rate {m3/s}', 'autosize')
    CALL AddToObjFld('Cooling Capacity Ratio Modifier Function of Temperature Curve Name',   &
       base + vrfzNameOff,' VRF Cooling Coil Cap-FT')
    CALL AddToObjFld('Cooling Capacity Modifier Curve Function of Flow Fraction Name',   &
       base + vrfzNameOff,' VRF Cooling Coil Cap-FF')
    CALL AddToObjStr('Coil Air Inlet Node', TRIM(coolCoilUnitInlet))
    CALL AddToObjStr('Coil Air Outlet Node', TRIM(coolCoilUnitOutlet))
    CALL AddToObjStr('Name of Water Storage Tank for Condensate Collection','',.TRUE.)
    !CURVE:BIQUADRATIC
    CALL WriteComment('VRF Cooling Coil Cap-FT curve from example file VariableRefrigerantFlow_5Zone.idf Sep 2013')
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + vrfzNameOff,' VRF Cooling Coil Cap-FT')
    CALL AddToObjStr('Coefficient1 Constant','0.504547273506488')
    CALL AddToObjStr('Coefficient2 x','0.0288891279198444')
    CALL AddToObjStr('Coefficient3 x**2','-0.000010819418650677')
    CALL AddToObjStr('Coefficient4 x**3','0.0000101359395177008')
    CALL AddToObjStr('Minimum Value of x','0.0')
    CALL AddToObjStr('Maximum Value of x','50.0')
    CALL AddToObjStr('Minimum Curve Output','0.5')
    CALL AddToObjStr('Maximum Curve Output','1.5')
    CALL AddToObjStr('Input Unit Type for X','Temperature')
    CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
    !CURVE:QUADRATIC
    CALL WriteComment('VRF Cooling Coil Cap-FF curve from example file VariableRefrigerantFlow_5Zone.idf Sep 2013')
    CALL CreateNewObj('Curve:Quadratic')
    CALL AddToObjFld('Name', base + vrfzNameOff,' VRF Cooling Coil Cap-FF')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Minimum Value of x','0.5')
    CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
  END IF

  ! Heating coil
  IF (.NOT. isHeatCoilNone) THEN
    CALL CreateNewObj('Coil:Heating:DX:VariableRefrigerantFlow')
    CALL AddToObjFld('Name', base + vrfzNameOff,' VRF Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + vrfzHeatAvailSchedNameOff,'')
    CALL AddToObjFld('Gross Rated Heating Capacity {W}', base + vrfzHeatCoilCapOff,'')
    CALL AddToObjStr('Rated Air Flow Rate {m3/s}', 'autosize')
    CALL AddToObjStr('Coil Air Inlet Node', TRIM(heatCoilInlet))
    CALL AddToObjStr('Coil Air Outlet Node', TRIM(heatCoilOutlet))
    CALL AddToObjFld('Heating Capacity Ratio Modifier Function of Temperature Curve Name',   &
       base + vrfzNameOff,' VRF Heating Coil Cap-FT')
    CALL AddToObjFld('Heating Capacity Modifier Function of Flow Fraction Curve Name',   &
       base + vrfzNameOff,' VRF Heating Coil Cap-FF',.TRUE.)
    !CURVE:CUBIC
    CALL WriteComment('VRF Cooling Coil Cap-FT curve from example file VariableRefrigerantFlow_5Zone.idf Sep 2013')
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + vrfzNameOff,' VRF Heating Coil Cap-FT')
    CALL AddToObjStr('Coefficient1 Constant','-0.390708928227928')
    CALL AddToObjStr('Coefficient2 x','0.261815023760162')
    CALL AddToObjStr('Coefficient3 x**2','-0.0130431603151873')
    CALL AddToObjStr('Coefficient4 x**3','0.000178131745997821')
    CALL AddToObjStr('Minimum Value of x','0.0')
    CALL AddToObjStr('Maximum Value of x','50.0')
    CALL AddToObjStr('Minimum Curve Output','0.5')
    CALL AddToObjStr('Maximum Curve Output','1.5')
    CALL AddToObjStr('Input Unit Type for X','Temperature')
    CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)
    !CURVE:CUBIC
    CALL WriteComment('VRF Cooling Coil Cap-FF curve from example file VariableRefrigerantFlow_5Zone.idf Sep 2013')
    CALL CreateNewObj(' Curve:Quadratic')
    CALL AddToObjFld('Name', base + vrfzNameOff,' VRF Heating Coil Cap-FF')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Minimum Value of x','0.5')
    CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
  END IF
  !Baseboards
  SELECT CASE (baseboardKind)
    CASE (baseboardHotWater)
      !ZoneHVAC:Baseboard:RadiantConvective:Water
      CALL CreateNewObj('ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Name', base + vrfzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + vrfzBaseboardAvailSchedNameOff,'')
      CALL AddToObjFld('Inlet Node Name', base + vrfzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Outlet Node Name', base + vrfzNameOff,' Baseboard HW Outlet')
      CALL AddToObjStr('Rated Average Water Temperature {C}','87.78')
      CALL AddToObjStr('Rated Water Mass Flow Rate {m3/s}','0.063')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + vrfzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Maximum Water Flow Rate','autosize')
      CALL AddToObjStr('Convergence Tolerance','0.001')
      CALL AddToObjStr('Fraction Radiant','0.0')
      CALL AddToObjStr('Fraction of Radiant Energy Incident on People','0.0',.TRUE.)
      !BRANCH (for baseboard)
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + vrfzNameOff,' Baseboard Heat HW Branch')
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Component Name', base + vrfzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Component Inlet Node Name', base + vrfzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + vrfzNameOff,' Baseboard HW Outlet',.TRUE.)
      CALL AddToStrList(TRIM(FldVal(base + vrfzNameOff)) // ' Baseboard Heat HW Branch',handleHeatingCoilBranch)
      IF ((compactHotLoopBase .EQ. 0) .OR. (numCompactBoiler .EQ. 0)) THEN
        CALL WriteError('Both a HVACTemplate:Plant:HotWaterLoop and a HVACTemplate:Plant:Boiler are needed ' // &
             'when using hot water baseboards.')
      END IF
    CASE (baseboardElectric)
      !BASEBOARD HEATER:Electric:Convective
      CALL CreateNewObj('ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Name', base + vrfzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + vrfzBaseboardAvailSchedNameOff,'')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + vrfzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Efficiency','1',.TRUE.)
  END SELECT
END DO

!======================================================================================================
! create Systems for HVACTemplate:System:VRF
!======================================================================================================
DO iSys = 1, numCompactSysVRF
  base =  compactSysVRFBase(iSys)

  ! set flags
  isCondWaterCooled = SameString(FldVal(base + vrfsCondenserTypeOff),'WaterCooled')
  isTstatCtrlTypeScheduled = SameString(FldVal(base + vrfsTstatPriorityCtrlTypeOff),'Scheduled')

  ! find base for controlled zone used by this system
  controlZoneBase = -1
  DO iZone = 1, numCompactZoneVRF
    IF (SameString(FldVal(compactZoneVRFBase(iZone) + vrfzNameOff), FldVal(base +  vrfsTstatZoneNameOff))) THEN
      controlZoneBase = compactZoneVRFBase(iZone)
      EXIT
    END IF
  END DO
  IF ((controlZoneBase .EQ. -1) .AND. SameString(FldVal(base + vrfsTstatPriorityCtrlTypeOff),'MasterThermostatPriority')) THEN
    CALL WriteError('In HVACTemplate:System:VRF "'//TRIM(FldVal(base + vrfsNameOff))//'" for the field Zone Name for '//&
   'Master Thermostat Location could not find a matching HVACTemplate:Zone:VRF named: '//&
    TRIM(FldVal(base +  vrfsTstatZoneNameOff)),msgWarning)
  END IF
  IF ((isTstatCtrlTypeScheduled) .AND. SameString(FldVal(base + vrfsTstatPrioritySchedNameOff),'')) THEN
    CALL WriteError('In HVACTemplate:System:VRF "'//TRIM(FldVal(base + vrfsNameOff))//'" the Master Thermostat  '//&
   'Priority Control Type = Scheduled, but the Thermostat Priority Schedule Name is blank.')
  END IF

  CALL CreateNewObj('AirConditioner:VariableRefrigerantFlow')
  CALL AddToObjFld('Heat Pump Name', base + vrfsNameOff,' VRF Heat Pump')
  CALL AddToObjFld('Availability Schedule Name', base + vrfsSysAvailSchedNameOff,'')
  CALL AddToObjFld('Gross Rated Total Cooling Capacity {W}', base + vrfsCoolCoilCapOff,'')
  CALL AddToObjFld('Gross Rated Cooling COP {W/W}', base + vrfsCoolCoilCOPOff,'')
  CALL AddToObjFld('Minimum Outdoor Temperature in Cooling Mode {C}', base + vrfsMinTempCoolingOff,'')
  CALL AddToObjFld('Maximum Outdoor Temperature in Cooling Mode {C}', base + vrfsMaxTempCoolingOff,'')
  CALL AddToObjFld('Cooling Capacity Ratio Modifier Function of Low Temperature Curve Name', base + vrfsNameOff,' VRFCoolCapFT')
  CALL AddToObjFld('Cooling Capacity Ratio Boundary Curve Name', base + vrfsNameOff,' VRFCoolCapFTBoundary')
  CALL AddToObjFld('Cooling Capacity Ratio Modifier Function of High Temperature Curve Name', base + vrfsNameOff,' VRFCoolCapFTHi')
  CALL AddToObjFld('Cooling Energy Input Ratio Modifier Function of Low Temperature Curve Name', base + vrfsNameOff,' VRFCoolEIRFT')
  CALL AddToObjFld('Cooling Energy Input Ratio Boundary Curve Name', base + vrfsNameOff,' VRFCoolEIRFTBoundary')
  CALL AddToObjFld('Cooling Energy Input Ratio Modifier Function of High Temperature Curve Name', base + vrfsNameOff, &
   ' VRFCoolEIRFTHi')
  CALL AddToObjFld('Cooling Energy Input Ratio Modifier Function of Low Part-Load Ratio Curve Name', base + vrfsNameOff, &
   ' CoolingEIRLowPLR')
  CALL AddToObjFld('Cooling Energy Input Ratio Modifier Function of High Part-Load Ratio Curve Name', base + vrfsNameOff, &
   ' CoolingEIRHiPLR')
  CALL AddToObjFld('Cooling Combination Ratio Correction Factor Curve Name', base + vrfsNameOff,' CoolingCombRatio')
  CALL AddToObjFld('Cooling Part-Load Fraction Correlation Curve Name', base + vrfsNameOff,' VRFCPLFFPLR')
  CALL AddToObjFld('Gross Rated Heating Capacity {W}', base + vrfsHPHeatCoilCapOff,'')
  CALL AddToObjFld('Rated Heating Capacity Sizing Ratio {W/W}', base + vrfsHeatCapSizRatOff,'')
  CALL AddToObjFld('Gross Rated Heating COP {W/W}', base + vrfsHPHeatCoilCOPOff,'')
  CALL AddToObjFld('Minimum Outdoor Temperature in Heating Mode {C}', base + vrfsHPHeatMinODBOff,'')
  CALL AddToObjFld('Maximum Outdoor Temperature in Heating Mode {C}', base + vrfsHPHeatMaxODBOff,'')
  CALL AddToObjFld('Heating Capacity Ratio Modifier Function of Low Temperature Curve Name', base + vrfsNameOff,' VRFHeatCapFT')
  CALL AddToObjFld('Heating Capacity Ratio Boundary Curve Name', base + vrfsNameOff,' VRFHeatCapFTBoundary')
  CALL AddToObjFld('Heating Capacity Ratio Modifier Function of High Temperature Curve Name', base + vrfsNameOff,' VRFHeatCapFTHi')
  CALL AddToObjFld('Heating Energy Input Ratio Modifier Function of Low Temperature Curve Name', base + vrfsNameOff,' VRFHeatEIRFT')
  CALL AddToObjFld('Heating Energy Input Ratio Boundary Curve Name', base + vrfsNameOff,' VRFHeatEIRFTBoundary')
  CALL AddToObjFld('Heating Energy Input Ratio Modifier Function of High Temperature Curve Name', base + vrfsNameOff, &
   ' VRFHeatEIRFTHi')
  CALL AddToObjStr('Heating Performance Curve Outdoor Temperature Type','WetBulbTemperature')
  CALL AddToObjFld('Heating Energy Input Ratio Modifier Function of Low Part-Load Ratio Curve Name', base + vrfsNameOff, &
   ' HeatingEIRLowPLR')
  CALL AddToObjFld('Heating Energy Input Ratio Modifier Function of High Part-Load Ratio Curve Name', base + vrfsNameOff, &
   ' HeatingEIRHiPLR')
  CALL AddToObjFld('Heating Combination Ratio Correction Factor Curve Name', base + vrfsNameOff,' HeatingCombRatio')
  CALL AddToObjFld('Heating Part-Load Fraction Correlation Curve Name', base + vrfsNameOff,' VRFCPLFFPLR')
  CALL AddToObjStr('Minimum Heat Pump Part-Load Ratio','0.25')
  CALL AddToObjFld('Zone Name for Master Thermostat Location', base + vrfsTstatZoneNameOff,'')
  CALL AddToObjFld('Master Thermostat Priority Control Type', base + vrfsTstatPriorityCtrlTypeOff,'')
  CALL AddToObjFld('Thermostat Priority Schedule Name', base + vrfsTstatPrioritySchedNameOff,'')
  CALL AddToObjFld('Zone Terminal Unit List Name', base + vrfsNameOff, 'TU List')
  CALL AddToObjFld('Heat Pump Waste Heat Recovery', base + vrfsHPWasteHeatRecoveryOff,'')
  CALL AddToObjFld('Equivalent Piping Length used for Piping Correction Factor in Cooling Mode {m}', &
   base + vrfsEquivPipLenCoolOff,'')
  CALL AddToObjFld('Vertical Height used for Piping Correction Factor {m}', base + vrfsEquivPipHeightOff,'')
  CALL AddToObjFld('Piping Correction Factor for Length in Cooling Mode Curve Name', base + vrfsNameOff, &
   ' CoolingLengthCorrectionFactor')
  CALL AddToObjStr('Piping Correction Factor for Height in Cooling Mode Coefficient','-0.000386')
  CALL AddToObjFld('Equivalent Piping Length used for Piping Correction Factor in Heating Mode {m}', &
   base + vrfsEquivPipLenHeatOff,'')
  CALL AddToObjStr('Piping Correction Factor for Length in Heating Mode Curve Name','')
  CALL AddToObjStr('Piping Correction Factor for Height in Heating Mode Coefficient','')
  CALL AddToObjFld('Crankcase Heater Power per Compressor {W}', base + vrfsCrankHtPwrOff,'')
  CALL AddToObjFld('Number of Compressors', base + vrfsNumComprOff,'')
  CALL AddToObjFld('Ratio of Compressor Size to Total Compressor Capacity', base + vrfsComprSizeRatioOff,'')
  CALL AddToObjFld('Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater {C}', base + vrfsCrankHtMaxODBOff,'')
  CALL AddToObjFld('Defrost Strategy', base + vrfsHPDefrTypeOff,'')
  CALL AddToObjFld('Defrost Control', base + vrfsHPDefrCtrlTypeOff,'')
  CALL AddToObjStr('Defrost Energy Input Ratio Modifier Function of Temperature Curve Name','')
  CALL AddToObjFld('Defrost Time Period Fraction', base + vrfsHPDefrTimeFracOff,'')
  CALL AddToObjFld('Resistive Defrost Heater Capacity {W}', base + vrfsHPDefrHeatCapOff,'')
  CALL AddToObjFld('Maximum Outdoor Dry-Bulb Temperature for Defrost Operation {C}', base + vrfsHPDefrMaxODBOff,'')
  CALL AddToObjFld('Condenser Type', base + vrfsCondenserTypeOff,'')
  CALL AddToObjFld('Condenser Inlet Node Name', base + vrfsNameOff,' Condenser Inlet Node')
  CALL AddToObjFld('Condenser Outlet Node Name', base + vrfsNameOff,' Condenser Outlet Node')
  CALL AddToObjFld('Water Condenser Volume Flow Rate (m3/s)', base + vrfsWaterCondVolFlowRateOff,'')
  CALL AddToObjFld('Evaporative Condenser Effectiveness {dimensionless}', base + vrfsEvapCondEffectOff,'')
  CALL AddToObjFld('Evaporative Condenser Air Flow Rate {m3/s}', base + vrfsEvapCondAirFlowRateOff,'')
  CALL AddToObjFld('Evaporative Condenser Pump Rated Power Consumption {W}', base + vrfsEvapCondPumpPwrOff,'')
  CALL AddToObjStr('Supply Water Storage Tank Name','')
  CALL AddToObjFld('Basin Heater Capacity {W/K}', base + vrfsBasinHeatCapOff,'')
  CALL AddToObjFld('Basin Heater Setpoint Temperature {C}', base + vrfsBasinHeatSetPtOff,'')
  CALL AddToObjFld('Basin Heater Operating Schedule Name', base + vrfsBasinHeatOperSchedNameOff,'')
  CALL AddToObjFld('Fuel Type', base + vrfsFuelTypeOff,'')
  CALL AddToObjFld('Minimum Outdoor Temperature in Heat Recovery Mode (C)', base + vrfsHeatRecMinODBOff,'')
  CALL AddToObjFld('Maximum Outdoor Temperature in Heat Recovery Mode (C)', base + vrfsHeatRecMaxODBOff,'')
  CALL AddToObjStr('Heat Recovery Cooling Capacity Modifier Curve Name','')
  CALL AddToObjStr('Initial Heat Recovery Cooling Capacity Fraction','')
  CALL AddToObjStr('Heat Recovery Cooling Capacity Time Constant (hr)','0.45')
  CALL AddToObjStr('Heat Recovery Cooling Energy Modifier Curve Name','')
  CALL AddToObjStr('Initial Heat Recovery Cooling Energy Fraction','')
  CALL AddToObjStr('Heat Recovery Cooling Energy Time Constant (hr)','')
  CALL AddToObjStr('Heat Recovery Heating Capacity Modifier Curve Name','')
  CALL AddToObjStr('Initial Heat Recovery Heating Capacity Fraction','')
  CALL AddToObjStr('Heat Recovery Heating Capacity Time Constant (hr)','0.45')
  CALL AddToObjStr('Heat Recovery Heating Eenrgy Modifier Curve Name','')
  CALL AddToObjStr('Initial Heat Recovery Heating Energy Fraction','')
  CALL AddToObjStr('Heat Recovery Heating Energy Time Constant (hr)','',.TRUE.)

  ! Terminal Unit List
  CALL CreateNewObj('ZoneTerminalUnitList')
  CALL AddToObjFld('Name', base + vrfsNameOff, 'TU List')
  CALL MakeSpecialIntList(zlVRFAll,base + vrfsNameOff)
  DO iInt = 1,numIntList - 1
    CALL AddToObjFld('Zone Terminal Unit Name', IntList(iInt),' VRF Terminal Unit')
  END DO
  IF (numIntList .GT. 0) THEN
    CALL AddToObjFld('Zone Terminal Unit Name', IntList(numIntList),' VRF Terminal Unit',.TRUE.)
  END IF

  !Condenser air or water connection
  IF (isCondWaterCooled) THEN
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + vrfsNameOff,' VRF Condenser Branch')
    CALL AddToStrList(TRIM(FldVal(base + vrfsNameOff)) // ' VRF Condenser Branch',handleMixedDemanBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','AirConditioner:VariableRefrigerantFlow')
    CALL AddToObjFld('Component Name', base + vrfsNameOff,' VRF Heat Pump')
    CALL AddToObjFld('Component Inlet Node Name', base + vrfsNameOff,' Condenser Inlet Node')
    CALL AddToObjFld('Component Outlet Node Name', base + vrfsNameOff,' Condenser Outlet Node',.TRUE.)
  ELSE
    !OUTSIDE AIR NODE
    CALL CreateNewObj('OutdoorAir:Node')
    CALL AddToObjFld('Name', base + vrfsNameOff,' Condenser Inlet Node')
    CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
  END IF

  ! Curve objects
  CALL WriteComment('VRF curves from example file VariableRefrigerantFlow_5Zone.idf Sep 2013')
  CALL CreateNewObj('Curve:Biquadratic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' VRFCoolCapFT')
  CALL AddToObjStr('Coefficient1 Constant','0.576856477')
  CALL AddToObjStr('Coefficient2 x','0.017450222')
  CALL AddToObjStr('Coefficient3 x**2','0.00058326')
  CALL AddToObjStr('Coefficient4 y','1.15563E-07')
  CALL AddToObjStr('Coefficient5 y**2','7.61504E-09')
  CALL AddToObjStr('Coefficient6 x*y','-1.1503E-08')
  CALL AddToObjStr('Minimum Value of x','15')
  CALL AddToObjStr('Maximum Value of x','24')
  CALL AddToObjStr('Minimum Value of y','-5')
  CALL AddToObjStr('Maximum Value of y','23')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Temperature')
  CALL AddToObjStr('Input Unit Type for Y','Temperature')
  CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)

  CALL CreateNewObj('Curve:Cubic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' VRFCoolCapFTBoundary')
  CALL AddToObjStr('Coefficient1 Constant','25.73473775')
  CALL AddToObjStr('Coefficient2 x','-0.03150043')
  CALL AddToObjStr('Coefficient3 x**2','-0.01416595')
  CALL AddToObjStr('Coefficient4 x**3','0')
  CALL AddToObjStr('Minimum Value of x','11')
  CALL AddToObjStr('Maximum Value of x','30')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Temperature')
  CALL AddToObjStr('Output Unit Type','Temperature',.TRUE.)

  CALL CreateNewObj('Curve:Biquadratic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' VRFCoolCapFTHi')
  CALL AddToObjStr('Coefficient1 Constant','0.6861551')
  CALL AddToObjStr('Coefficient2 x','0.0207869')
  CALL AddToObjStr('Coefficient3 x**2','0.0005445')
  CALL AddToObjStr('Coefficient4 y','-0.0016098')
  CALL AddToObjStr('Coefficient5 y**2','-6.254E-07')
  CALL AddToObjStr('Coefficient6 x*y','-0.000339')
  CALL AddToObjStr('Minimum Value of x','15')
  CALL AddToObjStr('Maximum Value of x','24')
  CALL AddToObjStr('Minimum Value of y','16')
  CALL AddToObjStr('Maximum Value of y','43')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Temperature')
  CALL AddToObjStr('Input Unit Type for Y','Temperature')
  CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)

  CALL CreateNewObj('Curve:Biquadratic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' VRFCoolEIRFT')
  CALL AddToObjStr('Coefficient1 Constant','0.989010541')
  CALL AddToObjStr('Coefficient2 x','-0.02347967')
  CALL AddToObjStr('Coefficient3 x**2','0.000199711')
  CALL AddToObjStr('Coefficient4 y','0.005968336')
  CALL AddToObjStr('Coefficient5 y**2','-1.0289E-07')
  CALL AddToObjStr('Coefficient6 x*y','-0.00015686')
  CALL AddToObjStr('Minimum Value of x','15')
  CALL AddToObjStr('Maximum Value of x','24')
  CALL AddToObjStr('Minimum Value of y','-5')
  CALL AddToObjStr('Maximum Value of y','23')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Temperature')
  CALL AddToObjStr('Input Unit Type for Y','Temperature')
  CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)

  CALL CreateNewObj('Curve:Cubic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' VRFCoolEIRFTBoundary')
  CALL AddToObjStr('Coefficient1 Constant','25.73473775')
  CALL AddToObjStr('Coefficient2 x','-0.03150043')
  CALL AddToObjStr('Coefficient3 x**2','-0.01416595')
  CALL AddToObjStr('Coefficient4 x**3','0')
  CALL AddToObjStr('Minimum Value of x','15')
  CALL AddToObjStr('Maximum Value of x','24')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Temperature')
  CALL AddToObjStr('Output Unit Type','Temperature',.TRUE.)

  CALL CreateNewObj('Curve:Biquadratic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' VRFCoolEIRFTHi')
  CALL AddToObjStr('Coefficient1 Constant','0.18293171')
  CALL AddToObjStr('Coefficient2 x','0.01635442')
  CALL AddToObjStr('Coefficient3 x**2','-0.0003505')
  CALL AddToObjStr('Coefficient4 y','0.02375694')
  CALL AddToObjStr('Coefficient5 y**2','0.00017854')
  CALL AddToObjStr('Coefficient6 x*y','-0.0006177')
  CALL AddToObjStr('Minimum Value of x','15')
  CALL AddToObjStr('Maximum Value of x','24')
  CALL AddToObjStr('Minimum Value of y','16')
  CALL AddToObjStr('Maximum Value of y','43')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Temperature')
  CALL AddToObjStr('Input Unit Type for Y','Temperature')
  CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)

  CALL CreateNewObj('Curve:Cubic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' CoolingEIRLowPLR')
  CALL AddToObjStr('Coefficient1 Constant','0.4628123')
  CALL AddToObjStr('Coefficient2 x','-1.0402406')
  CALL AddToObjStr('Coefficient3 x**2','2.17491')
  CALL AddToObjStr('Coefficient4 x**3','-0.5974817')
  CALL AddToObjStr('Minimum Value of x','0')
  CALL AddToObjStr('Maximum Value of x','1')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Temperature')
  CALL AddToObjStr('Output Unit Type','Temperature',.TRUE.)

  CALL CreateNewObj('Curve:Quadratic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' CoolingEIRHiPLR')
  CALL AddToObjStr('Coefficient1 Constant','1.0')
  CALL AddToObjStr('Coefficient2 x','0.0')
  CALL AddToObjStr('Coefficient3 x**2','0.0')
  CALL AddToObjStr('Minimum Value of x','1.0')
  CALL AddToObjStr('Maximum Value of x','1.5')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Dimensionless')
  CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)

  CALL CreateNewObj('Curve:Linear')
  CALL AddToObjFld('Name', base + vrfsNameOff,' CoolingCombRatio')
  CALL AddToObjStr('Coefficient1 Constant','0.618055')
  CALL AddToObjStr('Coefficient2 x','0.381945')
  CALL AddToObjStr('Minimum Value of x','1.0')
  CALL AddToObjStr('Maximum Value of x','1.5')
  CALL AddToObjStr('Minimum Curve Output','1.0')
  CALL AddToObjStr('Maximum Curve Output','1.2')
  CALL AddToObjStr('Input Unit Type for X','Dimensionless')
  CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)

  CALL CreateNewObj('CURVE:QUADRATIC')
  CALL AddToObjFld('Name', base + vrfsNameOff,' VRFCPLFFPLR')
  CALL AddToObjStr('Coefficient1 Constant','0.85')
  CALL AddToObjStr('Coefficient2 x','0.15')
  CALL AddToObjStr('Coefficient3 x**2','0.0')
  CALL AddToObjStr('Minimum Value of x','0.0')
  CALL AddToObjStr('Maximum Value of x','1.0')
  CALL AddToObjStr('Minimum Curve Output','0.85')
  CALL AddToObjStr('Maximum Curve Output','1.0')
  CALL AddToObjStr('Input Unit Type for X','Dimensionless')
  CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)

  CALL CreateNewObj('Curve:Biquadratic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' VRFHeatCapFT')
  CALL AddToObjStr('Coefficient1 Constant','1.014599599')
  CALL AddToObjStr('Coefficient2 x','-0.002506703')
  CALL AddToObjStr('Coefficient3 x**2','-0.000141599')
  CALL AddToObjStr('Coefficient4 y','0.026931595')
  CALL AddToObjStr('Coefficient5 y**2','1.83538E-06')
  CALL AddToObjStr('Coefficient6 x*y','-0.000358147')
  CALL AddToObjStr('Minimum Value of x','15')
  CALL AddToObjStr('Maximum Value of x','27')
  CALL AddToObjStr('Minimum Value of y','-20')
  CALL AddToObjStr('Maximum Value of y','15')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Temperature')
  CALL AddToObjStr('Input Unit Type for Y','Temperature')
  CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)

  CALL CreateNewObj('Curve:Cubic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' VRFHeatCapFTBoundary')
  CALL AddToObjStr('Coefficient1 Constant','-7.6000882')
  CALL AddToObjStr('Coefficient2 x','3.05090016')
  CALL AddToObjStr('Coefficient3 x**2','-0.1162844')
  CALL AddToObjStr('Coefficient4 x**3','0.0')
  CALL AddToObjStr('Minimum Value of x','15')
  CALL AddToObjStr('Maximum Value of x','27')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Temperature')
  CALL AddToObjStr('Output Unit Type','Temperature',.TRUE.)

  CALL CreateNewObj('Curve:Biquadratic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' VRFHeatCapFTHi')
  CALL AddToObjStr('Coefficient1 Constant','1.161134821')
  CALL AddToObjStr('Coefficient2 x','0.027478868')
  CALL AddToObjStr('Coefficient3 x**2','-0.00168795')
  CALL AddToObjStr('Coefficient4 y','0.001783378')
  CALL AddToObjStr('Coefficient5 y**2','2.03208E-06')
  CALL AddToObjStr('Coefficient6 x*y','-6.8969E-05')
  CALL AddToObjStr('Minimum Value of x','15')
  CALL AddToObjStr('Maximum Value of x','27')
  CALL AddToObjStr('Minimum Value of y','-10')
  CALL AddToObjStr('Maximum Value of y','15')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Temperature')
  CALL AddToObjStr('Input Unit Type for Y','Temperature')
  CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)

  CALL CreateNewObj('Curve:Biquadratic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' VRFHeatEIRFT')
  CALL AddToObjStr('Coefficient1 Constant','0.87465501')
  CALL AddToObjStr('Coefficient2 x','-0.01319754')
  CALL AddToObjStr('Coefficient3 x**2','0.00110307')
  CALL AddToObjStr('Coefficient4 y','-0.0133118')
  CALL AddToObjStr('Coefficient5 y**2','0.00089017')
  CALL AddToObjStr('Coefficient6 x*y','-0.00012766')
  CALL AddToObjStr('Minimum Value of x','15')
  CALL AddToObjStr('Maximum Value of x','27')
  CALL AddToObjStr('Minimum Value of y','-20')
  CALL AddToObjStr('Maximum Value of y','12')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Temperature')
  CALL AddToObjStr('Input Unit Type for Y','Temperature')
  CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)

  CALL CreateNewObj('Curve:Cubic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' VRFHeatEIRFTBoundary')
  CALL AddToObjStr('Coefficient1 Constant','-7.6000882')
  CALL AddToObjStr('Coefficient2 x','3.05090016')
  CALL AddToObjStr('Coefficient3 x**2','-0.1162844')
  CALL AddToObjStr('Coefficient4 x**3','0.0')
  CALL AddToObjStr('Minimum Value of x','15')
  CALL AddToObjStr('Maximum Value of x','27')
  CALL AddToObjStr('Minimum Curve Output','-20')
  CALL AddToObjStr('Maximum Curve Output','15')
  CALL AddToObjStr('Input Unit Type for X','Temperature')
  CALL AddToObjStr('Output Unit Type','Temperature',.TRUE.)

  CALL CreateNewObj('Curve:Biquadratic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' VRFHeatEIRFTHi')
  CALL AddToObjStr('Coefficient1 Constant','2.504005146')
  CALL AddToObjStr('Coefficient2 x','-0.05736767')
  CALL AddToObjStr('Coefficient3 x**2','4.07336E-05')
  CALL AddToObjStr('Coefficient4 y','-0.12959669')
  CALL AddToObjStr('Coefficient5 y**2','0.00135839')
  CALL AddToObjStr('Coefficient6 x*y','0.00317047')
  CALL AddToObjStr('Minimum Value of x','15')
  CALL AddToObjStr('Maximum Value of x','27')
  CALL AddToObjStr('Minimum Value of y','-10')
  CALL AddToObjStr('Maximum Value of y','15')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Temperature')
  CALL AddToObjStr('Input Unit Type for Y','Temperature')
  CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)

  CALL CreateNewObj('Curve:Cubic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' HeatingEIRLowPLR')
  CALL AddToObjStr('Coefficient1 Constant','0.1400093')
  CALL AddToObjStr('Coefficient2 x','0.6415002')
  CALL AddToObjStr('Coefficient3 x**2','0.1339047')
  CALL AddToObjStr('Coefficient4 x**3','0.0845859')
  CALL AddToObjStr('Minimum Value of x','0')
  CALL AddToObjStr('Maximum Value of x','1')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Dimensionless')
  CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)

  CALL CreateNewObj('Curve:Quadratic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' HeatingEIRHiPLR')
  CALL AddToObjStr('Coefficient1 Constant','2.4294355')
  CALL AddToObjStr('Coefficient2 x','-2.235887')
  CALL AddToObjStr('Coefficient3 x**2','0.8064516')
  CALL AddToObjStr('Minimum Value of x','1.0')
  CALL AddToObjStr('Maximum Value of x','1.5')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Dimensionless')
  CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)

  CALL CreateNewObj('Curve:Linear')
  CALL AddToObjFld('Name', base + vrfsNameOff,' HeatingCombRatio')
  CALL AddToObjStr('Coefficient1 Constant','0.96034')
  CALL AddToObjStr('Coefficient2 x','0.03966')
  CALL AddToObjStr('Minimum Value of x','1.0')
  CALL AddToObjStr('Maximum Value of x','1.5')
  CALL AddToObjStr('Minimum Curve Output','1.0')
  CALL AddToObjStr('Maximum Curve Output','1.023')
  CALL AddToObjStr('Input Unit Type for X','Dimensionless')
  CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)

  CALL CreateNewObj('Curve:Biquadratic')
  CALL AddToObjFld('Name', base + vrfsNameOff,' CoolingLengthCorrectionFactor')
  CALL AddToObjStr('Coefficient1 Constant','1.0693794')
  CALL AddToObjStr('Coefficient2 x','-0.0014951')
  CALL AddToObjStr('Coefficient3 x**2','2.56E-06')
  CALL AddToObjStr('Coefficient4 y','-0.1151104')
  CALL AddToObjStr('Coefficient5 y**2','0.0511169')
  CALL AddToObjStr('Coefficient6 x*y','-0.0004369')
  CALL AddToObjStr('Minimum Value of x','8')
  CALL AddToObjStr('Maximum Value of x','175')
  CALL AddToObjStr('Minimum Value of y','0.5')
  CALL AddToObjStr('Maximum Value of y','1.5')
  CALL AddToObjStr('Minimum Curve Output','')
  CALL AddToObjStr('Maximum Curve Output','')
  CALL AddToObjStr('Input Unit Type for X','Distance')
  CALL AddToObjStr('Input Unit Type for Y','Dimensionless')
  CALL AddToObjStr('Output Unit Type','Dimensionless',.TRUE.)

END DO
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE CreateFanCoil
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   March 2006
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a Fan Coil system.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for cooling coil type
INTEGER, PARAMETER :: ccChWater = 1
INTEGER, PARAMETER :: ccChWaterDetailed = 2
INTEGER, PARAMETER :: ccHXAssistedChWater = 3
          ! parameters for baseboard type
INTEGER, PARAMETER :: baseboardNone           = 1
INTEGER, PARAMETER :: HotWater       = 2
INTEGER, PARAMETER :: Electric       = 3
          ! parameters for fan coil capacity control type
INTEGER, PARAMETER :: capctrlConstFanVarFlow  = 1
INTEGER, PARAMETER :: capctrlCyclFan          = 2
INTEGER, PARAMETER :: capctrlVarFanVarFlow    = 3
INTEGER, PARAMETER :: capctrlVarFanConstFlow  = 4
INTEGER, PARAMETER :: capctrlMultiSpeedFan    = 5
INTEGER, PARAMETER :: capctrlASHRAE90VarFan   = 6

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for
INTEGER :: iZone
LOGICAL :: isTStatNameBlank
LOGICAL :: isSupplyMaxRateAutosize
LOGICAL :: isDedOutAirNameBlank
INTEGER :: coolCoilKind=0
INTEGER :: heatCoilKind=0
LOGICAL :: isOATypeDetailed = .FALSE.
LOGICAL :: isBaseboardNone = .FALSE.
INTEGER :: baseboardKind = 0
INTEGER :: capacityControlKind = 0
LOGICAL :: isCapacityControlKindBlank = .FALSE.
CHARACTER(len=1) :: NextSequenceNumber = '1'

DO iZone = 1, numCompactFanCoil
  base =  compactFanCoilBase(iZone)
  ! set flags
  isTStatNameBlank = (FldVal(base + fczTStatNameOff) .EQ. '')
  isSupplyMaxRateAutosize = SameString(FldVal(base + fczSupplyMaxRateOff),'autosize')
  isDedOutAirNameBlank = (FldVal(base + fczDedOutAirSysNameOff) .EQ. '')
  isOATypeDetailed = SameString(FldVal(base + fczOutAirTypeOff),'DetailedSpecification')
  isBaseboardNone = SameString(FldVal(base + fczBaseboardTypeOff),'None')
  IF (SameString(FldVal(base + fczCoolCoilKindOff),'ChilledWater')) THEN
    coolCoilKind = ccChWater
  ELSEIF (SameString(FldVal(base + fczCoolCoilKindOff),'ChilledWaterDetailedFlatModel')) THEN
    coolCoilKind = ccChWaterDetailed
  ELSEIF (SameString(FldVal(base + fczCoolCoilKindOff),'HeatExchangerAssistedChilledWater')) THEN
    coolCoilKind = ccHXAssistedChWater
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:FanCoil "'//TRIM(FldVal(base + fczNameOff))//'"'// &
                    ' in the Cooling Coil Type field: '//TRIM(FldVal(base + fczCoolCoilKindOff)))
  END IF
  IF (SameString(FldVal(base + fczHeatCoilKindOff),'HotWater')) THEN
    heatCoilKind = HotWater
  ELSEIF (SameString(FldVal(base + fczHeatCoilKindOff),'Electric')) THEN
    heatCoilKind = Electric
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:FanCoil "'//TRIM(FldVal(base + fczNameOff))//'"'// &
                    ' in the Heating Coil Type field: '//TRIM(FldVal(base + fczHeatCoilKindOff)))
  END IF
  !validate cooling design method
  IF (SameString(FldVal(base +  fczCoolDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  fczCoolDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:FanCoil "'//TRIM(FldVal(base + fczNameOff))//'"'// &
                    ' in the Zone Cooling Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + fczCoolDesignTempMethodOff)))
  END IF
  !validate heating design method
  IF (SameString(FldVal(base +  fczHeatDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  fczHeatDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:FanCoil "'//TRIM(FldVal(base + fczNameOff))//'"'// &
                    ' in the Zone Heating Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + fczHeatDesignTempMethodOff)))
  END IF
  !set the baseboard indicator
  IF (SameString(FldVal(base +  fczBaseboardTypeOff),'None')) THEN
    baseboardKind = baseboardNone
  ELSEIF (SameString(FldVal(base +  fczBaseboardTypeOff),'HotWater')) THEN
    baseboardKind = HotWater
  ELSEIF (SameString(FldVal(base +  fczBaseboardTypeOff),'Electric')) THEN
    baseboardKind = Electric
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:FanCoil "'//TRIM(FldVal(base + fczNameOff))//'"'// &
                    ' in the Baseboard Heating Type field: '//TRIM(FldVal(base + fczBaseboardTypeOff)))
  END IF

    !set the capacity control indicator
    ! if blank, set default based on DOAS status
  isCapacityControlKindBlank = (FldVal(base + fczCapCtrlTypeOff) .EQ. '')
  IF (isCapacityControlKindBlank) THEN
    IF (isDedOutAirNameBlank) THEN
      FldVal(base + fczCapCtrlTypeOff) = 'ConstantFanVariableFlow'
    ELSE
      FldVal(base + fczCapCtrlTypeOff) = 'CyclingFan'
    END IF
  END IF
  IF (SameString(FldVal(base +  fczCapCtrlTypeOff),'ConstantFanVariableFlow')) THEN
    capacityControlKind = capctrlConstFanVarFlow
  ELSEIF (SameString(FldVal(base +  fczCapCtrlTypeOff),'CyclingFan')) THEN
    capacityControlKind = capctrlCyclFan
  ELSEIF (SameString(FldVal(base +  fczCapCtrlTypeOff),'VariableFanVariableFlow')) THEN
    capacityControlKind = capctrlVarFanVarFlow
  ELSEIF (SameString(FldVal(base +  fczCapCtrlTypeOff),'VariableFanConstantFlow')) THEN
    capacityControlKind = capctrlVarFanConstFlow
  ELSEIF (SameString(FldVal(base +  fczCapCtrlTypeOff),'MultiSpeedFan')) THEN
    capacityControlKind = capctrlMultiSpeedFan
  ELSEIF (SameString(FldVal(base +  fczCapCtrlTypeOff),'ASHRAE90VariableFan')) THEN
    capacityControlKind = capctrlASHRAE90VarFan
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:FanCoil "'//TRIM(FldVal(base + fczNameOff))//'"'// &
                    ' in the Capacity Control Method field: '//TRIM(FldVal(base + fczCapCtrlTypeOff)))
  END IF

  ! warn if fan control is constant fan with a DOAS
  IF ((capacityControlKind == capctrlConstFanVarFlow) .AND. (.NOT. isDedOutAirNameBlank)) THEN
    CALL WriteError('In HVACTemplate:Zone:FanCoil "'//TRIM(FldVal(base + fczNameOff))//'"'// &
                    ' the Capacity Control Method is '//TRIM(FldVal(base + fczCapCtrlTypeOff)) // &
                    ' and the zone is served by a dedicated outdoor air system.',msgWarning)
  END IF

  !repeat per zone - only if fczTStatNameOff is not blank
  IF (.NOT. isTStatNameBlank) THEN
    !ZONE CONTROL:THERMOSTATIC ~ line 5
    CALL AddAlwaysSchedule('4')
    CALL CreateNewObj('ZoneControl:Thermostat')
    CALL AddToObjFld('Name', base + fczNameOff,' Thermostat')
    CALL AddToObjFld('Zone or ZoneList Name', base + fczNameOff,'')
    CALL AddToObjStr('Control Type Schedule Name','HVACTemplate-Always 4')
    CALL AddToObjStr('Control Object Type','ThermostatSetpoint:DualSetpoint')
    CALL AddToObjFld('Control Name', base + fczTStatNameOff,' Dual SP Control',.TRUE.)
  END IF
  !ZONE SIZING ~ line 12
  ! OADCV Changes
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:OutdoorAir')
    CALL AddToObjStr('Name', 'SZ DSOA '//FldVal(base+fczNameOff))
    CALL AddToObjFld('Outdoor Air Method', base + fczOutAirTypeOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Person {m3/s}', base + fczOutAirFlowPersonOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone Floor Area {m3/s-m2}', base + fczOutAirFlowAreaOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone {m3/s}', base + fczOutAirFlowZoneOff,'',.true.)
  END IF

  CALL CreateNewObj('Sizing:Zone')
  CALL AddToObjFld('Zone or ZoneList Name', base + fczNameOff,'')
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Input Method',base + fczCoolDesignTempMethodOff,'')
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', base + fczCoolSetPtDesignOff,'')
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Difference {delta C}',base + fczCoolDesignTempDiffOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Input Method',base + fczHeatDesignTempMethodOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', base + fczHeatSetPtDesignOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Difference {delta C}',base + fczHeatDesignTempDiffOff,'')
  CALL AddToObjStr('Zone Cooling Design Supply Air Humidity Ratio {kg-H20/kg-air}','0.008')
  CALL AddToObjStr('Zone Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + fczDesSpecOANameOff,'')
  ELSE
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','SZ DSOA '//FldVal(base+fczNameOff))
  END IF
  CALL AddToObjFld('Zone Heating Sizing Factor', base + fczZoneHeatSizingFactorOff,'')
  CALL AddToObjFld('Zone Cooling Sizing Factor', base + fczZoneCoolSizingFactorOff,'')
  IF (isSupplyMaxRateAutosize) THEN
    CALL AddToObjStr('Cooling Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Cooling Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
    CALL AddToObjStr('Heating Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Heating Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
  ELSE
    CALL AddToObjStr('Cooling Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Cooling Design Air Flow Rate {m3/s}', base + fczSupplyMaxRateOff,'')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
    CALL AddToObjStr('Heating Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Heating Design Air Flow Rate {m3/s}', base + fczSupplyMaxRateOff,'')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
  END IF
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Zone Air Distribution Object Name', base + fczDesSpecOADistNameOff,'',.true.)
  ELSE
    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+fczNameOff),.true.)
  END IF

  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:ZoneAirDistribution')
    CALL AddToObjStr('Name','SZ DSZAD '//FldVal(base+fczNameOff))
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Cooling Mode {}','1')
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Heating Mode {}','1',.true.)
  END IF
  !CONTROLLED ZONE EQUIP CONFIGURATION ~ line 38
  CALL CreateNewObj('ZoneHVAC:EquipmentConnections')
  CALL AddToObjFld('Zone Name', base + fczNameOff,'             ')
  CALL AddToObjFld('Zone Conditioning Equipment List Name', base + fczNameOff,' Equipment')
  IF (isDedOutAirNameBlank) THEN
    CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base + fczNameOff,' Supply Inlet')
  ELSE
    CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base + fczNameOff,' Inlets')
  END IF
  CALL AddToObjFld('Zone Air Exhaust Node or NodeList Name', base + fczNameOff,' Fan Coil Return')
  CALL AddToObjFld('Zone Air Node Name', base + fczNameOff,' Zone Air Node')
  CALL AddToObjFld('Zone Return Air Node Name', base + fczNameOff,' Return Outlet',.TRUE.)
  !NODELIST (not in spreadsheet added with DOAS)
  IF (.NOT. isDedOutAirNameBlank) THEN
    CALL CreateNewObj('NodeList')
    CALL AddToObjFld('Name', base + fczNameOff,' Inlets')
    CALL AddToObjFld('Node Name', base + fczNameOff,' Supply Inlet')
    CALL AddToObjFld('Node Name', base + fczNameOff,' DOAS Supply Inlet',.TRUE.)
  ENDIF
  !ZONE EQUIPMENT LIST ~ line 46
  CALL CreateNewObj('ZoneHVAC:EquipmentList')
  CALL AddToObjFld('Name', base + fczNameOff,' Equipment')
  CALL AddToObjStr('Load Distribution Scheme','SequentialLoad')

  NextSequenceNumber = '1'
  IF (.NOT. isDedOutAirNameBlank) THEN
    CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Zone Equipment Name', base + fczNameOff,' DOAS ATU')
    CALL AddToObjStr('Zone Equipment Cooling Sequence',trim(NextSequenceNumber))
    CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence',trim(NextSequenceNumber))
    CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
    CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.FALSE.)
  NextSequenceNumber = '2'
  END IF
  CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:FourPipeFanCoil')
  CALL AddToObjFld('Zone Equipment Name', base + fczNameOff,' Fan Coil')
  CALL AddToObjStr('Zone Equipment Cooling Sequence',trim(NextSequenceNumber))
  CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence',trim(NextSequenceNumber),.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',isBaseboardNone)
IF (.NOT. isBaseboardNone) THEN
    IF (isDedOutAirNameBlank) THEN
      NextSequenceNumber = '2'
    ELSE
      NextSequenceNumber = '3'
    END IF
    SELECT CASE (baseboardKind)
      CASE (HotWater)
        CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CASE (Electric)
        CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:Convective:Electric')
    END SELECT
        CALL AddToObjFld('Zone Equipment Name', base + fczNameOff,' Baseboard Heat')
        CALL AddToObjStr('Zone Equipment Cooling Sequence',trim(NextSequenceNumber))
        CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence',trim(NextSequenceNumber),.FALSE.)
        CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
        CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.TRUE.)
  END IF

  IF (.NOT. isDedOutAirNameBlank) THEN
  !***ZoneHVAC:AirDistributionUnit
    CALL CreateNewObj('ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Name', base + fczNameOff,' DOAS ATU')
    CALL AddToObjFld('Air Distribution Unit Outlet Node Name', base + fczNameOff,' DOAS Supply Inlet')
    CALL AddToObjStr('Air Terminal Object Type','AirTerminal:SingleDuct:VAV:NoReheat')
    CALL AddToObjFld('Air Terminal Name', base + fczNameOff,' DOAS Air Terminal',.TRUE.)
    !***AirTerminal:SingleDuct:VAV:NoReheat
    CALL CreateNewObj('AirTerminal:SingleDuct:VAV:NoReheat')
    CALL AddToObjFld('Name', base + fczNameOff,' DOAS Air Terminal')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Air Outlet Node Name', base + fczNameOff,' DOAS Supply Inlet')
    CALL AddToObjFld('Air Inlet Node Name', base + fczNameOff,' Zone Equip Inlet')
    CALL AddToObjStr('Maximum Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Zone Minimum Air Flow Input Method','Constant')
    CALL AddToObjStr('Constant Minimum Air Flow Fraction','1.0')
    CALL AddToObjStr('Fixed Minimum Air Flow Rate {m3/s}','')
    CALL AddToObjStr('Minimum Air Flow Fraction Schedule Name','')
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','',.TRUE.)
  ENDIF
  !FAN COIL UNIT:4 PIPE ~ line 53
  IF (isCapacityControlKindBlank) THEN
    CALL WriteComment('Capacity Control Method was blank, defaulted to '//TRIM(FldVal(base + fczCapCtrlTypeOff)))
  END IF
  IF (.NOT. isDedOutAirNameBlank) THEN
    CALL WriteComment('Setting fan coil outdoor air to zero, because zone is served by dedicated outdoor air system (DOAS)')
  END IF
  CALL CreateNewObj('ZoneHVAC:FourPipeFanCoil')
  CALL AddToObjFld('Name', base + fczNameOff,' Fan Coil')
  CALL AddToObjFld('Availability Schedule Name', base + fczSysAvailSchedNameOff,'')
  CALL AddToObjFld('Capacity Control Method', base + fczCapCtrlTypeOff,'')
  !CR8001
  !CALL AddToObjFld('maximum air flow rate {m3/s}', base + fczSupplyMaxRateOff,'')
  CALL AddToObjStr('Maximum Supply Air Flow Rate {m3/s}', 'autosize')
  IF (capacityControlKind == capctrlASHRAE90VarFan) THEN
    CALL AddToObjStr('Low Speed Supply Air Flow Ratio','0.5')
  ELSE
    CALL AddToObjFld('Low Speed Supply Air Flow Ratio', base + fczLowSpdSupplyRatioOff,'')
  END IF
  CALL AddToObjFld('Medium Speed Supply Air Flow Ratio', base + fczMedSpdSupplyRatioOff,'')
  IF (isDedOutAirNameBlank) THEN
    CALL AddToObjStr('Maximum Outdoor Air Flow Rate {m3/s}','autosize')
  ELSE
    CALL AddToObjStr('Maximum Outdoor Air Flow Rate {m3/s}','0.0')
  END IF
  CALL AddToObjFld('Outdoor Air Schedule Name', base + fczOutAirSchedNameOff,'')
  CALL AddToObjFld('Air Inlet Node Name', base + fczNameOff,' Fan Coil Return')
  CALL AddToObjFld('Air Outlet Node Name', base + fczNameOff,' Supply Inlet')
  ! deprecate and remove fields
!  CALL AddToObjFld('Outdoor Air Node Name', base + fczNameOff,' Outside Air Inlet')
!  CALL AddToObjFld('Relief Air Node Name', base + fczNameOff,' Relief Air Outlet')
  CALL AddToObjStr('Outdoor Air Mixer Object Type', 'OutdoorAir:Mixer')
  CALL AddToObjFld('Outdoor Air Mixer Name', base + fczNameOff,' OA Mixing Box')
  CALL AddToObjStr('Supply Air Fan Object Type', 'Fan:SystemModel')
  CALL AddToObjFld('Fan Name', base + fczNameOff,' Supply Fan')
  IF (coolCoilKind .EQ. ccChWater) THEN
    CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:Water')
    CALL AddToObjFld('Cooling Coil Name', base + fczNameOff,' Cooling Coil')
  ELSEIF (coolCoilKind .EQ. ccChWaterDetailed) THEN
    CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:Water:DetailedGeometry')
    CALL AddToObjFld('Cooling Coil Name', base + fczNameOff,' Cooling Coil')
  ELSEIF (coolCoilKind .EQ. ccHXAssistedChWater) THEN
    CALL AddToObjStr('Cooling Coil Object Type','CoilSystem:Cooling:Water:HeatExchangerAssisted')
    CALL AddToObjFld('Cooling Coil Name', base + fczNameOff,' Heat Exchanger Assisted Cooling Coil')
  END IF
  CALL AddToObjStr('Maximum Cold Water Flow Rate {m3/s}','autosize')
  CALL AddToObjStr('Minimum Cold Water Flow Rate {m3/s}','0')
  CALL AddToObjStr('Cooling Convergence Tolerance','0.001')
  IF (heatCoilKind .EQ. HotWater) THEN
    CALL AddToObjStr('Heating Coil Object Type','Coil:Heating:Water')
  ELSE
    CALL AddToObjStr('Heating Coil Object Type','Coil:Heating:Electric')
  END IF
  CALL AddToObjFld('Heating Coil Name', base + fczNameOff,' Heating Coil')
  CALL AddToObjStr('Maximum Hot Water Flow Rate {m3/s}','autosize')
  CALL AddToObjStr('Minimum Hot Water Flow Rate{m3/s}','0')
  IF (capacityControlKind == capctrlASHRAE90VarFan) THEN
    CALL AddToObjStr('Heating Convergence Tolerance','0.001')
    CALL AddToObjStr('Availability Manager List Name','')
    CALL AddToObjStr('Design Specification ZoneHVAC Sizing Object Name','')
    CALL AddToObjStr('Supply Air Fan Operating Mode Schedule Name','HVACTemplate-Always 1')
    CALL AddToObjStr('Minimum Supply Air Temperature in Cooling Mode {C}','autosize')
    CALL AddToObjStr('Maximum Supply Air Temperature in Heating Mode {C}','autosize',.true.)
  ELSE
    CALL AddToObjStr('Heating Convergence Tolerance','0.001',.true.)
  END IF
  ! Supply fan
  CALL CreateNewObj('Fan:SystemModel')
  CALL AddToObjFld('Name', base + fczNameOff,' Supply Fan')
  CALL AddToObjFld('Availability Schedule Name', base + fczSysAvailSchedNameOff,'')
  CALL AddToObjFld('Air Inlet Node Name', base + fczNameOff,' Mixed Air Outlet')
  CALL AddToObjFld('Air Outlet Node Name', base + fczNameOff,' Supply Fan Outlet')
  CALL AddToObjStr('Design Maximum Air Flow Rate {m3/s}','autosize')
  IF (capacityControlKind == capctrlVarFanConstFlow) THEN
    CALL AddToObjStr('Speed Control Method', 'Continuous')
  ELSE
    CALL AddToObjStr('Speed Control Method', 'Discrete')
  END IF
  CALL AddToObjStr('Electric Power Minimum Flow Rate Fraction', '0.0')
  CALL AddToObjFld('Design Pressure Rise {Pa}', base + fczSupplyPressureOff,'')
  CALL AddToObjFld('Motor Efficiency', base + fczSupplyFanMotorEffOff,'')
  CALL AddToObjFld('Motor in Airstream Fraction', base + fczSupplyFanMotorFracOff,'')
  CALL AddToObjStr('Design Electric Power Consumption {W}','autosize')
  CALL AddToObjStr('Design Power Sizing Method','TotalEfficiencyAndPressure')
  CALL AddToObjStr('Electric Power Per Unit Flow Rate {W/(m3/s)}','')
  CALL AddToObjStr('Electric Power Per Unit Flow Rate Per Unit Pressure {W/((m3/s)-Pa)}','')
  CALL AddToObjFld('Fan Total Efficiency', base + fczSupplyEfficiencyOff,'')
  IF ((capacityControlKind == capctrlConstFanVarFlow) .OR. (capacityControlKind == capctrlCyclFan)) THEN
    CALL AddToObjStr('Electric Power Function of Flow Fraction Curve Name','')
  ELSE
    CALL AddToObjFld('Electric Power Function of Flow Fraction Curve Name', base + fczNameOff,' Fan Power Curve')
  END IF
  CALL AddToObjStr('Night Ventilation Mode Pressure Rise {Pa}','')
  CALL AddToObjStr('Night Ventilation Mode Flow Fraction','')
  CALL AddToObjStr('Motor Loss Zone Name','')
  CALL AddToObjStr('Motor Loss Radiative Fraction','')
  CALL AddToObjStr('End-Use Subcategory','General')
  IF (capacityControlKind == capctrlMultiSpeedFan) THEN
    CALL AddToObjStr('Number of Speeds','3')
    CALL AddToObjFld('Speed 1 Flow Fraction', base + fczLowSpdSupplyRatioOff,'')
    CALL AddToObjStr('Speed 1 Electric Power Fraction','')
    CALL AddToObjStr('Speed 2 Flow Fraction','0.66')
    CALL AddToObjStr('Speed 2 Electric Power Fraction','')
    CALL AddToObjStr('Speed 3 Flow Fraction','1.0')
    CALL AddToObjStr('Speed 3 Electric Power Fraction','1.0',.TRUE.)
  ELSE IF (capacityControlKind == capctrlASHRAE90VarFan) THEN
    CALL AddToObjStr('Number of Speeds','2')
    CALL AddToObjFld('Speed 1 Flow Fraction', base + fczLowSpdSupplyRatioOff,'')
    CALL AddToObjStr('Speed 1 Electric Power Fraction','')
    CALL AddToObjStr('Speed 2 Flow Fraction','1.0')
    CALL AddToObjStr('Speed 2 Electric Power Fraction','1.0',.TRUE.)
  ELSE
    CALL AddToObjStr('Number of Speeds','1')
    CALL AddToObjStr('Speed 1 Flow Fraction','1.0')
    CALL AddToObjStr('Speed 1 Electric Power Fraction','1.0',.TRUE.)
  END IF
  IF ((capacityControlKind .NE. capctrlConstFanVarFlow) .AND. (capacityControlKind .NE. capctrlCyclFan)) THEN
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name',  base + fczNameOff,' Fan Power Curve')
    CALL AddToObjStr('Coefficient1 Constant','0.0015302446')
    CALL AddToObjStr('Coefficient2 x','0.0052080574')
    CALL AddToObjStr('Coefficient3 x**2','1.1086242')
    CALL AddToObjStr('Coefficient4 x**3','-0.11635563')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  END IF

    !OUTSIDE AIR MIXER ~ line 89
  CALL CreateNewObj('OutdoorAir:Mixer')
  CALL AddToObjFld('Name', base + fczNameOff,' OA Mixing Box')
  CALL AddToObjFld('Mixed Air Node Name', base + fczNameOff,' Mixed Air Outlet')
  CALL AddToObjFld('Outside Air Stream Node Name', base + fczNameOff,' Outside Air Inlet')
  CALL AddToObjFld('Relief Air Stream Node Name', base + fczNameOff,' Relief Air Outlet')
  CALL AddToObjFld('Return Air Stream Node Name', base + fczNameOff,' Fan Coil Return',.TRUE.)
  !OUTSIDE AIR INLET NODE LIST ~ line 96
  CALL CreateNewObj('OutdoorAir:NodeList')
  CALL AddToObjFld('Node or NodeList Name 1', base + fczNameOff,' Outside Air Inlet',.TRUE.)
  !COIL:Water:DetailedFlatCooling ~ line 118
  IF (coolCoilKind .EQ. ccChWaterDetailed) THEN
    CALL CreateNewObj('Coil:Cooling:Water:DetailedGeometry')
    CALL AddToObjFld('Name', base + fczNameOff,' Cooling Coil')
    CALL AddToObjFld('Availability Schedule Name', base + fczCoolAvailSchedNameOff,'')
    CALL AddToObjStr('Max Water Flow Rate of Coil {m3/s}','autosize')
    CALL AddToObjStr('Tube Outside Surf Area {m2}','autosize')
    CALL AddToObjStr('Total Tube Inside Area {m2}','autosize')
    CALL AddToObjStr('Fin Surface Area {m2}','autosize')
    CALL AddToObjStr('Minimum Air Flow Area {m2}','autosize')
    CALL AddToObjStr('Coil Depth {m}','autosize')
    CALL AddToObjStr('Fin Diameter {m}','autosize')
    CALL AddToObjStr('Fin Thickness {m}','0.0015')
    CALL AddToObjStr('Tube Inside Diameter {m}','0.01445')
    CALL AddToObjStr('Tube Outside Diameter {m}','0.0159')
    CALL AddToObjStr('Tube Thermal Conductivity {W/m-K}','386.0')
    CALL AddToObjStr('Fin Thermal Conductivity {W/m-K}','204.0')
    CALL AddToObjStr('Fin Spacing {m}','0.0018')
    CALL AddToObjStr('Tube Depth Spacing {m}','0.026')
    CALL AddToObjStr('Number of Tube Rows','4')
    CALL AddToObjStr('Number of Tubes per Row','autosize')
    CALL AddToObjFld('Coil Water Inlet Node', base + fczNameOff,' Cooling Coil ChW Inlet')
    CALL AddToObjFld('Coil Water Outlet Node', base + fczNameOff,' Cooling Coil ChW Outlet')
    CALL AddToObjFld('Air Inlet Node Name', base + fczNameOff,' Supply Fan Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + fczNameOff,' Cooling Coil Outlet',.TRUE.)
  ELSEIF (coolCoilKind .EQ. ccChWater) THEN
  !COIL:Water:Cooling ~ line 142
    CALL CreateNewObj('Coil:Cooling:Water')
    CALL AddToObjFld('Name', base + fczNameOff,' Cooling Coil')
    CALL AddToObjFld('Availability Schedule Name', base + fczCoolAvailSchedNameOff,'')
    CALL AddToObjStr('Design Water Flow Rate {m3/s}','autosize')
    !CR8001
    !CALL AddToObjFld('Design Air Flow Rate {m3/s}', base + fczSupplyMaxRateOff,'')
    CALL AddToObjStr('Design Air Flow Rate {m3/s}', 'autosize')
    CALL AddToObjStr('Design Inlet Water Temperature {C}','autosize')
    CALL AddToObjStr('Design Inlet Air Temperature {C}','autosize')
    CALL AddToObjStr('Design Outlet Air Temperature {C}','autosize')
    CALL AddToObjStr('Design Inlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
    CALL AddToObjStr('Design Outlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
    CALL AddToObjFld('Water Inlet Node Name', base + fczNameOff,' Cooling Coil ChW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + fczNameOff,' Cooling Coil ChW Outlet')
    CALL AddToObjFld('Air Inlet Node Name', base + fczNameOff,' Supply Fan Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + fczNameOff,' Cooling Coil Outlet')
    CALL AddToObjStr('Type of Analysis','DetailedAnalysis')
    CALL AddToObjStr('Heat Exchanger Configuration','CrossFlow',.TRUE.)
  ELSEIF (coolCoilKind .EQ. ccHXAssistedChWater) THEN
    !***CoilSystem:Cooling:Water:HeatExchangerAssisted
    CALL CreateNewObj('CoilSystem:Cooling:Water:HeatExchangerAssisted')
    CALL AddToObjFld('Name', base + fczNameOff,' Heat Exchanger Assisted Cooling Coil')
    CALL AddToObjStr('Heat Exchanger Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat Exchanger Name', base + fczNameOff,' Cooling Coil Heat Exchanger')
    CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:Water')
    CALL AddToObjFld('Cooling Coil Name', base + fczNameOff, ' Cooling Coil',.TRUE.)
    !***HeatExchanger:AirToAir:SensibleAndLatent
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Name', base + fczNameOff,' Cooling Coil Heat Exchanger')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjStr('Nominal Supply Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Sensible Effectiveness at 100% Heating Air Flow {dimensionless}','0.7')
    CALL AddToObjStr('Latent Effectiveness at 100% Heating Air Flow {dimensionless}','0.0')
    CALL AddToObjStr('Sensible Effectiveness at 75% Heating Air Flow {dimensionless}','0.75')
    CALL AddToObjStr('Latent Effectiveness at 75% Heating Air Flow {dimensionless}','0.0')
    CALL AddToObjStr('Sensible Effectiveness at 100% Cooling Air Flow {dimensionless}','0.7')
    CALL AddToObjStr('Latent Effectiveness at 100% Cooling Air Flow {dimensionless}','0.0')
    CALL AddToObjStr('Sensible Effectiveness at 75% Cooling Air Flow {dimensionless}','0.75')
    CALL AddToObjStr('Latent Effectiveness at 75% Cooling Air Flow {dimensionless}','0.0')
    CALL AddToObjFld('Supply Air Inlet Node Name', base + fczNameOff,' Supply Fan Outlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + fczNameOff,' HX Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + fczNameOff,' HX Exhaust Inlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + fczNameOff,' Cooling Coil Outlet')
    CALL AddToObjStr('Nominal Electric Power {W}','0')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','No')
    CALL AddToObjStr('Heat Exchanger Type','Plate')
    CALL AddToObjStr('Frost Control Type','None')
    CALL AddToObjStr('Threshold Temperature {C}','')
    CALL AddToObjStr('Initial Defrost Time Fraction {dimensionless}','')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase {1/K}','')
    ! MJW ???? - Not sure
    CALL AddToObjStr('Economizer Lockout','No',.TRUE.)
    CALL CreateNewObj('Coil:Cooling:Water')
    CALL AddToObjFld('Name', base + fczNameOff,' Cooling Coil')
    CALL AddToObjFld('Availability Schedule Name', base + fczCoolAvailSchedNameOff,'')
    CALL AddToObjStr('Design Water Flow Rate {m3/s}','autosize')
    !CR8001
    !CALL AddToObjFld('Design Air Flow Rate {m3/s}', base + fczSupplyMaxRateOff,'')
    CALL AddToObjStr('Design Air Flow Rate {m3/s}', 'autosize')
    CALL AddToObjStr('Design Inlet Water Temperature {C}','autosize')
    CALL AddToObjStr('Design Inlet Air Temperature {C}','autosize')
    CALL AddToObjStr('Design Outlet Air Temperature {C}','autosize')
    CALL AddToObjStr('Design Inlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
    CALL AddToObjStr('Design Outlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
    CALL AddToObjFld('Water Inlet Node Name', base + fczNameOff,' Cooling Coil ChW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + fczNameOff,' Cooling Coil ChW Outlet')
    CALL AddToObjFld('Air Inlet Node Name', base + fczNameOff,' HX Supply Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + fczNameOff,' HX Exhaust Inlet')
    CALL AddToObjStr('Type of Analysis','DetailedAnalysis')
    CALL AddToObjStr('Heat Exchanger Configuration','CrossFlow',.TRUE.)
  END IF
  !BRANCH ~ line 169
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + fczNameOff,' Cooling Coil ChW Branch')
  !store for later in the chilled water loop
  CALL AddToStrList(TRIM(FldVal(base + fczNameOff)) // ' Cooling Coil ChW Branch', handleCoolingCoilBranch)
  CALL AddToObjStr('Pressure Drop Curve Name','')
  IF (coolCoilKind .EQ. ccChWaterDetailed) THEN
    CALL AddToObjStr('Component Object Type','Coil:Cooling:Water:DetailedGeometry')
  ELSEIF (coolCoilKind .EQ. ccChWater) THEN
    CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:Water')
  ELSEIF (coolCoilKind .EQ. ccHXAssistedChWater) THEN
    CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:Water')
  END IF
  CALL AddToObjFld('Component Name', base + fczNameOff,' Cooling Coil')
  CALL AddToObjFld('Component Inlet Node Name', base + fczNameOff,' Cooling Coil ChW Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + fczNameOff,' Cooling Coil ChW Outlet',.TRUE.)
  IF (heatCoilKind .EQ. HotWater) THEN
    !COIL:Water:SimpleHeating ~ line
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjFld('Name', base + fczNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + fczHeatAvailSchedNameOff,'')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name ', base + fczNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + fczNameOff,' Heating Coil HW Outlet')
    CALL AddToObjFld('Air Inlet Node Name', base + fczNameOff,' Cooling Coil Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + fczNameOff,' Supply Inlet')
    CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
    CALL AddToObjStr('Rated Capacity','autosize')
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
    !BRANCH ~ line
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + fczNameOff,' Heating Coil HW Branch')
    !save this branch name for later use in plant
    CALL AddToStrList(TRIM(FldVal(base + fczNameOff)) // ' Heating Coil HW Branch',handleHeatingCoilBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + fczNameOff,' Heating Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + fczNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + fczNameOff,' Heating Coil HW Outlet',.TRUE.)
  ELSE
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjFld('Name', base + fczNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + fczHeatAvailSchedNameOff,'')
    CALL AddToObjStr('Efficiency','1')
    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    CALL AddToObjFld('Air Inlet Node Name', base + fczNameOff,' Cooling Coil Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + fczNameOff,' Supply Inlet')
    CALL AddToObjStr('Coil Temp Setpoint Node','',.TRUE.)
  END IF
  
  !Baseboards
  SELECT CASE (baseboardKind)
    CASE (HotWater)
      !ZoneHVAC:Baseboard:RadiantConvective:Water
      CALL CreateNewObj('ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Name', base + fczNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + fczBaseboardAvailSchedNameOff,'')
      CALL AddToObjFld('Inlet Node Name', base + fczNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Outlet Node Name', base + fczNameOff,' Baseboard HW Outlet')
      CALL AddToObjStr('Rated Average Water Temperature {C}','87.78')
      CALL AddToObjStr('Rated Water Mass Flow Rate {m3/s}','0.063')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + fczBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Maximum Water Flow Rate','autosize')
      CALL AddToObjStr('Convergence Tolerance','0.001')
      CALL AddToObjStr('Fraction Radiant','0.0')
      CALL AddToObjStr('Fraction of Radiant Energy Incident on People','0.0',.TRUE.)
      !BRANCH (for baseboard)
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + fczNameOff,' Baseboard Heat HW Branch')
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Component Name', base + fczNameOff,' Baseboard Heat')
      CALL AddToObjFld('Component Inlet Node Name', base + fczNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + fczNameOff,' Baseboard HW Outlet',.TRUE.)
      CALL AddToStrList(TRIM(FldVal(base + fczNameOff)) // ' Baseboard Heat HW Branch',handleHeatingCoilBranch)
      IF ((compactHotLoopBase .EQ. 0) .OR. (numCompactBoiler .EQ. 0)) THEN
        CALL WriteError('Both a HVACTemplate:Plant:HotWaterLoop and a HVACTemplate:Plant:Boiler are needed ' // &
             'when using hot water baseboards.')
      END IF
    CASE (Electric)
      !BASEBOARD HEATER:Electric:Convective
      CALL CreateNewObj('ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Name', base + fczNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + fczBaseboardAvailSchedNameOff,'')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + fczBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Efficiency','1',.TRUE.)
  END SELECT
END DO
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE CreatePurchasedAir
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a compact purchased air object

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for
INTEGER :: iPurchAir
LOGICAL :: isTStatNameBlank = .FALSE.
LOGICAL :: isSupplyCoolRateAutosize = .FALSE.
LOGICAL :: isSupplyHeatRateAutosize = .FALSE.
LOGICAL :: isHeatCapAutosize = .FALSE.
LOGICAL :: isCoolCapAutosize = .FALSE.
LOGICAL :: isOATypeNone = .FALSE.
LOGICAL :: isOATypeDetailed = .FALSE.
LOGICAL :: isDehumCtrlTypeHumidistat = .FALSE.
LOGICAL :: isHumidCtrlTypeHumidistat = .FALSE.
LOGICAL :: isAnyZoneAutosized = .FALSE.
LOGICAL :: isNoEconomizer = .FALSE.
LOGICAL :: isCoolFlowLimited = .FALSE.

! Check if any of the HVACTemplate:Zone:IdealLoadsAirSystem objects have an autosized field
isAnyZoneAutosized = .FALSE.
DO iPurchAir = 1, numCompactPurchAir
  base =  compactPurchAirBase(iPurchAir)
  isSupplyCoolRateAutosize = SameString(FldVal(base + pazMaxCoolAirFlowRateOff),'autosize')
  isSupplyHeatRateAutosize = SameString(FldVal(base + pazMaxHeatAirFlowRateOff),'autosize')
  isHeatCapAutosize = SameString(FldVal(base + pazMaxSensHeatCapOff),'autosize')
  isCoolCapAutosize = SameString(FldVal(base + pazMaxTotalCoolCapOff),'autosize')
  isOATypeNone = SameString(FldVal(base + pazOutAirTypeOff),'None')
  IF ((.NOT. isOATypeNone) .OR. isHeatCapAutosize .OR. isCoolCapAutosize &
       .OR. isSupplyCoolRateAutosize .OR. isSupplyHeatRateAutosize) THEN
    isAnyZoneAutosized = .TRUE.
  END IF
END DO

DO iPurchAir = 1, numCompactPurchAir
  base =  compactPurchAirBase(iPurchAir)
  ! set flags
  isTStatNameBlank = (FldVal(base + pazTStatNameOff) .EQ. '')
  isSupplyCoolRateAutosize = SameString(FldVal(base + pazMaxCoolAirFlowRateOff),'autosize')
  isSupplyHeatRateAutosize = SameString(FldVal(base + pazMaxHeatAirFlowRateOff),'autosize')
  isHeatCapAutosize = SameString(FldVal(base + pazMaxSensHeatCapOff),'autosize')
  isCoolCapAutosize = SameString(FldVal(base + pazMaxTotalCoolCapOff),'autosize')
  isOATypeNone = SameString(FldVal(base + pazOutAirTypeOff),'None')
  isOATypeDetailed = SameString(FldVal(base + pazOutAirTypeOff),'DetailedSpecification')
  IF (SameString(FldVal(base + pazHeatLimitOff),'NoLimit')) THEN
  ELSEIF (SameString(FldVal(base + pazHeatLimitOff),'LimitFlowRate')) THEN
    IF (SameString(FldVal(base + pazMaxHeatAirFlowRateOff),'')) THEN
      CALL WriteError('In HVACTemplate:Zone:IdealLoadsAirSystem "'//TRIM(FldVal(base + pazNameOff))//'"'// &
                      ' the Heating Limit field is '//TRIM(FldVal(base + pazHeatLimitOff))//' but the'// &
                      ' Maximum Heating Air Flow Rate field is blank. Enter a value or autosize in this field.')
    END IF
  ELSEIF (SameString(FldVal(base + pazHeatLimitOff),'LimitCapacity')) THEN
    IF (SameString(FldVal(base + pazMaxSensHeatCapOff),'')) THEN
      CALL WriteError('In HVACTemplate:Zone:IdealLoadsAirSystem "'//TRIM(FldVal(base + pazNameOff))//'"'// &
                      ' the Heating Limit field is '//TRIM(FldVal(base + pazHeatLimitOff))//' but the'// &
                      ' Maximum Sensible Heating Capacity field is blank. Enter a value or autosize in this field.')
    END IF
  ELSEIF (SameString(FldVal(base + pazHeatLimitOff),'LimitFlowRateAndCapacity')) THEN
    IF (SameString(FldVal(base + pazMaxSensHeatCapOff),'')) THEN
      CALL WriteError('In HVACTemplate:Zone:IdealLoadsAirSystem "'//TRIM(FldVal(base + pazNameOff))//'"'// &
                      ' the Heating Limit field is '//TRIM(FldVal(base + pazHeatLimitOff))//' but the'// &
                      ' Maximum Sensible Heating Capacity field is blank. Enter a value or autosize in this field.')
    END IF
    IF (SameString(FldVal(base + pazMaxHeatAirFlowRateOff),'')) THEN
      CALL WriteError('In HVACTemplate:Zone:IdealLoadsAirSystem "'//TRIM(FldVal(base + pazNameOff))//'"'// &
                      ' the Heating Limit field is '//TRIM(FldVal(base + pazHeatLimitOff))//' but the'// &
                      ' Maximum Heating Air Flow Rate field is blank. Enter a value or autosize in this field.')
    END IF
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:IdealLoadsAirSystem "'//TRIM(FldVal(base + pazNameOff))//'"'// &
                    ' in the Heating Limit field: '//TRIM(FldVal(base + pazHeatLimitOff)))
  END IF
  isCoolFlowLimited = .FALSE.
  IF (SameString(FldVal(base + pazCoolLimitOff),'NoLimit')) THEN
  ELSEIF (SameString(FldVal(base + pazCoolLimitOff),'LimitFlowRate')) THEN
    isCoolFlowLimited = .TRUE.
    IF (SameString(FldVal(base + pazMaxCoolAirFlowRateOff),'')) THEN
      CALL WriteError('In HVACTemplate:Zone:IdealLoadsAirSystem "'//TRIM(FldVal(base + pazNameOff))//'"'// &
                      ' the Cooling Limit field is '//TRIM(FldVal(base + pazCoolLimitOff))//' but the'// &
                      ' Maximum Cooling Air Flow Rate field is blank. Enter a value or autosize in this field.')
    END IF
  ELSEIF (SameString(FldVal(base + pazCoolLimitOff),'LimitCapacity')) THEN
    IF (SameString(FldVal(base + pazMaxTotalCoolCapOff),'')) THEN
      CALL WriteError('In HVACTemplate:Zone:IdealLoadsAirSystem "'//TRIM(FldVal(base + pazNameOff))//'"'// &
                      ' the Cooling Limit field is '//TRIM(FldVal(base + pazCoolLimitOff))//' but the'// &
                      ' Maximum Total Cooling Capacity field is blank. Enter a value or autosize in this field.')
    END IF
  ELSEIF (SameString(FldVal(base + pazCoolLimitOff),'LimitFlowRateAndCapacity')) THEN
    isCoolFlowLimited = .TRUE.
    IF (SameString(FldVal(base + pazMaxTotalCoolCapOff),'')) THEN
      CALL WriteError('In HVACTemplate:Zone:IdealLoadsAirSystem "'//TRIM(FldVal(base + pazNameOff))//'"'// &
                      ' the Cooling Limit field is '//TRIM(FldVal(base + pazCoolLimitOff))//' but the'// &
                      ' Maximum Total Cooling Capacity field is blank. Enter a value or autosize in this field.')
    END IF
    IF (SameString(FldVal(base + pazMaxCoolAirFlowRateOff),'')) THEN
      CALL WriteError('In HVACTemplate:Zone:IdealLoadsAirSystem "'//TRIM(FldVal(base + pazNameOff))//'"'// &
                      ' the Cooling Limit field is '//TRIM(FldVal(base + pazCoolLimitOff))//' but the'// &
                      ' Maximum Cooling Air Flow Rate field is blank. Enter a value or autosize in this field.')
    END IF
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:IdealLoadsAirSystem "'//TRIM(FldVal(base + pazNameOff))//'"'// &
                    ' in the Cooling Limit field: '//TRIM(FldVal(base + pazCoolLimitOff)))
  END IF
  IF (SameString(FldVal(base + pazDCVTypeOff),'None')) THEN
  ELSEIF (SameString(FldVal(base + pazDCVTypeOff),'OccupancySchedule')) THEN
  ELSEIF (SameString(FldVal(base + pazDCVTypeOff),'CO2Setpoint')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:IdealLoadsAirSystem "'//TRIM(FldVal(base + pazNameOff))//'"'// &
                    ' in the Demand Controlled Ventilation Type field: '//TRIM(FldVal(base + pazDCVTypeOff)))
  END IF
  IF (SameString(FldVal(base + pazDehumCtrlTypeOff),'ConstantSensibleHeatRatio')) THEN
  ELSEIF (SameString(FldVal(base + pazDehumCtrlTypeOff),'Humidistat')) THEN
    isDehumCtrlTypeHumidistat = .true.
  ELSEIF (SameString(FldVal(base + pazDehumCtrlTypeOff),'None')) THEN
  ELSEIF (SameString(FldVal(base + pazDehumCtrlTypeOff),'ConstantSupplyHumidityRatio')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:IdealLoadsAirSystem "'//TRIM(FldVal(base + pazNameOff))//'"'// &
                    ' in the Dehumidification Control Type field: '//TRIM(FldVal(base + pazDehumCtrlTypeOff)))
  END IF
  IF (SameString(FldVal(base + pazHumidCtrlTypeOff),'None')) THEN
  ELSEIF (SameString(FldVal(base + pazHumidCtrlTypeOff),'Humidistat')) THEN
    isHumidCtrlTypeHumidistat = .true.
  ELSEIF (SameString(FldVal(base + pazHumidCtrlTypeOff),'ConstantSupplyHumidityRatio')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:IdealLoadsAirSystem "'//TRIM(FldVal(base + pazNameOff))//'"'// &
                    ' in the Humidification Control Type field: '//TRIM(FldVal(base + pazHumidCtrlTypeOff)))
  END IF
  isNoEconomizer = .FALSE.
  IF (SameString(FldVal(base + pazEconomizerTypeOff),'NoEconomizer')) THEN
    isNoEconomizer = .TRUE.
  ELSEIF (SameString(FldVal(base + pazEconomizerTypeOff),'DifferentialDryBulb')) THEN
  ELSEIF (SameString(FldVal(base + pazEconomizerTypeOff),'DifferentialEnthalpy')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:IdealLoadsAirSystem "'//TRIM(FldVal(base + pazNameOff))//'"'// &
                    ' in the Outdoor Air Economizer Type field: '//TRIM(FldVal(base + pazEconomizerTypeOff)))
  END IF
  IF (SameString(FldVal(base + pazHeatRecTypeOff),'None')) THEN
  ELSEIF (SameString(FldVal(base + pazHeatRecTypeOff),'Sensible')) THEN
  ELSEIF (SameString(FldVal(base + pazHeatRecTypeOff),'Enthalpy')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:IdealLoadsAirSystem "'//TRIM(FldVal(base + pazNameOff))//'"'// &
                    ' in the Heat Recovery Type field: '//TRIM(FldVal(base + pazHeatRecTypeOff)))
  END IF

  !repeat per zone - only if pazTStatNameOff is not blank
  IF (.NOT. isTStatNameBlank) THEN
    !ZONE CONTROL:THERMOSTATIC
    CALL AddAlwaysSchedule('4')
    CALL CreateNewObj('ZoneControl:Thermostat')
    CALL AddToObjFld('Name', base +  pazNameOff,' Thermostat')
    CALL AddToObjFld('Zone or ZoneList Name', base +  pazNameOff,' ')
    CALL AddToObjStr('Control Type Schedule Name','HVACTemplate-Always 4')
    CALL AddToObjStr('Control Object Type','ThermostatSetpoint:DualSetpoint')
    CALL AddToObjFld('Control Name', base + pazTStatNameOff,' Dual SP Control',.TRUE.)
  END IF

  ! Humdistat(s) if needed
  !    Single humidistat if humidification and dehumidification are both active
  IF ((isHumidCtrlTypeHumidistat) .AND. (isDehumCtrlTypeHumidistat)) THEN
    CALL CreateNewObj('ZoneControl:Humidistat')
    CALL AddToObjFld('Name', base + pazNameOff,' Humidistat')
    CALL AddToObjFld('Zone Name', base + pazNameOff,'')
    CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name', &
                     'HVACTemplate-Always ' // TRIM(FldVal(base + pazHumidSetPtOff)))
    CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name', &
                     'HVACTemplate-Always ' // TRIM(FldVal(base + pazDehumSetPtOff)),.TRUE.)

    CALL AddAlwaysSchedule(FldVal(base + pazHumidSetPtOff))
    CALL AddAlwaysSchedule(FldVal(base + pazDehumSetPtOff))

  ELSE
    IF (isDehumCtrlTypeHumidistat) THEN
  !   Dehumidification humidistat
      CALL CreateNewObj('ZoneControl:Humidistat')
      CALL AddToObjFld('Name', base + pazNameOff,' Humidistat')
      CALL AddToObjFld('Zone Name', base + pazNameOff,'')
      CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name','HVACTemplate-Always 1')
      CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name', &
                       'HVACTemplate-Always ' // TRIM(FldVal(base + pazDehumSetPtOff)),.TRUE.)

      CALL AddAlwaysSchedule('1')
      CALL AddAlwaysSchedule(FldVal(base + pazDehumSetPtOff))
    ENDIF
    IF (isHumidCtrlTypeHumidistat) THEN
  !    Humidification humidistat
      CALL CreateNewObj('ZoneControl:Humidistat')
      CALL AddToObjFld('Name', base + pazNameOff,' Humidification Humidistat')
      CALL AddToObjFld('Zone Name', base + pazNameOff,'')
      CALL AddToObjStr('Humidifying Relative Humidity Setpoint Schedule Name', &
                       'HVACTemplate-Always ' // TRIM(FldVal(base + pazHumidSetPtOff)))
      CALL AddToObjStr('Dehumidifying Relative Humidity Setpoint Schedule Name','HVACTemplate-Always 100',.TRUE.)

      CALL AddAlwaysSchedule(FldVal(base + pazHumidSetPtOff))
      CALL AddAlwaysSchedule('100')
    ENDIF
  ENDIF


      !ZONE SIZING - If one or more zones need a sizing:zone object, add it for all zones to prevent warnings
      !              since zone sizing will autosize every zone that has a thermostat
  IF (isAnyZoneAutosized) THEN
    IF ((.NOT. isOATypeNone) .AND. (.NOT. isOATypeDetailed)) THEN
      CALL CreateNewObj('DesignSpecification:OutdoorAir')
      CALL AddToObjStr('Name', 'SZ DSOA '//FldVal(base+pazNameOff))
      CALL AddToObjFld('Outdoor Air Method', base + pazOutAirTypeOff,' ')
      CALL AddToObjFld('Outdoor Air Flow per Person {m3/s}', base + pazOutAirFlowPersonOff,'')
      CALL AddToObjFld('Outdoor Air Flow per Zone Floor Area {m3/s-m2}', base + pazOutAirFlowAreaOff,'')
      CALL AddToObjFld('Outdoor Air Flow per Zone {m3/s}', base + pazOutAirFlowZoneOff,'',.true.)
    END IF

    CALL CreateNewObj('Sizing:Zone')
    CALL AddToObjFld('Zone or ZoneList Name', base +  pazNameOff,'')
    CALL AddToObjStr('Zone Cooling Design Supply Air Temperature Input Method','SupplyAirTemperature')
    CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', base + pazMinCoolSupplyTempOff,'')
    CALL AddToObjStr('Zone Cooling Design Supply Air Temperature Difference {delta C}','')
    CALL AddToObjStr('Zone Heating Design Supply Air Temperature Input Method','SupplyAirTemperature')
    CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', base + pazMaxHeatSupplyTempOff,'')
    CALL AddToObjStr('Zone Heating Design Supply Air Temperature Difference {delta C}','')
    CALL AddToObjFld('Zone Cooling Design Supply Air Humidity Ratio {kg-H20/kg-air}',base + pazMinCoolSupplyHumRatOff,'')
    CALL AddToObjFld('Zone Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}',base + pazMaxHeatSupplyHumRatOff,'')
    IF (isOATypeDetailed) THEN
      CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + pazDesSpecOANameOff,'')
    ELSEIF (isOATypeNone) THEN
      CALL AddToObjStr('Design Specification Outdoor Air Object Name','')
    ELSE
      CALL AddToObjStr('Design Specification Outdoor Air Object Name','SZ DSOA '//FldVal(base+pazNameOff))
    END IF
    CALL AddToObjStr('Zone Heating Sizing Factor','')
    CALL AddToObjStr('Zone Cooling Sizing Factor','')
    CALL AddToObjStr('Cooling Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Cooling Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
    CALL AddToObjStr('Heating Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Heating Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','',.true.)
  END IF
!CONTROLLED ZONE EQUIP CONFIGURATION
  CALL CreateNewObj('ZoneHVAC:EquipmentConnections')
  CALL AddToObjFld('Zone Name', base +  pazNameOff,'')
  CALL AddToObjFld('Zone Conditioning Equipment List Name', base +  pazNameOff,' Equipment')
  CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base +  pazNameOff,' Ideal Loads Supply Inlet')
  CALL AddToObjStr('Zone Air Exhaust Node or NodeList Name','')
  CALL AddToObjFld('Zone Air Node Name', base +  pazNameOff,' Zone Air Node')
  CALL AddToObjFld('Zone Return Air Node Name', base +  pazNameOff,' Return Outlet',.TRUE.)
  !ZONE EQUIPMENT LIST
  CALL CreateNewObj('ZoneHVAC:EquipmentList')
  CALL AddToObjFld('Name', base +  pazNameOff,' Equipment')
  CALL AddToObjStr('Load Distribution Scheme','SequentialLoad')
  CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:IdealLoadsAirSystem')
  CALL AddToObjFld('Zone Equipment Name', base +  pazNameOff,' Ideal Loads Air System')
  CALL AddToObjStr('Zone Equipment Cooling Sequence','1')
  CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1',.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.TRUE.)
!PURCHASED AIR
  CALL CreateNewObj('ZoneHVAC:IdealLoadsAirSystem')
  CALL AddToObjFld('Name', base +  pazNameOff,' Ideal Loads Air System')
  CALL AddToObjFld('Availability Schedule Name', base + pazSysAvailSchedNameOff,'')
  CALL AddToObjFld('Zone Supply Air Node Name', base + pazNameOff,' Ideal Loads Supply Inlet')
  CALL AddToObjStr('Zone Exhaust Air Node Name','')
  CALL AddToObjStr('System Inlet Air Node Name','')
  CALL AddToObjFld('Maximum Heating Supply Air Temperature [C]', base + pazMaxHeatSupplyTempOff,'')
  CALL AddToObjFld('Minimum Cooling Supply Air Temperature [C]', base + pazMinCoolSupplyTempOff,'')
  CALL AddToObjFld('Maximum Heating Supply Air Humidity Ratio [kg-H20/kg-air]', base + pazMaxHeatSupplyHumRatOff,'')
  CALL AddToObjFld('Minimum Cooling Supply Air Humidity Ratio [kg-H20/kg-air]', base + pazMinCoolSupplyHumRatOff,'')
  CALL AddToObjFld('Heating Limit', base + pazHeatLimitOff,'')
  CALL AddToObjFld('Maximum Heating Air Flow Rate {m3/s}   ', base + pazMaxHeatAirFlowRateOff,'')
  CALL AddToObjFld('Maximum Sensible Heating Capacity {m3/s}   ', base + pazMaxSensHeatCapOff,'')
  CALL AddToObjFld('Cooling Limit', base + pazCoolLimitOff,'')
  CALL AddToObjFld('Maximum Cooling Air Flow Rate {m3/s}   ', base + pazMaxCoolAirFlowRateOff,'')
  CALL AddToObjFld('Maximum Total Cooling Capacity {m3/s}   ', base + pazMaxTotalCoolCapOff,'')
  CALL AddToObjFld('Heating Availability Schedule Name', base + pazHeatAvailSchedNameOff,'')
  CALL AddToObjFld('Cooling Availability Schedule Name', base + pazCoolAvailScheduleNameOff,'')
  CALL AddToObjFld('Dehumidification Control Type', base + pazDehumCtrlTypeOff,'')
  CALL AddToObjFld('Cooling Sensible Heat Ratio', base + pazCoolSHROff,'')
  CALL AddToObjFld('Humidification Control Type', base + pazHumidCtrlTypeOff,'')
  IF (isOATypeNone) THEN
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','')
    CALL AddToObjStr('Outdoor Air Inlet Node Name','')
  ELSE
    IF (isOATypeDetailed) THEN
      CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + pazDesSpecOANameOff,'')
    ELSE
      CALL AddToObjStr('Design Specification Outdoor Air Object Name','SZ DSOA '//FldVal(base+pazNameOff))
    END IF
    CALL AddToObjFld('Outdoor Air Inlet Node Name', base + pazNameOff,' Ideal Loads Outdoor Air Inlet')
  ENDIF
  CALL AddToObjFld('Demand Controlled Ventilation Type', base + pazDCVTypeOff,'')
  CALL AddToObjFld('Outdoor Air Economizer Type', base + pazEconomizerTypeOff,'')
  CALL AddToObjFld('Heat Recovery Type', base + pazHeatRecTypeOff,'')
  CALL AddToObjFld('Sensible Heat Recovery Effectiveness', base + pazHeatRecSenEffOff,'')
  CALL AddToObjFld('Latent Heat Recovery Effectiveness', base + pazHeatRecLatEffOff,'',.true.)
END DO
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE CreateBaseboard
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Michael Witte of GARD Analytics, Inc.
          !    DATE WRITTEN   September 2013
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a Baseboard Heat system.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for baseboard type
INTEGER, PARAMETER :: baseboardNone           = 1
INTEGER, PARAMETER :: baseboardHotWater       = 2
INTEGER, PARAMETER :: baseboardElectric       = 3

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for
INTEGER :: iZone
LOGICAL :: isTStatNameBlank = .FALSE.
LOGICAL :: isDedOutAirNameBlank = .FALSE.
LOGICAL :: isOATypeDetailed = .FALSE.
INTEGER :: baseboardKind = 0

DO iZone = 1, numCompactBaseboard
  base =  compactBaseboardBase(iZone)
  ! set flags
  isTStatNameBlank = (FldVal(base + bbzTStatNameOff) .EQ. '')
  isDedOutAirNameBlank = (FldVal(base + bbzDedOutAirSysNameOff) .EQ. '')
  isOATypeDetailed = SameString(FldVal(base + bbzOutAirTypeOff),'DetailedSpecification')
  !set the baseboard indicator
  IF (SameString(FldVal(base +  bbzBaseboardTypeOff),'HotWater')) THEN
    baseboardKind = baseboardHotWater
  ELSEIF (SameString(FldVal(base +  bbzBaseboardTypeOff),'Electric')) THEN
    baseboardKind = baseboardElectric
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:Baseboard "'//TRIM(FldVal(base + bbzNameOff))//'"'// &
                    ' in the Baseboard Heating Type field: '//TRIM(FldVal(base + bbzBaseboardTypeOff)))
  END IF

  !repeat per zone - only if bbzTStatNameOff is not blank
  IF (.NOT. isTStatNameBlank) THEN
    !ZONE CONTROL:THERMOSTATIC ~ line 5
    CALL AddAlwaysSchedule('4')
    CALL CreateNewObj('ZoneControl:Thermostat')
    CALL AddToObjFld('Name', base + bbzNameOff,' Thermostat')
    CALL AddToObjFld('Zone or ZoneList Name', base + bbzNameOff,'')
    CALL AddToObjStr('Control Type Schedule Name','HVACTemplate-Always 4')
    CALL AddToObjStr('Control Object Type','ThermostatSetpoint:DualSetpoint')
    CALL AddToObjFld('Control Name', base + bbzTStatNameOff,' Dual SP Control',.TRUE.)
  END IF
  !ZONE SIZING ~ line 12
  ! OADCV Changes
  IF ((.NOT. isDedOutAirNameBlank) .AND. (.NOT. isOATypeDetailed)) THEN
    CALL CreateNewObj('DesignSpecification:OutdoorAir')
    CALL AddToObjStr('Name', 'SZ DSOA '//FldVal(base+bbzNameOff))
    CALL AddToObjFld('Outdoor Air Method', base + bbzOutAirTypeOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Person {m3/s}', base + bbzOutAirFlowPersonOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone Floor Area {m3/s-m2}', base + bbzOutAirFlowAreaOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone {m3/s}', base + bbzOutAirFlowZoneOff,'',.true.)
  END IF

  CALL CreateNewObj('Sizing:Zone')
  CALL AddToObjFld('Zone or ZoneList Name', base + bbzNameOff,'')
  CALL AddToObjStr('Zone Cooling Design Supply Air Temperature Input Method','SupplyAirTemperature')
  CALL AddToObjStr('Zone Cooling Design Supply Air Temperature {C}', '12.5')
  CALL AddToObjStr('Zone Cooling Design Supply Air Temperature Difference {delta C}','')
  CALL AddToObjStr('Zone Heating Design Supply Air Temperature Input Method','SupplyAirTemperature')
  CALL AddToObjStr('Zone Heating Design Supply Air Temperature {C}', '50')
  CALL AddToObjStr('Zone Heating Design Supply Air Temperature Difference {delta C}','')
  CALL AddToObjStr('Zone Cooling Design Supply Air Humidity Ratio {kg-H20/kg-air}','0.008')
  CALL AddToObjStr('Zone Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isDedOutAirNameBlank) THEN
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','')
  ELSE
    IF (isOATypeDetailed) THEN
      CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + bbzDesSpecOANameOff,'')
    ELSE
      CALL AddToObjStr('Design Specification Outdoor Air Object Name','SZ DSOA '//FldVal(base+bbzNameOff))
    END IF
  END IF
  CALL AddToObjFld('Zone Heating Sizing Factor', base + bbzZoneHeatSizingFactorOff,'')
  CALL AddToObjStr('Zone Cooling Sizing Factor','')
  CALL AddToObjStr('Cooling Design Air Flow Method','DesignDay')
  CALL AddToObjStr('Cooling Design Air Flow Rate {m3/s}','0')
  CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
  CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}',' ')
  CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  CALL AddToObjStr('Heating Design Air Flow Method','DesignDay')
  CALL AddToObjStr('Heating Design Air Flow Rate {m3/s}','0')
  CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
  CALL AddToObjStr('Heating Maximum Air Flow {m3/s}',' ')
  CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
  IF (isDedOutAirNameBlank) THEN
    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','',.true.)
  ELSE
    IF (isOATypeDetailed) THEN
      CALL AddToObjFld('Design Specification Zone Air Distribution Object Name', base + bbzDesSpecOADistNameOff,'',.true.)
    ELSE
      CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+bbzNameOff),.true.)
      CALL CreateNewObj('DesignSpecification:ZoneAirDistribution')
      CALL AddToObjStr('Name','SZ DSZAD '//FldVal(base+bbzNameOff))
      CALL AddToObjStr('Zone Air Distribution Effectiveness in Cooling Mode {}','1')
      CALL AddToObjStr('Zone Air Distribution Effectiveness in Heating Mode {}','1',.true.)
    END IF
  END IF
  !CONTROLLED ZONE EQUIP CONFIGURATION
  CALL CreateNewObj('ZoneHVAC:EquipmentConnections')
  CALL AddToObjFld('Zone Name', base + bbzNameOff,'             ')
  CALL AddToObjFld('Zone Conditioning Equipment List Name', base + bbzNameOff,' Equipment')
  IF (isDedOutAirNameBlank) THEN
    CALL AddToObjStr('Zone Air Inlet Node or NodeList Name', '')
  ELSE
    CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base + bbzNameOff,' DOAS Supply Inlet')
  END IF
  CALL AddToObjStr('Zone Air Exhaust Node or NodeList Name', '')
  CALL AddToObjFld('Zone Air Node Name', base + bbzNameOff,' Zone Air Node')
  CALL AddToObjFld('Zone Return Air Node Name', base + bbzNameOff,' Return Outlet',.TRUE.)
  !ZONE EQUIPMENT LIST
  CALL CreateNewObj('ZoneHVAC:EquipmentList')
  CALL AddToObjFld('Name', base + bbzNameOff,' Equipment')
  CALL AddToObjStr('Load Distribution Scheme','SequentialLoad')
  IF (isDedOutAirNameBlank) THEN
    SELECT CASE (baseboardKind)
      CASE (baseboardHotWater)
        CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CASE (baseboardElectric)
        CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:Convective:Electric')
    END SELECT
    CALL AddToObjFld('Zone Equipment Name', base + uzNameOff,' Baseboard Heat')
    CALL AddToObjStr('Zone Equipment Cooling Sequence','1')
    CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1',.FALSE.)
    CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
    CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.TRUE.)
ELSE
    CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Zone Equipment Name', base + bbzNameOff,' DOAS ATU')
    CALL AddToObjStr('Zone Equipment Cooling Sequence','1')
    CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','1')
    CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
    CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.FALSE.)
  SELECT CASE (baseboardKind)
      CASE (baseboardHotWater)
        CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CASE (baseboardElectric)
        CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:Convective:Electric')
    END SELECT
    CALL AddToObjFld('Zone Equipment Name', base + uzNameOff,' Baseboard Heat')
    CALL AddToObjStr('Zone Equipment Cooling Sequence','2')
    CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence','2',.FALSE.)
    CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
    CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.TRUE.)
  !***ZoneHVAC:AirDistributionUnit
    CALL CreateNewObj('ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Name', base + bbzNameOff,' DOAS ATU')
    CALL AddToObjFld('Air Distribution Unit Outlet Node Name', base + bbzNameOff,' DOAS Supply Inlet')
    CALL AddToObjStr('Air Terminal Object Type','AirTerminal:SingleDuct:VAV:NoReheat')
    CALL AddToObjFld('Air Terminal Name', base + bbzNameOff,' DOAS Air Terminal',.TRUE.)
    !***AirTerminal:SingleDuct:VAV:NoReheat
    CALL CreateNewObj('AirTerminal:SingleDuct:VAV:NoReheat')
    CALL AddToObjFld('Name', base + bbzNameOff,' DOAS Air Terminal')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Air Outlet Node Name', base + bbzNameOff,' DOAS Supply Inlet')
    CALL AddToObjFld('Air Inlet Node Name', base + bbzNameOff,' Zone Equip Inlet')
    CALL AddToObjStr('Maximum Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Zone Minimum Air Flow Input Method','Constant')
    CALL AddToObjStr('Constant Minimum Air Flow Fraction','1.0')
    CALL AddToObjStr('Fixed Minimum Air Flow Rate {m3/s}','')
    CALL AddToObjStr('Minimum Air Flow Fraction Schedule Name','')
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','',.TRUE.)
  ENDIF
  !Baseboards
  SELECT CASE (baseboardKind)
    CASE (baseboardHotWater)
      !ZoneHVAC:Baseboard:RadiantConvective:Water
      CALL CreateNewObj('ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Name', base + bbzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + bbzBaseboardAvailSchedNameOff,'')
      CALL AddToObjFld('Inlet Node Name', base + bbzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Outlet Node Name', base + bbzNameOff,' Baseboard HW Outlet')
      CALL AddToObjStr('Rated Average Water Temperature {C}','87.78')
      CALL AddToObjStr('Rated Water Mass Flow Rate {m3/s}','0.063')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + bbzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Maximum Water Flow Rate','autosize')
      CALL AddToObjStr('Convergence Tolerance','0.001')
      CALL AddToObjStr('Fraction Radiant','0.0')
      CALL AddToObjStr('Fraction of Radiant Energy Incident on People','0.0',.TRUE.)
      !BRANCH (for baseboard)
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + bbzNameOff,' Baseboard Heat HW Branch')
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Component Name', base + bbzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Component Inlet Node Name', base + bbzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + bbzNameOff,' Baseboard HW Outlet',.TRUE.)
      CALL AddToStrList(TRIM(FldVal(base + bbzNameOff)) // ' Baseboard Heat HW Branch',handleHeatingCoilBranch)
      IF ((compactHotLoopBase .EQ. 0) .OR. (numCompactBoiler .EQ. 0)) THEN
        CALL WriteError('Both a HVACTemplate:Plant:HotWaterLoop and a HVACTemplate:Plant:Boiler are needed ' // &
             'when using hot water baseboards.')
      END IF
    CASE (baseboardElectric)
      !BASEBOARD HEATER:Electric:Convective
      CALL CreateNewObj('ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Name', base + bbzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + bbzBaseboardAvailSchedNameOff,'')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + bbzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Efficiency','1',.TRUE.)
  END SELECT
END DO
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE CreatePTAC
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   September 2007
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a packaged terminal air conditioner.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for heating coil type
INTEGER, PARAMETER :: hcElectric = 1
INTEGER, PARAMETER :: hcWater = 2
INTEGER, PARAMETER :: hcGas = 3
          ! parameters for fan mode type
INTEGER, PARAMETER :: fmCycling = 1
INTEGER, PARAMETER :: fmNeither = 3
          ! parameters for baseboard type
INTEGER, PARAMETER :: baseboardNone           = 1
INTEGER, PARAMETER :: baseboardHotWater       = 2
INTEGER, PARAMETER :: baseboardElectric       = 3
          ! parameters for capacity control
INTEGER, PARAMETER :: capctrlNone             = 1
INTEGER, PARAMETER :: capctrlASHRAE90VarFan   = 2

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for
INTEGER :: iZone
LOGICAL :: isTStatNameBlank
LOGICAL :: isSupplyCoolRateAutosize
LOGICAL :: isSupplyHeatRateAutosize
LOGICAL :: isSysAvailSchedBlank
INTEGER :: heatCoilKind=0
INTEGER :: fanMode=0
LOGICAL :: isDrawThru
LOGICAL :: isHeatCoilCapAutosize
LOGICAL :: isDedOutAirNameBlank
LOGICAL :: isOATypeDetailed = .FALSE.
LOGICAL :: isBaseboardNone = .FALSE.
INTEGER :: baseboardKind = 0
INTEGER :: capacityControlKind = 0
CHARACTER(len=1) :: NextSequenceNumber = '1'

DO iZone = 1, numCompactPTAC
  base =  compactPTACBase(iZone)
  ! set flags
  isTStatNameBlank = (FldVal(base + ptaczTStatNameOff) .EQ. '')
  isSupplyCoolRateAutosize = SameString(FldVal(base + ptaczSupplyCoolFlowRateOff),'autosize')
  isSupplyHeatRateAutosize = SameString(FldVal(base + ptaczSupplyHeatFlowRateOff),'autosize')
  isSysAvailSchedBlank = (FldVal(base + ptaczSysAvailSchedNameOff) .EQ. '')
  isHeatCoilCapAutosize = SameString(FldVal(base + ptaczHeatCoilCapOff),'autosize')
  isOATypeDetailed = SameString(FldVal(base + ptaczOutAirTypeOff),'DetailedSpecification')
  isBaseboardNone = SameString(FldVal(base + ptaczBaseboardTypeOff),'None')
  IF (SameString(FldVal(base + ptaczHeatCoilKindOff),'Electric')) THEN
    heatCoilKind = hcElectric
  ELSEIF (SameString(FldVal(base + ptaczHeatCoilKindOff),'HotWater')) THEN
    heatCoilKind = hcWater
  ELSEIF (SameString(FldVal(base + ptaczHeatCoilKindOff),'Gas')) THEN
    heatCoilKind = hcGas
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:PTAC "'//TRIM(FldVal(base + ptaczNameOff))//'"'// &
                    ' in the Heating Coil Type field: '//TRIM(FldVal(base + ptaczHeatCoilKindOff)))
  END IF

  ! Supply fan mode may be blank (default to Cycling) or a schedule name
  IF (SameString(FldVal(base + ptaczFanModeOff),'')) THEN
    fanMode = fmCycling
  ELSE
    fanMode = fmNeither
  END IF
  isDrawThru = SameString(FldVal(base + ptaczFanPlacementOff),'DrawThrough')
  isDedOutAirNameBlank = (FldVal(base + ptaczDedOutAirSysNameOff) .EQ. '')
  !validate cooling design method
  IF (SameString(FldVal(base +  ptaczCoolDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  ptaczCoolDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:PTAC "'//TRIM(FldVal(base + ptaczNameOff))//'"'// &
                    ' in the Zone Cooling Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + ptaczCoolDesignTempMethodOff)))
  END IF
  !validate heating design method
  IF (SameString(FldVal(base +  ptaczHeatDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  ptaczHeatDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:PTAC "'//TRIM(FldVal(base + ptaczNameOff))//'"'// &
                    ' in the Zone Heating Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + ptaczCoolDesignTempMethodOff)))
  END IF
  !set the baseboard indicator
  IF (SameString(FldVal(base +  ptaczBaseboardTypeOff),'None')) THEN
    baseboardKind = baseboardNone
  ELSEIF (SameString(FldVal(base +  ptaczBaseboardTypeOff),'HotWater')) THEN
    baseboardKind = baseboardHotWater
  ELSEIF (SameString(FldVal(base +  ptaczBaseboardTypeOff),'Electric')) THEN
    baseboardKind = baseboardElectric
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:PTAC "'//TRIM(FldVal(base + ptaczNameOff))//'"'// &
                    ' in the Baseboard Heating Type field: '//TRIM(FldVal(base + ptaczBaseboardTypeOff)))
  END IF
  IF (SameString(FldVal(base +  ptaczCapCtrlTypeOff),'SingleZoneVAV')) THEN
    capacityControlKind = capctrlASHRAE90VarFan
  ELSE IF (SameString(FldVal(base +  ptaczCapCtrlTypeOff),'None')) THEN
    capacityControlKind = capctrlNone
  END IF

  !repeat per zone - only if pazTStatNameOff is not blank
  IF (.NOT. isTStatNameBlank) THEN
    !ZONE CONTROL:THERMOSTATIC ~ line 5
    CALL AddAlwaysSchedule('4')
    CALL CreateNewObj('ZoneControl:Thermostat')
    CALL AddToObjFld('Name', base + ptaczNameOff,' Thermostat')
    CALL AddToObjFld('Zone or ZoneList Name', base + ptaczNameOff,'')
    CALL AddToObjStr('Control Type Schedule Name','HVACTemplate-Always 4')
    CALL AddToObjStr('Control Object Type','ThermostatSetpoint:DualSetpoint')
    CALL AddToObjFld('Control Name', base + ptaczTStatNameOff,' Dual SP Control',.TRUE.)
  END IF
  !ZONE SIZING ~ line 12
! OADCV Changes
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:OutdoorAir')
    CALL AddToObjStr('Name', 'SZ DSOA '//FldVal(base+ptaczNameOff))
    CALL AddToObjFld('Outdoor Air Method', base + ptaczOutAirTypeOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Person {m3/s}', base + ptaczOutAirFlowPersonOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone Floor Area {m3/s-m2}', base + ptaczOutAirFlowAreaOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone {m3/s}', base + ptaczOutAirFlowZoneOff,'',.true.)
  END IF

  CALL CreateNewObj('Sizing:Zone')
  CALL AddToObjFld('Zone or ZoneList Name', base + ptaczNameOff,'')
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Input Method',base + ptaczCoolDesignTempMethodOff,'')
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', base + ptaczCoolDesignTempOff,'')
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Difference {delta C}',base + ptaczCoolDesignTempDiffOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Input Method',base + ptaczHeatDesignTempMethodOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', base + ptaczHeatDesignTempOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Difference {delta C}',base + ptaczHeatDesignTempDiffOff,'')
  CALL AddToObjStr('Zone Cooling Design Supply Air Humidity Ratio {kg-H20/kg-air}','0.008')
  CALL AddToObjStr('Zone Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + ptaczDesSpecOANameOff,'')
  ELSE
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','SZ DSOA '//FldVal(base+ptaczNameOff))
  END IF
  CALL AddToObjFld('Zone Heating Sizing Factor', base + ptaczZoneHeatSizingFactorOff,'')
  CALL AddToObjFld('Zone Cooling Sizing Factor', base + ptaczZoneCoolSizingFactorOff,'')
  IF (isSupplyCoolRateAutosize) THEN
    CALL AddToObjStr('Cooling Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Cooling Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}','')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  ELSE
    CALL AddToObjStr('Cooling Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Cooling Design Air Flow Rate {m3/s}', base + ptaczSupplyCoolFlowRateOff,'')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}','')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  END IF
  IF (isSupplyHeatRateAutosize) THEN
    CALL AddToObjStr('Heating Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Heating Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}','')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
!    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+ptaczNameOff),.true.)
  ELSE
    CALL AddToObjStr('Heating Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Heating Design Air Flow Rate {m3/s}', base + ptaczSupplyHeatFlowRateOff,'')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}','')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
!    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+ptaczNameOff),.true.)
  END IF
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Zone Air Distribution Object Name', base + ptaczDesSpecOADistNameOff,'',.true.)
  ELSE
    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+ptaczNameOff),.true.)
  END IF

  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:ZoneAirDistribution')
    CALL AddToObjStr('Name','SZ DSZAD '//FldVal(base+ptaczNameOff))
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Cooling Mode {}','1')
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Heating Mode {}','1',.true.)
  END IF
  !CONTROLLED ZONE EQUIP CONFIGURATION ~ line 46
  CALL CreateNewObj('ZoneHVAC:EquipmentConnections')
  CALL AddToObjFld('Zone Name', base + ptaczNameOff,'')
  CALL AddToObjFld('Zone Conditioning Equipment List Name', base + ptaczNameOff,' Equipment')
  IF (isDedOutAirNameBlank) THEN
    CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base + ptaczNameOff,' PTAC Supply Inlet')
  ELSE
    CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base + ptaczNameOff,' Inlets')
  END IF
  CALL AddToObjFld('Zone Air Exhaust Node or NodeList Name', base + ptaczNameOff,' PTAC Return')
  CALL AddToObjFld('Zone Air Node Name', base + ptaczNameOff,' Zone Air Node')
  CALL AddToObjFld('Zone Return Air Node Name', base + ptaczNameOff,' Return Outlet',.TRUE.)
  !NODELIST (not in spreadsheet added with DOAS)
  IF (.NOT. isDedOutAirNameBlank) THEN
    CALL CreateNewObj('NodeList')
    CALL AddToObjFld('Name', base + ptaczNameOff,' Inlets')
    CALL AddToObjFld('Node Name', base + ptaczNameOff,' PTAC Supply Inlet')
    CALL AddToObjFld('Node Name', base + ptaczNameOff,' DOAS Supply Inlet',.TRUE.)
  ENDIF
  !ZONE EQUIPMENT LIST ~ line 54
  CALL CreateNewObj('ZoneHVAC:EquipmentList')
  CALL AddToObjFld('Name', base + ptaczNameOff,' Equipment')
  CALL AddToObjStr('Load Distribution Scheme','SequentialLoad')

  NextSequenceNumber = '1'
  IF (.NOT. isDedOutAirNameBlank) THEN
    CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Zone Equipment Name', base + ptaczNameOff,' DOAS ATU')
    CALL AddToObjStr('Zone Equipment Cooling Sequence',trim(NextSequenceNumber))
    CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence',trim(NextSequenceNumber))
    CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
    CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.FALSE.)
    NextSequenceNumber = '2'
  END IF
  CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:PackagedTerminalAirConditioner')
  CALL AddToObjFld('Zone Equipment Name', base + ptaczNameOff,' PTAC')
  CALL AddToObjStr('Zone Equipment Cooling Sequence',trim(NextSequenceNumber))
  CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence',trim(NextSequenceNumber),.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',isBaseboardNone)
IF (.NOT. isBaseboardNone) THEN
    IF (isDedOutAirNameBlank) THEN
      NextSequenceNumber = '2'
    ELSE
      NextSequenceNumber = '3'
    END IF
    SELECT CASE (baseboardKind)
      CASE (baseboardHotWater)
        CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CASE (baseboardElectric)
        CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:Convective:Electric')
    END SELECT
        CALL AddToObjFld('Zone Equipment Name', base + ptaczNameOff,' Baseboard Heat')
        CALL AddToObjStr('Zone Equipment Cooling Sequence',trim(NextSequenceNumber))
        CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence',trim(NextSequenceNumber),.FALSE.)
        CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
        CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.TRUE.)
      END IF

  IF (.NOT. isDedOutAirNameBlank) THEN
    !***ZoneHVAC:AirDistributionUnit
    CALL CreateNewObj('ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Name', base + ptaczNameOff,' DOAS ATU')
    CALL AddToObjFld('Air Distribution Unit Outlet Node Name', base + ptaczNameOff,' DOAS Supply Inlet')
    CALL AddToObjStr('Air Terminal Object Type','AirTerminal:SingleDuct:VAV:NoReheat')
    CALL AddToObjFld('Air Terminal Name', base + ptaczNameOff,' DOAS Air Terminal',.TRUE.)
    !***AirTerminal:SingleDuct:VAV:NoReheat
    CALL CreateNewObj('AirTerminal:SingleDuct:VAV:NoReheat')
    CALL AddToObjFld('Name', base + ptaczNameOff,' DOAS Air Terminal')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Air Outlet Node Name', base + ptaczNameOff,' DOAS Supply Inlet')
    CALL AddToObjFld('Air Inlet Node Name', base + ptaczNameOff,' Zone Equip Inlet')
    CALL AddToObjStr('Maximum Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Zone Minimum Air Flow Input Method','Constant')
    CALL AddToObjStr('Constant Minimum Air Flow Fraction','1.0')
    CALL AddToObjStr('Fixed Minimum Air Flow Rate {m3/s}','')
    CALL AddToObjStr('Minimum Air Flow Fraction Schedule Name','')
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','',.TRUE.)
  ENDIF
  !ZoneHVAC:PackagedTerminalAirConditioner ~ line 61
  IF (.NOT. isDedOutAirNameBlank) THEN
    CALL WriteComment('Set PTAC outdoor air to zero, because zone is served by dedicated outdoor air system (DOAS)')
  END IF
  CALL CreateNewObj('ZoneHVAC:PackagedTerminalAirConditioner')
  CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC')
  CALL AddToObjFld('Availability Schedule Name', base + ptaczSysAvailSchedNameOff,'')
  CALL AddToObjFld('Air Inlet Node Name', base + ptaczNameOff,' PTAC Return')
  CALL AddToObjFld('Air Outlet Node Name', base + ptaczNameOff,' PTAC Supply Inlet')
  CALL AddToObjStr('Outdoor Air Mixer Object Type','OutdoorAir:Mixer')
  CALL AddToObjFld('Outdoor Air Mixer Name', base + ptaczNameOff,' PTAC OA Mixing Box')
  !CR8001
  !CALL AddToObjFld('Supply air volumetric flow rate during cooling operation {m3/s}',   &
  !   base + ptaczSupplyCoolFlowRateOff,'')
  CALL AddToObjStr('Cooling Supply Air Flow Rate {m3/s}','autosize')
  !CR8001
  !CALL AddToObjFld('Supply air volumetric flow rate during heating operation {m3/s}',   &
  !   base + ptaczSupplyHeatFlowRateOff,'')
  CALL AddToObjStr('Heating Supply Air Flow Rate {m3/s}','autosize')
  IF (capacityControlKind == capctrlASHRAE90VarFan) THEN
    CALL AddToObjStr('No Load Supply Air Flow Rate {m3/s}','autosize')
  ELSE
    CALL AddToObjFld('No Load Supply Air Flow Rate {m3/s}',   &
     base + ptaczSupplyNoLoadFlowRateOff,'')
  END IF
  IF (isDedOutAirNameBlank) THEN
    CALL AddToObjStr('Cooling Outdoor Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Heating Outdoor Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('No Load Outdoor Air Flow Rate {m3/s}','autosize')
  ELSE
    CALL AddToObjStr('Cooling Outdoor Air Flow Rate {m3/s}','0.0')
    CALL AddToObjStr('Heating Outdoor Air Flow Rate {m3/s}','0.0')
    CALL AddToObjStr('No Load Outdoor Air Flow Rate {m3/s}','0.0')
  END IF
  CALL AddToObjStr('Supply Air Fan Object Type','Fan:OnOff')
  CALL AddToObjFld('Supply Air Fan Name', base + ptaczNameOff,' PTAC Supply Fan')
  SELECT CASE (heatCoilKind)
    CASE (hcElectric)
      CALL AddToObjStr('Heating Coil Object Type','Coil:Heating:Electric')
    CASE (hcWater)
      CALL AddToObjStr('Heating Coil Object Type','Coil:Heating:Water')
    CASE (hcGas)
      CALL AddToObjStr('Heating Coil Object Type','Coil:Heating:Fuel')
  END SELECT
  CALL AddToObjFld('Heating Coil Name', base + ptaczNameOff,' PTAC Heating Coil')
  CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:DX:SingleSpeed')
  CALL AddToObjFld('Cooling Coil Name', base + ptaczNameOff,' PTAC Cooling Coil')
  CALL AddToObjFld('Fan Placement', base + ptaczFanPlacementOff,'')
  IF (capacityControlKind == capctrlASHRAE90VarFan) THEN
    CALL AddToObjStr('Supply Air Fan Operating Mode Schedule Name','HVACTemplate-Always 1')
    CALL AddToObjStr('Availability Manager List Name','')
    CALL AddToObjStr('Design Specification ZoneHVAC Sizing Object Name','')
    CALL AddToObjStr('Capacity Control Method','SingleZoneVAV')
    CALL AddToObjStr('Minimum Supply Air Temperature in Cooling Mode {C}','autosize')
    CALL AddToObjStr('Maximum Supply Air Temperature in Heating Mode {C}','autosize',.true.)
  ELSE
    SELECT CASE (fanMode)
      CASE (fmCycling)
        CALL AddToObjStr('Supply Air Fan Operating Mode Schedule Name','HVACTemplate-Always 0',.TRUE.)
      CASE (fmNeither)
        CALL AddToObjFld('Supply Air Fan Operating Mode Schedule Name', base + ptaczFanModeOff,'',.TRUE.)
    END SELECT
    IF (fanMode .EQ. fmCycling) THEN
      CALL AddAlwaysSchedule('0')
    END IF
  END IF
  !FAN:SIMPLE:CONSTVOLUME ~ line 87
  CALL CreateNewObj('Fan:OnOff')
  CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC Supply Fan')
  IF (.NOT. isSysAvailSchedBlank) THEN
    CALL AddToObjFld('Availability Schedule Name', base + ptaczSysAvailSchedNameOff,'')
  ELSE
    CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
  END IF
  CALL AddToObjFld('Fan Efficiency', base + ptaczFanTotEfficiencyOff,'')
  CALL AddToObjFld('Pressure Rise {Pa}', base + ptaczFanPressureOff,'')
  CALL AddToObjStr('Maximum Flow Rate {m3/s}','autosize')
  CALL AddToObjFld('Motor Efficiency', base + ptaczFanMotorEfficiencyOff,'')
  CALL AddToObjStr('Motor in Airstream Fraction','1')
  IF (isDrawThru) THEN
    CALL AddToObjFld('Air Inlet Node Name', base + ptaczNameOff,' PTAC Heating Coil Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + ptaczNameOff,' PTAC Supply Inlet',.TRUE.)
  ELSE
    CALL AddToObjFld('Air Inlet Node Name', base + ptaczNameOff,' PTAC Mixed Air Outlet')
    CALL AddToObjFld('Air Outlet Node Name', base + ptaczNameOff,' PTAC Supply Fan Outlet',.TRUE.)
  END IF
  IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  !OUTSIDE AIR MIXER ~ line 100
  CALL CreateNewObj('OutdoorAir:Mixer')
  CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC OA Mixing Box')
  CALL AddToObjFld('Mixed Air Node Name', base + ptaczNameOff,' PTAC Mixed Air Outlet')
  CALL AddToObjFld('Outside Air Stream Node Name', base + ptaczNameOff,' PTAC Outside Air Inlet')
  CALL AddToObjFld('Relief Air Stream Node Name', base + ptaczNameOff,' PTAC Relief Air Outlet')
  CALL AddToObjFld('Return Air Stream Node Name', base + ptaczNameOff,' PTAC Return',.TRUE.)
  !OUTSIDE AIR NODE ~ line 107
  CALL CreateNewObj('OutdoorAir:Node')
  CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC Outside Air Inlet')
  CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
  !COIL:DX:CoolingBypassFactorEmpirical ~ line 111
  CALL CreateNewObj('Coil:Cooling:DX:SingleSpeed')
  CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC Cooling Coil')
  CALL AddToObjFld('Availability Schedule Name', base + ptaczCoolAvailSchedNameOff,'')
  CALL AddToObjFld('Gross Rated Total Cooling Capacity {W}', base + ptaczCoolCoilCapOff,'')
  CALL AddToObjFld('Gross Rated Sensible Heat Ratio', base + ptaczCoolCoilSHROff,'')
  CALL AddToObjFld('Rated COP', base + ptaczCoolCoilCOPOff,'')
  !CR8001
  !CALL AddToObjFld('Rated Air Volume Flow Rate {m3/s}', base + ptaczSupplyCoolFlowRateOff,'')
  CALL AddToObjStr('Rated Air Volume Flow Rate {m3/s}', 'autosize')
  CALL AddToObjStr('Rated Evaporator Fan Power per Volume Flow Rate {W/(m3/s)}','')
  IF (isDrawThru) THEN
    CALL AddToObjFld('Air Inlet Node Name', base + ptaczNameOff,' PTAC Mixed Air Outlet')
  ELSE
    CALL AddToObjFld('Air Inlet Node Name', base + ptaczNameOff,' PTAC Supply Fan Outlet')
  END IF
  CALL AddToObjFld('Air Outlet Node Name', base + ptaczNameOff,' PTAC Cooling Coil Outlet')
  CALL AddToObjFld('Total Cooling Capacity Modifier Curve (function of temperature)',   &
     base + ptaczNameOff,' PTAC Cooling Coil Cap-FT')
  CALL AddToObjFld('Total Cooling Capacity Modifier Curve (function of flow fraction)',   &
     base + ptaczNameOff,' PTAC Cooling Coil Cap-FF')
  CALL AddToObjFld('Energy Input Ratio Modifier Curve (function of temperature)',   &
     base + ptaczNameOff,' PTAC Cooling Coil EIR-FT')
  CALL AddToObjFld('Energy Input Ratio Modifier Curve (function of flow fraction)',   &
     base + ptaczNameOff,' PTAC Cooling Coil EIR-FF')
  CALL AddToObjFld('Part Load Fraction Correlation (function of part load ratio)',   &
     base + ptaczNameOff,' PTAC Cooling Coil Coil PLF')
  CALL AddToObjStr('Minimum Outdoor Dry-Bulb Temperature for Compressor Operation {C}','')
  CALL AddToObjStr('Nominal Time for Condensate Removal to Begin','0')
  CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity','0')
  CALL AddToObjStr('Maximum Cycling Rate','0')
  CALL AddToObjStr('Latent Capacity Time Constant','0')
  CALL AddToObjFld('Condenser Air Inlet Node Name', base + ptaczNameOff,' PTAC Cooling Coil Condenser Inlet')
  CALL AddToObjStr('Condenser Type','AirCooled')
  CALL AddToObjStr('Evaporative Condenser Effectiveness','0')
  CALL AddToObjStr('Evaporative Condenser Air Flow Rate','')
  CALL AddToObjStr('Evaporative Condenser Pump Rated Power Consumption','0')
  CALL AddToObjStr('Crankcase Heater Capacity','0')
  CALL AddToObjStr('Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater Operation','10',.TRUE.)
  !CURVE:BIQUADRATIC ~ line 141
  CALL CreateNewObj('Curve:Biquadratic')
  CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC Cooling Coil Cap-FT')
  CALL AddToObjStr('Coefficient1 Constant','0.942587793')
  CALL AddToObjStr('Coefficient2 x','0.009543347')
  CALL AddToObjStr('Coefficient3 x**2','0.00068377')
  CALL AddToObjStr('Coefficient4 y','-0.011042676')
  CALL AddToObjStr('Coefficient5 y**2','0.000005249')
  CALL AddToObjStr('Coefficient6 x*y','-0.00000972')
  CALL AddToObjStr('Minimum Value of x','12.77778')
  CALL AddToObjStr('Maximum Value of x','23.88889')
  CALL AddToObjStr('Minimum Value of y','18.0')
  CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
  !CURVE:QUADRATIC ~ line 155
  CALL CreateNewObj('Curve:Quadratic')
  CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC Cooling Coil Cap-FF')
  CALL AddToObjStr('Coefficient1 Constant','0.8')
  CALL AddToObjStr('Coefficient2 x','0.2')
  CALL AddToObjStr('Coefficient3 x**2','0')
  CALL AddToObjStr('Minimum Value of x','0.5')
  CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
  !CURVE:BIQUADRATIC ~ line 164
  CALL CreateNewObj('Curve:Biquadratic')
  CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC Cooling Coil EIR-FT')
  CALL AddToObjStr('Coefficient1 Constant','0.342414409')
  CALL AddToObjStr('Coefficient2 x','0.034885008')
  CALL AddToObjStr('Coefficient3 x**2','-0.0006237')
  CALL AddToObjStr('Coefficient4 y','0.004977216')
  CALL AddToObjStr('Coefficient5 y**2','0.000437951')
  CALL AddToObjStr('Coefficient6 x*y','-0.000728028')
  CALL AddToObjStr('Minimum Value of x','12.77778')
  CALL AddToObjStr('Maximum Value of x','23.88889')
  CALL AddToObjStr('Minimum Value of y','18.0')
  CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
  !CURVE:QUADRATIC ~ line 178
  CALL CreateNewObj('Curve:Quadratic')
  CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC Cooling Coil EIR-FF')
  CALL AddToObjStr('Coefficient1 Constant','1.1552')
  CALL AddToObjStr('Coefficient2 x','-0.1808')
  CALL AddToObjStr('Coefficient3 x**2','0.0256')
  CALL AddToObjStr('Minimum Value of x','0.5')
  CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
  !CURVE:QUADRATIC ~ line 187
  CALL CreateNewObj('Curve:Quadratic')
  CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC Cooling Coil Coil PLF')
  CALL AddToObjStr('Coefficient1 Constant','0.85')
  CALL AddToObjStr('Coefficient2 x','0.15')
  CALL AddToObjStr('Coefficient3 x**2','0')
  CALL AddToObjStr('Minimum Value of x','0')
  CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  !CURVE:QUADRATIC ~ line 196
  CALL CreateNewObj('OutdoorAir:Node')
  CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC Cooling Coil Condenser Inlet')
  CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
  SELECT CASE (heatCoilKind)
    CASE (hcElectric)
      !COIL:Electric:Heating ~ line 201
      CALL CreateNewObj('Coil:Heating:Electric')
      CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC Heating Coil')
      CALL AddToObjFld('Availability Schedule Name', base + ptaczHeatAvailSchedNameOff,'')
      CALL AddToObjStr('Efficiency','1')
      CALL AddToObjFld('Nominal Capacity of the Coil {W}', base + ptaczHeatCoilCapOff,'')
      CALL AddToObjFld('Air Inlet Node Name', base + ptaczNameOff,' PTAC Cooling Coil Outlet')
      IF (isDrawThru) THEN
        CALL AddToObjFld('Air Outlet Node Name', base + ptaczNameOff,' PTAC Heating Coil Outlet')
      ELSE
        CALL AddToObjFld('Air Outlet Node Name', base + ptaczNameOff,' PTAC Supply Inlet')
      END IF
      CALL AddToObjStr('Coil Temp Setpoint Node','',.TRUE.)
    CASE (hcGas)
      !COIL:Gas:Heating ~ line 213
      CALL CreateNewObj('Coil:Heating:Fuel')
      CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC Heating Coil')
      CALL AddToObjFld('Availability Schedule Name', base + ptaczHeatAvailSchedNameOff,'')
      CALL AddToObjStr('Fuel Type', 'Gas')
      CALL AddToObjFld('Gas Burner Efficiency', base + ptaczHeatEfficiencyOff,'')
      CALL AddToObjFld('Nominal Capacity of the Coil {W}', base + ptaczHeatCoilCapOff,'')
      CALL AddToObjFld('Air Inlet Node Name', base + ptaczNameOff,' PTAC Cooling Coil Outlet')
      IF (isDrawThru) THEN
        CALL AddToObjFld('Air Outlet Node Name', base + ptaczNameOff,' PTAC Heating Coil Outlet')
      ELSE
        CALL AddToObjFld('Air Outlet Node Name', base + ptaczNameOff,' PTAC Supply Inlet')
      END IF
      CALL AddToObjStr('Coil Temp Setpoint Node','')
      CALL AddToObjFld('Parasitic Electric Load {W}', base + ptaczHeatParasiticOff,'')
      CALL AddToObjFld('Part load fraction correlation (function of part load ratio)',   &
         base + ptaczNameOff,' PTAC Heating Coil PLF-FPLR')
      CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
      !CURVE:CUBIC ~ line 227
      CALL CreateNewObj('Curve:Cubic')
      CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC Heating Coil PLF-FPLR')
      CALL AddToObjStr('Coefficient1 Constant','0.8')
      CALL AddToObjStr('Coefficient2 x','0.2')
      CALL AddToObjStr('Coefficient3 x**2','0')
      CALL AddToObjStr('Coefficient4 x**3','0')
      CALL AddToObjStr('Minimum Value of x','0')
      CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
    CASE (hcWater)
      !CURVE:CUBIC ~ line 235
      CALL CreateNewObj('Coil:Heating:Water')
      CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC Heating Coil')
      CALL AddToObjFld('Availability Schedule Name', base + ptaczHeatAvailSchedNameOff,'')
      CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
      CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Water Inlet Node Name', base + ptaczNameOff,' PTAC Heating Coil HW Inlet')
      CALL AddToObjFld('Water Outlet Node Name', base + ptaczNameOff,' PTAC Heating Coil HW Outlet')
      CALL AddToObjFld('Air Inlet Node Name', base + ptaczNameOff,' PTAC Cooling Coil Outlet')
      IF (isDrawThru) THEN
        CALL AddToObjFld('Air Outlet Node Name', base + ptaczNameOff,' PTAC Heating Coil Outlet')
      ELSE
        CALL AddToObjFld('Air Outlet Node Name', base + ptaczNameOff,' PTAC Supply Inlet')
      END IF
      IF (isHeatCoilCapAutosize) THEN
        CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
      ELSE
        CALL AddToObjStr('Performance Input Method','NominalCapacity')
      END IF
      CALL AddToObjFld('Rated Capacity', base + ptaczHeatCoilCapOff,'')
      CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
      CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
      CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
      CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
      CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
      !CURVE:CUBIC ~ line 254
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + ptaczNameOff,' PTAC Heating Coil HW Branch')
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
      CALL AddToObjFld('Component Name', base + ptaczNameOff,' PTAC Heating Coil')
      CALL AddToObjFld('Component Inlet Node Name', base + ptaczNameOff,' PTAC Heating Coil HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + ptaczNameOff,' PTAC Heating Coil HW Outlet',.TRUE.)
      CALL AddToStrList(TRIM(FldVal(base + ptaczNameOff)) // ' PTAC Heating Coil HW Branch',handleHeatingCoilBranch)
  END SELECT
  !Baseboards
  SELECT CASE (baseboardKind)
    CASE (baseboardHotWater)
      !ZoneHVAC:Baseboard:RadiantConvective:Water
      CALL CreateNewObj('ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Name', base + ptaczNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + ptaczBaseboardAvailSchedNameOff,'')
      CALL AddToObjFld('Inlet Node Name', base + ptaczNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Outlet Node Name', base + ptaczNameOff,' Baseboard HW Outlet')
      CALL AddToObjStr('Rated Average Water Temperature {C}','87.78')
      CALL AddToObjStr('Rated Water Mass Flow Rate {m3/s}','0.063')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + ptaczBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Maximum Water Flow Rate','autosize')
      CALL AddToObjStr('Convergence Tolerance','0.001')
      CALL AddToObjStr('Fraction Radiant','0.0')
      CALL AddToObjStr('Fraction of Radiant Energy Incident on People','0.0',.TRUE.)
      !BRANCH (for baseboard)
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + ptaczNameOff,' Baseboard Heat HW Branch')
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Component Name', base + ptaczNameOff,' Baseboard Heat')
      CALL AddToObjFld('Component Inlet Node Name', base + ptaczNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + ptaczNameOff,' Baseboard HW Outlet',.TRUE.)
      CALL AddToStrList(TRIM(FldVal(base + ptaczNameOff)) // ' Baseboard Heat HW Branch',handleHeatingCoilBranch)
      IF ((compactHotLoopBase .EQ. 0) .OR. (numCompactBoiler .EQ. 0)) THEN
        CALL WriteError('Both a HVACTemplate:Plant:HotWaterLoop and a HVACTemplate:Plant:Boiler are needed ' // &
             'when using hot water baseboards.')
      END IF
    CASE (baseboardElectric)
      !BASEBOARD HEATER:Electric:Convective
      CALL CreateNewObj('ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Name', base + ptaczNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + ptaczBaseboardAvailSchedNameOff,'')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + ptaczBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Efficiency','1',.TRUE.)
  END SELECT
END DO
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE CreatePTHP
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   September 2007
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a packaged terminal heat pump.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for supplementary heating coil type
INTEGER, PARAMETER :: shcElectric = 1
INTEGER, PARAMETER :: shcGas = 2
INTEGER, PARAMETER :: shcWater = 3
          ! parameters for fan mode type
INTEGER, PARAMETER :: fmCycling = 1
INTEGER, PARAMETER :: fmNeither = 3
          ! parameters for baseboard type
INTEGER, PARAMETER :: baseboardNone           = 1
INTEGER, PARAMETER :: baseboardHotWater       = 2
INTEGER, PARAMETER :: baseboardElectric       = 3
          ! parameters for capacity control
INTEGER, PARAMETER :: capctrlNone             = 1
INTEGER, PARAMETER :: capctrlASHRAE90VarFan   = 2

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for
INTEGER :: iZone
LOGICAL :: isTStatNameBlank
LOGICAL :: isSupplyCoolRateAutosize
LOGICAL :: isSupplyHeatRateAutosize
LOGICAL :: isSysAvailSchedBlank
INTEGER :: suppHeatCoilKind=0
INTEGER :: fanMode=0
LOGICAL :: isDrawThru
LOGICAL :: isSuppHeatCoilCapAutosize
LOGICAL :: isDedOutAirNameBlank
CHARACTER(len=1) :: SchType
LOGICAL :: isOATypeDetailed = .FALSE.
LOGICAL :: isBaseboardNone = .FALSE.
INTEGER :: baseboardKind = 0
CHARACTER(len=1) :: NextSequenceNumber = '1'
INTEGER :: capacityControlKind = 0

DO iZone = 1, numCompactPTHP
  base =  compactPTHPBase(iZone)
  ! set flags
  isTStatNameBlank = (FldVal(base + pthpzTStatNameOff) .EQ. '')
  isSupplyCoolRateAutosize = SameString(FldVal(base + pthpzSupplyCoolFlowRateOff),'autosize')
  isSupplyHeatRateAutosize = SameString(FldVal(base + pthpzSupplyHeatFlowRateOff),'autosize')
  isSuppHeatCoilCapAutosize = SameString(FldVal(base + pthpzSuppHeatCoilCapOff),'autosize')
  isSysAvailSchedBlank = (FldVal(base + pthpzSysAvailSchedNameOff) .EQ. '')
  isOATypeDetailed = SameString(FldVal(base + pthpzOutAirTypeOff),'DetailedSpecification')
  isBaseboardNone = SameString(FldVal(base + pthpzBaseboardTypeOff),'None')
  IF (SameString(FldVal(base + pthpzSuppHeatCoilKindOff),'Electric')) THEN
    suppHeatCoilKind = shcElectric
  ELSEIF (SameString(FldVal(base + pthpzSuppHeatCoilKindOff),'Gas')) THEN
    suppHeatCoilKind = shcGas
  ELSEIF (SameString(FldVal(base + pthpzSuppHeatCoilKindOff),'HotWater')) THEN
    suppHeatCoilKind = shcWater
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:PTHP "'//TRIM(FldVal(base + pthpzNameOff))//'"'// &
                    ' in the Supplemental Heating Coil Type field: '//TRIM(FldVal(base + pthpzSuppHeatCoilKindOff)))
  END IF
  ! Supply fan mode may be blank (default to Cycling) or a schedule name
  IF (SameString(FldVal(base + pthpzFanModeOff),'')) THEN
    fanMode = fmCycling
  ELSE
    fanMode = fmNeither
  END IF
  isDrawThru = SameString(FldVal(base + pthpzFanPlacementOff),'DrawThrough')
  isDedOutAirNameBlank = (FldVal(base + pthpzDedOutAirSysNameOff) .EQ. '')
  !validate cooling design method
  IF (SameString(FldVal(base +  pthpzCoolDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  pthpzCoolDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:PTHP "'//TRIM(FldVal(base + pthpzNameOff))//'"'// &
                    ' in the Zone Cooling Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + pthpzCoolDesignTempMethodOff)))
  END IF
  !validate heating design method
  IF (SameString(FldVal(base +  pthpzHeatDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  pthpzHeatDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:PTHP "'//TRIM(FldVal(base + pthpzNameOff))//'"'// &
                    ' in the Zone Heating Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + pthpzCoolDesignTempMethodOff)))
  END IF
  !set the baseboard indicator
  IF (SameString(FldVal(base +  pthpzBaseboardTypeOff),'None')) THEN
    baseboardKind = baseboardNone
  ELSEIF (SameString(FldVal(base +  pthpzBaseboardTypeOff),'HotWater')) THEN
    baseboardKind = baseboardHotWater
  ELSEIF (SameString(FldVal(base +  pthpzBaseboardTypeOff),'Electric')) THEN
    baseboardKind = baseboardElectric
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:PTHP "'//TRIM(FldVal(base + pthpzNameOff))//'"'// &
                    ' in the Baseboard Heating Type field: '//TRIM(FldVal(base + pthpzBaseboardTypeOff)))
  END IF
  IF (SameString(FldVal(base +  pthpzCapCtrlTypeOff),'SingleZoneVAV')) THEN
    capacityControlKind = capctrlASHRAE90VarFan
  ELSE IF (SameString(FldVal(base +  pthpzCapCtrlTypeOff),'None')) THEN
    capacityControlKind = capctrlNone
  END IF

  !repeat per zone - only if pthpzTStatNameOff is not blank
  IF (.NOT. isTStatNameBlank) THEN
    !ZONE CONTROL:THERMOSTATIC ~ line 5
    CALL AddAlwaysSchedule('4')
    CALL CreateNewObj('ZoneControl:Thermostat')
    CALL AddToObjFld('Name', base + pthpzNameOff,' Thermostat')
    CALL AddToObjFld('Zone or ZoneList Name', base + pthpzNameOff,'')
    CALL AddToObjStr('Control Type Schedule Name','HVACTemplate-Always 4')
    CALL AddToObjStr('Control Object Type','ThermostatSetpoint:DualSetpoint')
    CALL AddToObjFld('Control Name', base + pthpzTStatNameOff,' Dual SP Control',.TRUE.)
  END IF
  !ZONE SIZING ~ line 12
! OADCV Changes
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:OutdoorAir')
    CALL AddToObjStr('Name', 'SZ DSOA '//FldVal(base+pthpzNameOff))
    CALL AddToObjFld('Outdoor Air Method', base + pthpzOutAirTypeOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Person {m3/s}', base + pthpzOutAirFlowPersonOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone Floor Area {m3/s-m2}', base + pthpzOutAirFlowAreaOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone {m3/s}', base + pthpzOutAirFlowZoneOff,'',.true.)
  END IF

  CALL CreateNewObj('Sizing:Zone')
  CALL AddToObjFld('Zone or ZoneList Name', base + pthpzNameOff,'')
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Input Method',base + pthpzCoolDesignTempMethodOff,'')
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', base + pthpzCoolDesignTempOff,'')
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Difference {delta C}',base + pthpzCoolDesignTempDiffOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Input Method',base + pthpzHeatDesignTempMethodOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', base + pthpzHeatDesignTempOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Difference {delta C}',base + pthpzHeatDesignTempDiffOff,'')
  CALL AddToObjStr('Zone Cooling Design Supply Air Humidity Ratio {kg-H20/kg-air}','0.008')
  CALL AddToObjStr('Zone Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + pthpzDesSpecOANameOff,'')
  ELSE
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','SZ DSOA '//FldVal(base+pthpzNameOff))
  END IF
  CALL AddToObjFld('Zone Heating Sizing Factor', base + pthpzZoneHeatSizingFactorOff,'')
  CALL AddToObjFld('Zone Cooling Sizing Factor', base + pthpzZoneCoolSizingFactorOff,'')
  IF (isSupplyCoolRateAutosize) THEN
    CALL AddToObjStr('Cooling Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Cooling Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}','')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  ELSE
    CALL AddToObjStr('Cooling Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Cooling Design Air Flow Rate {m3/s}', base + pthpzSupplyCoolFlowRateOff,'')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}','')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  END IF
  IF (isSupplyHeatRateAutosize) THEN
    CALL AddToObjStr('Heating Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Heating Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}','')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
!    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+pthpzNameOff),.true.)
  ELSE
    CALL AddToObjStr('Heating Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Heating Design Air Flow Rate {m3/s}', base + pthpzSupplyHeatFlowRateOff,'')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}','')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
!    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+pthpzNameOff),.true.)
  END IF
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Zone Air Distribution Object Name', base + pthpzDesSpecOADistNameOff,'',.true.)
  ELSE
    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+pthpzNameOff),.true.)
  END IF

  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:ZoneAirDistribution')
    CALL AddToObjStr('Name','SZ DSZAD '//FldVal(base+pthpzNameOff))
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Cooling Mode {}','1')
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Heating Mode {}','1',.true.)
  END IF
  !CONTROLLED ZONE EQUIP CONFIGURATION ~ line 46
  CALL CreateNewObj('ZoneHVAC:EquipmentConnections')
  CALL AddToObjFld('Zone Name', base + pthpzNameOff,' ')
  CALL AddToObjFld('Zone Conditioning Equipment List Name', base + pthpzNameOff,' Equipment')
  IF (isDedOutAirNameBlank) THEN
    CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base + pthpzNameOff,' PTHP Supply Inlet')
  ELSE
    CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base + pthpzNameOff,' Inlets')
  END IF
  CALL AddToObjFld('Zone Air Exhaust Node or NodeList Name', base + pthpzNameOff,' PTHP Return')
  CALL AddToObjFld('Zone Air Node Name', base + pthpzNameOff,' Zone Air Node')
  CALL AddToObjFld('Zone Return Air Node Name', base + pthpzNameOff,' Return Outlet',.TRUE.)
  !NODELIST (not in spreadsheet added with DOAS)
  IF (.NOT. isDedOutAirNameBlank) THEN
    CALL CreateNewObj('NodeList')
    CALL AddToObjFld('Name', base + pthpzNameOff,' Inlets')
    CALL AddToObjFld('Node Name', base + pthpzNameOff,' PTHP Supply Inlet')
    CALL AddToObjFld('Node Name', base + pthpzNameOff,' DOAS Supply Inlet',.TRUE.)
  ENDIF
  !ZONE EQUIPMENT LIST ~ line 54
  CALL CreateNewObj('ZoneHVAC:EquipmentList')
  CALL AddToObjFld('Name', base + pthpzNameOff,' Equipment')
  CALL AddToObjStr('Load Distribution Scheme','SequentialLoad')

  NextSequenceNumber = '1'
  IF (.NOT. isDedOutAirNameBlank) THEN
    CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Zone Equipment Name', base + pthpzNameOff,' DOAS ATU')
    CALL AddToObjStr('Zone Equipment Cooling Sequence',trim(NextSequenceNumber))
    CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence',trim(NextSequenceNumber))
    CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
    CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.FALSE.)
    NextSequenceNumber = '2'
  END IF
  CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:PackagedTerminalHeatPump')
  CALL AddToObjFld('Zone Equipment Name', base + pthpzNameOff,' PTHP')
  CALL AddToObjStr('Zone Equipment Cooling Sequence',trim(NextSequenceNumber))
  CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence',trim(NextSequenceNumber),.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',isBaseboardNone)
  IF (.NOT. isBaseboardNone) THEN
    IF (isDedOutAirNameBlank) THEN
      NextSequenceNumber = '2'
    ELSE
      NextSequenceNumber = '3'
    END IF
    SELECT CASE (baseboardKind)
      CASE (baseboardHotWater)
        CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CASE (baseboardElectric)
        CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:Convective:Electric')
    END SELECT
        CALL AddToObjFld('Zone Equipment Name', base + pthpzNameOff,' Baseboard Heat')
        CALL AddToObjStr('Zone Equipment Cooling Sequence',trim(NextSequenceNumber))
        CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence',trim(NextSequenceNumber),.FALSE.)
        CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
        CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.TRUE.)
  END IF

  IF (.NOT. isDedOutAirNameBlank) THEN
    !***ZoneHVAC:AirDistributionUnit
    CALL CreateNewObj('ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Name', base + pthpzNameOff,' DOAS ATU')
    CALL AddToObjFld('Air Distribution Unit Outlet Node Name', base + pthpzNameOff,' DOAS Supply Inlet')
    CALL AddToObjStr('Air Terminal Object Type','AirTerminal:SingleDuct:VAV:NoReheat')
    CALL AddToObjFld('Air Terminal Name', base + pthpzNameOff,' DOAS Air Terminal',.TRUE.)
    !***AirTerminal:SingleDuct:VAV:NoReheat
    CALL CreateNewObj('AirTerminal:SingleDuct:VAV:NoReheat')
    CALL AddToObjFld('Name', base + pthpzNameOff,' DOAS Air Terminal')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Air Outlet Node Name', base + pthpzNameOff,' DOAS Supply Inlet')
    CALL AddToObjFld('Air Inlet Node Name', base + pthpzNameOff,' Zone Equip Inlet')
    CALL AddToObjStr('Maximum Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Zone Minimum Air Flow Input Method','Constant')
    CALL AddToObjStr('Constant Minimum Air Flow Fraction','1.0')
    CALL AddToObjStr('Fixed Minimum Air Flow Rate {m3/s}','')
    CALL AddToObjStr('Minimum Air Flow Fraction Schedule Name','')
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','',.TRUE.)
  ENDIF
  !ZoneHVAC:PackagedTerminalHeatPump ~ line 61
  IF (isDedOutAirNameBlank) THEN
    CALL WriteComment('Set PTHP outdoor air to zero, because zone is served by dedicated outdoor air system (DOAS)')
  END IF
  CALL CreateNewObj('ZoneHVAC:PackagedTerminalHeatPump')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP')
  CALL AddToObjFld('Availability Schedule Name', base + pthpzSysAvailSchedNameOff,'')
  CALL AddToObjFld('Air Inlet Node Name', base + pthpzNameOff,' PTHP Return')
  CALL AddToObjFld('Air Outlet Node Name', base + pthpzNameOff,' PTHP Supply Inlet')
  CALL AddToObjStr('Outdoor Air Mixer Object Type','OutdoorAir:Mixer')
  CALL AddToObjFld('Outdoor Air Mixer Name', base + pthpzNameOff,' PTHP OA Mixing Box')
  !CR8001
  !CALL AddToObjFld('Supply air volumetric flow rate during cooling operation {m3/s}',   &
  !   base + pthpzSupplyCoolFlowRateOff,'')
  CALL AddToObjStr('Cooling Supply Air Flow Rate {m3/s}','autosize')
  !CR8001
  !CALL AddToObjFld('Supply air volumetric flow rate during heating operation {m3/s}',   &
  !   base + pthpzSupplyHeatFlowRateOff,'')
  CALL AddToObjStr('Heating Supply Air Flow Rate {m3/s}','autosize')
  IF (capacityControlKind == capctrlASHRAE90VarFan) THEN
    CALL AddToObjStr('No Load Supply Air Flow Rate {m3/s}','autosize')
  ELSE
    CALL AddToObjFld('No Load Supply Air Flow Rate {m3/s}',   &
     base + pthpzSupplyNoLoadFlowRateOff,'')
  END IF
  IF (isDedOutAirNameBlank) THEN
    CALL AddToObjStr('Cooling Outdoor Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Heating Outdoor Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('No Load Outdoor Air Flow Rate {m3/s}','autosize')
  ELSE
    CALL AddToObjStr('Cooling Outdoor Air Flow Rate {m3/s}','0.0')
    CALL AddToObjStr('Heating Outdoor Air Flow Rate {m3/s}','0.0')
    CALL AddToObjStr('No Load Outdoor Air Flow Rate {m3/s}','0.0')
  END IF
  CALL AddToObjStr('Supply Air Fan Object Type','Fan:OnOff')
  CALL AddToObjFld('Supply Air Fan Name', base + pthpzNameOff,' PTHP Supply Fan')
  CALL AddToObjStr('Heating Coil Object Type','Coil:Heating:DX:SingleSpeed')
  CALL AddToObjFld('Heating Coil Name', base + pthpzNameOff,' PTHP Heating Coil')
  CALL AddToObjStr('Heating Convergence Tolerance {dimensionless}','0.001')
  CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:DX:SingleSpeed')
  CALL AddToObjFld('Cooling Coil Name', base + pthpzNameOff,' PTHP Cooling Coil')
  CALL AddToObjStr('Cooling Convergence Tolerance {dimensionless}','0.001')
  SELECT CASE (suppHeatCoilKind)
    CASE (shcElectric)
      CALL AddToObjStr('Supplemental Heating Coil Object Type','Coil:Heating:Electric')
    CASE (shcGas)
      CALL AddToObjStr('Supplemental Heating Coil Object Type','Coil:Heating:Fuel')
    CASE (shcWater)
      CALL AddToObjStr('Supplemental Heating Coil Object Type','Coil:Heating:Water')
  END SELECT
  CALL AddToObjFld('Supplemental Heating Coil Name', base + pthpzNameOff,' PTHP Supp Heating Coil')
  CALL AddToObjStr('Maximum Supply Air Temperature from Supplemental Heater {C}','autosize')
  CALL AddToObjFld('Maximum Outdoor Dry-Bulb Temperature for Supplemental Heater Operation {C}',   &
     base + pthpzSuppHeatMaxODBOff,'')
  CALL AddToObjFld('Fan Placement', base + pthpzFanPlacementOff,'')
  SchType=' '
  IF (capacityControlKind == capctrlASHRAE90VarFan) THEN
    CALL AddToObjStr('Supply Air Fan Operating Mode Schedule Name','HVACTemplate-Always 1')
      SchType='1'
    CALL AddToObjStr('Availability Manager List Name','')
    CALL AddToObjStr('Design Specification ZoneHVAC Sizing Object Name','')
    CALL AddToObjStr('Capacity Control Method','SingleZoneVAV')
    CALL AddToObjStr('Minimum Supply Air Temperature in Cooling Mode {C}','autosize')
    CALL AddToObjStr('Maximum Supply Air Temperature in Heating Mode {C}','autosize',.true.)
  ELSE
    SELECT CASE (fanMode)
      CASE (fmCycling)
        CALL AddToObjStr('Supply Air Fan Operating Mode Schedule Name','HVACTemplate-Always 0',.TRUE.)
        SchType='0'
      CASE (fmNeither)
        CALL AddToObjFld('Supply Air Fan Operating Mode Schedule Name', base + pthpzFanModeOff,'',.TRUE.)
    END SELECT
    IF (fanMode .EQ. fmCycling) THEN
      CALL AddAlwaysSchedule('0')
    END IF
  END IF
  IF (SchType /= ' ') CALL AddAlwaysSchedule(SchType)
!end LKL changes
  !FAN:SIMPLE:ONOFF ~ line 95
  CALL CreateNewObj('Fan:OnOff')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Supply Fan')
  IF (.NOT. isSysAvailSchedBlank) THEN
    CALL AddToObjFld('Availability Schedule Name', base + pthpzSysAvailSchedNameOff,'')
  ELSE
    CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
  END IF
  CALL AddToObjFld('Fan Efficiency', base + pthpzFanTotEfficiencyOff,'')
  CALL AddToObjFld('Pressure Rise {Pa}', base + pthpzFanPressureOff,'')
  CALL AddToObjStr('Maximum Flow Rate {m3/s}','autosize')
  CALL AddToObjFld('Motor Efficiency', base + pthpzFanMotorEfficiencyOff,'')
  CALL AddToObjStr('Motor in Airstream Fraction','1')
  IF (isDrawThru) THEN
    CALL AddToObjFld('Air Inlet Node Name', base + pthpzNameOff,' PTHP Heating Coil Outlet')
  ELSE
    CALL AddToObjFld('Air Inlet Node Name', base + pthpzNameOff,' PTHP Mixed Air Outlet')
  END IF
  CALL AddToObjFld('Air Outlet Node Name', base + pthpzNameOff,' PTHP Supply Fan Outlet',.TRUE.)
  IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  !OUTSIDE AIR MIXER ~ line 108
  CALL CreateNewObj('OutdoorAir:Mixer')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP OA Mixing Box')
  CALL AddToObjFld('Mixed Air Node Name', base + pthpzNameOff,' PTHP Mixed Air Outlet')
  CALL AddToObjFld('Outside Air Stream Node Name', base + pthpzNameOff,' PTHP Outside Air Inlet')
  CALL AddToObjFld('Relief Air Stream Node Name', base + pthpzNameOff,' PTHP Relief Air Outlet')
  CALL AddToObjFld('Return Air Stream Node Name', base + pthpzNameOff,' PTHP Return',.TRUE.)
  !OUTSIDE AIR NODE ~ line 115
  CALL CreateNewObj('OutdoorAir:Node')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Outside Air Inlet')
  CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
  !COIL:DX:CoolingBypassFactorEmpirical ~ line 119
  CALL CreateNewObj('Coil:Cooling:DX:SingleSpeed')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Cooling Coil')
  CALL AddToObjFld('Availability Schedule Name', base + pthpzCoolAvailSchedNameOff,'')
  CALL AddToObjFld('Gross Rated Total Cooling Capacity {W}', base + pthpzCoolCoilCapOff,'')
  CALL AddToObjFld('Gross Rated Sensible Heat Ratio', base + pthpzCoolCoilSHROff,'')
  CALL AddToObjFld('Rated COP', base + pthpzCoolCoilCOPOff,'')
  !CR8001
  !CALL AddToObjFld('Rated Air Volume Flow Rate {m3/s}', base + pthpzSupplyCoolFlowRateOff,'')
  CALL AddToObjStr('Rated Air Volume Flow Rate {m3/s}', 'autosize')
  CALL AddToObjStr('Rated Evaporator Fan Power per Volume Flow Rate {W/(m3/s)}','')
  IF (isDrawThru) THEN
    CALL AddToObjFld('Air Inlet Node Name', base + pthpzNameOff,' PTHP Mixed Air Outlet')
  ELSE
    CALL AddToObjFld('Air Inlet Node Name', base + pthpzNameOff,' PTHP Supply Fan Outlet')
  END IF
  CALL AddToObjFld('Air Outlet Node Name', base + pthpzNameOff,' PTHP Cooling Coil Outlet')
  CALL AddToObjFld('Total Cooling Capacity Modifier Curve (function of temperature)',   &
     base + pthpzNameOff,' PTHP Cooling Coil Cap-FT')
  CALL AddToObjFld('Total Cooling Capacity Modifier Curve (function of flow fraction)',   &
     base + pthpzNameOff,' PTHP Cooling Coil Cap-FF')
  CALL AddToObjFld('Energy Input Ratio Modifier Curve (function of temperature)',   &
     base + pthpzNameOff,' PTHP Cooling Coil EIR-FT')
  CALL AddToObjFld('Energy Input Ratio Modifier Curve (function of flow fraction)',   &
     base + pthpzNameOff,' PTHP Cooling Coil EIR-FF')
  CALL AddToObjFld('Part Load Fraction Correlation (function of part load ratio)',   &
     base + pthpzNameOff,' PTHP Cooling Coil Coil PLF')
  CALL AddToObjStr('Minimum Outdoor Dry-Bulb Temperature for Compressor Operation {C}','')
  CALL AddToObjStr('Nominal Time for Condensate Removal to Begin','0')
  CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity','0')
  CALL AddToObjStr('Maximum Cycling Rate','0')
  CALL AddToObjStr('Latent Capacity Time Constant','0')
  CALL AddToObjFld('Condenser Air Inlet Node Name', base + pthpzNameOff,' PTHP Cooling Coil Condenser Inlet')
  CALL AddToObjStr('Condenser Type','AirCooled')
  CALL AddToObjStr('Evaporative Condenser Effectiveness','0')
  CALL AddToObjStr('Evaporative Condenser Air Flow Rate','')
  CALL AddToObjStr('Evaporative Condenser Pump Rated Power Consumption','0')
  CALL AddToObjStr('Crankcase Heater Capacity','0')
  CALL AddToObjStr('Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater Operation','10',.TRUE.)
  !CURVE:BIQUADRATIC ~ line 149
  CALL CreateNewObj('Curve:Biquadratic')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Cooling Coil Cap-FT')
  CALL AddToObjStr('Coefficient1 Constant','0.942587793')
  CALL AddToObjStr('Coefficient2 x','0.009543347')
  CALL AddToObjStr('Coefficient3 x**2','0.00068377')
  CALL AddToObjStr('Coefficient4 y','-0.011042676')
  CALL AddToObjStr('Coefficient5 y**2','0.000005249')
  CALL AddToObjStr('Coefficient6 x*y','-0.00000972')
  CALL AddToObjStr('Minimum Value of x','12.77778')
  CALL AddToObjStr('Maximum Value of x','23.88889')
  CALL AddToObjStr('Minimum Value of y','18.0')
  CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
  !CURVE:QUADRATIC ~ line 163
  CALL CreateNewObj('Curve:Quadratic')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Cooling Coil Cap-FF')
  CALL AddToObjStr('Coefficient1 Constant','0.8')
  CALL AddToObjStr('Coefficient2 x','0.2')
  CALL AddToObjStr('Coefficient3 x**2','0')
  CALL AddToObjStr('Minimum Value of x','0.5')
  CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
  !CURVE:BIQUADRATIC ~ line 172
  CALL CreateNewObj('Curve:Biquadratic')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Cooling Coil EIR-FT')
  CALL AddToObjStr('Coefficient1 Constant','0.342414409')
  CALL AddToObjStr('Coefficient2 x','0.034885008')
  CALL AddToObjStr('Coefficient3 x**2','-0.0006237')
  CALL AddToObjStr('Coefficient4 y','0.004977216')
  CALL AddToObjStr('Coefficient5 y**2','0.000437951')
  CALL AddToObjStr('Coefficient6 x*y','-0.000728028')
  CALL AddToObjStr('Minimum Value of x','12.77778')
  CALL AddToObjStr('Maximum Value of x','23.88889')
  CALL AddToObjStr('Minimum Value of y','18.0')
  CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
  !CURVE:QUADRATIC ~ line 186
  CALL CreateNewObj('Curve:Quadratic')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Cooling Coil EIR-FF')
  CALL AddToObjStr('Coefficient1 Constant','1.1552')
  CALL AddToObjStr('Coefficient2 x','-0.1808')
  CALL AddToObjStr('Coefficient3 x**2','0.0256')
  CALL AddToObjStr('Minimum Value of x','0.5')
  CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
  !CURVE:QUADRATIC ~ line 195
  CALL CreateNewObj('Curve:Quadratic')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Cooling Coil Coil PLF')
  CALL AddToObjStr('Coefficient1 Constant','0.85')
  CALL AddToObjStr('Coefficient2 x','0.15')
  CALL AddToObjStr('Coefficient3 x**2','0')
  CALL AddToObjStr('Minimum Value of x','0')
  CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  !OUTSIDE AIR NODE ~ line 204
  CALL CreateNewObj('OutdoorAir:Node')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Cooling Coil Condenser Inlet')
  CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
  !COIL:DX:HeatingEmpirical ~ line 208
  CALL CreateNewObj('Coil:Heating:DX:SingleSpeed')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Heating Coil')
  CALL AddToObjFld('Availability Schedule Name', base + pthpzHeatAvailSchedNameOff,'')
  CALL AddToObjFld('Rated Total Heating Capacity {W}', base + pthpzHPHeatCoilCapOff,'')
  CALL AddToObjFld('Rated COP', base + pthpzHPHeatCoilCOPOff,'')
  !CR8001
  !CALL AddToObjFld('Rated Air Volume Flow Rate {m3/s}', base + pthpzSupplyHeatFlowRateOff,'')
  CALL AddToObjStr('Rated Air Volume Flow Rate {m3/s}', 'autosize')
  CALL AddToObjStr('Rated Evaporator Fan Power Per Volume Flow Rate','')
  CALL AddToObjFld('Air Inlet Node Name', base + pthpzNameOff,' PTHP Cooling Coil Outlet')
  CALL AddToObjFld('Air Outlet Node Name', base + pthpzNameOff,' PTHP Heating Coil Outlet')
  CALL AddToObjFld('Total heating capacity modifier curve (function of temperature)',   &
     base + pthpzNameOff,' PTHP Heating Coil Cap-FT')
  CALL AddToObjFld('Total heating capacity modifier curve (function of flow fraction)',   &
     base + pthpzNameOff,' PTHP Heating Coil Cap-FF')
  CALL AddToObjFld('Energy input ratio modifier curve (function of temperature)',   &
     base + pthpzNameOff,' PTHP Heating Coil EIR-FT')
  CALL AddToObjFld('Energy input ratio modifier curve (function of flow fraction)',   &
     base + pthpzNameOff,' PTHP Heating Coil EIR-FF')
  CALL AddToObjFld('Part load fraction correlation (function of part load ratio)',   &
     base + pthpzNameOff,' PTHP Heating Coil PLF')
  CALL AddToObjFld('Defrost energy input ratio modifier curve (function of temperature)',   &
     base + pthpzNameOff,' PTHP Heating Coil DefrEIR-FT')
  CALL AddToObjFld('Minimum Outdoor Dry-bulb Temperature for Compressor Operation {C}', base + pthpzHPHeatMinODBOff,'')
  CALL AddToObjStr('Outdoor Dry-Bulb Temperature to Turn On Compressor','')
  CALL AddToObjFld('Maximum Outdoor Dry-bulb Temperature for Defrost Operation {C}', base + pthpzHPDefrMaxODBOff,'')
  CALL AddToObjStr('Crankcase Heater Capacity {W}','0')
  CALL AddToObjStr('Maximum Outdoor Dry-bulb Temperature for Crankcase Heater Operation {C}','0')
  CALL AddToObjFld('Defrost Strategy', base + pthpzHPDefrTypeOff,'')
  CALL AddToObjFld('Defrost Control', base + pthpzHPDefrCtrlTypeOff,'')
  CALL AddToObjFld('Defrost Time Period Fraction', base + pthpzHPDefrTimeFracOff,'')
  CALL AddToObjStr('Resistive Defrost Heater Capacity {W}','autosize',.TRUE.)
  !CURVE:CUBIC ~ line 234
  CALL CreateNewObj('Curve:Cubic')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Heating Coil Cap-FT')
  CALL AddToObjStr('Coefficient1 Constant','0.758746')
  CALL AddToObjStr('Coefficient2 x','0.027626')
  CALL AddToObjStr('Coefficient3 x**2','0.000148716')
  CALL AddToObjStr('Coefficient4 x**3','0.0000034992')
  CALL AddToObjStr('Minimum Value of x','-20.0')
  CALL AddToObjStr('Maximum Value of x','20.0',.TRUE.)
  !CURVE:CUBIC ~ line 243
  CALL CreateNewObj('Curve:Cubic')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Heating Coil Cap-FF')
  CALL AddToObjStr('Coefficient1 Constant','0.84')
  CALL AddToObjStr('Coefficient2 x','0.16')
  CALL AddToObjStr('Coefficient3 x**2','0.0')
  CALL AddToObjStr('Coefficient4 x**3','0.0')
  CALL AddToObjStr('Minimum Value of x','0.5')
  CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
  !CURVE:CUBIC ~ line 252
  CALL CreateNewObj('Curve:Cubic')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Heating Coil EIR-FT')
  CALL AddToObjStr('Coefficient1 Constant','1.19248')
  CALL AddToObjStr('Coefficient2 x','-0.0300438')
  CALL AddToObjStr('Coefficient3 x**2','0.00103745')
  CALL AddToObjStr('Coefficient4 x**3','-0.000023328')
  CALL AddToObjStr('Minimum Value of x','-20.0')
  CALL AddToObjStr('Maximum Value of x','20.0',.TRUE.)
  !CURVE:QUADRATIC ~ line 261
  CALL CreateNewObj('Curve:Quadratic')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Heating Coil EIR-FF')
  CALL AddToObjStr('Coefficient1 Constant','1.3824')
  CALL AddToObjStr('Coefficient2 x','-0.4336')
  CALL AddToObjStr('Coefficient3 x**2','0.0512')
  CALL AddToObjStr('Minimum Value of x','0.0')
  CALL AddToObjStr('Maximum Value of x','1.0',.TRUE.)
  !CURVE:QUADRATIC ~ line 269
  CALL CreateNewObj('Curve:Quadratic')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Heating Coil PLF')
  CALL AddToObjStr('Coefficient1 Constant','0.75')
  CALL AddToObjStr('Coefficient2 x','0.25')
  CALL AddToObjStr('Coefficient3 x**2','0.0')
  CALL AddToObjStr('Minimum Value of x','0.0')
  CALL AddToObjStr('Maximum Value of x','1.0',.TRUE.)
  !CURVE:BIQUADRATIC ~ line 277
  CALL CreateNewObj('Curve:Biquadratic')
  CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Heating Coil DefrEIR-FT')
  CALL AddToObjStr('Coefficient1 Constant','1')
  CALL AddToObjStr('Coefficient2 x','0')
  CALL AddToObjStr('Coefficient3 x**2','0')
  CALL AddToObjStr('Coefficient4 y','0')
  CALL AddToObjStr('Coefficient5 y**2','0')
  CALL AddToObjStr('Coefficient6 x*y','0')
  CALL AddToObjStr('Minimum Value of x','0')
  CALL AddToObjStr('Maximum Value of x','50')
  CALL AddToObjStr('Minimum Value of y','0')
  CALL AddToObjStr('Maximum Value of y','50',.TRUE.)
  SELECT CASE (suppHeatCoilKind)
    CASE (shcElectric)
      !COIL:Electric:Heating ~ line 290
      CALL CreateNewObj('Coil:Heating:Electric')
      CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Supp Heating Coil')
      CALL AddToObjFld('Availability Schedule Name', base + pthpzSuppHeatAvailSchedNameOff,'')
      CALL AddToObjStr('Efficiency','1')
      CALL AddToObjFld('Nominal Capacity of the Coil {W}', base + pthpzSuppHeatCoilCapOff,'')
      IF (isDrawThru) THEN
        CALL AddToObjFld('Air Inlet Node Name', base + pthpzNameOff,' PTHP Supply Fan Outlet')
      ELSE
        CALL AddToObjFld('Air Inlet Node Name', base + pthpzNameOff,' PTHP Heating Coil Outlet')
      END IF
      CALL AddToObjFld('Air Outlet Node Name', base + pthpzNameOff,' PTHP Supply Inlet')
      CALL AddToObjStr('Coil Temp Setpoint Node','',.TRUE.)
    CASE (shcGas)
      !COIL:Gas:Heating ~ line 302
      CALL CreateNewObj('Coil:Heating:Fuel')
      CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Supp Heating Coil')
      CALL AddToObjFld('Availability Schedule Name', base + pthpzSuppHeatAvailSchedNameOff,'')
      CALL AddToObjStr('Fuel Type', 'Gas')
      CALL AddToObjFld('Gas Burner Efficiency', base + pthpzSuppHeatEfficiencyOff,'')
      CALL AddToObjFld('Nominal Capacity of the Coil {W}', base + pthpzSuppHeatCoilCapOff,'')
      IF (isDrawThru) THEN
        CALL AddToObjFld('Air Inlet Node Name', base + pthpzNameOff,' PTHP Supply Fan Outlet')
      ELSE
        CALL AddToObjFld('Air Inlet Node Name', base + pthpzNameOff,' PTHP Heating Coil Outlet')
      END IF
      CALL AddToObjFld('Air Outlet Node Name', base + pthpzNameOff,' PTHP Supply Inlet')
      CALL AddToObjStr('Coil Temp Setpoint Node','')
      CALL AddToObjFld('Parasitic Electric Load {W}', base + pthpzSuppHeatParasiticOff,'')
      CALL AddToObjFld('Part load fraction correlation (function of part load ratio)',   &
         base + pthpzNameOff,' PTHP Supp Heating Coil PLF-FPLR')
      CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
      !CURVE:CUBIC ~ line 316
      CALL CreateNewObj('Curve:Cubic')
      CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Supp Heating Coil PLF-FPLR')
      CALL AddToObjStr('Coefficient1 Constant','0.8')
      CALL AddToObjStr('Coefficient2 x','0.2')
      CALL AddToObjStr('Coefficient3 x**2','0')
      CALL AddToObjStr('Coefficient4 x**3','0')
      CALL AddToObjStr('Minimum Value of x','0')
      CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
    CASE (shcWater)
      !CURVE:CUBIC ~ line 235
      CALL CreateNewObj('Coil:Heating:Water')
      CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Supp Heating Coil')
      CALL AddToObjFld('Availability Schedule Name', base + pthpzHeatAvailSchedNameOff,'')
      CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
      CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Water Inlet Node Name', base + pthpzNameOff,' PTHP Supp Heating Coil HW Inlet')
      CALL AddToObjFld('Water Outlet Node Name', base + pthpzNameOff,' PTHP Supp Heating Coil HW Outlet')
      IF (isDrawThru) THEN
        CALL AddToObjFld('Air Inlet Node Name', base + pthpzNameOff,' PTHP Supply Fan Outlet')
      ELSE
        CALL AddToObjFld('Air Inlet Node Name', base + pthpzNameOff,' PTHP Heating Coil Outlet')
      END IF
      CALL AddToObjFld('Air Outlet Node Name', base + pthpzNameOff,' PTHP Supply Inlet')
      IF (isSuppHeatCoilCapAutosize) THEN
        CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
      ELSE
        CALL AddToObjStr('Performance Input Method','NominalCapacity')
      END IF
      CALL AddToObjFld('Rated Capacity', base + pthpzSuppHeatCoilCapOff,'')
      CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
      CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
      CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
      CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
      CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + pthpzNameOff,' PTHP Supp Heating Coil HW Branch')
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
      CALL AddToObjFld('Component Name', base + pthpzNameOff,' PTHP Supp Heating Coil')
      CALL AddToObjFld('Component Inlet Node Name', base + pthpzNameOff,' PTHP Supp Heating Coil HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + pthpzNameOff,' PTHP Supp Heating Coil HW Outlet',.TRUE.)
      CALL AddToStrList(TRIM(FldVal(base + pthpzNameOff)) // ' PTHP Supp Heating Coil HW Branch',handleHeatingCoilBranch)
  END SELECT
  !Baseboards
  SELECT CASE (baseboardKind)
    CASE (baseboardHotWater)
      !ZoneHVAC:Baseboard:RadiantConvective:Water
      CALL CreateNewObj('ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Name', base + pthpzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + pthpzBaseboardAvailSchedNameOff,'')
      CALL AddToObjFld('Inlet Node Name', base + pthpzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Outlet Node Name', base + pthpzNameOff,' Baseboard HW Outlet')
      CALL AddToObjStr('Rated Average Water Temperature {C}','87.78')
      CALL AddToObjStr('Rated Water Mass Flow Rate {m3/s}','0.063')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + pthpzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Maximum Water Flow Rate','autosize')
      CALL AddToObjStr('Convergence Tolerance','0.001')
      CALL AddToObjStr('Fraction Radiant','0.0')
      CALL AddToObjStr('Fraction of Radiant Energy Incident on People','0.0',.TRUE.)
      !BRANCH (for baseboard)
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + pthpzNameOff,' Baseboard Heat HW Branch')
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Component Name', base + pthpzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Component Inlet Node Name', base + pthpzNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + pthpzNameOff,' Baseboard HW Outlet',.TRUE.)
      CALL AddToStrList(TRIM(FldVal(base + pthpzNameOff)) // ' Baseboard Heat HW Branch',handleHeatingCoilBranch)
      IF ((compactHotLoopBase .EQ. 0) .OR. (numCompactBoiler .EQ. 0)) THEN
        CALL WriteError('Both a HVACTemplate:Plant:HotWaterLoop and a HVACTemplate:Plant:Boiler are needed ' // &
             'when using hot water baseboards.')
      END IF
    CASE (baseboardElectric)
      !BASEBOARD HEATER:Electric:Convective
      CALL CreateNewObj('ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Name', base + pthpzNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + pthpzBaseboardAvailSchedNameOff,'')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + pthpzBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Efficiency','1',.TRUE.)
  END SELECT
END DO
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE CreateBoiler
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a compact boiler object

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for detailed boiler object type
INTEGER, PARAMETER :: blotHotWater = 1

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for boiler
INTEGER :: detblrBase=0  ! used as base for detailed boiler
INTEGER :: iBoiler
INTEGER :: iDetBoiler=0
INTEGER :: detblrObjType=0
INTEGER :: boilerloopbase=0
LOGICAL :: isBoilerHotWater
LOGICAL :: isBoilerCondHotWater
LOGICAL :: isHotPumpVariable
LOGICAL :: isCapacityAutosize
LOGICAL :: isFuelTypeValid
LOGICAL :: isPumpBranch
LOGICAL :: isPumpSchedNameBlank
CHARACTER(len=MaxAlphaLength) :: blrInletNode=''
CHARACTER(len=MaxAlphaLength) :: blrOutletNode=''
CHARACTER(len=MaxAlphaLength) :: pumpHead=''
CHARACTER(len=MaxAlphaLength) :: pumpControlType=''
CHARACTER(len=MaxAlphaLength) :: pumpSchedule=''

IF ((numCompactBoiler + numCompactBoilerOR) .EQ. 0) RETURN
isHotPumpVariable =  .FALSE.
IF (compactHotLoopBase .NE. 0) THEN
  isHotPumpVariable =  SameString(FldVal(compactHotLoopBase + hwpHWPumpConfigOff),'VariableFlow')
  isPumpBranch = SameString(FldVal(compactHotLoopBase + hwpHWPumpTypeOff),'PumpPerBoiler')
  isPumpSchedNameBlank = SameString(FldVal(compactHotLoopBase + hwpPumpSchedNameOff),'')
  pumpSchedule = FldVal(compactHotLoopBase + hwpPumpSchedNameOff)
  pumpHead = FldVal(compactHotLoopBase + hwpHWPumpHeadOff)
  pumpControlType = FldVal(compactHotLoopBase + hwpPumpControlTypeOff)
ELSEIF (compactMixedLoopBase .NE. 0) THEN
  isHotPumpVariable =  SameString(FldVal(compactMixedLoopBase + mwpPumpConfigOff),'VariableFlow')
  isPumpBranch = SameString(FldVal(compactMixedLoopBase + mwpPumpTypeOff),'PumpPerTowerOrBoiler')
  isPumpSchedNameBlank = SameString(FldVal(compactMixedLoopBase + mwpPumpSchedNameOff),'')
  pumpSchedule = FldVal(compactMixedLoopBase + mwpPumpSchedNameOff)
  pumpHead = FldVal(compactMixedLoopBase + mwpPumpHeadOff)
  pumpControlType = FldVal(compactMixedLoopBase + mwpPumpControlTypeOff)
ELSE
  RETURN  !with no loop - no boiler can be hooked up
END IF

! HVACTemplate:Plant:Boiler
DO iBoiler = 1, numCompactBoiler
  base =  compactBoilerBase(iBoiler)
  ! set flags
  !validate boiler type
  isBoilerHotWater = .FALSE.
  isBoilerCondHotWater = .FALSE.
  IF (SameString(FldVal(base + blrTypeOff),'DistrictHotWater')) THEN
  ELSEIF (SameString(FldVal(base + blrTypeOff),'HotWaterBoiler')) THEN
    isBoilerHotWater = .TRUE.
  ELSEIF (SameString(FldVal(base + blrTypeOff),'CondensingHotWaterBoiler')) THEN
    isBoilerCondHotWater = .TRUE.
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Plant:Boiler "'//TRIM(FldVal(base + blrNameOff))//'"'// &
                    ' in the Zone Heating Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + blrTypeOff)))
  END IF
  isCapacityAutosize = SameString(FldVal(base + blrCapacityOff),'autosize')
  isFuelTypeValid = .FALSE.
  IF (SameString(FldVal(base + blrFuelTypeOff),'Electricity')) isFuelTypeValid = .TRUE.
  IF (SameString(FldVal(base + blrFuelTypeOff),'NaturalGas')) isFuelTypeValid = .TRUE.
  IF (SameString(FldVal(base + blrFuelTypeOff),'PropaneGas')) isFuelTypeValid = .TRUE.
  IF (SameString(FldVal(base + blrFuelTypeOff),'FuelOil#1')) isFuelTypeValid = .TRUE.
  IF (SameString(FldVal(base + blrFuelTypeOff),'FuelOil#2')) isFuelTypeValid = .TRUE.
  IF (SameString(FldVal(base + blrFuelTypeOff),'Coal')) isFuelTypeValid = .TRUE.
  IF (SameString(FldVal(base + blrFuelTypeOff),'Diesel')) isFuelTypeValid = .TRUE.
  IF (SameString(FldVal(base + blrFuelTypeOff),'Gasoline')) isFuelTypeValid = .TRUE.
  IF (SameString(FldVal(base + blrFuelTypeOff),'OtherFuel1')) isFuelTypeValid = .TRUE.
  IF (SameString(FldVal(base + blrFuelTypeOff),'OtherFuel2')) isFuelTypeValid = .TRUE.
  IF (SameString(FldVal(base + blrFuelTypeOff),'')) isFuelTypeValid = .TRUE.
  IF (.NOT. isFuelTypeValid) THEN
    CALL WriteError('Invalid choice in HVACTemplate:Plant:Boiler in the Fuel Type field')
  END IF
  ! add Boiler Outlet Node to list of node names for HW setpoints
  IF ((compactHotLoopBase .NE. 0) .AND. &
      (SameString(FldVal(base + blrLoopTypeOff),'HotWater') .OR. SameString(FldVal(base + blrLoopTypeOff),''))) THEN
    CALL AddToStrList(TRIM(FldVal(base + blrNameOff)) // ' HW Outlet',handleBoilerOutletNode)
  ELSEIF (SameString(FldVal(base + blrLoopTypeOff),'MixedWater') .OR. SameString(FldVal(base + blrLoopTypeOff),'')) THEN
    CALL AddToStrList(TRIM(FldVal(base + blrNameOff)) // ' HW Outlet',handleMixedBoilerOutletNode)
  END IF
  ! write idf objects
  IF (isBoilerHotWater .OR. isBoilerCondHotWater) THEN
    !BOILER:SIMPLE ~line 4
    CALL CreateNewObj('Boiler:HotWater')
    CALL AddToObjFld('Name', base + blrNameOff,'')
    CALL AddToObjFld('Fuel Type', base + blrFuelTypeOff,'')
    CALL AddToObjFld('Nominal Capacity {W}', base + blrCapacityOff,'')
    CALL AddToObjFld('Nominal Thermal Efficiency', base + blrEfficOff,'')
    IF (isBoilerCondHotWater) THEN
      CALL AddToObjStr('Efficiency Curve Temperature Evaluation Variable', 'EnteringBoiler ')
      CALL AddToObjFld('Normalized Boiler Efficiency Curve Name', base + blrNameOff,' Condensing Boiler Efficiency Curve')
    ELSE
      CALL AddToObjStr('Efficiency Curve Temperature Evaluation Variable', 'LeavingBoiler')
      CALL AddToObjFld('Normalized Boiler Efficiency Curve Name', base + blrNameOff,' Efficiency Curve')
    END IF
    CALL AddToObjStr('Max Design Boiler Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Minimum Part Load Ratio',base + blrMinPLROff,'')
    CALL AddToObjFld('Maximum Part Load Ratio',base + blrMaxPLROff,'')
    CALL AddToObjFld('Optimum Part Load Ratio',base + blrOptPLROff,'')
    CALL AddToObjFld('Water Inlet Node Name', base + blrNameOff,' HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + blrNameOff,' HW Outlet')
    CALL AddToObjFld('Temp Upper Limit Water Outlet {C}',base + blrOutletUpLimOff,'')
    IF (isHotPumpVariable) THEN
      CALL AddToObjStr('Boiler Flow Mode','LeavingSetpointModulated')
    ELSE
      CALL AddToObjStr('Boiler Flow Mode','ConstantFlow')
    END IF
    CALL AddToObjStr('Parasitic Electric Load {W}','0')
    CALL AddToObjFld('Sizing Factor', base + blrSizeFactorOff,'',.TRUE.)

    IF (isBoilerCondHotWater) THEN
      ! Boiler curve from R.Raustad, FSEC, similar to curve from Fig. 9 of Viessmann Technical Series
      !  "Condensing Technology", 9446 803 - 1 GB 11/2004, viessmann.com
      CALL CreateNewObj('Curve:BiQuadratic')
      CALL AddToObjFld('Name', base + blrNameOff,' Condensing Boiler Efficiency Curve')
      CALL AddToObjStr('Coefficient1 Constant','1.124970374')
      CALL AddToObjStr('Coefficient2 x','0.014963852')
      CALL AddToObjStr('Coefficient3 x**2','-0.02599835')
      CALL AddToObjStr('Coefficient4 y','0.0')
      CALL AddToObjStr('Coefficient5 y**2','-1.40464E-6')
      CALL AddToObjStr('Coefficient6 x*y','-0.00153624')
      CALL AddToObjStr('Minimum Value of x','0.1')
      CALL AddToObjStr('Maximum Value of x','1.0')
      CALL AddToObjStr('Minimum Value of y','30.0')
      CALL AddToObjStr('Maximum Value of y','85.0',.TRUE.)
    ELSE
      CALL CreateNewObj('Curve:Quadratic')
      CALL AddToObjFld('Name', base + blrNameOff,' Efficiency Curve')
      CALL AddToObjStr('Coefficient1 Constant','0.97')
      CALL AddToObjStr('Coefficient2 x','0.0633')
      CALL AddToObjStr('Coefficient3 x**2','-0.0333')
      CALL AddToObjStr('Minimum Value of x','0.0')
      CALL AddToObjStr('Maximum Value of x','1.0',.TRUE.)
    END IF

    !BRANCH ~ line 24
    CALL CreateNewObj('Branch')
    IF ((compactHotLoopBase .NE. 0) .AND. &
        (SameString(FldVal(base + blrLoopTypeOff),'HotWater') .OR. SameString(FldVal(base + blrLoopTypeOff),''))) THEN
      CALL AddToObjFld('Name', base + blrNameOff,' HW Branch')
      CALL AddToStrList(TRIM(FldVal(base + blrNameOff)) // ' HW Branch',handleHWSupplyBranch)
    ELSEIF (SameString(FldVal(base + blrLoopTypeOff),'MixedWater') .OR. SameString(FldVal(base + blrLoopTypeOff),'')) THEN
      CALL AddToObjFld('Name', base + blrNameOff,' Mixed Branch')
      CALL AddToStrList(TRIM(FldVal(base + blrNameOff)) // ' Mixed Branch',handleMixedSupplyBranch)
    ENDIF
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isPumpBranch) THEN
      !not using branch pumps so only boiler is on the branch
      CALL AddToObjStr('Component Object Type','Boiler:HotWater')
      CALL AddToObjFld('Component Name', base + blrNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + blrNameOff,' HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + blrNameOff,' HW Outlet',.TRUE.)
    ELSE
      ! using a branch pump define the pump first followed by the boiler
      IF (isHotPumpVariable) THEN
        CALL AddToObjStr('Component Object Type','Pump:VariableSpeed')
      ELSE
        CALL AddToObjStr('Component Object Type','Pump:ConstantSpeed')
      END IF
      CALL AddToObjFld('Component Name', base + blrNameOff,' Branch Pump')
      CALL AddToObjFld('Component Inlet Node Name', base + blrNameOff,' HW Pump Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + blrNameOff,' HW Inlet')
      CALL AddToObjStr('Component Object Type','Boiler:HotWater')
      CALL AddToObjFld('Component Name', base + blrNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + blrNameOff,' HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + blrNameOff,' HW Outlet',.TRUE.)
    END IF
  ELSE
    !PURCHASED:HOT WATER ~ line 33
    CALL CreateNewObj('DistrictHeating')
    CALL AddToObjFld('Name', base + blrNameOff,'')
    CALL AddToObjFld('Plant_Loop_Inlet_Node', base + blrNameOff,' HW Inlet')
    CALL AddToObjFld('Plant_Loop_Outlet_Node', base + blrNameOff,' HW Outlet')
    IF (isCapacityAutosize) THEN
      CALL AddToObjStr('Nominal Capacity {W}','1000000000000000',.TRUE.)
    ELSE
      CALL AddToObjFld('Nominal Capacity {W}', base + blrCapacityOff,'',.TRUE.)
    END IF
    !BRANCH ~ line 40
    CALL CreateNewObj('Branch')
    IF (compactHotLoopBase .NE. 0) THEN
      CALL AddToObjFld('Name', base + blrNameOff,' HW Branch')
      CALL AddToStrList(TRIM(FldVal(base + blrNameOff)) // ' HW Branch',handleHWSupplyBranch)
    ELSE
      CALL AddToObjFld('Name', base + blrNameOff,' Mixed Branch')
      CALL AddToStrList(TRIM(FldVal(base + blrNameOff)) // ' Mixed Branch',handleMixedSupplyBranch)
    ENDIF
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isPumpBranch) THEN
      CALL AddToObjStr('Component Object Type','DistrictHeating')
      CALL AddToObjFld('Component Name', base + blrNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + blrNameOff,' HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + blrNameOff,' HW Outlet',.TRUE.)
    ELSE
      IF (isHotPumpVariable) THEN
        CALL AddToObjStr('Component Object Type','Pump:VariableSpeed')
      ELSE
        CALL AddToObjStr('Component Object Type','Pump:ConstantSpeed')
      END IF
      CALL AddToObjFld('Component Name', base + blrNameOff,' Branch Pump')
      CALL AddToObjFld('Component Inlet Node Name', base + blrNameOff,' HW Pump Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + blrNameOff,' HW Inlet')
      CALL AddToObjStr('Component Object Type','DistrictHeating')
      CALL AddToObjFld('Component Name', base + blrNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + blrNameOff,' HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + blrNameOff,' HW Outlet',.TRUE.)
    END IF
  END IF
  !Add branch pumps if required - the branch is created above
  IF (isPumpBranch) THEN
    IF (isHotPumpVariable) THEN
      CALL WriteComment('Pump part load coefficients based on simple parabola to represent a generic VSD drive.')
      CALL CreateNewObj('Pump:VariableSpeed')
      CALL AddToObjFld('Name', base + blrNameOff,' Branch Pump')
      CALL AddToObjFld('Inlet Node Name', base + blrNameOff,' HW Pump Inlet')
      CALL AddToObjFld('Outlet Node Name', base + blrNameOff,' HW Inlet')
      CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Rated Pump Head {Pa}', TRIM(pumpHead))
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      CALL AddToObjStr('Coefficient 1 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 2 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 3 of the Part Load Performance Curve','1.0')
      CALL AddToObjStr('Coefficient 4 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Min Flow Rate while operating in variable flow capacity {m3/s}','0')
      CALL AddToObjStr('Pump Control Type', TRIM(pumpControlType))
      CALL AddToObjStr('Pump Flow Rate Schedule Name', TRIM(pumpSchedule),.TRUE.)
    ELSE
      CALL CreateNewObj('Pump:ConstantSpeed')
      CALL AddToObjFld('Name', base + blrNameOff,' Branch Pump')
      CALL AddToObjFld('Inlet Node Name', base + blrNameOff,' HW Pump Inlet')
      CALL AddToObjFld('Outlet Node Name', base + blrNameOff,' HW Inlet')
      CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Rated Pump Head {Pa}', TRIM(pumpHead))
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      CALL AddToObjStr('Pump Control Type', TRIM(pumpControlType))
      CALL AddToObjStr('Pump Flow Rate Schedule Name', TRIM(pumpSchedule),.TRUE.)
    END IF
  END IF
END DO

! HVACTemplate:Plant:Boiler:ObjectReference
DO iBoiler = 1, numCompactBoilerOR
  base =  compactBoilerORBase(iBoiler)

  ! set flags
  IF (SameString(FldVal(base + blrorObjTypeOff),'Boiler:HotWater')) THEN
    detblrObjType = blotHotWater
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Plant:Boiler:ObjectReference "'//TRIM(FldVal(base + blrorNameOff))//'"'// &
                    ' in the Boiler Object Type field: '//TRIM(FldVal(base + blrorObjTypeOff)))
  END IF
  ! find referenced detailed boiler object
  detblrBase = -1
  ! Boiler:HotWater
  IF (detblrObjType == blotHotWater) THEN
    DO iDetBoiler = 1, numDetBoilerHW
      IF (SameString(FldVal(detBoilerHWBase(iDetBoiler) + blrhwNameOff), FldVal(base +  blrorObjNameOff))) THEN
        detblrBase = detBoilerHWBase(iDetBoiler)
        EXIT
      END IF
    END DO
    IF (detblrBase .NE. -1) THEN
      blrInletNode   = TRIM(FldVal(detblrBase + blrhwInletNodeOff))
      blrOutletNode  = TRIM(FldVal(detblrBase + blrhwOutletNodeOff))
    ELSE
      CALL WriteError('In HVACTemplate:Plant:Boiler:ObjectReference "'// TRIM(FldVal(base + blrorNameOff))//'"'// &
                      ' Referenced boiler not found: '//TRIM(FldVal(base + blrorObjTypeOff))//'="'// &
                      TRIM(FldVal(base + blrorObjNameOff))//'"')
      CYCLE
    ENDIF
  ENDIF

  ! check if node names in referenced boiler are all unique (within the boiler - not checking across boiler)
  IF (SameString(TRIM(blrInletNode),TRIM(blrOutletNode))) THEN
    CALL WriteError('In HVACTemplate:Plant:Boiler:ObjectReference "'// TRIM(FldVal(base + blrorNameOff))//'"'// &
                 ' Duplicate node name found in referenced boiler: '//TRIM(FldVal(base + blrorObjTypeOff))//'="'// &
                 TRIM(FldVal(base + blrorObjNameOff))//'" All boiler inlet and outlet node names must be unique.')
  END IF

  ! check if node names in referenced boiler are blank
  IF (SameString(TRIM(blrInletNode),'')) THEN
    CALL WriteError('In HVACTemplate:Plant:Boiler:ObjectReference "'// TRIM(FldVal(base + blrorNameOff))//'"'// &
      ' Blank Inlet Node Name found in referenced boiler: '//TRIM(FldVal(base + blrorObjTypeOff))//'="'// &
      TRIM(FldVal(base + blrorObjNameOff))//'" The user must specify all node names in the referenced boiler.')
  END IF
  IF (SameString(TRIM(blrOutletNode),'')) THEN
    CALL WriteError('In HVACTemplate:Plant:Boiler:ObjectReference "'// TRIM(FldVal(base + blrorNameOff))//'"'// &
      ' Blank Chilled Water Outlet Node Name found in referenced boiler: '//TRIM(FldVal(base + blrorObjTypeOff))//'="'// &
      TRIM(FldVal(base + blrorObjNameOff))//'" The user must specify all node names in the referenced boiler.')
  END IF
  ! add HW Outlet Node to list of node names for HW setpoints
  IF ((compactHotLoopBase .NE. 0) .AND. &
        (SameString(FldVal(base + blrorLoopTypeOff),'HotWater') .OR. SameString(FldVal(base + blrorLoopTypeOff),''))) THEN
    CALL AddToStrList(TRIM(blrOutletNode),handleBoilerOutletNode)
  ELSEIF (SameString(FldVal(base + blrorLoopTypeOff),'MixedWater') .OR. SameString(FldVal(base + blrorLoopTypeOff),'')) THEN
    CALL AddToStrList(TRIM(blrOutletNode),handleMixedBoilerOutletNode)
  END IF
  ! write branch objects
  !BRANCH ~line 50
  CALL CreateNewObj('Branch')
  IF ((compactHotLoopBase .NE. 0) .AND. &
        (SameString(FldVal(base + blrorLoopTypeOff),'HotWater') .OR. SameString(FldVal(base + blrorLoopTypeOff),''))) THEN
    CALL AddToObjFld('Name', base + blrorObjNameOff,' HW Branch')
    CALL AddToStrList(TRIM(FldVal(base + blrorObjNameOff)) // ' HW Branch',handleHWSupplyBranch)
  ELSEIF (SameString(FldVal(base + blrorLoopTypeOff),'MixedWater') .OR. SameString(FldVal(base + blrorLoopTypeOff),'')) THEN
    CALL AddToObjFld('Name', base + blrorObjNameOff,' Mixed Branch')
    CALL AddToStrList(TRIM(FldVal(base + blrorObjNameOff)) // ' Mixed Branch',handleMixedSupplyBranch)
  ENDIF
  CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isPumpBranch) THEN
      !not using branch pumps so only boiler is on the branch
    CALL AddToObjFld('Component Object Type',base + blrorObjTypeOff,'')
    CALL AddToObjFld('Component Name', base + blrorObjNameOff,'')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(blrInletNode))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(blrOutletNode),.TRUE.)
  ELSE
    ! using a branch pump define the pump first followed by the boiler
    IF (isHotPumpVariable) THEN
      CALL AddToObjStr('Component Object Type','Pump:VariableSpeed')
    ELSE
      CALL AddToObjStr('Component Object Type','Pump:ConstantSpeed')
    END IF
    CALL AddToObjFld('Component Name', base + blrorObjNameOff,' Branch Pump')
    CALL AddToObjFld('Component Inlet Node Name', base + blrorObjNameOff,' HW Pump Inlet')
    CALL AddToObjStr('Component Outlet Node Name', TRIM(blrInletNode))
    CALL AddToObjFld('Component Object Type',base + blrorObjTypeOff,'')
    CALL AddToObjFld('Component Name', base + blrorObjNameOff,'')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(blrInletNode))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(blrOutletNode),.TRUE.)
  ENDIF
  !Add branch pumps if required - the branch is created above
  IF (isPumpBranch) THEN
    IF (isHotPumpVariable) THEN
      CALL WriteComment('Pump part load coefficients based on simple parabola to represent a generic VSD drive.')
      CALL CreateNewObj('Pump:VariableSpeed')
      CALL AddToObjFld('Name', base + blrorObjNameOff,' Branch Pump')
      CALL AddToObjFld('Inlet Node Name', base + blrorObjNameOff,' HW Pump Inlet')
      CALL AddToObjStr('Outlet Node Name', TRIM(blrInletNode))
      CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Rated Pump Head {Pa}', TRIM(pumpHead))
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      CALL AddToObjStr('Coefficient 1 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 2 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 3 of the Part Load Performance Curve','1.0')
      CALL AddToObjStr('Coefficient 4 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Min Flow Rate while operating in variable flow capacity {m3/s}','0')
      CALL AddToObjStr('Pump Control Type', TRIM(pumpControlType))
      CALL AddToObjStr('Pump Flow Rate Schedule Name', TRIM(pumpSchedule),.TRUE.)
    ELSE
      CALL CreateNewObj('Pump:ConstantSpeed')
      CALL AddToObjFld('Name', base + blrorObjNameOff,' Branch Pump')
      CALL AddToObjFld('Inlet Node Name', base + blrorObjNameOff,' HW Pump Inlet')
      CALL AddToObjStr('Outlet Node Name', TRIM(blrInletNode))
      CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Rated Pump Head {Pa}', TRIM(pumpHead))
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      CALL AddToObjStr('Pump Control Type', TRIM(pumpControlType))
      CALL AddToObjStr('Pump Flow Rate Schedule Name', TRIM(pumpSchedule),.TRUE.)
    END IF
  END IF
END DO
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE CreateHotLoop
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for COMPACT HVAC:PLANT:HOT WATER LOOP

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for pump configuration type
INTEGER,PARAMETER :: pcVariable = 1
INTEGER,PARAMETER :: pcConstant = 2
          ! parameters for setpoint reset type
INTEGER,PARAMETER :: sprtNone = 1
INTEGER,PARAMETER :: sprtOutdoor = 2
          ! parameters for fluid type
INTEGER,PARAMETER :: ftWater = 1
INTEGER,PARAMETER :: ftEthyleneGlycol30 = 2
INTEGER,PARAMETER :: ftEthyleneGlycol40 = 3
INTEGER,PARAMETER :: ftEthyleneGlycol50 = 4
INTEGER,PARAMETER :: ftEthyleneGlycol60 = 5
INTEGER,PARAMETER :: ftPropyleneGlycol30 = 6
INTEGER,PARAMETER :: ftPropyleneGlycol40 = 7
INTEGER,PARAMETER :: ftPropyleneGlycol50 = 8
INTEGER,PARAMETER :: ftPropyleneGlycol60 = 9
INTEGER,PARAMETER :: ftSteam = 10

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for
LOGICAL :: isHwpHWOperTypeDefault
LOGICAL :: ishwpPumpSchedNameBlank
LOGICAL :: isHwpHWSetPtSchedName
LOGICAL :: isSupplyBypassYes
LOGICAL :: isDemandBypassYes
LOGICAL :: isMaxOutTempBlank
INTEGER :: PumpConfig=0
INTEGER :: iBoiler
INTEGER :: boilerBase
INTEGER :: iStr
INTEGER :: setPtResetType=0
INTEGER :: fluidType=0
INTEGER :: lastBranch=0
INTEGER :: lastSupplyBranch=0
INTEGER :: numBoilersThisLoop=0
LOGICAL :: isPumpHeader
LOGICAL :: isPumpBranch
LOGICAL :: isEndOfList
CHARACTER(len=1) :: pumpCount='1'

base =  compactHotLoopBase
IF (base .EQ. 0) RETURN
IF ((numCompactBoiler + numCompactBoilerOR) .EQ. 0) RETURN
! set flags
CALL validChoice(base + hwpPumpControlTypeOff,'Intermittent','Continuous','HVACTemplate:Plant:HotWaterLoop','Pump Control Type')
isHwpHWOperTypeDefault = SameString(FldVal(base + hwpHWOperTypeOff),'Default')
CALL validChoice(base + hwpHWOperTypeOff,'Default','UserDefined','HVACTemplate:Plant:HotWaterLoop','Plant Operation Scheme Type')
ishwpPumpSchedNameBlank = SameString(FldVal(base + hwpPumpSchedNameOff),'')
IF (SameString(FldVal(base +  hwpHWSetPtResetTypeOff), 'OutdoorAirTemperatureReset')) THEN
  setPtResetType = sprtOutdoor
ELSE IF (SameString(FldVal(base +  hwpHWSetPtResetTypeOff), 'None')) THEN
  setPtResetType = sprtNone
ELSE
  CALL WriteError('Invalid Hot Water Setpoint Reset Type.  None assumed.')
  setPtResetType = sprtNone
END IF
IF (SameString(FldVal(base + hwpHWPumpConfigOff),'VariableFlow')) THEN
  PumpConfig = pcVariable
ELSEIF (SameString(FldVal(base + hwpHWPumpConfigOff),'ConstantFlow')) THEN
  PumpConfig = pcConstant
ELSEIF (SameString(FldVal(base + hwpHWPumpConfigOff),'')) THEN
  PumpConfig = pcConstant
ELSE
  CALL WriteError('Invalid choice in HVACTemplate:Plant:HotWaterLoop in the Hot Water Pump Configuration field')
END IF
CALL validChoice(base + hwpSupplyBypassOff,'Yes','No','HVACTemplate:Plant:HotWaterLoop','Supply Side Bypass Pipe')
isSupplyBypassYes = SameString(FldVal(base + hwpSupplyBypassOff),'Yes')
CALL validChoice(base + hwpDemandBypassOff,'Yes','No','HVACTemplate:Plant:HotWaterLoop','Demand Side Bypass Pipe')
isDemandBypassYes = SameString(FldVal(base + hwpDemandBypassOff),'Yes')
IF (SameString(FldVal(base + hwpFluidTypeOff),'Water')) THEN
  FluidType = ftWater
ELSEIF (SameString(FldVal(base + hwpFluidTypeOff),'EthyleneGlycol30')) THEN
  FluidType = ftEthyleneGlycol30
ELSEIF (SameString(FldVal(base + hwpFluidTypeOff),'EthyleneGlycol40')) THEN
  FluidType = ftEthyleneGlycol40
ELSEIF (SameString(FldVal(base + hwpFluidTypeOff),'EthyleneGlycol50')) THEN
  FluidType = ftEthyleneGlycol50
ELSEIF (SameString(FldVal(base + hwpFluidTypeOff),'EthyleneGlycol60')) THEN
  FluidType = ftEthyleneGlycol60
ELSEIF (SameString(FldVal(base + hwpFluidTypeOff),'PropyleneGlycol30')) THEN
  FluidType = ftPropyleneGlycol30
ELSEIF (SameString(FldVal(base + hwpFluidTypeOff),'PropyleneGlycol40')) THEN
  FluidType = ftPropyleneGlycol40
ELSEIF (SameString(FldVal(base + hwpFluidTypeOff),'PropyleneGlycol50')) THEN
  FluidType = ftPropyleneGlycol50
ELSEIF (SameString(FldVal(base + hwpFluidTypeOff),'PropyleneGlycol60')) THEN
  FluidType = ftPropyleneGlycol60
!ELSEIF (SameString(FldVal(base + hwpFluidTypeOff),'Steam')) THEN
!  FluidType = ftSteam
ELSE
  CALL WriteError('Invalid choice in HVACTemplate:Plant:HotWaterLoop in the Fluid Type field')
END IF
isMaxOutTempBlank = SameString(FldVal(base + hwpMaxTOutOff),'')
IF (SameString(FldVal(base + hwpHWPumpTypeOff),'SinglePump')) THEN
  isPumpHeader = .FALSE.
  isPumpBranch = .FALSE.
ELSEIF (SameString(FldVal(base + hwpHWPumpTypeOff),'PumpPerBoiler')) THEN
  isPumpHeader = .FALSE.
  isPumpBranch = .TRUE.
ELSEIF (SameString(FldVal(base + hwpHWPumpTypeOff),'TwoHeaderedPumps')) THEN
  isPumpHeader = .TRUE.
  isPumpBranch = .FALSE.
  pumpCount = '2'
ELSEIF (SameString(FldVal(base + hwpHWPumpTypeOff),'ThreeHeaderedPumps')) THEN
  isPumpHeader = .TRUE.
  isPumpBranch = .FALSE.
  pumpCount = '3'
ELSEIF (SameString(FldVal(base + hwpHWPumpTypeOff),'FourHeaderedPumps')) THEN
  isPumpHeader = .TRUE.
  isPumpBranch = .FALSE.
  pumpCount = '4'
ELSEIF (SameString(FldVal(base + hwpHWPumpTypeOff),'FiveHeaderedPumps')) THEN
  isPumpHeader = .TRUE.
  isPumpBranch = .FALSE.
  pumpCount = '5'
ELSE
  CALL WriteError('Invalid choice in HVACTemplate:Plant:HotWaterLoop in the Hot Water Pump Type field')
END IF

!Sort the boilers by order of priority - only boilers serving a hot water loop, also count boilers on this loop
numBoilersThisLoop = 0
CALL ClearNumSortList
DO iBoiler = 1, numCompactBoiler
  boilerBase =  compactBoilerBase(iBoiler)  !store the base pointer to each boiler
  IF (SameString(FldVal(boilerBase + blrLoopTypeOff),'HotWater') .OR. SameString(FldVal(boilerBase + blrLoopTypeOff),'')) THEN
    CALL AddToNumSortList(StringToReal(FldVal(boilerBase + blrPriorityOff)),boilerBase)
    numBoilersThisLoop = numBoilersThisLoop + 1
  END IF
END DO
DO iBoiler = 1, numCompactBoilerOR
  boilerBase =  compactBoilerORBase(iBoiler)  !store the base pointer to each boiler
  IF (SameString(FldVal(boilerBase + blrorLoopTypeOff),'HotWater') .OR. SameString(FldVal(boilerBase + blrorLoopTypeOff),'')) THEN
    CALL AddToNumSortList(StringToReal(FldVal(boilerBase + blrorPriorityOff)),boilerBase)
    numBoilersThisLoop = numBoilersThisLoop + 1
  END IF
END DO
IF (numBoilersThisLoop == 0) THEN
  CALL WriteError('Severe:  In HVACTemplate:Plant:HotWaterLoop "'//TRIM(FldVal(base + hwpNameOff))//'"'// &
                  ' there are no boilers serving this loop.')
END IF
CALL SortIncreasingNumList
! write idf objects
!PLANT SIZING ~ line 4
CALL CreateNewObj('Sizing:Plant')
CALL AddToObjFld('Plant or Condenser Loop Name', base + hwpNameOff,' Hot Water Loop')
CALL AddToObjStr('Loop Type','Heating')
CALL AddToObjFld('Design loop exit temperature {C}', base + hwpHWSetPtDesignOff,'')
CALL AddToObjFld('Design loop delta T {deltaC}',base + hwpDesDeltaTOff,'',.TRUE.)
!PLANT LOOP ~ line 10
CALL CreateNewObj('PlantLoop')
CALL AddToObjFld('Name', base + hwpNameOff,' Hot Water Loop')
IF (fluidType .EQ. ftWater) THEN
  CALL AddToObjStr('Fluid Type','Water')
  CALL AddToObjStr('User Defined Fluid Type','')
ELSEIF (fluidType .EQ. ftSteam) THEN
  ! for now steam is not implemented. Use water instead.
  CALL AddToObjStr('Fluid Type','Water')
  CALL AddToObjStr('User Defined Fluid Type','')
ELSE
  CALL AddToObjStr('Fluid Type','UserDefinedFluidType')
  CALL AddToObjFld('User Defined Fluid Type', base + hwpNameOff,' Loop Fluid')
END IF
IF (isHwpHWOperTypeDefault) THEN
  CALL AddToObjFld('PlantEquipmentOperationSchemes Name', base + hwpNameOff,' Operation')
ELSE
  CALL AddToObjFld('PlantEquipmentOperationSchemes Name', base + hwpHWOperNameOff,'')
END IF
CALL AddToObjFld('Loop Temperature Setpoint Node Name', base + hwpNameOff,' HW Supply Outlet')
CALL AddToObjStr('Maximum Loop Temperature {C}','100')
CALL AddToObjStr('Minimum Loop Temperature {C}','10')
CALL AddToObjStr('Maximum Loop Volumetric Flow Rate {m3/s}','autosize')
CALL AddToObjStr('Minimum Loop Volumetric Flow Rate {m3/s}','0')
CALL AddToObjStr('volume of the plant loop {m3}','autosize')
CALL AddToObjFld('Plant Side Inlet Node Name', base + hwpNameOff,' HW Supply Inlet')
CALL AddToObjFld('Plant Side Outlet Node Name', base + hwpNameOff,' HW Supply Outlet')
CALL AddToObjFld('Plant Side Branch List Name', base + hwpNameOff,' HW Supply Side Branches')
CALL AddToObjFld('Plant Side Connector List Name', base + hwpNameOff,' HW Supply Side Connectors')
CALL AddToObjFld('Demand Side Inlet Node Name', base + hwpNameOff,' HW Demand Inlet')
CALL AddToObjFld('Demand Side Outlet Node Name', base + hwpNameOff,' HW Demand Outlet')
CALL AddToObjFld('Demand Side Branch List Name', base + hwpNameOff,' HW Demand Side Branches')
CALL AddToObjFld('Demand Side Connector List Name', base + hwpNameOff,' HW Demand Side Connectors')
CALL AddToObjFld('Load Distribution Scheme', base + hwpLdDistSchmOff,'')
IF (isMaxOutTempBlank) THEN
  CALL AddToObjStr('AvailabilityManagerAssignmentList','')
ELSE
  CALL AddToObjFld('AvailabilityManagerAssignmentList', base + hwpNameOff,' Availability List')
END IF
CALL AddToObjStr('Plant Loop Demand Calculation Scheme','SingleSetpoint',.TRUE.)
IF (isHwpHWOperTypeDefault) THEN
  !PLANT OPERATION SCHEMES ~ line 33
  CALL CreateNewObj('PlantEquipmentOperationSchemes')
  CALL AddToObjFld('Name', base + hwpNameOff,' Operation')
  CALL AddToObjStr('Control Scheme Object Type','PlantEquipmentOperation:HeatingLoad')
  CALL AddToObjFld('Control Scheme Name', base + hwpNameOff,' Operation All Hours')
  CALL AddToObjStr('Control Scheme Schedule Name','HVACTemplate-Always 1',.TRUE.)
  CALL AddAlwaysSchedule('1')
  !HEATING LOAD RANGE BASED OPERATION ~ line 39
  CALL CreateNewObj('PlantEquipmentOperation:HeatingLoad')
  CALL AddToObjFld('Name', base + hwpNameOff,' Operation All Hours')
  CALL AddToObjStr('Load Range Lower Limit 1 {W}','0')
  CALL AddToObjStr('Load Range Upper Limit 1 {W}','1000000000000000')
  CALL AddToObjFld('Priority Control 1 Equipment List Name', base + hwpNameOff,' All Equipment',.TRUE.)
  !PLANT EQUIPMENT LIST ~ line 45
  CALL CreateNewObj('PlantEquipmentList')
  CALL AddToObjFld('Name', base + hwpNameOff,' All Equipment')
  DO iBoiler = 1, numBoilersThisLoop
    boilerBase = NumSortList(iBoiler)%sortedPtr
    isEndOfList = (iBoiler .EQ. numBoilersThisLoop)
    IF (SameString(FldVal(boilerBase),'HVACTemplate:Plant:Boiler')) THEN
      !If blrTypeOff="Hot Water Boiler" then this field is "BOILER:SIMPLE"
      !If blrTypeOff="Purchased Hot Water" then this field is "PURCHASED:HOT WATER"
      IF (SameString(FldVal(boilerBase + blrTypeOff),'DistrictHotWater')) THEN
        CALL AddToObjStr('Equipment Object Type', 'DistrictHeating')
      ELSE
        CALL AddToObjStr('Equipment Object Type', 'Boiler:HotWater')
      END IF
      CALL AddToObjFld('Equipment Name', boilerBase + blrNameOff,'',isEndOfList)
    ELSEIF (SameString(FldVal(boilerBase),'HVACTemplate:Plant:Boiler:ObjectReference')) THEN
      CALL AddToObjFld('Equipment Object Type', boilerBase + blrorObjTypeOff,'')
      CALL AddToObjFld('Equipment Name', boilerBase + blrorObjNameOff,'',isEndOfList)
    ENDIF
  END DO
END IF
!If  hwpHWSetPtSchedNameOff is blank, then create schedule of hwpHWSetPtDesignOff
IF (FldVal(base + hwpHWSetPtSchedNameOff) .EQ. '') THEN
  isHwpHWSetPtSchedName = .FALSE.
! MJW 21 Sep 2010 - this is not needed here, done below
!  CALL AddAlwaysSchedule(FldVal(base + hwpHWSetPtDesignOff))
ELSE
  isHwpHWSetPtSchedName = .TRUE.
END IF

! Variable flow boilers need setpoints, make a nodelist regardless of flow type
CALL CreateNewObj('NodeList')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Setpoint Nodes')
!use list created in boiler object
DO iStr = 1,numStrList
  IF(StrList(iStr)%Handle .EQ. handleBoilerOutletNode) THEN
    CALL AddToObjStr('Node Name', StrList(iStr)%StrToSave)
  END IF
END DO
CALL AddToObjFld('Node Name', base + hwpNameOff,' HW Supply Outlet',.TRUE.)

IF (setPtResetType .EQ. sprtNone) THEN
  !SET POINT MANAGER:SCHEDULED ~ line 50
  IF (.NOT. isHwpHWSetPtSchedName) THEN
    CALL AddAlwaysSchedule(FldVal(base + hwpHWSetPtDesignOff))
  END IF
  CALL CreateNewObj('SetpointManager:Scheduled')
  CALL AddToObjFld('Name', base + hwpNameOff,' HW Temp Manager')
  CALL AddToObjStr('Control Variable','Temperature')
  IF (isHwpHWSetPtSchedName) THEN
    CALL AddToObjFld('Schedule Name', base + hwpHWSetPtSchedNameOff,'')
  ELSE
    CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + hwpHWSetPtDesignOff)))
  END IF
  CALL AddToObjFld('Setpoint Node or NodeList Name', base + hwpNameOff,' HW Supply Setpoint Nodes',.TRUE.)
ELSEIF (setPtResetType .EQ. sprtOutdoor) THEN
  CALL CreateNewObj('SetpointManager:OutdoorAirReset')
  CALL AddToObjFld('Name', base + hwpNameOff,' HW Temp Manager')
  CALL AddToObjStr('Control Variable','Temperature')
  CALL AddToObjFld('Set point at outside low {C}', base + hwpHWSetPtAtODBLoOff,'')
  CALL AddToObjFld('Outside low {C}', base + hwpHWODBLoOff,'')
  CALL AddToObjFld('Set point at outside high {C}', base + hwpHWSetPtAtODBHiOff,'')
  CALL AddToObjFld('Outside high {C}', base + hwpHWODBHiOff,'')
  CALL AddToObjFld('Setpoint Node or NodeList Name', base + hwpNameOff,' HW Supply Setpoint Nodes',.TRUE.)
END IF
!BRANCH LIST ~ line 56
CALL CreateNewObj('BranchList')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Side Branches')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Inlet Branch')
!Repeat this line for every boiler, use list created in boiler objects, also save lastSupplyBranch counter for use in later lists
DO iStr = 1,numStrList
  IF(StrList(iStr)%Handle .EQ. handleHWSupplyBranch) THEN
    CALL AddToObjStr('Branch Name', StrList(iStr)%StrToSave)
    lastSupplyBranch = iStr
  END IF
END DO
IF (isSupplyBypassYes) THEN
  CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Bypass Branch')
END IF
CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Outlet Branch',.TRUE.)
!CONNECTOR LIST ~ line 63
CALL CreateNewObj('ConnectorList')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Side Connectors')
CALL AddToObjStr('Connector Object Type','Connector:Splitter')
CALL AddToObjFld('Connector Name', base + hwpNameOff,' HW Supply Splitter')
CALL AddToObjStr('Connector Object Type','Connector:Mixer')
CALL AddToObjFld('Connector Name', base + hwpNameOff,' HW Supply Mixer',.TRUE.)
!SPLITTER ~ line 70
CALL CreateNewObj('Connector:Splitter')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Splitter')
CALL AddToObjFld('Inlet Branch Name', base + hwpNameOff,' HW Supply Inlet Branch')
!Repeat this line for every boiler, use list created in boiler objects
DO iStr = 1,lastSupplyBranch
  IF(StrList(iStr)%Handle .EQ. handleHWSupplyBranch) THEN
    isEndOfList = ((iStr == lastSupplyBranch) .AND. (.NOT. isSupplyBypassYes))
    CALL AddToObjStr('Outlet Branch Name', StrList(iStr)%StrToSave,isEndOfList)
  END IF
END DO
IF (isSupplyBypassYes) THEN
  CALL AddToObjFld('Outlet Branch Name', base + hwpNameOff,' HW Supply Bypass Branch',.TRUE.)
END IF
!MIXER ~ line 76
CALL CreateNewObj('Connector:Mixer')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Mixer')
CALL AddToObjFld('Outlet Branch Name', base + hwpNameOff,' HW Supply Outlet Branch')
!Repeat this line for every boiler, use list created in boiler objects
DO iStr = 1,lastSupplyBranch
  IF(StrList(iStr)%Handle .EQ. handleHWSupplyBranch) THEN
    isEndOfList = ((iStr == lastSupplyBranch) .AND. (.NOT. isSupplyBypassYes))
    CALL AddToObjStr('Outlet Branch Name', StrList(iStr)%StrToSave,isEndOfList)
  END IF
END DO
IF (isSupplyBypassYes) THEN
  CALL AddToObjFld('Inlet Branch Name', base + hwpNameOff,' HW Supply Bypass Branch',.TRUE.)
END IF

! This is out of order since pump logic always requires this branch and pipe
IF (isSupplyBypassYes) THEN
!BRANCH ~ line 104
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Bypass Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
  CALL AddToObjFld('Component Name', base + hwpNameOff,' HW Supply Side Bypass Pipe')
  CALL AddToObjFld('Component Inlet Node Name', base + hwpNameOff,' HW Supply Bypass Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + hwpNameOff,' HW Supply Bypass Outlet',.TRUE.)
  !PIPE ~line 113
  CALL CreateNewObj('Pipe:Adiabatic')
  CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Side Bypass Pipe')
  CALL AddToObjFld('Inlet Node Name', base + hwpNameOff,' HW Supply Bypass Inlet')
  CALL AddToObjFld('Outlet Node Name', base + hwpNameOff,' HW Supply Bypass Outlet',.TRUE.)
END IF

IF (isPumpBranch) THEN
  !if branch pumps are used just insert a pipe in the supply inlet branch
  CALL CreateNewObj('Pipe:Adiabatic')
  CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Side Inlet Pipe')
  CALL AddToObjFld('Inlet Node Name', base + hwpNameOff,' HW Supply Inlet')
  CALL AddToObjFld('Outlet Node Name', base + hwpNameOff,' HW Supply Pipe Outlet',.TRUE.)
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Inlet Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
  CALL AddToObjFld('Component Name', base + hwpNameOff,' HW Supply Side Inlet Pipe')
  CALL AddToObjFld('Component Inlet Node Name', base + hwpNameOff,' HW Supply Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + hwpNameOff,' HW Supply Pipe Outlet',.TRUE.)
ELSE
  ! normal pump arrangement with the pump on the supply inlet branch
  !PUMP:CONSTANT SPEED ~ line 82
  IF (PumpConfig .EQ. pcConstant) THEN
    IF (.NOT. isPumpHeader) THEN
      CALL CreateNewObj('Pump:ConstantSpeed')
      CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Pump')
      CALL AddToObjFld('Inlet Node Name', base + hwpNameOff,' HW Supply Inlet')
      CALL AddToObjFld('Outlet Node Name', base + hwpNameOff,' HW Pump Outlet')
      CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Rated Pump Head {Pa}', base + hwpHWPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      CALL AddToObjFld('Pump Control Type', base + hwpPumpControlTypeOff,'')
      IF (ishwpPumpSchedNameBlank) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', base + hwpPumpSchedNameOff,'',.TRUE.)
      END IF
    ELSE
      CALL CreateNewObj('HeaderedPumps:ConstantSpeed')
      CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Pump')
      CALL AddToObjFld('Inlet Node Name', base + hwpNameOff,' HW Supply Inlet')
      CALL AddToObjFld('Outlet Node Name', base + hwpNameOff,' HW Pump Outlet')
      CALL AddToObjStr('Total Rated Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Number of Pumps in Bank',pumpCount)
      CALL AddToObjStr('Flow Sequencing Control Scheme','Sequential')
      CALL AddToObjFld('Rated Pump Head {Pa}', base + hwpHWPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      CALL AddToObjFld('Pump Control Type', base + hwpPumpControlTypeOff,'')
      IF (ishwpPumpSchedNameBlank) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', base + hwpPumpSchedNameOff,'',.TRUE.)
      END IF
    END IF
    !BRANCH ~line 95
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Inlet Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isPumpHeader) THEN
      CALL AddToObjStr('Component Object Type','Pump:ConstantSpeed')
    ELSE
      CALL AddToObjStr('Component Object Type','HeaderedPumps:ConstantSpeed')
    END IF
    CALL AddToObjFld('Component Name', base + hwpNameOff,' HW Supply Pump')
    CALL AddToObjFld('Component Inlet Node Name', base + hwpNameOff,' HW Supply Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + hwpNameOff,' HW Pump Outlet',.TRUE.)
  ELSE
    !PUMP:VARIABLE SPEED ~ line 118
    IF (PumpConfig .EQ. pcVariable) THEN
      CALL WriteComment('Pump part load coefficients based on simple parabola to represent a generic VSD drive.')
    END IF
    IF (.NOT. isPumpHeader) THEN
      CALL CreateNewObj('Pump:VariableSpeed')
      CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Pump')
      CALL AddToObjFld('Inlet Node Name', base + hwpNameOff,' HW Supply Inlet')
      CALL AddToObjFld('Outlet Node Name', base + hwpNameOff,' HW Pump Outlet')
      CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Rated Pump Head {Pa}', base + hwpHWPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      !Curve based on VSD generic curve from user with adjustments to prevent neg values - Nov 2008
      !Can use something better when min flow rate is able to be autosized or entered as a fraction
      CALL AddToObjStr('Coefficient 1 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 2 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 3 of the Part Load Performance Curve','1.0')
      CALL AddToObjStr('Coefficient 4 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Min Flow Rate while operating in variable flow capacity {m3/s}','0')
      CALL AddToObjFld('Pump Control Type', base + hwpPumpControlTypeOff,'')
      IF (ishwpPumpSchedNameBlank) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', base + hwpPumpSchedNameOff,'',.TRUE.)
      END IF
    ELSE
      CALL CreateNewObj('HeaderedPumps:VariableSpeed')
      CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Pump')
      CALL AddToObjFld('Inlet Node Name', base + hwpNameOff,' HW Supply Inlet')
      CALL AddToObjFld('Outlet Node Name', base + hwpNameOff,' HW Pump Outlet')
      CALL AddToObjStr('Total Rated Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Number of Pumps in Bank',pumpCount)
      CALL AddToObjStr('Flow Sequencing Control Scheme','Sequential')
      CALL AddToObjFld('Rated Pump Head {Pa}', base + hwpHWPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      IF (PumpConfig .EQ. pcVariable) THEN
        !Curve based on VSD generic curve from user with adjustments to prevent neg values - Nov 2008
        !Can use something better when min flow rate is able to be autosized or entered as a fraction
        CALL AddToObjStr('Coefficient 1 of the Part Load Performance Curve','0.0')
        CALL AddToObjStr('Coefficient 2 of the Part Load Performance Curve','0.0')
        CALL AddToObjStr('Coefficient 3 of the Part Load Performance Curve','1.0')
        CALL AddToObjStr('Coefficient 4 of the Part Load Performance Curve','0.0')
      ELSE
        ! MJW Oct 2011 - I think this is currently dead code which was in place for the stepped pump option which is disabled
        CALL AddToObjStr('Coefficient 1 of the Part Load Performance Curve','0')
        CALL AddToObjStr('Coefficient 2 of the Part Load Performance Curve','1')
        CALL AddToObjStr('Coefficient 3 of the Part Load Performance Curve','0')
        CALL AddToObjStr('Coefficient 4 of the Part Load Performance Curve','0')
      END IF
      CALL AddToObjStr('Min Flow Rate while operating in variable flow capacity {m3/s}','0')
      CALL AddToObjFld('Pump Control Type', base + hwpPumpControlTypeOff,'')
      IF (ishwpPumpSchedNameBlank) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', base + hwpPumpSchedNameOff,'',.TRUE.)
      END IF
    END IF
    !BRANCH ~line 140
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Inlet Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isPumpHeader) THEN
      CALL AddToObjStr('Component Object Type','Pump:VariableSpeed')
    ELSE
      CALL AddToObjStr('Component Object Type','HeaderedPumps:VariableSpeed')
    END IF
    CALL AddToObjFld('Component Name', base + hwpNameOff,' HW Supply Pump')
    CALL AddToObjFld('Component Inlet Node Name', base + hwpNameOff,' HW Supply Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + hwpNameOff,' HW Pump Outlet',.TRUE.)
  END IF
END IF
!BRANCH ~ line 149
CALL CreateNewObj('Branch')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Outlet Branch')
CALL AddToObjStr('Pressure Drop Curve Name','')
CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
CALL AddToObjFld('Component Name', base + hwpNameOff,' HW Supply Outlet Pipe')
CALL AddToObjFld('Component Inlet Node Name', base + hwpNameOff,' HW Supply Outlet Pipe Inlet')
CALL AddToObjFld('Component Outlet Node Name', base + hwpNameOff,' HW Supply Outlet',.TRUE.)
!PIPE ~ line 158
CALL CreateNewObj('Pipe:Adiabatic')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Supply Outlet Pipe')
CALL AddToObjFld('Inlet Node Name', base + hwpNameOff,' HW Supply Outlet Pipe Inlet')
CALL AddToObjFld('Outlet Node Name', base + hwpNameOff,' HW Supply Outlet',.TRUE.)
!BRANCH LIST ~ line 163
CALL CreateNewObj('BranchList')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Demand Side Branches')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Demand Inlet Branch')
!The names of all of the branches for the heating coils have been previously gathered
DO iStr = 1, numStrList
  IF (StrList(iStr)%Handle .EQ. handleHeatingCoilBranch) THEN
    CALL AddToObjStr('Name', StrList(iStr)%StrToSave)
  END IF
END DO
IF (isDemandBypassYes) THEN
  CALL AddToObjFld('Name', base + hwpNameOff,' HW Demand Bypass Branch')
END IF
CALL AddToObjFld('Name', base + hwpNameOff,' HW Demand Outlet Branch',.TRUE.)
!CONNECTOR LIST ~ line 170
CALL CreateNewObj('ConnectorList')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Demand Side Connectors')
CALL AddToObjStr('Connector Object Type','Connector:Splitter')
CALL AddToObjFld('Connector Name', base + hwpNameOff,' HW Demand Splitter')
CALL AddToObjStr('Connector Object Type','Connector:Mixer')
CALL AddToObjFld('Connector Name', base + hwpNameOff,' HW Demand Mixer',.TRUE.)
!SPLITTER ~ line 177
CALL CreateNewObj('Connector:Splitter')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Demand Splitter')
CALL AddToObjFld('Inlet Branch Name', base + hwpNameOff,' HW Demand Inlet Branch')
IF (isDemandBypassYes) THEN
  CALL AddToObjFld('Outlet Branch Name', base + hwpNameOff,' HW Demand Bypass Branch')
END IF
!determine the last branch since it will require a semicolon after the object
DO iStr = 1, numStrList
  IF (StrList(iStr)%Handle .EQ. handleHeatingCoilBranch) THEN
    lastBranch = iStr
  END IF
END DO
!The names of all of the branches for the heating coils have been previously gathered
DO iStr = 1, numStrList
  IF (StrList(iStr)%Handle .EQ. handleHeatingCoilBranch) THEN
    CALL AddToObjStr('Name', StrList(iStr)%StrToSave,iStr .EQ. lastBranch)
  END IF
END DO
!MIXER ~ 183
CALL CreateNewObj('Connector:Mixer')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Demand Mixer')
CALL AddToObjFld('Outlet Branch Name', base + hwpNameOff,' HW Demand Outlet Branch')
IF (isDemandBypassYes) THEN
  CALL AddToObjFld('Inlet Branch Name', base + hwpNameOff,' HW Demand Bypass Branch')
END IF
!The names of all of the branches for the heating coils have been previously gathered
DO iStr = 1, numStrList
  IF (StrList(iStr)%Handle .EQ. handleHeatingCoilBranch) THEN
    CALL AddToObjStr('Name', StrList(iStr)%StrToSave,iStr .EQ. lastBranch)
  END IF
END DO
!BRANCH ~ line 189
CALL CreateNewObj('Branch')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Demand Inlet Branch')
CALL AddToObjStr('Pressure Drop Curve Name','')
CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
CALL AddToObjFld('Component Name', base + hwpNameOff,' HW Demand Inlet Pipe')
CALL AddToObjFld('Component Inlet Node Name', base + hwpNameOff,' HW Demand Inlet')
CALL AddToObjFld('Component Outlet Node Name', base + hwpNameOff,' HW Demand Inlet Pipe Outlet',.TRUE.)
!PIPE ~ line 198
CALL CreateNewObj('Pipe:Adiabatic')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Demand Inlet Pipe')
CALL AddToObjFld('Inlet Node Name', base + hwpNameOff,' HW Demand Inlet')
CALL AddToObjFld('Outlet Node Name', base + hwpNameOff,' HW Demand Inlet Pipe Outlet',.TRUE.)
!line numbers shifted after this point
!BRANCH ~ line 189
IF (isDemandBypassYes) THEN
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + hwpNameOff,' HW Demand Bypass Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
  CALL AddToObjFld('Component Name', base + hwpNameOff,' HW Demand Side Bypass Pipe')
  CALL AddToObjFld('Component Inlet Node Name', base + hwpNameOff,' HW Demand Bypass Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + hwpNameOff,' HW Demand Bypass Outlet',.TRUE.)
  !PIPE ~ line 198
  CALL CreateNewObj('Pipe:Adiabatic')
  CALL AddToObjFld('Name', base + hwpNameOff,' HW Demand Side Bypass Pipe')
  CALL AddToObjFld('Inlet Node Name', base + hwpNameOff,' HW Demand Bypass Inlet')
  CALL AddToObjFld('Outlet Node Name', base + hwpNameOff,' HW Demand Bypass Outlet',.TRUE.)
END IF
!BRANCH ~ line 203
CALL CreateNewObj('Branch')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Demand Outlet Branch')
CALL AddToObjStr('Pressure Drop Curve Name','')
CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
CALL AddToObjFld('Component Name', base + hwpNameOff,' HW Demand Outlet Pipe')
CALL AddToObjFld('Component Inlet Node Name', base + hwpNameOff,' HW Demand Outlet Pipe Inlet')
CALL AddToObjFld('Component Outlet Node Name', base + hwpNameOff,' HW Demand Outlet',.TRUE.)
!PIPE ~ line 212
CALL CreateNewObj('Pipe:Adiabatic')
CALL AddToObjFld('Name', base + hwpNameOff,' HW Demand Outlet Pipe')
CALL AddToObjFld('Inlet Node Name', base + hwpNameOff,' HW Demand Outlet Pipe Inlet')
CALL AddToObjFld('Outlet Node Name', base + hwpNameOff,' HW Demand Outlet',.TRUE.)
! if the user has specified a Maximum Outdoor Dry Bulb Temperatures
! then add an AvailabilityManagerAssignmentList
!             AvailabilityManager:HighTemperatureTurnOff
!             OutdoorAir:Node
IF (.NOT. isMaxOutTempBlank) THEN
  CALL CreateNewObj('AvailabilityManagerAssignmentList')
  CALL AddToObjFld('Name', base + hwpNameOff,' Availability List')
  CALL AddToObjStr('Availability Manager Object Type','AvailabilityManager:HighTemperatureTurnOff')
  CALL AddToObjFld('Availability Manager Name',base + hwpNameOff,' Availability High Temp Off',.TRUE.)
  CALL CreateNewObj('AvailabilityManager:HighTemperatureTurnOff')
  CALL AddToObjFld('Name',base + hwpNameOff,' Availability High Temp Off')
  CALL AddToObjFld('Sensor Node Name',base + hwpNameOff,' Outside Air Sensor')
  CALL AddToObjFld('Temperature',base + hwpMaxTOutOff,'',.TRUE.)
  CALL CreateNewObj('OutdoorAir:Node')
  CALL AddToObjFld('Name', base + hwpNameOff,' Outside Air Sensor')
  CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
END IF
! Create glycol fluid if not water or steam
IF ((fluidType .NE. ftWater) .AND. (fluidType .NE. ftSteam)) THEN
  CALL CreateNewObj('FluidProperties:GlycolConcentration')
  CALL AddToObjFld('Name', base + hwpNameOff,' Loop Fluid')
  SELECT CASE (fluidType)
    CASE (ftEthyleneGlycol30)
      CALL AddToObjStr('Glycol Type','EthyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.30',.TRUE.)
    CASE (ftEthyleneGlycol40)
      CALL AddToObjStr('Glycol Type','EthyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.40',.TRUE.)
    CASE (ftEthyleneGlycol50)
      CALL AddToObjStr('Glycol Type','EthyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.50',.TRUE.)
    CASE (ftEthyleneGlycol60)
      CALL AddToObjStr('Glycol Type','EthyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.60',.TRUE.)
    CASE (ftPropyleneGlycol30)
      CALL AddToObjStr('Glycol Type','PropyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.30',.TRUE.)
    CASE (ftPropyleneGlycol40)
      CALL AddToObjStr('Glycol Type','PropyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.40',.TRUE.)
    CASE (ftPropyleneGlycol50)
      CALL AddToObjStr('Glycol Type','PropyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.50',.TRUE.)
    CASE (ftPropyleneGlycol60)
      CALL AddToObjStr('Glycol Type','PropyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.60',.TRUE.)
    CASE DEFAULT
      CALL AddToObjStr('Glycol Type','EthyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.30',.TRUE.)
  END SELECT
END IF
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE CreateChiller
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a compact chiller object

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for chiller type
INTEGER, PARAMETER ::chtPurchased     = 1
INTEGER, PARAMETER ::chtCentrifugal   = 2
INTEGER, PARAMETER ::chtReciprocating = 3
INTEGER, PARAMETER ::chtScrew         = 4
          ! parameters for condenser type
INTEGER, PARAMETER ::condWater = 1
INTEGER, PARAMETER ::condAir   = 2
INTEGER, PARAMETER ::condEvap  = 3
          ! parameters for pump type
INTEGER, PARAMETER :: pumpVariable = 1
INTEGER, PARAMETER :: pumpConstant = 2
INTEGER, PARAMETER :: pumpConstPriVarSec = 3
INTEGER, PARAMETER :: pumpVarPriConstSec = 4
          ! parameters for detailed chiller object type
INTEGER, PARAMETER :: chotElecEIR = 1
INTEGER, PARAMETER :: chotElecRefEIR = 2
          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base=0  ! used as base for chiller
INTEGER :: detchlrBase=0  ! used as base for detailed chiller
INTEGER :: iChiller=0
INTEGER :: iDetChlr=0
LOGICAL :: isCapacityAutosize
INTEGER :: chlrType=0
INTEGER :: condType=0
INTEGER :: detchlrObjType=0
INTEGER :: loopPumpConfig=0
LOGICAL :: isBranchPriChwPump
LOGICAL :: isPriChwPumpVariable
CHARACTER(len=MaxAlphaLength) :: chwInletNode=''
CHARACTER(len=MaxAlphaLength) :: chwOutletNode=''
CHARACTER(len=MaxAlphaLength) :: cndwInletNode=''
CHARACTER(len=MaxAlphaLength) :: cndwOutletNode=''
CHARACTER(len=MaxAlphaLength) :: condensertype=''

! HVACTemplate:Plant:Chiller
IF ((numCompactChiller + numCompactChillerOR) .EQ. 0) RETURN
IF ((SameString(FldVal(compactChldLoopBase + cwpChWPumpConfigOff),'VariablePrimaryNoSecondary')) .OR. &
    (SameString(FldVal(compactChldLoopBase + cwpChWPumpConfigOff),'VariableFlow'))) THEN
  loopPumpConfig = pumpVariable
  isPriChwPumpVariable = .TRUE.
ELSEIF ((SameString(FldVal(compactChldLoopBase + cwpChWPumpConfigOff),'ConstantPrimaryNoSecondary')) .OR. &
        (SameString(FldVal(compactChldLoopBase + cwpChWPumpConfigOff),'ConstantFlow'))) THEN
  loopPumpConfig = pumpConstant
  isPriChwPumpVariable = .FALSE.
ELSEIF (SameString(FldVal(compactChldLoopBase + cwpChWPumpConfigOff),'ConstantPrimaryVariableSecondary')) THEN
  loopPumpConfig = pumpConstPriVarSec
  isPriChwPumpVariable = .FALSE.
ELSEIF (SameString(FldVal(compactChldLoopBase + cwpChWPumpConfigOff),'VariablePrimaryConstantSecondary')) THEN
  !  VariablePrimaryConstantSecondary is currently unsupported, in HVACTemplate:Plant:ChilledWaterLoop warning
  !  is issued and this is changed to VariablePrimaryNoSecondary
  !  loopPumpConfig = pumpVarPriConstSec
  loopPumpConfig = pumpVariable
  isPriChwPumpVariable = .TRUE.
ELSE
  loopPumpConfig = pumpConstant
  isPriChwPumpVariable = .FALSE.
END IF
isBranchPriChwPump = SameString(FldVal(compactChldLoopBase + cwpPriChWPumpTypeOff),'PumpPerChiller')
DO iChiller = 1, numCompactChiller
  base =  compactChillerBase(iChiller)
  ! set flags
  IF (SameString(FldVal(base + chlTypeOff),'DistrictChilledWater')) THEN
    chlrType = chtPurchased
  ELSEIF (SameString(FldVal(base + chlTypeOff),'ElectricCentrifugalChiller')) THEN
    chlrType = chtCentrifugal
  ELSEIF (SameString(FldVal(base + chlTypeOff),'ElectricReciprocatingChiller')) THEN
    chlrType = chtReciprocating
  ELSEIF (SameString(FldVal(base + chlTypeOff),'ElectricScrewChiller')) THEN
    chlrType = chtScrew
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Plant:Chiller in the Chiller Type field')
  END IF
  IF (SameString(FldVal(base + chlCondTypeOff),'WaterCooled')) THEN
    condType = condWater
  ELSEIF (SameString(FldVal(base + chlCondTypeOff),'AirCooled')) THEN
    condType = condAir
  ELSEIF (SameString(FldVal(base + chlCondTypeOff),'EvaporativelyCooled')) THEN
    condType = condEvap
  ELSEIF (SameString(FldVal(base + chlCondTypeOff),'')) THEN
    condType = condWater
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Plant:Chiller in the Condenser Type field')
  END IF
  ! add ChW Outlet Node to list of node names for ChW setpoints
  CALL AddToStrList(TRIM(FldVal(base + chlNameOff)) // ' ChW Outlet',handleChillerChWOutletNode)

  !for district cooling ignore the condenser type and assume air cooled
  IF (chlrType .EQ. chtPurchased) condType = condAir
  isCapacityAutosize = SameString(FldVal(base + chlCapacityOff),'autosize')
  ! write idf objects
  IF (chlrType .EQ. chtPurchased) THEN
    !PURCHASED:CHILLED WATER ~line 4
    CALL CreateNewObj('DistrictCooling')
    CALL AddToObjFld('Name', base + chlNameOff,'')
    CALL AddToObjFld('Plant_Loop_Inlet_Node', base + chlNameOff,' ChW Inlet')
    CALL AddToObjFld('Plant_Loop_Outlet_Node', base + chlNameOff,' ChW Outlet')
    IF (isCapacityAutosize) THEN
      CALL AddToObjStr('Nominal Capacity {W}','1000000000000000',.TRUE.)
    ELSE
      CALL AddToObjFld('Nominal Capacity {W}', base + chlCapacityOff,'',.TRUE.)
    END IF
    !BRANCH ~ line 11
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + chlNameOff,' ChW Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isBranchPriChwPump) THEN
      CALL AddToObjStr('Component Object Type','DistrictCooling')
      CALL AddToObjFld('Component Name', base + chlNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + chlNameOff,' ChW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + chlNameOff,' ChW Outlet',.TRUE.)
    ELSE
      ! using a branch pump define the pump first followed by the chiller/district cooling
      IF (isPriChwPumpVariable) THEN
        CALL AddToObjStr('Component Object Type','Pump:VariableSpeed')
      ELSE
        CALL AddToObjStr('Component Object Type','Pump:ConstantSpeed')
      END IF
      CALL AddToObjFld('Component Name', base + chlNameOff,' ChW Branch Pump')
      CALL AddToObjFld('Component Inlet Node Name', base + chlNameOff,' ChW Pump Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + chlNameOff,' ChW Inlet')
      CALL AddToObjStr('Component Object Type','DistrictCooling')
      CALL AddToObjFld('Component Name', base + chlNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + chlNameOff,' ChW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + chlNameOff,' ChW Outlet',.TRUE.)
    ENDIF
  ELSEIF  (chlrType .EQ. chtCentrifugal) THEN
    !CHILLER:ELECTRIC:EIR  ~line 20
    CALL CreateNewObj('Chiller:Electric:EIR')
    CALL AddToObjFld('Name', base + chlNameOff,'')
    CALL AddToObjFld('Reference Capacity {W}', base + chlCapacityOff,'')
    CALL AddToObjFld('Reference COP {W/W}', base + chlCOPOff,'')
    CALL AddToObjStr('Reference Leaving Chilled Water Temperature {C}','6.67')
    CALL AddToObjStr('Reference Entering Condenser Fluid Temperature {C}','29.4')
    CALL AddToObjStr('Reference Evaporator Volumetric Water Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Reference Condenser Volumetric Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Cooling Capacity Function of Temperature Curve Name', base + chlNameOff,' CentCapFT')
    CALL AddToObjFld('Electric Input to Cooling Output Ratio Function of Temperature Curve Name', base + chlNameOff,' CentEIRFT')
    CALL AddToObjFld('Electric Input to Cooling Output Ratio Function of Part Load Ratio Curve Name', base + chlNameOff, &
     ' CentEIRFPLR')
    CALL AddToObjFld('Minimum Part Load Ratio',base + chlMinPLROff,'')
    CALL AddToObjFld('Maximum Part Load Ratio',base + chlMaxPLROff,'')
    CALL AddToObjFld('Optimum Part Load Ratio',base + chlOptPLROff,'')
    CALL AddToObjFld('Minimum Unloading Ratio',base + chlMinUnloadRatOff,'')
    CALL AddToObjFld('Chilled Water Inlet Node Name', base + chlNameOff,' ChW Inlet')
    CALL AddToObjFld('Chilled Water Outlet Node Name', base + chlNameOff,' ChW Outlet')
    CALL AddToObjFld('Condenser Inlet Node Name', base + chlNameOff,' Cnd Inlet')
    CALL AddToObjFld('Condenser Outlet Node Name', base + chlNameOff,' Cnd Outlet')
    CALL AddToObjFld('Condenser Type', base + chlCondTypeOff,'')
    IF (condType .EQ. condWater) THEN
      CALL AddToObjStr('Condenser Fan Power Ratio {W/W}','0')
      CALL AddToObjStr('Compressor Motor Efficiency','1')
    ELSE
      CALL AddToObjStr('Condenser Fan Power Ratio {W/W}','0.035')
      CALL AddToObjStr('Compressor Motor Efficiency','1')
    END IF
    CALL AddToObjFld('Leaving Chilled Water Lower Temperature Limit {C}',base + chlLeavingLowLimOff,'')
    IF (isPriChwPumpVariable) THEN
      CALL AddToObjStr('Chiller Flow Mode','LeavingSetpointModulated')
    ELSE
      CALL AddToObjStr('Chiller Flow Mode','ConstantFlow')
    END IF
    CALL AddToObjStr('Design Heat Recovery Water Flow Rate {m3/s}','0')
    CALL AddToObjStr('Heat Recovery Inlet Node Name','')
    CALL AddToObjStr('Heat Recovery Outlet Node Name','')
    CALL AddToObjFld('Sizing Factor', base + chlSizeFactorOff,'',.TRUE.)

    !BRANCH ~line 50
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + chlNameOff,' ChW Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isBranchPriChwPump) THEN
      CALL AddToObjStr('Component Object Type','Chiller:Electric:EIR')
      CALL AddToObjFld('Component Name', base + chlNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + chlNameOff,' ChW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + chlNameOff,' ChW Outlet',.TRUE.)
    ELSE
      ! using a branch pump define the pump first followed by the chiller/district cooling
      IF (isPriChwPumpVariable) THEN
        CALL AddToObjStr('Component Object Type','Pump:VariableSpeed')
      ELSE
        CALL AddToObjStr('Component Object Type','Pump:ConstantSpeed')
      END IF
      CALL AddToObjFld('Component Name', base + chlNameOff,' ChW Branch Pump')
      CALL AddToObjFld('Component Inlet Node Name', base + chlNameOff,' ChW Pump Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + chlNameOff,' ChW Inlet')
      CALL AddToObjStr('Component Object Type','Chiller:Electric:EIR')
      CALL AddToObjFld('Component Name', base + chlNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + chlNameOff,' ChW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + chlNameOff,' ChW Outlet',.TRUE.)
    ENDIF
    IF (condType .EQ. condWater) THEN
      !BRANCH ~ line 59
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + chlNameOff,' CndW Branch')
      CALL AddToStrList(TRIM(FldVal(base + chlNameOff)) // ' CndW Branch',handleCondenserBranch)
      isChillerWithWaterCooled = .TRUE.
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','Chiller:Electric:EIR')
      CALL AddToObjFld('Component Name', base + chlNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + chlNameOff,' Cnd Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + chlNameOff,' Cnd Outlet',.TRUE.)
    END IF
    !CURVE:BIQUADRATIC ~ line 68
    CALL CreateNewObj('Curve:Biquadratic')
    CALL WriteComment('Cooling Capacity Function of Temperature Curve for open or hermetic water-cooled centrifugal chillers')
    CALL WriteComment('Same as DOE-2.1E HERM-CENT-CAP-FT (CCAPT3) and OPEN-CENT-CAP-FT (CCAPT1)')
    CALL AddToObjFld('Name', base + chlNameOff,' CentCapFT')
    CALL AddToObjStr('Coefficient1 Constant','0.257896')
    CALL AddToObjStr('Coefficient2 x','0.0389016')
    CALL AddToObjStr('Coefficient3 x**2','-0.00021708')
    CALL AddToObjStr('Coefficient4 y','0.0468684')
    CALL AddToObjStr('Coefficient5 y**2','-0.00094284')
    CALL AddToObjStr('Coefficient6 x*y','-0.00034344')
    CALL AddToObjStr('Minimum Value of x','5')
    CALL AddToObjStr('Maximum Value of x','10')
    CALL AddToObjStr('Minimum Value of y','24')
    CALL AddToObjStr('Maximum Value of y','35',.TRUE.)
    !CURVE:BIQUADRATIC ~ line 81
    CALL CreateNewObj('Curve:Biquadratic')
    CALL WriteComment('Energy Input to Cooling Output Ratio Function of Temperature Curve for open '//  &
       'or hermetic water-cooled centrifugal chillers')
    CALL WriteComment('Same as DOE-2.1E HERM-CENT-EIR-FT (EIRT3) and OPEN-CENT-EIR-FT (EIRT1)')
    CALL AddToObjFld('Name', base + chlNameOff,' CentEIRFT')
    CALL AddToObjStr('Coefficient1 Constant','0.933884')
    CALL AddToObjStr('Coefficient2 x','-0.058212')
    CALL AddToObjStr('Coefficient3 x**2','0.00450036')
    CALL AddToObjStr('Coefficient4 y','0.00243')
    CALL AddToObjStr('Coefficient5 y**2','0.000486')
    CALL AddToObjStr('Coefficient6 x*y','-0.001215')
    CALL AddToObjStr('Minimum Value of x','5')
    CALL AddToObjStr('Maximum Value of x','10')
    CALL AddToObjStr('Minimum Value of y','24')
    CALL AddToObjStr('Maximum Value of y','35',.TRUE.)
    !CURVE:QUADRATIC ~ line 94
    CALL CreateNewObj('Curve:Quadratic')
    CALL WriteComment('Energy Input to Cooling Output Ratio Function of Part Load Ratio Curve '//  &
       'for open or hermetic water-cooled centrifugal chillers')
    CALL WriteComment('Same as DOE-2.1E HERM-CENT-EIR-FPLR (EIRPLR3) and OPEN-CENT-EIR-FPLR (EIRPLR1)')
    CALL AddToObjFld('Name', base + chlNameOff,' CentEIRFPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.222903')
    CALL AddToObjStr('Coefficient2 x','0.313387')
    CALL AddToObjStr('Coefficient3 x**2','0.46371')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  ELSEIF  (chlrType .EQ. chtReciprocating) THEN
    !CHILLER:ELECTRIC:EIR ~ line 102
    CALL CreateNewObj('Chiller:Electric:EIR')
    CALL AddToObjFld('Name', base + chlNameOff,'')
    CALL AddToObjFld('Reference Capacity {W}', base + chlCapacityOff,'')
    CALL AddToObjFld('Reference COP {W/W}', base + chlCOPOff,'')
    CALL AddToObjStr('Reference Leaving Chilled Water Temperature {C}','6.67')
    CALL AddToObjStr('Reference Entering Condenser Fluid Temperature {C}','29.4')
    CALL AddToObjStr('Reference Evaporator Volumetric Water Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Reference Condenser Volumetric Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Cooling Capacity Function of Temperature Curve Name', base + chlNameOff,' RecipCapFT')
    CALL AddToObjFld('Electric Input to Cooling Output Ratio Function of Temperature Curve Name', base + chlNameOff,' RecipEIRFT')
    CALL AddToObjFld('Electric Input to Cooling Output Ratio Function of Part Load Ratio Curve Name', base + chlNameOff, &
     ' RecipEIRFPLR')
    CALL AddToObjFld('Minimum Part Load Ratio',base + chlMinPLROff,'')
    CALL AddToObjFld('Maximum Part Load Ratio',base + chlMaxPLROff,'')
    CALL AddToObjFld('Optimum Part Load Ratio',base + chlOptPLROff,'')
    CALL AddToObjFld('Minimum Unloading Ratio',base + chlMinUnloadRatOff,'')
    !next four lines fixes from 3/25/2005
    CALL AddToObjFld('Chilled Water Inlet Node Name', base + chlNameOff,' ChW Inlet')
    CALL AddToObjFld('Chilled Water Outlet Node Name', base + chlNameOff,' ChW Outlet')
    CALL AddToObjFld('Condenser Inlet Node Name', base + chlNameOff,' Cnd Inlet')
    CALL AddToObjFld('Condenser Outlet Node Name', base + chlNameOff,' Cnd Outlet')
    CALL AddToObjFld('Condenser Type', base + chlCondTypeOff,'')
    IF (condType .EQ. condWater) THEN
      CALL AddToObjStr('Condenser Fan Power Ratio {W/W}','')
      CALL AddToObjStr('Compressor Motor Efficiency','1')
    ELSE
      CALL AddToObjStr('Condenser Fan Power Ratio {W/W}','0.035')
      CALL AddToObjStr('Compressor Motor Efficiency','1')
    END IF
    CALL AddToObjFld('Leaving Chilled Water Lower Temperature Limit {C}',base + chlLeavingLowLimOff,'')
    IF (isPriChwPumpVariable) THEN
      CALL AddToObjStr('Chiller Flow Mode','LeavingSetpointModulated')
    ELSE
      CALL AddToObjStr('Chiller Flow Mode','ConstantFlow')
    END IF
    CALL AddToObjStr('Design Heat Recovery Water Flow Rate {m3/s}','0')
    CALL AddToObjStr('Heat Recovery Inlet Node Name','')
    CALL AddToObjStr('Heat Recovery Outlet Node Name','')
    CALL AddToObjFld('Sizing Factor', base + chlSizeFactorOff,'',.TRUE.)

    !BRANCH ~ line 132
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + chlNameOff,' ChW Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isBranchPriChwPump) THEN
      CALL AddToObjStr('Component Object Type','Chiller:Electric:EIR')
      CALL AddToObjFld('Component Name', base + chlNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + chlNameOff,' ChW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + chlNameOff,' ChW Outlet',.TRUE.)
    ELSE
      ! using a branch pump define the pump first followed by the chiller/district cooling
      IF (isPriChwPumpVariable) THEN
        CALL AddToObjStr('Component Object Type','Pump:VariableSpeed')
      ELSE
        CALL AddToObjStr('Component Object Type','Pump:ConstantSpeed')
      END IF
      CALL AddToObjFld('Component Name', base + chlNameOff,' ChW Branch Pump')
      CALL AddToObjFld('Component Inlet Node Name', base + chlNameOff,' ChW Pump Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + chlNameOff,' ChW Inlet')
      CALL AddToObjStr('Component Object Type','Chiller:Electric:EIR')
      CALL AddToObjFld('Component Name', base + chlNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + chlNameOff,' ChW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + chlNameOff,' ChW Outlet',.TRUE.)
    ENDIF
    IF (condType .EQ. condWater) THEN
      !BRANCH ~ line 141
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + chlNameOff,' CndW Branch')
      CALL AddToStrList(TRIM(FldVal(base + chlNameOff)) // ' CndW Branch',handleCondenserBranch)
      isChillerWithWaterCooled = .TRUE.
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','Chiller:Electric:EIR')
      CALL AddToObjFld('Component Name', base + chlNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + chlNameOff,' Cnd Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + chlNameOff,' Cnd Outlet',.TRUE.)
    END IF
    !CURVE:BIQUADRATIC ~ line 150
    CALL CreateNewObj('Curve:Biquadratic')
    CALL WriteComment('Cooling Capacity Function of Temperature Curve for open or hermetic water-cooled reciprocating chillers')
    CALL WriteComment('Same as DOE-2.1E HERM-REC-CAP-FT (CCAPT4) and OPEN-REC-CAP-FT (CCAPT2)')
    CALL AddToObjFld('Name', base + chlNameOff,' RecipCapFT')
    CALL AddToObjStr('Coefficient1 Constant','0.507883')
    CALL AddToObjStr('Coefficient2 x','0.145228')
    CALL AddToObjStr('Coefficient3 x**2','-0.00625644')
    CALL AddToObjStr('Coefficient4 y','-0.0011178')
    CALL AddToObjStr('Coefficient5 y**2','-0.0001296')
    CALL AddToObjStr('Coefficient6 x*y','-0.00028188')
    CALL AddToObjStr('Minimum Value of x','5')
    CALL AddToObjStr('Maximum Value of x','10')
    CALL AddToObjStr('Minimum Value of y','24')
    CALL AddToObjStr('Maximum Value of y','35',.TRUE.)
    !CURVE:BIQUADRATIC ~ line 163
    CALL CreateNewObj('Curve:Biquadratic')
    CALL WriteComment('Energy Input to Cooling Output Ratio Function of Temperature Curve '//  &
       'for open or hermetic water-cooled reciprocating chillers')
    CALL WriteComment('Same as DOE-2.1E HERM-REC-EIR-FT (EIRT4) and OPEN-REC-EIR-FT (EIRT2)')
    CALL AddToObjFld('Name', base + chlNameOff,' RecipEIRFT')
    CALL AddToObjStr('Coefficient1 Constant','1.03076')
    CALL AddToObjStr('Coefficient2 x','-0.103536')
    CALL AddToObjStr('Coefficient3 x**2','0.00710208')
    CALL AddToObjStr('Coefficient4 y','0.0093186')
    CALL AddToObjStr('Coefficient5 y**2','0.00031752')
    CALL AddToObjStr('Coefficient6 x*y','-0.00104328')
    CALL AddToObjStr('Minimum Value of x','5')
    CALL AddToObjStr('Maximum Value of x','10')
    CALL AddToObjStr('Minimum Value of y','24')
    CALL AddToObjStr('Maximum Value of y','35',.TRUE.)
    !CURVE:QUADRATIC ~ line 176
    CALL CreateNewObj('Curve:Quadratic')
    CALL WriteComment('Energy Input to Cooling Output Ratio Function of '//  &
       'Part Load Ratio Curve for open or hermetic water-cooled reciprocating chillers')
    CALL WriteComment('Same as DOE-2.1E HERM-REC-EIR-FPLR (EIRPLR4) and OPEN-REC-EIR-FPLR (EIRPLR2)')
    CALL AddToObjFld('Name', base + chlNameOff,' RecipEIRFPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.088065')
    CALL AddToObjStr('Coefficient2 x','1.137742')
    CALL AddToObjStr('Coefficient3 x**2','-0.225806')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  ELSEIF  (chlrType .EQ. chtScrew) THEN
    !CHILLER:ELECTRIC:EIR
    CALL CreateNewObj('Chiller:Electric:EIR')
    CALL AddToObjFld('Name', base + chlNameOff,'')
    CALL AddToObjFld('Reference Capacity {W}', base + chlCapacityOff,'')
    CALL AddToObjFld('Reference COP {W/W}', base + chlCOPOff,'')
    CALL AddToObjStr('Reference Leaving Chilled Water Temperature {C}','6.67')
    CALL AddToObjStr('Reference Entering Condenser Fluid Temperature {C}','29.4')
    CALL AddToObjStr('Reference Evaporator Volumetric Water Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Reference Condenser Volumetric Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Cooling Capacity Function of Temperature Curve Name', base + chlNameOff,' ScrewCapFT')
    CALL AddToObjFld('Electric Input to Cooling Output Ratio Function of Temperature Curve Name', base + chlNameOff,' ScrewEIRFT')
    CALL AddToObjFld('Electric Input to Cooling Output Ratio Function of Part Load Ratio Curve Name', base + chlNameOff, &
     ' ScrewEIRFPLR')
    CALL AddToObjFld('Minimum Part Load Ratio',base + chlMinPLROff,'')
    CALL AddToObjFld('Maximum Part Load Ratio',base + chlMaxPLROff,'')
    CALL AddToObjFld('Optimum Part Load Ratio',base + chlOptPLROff,'')
    CALL AddToObjFld('Minimum Unloading Ratio',base + chlMinUnloadRatOff,'')
    CALL AddToObjFld('Chilled Water Inlet Node Name', base + chlNameOff,' ChW Inlet')
    CALL AddToObjFld('Chilled Water Outlet Node Name', base + chlNameOff,' ChW Outlet')
    CALL AddToObjFld('Condenser Inlet Node Name', base + chlNameOff,' Cnd Inlet')
    CALL AddToObjFld('Condenser Outlet Node Name', base + chlNameOff,' Cnd Outlet')
    CALL AddToObjFld('Condenser Type', base + chlCondTypeOff,'')
    IF (condType .EQ. condWater) THEN
      CALL AddToObjStr('Condenser Fan Power Ratio {W/W}','0')
      CALL AddToObjStr('Compressor Motor Efficiency','1')
    ELSE
      CALL AddToObjStr('Condenser Fan Power Ratio {W/W}','0.035')
      CALL AddToObjStr('Compressor Motor Efficiency','1')
    END IF
    CALL AddToObjFld('Leaving Chilled Water Lower Temperature Limit {C}',base + chlLeavingLowLimOff,'')
    IF (isPriChwPumpVariable) THEN
      CALL AddToObjStr('Chiller Flow Mode','LeavingSetpointModulated')
    ELSE
      CALL AddToObjStr('Chiller Flow Mode','ConstantFlow')
    END IF
    CALL AddToObjStr('Design Heat Recovery Water Flow Rate {m3/s}','0')
    CALL AddToObjStr('Heat Recovery Inlet Node Name','')
    CALL AddToObjStr('Heat Recovery Outlet Node Name','')
    CALL AddToObjFld('Sizing Factor', base + chlSizeFactorOff,'',.TRUE.)
    !BRANCH
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + chlNameOff,' ChW Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isBranchPriChwPump) THEN
      CALL AddToObjStr('Component Object Type','Chiller:Electric:EIR')
      CALL AddToObjFld('Component Name', base + chlNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + chlNameOff,' ChW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + chlNameOff,' ChW Outlet',.TRUE.)
    ELSE
      ! using a branch pump define the pump first followed by the chiller/district cooling
      IF (isPriChwPumpVariable) THEN
        CALL AddToObjStr('Component Object Type','Pump:VariableSpeed')
      ELSE
        CALL AddToObjStr('Component Object Type','Pump:ConstantSpeed')
      END IF
      CALL AddToObjFld('Component Name', base + chlNameOff,' ChW Branch Pump')
      CALL AddToObjFld('Component Inlet Node Name', base + chlNameOff,' ChW Pump Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + chlNameOff,' ChW Inlet')
      CALL AddToObjStr('Component Object Type','Chiller:Electric:EIR')
      CALL AddToObjFld('Component Name', base + chlNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + chlNameOff,' ChW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + chlNameOff,' ChW Outlet',.TRUE.)
    ENDIF
    IF (condType .EQ. condWater) THEN
      !BRANCH
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + chlNameOff,' CndW Branch')
      CALL AddToStrList(TRIM(FldVal(base + chlNameOff)) // ' CndW Branch',handleCondenserBranch)
      isChillerWithWaterCooled = .TRUE.
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','Chiller:Electric:EIR')
      CALL AddToObjFld('Component Name', base + chlNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + chlNameOff,' Cnd Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + chlNameOff,' Cnd Outlet',.TRUE.)
      !CURVE:BIQUADRATIC
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('Cooling Capacity Function of Temperature Curve for water cooled screw chillers')
      CALL WriteComment('Based on COMNET 2010 Table 88. ')
      CALL AddToObjFld('Name', base + chlNameOff,' ScrewCapFT')
      CALL AddToObjStr('Coefficient1 Constant','0.81302462')
      CALL AddToObjStr('Coefficient2 x',      '-0.01425319')
      CALL AddToObjStr('Coefficient3 x**2',   '-0.00161799')
      CALL AddToObjStr('Coefficient4 y',       '0.02638438')
      CALL AddToObjStr('Coefficient5 y**2',   '-0.00091543')
      CALL AddToObjStr('Coefficient6 x*y',     '0.00169601')
      CALL AddToObjStr('Minimum Value of x','5')
      CALL AddToObjStr('Maximum Value of x','10')
      CALL AddToObjStr('Minimum Value of y','24')
      CALL AddToObjStr('Maximum Value of y','35',.TRUE.)
      !CURVE:BIQUADRATIC
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('Energy Input to Cooling Output Ratio Function of Temperature Curve water '//  &
         'cooled screw chillers ')
      CALL WriteComment('Based on COMNET 2010 Table 91.')
      CALL AddToObjFld('Name', base + chlNameOff,' ScrewEIRFT')
      CALL AddToObjStr('Coefficient1 Constant','0.63812315')
      CALL AddToObjStr('Coefficient2 x',       '0.00630389')
      CALL AddToObjStr('Coefficient3 x**2',    '0.00092334')
      CALL AddToObjStr('Coefficient4 y',      '-0.00455294')
      CALL AddToObjStr('Coefficient5 y**2',    '0.00082568')
      CALL AddToObjStr('Coefficient6 x*y',    '-0.00156152')
      CALL AddToObjStr('Minimum Value of x','5')
      CALL AddToObjStr('Maximum Value of x','10')
      CALL AddToObjStr('Minimum Value of y','24')
      CALL AddToObjStr('Maximum Value of y','35',.TRUE.)
      !CURVE:QUADRATIC
      CALL CreateNewObj('Curve:Quadratic')
      CALL WriteComment('Energy Input to Cooling Output Ratio Function of Part Load Ratio Curve '//  &
       'for water cooled screw chiller')
      CALL WriteComment('Based on COMNET 2010 Table 93.')
      CALL AddToObjFld('Name', base + chlNameOff,' ScrewEIRFPLR')
      CALL AddToObjStr('Coefficient1 Constant','0.33018833')
      CALL AddToObjStr('Coefficient2 x',       '0.23554291')
      CALL AddToObjStr('Coefficient3 x**2',    '0.46070828')
      CALL AddToObjStr('Minimum Value of x','0')
      CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
    ELSE
      !CURVE:BIQUADRATIC
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('Cooling Capacity Function of Temperature Curve for air cooled screw chillers')
      CALL WriteComment('Based on COMNET 2010 Table 87. ')
      CALL AddToObjFld('Name', base + chlNameOff,' ScrewCapFT')
      CALL AddToObjStr('Coefficient1 Constant', '0.98652509')
      CALL AddToObjStr('Coefficient2 x',        '0.04950184')
      CALL AddToObjStr('Coefficient3 x**2',    '-0.00029824')
      CALL AddToObjStr('Coefficient4 y',       '-0.00368714')
      CALL AddToObjStr('Coefficient5 y**2',    '-0.00004455')
      CALL AddToObjStr('Coefficient6 x*y',     '-0.00050103')
      CALL AddToObjStr('Minimum Value of x','5')
      CALL AddToObjStr('Maximum Value of x','10')
      CALL AddToObjStr('Minimum Value of y','24')
      CALL AddToObjStr('Maximum Value of y','35',.TRUE.)
      !CURVE:BIQUADRATIC
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('Energy Input to Cooling Output Ratio Function of Temperature Curve air '//  &
         'cooled screw chillers ')
      CALL WriteComment('Based on COMNET 2010 Table 90.')
      CALL AddToObjFld('Name', base + chlNameOff,' ScrewEIRFT')
      CALL AddToObjStr('Coefficient1 Constant', '0.57063012')
      CALL AddToObjStr('Coefficient2 x',        '0.01195520')
      CALL AddToObjStr('Coefficient3 x**2',    '-0.00052187')
      CALL AddToObjStr('Coefficient4 y',       '-0.00003406')
      CALL AddToObjStr('Coefficient5 y**2',     '0.00042091')
      CALL AddToObjStr('Coefficient6 x*y',     '-0.00060539')
      CALL AddToObjStr('Minimum Value of x','5')
      CALL AddToObjStr('Maximum Value of x','10')
      CALL AddToObjStr('Minimum Value of y','24')
      CALL AddToObjStr('Maximum Value of y','35',.TRUE.)
      !CURVE:QUADRATIC
      CALL CreateNewObj('Curve:Quadratic')
      CALL WriteComment('Energy Input to Cooling Output Ratio Function of Part Load Ratio Curve '//  &
       'for air cooled screw chiller')
      CALL WriteComment('Based on COMNET 2010 Table 92.')
      CALL AddToObjFld('Name', base + chlNameOff,' ScrewEIRFPLR')
      CALL AddToObjStr('Coefficient1 Constant','0.03648722')
      CALL AddToObjStr('Coefficient2 x',       '0.73474298')
      CALL AddToObjStr('Coefficient3 x**2',    '0.21994748')
      CALL AddToObjStr('Minimum Value of x','0')
      CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
    ENDIF
  END IF
  IF ((condType .EQ. condAir) .OR. (condType .EQ. condEvap)) THEN
    CALL CreateNewObj('OutdoorAir:Node')
    CALL AddToObjFld('Name', base + chlNameOff,' Cnd Inlet')
    CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
  END IF

  !Add branch pumps if required - the branch is created above
  IF (isBranchPriChwPump) THEN
    IF (isPriChwPumpVariable) THEN
      CALL CreateNewObj('Pump:VariableSpeed')
      CALL AddToObjFld('Name', base + chlNameOff,' ChW Branch Pump')
      CALL AddToObjFld('Inlet Node Name', base + chlNameOff,' ChW Pump Inlet')
      CALL AddToObjFld('Outlet Node Name', base + chlNameOff,' Chw Inlet')
      CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Rated Pump Head {Pa}', compactChldLoopBase + cwpPriChWPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      !Curve based on VSD generic curve from user with adjustments to prevent neg values - Nov 2008
      !Can use something better when min flow rate is able to be autosized or entered as a fraction
      CALL AddToObjStr('Coefficient 1 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 2 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 3 of the Part Load Performance Curve','1.0')
      CALL AddToObjStr('Coefficient 4 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Min Flow Rate while operating in variable flow capacity {m3/s}','0')
      CALL AddToObjFld('Pump Control Type', compactChldLoopBase + cwpPumpControlTypeOff,'')
      IF (SameString(FldVal(compactChldLoopBase + cwpPumpSchedNameOff),'')) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', compactChldLoopBase + cwpPumpSchedNameOff,'',.TRUE.)
      END IF
    ELSE !constant speed pumping
      CALL CreateNewObj('Pump:ConstantSpeed')
      CALL AddToObjFld('Name', base + chlNameOff,' ChW Branch Pump')
      CALL AddToObjFld('Inlet Node Name', base + chlNameOff,' ChW Pump Inlet')
      CALL AddToObjFld('Outlet Node Name', base + chlNameOff,' Chw Inlet')
      CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Rated Pump Head {Pa}', compactChldLoopBase + cwpPriChWPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      CALL AddToObjFld('Pump Control Type', compactChldLoopBase + cwpPumpControlTypeOff,'')
      IF (SameString(FldVal(compactChldLoopBase + cwpPumpSchedNameOff),'')) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', compactChldLoopBase + cwpPumpSchedNameOff,'',.TRUE.)
      END IF
    END IF
  END IF
END DO

! HVACTemplate:Plant:Chiller:ObjectReference
DO iChiller = 1, numCompactChillerOR
  base =  compactChillerORBase(iChiller)

  ! set flags
  IF (SameString(FldVal(base + chlorObjTypeOff),'Chiller:Electric:EIR')) THEN
    detchlrObjType = chotElecEIR
  ELSEIF (SameString(FldVal(base + chlorObjTypeOff),'Chiller:Electric:ReformulatedEIR')) THEN
    detchlrObjType = chotElecRefEIR
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Plant:Chiller:ObjectReference "'//TRIM(FldVal(base + chlorNameOff))//'"'// &
                    ' in the Chiller Object Type field: '//TRIM(FldVal(base + chlorObjTypeOff)))
  END IF
  ! find referenced detailed chiller object
  detchlrBase = -1
  SELECT CASE (detchlrObjType)
  ! Chiller:Electric:EIR
    CASE(chotElecEIR)
      DO iDetChlr = 1, numDetChillerEIR
        IF (SameString(FldVal(detChillerEIRBase(iDetChlr) + chleirNameOff), FldVal(base +  chlorObjNameOff))) THEN
          detchlrBase = detChillerEIRBase(iDetChlr)
          EXIT
        END IF
      END DO
    IF (detchlrBase .NE. -1) THEN
      chwInletNode   = TRIM(FldVal(detchlrBase + chleirChWInletNodeOff))
      chwOutletNode  = TRIM(FldVal(detchlrBase + chleirChWOutletNodeOff))
      cndwInletNode  = TRIM(FldVal(detchlrBase + chleirCondInletNodeOff))
      cndwOutletNode = TRIM(FldVal(detchlrBase + chleirCondOutletNodeOff))
      condensertype  = TRIM(FldVal(detchlrBase + chleirCondTypeOff))
    ENDIF
  ! Chiller:Electric:ReformulatedEIR
    CASE(chotElecRefEIR)
      DO iDetChlr = 1, numDetChillerREIR
        IF (SameString(FldVal(detChillerREIRBase(iDetChlr) + chlreirNameOff), FldVal(base +  chlorObjNameOff))) THEN
          detchlrBase = detChillerREIRBase(iDetChlr)
          EXIT
        END IF
      END DO
      IF (detchlrBase .NE. -1) THEN
        chwInletNode   = TRIM(FldVal(detchlrBase + chlreirChWInletNodeOff))
        chwOutletNode  = TRIM(FldVal(detchlrBase + chlreirChWOutletNodeOff))
        cndwInletNode  = TRIM(FldVal(detchlrBase + chlreirCondInletNodeOff))
        cndwOutletNode = TRIM(FldVal(detchlrBase + chlreirCondOutletNodeOff))
        condensertype  = 'WaterCooled'
      ENDIF
    CASE DEFAULT
  END SELECT
  IF (detchlrBase == -1) THEN
    CALL WriteError('In HVACTemplate:Plant:Chiller:ObjectReference "'// TRIM(FldVal(base + chlorNameOff))//'"'// &
                    ' Referenced chiller not found: '//TRIM(FldVal(base + chlorObjTypeOff))//'="'// &
                    TRIM(FldVal(base + chlorObjNameOff))//'"')
    CYCLE
  ENDIF

  IF (SameString(TRIM(condensertype),'WaterCooled')) THEN
    condType = condWater
  ELSEIF (SameString(TRIM(condensertype),'AirCooled')) THEN
    condType = condAir
  ELSEIF (SameString(TRIM(condensertype),'EvaporativelyCooled')) THEN
    condType = condEvap
  ELSEIF (SameString(TRIM(condensertype),'')) THEN
    condType = condWater
  ELSE
      CALL WriteError('In HVACTemplate:Plant:Chiller:ObjectReference "'// TRIM(FldVal(base + chlorNameOff))//'"'// &
                      ' Invalid Condenser Typee in referenced chiller: '//TRIM(FldVal(base + chlorObjTypeOff))//'="'// &
                      TRIM(FldVal(base + chlorObjNameOff))//'"')
  END IF

  ! check if node names in referenced chiller are all unique (within the chiller - not checking across chillers)
  IF (SameString(TRIM(chwInletNode),TRIM(chwOutletNode)) .OR. SameString(TRIM(chwInletNode),TRIM(cndwInletNode)) .OR. &
      SameString(TRIM(chwInletNode),TRIM(cndwOutletNode)) .OR. SameString(TRIM(chwOutletNode),TRIM(cndwInletNode)) .OR. &
      SameString(TRIM(chwOutletNode),TRIM(cndwOutletNode)) .OR. SameString(TRIM(cndwInletNode),TRIM(cndwOutletNode))) THEN
    CALL WriteError('In HVACTemplate:Plant:Chiller:ObjectReference "'// TRIM(FldVal(base + chlorNameOff))//'"'// &
                 ' Duplicate node name found in referenced chiller: '//TRIM(FldVal(base + chlorObjTypeOff))//'="'// &
                 TRIM(FldVal(base + chlorObjNameOff))//'" All chiller inlet and outlet node names must be unique.')
  END IF

  ! check if node names in referenced chiller are blank
  IF (SameString(TRIM(chwInletNode),'')) THEN
    CALL WriteError('In HVACTemplate:Plant:Chiller:ObjectReference "'// TRIM(FldVal(base + chlorNameOff))//'"'// &
      ' Blank Chilled Water Inlet Node Name found in referenced chiller: '//TRIM(FldVal(base + chlorObjTypeOff))//'="'// &
      TRIM(FldVal(base + chlorObjNameOff))//'" The user must specify all node names in the referenced chiller.')
  END IF
  IF (SameString(TRIM(chwOutletNode),'')) THEN
    CALL WriteError('In HVACTemplate:Plant:Chiller:ObjectReference "'// TRIM(FldVal(base + chlorNameOff))//'"'// &
      ' Blank Chilled Water Outlet Node Name found in referenced chiller: '//TRIM(FldVal(base + chlorObjTypeOff))//'="'// &
      TRIM(FldVal(base + chlorObjNameOff))//'" The user must specify all node names in the referenced chiller.')
  END IF
  IF (SameString(TRIM(cndwInletNode),'')) THEN
    CALL WriteError('In HVACTemplate:Plant:Chiller:ObjectReference "'// TRIM(FldVal(base + chlorNameOff))//'"'// &
      ' Blank Condenser Water Inlet Node Name found in referenced chiller: '//TRIM(FldVal(base + chlorObjTypeOff))//'="'// &
      TRIM(FldVal(base + chlorObjNameOff))//'" The user must specify all node names in the referenced chiller.')
  END IF
  IF (SameString(TRIM(cndwOutletNode),'')) THEN
    CALL WriteError('In HVACTemplate:Plant:Chiller:ObjectReference "'// TRIM(FldVal(base + chlorNameOff))//'"'// &
      ' Blank Condenser Water Outlet Node Name found in referenced chiller: '//TRIM(FldVal(base + chlorObjTypeOff))//'="'// &
      TRIM(FldVal(base + chlorObjNameOff))//'" The user must specify all node names in the referenced chiller.')
  END IF
  ! add ChW Outlet Node to list of node names for ChW setpoints
  CALL AddToStrList(TRIM(chwOutletNode),handleChillerChWOutletNode)
  ! write branch objects
  !BRANCH ~line 50
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + chlorObjNameOff,' ChW Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  IF (.NOT. isBranchPriChwPump) THEN
    CALL AddToObjFld('Component Object Type',base + chlorObjTypeOff,'')
    CALL AddToObjFld('Component Name', base + chlorObjNameOff,'')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(chwInletNode))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(chwOutletNode),.TRUE.)
  ELSE
    ! using a branch pump define the pump first followed by the chiller/district cooling
    IF (isPriChwPumpVariable) THEN
      CALL AddToObjStr('Component Object Type','Pump:VariableSpeed')
    ELSE
      CALL AddToObjStr('Component Object Type','Pump:ConstantSpeed')
    END IF
    CALL AddToObjFld('Component Name', base + chlorObjNameOff,' ChW Branch Pump')
    CALL AddToObjFld('Component Inlet Node Name', base + chlorObjNameOff,' ChW Pump Inlet')
    CALL AddToObjStr('Component Outlet Node Name', TRIM(chwInletNode))
    CALL AddToObjFld('Component Object Type',base + chlorObjTypeOff,'')
    CALL AddToObjFld('Component Name', base + chlorObjNameOff,'')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(chwInletNode))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(chwOutletNode),.TRUE.)
  ENDIF
  IF (condType .EQ. condWater) THEN
    !BRANCH ~ line 59
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + chlorObjNameOff,' CndW Branch')
    CALL AddToStrList(TRIM(FldVal(base + chlorObjNameOff)) // ' CndW Branch',handleCondenserBranch)
    isChillerWithWaterCooled = .TRUE.
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjFld('Component Object Type',base + chlorObjTypeOff,'')
    CALL AddToObjFld('Component Name', base + chlorObjNameOff,'')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(cndwInletNode))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(cndwOutletNode),.TRUE.)
  END IF
!  IF ((condType .EQ. condAir) .OR. (condType .EQ. condEvap)) THEN
!    CALL CreateNewObj('OutdoorAir:Node')
!    CALL AddToObjStr('Name', TRIM(cndwInletNode))
!    CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
!  END IF

  !Add branch pumps if required - the branch is created above
  IF (isBranchPriChwPump) THEN
    IF (isPriChwPumpVariable) THEN
      CALL CreateNewObj('Pump:VariableSpeed')
      CALL AddToObjFld('Name', base + chlorObjNameOff,' ChW Branch Pump')
      CALL AddToObjFld('Inlet Node Name', base + chlorObjNameOff,' ChW Pump Inlet')
      CALL AddToObjStr('Outlet Node Name', TRIM(chwInletNode))
      CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Rated Pump Head {Pa}', compactChldLoopBase + cwpPriChWPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      !Curve based on VSD generic curve from user with adjustments to prevent neg values - Nov 2008
      !Can use something better when min flow rate is able to be autosized or entered as a fraction
      CALL AddToObjStr('Coefficient 1 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 2 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 3 of the Part Load Performance Curve','1.0')
      CALL AddToObjStr('Coefficient 4 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Min Flow Rate while operating in variable flow capacity {m3/s}','0')
      CALL AddToObjFld('Pump Control Type', compactChldLoopBase + cwpPumpControlTypeOff,'')
      IF (SameString(FldVal(compactChldLoopBase + cwpPumpSchedNameOff),'')) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', compactChldLoopBase + cwpPumpSchedNameOff,'',.TRUE.)
      END IF
    ELSE !constant speed pumping
      CALL CreateNewObj('Pump:ConstantSpeed')
      CALL AddToObjFld('Name', base + chlorObjNameOff,' ChW Branch Pump')
      CALL AddToObjFld('Inlet Node Name', base + chlorObjNameOff,' ChW Pump Inlet')
      CALL AddToObjStr('Outlet Node Name', TRIM(chwInletNode))
      CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Rated Pump Head {Pa}', compactChldLoopBase + cwpPriChWPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      CALL AddToObjFld('Pump Control Type', compactChldLoopBase + cwpPumpControlTypeOff,'')
      IF (SameString(FldVal(compactChldLoopBase + cwpPumpSchedNameOff),'')) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', compactChldLoopBase + cwpPumpSchedNameOff,'',.TRUE.)
      END IF
    END IF
  END IF
END DO

END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE CreateTower
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a compact cooling tower object

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for cooling tower type
INTEGER, PARAMETER :: towerSingle = 1
INTEGER, PARAMETER :: towerTwo = 2
          ! parameters for detailed tower object type
INTEGER, PARAMETER :: twrotOneSpeed = 1
INTEGER, PARAMETER :: twrotTwoSpeed = 2
INTEGER, PARAMETER :: twrotVarSpeed = 3

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for boiler
INTEGER :: dettwrBase=0  ! used as base for detailed tower
INTEGER :: towerType=0
INTEGER :: iTower
INTEGER :: iDetTower=0
INTEGER :: dettwrObjType=0
LOGICAL :: isHiSpdCapAuto
LOGICAL :: isLoSpdCapAuto
LOGICAL :: isFreeCapAuto
LOGICAL :: isAllAuto
LOGICAL :: isAnyAuto
LOGICAL :: isPumpBranch
CHARACTER(len=MaxAlphaLength) :: twrInletNode=''
CHARACTER(len=MaxAlphaLength) :: twrOutletNode=''
CHARACTER(len=MaxAlphaLength) :: pumpHead=''
CHARACTER(len=MaxAlphaLength) :: pumpControlType=''
CHARACTER(len=MaxAlphaLength) :: pumpSchedule=''

IF ((numCompactTower + numCompactTowerOR) .EQ. 0) RETURN
IF (compactChldLoopBase .NE. 0) THEN
  isPumpBranch = SameString(FldVal(compactChldLoopBase + cwpCndPumpTypeOff),'PumpPerTower')
  pumpSchedule = FldVal(compactChldLoopBase + cwpPumpSchedNameOff)
  pumpHead = FldVal(compactChldLoopBase + cwpCndPumpHeadOff)
  pumpControlType = FldVal(compactChldLoopBase + cwpPumpControlTypeOff)
ELSEIF (compactMixedLoopBase .NE. 0) THEN
  isPumpBranch = SameString(FldVal(compactMixedLoopBase + mwpPumpTypeOff),'PumpPerTowerOrBoiler')
  pumpSchedule = FldVal(compactMixedLoopBase + mwpPumpSchedNameOff)
  pumpHead = FldVal(compactMixedLoopBase + mwpPumpHeadOff)
  pumpControlType = FldVal(compactMixedLoopBase + mwpPumpControlTypeOff)
ELSE
  RETURN  !with no loop - no tower can be hooked up
END IF

! HVACTemplate:Plant:Tower
DO iTower = 1, numCompactTower
  base =  compactTowerBase(iTower)
  ! set flags
  !This object if twrTypeOff="Single Speed"
  IF (SameString(FldVal(base + twrTypeOff),'SingleSpeed')) THEN
    towerType = towerSingle
  ELSEIF (SameString(FldVal(base + twrTypeOff),'TwoSpeed')) THEN
    towerType = towerTwo
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Plant:Tower in the Tower Type field')
  END IF
  !twrHiSpdCapacityOff=autosize OR  twrLoSpdCapacityOff=autosize OR twrFreeConvCapacityOff=autosize
  isHiSpdCapAuto = SameString(FldVal(base + twrHiSpdCapacityOff),'autosize')
  isLoSpdCapAuto = SameString(FldVal(base + twrLoSpdCapacityOff),'autosize')
  isFreeCapAuto =  SameString(FldVal(base + twrFreeConvCapacityOff),'autosize')
  ! add CndW Outlet Node to list of node names for CndW setpoints
  IF ((compactChldLoopBase .NE. 0) .AND. &
      (SameString(FldVal(base + twrLoopTypeOff),'ChilledWater') .OR. SameString(FldVal(base + twrLoopTypeOff),''))) THEN
    CALL AddToStrList(TRIM(FldVal(base + twrNameOff)) // ' CndW Outlet',handleTowerOutletNode)
  ELSEIF (SameString(FldVal(base + twrLoopTypeOff),'MixedWater') .OR. SameString(FldVal(base + twrLoopTypeOff),'')) THEN
    CALL AddToStrList(TRIM(FldVal(base + twrNameOff)) // ' CndW Outlet',handleMixedTowerOutletNode)
  END IF
  ! write idf objects
  IF (towerType .EQ. towerSingle) THEN
    !check if autosizing is used properly
    isAllAuto = isHiSpdCapAuto .AND. isFreeCapAuto
    isAnyAuto = isHiSpdCapAuto .OR. isFreeCapAuto
    IF (isAnyAuto .AND. (.NOT. isAllAuto)) THEN
      CALL WriteError('For a single speed tower the high speed capacity and free ' // &
                      'convection capacity both need to be specified or set to autosize.')
    END IF
    !COOLING TOWER:SINGLE SPEED ~ line 4
    CALL CreateNewObj('CoolingTower:SingleSpeed')
    CALL AddToObjFld('Name', base + twrNameOff,'')
    CALL AddToObjFld('Water Inlet Node Name', base + twrNameOff,' CndW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + twrNameOff,' CndW Outlet')
    IF (isHiSpdCapAuto .OR. isFreeCapAuto) THEN
      CALL AddToObjStr('Design Water Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Design Air Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Design Fan Power {W}', base + twrHiSpdFanPowerOff,'')
      CALL AddToObjStr('Design U-Factor Times Area Value {W/K}','autosize')
      CALL AddToObjStr('Free Convection Air Flow Rate {m3/s}','autocalculate')
      CALL AddToObjStr('Free Convection Air Flow Sizing Factor','')
      CALL AddToObjStr('Free Convection U-Factor Times Area Value {W/K}','autocalculate')
      CALL AddToObjStr('Free Convection U-Factor Times Area Value Sizing Factor','')
      CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
      CALL AddToObjStr('Heat Rejection Capacity and Nominal Capacity Sizing Ratio','')
      CALL AddToObjStr('Nominal Capacity {W}','')
      CALL AddToObjStr('Free Convection Capacity {W}','autocalculate')
      CALL AddToObjStr('Free Convection Nominal Capacity Sizing Factor','')
      CALL AddToObjStr('Design Inlet Air Dry-Bulb Temperature {C}','')
      CALL AddToObjStr('Design Inlet Air Wet-Bulb Temperature {C}','')
      CALL AddToObjStr('Design Approach Temperature {deltaT}','')
      CALL AddToObjStr('Design Range Temperature {deltaT}','')
      CALL AddToObjStr('Basin Heater Capacity {W/K}','')
      CALL AddToObjStr('Basin Heater Setpoint Temperature {C}','')
      CALL AddToObjStr('Basin Heater Operating Schedule Name','')
      CALL AddToObjStr('Evaporation Loss Mode','SaturatedExit')
      CALL AddToObjStr('Evaporation Loss Factor {percent/K}','')
      CALL AddToObjStr('Drift Loss Percent {percent}','0.008')
      CALL AddToObjStr('Blowdown Calculation Mode','ConcentrationRatio')
      CALL AddToObjStr('Blowdown Concentration Ratio','3')
      CALL AddToObjStr('Blowdown Makeup Water Usage Schedule Name','')
      CALL AddToObjStr('Supply Water Storage Tank Name','')
      CALL AddToObjFld('Outdoor Air Inlet Node Name', base + twrNameOff,' Cooling Tower Outdoor Air Inlet Node')
      CALL AddToObjStr('Capacity Control','FanCycling')
      CALL AddToObjStr('Number of Cells','')
      CALL AddToObjStr('Cell Control','')
      CALL AddToObjStr('Cell Minimum  Water Flow Rate Fraction','')
      CALL AddToObjStr('Cell Maximum Water Flow Rate Fraction','')
      CALL AddToObjFld('Sizing Factor', base + twrSizeFactorOff,'',.TRUE.)
    ELSE
      CALL AddToObjStr('Design Water Flow Rate {m3/s}','')
      CALL AddToObjStr('Design Air Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Design Fan Power {W}', base + twrHiSpdFanPowerOff,'')
      CALL AddToObjStr('Design U-Factor Times Area Value {W/K}','')
      CALL AddToObjStr('Free Convection Air Flow Rate {m3/s}','autocalculate')
      CALL AddToObjStr('Free Convection Air Flow Sizing Factor','')
      CALL AddToObjStr('Free Convection U-Factor Times Area Value {W/K}','')
      CALL AddToObjStr('Free Convection U-Factor Times Area Value Sizing Factor','')
      CALL AddToObjStr('Performance Input Method','NominalCapacity')
      CALL AddToObjStr('Heat Rejection Capacity and Nominal Capacity Sizing Ratio','')
      CALL AddToObjFld('Nominal Capacity {W}', base + twrHiSpdCapacityOff,'')
      CALL AddToObjFld('Free Convection Capacity {W}', base + twrFreeConvCapacityOff,'')
      CALL AddToObjStr('Free Convection Nominal Capacity Sizing Factor','')
      CALL AddToObjStr('Design Inlet Air Dry-Bulb Temperature {C}','')
      CALL AddToObjStr('Design Inlet Air Wet-Bulb Temperature {C}','')
      CALL AddToObjStr('Design Approach Temperature {deltaT}','')
      CALL AddToObjStr('Design Range Temperature {deltaT}','')
      CALL AddToObjStr('Basin Heater Capacity {W/K}','')
      CALL AddToObjStr('Basin Heater Setpoint Temperature {C}','')
      CALL AddToObjStr('Basin Heater Operating Schedule Name','')
      CALL AddToObjStr('Evaporation Loss Mode','SaturatedExit')
      CALL AddToObjStr('Evaporation Loss Factor {percent/K}','')
      CALL AddToObjStr('Drift Loss Percent {percent}','0.008')
      CALL AddToObjStr('Blowdown Calculation Mode','ConcentrationRatio')
      CALL AddToObjStr('Blowdown Concentration Ratio','3')
      CALL AddToObjStr('Blowdown Makeup Water Usage Schedule Name','')
      CALL AddToObjStr('Supply Water Storage Tank Name','')
      CALL AddToObjFld('Outdoor Air Inlet Node Name', base + twrNameOff,' Cooling Tower Outdoor Air Inlet Node')
      CALL AddToObjStr('Capacity Control','FanCycling')
      CALL AddToObjStr('Number of Cells','')
      CALL AddToObjStr('Cell Control','')
      CALL AddToObjStr('Cell Minimum  Water Flow Rate Fraction','')
      CALL AddToObjStr('Cell Maximum Water Flow Rate Fraction','')
      CALL AddToObjFld('Sizing Factor', base + twrSizeFactorOff,'',.TRUE.)
    END IF
!    !OutAir Node
    CALL CreateNewObj('OutdoorAir:Node')
    CALL AddToObjStr('Name', TRIM(FldVal(base + twrNameOff))//' Cooling Tower Outdoor Air Inlet Node')
    CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
    !BRANCH ~ line 29
    CALL CreateNewObj('Branch')
    IF ((compactChldLoopBase .NE. 0) .AND. &
        (SameString(FldVal(base + twrLoopTypeOff),'ChilledWater') .OR. SameString(FldVal(base + twrLoopTypeOff),''))) THEN
      CALL AddToObjFld('Name', base + twrNameOff,' CndW Branch')
      CALL AddToStrList(TRIM(FldVal(base + twrNameOff)) // ' CndW Branch',handleCondenserSupplyBranch)
    ELSEIF (SameString(FldVal(base + twrLoopTypeOff),'MixedWater') .OR. SameString(FldVal(base + twrLoopTypeOff),'')) THEN
      CALL AddToObjFld('Name', base + twrNameOff,' Mixed Branch')
      CALL AddToStrList(TRIM(FldVal(base + twrNameOff)) // ' Mixed Branch',handleMixedSupplyBranch)
    ENDIF
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isPumpBranch) THEN
      CALL AddToObjStr('Component Object Type','CoolingTower:SingleSpeed')
      CALL AddToObjFld('Component Name', base + twrNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + twrNameOff,' CndW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + twrNameOff,' CndW Outlet',.TRUE.)
    ELSE
      ! using a branch pump define the pump first followed by the tower
      CALL AddToObjStr('Component Object Type','Pump:ConstantSpeed')
      CALL AddToObjFld('Component Name', base + twrNameOff,' CndW Branch Pump')
      CALL AddToObjFld('Component Inlet Node Name', base + twrNameOff,' CndW Pump Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + twrNameOff,' CndW Inlet')
      CALL AddToObjStr('Component Object Type','CoolingTower:SingleSpeed')
      CALL AddToObjFld('Component Name', base + twrNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + twrNameOff,' CndW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + twrNameOff,' CndW Outlet',.TRUE.)
    ENDIF
  ELSE
    !check if autosizing is used properly
    isAllAuto = isHiSpdCapAuto .AND. isLoSpdCapAuto .AND. isFreeCapAuto
    isAnyAuto = isHiSpdCapAuto .OR.  isLoSpdCapAuto .OR.  isFreeCapAuto
    IF (isAnyAuto .AND. (.NOT. isAllAuto)) THEN
      CALL WriteError('For a two speed tower the high speed capacity the low speed capacity and free ' // &
                      'convection capacity all need to be specified or set to autosize.')
    END IF
    !COOLING TOWER:TWO SPEED ~ line 30
    CALL CreateNewObj('CoolingTower:TwoSpeed')
    CALL AddToObjFld('Name', base + twrNameOff,'')
    CALL AddToObjFld('Water Inlet Node Name', base + twrNameOff,' CndW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + twrNameOff,' CndW Outlet')
    IF (isHiSpdCapAuto .OR. isFreeCapAuto .OR. isLoSpdCapAuto) THEN
      CALL AddToObjStr('Design Water Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('High Fan Speed Air Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('High Fan Speed Fan Power {W}', base + twrHiSpdFanPowerOff,'')
      CALL AddToObjStr('High Fan Speed U-Factor Times Area Value {W/K}','autosize')
      CALL AddToObjStr('Low Fan Speed Air Flow Rate {m3/s}','autocalculate')
      CALL AddToObjStr('Low Fan Speed Air Flow Rate Sizing Factor {m3/s}','')
      CALL AddToObjFld('Low Fan Speed Fan Power {W}', base + twrLoSpdFanPowerOff,'')
      CALL AddToObjStr('Low Fan Speed Fan Power Sizing Factor {m3/s}','')
      CALL AddToObjStr('Low Fan Speed U-Factor Times Area Value {W/K}','autocalculate')
      CALL AddToObjStr('Low Fan Speed U-Factor Times Area Value Sizing Factor {m3/s}','')
      CALL AddToObjStr('Free Convection Regime Air Flow Rate {m3/s}','autocalculate')
      CALL AddToObjStr('Free Convection Regime Air Flow Rate Sizing Factor {m3/s}','')
      CALL AddToObjStr('Free Convection Regime U-Factor Times Area Value {W/K}','autocalculate')
      CALL AddToObjStr('Free Convection Regime U-Factor Times Area Value Sizing Factor {m3/s}','')
      CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
      CALL AddToObjStr('Heat Rejection Capacity and Nominal Capacity Sizing Ratio {m3/s}','')
      CALL AddToObjStr('High Speed Nominal Capacity {W}','')
      CALL AddToObjStr('Low Speed Nominal Capacity {W}','autocalculate')
      CALL AddToObjStr('Low Speed Nominal Capacity Sizing Factor {m3/s}','')
      CALL AddToObjStr('Free Convection Capacity {W}','autocalculate')
      CALL AddToObjStr('Free Convection Capacity Sizing Factor {m3/s}','')
      CALL AddToObjStr('Design Inlet Air Dry-Bulb Temperature {C}','')
      CALL AddToObjStr('Design Inlet Air Wet-Bulb Temperature {C}','')
      CALL AddToObjStr('Design Approach Temperature {deltaT}','')
      CALL AddToObjStr('Design Range Temperature {deltaT}','')
      CALL AddToObjStr('Basin Heater Capacity {W/K}','')
      CALL AddToObjStr('Basin Heater Setpoint Temperature {C}','')
      CALL AddToObjStr('Basin Heater Operating Schedule Name','')
      CALL AddToObjStr('Evaporation Loss Mode','SaturatedExit')
      CALL AddToObjStr('Evaporation Loss Factor {percent/K}','')
      CALL AddToObjStr('Drift Loss Percent {percent}','0.008')
      CALL AddToObjStr('Blowdown Calculation Mode','ConcentrationRatio')
      CALL AddToObjStr('Blowdown Concentration Ratio','3')
      CALL AddToObjStr('Blowdown Makeup Water Usage Schedule Name','')
      CALL AddToObjStr('Supply Water Storage Tank Name','')
      CALL AddToObjFld('Outdoor Air Inlet Node Name', base + twrNameOff,' Cooling Tower Outdoor Air Inlet Node')
      CALL AddToObjStr('Number of Cells','')
      CALL AddToObjStr('Cell Control','')
      CALL AddToObjStr('Cell Minimum  Water Flow Rate Fraction','')
      CALL AddToObjStr('Cell Maximum Water Flow Rate Fraction','')
      CALL AddToObjFld('Sizing Factor', base + twrSizeFactorOff,'',.TRUE.)
    ELSE
      CALL AddToObjStr('Design Water Flow Rate {m3/s}','')
      CALL AddToObjStr('High Fan Speed Air Flow Rate{m3/s}','autosize')
      CALL AddToObjFld('High Fan Speed Fan Power {W}', base + twrHiSpdFanPowerOff,'')
      CALL AddToObjStr('High Fan Speed U-Factor Times Area Value {W/K}','')
      CALL AddToObjStr('Low Fan Speed Air Flow Rate {m3/s}','autocalculate')
      CALL AddToObjStr('Low Fan Speed Air Flow Rate Sizing Factor {m3/s}','')
      CALL AddToObjFld('Low Fan Speed Fan Power {W}', base + twrLoSpdFanPowerOff,'')
      CALL AddToObjStr('Low Fan Speed Fan Power Sizing Factor {m3/s}','')
      CALL AddToObjStr('Low Fan Speed U-Factor Times Area Value {W/K}','')
      CALL AddToObjStr('Low Fan Speed U-Factor Times Area Value Sizing Factor {m3/s}','')
      CALL AddToObjStr('Free Convection Regime Air Flow Rate {m3/s}','autocalculate')
      CALL AddToObjStr('Free Convection Regime Air Flow Rate Sizing Factor {m3/s}','')
      CALL AddToObjStr('Free Convection Regime U-Factor Times Area Value {W/K}','')
      CALL AddToObjStr('Free Convection Regime U-Factor Times Area Value Sizing Factor {m3/s}','')
      CALL AddToObjStr('Performance Input Method','NominalCapacity')
      CALL AddToObjStr('Heat Rejection Capacity and Nominal Capacity Sizing Ratio {m3/s}','')
      CALL AddToObjFld('High Speed Nominal Capacity {W}', base + twrHiSpdCapacityOff,'')
      CALL AddToObjFld('Low Speed Nominal Capacity {W}', base + twrLoSpdCapacityOff,'')
      CALL AddToObjStr('Low Speed Nominal Capacity Sizing Factor {m3/s}','')
      CALL AddToObjFld('Free Convection Capacity {W}', base + twrFreeConvCapacityOff,'')
      CALL AddToObjStr('Free Convection Capacity Sizing Factor {m3/s}','')
      CALL AddToObjStr('Design Inlet Air Dry-Bulb Temperature {C}','')
      CALL AddToObjStr('Design Inlet Air Wet-Bulb Temperature {C}','')
      CALL AddToObjStr('Design Approach Temperature {deltaT}','')
      CALL AddToObjStr('Design Range Temperature {deltaT}','')
      CALL AddToObjStr('Basin Heater Capacity {W/K}','')
      CALL AddToObjStr('Basin Heater Setpoint Temperature {C}','')
      CALL AddToObjStr('Basin Heater Operating Schedule Name','')
      CALL AddToObjStr('Evaporation Loss Mode','SaturatedExit')
      CALL AddToObjStr('Evaporation Loss Factor {percent/K}','')
      CALL AddToObjStr('Drift Loss Percent {percent}','0.008')
      CALL AddToObjStr('Blowdown Calculation Mode','ConcentrationRatio')
      CALL AddToObjStr('Blowdown Concentration Ratio','3')
      CALL AddToObjStr('Blowdown Makeup Water Usage Schedule Name','')
      CALL AddToObjStr('Supply Water Storage Tank Name','')
      CALL AddToObjFld('Outdoor Air Inlet Node Name', base + twrNameOff,' Cooling Tower Outdoor Air Inlet Node')
      CALL AddToObjStr('Number of Cells','')
      CALL AddToObjStr('Cell Control','')
      CALL AddToObjStr('Cell Minimum  Water Flow Rate Fraction','')
      CALL AddToObjStr('Cell Maximum Water Flow Rate Fraction','')
      CALL AddToObjFld('Sizing Factor', base + twrSizeFactorOff,'',.TRUE.)
    END IF
    CALL CreateNewObj('OutdoorAir:Node')
    CALL AddToObjStr('Name', TRIM(FldVal(base + twrNameOff))//' Cooling Tower Outdoor Air Inlet Node')
    CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
    !BRANCH ~ line 71
    CALL CreateNewObj('Branch')
    IF ((compactChldLoopBase .NE. 0) .AND. &
        (SameString(FldVal(base + twrLoopTypeOff),'ChilledWater') .OR. SameString(FldVal(base + twrLoopTypeOff),''))) THEN
      CALL AddToObjFld('Name', base + twrNameOff,' CndW Branch')
      CALL AddToStrList(TRIM(FldVal(base + twrNameOff)) // ' CndW Branch',handleCondenserSupplyBranch)
    ELSEIF (SameString(FldVal(base + twrLoopTypeOff),'MixedWater') .OR. SameString(FldVal(base + twrLoopTypeOff),'')) THEN
      CALL AddToObjFld('Name', base + twrNameOff,' Mixed Branch')
      CALL AddToStrList(TRIM(FldVal(base + twrNameOff)) // ' Mixed Branch',handleMixedSupplyBranch)
    ENDIF
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isPumpBranch) THEN
      CALL AddToObjStr('Component Object Type','CoolingTower:TwoSpeed')
      CALL AddToObjFld('Component Name', base + twrNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + twrNameOff,' CndW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + twrNameOff,' CndW Outlet',.TRUE.)
    ELSE
      ! using a branch pump define the pump first followed by the tower
      CALL AddToObjStr('Component Object Type','Pump:ConstantSpeed')
      CALL AddToObjFld('Component Name', base + twrNameOff,' CndW Branch Pump')
      CALL AddToObjFld('Component Inlet Node Name', base + twrNameOff,' CndW Pump Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + twrNameOff,' CndW Inlet')
      CALL AddToObjStr('Component Object Type','CoolingTower:TwoSpeed')
      CALL AddToObjFld('Component Name', base + twrNameOff,'')
      CALL AddToObjFld('Component Inlet Node Name', base + twrNameOff,' CndW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + twrNameOff,' CndW Outlet',.TRUE.)
    END IF
  END IF
  !Add branch pumps if required - the branch is created above
  !Tower branch pumps are always constant speed
  IF (isPumpBranch) THEN
    CALL CreateNewObj('Pump:ConstantSpeed')
    CALL AddToObjFld('Name', base + twrNameOff,' CndW Branch Pump')
    CALL AddToObjFld('Inlet Node Name', base + twrNameOff,' CndW Pump Inlet')
    CALL AddToObjFld('Outlet Node Name', base + twrNameOff,' CndW Inlet')
    CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Rated Pump Head {Pa}', TRIM(pumpHead))
    CALL AddToObjStr('Rated Power Consumption {W}','autosize')
    CALL AddToObjStr('Motor Efficiency','0.9')
    CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
    CALL AddToObjStr('Pump Control Type', TRIM(pumpControlType))
    CALL AddToObjStr('Pump Flow Rate Schedule Name', TRIM(pumpSchedule),.TRUE.)
  END IF
END DO

! HVACTemplate:Plant:Tower:ObjectReference
DO iTower = 1, numCompactTowerOR
  base =  compactTowerORBase(iTower)

  ! set flags
  IF (SameString(FldVal(base + twrorObjTypeOff),'CoolingTower:SingleSpeed')) THEN
    dettwrObjType = twrotOneSpeed
  ELSEIF (SameString(FldVal(base + twrorObjTypeOff),'CoolingTower:TwoSpeed')) THEN
    dettwrObjType = twrotTwoSpeed
  ELSEIF (SameString(FldVal(base + twrorObjTypeOff),'CoolingTower:VariableSpeed')) THEN
    dettwrObjType = twrotVarSpeed
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Plant:Tower:ObjectReference "'//TRIM(FldVal(base + twrorNameOff))//'"'// &
                    ' in the tower Object Type field: '//TRIM(FldVal(base + twrorObjTypeOff)))
  END IF
  ! find referenced detailed tower object
  dettwrBase = -1
  SELECT CASE(dettwrObjType)
  ! Tower:SingleSpeed
    CASE(twrotOneSpeed)
      DO iDetTower = 1, numDetTowerSS
        IF (SameString(FldVal(detTowerSSBase(iDetTower) + twrssNameOff), FldVal(base +  twrorObjNameOff))) THEN
          dettwrBase = detTowerSSBase(iDetTower)
          EXIT
        END IF
      END DO
      IF (dettwrBase .NE. -1) THEN
        twrInletNode   = TRIM(FldVal(dettwrBase + twrssInletNodeOff))
        twrOutletNode  = TRIM(FldVal(dettwrBase + twrssOutletNodeOff))
      ENDIF
  ! Tower:TwoSpeed
    CASE(twrotTwoSpeed)
      DO iDetTower = 1, numDetTowerTS
        IF (SameString(FldVal(detTowerTSBase(iDetTower) + twrtsNameOff), FldVal(base +  twrorObjNameOff))) THEN
          dettwrBase = detTowerTSBase(iDetTower)
          EXIT
        END IF
      END DO
      IF (dettwrBase .NE. -1) THEN
        twrInletNode   = TRIM(FldVal(dettwrBase + twrtsInletNodeOff))
        twrOutletNode  = TRIM(FldVal(dettwrBase + twrtsOutletNodeOff))
      ENDIF
  ! Tower:VariableSpeed
    CASE(twrotVarSpeed)
      DO iDetTower = 1, numDetTowerVS
        IF (SameString(FldVal(detTowerVSBase(iDetTower) + twrvsNameOff), FldVal(base +  twrorObjNameOff))) THEN
          dettwrBase = detTowerVSBase(iDetTower)
          EXIT
        END IF
      END DO
      IF (dettwrBase .NE. -1) THEN
        twrInletNode   = TRIM(FldVal(dettwrBase + twrvsInletNodeOff))
        twrOutletNode  = TRIM(FldVal(dettwrBase + twrvsOutletNodeOff))
      ENDIF
    CASE DEFAULT
  END SELECT
  IF (dettwrBase == -1) THEN
    CALL WriteError('In HVACTemplate:Plant:Tower:ObjectReference "'// TRIM(FldVal(base + twrorNameOff))//'"'// &
                    ' Referenced tower not found: '//TRIM(FldVal(base + twrorObjTypeOff))//'="'// &
                    TRIM(FldVal(base + twrorObjNameOff))//'"')
    CYCLE
  ENDIF

  ! check if node names in referenced tower are all unique (within the tower - not checking across tower)
  IF (SameString(TRIM(twrInletNode),TRIM(twrOutletNode))) THEN
    CALL WriteError('In HVACTemplate:Plant:Tower:ObjectReference "'// TRIM(FldVal(base + twrorNameOff))//'"'// &
                 ' Duplicate node name found in referenced tower: '//TRIM(FldVal(base + twrorObjTypeOff))//'="'// &
                 TRIM(FldVal(base + twrorObjNameOff))//'" All tower inlet and outlet node names must be unique.')
  END IF

  ! check if node names in referenced tower are blank
  IF (SameString(TRIM(twrInletNode),'')) THEN
    CALL WriteError('In HVACTemplate:Plant:Tower:ObjectReference "'// TRIM(FldVal(base + twrorNameOff))//'"'// &
      ' Blank Inlet Node Name found in referenced tower: '//TRIM(FldVal(base + twrorObjTypeOff))//'="'// &
      TRIM(FldVal(base + twrorObjNameOff))//'" The user must specify all node names in the referenced tower.')
  END IF
  IF (SameString(TRIM(twrOutletNode),'')) THEN
    CALL WriteError('In HVACTemplate:Plant:Tower:ObjectReference "'// TRIM(FldVal(base + twrorNameOff))//'"'// &
      ' Blank Chilled Water Outlet Node Name found in referenced tower: '//TRIM(FldVal(base + twrorObjTypeOff))//'="'// &
      TRIM(FldVal(base + twrorObjNameOff))//'" The user must specify all node names in the referenced tower.')
  END IF
  ! add CndW Outlet Node to list of node names for CndW setpoints
  IF ((compactChldLoopBase .NE. 0) .AND. &
      (SameString(FldVal(base + twrorLoopTypeOff),'ChilledWater') .OR. SameString(FldVal(base + twrorLoopTypeOff),''))) THEN
    CALL AddToStrList(TRIM(twrOutletNode),handleTowerOutletNode)
  ELSEIF (SameString(FldVal(base + twrorLoopTypeOff),'MixedWater') .OR. SameString(FldVal(base + twrorLoopTypeOff),'')) THEN
    CALL AddToStrList(TRIM(twrOutletNode),handleMixedTowerOutletNode)
  END IF
  CALL CreateNewObj('OutdoorAir:Node')
  CALL AddToObjStr('Name', TRIM(FldVal(base + twrNameOff))//' Cooling Tower Outdoor Air Inlet Node')
  CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
  ! write branch objects
  !BRANCH ~line 50
  CALL CreateNewObj('Branch')
  IF ((compactChldLoopBase .NE. 0) .AND. &
      (SameString(FldVal(base + twrorLoopTypeOff),'ChilledWater') .OR. SameString(FldVal(base + twrorLoopTypeOff),''))) THEN
    CALL AddToObjFld('Name', base + twrorObjNameOff,' CndW Branch')
    CALL AddToStrList(TRIM(FldVal(base + twrorObjNameOff)) // ' CndW Branch',handleCondenserSupplyBranch)
  ELSEIF (SameString(FldVal(base + twrorLoopTypeOff),'MixedWater') .OR. SameString(FldVal(base + twrorLoopTypeOff),'')) THEN
    CALL AddToObjFld('Name', base + twrorObjNameOff,' Mixed Branch')
    CALL AddToStrList(TRIM(FldVal(base + twrorObjNameOff)) // ' Mixed Branch',handleMixedSupplyBranch)
  ENDIF
  CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isPumpBranch) THEN
      !not using branch pumps so only tower is on the branch
    CALL AddToObjFld('Component Object Type',base + twrorObjTypeOff,'')
    CALL AddToObjFld('Component Name', base + twrorObjNameOff,'')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(twrInletNode))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(twrOutletNode),.TRUE.)
  ELSE
    ! using a branch pump define the pump first followed by the tower
    CALL AddToObjStr('Component Object Type','Pump:ConstantSpeed')
    CALL AddToObjFld('Component Name', base + twrorObjNameOff,' CndW Branch Pump')
    CALL AddToObjFld('Component Inlet Node Name', base + twrorObjNameOff,' CndW Pump Inlet')
    CALL AddToObjStr('Component Outlet Node Name', TRIM(twrInletNode))
    CALL AddToObjFld('Component Object Type',base + twrorObjTypeOff,'')
    CALL AddToObjFld('Component Name', base + twrorObjNameOff,'')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(twrInletNode))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(twrOutletNode),.TRUE.)
  ENDIF
  !Tower branch pumps are always constant speed
  IF (isPumpBranch) THEN
    CALL CreateNewObj('Pump:ConstantSpeed')
    CALL AddToObjFld('Name', base + twrorObjNameOff,' CndW Branch Pump')
    CALL AddToObjFld('Inlet Node Name', base + twrorObjNameOff,' CndW Pump Inlet')
    CALL AddToObjStr('Outlet Node Name', TRIM(twrInletNode))
    CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Rated Pump Head {Pa}', TRIM(pumpHead))
    CALL AddToObjStr('Rated Power Consumption {W}','autosize')
    CALL AddToObjStr('Motor Efficiency','0.9')
    CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
    CALL AddToObjStr('Pump Control Type', TRIM(pumpControlType))
    CALL AddToObjStr('Pump Flow Rate Schedule Name', TRIM(pumpSchedule),.TRUE.)
  END IF
END DO
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE CreateChldLoop
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for COMPACT HVAC:PLANT:CHILLED WATER LOOP

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for pump type
INTEGER, PARAMETER :: pumpVariable = 1
INTEGER, PARAMETER :: pumpConstant = 2
INTEGER, PARAMETER :: pumpConstPriVarSec = 3
INTEGER, PARAMETER :: pumpVarPriConstSec = 4
          ! parameters for setpoint reset type
INTEGER,PARAMETER :: sprtNone = 1
INTEGER,PARAMETER :: sprtOutdoor = 2
          ! parameters for fluid type
INTEGER,PARAMETER :: ftWater = 1
INTEGER,PARAMETER :: ftEthyleneGlycol30 = 2
INTEGER,PARAMETER :: ftEthyleneGlycol40 = 3
INTEGER,PARAMETER :: ftEthyleneGlycol50 = 4
INTEGER,PARAMETER :: ftEthyleneGlycol60 = 5
INTEGER,PARAMETER :: ftPropyleneGlycol30 = 6
INTEGER,PARAMETER :: ftPropyleneGlycol40 = 7
INTEGER,PARAMETER :: ftPropyleneGlycol50 = 8
INTEGER,PARAMETER :: ftPropyleneGlycol60 = 9


          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for
LOGICAL :: isCwpWSetPtSchedName
INTEGER :: chillerBase
INTEGER :: iChiller
INTEGER :: towerBase
INTEGER :: iTower
INTEGER :: iStr
INTEGER :: loopPumpConfig=0
LOGICAL :: isCwpPumpSchedBlank
LOGICAL :: isCwpCndOperDefault
LOGICAL :: isCndTempControlWetBulb
LOGICAL :: isCndWSetPtSchedBlank
LOGICAL :: isCwpChWOperTypeUserDefined
LOGICAL :: isCndTempControlTypeValid
LOGICAL :: isCndOperSchemeValid
LOGICAL :: isChWSupplyBypassYes
LOGICAL :: isChWDemandBypassYes
LOGICAL :: isCndSupplyBypassYes
LOGICAL :: isCndDemandBypassYes
LOGICAL :: isMinOutTempBlank
INTEGER :: setPtResetType=0
INTEGER :: fluidType=0
INTEGER :: lastBranch=0
INTEGER :: numTowersThisLoop=0
LOGICAL :: isChwPrimPumpHeader
LOGICAL :: isChwPrimPumpBranch
LOGICAL :: isChwSecPumpHeader
LOGICAL :: isCndPumpHeader
LOGICAL :: isCndPumpBranch
LOGICAL :: isEndOfList
CHARACTER(len=1) :: pumpCountChwPrim='1'
CHARACTER(len=1) :: pumpCountChwSec='1'
CHARACTER(len=1) :: pumpCountCnd='1'

base =  compactChldLoopBase
IF (base .EQ. 0) RETURN
IF ((numCompactChiller + numCompactChillerOR) .EQ. 0) RETURN
! set flags
CALL validChoice(base + cwpPumpControlTypeOff,'Intermittent','Continuous','Plant:ChilledWaterLoop','Pump Control Type')
isCwpChWOperTypeUserDefined = SameString(FldVal(base + cwpChWOperTypeOff),'UserDefined')
CALL validChoice(base + cwpChWOperTypeOff,'Default','UserDefined','Plant:ChilledWaterLoop','Chiller Plant Operation Scheme Type')
isCwpPumpSchedBlank = SameString(FldVal(base + cwpPumpSchedNameOff),'')
isCwpCndOperDefault = SameString(FldVal(base + cwpCndOperTypeOff),'Default')
! validate Condenser Plant Operation Scheme Type
isCndOperSchemeValid = .FALSE.
IF (SameSTring(FldVal(base + cwpCndOperTypeOff),'Default')) isCndOperSchemeValid = .TRUE.
IF (SameSTring(FldVal(base + cwpCndOperTypeOff),'UserDefined')) isCndOperSchemeValid = .TRUE.
IF (SameSTring(FldVal(base + cwpCndOperTypeOff),'')) isCndOperSchemeValid = .TRUE.
IF (.NOT. isCndOperSchemeValid) THEN
  CALL WriteError('Invalid choice in HVACTemplate:Plant:ChilledWaterLoop in the Condenser Plant Operation Scheme Type field')
END IF
isCndTempControlWetBulb = SameString(FldVal(base + cwpCndTempControlTypeOff),'OutdoorWetBulbTemperature')
!validate cwpCndTempControlTypeOff
isCndTempControlTypeValid = .FALSE.
IF (SameString(FldVal(base + cwpCndTempControlTypeOff),'OutdoorWetBulbTemperature')) isCndTempControlTypeValid = .TRUE.
IF (SameString(FldVal(base + cwpCndTempControlTypeOff),'SpecifiedSetpoint')) isCndTempControlTypeValid = .TRUE.
IF (SameString(FldVal(base + cwpCndTempControlTypeOff),'')) isCndTempControlTypeValid = .TRUE.
IF (.NOT. isCndTempControlTypeValid) THEN
  CALL WriteError('Invalid choice in HVACTemplate:Plant:ChilledWaterLoop in the Condenser Water Temperature Control Type field')
END IF
isCndWSetPtSchedBlank = SameString(FldVal(base + cwpCndSetPtSchedNameOff),'')
IF ((SameString(FldVal(base + cwpChWPumpConfigOff),'VariablePrimaryNoSecondary')) .OR. &
    (SameString(FldVal(base + cwpChWPumpConfigOff),'VariableFlow'))) THEN
  loopPumpConfig = pumpVariable
ELSEIF ((SameString(FldVal(base + cwpChWPumpConfigOff),'ConstantPrimaryNoSecondary')) .OR. &
        (SameString(FldVal(base + cwpChWPumpConfigOff),'ConstantFlow'))) THEN
  loopPumpConfig = pumpConstant
ELSEIF (SameString(FldVal(base + cwpChWPumpConfigOff),'ConstantPrimaryVariableSecondary')) THEN
  loopPumpConfig = pumpConstPriVarSec
ELSEIF (SameString(FldVal(base + cwpChWPumpConfigOff),'VariablePrimaryConstantSecondary')) THEN
!  loopPumpConfig = pumpVarPriConstSec
  loopPumpConfig = pumpVariable
  CALL WriteError('Warning:  In HVACTemplate:Plant:ChilledWaterLoop "'//TRIM(FldVal(base + cwpNameOff))//'"'// &
                    ' Chilled Water Pump Configuration VariablePrimaryConstantSecondary is currently unsupported.'// &
                    ' Changed to VariablePrimaryNoSecondary.',msgWarning)
ELSE
  CALL WriteError('Invalid choice in HVACTemplate:Plant:ChilledWaterLoop in the Chilled Water Pump Configuration field')
END IF
IF (SameString(FldVal(base +  cwpChWSetPtResetTypeOff), 'OutdoorAirTemperatureReset')) THEN
  setPtResetType = sprtOutdoor
ELSE IF (SameString(FldVal(base +  cwpChWSetPtResetTypeOff), 'None')) THEN
  setPtResetType = sprtNone
ELSE
  CALL WriteError('Invalid Chilled Water Setpoint Reset Type.  None assumed.')
  setPtResetType = sprtNone
END IF
CALL validChoice(base + cwpChWSupplyBypassOff,'Yes','No','HVACTemplate:Plant:ChilledWaterLoop', &
 'Chilled Water Supply Side Bypass Pipe')
isChWSupplyBypassYes = SameString(FldVal(base + cwpChWSupplyBypassOff),'Yes')
CALL validChoice(base + cwpChwDemandBypassOff,'Yes','No','HVACTemplate:Plant:ChilledWaterLoop', &
 'Chilled Water Demand Side Bypass Pipe')
isChWDemandBypassYes = SameString(FldVal(base + cwpChwDemandBypassOff),'Yes')
CALL validChoice(base + cwpCndSupplyBypassOff,'Yes','No','HVACTemplate:Plant:ChilledWaterLoop', &
 'Condenser Water Supply Side Bypass Pipe')
isCndSupplyBypassYes = SameString(FldVal(base + cwpCndSupplyBypassOff),'Yes')
CALL validChoice(base + cwpCndDemandBypassOff,'Yes','No','HVACTemplate:Plant:ChilledWaterLoop', &
 'Condenser Water Demand Side Bypass Pipe')
isCndDemandBypassYes = SameString(FldVal(base + cwpCndDemandBypassOff),'Yes')
IF (SameString(FldVal(base + cwpFluidTypeOff),'Water')) THEN
  FluidType = ftWater
ELSEIF (SameString(FldVal(base + cwpFluidTypeOff),'EthyleneGlycol30')) THEN
  FluidType = ftEthyleneGlycol30
ELSEIF (SameString(FldVal(base + cwpFluidTypeOff),'EthyleneGlycol40')) THEN
  FluidType = ftEthyleneGlycol40
ELSEIF (SameString(FldVal(base + cwpFluidTypeOff),'EthyleneGlycol50')) THEN
  FluidType = ftEthyleneGlycol50
ELSEIF (SameString(FldVal(base + cwpFluidTypeOff),'EthyleneGlycol60')) THEN
  FluidType = ftEthyleneGlycol60
ELSEIF (SameString(FldVal(base + cwpFluidTypeOff),'PropyleneGlycol30')) THEN
  FluidType = ftPropyleneGlycol30
ELSEIF (SameString(FldVal(base + cwpFluidTypeOff),'PropyleneGlycol40')) THEN
  FluidType = ftPropyleneGlycol40
ELSEIF (SameString(FldVal(base + cwpFluidTypeOff),'PropyleneGlycol50')) THEN
  FluidType = ftPropyleneGlycol50
ELSEIF (SameString(FldVal(base + cwpFluidTypeOff),'PropyleneGlycol60')) THEN
  FluidType = ftPropyleneGlycol60
ELSE
  CALL WriteError('Invalid choice in HVACTemplate:Plant:ChilledWaterLoop in the Fluid Type field')
END IF
isMinOutTempBlank = SameString(FldVal(base + cwpMinTOutOff),'')
IF (SameString(FldVal(base + cwpPriChWPumpTypeOff),'SinglePump')) THEN
  isChwPrimPumpHeader = .FALSE.
  isChwPrimPumpBranch = .FALSE.
ELSEIF (SameString(FldVal(base + cwpPriChWPumpTypeOff),'PumpPerChiller')) THEN
  isChwPrimPumpHeader = .FALSE.
  isChwPrimPumpBranch = .TRUE.
ELSEIF (SameString(FldVal(base + cwpPriChWPumpTypeOff),'TwoHeaderedPumps')) THEN
  isChwPrimPumpHeader = .TRUE.
  isChwPrimPumpBranch = .FALSE.
  pumpCountChwPrim = '2'
ELSEIF (SameString(FldVal(base + cwpPriChWPumpTypeOff),'ThreeHeaderedPumps')) THEN
  isChwPrimPumpHeader = .TRUE.
  isChwPrimPumpBranch = .FALSE.
  pumpCountChwPrim = '3'
ELSEIF (SameString(FldVal(base + cwpPriChWPumpTypeOff),'FourHeaderedPumps')) THEN
  isChwPrimPumpHeader = .TRUE.
  isChwPrimPumpBranch = .FALSE.
  pumpCountChwPrim = '4'
ELSEIF (SameString(FldVal(base + cwpPriChWPumpTypeOff),'FiveHeaderedPumps')) THEN
  isChwPrimPumpHeader = .TRUE.
  isChwPrimPumpBranch = .FALSE.
  pumpCountChwPrim = '5'
ELSE
  CALL WriteError('Invalid choice in HVACTemplate:Plant:ChilledWaterLoop in the Chilled Water Primary Pump Type field')
END IF
IF (SameString(FldVal(base + cwpSecChWPumpTypeOff),'SinglePump')) THEN
  isChwSecPumpHeader = .FALSE.
ELSEIF (SameString(FldVal(base + cwpSecChWPumpTypeOff),'TwoHeaderedPumps')) THEN
  isChwSecPumpHeader = .TRUE.
  pumpCountChwSec = '2'
ELSEIF (SameString(FldVal(base + cwpSecChWPumpTypeOff),'ThreeHeaderedPumps')) THEN
  isChwSecPumpHeader = .TRUE.
  pumpCountChwSec = '3'
ELSEIF (SameString(FldVal(base + cwpSecChWPumpTypeOff),'FourHeaderedPumps')) THEN
  isChwSecPumpHeader = .TRUE.
  pumpCountChwSec = '4'
ELSEIF (SameString(FldVal(base + cwpSecChWPumpTypeOff),'FiveHeaderedPumps')) THEN
  isChwSecPumpHeader = .TRUE.
  pumpCountChwSec = '5'
ELSE
  CALL WriteError('Invalid choice in HVACTemplate:Plant:ChilledWaterLoop in the Chilled Water Secondary Pump Type field')
END IF
IF (SameString(FldVal(base + cwpCndPumpTypeOff),'SinglePump')) THEN
  isCndPumpHeader = .FALSE.
  isCndPumpBranch = .FALSE.
ELSEIF (SameString(FldVal(base + cwpCndPumpTypeOff),'PumpPerTower')) THEN
  isCndPumpHeader = .FALSE.
  isCndPumpBranch = .TRUE.
ELSEIF (SameString(FldVal(base + cwpCndPumpTypeOff),'TwoHeaderedPumps')) THEN
  isCndPumpHeader = .TRUE.
  isCndPumpBranch = .FALSE.
  pumpCountCnd = '2'
ELSEIF (SameString(FldVal(base + cwpCndPumpTypeOff),'ThreeHeaderedPumps')) THEN
  isCndPumpHeader = .TRUE.
  isCndPumpBranch = .FALSE.
  pumpCountCnd = '3'
ELSEIF (SameString(FldVal(base + cwpCndPumpTypeOff),'FourHeaderedPumps')) THEN
  isCndPumpHeader = .TRUE.
  isCndPumpBranch = .FALSE.
  pumpCountCnd = '4'
ELSEIF (SameString(FldVal(base + cwpCndPumpTypeOff),'FiveHeaderedPumps')) THEN
  isCndPumpHeader = .TRUE.
  isCndPumpBranch = .FALSE.
  pumpCountCnd = '5'
ELSE
  CALL WriteError('Invalid choice in HVACTemplate:Plant:ChilledWaterLoop in the Condenser Water Pump Type field')
END IF

!Sort the chillers by order of priority
CALL ClearNumSortList
DO iChiller = 1, numCompactChiller
  chillerBase =  compactChillerBase(iChiller)  !store the base pointer to each chiller
  CALL AddToNumSortList(StringToReal(FldVal(chillerBase + chlPriorityOff)),chillerBase)
END DO
DO iChiller = 1, numCompactChillerOR
  chillerBase =  compactChillerORBase(iChiller)  !store the base pointer to each chiller
  CALL AddToNumSortList(StringToReal(FldVal(chillerBase + chlorPriorityOff)),chillerBase)
END DO
CALL SortIncreasingNumList
! write idf objects
!PLANT SIZING ~line 4
CALL CreateNewObj('Sizing:Plant')
CALL AddToObjFld('Plant or Condenser Loop Name', base + cwpNameOff,' Chilled Water Loop')
CALL AddToObjStr('Loop Type','Cooling')
CALL AddToObjFld('Design loop exit temperature {C}', base + cwpChWSetPtDesignOff,'')
CALL AddToObjFld('Design loop delta T {deltaC}',base + cwpDesDeltaTOff,'',.TRUE.)
!PLANT LOOP ~line 10
CALL CreateNewObj('PlantLoop')
CALL AddToObjFld('Name', base + cwpNameOff,' Chilled Water Loop')
IF (fluidType .EQ. ftWater) THEN
  CALL AddToObjStr('Fluid Type','Water')
  CALL AddToObjStr('User Defined Fluid Type','')
ELSE
  CALL AddToObjStr('Fluid Type','UserDefinedFluidType')
  CALL AddToObjFld('User Defined Fluid Type', base + cwpNameOff,' Loop Fluid')
END IF
IF (isCwpChWOperTypeUserDefined) THEN
  CALL AddToObjFld('PlantEquipmentOperationSchemes Name', base + cwpChWOperNameOff,'')
ELSE
  CALL AddToObjFld('PlantEquipmentOperationSchemes Name', base + cwpNameOff,' Chiller Operation')
END IF
CALL AddToObjFld('Loop Temperature Setpoint Node Name', base + cwpNameOff,' ChW Supply Outlet')
CALL AddToObjStr('Maximum Loop Temperature {C}','98')
CALL AddToObjStr('Minimum Loop Temperature {C}','1')
CALL AddToObjStr('Maximum Loop Volumetric Flow Rate {m3/s}','autosize')
CALL AddToObjStr('Minimum Loop Volumetric Flow Rate {m3/s}','0')
CALL AddToObjStr('volume of the plant loop {m3}','autosize')
CALL AddToObjFld('Plant Side Inlet Node Name', base + cwpNameOff,' ChW Supply Inlet')
CALL AddToObjFld('Plant Side Outlet Node Name', base + cwpNameOff,' ChW Supply Outlet')
CALL AddToObjFld('Plant Side Branch List Name', base + cwpNameOff,' ChW Supply Side Branches')
CALL AddToObjFld('Plant Side Connector List Name', base + cwpNameOff,' ChW Supply Side Connectors')
CALL AddToObjFld('Demand Side Inlet Node Name', base + cwpNameOff,' ChW Demand Inlet')
CALL AddToObjFld('Demand Side Outlet Node Name', base + cwpNameOff,' ChW Demand Outlet')
CALL AddToObjFld('Demand Side Branch List Name', base + cwpNameOff,' ChW Demand Side Branches')
CALL AddToObjFld('Demand Side Connector List Name', base + cwpNameOff,' ChW Demand Side Connectors')
CALL AddToObjFld('Load Distribution Scheme', base + cwpChwLdDistSchmOff,'')
IF (isMinOutTempBlank) THEN
  CALL AddToObjStr('AvailabilityManagerAssignmentList','')
ELSE
  CALL AddToObjFld('AvailabilityManagerAssignmentList', base + cwpNameOff,' Availability List')
END IF
CALL AddToObjStr('Plant Loop Demand Calculation Scheme','SingleSetpoint')
IF ((loopPumpConfig .EQ. pumpConstPriVarSec) .OR. (loopPumpConfig .EQ. pumpVarPriConstSec)) THEN
  CALL AddToObjStr('Common Pipe Simulation','CommonPipe',.TRUE.)
ELSE
  CALL AddToObjStr('Common Pipe Simulation','None',.TRUE.)
END IF
IF (.NOT. isCwpChWOperTypeUserDefined) THEN
  !PLANT OPERATION SCHEMES ~line 37
  CALL CreateNewObj('PlantEquipmentOperationSchemes')
  CALL AddToObjFld('Name', base + cwpNameOff,' Chiller Operation')
  CALL AddToObjStr('Control Scheme Object Type','PlantEquipmentOperation:CoolingLoad')
  CALL AddToObjFld('Control Scheme Name', base + cwpNameOff,' Chiller Operation All Hours')
  CALL AddToObjStr('Control Scheme Schedule Name','HVACTemplate-Always 1',.TRUE.)
  CALL AddAlwaysSchedule('1')
  !COOLING LOAD RANGE BASED OPERATION ~line 43
  CALL CreateNewObj('PlantEquipmentOperation:CoolingLoad')
  CALL AddToObjFld('Name', base + cwpNameOff,' Chiller Operation All Hours')
  CALL AddToObjStr('Load Range Lower Limit 1 {W}','0')
  CALL AddToObjStr('Load Range Upper Limit 1 {W}','1000000000000000')
  CALL AddToObjFld('Priority Control 1 Equipment List Name', base + cwpNameOff,' All Chillers',.TRUE.)
  !PLANT EQUIPMENT LIST ~line 49
  CALL CreateNewObj('PlantEquipmentList')
  CALL AddToObjFld('Name', base + cwpNameOff,' All Chillers')
  !"Repeat this line for every COMPACT HVAC:PLANT:CHILLER object in order of chlPriorityOff from smallest to largest
  !If chlTypeOff=""Purchased Chilled Water"" then this field is ""PURCHASED:CHILLED WATER""
  !If chlTypeOff=""Electric Centrifugal Chiller"" then this field is ""CHILLER:ELECTRIC:EIR""
  !If chlTypeOff=""Electric Reciprocating Chiller"" then this field is ""CHILLER:ELECTRIC:EIR"""
  DO iChiller = 1, (numCompactChiller + numCompactChillerOR)
    chillerBase = NumSortList(iChiller)%sortedPtr
    isEndOfList = (iChiller .EQ. (numCompactChiller + numCompactChillerOR))
    IF (SameString(FldVal(chillerBase),'HVACTemplate:Plant:Chiller')) THEN
      IF (SameString(FldVal(chillerBase + chlTypeOff),'DistrictChilledWater')) THEN
        CALL AddToObjStr('Equipment Object Type', 'DistrictCooling')
      ELSEIF (SameString(FldVal(chillerBase + chlTypeOff),'ElectricCentrifugalChiller')) THEN
        CALL AddToObjStr('Equipment Object Type', 'Chiller:Electric:EIR')
      ELSEIF (SameString(FldVal(chillerBase + chlTypeOff),'ElectricReciprocatingChiller')) THEN
        CALL AddToObjStr('Equipment Object Type', 'Chiller:Electric:EIR')
      ELSEIF (SameString(FldVal(chillerBase + chlTypeOff),'ElectricScrewChiller')) THEN
        CALL AddToObjStr('Equipment Object Type', 'Chiller:Electric:EIR')
      ELSE
        CALL AddToObjStr('Equipment Object Type', 'Chiller:Electric:EIR')
      END IF
      CALL AddToObjFld('Equipment Name', chillerBase + chlNameOff,'',isEndOfList)
    ELSEIF (SameString(FldVal(chillerBase),'HVACTemplate:Plant:Chiller:ObjectReference')) THEN
      CALL AddToObjFld('Equipment Object Type', chillerBase + chlorObjTypeOff,'')
      CALL AddToObjFld('Equipment Name', chillerBase + chlorObjNameOff,'',isEndOfList)
    ENDIF
  END DO
END IF

  ! Variable flow chillers need setpoints, make a nodelist regardless of flow type
CALL CreateNewObj('NodeList')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Setpoint Nodes')
!use list created in chiller object
DO iStr = 1,numStrList
  IF(StrList(iStr)%Handle .EQ. handleChillerChWOutletNode) THEN
    CALL AddToObjStr('Node Name', StrList(iStr)%StrToSave)
  END IF
END DO
CALL AddToObjFld('Node Name', base + cwpNameOff,' ChW Supply Outlet',.TRUE.)

IF (setPtResetType .EQ. sprtNone) THEN
  !If  cwpChWSetPtSchedNameOff is blank, then create schedule of cwpChWSetPtDesignOff
  IF (FldVal(base + cwpChWSetPtSchedNameOff) .EQ. '') THEN
    isCwpWSetPtSchedName = .FALSE.
    CALL AddAlwaysSchedule(FldVal(base + cwpChWSetPtDesignOff))
  ELSE
    isCwpWSetPtSchedName = .TRUE.
  END IF
!SET POINT MANAGER:SCHEDULED ~line 54
  IF (.NOT. isCwpWSetPtSchedName) THEN
    CALL AddAlwaysSchedule(FldVal(base + cwpChWSetPtDesignOff))
  END IF
  CALL CreateNewObj('SetpointManager:Scheduled')
  CALL AddToObjFld('Name', base + cwpNameOff,' ChW Temp Manager')
  CALL AddToObjStr('Control Variable','Temperature')
  IF (isCwpWSetPtSchedName) THEN
    CALL AddToObjFld('Schedule Name', base + cwpChWSetPtSchedNameOff,'')
  ELSE
    CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + cwpChWSetPtDesignOff)))
  END IF
  CALL AddToObjFld('Setpoint Node or NodeList Name', base + cwpNameOff,' ChW Supply Setpoint Nodes',.TRUE.)
ELSEIF (setPtResetType .EQ. sprtOutdoor) THEN
  CALL CreateNewObj('SetpointManager:OutdoorAirReset')
  CALL AddToObjFld('Name', base + cwpNameOff,' ChW Temp Manager')
  CALL AddToObjStr('Control Variable','Temperature')
  CALL AddToObjFld('Set point at outside low {C}', base + cwpChWSetPtAtODBLoOff,'')
  CALL AddToObjFld('Outside low {C}', base + cwpChWODBLoOff,'')
  CALL AddToObjFld('Set point at outside high {C}', base + cwpChWSetPtAtODBHiOff,'')
  CALL AddToObjFld('Outside high {C}', base + cwpChWODBHiOff,'')
  CALL AddToObjFld('Setpoint Node or NodeList Name', base + cwpNameOff,' ChW Supply Setpoint Nodes',.TRUE.)
END IF
!BRANCH LIST ~line 60
CALL CreateNewObj('BranchList')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Side Branches')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Inlet Branch')
DO iChiller = 1, (numCompactChiller + numCompactChillerOR)
  IF (SameString(FldVal(NumSortList(iChiller)%sortedPtr),'HVACTemplate:Plant:Chiller')) THEN
    CALL AddToObjFld('Name', NumSortList(iChiller)%sortedPtr + chlNameOff,' ChW Branch')
  ELSEIF (SameString(FldVal(NumSortList(iChiller)%sortedPtr),'HVACTemplate:Plant:Chiller:ObjectReference')) THEN
    CALL AddToObjFld('Name', NumSortList(iChiller)%sortedPtr + chlorObjNameOff,' ChW Branch')
  ENDIF
END DO
IF (isChWSupplyBypassYes) THEN
  CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Bypass Branch')
END IF
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Outlet Branch',.TRUE.)
!CONNECTOR LIST ~ line 67
CALL CreateNewObj('ConnectorList')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Side Connectors')
CALL AddToObjStr('Connector Object Type','Connector:Splitter')
CALL AddToObjFld('Connector Name', base + cwpNameOff,' ChW Supply Splitter')
CALL AddToObjStr('Connector Object Type','Connector:Mixer')
CALL AddToObjFld('Connector Name', base + cwpNameOff,' ChW Supply Mixer',.TRUE.)
!SPLITTER ~ line 74
CALL CreateNewObj('Connector:Splitter')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Splitter')
CALL AddToObjFld('Inlet Branch Name', base + cwpNameOff,' ChW Supply Inlet Branch')
IF (isChWSupplyBypassYes) THEN
  CALL AddToObjFld('Outlet Branch Name', base + cwpNameOff,' ChW Supply Bypass Branch')
END IF
DO iChiller = 1, (numCompactChiller + numCompactChillerOR)
  isEndOfList = ( iChiller .EQ. (numCompactChiller + numCompactChillerOR))
  IF (SameString(FldVal(NumSortList(iChiller)%sortedPtr),'HVACTemplate:Plant:Chiller')) THEN
    CALL AddToObjFld('Outlet Branch Name', NumSortList(iChiller)%sortedPtr + chlNameOff,' ChW Branch', isEndOfList)
  ELSEIF (SameString(FldVal(NumSortList(iChiller)%sortedPtr),'HVACTemplate:Plant:Chiller:ObjectReference')) THEN
    CALL AddToObjFld('Outlet Branch Name', NumSortList(iChiller)%sortedPtr + chlorObjNameOff,' ChW Branch', isEndOfList)
  ENDIF
END DO

!MIXER ~ line 80
CALL CreateNewObj('Connector:Mixer')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Mixer')
CALL AddToObjFld('Outlet Branch Name', base + cwpNameOff,' ChW Supply Outlet Branch')
IF (isChWSupplyBypassYes) THEN
  CALL AddToObjFld('Inlet Branch Name', base + cwpNameOff,' ChW Supply Bypass Branch')
END IF
DO iChiller = 1, (numCompactChiller + numCompactChillerOR)
  isEndOfList = (iChiller .EQ. (numCompactChiller + numCompactChillerOR))
  IF (SameString(FldVal(NumSortList(iChiller)%sortedPtr),'HVACTemplate:Plant:Chiller')) THEN
    CALL AddToObjFld('Inlet Branch Name', NumSortList(iChiller)%sortedPtr + chlNameOff,' ChW Branch', isEndOfList)
  ELSEIF (SameString(FldVal(NumSortList(iChiller)%sortedPtr),'HVACTemplate:Plant:Chiller:ObjectReference')) THEN
    CALL AddToObjFld('Inlet Branch Name', NumSortList(iChiller)%sortedPtr + chlorObjNameOff,' ChW Branch', isEndOfList)
  ENDIF
END DO
! This is out of order since pump logic always requires this branch and pipe
!BRANCH ~ line 108
IF (isChWSupplyBypassYes) THEN
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Bypass Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
  CALL AddToObjFld('Component Name', base + cwpNameOff,' ChW Supply Side Bypass Pipe')
  CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' ChW Supply Bypass Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' ChW Supply Bypass Outlet',.TRUE.)
  !PIPE ~ line 117
  CALL CreateNewObj('Pipe:Adiabatic')
  CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Side Bypass Pipe')
  CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Supply Bypass Inlet')
  CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Supply Bypass Outlet',.TRUE.)
END IF
! Primary branch pumps
IF (isChwPrimPumpBranch) THEN
  !if branch pumps are used just insert a pipe in the supply inlet branch
  CALL CreateNewObj('Pipe:Adiabatic')
  CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Side Inlet Pipe')
  CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Supply Inlet')
  CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Supply Pipe Outlet',.TRUE.)
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Inlet Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
  CALL AddToObjFld('Component Name', base + cwpNameOff,' ChW Supply Side Inlet Pipe')
  CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' ChW Supply Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' ChW Supply Pipe Outlet',.TRUE.)
ELSE
  IF ((loopPumpConfig .EQ. pumpConstant) .OR. (loopPumpConfig .EQ. pumpConstPriVarSec)) THEN
    !PUMP:CONSTANT SPEED ~ line 86
    IF (.NOT. isChwPrimPumpHeader) THEN
      CALL CreateNewObj('Pump:ConstantSpeed')
      IF (loopPumpConfig .EQ. pumpConstant) THEN
        CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Pump')
        CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Supply Inlet')
        CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Pump Outlet')
      ELSEIF (loopPumpConfig .EQ. pumpConstPriVarSec) THEN
        CALL AddToObjFld('Name', base + cwpNameOff,' ChW Primary Supply Pump')
        CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Supply Inlet')
        CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Primary Pump Outlet')
      END IF
      CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Rated Pump Head {Pa}', base + cwpPriChWPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      CALL AddToObjFld('Pump Control Type', base + cwpPumpControlTypeOff,'')
      IF (isCwpPumpSchedBlank) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', base + cwpPumpSchedNameOff,'',.TRUE.)
      END IF
    ELSE
      CALL CreateNewObj('HeaderedPumps:ConstantSpeed')
      IF (loopPumpConfig .EQ. pumpConstant) THEN
        CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Pump')
        CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Supply Inlet')
        CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Pump Outlet')
      ELSEIF (loopPumpConfig .EQ. pumpConstPriVarSec) THEN
        CALL AddToObjFld('Name', base + cwpNameOff,' ChW Primary Supply Pump')
        CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Supply Inlet')
        CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Primary Pump Outlet')
      END IF
      CALL AddToObjStr('Total Rated Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Number of Pumps in Bank',pumpCountChwPrim)
      CALL AddToObjStr('Flow Sequencing Control Scheme','Sequential')
      CALL AddToObjFld('Rated Pump Head {Pa}', base + cwpPriChWPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      CALL AddToObjFld('Pump Control Type', base + cwpPumpControlTypeOff,'')
      IF (isCwpPumpSchedBlank) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', base + cwpPumpSchedNameOff,'',.TRUE.)
      END IF
    END IF
    !BRANCH ~ line 99
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Inlet Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isChwPrimPumpHeader) THEN
      CALL AddToObjStr('Component Object Type','Pump:ConstantSpeed')
    ELSE
      CALL AddToObjStr('Component Object Type','HeaderedPumps:ConstantSpeed')
    END IF
    IF (loopPumpConfig .EQ. pumpConstant) THEN
      CALL AddToObjFld('Component Name', base + cwpNameOff,' ChW Supply Pump')
      CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' ChW Supply Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' ChW Pump Outlet',.TRUE.)
    ELSEIF (loopPumpConfig .EQ. pumpConstPriVarSec) THEN
      CALL AddToObjFld('Name', base + cwpNameOff,' ChW Primary Supply Pump')
      CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' ChW Supply Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' ChW Primary Pump Outlet',.TRUE.)
    END IF
  ELSEIF ((loopPumpConfig .EQ. pumpVariable) .OR. (loopPumpConfig .EQ. pumpVarPriConstSec)) THEN
    !PUMP:VARIABLE SPEED ~ line 122
    CALL WriteComment('Pump part load coefficients based on simple parabola to represent a generic VSD drive.')
    IF (.NOT. isChwPrimPumpHeader) THEN
      CALL CreateNewObj('Pump:VariableSpeed')
      IF (loopPumpConfig .EQ. pumpVariable) THEN
        CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Pump')
        CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Supply Inlet')
        CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Pump Outlet')
      ELSEIF (loopPumpConfig .EQ. pumpVarPriConstSec) THEN
        CALL AddToObjFld('Name', base + cwpNameOff,' ChW Primary Supply Pump')
        CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Supply Inlet')
        CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Primary Pump Outlet')
      END IF
      CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Rated Pump Head {Pa}', base + cwpPriChWPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      !Curve based on VSD generic curve from user with adjustments to prevent neg values - Nov 2008
      !Can use something better when min flow rate is able to be autosized or entered as a fraction
      CALL AddToObjStr('Coefficient 1 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 2 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 3 of the Part Load Performance Curve','1.0')
      CALL AddToObjStr('Coefficient 4 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Min Flow Rate while operating in variable flow capacity {m3/s}','0')
      CALL AddToObjFld('Pump Control Type', base + cwpPumpControlTypeOff,'')
      IF (isCwpPumpSchedBlank) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', base + cwpPumpSchedNameOff,'',.TRUE.)
      END IF
    ELSE
      CALL CreateNewObj('HeaderedPumps:VariableSpeed')
      IF (loopPumpConfig .EQ. pumpVariable) THEN
        CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Pump')
        CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Supply Inlet')
        CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Pump Outlet')
      ELSEIF (loopPumpConfig .EQ. pumpVarPriConstSec) THEN
        CALL AddToObjFld('Name', base + cwpNameOff,' ChW Primary Supply Pump')
        CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Supply Inlet')
        CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Primary Pump Outlet')
      END IF
      CALL AddToObjStr('Total Rated Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Number of Pumps in Bank',pumpCountChwPrim)
      CALL AddToObjStr('Flow Sequencing Control Scheme','Sequential')
      CALL AddToObjFld('Rated Pump Head {Pa}', base + cwpPriChWPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      !Curve based on VSD generic curve from user with adjustments to prevent neg values - Nov 2008
      !Can use something better when min flow rate is able to be autosized or entered as a fraction
      CALL AddToObjStr('Coefficient 1 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 2 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 3 of the Part Load Performance Curve','1.0')
      CALL AddToObjStr('Coefficient 4 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Minimum Flow Rate {m3/s}','0')
      CALL AddToObjFld('Pump Control Type', base + cwpPumpControlTypeOff,'')
      IF (isCwpPumpSchedBlank) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', base + cwpPumpSchedNameOff,'',.TRUE.)
      END IF
    END IF
    !BRANCH ~ line 144
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Inlet Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isChwPrimPumpHeader) THEN
      CALL AddToObjStr('Component Object Type','Pump:VariableSpeed')
    ELSE
      CALL AddToObjStr('Component Object Type','HeaderedPumps:VariableSpeed')
    END IF
    IF (loopPumpConfig .EQ. pumpVariable) THEN
      CALL AddToObjFld('Component Name', base + cwpNameOff,' ChW Supply Pump')
      CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' ChW Supply Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' ChW Pump Outlet',.TRUE.)
    ELSEIF (loopPumpConfig .EQ. pumpVarPriConstSec) THEN
      CALL AddToObjFld('Name', base + cwpNameOff,' ChW Primary Supply Pump')
      CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' ChW Supply Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' ChW Primary Pump Outlet',.TRUE.)
    END IF
  END IF
END IF
!BRANCH ~ line 153
CALL CreateNewObj('Branch')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Outlet Branch')
CALL AddToObjStr('Pressure Drop Curve Name','')
CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
CALL AddToObjFld('Component Name', base + cwpNameOff,' ChW Supply Outlet Pipe')
CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' ChW Supply Outlet Pipe Inlet')
CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' ChW Supply Outlet',.TRUE.)
!PIPE ~ line 162
CALL CreateNewObj('Pipe:Adiabatic')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Supply Outlet Pipe')
CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Supply Outlet Pipe Inlet')
CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Supply Outlet',.TRUE.)
!deleted some line so line numbers may be too high after this point.
!BRANCH LIST ~ line 185
CALL CreateNewObj('BranchList')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Demand Side Branches')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Demand Inlet Branch')
!Repeat this line for every chilled water coil branch "* ChW Branch".
!Currently for COMPACT HVAC:SYSTEM:VAV, line 327 vsAirHandlerNameOff&" Cooling Coil ChW Branch"
DO iStr = 1, numStrList
  IF (StrList(iStr)%Handle .EQ. handleCoolingCoilBranch) THEN
    CALL AddToObjStr('Name', StrList(iStr)%StrToSave)
  END IF
END DO
IF (isChWDemandBypassYes) THEN
  CALL AddToObjFld('Name', base + cwpNameOff,' ChW Demand Bypass Branch')
END IF
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Demand Outlet Branch',.TRUE.)
!CONNECTOR LIST ~ line 192
CALL CreateNewObj('ConnectorList')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Demand Side Connectors')
CALL AddToObjStr('Connector Object Type','Connector:Splitter')
CALL AddToObjFld('Connector Name', base + cwpNameOff,' ChW Demand Splitter')
CALL AddToObjStr('Connector Object Type','Connector:Mixer')
CALL AddToObjFld('Connector Name', base + cwpNameOff,' ChW Demand Mixer',.TRUE.)
!SPLITTER ~ line 199
CALL CreateNewObj('Connector:Splitter')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Demand Splitter')
CALL AddToObjFld('Inlet Branch Name', base + cwpNameOff,' ChW Demand Inlet Branch')
IF (isChWDemandBypassYes) THEN
  CALL AddToObjFld('Outlet Branch Name', base + cwpNameOff,' ChW Demand Bypass Branch')
END IF
!determine the last branch since it will require a semicolon after the object
DO iStr = 1, numStrList
  IF (StrList(iStr)%Handle .EQ. handleCoolingCoilBranch) THEN
    lastBranch = iStr
  END IF
END DO
DO iStr = 1, numStrList
  IF (StrList(iStr)%Handle .EQ. handleCoolingCoilBranch) THEN
    CALL AddToObjStr('Outlet Branch Name', StrList(iStr)%StrToSave, iStr .EQ. lastBranch)
  END IF
END DO
!MIXER ~ line 205
CALL CreateNewObj('Connector:Mixer')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Demand Mixer')
CALL AddToObjFld('Outlet Branch Name', base + cwpNameOff,' ChW Demand Outlet Branch')
IF (isChWDemandBypassYes) THEN
  CALL AddToObjFld('Inlet Branch Name', base + cwpNameOff,' ChW Demand Bypass Branch')
END IF
DO iStr = 1, numStrList
  IF (StrList(iStr)%Handle .EQ. handleCoolingCoilBranch) THEN
    CALL AddToObjStr('Inlet Branch Name', StrList(iStr)%StrToSave, iStr .EQ. lastBranch)
  END IF
END DO
! begin section on secondary (coil) side of chilled water loop.
!BRANCH ~ line 211
CALL CreateNewObj('Branch')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Demand Inlet Branch')
CALL AddToObjStr('Pressure Drop Curve Name','')
IF (loopPumpConfig .EQ. pumpVarPriConstSec) THEN
  IF (.NOT. isChwSecPumpHeader) THEN
    CALL AddToObjStr('Component Object Type','Pump:ConstantSpeed')
  ELSE
    CALL AddToObjStr('Component Object Type','HeaderedPumps:ConstantSpeed')
  END IF
  CALL AddToObjFld('Component Name', base + cwpNameOff,' ChW Secondary Pump')
  CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' ChW Demand Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' ChW Secondary Pump Outlet',.TRUE.)
ELSEIF (loopPumpConfig .EQ. pumpConstPriVarSec) THEN
  IF (.NOT. isChwSecPumpHeader) THEN
    CALL AddToObjStr('Component Object Type','Pump:VariableSpeed')
  ELSE
    CALL AddToObjStr('Component Object Type','HeaderedPumps:VariableSpeed')
  END IF
  CALL AddToObjFld('Component Name', base + cwpNameOff,' ChW Secondary Pump')
  CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' ChW Demand Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' ChW Secondary Pump Outlet',.TRUE.)
ELSEIF ((loopPumpConfig .EQ. pumpConstant) .OR. (loopPumpConfig .EQ. pumpVariable)) THEN
  CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
  CALL AddToObjFld('Component Name', base + cwpNameOff,' ChW Demand Inlet Pipe')
  CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' ChW Demand Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' ChW Demand Inlet Pipe Outlet',.TRUE.)
END IF
IF (loopPumpConfig .EQ. pumpVarPriConstSec) THEN
  IF (.NOT. isChwSecPumpHeader) THEN
    CALL CreateNewObj('Pump:ConstantSpeed')
    CALL AddToObjFld('Name', base + cwpNameOff,' ChW Secondary Pump')
    CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Demand Inlet')
    CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Secondary Pump Outlet')
    CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Rated Pump Head {Pa}', base + cwpSecChWPumpHeadOff,'')
    CALL AddToObjStr('Rated Power Consumption {W}','autosize')
    CALL AddToObjStr('Motor Efficiency','0.9')
    CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
    CALL AddToObjFld('Pump Control Type', base + cwpPumpControlTypeOff,'')
    IF (isCwpPumpSchedBlank) THEN
      CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
    ELSE
      CALL AddToObjFld('Pump Flow Rate Schedule Name', base + cwpPumpSchedNameOff,'',.TRUE.)
    END IF
  ELSE
    CALL CreateNewObj('HeaderedPumps:ConstantSpeed')
    CALL AddToObjFld('Name', base + cwpNameOff,' ChW Secondary Pump')
    CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Demand Inlet')
    CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Secondary Pump Outlet')
    CALL AddToObjStr('Total Rated Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Number of Pumps in Bank',pumpCountChwSec)
    CALL AddToObjStr('Flow Sequencing Control Scheme','Sequential')
    CALL AddToObjFld('Rated Pump Head {Pa}', base + cwpSecChWPumpHeadOff,'')
    CALL AddToObjStr('Rated Power Consumption {W}','autosize')
    CALL AddToObjStr('Motor Efficiency','0.9')
    CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
    CALL AddToObjFld('Pump Control Type', base + cwpPumpControlTypeOff,'')
    IF (isCwpPumpSchedBlank) THEN
      CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
    ELSE
      CALL AddToObjFld('Pump Flow Rate Schedule Name', base + cwpPumpSchedNameOff,'',.TRUE.)
    END IF
  END IF
ELSEIF (loopPumpConfig .EQ. pumpConstPriVarSec) THEN
  CALL WriteComment('Pump part load coefficients based on simple parabola to represent a generic VSD drive.')
  IF (.NOT. isChwSecPumpHeader) THEN
    CALL CreateNewObj('Pump:VariableSpeed')
    CALL AddToObjFld('Name', base + cwpNameOff,' ChW Secondary Pump')
    CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Demand Inlet')
    CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Secondary Pump Outlet')
    CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Rated Pump Head {Pa}', base + cwpSecChWPumpHeadOff,'')
    CALL AddToObjStr('Rated Power Consumption {W}','autosize')
    CALL AddToObjStr('Motor Efficiency','0.9')
    CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
    !Curve based on VSD generic curve from user with adjustments to prevent neg values - Nov 2008
    !Can use something better when min flow rate is able to be autosized or entered as a fraction
    CALL AddToObjStr('Coefficient 1 of the Part Load Performance Curve','0.0')
    CALL AddToObjStr('Coefficient 2 of the Part Load Performance Curve','0.0')
    CALL AddToObjStr('Coefficient 3 of the Part Load Performance Curve','1.0')
    CALL AddToObjStr('Coefficient 4 of the Part Load Performance Curve','0.0')
    CALL AddToObjStr('Min Flow Rate while operating in variable flow capacity {m3/s}','0')
    CALL AddToObjFld('Pump Control Type', base + cwpPumpControlTypeOff,'')
    IF (isCwpPumpSchedBlank) THEN
      CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
    ELSE
      CALL AddToObjFld('Pump Flow Rate Schedule Name', base + cwpPumpSchedNameOff,'',.TRUE.)
    END IF
  ELSE
    CALL CreateNewObj('HeaderedPumps:VariableSpeed')
    CALL AddToObjFld('Name', base + cwpNameOff,' ChW Secondary Pump')
    CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Demand Inlet')
    CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Secondary Pump Outlet')
    CALL AddToObjStr('Total Rated Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Number of Pumps in Bank',pumpCountChwSec)
    CALL AddToObjStr('Flow Sequencing Control Scheme','Sequential')
    CALL AddToObjFld('Rated Pump Head {Pa}', base + cwpSecChWPumpHeadOff,'')
    CALL AddToObjStr('Rated Power Consumption {W}','autosize')
    CALL AddToObjStr('Motor Efficiency','0.9')
    CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
    !Curve based on VSD generic curve from user with adjustments to prevent neg values - Nov 2008
    !Can use something better when min flow rate is able to be autosized or entered as a fraction
    CALL AddToObjStr('Coefficient 1 of the Part Load Performance Curve','0.0')
    CALL AddToObjStr('Coefficient 2 of the Part Load Performance Curve','0.0')
    CALL AddToObjStr('Coefficient 3 of the Part Load Performance Curve','1.0')
    CALL AddToObjStr('Coefficient 4 of the Part Load Performance Curve','0.0')
    CALL AddToObjStr('Min Flow Rate while operating in variable flow capacity {m3/s}','0')
    CALL AddToObjFld('Pump Control Type', base + cwpPumpControlTypeOff,'')
    IF (isCwpPumpSchedBlank) THEN
      CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
    ELSE
      CALL AddToObjFld('Pump Flow Rate Schedule Name', base + cwpPumpSchedNameOff,'',.TRUE.)
    END IF
  END IF
ELSE   ! when loopPumpConfig is pumpConstant or when loopPumpConfig pumpVariable
  !PIPE ~ line 220
  CALL CreateNewObj('Pipe:Adiabatic')
  CALL AddToObjFld('Name', base + cwpNameOff,' ChW Demand Inlet Pipe')
  CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Demand Inlet')
  CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Demand Inlet Pipe Outlet',.TRUE.)
END IF
! SHIFTED LINE NUMBERS AFTER THIS POINT
!BRANCH ~ line 211
IF (isChWDemandBypassYes) THEN
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + cwpNameOff,' ChW Demand Bypass Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
  CALL AddToObjFld('Component Name', base + cwpNameOff,' ChW Demand Side Bypass Pipe')
  CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' ChW Demand Bypass Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' ChW Demand Bypass Outlet',.TRUE.)
  !PIPE ~ line 220
  CALL CreateNewObj('Pipe:Adiabatic')
  CALL AddToObjFld('Name', base + cwpNameOff,' ChW Demand Side Bypass Pipe')
  CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Demand Bypass Inlet')
  CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Demand Bypass Outlet',.TRUE.)
END IF
!BRANCH ~ line 225
CALL CreateNewObj('Branch')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Demand Outlet Branch')
CALL AddToObjStr('Pressure Drop Curve Name','')
CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
CALL AddToObjFld('Component Name', base + cwpNameOff,' ChW Demand Outlet Pipe')
CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' ChW Demand Outlet Pipe Inlet')
CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' ChW Demand Outlet',.TRUE.)
!PIPE ~ line 234
CALL CreateNewObj('Pipe:Adiabatic')
CALL AddToObjFld('Name', base + cwpNameOff,' ChW Demand Outlet Pipe')
CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' ChW Demand Outlet Pipe Inlet')
CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' ChW Demand Outlet',.TRUE.)
! this flag variable is actually set in the chiller routine which must be run first
! for this flag to be set properly.
IF (isChillerWithWaterCooled) THEN
  !Sort the towers by order of priority - only boilers serving a hot water loop, also count boilers on this loop
  numTowersThisLoop = 0
  CALL ClearNumSortList
  DO iTower = 1, numCompactTower
    towerBase =  compactTowerBase(iTower)  !store the base pointer to each tower
    IF (SameString(FldVal(towerBase + twrLoopTypeOff),'ChilledWater') .OR. SameString(FldVal(towerBase + twrLoopTypeOff),'')) THEN
      CALL AddToNumSortList(StringToReal(FldVal(towerBase + twrPriorityOff)),towerBase)
      numTowersThisLoop = numTowersThisLoop + 1
    END IF
  END DO
  DO iTower = 1, numCompactTowerOR
    towerBase =  compactTowerORBase(iTower)  !store the base pointer to each tower
    IF (SameString(FldVal(towerBase + twrorLoopTypeOff),'ChilledWater') .OR. SameString(FldVal(towerBase + twrorLoopTypeOff),'')) &
     THEN
      CALL AddToNumSortList(StringToReal(FldVal(towerBase + twrorPriorityOff)),towerBase)
      numTowersThisLoop = numTowersThisLoop + 1
    END IF
  END DO
  IF (numTowersThisLoop == 0) THEN
    CALL WriteError('Severe:  In HVACTemplate:Plant:ChilledWaterLoop "'//TRIM(FldVal(base + cwpNameOff))//'"'// &
                    ' there is one or more water cooled chiller(s) but there are no towers serving this loop.')
  END IF
  CALL SortIncreasingNumList
  !PLANT SIZING ~ line 242
  CALL CreateNewObj('Sizing:Plant')
  CALL AddToObjFld('Plant or Condenser Loop Name', base + cwpNameOff,' Condenser Water Loop')
  CALL AddToObjStr('Loop Type','Condenser')
  CALL AddToObjFld('Design loop exit temperature {C}', base + cwpCndSetPtDesignOff,'')
  CALL AddToObjStr('Design loop delta T {deltaC}','5.6',.TRUE.)
  !CONDENSER LOOP ~ line 248
  CALL CreateNewObj('CondenserLoop')
  CALL AddToObjFld('Name', base + cwpNameOff,' Condenser Water Loop')
  CALL AddToObjStr('Fluid Type','Water')
  CALL AddToObjStr('User Defined Fluid Type','')
  IF (isCwpCndOperDefault) THEN
    CALL AddToObjFld('Condenser Operation Scheme List Name', base + cwpNameOff,' Condenser Operation')
  ELSE
    CALL AddToObjFld('Condenser Operation Scheme List Name', base + cwpCndOperNameOff,'')
  END IF
! Condenser setpoint change.
!  IF (isCndTempControlWetBulb) THEN
!    CALL AddToObjStr('Condenser Loop Temperature Setpoint Node Name or Reference','AIR')
!  ELSE
!    CALL AddToObjFld('Condenser Loop Temperature Setpoint Node Name or Reference', base + cwpNameOff,' CndW Supply Outlet')
!  END IF
  CALL AddToObjFld('Condenser Loop Temperature Setpoint Node Name or Reference', base + cwpNameOff,' CndW Supply Outlet')
  CALL AddToObjStr('Maximum Loop Temperature {C}','80')
  CALL AddToObjStr('Minimum Loop Temperature {C}','5')
  CALL AddToObjStr('Maximum Loop Volumetric Flow Rate {m3/s}','autosize')
  CALL AddToObjStr('Minimum Loop Volumetric Flow Rate {m3/s}','0')
  CALL AddToObjStr('volume of the plant loop {m3}','autosize')
  CALL AddToObjFld('Condenser Side Inlet Node Name', base + cwpNameOff,' CndW Supply Inlet')
  CALL AddToObjFld('Condenser Side Outlet Node Name', base + cwpNameOff,' CndW Supply Outlet')
  CALL AddToObjFld('Condenser Side Branch List Name', base + cwpNameOff,' CndW Supply Side Branches')
  CALL AddToObjFld('Condenser Side Connector List Name', base + cwpNameOff,' CndW Supply Side Connectors')
  CALL AddToObjFld('Demand Side Inlet Node Name', base + cwpNameOff,' CndW Demand Inlet')
  CALL AddToObjFld('Demand Side Outlet Node Name', base + cwpNameOff,' CndW Demand Outlet')
  CALL AddToObjFld('Condenser Demand Side Branch List Name', base + cwpNameOff,' CndW Demand Side Branches')
  CALL AddToObjFld('Condenser Demand Side Connector List Name', base + cwpNameOff,' CndW Demand Side Connectors')
  CALL AddToObjFld('Load Distribution Scheme', base + cwpCndLdDistSchmOff,'',.TRUE.)
  IF (isCwpCndOperDefault) THEN
    !CONDENSER OPERATION SCHEMES ~ line 270
    CALL CreateNewObj('CondenserEquipmentOperationSchemes')
    CALL AddToObjFld('Name', base + cwpNameOff,' Condenser Operation')
    CALL AddToObjStr('Control Scheme Object Type','PlantEquipmentOperation:CoolingLoad')
    CALL AddToObjFld('Control Scheme Name', base + cwpNameOff,' Condenser Operation All Hours')
    CALL AddToObjStr('Control Scheme Schedule Name','HVACTemplate-Always 1',.TRUE.)
    CALL AddAlwaysSchedule('1')
    !COOLING LOAD RANGE BASED OPERATION ~ line 276
    CALL CreateNewObj('PlantEquipmentOperation:CoolingLoad')
    CALL AddToObjFld('Name', base + cwpNameOff,' Condenser Operation All Hours')
    CALL AddToObjStr('Load Range Lower Limit 1 {W}','0')
    CALL AddToObjStr('Load Range Upper Limit 1 {W}','1000000000000000')
    CALL AddToObjFld('Priority Control 1 Equipment List Name', base + cwpNameOff,' All Condensers',.TRUE.)
    !CONDENSER EQUIPMENT LIST
    CALL CreateNewObj('CondenserEquipmentList')
    CALL AddToObjFld('Name', base + cwpNameOff,' All Condensers')
    !Repeat this line for every COMPACT HVAC:PLANT:TOWER object in order of twrPriorityOff from smallest to largest
    !If twrTypeOff="Single Speed" then this field is "COOLING TOWER:SINGLE SPEED"
    !If twrTypeOff="Two Speed" then this field is "COOLING TOWER:TWO SPEED"
    DO iTower = 1, numTowersThisLoop
      towerBase = NumSortList(iTower)%sortedPtr
      isEndOfList = (iTower .EQ. numTowersThisLoop)
      IF (SameString(FldVal(towerBase),'HVACTemplate:Plant:Tower')) THEN
        IF (SameString(FldVal(towerBase + twrTypeOff),'SingleSpeed')) THEN
          CALL AddToObjStr('Equipment Object Type','CoolingTower:SingleSpeed')
        ELSEIF (SameString(FldVal(towerBase + twrTypeOff),'TwoSpeed')) THEN
          CALL AddToObjStr('Equipment Object Type','CoolingTower:TwoSpeed')
        ELSE
          CALL AddToObjStr('Equipment Object Type','CoolingTower:SingleSpeed')
        END IF
        CALL AddToObjFld('Equipment Name', towerBase + twrNameOff,'',isEndOfList)
      ELSEIF (SameString(FldVal(towerBase),'HVACTemplate:Plant:Tower:ObjectReference')) THEN
        CALL AddToObjFld('Equipment Object Type', towerBase + twrorObjTypeOff,'')
        CALL AddToObjFld('Equipment Name', towerBase + twrorObjNameOff,'',isEndOfList)
      ENDIF
    END DO
  END IF

  ! Make a nodelist, regardless of flow type, to put setpoint on tower outlets
  CALL CreateNewObj('NodeList')
  CALL AddToObjFld('Name', base + mwpNameOff,' CndW Supply Setpoint Nodes')
  !use list created in chiller object
  DO iStr = 1,numStrList
    IF(StrList(iStr)%Handle .EQ. handleTowerOutletNode) THEN
      CALL AddToObjStr('Node Name', StrList(iStr)%StrToSave)
    END IF
  END DO
  CALL AddToObjFld('Node Name', base + mwpNameOff,' CndW Supply Outlet',.TRUE.)

  IF (.NOT. isCndTempControlWetBulb) THEN
    !If  cwpCndWSetPtSchedNameOff is blank, then create schedule of cwpCndSetPtDesignOff
    IF (isCndWSetPtSchedBlank) THEN
      CALL AddAlwaysSchedule(FldVal(base + cwpCndSetPtDesignOff))
    END IF
    !SET POINT MANAGER:SCHEDULED ~ line 287
    IF (isCndWSetPtSchedBlank) THEN
      CALL AddAlwaysSchedule(FldVal(base + cwpCndSetPtDesignOff))
    END IF
    CALL CreateNewObj('SetpointManager:Scheduled')
    CALL AddToObjFld('Name', base + cwpNameOff,' CndW Temp Manager')
    CALL AddToObjStr('Control Variable','Temperature')
    IF (isCndWSetPtSchedBlank) THEN
      CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + cwpCndSetPtDesignOff)))
    ELSE
      CALL AddToObjFld('Schedule Name', base + cwpCndSetPtSchedNameOff,'')
    END IF
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + cwpNameOff,' CndW Supply Setpoint Nodes',.TRUE.)
  ELSE  ! wet bulb temp control
    CALL CreateNewObj('SetpointManager:FollowOutdoorAirTemperature')
    CALL AddToObjFld('Name', base + cwpNameOff,' CndW Temp Manager')
    CALL AddToObjStr('Control Variable','Temperature')
    CALL AddToObjStr('Reference Temperature Type','OutdoorAirWetBulb')
    CALL AddToObjStr('Offset Temperature Difference','0')
    CALL AddToObjStr('Maximum Setpoint Temperature {C}','80')
    CALL AddToObjStr('Minimum Setpoint Temperature {C}','5')
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + cwpNameOff,' CndW Supply Setpoint Nodes',.TRUE.)
  END IF
  !BRANCH LIST ~ line 293
  CALL CreateNewObj('BranchList')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Side Branches')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Inlet Branch')
  DO iTower = 1, (numTowersThisLoop)
    IF (SameString(FldVal(NumSortList(iTower)%sortedPtr),'HVACTemplate:Plant:Tower')) THEN
      CALL AddToObjFld('Name', NumSortList(iTower)%sortedPtr + twrNameOff,' CndW Branch')
    ELSEIF (SameString(FldVal(NumSortList(iTower)%sortedPtr),'HVACTemplate:Plant:Tower:ObjectReference')) THEN
      CALL AddToObjFld('Name', NumSortList(iTower)%sortedPtr + twrorObjNameOff,' CndW Branch')
    ENDIF
  END DO
  IF (isCndSupplyBypassYes) THEN
    CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Bypass Branch')
  END IF
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Outlet Branch',.TRUE.)
  !CONNECTOR LIST ~ line 300
  CALL CreateNewObj('ConnectorList')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Side Connectors')
  CALL AddToObjStr('Connector Object Type','Connector:Splitter')
  CALL AddToObjFld('Connector Name', base + cwpNameOff,' CndW Supply Splitter')
  CALL AddToObjStr('Connector Object Type','Connector:Mixer')
  CALL AddToObjFld('Connector Name', base + cwpNameOff,' CndW Supply Mixer',.TRUE.)
  !SPLITTER ~ line 307
  CALL CreateNewObj('Connector:Splitter')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Splitter')
  CALL AddToObjFld('Inlet Branch Name', base + cwpNameOff,' CndW Supply Inlet Branch')
  IF (isCndSupplyBypassYes) THEN
    CALL AddToObjFld('Outlet Branch Name', base + cwpNameOff,' CndW Supply Bypass Branch')
  END IF
  DO iTower = 1, (numTowersThisLoop)
    isEndOfList = (iTower .EQ. numTowersThisLoop)
    IF (SameString(FldVal(NumSortList(iTower)%sortedPtr),'HVACTemplate:Plant:Tower')) THEN
      CALL AddToObjFld('Outlet Branch Name', NumSortList(iTower)%sortedPtr + twrNameOff,' CndW Branch',isEndOfList)
    ELSEIF (SameString(FldVal(NumSortList(iTower)%sortedPtr),'HVACTemplate:Plant:Tower:ObjectReference')) THEN
      CALL AddToObjFld('Outlet Branch Name', NumSortList(iTower)%sortedPtr + twrorObjNameOff,' CndW Branch',isEndOfList)
    ENDIF
  END DO
  !MIXER ~ line 313
  CALL CreateNewObj('Connector:Mixer')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Mixer')
  CALL AddToObjFld('Outlet Branch Name', base + cwpNameOff,' CndW Supply Outlet Branch')
  IF (isCndSupplyBypassYes) THEN
    CALL AddToObjFld('Inlet Branch Name', base + cwpNameOff,' CndW Supply Bypass Branch')
  END IF
  DO iTower = 1, numTowersThisLoop
    isEndOfList = (iTower .EQ. numTowersThisLoop)
    IF (SameString(FldVal(NumSortList(iTower)%sortedPtr),'HVACTemplate:Plant:Tower')) THEN
      CALL AddToObjFld('Inlet Branch Name', NumSortList(iTower)%sortedPtr + twrNameOff,' CndW Branch',isEndOfList)
    ELSEIF (SameString(FldVal(NumSortList(iTower)%sortedPtr),'HVACTemplate:Plant:Tower:ObjectReference')) THEN
      CALL AddToObjFld('Inlet Branch Name', NumSortList(iTower)%sortedPtr + twrorObjNameOff,' CndW Branch',isEndOfList)
    ENDIF
  END DO
  IF (isCndPumpBranch) THEN
    ! if branch pumps are used just insert a pipe in the supply inlet branch
    CALL CreateNewObj('Pipe:Adiabatic')
    CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Side Inlet Pipe')
    CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' CndW Supply Inlet')
    CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' CndW Supply Pipe Outlet',.TRUE.)
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Inlet Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
    CALL AddToObjFld('Component Name', base + cwpNameOff,' CndW Supply Side Inlet Pipe')
    CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' CndW Supply Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' CndW Supply Pipe Outlet',.TRUE.)
  ELSE
    !PUMP:VARIABLE SPEED ~ line 319
    CALL WriteComment('Pump part load coefficients are linear to represent condenser pumps dedicated to each chiller.')
    IF (.NOT. isCndPumpHeader) THEN
      CALL CreateNewObj('Pump:VariableSpeed')
      CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Pump')
      CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' CndW Supply Inlet')
      CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' CndW Pump Outlet')
      CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Rated Pump Head {Pa}', base + cwpCndPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      ! MJW This pump PLR curve is linear under the assumption that condenser water flow will vary only by
      ! how many chillers are operating (multiple constant speed pumps)
      CALL AddToObjStr('Coefficient 1 of the Part Load Performance Curve','0')
      CALL AddToObjStr('Coefficient 2 of the Part Load Performance Curve','0')
      CALL AddToObjStr('Coefficient 3 of the Part Load Performance Curve','1')
      CALL AddToObjStr('Coefficient 4 of the Part Load Performance Curve','0')
      CALL AddToObjStr('Min Flow Rate while operating in variable flow capacity {m3/s}','0')
      CALL AddToObjFld('Pump Control Type', base + cwpPumpControlTypeOff,'')
      IF (isCwpPumpSchedBlank) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', base + cwpPumpSchedNameOff,'',.TRUE.)
      END IF
    ELSE
      CALL CreateNewObj('HeaderedPumps:VariableSpeed')
      CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Pump')
      CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' CndW Supply Inlet')
      CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' CndW Pump Outlet')
      CALL AddToObjStr('Total Rated Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Number of Pumps in Bank',pumpCountCnd)
      CALL AddToObjStr('Flow Sequencing Control Scheme','Sequential')
      CALL AddToObjFld('Rated Pump Head {Pa}', base + cwpCndPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      CALL AddToObjStr('Coefficient 1 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 2 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 3 of the Part Load Performance Curve','1.0')
      CALL AddToObjStr('Coefficient 4 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Min Flow Rate while operating in variable flow capacity {m3/s}','0')
      CALL AddToObjFld('Pump Control Type', base + cwpPumpControlTypeOff,'')
      IF (isCwpPumpSchedBlank) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', base + cwpPumpSchedNameOff,'',.TRUE.)
      END IF
    END IF
    !BRANCH ~ line 337
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Inlet Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isCndPumpHeader) THEN
      CALL AddToObjStr('Component Object Type','Pump:VariableSpeed')
    ELSE
      CALL AddToObjStr('Component Object Type','HeaderedPumps:VariableSpeed')
    END IF
    CALL AddToObjFld('Component Name', base + cwpNameOff,' CndW Supply Pump')
    CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' CndW Supply Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' CndW Pump Outlet',.TRUE.)
    END IF
  !BRANCH ~ line 346
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Outlet Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
  CALL AddToObjFld('Component Name', base + cwpNameOff,' CndW Supply Outlet Pipe')
  CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' CndW Supply Outlet Pipe Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' CndW Supply Outlet',.TRUE.)
  !PIPE ~ line 355
  CALL CreateNewObj('Pipe:Adiabatic')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Outlet Pipe')
  CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' CndW Supply Outlet Pipe Inlet')
  CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' CndW Supply Outlet',.TRUE.)
  !BRANCH ~ line 374
  IF (isCndSupplyBypassYes) THEN
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Bypass Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
    CALL AddToObjFld('Component Name', base + cwpNameOff,' CndW Supply Side Bypass Pipe')
    CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' CndW Supply Bypass Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' CndW Supply Bypass Outlet',.TRUE.)
    !PIPE ~ line 383
    CALL CreateNewObj('Pipe:Adiabatic')
    CALL AddToObjFld('Name', base + cwpNameOff,' CndW Supply Side Bypass Pipe')
    CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' CndW Supply Bypass Inlet')
    CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' CndW Supply Bypass Outlet',.TRUE.)
  END IF
  !BRANCH LIST ~ line 360
  CALL CreateNewObj('BranchList')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Demand Side Branches')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Demand Inlet Branch')
  !use list created in chiller object
  DO iStr = 1,numStrList
    IF(StrList(iStr)%Handle .EQ. handleCondenserBranch) THEN
      CALL AddToObjStr('Name', StrList(iStr)%StrToSave)
    END IF
  END DO
  IF (isCndDemandBypassYes) THEN
    CALL AddToObjFld('Name', base + cwpNameOff,' CndW Demand Bypass Branch')
  END IF
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Demand Outlet Branch',.TRUE.)
  !CONNECTOR LIST  ~ line 367
  CALL CreateNewObj('ConnectorList')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Demand Side Connectors')
  CALL AddToObjStr('Connector Object Type','Connector:Splitter')
  CALL AddToObjFld('Connector Name', base + cwpNameOff,' CndW Demand Splitter')
  CALL AddToObjStr('Connector Object Type','Connector:Mixer')
  CALL AddToObjFld('Connector Name', base + cwpNameOff,' CndW Demand Mixer',.TRUE.)
  !SPLITTER  ~ line 374
  CALL CreateNewObj('Connector:Splitter')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Demand Splitter')
  CALL AddToObjFld('Inlet Branch Name', base + cwpNameOff,' CndW Demand Inlet Branch')
  IF (isCndDemandBypassYes) THEN
    CALL AddToObjFld('Outlet Branch Name', base + cwpNameOff,' CndW Demand Bypass Branch')
  END IF
  !use list created in chiller object
  DO iStr = 1, numStrList
    IF (StrList(iStr)%Handle .EQ. handleCondenserBranch) THEN
      lastBranch = iStr
    END IF
  END DO
  DO iStr = 1,lastBranch
    IF(StrList(iStr)%Handle .EQ. handleCondenserBranch) THEN
      CALL AddToObjStr('Outlet Branch Name', StrList(iStr)%StrToSave, iStr .EQ. lastBranch)
    END IF
  END DO
  !MIXER ~ line 380
  CALL CreateNewObj('Connector:Mixer')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Demand Mixer')
  CALL AddToObjFld('Outlet Branch Name', base + cwpNameOff,' CndW Demand Outlet Branch')
  IF (isCndDemandBypassYes) THEN
    CALL AddToObjFld('Inlet Branch Name', base + cwpNameOff,' CndW Demand Bypass Branch')
  END IF
  !use list created in chiller object
  DO iStr = 1,lastBranch
    IF(StrList(iStr)%Handle .EQ. handleCondenserBranch) THEN
      CALL AddToObjStr('Inlet Branch Name', StrList(iStr)%StrToSave, iStr .EQ. lastBranch)
    END IF
  END DO
  !BRANCH ~ line 400
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Demand Inlet Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
  CALL AddToObjFld('Component Name', base + cwpNameOff,' CndW Demand Inlet Pipe')
  CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' CndW Demand Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' CndW Demand Inlet Pipe Outlet',.TRUE.)
  !PIPE ~ line 409
  CALL CreateNewObj('Pipe:Adiabatic')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Demand Inlet Pipe')
  CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' CndW Demand Inlet')
  CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' CndW Demand Inlet Pipe Outlet',.TRUE.)
  !Bigger shift after this point for line numbers
  !BRANCH ~ line 386
  IF (isCndDemandBypassYes) THEN
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + cwpNameOff,' CndW Demand Bypass Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
    CALL AddToObjFld('Component Name', base + cwpNameOff,' CndW Demand Side Bypass Pipe')
    CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' CndW Demand Bypass Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' CndW Demand Bypass Outlet',.TRUE.)
    !PIPE ~ line 395
    CALL CreateNewObj('Pipe:Adiabatic')
    CALL AddToObjFld('Name', base + cwpNameOff,' CndW Demand Side Bypass Pipe')
    CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' CndW Demand Bypass Inlet')
    CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' CndW Demand Bypass Outlet',.TRUE.)
  END IF
  !BRANCH ~ line 400
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Demand Outlet Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
  CALL AddToObjFld('Component Name', base + cwpNameOff,' CndW Demand Outlet Pipe')
  CALL AddToObjFld('Component Inlet Node Name', base + cwpNameOff,' CndW Demand Outlet Pipe Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + cwpNameOff,' CndW Demand Outlet',.TRUE.)
  !PIPE ~ line 409
  CALL CreateNewObj('Pipe:Adiabatic')
  CALL AddToObjFld('Name', base + cwpNameOff,' CndW Demand Outlet Pipe')
  CALL AddToObjFld('Inlet Node Name', base + cwpNameOff,' CndW Demand Outlet Pipe Inlet')
  CALL AddToObjFld('Outlet Node Name', base + cwpNameOff,' CndW Demand Outlet',.TRUE.)
END IF
! if the user has specified a Minimum Outdoor Dry Bulb Temperatures
! then add an AvailabilityManagerAssignmentList
!             AvailabilityManager:HighTemperatureTurnOff
!             OutdoorAir:Node
IF (.NOT. isMinOutTempBlank) THEN
  CALL CreateNewObj('AvailabilityManagerAssignmentList')
  CALL AddToObjFld('Name', base + cwpNameOff,' Availability List')
  CALL AddToObjStr('Availability Manager Object Type','AvailabilityManager:LowTemperatureTurnOff')
  CALL AddToObjFld('Availability Manager Name',base + hwpNameOff,' Availability Low Temp Off',.TRUE.)
  CALL CreateNewObj('AvailabilityManager:LowTemperatureTurnOff')
  CALL AddToObjFld('Name',base + cwpNameOff,' Availability Low Temp Off')
  CALL AddToObjFld('Sensor Node Name',base + cwpNameOff,' Outside Air Sensor')
  CALL AddToObjFld('Temperature',base + cwpMinTOutOff,'',.TRUE.)
  CALL CreateNewObj('OutdoorAir:Node')
  CALL AddToObjFld('Name', base + cwpNameOff,' Outside Air Sensor')
  CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
END IF
! Create glycol fluid if not water or steam
IF (fluidType .NE. ftWater) THEN
  CALL CreateNewObj('FluidProperties:GlycolConcentration')
  CALL AddToObjFld('Name', base + cwpNameOff,' Loop Fluid')
  SELECT CASE (fluidType)
    CASE (ftEthyleneGlycol30)
      CALL AddToObjStr('Glycol Type','EthyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.30',.TRUE.)
    CASE (ftEthyleneGlycol40)
      CALL AddToObjStr('Glycol Type','EthyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.40',.TRUE.)
    CASE (ftEthyleneGlycol50)
      CALL AddToObjStr('Glycol Type','EthyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.50',.TRUE.)
    CASE (ftEthyleneGlycol60)
      CALL AddToObjStr('Glycol Type','EthyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.60',.TRUE.)
    CASE (ftPropyleneGlycol30)
      CALL AddToObjStr('Glycol Type','PropyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.30',.TRUE.)
    CASE (ftPropyleneGlycol40)
      CALL AddToObjStr('Glycol Type','PropyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.40',.TRUE.)
    CASE (ftPropyleneGlycol50)
      CALL AddToObjStr('Glycol Type','PropyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.50',.TRUE.)
    CASE (ftPropyleneGlycol60)
      CALL AddToObjStr('Glycol Type','PropyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.60',.TRUE.)
    CASE DEFAULT
      CALL AddToObjStr('Glycol Type','EthyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.30',.TRUE.)
  END SELECT
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE CreateWaterAirHP
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2010
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a water to air heat pump.

          ! METHODOLOGY EMPLOYED:
          !    Based on PTHP code

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for supplementary heating coil type
INTEGER, PARAMETER :: shcElectric = 1
INTEGER, PARAMETER :: shcGas = 2
INTEGER, PARAMETER :: shcWater = 3
          ! parameters for fan mode type
INTEGER, PARAMETER :: fmCycling = 1
INTEGER, PARAMETER :: fmNeither = 3
          ! parameters for baseboard type
INTEGER, PARAMETER :: baseboardNone           = 1
INTEGER, PARAMETER :: baseboardHotWater       = 2
INTEGER, PARAMETER :: baseboardElectric       = 3

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base
INTEGER :: iZone

LOGICAL :: isTStatNameBlank
LOGICAL :: isSupplyCoolRateAutosize
LOGICAL :: isSupplyHeatRateAutosize
LOGICAL :: isSysAvailSchedBlank
INTEGER :: suppHeatCoilKind
INTEGER :: fanMode
LOGICAL :: isDrawThru
LOGICAL :: isCoolCoilCapAutosize
LOGICAL :: isSHRAutosize
LOGICAL :: isDedOutAirNameBlank
CHARACTER(len=1) :: SchType
LOGICAL :: isOATypeDetailed = .FALSE.
LOGICAL :: isBaseboardNone = .FALSE.
INTEGER :: baseboardKind = 0
CHARACTER(len=1) :: NextSequenceNumber = '1'

DO iZone = 1, numCompactWaterAirHP
  base =  compactWaterAirHPBase(iZone)
  ! set flags
  isTStatNameBlank = (FldVal(base + wahpTStatNameOff) .EQ. '')
  isSupplyCoolRateAutosize = SameString(FldVal(base + wahpSupplyCoolFlowRateOff),'autosize')
  isSupplyHeatRateAutosize = SameString(FldVal(base + wahpSupplyHeatFlowRateOff),'autosize')
  isCoolCoilCapAutosize = SameString(FldVal(base + wahpCoolCoilCapOff),'autosize')
  isSHRAutosize = SameString(FldVal(base + wahpCoolCoilSHROff),'autosize')
  isSysAvailSchedBlank = (FldVal(base + wahpSysAvailSchedNameOff) .EQ. '')
  isOATypeDetailed = SameString(FldVal(base + wahpOutAirTypeOff),'DetailedSpecification')
  isBaseboardNone = SameString(FldVal(base + wahpBaseboardTypeOff),'None')
  ! Supply fan mode may be blank (default to Cycling) or a schedule name
  IF (SameString(FldVal(base + wahpFanModeOff),'')) THEN
    fanMode = fmCycling
  ELSE
    fanMode = fmNeither
  END IF
  isDrawThru = SameString(FldVal(base + wahpFanPlacementOff),'DrawThrough')
  isDedOutAirNameBlank = (FldVal(base + wahpDedOutAirSysNameOff) .EQ. '')
  IF (SameString(FldVal(base + wahpSuppHeatCoilKindOff),'Electric')) THEN
    suppHeatCoilKind = shcElectric
  ELSEIF (SameString(FldVal(base + wahpSuppHeatCoilKindOff),'HotWater')) THEN
    suppHeatCoilKind = shcWater
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:WaterToAirHeatPump "'//TRIM(FldVal(base + wahpNameOff))//'"'// &
                    ' in the Supplemental Heating Coil Type field: '//TRIM(FldVal(base + wahpSuppHeatCoilKindOff)))
    suppHeatCoilKind=0
  END IF
  !validate cooling design method
  IF (SameString(FldVal(base +  wahpCoolDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  wahpCoolDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:WaterToAirHeatPump "'//TRIM(FldVal(base + wahpNameOff))//'"'// &
                    ' in the Zone Cooling Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + wahpCoolDesignTempMethodOff)))
  END IF
  !validate heating design method
  IF (SameString(FldVal(base +  wahpHeatDesignTempMethodOff),'SupplyAirTemperature')) THEN
  ELSEIF (SameString(FldVal(base +  wahpHeatDesignTempMethodOff),'TemperatureDifference')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:WaterToAirHeatPump "'//TRIM(FldVal(base + wahpNameOff))//'"'// &
                    ' in the Zone Heating Design Supply Air Temperature Input Method field: '// &
                    TRIM(FldVal(base + wahpCoolDesignTempMethodOff)))
  END IF
  !validate heat pump coil water flow mode
  IF (SameString(FldVal(base +  wahpHPCoilWaterFlowMode),'Cycling')) THEN
  ELSEIF (SameString(FldVal(base +  wahpHPCoilWaterFlowMode),'Constant')) THEN
  ELSEIF (SameString(FldVal(base +  wahpHPCoilWaterFlowMode),'ConstantOnDemand')) THEN
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:WaterToAirHeatPump "'//TRIM(FldVal(base + wahpNameOff))//'"'// &
                    ' in the Heat Pump Coil Water Flow Mode field: '// &
                    TRIM(FldVal(base + wahpHPCoilWaterFlowMode)))
  END IF
  !set the baseboard indicator
  IF (SameString(FldVal(base +  wahpBaseboardTypeOff),'None')) THEN
    baseboardKind = baseboardNone
  ELSEIF (SameString(FldVal(base +  wahpBaseboardTypeOff),'HotWater')) THEN
    baseboardKind = baseboardHotWater
  ELSEIF (SameString(FldVal(base +  wahpBaseboardTypeOff),'Electric')) THEN
    baseboardKind = baseboardElectric
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:Zone:WaterToAirHeatPump "'//TRIM(FldVal(base + wahpNameOff))//'"'// &
                    ' in the Baseboard Heating Type field: '//TRIM(FldVal(base + wahpBaseboardTypeOff)))
  END IF

  !repeat per zone - only if wahpTStatNameOff is not blank
  IF (.NOT. isTStatNameBlank) THEN
    !***ZoneControl:Thermostat
    CALL AddAlwaysSchedule('4')
    CALL CreateNewObj('ZoneControl:Thermostat')
    CALL AddToObjFld('Name', base + wahpNameOff,' Thermostat')
    CALL AddToObjFld('Zone or ZoneList Name', base + wahpNameOff,'')
    CALL AddToObjStr('Control Type Schedule Name','HVACTemplate-Always 4')
    CALL AddToObjStr('Control Object Type','ThermostatSetpoint:DualSetpoint')
    CALL AddToObjFld('Control Name', base + wahpTStatNameOff,' Dual SP Control',.TRUE.)
  END IF
! OADCV Changes
  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:OutdoorAir')
    CALL AddToObjStr('Name', 'SZ DSOA '//FldVal(base+wahpNameOff))
    CALL AddToObjFld('Outdoor Air Method', base + wahpOutAirTypeOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Person {m3/s}', base + wahpOutAirFlowPersonOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone Floor Area {m3/s-m2}', base + wahpOutAirFlowAreaOff,'')
    CALL AddToObjFld('Outdoor Air Flow per Zone {m3/s}', base + wahpOutAirFlowZoneOff,'',.true.)
  END IF

  !***Sizing:Zone
  CALL CreateNewObj('Sizing:Zone')
  CALL AddToObjFld('Zone or ZoneList Name', base + wahpNameOff,'')
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Input Method',base + wahpCoolDesignTempMethodOff,'')
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature {C}', base + wahpCoolDesignTempOff,'')
  CALL AddToObjFld('Zone Cooling Design Supply Air Temperature Difference {delta C}',base + wahpCoolDesignTempDiffOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Input Method',base + wahpHeatDesignTempMethodOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature {C}', base + wahpHeatDesignTempOff,'')
  CALL AddToObjFld('Zone Heating Design Supply Air Temperature Difference {delta C}',base + wahpHeatDesignTempDiffOff,'')
  CALL AddToObjStr('Zone Cooling Design Supply Air Humidity Ratio {kg-H20/kg-air}','0.008')
  CALL AddToObjStr('Zone Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}','0.008')
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Outdoor Air Object Name', base + wahpDesSpecOANameOff,'')
  ELSE
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','SZ DSOA '//FldVal(base+wahpNameOff))
  END IF
  CALL AddToObjFld('Zone Heating Sizing Factor', base + wahpZoneHeatSizingFactorOff,'')
  CALL AddToObjFld('Zone Cooling Sizing Factor', base + wahpZoneCoolSizingFactorOff,'')
  IF (isSupplyCoolRateAutosize) THEN
    CALL AddToObjStr('Cooling Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Cooling Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}','')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  ELSE
    CALL AddToObjStr('Cooling Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Cooling Design Air Flow Rate {m3/s}', base + wahpSupplyCoolFlowRateOff,'')
    CALL AddToObjStr('Cooling Minimum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Cooling Minimum Air Flow {m3/s}','')
    CALL AddToObjStr('Cooling Minimum Air Flow Fraction {}','0')
  END IF
  IF (isSupplyHeatRateAutosize) THEN
    CALL AddToObjStr('Heating Design Air Flow Method','DesignDay')
    CALL AddToObjStr('Heating Design Air Flow Rate {m3/s}','0')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}','')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
!    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+wahpNameOff),.true.)
  ELSE
    CALL AddToObjStr('Heating Design Air Flow Method','Flow/Zone')
    CALL AddToObjFld('Heating Design Air Flow Rate {m3/s}', base + wahpSupplyHeatFlowRateOff,'')
    CALL AddToObjStr('Heating Maximum Air Flow per Zone Floor Area {m3/s-m2}',' ')
    CALL AddToObjStr('Heating Maximum Air Flow {m3/s}','')
    CALL AddToObjStr('Heating Maximum Air Flow Fraction {}','0')
!    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+wahpNameOff),.true.)
  END IF
  IF (isOATypeDetailed) THEN
    CALL AddToObjFld('Design Specification Zone Air Distribution Object Name', base + wahpDesSpecOADistNameOff,'',.true.)
  ELSE
    CALL AddToObjStr('Design Specification Zone Air Distribution Object Name','SZ DSZAD '//FldVal(base+wahpNameOff),.true.)
  END IF

  IF (.NOT. isOATypeDetailed) THEN
    CALL CreateNewObj('DesignSpecification:ZoneAirDistribution')
    CALL AddToObjStr('Name','SZ DSZAD '//FldVal(base+wahpNameOff))
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Cooling Mode {}','1')
    CALL AddToObjStr('Zone Air Distribution Effectiveness in Heating Mode {}','1',.true.)
  END IF
  !***ZoneHVAC:EquipmentConnections
  CALL CreateNewObj('ZoneHVAC:EquipmentConnections')
  CALL AddToObjFld('Zone Name', base + wahpNameOff,' ')
  CALL AddToObjFld('Zone Conditioning Equipment List Name', base + wahpNameOff,' Equipment')
  IF (isDedOutAirNameBlank) THEN
    CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base + wahpNameOff,' WAHP Supply Inlet')
  ELSE
    CALL AddToObjFld('Zone Air Inlet Node or NodeList Name', base + wahpNameOff,' Inlets')
  END IF
  CALL AddToObjFld('Zone Air Exhaust Node or NodeList Name', base + wahpNameOff,' WAHP Return')
  CALL AddToObjFld('Zone Air Node Name', base + wahpNameOff,' Zone Air Node')
  CALL AddToObjFld('Zone Return Air Node Name', base + wahpNameOff,' Return Outlet',.TRUE.)
  !***ZoneHVAC:EquipmentList
  !NODELIST (not in spreadsheet added with DOAS)
  IF (.NOT. isDedOutAirNameBlank) THEN
    CALL CreateNewObj('NodeList')
    CALL AddToObjFld('Name', base + wahpNameOff,' Inlets')
    CALL AddToObjFld('Node Name', base + wahpNameOff,' WAHP Supply Inlet')
    CALL AddToObjFld('Node Name', base + wahpNameOff,' DOAS Supply Inlet',.TRUE.)
  ENDIF
  !***ZoneHVAC:EquipmentList
  CALL CreateNewObj('ZoneHVAC:EquipmentList')
  CALL AddToObjFld('Name', base + wahpNameOff,' Equipment')
  CALL AddToObjStr('Load Distribution Scheme','SequentialLoad')

  NextSequenceNumber = '1'
  IF (.NOT. isDedOutAirNameBlank) THEN
    CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Zone Equipment Name', base + wahpNameOff,' DOAS ATU')
    CALL AddToObjStr('Zone Equipment Cooling Sequence',trim(NextSequenceNumber))
    CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence',trim(NextSequenceNumber))
    CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
    CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.FALSE.)
    NextSequenceNumber = '2'
  END IF
  CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:WaterToAirHeatPump')
  CALL AddToObjFld('Zone Equipment Name', base + wahpNameOff,' WAHP')
  CALL AddToObjStr('Zone Equipment Cooling Sequence',trim(NextSequenceNumber))
  CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence',trim(NextSequenceNumber), .FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
  CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',isBaseboardNone)
  IF (.NOT. isBaseboardNone) THEN
    IF (isDedOutAirNameBlank) THEN
      NextSequenceNumber = '2'
    ELSE
      NextSequenceNumber = '3'
    END IF
    SELECT CASE (baseboardKind)
      CASE (baseboardHotWater)
        CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CASE (baseboardElectric)
        CALL AddToObjStr('Zone Equipment Object Type','ZoneHVAC:Baseboard:Convective:Electric')
    END SELECT
        CALL AddToObjFld('Zone Equipment Name', base + wahpNameOff,' Baseboard Heat')
        CALL AddToObjStr('Zone Equipment Cooling Sequence',trim(NextSequenceNumber))
        CALL AddToObjStr('Zone Equipment Heating or No-Load Sequence',trim(NextSequenceNumber),.FALSE.)
        CALL AddToObjStr('Zone Equipment Sequential Cooling Fraction','1',.FALSE.)
        CALL AddToObjStr('Zone Equipment Sequential Heating Fraction','1',.TRUE.)
  END IF

  IF (.NOT. isDedOutAirNameBlank) THEN
    !***ZoneHVAC:AirDistributionUnit
    CALL CreateNewObj('ZoneHVAC:AirDistributionUnit')
    CALL AddToObjFld('Name', base + wahpNameOff,' DOAS ATU')
    CALL AddToObjFld('Air Distribution Unit Outlet Node Name', base + wahpNameOff,' DOAS Supply Inlet')
    CALL AddToObjStr('Air Terminal Object Type','AirTerminal:SingleDuct:VAV:NoReheat')
    CALL AddToObjFld('Air Terminal Name', base + wahpNameOff,' DOAS Air Terminal',.TRUE.)
    !***AirTerminal:SingleDuct:VAV:NoReheat
    CALL CreateNewObj('AirTerminal:SingleDuct:VAV:NoReheat')
    CALL AddToObjFld('Name', base + wahpNameOff,' DOAS Air Terminal')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Air Outlet Node Name', base + wahpNameOff,' DOAS Supply Inlet')
    CALL AddToObjFld('Air Inlet Node Name', base + wahpNameOff,' Zone Equip Inlet')
    CALL AddToObjStr('Maximum Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Zone Minimum Air Flow Input Method','Constant')
    CALL AddToObjStr('Constant Minimum Air Flow Fraction','1.0')
    CALL AddToObjStr('Fixed Minimum Air Flow Rate {m3/s}','')
    CALL AddToObjStr('Minimum Air Flow Fraction Schedule Name','')
    CALL AddToObjStr('Design Specification Outdoor Air Object Name','',.TRUE.)
  ENDIF
  !***ZoneHVAC:WaterToAirHeatPump
  IF (isDedOutAirNameBlank) THEN
    CALL WriteComment('Set heat pump outdoor air to zero, because zone is served by dedicated outdoor air system (DOAS)')
  END IF
  CALL CreateNewObj('ZoneHVAC:WaterToAirHeatPump')
  CALL AddToObjFld('Name', base + wahpNameOff,' WAHP')
  CALL AddToObjFld('Availability Schedule Name', base + wahpSysAvailSchedNameOff,'')
  CALL AddToObjFld('Air Inlet Node Name', base + wahpNameOff,' WAHP Return')
  CALL AddToObjFld('Air Outlet Node Name', base + wahpNameOff,' WAHP Supply Inlet')
  CALL AddToObjStr('Outdoor Air Mixer Object Type','OutdoorAir:Mixer')
  CALL AddToObjFld('Outdoor Air Mixer Name', base + wahpNameOff,' WAHP OA Mixing Box')
  !CR8001
  !CALL AddToObjFld('Supply air volumetric flow rate during cooling operation {m3/s}',   &
  !   base + wahpSupplyCoolFlowRateOff,'')
  CALL AddToObjStr('Cooling Supply Air Flow Rate {m3/s}','autosize')
  !CR8001
  !CALL AddToObjFld('Supply air volumetric flow rate during heating operation {m3/s}',   &
  !   base + wahpSupplyHeatFlowRateOff,'')
  CALL AddToObjStr('Heating Supply Air Flow Rate {m3/s}','autosize')
  CALL AddToObjFld('No Load Supply Air Flow Rate {m3/s}',   &
     base + wahpSupplyNoLoadFlowRateOff,'')
  IF (isDedOutAirNameBlank) THEN
    CALL AddToObjStr('Cooling Outdoor Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('Heating Outdoor Air Flow Rate {m3/s}','autosize')
    CALL AddToObjStr('No Load Outdoor Air Flow Rate {m3/s}','autosize')
  ELSE
    CALL AddToObjStr('Cooling Outdoor Air Flow Rate {m3/s}','0.0')
    CALL AddToObjStr('Heating Outdoor Air Flow Rate {m3/s}','0.0')
    CALL AddToObjStr('No Load Outdoor Air Flow Rate {m3/s}','0.0')
  END IF
  CALL AddToObjStr('Supply Air Fan Object Type','Fan:OnOff')
  CALL AddToObjFld('Supply Air Fan Name', base + wahpNameOff,' WAHP Supply Fan')
  CALL AddToObjStr('Heating Coil Object Type','Coil:Heating:WaterToAirHeatPump:EquationFit')
  CALL AddToObjFld('Heating Coil Name', base + wahpNameOff,' WAHP Heating Coil')
  CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:WaterToAirHeatPump:EquationFit')
  CALL AddToObjFld('Cooling Coil Name', base + wahpNameOff,' WAHP Cooling Coil')
  CALL AddToObjFld('Maximum Cycling Rate {cycles/hr}',base + wahpMaxCycRateOff,'')
  CALL AddToObjFld('Heat Pump Time Constant {s}',base + wahpTimeConstOff,'')
  CALL AddToObjFld('Fraction of On-Cycle Power Use',base + wahpFractOnPowerOff,'')
  CALL AddToObjFld('Heat Pump Fan Delay Time {s}',base + wahpFanDelayOff,'')
  IF (suppHeatCoilKind == shcElectric) THEN
    CALL AddToObjStr('Supplemental Heating Coil Object Type','Coil:Heating:Electric')
  ELSEIF (suppHeatCoilKind == shcWater) THEN
    CALL AddToObjStr('Supplemental Heating Coil Object Type','Coil:Heating:Water')
  ENDIF
  CALL AddToObjFld('Supplemental Heating Coil Name', base + wahpNameOff,' WAHP Supp Heating Coil')
  CALL AddToObjStr('Maximum Supply Air Temperature from Supplemental Heater {C}','autosize')
  CALL AddToObjStr('Maximum Outdoor Dry-Bulb Temperature for Supplemental Heater Operation {C}','20.0')
  CALL AddToObjFld('Outdoor Dry-Bulb Temperature Sensor Node Name', base + wahpNameOff,' WAHP Outside Air Inlet')
  CALL AddToObjFld('Fan Placement', base + wahpFanPlacementOff,'')
  SELECT CASE (fanMode)
    CASE (fmCycling)
      CALL AddToObjStr('Supply Air Fan Operating Mode Schedule Name','HVACTemplate-Always 0',.FALSE.)
    CASE (fmNeither)
      CALL AddToObjFld('Supply Air Fan Operating Mode Schedule Name', base + wahpFanModeOff,'',.FALSE.)
  END SELECT
  CALL AddToObjStr('Availability Manager List Name','')
  CALL AddToObjFld('Heat Pump Coil Water Flow Mode',base + wahpHPCoilWaterFlowMode,'',.TRUE.)
  SELECT CASE (fanMode)
    CASE (fmCycling)
      CALL AddAlwaysSchedule('0')
    CASE DEFAULT
  END SELECT
  !***Fan:OnOff
  CALL CreateNewObj('Fan:OnOff')
  CALL AddToObjFld('Name', base + wahpNameOff,' WAHP Supply Fan')
  IF (.NOT. isSysAvailSchedBlank) THEN
    CALL AddToObjFld('Availability Schedule Name', base + wahpSysAvailSchedNameOff,'')
  ELSE
    CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
  END IF
  CALL AddToObjFld('Fan Efficiency', base + wahpFanTotEfficiencyOff,'')
  CALL AddToObjFld('Pressure Rise {Pa}', base + wahpFanPressureOff,'')
  CALL AddToObjStr('Maximum Flow Rate {m3/s}','autosize')
  CALL AddToObjFld('Motor Efficiency', base + wahpFanMotorEfficiencyOff,'')
  CALL AddToObjStr('Motor in Airstream Fraction','1')
  IF (isDrawThru) THEN
    CALL AddToObjFld('Air Inlet Node Name', base + wahpNameOff,' WAHP Heating Coil Outlet')
  ELSE
    CALL AddToObjFld('Air Inlet Node Name', base + wahpNameOff,' WAHP Mixed Air Outlet')
  END IF
  CALL AddToObjFld('Air Outlet Node Name', base + wahpNameOff,' WAHP Supply Fan Outlet',.TRUE.)
  IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  !***OutdoorAir:Mixer
  CALL CreateNewObj('OutdoorAir:Mixer')
  CALL AddToObjFld('Name', base + wahpNameOff,' WAHP OA Mixing Box')
  CALL AddToObjFld('Mixed Air Node Name', base + wahpNameOff,' WAHP Mixed Air Outlet')
  CALL AddToObjFld('Outside Air Stream Node Name', base + wahpNameOff,' WAHP Outside Air Inlet')
  CALL AddToObjFld('Relief Air Stream Node Name', base + wahpNameOff,' WAHP Relief Air Outlet')
  CALL AddToObjFld('Return Air Stream Node Name', base + wahpNameOff,' WAHP Return',.TRUE.)
  !***OutdoorAir:Node
  CALL CreateNewObj('OutdoorAir:Node')
  CALL AddToObjFld('Name', base + wahpNameOff,' WAHP Outside Air Inlet')
  CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
  !*** Coil:Cooling:WaterToAirHeatPump:EquationFit
  ! Warn if cooling capacity and SHR inputs are not consistent
  IF (isCoolCoilCapAutosize .AND. (.NOT. isSHRAutosize)) THEN
    CALL WriteError('Warning:  In HVACTemplate:Zone:WaterToAirHeatPump "'//TRIM(FldVal(base + wahpNameOff))//'"'// &
                      ' the Cooling Coil Rated Capacity is autosized, so the Cooling Coil Gross Rated Sensible Heat Ratio '// &
                      ' will also be autosized. The specified value for Cooling Coil Gross Rated Sensible Heat Ratio'// &
                      ' will be ignored.  Autosize both or specify values for both.',msgWarning)
  ENDIF
  IF ((.NOT.isCoolCoilCapAutosize) .AND. isSHRAutosize) THEN
    CALL WriteError('Warning:  In HVACTemplate:Zone:WaterToAirHeatPump "'//TRIM(FldVal(base + wahpNameOff))//'"'// &
                      ' the Cooling Coil Rated Capacity will not be used when the Cooling Coil Gross Rated Sensible Heat Ratio'// &
                      ' is autosized. Autosize both or specify values for both.',msgWarning)
  ENDIF
  CALL WriteComment('Performance coefficients assume rated Capacities and COP are at:')
  CALL WriteComment('26.7C (80F) entering air dry-bulb temperature, 19.4C (67F) entering air wet-bulb temperature')
  CALL WriteComment('and 29.4C (85F) entering water temperature')
  CALL CreateNewObj('Coil:Cooling:WaterToAirHeatPump:EquationFit')
  CALL AddToObjFld('Name', base + wahpNameOff,' WAHP Cooling Coil')
  CALL AddToObjFld('Water Inlet Node Name', base + wahpNameOff,' WAHP Cooling Water Inlet Node')
  CALL AddToObjFld('Water Outlet Node Name', base + wahpNameOff,' WAHP Cooling Water Outlet Node')
  IF (isDrawThru) THEN
    CALL AddToObjFld('Air Inlet Node Name', base + wahpNameOff,' WAHP Mixed Air Outlet')
  ELSE
    CALL AddToObjFld('Air Inlet Node Name', base + wahpNameOff,' WAHP Supply Fan Outlet')
  END IF
  CALL AddToObjFld('Air Outlet Node Name', base + wahpNameOff,' WAHP Cooling Coil Outlet')
  CALL AddToObjStr('Rated Air Flow Rate {m3/s}','Autosize')
  CALL AddToObjStr('Rated Water Flow Rate {m3/s}','Autosize')
  CALL AddToObjFld('Rated Total Cooling Capacity {W}',base + wahpCoolCoilCapOff,'')
  IF (isCoolCoilCapAutosize .OR. isSHRAutosize) THEN
    CALL AddToObjStr('Rated Sensible Cooling Capacity {W}','Autosize')
  ELSE
    CALL AddToObjNum('Rated Sensible Cooling Capacity {W}',StringToReal(FldVal(base + wahpCoolCoilCapOff)) * &
     StringToReal(FldVal(base + wahpCoolCoilSHROff)))
  ENDIF
  CALL AddToObjFld('Rated Cooling Coefficient of Performance',base + wahpCoolCoilCOPOff,'')
  CALL AddToObjStr('Total Cooling Capacity Coefficient 1','-9.149069561')
  CALL AddToObjStr('Total Cooling Capacity Coefficient 2','10.87814026')
  CALL AddToObjStr('Total Cooling Capacity Coefficient 3','-1.718780157')
  CALL AddToObjStr('Total Cooling Capacity Coefficient 4','0.746414818')
  CALL AddToObjStr('Total Cooling Capacity Coefficient 5','0.0')
  CALL AddToObjStr('Sensible Cooling Capacity Coefficient 1','-5.462690012')
  CALL AddToObjStr('Sensible Cooling Capacity Coefficient 2','17.95968138')
  CALL AddToObjStr('Sensible Cooling Capacity Coefficient 3','-11.87818402')
  CALL AddToObjStr('Sensible Cooling Capacity Coefficient 4','-0.980163419')
  CALL AddToObjStr('Sensible Cooling Capacity Coefficient 5','0.767285761')
  CALL AddToObjStr('Sensible Cooling Capacity Coefficient 6','0.0')
  CALL AddToObjStr('Cooling Power Consumption Coefficient 1','-3.205409884')
  CALL AddToObjStr('Cooling Power Consumption Coefficient 2','-0.976409399')
  CALL AddToObjStr('Cooling Power Consumption Coefficient 3','3.97892546')
  CALL AddToObjStr('Cooling Power Consumption Coefficient 4','0.938181818')
  CALL AddToObjStr('Cooling Power Consumption Coefficient 5','0.0')
  CALL AddToObjStr('Nominal Time for Condensate Removal to Begin {s}','0')
  CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity {dimensionless}','0',.TRUE.)
  !***Coil:Heating:WaterToAirHeatPump:EquationFit
  CALL WriteComment('Performance coefficients assume rated Capacity and COP are at:')
  CALL WriteComment('21.11C (70F) entering air dry-bulb temperature and 21.11C (70F) entering water temperature')
  CALL CreateNewObj('Coil:Heating:WaterToAirHeatPump:EquationFit')
  CALL AddToObjFld('Name', base + wahpNameOff,' WAHP Heating Coil')
  CALL AddToObjFld('Water Inlet Node Name', base + wahpNameOff,' WAHP Heating Water Inlet Node')
  CALL AddToObjFld('Water Outlet Node Name', base + wahpNameOff,' WAHP Heating Water Outlet Node')
  CALL AddToObjFld('Air Inlet Node Name', base + wahpNameOff,' WAHP Cooling Coil Outlet')
  CALL AddToObjFld('Air Outlet Node Name', base + wahpNameOff,' WAHP Heating Coil Outlet')
  CALL AddToObjStr('Rated Air Flow Rate {m3/s}','Autosize')
  CALL AddToObjStr('Rated Water Flow Rate {m3/s}','Autosize')
  CALL AddToObjFld('Rated Heating Capacity {W}',base + wahpHPHeatCoilCapOff,'')
  CALL AddToObjFld('Rated Heating Coefficient of Performance',base + wahpHPHeatCoilCOPOff,'')
  CALL AddToObjStr('Heating Capacity Coefficient 1','-1.361311959')
  CALL AddToObjStr('Heating Capacity Coefficient 2','-2.471798046')
  CALL AddToObjStr('Heating Capacity Coefficient 3','4.173164514')
  CALL AddToObjStr('Heating Capacity Coefficient 4','0.640757401')
  CALL AddToObjStr('Heating Capacity Coefficient 5','0.0')
  CALL AddToObjStr('Heating Power Consumption Coefficient 1','-2.176941116')
  CALL AddToObjStr('Heating Power Consumption Coefficient 2','0.832114286')
  CALL AddToObjStr('Heating Power Consumption Coefficient 3','1.570743399')
  CALL AddToObjStr('Heating Power Consumption Coefficient 4','0.690793651')
  CALL AddToObjStr('Heating Power Consumption Coefficient 5','0.0',.TRUE.)
  SELECT CASE (suppHeatCoilKind)
    CASE (shcElectric)
      !***Coil:Heating:Electric
      CALL CreateNewObj('Coil:Heating:Electric')
      CALL AddToObjFld('Name', base + wahpNameOff,' WAHP Supp Heating Coil')
      CALL AddToObjFld('Availability Schedule Name', base + wahpSuppHeatAvailSchedNameOff,'')
      CALL AddToObjStr('Efficiency','1')
      CALL AddToObjFld('Nominal Capacity of the Coil {W}', base + wahpSuppHeatCoilCapOff,'')
      IF (isDrawThru) THEN
        CALL AddToObjFld('Air Inlet Node Name', base + wahpNameOff,' WAHP Supply Fan Outlet')
      ELSE
        CALL AddToObjFld('Air Inlet Node Name', base + wahpNameOff,' WAHP Heating Coil Outlet')
      END IF
      CALL AddToObjFld('Air Outlet Node Name', base + wahpNameOff,' WAHP Supply Inlet')
      CALL AddToObjStr('Coil Temp Setpoint Node','',.TRUE.)
    CASE (shcWater)
      !CURVE:CUBIC ~ line 235
      CALL CreateNewObj('Coil:Heating:Water')
      CALL AddToObjFld('Name', base + wahpNameOff,' WAHP Supp Heating Coil')
      CALL AddToObjFld('Availability Schedule Name', base + wahpSuppHeatAvailSchedNameOff,'')
      CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
      CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Water Inlet Node Name', base + wahpNameOff,' WAHP Supp Heating Coil HW Inlet')
      CALL AddToObjFld('Water Outlet Node Name', base + wahpNameOff,' WAHP Supp Heating Coil HW Outlet')
      IF (isDrawThru) THEN
        CALL AddToObjFld('Air Inlet Node Name', base + wahpNameOff,' WAHP Supply Fan Outlet')
      ELSE
        CALL AddToObjFld('Air Inlet Node Name', base + wahpNameOff,' WAHP Heating Coil Outlet')
      END IF
      CALL AddToObjFld('Air Outlet Node Name', base + wahpNameOff,' WAHP Supply Inlet')
! For future use - if supp heat capacity field is added to this system type
!      IF (isSuppHeatCoilCapAutosize) THEN
!        CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
!      ELSE
!        CALL AddToObjStr('Performance Input Method','NominalCapacity')
!      END IF
!      CALL AddToObjFld('Rated Capacity', base + wahpHeatCoilCapOff,'')
      CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
      CALL AddToObjStr('Rated Capacity','autosize')
      CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
      CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
      CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
      CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
      CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + wahpNameOff,' WAHP Supp Heating Coil HW Branch')
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
      CALL AddToObjFld('Component Name', base + wahpNameOff,' WAHP Supp Heating Coil')
      CALL AddToObjFld('Component Inlet Node Name', base + wahpNameOff,' WAHP Supp Heating Coil HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + wahpNameOff,' WAHP Supp Heating Coil HW Outlet',.TRUE.)
      CALL AddToStrList(TRIM(FldVal(base + wahpNameOff)) // ' WAHP Supp Heating Coil HW Branch',handleHeatingCoilBranch)
    CASE DEFAULT
  END SELECT
  !***Branch
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + wahpNameOff,' Cooling Condenser Branch')
  CALL AddToStrList(TRIM(FldVal(base + wahpNameOff)) // ' Cooling Condenser Branch',handleMixedDemanBranch)
  CALL AddToObjStr('Pressure Drop Curve Name','')
  CALL AddToObjStr('Component Object Type','Coil:Cooling:WaterToAirHeatPump:EquationFit')
  CALL AddToObjFld('Component Name', base + wahpNameOff,' WAHP Cooling Coil')
  CALL AddToObjFld('Component Inlet Node Name', base + wahpNameOff,' WAHP Cooling Water Inlet Node')
  CALL AddToObjFld('Component Outlet Node Name', base + wahpNameOff,' WAHP Cooling Water Outlet Node',.TRUE.)
  !***Branch
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + wahpNameOff,' Heating Condenser Branch')
  CALL AddToStrList(TRIM(FldVal(base + wahpNameOff)) // ' Heating Condenser Branch',handleMixedDemanBranch)
  CALL AddToObjStr('Pressure Drop Curve Name','')
  CALL AddToObjStr('Component Object Type','Coil:Heating:WaterToAirHeatPump:EquationFit')
  CALL AddToObjFld('Component Name', base + wahpNameOff,' WAHP Heating Coil')
  CALL AddToObjFld('Component Inlet Node Name', base + wahpNameOff,' WAHP Heating Water Inlet Node')
  CALL AddToObjFld('Component Outlet Node Name', base + wahpNameOff,' WAHP Heating Water Outlet Node',.TRUE.)
  !Baseboards
  SELECT CASE (baseboardKind)
    CASE (baseboardHotWater)
      !ZoneHVAC:Baseboard:RadiantConvective:Water
      CALL CreateNewObj('ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Name', base + wahpNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + wahpBaseboardAvailSchedNameOff,'')
      CALL AddToObjFld('Inlet Node Name', base + wahpNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Outlet Node Name', base + wahpNameOff,' Baseboard HW Outlet')
      CALL AddToObjStr('Rated Average Water Temperature {C}','87.78')
      CALL AddToObjStr('Rated Water Mass Flow Rate {m3/s}','0.063')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + wahpBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Maximum Water Flow Rate','autosize')
      CALL AddToObjStr('Convergence Tolerance','0.001')
      CALL AddToObjStr('Fraction Radiant','0.0')
      CALL AddToObjStr('Fraction of Radiant Energy Incident on People','0.0',.TRUE.)
      !BRANCH (for baseboard)
      CALL CreateNewObj('Branch')
      CALL AddToObjFld('Name', base + wahpNameOff,' Baseboard Heat HW Branch')
      CALL AddToObjStr('Pressure Drop Curve Name','')
      CALL AddToObjStr('Component Object Type','ZoneHVAC:Baseboard:RadiantConvective:Water')
      CALL AddToObjFld('Component Name', base + wahpNameOff,' Baseboard Heat')
      CALL AddToObjFld('Component Inlet Node Name', base + wahpNameOff,' Baseboard HW Inlet')
      CALL AddToObjFld('Component Outlet Node Name', base + wahpNameOff,' Baseboard HW Outlet',.TRUE.)
      CALL AddToStrList(TRIM(FldVal(base + wahpNameOff)) // ' Baseboard Heat HW Branch',handleHeatingCoilBranch)
      IF ((compactHotLoopBase .EQ. 0) .OR. (numCompactBoiler .EQ. 0)) THEN
        CALL WriteError('Both a HVACTemplate:Plant:HotWaterLoop and a HVACTemplate:Plant:Boiler are needed ' // &
             'when using hot water baseboards.')
      END IF
    CASE (baseboardElectric)
      !BASEBOARD HEATER:Electric:Convective
      CALL CreateNewObj('ZoneHVAC:Baseboard:Convective:Electric')
      CALL AddToObjFld('Name', base + wahpNameOff,' Baseboard Heat')
      CALL AddToObjFld('Availability Schedule Name', base + wahpBaseboardAvailSchedNameOff,'')
      CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
      CALL AddToObjFld('Heating Design Capacity {W}', base + wahpBaseboardCapOff,'')
      CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
      CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
      CALL AddToObjStr('Efficiency','1',.TRUE.)
  END SELECT
END DO
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE CreateMixedLoop
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2010
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for HVACTemplate:Plant:MixedWaterLoop

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    Based on CreateChldLoop

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for pump configuration type
INTEGER,PARAMETER :: pcVariable = 1
INTEGER,PARAMETER :: pcConstant = 2
          ! parameters for fluid type
INTEGER,PARAMETER :: ftWater = 1
INTEGER,PARAMETER :: ftEthyleneGlycol30 = 2
INTEGER,PARAMETER :: ftEthyleneGlycol40 = 3
INTEGER,PARAMETER :: ftEthyleneGlycol50 = 4
INTEGER,PARAMETER :: ftEthyleneGlycol60 = 5
INTEGER,PARAMETER :: ftPropyleneGlycol30 = 6
INTEGER,PARAMETER :: ftPropyleneGlycol40 = 7
INTEGER,PARAMETER :: ftPropyleneGlycol50 = 8
INTEGER,PARAMETER :: ftPropyleneGlycol60 = 9
INTEGER,PARAMETER :: ftSteam = 10

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base
INTEGER :: towerBase
INTEGER :: iTower
INTEGER :: boilerBase
INTEGER :: iBoiler
INTEGER :: iStr
LOGICAL :: isMwpOperTypeDefault
LOGICAL :: isMwpPumpSchedNameBlank
LOGICAL :: isMwpHiSetPtSchedNameBlank
LOGICAL :: isMwpLoSetPtSchedNameBlank
LOGICAL :: isSupplyBypassYes
LOGICAL :: isDemandBypassYes
INTEGER :: PumpConfig
INTEGER :: fluidType=0
INTEGER :: lastBranch=0
INTEGER :: lastSupplyBranch=0
INTEGER :: numBoilersThisLoop=0
INTEGER :: numTowersThisLoop=0
LOGICAL :: isPumpHeader
LOGICAL :: isPumpBranch
LOGICAL :: isTowerPresent
LOGICAL :: isBoilerPresent
LOGICAL :: isEndOfList
CHARACTER(len=1) :: pumpCount='1'

INTEGER,ALLOCATABLE, DIMENSION(:) :: sortedBoilers
INTEGER,ALLOCATABLE, DIMENSION(:) :: sortedTowers

base =  compactMixedLoopBase
IF (base .EQ. 0) RETURN
 !must have either a boiler or tower
IF ((numCompactTower + numCompactTowerOR + numCompactBoiler + numCompactBoilerOR) .EQ. 0 ) RETURN
! set flags
CALL validChoice(base + mwpPumpControlTypeOff,'Intermittent','Continuous','Plant:HotWaterLoop','Pump Control Type')
isMwpOperTypeDefault = SameString(FldVal(base + mwpOperTypeOff),'Default')
isMwpPumpSchedNameBlank = SameString(FldVal(base + mwpPumpSchedNameOff),'')
IF (SameString(FldVal(base + mwpPumpConfigOff),'VariableFlow')) THEN
  PumpConfig = pcVariable
ELSEIF (SameString(FldVal(base + mwpPumpConfigOff),'ConstantFlow')) THEN
  PumpConfig = pcConstant
ELSEIF (SameString(FldVal(base + mwpPumpConfigOff),'')) THEN
  PumpConfig = pcConstant
ELSE
  CALL WriteError('Invalid choice in HVACTemplate:Plant:MixedWaterLoop in the Hot Water Pump Configuration field')
END IF
isMwpHiSetPtSchedNameBlank = SameString(FldVal(base + mwpHiSetPtSchedNameOff),'')
isMwpLoSetPtSchedNameBlank = SameString(FldVal(base + mwpLoSetPtSchedNameOff),'')
CALL validChoice(base + mwpSupplyBypassOff,'Yes','No','HVACTemplate:Plant:MixedWaterLoop','Supply Side Bypass Pipe')
isSupplyBypassYes = SameString(FldVal(base + mwpSupplyBypassOff),'Yes')
CALL validChoice(base + mwpDemandBypassOff,'Yes','No','HVACTemplate:Plant:MixedWaterLoop','Demand Side Bypass Pipe')
isDemandBypassYes = SameString(FldVal(base + mwpDemandBypassOff),'Yes')
IF (SameString(FldVal(base + mwpFluidTypeOff),'Water')) THEN
  FluidType = ftWater
ELSEIF (SameString(FldVal(base + mwpFluidTypeOff),'EthyleneGlycol30')) THEN
  FluidType = ftEthyleneGlycol30
ELSEIF (SameString(FldVal(base + mwpFluidTypeOff),'EthyleneGlycol40')) THEN
  FluidType = ftEthyleneGlycol40
ELSEIF (SameString(FldVal(base + mwpFluidTypeOff),'EthyleneGlycol50')) THEN
  FluidType = ftEthyleneGlycol50
ELSEIF (SameString(FldVal(base + mwpFluidTypeOff),'EthyleneGlycol60')) THEN
  FluidType = ftEthyleneGlycol60
ELSEIF (SameString(FldVal(base + mwpFluidTypeOff),'PropyleneGlycol30')) THEN
  FluidType = ftPropyleneGlycol30
ELSEIF (SameString(FldVal(base + mwpFluidTypeOff),'PropyleneGlycol40')) THEN
  FluidType = ftPropyleneGlycol40
ELSEIF (SameString(FldVal(base + mwpFluidTypeOff),'PropyleneGlycol50')) THEN
  FluidType = ftPropyleneGlycol50
ELSEIF (SameString(FldVal(base + mwpFluidTypeOff),'PropyleneGlycol60')) THEN
  FluidType = ftPropyleneGlycol60
ELSE
  CALL WriteError('Invalid choice in HVACTemplate:Plant:MixedWaterLoop in the Fluid Type field')
END IF
IF (SameString(FldVal(base + mwpPumpTypeOff),'SinglePump')) THEN
  isPumpHeader = .FALSE.
  isPumpBranch = .FALSE.
ELSEIF (SameString(FldVal(base + mwpPumpTypeOff),'PumpPerTowerOrBoiler')) THEN
  isPumpHeader = .FALSE.
  isPumpBranch = .TRUE.
ELSEIF (SameString(FldVal(base + mwpPumpTypeOff),'TwoHeaderedPumps')) THEN
  isPumpHeader = .TRUE.
  isPumpBranch = .FALSE.
  pumpCount = '2'
ELSEIF (SameString(FldVal(base + mwpPumpTypeOff),'ThreeHeaderedPumps')) THEN
  isPumpHeader = .TRUE.
  isPumpBranch = .FALSE.
  pumpCount = '3'
ELSEIF (SameString(FldVal(base + mwpPumpTypeOff),'FourHeaderedPumps')) THEN
  isPumpHeader = .TRUE.
  isPumpBranch = .FALSE.
  pumpCount = '4'
ELSEIF (SameString(FldVal(base + mwpPumpTypeOff),'FiveHeaderedPumps')) THEN
  isPumpHeader = .TRUE.
  isPumpBranch = .FALSE.
  pumpCount = '5'
ELSE
  CALL WriteError('Invalid choice in HVACTemplate:Plant:HotWaterLoop in the Hot Water Pump Type field')
END IF
isTowerPresent = ((numCompactTower + numCompactTowerOR) .GT. 0)
! If there is an HVACTemplate:Plant:HotWaterLoop present, all boilers with blank loop type will be placed there
isBoilerPresent = ((numCompactBoiler + numCompactBoilerOR) .GT. 0)
IF(((numCompactBoiler + numCompactBoilerOR) .GT. 0) .AND. (compactHotLoopBase .NE. 0)) THEN
  CALL WriteError('In HVACTemplate:Plant:MixedWaterLoop "'//TRIM(FldVal(base + mwpNameOff))//'"'// &
                  ' an HVACTemplate:Plant:HotWaterLoop is also present.  All boilers with blank Template Loop Type field'// &
                  ' will be connected to the Hot Water Loop.',msgWarning)
ENDIF


! Sort the boilers and towers just once and store in arrays since both lists
! are needed several times - only boilers serving a hot water loop, also count boilers on this loop
IF (isBoilerPresent) THEN
  ALLOCATE(sortedBoilers(numCompactBoiler + numCompactBoilerOR))
  numBoilersThisLoop = 0
  CALL ClearNumSortList
  DO iBoiler = 1, numCompactBoiler
    boilerBase =  compactBoilerBase(iBoiler)  !store the base pointer to each boiler
    IF (SameString(FldVal(boilerBase + blrLoopTypeOff),'MixedWater') .OR. SameString(FldVal(boilerBase + blrLoopTypeOff),'')) THEN
      CALL AddToNumSortList(StringToReal(FldVal(boilerBase + blrPriorityOff)),boilerBase)
      numBoilersThisLoop = numBoilersThisLoop + 1
    END IF
  END DO
  DO iBoiler = 1, numCompactBoilerOR
    boilerBase =  compactBoilerORBase(iBoiler)  !store the base pointer to each boiler
    IF (SameString(FldVal(boilerBase + blrorLoopTypeOff),'MixedWater') .OR. SameString(FldVal(boilerBase + blrorLoopTypeOff),'')) &
     THEN
      CALL AddToNumSortList(StringToReal(FldVal(boilerBase + blrorPriorityOff)),boilerBase)
      numBoilersThisLoop = numBoilersThisLoop + 1
    END IF
  END DO
  CALL SortIncreasingNumList
  DO iBoiler = 1, numBoilersThisLoop
    sortedBoilers(iBoiler) = NumSortList(iBoiler)%sortedPtr
  END DO
ENDIF
isBoilerPresent = (numBoilersThisLoop .GT. 0)
IF (isTowerPresent) THEN
  ALLOCATE(sortedTowers(numCompactTower + numCompactTowerOR))
  numTowersThisLoop = 0
  CALL ClearNumSortList
  DO iTower = 1, numCompactTower
    towerBase =  compactTowerBase(iTower)  !store the base pointer to each tower
    IF (SameString(FldVal(towerBase + twrLoopTypeOff),'MixedWater') .OR. SameString(FldVal(towerBase + twrLoopTypeOff),'')) THEN
      CALL AddToNumSortList(StringToReal(FldVal(towerBase + twrPriorityOff)),towerBase)
      numTowersThisLoop = numTowersThisLoop + 1
    END IF
  END DO
  DO iTower = 1, numCompactTowerOR
    towerBase =  compactTowerORBase(iTower)  !store the base pointer to each tower
    IF (SameString(FldVal(towerBase + twrorLoopTypeOff),'MixedWater') .OR. SameString(FldVal(towerBase + twrorLoopTypeOff),'')) THEN
      CALL AddToNumSortList(StringToReal(FldVal(towerBase + twrorPriorityOff)),towerBase)
      numTowersThisLoop = numTowersThisLoop + 1
    END IF
  END DO
  CALL SortIncreasingNumList
  DO iTower = 1, numTowersThisLoop
    sortedTowers(iTower) = NumSortList(iTower)%sortedPtr
  END DO
ENDIF
IF (numTowersThisLoop+numBoilersThisLoop == 0) THEN
  CALL WriteError('Severe:  In HVACTemplate:Plant:MixedWaterLoop "'//TRIM(FldVal(base + mwpNameOff))//'"'// &
                  ' there are no towers or boilers serving this loop.')
END IF
!***Sizing:Plant
! Do assuming the condenser sets the required flow rates.
CALL CreateNewObj('Sizing:Plant')
CALL AddToObjFld('Plant or Condenser Loop Name', base + mwpNameOff,' Mixed Water Loop')
CALL AddToObjStr('Loop Type','Condenser')
CALL AddToObjFld('Design loop exit temperature {C}', base + mwpHiSetPtDesignOff,'')
CALL AddToObjFld('Design loop delta T {deltaC}',base + mwpDesDeltaTOff,'',.TRUE.)
!***PlantLoop
CALL CreateNewObj('PlantLoop')
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Water Loop')
IF (fluidType .EQ. ftWater) THEN
  CALL AddToObjStr('Fluid Type','Water')
  CALL AddToObjStr('User Defined Fluid Type','')
ELSE
  CALL AddToObjStr('Fluid Type','UserDefinedFluidType')
  CALL AddToObjFld('User Defined Fluid Type', base + mwpNameOff,' Loop Fluid')
END IF
IF (isMwpOperTypeDefault) THEN
  CALL AddToObjFld('PlantEquipmentOperationSchemes Name', base + mwpNameOff,' Operation')
ELSE
  CALL AddToObjFld('PlantEquipmentOperationSchemes Name', base + mwpOperNameOff,'')
END IF
CALL AddToObjFld('Loop Temperature Setpoint Node Name', base + mwpNameOff,' Mixed Supply Outlet')
CALL AddToObjStr('Maximum Loop Temperature {C}','100')
CALL AddToObjStr('Minimum Loop Temperature {C}','10')
CALL AddToObjStr('Maximum Loop Volumetric Flow Rate {m3/s}','autosize')
CALL AddToObjStr('Minimum Loop Volumetric Flow Rate {m3/s}','0')
CALL AddToObjStr('volume of the plant loop {m3}','autosize')
CALL AddToObjFld('Plant Side Inlet Node Name', base + mwpNameOff,' Mixed Supply Inlet')
CALL AddToObjFld('Plant Side Outlet Node Name', base + mwpNameOff,' Mixed Supply Outlet')
CALL AddToObjFld('Plant Side Branch List Name', base + mwpNameOff,' Mixed Supply Side Branches')
CALL AddToObjFld('Plant Side Connector List Name', base + mwpNameOff,' Mixed Supply Side Connectors')
CALL AddToObjFld('Demand Side Inlet Node Name', base + mwpNameOff,' Mixed Demand Inlet')
CALL AddToObjFld('Demand Side Outlet Node Name', base + mwpNameOff,' Mixed Demand Outlet')
CALL AddToObjFld('Demand Side Branch List Name', base + mwpNameOff,' Mixed Demand Side Branches')
CALL AddToObjFld('Demand Side Connector List Name', base + mwpNameOff,' Mixed Demand Side Connectors')
CALL AddToObjFld('Load Distribution Scheme', base + mwpLdDistSchmOff,'')
CALL AddToObjStr('AvailabilityManagerAssignmentList','')
CALL AddToObjStr('Plant Loop Demand Calculation Scheme','DualSetPointDeadband',.TRUE.)
IF (isMwpOperTypeDefault) THEN
  !***PlantEquipmentOperationSchemes
  CALL CreateNewObj('PlantEquipmentOperationSchemes')
  CALL AddToObjFld('Name', base + mwpNameOff,' Operation')
  IF ((isTowerPresent) .AND. (isBoilerPresent)) THEN
    CALL AddToObjStr('Control Scheme Object Type','PlantEquipmentOperation:HeatingLoad')
    CALL AddToObjFld('Control Scheme Name', base + mwpNameOff,' Heat Operation All Hours')
    CALL AddToObjStr('Control Scheme Schedule Name','HVACTemplate-Always 1')
    CALL AddToObjStr('Control Scheme Object Type','PlantEquipmentOperation:CoolingLoad')
    CALL AddToObjFld('Control Scheme Name', base + mwpNameOff,' Cool Operation All Hours')
    CALL AddToObjStr('Control Scheme Schedule Name','HVACTemplate-Always 1',.TRUE.)
  ELSEIF ((isTowerPresent) .AND. (.NOT. isBoilerPresent)) THEN
    CALL AddToObjStr('Control Scheme Object Type','PlantEquipmentOperation:CoolingLoad')
    CALL AddToObjFld('Control Scheme Name', base + mwpNameOff,' Cool Operation All Hours')
    CALL AddToObjStr('Control Scheme Schedule Name','HVACTemplate-Always 1',.TRUE.)
  ELSEIF ((.NOT. isTowerPresent) .AND. (isBoilerPresent)) THEN
    CALL AddToObjStr('Control Scheme Object Type','PlantEquipmentOperation:HeatingLoad')
    CALL AddToObjFld('Control Scheme Name', base + mwpNameOff,' Heat Operation All Hours')
    CALL AddToObjStr('Control Scheme Schedule Name','HVACTemplate-Always 1',.TRUE.)
  ENDIF
  CALL AddAlwaysSchedule('1')
  !***PlantEquipmentOperation:HeatingLoad
  IF (isBoilerPresent) THEN
    CALL CreateNewObj('PlantEquipmentOperation:HeatingLoad')
    CALL AddToObjFld('Name', base + mwpNameOff,' Heat Operation All Hours')
    CALL AddToObjStr('Load Range Lower Limit 1 {W}','0')
    CALL AddToObjStr('Load Range Upper Limit 1 {W}','1000000000000000')
    CALL AddToObjFld('Priority Control 1 Equipment List Name', base + mwpNameOff,' All Heating Equipment',.TRUE.)
  END IF
  !***PlantEquipmentOperation:CoolingLoad
  IF (isTowerPresent) THEN
    CALL CreateNewObj('PlantEquipmentOperation:CoolingLoad')
    CALL AddToObjFld('Name', base + mwpNameOff,' Cool Operation All Hours')
    CALL AddToObjStr('Load Range Lower Limit 1 {W}','0')
    CALL AddToObjStr('Load Range Upper Limit 1 {W}','1000000000000000')
    CALL AddToObjFld('Priority Control 1 Equipment List Name', base + mwpNameOff,' All Cooling Equipment',.TRUE.)
  END IF
  !***PlantEquipmentList
  IF (isBoilerPresent) THEN
    CALL CreateNewObj('PlantEquipmentList')
    CALL AddToObjFld('Name', base + mwpNameOff,' All Heating Equipment')
    !Sort the boilers by order of priority
    DO iBoiler = 1, numBoilersThisLoop
      boilerBase = sortedBoilers(iBoiler)
      isEndOfList = (iBoiler .EQ. numBoilersThisLoop)
      IF (SameString(FldVal(boilerBase),'HVACTemplate:Plant:Boiler')) THEN
        !If blrTypeOff="Hot Water Boiler" then this field is "BOILER:SIMPLE"
        !If blrTypeOff="Purchased Hot Water" then this field is "PURCHASED:HOT WATER"
        IF (SameString(FldVal(boilerBase + blrTypeOff),'DistrictHotWater')) THEN
          CALL AddToObjStr('Equipment Object Type', 'DistrictHeating')
        ELSE
          CALL AddToObjStr('Equipment Object Type', 'Boiler:HotWater')
        END IF
        CALL AddToObjFld('Equipment Name', boilerBase + blrNameOff,'',isEndOfList)
      ELSEIF (SameString(FldVal(boilerBase),'HVACTemplate:Plant:Boiler:ObjectReference')) THEN
        CALL AddToObjFld('Equipment Object Type', boilerBase + blrorObjTypeOff,'')
        CALL AddToObjFld('Equipment Name', boilerBase + blrorObjNameOff,'',isEndOfList)
      ENDIF
    END DO
  END IF
  !***PlantEquipmentList
  IF (isTowerPresent) THEN
    CALL CreateNewObj('PlantEquipmentList')
    CALL AddToObjFld('Name', base + mwpNameOff,' All Cooling Equipment')
    !Sort the towers by order of priority
    DO iTower = 1, numTowersThisLoop
      towerBase = sortedTowers(iTower)
      isEndOfList = (iTower .EQ. numTowersThisLoop)
      IF (SameString(FldVal(towerBase),'HVACTemplate:Plant:Tower')) THEN
        IF (SameString(FldVal(towerBase + twrTypeOff),'SingleSpeed')) THEN
          CALL AddToObjStr('Equipment Object Type','CoolingTower:SingleSpeed')
        ELSEIF (SameString(FldVal(towerBase + twrTypeOff),'TwoSpeed')) THEN
          CALL AddToObjStr('Equipment Object Type','CoolingTower:TwoSpeed')
        ELSE
          CALL AddToObjStr('Equipment Object Type','CoolingTower:SingleSpeed')
        END IF
        CALL AddToObjFld('Equipment Name', towerBase + twrNameOff,'',isEndOfList)
      ELSEIF (SameString(FldVal(towerBase),'HVACTemplate:Plant:Tower:ObjectReference')) THEN
        CALL AddToObjFld('Equipment Object Type', towerBase + twrorObjTypeOff,'')
        CALL AddToObjFld('Equipment Name', towerBase + twrorObjNameOff,'',isEndOfList)
      ENDIF
    END DO
  END IF
END IF
IF (isMwpHiSetPtSchedNameBlank) THEN
  CALL AddAlwaysSchedule(FldVal(base + mwpHiSetPtDesignOff))
END IF
IF (isMwpLoSetPtSchedNameBlank) THEN
  CALL AddAlwaysSchedule(FldVal(base + mwpLoSetPtDesignOff))
END IF

! Variable flow boilers need setpoints, make a nodelist regardless of flow type, also include towers
CALL CreateNewObj('NodeList')
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Supply Setpoint Nodes')
!use list created in boiler object
IF (isBoilerPresent) THEN
  DO iStr = 1,numStrList
    IF(StrList(iStr)%Handle .EQ. handleMixedBoilerOutletNode) THEN
      CALL AddToObjStr('Node Name', StrList(iStr)%StrToSave)
    END IF
  END DO
ENDIF
!use list created in tower object
IF (isTowerPresent) THEN
  DO iStr = 1,numStrList
    IF(StrList(iStr)%Handle .EQ. handleMixedTowerOutletNode) THEN
      CALL AddToObjStr('Node Name', StrList(iStr)%StrToSave)
    END IF
  END DO
ENDIF
CALL AddToObjFld('Node Name', base + mwpNameOff,' Mixed Supply Outlet',.TRUE.)

!***SetpointManager:Scheduled:DualSetpoint
CALL CreateNewObj('SetpointManager:Scheduled:DualSetpoint')
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Temp Manager')
CALL AddToObjStr('Control Variable','Temperature')
IF (isMwpHiSetPtSchedNameBlank) THEN
  CALL AddToObjStr('High Setpoint Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + mwpHiSetPtDesignOff)))
ELSE
  CALL AddToObjFld('High Setpoint Schedule Name', base + mwpHiSetPtSchedNameOff,'')
END IF
IF (isMwpLoSetPtSchedNameBlank) THEN
  CALL AddToObjStr('Low Setpoint Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + mwpLoSetPtDesignOff)))
ELSE
  CALL AddToObjFld('Low Setpoint Schedule Name', base + mwpLoSetPtSchedNameOff,'')
END IF
CALL AddToObjFld('Setpoint Node or NodeList Name', base + mwpNameOff,' Mixed Supply Setpoint Nodes',.TRUE.)
!***BranchList
CALL CreateNewObj('BranchList')
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Supply Side Branches')
CALL AddToObjFld('Branch Name', base + mwpNameOff,' Mixed Supply Inlet Branch')
!Repeat this line for every boiler or tower
!use list created in boiler and tower objects, also save lastSupplyBranch counter for use in later lists
DO iStr = 1,numStrList
  IF(StrList(iStr)%Handle .EQ. handleMixedSupplyBranch) THEN
    CALL AddToObjStr('Branch Name', StrList(iStr)%StrToSave)
    lastSupplyBranch = iStr
  END IF
END DO
IF (isSupplyBypassYes) THEN
  CALL AddToObjFld('Branch Name', base + mwpNameOff,' Mixed Supply Bypass Branch')
END IF
CALL AddToObjFld('Branch Name', base + mwpNameOff,' Mixed Supply Outlet Branch',.TRUE.)
!***ConnectorList
CALL CreateNewObj('ConnectorList')
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Supply Side Connectors')
CALL AddToObjStr('Connector Object Type','Connector:Splitter')
CALL AddToObjFld('Connector Name', base + mwpNameOff,' Mixed Supply Splitter')
CALL AddToObjStr('Connector Object Type','Connector:Mixer')
CALL AddToObjFld('Connector Name', base + mwpNameOff,' Mixed Supply Mixer',.TRUE.)
!***Connector:Splitter
CALL CreateNewObj('Connector:Splitter')
CALL AddToObjFld('Name', base + hwpNameOff,' Mixed Supply Splitter')
CALL AddToObjFld('Inlet Branch Name', base + mwpNameOff,' Mixed Supply Inlet Branch')
!Repeat this line for every boiler or tower, use list created in boiler and tower objects
DO iStr = 1,lastSupplyBranch
  IF(StrList(iStr)%Handle .EQ. handleMixedSupplyBranch) THEN
    isEndOfList = ((iStr == lastSupplyBranch) .AND. (.NOT. isSupplyBypassYes))
    CALL AddToObjStr('Outlet Branch Name', StrList(iStr)%StrToSave,isEndOfList)
  END IF
END DO
IF (isSupplyBypassYes) THEN
  CALL AddToObjFld('Outlet Branch Name', base + mwpNameOff,' Mixed Supply Bypass Branch',.TRUE.)
END IF
!***Connector:Mixer
CALL CreateNewObj('Connector:Mixer')
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Supply Mixer')
CALL AddToObjFld('Outlet Branch Name', base + mwpNameOff,' Mixed Supply Outlet Branch')
!Repeat this line for every boiler or tower, use list created in boiler and tower objects
DO iStr = 1,lastSupplyBranch
  IF(StrList(iStr)%Handle .EQ. handleMixedSupplyBranch) THEN
    isEndOfList = ((iStr == lastSupplyBranch) .AND. (.NOT. isSupplyBypassYes))
    CALL AddToObjStr('Inlet Branch Name', StrList(iStr)%StrToSave,isEndOfList)
  END IF
END DO
IF (isSupplyBypassYes) THEN
  CALL AddToObjFld('Inlet Branch Name', base + mwpNameOff,' Mixed Supply Bypass Branch',.TRUE.)
END IF
! This is out of order since pump logic always requires this branch and pipe
!***Branch
IF (isSupplyBypassYes) THEN
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Supply Bypass Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
  CALL AddToObjFld('Component Name', base + mwpNameOff,' Mixed Supply Side Bypass Pipe')
  CALL AddToObjFld('Component Inlet Node Name', base + mwpNameOff,' Mixed Supply Bypass Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + mwpNameOff,' Mixed Supply Bypass Outlet',.TRUE.)
  !***Pipe:Adiabatic
  CALL CreateNewObj('Pipe:Adiabatic')
  CALL AddToObjFld('Name', base + hwpNameOff,' Mixed Supply Side Bypass Pipe')
  CALL AddToObjFld('Inlet Node Name', base + mwpNameOff,' Mixed Supply Bypass Inlet')
  CALL AddToObjFld('Outlet Node Name', base + mwpNameOff,' Mixed Supply Bypass Outlet',.TRUE.)
END IF
!***Pump:ConstantSpeed
IF (isPumpBranch) THEN
  !if branch pumps are used just insert a pipe in the supply inlet branch
  CALL CreateNewObj('Pipe:Adiabatic')
  CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Supply Side Inlet Pipe')
  CALL AddToObjFld('Inlet Node Name', base + mwpNameOff,' Mixed Supply Inlet')
  CALL AddToObjFld('Outlet Node Name', base + mwpNameOff,' Mixed Supply Pipe Outlet',.TRUE.)
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Supply Inlet Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
  CALL AddToObjFld('Component Name', base + mwpNameOff,' Mixed Supply Side Inlet Pipe')
  CALL AddToObjFld('Component Inlet Node Name', base + mwpNameOff,' Mixed Supply Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + mwpNameOff,' Mixed Supply Pipe Outlet',.TRUE.)
ELSE
  IF (PumpConfig .EQ. pcConstant) THEN
    IF (.NOT. isPumpHeader) THEN
      !***Pump:ConstantSpeed
      CALL CreateNewObj('Pump:ConstantSpeed')
      CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Supply Pump')
      CALL AddToObjFld('Inlet Node Name', base + mwpNameOff,' Mixed Supply Inlet')
      CALL AddToObjFld('Outlet Node Name', base + mwpNameOff,' Mixed Pump Outlet')
      CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Rated Pump Head {Pa}', base + mwpPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      CALL AddToObjFld('Pump Control Type', base + mwpPumpControlTypeOff,'')
      IF (isMwpPumpSchedNameBlank) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', base + mwpPumpSchedNameOff,'',.TRUE.)
      END IF
    ELSE
      !***HeaderedPumps:ConstantSpeed
      CALL CreateNewObj('HeaderedPumps:ConstantSpeed')
      CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Supply Pump')
      CALL AddToObjFld('Inlet Node Name', base + mwpNameOff,' Mixed Supply Inlet')
      CALL AddToObjFld('Outlet Node Name', base + mwpNameOff,' Mixed Pump Outlet')
      CALL AddToObjStr('Total Rated Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Number of Pumps in Bank',pumpCount)
      CALL AddToObjStr('Flow Sequencing Control Scheme','Sequential')
      CALL AddToObjFld('Rated Pump Head {Pa}', base + mwpPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      CALL AddToObjFld('Pump Control Type', base + mwpPumpControlTypeOff,'')
      IF (isMwpPumpSchedNameBlank) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', base + mwpPumpSchedNameOff,'',.TRUE.)
      END IF
    END IF
    !***Branch
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Supply Inlet Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isPumpHeader) THEN
      CALL AddToObjStr('Component Object Type','Pump:ConstantSpeed')
    ELSE
      CALL AddToObjStr('Component Object Type','HeaderedPumps:ConstantSpeed')
    END IF
    CALL AddToObjFld('Component Name', base + mwpNameOff,' Mixed Supply Pump')
    CALL AddToObjFld('Component Inlet Node Name', base + mwpNameOff,' Mixed Supply Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + mwpNameOff,' Mixed Pump Outlet',.TRUE.)
  ELSE
    !***Pump:VariableSpeed
    CALL WriteComment('Pump part load coefficients based on simple parabola to represent a generic VSD drive.')
    IF (.NOT. isPumpHeader) THEN
      CALL CreateNewObj('Pump:VariableSpeed')
      CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Supply Pump')
      CALL AddToObjFld('Inlet Node Name', base + mwpNameOff,' Mixed Supply Inlet')
      CALL AddToObjFld('Outlet Node Name', base + mwpNameOff,' Mixed Pump Outlet')
      CALL AddToObjStr('Rated Volumetric Flow Rate {m3/s}','autosize')
      CALL AddToObjFld('Rated Pump Head {Pa}', base + mwpPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      !Curve based on VSD generic curve from user with adjustments to prevent neg values - Nov 2008
      !Can use something better when min flow rate is able to be autosized or entered as a fraction
      CALL AddToObjStr('Coefficient 1 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 2 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 3 of the Part Load Performance Curve','1.0')
      CALL AddToObjStr('Coefficient 4 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Min Flow Rate while operating in variable flow capacity {m3/s}','0')
      CALL AddToObjFld('Pump Control Type', base + mwpPumpControlTypeOff,'')
      IF (isMwpPumpSchedNameBlank) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', base + mwpPumpSchedNameOff,'',.TRUE.)
      END IF
    ELSE
      !***HeaderedPumps:VariableSpeed
      CALL CreateNewObj('HeaderedPumps:VariableSpeed')
      CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Supply Pump')
      CALL AddToObjFld('Inlet Node Name', base + mwpNameOff,' Mixed Supply Inlet')
      CALL AddToObjFld('Outlet Node Name', base + mwpNameOff,' Mixed Pump Outlet')
      CALL AddToObjStr('Total Rated Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Number of Pumps in Bank',pumpCount)
      CALL AddToObjStr('Flow Sequencing Control Scheme','Sequential')
      CALL AddToObjFld('Rated Pump Head {Pa}', base + mwpPumpHeadOff,'')
      CALL AddToObjStr('Rated Power Consumption {W}','autosize')
      CALL AddToObjStr('Motor Efficiency','0.9')
      CALL AddToObjStr('Fraction of Motor Inefficiencies to Fluid Stream','0')
      !Curve based on VSD generic curve from user with adjustments to prevent neg values - Nov 2008
      !Can use something better when min flow rate is able to be autosized or entered as a fraction
      CALL AddToObjStr('Coefficient 1 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 2 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Coefficient 3 of the Part Load Performance Curve','1.0')
      CALL AddToObjStr('Coefficient 4 of the Part Load Performance Curve','0.0')
      CALL AddToObjStr('Min Flow Rate while operating in variable flow capacity {m3/s}','0')
      CALL AddToObjFld('Pump Control Type', base + mwpPumpControlTypeOff,'')
      IF (isMwpPumpSchedNameBlank) THEN
        CALL AddToObjStr('Pump Flow Rate Schedule Name','',.TRUE.)
      ELSE
        CALL AddToObjFld('Pump Flow Rate Schedule Name', base + mwpPumpSchedNameOff,'',.TRUE.)
      END IF
    END IF
    !***Branch
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Supply Inlet Branch')
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF (.NOT. isPumpHeader) THEN
      CALL AddToObjStr('Component Object Type','Pump:VariableSpeed')
    ELSE
      CALL AddToObjStr('Component Object Type','HeaderedPumps:VariableSpeed')
    END IF
    CALL AddToObjFld('Component Name', base + mwpNameOff,' Mixed Supply Pump')
    CALL AddToObjFld('Component Inlet Node Name', base + mwpNameOff,' Mixed Supply Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + mwpNameOff,' Mixed Pump Outlet',.TRUE.)
  END IF
END IF
!***Branch
CALL CreateNewObj('Branch')
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Supply Outlet Branch')
CALL AddToObjStr('Pressure Drop Curve Name','')
CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
CALL AddToObjFld('Component Name', base + mwpNameOff,' Mixed Supply Outlet Pipe')
CALL AddToObjFld('Component Inlet Node Name', base + mwpNameOff,' Mixed Supply Outlet Pipe Inlet')
CALL AddToObjFld('Component Outlet Node Name', base + mwpNameOff,' Mixed Supply Outlet',.TRUE.)
!***Pipe:Adiabatic
CALL CreateNewObj('Pipe:Adiabatic')
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Supply Outlet Pipe')
CALL AddToObjFld('Inlet Node Name', base + mwpNameOff,' Mixed Supply Outlet Pipe Inlet')
CALL AddToObjFld('Outlet Node Name', base + mwpNameOff,' Mixed Supply Outlet',.TRUE.)
!***BranchList
CALL CreateNewObj('BranchList')
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Demand Side Branches')
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Demand Inlet Branch')
!The names of all of the branches for the heating coils have been previously gathered
DO iStr = 1, numStrList
  IF (StrList(iStr)%Handle .EQ. handleMixedDemanBranch) THEN
    CALL AddToObjStr('Name', StrList(iStr)%StrToSave)
  END IF
END DO
IF (isDemandBypassYes) THEN
  CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Demand Bypass Branch')
END IF
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Demand Outlet Branch',.TRUE.)
!***ConnectorList
CALL CreateNewObj('ConnectorList')
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Demand Side Connectors')
CALL AddToObjStr('Connector Object Type','Connector:Splitter')
CALL AddToObjFld('Connector Name', base + mwpNameOff,' Mixed Demand Splitter')
CALL AddToObjStr('Connector Object Type','Connector:Mixer')
CALL AddToObjFld('Connector Name', base + mwpNameOff,' Mixed Demand Mixer',.TRUE.)
!***Connector:Splitter
CALL CreateNewObj('Connector:Splitter')
CALL AddToObjFld('Name', base + hwpNameOff,' Mixed Demand Splitter')
CALL AddToObjFld('Inlet Branch Name', base + hwpNameOff,' Mixed Demand Inlet Branch')
IF (isDemandBypassYes) THEN
  CALL AddToObjFld('Outlet Branch Name', base + hwpNameOff,' Mixed Demand Bypass Branch')
END IF
!The names of all of the branches for the heating coils have been previously gathered
DO iStr = 1, numStrList
  IF (StrList(iStr)%Handle .EQ. handleMixedDemanBranch) THEN
    lastBranch = iStr
  END IF
END DO
DO iStr = 1, numStrList
  IF (StrList(iStr)%Handle .EQ. handleMixedDemanBranch) THEN
    CALL AddToObjStr('Name', StrList(iStr)%StrToSave, iStr .EQ. lastBranch)
  END IF
END DO
!***Connector:Mixer
CALL CreateNewObj('Connector:Mixer')
CALL AddToObjFld('Name', base + hwpNameOff,' Mixed Demand Mixer')
CALL AddToObjFld('Outlet Branch Name', base + mwpNameOff,' Mixed Demand Outlet Branch')
IF (isDemandBypassYes) THEN
  CALL AddToObjFld('Inlet Branch Name', base + mwpNameOff,' Mixed Demand Bypass Branch')
END IF
!The names of all of the branches for the heating coils have been previously gathered
DO iStr = 1, numStrList
  IF (StrList(iStr)%Handle .EQ. handleMixedDemanBranch) THEN
    CALL AddToObjStr('Name', StrList(iStr)%StrToSave, iStr .EQ. lastBranch)
  END IF
END DO
!***Branch
CALL CreateNewObj('Branch')
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Demand Inlet Branch')
CALL AddToObjStr('Pressure Drop Curve Name','')
CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
CALL AddToObjFld('Component Name', base + mwpNameOff,' Mixed Demand Inlet Pipe')
CALL AddToObjFld('Component Inlet Node Name', base + mwpNameOff,' Mixed Demand Inlet')
CALL AddToObjFld('Component Outlet Node Name', base + mwpNameOff,' Mixed Demand Inlet Pipe Outlet',.TRUE.)
!***Pipe:Adiabatic
CALL CreateNewObj('Pipe:Adiabatic')
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Demand Inlet Pipe')
CALL AddToObjFld('Inlet Node Name', base + mwpNameOff,' Mixed Demand Inlet')
CALL AddToObjFld('Outlet Node Name', base + mwpNameOff,' Mixed Demand Inlet Pipe Outlet',.TRUE.)
!***Branch
IF (isDemandBypassYes) THEN
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Demand Bypass Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
  CALL AddToObjFld('Component Name', base + mwpNameOff,' Mixed Demand Side Bypass Pipe')
  CALL AddToObjFld('Component Inlet Node Name', base + mwpNameOff,' Mixed Demand Bypass Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + mwpNameOff,' Mixed Demand Bypass Outlet',.TRUE.)
  !***Pipe:Adiabatic
  CALL CreateNewObj('Pipe:Adiabatic')
  CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Demand Side Bypass Pipe')
  CALL AddToObjFld('Inlet Node Name', base + mwpNameOff,' Mixed Demand Bypass Inlet')
  CALL AddToObjFld('Outlet Node Name', base + mwpNameOff,' Mixed Demand Bypass Outlet',.TRUE.)
END IF
!***Branch
CALL CreateNewObj('Branch')
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Demand Outlet Branch')
CALL AddToObjStr('Pressure Drop Curve Name','')
CALL AddToObjStr('Component Object Type','Pipe:Adiabatic')
CALL AddToObjFld('Component Name', base + mwpNameOff,' Mixed Demand Outlet Pipe')
CALL AddToObjFld('Component Inlet Node Name', base + mwpNameOff,' Mixed Demand Outlet Pipe Inlet')
CALL AddToObjFld('Component Outlet Node Name', base + mwpNameOff,' Mixed Demand Outlet',.TRUE.)
!***Pipe:Adiabatic
CALL CreateNewObj('Pipe:Adiabatic')
CALL AddToObjFld('Name', base + mwpNameOff,' Mixed Demand Outlet Pipe')
CALL AddToObjFld('Inlet Node Name', base + mwpNameOff,' Mixed Demand Outlet Pipe Inlet')
CALL AddToObjFld('Outlet Node Name', base + mwpNameOff,' Mixed Demand Outlet',.TRUE.)
! Create glycol fluid if not water
IF (fluidType .NE. ftWater) THEN
  CALL CreateNewObj('FluidProperties:GlycolConcentration')
  CALL AddToObjFld('Name', base + mwpNameOff,' Loop Fluid')
  SELECT CASE (fluidType)
    CASE (ftEthyleneGlycol30)
      CALL AddToObjStr('Glycol Type','EthyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.30',.TRUE.)
    CASE (ftEthyleneGlycol40)
      CALL AddToObjStr('Glycol Type','EthyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.40',.TRUE.)
    CASE (ftEthyleneGlycol50)
      CALL AddToObjStr('Glycol Type','EthyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.50',.TRUE.)
    CASE (ftEthyleneGlycol60)
      CALL AddToObjStr('Glycol Type','EthyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.60',.TRUE.)
    CASE (ftPropyleneGlycol30)
      CALL AddToObjStr('Glycol Type','PropyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.30',.TRUE.)
    CASE (ftPropyleneGlycol40)
      CALL AddToObjStr('Glycol Type','PropyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.40',.TRUE.)
    CASE (ftPropyleneGlycol50)
      CALL AddToObjStr('Glycol Type','PropyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.50',.TRUE.)
    CASE (ftPropyleneGlycol60)
      CALL AddToObjStr('Glycol Type','PropyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.60',.TRUE.)
    CASE DEFAULT
      CALL AddToObjStr('Glycol Type','EthyleneGlycol')
      CALL AddToObjStr('User Defined Glycol Name','')
      CALL AddToObjStr('Glycol Concentration','0.30',.TRUE.)
  END SELECT
END IF
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE CreateDOAS
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer and Michael Witte of GARD Analytics, Inc.
          !    DATE WRITTEN   April 2010-May 2011
          !    MODIFIED
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Creates the new objects for a dedicated outdoor air supply system
          !    that is connected to a zonal system such as:
          !       HVACTemplate:Zone:FanCoil
          !       HVACTemplate:Zone:PTAC
          !       HVACTemplate:Zone:PTHP
          !       HVACTemplate:Zone:Unitary
          !       HVACTemplate:Zone:WaterToAirHeatPump

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          ! parameters for coil type (heating)
INTEGER,PARAMETER :: ctHotWater = 1
INTEGER,PARAMETER :: ctElectric = 2
INTEGER,PARAMETER :: ctGas      = 3
INTEGER,PARAMETER :: ctNone     = 4
          ! parameters for coil type (cooling)
INTEGER, PARAMETER :: ccChWater = 1
INTEGER, PARAMETER :: ccChWaterDetailed = 2
INTEGER, PARAMETER :: ccTwoSpeedDX = 3
INTEGER, PARAMETER :: ccTwoStageDX = 4
INTEGER, PARAMETER :: ccTwoStageHumidControlDX = 5
INTEGER, PARAMETER :: ccWaterCoilSystemHX = 6
INTEGER, PARAMETER :: ccDXcoilSystemHX = 7
INTEGER, PARAMETER :: ccNone = 8
          ! parameters for cool setpoint control type
INTEGER,PARAMETER :: cspFixed    = 1
INTEGER,PARAMETER :: cspScheduled = 2
INTEGER,PARAMETER :: cspOAreset = 3
          ! parameters for heat setpoint control type
INTEGER,PARAMETER :: hspFixed    = 1
INTEGER,PARAMETER :: hspScheduled = 2
INTEGER,PARAMETER :: hspOAreset = 3
          ! parameters for supply fan coef type
!? INTEGER,PARAMETER :: sfcInletVane = 1
!? INTEGER,PARAMETER :: sfcDischgDamp = 2
!? INTEGER,PARAMETER :: sfcVSD = 3
!? INTEGER,PARAMETER :: sfcAppG = 4
          ! parameters for heat recovery type
INTEGER,PARAMETER :: htrecNone = 1
INTEGER,PARAMETER :: htrecSens = 2
INTEGER,PARAMETER :: htrecEnth = 3
          ! parameters for supply fan type
INTEGER,PARAMETER :: sfpBlowThru = 1
INTEGER,PARAMETER :: sfpDrawThru = 2
          ! parameters for dehumidification type
INTEGER, PARAMETER :: dehumidNone           = 1
INTEGER, PARAMETER :: dehumidCoolRhtHtgCoil = 2
INTEGER, PARAMETER :: dehumidCoolRhtDesuper = 3
INTEGER, PARAMETER :: dehumidMultimode      = 4
          ! parameters for humidification type
INTEGER, PARAMETER :: humidifyNone      = 1
INTEGER, PARAMETER :: humidifyElecSteam = 2

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: base  ! used as base for
INTEGER :: iSys
INTEGER :: i
LOGICAL :: isSupplyFlowRateAutosize
INTEGER :: heatCoilType=0
! INTEGER :: preHeatCoilType=0
INTEGER :: coolCoilKind=0
! REAL    :: minAirRatio
LOGICAL :: isSysAvailSchedBlank
LOGICAL :: isCoolStPtSchedBlank
LOGICAL :: isHeatStPtSchedBlank
! INTEGER :: econoKind=0
! INTEGER :: lockoutKind=0
INTEGER :: coolSetPtCtrlKind=0
INTEGER :: heatSetPtCtrlKind=0
! INTEGER :: supFanCoef=0
INTEGER :: heatRecovery=0
INTEGER :: iInt
INTEGER :: supFanPlacement=0
REAL    :: htRecSens75
REAL    :: htRecLat75
LOGICAL :: isDehumidifyNone
INTEGER :: dehumidCtrlKind = 0
LOGICAL :: isHumidifierNone
INTEGER :: humidifierKind = 0
LOGICAL :: isCoolCapAutosize
LOGICAL :: isSHRAutosize
LOGICAL :: isCoolCoilChW
LOGICAL :: isCoolCoilDX

! Node Names
CHARACTER(len=MaxAlphaLength) :: lastOutlet=''
CHARACTER(len=MaxAlphaLength) :: fanInlet=''
CHARACTER(len=MaxAlphaLength) :: fanOutlet=''
CHARACTER(len=MaxAlphaLength) :: coolCoilUnitInlet=''
CHARACTER(len=MaxAlphaLength) :: coolCoilUnitOutlet=''
CHARACTER(len=MaxAlphaLength) :: heatCoilInlet=''
CHARACTER(len=MaxAlphaLength) :: heatCoilOutlet=''
CHARACTER(len=MaxAlphaLength) :: reheatCoilInlet=''
CHARACTER(len=MaxAlphaLength) :: reheatCoilOutlet=''
CHARACTER(len=MaxAlphaLength) :: humidifierInlet=''
CHARACTER(len=MaxAlphaLength) :: humidifierOutlet=''
CHARACTER(len=MaxAlphaLength) :: airloopOutlet=''
CHARACTER(len=MaxAlphaLength) :: dxCoilObjectType='None'
CHARACTER(len=MaxAlphaLength) :: scheduleName=''

!======================================================================================================
! create Systems
!======================================================================================================
DO iSys = 1, numCompactDedOutAir
  base =  compactDedOutAirBase(iSys)
  ! set flags
  isSupplyFlowRateAutosize = SameString(FldVal(base + doasSupplyRateOff),'autosize')
  isSysAvailSchedBlank = (FldVal(base + doasSysAvailSchedNameOff) .EQ. '')
  isCoolStPtSchedBlank = (FldVal(base + doasCoolSetPtSchedNameOff) .EQ. '')
  isCoolCapAutosize = SameString(FldVal(base + doasCoolCoilCapOff),'autosize')
  isSHRAutosize = SameString(FldVal(base + doasCoolCoilSHROff),'autosize')
  isHeatStPtSchedBlank = (FldVal(base + doasHeatSetPtSchedNameOff) .EQ. '')
  !check cooling coil kind
  IF (SameString(FldVal(base +  doasCoolCoilKindOff),'ChilledWater')) THEN
    coolCoilKind  = ccChWater
    isCoolCoilChW = .TRUE.
    isCoolCoilDX  = .FALSE.
  ELSEIF (SameString(FldVal(base + doasCoolCoilKindOff),'ChilledWaterDetailedFlatModel')) THEN
    coolCoilKind = ccChWaterDetailed
    isCoolCoilChW = .TRUE.
    isCoolCoilDX  = .FALSE.
  ELSEIF (SameString(FldVal(base + doasCoolCoilKindOff),'TwoSpeedDX')) THEN
    coolCoilKind = ccTwoSpeedDX
    isCoolCoilChW = .FALSE.
    isCoolCoilDX  = .TRUE.
  ELSEIF (SameString(FldVal(base + doasCoolCoilKindOff),'TwoStageDX')) THEN
    coolCoilKind = ccTwoStageDX
    isCoolCoilChW = .FALSE.
    isCoolCoilDX  = .TRUE.
  ELSEIF (SameString(FldVal(base + doasCoolCoilKindOff),'TwoStageHumidityControlDX')) THEN
    coolCoilKind = ccTwoStageHumidControlDX
    isCoolCoilChW = .FALSE.
    isCoolCoilDX  = .TRUE.
  ELSEIF (SameString(FldVal(base + doasCoolCoilKindOff),'HeatExchangerAssistedChilledWater')) THEN
    coolCoilKind = ccWaterCoilSystemHX
    isCoolCoilChW = .TRUE.
    isCoolCoilDX  = .FALSE.
  ELSEIF (SameString(FldVal(base + doasCoolCoilKindOff),'HeatExchangerAssistedDX')) THEN
    coolCoilKind = ccDXcoilSystemHX
    isCoolCoilChW = .FALSE.
    isCoolCoilDX  = .TRUE.
  ELSEIF (SameString(FldVal(base + doasCoolCoilKindOff),'None')) THEN
    coolCoilKind = ccNone
    isCoolCoilChW = .FALSE.
    isCoolCoilDX  = .FALSE.
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DedicatedOutdoorAir  "'//TRIM(FldVal(base + doasNameOff))//'"'// &
                    ' in the Cooling Coil Type field: '//TRIM(FldVal(base + doasCoolCoilKindOff)))
  END IF
  !check cooling coil setpoint control type
  IF (SameString(FldVal(base +  doasCoolSetPtCtrlTypeOff),'FixedSetpoint')) THEN
    coolSetPtCtrlKind = cspFixed
  ELSEIF (SameString(FldVal(base + doasCoolSetPtCtrlTypeOff),'Scheduled')) THEN
    coolSetPtCtrlKind = cspScheduled
  ELSEIF (SameString(FldVal(base + doasCoolSetPtCtrlTypeOff),'OutdoorAirTemperatureReset')) THEN
    coolSetPtCtrlKind = cspOAreset
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DedicatedOutdoorAir  "'//TRIM(FldVal(base + doasNameOff))//'"'// &
                    ' in the Cooling Coil Setpoint Control Type field: '//TRIM(FldVal(base + doasCoolSetPtCtrlTypeOff)))
  END IF
  !set heating coil indicator
  IF (SameString(FldVal(base + doasHeatCoilKindOff),'HotWater')) THEN
    heatCoilType = ctHotWater
  ELSEIF (SameString(FldVal(base + doasHeatCoilKindOff),'Electric')) THEN
    heatCoilType = ctElectric
  ELSEIF (SameString(FldVal(base + doasHeatCoilKindOff),'Gas')) THEN
    heatCoilType = ctGas
  ELSEIF (SameString(FldVal(base + doasHeatCoilKindOff),'None')) THEN
    heatCoilType = ctNone
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DedicatedOutdoorAir in the Heating Coil Type field')
  END IF
  !check heating coil setpoint control type
  IF (SameString(FldVal(base +  doasHeatSetPtCtrlTypeOff),'FixedSetpoint')) THEN
    heatSetPtCtrlKind = hspFixed
  ELSEIF (SameString(FldVal(base + doasHeatSetPtCtrlTypeOff),'Scheduled')) THEN
    heatSetPtCtrlKind = hspScheduled
  ELSEIF (SameString(FldVal(base + doasHeatSetPtCtrlTypeOff),'OutdoorAirTemperatureReset')) THEN
    heatSetPtCtrlKind = hspOAreset
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DedicatedOutdoorAir  "'//TRIM(FldVal(base + doasNameOff))//'"'// &
                    ' in the Heating Coil Setpoint Control Type field: '//TRIM(FldVal(base + doasHeatSetPtCtrlTypeOff)))
  END IF
  !set the Heat Recovery Type indicator
  IF (SameString(FldVal(base + doasHeatRecTypeOff),'Sensible')) THEN
    heatRecovery = htrecSens
  ELSEIF (SameString(FldVal(base + doasHeatRecTypeOff),'Enthalpy')) THEN
    heatRecovery = htrecEnth
  ELSEIF (SameString(FldVal(base + doasHeatRecTypeOff),'None')) THEN
    heatRecovery = htrecNone
  ELSE
    heatRecovery = htrecNone
    CALL WriteError('Invalid choice in HVACTemplate:System:DedicatedOutdoorAir in the Heating Coil Setpoint Reset Type field')
  ENDIF
  IF (SameString(FldVal(base + doasSupplyFanPlaceOff), 'BlowThrough')) THEN
    supFanPlacement = sfpBlowThru
  ELSEIF (SameString(FldVal(base + doasSupplyFanPlaceOff), 'DrawThrough')) THEN
    supFanPlacement = sfpDrawThru
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DedicatedOutdoorAir "'//TRIM(FldVal(base + doasNameOff))//'"'// &
                    ' in the Supply Fan Placement field: '//TRIM(FldVal(base + vsSupplyFanPlaceOff)))
  END IF
  !set the dehumidification indicators
  isDehumidifyNone = SameString(FldVal(base + doasDehumCtrlTypeOff),'None')
  IF (SameString(FldVal(base +  doasDehumCtrlTypeOff),'None')) THEN
    dehumidCtrlKind = dehumidNone
  ELSEIF (SameString(FldVal(base +  doasDehumCtrlTypeOff),'CoolReheatHeatingCoil')) THEN
    dehumidCtrlKind = dehumidCoolRhtHtgCoil
  ELSEIF (SameString(FldVal(base +  doasDehumCtrlTypeOff),'CoolReheatDesuperheater')) THEN
    dehumidCtrlKind = dehumidCoolRhtDesuper
  ELSEIF (SameString(FldVal(base +  doasDehumCtrlTypeOff),'Multimode')) THEN
    dehumidCtrlKind = dehumidMultimode
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DedicatedOutdoorAir "'//TRIM(FldVal(base + doasNameOff))//'"'// &
                    ' in the Dehumidification Control Type field: '//TRIM(FldVal(base + doasDehumCtrlTypeOff)))
  END IF
  ! check for valid combination of cooling coil type and humidity control type
  IF ((dehumidCtrlKind .EQ. dehumidMultimode) .AND. (coolCoilKind .NE. ccTwoStageHumidControlDX)) THEN
    CALL WriteError('In HVACTemplate:System:DedicatedOutdoorAir "'//TRIM(FldVal(base + doasNameOff))//'"'// &
                    ' the Dehumidification Control Type field='//TRIM(FldVal(base + doasDehumCtrlTypeOff))// &
                    ' is not applicable for Cooling Coil Type='//TRIM(FldVal(base + doasCoolCoilKindOff))// &
                    '. Dehumidification Control Type reset to None.',msgWarning)
    dehumidCtrlKind = dehumidNone
    isDehumidifyNone = .TRUE.
  END IF
  IF ((dehumidCtrlKind .EQ. dehumidCoolRhtDesuper) .AND. .NOT. isCoolCoilDX) THEN
    CALL WriteError('In HVACTemplate:System:DedicatedOutdoorAir "'//TRIM(FldVal(base + doasNameOff))//'"'// &
                    ' the Dehumidification Control Type field='//TRIM(FldVal(base + doasDehumCtrlTypeOff))// &
                    ' is not applicable for Cooling Coil Type='//TRIM(FldVal(base + doasCoolCoilKindOff))// &
                    '. Dehumidification Control Type reset to None.',msgWarning)
    dehumidCtrlKind = dehumidNone
    isDehumidifyNone = .TRUE.
  END IF
  IF ((dehumidCtrlKind .EQ. dehumidCoolRhtHtgCoil) .AND. (heatCoilType .EQ. ctNone)) THEN
    CALL WriteError('In HVACTemplate:System:DedicatedOutdoorAir "'//TRIM(FldVal(base + doasNameOff))//'"'// &
                    ' the Dehumidification Control Type field='//TRIM(FldVal(base + doasDehumCtrlTypeOff))// &
                    ' but there is no heating coil in this system to provide reheat. Cold supply temps may result.',msgWarning)
  END IF

  !set the humidifier indicators
  isHumidifierNone = SameString(FldVal(base + doasHumidCtrlTypeOff),'None')
  IF (SameString(FldVal(base +  doasHumidCtrlTypeOff),'None')) THEN
    humidifierKind = humidifyNone
  ELSEIF (SameString(FldVal(base +  doasHumidCtrlTypeOff),'ElectricSteam')) THEN
    humidifierKind = humidifyElecSteam
  ELSE
    CALL WriteError('Invalid choice in HVACTemplate:System:DedicatedOutdoorAir "'//TRIM(FldVal(base + doasNameOff))//'"'// &
                    ' in the Humidifier Type field: '//TRIM(FldVal(base + doasHumidCtrlTypeOff)))
  END IF
  IF (.NOT. (SameString(FldVal(base +  doasAirOutletTypeOff),'DirectIntoZone'))) THEN
    CALL WriteError('Invalid choice in HVACTemplate:System:DedicatedOutdoorAir  "'//TRIM(FldVal(base + doasNameOff))//'"'// &
                    ' in the Air Outlet Type field: '//TRIM(FldVal(base + doasAirOutletTypeOff)) // &
                    ' The field can only contain the value: DirectIntoZone.')
  END IF

  !set component inlet and outlet node names, including system name from doasNameOff as prefix
  ! First component will always be AirLoopHVAC:OutdoorAirSystem with outlet node name <doasNameOff> Mixed Air Outlet
  lastOutlet = TRIM(FldVal(base + doasNameOff))//' Mixed Air Outlet'

  ! humidifier
  IF (.NOT. isHumidifierNone) THEN
    humidifierInlet  = TRIM(lastOutlet)
    humidifierOutlet = TRIM(FldVal(base + doasNameOff))//' Humidifier Outlet'
    lastOutlet       = TRIM(humidifierOutlet)
  END IF

  ! blow thru fan
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    fanInlet         = TRIM(lastOutlet)
    fanOutlet        = TRIM(FldVal(base + doasNameOff))//' Supply Fan Outlet'
    lastOutlet       = TRIM(fanOutlet)
  END IF

  ! cooling coil
  IF (coolCoilKind .NE. ccNone) THEN
    coolCoilUnitInlet = TRIM(lastOutlet)
    IF ((coolCoilKind .EQ. ccWaterCoilSystemHX) .OR. (coolCoilKind .EQ. ccDXcoilSystemHX)) THEN
      ! cooling coil with heat exchanger assist
      coolCoilUnitOutlet = TRIM(FldVal(base + doasNameOff))//' Cooling Coil HX Unit Outlet'
    ELSE
      ! simple cooling coil (without heat exchanger assist)
      coolCoilUnitOutlet = TRIM(FldVal(base + doasNameOff))//' Cooling Coil Outlet'
    END IF
    lastOutlet       = TRIM(coolCoilUnitOutlet)
  ELSE ! If there is no cooling coil, set to fan outlet in case the cooling setpoint is needed for heat recovery controls
      coolCoilUnitOutlet = TRIM(FldVal(base + doasNameOff))//' Supply Fan Outlet'
  END IF

  ! desuperheater reheat coil
  IF (dehumidCtrlKind .EQ. dehumidCoolRhtDesuper) THEN
    reheatCoilInlet    = TRIM(lastOutlet)
    reheatCoilOutlet   = TRIM(FldVal(base + doasNameOff))//' Reheat Coil Outlet'
    lastOutlet       = TRIM(reheatCoilOutlet)
  END IF

  ! heating coil
  IF (heatCoilType .NE. ctNone) THEN
    heatCoilInlet    = TRIM(lastOutlet)
    heatCoilOutlet   = TRIM(FldVal(base + doasNameOff))//' Heating Coil Outlet'
    lastOutlet       = TRIM(heatCoilOutlet)
  END IF

  ! draw thru fan
  IF (supFanPlacement .EQ. sfpDrawThru) THEN
    fanInlet         = TRIM(lastOutlet)
    fanOutlet        = TRIM(FldVal(base + doasNameOff))//' Supply Fan Outlet'
    lastOutlet       = TRIM(fanOutlet)
  END IF

  ! set AirLoopHVAC Supply Side Outlet Node Name
  airloopOutlet      = TRIM(lastOutlet)

  !***Sizing:System
  CALL CreateNewObj('Sizing:System')
  CALL AddToObjFld('AirLoop Name', base + doasNameOff,' ')
  CALL AddToObjStr('Type of Load to Size On','VentilationRequirement')
  CALL AddToObjFld('Design Outdoor Air Flow Rate {m3/s}', base + doasSupplyRateOff,' ')
  !Currently support only constant volume
  CALL AddToObjStr('Central Heating Maximum System Air Flow Ratio','1.0')
  CALL AddToObjStr('Preheat Design Temperature {C}', '2')
  CALL AddToObjStr('Preheat Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjStr('Precool Design Temperature {C}','11')
  CALL AddToObjStr('Precool Design Humidity Ratio {kg-H2O/kg-air}','0.008')
  CALL AddToObjFld('Central Cooling Design Supply Air Temperature {C}', base + doasCoolSetPtDesignOff,'')
  CALL AddToObjFld('Central Heating Design Supply Air Temperature {C}', base + doasHeatSetPtDesignOff,'')
  CALL AddToObjStr('Type of Zone Sum to Use','NonCoincident')
  CALL AddToObjStr('100% Outdoor Air in Cooling','Yes')
  CALL AddToObjStr('100% Outdoor Air in Heating','Yes')
  CALL AddToObjFld('Central Cooling Design Supply Air Humidity Ratio {kg-H2O/kg-air}', base + doasDehumSetPtOff,'')
  CALL AddToObjFld('Central Heating Design Supply Air Humidity Ratio {kg-H2O/kg-air}', base + doasHumidSetPtOff,'')
  IF (isSupplyFlowRateAutosize) THEN
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','DesignDay')
    CALL AddToObjStr('Cooling Supply Air Flow Rate {m3/s}','0')
    ! New fields for v8.2 next three rows
    CALL AddToObjStr('Cooling Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
    CALL AddToObjStr('Cooling Fraction of Autosized Cooling Supply Air Flow Rate','')
    CALL AddToObjStr('Cooling Supply Air Flow Rate Per Unit Cooling Capacity {m3/s-W}','')
    CALL AddToObjStr('Heating Supply Air Flow Rate Method','DesignDay')
    CALL AddToObjStr('Heating Supply Air Flow Rate {m3/s}','0')
  ELSE
    CALL AddToObjStr('Cooling Supply Air Flow Rate Method','Flow/System')
    CALL AddToObjFld('Cooling Supply Air Flow Rate {m3/s}', base + doasSupplyRateOff,' ')
    ! New fields for v8.2 next three rows
    CALL AddToObjStr('Cooling Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
    CALL AddToObjStr('Cooling Fraction of Autosized Cooling Supply Air Flow Rate','')
    CALL AddToObjStr('Cooling Supply Air Flow Rate Per Unit Cooling Capacity {m3/s-W}','')
    CALL AddToObjStr('Heating Supply Air Flow Rate Method','Flow/System')
    CALL AddToObjFld('Heating Supply Air Flow Rate {m3/s}', base + doasSupplyRateOff,' ')
  END IF
  ! New fields for v8.2 next four rows
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Floor Area {m3/s-m2}','')
  CALL AddToObjStr('Heating Fraction of Autosized Heating Supply Air Flow Rate','')
  CALL AddToObjStr('Heating Fraction of Autosized Cooling Supply Air Flow Rate','')
  CALL AddToObjStr('Heating Supply Air Flow Rate Per Unit Heating Capacity {m3/s-W}','')
  CALL AddToObjStr('System Outdoor Air Method','ZoneSum')
  CALL AddToObjStr('Zone Maximum Outdoor Air Fraction','1.0')
  CALL AddToObjStr('Cooling Design Capacity Method','CoolingDesignCapacity')
  CALL AddToObjStr('Cooling Design Capacity {W}','autosize')
  CALL AddToObjStr('Cooling Design Capacity Per Floor Area {W/m2}','')
  CALL AddToObjStr('Fraction of Autosized Cooling Design Capacity','')
  CALL AddToObjStr('Heating Design Capacity Method','HeatingDesignCapacity')
  CALL AddToObjStr('Heating Design Capacity {W}','autosize')
  CALL AddToObjStr('Heating Design Capacity Per Floor Area {W/m2}','')
  CALL AddToObjStr('Fraction of Autosized Heating Design Capacity','')
  CALL AddToObjStr('Central Cooling Capacity Control Method','OnOff',.true.)
  !***AirLoopHVAC
  CALL CreateNewObj('AirLoopHVAC')
  CALL AddToObjFld('Name', base + doasNameOff,' ')
  IF ((coolCoilKind .EQ. ccChWater) .OR. (coolCoilKind .EQ. ccChWaterDetailed) .OR. (coolCoilKind .EQ. ccWaterCoilSystemHX) &
       .OR. (heatCoilType .EQ. ctHotWater)) THEN
    CALL AddToObjFld('Controller List Name', base + doasNameOff,' Controllers')
  ELSE
    CALL AddToObjStr('Controller List Name','')
  ENDIF
  CALL AddToObjFld('Availability Manager List Name', base + doasNameOff,' Availability Managers')
  CALL AddToObjStr('Design Supply Air Flow Rate {m3/s}', 'autosize')
  CALL AddToObjFld('Branch List Name', base + doasNameOff,' Branches')
  CALL AddToObjStr('Connector List Name','')
  CALL AddToObjFld('Supply Side Inlet Node Name', base + doasNameOff,' Air Loop Inlet')
  CALL AddToObjFld('Demand Side Outlet Node Name', base + doasNameOff,' Return Air Outlet')
  CALL AddToObjFld('Demand Side Inlet Node Names', base + doasNameOff,' Supply Path Inlet')
  CALL AddToObjStr('Supply Side Outlet Node Names', TRIM(airloopOutlet),.TRUE.)
  !if there is a chilled water coil
  IF ((coolCoilKind .EQ. ccChWater) .OR. (coolCoilKind .EQ. ccChWaterDetailed) .OR. (coolCoilKind .EQ. ccWaterCoilSystemHX)) THEN
  !if doasHeatCoilKindOff=Hot Water
    IF (heatCoilType .EQ. ctHotWater) THEN
      !***AirLoopHVAC:ControllerList
      CALL CreateNewObj('AirLoopHVAC:ControllerList')
      CALL AddToObjFld('Name', base + doasNameOff,' Controllers')
      CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
      CALL AddToObjFld('Controller Name', base + doasNameOff,' Cooling Coil Controller')
      CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
      CALL AddToObjFld('Controller Name', base + doasNameOff,' Heating Coil Controller',.TRUE.)
    ELSE
      !***AirLoopHVAC:ControllerList
      CALL CreateNewObj('AirLoopHVAC:ControllerList')
      CALL AddToObjFld('Name', base + doasNameOff,' Controllers')
      CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
      CALL AddToObjFld('Controller Name', base + doasNameOff,' Cooling Coil Controller',.TRUE.)
    END IF
  ELSE
    IF (heatCoilType .EQ. ctHotWater) THEN
      !***AirLoopHVAC:ControllerList
      CALL CreateNewObj('AirLoopHVAC:ControllerList')
      CALL AddToObjFld('Name', base + doasNameOff,' Controllers')
      CALL AddToObjStr('Controller Object Type','Controller:WaterCoil')
      CALL AddToObjFld('Controller Name', base + doasNameOff,' Heating Coil Controller',.TRUE.)
    END IF
  END IF
  !***BranchList
  CALL CreateNewObj('BranchList')
  CALL AddToObjFld('Name', base + doasNameOff,' Branches')
  CALL AddToObjFld('Branch Name', base + doasNameOff,' Main Branch',.TRUE.)
  !***Branch
  CALL CreateNewObj('Branch')
  CALL AddToObjFld('Name', base + doasNameOff,' Main Branch')
  CALL AddToObjStr('Pressure Drop Curve Name','')
  CALL AddToObjStr('Component Object Type','AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Component Name', base + doasNameOff,' OA System')
  CALL AddToObjFld('Component Inlet Node Name', base + doasNameOff,' Air Loop Inlet')
  CALL AddToObjFld('Component Outlet Node Name', base + doasNameOff,' Mixed Air Outlet')
  IF (humidifierKind .EQ. humidifyElecSteam) THEN
    CALL AddToObjStr('Component Object Type','Humidifier:Steam:Electric')
    CALL AddToObjFld('Component Name', base + doasNameOff,' Humidifier')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(humidifierInlet))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(humidifierOutlet))
  END IF
  IF (supFanPlacement .EQ. sfpBlowThru) THEN
    CALL AddToObjStr('Component Object Type','Fan:VariableVolume')
    CALL AddToObjFld('Component Name', base + doasNameOff,' Supply Fan')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(fanInlet))
    IF ((coolCoilKind .NE. ccNone) .OR. (heatCoilType .NE. ctNone)) THEN
      CALL AddToObjStr('Component Outlet Node Name', TRIM(fanOutlet))
    ELSE
      CALL AddToObjStr('Component Outlet Node Name', TRIM(fanOutlet),.TRUE.)
    END IF
  END IF
  IF (coolCoilKind .NE. ccNone) THEN
    IF (coolCoilKind .EQ. ccChWater) THEN
      CALL AddToObjStr('Component Object Type','Coil:Cooling:Water')
      CALL AddToObjFld('Component Name', base + doasNameOff,' Cooling Coil')
    ELSEIF (coolCoilKind .EQ. ccChWaterDetailed) THEN
      CALL AddToObjStr('Component Object Type','Coil:Cooling:Water:DetailedGeometry')
      CALL AddToObjFld('Component Name', base + doasNameOff,' Cooling Coil')
    ELSEIF (coolCoilKind .EQ. ccTwoSpeedDX) THEN
      CALL AddToObjStr('Component Object Type','CoilSystem:Cooling:DX')
      CALL AddToObjFld('Component Name', base + doasNameOff,' DX Cooling System')
    ELSEIF (coolCoilKind .EQ. ccTwoStageDX) THEN
      CALL AddToObjStr('Component Object Type','CoilSystem:Cooling:DX')
      CALL AddToObjFld('Component Name', base + doasNameOff,' DX Cooling System')
    ELSEIF (coolCoilKind .EQ. ccTwoStageHumidControlDX) THEN
      CALL AddToObjStr('Component Object Type','CoilSystem:Cooling:DX')
      CALL AddToObjFld('Component Name', base + doasNameOff,' DX Cooling System')
    ELSEIF (coolCoilKind .EQ. ccWaterCoilSystemHX) THEN
      CALL AddToObjStr('Component Object Type','CoilSystem:Cooling:Water:HeatExchangerAssisted')
      CALL AddToObjFld('Component Name', base + doasNameOff,' Heat Exchanger Assisted Cooling Coil')
    ELSEIF (coolCoilKind .EQ. ccDXcoilSystemHX) THEN
      CALL AddToObjStr('Component Object Type','CoilSystem:Cooling:DX')
      CALL AddToObjFld('Component Name', base + doasNameOff,' DX Cooling System')
    END IF
    CALL AddToObjStr('Component Inlet Node Name', TRIM(coolCoilUnitInlet))
    IF ((heatCoilType .EQ. ctNone) .AND. (supFanPlacement .NE. sfpDrawThru)) THEN
      CALL AddToObjStr('Component Outlet Node Name', TRIM(coolCoilUnitOutlet),.TRUE.)
    ELSE
      CALL AddToObjStr('Component Outlet Node Name', TRIM(coolCoilUnitOutlet))
    END IF
  END IF
  IF (dehumidCtrlKind .EQ. dehumidCoolRhtDesuper) THEN
    CALL AddToObjStr('Component Object Type','Coil:Heating:Desuperheater')
    CALL AddToObjFld('Component Name', base + doasNameOff,' Reheat Coil')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(reheatCoilInlet))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(reheatCoilOutlet))
  END IF
  IF (heatCoilType .NE. ctNone) THEN
    IF (heatCoilType .EQ. ctHotWater) THEN
      CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    ELSEIF (heatCoilType .EQ. ctElectric) THEN
      CALL AddToObjStr('Component Object Type','Coil:Heating:Electric')
    ELSEIF (heatCoilType .EQ. ctGas) THEN
      CALL AddToObjStr('Component Object Type','Coil:Heating:Fuel')
    END IF
    CALL AddToObjFld('Component Name', base + doasNameOff,' Heating Coil')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(heatCoilInlet))
    IF (supFanPlacement .EQ. sfpDrawThru) THEN
      CALL AddToObjStr('Component Outlet Node Name', TRIM(heatCoilOutlet))
    ELSE
      CALL AddToObjStr('Component Outlet Node Name', TRIM(heatCoilOutlet),.TRUE.)
    END IF
  END IF
  IF (supFanPlacement .EQ. sfpDrawThru) THEN
    CALL AddToObjStr('Component Object Type','Fan:VariableVolume')
    CALL AddToObjFld('Component Name', base + doasNameOff,' Supply Fan')
    CALL AddToObjStr('Component Inlet Node Name', TRIM(fanInlet))
    CALL AddToObjStr('Component Outlet Node Name', TRIM(fanOutlet),.TRUE.)
  END IF
  !***AirLoopHVAC:SupplyPath
  CALL CreateNewObj('AirLoopHVAC:SupplyPath')
  CALL AddToObjFld('Name', base + doasNameOff,' Supply Path')
  CALL AddToObjFld('Supply Air Path Inlet Node Name', base + doasNameOff,' Supply Path Inlet')
  CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneSplitter')
  CALL AddToObjFld('Component Name', base + doasNameOff,' Zone Splitter',.TRUE.)
  !***AirLoopHVAC:ZoneSplitter
  CALL CreateNewObj('AirLoopHVAC:ZoneSplitter')
  CALL AddToObjFld('Name', base + doasNameOff,' Zone Splitter')
  CALL AddToObjFld('Inlet Node Name', base + doasNameOff,' Supply Path Inlet')
  CALL MakeSpecialIntList(zlDOAS, base + doasNameOff)
  DO i = 1,numIntList
    CALL AddToObjFld('Outlet Node Name', IntList(i),' Zone Equip Inlet', i .EQ. numIntList)
  END DO
  !***AirLoopHVAC:ReturnPath
  CALL CreateNewObj('AirLoopHVAC:ReturnPath')
  CALL AddToObjFld('Name', base + doasNameOff,' Return Path')
  CALL AddToObjFld('Return Air Path Outlet Node Name', base + doasNameOff,' Return Air Outlet')
  CALL AddToObjStr('Component Object Type','AirLoopHVAC:ZoneMixer')
  CALL AddToObjFld('Component Name', base + doasNameOff,' Zone Mixer',.TRUE.)
  !***AirLoopHVAC:ZoneMixer
  CALL CreateNewObj('AirLoopHVAC:ZoneMixer')
  CALL AddToObjFld('Name', base + doasNameOff,' Zone Mixer')
  CALL AddToObjFld('Outlet Node Name', base + doasNameOff,' Return Air Outlet')
  !the IntList holds the list of pointers to the zone names
  CALL MakeSpecialIntList(zlDOAS, base + doasNameOff)
  DO i = 1,numIntList
    CALL AddToObjFld('Inlet Node Name', IntList(i),' Return Outlet', i .EQ. numIntList)
  END DO
  !***AvailabilityManagerAssignmentList
  CALL CreateNewObj('AvailabilityManagerAssignmentList')
  CALL AddToObjFld('Name', base + doasNameOff,' Availability Managers')
  CALL AddToObjStr('Availability Manager Object Type','AvailabilityManager:Scheduled')
  CALL AddToObjFld('Availability Manager Name', base + doasNameOff,' Availability',.TRUE.)
  !***AvailabilityManager:Scheduled
  CALL CreateNewObj('AvailabilityManager:Scheduled')
  CALL AddToObjFld('Name', base + doasNameOff,' Availability')
  IF (isSysAvailSchedBlank) THEN
    CALL AddToObjStr('Schedule Name','HVACTemplate-Always 1',.TRUE.)
    CALL AddAlwaysSchedule('1')
  ELSE
    CALL AddToObjFld('Schedule Name', base + doasSysAvailSchedNameOff,'',.TRUE.)
  END IF

  ! Cooling coil setpoint manager (If there is no cooling coil and no heating coil but there is heat recovery, then need this)
  IF ((coolCoilKind .NE. ccNone) .OR. ((heatRecovery .NE. htrecNone) .AND. heatCoilType .EQ. ctNone)) THEN
    CALL CreateNewObj('NodeList')
    CALL AddToObjFld('Name', base + doasNameOff,' Cooling Setpoint Nodes')
    IF (dehumidCtrlKind .EQ. dehumidCoolRhtDesuper) THEN
      CALL AddToObjStr('Node Name', TRIM(reheatCoilOutlet))
    END IF
    CALL AddToObjStr('Node Name', TRIM(coolCoilUnitOutlet),.TRUE.)
    IF ((coolSetPtCtrlKind .EQ. cspFixed) .OR. (coolSetPtCtrlKind .EQ. cspScheduled)) THEN
      IF ((coolSetPtCtrlKind .EQ. cspFixed) .OR. (isCoolStPtSchedBlank)) THEN
        CALL AddAlwaysSchedule(FldVal(base + doasCoolSetPtDesignOff))
      END IF
      !***SetpointManager:Scheduled
      CALL CreateNewObj('SetpointManager:Scheduled')
      CALL AddToObjFld('Name', base + doasNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      IF ((coolSetPtCtrlKind .EQ. cspFixed) .OR. (isCoolStPtSchedBlank)) THEN
        CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + doasCoolSetPtDesignOff)))
      ELSE
        CALL AddToObjFld('Schedule Name', base + doasCoolSetPtSchedNameOff,' ')
      END IF
    ELSE IF (coolSetPtCtrlKind .EQ. cspOAreset) THEN
      !***SetpointManager:OutdoorAirReset
      CALL CreateNewObj('SetpointManager:OutdoorAirReset')
      CALL AddToObjFld('Name', base + doasNameOff,' Cooling Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('Setpoint at Outdoor Low Temperature {C}', base + doasCoolSetPtAtODBLoOff,'')
      CALL AddToObjFld('Outdoor Low Temperature {C}', base + doasCoolResetODBLoOff,'')
      CALL AddToObjFld('Setpoint at Outdoor High Temperature {C}', base + doasCoolSetPtAtODBHiOff,'')
      CALL AddToObjFld('Outdoor High Temperature {C}', base + doasCoolResetODBHiOff,'')
    END IF
    IF (supFanPlacement .EQ. sfpDrawThru) THEN
      CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(airloopOutlet),.TRUE.)
    ELSE
      CALL AddToObjFld('Setpoint Node or NodeList Name', base + doasNameOff,' Cooling Setpoint Nodes',.TRUE.)
    END IF
    IF ((supFanPlacement .EQ. sfpDrawThru) .AND. (coolCoilKind .NE. ccNone)) THEN
      !***SetpointManager:MixedAir
      CALL CreateNewObj('SetpointManager:MixedAir')
      CALL AddToObjFld('Name', base + doasNameOff,' Cooling Coil Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjStr('Reference Setpoint Node Name', TRIM(airloopOutlet))
      CALL AddToObjStr('Fan Inlet Node Name', TRIM(fanInlet))
      CALL AddToObjStr('Fan Outlet Node Name', TRIM(fanOutlet))
      CALL AddToObjFld('Setpoint Node or NodeList Name', base + doasNameOff,' Cooling Setpoint Nodes',.TRUE.)
    END IF
  END IF

  SELECT CASE (coolCoilKind)
    CASE (ccChWaterDetailed)
      !***Coil:Cooling:Water:DetailedGeometry
      CALL CreateNewObj('Coil:Cooling:Water:DetailedGeometry')
      CALL AddToObjFld('Name', base + doasNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + doasCoolAvailSchedNameOff,' ')
      CALL AddToObjStr('Max Water Flow Rate of Coil {m3/s}','autosize')
      CALL AddToObjStr('Tube Outside Surf Area {m2}','autosize')
      CALL AddToObjStr('Total Tube Inside Area {m2}','autosize')
      CALL AddToObjStr('Fin Surface Area {m2}','autosize')
      CALL AddToObjStr('Minimum Air Flow Area {m2}','autosize')
      CALL AddToObjStr('Coil Depth {m}','autosize')
      CALL AddToObjStr('Fin Diameter {m}','autosize')
      CALL AddToObjStr('Fin Thickness {m}','0.0015')
      CALL AddToObjStr('Tube Inside Diameter {m}','0.01445')
      CALL AddToObjStr('Tube Outside Diameter {m}','0.0159')
      CALL AddToObjStr('Tube Thermal Conductivity {W/m-K}','386.0')
      CALL AddToObjStr('Fin Thermal Conductivity {W/m-K}','204.0')
      CALL AddToObjStr('Fin Spacing {m}','0.0018')
      CALL AddToObjStr('Tube Depth Spacing {m}','0.026')
      CALL AddToObjStr('Number of Tube Rows','4')
      CALL AddToObjStr('Number of Tubes per Row','autosize')
      CALL AddToObjFld('Coil Water Inlet Node', base + doasNameOff,' Cooling Coil ChW Inlet')
      CALL AddToObjFld('Coil Water Outlet Node', base + doasNameOff,' Cooling Coil ChW Outlet')
      CALL AddToObjStr('Air Inlet Node Name', TRIM(coolCoilUnitInlet))
      CALL AddToObjStr('Air Outlet Node Name',TRIM(coolCoilUnitOutlet),.TRUE.)
      !***Controller:WaterCoil
      CALL CreateNewObj('Controller:WaterCoil')
      CALL AddToObjFld('Name', base + doasNameOff,' Cooling Coil Controller')
      IF (isDehumidifyNone) THEN
        CALL AddToObjStr('Control Variable','Temperature')
      ELSE
        CALL AddToObjStr('Control Variable','TemperatureAndHumidityRatio')
      END IF
      CALL AddToObjStr('Action','Reverse')
      CALL AddToObjStr('Actuator Variable','Flow')
      CALL AddToObjStr('Sensor Node Name', TRIM(coolCoilUnitOutlet))
      CALL AddToObjFld('Actuator Node Name', base + doasNameOff,' Cooling Coil ChW Inlet')
      CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
      CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
      CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)
    CASE (ccChWater)
      !***Coil:Cooling:Water
      CALL CreateNewObj('Coil:Cooling:Water')
      CALL AddToObjFld('Name', base + doasNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + doasCoolAvailSchedNameOff,' ')
      CALL AddToObjStr('Design Water Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Design Air Flow Rate {m3/s}', 'autosize')
      CALL AddToObjStr('Design Inlet Water Temperature {C}','autosize')
      CALL AddToObjStr('Design Inlet Air Temperature {C}','autosize')
      CALL AddToObjStr('Design Outlet Air Temperature {C}','autosize')
      CALL AddToObjStr('Design Inlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
      CALL AddToObjStr('Design Outlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
      CALL AddToObjFld('Water Inlet Node Name', base + doasNameOff,' Cooling Coil ChW Inlet')
      CALL AddToObjFld('Water Outlet Node Name', base + doasNameOff,' Cooling Coil ChW Outlet')
      CALL AddToObjStr('Air Inlet Node Name', TRIM(coolCoilUnitInlet))
      CALL AddToObjStr('Air Outlet Node Name',TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('Type of Analysis','DetailedAnalysis')
      CALL AddToObjStr('Heat Exchanger Configuration','CrossFlow',.TRUE.)
      !***Controller:WaterCoil
      CALL CreateNewObj('Controller:WaterCoil')
      CALL AddToObjFld('Name', base + doasNameOff,' Cooling Coil Controller')
      IF (isDehumidifyNone) THEN
        CALL AddToObjStr('Control Variable','Temperature')
      ELSE
        CALL AddToObjStr('Control Variable','TemperatureAndHumidityRatio')
      END IF
      CALL AddToObjStr('Action','Reverse')
      CALL AddToObjStr('Actuator Variable','Flow')
      CALL AddToObjStr('Sensor Node Name', TRIM(coolCoilUnitOutlet))
      CALL AddToObjFld('Actuator Node Name', base + doasNameOff,' Cooling Coil ChW Inlet')
      CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
      CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
      CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)
    CASE (ccTwoSpeedDX)
      !***CoilSystem:Cooling:DX
      dxCoilObjectType = 'Coil:Cooling:DX:TwoSpeed'
      CALL CreateNewObj('CoilSystem:Cooling:DX')
      CALL AddToObjFld('Name',base + doasNameOff,' DX Cooling System')
      CALL AddToObjFld('Availability Schedule Name', base + doasCoolAvailSchedNameOff,' ')
      CALL AddToObjStr('DX Cooling Coil System Inlet Node Name', TRIM(coolCoilUnitInlet))
      CALL AddToObjStr('DX Cooling Coil System Outlet Node Name', TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('DX Cooling Coil System Sensor Node Name', TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:DX:TwoSpeed')
      CALL AddToObjFld('Cooling Coil Name',base + doasNameOff,' Cooling Coil')
      IF ((dehumidCtrlKind .EQ. dehumidCoolRhtHtgCoil) .OR. (dehumidCtrlKind .EQ. dehumidCoolRhtDesuper)) THEN
        CALL AddToObjStr('Dehumidification Control Type','CoolReheat')
      ELSE
        CALL AddToObjStr('Dehumidification Control Type','None')
      END IF
      CALL AddToObjStr('Run on Sensible Load','Yes')
      ! CR8813 - MJW - Delete isHumidifierNone IF (.NOT. isDehumidifyNone .OR. .NOT. isHumidifierNone) THEN
      IF (.NOT. isDehumidifyNone) THEN
        CALL AddToObjStr('Run on Latent Load','Yes')
      ELSE
        CALL AddToObjStr('Run on Latent Load','No')
      END IF
      CALL AddToObjStr('Use Outdoor Air DX Cooling Coil','Yes')
      CALL AddToObjStr('Outdoor Air DX Cooling Coil Leaving Minimum Air Temperature','2.0', .TRUE.)
      !***Coil:Cooling:DX:TwoSpeed
      CALL CreateNewObj('Coil:Cooling:DX:TwoSpeed')
      CALL AddToObjFld('Name', base + doasNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + doasCoolAvailSchedNameOff,' ')
      CALL AddToObjFld('Rated High Speed Total Cooling Capacity', base + doasCoolCoilCapOff, '')
      CALL AddToObjFld('Rated High Speed Sensible Heat Ratio', base + doasCoolCoilSHROff, '')
      CALL AddToObjFld('Rated High Speed COP', base + doasCoolCoilCOPOff, '')
      CALL AddToObjStr('Rated High Speed Air Flow Rate', 'autosize')
      ! DX Ratings - Jul 2012
      CALL AddToObjStr('Unit Internal Static Air Pressure','')
      CALL AddToObjStr('Air Inlet Node Name', TRIM(coolCoilUnitInlet))
      CALL AddToObjStr('Air Outlet Node Name',TRIM(coolCoilUnitOutlet))
      CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
         base + doasNameOff,' DX Coil Cap-FT')
      CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
         base + doasNameOff,' DX Coil Cap-FF')
      CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
         base + doasNameOff,' DX Coil EIR-FT')
      CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
         base + doasNameOff,' DX Coil EIR-FF')
      CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
         base + doasNameOff,' DX Coil PLF')
      IF (isCoolCapAutosize) THEN
        CALL AddToObjStr('Rated Low Speed Total Cooling Capacity','Autosize')
      ELSE
        ! If high speed capacity is specified, assume that low speed is 1/3 of high speed capacity, same assumption that autosizing uses
        CALL AddToObjNum('Rated Low Speed Total Cooling Capacity',StringToReal(FldVal(base + doasCoolCoilCapOff)) * 0.33)
      END IF
      ! assume low speed sensible heat ratio same as high speed
      CALL AddToObjFld('Rated Low Speed Sensible Heat Ratio', base + doasCoolCoilSHROff, '')
      CALL AddToObjNum('Rated Low Speed COP',StringToReal(FldVal(base + doasCoolCoilCOPOff)) * 1.5)
      ! assume low speed air flow rate is automatically sized
      CALL AddToObjStr('Rated Low Speed Air Flow Rate','Autosize')
      ! assume the same curves as full speed
      CALL AddToObjFld('Low Speed Total Cooling Capacity Function of Temperature Curve Name',   &
         base + doasNameOff,' DX Coil Low Cap-FT')
      CALL AddToObjFld('Low Speed Energy Input Ratio Function of Temperature Curve Name',   &
         base + doasNameOff,' DX Coil Low EIR-FT',.TRUE.)
      !***Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('DOE-2.1E, SDL-C4')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil Cap-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.476428')
      CALL AddToObjStr('Coefficient2 x','0.401147E-01')
      CALL AddToObjStr('Coefficient3 x**2','0.226411E-03')
      CALL AddToObjStr('Coefficient4 y','-0.827136E-03')
      CALL AddToObjStr('Coefficient5 y**2','-0.732240E-05')
      CALL AddToObjStr('Coefficient6 x*y','-0.446278E-03')
      CALL AddToObjStr('Minimum Value of x','12.77778')
      CALL AddToObjStr('Maximum Value of x','23.88889')
      CALL AddToObjStr('Minimum Value of y','23.88889')
      CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
      !***Curve:Cubic
      CALL CreateNewObj('Curve:Cubic')
      CALL WriteComment('DOE-2.1E, SDL-C78')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil Cap-FF')
      CALL AddToObjStr('Coefficient1 Constant','.47278589')
      CALL AddToObjStr('Coefficient2 x','1.2433415')
      CALL AddToObjStr('Coefficient3 x**2','-1.0387055')
      CALL AddToObjStr('Coefficient4 x**3','.32257813')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !***Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('DOE-2.1E, SDL-C82')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil EIR-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.632475')
      CALL AddToObjStr('Coefficient2 x',' -0.121321E-01')
      CALL AddToObjStr('Coefficient3 x**2','0.507773E-03')
      CALL AddToObjStr('Coefficient4 y','0.155377E-01')
      CALL AddToObjStr('Coefficient5 y**2','0.272840E-03')
      CALL AddToObjStr('Coefficient6 x*y','-0.679201E-03')
      CALL AddToObjStr('Minimum Value of x','12.77778')
      CALL AddToObjStr('Maximum Value of x','23.88889')
      CALL AddToObjStr('Minimum Value of y','23.88889')
      CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
      !***Curve:Cubic
      CALL CreateNewObj('Curve:Cubic')
      CALL WriteComment('DOE-2.1E, SDL-C82')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil EIR-FF')
      CALL AddToObjStr('Coefficient1 Constant','1.0079484')
      CALL AddToObjStr('Coefficient2 x','.34544129')
      CALL AddToObjStr('Coefficient3 x**2','-.6922891')
      CALL AddToObjStr('Coefficient4 x**3','.33889943')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !***Curve:Quadratic
      CALL CreateNewObj('Curve:Quadratic')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil PLF')
      CALL AddToObjStr('Coefficient1 Constant','0.85')
      CALL AddToObjStr('Coefficient2 x','0.15')
      CALL AddToObjStr('Coefficient3 x**2','0.0')
      CALL AddToObjStr('Minimum Value of x','0.0')
      CALL AddToObjStr('Maximum Value of x','1.0',.TRUE.)
      !***Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('DOE-2.1E, SDL-C4')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil Low Cap-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.476428E+00')
      CALL AddToObjStr('Coefficient2 x','0.401147E-01')
      CALL AddToObjStr('Coefficient3 x**2','0.226411E-03')
      CALL AddToObjStr('Coefficient4 y','-0.827136E-03')
      CALL AddToObjStr('Coefficient5 y**2','-0.732240E-05')
      CALL AddToObjStr('Coefficient6 x*y','-0.446278E-03')
      CALL AddToObjStr('Minimum Value of x','12.77778')
      CALL AddToObjStr('Maximum Value of x','23.88889')
      CALL AddToObjStr('Minimum Value of y','23.88889')
      CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
      !***Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil Low EIR-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.774645E+00')
      CALL AddToObjStr('Coefficient2 x','-0.343731E-01')
      CALL AddToObjStr('Coefficient3 x**2','0.783173E-03')
      CALL AddToObjStr('Coefficient4 y','0.146596E-01')
      CALL AddToObjStr('Coefficient5 y**2','0.488851E-03')
      CALL AddToObjStr('Coefficient6 x*y','-0.752036E-03')
      CALL AddToObjStr('Minimum Value of x','12.77778')
      CALL AddToObjStr('Maximum Value of x','23.88889')
      CALL AddToObjStr('Minimum Value of y','23.88889')
      CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
    CASE (ccTwoStageDX, ccTwoStageHumidControlDX)
      dxCoilObjectType = 'Coil:Cooling:DX:TwoStageWithHumidityControlMode'
      !***CoilSystem:Cooling:DX
      CALL CreateNewObj('CoilSystem:Cooling:DX')
      CALL AddToObjFld('Name',base + doasNameOff,' DX Cooling System')
      CALL AddToObjFld('Availability Schedule Name', base + doasCoolAvailSchedNameOff,' ')
      CALL AddToObjStr('DX Cooling Coil System Inlet Node Name', TRIM(coolCoilUnitInlet))
      CALL AddToObjStr('DX Cooling Coil System Outlet Node Name', TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('DX Cooling Coil System Sensor Node Name', TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:DX:TwoStageWithHumidityControlMode')
      CALL AddToObjFld('Cooling Coil Name',base + doasNameOff,' Cooling Coil')
      IF ((dehumidCtrlKind .EQ. dehumidCoolRhtHtgCoil) .OR. (dehumidCtrlKind .EQ. dehumidCoolRhtDesuper)) THEN
        CALL AddToObjStr('Dehumidification Control Type','CoolReheat')
      ELSE IF (dehumidCtrlKind .EQ. dehumidMultimode) THEN
        CALL AddToObjStr('Dehumidification Control Type','Multimode')
      ELSE
        CALL AddToObjStr('Dehumidification Control Type','None')
      END IF
      CALL AddToObjStr('Run on Sensible Load','Yes')
      ! CR8813 - MJW - Delete isHumidifierNone IF (.NOT. isDehumidifyNone .OR. .NOT. isHumidifierNone) THEN
      IF (.NOT. isDehumidifyNone) THEN
        CALL AddToObjStr('Run on Latent Load','Yes')
      ELSE
        CALL AddToObjStr('Run on Latent Load','No')
      END IF
      CALL AddToObjStr('Use Outdoor Air DX Cooling Coil','Yes')
      CALL AddToObjStr('Outdoor Air DX Cooling Coil Leaving Minimum Air Temperature','2.0', .TRUE.)
      !***Coil:Cooling:DX:TwoStageWithHumidityControlMode
      CALL CreateNewObj('Coil:Cooling:DX:TwoStageWithHumidityControlMode')
      CALL AddToObjFld('Name', base + doasNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + doasCoolAvailSchedNameOff,' ')
      CALL AddToObjStr('Air Inlet Node Name', TRIM(coolCoilUnitInlet))
      CALL AddToObjStr('Air Outlet Node Name',TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('Crankcase Heater Capacity','')
      CALL AddToObjStr('Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater','')
      CALL AddToObjStr('Number of Capacity Stages','2')
      IF (coolCoilKind .EQ. ccTwoStageHumidControlDX) THEN
        CALL AddToObjStr('Number of Enhanced Dehumidification Modes','1')
      ELSE
        CALL AddToObjStr('Number of Enhanced Dehumidification Modes','0')
      END IF
      CALL AddToObjStr('Normal Mode Stage 1 Coil Performance Object Type','CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Normal Mode Stage 1 Coil Performance Name',base + doasNameOff,' Standard Perf 1')
      CALL AddToObjStr('Normal Mode Stage 1+2 Coil Performance Object Type','CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Normal Mode Stage 1+2 Coil Performance Name',base + doasNameOff,' Standard Perf 1+2')
      IF (coolCoilKind .EQ. ccTwoStageHumidControlDX) THEN
        CALL AddToObjStr('Dehumidification Mode 1 Stage 1 Coil Performance Object Type','CoilPerformance:DX:Cooling')
        CALL AddToObjFld('Dehumidification Mode 1 Stage 1 Coil Performance Name',base + doasNameOff,' Dehumid Perf 1')
        CALL AddToObjStr('Dehumidification Mode 1 Stage 1+2 Coil Performance Object Type','CoilPerformance:DX:Cooling')
        CALL AddToObjFld('Dehumidification Mode 1 Stage 1+2 Coil Performance Name',base + doasNameOff,' Dehumid Perf 1+2')
      ELSE
        CALL AddToObjStr('Dehumidification Mode 1 Stage 1 Coil Performance Object Type',' ')
        CALL AddToObjStr('Dehumidification Mode 1 Stage 1 Coil Performance Name',' ')
        CALL AddToObjStr('Dehumidification Mode 1 Stage 1+2 Coil Performance Object Type',' ')
        CALL AddToObjStr('Dehumidification Mode 1 Stage 1+2 Coil Performance Name',' ')
      END IF
      CALL AddToObjStr('Supply Water Storage Tank Name','')
      CALL AddToObjStr('Condensate Collection Water Storage Tank Name','',.TRUE.)
      !***CoilPerformance:DX:Cooling
      ! Standard - Stage 1
      CALL CreateNewObj('CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Name',base + doasNameOff,' Standard Perf 1')
      IF (isCoolCapAutosize) THEN
        CALL AddToObjStr('Rated Total Cooling Capacity', 'autosize')
      ELSE
        ! half the capacity if provided
        CALL AddToObjNum('Rated Total Cooling Capacity', 0.5 * StringToReal(FldVal(base + doasCoolCoilCapOff)))
      END IF
      CALL AddToObjFld('Rated Sensible Heat Ratio', base + doasCoolCoilSHROff, '')
      CALL AddToObjFld('Rated COP', base + doasCoolCoilCOPOff, '')
      CALL AddToObjStr('Rated Air Flow Rate', 'autosize')
      CALL AddToObjStr('Fraction of Air Flow Bypassed Around Coil', '0.5')
      CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
         base + doasNameOff,' DX Coil Std Cap-FT')
      CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
         base + doasNameOff,' DX Coil Cap-FF')
      CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
         base + doasNameOff,' DX Coil Std EIR-FT')
      CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
         base + doasNameOff,' DX Coil EIR-FF')
      CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
         base + doasNameOff,' DX Coil PLF')
      CALL AddToObjStr('Nominal Time for Condensate Removal to Begin', '')
      CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity', '')
      CALL AddToObjStr('Maximum Cycling Rate', '')
      CALL AddToObjStr('Latent Capacity Time Constant', '',.TRUE.)
      !***CoilPerformance:DX:Cooling
      ! Standard - Stage 1 + 2
      CALL CreateNewObj('CoilPerformance:DX:Cooling')
      CALL AddToObjFld('Name',base + doasNameOff,' Standard Perf 1+2')
      CALL AddToObjFld('Rated Total Cooling Capacity', base + doasCoolCoilCapOff,'')
      CALL AddToObjFld('Gross Rated Sensible Heat Ratio', base + doasCoolCoilSHROff, '')
      CALL AddToObjFld('Rated COP', base + doasCoolCoilCOPOff, '')
      CALL AddToObjStr('Rated Air Flow Rate', 'autosize')
      CALL AddToObjStr('Fraction of Air Flow Bypassed Around Coil', '0.0')
      CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
         base + doasNameOff,' DX Coil Std Cap-FT')
      CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
         base + doasNameOff,' DX Coil Cap-FF')
      CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
         base + doasNameOff,' DX Coil Std EIR-FT')
      CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
         base + doasNameOff,' DX Coil EIR-FF')
      CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
         base + doasNameOff,' DX Coil PLF')
      CALL AddToObjStr('Nominal Time for Condensate Removal to Begin', '')
      CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity', '')
      CALL AddToObjStr('Maximum Cycling Rate', '')
      CALL AddToObjStr('Latent Capacity Time Constant', '',.TRUE.)

      IF (coolCoilKind .EQ. ccTwoStageHumidControlDX) THEN
        !***CoilPerformance:DX:Cooling
        ! Dehumidification (subcool) - Stage 1
        CALL CreateNewObj('CoilPerformance:DX:Cooling')
        CALL AddToObjFld('Name',base + doasNameOff,' Dehumid Perf 1')
        IF (isCoolCapAutosize) THEN
          CALL AddToObjStr('Rated Total Cooling Capacity', 'autosize')
        ELSE
          ! half the capacity if provided
          CALL AddToObjNum('Rated Total Cooling Capacity', 0.5 * StringToReal(FldVal(base + doasCoolCoilCapOff)))
        END IF
        IF (isSHRAutosize) THEN
          CALL AddToObjStr('Gross Rated Sensible Heat Ratio','autosize')
        ELSE
          CALL AddToObjNum('Gross Rated Sensible Heat Ratio', 0.9 * StringToReal(FldVal(base + doasCoolCoilSHROff)))
        ENDIF
        CALL AddToObjNum('Rated COP', 0.9 * StringToReal(FldVal(base + doasCoolCoilCOPOff)))
        CALL AddToObjStr('Rated Air Flow Rate', 'autosize')
        CALL AddToObjStr('Fraction of Air Flow Bypassed Around Coil', '0.5')
        CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
           base + doasNameOff,' DX Coil Dehum Cap-FT')
        CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
           base + doasNameOff,' DX Coil Cap-FF')
        CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
           base + doasNameOff,' DX Coil Dehum EIR-FT')
        CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
           base + doasNameOff,' DX Coil EIR-FF')
        CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
           base + doasNameOff,' DX Coil PLF')
        CALL AddToObjStr('Nominal Time for Condensate Removal to Begin', '')
        CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity', '')
        CALL AddToObjStr('Maximum Cycling Rate', '')
        CALL AddToObjStr('Latent Capacity Time Constant', '',.TRUE.)
        !***CoilPerformance:DX:Cooling
        ! Dehumidification (subcool) - Stage 1 + 2
        CALL CreateNewObj('CoilPerformance:DX:Cooling')
        CALL AddToObjFld('Name',base + doasNameOff,' Dehumid Perf 1+2')
        CALL AddToObjFld('Rated Total Cooling Capacity', base + doasCoolCoilCapOff,'')
        IF (isSHRAutosize) THEN
          CALL AddToObjStr('Gross Rated Sensible Heat Ratio','autosize')
        ELSE
          CALL AddToObjNum('Gross Rated Sensible Heat Ratio', 0.9 * StringToReal(FldVal(base + doasCoolCoilSHROff)))
        ENDIF
        CALL AddToObjNum('Rated COP', 0.9 * StringToReal(FldVal(base + doasCoolCoilCOPOff)))
        CALL AddToObjStr('Rated Air Flow Rate', 'autosize')
        CALL AddToObjStr('Fraction of Air Flow Bypassed Around Coil', '0.0')
        CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
           base + doasNameOff,' DX Coil Dehum Cap-FT')
        CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
           base + doasNameOff,' DX Coil Cap-FF')
        CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
           base + doasNameOff,' DX Coil Dehum EIR-FT')
        CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
           base + doasNameOff,' DX Coil EIR-FF')
        CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
           base + doasNameOff,' DX Coil PLF')
        CALL AddToObjStr('Nominal Time for Condensate Removal to Begin', '')
        CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity', '')
        CALL AddToObjStr('Maximum Cycling Rate', '')
        CALL AddToObjStr('Latent Capacity Time Constant', '',.TRUE.)
      ENDIF
      !***Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil Std Cap-FT')
      CALL AddToObjStr('Coefficient1 Constant','1.067939449')
      CALL AddToObjStr('Coefficient2 x','-0.031261829')
      CALL AddToObjStr('Coefficient3 x**2','0.001974308')
      CALL AddToObjStr('Coefficient4 y','-0.002726426')
      CALL AddToObjStr('Coefficient5 y**2','-5.52654E-05')
      CALL AddToObjStr('Coefficient6 x*y','-6.31169E-05')
      CALL AddToObjStr('Minimum Value of x','-100')
      CALL AddToObjStr('Maximum Value of x','100')
      CALL AddToObjStr('Minimum Value of y','-100')
      CALL AddToObjStr('Maximum Value of y','100',.TRUE.)
      IF (coolCoilKind .EQ. ccTwoStageHumidControlDX) THEN
        !***Curve:Biquadratic
        CALL CreateNewObj('Curve:Biquadratic')
        CALL AddToObjFld('Name', base + doasNameOff,' DX Coil Dehum Cap-FT')
        CALL AddToObjStr('Coefficient1 Constant','0.596779741')
        CALL AddToObjStr('Coefficient2 x','0.034216637')
        CALL AddToObjStr('Coefficient3 x**2','0.000113924')
        CALL AddToObjStr('Coefficient4 y','-0.00375859')
        CALL AddToObjStr('Coefficient5 y**2','-9.17495E-05')
        CALL AddToObjStr('Coefficient6 x*y','-8.98373E-05')
        CALL AddToObjStr('Minimum Value of x','-100')
        CALL AddToObjStr('Maximum Value of x','100')
        CALL AddToObjStr('Minimum Value of y','-100')
        CALL AddToObjStr('Maximum Value of y','100',.TRUE.)
      ENDIF
      !***Curve:Quadratic
      CALL CreateNewObj('Curve:Quadratic')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil Cap-FF')
      CALL AddToObjStr('Coefficient1 Constant','0.8')
      CALL AddToObjStr('Coefficient2 x','0.2')
      CALL AddToObjStr('Coefficient3 x**2','0.0')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !***Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil Std EIR-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.174059889')
      CALL AddToObjStr('Coefficient2 x','0.022281508')
      CALL AddToObjStr('Coefficient3 x**2','-0.000134077')
      CALL AddToObjStr('Coefficient4 y','0.028298025')
      CALL AddToObjStr('Coefficient5 y**2','0.00048510')
      CALL AddToObjStr('Coefficient6 x*y','-0.001677095')
      CALL AddToObjStr('Minimum Value of x','-100')
      CALL AddToObjStr('Maximum Value of x','100')
      CALL AddToObjStr('Minimum Value of y','-100')
      CALL AddToObjStr('Maximum Value of y','100',.TRUE.)
      IF (coolCoilKind .EQ. ccTwoStageHumidControlDX) THEN
        !***Curve:Biquadratic
        CALL CreateNewObj('Curve:Biquadratic')
        CALL AddToObjFld('Name', base + doasNameOff,' DX Coil Dehum EIR-FT')
        CALL AddToObjStr('Coefficient1 Constant','0.435347586')
        CALL AddToObjStr('Coefficient2 x','0.004015641')
        CALL AddToObjStr('Coefficient3 x**2','0.000604235')
        CALL AddToObjStr('Coefficient4 y','0.015824043')
        CALL AddToObjStr('Coefficient5 y**2','0.000747287')
        CALL AddToObjStr('Coefficient6 x*y','-0.001779745')
        CALL AddToObjStr('Minimum Value of x','-100')
        CALL AddToObjStr('Maximum Value of x','100')
        CALL AddToObjStr('Minimum Value of y','-100')
        CALL AddToObjStr('Maximum Value of y','100',.TRUE.)
      ENDIF
      !***Curve:Quadratic
      CALL CreateNewObj('Curve:Quadratic')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil EIR-FF')
      CALL AddToObjStr('Coefficient1 Constant','1.1552')
      CALL AddToObjStr('Coefficient2 x','-0.1808')
      CALL AddToObjStr('Coefficient3 x**2','0.0256')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !***Curve:Quadratic
      CALL CreateNewObj('Curve:Quadratic')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil PLF')
      CALL AddToObjStr('Coefficient1 Constant','0.85')
      CALL AddToObjStr('Coefficient2 x','0.15')
      CALL AddToObjStr('Coefficient3 x**2','0.0')
      CALL AddToObjStr('Minimum Value of x','0.0')
      CALL AddToObjStr('Maximum Value of x','1.0',.TRUE.)
    CASE (ccWaterCoilSystemHX)
      !***CoilSystem:Cooling:Water:HeatExchangerAssisted
      CALL CreateNewObj('CoilSystem:Cooling:Water:HeatExchangerAssisted')
      CALL AddToObjFld('Name', base + doasNameOff,' Heat Exchanger Assisted Cooling Coil')
      CALL AddToObjStr('Heat Exchanger Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
      CALL AddToObjFld('Heat Exchanger Name', base + doasNameOff,' Cooling Coil Heat Exchanger')
      CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:Water')
      CALL AddToObjFld('Cooling Coil Name', base + doasNameOff, ' Cooling Coil',.TRUE.)
      !***HeatExchanger:AirToAir:SensibleAndLatent
      CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
      CALL AddToObjFld('Name', base + doasNameOff,' Cooling Coil Heat Exchanger')
      CALL AddToObjStr('Availability Schedule Name',' ')
      CALL AddToObjStr('Nominal Supply Air Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Sensible Effectiveness at 100% Heating Air Flow {dimensionless}','0.7')
      CALL AddToObjStr('Latent Effectiveness at 100% Heating Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Sensible Effectiveness at 75% Heating Air Flow {dimensionless}','0.75')
      CALL AddToObjStr('Latent Effectiveness at 75% Heating Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Sensible Effectiveness at 100% Cooling Air Flow {dimensionless}','0.7')
      CALL AddToObjStr('Latent Effectiveness at 100% Cooling Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Sensible Effectiveness at 75% Cooling Air Flow {dimensionless}','0.75')
      CALL AddToObjStr('Latent Effectiveness at 75% Cooling Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Supply Air Inlet Node Name', TRIM(coolCoilUnitInlet))
      CALL AddToObjFld('Supply Air Outlet Node Name', base + doasNameOff,' HX Supply Outlet Node')
      CALL AddToObjFld('Exhaust Air Inlet Node Name', base + doasNameOff,' Cooling Coil Outlet Node')
      CALL AddToObjStr('Exhaust Air Outlet Node Name', TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('Nominal Electric Power {W}','0')
      CALL AddToObjStr('Supply Air Outlet Temperature Control','No')
      CALL AddToObjStr('Heat Exchanger Type','Plate')
      CALL AddToObjStr('Frost Control Type','None')
      CALL AddToObjStr('Threshold Temperature {C}','')
      CALL AddToObjStr('Initial Defrost Time Fraction {dimensionless}','')
      CALL AddToObjStr('Rate of Defrost Time Fraction Increase {1/K}','')
      ! MJW ???? - Not sure
      CALL AddToObjStr('Economizer Lockout','No',.TRUE.)
      !***Coil:Cooling:Water
      CALL CreateNewObj('Coil:Cooling:Water')
      CALL AddToObjFld('Name', base + doasNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + doasCoolAvailSchedNameOff,' ')
      CALL AddToObjStr('Design Water Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Design Air Flow Rate {m3/s}', 'autosize')
      CALL AddToObjStr('Design Inlet Water Temperature {C}','autosize')
      CALL AddToObjStr('Design Inlet Air Temperature {C}','autosize')
      CALL AddToObjStr('Design Outlet Air Temperature {C}','autosize')
      CALL AddToObjStr('Design Inlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
      CALL AddToObjStr('Design Outlet Air Humidity Ratio {kg-H2O/kg-air}','autosize')
      CALL AddToObjFld('Water Inlet Node Name', base + doasNameOff,' Cooling Coil ChW Inlet')
      CALL AddToObjFld('Water Outlet Node Name', base + doasNameOff,' Cooling Coil ChW Outlet')
      CALL AddToObjFld('Air Inlet Node Name', base + doasNameOff,' HX Supply Outlet Node')
      CALL AddToObjFld('Air Outlet Node Name', base + doasNameOff,' Cooling Coil Outlet Node')
      CALL AddToObjStr('Type of Analysis','DetailedAnalysis')
      CALL AddToObjStr('Heat Exchanger Configuration','CrossFlow',.TRUE.)
      !***Controller:WaterCoil
      CALL CreateNewObj('Controller:WaterCoil')
      CALL AddToObjFld('Name', base + doasNameOff,' Cooling Coil Controller')
      IF (isDehumidifyNone) THEN
        CALL AddToObjStr('Control Variable','Temperature')
      ELSE
        CALL AddToObjStr('Control Variable','TemperatureAndHumidityRatio')
      END IF
      CALL AddToObjStr('Action','Reverse')
      CALL AddToObjStr('Actuator Variable','Flow')
      CALL AddToObjStr('Sensor Node Name', TRIM(coolCoilUnitOutlet))
      CALL AddToObjFld('Actuator Node Name', base + doasNameOff,' Cooling Coil ChW Inlet')
      CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
      CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
      CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)
    CASE (ccDXcoilSystemHX)
      dxCoilObjectType = 'Coil:Cooling:DX:SingleSpeed'
      !***CoilSystem:Cooling:DX
      CALL CreateNewObj('CoilSystem:Cooling:DX')
      CALL AddToObjFld('Name',base + doasNameOff,' DX Cooling System')
      CALL AddToObjFld('Availability Schedule Name', base + doasCoolAvailSchedNameOff,' ')
      CALL AddToObjStr('DX Cooling Coil System Inlet Node Name', TRIM(coolCoilUnitInlet))
      CALL AddToObjStr('DX Cooling Coil System Outlet Node Name', TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('DX Cooling Coil System Sensor Node Name', TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('Cooling Coil Object Type','CoilSystem:Cooling:DX:HeatExchangerAssisted')
      CALL AddToObjFld('Cooling Coil Name',base + doasNameOff,' Heat Exchanger Assisted Cooling Coil')
      IF ((dehumidCtrlKind .EQ. dehumidCoolRhtHtgCoil) .OR. (dehumidCtrlKind .EQ. dehumidCoolRhtDesuper)) THEN
        CALL AddToObjStr('Dehumidification Control Type','CoolReheat')
      ELSE
        CALL AddToObjStr('Dehumidification Control Type','None')
      END IF
      CALL AddToObjStr('Run on Sensible Load','Yes')
      ! CR8813 - MJW - Delete isHumidifierNone IF (.NOT. isDehumidifyNone .OR. .NOT. isHumidifierNone) THEN
      IF (.NOT. isDehumidifyNone) THEN
        CALL AddToObjStr('Run on Latent Load','Yes')
      ELSE
        CALL AddToObjStr('Run on Latent Load','No')
      END IF
      CALL AddToObjStr('Use Outdoor Air DX Cooling Coil','Yes')
      CALL AddToObjStr('Outdoor Air DX Cooling Coil Leaving Minimum Air Temperature','2.0', .TRUE.)
      !***CoilSystem:Cooling:DX:HeatExchangerAssisted
      CALL CreateNewObj('CoilSystem:Cooling:DX:HeatExchangerAssisted')
      CALL AddToObjFld('Name', base + doasNameOff,' Heat Exchanger Assisted Cooling Coil')
      CALL AddToObjStr('Heat Exchanger Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
      CALL AddToObjFld('Heat Exchanger Name', base + doasNameOff,' Cooling Coil Heat Exchanger')
      CALL AddToObjStr('Cooling Coil Object Type','Coil:Cooling:DX:SingleSpeed')
      CALL AddToObjFld('Cooling Coil Name', base + doasNameOff,' Cooling Coil',.TRUE.)
      !***HeatExchanger:AirToAir:SensibleAndLatent
      CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
      CALL AddToObjFld('Name', base + doasNameOff,' Cooling Coil Heat Exchanger')
      CALL AddToObjStr('Availability Schedule Name',' ')
      CALL AddToObjStr('Nominal Supply Air Flow Rate {m3/s}','autosize')
      CALL AddToObjStr('Sensible Effectiveness at 100% Heating Air Flow {dimensionless}','0.7')
      CALL AddToObjStr('Latent Effectiveness at 100% Heating Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Sensible Effectiveness at 75% Heating Air Flow {dimensionless}','0.75')
      CALL AddToObjStr('Latent Effectiveness at 75% Heating Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Sensible Effectiveness at 100% Cooling Air Flow {dimensionless}','0.7')
      CALL AddToObjStr('Latent Effectiveness at 100% Cooling Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Sensible Effectiveness at 75% Cooling Air Flow {dimensionless}','0.75')
      CALL AddToObjStr('Latent Effectiveness at 75% Cooling Air Flow {dimensionless}','0.0')
      CALL AddToObjStr('Supply Air Inlet Node Name', TRIM(coolCoilUnitInlet))
      CALL AddToObjFld('Supply Air Outlet Node Name', base + doasNameOff,' HX Supply Outlet Node')
      CALL AddToObjFld('Exhaust Air Inlet Node Name', base + doasNameOff,' Cooling Coil Outlet Node')
      CALL AddToObjStr('Exhaust Air Outlet Node Name', TRIM(coolCoilUnitOutlet))
      CALL AddToObjStr('Nominal Electric Power {W}','0')
      CALL AddToObjStr('Supply Air Outlet Temperature Control','No')
      CALL AddToObjStr('Heat Exchanger Type','Plate')
      CALL AddToObjStr('Frost Control Type','None')
      CALL AddToObjStr('Threshold Temperature {C}','')
      CALL AddToObjStr('Initial Defrost Time Fraction {dimensionless}','')
      CALL AddToObjStr('Rate of Defrost Time Fraction Increase {1/K}','')
      ! MJW ???? - Not sure
      CALL AddToObjStr('Economizer Lockout','No',.TRUE.)
      !***Coil:Cooling:DX:SingleSpeed
      CALL CreateNewObj('Coil:Cooling:DX:SingleSpeed')
      CALL AddToObjFld('Name', base + doasNameOff,' Cooling Coil')
      CALL AddToObjFld('Availability Schedule Name', base + doasCoolAvailSchedNameOff,' ')
      CALL AddToObjFld('Rated Total Cooling Capacity', base + doasCoolCoilCapOff, '')
      CALL AddToObjFld('Gross Rated Sensible Heat Ratio', base + doasCoolCoilSHROff, '')
      CALL AddToObjFld('Rated COP', base + doasCoolCoilCOPOff, '')
      CALL AddToObjStr('Rated Air Flow Rate', 'autosize')
      CALL AddToObjStr('Rated Evaporator Fan Power per Volume Flow Rate {W/(m3/s)}','')
      CALL AddToObjFld('Air Inlet Node Name', base + doasNameOff,' HX Supply Outlet Node')
      CALL AddToObjFld('Air Outlet Node Name', base + doasNameOff,' Cooling Coil Outlet Node')
      CALL AddToObjFld('Total Cooling Capacity Function of Temperature Curve Name',   &
         base + doasNameOff,' DX Coil Cap-FT')
      CALL AddToObjFld('Total Cooling Capacity Function of Flow Fraction Curve Name',   &
         base + doasNameOff,' DX Coil Cap-FF')
      CALL AddToObjFld('Energy Input Ratio Function of Temperature Curve Name',   &
         base + doasNameOff,' DX Coil EIR-FT')
      CALL AddToObjFld('Energy Input Ratio Function of Flow Fraction Curve Name',   &
         base + doasNameOff,' DX Coil EIR-FF')
      CALL AddToObjFld('Part Load Fraction Correlation Curve Name',   &
         base + doasNameOff,' DX Coil PLF')
      CALL AddToObjStr('Minimum Outdoor Dry-Bulb Temperature for Compressor Operation {C}','')
      CALL AddToObjStr('Nominal Time for Condensate Removal to Begin','0')
      CALL AddToObjStr('Ratio of Initial Moisture Evaporation Rate and Steady State Latent Capacity','0')
      CALL AddToObjStr('Maximum Cycling Rate','0')
      CALL AddToObjStr('Latent Capacity Time Constant','0')
      CALL AddToObjFld('Condenser Air Inlet Node Name',base + doasNameOff,' Cooling Coil Condenser Inlet')
      CALL AddToObjStr('Condenser Type','AirCooled')
      CALL AddToObjStr('Evaporative Condenser Effectiveness','0')
      CALL AddToObjStr('Evaporative Condenser Air Flow Rate','')
      CALL AddToObjStr('Evaporative Condenser Pump Rated Power Consumption','0')
      CALL AddToObjStr('Crankcase Heater Capacity','0')
      CALL AddToObjStr('Maximum Outdoor Dry-Bulb Temperature for Crankcase Heater Operation','10',.TRUE.)
      !***Curve:Biquadratic
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('DOE-2.1E, COOL-CAP-FT for PTAC w/ SI temps')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil Cap-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.942587793')
      CALL AddToObjStr('Coefficient2 x','0.009543347')
      CALL AddToObjStr('Coefficient3 x**2','0.00068377')
      CALL AddToObjStr('Coefficient4 y','-0.011042676')
      CALL AddToObjStr('Coefficient5 y**2','0.000005249')
      CALL AddToObjStr('Coefficient6 x*y','-0.00000972')
      CALL AddToObjStr('Minimum Value of x','12.77778')
      CALL AddToObjStr('Maximum Value of x','23.88889')
      CALL AddToObjStr('Minimum Value of y','18.0')
      CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
      !CURVE:QUADRATIC
      CALL CreateNewObj('Curve:Quadratic')
      CALL WriteComment('DOE-2.1E, RATED-CCAP-FFLOW for PTAC')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil Cap-FF')
      CALL AddToObjStr('Coefficient1 Constant','0.8')
      CALL AddToObjStr('Coefficient2 x','0.2')
      CALL AddToObjStr('Coefficient3 x**2','0')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !CURVE:BIQUADRATIC
      CALL CreateNewObj('Curve:Biquadratic')
      CALL WriteComment('DOE-2.1E, COOL-EIR-FT for PTAC w/ SI temps')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil EIR-FT')
      CALL AddToObjStr('Coefficient1 Constant','0.342414409')
      CALL AddToObjStr('Coefficient2 x','0.034885008')
      CALL AddToObjStr('Coefficient3 x**2','-0.0006237')
      CALL AddToObjStr('Coefficient4 y','0.004977216')
      CALL AddToObjStr('Coefficient5 y**2','0.000437951')
      CALL AddToObjStr('Coefficient6 x*y','-0.000728028')
      CALL AddToObjStr('Minimum Value of x','12.77778')
      CALL AddToObjStr('Maximum Value of x','23.88889')
      CALL AddToObjStr('Minimum Value of y','18.0')
      CALL AddToObjStr('Maximum Value of y','46.11111',.TRUE.)
      !CURVE:QUADRATIC
      CALL CreateNewObj('Curve:Quadratic')
      CALL WriteComment('DOE-2.1E, RATED-CEIR-FFLOW for PTAC')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil EIR-FF')
      CALL AddToObjStr('Coefficient1 Constant','1.1552')
      CALL AddToObjStr('Coefficient2 x','-0.1808')
      CALL AddToObjStr('Coefficient3 x**2','0.0256')
      CALL AddToObjStr('Minimum Value of x','0.5')
      CALL AddToObjStr('Maximum Value of x','1.5',.TRUE.)
      !CURVE:QUADRATIC
      CALL CreateNewObj('Curve:Quadratic')
      CALL WriteComment('PLF = l.- Cd(1.-PLR) where Cd = 0.15')
      CALL AddToObjFld('Name', base + doasNameOff,' DX Coil PLF')
      CALL AddToObjStr('Coefficient1 Constant','0.85')
      CALL AddToObjStr('Coefficient2 x','0.15')
      CALL AddToObjStr('Coefficient3 x**2','0')
      CALL AddToObjStr('Minimum Value of x','0')
      CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
      !OutAir Node
      CALL CreateNewObj('OutdoorAir:Node')
      CALL AddToObjFld('Name',base + doasNameOff,' Cooling Coil Condenser Inlet')
      CALL AddToObjStr('Height Above Ground','-1',.TRUE.)
    CASE (ccNone)
      ! nothing to do if no cooling coil specified
  END SELECT

  !***Branch
  IF ((coolCoilKind .EQ. ccChWater) .OR. (coolCoilKind .EQ. ccChWaterDetailed) .OR. (coolCoilKind .EQ. ccWaterCoilSystemHX)) THEN
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + doasNameOff,' Cooling Coil ChW Branch')
    !store for later in the chilled water loop
    CALL AddToStrList(TRIM(FldVal(base + doasNameOff)) // ' Cooling Coil ChW Branch', handleCoolingCoilBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    IF ((coolCoilKind .EQ. ccChWater) .OR. (coolCoilKind .EQ. ccWaterCoilSystemHX)) THEN
      CALL AddToObjStr('Component Object Type','Coil:Cooling:Water')
    ELSEIF (coolCoilKind .EQ. ccChWaterDetailed) THEN
      CALL AddToObjStr('Component Object Type','Coil:Cooling:Water:DetailedGeometry')
    END IF
    CALL AddToObjFld('Component Name', base + doasNameOff,' Cooling Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + doasNameOff,' Cooling Coil ChW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + doasNameOff,' Cooling Coil ChW Outlet',.TRUE.)
  ENDIF

  ! Reheat coil - if dehumidification control is CoolReheatDesuperheater
  IF (dehumidCtrlKind .EQ. dehumidCoolRhtDesuper) THEN
    CALL CreateNewObj('Coil:Heating:Desuperheater')
    CALL AddToObjFld('Name', base + doasNameOff,' Reheat Coil')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjStr('Heat Reclaim Recovery Efficiency','0.3')
    CALL AddToObjStr('Air Inlet Node Name', TRIM(reheatCoilInlet))
    CALL AddToObjStr('Air Outlet Node Name', TRIM(reheatCoilOutlet))
    CALL AddToObjStr('Heating Source Object Type',TRIM(dxCoilObjectType))
    CALL AddToObjFld('Heating Source Name', base + doasNameOff,' Cooling Coil')
    CALL AddToObjStr('Temperature Setpoint Node Name',TRIM(reheatCoilOutlet))
    CALL AddToObjStr('Parasitic Electric Load {W}','',.TRUE.)
  ENDIF

  IF (heatCoilType .EQ. ctHotWater) THEN
    !***Coil:Heating:Water
    CALL CreateNewObj('Coil:Heating:Water')
    CALL AddToObjFld('Name', base + doasNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + doasHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('U-Factor Times Area Value {W/K}','autosize')
    CALL AddToObjStr('Maximum Water Flow Rate {m3/s}','autosize')
    CALL AddToObjFld('Water Inlet Node Name ', base + doasNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Water Outlet Node Name', base + doasNameOff,' Heating Coil HW Outlet')
    CALL AddToObjStr('Air Inlet Node Name',  TRIM(heatCoilInlet))
    CALL AddToObjStr('Air Outlet Node Name',  TRIM(heatCoilOutlet))
    CALL AddToObjStr('Performance Input Method','UFactorTimesAreaAndDesignWaterFlowRate')
    CALL AddToObjStr('Nominal Capacity','autosize')
    CALL AddToObjStr('Rated Inlet Water Temperature','82.2')
    CALL AddToObjStr('Rated Inlet Air Temperature','16.6')
    CALL AddToObjStr('Rated Outlet Water Temperature','71.1')
    CALL AddToObjStr('Rated Outlet Air Temperature','32.2')
    CALL AddToObjStr('Rated Ratio for Air and Water Convection','0.5',.true.)
    !***Branch
    CALL CreateNewObj('Branch')
    CALL AddToObjFld('Name', base + doasNameOff,' Heating Coil HW Branch')
    !save this branch name for later use in plan
    CALL AddToStrList(TRIM(FldVal(base + doasNameOff)) // ' Heating Coil HW Branch',handleHeatingCoilBranch)
    CALL AddToObjStr('Pressure Drop Curve Name','')
    CALL AddToObjStr('Component Object Type','Coil:Heating:Water')
    CALL AddToObjFld('Component Name', base + doasNameOff,' Heating Coil')
    CALL AddToObjFld('Component Inlet Node Name', base + doasNameOff,' Heating Coil HW Inlet')
    CALL AddToObjFld('Component Outlet Node Name', base + doasNameOff,' Heating Coil HW Outlet',.TRUE.)
    !***Controller:WaterCoil
    CALL CreateNewObj('Controller:WaterCoil')
    CALL AddToObjFld('Name', base + doasNameOff,' Heating Coil Controller')
    CALL AddToObjStr('Control Variable','Temperature')
    CALL AddToObjStr('Action','Normal')
    CALL AddToObjStr('Actuator Variable','Flow')
    CALL AddToObjStr('Sensor Node Name', TRIM(heatCoilOutlet))
    CALL AddToObjFld('Actuator Node Name', base + doasNameOff,' Heating Coil HW Inlet')
    CALL AddToObjStr('Controller Convergence Tolerance {deltaC}','autosize')
    CALL AddToObjStr('Maximum Actuated Flow {m3/s}','autosize')
    CALL AddToObjStr('Minimum Actuated Flow {m3/s}','0',.TRUE.)
  END IF
  IF (heatCoilType .EQ. ctElectric) THEN
    !***Coil:Heating:Electric
    CALL CreateNewObj('Coil:Heating:Electric')
    CALL AddToObjFld('Name', base + doasNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + doasHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Efficiency','1')
    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    CALL AddToObjStr('Air Inlet Node Name',  TRIM(heatCoilInlet))
    CALL AddToObjStr('Air Outlet Node Name',  TRIM(heatCoilOutlet))
    CALL AddToObjStr('Coil Temp Setpoint Node', TRIM(heatCoilOutlet),.TRUE.)
  END IF
  IF (heatCoilType .EQ. ctGas) THEN
    !***Coil:Heating:Fuel
    CALL CreateNewObj('Coil:Heating:Fuel')
    CALL AddToObjFld('Name', base + doasNameOff,' Heating Coil')
    CALL AddToObjFld('Availability Schedule Name', base + doasHeatAvailSchedNameOff,' ')
    CALL AddToObjStr('Fuel Type', 'Gas')
    CALL AddToObjFld('Gas Burner Efficiency', base + doasHeatEffiencyOff,' ')
    CALL AddToObjStr('Nominal Capacity of the Coil {W}','autosize')
    CALL AddToObjStr('Air Inlet Node Name',  TRIM(heatCoilInlet))
    CALL AddToObjStr('Air Outlet Node Name',  TRIM(heatCoilOutlet))
    CALL AddToObjStr('Coil Temp Setpoint Node', TRIM(heatCoilOutlet))
    CALL AddToObjFld('Parasitic Electric Load {W}', base + doasHeatParasiticOff,' ')
    CALL AddToObjFld('Part load fraction correlation (function of part load ratio)',   &
       base + doasNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Parasitic Gas Load {W}','0',.TRUE.)
    !***Curve:Cubic
    CALL CreateNewObj('Curve:Cubic')
    CALL AddToObjFld('Name', base + doasNameOff,' Heating Coil PLF-FPLR')
    CALL AddToObjStr('Coefficient1 Constant','0.8')
    CALL AddToObjStr('Coefficient2 x','0.2')
    CALL AddToObjStr('Coefficient3 x**2','0')
    CALL AddToObjStr('Coefficient4 x**3','0')
    CALL AddToObjStr('Minimum Value of x','0')
    CALL AddToObjStr('Maximum Value of x','1',.TRUE.)
  END IF

  ! Heating coil setpoint manager
  IF (heatCoilType .NE. ctNone) THEN
    IF ((heatSetPtCtrlKind .EQ. hspFixed) .OR. (heatSetPtCtrlKind .EQ. hspScheduled)) THEN
      IF ((heatSetPtCtrlKind .EQ. hspFixed) .OR. (isheatStPtSchedBlank)) THEN
        CALL AddAlwaysSchedule(FldVal(base + doasHeatSetPtDesignOff))
      END IF
      !***SetpointManager:Scheduled
      IF (supFanPlacement .EQ. sfpDrawThru) THEN
        CALL WriteComment('NOTE:  Heating supply air setpoint is placed on Supply Path ' // &
                        'Inlet node to allow separate heating and cooling setpoints.')
        CALL WriteComment('SetpointManager:MixedAir <system name> Heating Coil Air ' // &
                        'Temp Manager adjusts this setpoint for fan heat and places it on the Heating Coil Outlet')
      END IF
      CALL CreateNewObj('SetpointManager:Scheduled')
      CALL AddToObjFld('Name', base + doasNameOff,' Heating Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      IF ((heatSetPtCtrlKind .EQ. hspFixed) .OR. (isheatStPtSchedBlank)) THEN
        CALL AddToObjStr('Schedule Name', 'HVACTemplate-Always ' // TRIM(FldVal(base + doasHeatSetPtDesignOff)))
      ELSE
        CALL AddToObjFld('Schedule Name', base + doasHeatSetPtSchedNameOff,'')
      END IF
    ELSE IF (heatSetPtCtrlKind .EQ. hspOAreset) THEN
      !***SetpointManager:OutdoorAirReset
      IF (supFanPlacement .EQ. sfpDrawThru) THEN
        CALL WriteComment('NOTE:  Heating supply air setpoint is placed on Supply Path ' // &
                        'Inlet node to allow separate heating and cooling setpoints.')
        CALL WriteComment('SetpointManager:MixedAir <system name> Heating Coil Air ' // &
                        'Temp Manager adjusts this setpoint for fan heat and places it on the Heating Coil Outlet')
      END IF
      CALL CreateNewObj('SetpointManager:OutdoorAirReset')
      CALL AddToObjFld('Name', base + doasNameOff,' Heating Supply Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('Setpoint at Outdoor Low Temperature {C}', base + doasHeatSetPtAtODBLoOff,'')
      CALL AddToObjFld('Outdoor Low Temperature {C}', base + doasHeatResetODBLoOff,'')
      CALL AddToObjFld('Setpoint at Outdoor High Temperature {C}', base + doasHeatSetPtAtODBHiOff,'')
      CALL AddToObjFld('Outdoor High Temperature {C}', base + doasHeatResetODBHiOff,'')
    END IF
    IF (supFanPlacement .EQ. sfpDrawThru) THEN
      CALL AddToObjFld('Setpoint Node or NodeList Name', base + doasNameOff,' Supply Path Inlet',.TRUE.)
    ELSE
      CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(heatCoilOutlet),.TRUE.)
    END IF
    ! Only needed if system is draw through and there is a heating coil
    IF (supFanPlacement .EQ. sfpDrawThru) THEN
      !***SetpointManager:MixedAir
      CALL CreateNewObj('SetpointManager:MixedAir')
      CALL AddToObjFld('Name', base + doasNameOff,' Heating Coil Air Temp Manager')
      CALL AddToObjStr('Control Variable','Temperature')
      CALL AddToObjFld('Reference Setpoint Node Name', base + doasNameOff,' Supply Path Inlet')
      CALL AddToObjStr('Fan Inlet Node Name', TRIM(fanInlet))
      CALL AddToObjStr('Fan Outlet Node Name', TRIM(fanOutlet))
      CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(heatCoilOutlet),.TRUE.)
    END IF
  END IF
  !***Fan:VariableVolume
  CALL CreateNewObj('Fan:VariableVolume')
  CALL AddToObjFld('Name', base + doasNameOff,' Supply Fan')
  IF (isSysAvailSchedBlank) THEN
    CALL AddToObjStr('Availability Schedule Name','HVACTemplate-Always 1')
  ELSE
    CALL AddToObjFld('Availability Schedule Name', base + doasSysAvailSchedNameOff,' ')
  END IF
  CALL AddToObjFld('Fan Efficiency', base + doasSupplyEfficiencyOff,' ')
  CALL AddToObjFld('Pressure Rise {Pa}', base + doasSupplyPressureOff,' ')
  CALL AddToObjStr('Maximum Flow Rate {m3/s}', 'autosize')
  CALL AddToObjStr('Fan Minimum Flow Rate Input Method For Fan Power', 'Fraction')
  CALL AddToObjStr('Minimum Flow Fraction For Fan Power', '0.0')
  CALL AddToObjStr('Minimum Air Flow Rate For Fan Power {m3/s}', ' ')
  CALL AddToObjFld('Motor Efficiency', base + doasSupplyFanMotorEffOff,' ')
  CALL AddToObjFld('Motor in Airstream Fraction', base + doasSupplyFanFractionOff,' ')
  !use variable speed drive coefficients
  CALL AddToObjStr('Fan Power Coefficient 1','0.0015302446')
  CALL AddToObjStr('Fan Power Coefficient 2','0.0052080574')
  CALL AddToObjStr('Fan Power Coefficient 3','1.1086242')
  CALL AddToObjStr('Fan Power Coefficient 4','-0.11635563')
  CALL AddToObjStr('Fan Power Coefficient 5','0')
  CALL AddToObjStr('Comp Inlet Node Name', TRIM(fanInlet))
  CALL AddToObjStr('Comp Outlet Node Name', TRIM(fanOutlet),.TRUE.)
  IF (isSysAvailSchedBlank) CALL AddAlwaysSchedule('1')
  ! Dehumidification set point manager
  IF (.NOT. isDehumidifyNone) THEN
    ! first check if the schedule name is used or if value should be used.
    IF (FldVal(base + doasDehumSetPtSchedNameOff) .EQ. '') THEN
      scheduleName = 'HVACTemplate-Always ' // TRIM(FldVal(base + doasDehumSetPtOff))
      CALL AddAlwaysSchedule(FldVal(base + doasDehumSetPtOff))
    ELSE
      scheduleName = TRIM(FldVal(base + doasDehumSetPtSchedNameOff))
    ENDIF
    CALL CreateNewObj('SetpointManager:Scheduled')
    CALL AddToObjFld('Name', base + doasNameOff,' Dehumidification Setpoint Manager')
    CALL AddToObjStr('Control Variable','MaximumHumidityRatio')
    CALL AddToObjStr('Schedule Name', TRIM(scheduleName))
    CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(coolCoilUnitOutlet),.TRUE.)
  END IF
  ! Humidifier set point manager and humidifier
  IF (.NOT. isHumidifierNone) THEN
    ! first check if the schedule name is used or if value should be used.
    IF (FldVal(base + doasHumidSetPtSchedNameOff) .EQ. '') THEN
      scheduleName = 'HVACTemplate-Always ' // TRIM(FldVal(base + doasHumidSetPtOff))
      CALL AddAlwaysSchedule(FldVal(base + doasHumidSetPtOff))
    ELSE
      scheduleName = TRIM(FldVal(base + doasHumidSetPtSchedNameOff))
    ENDIF
    CALL CreateNewObj('SetpointManager:Scheduled')
    CALL AddToObjFld('Name', base + doasNameOff,' Humidification Setpoint Manager')
    CALL AddToObjStr('Control Variable','MinimumHumidityRatio')
    CALL AddToObjStr('Schedule Name', TRIM(scheduleName))
    CALL AddToObjStr('Setpoint Node or NodeList Name', TRIM(humidifierOutlet),.TRUE.)
    IF (humidifierKind .EQ. humidifyElecSteam) THEN
      !***Humidifier:Steam:Electric
      CALL CreateNewObj('Humidifier:Steam:Electric')
      CALL AddToObjFld('Name', base + doasNameOff,' Humidifier')
      CALL AddToObjFld('Availability Schedule Name', base + doasHumidAvailSchedNameOff,' ')
      CALL AddToObjFld('Rated Capacity {m3/s}', base + doasHumidCapOff,'')
      CALL AddToObjFld('Rated Power {W}', base + doasHumidRatedElecOff,'')
      CALL AddToObjStr('rated fan power {W}','0')
      CALL AddToObjStr('Standby power {W}','0')
      CALL AddToObjStr('Air Inlet Node Name', TRIM(humidifierInlet))
      CALL AddToObjStr('Air Outlet Node Name', TRIM(humidifierOutlet),.TRUE.)
    END IF
  END IF
  !***OutdoorAir:NodeList
  CALL CreateNewObj('OutdoorAir:NodeList')
  CALL AddToObjFld('Node or NodeList Name 1', base + doasNameOff,' Outdoor Air Inlet',.TRUE.)
  !***AirLoopHVAC:OutdoorAirSystem
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem')
  CALL AddToObjFld('Name', base + doasNameOff,' OA System')
  CALL AddToObjFld('Controller List Name', base + doasNameOff,' OA System Controllers')
  CALL AddToObjFld('Outdoor Air Equipment List Name', base + doasNameOff,' OA System Equipment')
  CALL AddToObjFld('Availability Manager List Name', base + doasNameOff,' Availability Managers',.TRUE.)
  !***AirLoopHVAC:ControllerList
  CALL CreateNewObj('AirLoopHVAC:ControllerList')
  CALL AddToObjFld('Name', base + doasNameOff,' OA System Controllers')
  CALL AddToObjStr('Controller Object Type','Controller:OutdoorAir')
  CALL AddToObjFld('Controller Name', base + doasNameOff,' OA Controller',.TRUE.)
  !***AirLoopHVAC:OutdoorAirSystem:EquipmentList
  CALL CreateNewObj('AirLoopHVAC:OutdoorAirSystem:EquipmentList')
  CALL AddToObjFld('Name', base + doasNameOff,' OA System Equipment')
  IF (heatRecovery .EQ. htrecSens) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + doasNameOff,' Heat Recovery')
  ELSEIF (heatRecovery .EQ. htrecEnth) THEN
    CALL AddToObjStr('Component Object Type','HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Component Name', base + doasNameOff,' Heat Recovery')
  END IF
  CALL AddToObjStr('Component Object Type','OutdoorAir:Mixer')
  CALL AddToObjFld('Component Name', base + doasNameOff,' OA Mixing Box',.TRUE.)
  !***OutdoorAir:Mixer
  CALL CreateNewObj('OutdoorAir:Mixer')
  CALL AddToObjFld('Name', base + doasNameOff,' OA Mixing Box')
  CALL AddToObjFld('Mixed Air Node Name', base + doasNameOff,' Mixed Air Outlet')
  IF (heatRecovery .EQ. htrecNone) THEN
    CALL AddToObjFld('Outdoor Air Stream Node Name', base + doasNameOff,' Outdoor Air Inlet')
  END IF
  IF (heatRecovery .NE. htrecNone) THEN
    CALL AddToObjFld('Outdoor Air Stream Node Name', base + doasNameOff,' Heat Recovery Supply Outlet')
  ENDIF
  CALL AddToObjFld('Relief Air Stream Node Name', base + doasNameOff,' Relief Air Outlet')
  CALL AddToObjFld('Return Air Stream Node Name', base + doasNameOff,' Air Loop Inlet',.TRUE.)
  !***Controller:OutdoorAir
  CALL CreateNewObj('Controller:OutdoorAir')
  CALL AddToObjFld('Name', base + doasNameOff,' OA Controller')
  CALL AddToObjFld('Relief Air Outlet Node Name', base + doasNameOff,' Relief Air Outlet')
  CALL AddToObjFld('Return Air Node Name', base + doasNameOff,' Air Loop Inlet')
  CALL AddToObjFld('Mixed Air Node Name', base + doasNameOff,' Mixed Air Outlet')
  CALL AddToObjFld('Actuator Node Name', base + doasNameOff,' Outdoor Air Inlet')
  ! the minimum and maximum are the same volume of air
  CALL AddToObjFld('Minimum Outdoor Air Flow Rate {m3/s}', base + doasSupplyRateOff,' ')
  CALL AddToObjFld('Maximum Outdoor Air Flow Rate {m3/s}', base + doasSupplyRateOff,' ')
  IF (heatRecovery .EQ. htrecNone) THEN
  ! If heat recovery is selected, then set economizer control to bypass
  ! otherwise, set to modulate OA flow
    CALL AddToObjStr('Economizer Control Type', 'NoEconomizer')
    CALL AddtoObjStr('Economizer Control Action Type','ModulateFlow')
    CALL AddToObjStr('Economizer Maximum Limit Dry-Bulb Temperature {C}',' ')
    CALL AddToObjStr('Economizer Maximum Limit Enthalpy {J/kg}',' ')
    CALL AddToObjStr('Economizer Maximum Limit Dewpoint Temperature (C)',' ')
    CALL AddToObjStr('Electronic Enthalpy Limit Curve Name',' ')
    CALL AddToObjStr('Economizer Minimum Limit Dry-Bulb Temperature {C}',' ')
  ELSEIF (heatRecovery .EQ. htrecSens) THEN
    CALL AddToObjStr('Economizer Control Type', 'DifferentialDryBulb')
    CALL AddtoObjStr('Economizer Control Action Type','MinimumFlowWithBypass')
    CALL AddToObjStr('Economizer Maximum Limit Dry-Bulb Temperature {C}',' ')
    CALL AddToObjStr('Economizer Maximum Limit Enthalpy {J/kg}',' ')
    CALL AddToObjStr('Economizer Maximum Limit Dewpoint Temperature (C)',' ')
    CALL AddToObjStr('Electronic Enthalpy Limit Curve Name',' ')
    CALL AddToObjFld('Economizer Minimum Limit Dry-Bulb Temperature {C}', base + doasHeatSetPtDesignOff,'')
  ELSEIF (heatRecovery .EQ. htrecEnth) THEN
    CALL AddToObjStr('Economizer Control Type', 'DifferentialEnthalpy')
    CALL AddtoObjStr('Economizer Control Action Type','MinimumFlowWithBypass')
    CALL AddToObjStr('Economizer Maximum Limit Dry-Bulb Temperature {C}',' ')
    CALL AddToObjStr('Economizer Maximum Limit Enthalpy {J/kg}',' ')
    CALL AddToObjStr('Economizer Maximum Limit Dewpoint Temperature (C)',' ')
    CALL AddToObjStr('Electronic Enthalpy Limit Curve Name',' ')
    CALL AddToObjFld('Economizer Minimum Limit Dry-Bulb Temperature {C}', base + doasHeatSetPtDesignOff,'')
  ENDIF
  CALL AddToObjStr('Lockout Type', 'NoLockout')
  CALL AddToObjStr('Minimum Limit Type','ProportionalMinimum')
  CALL AddToObjStr('Minimum Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Minimum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Maximum Fraction of Outdoor Air Schedule Name',' ')
  CALL AddToObjStr('Mechanical Ventilation Controller Name','',.TRUE.)
  IF (heatRecovery .EQ. htrecSens) THEN
    htRecSens75 = StringToReal(FldVal(base + doasHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    !***HeatExchanger:AirToAir:SensibleAndLatent
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + doasNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Nominal supply air flow rate {m3/s}', base + doasSupplyRateOff,'')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + doasHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Heating Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Heating Air Flow','0')
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + doasHeatRecSenEffOff,'')
    CALL AddToObjStr('Latent Effectiveness at 100% Cooling Air Flow','0')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow', htRecSens75)
    CALL AddToObjStr('Latent Effectiveness at 75% Cooling Air Flow','0')
    CALL AddToObjFld('Supply Air Inlet Node Name', base + doasNameOff,' Outdoor Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + doasNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + doasNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + doasNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','250')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjFld('Heat Exchanger Type', base + doasHXTypeOff,'')
    CALL AddToObjFld('Frost Control Type', base + doasFrostControlOff,'')
    CALL AddToObjStr('Threshold Temperature','1.7')
    CALL AddToObjStr('Initial Defrost Time Fraction','0.083')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','0.012')
    CALL AddToObjStr('Economizer Lockout','Yes',.TRUE.)
  ELSEIF (heatRecovery .EQ. htrecEnth) THEN
    htRecSens75 = StringToReal(FldVal(base + doasHeatRecSenEffOff)) + 0.05
    IF (htRecSens75 .GT. 1.) htRecSens75 = 1.
    htRecLat75 = StringToReal(FldVal(base + doasHeatRecLatEffOff)) + 0.05
    IF (htRecLat75 .GT. 1.) htRecLat75 = 1.
    !***HeatExchanger:AirToAir:SensibleAndLatent
    CALL CreateNewObj('HeatExchanger:AirToAir:SensibleAndLatent')
    CALL AddToObjFld('Heat exchanger name', base + doasNameOff,' Heat Recovery')
    CALL AddToObjStr('Availability Schedule Name',' ')
    CALL AddToObjFld('Nominal supply air flow rate {m3/s}', base + doasSupplyRateOff,'')
    CALL AddToObjFld('Sensible Effectiveness at 100% Heating Air Flow', base + doasHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Heating Air Flow', base + doasHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Heating Air Flow', htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Heating Air Flow', htRecLat75)
    CALL AddToObjFld('Sensible Effectiveness at 100% Cooling Air Flow', base + doasHeatRecSenEffOff,'')
    CALL AddToObjFld('Latent Effectiveness at 100% Cooling Air Flow', base + doasHeatRecLatEffOff,'')
    CALL AddToObjNum('Sensible Effectiveness at 75% Cooling Air Flow', htRecSens75)
    CALL AddToObjNum('Latent Effectiveness at 75% Cooling Air Flow', htRecLat75)
    CALL AddToObjFld('Supply Air Inlet Node Name', base + doasNameOff,' Outdoor Air Inlet')
    CALL AddToObjFld('Supply Air Outlet Node Name', base + doasNameOff,' Heat Recovery Supply Outlet')
    CALL AddToObjFld('Exhaust Air Inlet Node Name', base + doasNameOff,' Relief Air Outlet')
    CALL AddToObjFld('Exhaust Air Outlet Node Name', base + doasNameOff,' Heat Recovery Relief Outlet')
    CALL AddToObjStr('Nominal electric power {W}','0')
    CALL AddToObjStr('Supply Air Outlet Temperature Control','Yes')
    CALL AddToObjFld('Heat Exchanger Type', base + doasHXTypeOff,'')
    CALL AddToObjFld('Frost Control Type', base + doasFrostControlOff,'')
    CALL AddToObjStr('Threshold Temperature','1.7')
    CALL AddToObjStr('Initial Defrost Time Fraction','0.083')
    CALL AddToObjStr('Rate of Defrost Time Fraction Increase','0.012')
    CALL AddToObjStr('Economizer Lockout','Yes',.TRUE.)
  END IF
  IF (heatRecovery .NE. htrecNone) THEN
    !***SetpointManager:MixedAir for Heat Recovery Outlet - Reference heating coil setpoint
    IF (heatCoilType .EQ. ctNone) THEN
      CALL WriteError('Warning:  In HVACTemplate:System:DedicatedOutdoorAir "'//TRIM(FldVal(base + doasNameOff))//'"'// &
                        ' there is heat recovery with no heating coil. The heat recovery heating mode will be'// &
                        ' controlled to not exceed the cooling setpoint.',msgWarning)
    END IF
    CALL CreateNewObj('SetpointManager:MixedAir')
    CALL AddToObjFld('Name', base + doasNameOff,' Heat Recovery Air Temp Manager')
    CALL AddToObjStr('Control Variable','Temperature')
    IF (heatCoilType .NE. ctNone) THEN
     IF (supFanPlacement .EQ. sfpDrawThru) THEN
        CALL AddToObjFld('Reference Setpoint Node Name', base + doasNameOff,' Supply Path Inlet')
      ELSE
        CALL AddToObjStr('Reference Setpoint Node Name', TRIM(heatCoilOutlet))
      END IF
    ELSE
    ! With no heating coil, there is no heating setpoint, so use the cooling setpoint instead
     IF (supFanPlacement .EQ. sfpDrawThru) THEN
        CALL AddToObjStr('Reference Setpoint Node Name', TRIM(airloopOutlet))
      ELSE
        CALL AddToObjStr('Reference Setpoint Node Name', TRIM(coolCoilUnitOutlet))
      END IF
    END IF
    CALL AddToObjStr('Fan Inlet Node Name', TRIM(fanInlet))
    CALL AddToObjStr('Fan Outlet Node Name', TRIM(fanOutlet))
    CALL AddToObjFld('Setpoint Node or NodeList Name', base + doasNameOff,' Heat Recovery Supply Outlet',.TRUE.)
    !***SetpointManager:MixedAir for Economizer heat recovery bypass control - Reference cooling coil setpoint
    CALL CreateNewObj('SetpointManager:MixedAir')
    CALL AddToObjFld('Name', base + doasNameOff,' Heat Recovery Economizer Temp Manager')
    CALL AddToObjStr('Control Variable','Temperature')
    CALL AddToObjStr('Reference Setpoint Node Name', TRIM(airloopOutlet))
    CALL AddToObjStr('Fan Inlet Node Name', TRIM(fanInlet))
    CALL AddToObjStr('Fan Outlet Node Name', TRIM(fanOutlet))
    CALL AddToObjFld('Setpoint Node or NodeList Name',  base + doasNameOff,' Mixed Air Outlet',.TRUE.)
  END IF
END DO
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE writeErrorObjects
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   October 2005
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Writes the errors found during the run to the IDF file

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
CHARACTER(len=MaxErrorLength) :: curErrMsg
INTEGER :: iMsgs
INTEGER :: lenMsg
INTEGER :: charPerLine
INTEGER :: startOfLine
INTEGER :: endOfLine

charPerLine = 75
DO iMsgs = 1, numErrMsgs
  curErrMsg = ErrMsgs(iMsgs)%msgText
  CALL CreateNewObj('Output:PreprocessorMessage')
  CALL AddToObjStr('Preprocessor Name','ExpandObjects')
  IF (ErrMsgs(iMsgs)%msgKind .EQ. msgWarning) THEN
    CALL AddToObjStr('Error Severity','Warning')
  ELSE
    CALL AddToObjStr('Error Severity','Fatal')
  END IF
  lenMsg = LEN_TRIM(curErrMsg)
  startOfLine = 1
  endOfLine = startOfLine + charPerLine
  DO
    DO WHILE (curErrMsg(endOfLine:endOfLine) .NE. ' ')
      endOfLine = endOfLine - 1
      IF (endOfLine .EQ. startOfLine) THEN
        endOfLine = startOfLine + charPerLine
        EXIT
      END IF
    END DO
    IF (endOfLine .LT. lenMsg) THEN
      CALL AddToObjStr('message line',curErrMsg(startOfLine:endOfLine))
    ELSE
      CALL AddToObjStr('message line',curErrMsg(startOfLine:endOfLine),.TRUE.)
    END IF
    startOfLine = endOfLine + 1
    endOfLine = startOfLine + charPerLine
    IF (startOfLine .GT. lenMsg) EXIT
  END DO
END DO
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE WriteBasement
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2010
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Writes the entire set of basement objects without modification.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: ios = 0
INTEGER :: fh = 0

IF (ghtControlBase .NE. 0) THEN
  IF (SameString(FldVal(ghtControlBase + ghtCtrlBasementOff), 'Yes')) THEN
    OPEN (unit=fh, file='BasementGHTIn.idf', action='WRITE',iostat=ios)
    IF (ios .EQ. 0) THEN
      CALL WriteGroupOfOrigObj('SimParameters',fh, ghtBsSimParamBase, numGhtBsSimParam, ghtBsSimParamLastOff)
      CALL WriteGroupOfOrigObj('MatlProps',fh, ghtBaseBsMatlProps, numGhtBsMatlProps, ghtBsMatlPropsLastOff)
      CALL WriteGroupOfOrigObj('Insulation',fh, ghtBsInsulBase, numGhtBsInsul, ghtBsInsulLastOff)
      CALL WriteGroupOfOrigObj('SurfaceProps',fh, ghtBsSurfPropsBase, numGhtBsSurfProps, ghtBsSurfPropsLastOff)
      CALL WriteGroupOfOrigObj('BldgData',fh, ghtBsBldgDataBase, numGhtBsBldgData, ghtBsBldgDataLastOff)
      CALL WriteGroupOfOrigObj('Interior',fh, ghtBsInteriorBase, numGhtBsInterior, ghtBsInteriorLastOff)
      CALL WriteGroupOfOrigObj('ComBldg',fh, ghtBsComBldgBase, numGhtBsComBldg, ghtBsComBldgLastOff)
      CALL WriteGroupOfOrigObj('EquivSlab',fh, ghtBsEqvSlbBase, numGhtBsEqvSlb, ghtBsEqvSlbLastOff)
      CALL WriteGroupOfOrigObj('EquivAutoGrid',fh, ghtBsEqvAutoGridBase, numGhtBsEqvAutoGrid, ghtBsEqvAutoGridLastOff)
      CALL WriteGroupOfOrigObj('AutoGrid',fh, ghtBsAutoGridBase, numGhtBsAutoGrid, ghtBsAutoGridLastOff)
      CALL WriteGroupOfOrigObj('ManualGrid',fh, ghtBsManGridBase, numGhtBsManGrid, ghtBsManGridLastOff)
      CALL WriteGroupOfOrigObj('XFACE',fh, ghtBsXFaceBase, numGhtBsXFace, ghtBsXFaceLastOff)
      CALL WriteGroupOfOrigObj('YFACE',fh, ghtBsYFaceBase, numGhtBsYFace, ghtBsYFaceLastOff)
      CALL WriteGroupOfOrigObj('ZFACE',fh, ghtBsZFaceBase, numGhtBsZFace, ghtBsZFaceLastOff)
      CLOSE(fh)
    ELSE
      CALL AddToErrMsg('error opening BasementGHTIn.idf file')
    END IF
    ! check if the Site:GroundTemperature:BuildingSurface already exist in the file and if they
    ! do add a warning message
    IF (gndTempBuildSurfBase .NE. 0) THEN
      CALL AddToErrMsg('The Site:GroundTemperature:BuildingSurface object present in the ' // &
        'file will not be used. Intead the values from Slab will be used.',msgWarning)
    END IF
  END IF
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE WriteSlab
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2010
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Writes the entire set of slab objects without modification.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
INTEGER :: ios = 0
INTEGER :: fh = 0

IF (ghtControlBase .NE. 0) THEN
  IF (SameString(FldVal(ghtControlBase + ghtCtrlSlabOff), 'Yes')) THEN
    OPEN (unit=fh, file='GHTIn.idf', action='WRITE',iostat=ios)
    IF (ios .EQ. 0) THEN
      CALL WriteGroupOfOrigObj('Materials',fh, ghtSlbMaterialBase, numGhtSlbMaterial, ghtSlbMaterialLastOff)
      CALL WriteGroupOfOrigObj('MatlProps',fh, ghtSlbMatlPropsBase, numGhtSlbMatlProps, ghtSlbMatlPropsLastOff)
      CALL WriteGroupOfOrigObj('BoundConds',fh, ghtSlbBoundCondsBase, numGhtSlbBoundConds, ghtSlbBoundCondsLastOff)
      CALL WriteGroupOfOrigObj('BldgProps',fh, ghtSlbBldgPropsBase, numGhtSlbBldgProps, ghtSlbBldgPropsLastOff)
      CALL WriteGroupOfOrigObj('Insulation',fh, ghtSlbInsulBase, numGhtSlbInsul, ghtSlbInsulLastOff)
      CALL WriteGroupOfOrigObj('EquivalentSlab',fh, ghtSlbEqvSlbBase, numGhtSlbEqvSlb, ghtSlbEqvSlbLastOff)
      CALL WriteGroupOfOrigObj('AutoGrid',fh, ghtSlbAutoGridBase, numGhtSlbAutoGrid, ghtSlbAutoGridLastOff)
      CALL WriteGroupOfOrigObj('ManualGrid',fh, ghtSlbManGridBase, numGhtSlbManGrid, ghtSlbManGridLastOff)
      CALL WriteGroupOfOrigObj('XFACE',fh, ghtSlbXFaceBase, numGhtSlbXFace, ghtSlbXFaceLastOff)
      CALL WriteGroupOfOrigObj('YFACE',fh, ghtSlbYFaceBase, numGhtSlbYFace, ghtSlbYFaceLastOff)
      CALL WriteGroupOfOrigObj('ZFACE',fh, ghtSlbZFaceBase, numGhtSlbZFace, ghtSlbZFaceLastOff)
      CLOSE(fh)
    ELSE
      CALL AddToErrMsg('error opening slab.int file')
    END IF
  END IF
END IF
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE CreateGroundTemperatures
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   February 2010
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Eacho the ground temperature object if the slab and basement proprocessors
          !    are not going to be used.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
LOGICAL :: doWriteTemperatures
INTEGER :: base

doWriteTemperatures = .TRUE.
IF (ghtControlBase .NE. 0) THEN
  IF (SameString(FldVal(ghtControlBase + ghtCtrlSlabOff), 'Yes')) doWriteTemperatures = .FALSE.
  IF (SameString(FldVal(ghtControlBase + ghtCtrlBasementOff), 'Yes')) doWriteTemperatures = .FALSE.
END IF
IF (gndTempBuildSurfBase .EQ. 0) doWriteTemperatures = .FALSE.
IF (doWriteTemperatures) THEN
  base =  gndTempBuildSurfBase
  CALL CreateNewObj('Site:GroundTemperature:BuildingSurface')
  CALL AddToObjFld('January Ground Temperature', base + gndTempSurfJanOff,'')
  CALL AddToObjFld('February Ground Temperature', base + gndTempSurfFebOff,'')
  CALL AddToObjFld('March Ground Temperature', base + gndTempSurfMarOff,'')
  CALL AddToObjFld('April Ground Temperature', base + gndTempSurfAprOff,'')
  CALL AddToObjFld('May Ground Temperature', base + gndTempSurfMayOff,'')
  CALL AddToObjFld('June Ground Temperature', base + gndTempSurfJunOff,'')
  CALL AddToObjFld('July Ground Temperature', base + gndTempSurfJulOff,'')
  CALL AddToObjFld('August Ground Temperature', base + gndTempSurfAugOff,'')
  CALL AddToObjFld('September Ground Temperature', base + gndTempSurfSepOff,'')
  CALL AddToObjFld('October Ground Temperature', base + gndTempSurfOctOff,'')
  CALL AddToObjFld('November Ground Temperature', base + gndTempSurfNovOff,'')
  CALL AddToObjFld('December Ground Temperature', base + gndTempSurfDecOff,'',.TRUE.)
END IF
END SUBROUTINE


!----------------------------------------------------------------------------------
SUBROUTINE CreateDetailedBuildingSurface
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   January 2010
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Inserts the appropriate other side coefficients into the
          !    detailed building surfaces that correspond wiht the results
          !    of the slab and basement preprocessors.

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
          !    na

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na

INTEGER :: base  ! used as base for
INTEGER :: lastNonBlank
INTEGER :: iSurface
INTEGER :: jVertex
CHARACTER(len=MaxAlphaLength) :: outBndCond
LOGICAL :: isSlabProgRun = .FALSE.
LOGICAL :: isBasementProgRun = .FALSE.

! check if slab or basement is preprocessors have been asked to run.
IF (ghtControlBase .NE. 0) THEN
  IF (SameString(FldVal(ghtControlBase + ghtCtrlSlabOff),'YES')) THEN
    isSlabProgRun = .TRUE.
  END IF
  IF (SameString(FldVal(ghtControlBase + ghtCtrlBasementOff),'YES')) THEN
    isBasementProgRun = .TRUE.
  END IF
END IF

!BuildingSurface:Detailed
DO iSurface = 1, numSurfBldgDetl
  base =  surfBldgDetlBase(iSurface)
  ! write object
  CALL CreateNewObj('BuildingSurface:Detailed')
  CALL AddToObjFld('Name', base + buildSurfNameOff,'')
  CALL AddToObjFld('Surface Type', base + buildSurfTypeOff,'')
  CALL AddToObjFld('Construction Name', base + buildSurfConstructionOff,'')
  CALL AddToObjFld('Zone Name', base + buildSurfZoneOff,'')
  outBndCond = FldVal(base + buildSurfOutBndCondOff)
  IF (SameString(outBndCond,'GroundSlabPreprocessorAverage')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefSlabAverage')
    IF (.NOT. isSlabProgRun) CALL makeSlabErr('BuildingSurface:Detailed',base + buildSurfNameOff)
  ELSEIF (SameString(outBndCond,'GroundSlabPreprocessorCore')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefSlabCore')
    IF (.NOT. isSlabProgRun) CALL makeSlabErr('BuildingSurface:Detailed',base + buildSurfNameOff)
  ELSEIF (SameString(outBndCond,'GroundSlabPreprocessorPerimeter')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefSlabPerimeter')
    IF (.NOT. isSlabProgRun) CALL makeSlabErr('BuildingSurface:Detailed',base + buildSurfNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorAverageWall')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementAvgWall')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('BuildingSurface:Detailed',base + buildSurfNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorAverageFloor')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementAvgFloor')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('BuildingSurface:Detailed',base + buildSurfNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorUpperWall')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementUpperWall')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('BuildingSurface:Detailed',base + buildSurfNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorLowerWall')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementLowerWall')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('BuildingSurface:Detailed',base + buildSurfNameOff)
  ELSE
    CALL AddToObjFld('Outside Boundary Condition', base + buildSurfOutBndCondOff,'')
    CALL AddToObjFld('Outside Boundary Condition Object', base + buildSurfOutBndCondNameOff,'')
  END IF
  CALL AddToObjFld('Sun Exposure', base + buildSurfSunExpOff,'')
  CALL AddToObjFld('Wind Exposure', base + buildSurfWindExpOff,'')
  CALL AddToObjFld('View Factor to Ground', base + buildSurfViewGroundOff,'')
  CALL AddToObjFld('Number of Vertices', base + buildSurfNumVertOff,'')
  !search for last non-blank field
  lastNonBlank = buildSurfLastZVertexOff
  DO jVertex = buildSurfLastZVertexOff,buildSurfX1Off,-1
    IF ((FldVal(base + jVertex)) .NE. '') THEN
      lastNonBlank = jVertex
      EXIT
    END IF
  END DO
  DO jVertex = buildSurfX1off, lastNonBlank
    ! the following is complex because it is building up the the Vertex 1 X-coordinate string
    ! based on the index of the loop
    CALL AddToObjFld('Vertex '// &
      TRIM(IntToStr(1 + INT((jVertex - buildSurfX1off)/3))) // ' ' // &
      CHAR(88 + MOD(jVertex - buildSurfX1off,3)) // '-coordinate', &
      base + jVertex,'', jVertex .EQ. lastNonBlank)
  END DO
END DO
!Wall:Detailed
DO iSurface = 1, numSurfWallDetl
  base =  surfWallDetlBase(iSurface)
  ! write object
  CALL CreateNewObj('Wall:Detailed')
  CALL AddToObjFld('Name', base + wallDetlNameOff,'')
  CALL AddToObjFld('Construction Name', base + wallDetlConstructionOff,'')
  CALL AddToObjFld('Zone Name', base + wallDetlZoneOff,'')
  outBndCond = FldVal(base + wallDetlOutBndCondOff)
  IF (SameString(outBndCond,'GroundSlabPreprocessorAverage')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefSlabAverage')
    IF (.NOT. isSlabProgRun) CALL makeSlabErr('Wall:Detailed',base + wallDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundSlabPreprocessorCore')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefSlabCore')
    IF (.NOT. isSlabProgRun) CALL makeSlabErr('Wall:Detailed',base + wallDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundSlabPreprocessorPerimeter')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefSlabPerimeter')
    IF (.NOT. isSlabProgRun) CALL makeSlabErr('Wall:Detailed',base + wallDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorAverageWall')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementAvgWall')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('Wall:Detailed',base + wallDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorAverageFloor')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementAvgFloor')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('Wall:Detailed',base + wallDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorUpperWall')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementUpperWall')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('Wall:Detailed',base + wallDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorLowerWall')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementLowerWall')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('Wall:Detailed',base + wallDetlNameOff)
  ELSE
    CALL AddToObjFld('Outside Boundary Condition', base + wallDetlOutBndCondOff,'')
    CALL AddToObjFld('Outside Boundary Condition Object', base + wallDetlOutBndCondNameOff,'')
  END IF
  CALL AddToObjFld('Sun Exposure', base + wallDetlSunExpOff,'')
  CALL AddToObjFld('Wind Exposure', base + wallDetlWindExpOff,'')
  CALL AddToObjFld('View Factor to Ground', base + wallDetlViewGroundOff,'')
  CALL AddToObjFld('Number of Vertices', base + wallDetlNumVertOff,'')
  !search for last non-blank field
  lastNonBlank = wallDetlLastZVertexOff
  DO jVertex = wallDetlLastZVertexOff,wallDetlX1Off,-1
    IF ((FldVal(base + jVertex)) .NE. '') THEN
      lastNonBlank = jVertex
      EXIT
    END IF
  END DO
  DO jVertex = wallDetlX1Off, lastNonBlank
    ! the following is complex because it is building up the the Vertex 1 X-coordinate string
    ! based on the index of the loop
    CALL AddToObjFld('Vertex '// &
      TRIM(IntToStr(1 + INT((jVertex - wallDetlX1Off)/3))) // ' ' // &
      CHAR(88 + MOD(jVertex - wallDetlX1Off,3)) // '-coordinate', &
      base + jVertex,'', jVertex .EQ. lastNonBlank)
  END DO
END DO
!RoofCeiling:Detailed
DO iSurface = 1, numSurfRfClDetl
  base =  surfRfClDetlBase(iSurface)
  ! write object
  CALL CreateNewObj('RoofCeiling:Detailed')
  CALL AddToObjFld('Name', base + rfClDetlNameOff,'')
  CALL AddToObjFld('Construction Name', base + rfClDetlConstructionOff,'')
  CALL AddToObjFld('Zone Name', base + rfClDetlZoneOff,'')
  outBndCond = FldVal(base + rfClDetlOutBndCondOff)
  IF (SameString(outBndCond,'GroundSlabPreprocessorAverage')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefSlabAverage')
    IF (.NOT. isSlabProgRun) CALL makeSlabErr('RoofCeiling:Detailed',base + rfClDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundSlabPreprocessorCore')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefSlabCore')
    IF (.NOT. isSlabProgRun) CALL makeSlabErr('RoofCeiling:Detailed',base + rfClDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundSlabPreprocessorPerimeter')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefSlabPerimeter')
    IF (.NOT. isSlabProgRun) CALL makeSlabErr('RoofCeiling:Detailed',base + rfClDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorAverageWall')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementAvgWall')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('RoofCeiling:Detailed',base + rfClDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorAverageFloor')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementAvgFloor')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('RoofCeiling:Detailed',base + rfClDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorUpperWall')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementUpperWall')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('RoofCeiling:Detailed',base + rfClDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorLowerWall')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementLowerWall')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('RoofCeiling:Detailed',base + rfClDetlNameOff)
  ELSE
    CALL AddToObjFld('Outside Boundary Condition', base + rfClDetlOutBndCondOff,'')
    CALL AddToObjFld('Outside Boundary Condition Object', base + rfClDetlOutBndCondNameOff,'')
  END IF
  CALL AddToObjFld('Sun Exposure', base + rfClDetlSunExpOff,'')
  CALL AddToObjFld('Wind Exposure', base + rfClDetlWindExpOff,'')
  CALL AddToObjFld('View Factor to Ground', base + rfClDetlViewGroundOff,'')
  CALL AddToObjFld('Number of Vertices', base + rfClDetlNumVertOff,'')
  !search for last non-blank field
  lastNonBlank = rfClDetlLastZVertexOff
  DO jVertex = rfClDetlLastZVertexOff,rfClDetlX1Off,-1
    IF ((FldVal(base + jVertex)) .NE. '') THEN
      lastNonBlank = jVertex
      EXIT
    END IF
  END DO
  DO jVertex = rfClDetlX1Off, lastNonBlank
    ! the following is complex because it is building up the the Vertex 1 X-coordinate string
    ! based on the index of the loop
    CALL AddToObjFld('Vertex '// &
      TRIM(IntToStr(1 + INT((jVertex - rfClDetlX1Off)/3))) // ' ' // &
      CHAR(88 + MOD(jVertex - rfClDetlX1Off,3)) // '-coordinate', &
      base + jVertex,'', jVertex .EQ. lastNonBlank)
  END DO
END DO
!Floor:Detailed
DO iSurface = 1, numSurfFlrDetl
  base =  surfFlrDetlBase(iSurface)
  ! write object
  CALL CreateNewObj('Floor:Detailed')
  CALL AddToObjFld('Name', base + flrDetlNameOff,'')
  CALL AddToObjFld('Construction Name', base + flrDetlConstructionOff,'')
  CALL AddToObjFld('Zone Name', base + flrDetlZoneOff,'')
  outBndCond = FldVal(base + flrDetlOutBndCondOff)
  IF (SameString(outBndCond,'GroundSlabPreprocessorAverage')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefSlabAverage')
    IF (.NOT. isSlabProgRun) CALL makeSlabErr('Floor:Detailed',base + flrDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundSlabPreprocessorCore')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefSlabCore')
    IF (.NOT. isSlabProgRun) CALL makeSlabErr('Floor:Detailed',base + flrDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundSlabPreprocessorPerimeter')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefSlabPerimeter')
    IF (.NOT. isSlabProgRun) CALL makeSlabErr('Floor:Detailed',base + flrDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorAverageWall')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementAvgWall')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('Floor:Detailed',base + flrDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorAverageFloor')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementAvgFloor')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('Floor:Detailed',base + flrDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorUpperWall')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementUpperWall')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('Floor:Detailed',base + flrDetlNameOff)
  ELSEIF (SameString(outBndCond,'GroundBasementPreprocessorLowerWall')) THEN
    CALL AddToObjStr('Outside Boundary Condition', 'OtherSideCoefficients')
    CALL AddToObjStr('Outside Boundary Condition Object', 'surfPropOthSdCoefBasementLowerWall')
    IF (.NOT. isBasementProgRun) CALL makeBasementErr('Floor:Detailed',base + flrDetlNameOff)
  ELSE
    CALL AddToObjFld('Outside Boundary Condition', base + flrDetlOutBndCondOff,'')
    CALL AddToObjFld('Outside Boundary Condition Object', base + flrDetlOutBndCondNameOff,'')
  END IF
  CALL AddToObjFld('Sun Exposure', base + flrDetlSunExpOff,'')
  CALL AddToObjFld('Wind Exposure', base + flrDetlWindExpOff,'')
  CALL AddToObjFld('View Factor to Ground', base + flrDetlViewGroundOff,'')
  CALL AddToObjFld('Number of Vertices', base + flrDetlNumVertOff,'')
  !search for last non-blank field
  lastNonBlank = flrDetlLastZVertexOff
  DO jVertex = flrDetlLastZVertexOff,flrDetlX1Off,-1
    IF ((FldVal(base + jVertex)) .NE. '') THEN
      lastNonBlank = jVertex
      EXIT
    END IF
  END DO
  DO jVertex = flrDetlX1Off, lastNonBlank
    ! the following is complex because it is building up the the Vertex 1 X-coordinate string
    ! based on the index of the loop
    CALL AddToObjFld('Vertex '// &
      TRIM(IntToStr(1 + INT((jVertex - flrDetlX1Off)/3))) // ' ' // &
      CHAR(88 + MOD(jVertex - flrDetlX1Off,3)) // '-coordinate', &
      base + jVertex,'', jVertex .EQ. lastNonBlank)
  END DO
END DO
END SUBROUTINE

!----------------------------------------------------------------------------------
SUBROUTINE makeSlabErr(surfObjKind,surfNameRef)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   March 2012
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Create an error message when slab program has not been run
          !    but ground conditions are referenced .

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN)  :: surfObjKind
INTEGER,INTENT(IN)            :: surfNameRef

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
CALL WriteError('Ground conditions are referenced in a ' // TRIM(surfObjKind) // &
     ' named "' // TRIM(FldVal(surfNameRef)) // '" but the Slab preprocessor was not run. ' // &
     ' Check the GroundHeatTransfer:Control object and see if the Slab preprocessor was set to Yes.')
END SUBROUTINE makeSlabErr

SUBROUTINE makeBasementErr(surfObjKind,surfNameRef)
          ! SUBROUTINE INFORMATION:
          !    AUTHOR         Jason Glazer of GARD Analytics, Inc.
          !    DATE WRITTEN   March 2012
          !    MODIFIED       na
          !    RE-ENGINEERED  na

          ! PURPOSE OF THIS SUBROUTINE:
          !    Create an error message when basement program has not been run
          !    but ground conditions are referenced .

          ! METHODOLOGY EMPLOYED:

          ! REFERENCES:
          !    na

          ! USE STATEMENTS:

IMPLICIT NONE    ! Enforce explicit typing of all variables in this routine


          ! SUBROUTINE ARGUMENT DEFINITIONS:
CHARACTER(len=*), INTENT(IN)  :: surfObjKind
INTEGER,INTENT(IN)            :: surfNameRef

          ! SUBROUTINE PARAMETER DEFINITIONS:
          !    na

          ! INTERFACE BLOCK SPECIFICATIONS
          !    na

          ! DERIVED TYPE DEFINITIONS
          !    na

          ! SUBROUTINE LOCAL VARIABLE DECLARATIONS:
          !    na
CALL WriteError('Ground conditions are referenced in a ' // TRIM(surfObjKind) // &
     ' named "' // TRIM(FldVal(surfNameRef)) // '" but the Basement preprocessor was not run. ' // &
     ' Check the GroundHeatTransfer:Control object and see if the Basement preprocessor was set to Yes.')
END SUBROUTINE makeBasementErr


!     NOTICE
!
!     Copyright � 1996-2010 The Board of Trustees of the University of Illinois
!     and The Regents of the University of California through Ernest Orlando Lawrence
!     Berkeley National Laboratory.  All rights reserved.
!
!     Portions of the EnergyPlus software package have been developed and copyrighted
!     by other individuals, companies and institutions.  These portions have been
!     incorporated into the EnergyPlus software package under license.   For a complete
!     list of contributors, see "Notice" located in EnergyPlus.f90.
!
!     NOTICE: The U.S. Government is granted for itself and others acting on its
!     behalf a paid-up, nonexclusive, irrevocable, worldwide license in this data to
!     reproduce, prepare derivative works, and perform publicly and display publicly.
!     Beginning five (5) years after permission to assert copyright is granted,
!     subject to two possible five year renewals, the U.S. Government is granted for
!     itself and others acting on its behalf a paid-up, non-exclusive, irrevocable
!     worldwide license in this data to reproduce, prepare derivative works,
!     distribute copies to the public, perform publicly and display publicly, and to
!     permit others to do so.
!
!     TRADEMARKS: EnergyPlus is a trademark of the US Department of Energy.
!

END PROGRAM epfilter
